(* Copyright (C) 2014, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)

(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)


(*The portions of this file that are marked "Schepler" were
 copied and pasted from Daniel Schepler's "Zorn's Lemma."  Mine
 are marked "Wyckoff."*)

Require Import Description.
Require Import FunctionalExtensionality.
Require Import Classical.
Require Import List.
Require Import Ensembles.
Require Import Basics.
Require Image.
Require Import Omega.
Require Import Setoid.
Require Import Equality.
Require Import EqDec.
Require Import DecidableDec.

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

(*Wyckoff*)
Lemma p_equal :
  forall {T} {P:T->Prop} {x y:T},
    P x -> x = y ->
    P y.
  intros; subst; auto.
Qed.

(*This is helpful for FiniteMaps functionality.*)
(*Wyckoff*)
Definition f_pr {T U V:Type} (p:T->U->V) : (T*U) -> V :=
  fun pr:T*U => p (fst pr) (snd pr).

(*Wyckoff*)
Definition f_no_pr {T U V:Type} (p:(T*U)->V) : T -> U -> V :=
  fun (x:T) (y:U) => p (x, y).

(*Wyckoff*)
Lemma f_pr_f_no_pr : forall {T U V:Type} (p:(T*U)->V),
                         p = f_pr (f_no_pr p).
intros T U V p. unfold f_pr. unfold f_no_pr.
apply functional_extensionality.
intros pr; simpl. rewrite (surjective_pairing pr).
simpl. reflexivity.
Qed.

(*Wyckoff*)
Lemma f_no_pr_f_pr : forall {T U V:Type} (p:T->U->V),
                         p = (f_no_pr (f_pr p)).
intros T U V p. unfold f_pr. unfold f_no_pr. 
simpl.
apply functional_extensionality. 
intro x. apply functional_extensionality.
intro y. reflexivity.
Qed.

   
                                    
(*Wyckoff*)
Definition dep_funT {T:Type} (U:T->Type) : T-> Type :=
  fun x => U x.

(*Wyckoff*)
Lemma f_equal_to_existT :
  forall {T} {U:T->Type} {x y}
         (f:forall c, dep_funT U c) (pf:x = y),
    existT _ x (f x) =
    existT _ y (f y).
  intros; subst; auto.
Qed.


(*Wyckoff*)
Definition dep_fun {T:Type} (U:T->Type) : Type :=
  forall x:T, (dep_funT U x).


(*Wyckoff*)
Lemma f_equal_dep_app' :
  forall {T:Type} {x y:T} (U:T->Type)
         (f:dep_fun U) (pf:x = y),
    existT _ x (f x) =
    existT _ y (f y).
  intros; subst; auto.
Qed.

(*Wyckoff*)
Definition prop_dep_fun {T:Type} (P:T->Prop) (U:Type): Type :=
  forall x:T, P x -> U.

(*Wyckoff*)
Definition prop_dep_fun_sig {T P U} (f:prop_dep_fun P U) :
  {x:T|P x} -> U :=
  fun x => f (proj1_sig x) (proj2_sig x).

(*Wyckoff*)
Definition prop_dep_fun2 {T:Type} (P:T->Prop) (Q:T->Prop) (U:Type) : Type :=
  forall x:T, P x -> Q x -> U.

(*Wyckoff*)
Definition prop_dep_fun2_sig {T P Q U} (f:prop_dep_fun2 P Q U) :
  {x:T|P x /\ Q x} -> U :=
  fun x => let (pfp, pfq) := proj2_sig x in
           f (proj1_sig x) pfp pfq.

(*Wyckoff*)
Definition prop_dep_funT {T:Type} (P:T->Prop) : Type :=
  forall x:T, P x -> Type.

(*Wyckoff*)
Definition prop_dep_funT2 {T:Type} (P:T->Prop) (Q:T->Prop) : Type :=
  forall x:T, P x -> Q x -> Type.

(*Wyckoff*)
Definition prop_dep_fun_dep {T} {U:T->Type}
           (P:forall x, U x -> Prop) V : Type :=
  forall (x:T) (y:U x), P x y -> V.

(*Wyckoff*)
Definition prop_dep_fun2T {T U}
           (P:T->U->Prop) (V:Type) : Type :=
  forall x y, P x y -> V.


(*Wyckoff*)
Definition prop_dep_fun2T' {T U}
           (P:T->U->Prop) (V:T->U->Type) : Type :=
  forall x y, P x y -> V x y.



  
(*Wyckoff*)
Definition prop_dep_fun_nest2
           {T:Type} {P:T->Prop}
           (Q:prop_dep_funT P) U : Type :=
  forall (x:T) (px:P x), Q x px -> U.

(*Wyckoff*)
Lemma f_equal_dep :
  forall {T:Type} {P:T->Prop} {U:Type}
         (f:prop_dep_fun P U)
         (x y:T)
         (pfx:P x) (pfy:P y),
    x = y ->
    f x pfx = f y pfy.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma f_equal_dep2T :
  forall {T U:Type} {P:T->U->Prop} {V:Type}
         (f:prop_dep_fun2T P V)
         (x x':T)
         (y y':U)
         (pf:P x y) (pf':P x' y'),
  x = x' ->
  y = y' ->
  f x y pf = f x' y' pf'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Definition prop_dep_set {T:Set} (P:T->Prop) (U:Set) : Set :=
  forall x:T, P x -> U.


(*Wyckoff*)
Definition prop_dep_set_dep {T:Type} (P:T->Prop) (U:T->Set) : Type :=
  forall x:T, P x -> U x.

(*Wyckoff*)
Definition forall_from_S {n U} (f:forall i (pfi:i < S n), U) :
  forall i (pfi:i < n), U :=
  fun i pfi => f i (lt_S _ _ pfi).

(*Wyckoff*)
Definition forall_to_S {n U} (f:forall i (pfi:i < S n), U) :
  forall i (pfi:i < n), U :=
  fun i pfi => f (S i) (lt_n_S _ _ pfi).


(*Wyckoff*)
Definition seg_fun (n:nat) (U:Type) : Type :=
  prop_dep_fun (fun i => i < n) U.

(*Wyckoff*)
Definition seg_fun' (n:nat) (U:Set) : Set :=
  prop_dep_set (fun i => i < n) U.

(*Wyckoff*)
Definition seg_fun0 U : seg_fun 0 U :=
  fun _ pf => False_rect _ (lt_n_0 _ pf).

(*Wyckoff*)
Lemma seg_fun0_compat :
  forall U (f:seg_fun 0 U),
    seg_fun0 U = f.
  intros; apply functional_extensionality_dep;
  intro; apply functional_extensionality;
  intro; omega.
Qed.

(*Wyckoff*)
(*This returns a function g, such that g i _ = f i _, for i < S n -- 
i.e. the value at [n] is lost.*)
Definition seg_fun_from_S {n:nat} {U:Type} (f:seg_fun (S n) U) :
  seg_fun n U :=
  fun (m:nat) (pfm:m < n) => f m (lt_S _ _ pfm).



(*Wyckoff*)
Lemma seg_fun_from_S_compat :
  forall {U:Type} {n:nat} (f:seg_fun (S n) U) (i:nat) (pf:i < n),
    seg_fun_from_S f i pf = f i (lt_S _ _ pf).
  auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_from_S_compat' : 
  forall {U:Type} {n:nat} (f:seg_fun (S n) U) (i:nat)
         (pfi:i < n) (pfi':i < S n),
    f i pfi' = seg_fun_from_S f i pfi.
  unfold seg_fun_from_S; intros; f_equal; apply proof_irrelevance.
Qed.



(*Wyckoff*)
(* This function also derives a function of widith [n] from 
   one of [S n], but this time, 
   the values are translated up one shift.
   The returned function g obeys -- g i _ = f (S i) _. *)
Definition seg_fun_to_S {n:nat} {U:Type} (f:seg_fun (S n) U) :
  seg_fun n U :=
  fun i pf =>
    f (S i) (lt_n_S _ _ pf).

Lemma seg_fun_to_S_compat :
  forall {U:Type} {i n:nat} (f:seg_fun (S n) U)
         (pfi:i < n)
         (pfi':S i < S n),
    seg_fun_to_S f i pfi =
    f (S i) pfi'.
  unfold seg_fun_to_S; intros; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma seg_fun_to_S_compat' :
  forall {U:Type} {i n:nat} (f:seg_fun (S n) U)
         (pfi:S i < S n),
    f (S i) pfi =
    (seg_fun_to_S f) i (lt_S_n _ _ pfi).
  unfold seg_fun_to_S; intros; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma seg_fun_to_S_from_S :
  forall {U n} (f:seg_fun (S (S n)) U),
    seg_fun_to_S (seg_fun_from_S f) =
    seg_fun_from_S (seg_fun_to_S f).
  intros; apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality.
  intro hi.
  unfold seg_fun_to_S, seg_fun_from_S.
  f_equal; apply proof_irrelevance.
Qed.


(*Wyckoff*)
Definition seg_fun_cons {U n} (y:U) (f:seg_fun n U) :
  seg_fun (S n) U :=
  fun i =>
    match i with
      | 0 => fun _ => y
      | S i' => fun pf =>
                  let pf' := lt_S_n _ _ pf in
                  f i' pf'
    end.

(*Wyckoff*)
Lemma seg_fun_cons0 :
  forall {U n} (f:seg_fun n U) (y:U) (pf:0 < S n),
  seg_fun_cons y f 0 pf = y.
  unfold seg_fun_cons; intros; auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_cons_S :
  forall {U n} (f:seg_fun n U) (y:U) (i:nat) (pf:S i < S n),
    seg_fun_cons y f (S i) pf =
    f i (lt_S_n _ _ pf).
  unfold seg_fun_cons; intros; auto.
Qed.
  
(*Wyckoff*)
Lemma seg_fun_decompose_cons :
  forall {U n} (f:seg_fun (S n) U),
    f = seg_fun_cons (f 0 (lt_0_Sn _)) (seg_fun_to_S f).
  intros; apply functional_extensionality_dep; intro i;
  apply functional_extensionality; intro; simpl.
  unfold seg_fun_cons, seg_fun_to_S.
  destruct i; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma seg_fun_to_S_cons :
  forall {U n} y (f:seg_fun n U),
    seg_fun_to_S (seg_fun_cons y f) = f.
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro;
  unfold seg_fun_to_S; rewrite seg_fun_cons_S.
  f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma seg_fun_from_S_cons :
  forall {U n} y (f:seg_fun (S n) U),
    seg_fun_from_S (seg_fun_cons y f) =
    seg_fun_cons y (seg_fun_from_S f).
  intros; apply functional_extensionality_dep; intro i;
  apply functional_extensionality; intro hi; f_equal.
  destruct i as [|i]; simpl.
  rewrite seg_fun_from_S_compat, seg_fun_cons0; auto.
  do 2 rewrite seg_fun_from_S_compat.
  rewrite seg_fun_cons_S; f_equal; apply proof_irrelevance.
Qed.


(*Wyckoff*)
Definition seg_fun_from_pred {U n} : seg_fun n U -> seg_fun (pred n) U :=
  match n with
    | 0 => fun _ => seg_fun0 U
    | S n' => fun f => seg_fun_from_S f
  end.

(*Wyckoff*)
Definition seg_fun_to_pred {U n} : seg_fun n U -> seg_fun (pred n) U :=
  match n with
    | 0 => fun _ => seg_fun0 U
    | S n' => fun f => seg_fun_to_S f
  end.

(*Wyckoff*)
Lemma seg_fun_to_pred_to_S :
  forall {U n} (f:seg_fun (S (S n)) U), 
    seg_fun_to_S (seg_fun_to_pred f) =
    seg_fun_to_pred (seg_fun_to_S f).
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro;
  unfold seg_fun_to_pred, seg_fun_to_S; auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_from_pred_from_S :
  forall {U n} (f:seg_fun (S (S n)) U), 
    seg_fun_from_S (seg_fun_from_pred f) =
    seg_fun_from_pred (seg_fun_from_S f).
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro;
  unfold seg_fun_from_pred, seg_fun_from_S; auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_to_pred_from_S :
  forall {U n} (f:seg_fun (S (S n)) U), 
    seg_fun_to_pred (seg_fun_from_S f) =
    seg_fun_from_S (seg_fun_to_pred f).
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro;
  unfold seg_fun_to_pred, seg_fun_from_S, seg_fun_to_S; auto;
  f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma seg_fun_from_pred_to_S :
  forall {U n} (f:seg_fun (S (S n)) U), 
    seg_fun_from_pred (seg_fun_to_S f) =
    seg_fun_to_S (seg_fun_from_pred f).
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro;
  unfold seg_fun_from_pred, seg_fun_from_S, seg_fun_to_S; auto;
  f_equal; apply proof_irrelevance.
Qed.
  
(*See FunctionProperties2 for [seg_fun_[to/from|_pred] properties*)

(*Wyckoff*)
Definition seg_fun_dep {n} (fU:seg_fun n Type) : Type :=
  forall i (pfi:i < n), fU i pfi.

(*Wyckoff*)
Definition seg_fun_dep_from_S {n} :
  forall {fU:seg_fun (S n) Type}
         (f:seg_fun_dep fU),
    seg_fun_dep (seg_fun_from_S fU) :=
  fun fU f =>
    (fun i (pfi:i < n) => f i _).

(*Wyckoff*)
Definition seg_fun_dep_to_S {n} :
  forall {fU:seg_fun (S n) Type}
         (f:seg_fun_dep fU),
    seg_fun_dep (seg_fun_to_S fU) :=
  fun fU f =>
    (fun i (pfi:i < n) => f (S i) (lt_n_S _ _ _)).

(*Wyckoff*)
Definition seg_fun_depT {S n} (U:S->Type) (f:seg_fun n S) : Type :=
  forall i (pfi:i < n), U (f i pfi).

(*Wyckoff*)
Definition seg_fun_depT_to_S {T n} {U:T->Type} {f:seg_fun (S n) T}
           (g:seg_fun_depT U f) :
  seg_fun_depT U (seg_fun_to_S f) :=
  fun i (pfi:i < n) => g (S i) (lt_n_S _ _ pfi).

(*Wyckoff*)
Definition seg_fun_nT n T (U:Type->nat->Type) : Type :=
  forall i (pfi:i < n) (u:U T n), Type.

(*Wyckoff*)
Definition type_nat_to_S (U:Type->nat->Type) : Type -> nat -> Type:=
  fun T n => U T (S n).

(*Wyckoff*)
Definition seg_fun_nT_to_S {n T} {U:Type->nat->Type}
           (f:seg_fun_nT (S n) T U) :
  seg_fun_nT n T (type_nat_to_S U) :=
  fun i (pfi:i < n) (u:U T (S n)) =>
    f (S i) (lt_n_S _ _ pfi) u.

(*Wyckoff*)
Definition seg_fun_nT_from_S {n T} {U:Type->nat->Type}
           (f:seg_fun_nT (S n) T U) :
  seg_fun_nT n T (type_nat_to_S U) :=
  fun i (pfi:i < n) (u:U T (S n)) =>
    f i (lt_S _ _ pfi) u.


(*Wyckoff*)
Fixpoint seg_fun_app {U n} :
  seg_fun n U -> U ->
  seg_fun (S n) U :=
  match n with
    | 0 => fun _ y _ _ => y
    | S n' => fun f y i =>
                let f' := seg_fun_to_S f in
                let g := seg_fun_app f' y in
                match i with
                  | 0 => fun _ => f 0 (lt_0_Sn _)
                  | S i' =>
                    fun pfi =>
                      seg_fun_cons (f 0 (lt_0_Sn _)) g (S i') pfi
                end
  end.


(*Wyckoff*)
Lemma seg_fun_app_lt :
  forall {U n} (f:seg_fun n U) y i (pfi:i < S n)
         (pfi':i < n),
    seg_fun_app f y i pfi = f i pfi'.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros f y i hi hi'.
  destruct i as [|i].
  f_equal; apply proof_irrelevance.
  specialize (ih (seg_fun_to_S f) y _ (lt_S_n _ _ hi) (lt_S_n _ _ hi')).
  rewrite ih.
  unfold seg_fun_to_S.
  f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma seg_fun_app_n :
  forall {U n} (f:seg_fun n U) y (pfn:n < S n),
    seg_fun_app f y n pfn = y.
  intros ? n; induction n; simpl; auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_decompose_app :
  forall {U n} (f:seg_fun (S n) U),
    f = seg_fun_app (seg_fun_from_S f) (f n (lt_n_Sn _)).
  intros U n f; apply functional_extensionality_dep; intro i;
  apply functional_extensionality; intro hi; simpl.
  red in hi.
  destruct (le_lt_eq_dec _ _ (le_S_n _ _ hi)) as [h1 | h1].
  erewrite seg_fun_app_lt.
  unfold seg_fun_from_S; f_equal; apply proof_irrelevance.
  subst.
  erewrite seg_fun_app_n.
  f_equal; apply proof_irrelevance.

  Grab Existential Variables.
  assumption.
Qed.

(*Wyckoff*)
Lemma seg_fun_from_S_app :
  forall {U n} (hdu:type_eq_dec U)
         (f:seg_fun n U) b,
    seg_fun_from_S (seg_fun_app f b) = f.
  intros; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro;
  unfold seg_fun_from_S; erewrite seg_fun_app_lt.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma seg_fun_to_S_app :
  forall {U n} (hdu:type_eq_dec U)
         (f:seg_fun (S n) U) b,
    seg_fun_to_S (seg_fun_app f b) =
    seg_fun_app (seg_fun_to_S f) b.
  intros U n hdu f b; apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro h1.
  unfold seg_fun_to_S.
  simpl.
  red in h1.
  destruct (le_lt_eq_dec _  _ (le_S_n _ _ h1)) as [h2|h2].
  do 2 erewrite seg_fun_app_lt.
  unfold seg_fun_to_S.
  reflexivity.
  subst.
  do 2 erewrite seg_fun_app_n.
  reflexivity.

  Grab Existential Variables.
  assumption.
Qed.

(*Wyckoff*)
Definition seg_fun_to_0S {U n} : seg_fun n U -> seg_fun n U :=
  match n with
    | 0 => fun _ _ pf => False_rect _ (lt_n_0 _ pf)
    | S n' => fun f i pfi => 
                seg_fun_from_S (seg_fun_cons (f 0 (lt_0_Sn _)) f) i pfi
  end.


(*Wyckoff*)
Lemma seg_fun_to_0S_0 :
  forall {U n} (f:seg_fun n U) (pf:0 < n),
    seg_fun_to_0S f 0 pf = f 0 pf.
  intros U n. 
  destruct n as [|n].
  intros ? h1.
  contradict (lt_irrefl _ h1).
  intros; simpl.
  rewrite seg_fun_from_S_compat.
  rewrite seg_fun_cons0.
  f_equal.
  apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma seg_fun_to_0S_S :
  forall {U n} (f:seg_fun n U) (i:nat) (pf:S i < n),
    seg_fun_to_0S f (S i) pf = f i (lt_trans _ _ _ (lt_n_Sn _) pf).
  unfold seg_fun_to_0S.
  destruct n as [|n].
  intros ? ? h1; contradict (lt_n_0 _ h1).
  intros.
  unfold seg_fun_from_S.
  simpl.
  f_equal.
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Definition forall_dep_from_S
           {n} {ft:seg_fun (S n) Type}
           (f:forall i (pfi:i < S n), ft i pfi) :
  forall i (pfi:i < n), seg_fun_from_S ft i pfi :=
  fun i pfi => f i (lt_S _ _ pfi).


(*Wyckoff*)
Definition forall_dep_to_S
           {n} {ft:seg_fun (S n) Type}
           (f:forall i (pfi:i < S n), ft i pfi) :
  forall i (pfi:i < n), seg_fun_to_S ft i pfi :=
  fun i pfi => f (S i) (lt_n_S _ _ pfi).


(*Wyckoff*)
Fixpoint in_seg_fun {n U}
         (y:U) : seg_fun n U -> Prop :=
  match n with
    | 0 => fun _ => False
    | S n' => fun f => f 0 (lt_0_Sn _) = y \/ in_seg_fun y (seg_fun_to_S f)
  end.

(*Decidable version of the above further below.*)

(*Wyckoff*)

(*Returns the first number from 0 to [(pred n)] whose value is [y].
If [y] is not in the image, then it returns [n] as a default.*)
Fixpoint inv_seg_fun {U n} (pfdu:type_eq_dec U) (y:U) :
  seg_fun n U -> nat :=
  match n with
    | 0 => fun _ => 0
    | S n' => fun f =>
                if (pfdu (f 0 (lt_0_Sn _)) y) then
                  0
                else
                  S (inv_seg_fun pfdu y (seg_fun_to_S f))
  end.


(*Wyckoff*)
Lemma inv_seg_fun0:
  forall {U} (pfdu:type_eq_dec U) (y:U) (f:seg_fun 0 U),
         inv_seg_fun pfdu y f = 0.
  auto.
Qed.

(*Wyckoff*)
Lemma inv_seg_fun0':
  forall {U n} (pfdu:type_eq_dec U) (f:seg_fun n U) (pf0:0 < n),
    inv_seg_fun pfdu (f 0 pf0) f = 0.
  intros U n.
  destruct n as [|n].
  intros; omega. 
  intros hdu f h0; simpl.
  destruct (hdu (f 0 (lt_0_Sn _)) (f 0 h0)) as [|h1]; auto.
  contradict h1; f_equal; apply proof_irrelevance.
Qed.




(*Wyckoff*)
Lemma inv_seg_fun_le :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) i (pfi:i < n),
    inv_seg_fun pfdu (f i pfi) f <= i.
  intros U n.
  induction n as [|n ih]; simpl; auto with arith.
  intros hdu f i hi.
  destruct (hdu (f 0 (lt_0_Sn n)) (f i hi)) as [| h1]; auto with arith.
  destruct i as [|i].
  contradict h1; auto.
  f_equal; apply proof_irrelevance.
  apply le_n_S.
  pose proof (lt_S_n _ _ hi) as h2.
  specialize (ih hdu (seg_fun_to_S f) i h2).
  unfold seg_fun_to_S in ih.
  erewrite f_equal_dep.
  apply ih.
  reflexivity.
Qed.

(*Wyckoff*)
Lemma inv_seg_fun_weak_bound :
  forall {U n} (pfdu:type_eq_dec U)
         (f:seg_fun n U) i (pfi:i < n) y,
    y = f i pfi ->
    inv_seg_fun pfdu y f <= i.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros hdu f i hi ? ?; subst.
  destruct (hdu (f 0 (lt_0_Sn n)) (f i hi)) as [|h1]; auto with arith.
  destruct i as [|i].
  contradict h1; f_equal; apply proof_irrelevance.
  apply le_n_S.
  unfold seg_fun_to_S.
  erewrite f_equal_dep.
  eapply ih; auto with arith.
  reflexivity.

  Grab Existential Variables.
  auto with arith.
Qed.
  


(*Wyckoff*)
Definition seg_pred (n:nat) (P:nat->Prop) : Type :=
  forall i (pf:i < n), P i.

(*Wyckoff*)
Definition seg_pred_from_S {n:nat} {P:nat->Prop}
           (f:seg_pred (S n) P) : seg_pred n P :=
  fun i (pf:i < n) =>
    f i (lt_trans _ _ _ pf (lt_n_Sn _)).

(*Wyckoff*)
Definition seg_pred_type n : Type :=
  forall i (pf:i < n), Prop.

(*Wyckoff*)
Definition seg_pred_type_from_S {n}
           (f:seg_pred_type (S n)) : seg_pred_type n :=
  fun i (pf:i < n) =>
    f i (lt_trans _ _ _ pf (lt_n_Sn _)).

(*Wyckoff*)
Definition seg_pred_dec
           {n:nat} (pf:seg_pred_type n) : Type :=
  (forall i (pflt:i < n), {pf i pflt} + {~ pf i pflt}).

(*Wyckoff*)
Lemma seg_pred_dec_irrelevance : 
  forall {n:nat} {pf:seg_pred_type n}
         (pfs pfs':seg_pred_dec pf),
    pfs = pfs'.
  intros n h1 h2 h3.
  apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality_dep.
  intro hi.
  destruct (h2 i hi), (h3 i hi); f_equal; try apply proof_irrelevance; try contradiction.
Qed.


(*Wyckoff*)
Definition seg_fun_eq_dec 
           {U:Type} {n:nat}
           (f:seg_fun n U) : Type :=
  forall i j (pfi:i < n) (pfj:j < n),
    {f i pfi = f j pfj} +  {f i pfi <> f j pfj}.

(*Wyckoff*)
Lemma seg_fun_eq_dec_irrelevance :
  forall {U:Type} {n:nat}
         {f:seg_fun n U}
         (pfsf pfsf':seg_fun_eq_dec f),
    pfsf = pfsf'.
  intros U n f h1 h2.
  apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality_dep.
  intro j.
  apply functional_extensionality_dep.
  intro hi.
  apply functional_extensionality_dep.
  intro hj.
  destruct (h1 i j hi hj), (h2 i j hi hj); f_equal; try apply proof_irrelevance; try contradiction.
Qed.




(*Wyckoff*)
Definition seg_fun_rest {m n:nat} {U:Type} (f:seg_fun n U) (pfm:m < n) :
  seg_fun m U :=
  fun (i:nat) (pfi:i < m) => f i (lt_trans _ _ _ pfi pfm).


(*Wyckoff*)
Lemma seg_fun_rest_compat :
  forall {m n:nat} {U:Type} (f:seg_fun n U) (pfm:m < n)
         (i:nat) (pfi:i < m),
    seg_fun_rest f pfm i pfi =
    f i (lt_trans _ _ _ pfi pfm).
  unfold seg_fun_rest; intros; auto.
Qed.


(*Wyckoff*)
Definition fun_to_seg {U:Type} (f:nat->U) (n:nat) :
  seg_fun n U :=
  fun (i:nat) (pfi:i < n) => f i.

(*Wyckoff*)
Definition fun_to_seg_compat :
  forall {U:Type} (f:nat->U) (n:nat)
         (i:nat) (pf:i < n),
    fun_to_seg f n i pf = f i.
  auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_from_S_fun_to_seg :
  forall {U:Type} (f:nat->U) (n:nat),
    seg_fun_from_S (fun_to_seg f (S n)) =
    fun_to_seg f n.
  unfold seg_fun_from_S, fun_to_seg; intros; auto.
Qed.

(*Wyckoff*)
Definition id_seg n : seg_fun n nat := fun i _ => i.

(*Wyckoff*)
Definition inj_seg
           {U:Type} {n:nat} (f:seg_fun n U) : Prop :=
  forall i j
         (pfi:i < n)
         (pfj:j < n),
    f i pfi = f j pfj -> i = j.

(*Wyckoff*)
Lemma inj_from_S : 
  forall {U:Type} {n:nat} (f:seg_fun (S n) U),
    inj_seg f ->
    inj_seg (seg_fun_from_S f).
  intros ? ? ? h1; red in h1; red; intros; eapply h1.
  do 2 rewrite seg_fun_from_S_compat in H.
  apply H.
Qed.

(*Wyckoff*)
Definition surj_seg
           {U:Type} {n:nat} (f:seg_fun n U) : Prop :=
  forall y:U,
  exists i pf,
    f i pf = y.


(*Wyckoff*)
Inductive seg_fun_endo {n} : seg_fun n nat -> nat -> Prop :=
| seg_fun_endo_intro : 
    forall r,
      forall (f:seg_fun n nat),
        (forall i (pf:i < n),
           f i pf < r) -> seg_fun_endo f r.


(*Wyckoff*)
Inductive inj_seg_fun_endo {n} : seg_fun n nat -> nat -> Prop :=
| inj_seg_fun_endo_intro :
    forall (f:seg_fun n nat) r,
      seg_fun_endo f r ->
      (forall i j (pfi:i < n) (pfj:j < n),
         f i pfi = f j pfj -> i = j) ->
      inj_seg_fun_endo f r.

(*Wyckoff*)
Inductive surj_seg_fun_endo {n} : seg_fun n nat -> nat -> Prop :=
| surj_seg_fun_endo_intro : 
    forall (f:seg_fun n nat) r,
      seg_fun_endo f r ->
      (forall j, j < r ->
                 exists s (pf:s < n),
                   f s pf = j) ->
      surj_seg_fun_endo f r.


(*Wyckoff*)
Lemma inj_seg_fun_endo_endo :
  forall {n m} (f:seg_fun n nat),
    inj_seg_fun_endo f m ->
    seg_fun_endo f m.
  intros ? ? ? h1; destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma surj_seg_fun_endo_endo :
  forall {n m} (f:seg_fun n nat),
    surj_seg_fun_endo f m ->
    seg_fun_endo f m.
  intros ? ? ? h1; destruct h1; auto.
Qed.


(*Wyckoff*)
Lemma seg_fun_endo_to_S :
  forall {n} (f:seg_fun (S n) nat),
    seg_fun_endo f (S n) ->
    seg_fun_endo (seg_fun_to_S f) (S n).
  intros n f h1.  
  dependent induction h1. 
  econstructor. 
  intros; apply H.
Qed.

(*[inj] analouges are in FunctionProperties2.*)  

(*Wyckoff*)
Definition lt_seg_fun_endo :
  forall {n r}
         (f:seg_fun n nat),
    seg_fun_endo f r ->
    forall i (pf:i < n),
      f i pf < r.
  intros ? ? ? h1; destruct h1; auto.
Qed.



Fixpoint in_seg_fun_dec {n U} 
         (y:U) : seg_fun n U -> Set :=
  match n with
    | 0 => fun _ => False
    | S n' => fun f => sumor (in_seg_fun_dec y (seg_fun_to_S f)) (f 0 (lt_0_Sn _) = y)
  end.

Lemma in_seg_fun_dec_functional :
  forall {n U}
         (y y':U)
         (f g:seg_fun n U),
    y = y' ->
    f = g ->
    in_seg_fun_dec y  f ->
    in_seg_fun_dec y' g.
  intros; subst; auto.
Qed.



Lemma in_seg_fun_dec0_from_S :
  forall {U n} (f:seg_fun (S (S n)) U),
    in_seg_fun_dec (f 0 (lt_0_Sn _)) (seg_fun_from_S f).
  intros; right; unfold seg_fun_from_S; f_equal; apply proof_irrelevance.
Qed.


Lemma in_seg_fun_impl_dec :
  forall {U n} (pfdu:type_eq_dec U)
         (y:U) (f:seg_fun n U),
    in_seg_fun y f ->
    in_seg_fun_dec y f.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; contradiction.
  intros hdu y f h1.
  destruct (hdu (f 0 (lt_0_Sn n)) y) as [|h2]; subst.
  right; auto.
  left.
  apply ih; auto.
  destruct h1; subst; auto.
  contradict h2; auto.
Qed.

Lemma in_seg_fun_dec_impl :
  forall {U n}
         (y:U) (f:seg_fun n U),
    in_seg_fun_dec y f ->
    in_seg_fun y f.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros y f h1.
  destruct h1 as [h1|]; subst.
  apply ih in h1.
  right; auto.
  left; auto.
Qed.


Lemma in_seg_fun_dec_compat :
  forall {U n} (f:seg_fun n U) i (pfi:i < n),
    in_seg_fun_dec (f i pfi) f.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros f i h1.
  destruct i as [|i].
  right; f_equal; apply proof_irrelevance.
  pose proof (lt_S_n _ _ h1) as h2.
  specialize (ih (seg_fun_to_S f) i h2).
  left; auto.
  unfold seg_fun_to_S in ih at 1.
  erewrite f_equal_dep.
  apply ih.
  reflexivity.
Qed.

Lemma in_seg_fun_dec_dec :
  forall {U n} (pfdu:type_eq_dec U) (f:seg_fun n U) y,
    (in_seg_fun_dec y f) + {in_seg_fun_dec y f -> False}.
  intros U n hdu.
  induction n as [|n ih]; simpl.
  intros; right; auto.
  intros f y.
  specialize (ih (seg_fun_to_S f) y).
  destruct ih as [h1 | h1].
  left.
  left; auto.
  destruct (hdu (f 0 (lt_0_Sn _)) y) as [h2 | h2].
  left; right; auto.
  right.
  intro h3.
  destruct h3; auto.
Qed.  

Lemma in_seg_fun_dec' :
  forall {U n} (pfdu:type_eq_dec U) (f:seg_fun n U) y,
    {in_seg_fun y f} + {~in_seg_fun y f}.
  intros U n hdu f y.
  pose proof (in_seg_fun_dec_dec hdu f y) as h1.
  destruct h1 as [h1 | h1].
  left.
  apply in_seg_fun_dec_impl; auto.
  right.
  intro h2.
  apply in_seg_fun_impl_dec in h2; auto.
Qed.
  



Lemma in_seg_fun_compat :
  forall {U n} (f:seg_fun n U) i (pfi:i < n),
    in_seg_fun (f i pfi) f.
  intros U n f i hi.
  apply in_seg_fun_dec_impl.
  apply in_seg_fun_dec_compat.
Qed.


Lemma in_seg_fun_dec_to_S :
  forall {U n} y (f:seg_fun (S n) U),
    in_seg_fun_dec y (seg_fun_to_S f) ->
    in_seg_fun_dec y f.
  intros; simpl; left; auto.
Qed.

Lemma in_seg_fun_to_S :
  forall {U n} y (f:seg_fun (S n) U),
    in_seg_fun y (seg_fun_to_S f) ->
    in_seg_fun y f.
  intros; simpl; right; auto.
Qed.
  



Lemma in_seg_fun_dec_to_S_neq :
  forall {U n} y (f:seg_fun (S n) U),
    in_seg_fun_dec y f ->
    y <> f 0 (lt_0_Sn _) ->
    in_seg_fun_dec y (seg_fun_to_S f).
  intros U n y f h1 h2; simpl; simpl in h1.
  destruct h1; subst; auto.
  contradict h2; auto.
Qed.

Lemma in_seg_fun_to_S_neq :
  forall {U n} y (f:seg_fun (S n) U),
    in_seg_fun y f ->
    y <> f 0 (lt_0_Sn _) ->
    in_seg_fun y (seg_fun_to_S f).
  intros ? ? ? ? h1 h2; simpl in h1.
  destruct h1; subst; auto.
  contradict h2; auto.
Qed.



(*Experiment with alternate primed version.*)
Lemma in_seg_fun_from_S :
  forall {n U} y (f:seg_fun (S n) U),
    in_seg_fun_dec y (seg_fun_from_S f) ->
    in_seg_fun_dec y f.
  intros n.
  induction n as [|n ih]; simpl. 
  intros; contradiction. 
  intros U y f h1.
  destruct h1 as [h1|]; subst.
  rewrite seg_fun_to_S_from_S in h1.
  specialize (ih _ _ _ h1).
  apply in_seg_fun_dec_to_S in ih; auto.
  right;  unfold seg_fun_from_S.
  f_equal; apply proof_irrelevance.
Qed.


Lemma in_seg_fun_dec_to_S_from_S :
  forall {U n} y (f:seg_fun (S n) U),
    in_seg_fun_dec y (seg_fun_to_S f) ->
    y <> f n (lt_n_Sn _) ->
    in_seg_fun_dec y (seg_fun_from_S f).
  intros U n .
  induction n as [|n ih]; simpl.
  intros; contradiction.
  intros y f h1 h2.
  destruct h1 as [h1 | h1]. 
  apply ih in h1.
  left; auto.
  rewrite seg_fun_to_S_from_S; auto.
  unfold seg_fun_to_S. 
  erewrite f_equal_dep.
  apply h2. 
  reflexivity.
  subst.
  destruct (zerop n) as [|h3]; subst.
  unfold seg_fun_to_S in h2.
  contradict h2; apply f_equal_dep; reflexivity.
  unfold seg_fun_to_S; unfold seg_fun_to_S in ih.
  pose proof (in_seg_fun_dec_compat (seg_fun_from_S f) 1
                                (lt_n_S _ _ h3)) as h4.
  simpl in h4.
  destruct h4 as [h4 | h4]. 
  left. 
  rewrite seg_fun_from_S_compat in h4; auto.
  erewrite f_equal_dep.
  apply h4.
  reflexivity.
  rewrite h4.
  right.
  unfold seg_fun_from_S. 
  erewrite f_equal_dep.
  reflexivity.
  reflexivity.
Qed.

Lemma in_seg_fun_from_S_neq :
  forall {U n} y (f:seg_fun (S n) U),
    in_seg_fun_dec y f ->
    y <> f n (lt_n_Sn _) ->
    in_seg_fun_dec y (seg_fun_from_S f). 
  intros U n y f h1 h2; simpl in h1. 
  destruct n as [|n].
  destruct h1; subst; auto.
  contradict h2; auto. 
  f_equal; apply proof_irrelevance.
  destruct h1 as [h1 | h1].
  apply in_seg_fun_dec_to_S_from_S in h1; auto.
  subst.
  apply in_seg_fun_dec0_from_S.
Qed.

  
Lemma in_seg_fun_iff :
  forall {U n} y (f:seg_fun n U),
    in_seg_fun y f <->
    exists i (pfi:i < n),
      f i pfi = y.
  intros U n y f; split; intro h1.
  revert h1. revert f y.
  induction n as [|n ih]; simpl.
  intros; contradiction.
  intros f y h1.
  destruct h1 as [|h1]; subst.
  exists 0, (lt_0_Sn _); auto.
  apply ih in h1.
  destruct h1 as [i [h1 ?]]; subst.
  exists (S i), (lt_n_S _ _ h1).
  unfold seg_fun_to_S; auto.
  destruct h1 as [i [h2 ?]]; subst.
  apply in_seg_fun_dec_impl.
  apply in_seg_fun_dec_compat.
Qed.


(*Wyckoff*)
Lemma in_seg_fun_lt :
  forall {n b} (f:seg_fun n nat) y,
    seg_fun_endo f b ->
    in_seg_fun y f->
    y < b.
  intros n b f y h2 h3.
  rewrite in_seg_fun_iff in h3.
  destruct h3 as [i [h3 ?]]; subst.
  destruct h2; auto.
Qed.




(*Wyckoff*)
Lemma seg_fun_endo1_eq :
  forall {n} (f:seg_fun n nat),
    seg_fun_endo f 1 ->
    forall i pfi,
      f i pfi = 0.
  intros ? ? h1.
  inversion h1 as [? ? h2 h3]; subst.
  intros i hi.
  specialize (h2 i hi).
  red in h2.
  apply le_S_n in h2; auto with arith.
Qed.

(*Wyckoff*)
Definition lt_surj_seg_fun_endo :
  forall {n r}
         (f:seg_fun n nat),
    surj_seg_fun_endo f r ->
    forall i (pf:i < n),
      f i pf < r.
  intros ? ? ? h1; destruct h1; intros; apply lt_seg_fun_endo; auto.
Qed.

(*Wyckoff*)
Lemma inj_seg_fun_endo_compat :
  forall {n r} (f:seg_fun n nat),
    seg_fun_endo f r ->
    inj_seg f ->
    inj_seg_fun_endo f r.
  intros; constructor; auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_to_ind :
  forall {U n} (f:seg_fun n U)
         (P:forall m (f:seg_fun m U), Prop),
    P 0 (seg_fun0 U) ->
    (forall (b:U) (m:nat) (f:seg_fun m U),
       P m f ->
       P (S m) (seg_fun_cons b f)) ->
    P n f.
  intros U n.
  induction n as [|n ih]; simpl. 
  intros f P h1 h2.
  erewrite seg_fun0_compat in h1.
  apply h1.
  intros f P h1 h2.
  rewrite seg_fun_decompose_cons.
  apply h2; auto.
  apply ih; auto.
Qed.

(*Wyckoff*)
Lemma seg_fun_from_ind :
  forall {U n} (f:seg_fun n U)
         (P:forall m (f:seg_fun m U), Prop),
    P 0 (seg_fun0 U) ->
    (forall (b:U) (m:nat) (f:seg_fun m U),
       P m f ->
       P (S m) (seg_fun_app f b)) ->
    P n f.
  intros U n.
  induction n as [|n ih]; simpl.
  intros f P h2 h3.
  erewrite <- seg_fun0_compat; auto.
  intros f P h1 h2.
  rewrite seg_fun_decompose_app.
  apply h2.
  apply ih; auto.
Qed.




(*Wyckoff*)
Definition seq_fun (n w:nat) (U:Type) : Type :=
  forall m:nat, n <= m < n+w -> U.

(*Wyckoff*)
Definition seg_fun_to_seq {m:nat} {U:Type} (f:seg_fun m U) : seq_fun 0 m U :=
  (fun m pf => let (_, pf') := pf in (f m pf')).


(*See [ArithUtilities] for [seg_S_tl], [seq_fun_from_S], and other
derived lemmas.*)

(*Wyckoff*)
Definition pred_impl {T} : relation (T->Prop) :=
  fun P Q => forall x, P x -> Q x.

(*Wyckoff*)
Definition impl_pred {T} : relation (T->Prop) :=
  fun P Q => pred_impl Q P.

(*Wyckoff*)
Definition iff_pred {T} : relation (T->Prop) :=
  fun P Q => pred_impl P Q /\ impl_pred P Q.

(*Wyckoff*)
Lemma iff_pred1 :
  forall {T} (P Q:T->Prop),
    iff_pred P Q ->
    pred_impl P Q.
  intros ? ? ? h1; destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma iff_pred2 :
  forall {T} (P Q:T->Prop),
    iff_pred P Q ->
    impl_pred P Q.
  intros ? ? ? h1; destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma iff_pred_iff :
  forall {T} (P Q:T->Prop),
    iff_pred P Q <->
    (forall x, P x <-> Q x).
  intros; red; split; intro h1; red in h1.
  destruct h1 as [h1 h2].
  intro x.
  red in h1. do 2 red in h2.
  specialize (h1 x); specialize (h2 x); split; auto.
  red; split; red; try red; intros x hx; specialize (h1 x); tauto.
Qed.
  

(*Wyckoff*)
Lemma refl_pred_impl :
  forall T,
    Reflexive (@pred_impl T).
  intro; red; intro; red; auto.
Qed.

(*Wyckoff*)
Lemma trans_pred_impl :
  forall T,
    Transitive (@pred_impl T).
  intro; red; intro; red; auto.
Qed.

(*Wyckoff*)
Lemma refl_impl_pred :
  forall T,
    Reflexive (@impl_pred T).
  intro; red; intro; do 2 red; auto.
Qed.

(*Wyckoff*)
Lemma trans_impl_pred :
  forall T,
    Transitive (@impl_pred T).
  intro; red; intro; do 2 red; auto.
Qed.

(*Wyckoff*)
Lemma refl_iff_pred :
  forall T,
    Reflexive (@iff_pred T).
  intros; red; intros; red; split; [apply refl_pred_impl | apply refl_impl_pred].
Qed.

(*Wyckoff*)
Lemma sym_iff_pred :
  forall T,
    Symmetric (@iff_pred T).
  intros; red; intro; red; intros ? h1; destruct h1 as [h1 h2];
  red in h1, h2; split; [apply refl_pred_impl | apply refl_impl_pred]; auto.
Qed.

(*Wyckoff*)
Lemma trans_iff_pred :
  forall T,
    Transitive (@iff_pred T).
  intro; red; intros ? ? ? h1 h2; destruct h1 as [h1a h1b], h2 as [h2a h2b]; red; split.
  eapply trans_pred_impl.
  apply h1a.
  auto.
  eapply trans_impl_pred.
  apply h1b.
  auto.
Qed.

(*Wyckoff*)
Add Parametric Relation T :
  (T->Prop)
  (@impl_pred T)
    reflexivity proved by (@refl_impl_pred T)
    transitivity proved by (@trans_impl_pred T)
      as impl_pred_rel.

(*Wyckoff*)
Add Parametric Relation T :
  (T->Prop)
  (@pred_impl T)
    reflexivity proved by (@refl_pred_impl T)
    transitivity proved by (@trans_pred_impl T)
      as pred_impl_rel.

(*Wyckoff*)
Add Parametric Relation T :
  (T->Prop)
  (@iff_pred T)
    reflexivity proved by (@refl_iff_pred T)
    symmetry proved by (@sym_iff_pred T)
    transitivity proved by (@trans_iff_pred T)
      as iff_pred_rel.

(*Wyckoff*)
Definition eq_subst_l {T} {x x':T} (pfe:x = x')
           (f:forall y, x = y) :
  forall y, x' = y.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Definition eq_subst_r {T} {y y':T} (pfe:y = y')
           (f:forall x, x = y) :
  forall x, x = y.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Definition neq_subst_l {T} {x x':T} (pfe:x = x')
           (f:forall y, x <> y) :
  forall y, x' <> y.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Definition neq_subst_r {T} {y y':T} (pfe:y = y')
           (f:forall x, x <> y) :
  forall x, x <> y.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma iff_pred_eq_l :
  forall {T} {x x':T} (pfe:x = x'),
    iff_pred (fun c => x = c) (fun c => x' = c).
  intros; subst; reflexivity.
Qed.

(*Wyckoff*)
Lemma iff_pred_eq_r :
  forall {T} {y y':T} (pfe:y = y'),
    iff_pred (fun c => c = y) (fun c => c = y').
  intros; subst; reflexivity.
Qed.

(*Wyckoff*)
Lemma iff_pred_neq_l :
  forall {T} {x x':T} (pfe:x = x'),
    iff_pred (fun c => x <> c) (fun c => x' <> c).
  intros; subst; reflexivity.
Qed.

(*Wyckoff*)
Lemma iff_pred_neq_r :
  forall {T} {y y':T} (pfe:y = y'),
    iff_pred (fun c => c <> y) (fun c => c <> y').
  intros; subst; reflexivity.
Qed.

          


(*Wyckoff*)
Definition proj1_sig_fun 
  {T U:Type} {P:U->Prop} (f:T->{x:U | P x}) :
    T->U :=
  fun x => proj1_sig (f x).


(*Wyckoff*)
Definition proj1_sig_fun_compat : 
  forall {T U:Type} {P:U->Prop} (f:T->{x:U | P x})
         (x:T),
    proj1_sig_fun f x = proj1_sig (f x).
  unfold proj1_sig_fun; intros; auto.
Qed.

(*Wyckoff*)
Definition proj1_sig_fun1 {T U} {P:T->Prop}
           (Q:{x:T | P x} -> U)
           (pfp:forall x:T, P x) :
  T -> U :=
  fun x => Q (exist P x (pfp x)).


(*Wyckoff*)
Definition unit_set_pred {T:Type} (U:T->Set) : Prop :=
  forall (x:T) (a b:U x),
    a = b.

(*Wyckoff*)
Definition fexcl_mid {T:Type} (P:T->Prop) : T -> Set :=
  (fun x => {P x} + {~P x}).

(*Wyckoff*)
Lemma unit_set_pred_fexcl_mid : 
  forall {T:Type} (P:T->Prop),
    unit_set_pred (fexcl_mid P).
  intros.
  red.
  intros x a b.
  destruct a, b; try contradiction; auto; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Definition inj_dep {T U:Type} {P:T->Prop}
           (f:prop_dep_fun P U) : Prop :=
  forall (x y:T) (pfx:P x) (pfy:P y),
    f x pfx = f y pfy ->
    x = y.

(*Wyckoff*)
Definition inj_dep_contra {T U:Type} {P:T->Prop}
           (f:prop_dep_fun P U) : Prop :=
  forall (x y:T) (pfx:P x) (pfy:P y),
    x <> y ->
    f x pfx <> f y pfy.

(*Wyckoff*)
Lemma inj_dep_impl_contra :
  forall {T U:Type} {P:T->Prop}
         (f:prop_dep_fun P U),
    inj_dep f ->
    inj_dep_contra f.
  intros ? ? ? ? h1; red in h1; red; intros ? ? ? ? ? h5; apply h1 in h5; auto.
Qed.

(*Wyckoff*)
Lemma contra_impl_inj_dep :
  forall {T U:Type} {P:T->Prop}
         (pfdt:type_eq_dec T)
         (f:prop_dep_fun P U),
    inj_dep_contra f ->
    inj_dep f.
  intros ? ? ? hdt f h1; red in h1; red; intros ? ? ? ? h5;
  destruct (hdt x y) as [|h6]; auto; eapply h1 in h6;
  rewrite h5 in h6 at 1.
  contradict h6; reflexivity.
Qed.

(*Wyckoff*)
Lemma inj_seg_fun0 :
  forall U,
    inj_dep (seg_fun0 U).
  intro; red; intros; omega.
Qed.

(*Wyckoff*)
Definition surj_dep {T U:Type} {P:T->Prop}
           (f:prop_dep_fun P U) : Prop :=
  forall y:U,
  exists (x:T) (pfx:P x),
    f x pfx = y.

(*Wyckoff*)
Definition bij_dep {T U:Type} {P:T->Prop}
           (f:prop_dep_fun P U) : Prop :=
  inj_dep f /\ surj_dep f.

(*Wyckoff*)
Lemma bij_impl_inj_dep :
  forall {T U:Type} {P:T->Prop}
         (f:prop_dep_fun P U),
    bij_dep f ->
    inj_dep f.
  intros ? ? ? ? h1; destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma bij_impl_surj_dep :
  forall {T U:Type} {P:T->Prop}
         (f:prop_dep_fun P U),
    bij_dep f ->
    surj_dep f.
  intros ? ? ? ? h1; destruct h1; auto.
Qed.


(*Wyckoff*)
(*Dependent invertibility.*)
Inductive inv_dep {T U:Type} {P:T->Prop} 
          (f:prop_dep_fun P U) : Prop :=
| inv_dep_intro :
    forall (g:U->T)
           (pfp:forall u, P (g u)),
      (forall (x:T) (pfx:P x), g (f x pfx) = x) ->
      (forall (y:U), f (g y) (pfp y) = y) ->
      inv_dep f.

(*Wyckoff*)
Definition fun_in_dep_fun_compose {T U V:Type} {P:T->Prop}
           (f:prop_dep_fun P U)
           (g:U->V) :
  prop_dep_fun P V :=
  fun x pf => g (f x pf).


(*Wyckoff*)
Definition proj1_sig_fun_dep
           {T U:Type} {P:T->Prop} {Q:U->Prop}
           (f:prop_dep_fun P {x:U | Q x}) :
  prop_dep_fun P U :=
  fun_in_dep_fun_compose f (@proj1_sig _ _).


           
(*See [FunctionProperties2] for more on [inv_dep].*)


(*Wyckoff*)
Lemma f_equal_prop_dep :
  forall {T U:Type} {P:T->Prop}
         {x y:T} (f:prop_dep_fun P U)
         (pfeq:x = y)(pfx:P x) (pfy:P y),
    f x pfx = f y pfy.
  intros ? ? ? ? ? ? ?  h1 h2; subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.


(*Wyckoff*)
Lemma f_equal_prop_dep2 :
  forall {T U:Type} {P Q:T->Prop}
         {x y:T} (f:prop_dep_fun2 P Q U)
         (pfeq:x = y)(pfpx:P x) (pfpy:P y) (pfqx:Q x) (pfqy:Q y),
    f x pfpx pfqx = f y pfpy pfqy.
  intros ? ? ? ? ? ? ? ? h1 h2 h3 h4; subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
  pose proof (proof_irrelevance _ h3 h4); subst; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U)
         (m m':nat)
         (pf:  m  < n)
         (pf': m' < n),
    m = m' ->
    f m pf = f m' pf'.
  intros; apply f_equal_prop_dep; auto.
Qed.

(*Wyckoff*)
Lemma f_equal_seg_fun_depT : 
  forall {S n} {U:S -> Type}
         {f:seg_fun n S}
         (g:seg_fun_depT U f)
         i (pfi pfi':i < n)
         (pfe:f i pfi = f i pfi'),
    existT _ (f i pfi) (g i pfi) = 
    existT _ (f i pfi') (g i pfi').
  intros S n;
  induction n as [|n ih]; simpl; intros U f g i h1 h2 h3; try omega.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.


(*Wyckoff*)
Definition id_prop_dep {T:Type} (P:T->Prop) : prop_dep_fun P T :=
  fun x _ => x.

(*Wyckoff*)
Definition id_prop_dep2 {T:Type} (P:T->Prop) (Q:T->Prop) : prop_dep_fun2 P Q T :=
  fun x _ _ => x.

(*Wyckoff*)
Definition id_prop_dep_sig {T:Type} (P:T->Prop) : {x:T|P x} -> T :=
  prop_dep_fun_sig (id_prop_dep P).

(*Wyckoff*)
Definition id_prop_dep2_sig {T:Type} (P:T->Prop) (Q:T->Prop) :
  {x:T | P x /\ Q x} -> T :=
  prop_dep_fun2_sig (id_prop_dep2 P Q).

(*Wyckoff*)
Definition id_prop_dep_sig' {T} (P:T->Prop) :
  prop_dep_fun P {x:T | P x} :=
  fun i pf => exist P i pf.


(*Wyckoff*)
Lemma prop_dep_fun_ext :
  forall {T U:Type} (P:T->Prop) (g g':prop_dep_fun P U),
    (forall (x:T) (pf:P x), g x pf = g' x pf) -> g = g'.
  intros T U P g g' h1.
  apply functional_extensionality_dep.
  intro x. apply functional_extensionality.
  intro h2.
  apply h1.
Qed.

(*Wyckoff*)
Lemma prop_dep_fun2_ext :
  forall {T U:Type} (P Q:T->Prop) (g g':prop_dep_fun2 P Q U),
    (forall (x:T) (pfp:P x) (pfq:Q x),
       g x pfp pfq = g' x pfp pfq) -> g = g'.
  intros T U P Q g g' h1. 
  apply functional_extensionality_dep.
  intro x. apply functional_extensionality.
  intro h2.
  apply functional_extensionality.
  intro h3.
  apply h1.
Qed.

(*Wyckoff*)
Lemma prop_dep_set_dep_ext :
  forall {T:Type} (P:T->Prop) (U:T->Set)
         (f g:prop_dep_set_dep P U),
    unit_set_pred U ->
    f = g.
  intros ? ? ? ? ? h1.
  apply functional_extensionality_dep.
  intro.
  apply functional_extensionality.
  intro.
  apply h1.
Qed.

(*Wyckoff*)
Lemma seg_fun_ext :
  forall {U:Type} (n:nat) (g g':seg_fun n U),
    (forall (m:nat) (pf:m < n), g m pf = g' m pf) -> g = g'.
  intros; apply prop_dep_fun_ext; auto.
Qed.


(*Wyckoff*)
Definition prop_dep_prop {T:Type} (P:T->Prop) : Type :=
  forall x:T, P x -> Prop.


(*Wyckoff*)
Definition prop_dep_prop_u {T:Type} (P:T->Prop) (U:Type) : Type :=
  forall x:T, P x -> U -> Prop.


(* Possible name clash with something in [FunctionalExtensionality],
   which should be imported before this file or something. *)
(*Wyckoff*)
Lemma equal_f :
  forall {T U:Type} {f g:T->U},
    f = g ->
    forall (x:T),
      f x = g x.
  intros; subst; auto.
Qed.

Lemma equal_f_dep :
  forall {T U} {P:T->Prop}
         (f g:prop_dep_fun P U),
    f = g ->
    forall x (pf:P x),
      f x pf = g x pf.
  intros; subst; auto.
Qed.



(*Wyckoff*)
Definition not_pred {T:Type} (P:T->Prop) := fun x => ~ P x.

(*Schepler*)
Definition injective {X Y:Type} (f:X->Y) :=
  forall x1 x2:X, f x1 = f x2 -> x1 = x2.
(*Schepler*)
Definition surjective {X Y:Type} (f:X->Y) :=
  forall y:Y, exists x:X, f x = y.
(*Schepler*)
Definition bijective {X Y:Type} (f:X->Y) :=
  injective f /\ surjective f.

(*Wyckoff*)
Lemma bij_impl_inj :
  forall  {X Y:Type} (f:X->Y),
    bijective f ->
    injective f.
  intros ? ? ? h1; destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma bij_impl_surj :
  forall  {X Y:Type} (f:X->Y),
    bijective f ->
    surjective f.
  intros ? ? ? h1; destruct h1; auto.
Qed.


Notation Injective := injective.

(*Wyckoff*)
Definition inj_contra {X Y:Type} (f:X->Y) : Prop :=
  forall x1 x2:X, x1 <> x2 -> f x1 <> f x2.

(*Wyckoff*)
Lemma inj_impl_contra :
  forall {X Y:Type} (f:X->Y),
    injective f ->
    inj_contra f.
  intros; red; intros ? ? h2; contradict h2; auto.
Qed.

(*Wyckoff*)
Lemma contra_impl_inj :
  forall {X Y:Type} (pfdt:type_eq_dec X) (f:X->Y),
    inj_contra f ->
    injective f.
  intros ? ? ? ? h1; red; intros x y h2.
  destruct (pfdt x y) as [|h3]; auto.
  apply h1 in h3; contradiction.
Qed.


(*Wyckoff*)
Lemma inj_image_fun_prop_compat : 
  forall {T U:Type} (f:T->U),
    injective f <->
    Image.injective _ _ f.
intros; split; auto.
Qed.

(*Schepler*)
Inductive invertible {X Y:Type} (f:X->Y) : Prop :=
  | intro_invertible: forall g:Y->X,
  (forall x:X, g (f x) = x) -> (forall y:Y, f (g y) = y) ->
  invertible f.

(*Schepler*)
Lemma unique_inverse: forall {X Y:Type} (f:X->Y), invertible f ->
  exists! g:Y->X, (forall x:X, g (f x) = x) /\
             (forall y:Y, f (g y) = y).
Proof.
intros.
destruct H.
exists g.
red; split.
tauto.
intros.
destruct H1.
extensionality y.
transitivity (g (f (x' y))).
rewrite H2.
reflexivity.
rewrite H.
reflexivity.
Qed.


(*Schepler*)
Definition function_inverse {X Y:Type} (f:X->Y)
  (i:invertible f) : { g:Y->X | (forall x:X, g (f x) = x) /\
                                (forall y:Y, f (g y) = y) }
  :=
     (constructive_definite_description _
      (unique_inverse f i)).

(*Schepler*)
Lemma bijective_impl_invertible: forall {X Y:Type} (f:X->Y),
  bijective f -> invertible f.
Proof.
intros.
destruct H.
assert (forall y:Y, {x:X | f x = y}).
intro.
apply constructive_definite_description.
pose proof (H0 y).
destruct H1.
exists x.
red; split.
assumption.
intros.
apply H.
transitivity y.
auto with *.
auto with *.
pose (g := fun y:Y => proj1_sig (X0 y)).
pose proof (fun y:Y => proj2_sig (X0 y)).
simpl in H1.
exists g.
intro.
apply H.
unfold g; rewrite H1.
reflexivity.
intro.
unfold g; apply H1.
Qed.

(*Schepler*)
Lemma invertible_impl_bijective: forall {X Y:Type} (f:X->Y),
  invertible f -> bijective f.
Proof.
intros.
destruct H.
split.
red; intros.
congruence.
red; intro.
exists (g y).
apply H0.
Qed.

(*Wyckoff*)
Lemma invertible_impl_injective:
  forall {X Y:Type} (f:X->Y),
    invertible f -> injective f.
  intros ? ? ? h1; apply invertible_impl_bijective in h1; destruct h1; auto.
Qed.

(*Wyckoff*)
Lemma invertible_impl_surjective:
  forall {X Y:Type} (f:X->Y),
    invertible f -> surjective f.
  intros ? ? ? h1; apply invertible_impl_bijective in h1; destruct h1; auto.
Qed.



(*Wyckoff*)
Lemma invertible_impl_inv_invertible: forall {X Y:Type} (f:X->Y)
  (pf:invertible f), invertible (proj1_sig (function_inverse _ pf)).
intros X Y f h1.
destruct (function_inverse f h1) as [g h2].
simpl.
destruct h2 as [h2l h2r].
apply (intro_invertible g f h2r h2l).
Qed.

(*Wyckoff*)
Lemma distinct_functions_differ_at_point: forall {X Y:Type} (f g:X -> Y), 
  f <> g -> (exists x:X, f x <> g x).
intros X Y f g h1.
apply NNPP.
intro h2.
pose proof (not_ex_all_not _ _ h2) as h3.
simpl in h3.
assert (h4:forall x:X, f x = g x).
  intro x.
  specialize (h3 x).
  tauto.
pose proof (functional_extensionality _ _ h4).
contradiction.
Qed.

(*Wyckoff*)
Lemma bij_id : forall T,
                 @bijective T T id.
intro T.
red; split; red; intros; auto.
exists y; auto.
Qed.

(*Wyckoff*)
Lemma invertible_id :
  forall T,
    @invertible T T id.
  intro; apply bijective_impl_invertible.
  apply bij_id.
Qed.

(*Wyckoff*)
Lemma inj_id_dep :
  forall {T} (P:T->Prop),
    inj_dep (id_prop_dep P).
  unfold id_prop_dep; intros; red; auto.
Qed.

(*Wyckoff*)
Lemma bij_compose : 
  forall {T U V:Type} (f:T->U) (g:U->V),
    bijective f -> bijective g ->
    bijective (compose g f).
intros T U V f g h1 h2.
pose proof h1 as h1'. apply bijective_impl_invertible in h1'.
destruct h1 as [h1l h1r].
destruct h2 as [h2l h2r].
red. split.
red.
intros x y h3.
apply h2l in h3.
apply h1l; auto.
red. intros y.
specialize (h2r y).
destruct h2r as [x h2r].
subst.
exists ((proj1_sig (function_inverse _ h1')) x).
unfold compose.
rewrite (match (proj2_sig (function_inverse f h1')) with
           | conj _ pf => pf 
         end).
reflexivity.
Qed.

(*Wyckoff*)
Lemma invertible_compose : 
  forall {T U V:Type} (f:T->U) (g:U->V),
    invertible f -> invertible g ->
    invertible (compose g f).
  intros; apply bijective_impl_invertible;
  apply bij_compose;  apply invertible_impl_bijective; auto.
Qed.

(*Wyckoff*) 
Lemma inj_compose :   
  forall {T U V:Type} (f:T->U) (g:U->V),
    injective f -> injective g ->
    injective (compose g f).
intros T U V f g h1 h2.
unfold compose. red.
intros x y h3.
apply h2 in h3. apply h1 in h3.
assumption.
Qed.

(*Wyckoff*)
Lemma inj_id : 
  forall (T:Type),
    injective (id (A:=T)).
intro; red; intros; simpl; auto.
Qed.


(*Wyckoff*)
Lemma inj_dep_sig :
  forall {T U} {P:T->Prop} (f:prop_dep_fun P U),
    inj_dep f->
    injective (prop_dep_fun_sig f).
  unfold prop_dep_fun_sig; intros T U P f h1.
  red in h1; red.
  intros x y h2; destruct x as [x hx], y as [y hy]; simpl in h2.
  apply h1 in h2; subst; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma surj_dep_sig :
  forall {T U} {P:T->Prop} (f:prop_dep_fun P U),
    surj_dep f->
    surjective (prop_dep_fun_sig f).
  unfold prop_dep_fun_sig;
  intros T U P f h1; red in h1; red; intro y; specialize (h1 y).
  destruct h1 as [x [hx ?]]; subst.
  exists (exist _ _ hx); simpl; auto.
Qed.
  
(*Wyckoff*)
Lemma bij_dep_sig :
  forall {T U} {P:T->Prop} (f:prop_dep_fun P U),
    bij_dep f->
    bijective (prop_dep_fun_sig f).
  intros ? ? ? ? h1; destruct h1; red; split;
  [apply inj_dep_sig; auto |
   apply surj_dep_sig; auto].
Qed.

    

(*Wyckoff*)
Definition false_fun U : False -> U :=
  fun x => False_rect _ x.

(*wyckoff*)
Lemma false_fun_compat :
  forall {U} (f:False -> U),
    f = false_fun U.
  intros; apply functional_extensionality; intro; contradiction.
Qed.

(*Wyckoff*)
Lemma inj_false :
  forall U,
    injective (false_fun U).
  intros; red; simpl; intro; contradiction.
Qed.

(*Wyckoff*)
Lemma inj_fun_false :
  forall {U} (f:U->False),
    injective f.
  intros ? f; red.
  intros x ?.
  pose (f x); contradiction.
Qed.

(*Wyckoff*)
Lemma surj_fun_false :
  forall {U} (f:U->False),
    surjective f.
  intros U f; red.
  intro; contradiction.
Qed.

(*Wyckoff*)
Lemma bij_fun_false :
  forall {U} (f:U->False),
    bijective f.
  intros; red; split; [apply inj_fun_false | apply surj_fun_false].
Qed.

(*Wyckoff*)
Lemma inv_fun_false :
  forall {U} (f:U->False),
    invertible f.
  intros; apply bijective_impl_invertible; apply bij_fun_false.
Qed.


(*Wyckoff*)
Lemma inj_false_fun_false :
  forall U,
    injective (false_fun (U->False)).
  intro; red; intros; contradiction.
Qed.


(*Wyckoff*)
(*iterate unary operator*)
Fixpoint iter_uo {T:Type} (uo:T->T) (x:T) (n:nat) : T :=
  match n with
    | 0 => x 
    | S n => (uo (iter_uo uo x n))
  end.

(*Wyckoff*)
Lemma snd_iter_uo_const :
  forall {T U:Type} (f:T*U->T*U) (u:U),
    (forall pr:T*U, 
       snd (f pr) = u) ->
    forall (pr:T*U) (n:nat),
      snd (iter_uo f pr (S n)) = u.
  intros T U f u h1 pr n.
  revert pr h1. revert u f.
  induction n as [|n ih]; simpl.
  intros u f pr h1.
  apply h1; auto.
  intros u f pr h1.
  specialize (ih _ _ pr h1).
  simpl in ih.
  apply h1.
Qed.


Section Relations.

(*Wyckoff*)
Lemma lt_trans' :
  Transitive lt.
  red; intros; eapply lt_trans. apply H. apply H0.
Qed.


    
(*Wyckoff*)
Definition irrefl {T:Type} (R:T->T->Prop) : Prop :=
  forall x:T, ~ R x x.

(*Wyckoff*)
Inductive Strict_Order {T:Type} (R:T->T->Prop) :=
  strict_order_intro : Transitive R -> irrefl R ->
                       Strict_Order R.

(*Schepler*)
Definition total_strict_order {T:Type} (R:relation T) : Prop :=
  forall x y:T, R x y \/ x = y \/ R y x.

(*Wyckoff*)
Definition tso_dec {T:Type} (R:relation T) : Type :=
  forall x y:T, {R x y} + {x = y} + {R y x}.

(*Wyckoff*)
Definition Rlem_dec {T:Type} (R:relation T) : Type :=
  forall x y:T, {R x y} + {~R x y}.

(*Wyckoff*)
Lemma lt_irrefl_neq :
  forall {T:Type} (R:relation T) (a b:T),
    irrefl R ->
    R a b ->
    a <> b.
  intros; intro; subst; apply H in H0; auto.
Qed.

(*Wyckoff*)
Lemma gt_irrefl_neq :
  forall {T:Type} (R:relation T) (a b:T),
    irrefl R ->
    R b a->
    a <> b.
  intros; intro; subst; apply H in H0; auto.
Qed.


(*Wyckoff*)
Lemma tso_dec_total_strict_order :
  forall {T:Type} (R:relation T),
    tso_dec R ->
    total_strict_order R.
  intros T R h1.
  red in h1; red.
  intros x y.
  specialize (h1 x y).
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 | h1]; subst.
  left; auto.
  right; left; auto.
  right; auto.
Qed.

(*Wyckoff*)
Lemma irrefl_tso_dec_eq_dec :
  forall {T:Type} (R:relation T),
    irrefl R ->
    tso_dec R ->
    type_eq_dec T.
  intros T R h0 h1.
  red in h1; red.
  intros x y.
  specialize (h1 x y).
  destruct h1 as [[h1 | h1] | h1].
  right.
  intro; subst.
  apply h0 in h1; auto.
  left; auto.
  right.
  intro; subst.
  apply h0 in h1; auto.
Qed.


(*Wyckoff*)
Inductive sortable {T:Type} (R:relation T) : Prop :=
 | sort_intro : 
    total_strict_order R -> Transitive R -> irrefl R -> sortable R.

(*Wyckoff*)
Inductive sort_dec {T:Type} (R:relation T) : Type :=
 | sort_dec_intro : 
    tso_dec R -> Transitive R -> irrefl R -> sort_dec R.

(*Wyckoff*)
Lemma sort_dec_sortable :
  forall {T:Type} (R:relation T),
    sort_dec R ->
    sortable R.
  intros T R h1.
  destruct h1; constructor; auto.
  apply tso_dec_total_strict_order; auto.
Qed.
  
  
(*Wyckoff*)
Lemma sort_dec_tso_dec :
  forall {T:Type} (R:relation T),
    sort_dec R ->
    tso_dec R.
  intros ? ?  h1.
  destruct h1; auto.
Defined.



(*Wyckoff*)
Lemma sort_dec_trans :
  forall {T:Type} (R:relation T),
    sort_dec R ->
    Transitive R.
  intros ? ?  h1.
  destruct h1; auto.
Defined.

(*Wyckoff*)
Lemma sort_dec_irrefl :
  forall {T:Type} (R:relation T),
    sort_dec R ->
    irrefl R.
  intros ? ?  h1.
  destruct h1; auto.
Defined.

(*Wyckoff*)
Lemma sort_dec_so :
  forall {T:Type} (R:relation T),
    sort_dec R ->
    Strict_Order R.
  intros ? ? h1.
  destruct h1; constructor; auto.
Defined.




(*Wyckoff*)
Lemma sort_dec_eq :
  forall {T:Type} (R:relation T) (pf pf':sort_dec R),
    pf = pf'.
  intros T R h1 h2.
  destruct h1 as [ha hb hc], h2 as [ha' hb' hc'].
  assert (ha = ha').
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality_dep.
  intro y.
  destruct (ha x y) as [h1 | h1], (ha' x y) as [h2 | h2]; f_equal;
  try apply proof_irrelevance.
  destruct h1 as [h1 | h1], h2 as [h2 | h2]; f_equal; try apply proof_irrelevance; subst.
  pose proof (hc _  h1); contradiction.
  pose proof (hc _  h2); contradiction.
  destruct h1 as [h1 | h1]; subst.
  pose proof (hb _ _ _ h1 h2) as h3.
  apply hc in h3.
  contradiction.
  pose proof (hc _ h2); contradiction.
  destruct h2 as [h2 | h2]; subst.
  pose proof (hb _ _ _ h1 h2) as h3.
  pose proof (hc _ h3); contradiction.
  pose proof (hc _ h1); contradiction.
  subst.
  f_equal; apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma tso_irrefl_eq_dec :
  forall {T:Type} (R:relation T),
    tso_dec R -> irrefl R -> type_eq_dec T.
  intros T R h1 h3.
  red in h1, h3; red.
  intros x y.
  specialize (h1 x y).
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 |]; subst.
  right.
  intro; subst.
  apply h3 in h1; auto.
  left; auto.
  right.
  intro; subst.
  apply h3 in h1; auto.
Qed.


(*Wyckoff*)
Lemma sort_dec_eq_dec :
  forall {T:Type} (R:relation T),
    sort_dec R -> type_eq_dec T.
  intros T R h1.
  destruct h1 as [h1 ? ?]; eapply tso_irrefl_eq_dec; try apply h1; auto.
Qed.


(*Wyckoff*)
Lemma sort_dec_tso_dec_eq :
  forall {T:Type} (R:relation T) (pfso:sort_dec R) (pft:tso_dec R),
    sort_dec_tso_dec R pfso = pft.
  intros T R h1 h2.
  unfold sort_dec_tso_dec.
  destruct h1 as [h1 h3 h4].
  apply functional_extensionality_dep. intro x.
  apply functional_extensionality_dep. intro y.
  destruct (h1 x y) as [h5 | h5], (h2 x y) as [h6 | h6]; f_equal;
  try destruct h5 as [h5 | h5]; try destruct h6 as [h6 | h6]; f_equal; subst; try apply proof_irrelevance.
  pose proof (h4 _ h5); contradiction.
  pose proof (h4 _ h6); contradiction.
  pose proof (h3 _ _ _ h5 h6) as h7.
  apply h4 in h7; contradiction.
  pose proof (h4 _ h6); contradiction.
  pose proof (h3 _ _ _ h5 h6) as h7.
  apply h4 in h7; contradiction.
  pose proof (h4 _ h5); contradiction.
Qed.


(*Wyckoff*)
(*Uses LEM*)
Lemma tso_dec_compat :
  forall {T:Type} (R:relation T),
    total_strict_order R ->
    tso_dec R.
  intros T R h1 x y. 
  red in h1. specialize (h1 x y). 
  apply or_to_sumbool in h1.
  destruct h1 as [h1 | h1].
  left; left; auto.
  apply or_to_sumbool in h1.
  destruct h1 as [h1 | h1].
  left; right; auto.
  right; auto.
Qed.

  
(*Wyckoff*)
(*(R Weak decidable*)
Definition Rwd {T:Type} (R:relation T) : T -> T-> Type :=
  (fun x y => {R x y} + {x = y}).

  (*Wyckoff*)
Lemma R_impl_wd :
  forall {T:Type} (R:relation T) (x y:T),
    R x y -> Rwd R x y.
  intros; left; auto.
Qed.

(*Wyckoff*)
Lemma eq_impl_wd :
  forall {T:Type} (R:relation T) (x y:T),
    x = y -> Rwd R x y.
  intros; right; auto.
Qed.

(*Wyckoff*)
Lemma refl_Rwd :
  forall {T:Type} (R:relation T) (x:T),
    Rwd R x x.
  intros; right; auto.
Qed.

(*Wyckoff*)
Definition Rw {T:Type} (R:relation T) : T -> T-> Prop :=
    (fun x y => R x y \/ x = y).

(*Wyckoff*)
Lemma R_impl_w :
  forall {T:Type} (R:relation T) (x y:T),
      R x y -> Rw R x y.
  intros; left; auto.
Qed.

(*Wyckoff*)
Lemma eq_impl_w :
  forall {T:Type} (R:relation T) (x y:T),
    x = y -> Rw R x y.
  intros; right; auto.
Qed.

(*Wyckoff*)
Lemma refl_Rw :
  forall {T:Type} (R:relation T),
      Reflexive (Rw R).
  intros; right; auto.
Qed.

(*Wyckoff*)
Lemma trans_Rw :
    forall {T:Type} (R:relation T),
      Transitive R ->
      Transitive (Rw R).
  intros T R ht; red; intros x y z h1 h2.
  red in h1, h2; red.
  destruct h1 as [h1|], h2 as [h2|]; subst.
  left; eapply ht. apply h1. assumption.
  left; auto.
  left; auto.
  right; auto.
Qed.

(*Wyckoff*)
  Lemma trans_R_Rw :
    forall {T:Type} (R:relation T),
      Transitive R ->
      forall x y z:T,
        R x y -> Rw R y z -> R x z.
    intros T R h1 x y z h2 h3.
    red in h3.
    destruct h3 as [h3|]; subst.
    eapply h1. apply h2. assumption.
    assumption.
  Qed.

  (*Wyckoff*)
  Lemma trans_Rw_R :
    forall {T:Type} (R:relation T),
      Transitive R ->
      forall x y z:T,
        Rw R x y -> R y z -> R x z.
    intros T R h1 x y z h2 h3.
    red in h2. destruct h2 as [h2|]; subst.
    eapply h1. apply h2. assumption.
    assumption.
  Qed.

    
  (*Wyckoff*)
  Lemma antisym_Rw :
    forall {T:Type} (R:relation T),
      Strict_Order R ->
      forall (x y:T),
        Rw R x y -> Rw R y x -> x = y.
    intros T R h1 x y h2 h3.
    red in h2, h3.
    destruct h2 as [h2|], h3 as [h3|]; auto.
    destruct h1 as [h1 h4].
    specialize (h1 _ _ _ h2 h3).
    apply h4 in h1; contradiction.
  Qed.

  (*Wyckoff*)
  Lemma Rwd_impl_w :
    forall {T:Type} (R:relation T) (x y:T),
      Rwd R x y -> Rw R x y.
    intros T R x y h1.
    red in h1; red.
    destruct h1; [left; auto | right; auto].
  Qed.

  (*Wyckoff*)
  Lemma Rw_le_iff :
    forall (m n:nat), Rw lt m n <-> m <= n.
    unfold Rw; intros; split; omega.
  Qed.

  (*Wyckoff*)
  Definition Rwlem_dec {T:Type} (R:relation T) : Type :=
    forall x y:T, {Rw R x y} + {~Rw R x y}.
    
  (*Wyckoff*)
  Lemma R_iffn :
    forall {T:Type} (R:relation T) (pfso:sort_dec R) (a b:T),
      R a b <-> ~ Rw R b a.
    intros T R h1 a b; split; intro h2.
    intro h3.
    pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h3 h2) as h4.
    apply (sort_dec_irrefl _ h1) in h4; auto.
    destruct (sort_dec_tso_dec _ h1 a b) as [h3 | h3].
    destruct h3; auto; subst; auto.
    contradict h2; right; auto.
    contradict h2; left; auto.
  Qed.

  (*Wyckoff*)
  Lemma Rw_iffn :
    forall {T:Type} (R:relation T) (pfso:sort_dec R) (a b:T),
      Rw R a b <-> ~ R b a.
    intros T R h1 a b; split; intro h2.
    intro h3.
    pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h2 h3) as h4.
    apply (sort_dec_irrefl _ h1) in h4; auto.
    destruct (sort_dec_tso_dec _ h1 a b) as [h3 | h3].
    destruct h3; auto; subst; auto.
    left; auto.
    right; auto.
    contradiction.
  Qed.
 
  (*Wyckoff*)
  Lemma sort_dec_Rlem_dec :
    forall {T:Type} (R:relation T),
      sort_dec R ->
      Rlem_dec R.
    intros T R h1.
    destruct h1 as [h1 h2 h3].
    red; intros x y.
    destruct (h1 x y) as [h4 | h4].
    destruct h4 as [h4 | h4].
    left; auto. 
    subst.
    right; intro h4.
    apply h3 in h4; auto.
    right.
    intro h5.
    pose proof (h2 _ _ _ h4 h5) as h6.
    apply h3 in h6; auto.
  Qed.

   
  (*Wyckoff*)
  Lemma sort_dec_Rwlem_dec :
    forall {T:Type} (R:relation T),
      sort_dec R ->
      Rwlem_dec R.
    intros T R h1.
    destruct h1 as [h1 h2 h3].
    red; intros x y.
    destruct (h1 x y) as [h4 | h4].
    destruct h4 as [h4 | h4]. 
    left; left; auto.
    subst.
    left; right; auto. 
    right.
    intro h5.
    pose proof (trans_Rw_R R h2 _ _ _ h5 h4) as h6.
    apply h3 in h6; auto.
  Qed.

  
(*Wyckoff*)
Definition exist_rel {T:Type} (R:relation T) (P:T->Prop) :
  {t:T | P t} -> {t:T | P t} -> Prop :=
  (fun a b => R (proj1_sig a) (proj1_sig b)).

(*Wyckoff*)
Lemma well_founded_exist_rel : 
  forall {T:Type} (R:relation T) (P:T->Prop),
    well_founded R -> well_founded (exist_rel R P).
intros T R P h1.
red. red in h1.
intro x.
destruct x as [x h2].
specialize (h1 x). 
induction h1 as [x h3 h4].
constructor.
intros y h5. destruct y as [y h6].  red in h5. simpl in h5. 
apply h4; auto.
Qed.


(*Wyckoff*)
(*This is used in well-founded-induction types, to sync up 
  to iterators through the relation to prove statements or define
  values.*)
Definition acc_step_exist_rel 
           {T U:Type} (R:relation T) (P:T->Prop) (x:T)
           (pfp:P x) (pff:forall y:T, R y x -> U)
           (y:{y':T | P y'}) 
           (pfr':(exist_rel R P) y (exist _ _ pfp)) : U :=
  pff (proj1_sig y) pfr'.
                             

(*Wyckoff*)
Definition minR 
  {T:Type} (R:relation T) (pf:tso_dec R)
         : T -> T -> T :=
  (fun (x y:T) =>if (pf x y) then x else y).

(*Wyckoff*)
Definition maxR
  {T:Type} (R:relation T) (pf:tso_dec R)
         : T -> T -> T :=
  (fun (x y:T) =>if (pf x y) then y else x).

(*Wyckoff*)
Lemma minR_idem :
  forall {T:Type} (R:relation T) (pf:tso_dec R) (c:T),
    minR R pf c c = c.
  intros T R h1 c; unfold minR. destruct (h1 c c); auto.
Qed.


(*Wyckoff*)
Lemma maxR_idem :
  forall {T:Type} (R:relation T) (pf:tso_dec R) (c:T),
    maxR R pf c c = c.
    intros T R h1 c; unfold maxR. destruct (h1 c c); auto.
Qed.

(*Wyckoff*)
Lemma maxR_comm :
  forall {T:Type} (R:relation T) (pf:sort_dec R) (a b:T),
    let pftso := sort_dec_tso_dec _ pf in
    maxR R pftso a b =
    maxR R pftso b a. 
  intros T R h1 a b htso. unfold maxR.
  destruct (htso a b) as [h2 | h2]. 
  destruct h2 as [h2|]; subst.
  destruct (htso b a) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  reflexivity.
  reflexivity.
  destruct (htso b b) as [h2 | h2]; auto.
  destruct (htso b a) as [h3 | h3]; auto.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
Qed.


(*Wyckoff*)
Lemma minR_comm :
  forall {T:Type} (R:relation T) (pf:sort_dec R) (a b:T),
    let pftso := sort_dec_tso_dec _ pf in
    minR R pftso a b =
    minR R pftso b a.
  intros T R h1 a b htso. unfold minR.
  destruct (htso a b) as [h2 | h2]. 
  destruct h2 as [h2|]; subst.
  destruct (htso b a) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  reflexivity.
  reflexivity.
  destruct (htso b b) as [h2 | h2]; auto.
  destruct (htso b a) as [h3 | h3]; auto.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
Qed.


(*Wyckoff*)
Lemma minR_compat_l :
  forall {T:Type}  (R:relation T) (pf:tso_dec R) (x y:T),
    Rwd R (minR R pf x y) x.
  intros T R h1 x y.
  red.
  unfold minR.
  destruct (h1 x y); [right; auto | left; auto].
Qed.

(*Wyckoff*)
Lemma minR_compat_r :
  forall {T:Type}  (R:relation T) (pf:tso_dec R) (x y:T),
    Rwd R (minR R pf x y) y.
  intros T R h1 x y.
  red.
  unfold minR.
  destruct (h1 x y); auto; try right; auto. 
Qed.

(*Wyckoff*)
Lemma maxR_compat_l :
  forall {T:Type}  (R:relation T) (pf:tso_dec R) (x y:T),
    Rwd R x (maxR R pf x y).
  intros T R h1 x y.
  red.
  unfold maxR.
  destruct (h1 x y); auto; try right; auto.
Qed.

(*Wyckoff*)
Lemma maxR_compat_r  : 
  forall {T:Type}  (R:relation T) (pf:tso_dec R) (x y:T),
    Rwd R y (maxR R pf x y).
  intros T R h1 x y.
  red.
  unfold maxR.
  destruct (h1 x y); [right; auto | left; auto].
Qed.

(*Wyckoff*)
Lemma minR_functional : 
  forall {T:Type} (R:relation T) (pf:tso_dec R) (x x' y y':T),
    x = x' ->
    y = y' ->
    minR R pf x y = 
    minR R pf x' y'.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Lemma maxR_functional : 
  forall {T:Type} (R:relation T) (pf:tso_dec R) (x x' y y':T),
    x = x' ->
    y = y' ->
    maxR R pf x y = 
    maxR R pf x' y'.
  intros; subst; auto.
Qed.


(*Wyckoff*)
Lemma lt_minR :
  forall {T:Type} (R:relation T) (pf:tso_dec R) (pft:Transitive R)
         (x y z:T),
    R x (minR R pf y z) ->
    R x y /\ R x z.
  intros T R h1 h2 x y z h3.
  unfold minR in h3.  
  destruct (h1 y z) as [h4 | h4].
  destruct h4 as [h5 | h5].
  split; auto.
  eapply h2.
  apply h3. assumption.
  subst.
  split; auto.
  pose proof (h2 _ _ _ h3 h4) as h5.
  split; auto.
Qed.

(*Wyckoff*)
Lemma minR_lt :
  forall {T:Type} (R:relation T) (pf:tso_dec R)
         (x y z:T),
    R (minR R pf x y) z ->
    R x z \/ R y z.
  intros T R h1 x y z h3.
  unfold minR in h3. 
  destruct (h1 x y) as [h4 | h4].
  left; auto.
  right; auto.
Qed.

(*Wyckoff*)
Lemma lt_maxR :
  forall {T:Type} (R:relation T) (pf:tso_dec R)
         (x y z:T),
    R x (maxR R pf y z) ->
    R x y \/ R x z.
  intros T R h1 x y z h2.
  unfold maxR in h2.
  destruct (h1 y z) as [h3 | h3].
  right; auto.
  left; auto.
Qed.

(*Wyckoff*)
Lemma maxR_lt :
  forall {T:Type} (R:relation T) (pf:tso_dec R) (pfso:Transitive R)
         (x y z:T),
    R (maxR R pf x y) z ->
    R x z /\ R y z.
  intros T R h1 ht x y z h2.
  unfold maxR in h2. 
  destruct (h1 x y) as [h3 | h3].
  destruct h3 as [h3|]; subst. 
  split; auto.
  eapply ht.
  apply h3. assumption.
  split; auto.
  split; auto.
  eapply ht.
  apply h3.
  assumption.
Qed.

(*Wyckoff*)
Lemma lt_minR_eq :
  forall {T:Type} (R:relation T) (pfs:sort_dec R) (a b:T),
    R a b ->
    let pft := sort_dec_tso_dec _ pfs in 
    minR R pft a b = a.
  intros T R h1 a b h2 h3.
  unfold minR.
  destruct (h3 a b) as [h4 | h4]; auto.
  pose proof (sort_dec_trans R h1 _ _ _ h2 h4) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
Qed.


(*Wyckoff*)
Lemma gt_minR_eq :
  forall {T:Type} (R:relation T) (pfs:sort_dec R) (a b:T),
    R b a ->
    let pft := sort_dec_tso_dec _ pfs in 
    minR R pft a b = b.
  intros T R h1 a b h2 h3.
  unfold minR. 
  destruct (h3 a b) as [h4 | h4]; auto.
  destruct h4 as [h4|]; subst.
  pose proof (sort_dec_trans R h1 _ _ _ h4 h2) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
Qed.


(*Wyckoff*)
Lemma lt_maxR_eq :
  forall {T:Type} (R:relation T) (pfs:sort_dec R) (a b:T),
    R a b ->
    let pft := sort_dec_tso_dec _ pfs in 
    maxR R pft a b = b.
  intros T R h1 a b h2 h3.
  unfold maxR.
  destruct (h3 a b) as [h4 | h4]; auto.
  pose proof (sort_dec_trans R h1 _ _ _ h2 h4) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
Qed.


(*Wyckoff*)
Lemma gt_maxR_eq :
  forall {T:Type} (R:relation T) (pfs:sort_dec R) (a b:T),
    R b a ->
    let pft := sort_dec_tso_dec _ pfs in 
    maxR R pft a b = a.
  intros T R h1 a b h2 h3.
  unfold maxR. 
  destruct (h3 a b) as [h4 | h4]; auto.
  destruct h4 as [h4|]; subst.
  pose proof (sort_dec_trans R h1 _ _ _ h4 h2) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
Qed.

(*Wyckoff*)
Lemma le_minR_maxR :
  forall {T:Type} (R:relation T) (pfs:sort_dec R) (a b:T),
    let pft := sort_dec_tso_dec _ pfs in 
    Rw R (minR R pft a b) (maxR R pft a b). 
  intros T R h1 a b h2.
  unfold minR, maxR.
  destruct (h2 a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  left; auto.
  right; auto.
  left; auto.
Qed.


(*Wyckoff*)
Lemma minR_eq_iff :
  forall {T:Type} (R:relation T) (pf:sort_dec R) 
         (x y z:T),
    let pftso := sort_dec_tso_dec _ pf in
    minR R pftso x y = z <->
    (x = z /\ Rw R x y) \/
    (y = z /\ Rw R y x).
  intros T R h1 x y z ht; split; intro h2; subst; unfold minR.    
  destruct (ht x y) as [h2 | h2].
  left; auto.
  split; auto.
  destruct h2; subst; auto. left; auto. right; auto.
  right.
  split; auto.
  left; auto.
  destruct h2 as [h2 | h2].
  destruct (ht x y) as [h3 | h3].
  destruct h2; auto.
  destruct h2 as [? h2]; subst.
  pose proof (trans_R_Rw _ (sort_dec_trans _ h1)  _ _ _ h3 h2) as h4.
  apply (sort_dec_irrefl R h1) in h4.
  contradiction.
  destruct h2 as [? h3]; subst.
  destruct (ht x z) as [h4 | h4].
  destruct h4 as [h4 | h4].
  pose proof (trans_R_Rw _ (sort_dec_trans _ h1)  _ _ _ h4 h3) as h5.
  apply (sort_dec_irrefl _ h1) in h5.
  contradiction.
  assumption.
  reflexivity.
Qed.


(*Wyckoff*)
Lemma maxR_eq_iff :
  forall {T:Type} (R:relation T) (pf:sort_dec R) 
         (x y z:T),
    let pftso := sort_dec_tso_dec _ pf in
    maxR R pftso x y = z <->
    (x = z /\ Rw R y x) \/
    (y = z /\ Rw R x y).
  intros T R h1 x y z ht; split; intro h2; subst; unfold maxR.    
  destruct (ht x y) as [h2 | h2]. 
  right; auto.
  split; auto.  
  destruct h2; subst; auto. left; auto. right; auto.
  left.
  split; auto.
  left; auto.
  destruct h2 as [h2 | h2].
  destruct (ht x y) as [h3 | h3].
  destruct h2 as [? h2]; subst.
  destruct h3 as [h3 | h3].
  pose proof (trans_R_Rw _ (sort_dec_trans _ h1)  _ _ _ h3 h2) as h4.
  apply (sort_dec_irrefl R h1) in h4.
  contradiction.
  auto. 
  destruct h2 as [? h4]; subst.
  reflexivity.
  destruct h2 as [? h2]; subst.
  destruct (ht x z) as [h4 | h4].
  reflexivity.
  pose proof (trans_R_Rw _ (sort_dec_trans _ h1)  _ _ _ h4 h2) as h5.
  apply (sort_dec_irrefl _ h1) in h5.
  contradiction.
Qed.


(*Wyckoff*)
Lemma tso_lt :
  total_strict_order lt.
  red. intros x y.
  destruct (lt_eq_lt_dec x y) as [h1 | h1].
  destruct h1 as [h1 | h1].
  left; auto.
  right; left; auto.
  right; right; auto.
Qed.

(*Wyckoff*)
Lemma lt_sortable :
  sortable lt.
  apply (sort_intro _ tso_lt lt_trans lt_irrefl).
Defined.


(*Wyckoff*)
Lemma strict_order_not_symm : 
  forall {T:Type} (R:T->T->Prop),
    Strict_Order R -> 
    forall x y:T, R x y -> ~ R y x.
intros T R h1 x y h2. intro h3.
destruct h1 as [h1a h1b].
red in h1a.
specialize (h1a _ _ _ h2 h3).
red in h1b. specialize (h1b x).
contradiction.
Qed.

(*Wyckoff*)
Lemma transitive_exist_rel : 
  forall {T:Type} (R:relation T),
    Transitive R ->
    forall (P:T->Prop),
      Transitive (exist_rel R P).
  intros T R h1 P.
red in h1. red.
intros x y z h0 h0'.
destruct x  as [x h2], y as [y h3], z as [z h4].
red in h0, h0'. simpl in h0, h0'.
specialize (h1 _ _ _ h0 h0').
red. simpl.
assumption.
Qed.

(*Wyckoff*)
Lemma irrefl_exist_rel :
  forall {T:Type} (R:relation T) (P:T->Prop),
    irrefl R ->
    irrefl (exist_rel R P).
intros T R P h1.
red in h1. red.
intro x. destruct x as [x h2].
specialize (h1 x).
intro h3. red in h3.
simpl in h3.
contradiction.
Qed.


(*Wyckoff*)
Lemma strict_order_exist_rel : 
  forall {T:Type} (R:relation T),
    Strict_Order R ->
    forall (P:T->Prop),
      Strict_Order (exist_rel R P).
intros T R h1 P. destruct h1.
constructor; try apply irrefl_exist_rel; try apply transitive_exist_rel; auto.
Qed.

(*Wyckoff*)
Lemma so_lt_nat :
  Strict_Order lt.
  constructor.
  red; intro; apply lt_trans.
  red; intros; apply lt_irrefl.
Qed.

(*Maybe replace [tso_dec_lt_nat] with [lt_nat_tso_dec]*)
(*Wyckoff*)
Lemma tso_dec_lt_nat :
  tso_dec lt.
  red; apply lt_eq_lt_dec.
Qed.

(*Wyckoff*)
Lemma sort_dec_lt_nat :
  sort_dec lt.
  constructor.
  apply tso_dec_lt_nat.
  red; apply lt_trans.
  red; apply lt_irrefl.
Qed.

(*Wyckoff*)
Lemma tso_dec_lt_nat_irrelevance :
  forall (pf pf':forall x y:nat, {x < y} + {x = y} + {y < x}),
    pf = pf'.
  intros h1 h2.
  apply functional_extensionality_dep.
  intro m. apply functional_extensionality_dep.
  intro n.
  destruct (h1 m n) as [h3 | h3], (h2 m n) as [h4 | h4]; f_equal.
  destruct h3 as [h3 | h3], h4 as [h4 | h4]; f_equal; try apply proof_irrelevance; subst; try omega.
  destruct h3; omega.
  destruct h4; omega.
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma Rlem_dec_lt_nat :
  Rlem_dec lt.
  apply sort_dec_Rlem_dec;
  apply sort_dec_lt_nat.
Qed.

(*Wyckoff*)
Definition Rcompose
           {T U:Type} (R:relation U) (f:T->U) : relation T :=
  fun x y => R (f x) (f y).

(*Wyckoff*)
Lemma Rcompose_tso_dec : 
  forall {T U:Type} (R:relation U) (f:T->U),
    injective f ->
    tso_dec R ->
    tso_dec (Rcompose R f).
  unfold Rcompose. intros T U R f h0 h1.
  red in h1; red.
  intros x y.
  specialize (h1 (f x) (f y)); auto.
  destruct h1 as [[h1 | h1] | h1].
  left; auto.
  apply h0 in h1; subst.
  left; right; auto.
  right; auto.
Qed.

(*Wyckoff*)
Lemma Rcompose_trans : 
  forall {T U:Type} (R:relation U) (f:T->U),
    Transitive R ->
    Transitive (Rcompose R f).
  unfold Rcompose. intros T U R f h1.
  red in h1; red.
  intros x y z h2 h3.
  specialize (h1 (f x) (f y) (f z) h2 h3); auto.
Qed.

(*Wyckoff*)
Lemma Rcompose_irrefl : 
  forall {T U:Type} (R:relation U) (f:T->U),
    irrefl R ->
    irrefl (Rcompose R f).
  unfold Rcompose. intros T U R f h1.
  red in h1; red.
  intro x.
  apply (h1 (f x)); auto.
Qed.


(*Wyckoff*)
Definition Rcompose_sort_dec : 
  forall {T U:Type} (R:relation U) (f:T->U),
    injective f ->
    sort_dec R ->
    sort_dec (Rcompose R f).
  intros T U R f h0 h1.
  destruct h1 as [h1 h2 h3].
  constructor.
  apply Rcompose_tso_dec; auto.
  apply Rcompose_trans; auto.
  apply Rcompose_irrefl; auto.
Qed.

Inductive disagrees {T U} : relation (T->U) :=
| disagrees_intro : forall c f g, f c <> g c -> disagrees f g.
(*Do a min [seg_t] function.*)

Inductive disagrees_at {T U} {f g:T->U} : disagrees f g -> T -> Prop :=
| disagrees_at_intro : forall c (pfn:f c <> g c),
                         disagrees_at (disagrees_intro c f g pfn) c.
  
(*See FunctionProperties2 for a minimal version of [disagrees_at], 
and decidable lemmas.*)


Definition pred_eq {T} : relation (T->Prop) :=
  fun P Q =>
    forall x:T, P x <-> Q x.

Lemma refl_pred_eq :
  forall T, Reflexive (@pred_eq T).
  unfold Reflexive, pred_eq; intros; tauto.
Qed.

Lemma sym_pred_eq :
  forall T, Symmetric (@pred_eq T).
  unfold Symmetric, pred_eq; intros ? ? ? h1 ?; rewrite h1; tauto.
Qed.

Lemma trans_pred_eq :
  forall T, Transitive (@pred_eq T).
  unfold Transitive, pred_eq; intros ? ? ? ? h1 h2 ?; rewrite h1; auto.
Qed.

Add Parametric Relation {T} :
  (T->Prop)
    (@pred_eq T)
    reflexivity proved by (refl_pred_eq T)
    symmetry proved by (sym_pred_eq T)
    transitivity proved by (trans_pred_eq T)
      as pred_eq_eq.

Definition proj1_eq' {T} : relation {P:T->Prop & {x:T|P x}} :=
  fun x y => proj1_sig (projT2 x) = proj1_sig (projT2 y).

Lemma refl_proj1_eq' :
  forall T, Reflexive (@proj1_eq' T).
  unfold Reflexive, proj1_eq'; intros; auto.
Qed.

Lemma sym_proj1_eq' :
  forall T, Symmetric (@proj1_eq' T).
  unfold Symmetric, proj1_eq'; intros; auto.
Qed.

(*Wyckoff*)
Lemma trans_proj1_eq' :
  forall T, Transitive (@proj1_eq' T).
  unfold Transitive, proj1_eq'; intros T x y z; auto.
  destruct x, y, z; simpl; auto.
  destruct s, s0, s1; simpl.
  intros; subst; auto.
Qed.

(*Wyckoff*)
Add Parametric Relation {T} :
  {P:T->Prop & {x:T|P x}} 
    (@proj1_eq' T)
    reflexivity proved by (@refl_proj1_eq' T)
    symmetry proved by (@sym_proj1_eq' T)
    transitivity proved by (@trans_proj1_eq' T)
      as proj1_eq_eq'.

(*Wyckoff*)
Definition proj1_eq {T} (P:T->Prop) : relation {x:T|P x} :=
  fun x y => proj1_sig x = proj1_sig y.

(*Wyckoff*)
Lemma refl_proj1_eq :
  forall {T P},
    Reflexive (@proj1_eq T P).
  unfold Reflexive, proj1_eq; intros; auto.
Qed.

(*Wyckoff*)
Lemma sym_proj1_eq :
  forall {T P},
    Symmetric (@proj1_eq T P).
  unfold Symmetric, proj1_eq; intros; auto.
Qed.

(*Wyckoff*)
Lemma trans_proj1_eq :
  forall {T P},
    Transitive (@proj1_eq T P).
  unfold Transitive, proj1_eq; intros; congruence.
Qed.



(*Wyckoff*)
Definition incrR_fun {T U:Type} (R:relation T) (R':relation U)
           (f:T->U) : Prop :=
  forall a b:T, R a b -> R' (f a) (f b).

(*Wyckoff*)
Lemma incrR_fun_inj :
  forall {T U:Type} (R:relation T) (R':relation U)
         (f:T->U),
    tso_dec R ->
    irrefl R'->
    incrR_fun R R' f ->
    injective f.
  intros T U R R' f htso hir h1.
  red; intros x y h2. 
  red in h1.  
  destruct (htso x y) as [[h3|h3]|h3]; try apply h1 in h3; auto.
  rewrite h2 in h3.
  apply hir in h3; contradiction.
  rewrite h2 in h3.
  apply hir in h3; contradiction.
Qed.



(*Wyckoff*)
Definition incr_fun (f:nat->nat) : Prop :=
  forall n:nat, f n < f (S n).

(*Wyckoff*)
Definition incr_fun' (f:nat->nat) : Prop :=
  incrR_fun lt lt f.

(*Wyckoff*)
Lemma incr_fun_iff :
  forall (f:nat->nat),
    incr_fun f <-> incr_fun' f.
  intro f.
  split; intro h1; red in h1; red.
  intros a b h2.
  revert h2. revert a h1. revert  f.
  induction b  as [|b ih]; simpl.
  intros; omega.
  intros f a h1 h2.
  specialize (ih f a h1).
  red in h2.
  apply le_S_n in h2.
  apply le_lt_eq_dec in h2.
  destruct h2 as [h2 | h2].
  specialize (ih h2).
  specialize (h1 b).
  omega.
  subst.
  apply h1.

  intro; apply h1; auto with arith.
Qed.

(*Wyckoff*)
Lemma incr_fun_S :
  incr_fun S.
  red; intros; auto with arith.
Qed.


(*Wyckoff*)
Lemma incr_fun_inj :
  forall (f:nat -> nat),
    incr_fun f ->
    injective f.
  intros ? h1.
  rewrite incr_fun_iff in h1.
  unfold incr_fun' in h1.
  apply incrR_fun_inj in h1; auto.
  apply tso_dec_lt_nat.
  red; apply lt_irrefl.
Qed.


(*Wyckoff*)
Definition weak_incrR_fun {T U:Type} (R:relation T) (R':relation U)
           (f:T->U) : Prop :=
  forall a b:T, R a b -> Rw R' (f a) (f b).


(*Wyckoff*)
Lemma incrR_fun_impl_weak :
  forall {T U:Type} (R:relation T) (R':relation U)
         (f:T->U),
    incrR_fun R R' f ->
    weak_incrR_fun R R' f.
  intros T U R R' f h1.
  red in h1. red.
  intros; left; auto.
Qed.


(*Wyckoff*)
Definition weak_incr_fun (f:nat->nat) : Prop :=
  forall n:nat, f n <= f (S n).

(*Wyckoff*)
Definition weak_incr_fun' (f:nat->nat) : Prop :=
  weak_incrR_fun lt lt f.

(*Wyckoff*)
Definition weak_incr_fun_iff :
  forall (f:nat->nat),
    weak_incr_fun f <-> weak_incr_fun' f.
  intros f; split; intro h1.
  red in h1; red.
  intros a b h2. revert h2. revert a h1. revert f.
  induction b as [|b ih]; simpl.
  intros; omega.
  intros f a h1 h2.
  specialize (ih f a h1).
  red in h2.
  apply le_S_n in h2.
  apply le_lt_eq_dec in h2.
  destruct h2 as [h2 | h2].
  specialize (ih h2).
  specialize (h1 b).  
  rewrite Rw_le_iff in ih.
  rewrite Rw_le_iff.
  omega.
  subst.
  rewrite Rw_le_iff.
  apply h1.

  red; red in h1; intro m; specialize (h1 m (S m)); rewrite Rw_le_iff in h1; auto. 
Qed.

(*Wyckoff*)
Lemma incr_fun_impl_weak :
  forall (f:nat->nat),
    incr_fun f ->
    weak_incr_fun f.
  intros f h1.
  red in h1. red.
  intro; auto with arith.
Qed.

(*Wyckoff*)
Lemma weak_incr_fun_pred :
  weak_incr_fun pred.
  red; intros; simpl; auto with arith.
Qed.

(*Wyckoff*)
Lemma weak_incr_fun_S :
  weak_incr_fun S.
  apply incr_fun_impl_weak.
  apply incr_fun_S.
Qed.


End Relations.

(*Wyckoff*)
Definition P_to_dep {T} (P Q:T->Prop) : {x:T | Q x} -> Prop :=
  fun p =>
    let (x, _) := p in
    P x.


(*Wyckoff*)
(*This is a lot of how-do-you-do about merely substituting 
one value of a [seg_fun] [f] at a given integer [k] with 
the value [y:U], and all of [f]s other points remaining 
the same.  The simpler way to do it with [eq_nat_dec] 
is inappropriate for arbitrary computation
(i.e. it involves equality in [Prop], rather than 
[Set].*)
(*Wyckoff*)
Fixpoint seg_fun_subst
         {U:Type} {n:nat} :
  seg_fun n U -> U ->
  forall k, k < n -> seg_fun n U :=
  match n
  with
    | O => fun _ _ _  pf _ _ => False_rect _ (lt_n_0 _ pf)
    | S n' => fun f y k =>
                let f' := seg_fun_to_S f in
                match k with
                  | 0 =>
                    fun pfk i =>
                      match i with
                        | 0 => fun _ => y
                        | S i' => fun pfi => f i pfi
                      end
                  | S k' =>
                    fun pfk =>
                      let pfk' := lt_S_n _ _ pfk in
                      fun i =>
                        match i with
                          | 0 =>
                            fun pfi => f 0 pfi
                          | S i' =>
                            fun pfi =>
                              let pfi' := lt_S_n _ _ pfi in
                              seg_fun_subst f' y k' pfk' i' pfi'
                        end
                end
  end.

(*Wyckoff*)
Lemma seg_fun_subst_compat :
  forall {U n} (f:seg_fun n U) y
         k (pfk:k < n) i (pfi:i < n),
    let f' := seg_fun_subst f y k pfk in
    (i = k -> f' i pfi = y) /\
    (i <> k -> f' i pfi = f i pfi).
  intros U n.
  induction n as [|n ih].
  intros; omega. 
  intros f y k.
  destruct k as [|k]. 
  intros hk i.
  destruct i as [|i]; simpl.
  intros; tauto.
  intros.
  split.
  intro; discriminate.
  intro; auto. 
  intros hk i.
  destruct i as [|i].
  intros; split; auto.
  intro; discriminate.
  intros hi f'.
  pose proof (lt_S_n _ _ hk) as hk'.
  pose proof (lt_S_n _ _ hi) as hi'.
  specialize (ih (seg_fun_to_S f) y k hk' i hi').
  simpl in ih.
  destruct ih as [h2 h3].
  split.
  intro h4.
  inversion h4; subst.
  specialize (h2 (eq_refl _)).
  rewrite <- h2.
  unfold f'.
  simpl.
  f_equal.
  apply proof_irrelevance.
  apply proof_irrelevance.
  intro h4. 
  unfold f'.
  simpl.
  pose proof (proof_irrelevance _ hi' (lt_S_n _ _ hi)); subst.
  pose proof (proof_irrelevance _ hk' (lt_S_n _ _ hk)); subst.
  rewrite h3; auto with arith.
  unfold seg_fun_to_S.
  f_equal.
  apply proof_irrelevance.
Qed.

(*Wyckoff*)
Fixpoint seg_fun_prod_type {n} : seg_fun n Type -> Type :=
  match n with
    | 0 => fun _ => unit
    | S n' => fun f =>
                let f' := seg_fun_to_S f in
                ((f 0 (lt_0_Sn _)) * (seg_fun_prod_type f'))%type
  end.



(*Wyckoff*)
Definition seg_fun_prod_type_to_S {n}
           {f:seg_fun (S n) Type}
           (g:seg_fun_prod_type f) :
  seg_fun_prod_type (seg_fun_to_S f) :=
  snd g.

(*Wyckoff*)
Lemma seg_fun_prod_type_S_eq :
  forall {n}
         {f:seg_fun (S n) Type}
         (g:seg_fun_prod_type f),
    g = (fst g,
         seg_fun_prod_type_to_S g).
  intros; apply surjective_pairing.
Qed.

(*Wyckoff*)
Fixpoint seg_fun_sum_type {n} : seg_fun n Type -> Type :=
  match n with
    | 0 => fun _ => unit
    | S n' => fun f =>
                let f' := seg_fun_to_S f in
                ((f 0 (lt_0_Sn _)) + (seg_fun_sum_type f'))%type
  end.


(*Wyckoff*)
Fixpoint seg_fun_prod {n} :
  forall {f:seg_fun n Type}
         (g:seg_fun_dep f),
    seg_fun_prod_type f :=
  match n with
    | 0 => fun (f:seg_fun 0 Type) _ => tt
    | S n' => fun f g =>
                let f' := seg_fun_to_S f in
                let g' := seg_fun_dep_to_S g in
                (g 0 (lt_0_Sn _),
                 seg_fun_prod g')
  end.

(*Wyckoff*)
Definition seg_fun_prod_from_S
           {n}
           {f:seg_fun (S n) Type}
           (g:seg_fun_dep f) :
  seg_fun_prod_type (seg_fun_from_S f) :=
  seg_fun_prod (seg_fun_dep_from_S g).

(*Wyckoff*)
Definition seg_fun_prod_to_S
           {n}
           {f:seg_fun (S n) Type}
           (g:seg_fun_dep f) :
  seg_fun_prod_type (seg_fun_to_S f) :=
  seg_fun_prod (seg_fun_dep_to_S g).

(*Wyckoff*)
Lemma seg_fun_prod_S_eq :
  forall {n}
         {f:seg_fun (S n) Type}
         (g:seg_fun_dep f),
    seg_fun_prod g = (g 0 (lt_0_Sn _), seg_fun_prod_to_S g).
  auto.
Qed.

(*Wyckoff*)
Definition seg_fun_dep_pf
           {U n}
           (f:seg_fun n U)
           (b:forall i (pfi:i < n), i < n) :
  forall i (pfi:i < n), U :=
  fun i pfi => f i (b i pfi).


(*Wyckoff*)
Lemma seg_fun_dep_pf_eq :
  forall {U n}
         (f:seg_fun n U)
         (b:forall i (pfi:i < n), i < n),
    seg_fun_dep_pf f b = f.
  unfold seg_fun_dep_pf; intros; apply functional_extensionality_dep;
  intro; apply functional_extensionality_dep;
  intro; f_equal; apply proof_irrelevance.
Qed.

(*Wyckoff*)
Lemma seg_fun_dep_pf_eq2 : 
  forall {U n}
         (f:seg_fun n U)
         (b c:forall i (pfi:i < n), i < n),
    seg_fun_dep_pf f b =
    seg_fun_dep_pf f c.
  unfold seg_fun_dep_pf; intros;
  apply functional_extensionality_dep; intro;
  apply functional_extensionality_dep; intro;
  apply f_equal_dep; auto.
Qed.



(*Wyckoff*)
Definition dep_transfer_seg_fun
           {U n}
           {f g:seg_fun n U}
           (pfe:f = g)
           (V:forall m, seg_fun m U -> Type)
           (Vs:V _ f) :
  V _ g.
  intros; subst; auto.
Defined.


(*Wyckoff*)
Definition add_seg_to_identify n : seg_fun (S n) nat :=
  match n with
    | 0 => fun _ pf => 0 
    | S n' => 
      fun i =>
        match i with
          | 0 => fun _ => 0
          | S i' => fun _ =>  i'
        end
  end.

(*Wyckoff*)
Lemma lt_add_seg_to_identify :
  forall n i pf,
    0 < n ->
    add_seg_to_identify n i pf < n.
  intro n. 
  induction n as [|n ih]; auto; simpl.
  intros i ?.
  destruct i; simpl; auto with arith.
Qed.

(*Wyckoff*)
Definition tl_seg_for_fictitious n : seg_fun (pred n)  nat :=
  match n with
    | 0 =>
      fun _ pf => False_rect _ (lt_n_0 _ pf)
    | S n' =>
      match n' with
        | 0 => 
          fun _ pf => False_rect _ (lt_n_0 _ pf)
        | S _ => 
          fun i _ => S i
      end
  end.

(*Wyckoff*)
Lemma lt_tl_seg_for_fictitious :
  forall n i pf,
    tl_seg_for_fictitious n i pf < n.
  intro n; destruct n; simpl; auto with arith. 
  intros ? h1; contradict (lt_n_0 _ h1).
  destruct n as [|n]; auto with arith.
  intros ? h1; contradict (lt_n_0 _ h1).
Qed.

(*Wyckoff*)
Lemma inj_option_map : 
  forall {T U} (f:T->U),
    injective f ->
    injective (option_map f).
  intros T U f h1; red in h1; red; intros x y h2; destruct x as [x|], y as [y|]; simpl in h2; try discriminate; auto.
  inversion h2.
  apply h1 in H0; subst; auto.
Qed.

(*Wyckoff*)
Lemma option_map_inj :
  forall {T U} (f:T->U),
    injective (option_map f) ->
    injective f.
  intros T U f h1; red in h1; red; intros x y h2.
  specialize (h1 (Some x) (Some y)).
  simpl in h1.
  rewrite h2 in h1.
  specialize (h1 eq_refl).
  inversion h1; auto.
Qed.

(*Wyckoff*)
Lemma surj_option_map : 
  forall {T U} (f:T->U),
    surjective f ->
    surjective (option_map f).
  intros T u f h1; red in h1; red.
  intro y.
  destruct y as [y|].
  specialize (h1 y).
  destruct h1 as [x]; subst.
  exists (Some x); simpl; auto.
  exists None; simpl; auto.
Qed.

(*Wyckoff*)
Lemma option_map_surj : 
  forall {T U} (f:T->U),
    surjective (option_map f) ->
    surjective f.
  intros T U f h1; red in h1; red.
  intro y; specialize (h1 (Some y)).
  destruct h1 as [x h2].
  destruct x as [x|]; simpl in h2.
  inversion h2; subst.
  exists x; auto.
  discriminate.
Qed.

(*Wyckoff*)
Lemma bij_option_map : 
  forall {T U} (f:T->U),
    bijective f ->
    bijective (option_map f).
  intros ? ? ? h1; red in h1; red; split; destruct h1;
  [apply inj_option_map | apply surj_option_map]; auto.
Qed.

(*Wyckoff*)
Lemma option_map_bij : 
  forall {T U} (f:T->U),
    bijective (option_map f) ->
    bijective f.
  intros ? ? ? h1; red in h1; red; split; destruct h1 as [h1 h2];
  [apply option_map_inj in h1 | apply option_map_surj in h2]; auto.
Qed.

(*Wyckoff*)
Lemma invertible_option_map : 
  forall {T U} (f:T->U),
    invertible f ->
    invertible (option_map f).
  intros ? ? ? h1; apply bijective_impl_invertible;
  apply invertible_impl_bijective in h1; auto;
  apply bij_option_map; auto.
Qed.

(*Wyckoff*)
Lemma option_map_invertible : 
  forall {T U} (f:T->U),
    invertible (option_map f) ->
    invertible f.
  intros ? ? ? h1;
  apply invertible_impl_bijective in h1; 
  apply bijective_impl_invertible;
  apply option_map_bij in h1; auto.
Qed.

(*In pre-image of function.*)
(*Wyckoff*)
Inductive in_pre_aux {X S} (v:X->S) : S -> Prop :=
  in_pre_intro : forall x, in_pre_aux v (v x).

(*Wyckoff*)
Definition in_pre {X S} (s:S) (v:X->S) : Prop :=
  in_pre_aux v s.

(*Wyckoff*)
Inductive inh_ind_var {X S} (v:X->S) : Prop :=
  inh_ind_var_intro : forall s, in_pre s v -> inh_ind_var v.

