(* Copyright (C) 2014-2019, Daniel Wyckoff*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- version 8.4 pl4 Coq.*)

Require Export Arith.
Require Import Omega.
Require Import NPeano.
Require Import ProofIrrelevance.
Require Import LogicUtilities.
Require Import Ensembles.
Require Import Even.
Require Import Equality.
Require Import Description.
Require Import FunctionProperties.
Require Import EqDec.
Require Import FunctionalExtensionality.
Require Import DecidableDec.

Notation le_lt := le_lt_n_Sm.
Notation lt_le := lt_n_Sm_le.


Section Trivialities.

Lemma disc_nat :
  forall n,
    S n <> 0.
  auto.
Qed.
  
Lemma pred0 :
  pred 0 = 0.
  auto.
Qed.

Lemma lt_0_1 :
  0 < 1.
  auto with arith.
Qed.
  
Lemma lt_0_2 :
  0 < 2.
  auto with arith.
Qed.

Lemma lt_1_2 :
  1 < 2.
  auto with arith.
Qed.

Lemma le0 :
  forall (m:nat),
    m <= 0 ->
    m = 0.
  intro m; destruct m; auto with arith.
Qed.

Lemma le0' :
  forall (m:nat),
    m = 0 ->
    m <= 0.
  intros; subst; auto with arith.
Qed.

Lemma le_1_S :
  forall n,
    1 <= S n.
  auto with arith.
Qed.

Lemma ge_S_1 :
  forall n,
    S n >= 1.
  auto with arith.
Qed.


Lemma lt1 : forall (n:nat), n < 1 -> n = 0.
  intros; omega.
Qed.

Lemma lt1' :
  forall (n:nat), n = 0 -> n < 1.
  intros; subst; auto.
Qed.

Lemma lt_1_SSn :
  forall n,
    1 < (S (S n)).
  intros; auto with arith.
Qed.

Lemma lt2_dec : forall (n:nat), n < 2 -> {n = 0} + {n = 1}.
  intros n h1.
  destruct n as [|n].
  left; auto.
  right.
  apply lt_S_n in h1.
  apply lt1 in h1; subst; auto.
Qed.

Lemma Olt :
  forall m n,
    m < n -> 0 < n.
  intros; omega.
Qed.

Lemma gt0 : 
  forall n:nat, 0 < n ->
                n = S (pred n).
  intros; omega.
Qed.


Lemma gt1 : 
  forall n:nat, 1 < n ->
                n = S (S (pred (pred n))).
intro n.
induction n; omega.
Qed.

Lemma lt_pred' :
  forall m n,
    m < pred n ->
    m < n.
  intros; omega.
Qed.

Lemma lt_pred_S :
  forall m n,
    m < pred n ->
    S m < n.
  intros; omega.
Qed.

Lemma lt_le_pred :
  forall m n,
    m < n ->
    m <= pred n.
  intros; omega.
Qed.

Lemma le_S_pred :
  forall m n,
    m <= S n ->
    pred m <= n.
    intros; omega.
Qed.

Lemma le_pred_S :
  forall m n,
    pred m <= n ->
    m <= S n.
    intros; omega.
Qed.


Lemma ge_S_n :
  forall m n,
    S m >= S n ->
    m >= n.
  auto with arith.
Qed.

Lemma ge_n_S :
  forall m n,
    m >= n ->
    S m >= S n.
  auto with arith.
Qed.

Lemma ge_Sn_n :
  forall n,
    S n >= n.
  auto with arith.
Qed.
  
  
Lemma eq_le :
  forall (m n:nat),
    m = n ->
    m <= n.
  intros; subst; auto.
Qed.

Lemma eq_le' :
  forall (m n:nat),
    m = n ->
    n <= m.
  intros; subst; auto.
Qed.

Lemma eq_plus_l :
  forall m n,
    n = n + m ->
    m = 0.
  intros; omega.
Qed.

Lemma eq_plus_r :
  forall m n,
    n = m + n ->
    m = 0.
  intros; omega.
Qed.


Lemma lt_gt :
  forall m n,
    m < n -> n > m.
  auto.
Qed.

Lemma gt_lt :
  forall m n,
    m > n -> n < m.
  auto.
Qed.

Lemma le_ge :
  forall m n,
    m <= n -> n >= m.
  auto.
Qed.

Lemma ge_le :
  forall m n,
    m >= n -> n <= m.
  auto.
Qed.

Lemma nlt_ngt :
  forall m n,
    ~ m < n -> ~ n > m.
  auto.
Qed.

Lemma ngt_nlt :
  forall m n,
    ~ m > n -> ~ n < m.
  auto.
Qed.

Lemma nle_nge :
  forall m n,
    ~ m <= n -> ~ n >= m.
  auto.
Qed.

Lemma nge_nle :
  forall m n,
    ~ m >= n -> ~ n <= m.
  auto.
Qed.

Lemma lt_not_le :
  forall m n,
    m < n -> ~n <= m.
  intros; omega.
Qed.

Lemma not_le_lt :
  forall m n,
    ~n <= m -> m < n.
  intros; omega.
Qed.

Lemma gt_not_ge :
  forall m n,
    m > n -> ~n >= m.
  intros; omega.
Qed.

Lemma not_ge_gt :
  forall m n,
    ~n >= m -> m > n.
  intros; omega.
Qed.


                  
Lemma lt_impl_le :
  forall m n,
    m < n ->
    m <= n.
  auto with arith.
Qed.

Lemma lt_neq :
  forall m n,
    m < n ->
    m <> n.
  intros; intro; subst; apply lt_irrefl in H; auto.
Qed.

Lemma gt_neq :
  forall m n,
    m > n ->
    m <> n.
  intros; intro; subst; apply lt_irrefl in H; auto.
Qed.

Lemma neq_lt0 :
  forall n,
    0 <> n ->
    0 < n.
  auto with arith.
Qed.

Lemma lt_neq_impl :
  forall m n, 
    m < n ->
    m <> pred n ->
    m < pred n.
  intros; omega.
Qed.


Lemma neq_dec :
  forall m n,
    m <> n ->
    {m < n} + {n < m}.
  intros m n h1.
  destruct (lt_eq_lt_dec m n) as [[h2 | h2] | h2]; subst.
  left; auto.
  contradict h1; auto.
  right; auto.
Qed.




Lemma ge1_lt : 
  forall (m w:nat),
    1 <= m < 1 + w ->
    m < S w.
  intros; omega.
Qed.

Lemma neq0_iff :
  forall (n:nat),
    n <> 0 <-> 0 < n.
  intro n; destruct n; split; auto with arith.
Qed.


Lemma not_not0_iff :
  forall (n:nat),
    ~ n <> 0 <-> n = 0.
  intro n; destruct n; omega.
Qed.


Lemma not_lt0_iff :
  forall (n:nat),
    ~ 0 < n <-> n = 0.
  intro; omega.
Qed.

Lemma not_lt_le :
  forall (m n:nat),
    ~ n < m -> m <= n.
  intros; omega.
Qed.


Lemma not_not_nat :
  forall (m n:nat),
    ~ n <> m -> n = m.
  intros; omega.
Qed.


Lemma nat_eq_neq :
  forall (m n p:nat),
    m = n /\ n <> p -> m <> p.
  intros; omega.
Qed.

Lemma nat_eq_neq_iff :
  forall (m n p:nat),
    m = n /\ n <> p <-> m = n /\ m <> p.
  intros; split; omega.
Qed.

Lemma S_eq_iff :
  forall m n:nat,
    S m = n <->
    m = pred n /\ n > 0.
  intros; split; omega.
Qed.

Lemma lt_eq_trans :
  forall (m n p:nat),
    m < n -> n = p -> m < p.
  intros; subst; auto with arith.
Qed.

Lemma le_eq_trans :
  forall (m n p:nat),
    m <= n -> n = p -> m <= p.
  intros; subst; auto with arith.
Qed.


Lemma eq_lt_trans :
  forall (m n p:nat),
    m = n ->
    n < p ->
    m < p.
  intros; subst; auto.
Qed.


Lemma eq_le_trans :
  forall (m n p:nat),
    m = n ->
    n <= p ->
    m <= p.
  intros; subst; auto.
Qed.

Lemma mult2 :
  forall n:nat,
    2 * n = n + n.
  intro; omega.
Qed.

Lemma lt_mod :
  forall k n
         (pfn : 0 < n),
    k mod n < n.
  intros k n h1. 
  pose proof (mod_bound_pos k n (le_0_n _) h1) as h2.
  destruct h2; auto with arith.
Qed.

Lemma lt_mod' :
  forall r k n
         (pfn : r < n),
    k mod n < n.
  intros; apply lt_mod; omega.
Qed.

Lemma le_mod :
  forall k n,
    k mod n <= k.
  intros k n.
  destruct (zerop n) as [|h1]; subst; simpl; auto with arith.
  apply neq0_iff in h1.
  pose proof (div_mod k n h1) as h2.
  pose proof (le_plus_l (k mod n) (n * (k / n))) as h3.
  rewrite plus_comm in h3.
  rewrite <- h2 in h3; auto.
Qed.


Lemma le_pred_sum :
  forall i n m,
    i < n ->
    i <= m + (pred n).
  intros; omega.
Qed.


Lemma le_pred_sum' :
  forall i n m,
    i < n ->
    i <= m + n - 1.
  intros; omega.
Qed.

Lemma le_pred_sum_minus :
  forall i m n,
    i < n ->
    m <= m + (pred n) - (pred i).
  intros; omega.
Qed.

(*This is the symmetric equality-flip of [minus_n_n].*)
Lemma minus_same :
  forall n,
    n - n = 0.
  auto with arith.
Qed.

Lemma sum_pred_minus :
  forall i m n,
    0 < i ->
    i < n ->
    m + pred n - pred i - m = n - i.
  intros; omega.
Qed.


Lemma S_pred_plus_minus :
  forall n i,
    0 < i ->
    i < n ->
    S (pred i) + (n - i) = n.
  intros; omega.
Qed.

Lemma le_pred_sum_pos :
  forall m n,
    0 < m ->
    0 < n ->
    m <= pred (m + n).
  intros; omega.
Qed.

Lemma S_pred_sum_minus :
  forall m n,
    0 < m ->
    0 < n ->
    S (pred (m + n) - m) = n.
  intros; omega.
Qed.

  
  
Lemma O_mod :
  forall n,
    0 mod n = 0.
  intro n.
  induction n; simpl; auto with arith.
Qed.

Lemma O_div :
  forall n,
    0 / n = 0.
  intro n.
  induction n; simpl; auto with arith.
Qed.

Lemma plus_S_shift_l : forall (m n:nat), (S m + n) = S (m + n).
  intros; omega.
Qed.


Lemma S_compat : forall (m:nat), m + 1 = S m.
intros m'. induction m' as [|m h1]; auto with arith.
rewrite <- h1 at 2.
rewrite h1.
rewrite plus_comm.
rewrite plus_Snm_nSm.
auto with arith.
Qed.

Lemma plus_S_shift_r : forall (m n:nat), (m + S n) = S (m + n).
intros m n.
assert (h1:S n = 1 + n). auto with arith.
rewrite h1.
rewrite plus_permute.
rewrite plus_comm.
simpl. auto with sets.
apply S_compat.
Qed.

Lemma plus_eq_iff :
  forall i j,
    i = i + j <->
    j = 0.
  intros; omega.
Qed.

Lemma minus_eq_iff :
  forall i j,
    i = i - j <->
    i = 0 \/ j = 0.
  intros; omega.
Qed.



Lemma le1_dec :
  forall n,
    n <= 1 ->
    {n = 0} + {n = 1}.
  intros n h1.
  destruct n as [|n].
  left; auto.
  apply le_S_n in h1.
  apply le0 in h1; subst.
  right; auto.
Qed.

Lemma le2_dec :
  forall n,
    n <= 2 ->
    {n = 0} + {n = 1} + {n = 2}.
  intros n h1.
  destruct n as [|n].
  left; auto.
  apply le_S_n in h1.
  apply le1_dec in h1.
  destruct h1; subst.
  left; right; auto.
  right; auto.
Qed.


Lemma S_plus_S : 
  forall m n,
    S m + S n =
    S (S (m + n)).
  intros; omega.
Qed.
  
End Trivialities.


Lemma mono_pred_lt :
  forall (m n:nat),
    m < n -> 0 < m  ->
    pred m < pred n.
  intros m n h1 h2.
  apply lt_S_n.
  rewrite <- (S_pred _ _ h2); auto.
  assert (h3:0 < n). eapply lt_trans. apply h2. assumption.
  rewrite <- (S_pred _ _ h3).
  assumption.
Qed.


Section Minus.


Lemma zero_minus :
  forall (m:nat),
    0 - m = 0.
  intro; auto with arith.
Qed.

Lemma lt_minus_anti :
  forall i j n,
    i < j ->
    j <= n ->
    n - i > n - j.
  intros; omega.
Qed.

Lemma minus_gt_anti :
  forall i j n,
    j <= n ->
    n - i > n - j ->
    i < j.
  intros; omega.
Qed.

Lemma gt_minus_anti :
  forall i j n,
    i > j ->
    i <= n ->
    n - i < n - j.
  intros; omega.
Qed.

Lemma minus_lt_anti :
  forall i j n,
    i <= n ->
    n - i < n - j ->
    i > j.
  intros; omega.
Qed.


Lemma le_minus_anti :
  forall i j n,
    i <= j ->
    n - i >= n - j.
  intros; omega.
Qed.


Lemma minus_ge_anti :
  forall i j n,
    i <= n ->
    n - i >= n - j ->
    i <= j.
  intros; omega.
Qed.

Lemma ge_minus_anti :
  forall i j n,
    i >= j ->
    n - i <= n - j.
  intros; omega.
Qed.

Lemma minus_le_anti :
  forall i j n,
    j <= n ->
    n - i <= n - j ->
    i >= j.
  intros; omega.
Qed.

Lemma S_minus_S :
  forall (m n:nat),
    S m - S n = m - n.
  intro m.
  induction m as [|m ih].
  intros; omega.
  intro n.
  specialize (ih n).
  omega.
Qed.


Lemma one_minus :
  forall m,
    0 < m ->
    1 - m = 0.
  intro m.
  destruct m as [|m].
  intro h1.
  apply lt_irrefl in h1; contradiction.
  intro.
  rewrite S_minus_S.
  apply zero_minus.
Qed.
  
Lemma S_minus_shift :
  forall (m n:nat),
    n <= m ->
    (S m - n) = S (m - n).
  intros m n. revert m. 
  induction n as [|n ih].
  intro; omega.
  intros m h1.
  destruct m as [|m].
  omega.
  apply le_S_n in h1.
  specialize (ih _ h1).
  do 2 rewrite S_minus_S; auto.
Qed.

Lemma le_S_minus_shift :
  forall (m n:nat),
    n <= m ->
    S (m - n) <= S m - n.
  intros; omega.
Qed.




Lemma S_minus_pred : 
  forall m n,
    S n <= m ->
    m - S n =
    pred (m - n).
  intros; omega.
Qed.

  
Lemma le_minus_iff :
  forall (m n:nat),
    m <= n <-> m - n = 0.
  intuition.
Qed.

Lemma minus_plus_le_cancel :
  forall m n p,
    m <= n ->
    n = m + p ->
    n - m = p.
  intros; omega.
Qed.

Lemma minus_plus_lt_cancel :
  forall m n p,
    m < n ->
    n = m + p ->
    n - m = p.
  intros; omega.
Qed.

Lemma lt_iff : 
  forall p q,
    p < q <->
    q - p > 0.
  intuition.
Qed.

Lemma minus_lt_compat_r :
  forall i j n,
    n < j ->
    i < j ->
    i - n < j - n.
  intros; omega.
Qed.

Lemma plus_minus_assoc : 
  forall (m n p:nat),
    p <= n ->
    m + (n - p) = m + n - p.
intros m n p.
induction m as [|m h1]; destruct n as [|n]; destruct p as [|p];
simpl; auto.
intro h2.
omega.
intro h2.
specialize (h1 h2).
assert (h3:p <= n). omega. clear h2.
rewrite S_minus_S in h1.
rewrite h1.
rewrite minus_Sn_m.
rewrite S_minus_S. 
reflexivity.
omega.
Qed.


Lemma minus_plus_assoc : 
  forall (m n p:nat),
    n + p <= m ->
    m - (n + p) = m - n - p.
  intro m.
  induction m as [|m ih].
  intros; do 2 rewrite zero_minus; auto.
  intros n p h1.
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1 | h1]. 
  rewrite S_minus_shift; auto with arith.
  specialize (ih n p).
  hfirst ih.
  auto with arith.
  specialize (ih hf).
  rewrite ih.
  rewrite S_minus_shift, <- (S_minus_shift (m - n) p); auto; omega.
  omega.
Qed.


Lemma minus_minus_assoc : 
  forall (m n p:nat),
    n - p <= m ->
    p <= n ->
    n <= m ->
    m - (n - p) = m - n + p.
  intros; omega.
Qed.

Lemma plus_minus_same_r :
  forall m n:nat,
    m + n - m =
    n.
  intros; omega.
Qed.


Lemma plus_minus_same_l :
  forall m n:nat,
    n + m - m =
    n.
  intros; omega.
Qed.


Lemma plus_minus_same' :
  forall m n:nat,
    m <= n ->
    m + (n - m) =
    n.
  intros; omega.
Qed.

Lemma minus_plus_same : 
  forall m n,
    m <= n ->
    n - m + m = n.
  intros; omega.
Qed.


Lemma minus_inj :
  forall (m n p:nat),
    n <= m ->
    p <= m ->
    m - n = m - p ->
    n = p.
  intro m.
  induction m as [|m ih].
  intros n p h1 h2.
  apply le0 in h1.
  apply le0 in h2; subst; auto.
  intros n p h1 h2 h3.
  apply le_lt_eq_dec in h1.
  apply le_lt_eq_dec in h2.
  destruct h1 as [h1|], h2 as [h2|]; subst.
  red in h1, h2.
  apply le_S_n in h1.
  apply le_S_n in h2.
  specialize (ih _ _ h1 h2).
  do 2 rewrite S_minus_shift in h3; auto.
  omega.
  omega.
  reflexivity.
Qed.

Lemma le_minus_inj :
  forall (m n p:nat),
    p <= m ->
    p <= n ->
    m - p <= n - p ->
    m <= n.
  intros; omega.
Qed.


Lemma minus_inj_neq :
  forall (m n p:nat),
    m - n <> m - p ->
    n <> p.
  intros m n p h1 h2.
  subst.
  contradict h1; auto.
Qed.

Lemma le_minus_0 :
  forall m n,
    m <= n ->
    m - n = 0.
  intros; omega.
Qed.


Lemma minus_same_diff :
  forall i j n,
    i <= n ->
    j <= n ->
    n - i - (n - j) = j - i.
  intros; omega.
Qed.


Lemma le_plus_minus_same_iff : 
    forall i j k,
      i + j - k <= i <-> 
      j <= k.
  intros; omega.
Qed.

Lemma plus_minus_lt :
  forall i j k,
    j < k ->
    i + j - k =
    i - (k - j).
  intros; omega.
Qed.

Lemma plus_minus_le :
  forall i j k,
    j <= k ->
    i + j - k =
    i - (k - j).
  intros; omega.
Qed.

Lemma minus_minus_lt :
  forall i j k,
    j < k ->
    i - k - j =
    (i - j) - k.
  intros; omega.
Qed.

Lemma minus_minus_le :
  forall i j k,
    j <= k ->
    i - k - j =
    (i - j) - k.
  intros; omega.
Qed.

Lemma minus_minus_lt' :
  forall i j k,
    j < i ->
    i - k - j =
    (i - j) - k.
  intros; omega.
Qed.

Lemma minus_minus_le' :
  forall i j k,
    j <= i ->
    i - k - j =
    (i - j) - k.
  intros; omega.
Qed.

                 
Lemma plus_minus_minus_lt :
  forall i j k m,
    m < i ->
    i + j - k - m =
    i - m + j - k.
  intros; omega.
Qed.

Lemma plus_minus_minus_le :
  forall i j k m,
    m <= i ->
    i + j - k - m =
    i - m + j - k.
  intros; omega.
Qed.

Lemma plus_minus_minus_same :
  forall i j k,
    i + j - i - k =
    j - k.
  intros; omega.
Qed.

Lemma plus_minus_minus_same':
  forall i j k,
    i + j - k - i =
    j - k.
  intros; omega.
Qed.


Lemma plus_minus_minus_same'':
  forall i j k,
    i + j - k - j =
    i - k.
  intros; omega.
Qed.



Lemma plus_minus_eq_iff :
  forall i j k,
    i = i + j - k <->
    (j < k /\ i = 0) \/ j = k.
  intros; omega.
Qed.

Lemma plus_minus_plus_same_lt :
  forall i j k,
    k < j ->
      i + j - (i + k) =
      j - k.
  intros; omega.
Qed.

Lemma plus_minus_plus_same_le :
  forall i j k,
    k <= j ->
    i + j - (i + k) =
      j - k.
  intros; omega.
Qed.

Lemma plus_minus_plus_same_lt' :
    forall i j k,
      k < i ->
      i + j - (k + j) = i - k.
    intros; omega.
Qed.

Lemma plus_minus_plus_same_le' :
    forall i j k,
      k <=i ->
      i + j - (k + j) = i - k.
    intros; omega.
Qed.


Lemma minus_minus_plus_minus_same_le :
  forall i j k,
    j <= k ->
    k <= i + j ->
    i - (i + j - k) =
    k - j.
  intros; omega.
Qed.

Lemma plus_minus_anti :
  forall i j k,
    j <= i ->
    i - j + k + j - i =
    k.
    intros; omega.
Qed.



Lemma minus_plus_le :
  forall i j k,
    i - j + k <= i + k.
  intros; omega.
Qed.

Lemma plus_minus_minus_same_le :
  forall i j k,
    k <= i ->
    i - k <= j ->
    i + (j - (i - k)) =
    j + k.
  intros; omega.
Qed.

Lemma plus_plus_plus_minus_same :
    forall i j k,
      i + j + k - i =
      j + k.
    intros; omega.
Qed.


Lemma plus_minus_plus_diff_lt :
  forall i j k m,
    k < i ->
    m < j ->
    i + j - (k + m) =
    (i - k) + (j - m).
  intros; omega.
Qed.


Lemma plus_minus_plus_diff_le :
  forall i j k m,
    k <= i ->
    m <= j ->
    i + j - (k + m) =
    (i - k) + (j - m).
  intros; omega.
Qed.

Lemma plus_minus_plus_diff_lt_le :
  forall i j k m,
    k < i ->
    m <= j ->
    i + j - (k + m) =
    (i - k) + (j - m).
  intros; omega.
Qed.


Lemma plus_minus_plus_diff_le_lt :
  forall i j k m,
    k <= i ->
    m < j ->
    i + j - (k + m) =
    (i - k) + (j - m).
  intros; omega.
Qed.


Lemma plus_minus_plus_diff_lt' :
  forall i j k m,
    k < i ->
    j < m ->
    i + j - (k + m) =
    (i - k) - (m - j).
  intros; omega.
Qed.


Lemma plus_minus_plus_diff_le' :
  forall i j k m,
    k <= i ->
    j <= m ->
    i + j - (k + m) =
    (i - k) - (m - j).
  intros; omega.
Qed.

Lemma plus_minus_plus_diff_lt_le' :
  forall i j k m,
    k < i ->
    j <= m ->
    i + j - (k + m) =
    (i - k) - (m - j).
  intros; omega.
Qed.


Lemma plus_minus_plus_diff_le_lt' :
  forall i j k m,
    k <= i ->
    j < m ->
    i + j - (k + m) =
    (i - k) - (m - j).
  intros; omega.
Qed.

Lemma minus_lt_compat :
  forall i j k m,
    i < j ->
    k < m ->
    m < i ->
    i - m < j - k.
  intros; omega.
Qed.

Lemma minus_le_compat :
  forall i j k m,
    i <= j ->
    k <= m ->
    i - m <= j - k.
  intros; omega.
Qed.

Lemma minus_le_lt_compat :

  forall i j k m,
    i <= j ->
    k < m ->
    m <= i ->
    i - m < j - k.
  intros; omega.
Qed.

Lemma minus_lt_le_compat :
  forall i j k m,
    i < j ->
    k <= m ->
    m <= i ->
    i - m < j - k.
  intros; omega.
Qed.
 
Lemma minus_minus_minus_diff_le :
  forall i j k m, 
    k <= i ->
    j <= m ->
    j <= i ->
    m <= k ->
    k - m <= i - j ->
    i - j - (k - m) =
    i - k + (m - j).
  intros; omega.
Qed.

(*I'm spelling out a lot of omegas for efficiency!*) 
Lemma minus_minus_eq_iff :
  forall m n p,
    m - n = m - p <->
    (m <= n /\ m <= p  \/
     ((m > n /\ p <= m  \/ 
      m > p /\ n <= m) /\
      n = p)). 
  intros m n p. revert n m.     
  induction p as [|p ih]; simpl; intros m n. 
  rewrite <- minus_n_O, eq_sym_iff, minus_eq_iff.
  split; intro h1.
  destruct h1; subst.
  destruct (zerop m) as [|hn]; left; auto with arith.
  destruct (zerop n) as [|hn]; subst.
  left; auto with arith.
  right; auto with arith.
  destruct h1 as [h1|h1].
  destruct h1 as [? h1].
  apply le0 in h1; left; auto. 
  destruct h1; right; auto.
  destruct m as [|m], n as [|n];
    split; intro h1; auto with arith.
  rewrite <- minus_n_O, minus_eq_iff in h1.
  destruct h1; discriminate.
  destruct h1 as [h1|h1].
  destruct h1 as [h1 h2 h1].
  apply le_Sn_0 in h1; contradiction.
  destruct h1; discriminate.
  do 2 rewrite S_minus_S in h1.
  rewrite ih in h1. 
  intuition.
  do 2 rewrite S_minus_S; rewrite ih; intuition.
Qed.
  
(*Am spelling out a slow [intros; omega]!*)
Lemma minus_minus_minus_diff_le' :
  forall i j k m, 
    k <= i ->
    m <= j ->
    j <= i ->
    m <= k ->
    k - m <= i - j ->
    i - j - (k - m) =
    i - k - (j - m).
  intros i j k m h1 h2 h3 h4 h5.   
  pose proof (minus_minus_eq_iff i (j + (k - m)) (k + (j - m))) as h6.
  do 2 rewrite minus_plus_assoc in h6.
  rewrite h6. 
  destruct (le_lt_dec i (j +(k - m))) as [h7 | h7]. 
  left.
  split; auto. 
  rewrite plus_minus_assoc; auto. 
  rewrite plus_minus_assoc in h7; auto.
  rewrite plus_comm; auto.
  right.
  split.
  left. 
  split; auto.   
  rewrite plus_minus_assoc; auto. 
  rewrite plus_minus_assoc in h7; auto.
  rewrite plus_comm; auto with arith.
  rewrite plus_minus_assoc; auto. 
  rewrite (plus_minus_assoc k j m); auto.
  rewrite plus_comm; auto.
  
  pose proof (plus_le_compat_l _ _ j h5) as h7.
  rewrite plus_minus_assoc in h7; auto.
  rewrite plus_comm in h7.   
  rewrite <- plus_minus_assoc in h7; auto.  
  rewrite (plus_minus_assoc j i j) in h7; auto.
  rewrite <- (plus_comm i j) in h7.
  rewrite <- (plus_minus_assoc i j j) in h7; auto.
  rewrite minus_same, <- plus_n_O in h7; auto. 
  pose proof (plus_le_compat_l _ _ j h5) as h7. 
  rewrite plus_minus_assoc in h7; auto.
  rewrite <- plus_minus_assoc in h7; auto.  
  rewrite (plus_minus_assoc j i j) in h7; auto.
  rewrite <- (plus_comm i j) in h7.
  rewrite <- (plus_minus_assoc i j j) in h7; auto.
  rewrite minus_same, <- plus_n_O in h7; auto.
  pose proof (plus_le_compat_l _ _ j h5) as h7. 
  rewrite plus_minus_assoc in h7; auto.
  rewrite <- plus_minus_assoc in h7; auto.  
  rewrite (plus_minus_assoc j i j) in h7; auto.
  rewrite <- (plus_comm i j) in h7.
  rewrite <- (plus_minus_assoc i j j) in h7; auto.
  rewrite minus_same, <- plus_n_O in h7; auto. 
Qed.


Lemma minus_minus_plus_same_lt :
  forall i j k,
    k < j ->
    j - k < i ->
    i - (j - k) + j =
    i + k.
  intros; omega.
Qed.


Lemma minus_minus_plus_same_le :
  forall i j k,
    k <= j ->
    j - k <= i ->
    i - (j - k) + j =
    i + k.
  intros; omega.
Qed.


Lemma minus_plus_plus_same_lt :
  forall i j k,
    j < i ->
    i - j + k + j = i + k.
  intros; omega.
Qed.


Lemma minus_plus_plus_same_le :
  forall i j k,
    j <= i ->
    i - j + k + j = i + k.
  intros; omega.
Qed.


End Minus.

Section S.

Lemma S_minus :
  forall n,
    S n - n = 1.
  intros; omega.
Qed.

Lemma S_inj : forall (m n:nat), S m = S n -> m = n.
intros ? ? h1; injection h1; auto.
Qed.

Lemma S_inj_neq :
  forall (m n:nat), S m <> S n -> m <> n.
  intros m n h1 h2.
  subst.
  contradict h1; auto.
Qed.
  

Lemma le_times_S :
  forall i j,
    i <= i * S j.
  intros; rewrite <- mult_1_r at 1;
  apply mult_le_compat; auto with arith.
Qed.

Lemma S_eq_plus :
  forall m n,
    S m = S m + n ->
    m = m + n.
  intros; omega.
Qed.


End S.


Section Dec.
Lemma nat_eq_dec : type_eq_dec nat.
  red; intros m n.
  destruct (lt_eq_lt_dec m n) as [h1 | h1].
  destruct h1 as [h1 | h1].
  right; intro; subst.
  apply lt_irrefl in h1; auto.
  subst.
  left; auto.
  right; intro; subst.
  apply lt_irrefl in h1; auto.
Qed.


Lemma neq_nat_dec :
  forall (m n:nat),
    m <> n ->
    {m < n} + {n < m}.
  intros m.
  induction m as [|m ih].
  intros; left; auto with arith.
  intros n h1.
  destruct n as [|n].
  right; auto with arith.
  apply S_inj_neq in h1.
  specialize (ih _ h1).
  destruct ih; [left; auto with arith | right; auto with arith].
Qed.

Lemma lt_dec :
  forall m n,
    m < n ->
    {m < pred n} + {m = pred n}.
  intros m n h1.
  destruct n as [|n].
  apply lt_n_O in h1; contradiction.
  simpl.
  red in h1.
  apply le_S_n in h1.
  apply le_lt_eq_dec in h1; auto.
Qed.

Lemma le_dec' :
  forall m n,
    m <= n ->
    {m <= pred n} + {m = n}.
  intros m n. revert m.
  induction n as [|n ih].
  intros m h1; apply le0 in h1; subst.
  right; auto.
  simpl; intros m h1. 
  destruct m as [|m].
  left; auto with arith.
  apply le_S_n in h1.
  destruct (zerop n) as [|h3]; subst.
  right; auto with arith.
  specialize (ih _ h1).
  destruct ih as [h2|]; subst.
  left; omega.
  right; auto.
Qed.
  
  

Lemma pred0_dec : forall n:nat, pred n = O -> {n = 0} + {n = 1}.
intro n.
induction n; simpl; auto; omega.
Qed.

Lemma lt_le_dec :
  forall m n:nat,
    {m < n} + {n <= m}.
  intros m n.
  destruct (lt_eq_lt_dec m n) as [h1 | h1]. destruct h1 as [h1 | h1].
  left; auto. subst. right; auto with arith. right; auto with arith.
Qed.

Lemma gt_ge_dec :
  forall m n:nat,
    {m > n} + {n >= m}.
  intros; apply lt_le_dec.
Qed.



Lemma lt_S_dec :
  forall m n:nat,
    m < S n ->
    {m < n} + {m = n}.
  intros m n h1.
  destruct (lt_le_dec m n)  as [h2 | h2].
  left; auto.
  right. omega.
Qed.

Lemma lt_ex_dec :
  forall (P:nat->Prop) (n:nat),
    pred_dec P ->
    {exists m:nat, m < n /\ P m} + {~exists m:nat, m < n /\ P m}.
  intros P n.
  induction n as [|n ih]; simpl.
  intros; right; intro h1.
  destruct h1 as [? [? ?]]; omega.
  intro h1. 
  specialize (ih h1).
  destruct ih as [h2 | h2].
  left.
  destruct h2 as [m [h2 h4]].
  exists m.
  split; auto with arith.
  specialize (h1 n).
  destruct h1 as [h1 | h1].
  left.
  exists n.
  split; auto with arith.
  right.
  intro h3.
  destruct h3 as [m [h3 h4]].
  destruct (nat_eq_dec m n) as [|h5]; subst.
  contradiction.
  contradict h2.
  exists m.
  split; try omega; auto with arith.
Qed.
  
Lemma lt_ex_dep_nat_dec :
  forall {n:nat} (P:pred_dep_nat n),
    pred_dep_nat_dec P ->
    {exists (m:nat) (pfm:m < n), P m pfm} +
    {~exists (m:nat) (pfm:m < n), P m pfm}.
  intro n.
  induction n as [|n ih]; simpl.
  intros; right; intro h1.
  destruct h1 as [? [? ?]]; omega.
  intros P h1.
  specialize (ih (pred_dep_nat_pred P) (pred_dep_nat_dec_pred _ h1)).
  destruct ih as [h2 | h2].
  left.
  destruct h2 as [m [h2 h3]].
  exists m, (lt_trans _ _ _ h2 (lt_n_Sn _)).
  red in h3; auto.
  destruct (h1 n (lt_n_Sn _)) as [h3 | h3].
  left.
  exists n, (lt_n_Sn _); auto.
  right.
  intro h4.
  destruct h4 as [m [h4 h5]].
  red in h4.
  pose proof (le_S_n _ _ h4) as h6.
  destruct (le_lt_eq_dec _ _ h6) as [h7 |]; subst.
  contradict h2.
  exists m, h7; red. 
  erewrite f_equal_dep.
  apply h5.
  reflexivity.
  contradict h3.
  erewrite f_equal_dep.
  apply h5.
  reflexivity.
Qed.

End Dec.


Section Bound.
  
(*This sets a weak bound [b] for the identity function on nat.
So, [(nat_bound_weak b):nat->nat, i => i], for [i <= b].  
So [nat_bound_weak 3 1 = 1], [nat_bound_weak 3 2 = 2], 
[nat_bound 3 3 = 3], [nat_bound 3 8 = 3].*)

Fixpoint nat_bound_weak b : nat -> nat :=
  match b with
    | 0 => fun _ => 0
    | S b' =>
      fun i =>
        match i with
          | 0 => 0
          | S i' => S (nat_bound_weak b' i')
        end
  end.


(*This sets a strong bound [b] for the identity function on nat.
So, [(nat_bound_strong b):nat->nat, i => i], for [i < b].  
So [nat_bound 3 1 = 1], [nat_bound 3 2 = 2], 
[nat_bound 3 3 = 2], [nat_bound 3 8 = 2].*)

Definition nat_bound b : nat -> nat :=
  fun i => nat_bound_weak (pred b) i.
 
Lemma nat_bound_weak_eq_same_iff :
  forall n i,
    nat_bound_weak n i = i <->
    i <= n.
  intro n; induction n as [|n ih]; simpl.
  intuition. 
  intro i; destruct i as [|i]; simpl.
  intuition.
  split; intro h1.
  apply S_inj in h1;
    apply le_n_S.
  rewrite ih in h1; auto.
  apply le_S_n in h1; f_equal.
  rewrite <- ih in h1; auto.
Qed.


Lemma nat_bound_eq_same_iff :
  forall n i,
    nat_bound n i = i <->
    i < n \/
    (n = 0 /\ i = 0).
  unfold nat_bound; intros; rewrite nat_bound_weak_eq_same_iff.
  omega.
Qed.
 
Lemma nat_bound_weak_eq_iff :
  forall n i m,
    nat_bound_weak n i = m <->
    (i <= n /\ i = m \/ i > n /\ n = m).
  intro n; induction n as [|n ih]; simpl.
  intros i m; split; intro h1; subst.
  destruct i as [|i].
  left; auto.
  right; auto with arith.
  destruct h1 as [h1|h1].
  destruct h1; subst.
  apply le0 in H; auto.
  tauto.
  intro i.
  destruct i as [|i].
  intro m.
  split; intro; subst.
  left; auto with arith.
  destruct H; omega.
  intro m.
  destruct m as [|m].
  intuition.
  split; intro h1.
  apply S_inj in h1; subst.
  specialize (ih i n).
  destruct (le_lt_dec i n) as [h1|h1]. 
  left.
  split; auto with arith.
  f_equal.
  rewrite <- nat_bound_weak_eq_same_iff in h1; auto.
  right.
  split; auto with arith.
  f_equal.
  symmetry.
  rewrite ih.
  tauto.
  f_equal.
  destruct h1 as [h1|h1].
  destruct h1 as [h1 h2].
  apply S_inj in h2; subst.
  apply le_S_n in h1.
  rewrite <- nat_bound_weak_eq_same_iff in h1; auto.
  destruct h1 as [h1 h2].
  apply S_inj in h2; subst.
  rewrite ih.
  right; auto with arith.
Qed.


Lemma nat_bound_eq_iff :
  forall n i m,
    nat_bound n i = m <->
    (i < n /\ i = m \/ i >= n /\ pred n = m).
  unfold nat_bound; intros n i m; rewrite nat_bound_weak_eq_iff;
  omega.
Qed.

Lemma le_bound :
  forall n i,
    nat_bound_weak n i <= n.
  unfold nat_bound; intro n; induction n as [|n ih]; simpl;
  auto with arith;
  intro i; destruct i; auto with arith.
Qed.


Lemma lt_bound :
  forall n i,
    nat_bound (S n) i < S n.
  unfold nat_bound; intros; red; apply le_n_S; simpl; apply le_bound.
Qed.


End Bound.

Section Inj.

Lemma pred_inj0: forall (m n:nat),
                   0 < m -> 0 < n -> pred m = pred n -> m = n.
  intros m n; destruct m, n; omega.
Qed.

Lemma pred_inj1: forall (m n:nat),
                   m <> 1 -> n <> 1 -> pred m = pred n -> m = n.
  intros m n; destruct m, n; omega.
Qed.



Lemma pred_inj': forall (m n:nat), pred m > 0 -> pred m = pred n -> m = n.
intros m n h0.
induction m as [|m h1]; destruct n as [|n].
auto.
simpl in h0. omega.
simpl.
intro h2.
subst.
simpl in h0. omega.
intro h2.
simpl in h2.
f_equal. auto.
Qed.

Lemma pred_inj_lt :
  forall (m n:nat),
    pred m < pred n ->
    m < n.
  intros m n h1.
  destruct m as [|m]; destruct n as [|n]; simpl; auto.
  simpl in h1.
  apply lt_n_0 in h1. contradiction.
  simpl in h1.
  apply lt_n_S; auto.
Qed.

Lemma lt_S_pred :
  forall m n,
    S m < n ->
    m < pred n.
  intros; omega.
Qed.

Lemma lt_S_pred' :
  forall m n,
    0 < m ->
    m < S n ->
    pred m < n.
  intros; omega.
Qed.


Lemma lt_S_neq :
  forall m n,
    m < S n ->
    m <> n ->
    m < n.
  intros; omega.
Qed.


Lemma pred_pred_lt :
  forall n:nat,
    0 < n ->
    pred (pred n) < n.
  intros; omega.
Qed.

Lemma pred_minus_pred_pred_lt :
  forall n,
    n < 2 ->
    pred n - (pred (pred n)) = 0.
  intro; omega.
Qed.

Lemma pred_minus_pred_pred_ge :
  forall n,
    n >= 2 ->
    pred n - (pred (pred n)) = 1.
  intro; omega.
Qed.


  

Lemma pred_minus :
  forall n m,
    m <= n ->
    pred (n - m) =
    pred n - m. 
  intros; omega.
Qed.

Lemma plus_inj_l :
  forall m n p:nat,
    m + n = m + p ->
    n = p.
  intros; omega.
Qed.


Lemma plus_inj_r :
  forall m n p:nat,
    m + p = n + p ->
    m = n.
  intros; omega.
Qed.


End Inj.

  
Lemma le_prop_seg_one_t : 
  forall (m n:nat),
    m <= n ->
    forall a:nat,
      1 <= a <= m ->
      1 <= a <= n.
intros; omega.
Qed.

 
Lemma lt_prop_seg_one_t : 
  forall (m n:nat),
    m < n ->
    forall a:nat,
      1 <= a <= m ->
      1 <= a <= n.
intros; omega.
Qed.


Lemma lt_le_prop_seg_one_t : 
  forall (m n:nat),
    m < n ->
    forall a:nat,
      m < a <= n ->
      1 <= a <= n.
intros; omega.
Qed.


Lemma lt_le_prop_seg_one_t' : 
  forall (m n:nat),
    m <= n ->
    forall a:nat,
      m < a <= n ->
      1 <= a <= n.
intros; omega.
Qed.



Lemma inj_minus :
  forall (m n p:nat),
    p <= m ->
    p <= n ->
    m - p = n - p ->
    m = n.
  intros m n p.
  revert m n.
  induction p as [|p ih]; simpl.
  intros ? ? ? ? h3.
  do 2 rewrite <- minus_n_O in h3; auto.
  intros m n h1 h2 h3.
  do 2 rewrite S_minus_pred in h3; auto.
  apply pred_inj0 in h3; omega.
Qed.

(*This is a bit more handy than [NPeano]'s existential version [div], 
  in that you avoid an extra variable and existential. 
  Also, not too many properties
  were derived in that version, so I lose little.*)
Inductive dv' (a:nat) : nat -> Prop :=
  dv_intro : forall n, dv' a (a * n).

Infix "dv" := dv' (at level 40, no associativity).

Section Times.

  
Lemma mult_S_lt_l :
  forall i j k,
    S i * j < S i * k ->
    j < k.
  intros i j k h1.
  destruct (lt_eq_lt_dec j k) as [[h2 | h2] | h2]; auto.
  subst.
  apply lt_irrefl in h1; contradiction.
  pose proof (mult_lt_compat_l _ _ (S i) h2 (lt_0_Sn _)) as h3.
  pose proof (lt_trans _ _ _ h1 h3) as h4.
  apply lt_irrefl in h4.
  contradiction.
Qed.

  
Lemma mult_S_lt_r :
  forall i j k,
    j * S i < k * S i ->
    j < k.
  intros i j k h1.
  destruct (lt_eq_lt_dec j k) as [[h2 | h2] | h2]; auto.
  subst.
  apply lt_irrefl in h1; contradiction.
  pose proof (mult_lt_compat_r _ _ (S i) h2 (lt_0_Sn _)) as h3.
  pose proof (lt_trans _ _ _ h1 h3) as h4.
  apply lt_irrefl in h4.
  contradiction.
Qed.



Lemma lt_mult_iff :
  forall i j,
    i * j < i <->
    (j = 0 /\ 0 < i).
  intros i j; split; intro h1. 
  destruct j as [|j].
  rewrite mult_0_r in h1; auto.
  pose proof (le_times_S i j) as h2.
  pose proof (le_lt_trans _ _ _ h2 h1) as h3.
  apply lt_irrefl in h3; contradiction.
  destruct h1 as [? h1]; subst.
  rewrite mult_0_r; auto.
Qed.


Lemma S_times_l :
  forall x y,
    S x * y =
    x * y + y.
  intros x y.
  rewrite <- S_compat.
  rewrite mult_plus_distr_r.
  rewrite mult_1_l; auto.
Qed.

Lemma S_times_l' :
  forall x y,
    S x * y =
    y + x * y.
  intros; rewrite plus_comm; apply S_times_l.
Qed.

Lemma S_times_r :
  forall x y,
    y * S x =
    y * x + y.
  intros x y.
  rewrite <- S_compat.
  rewrite mult_plus_distr_l.
  rewrite mult_1_r; auto.
Qed.

Lemma S_times_r' :
  forall x y,
    y * S x =
    y + y * x.
  intros; rewrite plus_comm; apply S_times_r.
Qed.



Lemma S_times_S :
  forall x y,
    S x * S y =
    S (x * y + x + y).
  intros x y. 
  rewrite <- S_compat.
  rewrite <- (S_compat y).
  rewrite mult_plus_distr_r.
  rewrite mult_1_l.
  rewrite mult_plus_distr_l.
  rewrite S_compat, mult_1_r.
  rewrite <- plus_S_shift_r; auto.
Qed.


Lemma minus_dist_r : 
  forall i j m,
    i * m - j * m =
    (i - j) * m.
  intros i j m. revert i j.
  induction m as [|m ih].
  intros.
  do 3 rewrite mult_0_r; auto with arith.
  intros i j.
  destruct (lt_eq_lt_dec i j) as [[h1 | h1] | h1].
  pose proof (mult_S_lt_compat_l m i j h1) as h2.
  rewrite (mult_comm _ i), (mult_comm _ j) in h2.
  apply lt_le_weak in h2.
  rewrite le_minus_iff in h2.
  rewrite h2.
  apply lt_le_weak in h1.
  rewrite le_minus_iff in h1.
  rewrite h1.
  rewrite mult_0_l; auto.
  subst.
  do 2 rewrite <- minus_n_n.
  rewrite mult_0_l; auto.
  pose proof (mult_S_lt_compat_l m j i h1) as h2.
  rewrite mult_comm, (mult_comm j), (mult_comm (i - j)).
  do 3 rewrite S_times_l.
  specialize (ih i j).
  rewrite mult_comm, (mult_comm j), (mult_comm (i - j)) in ih.
  rewrite <- ih.
  rewrite plus_minus_assoc; auto with arith.
  destruct m as [|m].
  do 2 rewrite mult_0_l.
  do 2 rewrite plus_0_l; auto with arith. 
  pose proof (plus_minus_assoc i (S m*i) (S m*j)) as h3.
  hfirst h3. 
  apply mult_le_compat_l; auto with arith.
  specialize (h3 hf).
  omega.
Qed.

Lemma minus_dist_l : 
  forall i j m,
    m * i - m * j =
    m * (i - j).
  intros i j m.
  rewrite (mult_comm m i), (mult_comm m j), (mult_comm m (i - j)).
  apply minus_dist_r.
Qed.


Lemma times_S_eq_0_iff :
  forall i j,
    i * S j = 0 <->
    i = 0.
  intro i.
  destruct i as [|i].
  intro. rewrite mult_0_l; tauto.
  intro j.
  rewrite S_times_S.
  split; intro h1; discriminate h1.
Qed.

Lemma S_times_eq_iff :
  forall i j k,
    S i * j = S i * k <->
    j = k.
  intros i j k; split; intro h1.
  destruct (lt_eq_lt_dec j k) as [[h2 | h2] | h2].
  pose proof (mult_S_lt_compat_l i j k h2) as h3.
  rewrite h1 in h3.
  apply lt_irrefl in h3; contradiction.
  assumption.
  pose proof (mult_S_lt_compat_l i k j h2) as h3.
  rewrite <- h1 in h3.
  apply lt_irrefl in h3; contradiction.
  subst; auto.
Qed.


Lemma lt_times_S_plus_S :
  forall x y z,
    x < x * S y + S z.
  intro x.
  induction x as [|x ih]; simpl; auto with arith.
  intros y z.
  apply lt_n_S.
  rewrite (plus_comm y (x * S y)).
  rewrite <- plus_assoc.
  rewrite plus_S_shift_r.
  apply ih.
Qed.

Lemma lt_plus_S :
  forall x y,
    x < x + S y.
  intros; omega.
Qed.


Lemma times_le_iff : 
  forall i j,
    i * j <= i <->
    (i = 0 \/ j = 0 \/ j = 1).
  intros i j; split; intro h1.
  destruct i as [|i].
  left; auto.
  right.
  destruct j as [|j].
  left; auto.
  right.
  rewrite S_times_S in h1.
  apply le_S_n in h1.
  f_equal.
  destruct j as [|j]; auto.
  rewrite <- plus_assoc in h1.
  rewrite plus_S_shift_r in h1.
  pose proof (lt_times_S_plus_S i j (i + j)).
  omega.
  destruct h1 as [h1 | h1]; subst.
  rewrite mult_0_l; auto with arith.
  destruct h1; subst.
  rewrite mult_0_r; auto with arith.
  rewrite mult_1_r; auto with arith.
Qed.


Lemma eq_times_iff :
  forall x y,
    x = x * y <->
    x = 0 \/
    (x = 0 /\ y = 0) \/
    y = 1.
  intros x y; split; intro h1.
  destruct x  as [|x]. 
  left; auto.
  right.
  destruct y as [|y].
  left; auto.
  rewrite mult_0_r in h1.
  discriminate.
  right. 
  f_equal. 
  rewrite S_times_S in h1.
  apply S_inj in h1.
  destruct y as [|y]; auto.
  rewrite <- plus_assoc in h1.
  rewrite (plus_S_shift_r x y) in h1.
  pose proof (lt_times_S_plus_S x y (x + y)).
  omega. 
  destruct h1 as [h1 | [[h0 h1] | h1]]; subst.
  rewrite mult_0_l; auto.
  rewrite mult_0_r; auto.
  rewrite mult_1_r; auto.
Qed.

Lemma eq_times_plus_iff :
  forall x y z,
    x = x * y + z <->
    (x = 0 /\ z = 0) \/
    (y = 1 /\ z = 0) \/
    (y = 0 /\ x = z).
  intros x y z; split; intro h1.
  destruct x as [|x].
  left.
  rewrite mult_0_l in h1.
  rewrite plus_0_l in h1; auto.
  right. 
  destruct y as [|y].
  right. 
  rewrite mult_0_r in h1.
  rewrite plus_0_l in h1; auto.
  destruct z as [|z]. 
  rewrite plus_0_r in h1.
  rewrite eq_times_iff in h1.
  destruct h1 as [h1 | h1].
  discriminate.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  left; auto.
  rewrite S_times_S in h1.
  rewrite  plus_S_shift_l in h1.
  apply S_inj in h1.
  rewrite <- (mult_1_r x) in h1 at 3.
  rewrite <- mult_plus_distr_l in h1.
  rewrite S_compat in h1.
  rewrite <- plus_assoc in h1.
  rewrite plus_S_shift_r in h1.
  pose proof (lt_times_S_plus_S x y (y + z)).
  omega.
  destruct h1 as [[h0 h1] | [[h0 h1] | h1]]; subst.
  rewrite mult_0_l; auto with arith.
  rewrite mult_1_r; auto with arith.
  destruct h1; subst.
  rewrite mult_0_r; auto.
Qed.

Lemma one_eq_S_times_plus_iff :
  forall x y z,
    1 = (S x) * y + z <->
    (x = 0 /\ ((y = 0 /\ z = 1) \/ (y = 1 /\ z = 0))) \/
    (y = 0 /\ z = 1).
  intros x y z; split; intro h1.
  destruct x as [|x].
  left.
  rewrite mult_1_l in h1.
  split; auto.
  destruct y as [|y].
  left.
  split; auto with arith.
  right.
  destruct z as [|z].
  rewrite plus_0_r in h1.
  split; auto.
  omega.
  right.
  destruct y as [|y].
  rewrite mult_0_r in h1.
  rewrite plus_0_l in h1; auto.
  rewrite S_times_S in h1.
  rewrite plus_S_shift_l in h1.
  apply S_inj in h1.
  rewrite <- plus_assoc in h1.
  rewrite <- plus_assoc in h1.
  rewrite plus_S_shift_l in h1.
  rewrite plus_comm in h1.
  rewrite plus_S_shift_l in h1.
  discriminate.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1]; subst.
  destruct h1 as [[? ?] | [? ?]]; subst.
  rewrite mult_0_r; auto with arith.
  rewrite mult_1_r, plus_0_r; auto with arith.
  destruct h1; subst.
  rewrite mult_0_r; auto with arith.
Qed.

Lemma minus_eq_times_iff :
    forall i j k,
      i - j = i * k <->
      (i = 0 \/
       (j = 0 /\
        (i = 0 \/
         (i = 0 /\ k = 0) \/
         k = 1)) \/
       (k = 0 /\ i <= j)).
  intros i j k; split; intro h1.
  destruct i as [|i].
  left; auto.
  right.
  destruct j as [|j].
  rewrite <- minus_n_O in h1.
  rewrite eq_times_iff in h1.
  destruct h1 as [h1 | h1].
  discriminate.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  subst.
  left.
  split; auto.
  right.
  pose proof (le_minus (S i) (S j)) as h2.
  rewrite h1 in h2.
  rewrite times_le_iff in h2.
  destruct h2 as [h2 | h2].
  discriminate.
  destruct h2; subst.
  rewrite mult_0_r in h1.
  rewrite <- le_minus_iff in h1; auto.
  rewrite mult_1_r in h1.
  symmetry in h1.
  rewrite minus_eq_iff in h1.
  destruct h1; discriminate.
  destruct h1 as [h1 | h1]; subst.
  rewrite zero_minus, mult_0_l; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1]; subst.
  destruct h1 as [|h1]; subst.
  rewrite <- minus_n_n, mult_0_l; auto.
  destruct h1 as [h1 | h1].
  destruct h1; subst.
  rewrite <- minus_n_n, mult_0_l; auto.
  subst.
  rewrite minus_n_O, mult_1_r; auto.
  destruct h1 as [? h1]; subst.
  rewrite mult_0_r.
  rewrite <- le_minus_iff; auto.
Qed.

Lemma minus_eq_times_plus_iff :
  forall i j k m,
    i - j = i * k + m <->
      (i = 0 /\ m = 0) \/
      (j = 0 /\
       ((i = 0 /\ m = 0) \/
        (k = 1 /\ m = 0) \/
        (k = 0 /\ i = m))) \/
      (k = 0 /\ m = i - j) \/
      (m = 0 /\
       (i = 0 \/
        (j = 0 /\
         (i = 0 \/
          (i = 0 /\ k = 0) \/
          k = 1)) \/
        (k = 0 /\ i <= j))).
  intros i j k m.
  destruct m as [|m].
  rewrite plus_0_r.
  rewrite minus_eq_times_iff.
  split; intro h1.
  right.
  right.
  right.
  split; auto.
  destruct h1 as [h1 | h1].
  destruct h1; subst.
  left; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1]; subst.
  destruct h1 as [h1 | h1].
  destruct h1; subst.
  left; auto.
  destruct h1 as [h1 | h1].
  destruct h1; subst.
  right.
  left.
  split; auto.
  destruct h1; subst.
  left; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1]; subst.
  right.
  right.
  symmetry in h1.
  rewrite <- le_minus_iff in h1; auto.
  destruct h1 as [h1 h2].
  destruct h2 as [h2 | h2]; subst.
  left; auto.
  destruct h2 as [h2 | h2].
  destruct h2 as [? h2]; subst.
  destruct h2 as [h2 | h2]; subst.
  left; auto.
  destruct h2 as [h2 | h2].
  destruct h2; subst.
  left; auto.
  subst.
  right.
  left.
  split; auto.
  auto. 
  split; intro h1.
  right.
  destruct j as [|j].
  rewrite <- minus_n_O in h1.
  rewrite eq_times_plus_iff in h1.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  destruct h1; subst.
  left.
  auto.
  right.
  destruct k as [|k]. 
  rewrite mult_0_r, plus_0_l in h1.
  left; auto.
  right.
  pose proof (lt_times_S_plus_S i k m) as h2.
  pose proof (le_minus i (S j)) as h3.
  pose proof (le_lt_trans _ _ _ h3 h2) as h4.
  rewrite h1 in h4.
  apply lt_irrefl in h4; contradiction.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1]; subst.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  destruct h1; subst.
  rewrite minus_n_O, mult_0_r, plus_0_l; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1]; subst.
  rewrite mult_0_r, plus_0_l; auto.
  destruct h1; discriminate.
Qed.

Lemma times_eq_times_iff :
  forall i j k,
    i * j =
    i * k <->
    i = 0 \/
    j = k.
  intros i j k.
  destruct i as [|i].
  do 2 rewrite mult_0_l; tauto.
  split; intro h1.
  right.
  rewrite S_times_eq_iff in h1; auto.
  destruct h1 as [h1|]; try discriminate h1; subst; auto.
Qed.

Lemma mult_inj :
  forall i j n,
    0 < n ->
    n * i = n * j ->
    i = j.
  intros i j n h1 h2.
  destruct (lt_eq_lt_dec i j) as [[h3 | h3] | h3].
  pose proof (mult_lt_compat_l i j n h3 h1) as h4.
  rewrite h2 in h4.
  apply lt_irrefl in h4; contradiction.
  assumption.
  pose proof (mult_lt_compat_l j i n h3 h1) as h4.
  rewrite h2 in h4.
  apply lt_irrefl in h4; contradiction.
Qed.
  

Section Divides.

Lemma dv0 :
  forall n,
    n dv 0.
  intro n.
  rewrite <- (mult_0_r n).
  constructor.
Qed.

Lemma one_dv :
  forall n,
    1 dv n.
  intro n.
  rewrite <- (mult_1_l).
  constructor.
Qed.



  
Lemma dv_refl :
  forall a,
    a dv a.
  intro a.
  rewrite <- mult_1_r at 2.
  constructor.
Qed.

  
Lemma dv_trans :
  forall a b c,
    a dv b -> b dv c -> a dv c.
  intros a b c h1 h2.
  destruct h1, h2.
  rewrite <- mult_assoc.
  constructor.
Qed.

Lemma dv_antisym :
  forall a b,
    a dv b -> b dv a -> a = b.
  intros a b h1 h2.  
  destruct h1.
  inversion h2 as [c h3].
  rewrite h3.
  rewrite <- mult_assoc in h3.
  symmetry in h3.
  rewrite eq_times_iff in h3.
  destruct h3 as [h3 | h3]; subst.
  rewrite mult_0_l; auto.
  destruct h3 as [h3 | h3].
  destruct h3; subst.
  rewrite mult_0_l; auto.
  apply mult_is_one in h3.
  destruct h3; subst.
  rewrite mult_1_r; auto.
Qed.

Lemma dv_le :
  forall a b,
    a dv b ->
    b = 0 \/
    a <= b.
  intros a b h1.
  destruct h1.
  destruct n as [|n].
  left.
  rewrite mult_0_r; auto.
  right.
  pose proof (mult_O_le a (S n)) as h1.
  destruct h1 as [h1 | h1].
  discriminate.
  rewrite mult_comm; auto.
Qed.


Lemma times_eq_times_plus_div :
  forall i j k m,
    i * j = i * k + m ->
    i dv m.
  intros i j k m h1.
  destruct i as [|i].
  do 2 rewrite mult_0_l in h1.
  rewrite plus_0_l in h1; subst.
  rewrite <- mult_1_r.
  constructor.
  destruct (lt_eq_lt_dec j k) as [[h2 | h2] | h2]; subst.
  pose proof (mult_S_lt_compat_l i j k h2) as h3.
  rewrite h1 in h3.
  destruct m as [|m].
  rewrite plus_0_r in h3.
  apply lt_irrefl in h3; contradiction.
  omega.
  destruct m as [|m].
  rewrite <- (mult_0_r (S i)).
  constructor.
  omega.
  pose proof (mult_S_lt_compat_l i k j h2) as h3. 
  pose proof (minus_plus_lt_cancel _ _ _ h3 h1); subst. 
  rewrite (mult_comm _ j), (mult_comm _ k).
  rewrite minus_dist_r.
  rewrite mult_comm.
  constructor.
Qed.

Lemma mult_dv_iff :
  forall i n,
    i * n dv n <->
    n = 0 \/ i = 1.
  intros i n; split; intro h1.
  inversion h1. 
  rewrite <- mult_assoc, mult_comm in H.
  symmetry in H.
  rewrite <- mult_assoc in H.
  rewrite eq_times_iff in H.
  destruct H as [h2 | h2]; subst.
  left.
  rewrite mult_0_r, mult_0_l; auto.
  destruct h2 as [h2 | h2].
  destruct h2 as [? h2]; subst.
  left.
  rewrite mult_0_r, mult_0_l; auto.
  apply mult_is_one in h2.
  destruct h2; subst.
  right; auto.
  destruct h1; subst.
  rewrite mult_0_r.
  rewrite <- (mult_1_r 0) at 2.
  constructor.
  rewrite mult_1_l.
  rewrite <- (mult_1_r n) at 2.
  constructor.
Qed.


Lemma plus_mult_dv_iff :
  forall i j n,
    (i + j * n) dv n <->
    i = 0 /\ (j = 1 \/ n = 0) \/
    (j = 0 \/ n = 0) /\ i dv n.
  intros i j n. 
  destruct i as [|i].
  rewrite plus_0_l.
  split; intro h1.
  rewrite mult_dv_iff in h1.
  destruct h1; subst.
  left.
  split; auto.
  left.
  split; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1].
  destruct h1; subst. 
  rewrite mult_1_l.
  apply dv_refl.
  rewrite mult_0_r.
  apply dv_refl.
  destruct h1 as [? h1]; subst.
  destruct H; subst.
  rewrite mult_0_l; auto.
  rewrite mult_0_r; auto.
  split; intro h1.
  right.
  inversion h1.
  symmetry in H.
  rewrite mult_plus_distr_r, plus_comm in H.
  rewrite (mult_comm j), <- mult_assoc in H.
  rewrite eq_times_plus_iff in H.
  destruct H as [h2 | h2].
  destruct h2; subst.
  apply mult_is_O in H0.
  destruct H0; subst.
  discriminate.
  split.
  right.
  rewrite mult_0_r; auto.
  rewrite mult_0_r; auto.
  apply dv0.
  destruct h2 as [h2 | h2].
  destruct h2 as [h2 h3].
  apply mult_is_one in h2.
  destruct h2; subst.
  rewrite mult_1_r in h3.
  discriminate.
  destruct h2 as [h2 h3].
  apply mult_is_O in h2.
  destruct h2; subst.
  split.
  left; auto.
  rewrite mult_0_l, plus_0_r.
  constructor.
  split.
  right.
  rewrite mult_0_r; auto.
  rewrite mult_0_r; auto.
  apply dv0.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  destruct h1 as [h1 h2].
  destruct h1; subst.
  rewrite mult_0_l, plus_0_r; auto.
  apply dv0.
Qed.

Lemma dv_plus :
  forall i m n,
    i dv m ->
    i dv n ->
    i dv (m + n).
  intros i j n h1 h2.
  destruct h1, h2.
  rewrite <- mult_plus_distr_l.
  constructor.
Qed.

Lemma dv_minus :
  forall i m n,
    i dv m ->
    i dv n ->
    i dv (m - n).
  intros i j n h1 h2.
  destruct h1, h2.
  rewrite minus_dist_l.
  constructor.
Qed.

Lemma gt_not_div :
  forall i n,
    n > 0 ->
    i > n ->
    ~i dv n. 
  intros i n h0 h1 h2.
  destruct h2.
  destruct n as [|n].
  rewrite mult_0_r in h0.
  apply lt_irrefl in h0; auto.
  rewrite <- (mult_1_r i) in h1 at 1.
  destruct i as [|i].
  rewrite mult_0_l in h0.
  apply lt_irrefl in h0; auto.
  apply mult_S_lt_l in h1.
  apply lt_S_n in h1.
  apply lt_n_0 in h1; auto.
Qed.




Lemma times_div :
  forall i j,
    0 < i ->
    0 < j ->
    i * j / i = j.
  intros i j h1 h2.
  rewrite (S_pred _ _ h1).
  simpl.
  pose proof (divmod_spec (j + pred i * j) (pred i ) 0 (pred i) (le_refl _)) as h3.
  destruct (divmod (j + pred i * j) (pred i ) 0 (pred i)) as [q u].
  rewrite <- minus_n_n, mult_0_r, plus_0_r, plus_0_r, <- (S_pred _ _ h1) in h3.
  destruct h3 as [h3 h4].
  simpl.
  rewrite <- (mult_1_l j) in h3 at 1.
  rewrite <- mult_plus_distr_r in h3.
  rewrite plus_comm in h3.
  rewrite S_compat in h3.
  rewrite <- (S_pred _ _ h1) in h3. 
  pose proof h3 as h5.
  apply times_eq_times_plus_div in h5.
  pose proof (dv_le _ _ h5) as h6.
  destruct h6 as [h6 | h6].
  rewrite h6, plus_0_r in h3.
  apply mult_inj in h3; auto.
  omega.
Qed.



Lemma times_mod_r : 
  forall i n,
    (i * n) mod n = 0.
  intros i n.
  destruct n as [|n]; simpl; auto.
  pose proof (divmod_spec (i * S n) n 0 n (le_refl _)) as h1.
  destruct (divmod (i * S n) n 0 n) as [q u].
  simpl.
  rewrite mult_0_r, <- minus_n_n, plus_0_r, plus_0_r in h1.
  destruct h1 as [h1 h2].
  rewrite mult_comm in h1 at 1.
  pose proof h1 as h3.
  apply times_eq_times_plus_div in h3.
  pose proof (dv_le (S n) (n - u) h3) as h4.
  destruct h4 as [|h4]; auto with arith.
  omega.
Qed.


Lemma times_mod_l : 
  forall i n,
    (n * i) mod n = 0.
  intros; rewrite mult_comm, times_mod_r; auto.
Qed.



Lemma dv_iff :
  forall m n,
    0 < m ->
    (m dv n <->
     (n mod m = 0)).
  intros m n h0; split; intro h1.
  destruct h1. 
  rewrite mult_comm.
  apply times_mod_r.
  rewrite <- neq0_iff in h0.
  pose proof (div_mod n m h0) as h2.
  rewrite h1, plus_0_r in h2.
  rewrite h2.
  constructor.
Qed.

Lemma mod_eq0_iff :
  forall m n,
    m mod n = 0 <->
    n = 0 \/ n dv m.
  intros m n.
  destruct (zerop n) as [|h1]; subst; simpl.
  tauto.
  rewrite <- dv_iff; auto.
  split; intro h2.
  right; auto.
  destruct h2; subst; auto.
  apply lt_irrefl in h1; contradiction.
Qed.


Lemma mod1 :
  forall n,
    n mod 1 = 0.
  intro n.
  rewrite mod_eq0_iff.
  right.
  apply one_dv.
Qed.


Lemma times_eq_iff : 
  forall i j k,
    0 < i ->
    0 < j ->
    (i * j = k <->
     (j = k / i /\ i dv k)).
  intros i j k h1; split; intro h2.   
  subst.
  rewrite times_div; auto.
  split; auto.
  constructor.
  destruct h2; subst.
  destruct H1.
  rewrite times_div; auto.
  destruct n; simpl in H.
  rewrite mult_0_r, O_div in H.
  apply lt_irrefl in H.
  contradiction.
  auto with arith.
Qed.


 
  
Lemma div_eq_0_iff :
  forall i n,
    i / n = 0 <->
    n = 0 \/ i < n.
  intros i n; split; intro h1.
  destruct n as [|n].
  left; auto.
  simpl in h1.
  right.
  pose proof (divmod_spec i n 0 n (le_refl _)) as h2.
  destruct (divmod i n 0 n) as [q u].
  simpl in h1; subst.
  rewrite <- minus_n_n, mult_0_r, plus_0_r, plus_0_r, plus_0_l in h2.
  destruct h2; subst.
  omega. 
  destruct h1 as [|h0]; subst.
  simpl; auto.
  destruct n as [|n].
  simpl; auto.
  simpl. 
  pose proof (divmod_spec i n 0 n (le_refl _)) as h2.
  destruct (divmod i n 0 n) as [q u].
  rewrite <- minus_n_n, mult_0_r, plus_0_r, plus_0_r in h2.
  destruct h2 as [h2 h3].
  subst.
  simpl.
  pose proof (le_plus_trans (S n * q) (S n * q) (n - u) (le_refl _)) as h4.
  pose proof (le_lt_trans _ _ _ h4 h0) as h5.
  rewrite lt_mult_iff in h5.
  destruct h5; auto.
Qed.


Lemma lt_mod_eq :
  forall i n,
    i < n ->
    i mod n = i.
  intros i n h1.
  pose proof (div_eq_0_iff i n) as h3.
  hr h3.
  right; auto.
  rewrite <- h3 in hr.
  pose proof (div_mod i n) as h4.
  destruct (zerop n) as [|h5]; subst.
  apply lt_n_0 in h1.
  contradiction.
  rewrite <- neq0_iff in h5.
  specialize (h4 h5).
  rewrite hr, mult_0_r, plus_0_l in h4; auto.
Qed.


Lemma eq_div_iff :
  forall i j n,
    i = j / n <->
    n = 0 /\ i = 0 \/ 0 < n /\ j - j mod n = i * n.
  intros i j n; split; intro h1; subst.   
  destruct (zerop n) as [|h1]; subst.  
  simpl.
  left; auto. 
  right.
  split; auto.
  rewrite (S_pred _ _ h1) at 1 2.
  simpl.
  pose proof (divmod_spec j (pred n) 0 (pred n) (le_refl _)) as h2.   
  destruct (divmod j (pred n) 0 (pred n)) as [q u].
  simpl. 
  rewrite <- (S_pred _ _ h1), <- minus_n_n, plus_0_r, mult_0_r, plus_0_r in h2.   
  destruct h2 as [h0 h2]; subst. 
  rewrite <- (plus_minus_assoc (n*q) (pred n - u) _ (le_refl _)).
  rewrite <- minus_n_n. 
  rewrite plus_0_r. 
  rewrite mult_comm; auto. 
  destruct h1 as [h1 | h1].
  destruct h1; subst; simpl; auto.
  destruct h1 as [h1 h2].
  rewrite (S_pred _ _ h1) in h2 at 1.
  rewrite (S_pred _ _ h1).
  simpl in h2; simpl.
  pose proof (divmod_spec j (pred n) 0 (pred n) (le_refl _)) as h3.
  destruct (divmod j (pred n) 0 (pred n)) as [q u].
  simpl.
  simpl in h2, h3.
  rewrite <- minus_n_n, plus_0_r, mult_0_r, plus_0_r in h3.
  destruct h3 as [h3 h4].
  rewrite h3 in h2.
  rewrite <- (plus_minus_assoc (q + pred n*q) (pred n - u) _ (le_refl _)) in h2.
  rewrite <- minus_n_n, plus_0_r in h2.
  rewrite <- (mult_1_l q) in h2 at 1.
  rewrite <- mult_plus_distr_r in h2.
  rewrite plus_comm in h2.
  rewrite S_compat in h2.
  rewrite <- (S_pred _ _ h1) in h2.
  rewrite (mult_comm i) in h2.
  eapply mult_inj in h2; auto.
Qed.

  
Lemma times_div_eq_iff :
  forall i n,
    n * (i / n) = i <->
    n dv i.
  intros i n; split; intro h1. 
  unfold div in h1.
  destruct n as [|n].
  rewrite mult_0_r in h1; subst.
  apply dv0.
  pose proof (divmod_spec i n 0 n (le_refl _)) as h2.
  destruct (divmod i n 0 n) as [q u].
  rewrite <- h1.
  constructor.
  destruct h1.
  destruct n as [|n], n0 as [|n0].
  simpl; auto.
  simpl; auto.
  rewrite mult_0_r, O_div, mult_0_r; auto.
  rewrite times_div; auto with arith.
Qed.



End Divides.


Lemma le_div :
  forall i j n,
    i <= j ->
    i / n <= j / n.
  intros i j n h1. 
  unfold div.
  destruct n as [|n]; auto with arith.
  pose proof (divmod_spec i n 0 n (le_refl _)) as h2.
  pose proof (divmod_spec j n 0 n (le_refl _)) as h3.
  destruct (divmod i n 0 n) as [q u],
                               (divmod j n 0 n) as [q' u'].
  rewrite mult_0_r, plus_0_r, <- (minus_n_n), plus_0_r in h2, h3. 
  destruct h2 as [h2 hu], h3 as [h3 hu']; subst.
  simpl.
  assert (h2:n - u <= n). omega.
  assert (h3:n - u' <= n). omega.
  pose proof (plus_le_compat_r  _ _ (S n * q') h3) as h4.
  rewrite plus_comm in h4.
  pose proof (le_trans _ _ _ h1 h4) as h5.
  pose proof (minus_le_compat_r _ _ (n - u) h5) as h6.
  rewrite <- plus_minus_assoc, <- minus_n_n, plus_0_r in h6; auto.
  rewrite plus_comm in h6.
  rewrite minus_minus_assoc, <- plus_minus_assoc, <- minus_n_n,
  plus_0_r in h6; auto with arith.
  apply le_lt in hu. 
  pose proof (plus_lt_compat_r _ _ (S n * q') hu) as h7.
  rewrite <- (mult_1_r (S n)) in h7 at 2.
  rewrite <- mult_plus_distr_l in h7.
  rewrite plus_comm in h6.
  pose proof (le_lt_trans _ _ _ h6 h7) as h8.
  apply mult_S_lt_l in h8; auto with arith.
  pose proof (le_minus n u) as h7.
  pose proof (le_plus_l n (S n * q')) as h8.
  rewrite plus_comm in h8; auto with arith.
  eapply le_trans.
  apply h7.
  assumption.
Qed.  
  
          
Lemma times_eq_times_plus_iff :
  forall i j k m,
    i * j = i * k + m <->
    ((i = 0 \/ j = k) /\ m = 0) \/
    i > 0 /\ j > k /\ m = i * (j - k).
  intros i j k m.
  destruct (zerop i) as [|h1]; subst.
  do 2 rewrite mult_0_l.
  rewrite plus_0_l. 
  split; intro h1; subst. 
  left.
  split; auto.
  destruct h1 as [h1 | h1]; auto.
  destruct h1; auto.
  destruct h1 as [h1].
  apply lt_irrefl in h1; contradiction.
  split; intro h2.  
  destruct (lt_eq_lt_dec j k) as [[h3 | h3] | h3].
  pose proof (mult_S_lt_compat_l (pred i) _ _ h3) as h4.
  rewrite <- (S_pred _ _ h1) in h4.
  right.
  omega.
  subst.
  left.
  split; auto.
  omega.
  right.
  split; auto.
  pose proof (mult_S_lt_compat_l (pred i) _ _ h3) as h4.
  rewrite <- (S_pred _ _ h1) in h4.
  pose proof (f_equal (fun c => minus c (i * k)) h2) as h5. 
  simpl in h5.
  rewrite plus_comm in h5.
  rewrite <- plus_minus_assoc in h5; auto with arith.
  rewrite <- minus_n_n, plus_0_r in h5.
  rewrite mult_comm, (mult_comm i k), minus_dist_r in h5.
  rewrite times_eq_iff in h5.
  destruct h5 as [h5 h6]. 
  inversion h6.
  subst.
  destruct (zerop n) as [|hn]; subst.
  rewrite mult_0_r in h1.
  rewrite O_div in h1.
  apply lt_irrefl in h1.
  contradiction.
  rewrite times_div in h1; auto.
     
  split; auto.
  rewrite times_div.
  apply mult_comm.
  omega.
  assumption.
  omega.
  omega.
  assumption.

  destruct h2 as [h2 | h2].
  destruct h2 as [[? | ?] ?]; subst.
  apply lt_irrefl in h1; contradiction.
  auto with arith.
  destruct h2 as [h2 [h3 h4]].    
  apply (inj_minus _ _ (i * k)).
  apply mult_le_compat_l; auto with arith.
  auto with arith. 
  rewrite (mult_comm i j), (mult_comm i k).
  rewrite  minus_dist_r.
  rewrite (mult_comm (j - k)).
  rewrite <- h4.
  rewrite plus_comm.
  rewrite <- plus_minus_assoc; auto with arith.
  rewrite <- minus_n_n, plus_0_r.
  reflexivity.
Qed.


Lemma times_eq_times_minus_iff :
  forall i j k m,
    i * j = i * k - m <->
    i = 0 \/
    (j = 0 /\ i * k <= m) \/ 
    ((i > 0 /\ j > 0) /\
     (j = k /\ m = 0 \/
      j < k /\ m = i * (k - j))).
  intros i j k m.
  destruct (zerop i) as [|h1]; subst.
  do 2 rewrite mult_0_l.
  simpl.
  tauto.
  destruct (zerop j) as [|hj]; subst.
  rewrite mult_0_r, <- minus_n_O.
  split; intro h2.
  symmetry in h2.
  rewrite <- le_minus_iff in h2. 
  right.  
  left; auto.
  destruct h2 as [h2 | h2].
  subst. 
  apply lt_irrefl in h1; contradiction.
  destruct h2 as [h2 | h2].
  destruct h2 as [? h2].
  rewrite le_minus_iff in h2; auto.
  destruct h2 as [hg [h2 | h2]].
  destruct hg as [? hg].
  apply lt_irrefl in hg; contradiction.
  destruct hg as [? hg].
  apply lt_irrefl in hg.
  contradiction.
  split; intro h2.   
  destruct (lt_eq_lt_dec j k) as [[h3 | h3] | h3].
  pose proof (mult_S_lt_compat_l (pred i) _ _ h3) as h4.
  rewrite <- (S_pred _ _ h1) in h4.    
  right.
  right.
  split; auto.
  right.
  split; auto.
  pose proof (f_equal (fun c => c + m) h2) as h5. 
  simpl in h5.  
  rewrite <- minus_minus_assoc, <- minus_n_n, <- minus_n_O in h5.
  rewrite plus_comm in h5.
  pose proof (f_equal (fun c => c - i * j) h5) as h6.
  simpl in h6. 
  rewrite <- plus_minus_assoc, <- minus_n_n, plus_0_r in h6.
  rewrite minus_dist_l in h6; auto. 
  auto. 
  rewrite <- minus_n_n; auto with arith.
  auto.
  destruct (le_lt_dec m (i * k) ) as [h6 | h6]; auto.
  apply lt_le_weak in h6.
  rewrite le_minus_iff in h6.
  rewrite h6, plus_0_l in h5.
  symmetry in h5.
  rewrite plus_comm, plus_eq_iff in h5.
  apply mult_is_O in h5; destruct h5; subst.
  apply lt_irrefl in h1.
  contradiction.
  apply lt_irrefl in hj; contradiction.
  subst.
  rewrite minus_eq_iff in h2.
  destruct h2 as [h2 | h2].
  apply mult_is_O in h2.
  destruct h2; subst.
  apply lt_irrefl in h1; contradiction.
  apply lt_irrefl in hj; contradiction.
  subst.
  right. 
  right.
  split; auto.
  right.
  right.
  split; auto.
  right.
  pose proof (mult_lt_compat_l _ _ i h3 h1) as h4.
  pose proof (le_minus (i * k) m) as h5.
  rewrite <- h2 in h5.
  pose proof (lt_le_trans _ _ _ h4 h5) as h6.
  apply lt_irrefl in h6; contradiction.
  destruct h2 as [? | h2]; subst.
  apply lt_irrefl in h1; contradiction.
  destruct h2 as [h2 | h2].
  destruct h2 as [? h2]; subst.
  apply lt_irrefl in hj; contradiction.
  destruct h2 as [hg [h2 | h2]].
  destruct h2 as [? h2]; subst.
  rewrite <- minus_n_O; auto.
  destruct h2 as [? h2]; subst.
  rewrite <- minus_dist_l.
  rewrite minus_minus_assoc.
  rewrite <- minus_n_n, plus_0_l; auto.
  apply le_minus.
  apply mult_le_compat_l; auto with arith.
  auto.
Qed.                       


Lemma times_eq_times_plus_minus_iff : 
  forall i j k m n,
    i * j = i * k + m - n <->
    (i = 0 /\ m <= n) \/ 
    (j = k /\ m = n) \/
    (j = 0 /\ i * k + m <= n) \/
    (i > 0 /\ j > 0 /\
     ((j > k /\
       (n < m /\ m = i * (j - k) + n)) \/
      (j < k /\
       (m <= n /\
        n = i * (k - j) + m)))).
  intros i j k m n; split; intro h1.
  destruct (zerop i) as [|h2]; subst.
  do 2 rewrite mult_0_l in h1.
  rewrite plus_0_l in h1.
  symmetry in h1.
  rewrite <- le_minus_iff in h1.
  left.
  split; auto.
  destruct (zerop j) as [|h3]; subst.
  rewrite mult_0_r in h1.
  symmetry in h1.
  rewrite <- le_minus_iff in h1.
  right.
  right.
  left.
  split; auto. 
  destruct (lt_eq_lt_dec j k) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  destruct (le_lt_dec m n) as [h5 | h5].
  pose proof h4 as h4'.
  rewrite lt_iff in h4; auto.
  pose proof (f_equal (fun c => c - (i * j)) h1) as h9.
  simpl in h9.  
  rewrite <- minus_n_n in h9.
  rewrite plus_minus_le in h9; auto.
  rewrite minus_minus_lt' in h9. 
  rewrite minus_dist_l in h9.
  symmetry in h9. 
  rewrite <- le_minus_iff in h9.
  right.
  right.
  right.
  split; auto.
  split; auto.
  destruct (le_lt_dec (i *k + m) n) as [hl | hl]; auto.
  rewrite le_minus_iff in hl.
  rewrite hl in h1.
  apply mult_is_O in h1.
  destruct h1; subst. 
  apply lt_irrefl in h2; contradiction.
  apply lt_irrefl in h3; contradiction.
  right.
  split; auto.
  split; auto.
  pose proof (f_equal (minus (i * k)) h1) as h10.
  rewrite minus_minus_plus_minus_same_le in h10; auto with arith.
  rewrite minus_dist_l in h10.
  pose proof (f_equal (fun c => c + m) h10) as he.
  simpl in he.
  rewrite minus_plus_same in he; auto with arith.

  rewrite (S_pred _ _ h2).
  apply mult_S_lt_compat_l; auto.
  pose proof (f_equal (fun c => c - (i * j)) h1) as h9.
  simpl in h9. 
  rewrite <- minus_n_n in h9; auto.
  rewrite plus_minus_minus_lt in h9.
  rewrite minus_dist_l in h9.
  rewrite <- plus_minus_assoc in h9; auto with arith.
  symmetry in h9.
  apply plus_is_O in h9.
  destruct h9 as [h9 h10].
  apply mult_is_O in h9.
  destruct h9 as [|h9].
  subst.
  apply lt_irrefl in h2; contradiction.
  rewrite <- le_minus_iff in h9, h10.
  pose proof (lt_le_trans _ _ _ h4 h9) as h11.
  apply lt_irrefl in h11.
  contradiction. 
  rewrite (S_pred _ _ h2).
  apply mult_S_lt_compat_l; auto.
  rewrite plus_minus_eq_iff in h1.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 h4].
  apply mult_is_O in h4.
  destruct h4; subst.
  apply lt_irrefl in h2; contradiction.
  apply lt_irrefl in h3; contradiction.  
  subst. 
  right.
  left; auto. 
  right.
  right.
  right.
  split; auto.
  split; auto.
  left; auto.
  split; auto.  
  destruct (le_lt_dec m n) as [h5 | h5].
  rewrite plus_minus_le in h1; auto.
  pose proof (f_equal (fun c => c + (n - m)) h1) as h6.
  simpl in h6.   
  rewrite <- minus_minus_assoc, minus_same, <- minus_n_O in h6; auto.
  pose proof (f_equal (minus (i * j)) h6) as h7.
  pose proof (le_plus_l (i * j) (n - m)) as h8.
  rewrite le_minus_iff in h8.
  rewrite h8 in h7.
  symmetry in h7.
  rewrite <- le_minus_iff in h7.
  rewrite (S_pred _ _ h2) in h7.
  apply mult_S_le_reg_l in h7.
  pose proof (lt_le_trans _ _ _ h4 h7) as h9.
  apply lt_irrefl in h9; contradiction.
  rewrite minus_same; auto with arith.
  destruct (le_lt_dec (n - m) (i * k)) as [h7 | h7]; auto.
  apply lt_le_weak in h7.
  rewrite le_minus_iff in h7.
  rewrite h7 in h1.
  apply mult_is_O in h1.
  destruct h1; subst.
  apply lt_irrefl in h2; contradiction.
  apply lt_irrefl in h3; contradiction.
  split; auto.
  rewrite <- plus_minus_assoc in h1; auto with arith.
  pose proof (f_equal (fun c => c - i * k) h1) as h6.
  simpl in h6.
  rewrite plus_comm in h6.
  rewrite <- plus_minus_assoc, minus_same, plus_0_r in h6; auto.
  rewrite minus_dist_l in h6.
  rewrite h6.
  rewrite minus_plus_same; auto with arith.
  destruct h1 as [h1 | h1]; subst.
  destruct h1 as [? h1]; subst.
  do 2 rewrite mult_0_l.
  rewrite plus_0_l.
  rewrite le_minus_iff in h1; auto.
  destruct h1 as [h1 | h1].
  destruct h1; subst.
  rewrite plus_minus_le; auto.
  rewrite minus_same, <- minus_n_O; auto.
  destruct h1 as [h1 | h1].
  destruct h1 as [? h1]; subst.
  rewrite mult_0_r. 
  rewrite le_minus_iff in h1; auto. 
  destruct h1 as [h1 [h2 h3]].
  destruct h3 as [h3 | h3].
  destruct h3 as [h3 [h4 h5]].   
  pose proof (f_equal (fun c => c - n) h5) as h6.
  simpl in h6.
  rewrite plus_minus_same_l in h6. 
  rewrite <- plus_minus_assoc; auto with arith.
  rewrite h6.
  rewrite <- minus_dist_l.
  rewrite plus_minus_same'; auto.
  apply mult_le_compat_l; auto with arith.  
  destruct h3 as [h3 [h4 h5]].
  rewrite h5.
  rewrite <- minus_dist_l.
  rewrite minus_plus_assoc.
  rewrite plus_minus_minus_same''.
  rewrite minus_minus_assoc, <- minus_n_n; auto with arith.
  apply le_minus.
  apply minus_plus_le.
Qed.


Lemma times_plus_eq_times_plus_iff :
  forall i j k m n,
    i * j + m =
    i * k + n <->
    ((i = 0 \/ j = k) /\ m = n) \/
    (i > 0 /\
     (j < k /\ m = i*(k - j) + n) \/
     (k < j /\ n = i*(j - k) + m)).
  intros i j k m n. 
  destruct (zerop i) as [|hi]; subst.
  do 2 rewrite mult_0_l, plus_0_l.
  split; intro h1; subst.
  left.
  split; auto.
  destruct h1 as [h1 | h1].
  destruct h1; subst.
  reflexivity.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 h2].
  apply lt_irrefl in h1; contradiction. 
  rewrite mult_0_l, plus_0_l in h1.
  destruct h1; auto.
  split; intro h1.
  destruct (lt_eq_lt_dec j k) as [[h2 | h2] | h2].
  right. 
  left.
  split; auto with arith.
  split; auto.
  apply (plus_inj_r _ _ (i*j)).
  rewrite plus_comm, h1.
  rewrite <- plus_assoc, (plus_comm n (i*j)), plus_assoc.
  rewrite <- mult_plus_distr_l.
  rewrite <- minus_minus_assoc, <- minus_n_n, <- minus_n_O; auto with arith.
  rewrite <- minus_n_n; auto with arith.
  subst.
  left.
  split.
  right; auto.
  omega.
  right.
  right.
  split; auto.
  rewrite mult_comm.
  rewrite <- minus_dist_r. 
  rewrite plus_comm.
  rewrite plus_minus_assoc.
  rewrite mult_comm, plus_comm.
  rewrite h1.
  rewrite plus_comm, <- plus_minus_assoc.
  rewrite mult_comm, <- minus_n_n, plus_0_r; auto.
  rewrite mult_comm; auto with arith.
  apply mult_le_compat_r; auto with arith.

  destruct h1 as [h1 | h1].
  destruct h1 as [[? | ?]  ?]; subst.
  apply lt_irrefl in hi; contradiction.
  reflexivity.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 [h2 h3]]; subst.
  rewrite (mult_comm _ (k - j)), <- minus_dist_r. 
  rewrite plus_assoc.
  rewrite plus_minus_assoc.
  rewrite (plus_comm (i*j) (k*i)).
  rewrite <- plus_minus_assoc.
  rewrite (mult_comm i j), <- minus_n_n, plus_0_r.
  rewrite mult_comm; auto.
  rewrite mult_comm; auto.
  apply mult_le_compat_r; auto with arith.
  destruct h1 as [h1]; subst.
  rewrite (mult_comm _ (j - k)). 
  rewrite <- minus_dist_r.
  rewrite plus_assoc, plus_minus_assoc.
  rewrite (plus_comm (i*k)).
  rewrite <- plus_minus_assoc.
  rewrite (mult_comm i k), <- minus_n_n, plus_0_r, mult_comm; auto.
  rewrite mult_comm; auto.
  apply mult_le_compat_r; auto with arith.
Qed.       


End Times.
 
Lemma dv_plus_rev_r :
  forall i m n,
    i dv m ->
    i dv (m + n) ->
    i dv n.
  intros i m n h1 h2.
  destruct h1.
  inversion h2.
  rewrite times_eq_times_plus_iff in H0.
  destruct H0 as [h3 | h3].
  destruct h3 as [[? | ?] ?]; subst.
  apply dv0.
  apply dv0.
  destruct h3 as [? [? ?]]; subst.
  constructor.
Qed.

Lemma dv_plus_rev_l :
  forall i m n,
    i dv n ->
    i dv (m + n) ->
    i dv m.
  intros; eapply dv_plus_rev_r.
  apply H.
  rewrite plus_comm; auto.
Qed.
 

Lemma dv_minus_rev_l :
  forall i m n,
    i dv m ->
    i dv (m - n) ->
    i = 0 \/ m <= n \/ i dv n.
  intros i m n h1 h2. 
  destruct h1.
  inversion h2.  
  rewrite times_eq_times_minus_iff in H0.
  destruct H0 as [|h3]; subst.
  left ;auto. 
  destruct h3 as [h3 | h3].
  destruct h3 as [? h3]; subst.
  right.
  left; auto.
  destruct h3 as [h3 h4].
  destruct h3 as [h3 h5].
  destruct h4 as [h4 | h4].
  destruct h4; subst.
  right.
  right.
  apply dv0.
  right.
  destruct h4 as [h4]; subst.
  right.
  constructor.
Qed.

Lemma dv_minus_rev_r :
  forall i m n,
    i dv n ->
    i dv (m - n) ->
    m <= n \/ i dv m.
  intros i m n h1 h2.
  destruct h1.
  inversion h2.
  pose proof (f_equal (fun c => c + (i * n)) H0) as h3.
  simpl in h3.
  destruct (le_lt_dec (i * n) m) as [h4 | h4].
  rewrite <- minus_minus_assoc, <- minus_n_n, <- minus_n_O in h3; subst; auto with arith.
  rewrite <- mult_plus_distr_l.
  right.
  constructor.
  rewrite <- minus_n_n; auto with arith.
  apply lt_le_weak in h4.
  rewrite le_minus_iff in h4.
  symmetry in h3.
  rewrite h4, plus_0_l, plus_comm, plus_eq_iff in h3.
  apply mult_is_O in h3; destruct h3; subst.
  rewrite mult_0_l, <- minus_n_O in h4; subst.
  right.
  apply dv0.
  rewrite mult_0_r in H0.
  symmetry in H0.
  rewrite <- le_minus_iff in H0.
  left; auto.
Qed.

  
Lemma not_dv_plus_l :
  forall i m n,
    ~i dv m ->
    i dv n ->
    ~i dv (m + n).
  intros i m n h1 h2 h3.
  pose proof (dv_plus_rev_l _ _ _ h2 h3) as h4; contradiction.
Qed.

Lemma not_dv_plus_r :
  forall i m n,
    ~i dv n ->
    i dv m ->
    ~i dv (m + n).
  intros i m n h1 h2 h3.
  pose proof (dv_plus_rev_r _ _ _ h2 h3) as h4; contradiction.
Qed.

  
Lemma dv_not_plus_l :
  forall i m n,
    i dv m ->
    ~i dv (m + n) ->
    ~i dv n.
  intros i m n h1 h2 h3.
  pose proof (dv_plus _ _ _ h1 h3); contradiction.
Qed.

  
Lemma dv_not_plus_r :
  forall i m n,
    i dv n ->
    ~i dv (m + n) ->
    ~i dv m.
  intros i m n h1 h2 h3.
  pose proof (dv_plus _ _ _ h3 h1); contradiction.
Qed.


Lemma not_dv_minus_l :
  forall i m n,
    ~i dv m ->
    i dv n ->
    m <= n \/ ~i dv (m - n).
  intros i m n h1 h2.
  destruct (le_lt_dec m n) as [h3 | h3].
  left; auto.
  right.
  intro h4.
  pose proof (dv_minus_rev_r _ _ _ h2 h4) as h5.
  destruct h5 as [h5 | h5].
  pose proof (lt_le_trans _ _ _ h3 h5) as h6.
  apply lt_irrefl in h6; contradiction.
  contradiction.
Qed.

Lemma not_dv_minus_r :
  forall i m n,
    ~i dv n ->
    i dv m ->
    i = 0 \/ m <= n \/ ~i dv (m - n).
  intros i m n h1 h2.
  destruct (zerop i) as [|hz].
  left; auto.
  right.
  destruct (le_lt_dec m n) as [h3 | h3].
  left; auto.
  right.
  intro h4.
  pose proof (dv_minus_rev_l _ _ _ h2 h4) as h5.
  destruct h5 as [h5 | h5].
  subst.
  apply lt_irrefl in hz; contradiction.
  destruct h5 as [h5 | h5].
  pose proof (le_lt_trans _ _ _  h5 h3) as h6.
  apply lt_irrefl in h6; auto.
  contradiction.
Qed.

Lemma dv_not_minus_l :
  forall i m n,
    i dv m ->
    ~i dv (m - n) ->
    ~i dv n.
  intros i m n h1 h2 h3.
  pose proof (dv_minus _ _ _ h1 h3) as h4.
  contradiction.
Qed.

Lemma dv_not_minus_r :
  forall i m n,
    i dv n ->
    ~i dv (m - n) ->
    ~i dv m.
  intros i m n h1 h2 h3.
  pose proof (dv_minus _ _ _ h3 h1) as h4.
  contradiction.
Qed.



Lemma not_dv_plus_rev_l :
  forall i m n,
    ~i dv m ->
    i dv (m + n) ->
    ~i dv n.
  intros i m n h1 h2.
  intro h3.
  pose proof (dv_plus_rev_l _ _ _ h3 h2) as h4; contradiction.
Qed.

Lemma not_dv_plus_rev_r :
  forall i m n,
    ~i dv n ->
    i dv (m + n) ->
    ~i dv m.
  intros i m n h1 h2.
  intro h3.
  pose proof (dv_plus_rev_r _ _ _ h3 h2) as h4; contradiction.
Qed.

Lemma not_dv_minus_rev_l :
  forall i m n,
    ~i dv m ->
    i dv (m - n) ->
    m <= n \/ ~i dv n.
  intros i m n h1 h2.
  destruct (le_lt_dec m n) as [hl | hl].
  left; auto.
  right.
  intro h3.
  pose proof (dv_minus_rev_r _ _ _ h3 h2) as h4.
  destruct h4 as [h4 | h4].
  pose proof (lt_le_trans _ _ _ hl h4) as h5.
  apply lt_irrefl in h5; auto.
  contradiction.
Qed.


Lemma not_dv_minus_rev_r :
  forall i m n,
    ~i dv n ->
    i dv (m - n) ->
    i = 0 \/ m <= n \/ ~i dv m.
  intros i m n h1 h2.
  destruct (zerop i) as [hz | hz].
  left; auto.
  right.
  destruct (le_lt_dec m n) as [hl | hl].
  left; auto.
  right.
  intro h3.
  pose proof (dv_minus_rev_l _ _ _ h3 h2) as h4.
  destruct h4 as [h4 | h4].
  subst.
  apply lt_irrefl in hz; contradiction.
  destruct h4 as [h4 | h4].
  pose proof (lt_le_trans _ _ _ hl h4) as h5.
  apply lt_irrefl in h5; auto.
  contradiction.
Qed. 



  


Lemma minus_div_le :
  forall i j n,
    (i - j) / n <= i / n - j / n.
  intros i j n.
  destruct (le_lt_dec i j) as [hlt | hlt].
  rewrite le_minus_iff in hlt.
  rewrite hlt, O_div; auto with arith.
  unfold div.
  destruct n as [|n]; auto. 
  pose proof (divmod_spec (i - j) n 0 n (le_refl _)) as h1.
  pose proof (divmod_spec i n 0 n (le_refl _)) as h2.
  pose proof (divmod_spec j n 0 n (le_refl _)) as h3.
  destruct (divmod (i - j) n 0 n) as [q u],
                                     (divmod i n 0 n) as [q' u'],
                                                         (divmod j n 0 n) as [q0 u0].
  rewrite <- minus_n_n, mult_0_r, plus_0_r, plus_0_r in h1, h2, h3.
  simpl.
  destruct h1 as [h1 h4], h2 as [h2 h5], h3 as [h3 h6]; subst.
  rewrite minus_plus_assoc in h1.
  pose proof (le_minus n u0) as h7.
  pose proof (le_minus n u') as h8.
  pose proof (le_minus n u) as h9.
  red in hlt.
  destruct (le_lt_dec u' u0) as [hl | hl].
  pose proof (minus_le_compat_r _ _ (n - u0) hlt) as h10.
  rewrite <- plus_S_shift_r in h10.
  rewrite <- plus_minus_assoc in h10.
  rewrite S_minus in h10.
  rewrite S_compat in h10. 
  apply le_lt in h10.
  apply lt_S_n in h10.
  rewrite <- plus_minus_assoc, minus_same_diff in h10; auto.
  pose proof (le_minus u0 u') as h11.
  pose proof (le_trans _ _ _ h11 h6) as h12.
  pose proof (plus_le_compat_r _ _ (S n * q') h12) as h13.
  apply le_lt in h13.
  rewrite <- plus_S_shift_l in h13.
  rewrite plus_comm in h13.
  pose proof (lt_trans _ _ _ h10 h13) as h14.
  rewrite <- (mult_1_r (S n)) in h14 at 2.
  rewrite <- mult_plus_distr_l in h14.
  rewrite plus_comm, S_compat in h14.
  apply mult_S_lt_l in h14.
  apply lt_le in h14.
  rewrite plus_comm in h1.
  rewrite <- plus_minus_assoc in h1.
  rewrite minus_dist_l in h1.
  rewrite plus_comm in h1.
  rewrite <- plus_minus_assoc, minus_same_diff in h1.
  rewrite times_plus_eq_times_plus_iff in h1.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 h2].
  destruct h1; subst; auto with arith.
  discriminate.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 [h2 h3]].
  rewrite h3 in h12.
  pose proof (le_plus_l (S n * (q - (q' - q0))) (n- u)) as h15.
  pose proof (le_trans _ _ _ h15 h12) as h16.
  rewrite lt_iff in h2.
  do 2 red in h2.
  pose proof (mult_le_compat_r _ _ (S n) h2) as h17.
  rewrite mult_1_l in h17.
  rewrite mult_comm in h17.
  pose proof (le_trans _ _ _ h17 h16) as h18.
  apply le_Sn_n in h18. 
  contradiction.
  destruct h1; auto with arith.
  assumption.
  assumption.  
  apply ge_minus_anti; auto.
  apply mult_le_compat_l; auto.
  apply ge_minus_anti; auto. 
  auto with arith.  
  change (S n * q0 + (n - u0) < S n * q' + (n - u')) in hlt.
  pose proof hl as h10.
       
  apply (minus_lt_compat_r _ _ (n - u0)) in hlt.   
  rewrite <- plus_minus_assoc in hlt; auto.
  rewrite <- minus_n_n, plus_0_r in hlt.
  pose proof (le_minus (S n * q' + (n- u')) (n - u0)) as h11.
  pose proof (lt_le_trans _ _ _ hlt h11) as h12.
  pose proof (plus_le_compat_r  _ _ (S n  * q') h8) as h13.
  rewrite plus_comm in h13.
  pose proof (lt_le_trans _ _ _ h12 h13) as h14.
  pose proof (lt_n_Sn n) as h15.
  pose proof (plus_lt_compat_r _ _ (S n * q') h15) as h16.
  pose proof (lt_trans _ _ _ h14 h16) as h17.
  rewrite <- (mult_1_r (S n)) in h17 at 2.
  rewrite <- mult_plus_distr_l in h17.
  apply mult_S_lt_l in h17. 
  rewrite plus_comm, S_compat in h17. 
  apply lt_le in h17.
  rewrite plus_comm in h1.
  rewrite <- plus_minus_assoc in h1.
  rewrite minus_dist_l in h1.
  apply le_lt_eq_dec in h17.
  destruct h17 as [h17 | h17]; subst. 
  pose proof h17 as h17'.
  rewrite lt_iff in h17'. 
  rewrite plus_comm in h1.
  pose proof (f_equal (fun c => c + (n - u0)) h1) as h18.
  simpl in h18.  
  change (S n * (q' - q0) + (n - u') - (n - u0) + (n - u0) =
          S n * q + (n - u) + (n - u0)) in h18. 
  rewrite <- minus_minus_assoc in h18.
  rewrite <- minus_n_n, <- minus_n_O in h18.
  rewrite <- plus_assoc in h18.
  rewrite times_plus_eq_times_plus_iff in h18.
  destruct h18 as [h18 | h18].
  destruct h18 as [h18 h19].
  destruct h18; subst.
  discriminate. 
  auto. 
  destruct h18 as [h18 | h18].
  destruct h18 as [h18 h19].
  destruct h19 as [h19 h20].
  pose proof (le_minus n u') as h21.
  pose proof (lt_n_Sn n) as h22.
  pose proof (le_lt_trans _ _ _  h21 h22) as h23.
  pose proof h19 as h24.
  rewrite lt_iff in h24.
  pose proof (mult_le_compat_l _ _ (S n) h24) as h25.
  rewrite mult_1_r in h25. 
  pose proof (lt_le_trans _ _ _ h23 h25) as h26. 
  pose proof (le_plus_l (S n * (q - (q' - q0))) (n - u + (n - u0))) as h27. 
  pose proof (lt_le_trans _ _ _ h26 h27) as h28.
  rewrite h20 in h28.
  apply lt_irrefl in h28.
  contradiction.    
  destruct h18; auto with arith.
  rewrite <- minus_n_n; auto with arith.
  auto. 
  pose proof (mult_le_compat_l _ _ (S n) h17') as h19.
  rewrite mult_1_r in h19. 
  pose proof (le_plus_l (S n * (q'  - q0)) (n - u')) as h20.
  pose proof (le_trans _ _ _ h19 h20) as h21.
  pose proof (le_minus n u0) as h22.
  pose proof (lt_n_Sn n) as h23. 
  pose proof (lt_le_trans _ _ _ h23 h21) as h24. 
  pose proof (le_lt_trans _ _ _ h22 h24); auto with arith.
  rewrite <- minus_n_n, mult_0_r, plus_0_r in h1.
  apply (gt_minus_anti _ _ n) in h10.
  apply lt_le_weak in h10.
  rewrite le_minus_iff in h10.
  rewrite h10 in h1.
  simpl in h1.
  symmetry in h1.
  apply plus_is_O in h1.
  destruct h1 as [h1 h17].
  apply plus_is_O in h1.
  destruct h1; subst.
  rewrite <- minus_n_n; auto. 
  assumption.
  apply mult_le_compat_l; auto.  
  destruct q' as [|q']; subst. 
  rewrite mult_0_r, plus_0_l in h1.
  destruct q0 as [|q0].
  rewrite mult_0_r, plus_0_l, plus_0_l in hlt.
  pose proof (gt_minus_anti _ _ n h10 h5) as h11.
  pose proof (lt_trans _ _ _ h11 hlt) as h12.
  apply lt_irrefl in h12; contradiction.
  rewrite mult_0_r, plus_0_l in hlt.
  pose proof (lt_n_Sn n) as h11.
  pose proof (le_lt_trans _ _ _ h8 h11) as h12.
  pose proof (mult_O_le (S n) (S q0)) as h13.
  rewrite mult_comm in h13.
  destruct h13 as [|h13].
  discriminate.
  pose proof (lt_le_trans _ _ _ h12 h13) as h14.
  pose proof (le_plus_l (S n * S q0) (n - u0)) as h15.
  pose proof (lt_le_trans _ _ _ h14 h15) as h16.
  rewrite hlt in h16.
  apply lt_irrefl in h16.
  contradiction.
  pose proof (lt_n_Sn n) as h11.
  pose proof (le_lt_trans _ _ _ h7 h11) as h12.
  pose proof (mult_O_le (S n) (S q')) as h13.
  destruct h13 as [h13 | h13].
  discriminate.
  pose proof (lt_le_trans _ _ _ h12 h13) as h14.
  pose proof (le_plus_l (S q' * S n) (n - u')) as h15.
  pose proof (lt_le_trans _ _ _ h14 h15) as h16.
  rewrite mult_comm in h16; auto.
  auto with arith.
Qed.
  



Definition seg_pred_dec_from_S 
           {n:nat}
           {pf:seg_pred_type (S n)}
           (f:seg_pred_dec pf) : 
  seg_pred_dec (fun i (pflt:i < n) =>
                 (pf _ (lt_trans _ _ _ pflt (lt_n_Sn _)))) :=
  (fun i (pflt:i < n) =>
     f i (lt_trans _ _ _ pflt (lt_n_Sn _))).

Definition seg_fun_eq_dec_from_S 
           {U n}
           {f:seg_fun (S n) U}
           (pf:seg_fun_eq_dec f) : 
  seg_fun_eq_dec (seg_fun_from_S f) := 
  (fun i j (pfi:i < n) (pfj:j < n) =>
     pf i j
        (lt_S _ _ pfi)
        (lt_S _ _ pfj)).



Lemma seg_pred_dec_from_S_compat : 
  forall U i n (f:seg_fun (S n) U)
         (pfi:i < n) (pfi':i < S n)
         (pfs:seg_pred_dec (fun j (pfj:j < S n) => f i pfi' = f j pfj)),
    seg_pred_dec
      (fun (j : nat) (pfj : j < n) =>
         seg_fun_from_S f i pfi = seg_fun_from_S f j pfj).
  intros U i n f hi hi' h3.
  red in h3; red.
  intros j hj.
  unfold seg_fun_from_S. 
  specialize (h3 j (lt_S _ _ hj)); auto.
  destruct h3 as [h3 | h3].
  left.
  erewrite f_equal_dep.
  rewrite h3 at 1; auto.
  reflexivity.
  right.
  contradict h3.
  erewrite f_equal_dep.
  rewrite h3 at 1; auto.
  reflexivity.
Qed.


Lemma seg_fun_eq_dec_from_S_compat : 
  forall {U n}
         {f:seg_fun (S n) U},
         seg_fun_eq_dec f ->
         seg_fun_eq_dec (seg_fun_from_S f).
  intros n f pf h1.
  red in h1; red.
  intros i j hi hj.
  unfold seg_pred_type_from_S.
  specialize (h1 i j
                 (lt_S _ _  hi)
                 (lt_S _ _ hj)); auto.
Qed.


Lemma seg_pred_dec_from_S_compat' : 
  forall U i n (f:seg_fun (S n) U)
         (pfi:i < n),
  forall (pfi':i < S n) (pfs:seg_pred_dec (fun j (pfj:j < S n) => f i pfi' = f j pfj)),
    seg_pred_dec
      (fun (j : nat) (pflt : j < n) =>
         f i pfi' = f j (lt_S _ _ pflt)).
  intros U i n f hi hi' h3.
  red in h3; red.
  intros j hj.
  unfold seg_fun_from_S.  
  specialize (h3 j (lt_S _ _  hj)); auto .
Qed.


(*This section contains boolean analogues of many 
familiar arithmetical predicates.*)
Section NatAnalogues.

  Fixpoint beq_nat (a b:nat) : bool :=
  match a with
    | 0 =>
      match b with
        | 0 => true
        | S b' => false
      end
    | S a' =>
      match b with
        | 0 => false
        | S b' => beq_nat a' b'
      end
  end.

  Fixpoint lt_nat (a b:nat) : bool :=
    match a with
      | 0 =>
        match b with
          | 0 => false
          | S b' => true
        end
      | S a' =>
        match b with
          | 0 => false
          | S b' => lt_nat a' b'
        end
    end.

  Fixpoint gt_nat (a b:nat) : bool :=
    match a with
      | 0 =>
        false
      | S a' =>
        match b with
          | 0 => true
          | S b' => gt_nat a' b'
        end
    end.

  Fixpoint le_nat (a b:nat) : bool :=
    match a with
      | 0 => true
      | S a' =>
        match b with
          | 0 => false
          | S b' => le_nat a' b'
        end
    end.

  Fixpoint ge_nat (a b:nat) : bool :=
    match a with
      | 0 =>
        match b with
          | 0 => true
          | S b' => false
        end
      | S a' =>
        match b with
          | 0 => true
          | S b' => ge_nat a' b'
        end
    end.

  
  Lemma beq_nat_sym :
    forall a b,
      beq_nat a b = beq_nat b a.
    intro a; induction a; intro b; destruct b; simpl; auto.
  Qed.
  
  Lemma lt_nat_gt :
    forall a b,
      lt_nat a b = gt_nat b a.
    intro a; induction a; simpl; auto; intro b; destruct b; simpl; auto.
  Qed.

  
  Lemma le_nat_ge :
    forall a b,
      le_nat a b = ge_nat b a.
    intro a; induction a; simpl; auto; intro b; destruct b; simpl; auto.
  Qed.

  Lemma not_lt_nat :
    forall a b,
      negb (lt_nat a b) = orb (beq_nat a b) (gt_nat a b).
    intro a; induction a; simpl; auto; intro b; destruct b; simpl; auto.
  Qed.

  Lemma not_lt_nat' :
    forall a b,
      negb (lt_nat a b) = ge_nat a b.
    intro a; induction a; simpl; auto; intro b; destruct b; simpl; auto.
  Qed.


    
  Lemma not_gt_nat :
    forall a b,
      negb (gt_nat a b) = orb (beq_nat a b) (lt_nat a b).
    intro a; induction a; simpl; auto; intro b; destruct b; simpl; auto.
  Qed.

  Lemma not_gt_nat' :
    forall a b,
      negb (gt_nat a b) = le_nat a b.
    intro a; induction a; simpl; auto; intro b; destruct b; simpl; auto.
  Qed.

  
  
  Lemma lt_eq_lt_nat :
    forall a b,
      (orb (orb (lt_nat a b) (beq_nat a b)) (lt_nat b a)) = true.
    intro a; induction a; simpl; auto; intro b; destruct b; simpl; auto.
  Qed.

  
  Lemma beq_nat_compat :
    forall a b,
      beq_nat a b = true <->
      a = b.
    intro a; induction a; intro b; destruct b; simpl; auto; split; intro h1; try discriminate h1; auto.
    f_equal; auto.
    rewrite IHa in h1; auto.
    apply S_inj in h1; subst.
    rewrite IHa; auto.
  Qed.

 
  Lemma lt_nat_compat :
    forall a b,
      lt_nat a b = true <->
      a < b.
    intro a; induction a as [|a]; intro b; destruct b as [|b]; simpl;
    split; intro h1; try discriminate; try contradict (lt_n_0 _ h1);
    auto with arith.
    apply lt_n_S; auto.
    apply IHa; auto.
    apply lt_S_n in h1.
    rewrite <- IHa in h1; auto.
  Qed.
  
  Lemma gt_nat_compat :
    forall a b,
      gt_nat a b = true <->
      a > b.
    intros ? ?; split; intro h1.
    rewrite <- lt_nat_gt in h1; apply lt_gt; apply lt_nat_compat; auto.
    rewrite <- lt_nat_gt; apply lt_gt in h1; apply lt_nat_compat in h1; auto.
  Qed.

  Lemma le_nat_compat :
    forall a b,
      le_nat a b = true <->
      a <= b.
    intros a; induction a as [|a ih]; simpl.
    intuition.
    intro b; destruct b.
    intuition.
    split; intro h1.
    rewrite ih in h1; auto with arith.
    apply le_S_n in h1.
    rewrite <- ih in h1; auto.
  Qed.

  Lemma ge_nat_compat :
    forall a b,
      ge_nat a b = true <->
      a >= b.
    intros ? ?; split; intro h1.
    rewrite <- le_nat_ge in h1; apply le_ge; apply le_nat_compat; auto.
    rewrite <- le_nat_ge; apply le_ge in h1; apply le_nat_compat in h1; auto.
  Qed.
 

  
  Lemma beq_nat_same :
    forall r, beq_nat r r = true.
    intro r; induction r; auto.
  Qed.

  Lemma lt_nat_same :
    forall r, lt_nat r r = false.
    intro r; induction r; auto.
  Qed.
                             
  Lemma gt_nat_same :
    forall r, gt_nat r r = false.
    intro r; induction r; auto.
  Qed.

  Lemma le_nat_same :
    forall r, le_nat r r = true.
    intro r; induction r; auto.
  Qed.

  Lemma ge_nat_same :
    forall r, ge_nat r r = true.
    intro r; induction r; auto.
  Qed.

    
    
  Lemma neq_nat_compat :
    forall a b,
      beq_nat a b = false <->
      a <> b.
    intro a; induction a as [|a ih]; simpl; intro b; destruct b; split; intro h1; try discriminate; auto.
    contradict h1; auto.
    intro h2.
    apply S_inj in h2.
    rewrite ih in h1; contradiction.
    rewrite ih.
    auto with arith.
  Qed.
  
  
  
End NatAnalogues.



Lemma lt_decompose :
  forall (m n:nat),
    m < n ->
    exists p,
      n = m + S p.
  intros m n. revert m.
  induction n as [|a ih].
  intros ? h1. apply lt_n_0 in h1; contradiction.
  intros m h1.
  red in h1.
  apply le_S_n in h1.
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1 |]; subst.
  apply ih in h1.
  destruct h1 as [p]; subst.
  exists (S p); omega.
  exists 0; omega.
Qed.

Lemma minus_plus' :
  forall m n:nat,
    m < n ->
    m + (n - m) = n.
  intros; omega.
Qed.





Lemma mono_pred_le :
  forall (m n:nat),
    m <= n ->
    pred m <= pred n.
  intros; omega.
Qed.

Lemma mono_pred_ge :
  forall (m n:nat),
    m >= n ->
    pred m >= pred n.
  intros; omega.
Qed.



Lemma n_minus_pred_n : forall (n:nat), 0 < pred n -> n - pred n = 1.
intro n.
induction n as [|n h1].
simpl.
intro h1. contradict h1. auto with arith.
simpl.
intro h2.
destruct n. auto.
rewrite <- S_compat.
auto with arith.
Qed.



(*For use in [map_seq] in [ListUtilities].*)
Definition seg_S_tl {n:nat} {U:Type} (f:seg_fun (S n) U) :
  seq_fun 1 n U :=
  fun  (m:nat) (pf:1 <= m < 1 + n) => f m (ge1_lt m n pf).

Lemma lt_seq :
    forall (m n w:nat),
    m < w ->
    n <= n + m < n + w.
  auto with arith.
Qed.

Lemma seq_le :
  forall m n w,
    n <= n + m < n + w ->
    n <= n + m.
  intros ? ? ? h1; destruct h1; auto.
Qed.
    
Lemma seq_lt :
  forall m n w,
    n <= n + m < n + w ->
    n + m < n + w.
intros ? ? ? h1; destruct h1; auto.
Qed.
  
Lemma seq_lt' :
  forall m n w,
    n <= n + m < n + w ->
    n < n + w.
intros; omega.
Qed.
  

  
Lemma lt_seq_iff :
  forall (m n w:nat),
    m < w <->
    n <= n + m < n + w.
  intros; split; omega.
Qed.

Lemma lt_seq' :
  forall (m n w:nat),
    n <= m < n + w ->
    (n <= m /\ m - n < w).
  intros; omega.
Qed.
  
Lemma lt_seq_iff' :
  forall (m n w:nat),
    n <= m < n + w <->
    (n <= m /\ m - n < w).
  intros; split; omega.
Qed.

Lemma seq_S :
  forall (m n w:nat),
    n <= m < n + w ->
    n <= m < n + S w.
  intros; omega.
Qed.

Lemma seq_S_w :
  forall (n w:nat),
    n <= n + w < n + S w.
  auto with arith.
Qed.

Lemma le_lt_S :
  forall i n w,
    S n <= i < S n + pred w ->
    n <= i < n + w.
  intros; omega.
Qed.

Lemma one_le_lt :
  forall i n,
    1 <= i < 1 + pred n ->
    i < n.
  intros; omega.
Qed.

Lemma zero_le_lt :
  forall i m,
    0 <= i < 0 + m ->
    i < m.
  intros; omega.
Qed.

Lemma seg_fun_to_seq_compat :
  forall {m:nat} {U:Type} (f:seg_fun m U) (i:nat)
         (pf:0 <= i < 0 + m),
    seg_fun_to_seq f i pf = f i (zero_le_lt _ _ pf).
  intros m u f i h1; unfold seg_fun_to_seq; destruct h1; f_equal; apply proof_irrelevance.
Qed.


Definition seq_fun_from_S {U:Type} {n w:nat} (f:seq_fun n (S w) U) :
  seq_fun n w U :=
  fun m pf => f m (seq_S _ _ _ pf).

Definition seq_fun_to_seg  {U:Type} {n w:nat} (f:seq_fun n w U) :
  seg_fun w U :=
  fun m pf => f _ (lt_seq _ _ _  pf).


Definition seg_fun_rest_seq
           {m n:nat} {U:Type} (f:seg_fun n U) (pfm:m < n) :
  seq_fun m (n - m) U :=
  let pfe := minus_plus' m n pfm in
  (fun i pf => let pf' := proj2 pf in 
               let pflt := lt_eq_trans _ _ _ pf' pfe in
               f i pflt).

Definition seq_fun_tl
           {n w:nat} {U:Type}
           (f:seq_fun n w U) : seq_fun (S n) (pred w) U :=
  fun i pf => f i (le_lt_S _ _ _ pf).

Definition seg_fun_tl
           {n:nat} {U:Type}
           (f:seg_fun n U) : seq_fun 1 (pred n) U :=
  seq_fun_tl (seg_fun_to_seq f).

Lemma seq_fun_from_S_compat :
  forall {U:Type} {n w:nat} (f:seq_fun n (S w) U) i
         (pf:n <= i < n + w),
    seq_fun_from_S f i pf = f i (seq_S _ _ _ pf).
  unfold seq_fun_from_S; intros; reflexivity.
Qed.




Lemma seq_fun_tl_compat :
  forall {n w:nat} {U:Type} (f:seq_fun n w U)
         (i:nat) (pf:S n <= i < S n + pred w),
    seq_fun_tl f i pf = f i (le_lt_S _ _ _ pf).
  unfold seq_fun_tl; intros; f_equal.
Qed.


Lemma seg_fun_tl_compat :
  forall {n w:nat} {U:Type} (f:seg_fun n U)
         (i:nat) (pf:1 <= i < 1 + pred n),
    seg_fun_tl f i pf = f i (one_le_lt _ _ pf).
  unfold seg_fun_tl; intros; rewrite seq_fun_tl_compat, seg_fun_to_seq_compat; f_equal; apply proof_irrelevance.
Qed.


Lemma seq_fun_eq_seg :
  forall {U:Type} {n w:nat}  (f:seq_fun n w U),
    f = fun m pf =>
          let pf' := proj2 (lt_seq' _ _ _ pf) in
          (seq_fun_to_seg f) _ pf'.
  intros U n w. revert n.
  induction w as [|w]; simpl; auto.
  intros n f.
  apply functional_extensionality_dep.
  intro m.
  apply functional_extensionality.
  intros; omega.
  intros n f.
  apply functional_extensionality_dep.
  intro m.
  apply functional_extensionality.
  intros h1.
  unfold seq_fun_to_seg.
  apply f_equal_prop_dep.
  omega.
Qed.

(*[Not HalfOpen]*)
Lemma no_ho :
  forall (m n:nat),
    ~ m <= n < m + 0.
  intros; omega.
Qed.

Lemma O_lt_pred_n : forall (n:nat), 0 < pred n -> n > 1.
intros; induction n; omega.
Qed.

Lemma lt_pred_n :
  forall m n:nat,
    m < n -> pred n < n.
  intros; omega.
Qed.


Lemma O_lt_pred_n_lt_n : forall (n:nat), 0 < pred n -> pred n < n.
intros n h1.  induction n; omega.
Qed.


Lemma O_pred_n_eq_1 : forall n:nat, 0 < n -> 0 = pred n -> n = 1.
intro n.
induction n; simpl; auto; omega.
Qed.

Definition abs_diff (a b:nat) := max a b - min a b.  

Lemma lt_sum_iff :
  forall n a b, 
    n < a + b <->
    n - min a b < max a b.
  intros n a b.
  destruct (le_lt_dec a b) as [h2 | h2].
  rewrite min_l, max_r; omega.
  rewrite min_r, max_l; omega.
Qed.


Lemma min_comm : 
  forall (m n:nat),
    min m n = min n m.
intros m n.
destruct (le_lt_dec m n) as [h1 | h1].
rewrite (min_l m), (min_r n); auto.
rewrite (min_r m), (min_l n); auto with arith.
Qed.

Lemma max_comm : 
  forall (m n:nat),
    max m n = max n m.
  intros m n.
  destruct (le_lt_dec m n) as [h1 | h1].
  rewrite (max_r m), (max_l n); auto.
  rewrite (max_l m), (max_r n); auto with arith.
Qed.


Lemma min_compat_l :
  forall (m n:nat),
    min m n <= m.
  intros; destruct (le_lt_dec m n).
  rewrite min_l; auto.
  rewrite min_r; auto with arith.
Qed.

Lemma min_compat_r :
  forall (m n:nat),
    min m n <= n.
  intros; destruct (le_lt_dec m n). 
  rewrite min_l; auto.
  rewrite min_r; auto with arith.
Qed.


Lemma min_O : 
  forall (m:nat),
    min m O = O.
  intro; apply min_r; apply le_0_n.
Qed.

Lemma max_O : 
  forall (m:nat),
    max m O = m.
  intro; apply max_l; apply le_0_n.
Qed.

Lemma min_S :
  forall (m n:nat),
    min (S m) (S n) = S (min m n).
  intros; simpl; auto.
Qed.


Lemma le_min_iff :
  forall (m n a:nat),
    a <= min m n <-> (a <= m /\ a <= n).
intros m n a.
destruct (le_lt_dec m n).
rewrite min_l; omega.
rewrite min_r; omega.
Qed.


Lemma min_le_iff : 
  forall (m n a:nat),
    min m n <= a <-> (m <= a \/ n <= a).
intros m n a.
destruct (le_lt_dec m n) as [h1 | h1].
rewrite min_l; omega.
rewrite min_r; omega.
Qed.

Lemma le_max_iff : 
  forall (m n a:nat), 
    a <= max m n <-> (a <= m \/ a <= n).
intros m n a.
destruct (le_lt_dec m n) as [h1 | h1].
rewrite max_r; omega.
rewrite max_l; omega.
Qed.


Lemma max_le_iff : 
  forall (m n a:nat), 
    max m n <= a <-> (m <= a /\ n <= a).
intros m n a.
destruct (le_lt_dec m n) as [h1 | h1].
rewrite max_r; omega.
rewrite max_l; omega.
Qed.

Lemma min_max_compat : 
  forall (a b:nat),
    min (max a b) a = a.
intros.
destruct (le_lt_dec a b) as [h2 | h3].
rewrite max_r, min_r; auto.
rewrite max_l, min_l; auto with arith.
Qed.





Lemma O_lt_pow : forall (b n:nat), 0 < b -> 0 < b^n.
intros b n. revert b.
induction n as [|n h1].
simpl. auto with arith.
simpl. simpl in h1.
simpl. auto.
intro b.
simpl. specialize (h1 b).
induction b as [|b h2].
omega.
intro h3. specialize (h1 h3).
pose proof mult_lt_compat_r.
rewrite mult_comm.
assert (h4:0 = 0 * S b). ring.
rewrite h4.
apply mult_lt_compat_r.
assumption.
assumption.
Qed.

Lemma pow_mono : forall (b m n:nat), 0 < b -> m <= n -> b^m <= b^n.
intros b m n. revert b m.
induction n as [|n h1]. simpl.
intros b m h1 h2.
assert (h3:m = 0). auto with arith.
subst. simpl. auto.
intros b m. revert b.
rename h1 into hind.
induction m as [|m h1]. 
intros b h1 h2.
pose proof (O_lt_pow _ (S n) h1) as h3.
simpl. simpl in h3. omega.
intros b h2 h3.
apply le_S_n in h3.
specialize (hind _ _ h2 h3).
simpl.
pose proof (mult_le_compat_r (b ^ m) (b ^n) b hind) as h4.
rewrite (mult_comm b (b ^ n)).
rewrite mult_comm.
assumption.
Qed.


Lemma k_nat_ind : 
  forall (k:nat) (P:{n:nat | k <= n}->Prop),
    P (exist _ _ (le_refl _)) ->
    (forall n, P n -> P (exist _ _ (le_S _ _ (proj2_sig n)))) ->
    forall n, P n.
intro k.
induction k as [|k h1].
intros P h1 h2 n. destruct n as [n h3]. 
revert h3 h2.  revert h1. revert P.
induction n as [|n h4]. 
intros P h4 h5 h6.
pose proof (proof_irrelevance _ h5 (le_refl 0)); subst; auto.
intros P h1 h2 h3.
specialize (h4 P h1 (le_0_n _) h3).
specialize (h3 _ h4). simpl in h3.
assert (h5:h2 = le_S 0 n (le_0_n n)).  apply proof_irrelevance. subst.
assumption.
intros P h2 h3 n.
destruct n as [n h4].
pose (fun x:{n:nat | k <= n} => P (exist _ _ (le_n_S _ _ (proj2_sig x)))) as P'.
assert (h5: P' (exist (fun n : nat => k <= n) k (le_refl k))).
  unfold P'. simpl.
  assert (h6:exist _ _ (le_refl (S k)) = exist _ _ (le_n_S k k (le_refl k))).
    apply proj1_sig_injective; auto.
  rewrite <- h6 at 1. assumption.

assert (h6: forall n : {n : nat | k <= n},
        P' n ->
        P'
          (exist (fun n0 : nat => k <= n0) (S (proj1_sig n))
             (le_S k (proj1_sig n) (proj2_sig n)))).
  intros x h7. 
  specialize (h3 _ h7). simpl in h3. 
  unfold P'. simpl.
  assert (h8:exist _ _  (le_S (S k) (S (proj1_sig x))
               (le_n_S k (proj1_sig x) (proj2_sig x))) = 
             exist _ _ (le_n_S k (S (proj1_sig x)) (le_S k (proj1_sig x) (proj2_sig x)))).
    apply proj1_sig_injective; auto.
  rewrite <- h8 at 1.
  assumption.
pose proof (le_pred _ _ h4) as h4'. simpl in h4'.
specialize (h1 P' h5 h6 (exist _ _ h4')).
unfold P' in h1. simpl in h1.
assert (h7:0 < n). omega.
pose proof (S_pred _ _ h7) as h8.
assert (h9:exist _ _ h4 = exist _ _ (le_n_S k (pred n) h4')).
  apply proj1_sig_injective. simpl.
  assumption.
rewrite h9 at 1.
assumption.
Qed.


Lemma divmod_x0S0 :
  forall x q,
    divmod x 0 (S q) 0 =
    (S (fst (divmod x 0 q 0)), 0).
  intro x; induction x; simpl; auto.
Qed.

Lemma divmod_x000 :
  forall x,
    divmod x 0 0 0 =
    (x, 0).
  intro x. 
  induction x as [|x ih]; simpl; auto.
  rewrite divmod_x0S0.
  rewrite ih; simpl; auto.
Qed.

Lemma divmod_0y0u :
  forall y u,
    divmod 0 y 0 u =
    (0, u).
  intros  y u.
  revert y.
  induction u as [|u]; simpl; auto.
Qed.

Lemma divmod_S00S :
  forall x u,
    divmod (S x) 0 0 (S u) =
    divmod x 0 0 u.
  auto.
Qed.


Lemma divmod_SyS0 :
  forall x y q,
    divmod (S x) y (S q) 0 =
    divmod x y (S (S q)) y.
  auto.
Qed.

Lemma divmod_Sy00 :
  forall x y,
    divmod (S x) y 0 0 =
    divmod x y 1 y.
  auto.
Qed.

Lemma divmod_Sy0S :
  forall x y u,
    divmod (S x) y 0 (S u) =
    divmod x y 0 u.
  auto.
Qed.

Lemma divmod_Sx0x :
  forall x,
    divmod (S x) x 0 x = (1, x).
  intro x.
  pose proof (divmod_spec (S x) x 0 x (le_refl _)) as h1. 
  rewrite <- mult_n_O in h1.
  rewrite <- plus_n_O in h1.
  rewrite minus_same in h1.
  rewrite <- plus_n_O in h1.
  destruct (divmod (S x) x 0 x) as [f s].
  destruct h1 as [h1 h2].
  rewrite eq_times_plus_iff in h1.
  destruct h1 as [h1 | [h1 | h1]].
  destruct h1; discriminate.
  destruct h1; subst.
  rewrite <- le_minus_iff in H0.
  apply injective_projections; auto with arith.
  destruct h1 as [h1 h3]; subst.
  omega.
Qed.  

Lemma mod_functional :
  forall i j n,
    i = j ->
    i mod n = j mod n.
  intros i j n ?; subst; auto.
Qed.


Lemma mod_same :
  forall x,
    x mod x = 0.
  intro x.
  unfold modulo.
  destruct x as [|x]; auto.
  pose proof (divmod_spec (S x) x 0 x (le_refl _)) as h1.
  destruct (divmod (S x) x 0 x).
  rewrite minus_same in h1.
  rewrite mult_0_r in h1.
  do 2 rewrite plus_0_r in h1.
  destruct h1 as [h1 h2].
  rewrite eq_times_plus_iff in h1.
  destruct h1 as [h1 | [h1 | h1]].
  destruct h1; discriminate.
  destruct h1 as [? h1]; subst.
  simpl; auto.
  destruct h1 as [? h1]; subst; simpl.
  omega.
Qed.

Lemma one_mod :
  1 mod 1 = 0.
  simpl; auto.
Qed.

Lemma one_mod_S :
  forall n,
    n > 1 ->
    1 mod n = 1.
  intros n h1.
  destruct n as [|n]; simpl.
  apply lt_n_0 in h1; contradiction.
  destruct n as [|n].
  apply lt_irrefl in h1; contradiction.
  simpl. 
  destruct n as [|n]; auto.
  rewrite S_minus; auto.
Qed.


Lemma S_mod :
  forall x y,
    S x mod y =
    (x mod y + 1 mod y) mod y.
  intros x y.
  unfold modulo.
  destruct y as [|y]; auto.
  pose proof (divmod_spec (S x) y 0 y (le_refl _)) as h1.
  destruct (divmod (S x) y 0 y) as [q u].
  rewrite minus_same, mult_0_r in h1.
  do 2 rewrite plus_0_r in h1.
  destruct h1 as [h1 h2].
  pose proof (divmod_spec x y 0 y (le_refl _)) as h3.
  destruct (divmod x y 0 y) as [q' u'].
  rewrite minus_same, mult_0_r in h3.
  do 2 rewrite plus_0_r in h3.
  destruct h3 as [h3 h4].
  pose proof (divmod_spec 1 y 0 y (le_refl _)) as h5.
  destruct (divmod 1 y 0 y) as [r v].
  rewrite minus_same, mult_0_r in h5.
  do 2 rewrite plus_0_r in h5.
  destruct h5 as [h5 h6].
  pose proof (divmod_spec (y - snd (q', u') + (y - snd (r, v))) y 0 y (le_refl _)) as h7.
  destruct (divmod (y - snd (q', u') + (y - snd (r, v))) y 0 y) as [r' v'].
  simpl.
  simpl in h7.
  rewrite minus_same, mult_0_r in h7.
  do 2 rewrite plus_0_r in h7.
  destruct h7 as [h7 h8].
  f_equal.
  subst. 
  rewrite one_eq_S_times_plus_iff in h5.
  destruct h5 as [h5 | h5].
  destruct h5 as [? h5]; subst.
  apply le0 in h2; subst.
  apply le0 in h8; subst; auto.
  destruct h5 as [? h5]; subst.
  pose proof (f_equal (plus v) h5) as h9.
  rewrite <- le_plus_minus in h9; auto.
  subst.
  rewrite h5 in h7.
  rewrite <- (mult_1_l r') in h7 at 1.
  rewrite <- mult_plus_distr_r in h7.
  rewrite plus_comm in h7. 
  rewrite (plus_minus_assoc 1 (v+1) u') in h7; auto.
  rewrite minus_eq_times_plus_iff in h7.
  destruct h7 as [h7 | h7].
  destruct h7 as [h7 h9].
  rewrite S_compat, plus_S_shift_r in h7.
  discriminate. 
  destruct h7 as [h7 | h7].
  destruct h7 as [? h7]; subst.
  destruct h7 as [h7 | h7].
  destruct h7 as [h7].
  rewrite S_compat, plus_S_shift_r in h7.
  discriminate.
  destruct h7 as [h7 | h7].
  destruct h7 as [? h7]; subst.
  rewrite <- le_minus_iff in h7.
  pose proof (le_antisym _ _ h7 h8); subst.
  rewrite S_compat in h1.
  rewrite <- minus_n_O in h1.
  destruct u as [|u].
  rewrite <- minus_n_O in h1. 
  rewrite (plus_S_shift_r (S (S v) * q) v) in h1.
  apply S_inj in h1.
  rewrite times_plus_eq_times_plus_iff in h1.
  destruct h1 as [h1 | [h1 | h1]].
  destruct h1; omega.
  destruct h1 as [h1 h9].
  destruct h9 as [h9 h10].
  pose proof (lt_n_Sn (S v)) as h11.
  pose proof (mult_O_le (S (S v)) (q - q')) as h12.
  destruct h12 as [? | h12].
  omega. 
  rewrite mult_comm in h12.
  pose proof (le_plus_l (S (S v) * (q - q')) v) as h13.
  rewrite <- h10 in h13.
  pose proof (lt_le_trans _ _ _ h11 h12) as h14.
  pose proof (lt_le_trans _ _ _ h14 h13); omega.
  destruct h1 as [h1 h9].
  pose proof (lt_n_Sn v) as h11.
  pose proof (lt_n_Sn (S v)) as h11'.
  pose proof (mult_O_le (S (S v)) (q' - q)) as h12.
  destruct h12 as [? | h12].
  omega.
  rewrite mult_comm in h12.
  pose proof (le_plus_l (S (S v) * (q' - q)) (S v)) as h13.
  rewrite <- h9 in h13.
  pose proof (lt_trans _ _ _ h11 h11') as h11a.
  pose proof (lt_le_trans _ _ _ h11a h12) as h14.
  pose proof (lt_le_trans _ _ _ h14 h13); omega.
  rewrite S_minus_S in h1.
  rewrite <- plus_S_shift_r in h1.
  rewrite <- (mult_1_r (S (S v))) in h1 at 2.
  rewrite <- mult_plus_distr_l in h1.
  rewrite times_eq_times_plus_iff in h1.
  destruct h1 as [h1 | h1].
  destruct h1; omega.
  destruct h1 as [h1 [h10 he]].
  pose proof (lt_n_Sn v) as h11.
  pose proof (lt_n_Sn (S v)) as h11'.
  pose proof (lt_trans _ _ _ h11 h11') as h12.
  pose proof (mult_O_le (S (S v)) (q' + 1 - q)) as h13.
  destruct h13 as [? | h13].
  omega.
  rewrite mult_comm in h13.
  pose proof (lt_le_trans _ _ _ h12 h13) as h14.
  rewrite <- he in h14.
  omega.
  destruct h7 as [? h7]; subst.
  rewrite S_compat in h7.
  rewrite plus_comm, S_compat in h7.
  omega.
  destruct h7 as [h7 | h7].
  destruct h7 as [? h7]; subst.
  apply le_lt_eq_dec in h2.
  destruct h2 as [h2 | h2].
  rewrite S_compat in h1.
  pose proof (S_minus_shift v u) as h9.
  hfirst h9.
  red in h2.
  rewrite S_compat in h2.
  apply le_S_n in h2; auto.
  specialize (h9 hf).
  rewrite h9 in h1.
  rewrite plus_S_shift_r in h1.
  apply S_inj in h1.
  rewrite times_plus_eq_times_plus_iff in h1.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 h10].
  destruct h1; subst.
  discriminate.
  rewrite S_compat in h7. 
  rewrite <- plus_minus_assoc in h7; auto with arith.
  rewrite h10 in h7.
  rewrite plus_comm, S_compat in h7.
  rewrite <- S_minus_shift in h7.
  apply minus_inj in h7; auto.
  rewrite S_compat in h8; auto.
  assumption. 
  rewrite S_compat in h4; auto.
  rewrite S_compat in h7.
  rewrite <- plus_minus_assoc in h7.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 h10].
  destruct h10 as [h11 h12].
  rewrite h12 in h7.
  rewrite plus_comm in h7.
  rewrite <- plus_assoc in h7.
  rewrite S_compat in h7.
  pose proof (mult_O_le (S (S v)) (q - q')) as h13. 
  destruct h13 as [h13 | h13].
  rewrite lt_iff in h11.
  rewrite h13 in h11.
  apply lt_irrefl in h11; contradiction.
  rewrite mult_comm in h13. 
  pose proof (plus_le_compat_r _ _ (S (v - u)) h13) as h14.
  rewrite <- h7 in h14.
  pose proof (le_plus_l (S (S v)) (S (v - u))) as h15.
  pose proof (le_trans _ _ _ h15 h14) as h16.
  pose proof (le_minus (S v) v') as h17.
  pose proof (le_trans _ _ _ h16 h17) as h18.
  apply le_Sn_n in h18.
  contradiction.
  destruct h1 as [h1 h10].
  rewrite plus_comm in h7.
  rewrite <- minus_minus_assoc in h7.
  apply minus_inj in h7; subst.
  pose proof (le_minus v u) as h11.
  pose proof (mult_O_le (S (S v)) (q' - q)) as h12.
  destruct h12 as [h12 | h12].
  rewrite lt_iff in h1.
  rewrite h12 in h1.
  apply lt_irrefl in h1.
  contradiction.
  rewrite mult_comm in h12.
  pose proof (le_plus_r (S v - u') (S (S v) * (q' - q))) as h13.
  rewrite plus_comm in h13.
  rewrite <- h10 in h13.
  pose proof (lt_n_Sn v) as h14.
  pose proof (lt_n_Sn (S v)) as h15.
  pose proof (lt_trans _ _ _ h14 h15) as h16.
  pose proof (lt_le_trans _ _ _ h16 h12) as h17.
  pose proof (lt_le_trans _ _ _ h17 h13) as h18.
  pose proof (le_lt_trans _ _ _ h11 h18) as h19.
  apply lt_irrefl in h19.
  contradiction.
  rewrite S_compat in h8; auto.
  rewrite S_compat in h4.
  eapply le_trans.
  apply le_minus.
  assumption.
  rewrite S_compat in h4.
  eapply le_trans.
  apply le_minus.
  assumption.
  destruct (zerop u') as [|h11]; subst.
  rewrite <- minus_n_O in h7.
  pose proof (le_minus (S v) v') as h11.
  rewrite S_compat in h7.
  pose proof (lt_n_Sn (S v)) as h12.
  pose proof (le_lt_trans _ _ _ h11 h12) as h13.
  rewrite h7 in h13.
  apply lt_irrefl in h13.
  contradiction.
  red in h11; auto.
  rewrite S_compat in h4; auto.
  rewrite S_compat in h4; auto.
  subst.
  rewrite <- minus_n_n, plus_0_r in h1.
  destruct (zerop q) as [h9 | h9]; subst.
  rewrite mult_0_r in h1.
  discriminate.
  pose proof (le_times_S (S (v+1)) (pred q)) as h10.
  rewrite <- (S_pred _ _ h9) in h10.
  rewrite <- h1 in h10.
  destruct (zerop q') as [h11 | h11]; subst.
  rewrite mult_0_r in h1.
  rewrite plus_0_l in h1.
  rewrite <- S_minus_shift in h1.
  pose proof (le_minus (S (v+1)) u') as h11.
  rewrite h1 in h11.
  pose proof (le_times_S (S (v + 1)) (pred q)) as h12.
  rewrite <- (S_pred _ _ h9) in h12.
  pose proof (le_antisym _ _ h11 h12) as h13.
  rewrite <- mult_1_r in h13.
  apply mult_inj in h13; subst.
  rewrite mult_1_r in h1.
  symmetry in h1.
  rewrite minus_eq_iff in h1.
  destruct h1; subst.
  discriminate.
  rewrite <- minus_n_O in h7.
  rewrite S_compat in h7.
  rewrite plus_comm in h7.
  rewrite S_compat in h7.
  pose proof (le_minus (S v) v') as h13.
  rewrite h7 in h13.
  apply le_Sn_n in h13.
  contradiction.
  auto with arith.
  assumption.
  rewrite <- plus_S_shift_r in h1.
  symmetry in h1.
  rewrite times_eq_times_plus_iff in h1.
  destruct h1 as [h1 | h1].
  destruct h1; discriminate.
  destruct h1 as [h1 h2].
  destruct h2 as [h2 h3].
  symmetry in h3.
  rewrite <- S_minus_shift in h3. 
  pose proof (le_minus (S (v + 1)) u') as h12.
  rewrite <- h3 in h12.
  pose proof (mult_O_le (S (v + 1)) (q - q')) as h13.
  destruct h13 as [h13 | h13].
  apply gt_lt in h2.
  rewrite lt_iff in h2.
  rewrite h13 in h2.
  apply lt_irrefl in h2; contradiction.
  rewrite mult_comm in h13.
  pose proof (le_antisym _ _ h12 h13) as h14.
  rewrite <- mult_1_r in h14.
  apply mult_inj in h14.
  pose proof (f_equal (plus q') h14) as h15.
  rewrite <- le_plus_minus in h15.
  subst. 
  rewrite minus_plus, mult_1_r in h3. 
  rewrite minus_eq_iff in h3.
  destruct h3; subst.
  discriminate.
  rewrite <- minus_n_O in h7.
  rewrite S_compat in h7.
  pose proof (le_minus (S v) v') as h15.
  rewrite h7 in h15.
  rewrite plus_comm, S_compat in h15.
  apply le_Sn_n in h15.
  contradiction.
  auto with arith.
  auto with arith.
  assumption.
  destruct h7 as [h7 h10].
  pose proof (f_equal (plus v') h7) as h11.
  rewrite plus_minus_assoc in h11; auto.
  rewrite minus_plus, plus_0_r in h11.
  subst.
  destruct h10 as [h10 | h10].
  rewrite plus_comm, S_compat in h10.
  discriminate.
  destruct h10 as [h10 | h10].
  destruct h10 as [h10 h11]; subst.
  destruct h11 as [h11 | h11].
  rewrite plus_comm, S_compat in h11.
  discriminate.
  destruct h11 as [h11 | h11]; subst.
  destruct h11 as [h11].
  rewrite plus_comm, S_compat in h11.
  discriminate.   
  rewrite <- minus_n_O, S_compat in h1.
  rewrite plus_minus_assoc in h1.
  rewrite  (plus_S_shift_r (S (S v) * q) v) in h1.
  rewrite S_minus_shift in h1. 
  apply S_inj in h1.
  pose proof (f_equal (plus u) h1) as h9.
  rewrite plus_comm in h9.
  rewrite <- plus_assoc in h9.
  rewrite plus_minus_assoc in h9.
  rewrite minus_plus in h9.  
  rewrite times_plus_eq_times_plus_iff in h9. 
  destruct h9 as [h9 | h9].
  destruct h9 as [? h9].
  pose proof (le_plus_l (S v) u) as h10.
  rewrite h9 in h10.
  apply le_Sn_n in h10.
  contradiction.   
  destruct h9 as [h9 | h9].
  destruct h9 as [h9 h10].
  destruct h10 as [h10 h11].
  destruct u as [|u].
  rewrite plus_0_r in h11. 
  pose proof (mult_O_le (S (S v)) (q - q')) as h12.
  destruct h12 as [h12 | h12].
  rewrite lt_iff in h10.
  rewrite h12 in h10.
  apply lt_n_0 in h10.
  contradiction.
  rewrite mult_comm in h12.
  pose proof (le_plus_l (S (S v) * (q - q')) v) as h13.
  pose proof (le_trans _ _ _ h12 h13) as h14.
  rewrite <- h11 in h14.
  apply le_Sn_n in h14.
  contradiction.   
  rewrite <- plus_Snm_nSm in h11.
  rewrite <-  (mult_1_r (S (S v))) in h11 at 1.
  rewrite times_plus_eq_times_plus_iff in h11.
  destruct h11 as [h11 | h11].  
  destruct h11 as [h11].
  destruct h11 as [|h11].
  discriminate.
  subst.
  rewrite S_compat; auto. 
  destruct h11 as [h11 | h11].
  destruct h11 as [h11 h12]. 
  destruct h12 as [h12 h13]. 
  rewrite h13 in h1. 
  rewrite S_compat in h2.
  apply le_S_n in h2.
  rewrite h13 in h2.
  pose proof h12 as h12'.
  rewrite lt_iff in h12'.
  
  assert (hlt : S (S v) * (q - q' -1) > 0).
    rewrite (S_pred _ _ h11), (S_pred _ _ h12').
    rewrite S_times_S.
    auto with arith.
  pose proof (lt_plus_S v (pred (S (S v) * (q - q' - 1)))) as h14.
  rewrite <- (S_pred _ _ hlt) in h14.
  rewrite plus_comm in h14.
  pose proof (lt_le_trans _ _ _ h14 h2) as h15.
  apply lt_irrefl in h15.
  contradiction.
  destruct h11 as [h11 h12]. 
  red in h11. 
  apply le_S_n in h11.
  apply le0 in h11.
  rewrite <- le_minus_iff in h11.
  pose proof (lt_le_trans _ _ _ h10 h11) as h13.
  apply lt_irrefl in h13; contradiction.
  destruct h9 as [h9 h10].
  pose proof (lt_n_Sn v) as h11.
  pose proof (le_plus_l (S v) u) as h12.
  pose proof (lt_le_trans _ _ _ h11 h12) as h13.
  pose proof (le_plus_l (S v + u) (S (S v) * (q' - q))) as h14.
  pose proof (lt_le_trans _ _ _ h13 h14) as h15.
  rewrite plus_comm in h15.
  rewrite <- h10 in h15.  
  apply lt_irrefl in h15; contradiction.

  rewrite S_compat in h2.
  destruct q as [|q].
  rewrite mult_0_r in h9, h1.
  rewrite plus_0_l in h1, h9.
  pose proof (le_minus v u) as h10.
  pose proof (lt_n_Sn v) as h11.
  pose proof (le_lt_trans _ _ _ h10 h11) as h12.
  pose proof (le_plus_l (S v) (S (S v) * q')) as h13.
  pose proof (lt_le_trans _ _ _ h12 h13) as h14.
  rewrite plus_comm in h14.
  rewrite h1 in h14.
  apply lt_irrefl in h14; contradiction.

  pose proof (mult_O_le (S (S v)) (S q)) as h12. 
  pose proof (lt_n_Sn (S v)) as h10.
  pose proof (le_lt_trans _ _ _ h2 h10) as h11.
 
  destruct h12 as [|h12]; subst.
  discriminate.
  pose proof (lt_le_trans _ _ _ h11 h12) as h13.
  rewrite mult_comm in h13.
  pose proof (le_plus_l (S (S v) * S q) v) as h14.
  pose proof (lt_le_trans _ _ _ h13 h14); auto with arith.
 
  rewrite S_compat in h2. 
  destruct q as [|q].
  rewrite mult_0_r in h1.
  rewrite plus_0_l in h1.  
  pose proof (le_minus (S v) u) as h10.
  pose proof (le_plus_l (S v) (S (S v) * q')) as h11.
  pose proof (le_trans _ _ _ h10 h11) as h12.
  rewrite plus_comm in h12.
  pose proof (lt_n_Sn (S (S v) * q' + S v)) as h13.
  pose proof (le_lt_trans _ _ _ h12 h13) as h14.
  rewrite h1 in h14.
  apply lt_irrefl in h14.
  contradiction.
  pose proof (lt_n_Sn (S v)) as h9.
  pose proof (le_lt_trans _ _ _ h2 h9) as h10.
  pose proof (mult_O_le (S (S v)) (S q)) as h11.
  destruct h11 as [|h11].
  discriminate.
  pose proof (le_plus_l (S q * S (S v)) v) as h12.
  pose proof (le_trans _ _ _ h11 h12) as h13.
  rewrite mult_comm in h13.
  pose proof (lt_le_trans _ _ _ h10 h13); auto with arith.
  rewrite S_compat in h2; auto.
  destruct h10 as [? h10]; subst.
  rewrite plus_comm in h10.
  rewrite S_compat in h10.
  pose proof (le_trans _ _ _ h10 h4) as h11.
  apply le_Sn_n in h11.
  contradiction.
Qed.


Lemma gt_one_mod :
  forall n,
    n > 1 ->
  1 mod n = 1.
  intros; rewrite S_mod, O_mod, plus_0_l, lt_mod_eq, lt_mod_eq; auto.
  apply lt_mod; auto with arith.
Qed.

Lemma plus_times_mod_r :
  forall i j n,
    (i + n * j) mod n = i mod n.
  intros i j n.
  unfold modulo.
  destruct n as [|n]; auto.
  pose proof (divmod_spec (i + S n * j) n 0 n (le_refl _)) as h1.
  pose proof (divmod_spec i n 0 n (le_refl _)) as h2.
  destruct (divmod (i + S n * j) n 0 n) as [q u],
                                           (divmod i n 0 n) as [q' u'].
  rewrite mult_0_r, minus_same, plus_0_r, plus_0_r in h1, h2.
  simpl.
  destruct h1 as [h1 h3], h2 as [h2 h4].
  subst.
  rewrite <- plus_assoc in h1.
  rewrite (plus_comm (n - u')) in h1.
  rewrite plus_assoc in h1.
  rewrite <- mult_plus_distr_l in h1.
  rewrite times_plus_eq_times_plus_iff in h1.
  destruct h1 as [h1 | h1].
  destruct h1; auto.
  destruct h1 as [h1 | h1].  
  destruct h1 as [h1 [h2 h5]].
  pose proof (le_minus n u') as h6.
  rewrite h5 in h6.
  pose proof (mult_O_le (S n) (q - (q' + j))) as h8.
  destruct h8 as [h8 | h8].
  rewrite <- le_minus_iff in h8.
  pose proof (le_lt_trans _ _ _ h8 h2) as h9.
  apply lt_irrefl in h9; contradiction.
  rewrite mult_comm in h8.
  pose proof (le_plus_l (S n * (q - (q' + j))) (n - u)) as h9.
  pose proof (le_trans _ _ _ h9 h6) as h10.
  pose proof (le_trans _ _ _ h8 h10) as h11.
  apply le_Sn_n in h11.
  contradiction.
  destruct h1 as [h1 h2].
  pose proof (mult_O_le (S n) (q' + j - q)) as h5.
  destruct h5 as [h5 | h5].
  rewrite <- le_minus_iff in h5.
  pose proof (lt_le_trans _ _ _ h1 h5) as h6.
  apply lt_irrefl in h6.
  contradiction.
  rewrite mult_comm in h5.
  pose proof (le_plus_l (S n * (q' + j - q)) (n - u')) as h6.
  pose proof (le_trans _ _ _ h5 h6) as h7.
  rewrite <- h2 in h7.
  pose proof (le_minus n u) as h8.
  pose proof (le_trans _ _ _ h7 h8) as h9.
  apply le_Sn_n in h9.
  contradiction.
Qed.

Lemma plus_times_mod_  :
  forall i j n,
    (n * j + i) mod n = i mod n.
  intros; rewrite plus_comm; apply plus_times_mod_r.
Qed.

Lemma minus_times_mod_r :
  forall i j n,
    i >= n * j ->
    (i - n * j) mod n = i mod n.
  intros i j n h1.
  unfold modulo.
  destruct n as [|n]; auto. 
  pose proof (divmod_spec (i - S n * j) n 0 n (le_refl _)) as h2.
  pose proof (divmod_spec i n 0 n (le_refl _)) as h3.
  destruct (divmod (i - S n * j) n 0 n) as [u v],
                                           (divmod i n 0 n) as [u' v'].
  simpl.  
  rewrite mult_0_r, minus_same, plus_0_r, plus_0_r in h2, h3.
  destruct h2 as [h2 h4], h3 as [h3 h5]; subst.
  pose proof (f_equal (fun c => c + (S n * j)) h2) as h3.
  simpl in h3.
  do 3 rewrite  <- S_times_l' in h3. 
  rewrite minus_plus_same, <- plus_assoc, (plus_comm _ (S n * j)), 
  plus_assoc, <- mult_plus_distr_l in h3; auto. 
  rewrite times_plus_eq_times_plus_iff in h3.
  destruct h3 as [h3 | h3].
  destruct h3 as [h3 h6].
  destruct h3; subst.
  discriminate.   
  rewrite mult_plus_distr_l, <- plus_assoc, plus_comm,
  plus_plus_plus_minus_same, (plus_comm _ (n - v)) in h2.
  apply plus_inj_r in h2; auto.
  destruct h3 as [h3 | h3].
  destruct h3 as [h3 h6].
  destruct h6 as [h6 h7].
  rewrite lt_iff in h6.
  pose proof (le_minus n v') as h8.
  pose proof (le_lt_trans _ _ _ h8 (lt_n_Sn _)) as h9.
  pose proof (mult_O_le (S n) (u + j - u')) as h10.
  destruct h10 as [h10 | h10].
  rewrite h10 in h6. 
  apply lt_irrefl in h6; contradiction.   
  rewrite mult_comm in h10.
  pose proof (le_plus_l (S n * (u + j - u')) (n - v)) as h11.
  pose proof (le_trans _ _ _ h10 h11) as h12.
  rewrite <- h7 in h12.
  pose proof (le_lt_trans _ _ _ h12 h9) as h13.
  apply lt_irrefl in h13; contradiction.
  destruct h3 as [h3 h6].
  rewrite plus_comm in h2.
  pose proof (mult_O_le (S n) (u' - (u + j))) as h7.
  destruct h7 as [h7 | h7].
  rewrite <- le_minus_iff in h7.
  pose proof (le_lt_trans _ _ _ h7 h3) as h8.
  apply lt_irrefl in h8; contradiction.
  pose proof (le_plus_l (S n * (u' - (u + j))) (n - v')) as h8.
  rewrite <- h6, mult_comm in h8.
  pose proof (le_trans _ _ _ h7 h8) as h9.
  pose proof (le_minus n v) as h10.
  pose proof (le_trans _ _ _ h9 h10) as h11.
  apply le_Sn_n in h11.
  contradiction.
Qed.

Lemma mod_eq_iff :
  forall i j n,
    n > 0 ->
    i < j ->
    (i mod n = j mod n <->
     (n dv (j - i))).
  intros i j n h1 hlt. 
  pose proof h1 as h1'.
  apply gt_lt in h1'.
  rewrite <- neq0_iff in h1'.
  split; intro h2.
  pose proof (div_mod i n h1') as h3.
  pose proof (div_mod j n h1') as h4.
  rewrite h2 in h3. 
  rewrite h3.
  rewrite h4 at 1.  
  rewrite (plus_comm (n*(i/n)) (j mod n)). 
  rewrite minus_plus_assoc. 
  rewrite <- plus_minus_assoc.
  rewrite <- minus_n_n, plus_0_r.
  rewrite mult_comm, (mult_comm n).
  rewrite minus_dist_r.
  rewrite mult_comm.
  constructor.
  auto with arith.
  rewrite plus_comm.
  apply plus_le_compat_r.
  apply mult_le_compat_l.
  apply le_div; auto with arith. 
  inversion h2. 
  pose proof (div_mod i n h1') as h3.
  pose proof (div_mod j n h1') as h4.
  rewrite h3, h4 in H0.
  destruct (le_lt_dec (i mod n) (j mod n)) as [h5 | h5].
  rewrite plus_comm in H0. 
  rewrite minus_plus_assoc in H0.
  rewrite <- plus_minus_assoc in H0.
  rewrite minus_dist_l in H0.
  rewrite plus_comm in H0.
  rewrite times_eq_times_plus_minus_iff in H0.
  destruct H0 as [h6 | h6].
  destruct h6 as [h6].
  rewrite h6 in h1.
  apply lt_irrefl in h1; contradiction.
  destruct h6 as [h6 | h6].
  destruct h6; auto.
  destruct h6 as [h6 | h6].
  destruct h6 as [h6 h7]. 
  pose proof (lt_mod i n h1) as h9. 
  pose proof (le_lt_trans _ _ _ h7 h9) as h10.
  pose proof (le_plus_l (n * (j/n - i/n)) (j mod n)) as h11.
  pose proof (le_lt_trans _ _ _ h11 h10) as h12.
  rewrite <- (mult_1_r n) in h12 at 4.
  rewrite (S_pred _ _ h1) in h12.
  apply mult_S_lt_l in h12.
  red in h12.
  apply le_S_n in h12.
  apply le0 in h12.
  rewrite <- (S_pred _ _ h1) in h12.
  pose proof (minus_div_le j i n) as h13.
  rewrite h12 in h13.
  apply le0 in h13.
  rewrite div_eq_0_iff in h13.
  destruct h13 as [h13 | h13].
  rewrite h13 in h1.
  apply lt_irrefl in h1; contradiction.
  pose proof (dv_le _ _ h2) as h14.
  destruct h14 as [h14 | h14].
  rewrite <- le_minus_iff in h14.
  pose proof (lt_le_trans _ _ _ hlt h14) as h15.
  apply lt_irrefl in h15; contradiction.
  pose proof (lt_le_trans _ _ _ h13 h14) as h15.
  apply lt_irrefl in h15; contradiction.
  destruct h6 as [h6 [h7 h8]].
  destruct h8 as [h8 | h8].
  destruct h8 as [h8 [h9 h10]].
  rewrite h10 in h4.
  rewrite plus_assoc in h4.
  rewrite <- mult_plus_distr_l in h4.
  rewrite plus_minus_minus_same_le in h4; auto with arith.
  rewrite mult_plus_distr_l in h4.
  rewrite <- plus_assoc in h4.
  rewrite <- h3 in h4.
  rewrite h4. 
  rewrite plus_comm.
  rewrite plus_times_mod_r; auto.
  apply le_div; auto with arith.
  destruct h8 as [h8 h9].
  destruct h9 as [h9 h10].
  apply le_antisym; auto.
  apply mult_le_compat_l.
  apply le_div; auto with arith.
  rewrite plus_comm.
  apply plus_le_compat; auto.
  apply mult_le_compat_l.
  apply le_div; auto with arith.
  rewrite plus_minus_plus_diff_le_lt' in H0.
  rewrite minus_dist_l in H0.
  rewrite times_eq_times_minus_iff in H0.
  destruct H0 as [h6 | h6].
  rewrite h6 in h1.
  apply lt_irrefl in h1; contradiction.
  destruct h6 as [h6 | h6].
  destruct h6 as [h6 h7]. 
  destruct (zerop (j / n - i / n)) as [h8 | h8].
  pose proof (minus_div_le j i n) as h9.
  rewrite h8 in h9.
  apply le0 in h9.
  rewrite div_eq_0_iff in h9.
  destruct h9 as [h9 | h9].
  rewrite h9 in h1.
  apply lt_irrefl in h1.
  contradiction. 
  pose proof (dv_le _ _ h2) as h10.
  destruct h10 as [h10 | h10].
  rewrite <- le_minus_iff in h10.
  pose proof (lt_le_trans _ _ _ hlt h10) as h11.
  apply lt_irrefl in h11; contradiction.
  pose proof (lt_le_trans _ _ _ h9 h10) as h11.
  apply lt_irrefl in h11; contradiction.
  pose proof (mult_O_le n (j / n - i / n)) as h9.
  destruct h9 as [h9 | h9].
  rewrite h9 in h8. 
  apply lt_irrefl in h8; contradiction.
  rewrite mult_comm in h9.
  pose proof (lt_mod i n h1) as h10.
  pose proof (le_minus (i mod n) (j mod n)) as h11.
  pose proof (lt_le_trans _ _ _ h10 h9) as h12.
  pose proof (le_lt_trans _ _ _ h11 h12) as h13.
  pose proof (lt_le_trans _ _ _ h13 h7) as h14. 
  apply lt_irrefl in h14; contradiction. 
  destruct h6 as [h6 h7].
  destruct h6 as [? h6].
  destruct h7 as [h7 | h7].
  destruct h7 as [h7 h8].
  rewrite <- le_minus_iff in h8. 
  pose proof (lt_le_trans _ _ _ h5 h8) as h9. 
  apply lt_irrefl in h9; contradiction. 
  destruct h7 as [h7 h8]. 
  pose proof (mult_O_le n (j / n - i / n - n0)) as h9.
  destruct h9 as [h9 | h9]. 
  rewrite lt_iff in h7. 
  rewrite h9 in h7. 
  apply lt_irrefl in h7. 
  contradiction. 
  rewrite mult_comm in h9.
  rewrite <- h8 in h9.
  pose proof (lt_mod i n h1) as h10.
  pose proof (le_minus (i mod n) (j mod n)) as h11.
  pose proof (le_lt_trans _ _ _ h9 (le_lt_trans _ _ _ h11 h10)) as h12.
  apply lt_irrefl in h12; contradiction.
  apply mult_le_compat_l.
  apply le_div; auto with arith.
  assumption.
Qed.


Lemma plus_mod :
  forall i j n,
    (i + j) mod n =
    (i mod n + j mod n) mod n.
  intros i j n.
  destruct (zerop n) as [|h1]; subst.
  simpl; auto.
  pose proof h1 as h1'.
  rewrite <- neq0_iff in h1.
  pose proof (le_mod i n) as hli.
  pose proof (le_mod j n) as hlj.
  pose proof (plus_le_compat _ _ _ _ hli hlj) as hij.
  apply le_lt_eq_dec in hij.
  destruct hij as [hij | hij].
  pose proof (div_mod i n h1) as hi.
  pose proof (div_mod j n h1) as hj.
  pose proof (f_equal (fun c => c - i mod n) hi) as h4.
  pose proof (f_equal (fun c => c - j mod n) hj) as h5.
  simpl in h4, h5.
  rewrite plus_minus_same_l in h4, h5.
  symmetry.
  rewrite mod_eq_iff. 
  rewrite plus_minus_plus_diff_le, h4, h5.
  rewrite <- mult_plus_distr_l.
  constructor.
  apply le_mod.
  apply le_mod. 
  assumption.
  assumption.
  rewrite hij; auto.
Qed.

Lemma minus_mod_le :
  forall i j n,
    j <= i ->
    j mod n <= i mod n ->
    (i - j) mod n =
    i mod n - j mod n.
  intros i j n hij hijn.
  destruct (zerop n) as [|h1]; subst; simpl; auto.
  pose proof h1 as h1' .
  rewrite <- neq0_iff in h1.
  pose proof (div_mod i n h1) as h2.
  pose proof (div_mod j n h1) as h3.
  symmetry.
  pose proof (f_equal (fun c => c - (n * (i / n))) h2) as h7.
  simpl in h7.
  rewrite plus_minus_same_r in h7.
  pose proof (f_equal (fun c => c - (n * (j / n))) h3) as h8.
  simpl in h8.
  rewrite plus_minus_same_r in h8.
  rewrite <- (lt_mod_eq (i mod n - j mod n) n). 
  rewrite <- h7, <- h8, minus_minus_minus_diff_le'.
  rewrite minus_dist_l.
  apply minus_times_mod_r. 
  rewrite h2, h3 at 1.
  rewrite plus_minus_plus_diff_le.
  rewrite minus_dist_l.
  apply le_plus_l.
  apply mult_le_compat_l.
  apply le_div; auto.
  assumption.
  assumption.
  apply mult_le_compat_l.
  apply le_div; auto.
  rewrite h2 at 2.
  apply le_plus_l.
  rewrite h3 at 2.
  apply le_plus_l.
  rewrite h8, h7; auto.
  eapply le_lt_trans.
  apply le_minus.
  apply lt_mod; auto.
Qed.


Lemma minus_mod_lt :
  forall i j n,
    j <= i ->
    i mod n < j mod n ->
    (i - j) mod n =
    n - (j mod n - i mod n).
  intros i j n h1 h2.  
  destruct (zerop n) as [|h3]; subst; simpl; auto.
  pose proof h3 as h3'.
  rewrite <- neq0_iff in h3'.
  pose proof (div_mod (i - j) n h3') as h4.
  pose proof (div_mod i n h3') as h5.
  pose proof (div_mod j n h3') as h6. 
  rewrite h5, h6 in h4 at 1.
  rewrite plus_minus_plus_diff_le' in h4; auto with arith.
  rewrite minus_dist_l in h4. 
  symmetry in h4.
  pose proof (f_equal (fun c => c + (j mod n - i mod n)) h4) as h7.
  simpl in h7.
  pose proof (le_div j i n h1) as h8.
  apply le_lt_eq_dec in h8.  
  destruct h8 as [h8 | h8].
  rewrite minus_plus_same in h7.
  rewrite <- plus_assoc in h7.
  symmetry in h7.
  rewrite times_eq_times_plus_iff in h7.
  destruct h7 as [h7 | h7].
  destruct h7 as [h7 h9].
  destruct h7 as [h7 | h7].
  rewrite h7 in h3'; contradict h3'; auto. 
  apply plus_is_O in h9.
  destruct h9 as [h9 h10].
  rewrite <- le_minus_iff in h10.
  pose proof (lt_le_trans _ _ _ h2 h10) as h11. 
  apply lt_irrefl in h11; contradiction.
  destruct h7 as [h7 [h9 h10]].
  pose proof (lt_mod (i - j) n h3) as h11.
  pose proof (le_minus (j mod n) (i mod n)) as h12.
  pose proof (lt_mod j n h3) as h13.
  pose proof (le_lt_trans _ _ _ h12 h13) as h14.
  pose proof (plus_lt_compat _ _ _ _ h11 h14) as h15.
  rewrite <- mult2 in h15.
  rewrite h10 in h15. 
  rewrite (mult_comm 2) in h15.
  rewrite (S_pred _ _ h3) in h15 at 1 5.
  apply mult_S_lt_l in h15.
  apply gt_lt in h9.
  rewrite lt_iff in h9.
  rewrite (S_pred _ _ h9) in h15.
  apply lt_S_n in h15.
  apply lt1 in h15.
  apply pred0_dec in h15.
  destruct h15 as [h15 | h15].
  rewrite h15 in h9.
  apply lt_irrefl in h9; contradiction.
  rewrite h15 in h10.
  rewrite mult_1_r in h10.
  rewrite <- h10 at 2.
  rewrite plus_minus_same_l; auto.
  rewrite lt_iff in h8.
  pose proof (lt_mod j n h3) as h9.
  pose proof (le_minus (j mod n) (i mod n)) as h10.
  pose proof (le_lt_trans _ _ _ h10 h9) as h11.
  pose proof (mult_O_le n (i / n - j / n)) as h12.
  destruct h12 as [h12 | h13].
  rewrite h12 in h8.
  apply lt_irrefl in h8.
  contradiction.
  apply lt_le_weak.
  eapply lt_le_trans.
  apply h11.
  rewrite mult_comm; auto.
  rewrite h8 in h6.
  pose proof (plus_lt_compat_l _ _ (n * (i / n)) h2) as h9.
  rewrite <- h5, <- h6 in h9.
  pose proof (lt_le_trans _ _ _ h9 h1) as h10.
  apply lt_irrefl in h10.
  contradiction.
  apply mult_le_compat_l.
  apply le_div; auto.
Qed.

  
Lemma mod_plus_mod_l :
  forall i j n,
    ((i mod n) + j) mod n =
    (i + j) mod n.
  intros i j n.
  destruct (zerop n) as [|hn]; subst; simpl; auto.
  pose proof (le_mod i n) as hl.
  apply le_lt_eq_dec in hl.
  destruct hl as [hl|he].
  pose proof hn as hn'.
  rewrite <- neq0_iff in hn.
  rewrite mod_eq_iff, plus_minus_plus_same_le'.
  pose proof (div_mod i n hn) as h1.
  rewrite h1 at 1.
  rewrite plus_minus_same_l.
  constructor.
  apply le_mod.
  assumption.
  apply plus_lt_compat_r; auto.
  rewrite he; auto.
Qed.

Lemma mod_plus_mod_r :
  forall i j n,
    (j + (i mod n)) mod n =
    (j + i) mod n.
  intros; rewrite plus_comm, mod_plus_mod_l, plus_comm; auto.
Qed.

Lemma mod_mod :
  forall i n,
    (i mod n) mod n = i mod n.
  intros i n.
  destruct (zerop n) as [|h1]; subst; simpl; auto.
  intros; apply lt_mod_eq.
  apply lt_mod; auto.
Qed.


Lemma plus_mod_inj_r :
  forall i j k n
         (pfn : 0 < n),
    (i + k) mod n =
    (j + k) mod n ->
    i mod n =
    j mod n.
  intros i j k n h1 h2.
  destruct (lt_eq_lt_dec i j) as [h3 | h3].
  destruct h3 as [h3 | h3].
  rewrite mod_eq_iff in h2; auto.
  rewrite plus_minus_plus_diff_lt_le' in h2; auto with arith.
  rewrite minus_same, <- minus_n_O in h2.
  rewrite mod_eq_iff; auto.
  omega.
  subst; auto.
  symmetry.
  symmetry in h2.
  rewrite mod_eq_iff in h2.
  rewrite plus_minus_plus_diff_lt_le' in h2; auto with arith.
  rewrite minus_same, <- minus_n_O in h2.
  rewrite mod_eq_iff; auto.
  assumption.
  omega.
Qed.

Lemma plus_mod_inj_l :
  forall i j k n
         (pfn : 0 < n),
    (k + i) mod n =
    (k + j) mod n ->
    i mod n =
    j mod n.
  intros i j k n h1 h2.
  rewrite plus_comm, (plus_comm k j) in h2.
  apply plus_mod_inj_r in h2; auto.
Qed.

Definition shift_seg_fun (n k:nat) : seg_fun n nat :=
  fun i (pf:i < n) => (i + k) mod n.

Lemma lt_shift_seg_fun :
  forall n k i (pf:i < n),
    shift_seg_fun n k i pf < n.
  unfold shift_seg_fun; intros; apply lt_mod; omega.
Qed.

Definition left_shift_seg_fun (n:nat) : seg_fun n nat :=
  shift_seg_fun n (pred n).

Definition right_shift_seg_fun (n:nat) : seg_fun n nat :=
  shift_seg_fun n 1.

Lemma lt_left_shift_seg_fun :
  forall n i (pf:i < n),
    left_shift_seg_fun n i pf < n.
  unfold left_shift_seg_fun; intros; apply lt_shift_seg_fun.
Qed.


Lemma lt_right_shift_seg_fun :
  forall n i (pf:i < n),
    right_shift_seg_fun n i pf < n.
  unfold left_shift_seg_fun; intros; apply lt_shift_seg_fun.
Qed.


Lemma inj_shift_seg_fun :
  forall n k,
    inj_seg (shift_seg_fun n k).
  intros n k.
  red.
  intros i j h1 h2 h3. 
  unfold shift_seg_fun in h3.
  apply plus_mod_inj_r in h3.
  do 2 rewrite lt_mod_eq in h3; auto; subst.
  omega.
Qed.


Lemma inj_left_shift_seg_fun :
  forall n,
    inj_seg (left_shift_seg_fun n).
  unfold left_shift_seg_fun; intros; apply inj_shift_seg_fun.
Qed.

Lemma inj_right_shift_seg_fun :
  forall n,
    inj_seg (right_shift_seg_fun n).
  unfold left_shift_seg_fun; intros; apply inj_shift_seg_fun.
Qed.


Lemma shift_seg_fun0 :
  forall n, 
    shift_seg_fun n 0 = id_seg n.
  unfold shift_seg_fun, id_seg.
  intro.
  apply functional_extensionality_dep.
  intro.
  apply functional_extensionality.
  intro.
  rewrite plus_0_r, lt_mod_eq; auto.
Qed.



(*This would go in FunctionProperties if not for dependencies.*)
Definition shift_functor
           {U:Type} {n:nat} (f:seg_fun n U) (k:nat) (pf0 : 0 < n)
: seg_fun n U :=
  fun i (pf:i < n) => f ((i + k) mod n) (lt_mod _ _ pf0).

Definition left_shift_functor
           {U:Type} {n:nat} (f:seg_fun n U) (pf0 : 0 < n) :
  seg_fun n U :=
  shift_functor f (pred n) pf0.

Definition right_shift_functor
           {U:Type} {n:nat} (f:seg_fun n U) (pf0 : 0 < n) :
  seg_fun n U :=
  shift_functor f 1 pf0.

Lemma inj_shift_functor :
  forall {U:Type} {n:nat} (f:seg_fun n U) (k:nat) (pf0 : 0 < n),
    inj_seg f ->
    inj_seg (shift_functor f k pf0).
  intros U n f k h1 h2. 
  red in h2; red.
  intros i j hi hj.
  unfold shift_functor.
  intro h3. 
  apply h2 in h3.
  apply plus_mod_inj_r in h3; auto.
  rewrite <- neq0_iff in h1.
  pose proof (div_eq_0_iff i n) as h4.
  pose proof (div_eq_0_iff j n) as h5.
  hr h4.
  right; auto.
  hr h5.
  right; auto.
  rewrite <- h4 in hr.
  rewrite <- h5 in hr0.
  pose proof (div_mod i n h1) as h6.
  pose proof (div_mod j n h1) as h7.
  rewrite hr in h6.
  rewrite hr0 in h7.
  rewrite mult_0_r, plus_0_l in h6, h7; auto.
  rewrite h6, h7; auto.
Qed.

Lemma inj_left_shift_functor :
  forall {U:Type} {n:nat} (f:seg_fun n U) (pf0 : 0 < n),
    inj_seg f ->
    inj_seg (left_shift_functor f pf0).
  unfold left_shift_functor; intros; apply inj_shift_functor; auto.
Qed.

Lemma inj_right_shift_functor :
  forall {U:Type} {n:nat} (f:seg_fun n U) (pf0 : 0 < n),
    inj_seg f ->
    inj_seg (right_shift_functor f pf0).
  unfold right_shift_functor; intros; apply inj_shift_functor; auto.
Qed.

  


Section EvenOdd.

Scheme even_ind := Minimality for even Sort Prop
  with odd_ind  := Minimality for odd Sort Prop.

Lemma even_decompose : 
  forall n:nat,
    even n ->
    exists k:nat, n = (2 * k)%nat.
apply (even_ind (fun n => exists k, n = (2*k)%nat) (fun n => exists k, n = (2*k + 1)%nat)).
exists 0%nat. omega.
intros n h1 h2.
destruct h2 as [k h2]. subst.
exists (k + 1)%nat. omega.
intros n h1 h2.
destruct h2 as [k h2]. subst.
exists k. omega.
Qed.

Lemma odd_decompose : 
  forall n:nat,
    odd n ->
    exists k:nat, n = (2 * k + 1)%nat.
pose proof (even_ind (fun n => exists k, S n = (2*k + 1)%nat) (fun n => exists k, S n = (2*k)%nat)) as h0.
cut (forall n:nat, even n -> exists k:nat, S n = (2 * k + 1)%nat).
intros h1 n h2. 
destruct h2 as [n h2].
apply h1; auto.
apply h0.
exists 0%nat. omega.
intros n h1 h2. destruct h2 as [k h2]. rewrite h2.
exists k. omega.
intros n h1 h2. destruct h2 as [k h2]. rewrite h2.
exists (k + 1)%nat. omega.
Qed.

Lemma odd_1 : 
  odd 1.
do 2 constructor.
Qed.


Lemma even_2 : 
  even 2.
do 3 constructor.
Qed.


Lemma odd_eq_1_dec : 
  forall (n:nat),
    odd n ->
    {n = 1%nat} + {(n > 1)%nat}.
intros n h1.
destruct (lt_eq_lt_dec n 1) as [[h2 | h2] | h2].
assert (h3:n = 0%nat). omega. 
subst. 
contradict (not_even_and_odd _ even_O h1).
left; auto.
right; auto.
Qed.

Lemma odd_ge_1 :
  forall (n:nat),
    odd n ->
    1 <= n.
  intros n h1.
  destruct (odd_eq_1_dec _ h1); subst; auto with arith.
Qed.


Lemma odd_Sn_iff :
  forall (n:nat),
    odd (S n) <->
    even n.
intro n.
destruct n as [|n].
split. intro h1. destruct h1. constructor.
intro; constructor; auto.
split.
intro h1.
inversion h1. subst.
assumption.
intro h1.
constructor.
assumption.
Qed.

Lemma even_Sn_iff :
  forall (n:nat),
    even (S n) <->
    odd n.
intro n. 
destruct n as [|n].
split. intro h1. inversion h1.
contradict (not_even_and_odd _ even_O H0).
intro h1. 
contradict (not_even_and_odd _ even_O h1).
split.
intro h1.
inversion h1. assumption.
intro h1. constructor.
assumption.
Qed.

Lemma odd_pred_iff : 
  forall (n:nat),
    (0 < n)%nat ->
    (odd (pred n) <->
    even n).
intros n h0.
destruct n as [|n].
omega.
simpl.
rewrite even_Sn_iff.
tauto.
Qed.


Lemma even_pred_iff : 
  forall (n:nat),
    (0 < n)%nat ->
    (even (pred n) <->
    odd n).
intros n h0.
destruct n as [|n].
omega.
simpl.
rewrite odd_Sn_iff.
tauto.
Qed.


Lemma le_even_odd : 
  forall (m n:nat),
    even m -> odd n ->
    (m <= n)%nat ->
    (m <= pred n)%nat.
intros m n h1 h2 h3.
apply even_decompose in h1.
apply odd_decompose in h2.
destruct h1 as [k h1], h2 as [k' h2].
omega.
Qed.

Lemma le_odd_even : 
  forall (m n:nat),
    odd m -> even n ->
    (m <= n)%nat ->
    (m <= pred n)%nat.
intros m n h1 h2 h3.
apply odd_decompose in h1.
apply even_decompose in h2.
destruct h1, h2; omega.
Qed.

Lemma odd_gt_0 : 
  forall n:nat, 
    odd n -> 0 < n.
intros n h1.
destruct (zerop n) as [h2 |]; auto.
subst.
inversion h1.
Qed.


End EvenOdd.


(*This section defines the natural numbers from 1,2, ... (excluding 0) 
and is built so that inductive proofs involving it are more 
natural than using a lower-bounded sigma type.*)
Section Nat1.

Inductive nat1 : nat -> Set:=
  | nat1_1 : nat1 1
  | nat1_S : forall {n:nat}, nat1 n -> nat1 (S n).

Inductive nat1t : Set :=
  nat1t_intro : forall {n':nat} (n:nat1 n'), nat1t.

Definition nat1t_S (n:nat1t) : nat1t.
destruct n as [n' n].
refine (nat1t_intro (nat1_S n)).
Defined.


Lemma not_nat1_0 : 
  forall n:nat1 0,
    False.
intro n.
inversion n.
Qed.

Definition nat_of_1t (n:nat1t) : nat.
destruct n as [n' n].
refine n'.
Defined.


Lemma nat1_functional : 
  forall  (n' : nat) (n : nat1 n') (m : nat1 n'),
    m = n.
intros n' n m. 
dependent induction n; dependent induction m; auto.
pose proof (not_nat1_0 m) as h1.
contradiction.
pose proof (not_nat1_0 n0) as h1.
contradiction.
f_equal.
apply IHn.
Qed.


Lemma nat1t_functional : 
  forall (m n:nat1t),
    (let (m'', _) := m in
    let (n'', _) := n in 
    m'' = n'') -> m = n.
intros m n. simpl.
destruct m as [m' m], n as [n' n].
intro; subst.
f_equal.
apply nat1_functional.
Qed.


Lemma nat_of_1t_inj :
  Injective nat_of_1t.
red. intros x y h1.
unfold nat_of_1t in h1.
destruct x as [x h2], y as [y h3].
subst.
f_equal.
apply nat1_functional.
Qed.

Lemma nat_of_1t_S : 
  forall (n:nat1t),
    nat_of_1t (nat1t_S n) = 
    S (nat_of_1t n).
intro n.
destruct n; auto.
Qed.


Definition plus1 {m' n':nat} (m:nat1 m') (n:nat1 n') : nat1 (m'+n').
induction m as [|m'].
eapply nat1_S; auto.
simpl.
constructor; auto.
Qed.

Definition plus1t (m n:nat1t) : nat1t.
destruct m as [m' m], n as [n' n].
pose (plus1 m n) as p.
eapply nat1t_intro.
refine p.
Defined.


Definition times1 {m' n':nat} (m:nat1 m') (n:nat1 n') : nat1 (m'*n').
induction m as [|m' s' s].
rewrite mult_1_l; auto.
rewrite mult_comm.
rewrite <- mult_n_Sm.
pose (plus1 n s) as p.
rewrite plus_comm in p.
rewrite mult_comm in p.
refine p.
Defined.

Definition times1t (m n:nat1t) : nat1t.
destruct m as [m' m], n as [n' n].
pose (times1 m n) as p.
eapply nat1t_intro.
refine p.
Defined.


Definition n1_1 : nat1 1.
constructor.
Defined.

Definition n1t_1 : nat1t := (nat1t_intro (n1_1)).

Definition n1_2 : nat1 2.
do 2 constructor. 
Defined.

Definition n1t_2 : nat1t := (nat1t_intro (n1_2)).





Definition even1 {n':nat} (n:nat1 n') : Prop  := even n'.
Definition odd1 {n':nat} (n:nat1 n') : Prop  := odd n'.

Definition even1t (n:nat1t) : Prop.
destruct n as [n' n].
refine (even1 n).
Defined.


Definition odd1t (n:nat1t) : Prop.
destruct n as [n' n].
refine (odd1 n).
Defined.






Lemma even_odd_dec1 : 
  forall {n':nat} (n:nat1 n'),
    {even1 n} + {odd1 n}.
intros n' n. 
destruct (even_odd_dec n') as [h1 | h1].
left. red. assumption.
right. red. assumption.
Qed.

Lemma even_odd_dec_1t : 
  forall (n:nat1t),
    {even1t n} + {odd1t n}.
intro n.
destruct n as [n' n].
apply (even_odd_dec1 n).
Qed.

Lemma not_even_and_odd_1t : 
  forall n:nat1t,
    even1t n -> odd1t n -> False.
intros n h1 h2.
destruct n as [n' n].
red in h1, h2.
apply (not_even_and_odd _ h1 h2).
Qed.


Lemma odd1_1 : odd1 (n1_1).
  red. apply odd_1.
Qed.

Lemma odd1t_1 : odd1t (n1t_1).
unfold n1t_1, odd1t.
apply odd1_1.
Defined.


Lemma even1_2 : even1 (n1_2).
  red. apply even_2.
Qed.

Lemma even1t_2 : even1t (n1t_2).
unfold n1t_2, even1t.
apply even1_2.
Qed.



Lemma odd_Sn_iff1 :
  forall {n':nat} (n:nat1 n'),
    odd1 (nat1_S n) <->
    even1 n.
intros n' n.
split.
intro h1. red in h1. red.
rewrite <- odd_Sn_iff; auto.
intro h1. red in h1. red.
rewrite odd_Sn_iff; auto.
Qed.

Lemma odd_Sn_iff1t :
  forall (n:nat1t), 
    odd1t (nat1t_S n) <->
    even1t n.
intro n.
unfold nat1t_S.
destruct n as [n' n].
simpl.
apply odd_Sn_iff1.
Qed.


Lemma even_Sn_iff1 :
  forall {n':nat} (n:nat1 n'),
    even1 (nat1_S n) <->
    odd1 n.
intros n' n.
split.
intro h1. red in h1. red.
rewrite <- even_Sn_iff; auto.
intro h1. red in h1. red.
rewrite even_Sn_iff; auto.
Qed.

Lemma even_Sn_iff1t :
  forall (n:nat1t),
    even1t (nat1t_S n) <->
    odd1t n.
intro n.
unfold nat1t_S.
destruct n as [n' n].
simpl.
apply even_Sn_iff1.
Qed.



Definition pred1 {n':nat} (n:nat1 n') : 
  nat1 (match n' with 
                | 0 => 1
                | 1 => 1
                | k => pred k
              end).
destruct n as [|n' n].
refine n1_1.
destruct n as [|n' n].
simpl. 
refine n1_1.
simpl.
apply nat1_S; auto.
Defined.

Definition pred1t (n:nat1t) : nat1t.
destruct n as [n' n].
refine (nat1t_intro (pred1 n)).
Defined.



Lemma gt0_to_1t_ex : 
  forall {n:nat}, 0 < n ->
    exists! n1t:nat1t,
      nat_of_1t n1t = n.
intros n h1.
induction n as [|n h2].
contradict (lt_irrefl _ h1).
destruct (zerop n) as [h3 | h3].
subst.
exists n1t_1.
red; simpl. split; auto.
intros n h3. 
destruct n as [n' n].
simpl in h3.
subst.
pose proof (nat1t_functional n1t_1 (nat1t_intro n)) as h3.
simpl in h3.
apply h3; auto.
specialize (h2 h3).
destruct h2 as [n1t h2].
red in h2.
destruct h2 as [h2 h4].
subst.
exists (nat1t_S n1t).
red; simpl; split; auto.
unfold nat1t_S.
destruct n1t.
simpl.
reflexivity.
intros n h5.
unfold nat1t_S.
destruct n1t.
simpl in h5.
specialize (h4 (pred1t n)).
simpl in h4.
unfold nat_of_1t in h4.
destruct n as [n h6].
simpl in h4.
destruct n as [|n].
contradict (not_nat1_0 h6).
destruct n as [|n].
simpl in h5.
assert (h7:n' = 0). omega.
subst.
contradict (not_nat1_0 n0).
simpl in h4.
unfold nat_of_1t in h5.
apply S_inj in h5.
specialize (h4 h5).
subst.
injection h4.
intro h7.
unfold pred1 in h7.
simpl in h7.
dependent induction h6.
dependent induction h6.
f_equal.
f_equal.
apply inj_pair2 in h7.
subst.
simpl in h4.
unfold n1_1.
reflexivity.
f_equal. f_equal.
apply inj_pair2 in h7.
assumption.
Qed.

Definition gt0_to_1t {n:nat} (pf:0 < n) :=
  proj1_sig (constructive_definite_description _ (gt0_to_1t_ex pf)).

Lemma gt0_to_1t_compat : 
  forall {n:nat} (pf:0 < n),
    let n1t := gt0_to_1t pf in
      nat_of_1t n1t = n.
unfold gt0_to_1t.
intros.
destruct constructive_definite_description.
simpl.
assumption.
Qed.

Lemma gt0_to_1t_S :
  forall (n:nat) (pf:0 < n) (pf':0 < S n),
    gt0_to_1t pf' = nat1t_S (gt0_to_1t pf).
intros n h1 h2. 
pose proof (gt0_to_1t_compat h2) as h3. simpl in h3.
pose proof (gt0_to_1t_compat h1) as h4. simpl in h4.
apply nat_of_1t_inj.
rewrite nat_of_1t_S.
rewrite h3, h4.
reflexivity.
Qed.

Lemma gt0_to_1t_inj : 
  forall (m n:nat) (pfm:0 < m) (pfn:0 < n),
    gt0_to_1t pfm = gt0_to_1t pfn ->
    m = n.
intros m n h1 h2 h3.
pose proof (gt0_to_1t_compat h1) as h4.
pose proof (gt0_to_1t_compat h2) as h5.
simpl in h4, h5.
rewrite <- h4,  <- h5 at 1.
f_equal.
assumption.
Qed.


Lemma even_gt0_to_1t_iff : 
  forall (n:nat) (pf:0 < n),
    even n <-> even1t (gt0_to_1t pf).
intros n h1.
pose proof (gt0_to_1t_compat h1) as h2.
simpl in h2. 
rewrite <- h2 at 1. 
unfold even1t.
unfold nat_of_1t.
destruct (gt0_to_1t h1).
subst.
unfold even1.
tauto.
Qed.


Definition odd_to_1t {n:nat} (pf:odd n) : nat1t.
pose proof (odd_gt_0 _ pf) as h1.
refine (gt0_to_1t h1).
Defined.

Lemma odd_to_1t_compat : 
  forall {n:nat} (pf:odd n),
    odd1t (odd_to_1t pf).
intros n h1.
red.
unfold odd_to_1t.
unfold gt0_to_1t.
destruct constructive_definite_description as [n' h2].
simpl.
destruct n' as [n'' n'].
red. simpl in h2. subst.
assumption.
Qed.


    


Definition nat1t_lt (m n:nat1t) : Prop.
destruct m as [m' m], n as [n' n].
refine (m' < n').
Defined.

Definition nat1t_le (m n:nat1t) : Prop.
destruct m as [m' m], n as [n' n].
refine (m' <= n').
Defined.

Lemma nat1t_le_iff : 
  forall (m n:nat1t),
    nat1t_le m n <->
    let (m', _) := m in 
    let (n', _) := n in
    m' <= n'.
intros m n.
destruct m as [m h1], n as [n h2].
unfold nat1t_le.
tauto.
Qed.

Lemma nat1t_lt_iff : 
  forall (m n:nat1t),
    nat1t_lt m n <->
    let (m', _) := m in 
    let (n', _) := n in
    m' < n'.
intros m n.
destruct m as [m h1], n as [n h2].
unfold nat1t_lt.
tauto.
Qed.

Lemma le_nat_from_1t_iff :
  forall (m:nat) (n:nat1t) (pf:0 < m),
    m <= nat_of_1t n <->
    nat1t_le (gt0_to_1t pf) n.
intros m n h1.
pose proof (gt0_to_1t_compat h1) as h2.
simpl in h2.
unfold nat1t_le.
destruct (gt0_to_1t h1).
destruct n as [a b].
subst.
simpl.
tauto.
Qed.


Lemma odd_pred_iff1 :
  forall {n':nat} (n:nat1 n'),
    1 < n' ->
    ((odd1 (pred1 n) <->
    even1 n)).
intros n' n h0.
split.
intro h1. red in h1. red.
destruct n' as [|n'].
constructor.
destruct n' as [|n'].
omega.
simpl in h1.
rewrite <- odd_pred_iff.
simpl.
assumption.
omega.
intro h1. red in h1. red.
destruct n' as [|n'].
constructor. assumption.
destruct n' as [|n'].
omega.
rewrite odd_pred_iff; auto.
omega.
Qed.

Lemma odd_pred_iff1t :
  forall (n:nat1t),
    nat1t_lt n1t_1 n ->
    ((odd1t (pred1t n) <->
    even1t n)).
intros n h1.
red in h1.
destruct n as [n' n].
simpl.
unfold n1t_1 in h1.
apply odd_pred_iff1; auto.
Qed.



Lemma even_pred_iff1 :
  forall {n':nat} (n:nat1 n'),
    1 < n' ->
    (even1 (pred1 n) <->
    odd1 n).
intros n' n h0.
split.
intro h1. red in h1. red.
destruct n' as [|n'].
omega.
destruct n' as [|n'].
omega.
simpl in h1.
rewrite <- even_pred_iff.
simpl.
assumption.
omega.
intro h1. red in h1. red.
destruct n' as [|n'].
constructor. assumption.
destruct n' as [|n'].
omega.
rewrite even_pred_iff; auto.
omega.
Qed.


Lemma even_pred_iff1t :
  forall (n:nat1t),
    nat1t_lt n1t_1 n ->
    (even1t (pred1t n) <->
    odd1t n).
intros n h1.
unfold nat1t_lt, n1t_1 in h1.
destruct n as [n' n].
simpl.
apply even_pred_iff1; auto.
Qed.




Lemma gt1t_dec :
  forall (n:nat1t),
    {n = n1t_1} + {nat1t_lt n1t_1 n}.
intro n. 
destruct n as [n' n].
destruct (lt_eq_lt_dec n' 1) as [h2 | h2].
destruct h2 as [h2 | h2].
apply lt1 in h2. subst.
pose proof (not_nat1_0 n).
contradiction.
subst.
left.
unfold n1t_1, n1_1.
f_equal.
apply nat1_functional.
right.
red. 
simpl.
assumption.
Qed.




Section EvenOddNat1.

Inductive even_nat1t : Set :=
  | even_nat1t_intro :
      forall (n:nat1t), even1t n -> even_nat1t.

Inductive odd_nat1t : Set :=
  | odd_nat1t_intro :
      forall (n:nat1t), odd1t n -> odd_nat1t.


Definition odd_to_on1t {n:nat} (pf:odd n) : odd_nat1t :=
  odd_nat1t_intro _ (odd_to_1t_compat pf).


Lemma even_nat1t_functional : 
  forall (m n:nat1t) (pfm:even1t m) (pfn:even1t n),
    m = n -> even_nat1t_intro m pfm = even_nat1t_intro n pfn.
  intros; subst; pose proof (proof_irrelevance _ pfn pfm); subst; auto.
Qed.


Lemma odd_nat1t_functional : 
  forall (m n:nat1t) (pfm:odd1t m) (pfn:odd1t n),
    m = n -> odd_nat1t_intro m pfm = odd_nat1t_intro n pfn.
  intros; subst; pose proof (proof_irrelevance _ pfn pfm); subst; auto.
Qed.


Definition even_nat1ts := Full_set even_nat1t.
Definition odd_nat1ts := Full_set odd_nat1t.

Definition nat_of_en1t (n:even_nat1t) : nat.
destruct n as [n n'].
refine (nat_of_1t n).
Defined.

Definition nat_of_on1t (n:odd_nat1t) : nat.
destruct n as [n n'].
refine (nat_of_1t n).
Defined.



Definition en1t_lt : even_nat1t -> even_nat1t -> Prop.
intros m n.
destruct m as [m h1], n as [n h2].
refine (nat1t_lt m n).
Defined.


Definition on1t_lt : odd_nat1t -> odd_nat1t -> Prop.
intros m n.
destruct m as [m h1], n as [n h2].
refine (nat1t_lt m n).
Defined.


Definition en1t_le : even_nat1t -> even_nat1t -> Prop.
intros m n.
destruct m as [m h1], n as [n h2].
refine (nat1t_le m n).
Defined.


Definition on1t_le : odd_nat1t -> odd_nat1t -> Prop.
intros m n.
destruct m as [m h1], n as [n h2].
refine (nat1t_le m n).
Defined.


Lemma on1t_le_refl : 
  forall n:odd_nat1t, 
    on1t_le n n.
intro n.
red. 
destruct n as [n h1].
red.
destruct n.
auto with arith.
Qed.


Lemma en1t_le_refl : 
  forall n:even_nat1t, 
    en1t_le n n.
intro n.
red. 
destruct n as [n h1].
red.
destruct n.
auto with arith.
Qed.

Lemma lt_1_nat1t : 
  forall n:nat1t,
    1 <= nat_of_1t n.
intro n.
destruct n as [n' n].
simpl.
destruct (lt_eq_lt_dec 1 n') as [h1 | h1].
destruct h1; auto with arith. subst.
auto with arith.
assert (n' = 0). omega.
subst.
inversion n.
Qed.


Lemma le_1_odd1t : 
  forall (n:odd_nat1t),
    1 <= nat_of_on1t n.
intro n.
destruct n as [n h1]. 
apply lt_1_nat1t; auto.
Qed.

Lemma onep_on1t : 
  forall n:odd_nat1t,
    {1 = nat_of_on1t n} + 
    {1 < nat_of_on1t n}.
intro n.
pose proof (le_1_odd1t n) as h1.
apply le_lt_eq_dec in h1.
destruct h1.
right; auto.
left; auto.
Qed.


Lemma en1t_ge_2 :
  forall m:even_nat1t,
    2 <= nat_of_en1t m.
intro m.
destruct m as [m' h1].
destruct m' as [m'' m'].
red in h1.
simpl.
red in h1.
apply even_decompose in h1.
destruct h1 as [k h1].
subst.
destruct k as [|k].
simpl in m'.
inversion m'.
omega.
Qed.

Lemma en1t_gt_1 :
  forall m:even_nat1t,
    1 < nat_of_en1t m.
intro m.
pose proof (en1t_ge_2 m) as h1.
destruct m as [m' h2].
auto with arith.
Qed.


Lemma en1t_ge_1 :
  forall m:even_nat1t,
    1 <= nat_of_en1t m.
intro m.
pose proof (en1t_gt_1 m) as h1.
destruct m as [m' h2].
auto with arith.
Qed.


Lemma nat_of_en1t_le_compat : 
  forall (m n:even_nat1t),
    nat_of_en1t m <= nat_of_en1t n ->
    en1t_le m n.
intros m n h1.
red.
destruct m as [m h2], n as [n h3].
simpl in h1.
red.
destruct m as [m' m], n as [n' n].
simpl in h1.
assumption.
Qed.


Lemma nat_of_on1t_le_compat :
  forall (m n:odd_nat1t),
    nat_of_on1t m <= nat_of_on1t n ->
    on1t_le m n.
intros m n h1.
red.
destruct m as [m h2], n as [n h3].
simpl in h1.
red.
destruct m as [m' m], n as [n' n].
simpl in h1.
assumption.
Qed.



Lemma even_nat_of_en1t : 
  forall (n:even_nat1t),
    even (nat_of_en1t n).
intro n.
pose proof (en1t_ge_1 n) as h1.
assert (h2:0 < nat_of_en1t n).
  auto with arith.
rewrite (even_gt0_to_1t_iff _ h2).
destruct n as [n h3].
unfold gt0_to_1t.
destruct constructive_definite_description as [n' h4].
simpl.
simpl in h4.
apply nat_of_1t_inj in h4. 
subst.
assumption.
Qed.


Lemma not_odd_nat_of_en1t : 
  forall (n:even_nat1t),
    ~odd (nat_of_en1t n).
intro n.
pose proof (even_nat_of_en1t n) as h1.
intro h2.
apply (not_even_and_odd _ h1 h2).
Qed.


Lemma odd_nat_of_on1t : 
  forall (n:odd_nat1t),
    odd (nat_of_on1t n).
intro n.
unfold nat_of_on1t.
destruct n as [n' h1].
red in h1.
destruct n' as [n'' n'].
simpl.
red in h1.
assumption.
Qed.

Lemma not_even_nat_of_on1t : 
  forall (n:odd_nat1t),
    ~even (nat_of_on1t n).
intro n.
pose proof (odd_nat_of_on1t n) as h1.
intro h2.
apply (not_even_and_odd _ h2 h1).
Qed.


Lemma match_gt0_to_1t_even_iff : 
  forall (m:nat) (pf:0 < m),
         (match gt0_to_1t pf with
           | nat1t_intro n' n0 => even1 n0
          end) <->
         even (nat_of_1t (gt0_to_1t pf)).
intros m h0.
unfold gt0_to_1t.
destruct constructive_definite_description as [n h1].
simpl.
subst.
destruct n as [n' n].
simpl.
unfold even1.
tauto.
Qed.


Lemma nat_of_1t_pred_compat : 
  forall (n:nat1t),
    1 < nat_of_1t n ->
    nat_of_1t (pred1t n) = pred (nat_of_1t n).
intros n h0.
unfold nat_of_1t, pred1t.
destruct n as [n h1].
destruct n.
inversion h1.
dependent induction h1.
simpl in h0.
contradict (lt_irrefl _ h0).
destruct n as [|n].
simpl in h0.
contradict (lt_irrefl _ h0).
reflexivity.
Qed.



Lemma on1t_pred_ex : 
  forall n:odd_nat1t,
    1 < nat_of_on1t n ->
    exists! m:even_nat1t,
    nat_of_en1t m = pred (nat_of_on1t n).
intros n.
destruct n as [n' h0].
intro h1.
pose (pred1t n') as m.
pose proof (even_pred_iff1t n').
simpl in h1. 
assert (h1':nat1t_lt n1t_1 n').
  red. simpl. destruct n'.
  assumption.
pose proof (even_pred_iff1t n' h1') as h2. 
pose proof (iff2 h2 h0) as h0'.
exists (even_nat1t_intro _ h0'). 
red; split; auto.
simpl.
apply nat_of_1t_pred_compat; auto.
intros e h3.
destruct e as [e h4].
subst.
f_equal.
simpl in h3.
rewrite <- nat_of_1t_pred_compat in h3; auto.
pose proof nat_of_1t_inj as h5.
red in h5.
specialize (h5 _ _ h3).
subst.
f_equal.
apply proof_irrelevance.
Qed.


Definition on1t_pred (n:odd_nat1t)
           (pf: 1 < nat_of_on1t n) : even_nat1t :=
  proj1_sig 
    (constructive_definite_description _ (on1t_pred_ex n pf)).


Lemma on1t_pred_compat : 
  forall (n:odd_nat1t) (pf:1 < nat_of_on1t n),
    let m := on1t_pred n pf in
    nat_of_en1t m = pred (nat_of_on1t n).
unfold on1t_pred.
intros.
destruct constructive_definite_description.
simpl.
assumption.
Qed.

(*This defaults to "2" when passed an argument of "1"*)
Definition on1t_pred' (n:odd_nat1t) : even_nat1t.
destruct (onep_on1t n) as [h1 | h1].
refine (even_nat1t_intro _ even1t_2).
refine (on1t_pred _ h1).
Defined.


Lemma en1t_pred_ex : 
  forall n:even_nat1t,
    exists! m:odd_nat1t,
    nat_of_on1t m = pred (nat_of_en1t n).
intros n.
pose proof (en1t_gt_1 n) as h0.
destruct n as [n' n].
assert (h1':nat1t_lt n1t_1 n').
  red. simpl.
  destruct n' as [n'' n'].
  simpl in h0.
  assumption.
pose proof (odd_pred_iff1t _ h1') as h2. 
pose proof (iff2 h2 n) as h0'.
exists (odd_nat1t_intro _ h0'). 
red; split; auto.
simpl.
apply nat_of_1t_pred_compat; auto.
intros e h3.
destruct e as [e h4].
subst.
f_equal.
simpl in h3.
rewrite <- nat_of_1t_pred_compat in h3; auto.
pose proof nat_of_1t_inj as h5.
red in h5.
specialize (h5 _ _ h3).
subst.
f_equal.
apply proof_irrelevance.
Qed.


Definition en1t_pred (n:even_nat1t) : odd_nat1t :=
  proj1_sig 
    (constructive_definite_description _ (en1t_pred_ex n)).

Lemma en1t_pred_compat : 
  forall (n:even_nat1t) (pf:1 < nat_of_en1t n),
    let m := en1t_pred n in
    nat_of_on1t m = pred (nat_of_en1t n).
unfold en1t_pred.
intros.
destruct constructive_definite_description.
simpl.
assumption.
Qed.



Lemma nat_of_en1t_le_pred_compat : 
  forall (m:even_nat1t) (n:odd_nat1t),
    nat_of_en1t m <= nat_of_on1t n ->
    en1t_le m (on1t_pred' n).
intros m n h1.
unfold on1t_pred'.
destruct onep_on1t as [h2 | h2].
red.
destruct m as [m' m].
rewrite <- h2 in h1.
unfold nat_of_en1t in h1.
red.
destruct m' as [m'' m'].
simpl.
simpl in h1.
auto with arith.
red.
destruct m as [m' m].
unfold on1t_pred. 
destruct constructive_definite_description as [n' h3].
simpl.
destruct n' as [n'' n']. simpl in h2.
simpl in h1.
red.
destruct m' as [m'' m'].
destruct n'' as [n''' n''].
simpl in h1.
simpl in h3.
subst.
red in m. red in m.
assert (h3: m'' <> nat_of_on1t n).
  intro h3.
  subst.
  destruct n as [o h3].
  simpl in m.
  red in h3.
  destruct o as [o' o].
  red in h3. simpl in m.
  apply (not_even_and_odd _ m h3).
omega.
Qed.


Lemma nat_of_on1t_le_pred_compat :
  forall (m:odd_nat1t) (n:even_nat1t),
    nat_of_on1t m <= nat_of_en1t n ->
    on1t_le m (en1t_pred n).
intros m n h1.
unfold en1t_pred.
destruct m as [m' m].
unfold on1t_pred. 
destruct constructive_definite_description as [n' h3].
simpl.
destruct n' as [n'' n'].
simpl in h1.
red.
destruct m' as [m'' m'].
destruct n'' as [n''' n''].
simpl in h1.
simpl in h3.
subst.
red in m. red in m.
assert (h3: m'' <> nat_of_en1t n).
  intro h3.
  subst.
  destruct n as [o h3].
  simpl in m.
  red in h3.
  destruct o as [o' o].
  red in h3. simpl in m.
  apply (not_even_and_odd _ h3 m).
omega.
Qed.



Lemma en1t_le_iff : 
  forall (m n:even_nat1t),
    en1t_le m n <->
    let (m', _) := m in
    let (n', _) := n in
    nat1t_le m' n'.
intros m n.
destruct m as [m h1], n as [n h2].
split.
intro h3.
red in h3.
assumption.
intro h3. red.
assumption.
Qed.


End EvenOddNat1.

Lemma S_le_1 :
  forall (n:nat), 1 <= S n.
  intros; auto with arith.
Qed.


Lemma st01_pf : 0 < 1.
  auto with arith.
Qed.


Lemma st02_pf : 0 < 2.
  auto with arith.
Qed.
  
Lemma st12_pf : 1 < 2.
  auto with arith.
Qed.



(*For use in conjunction with [seg_one_t] defined in SetUtilities.*)
Lemma sot11_pf : 1 <= 1 <= 1.
  auto with arith.
Qed.


Lemma sot12_pf : 1 <= 1 <= 2.
  auto with arith.
Qed.

Lemma sot22_pf : 1 <= 2 <= 2.
  auto with arith.
Qed.


Lemma sot_cond : 
  forall n:nat1t,
    let (n', _) := n in
    1 <= n' <= n'.
  intro n.
  pose proof (lt_1_nat1t n) as h1.
  destruct n. simpl in h1.
  auto with arith.
Qed.


Lemma lt_sot_iff :
  forall (m n:nat),
    m < n <->
    1 <= S m <= n.
  intros; omega.
Qed.

End Nat1.




Section Integers.

Require Import ZArith.

Lemma Zpos_inj :
  forall p q,
    Zpos p = Zpos q ->
    p = q.
  intros ? ? h1.
  inversion h1; auto.
Qed.

Lemma Zneg_inj :
  forall p q,
    Zneg p = Zneg q ->
    p = q.
  intros ? ? h1.
  inversion h1; auto.
Qed.



Lemma Zlt_impl_Zle :
  forall (a b:Z),
    (a < b)%Z -> (a <= b)%Z.
intros; omega.
Qed.

Lemma Z_decompose : 
  forall (a:Z),
    (a = Z.sgn a * (Zpos (Z.to_pos (Z.abs a))))%Z.
  intro a; destruct a; simpl; auto.
Qed.


Lemma Zlt_neg :
  forall (p q:positive),
    (Z.pos p < Z.pos q)%Z ->
    (Z.neg p > Z.neg q)%Z.
  intros p q h1.
  inversion h1.
  red.
  simpl.
  rewrite H0.
  simpl; auto.
Qed.

  
Lemma Zlt_mult_mono :
  forall p q r,
    (Z.pos p < Z.pos q)%Z ->
    (Z.pos p * Z.pos r < 
     Z.pos q * Z.pos r)%Z.
  intros  p q r h1.
  red; red in h1.
  simpl; simpl in h1.
  rewrite Pos.mul_comm.
  rewrite (Pos.mul_comm q r).
  rewrite Pos.mul_compare_mono_l; auto.
Qed.


Lemma Zgt_mult_mono :
  forall p q r,
    (Z.pos p > Z.pos q)%Z ->
    (Z.pos p * Z.pos r >
     Z.pos q * Z.pos r)%Z.
  intros  p q r h1.
  red; red in h1.
  simpl; simpl in h1.
  rewrite Pos.mul_comm.
  rewrite (Pos.mul_comm q r).
  rewrite Pos.mul_compare_mono_l; auto.
Qed.



Lemma Zlt_mult_neg :
  forall (a b:Z) (p:positive),
    (a < b)%Z ->
    (a * Z.neg p > b * Z.neg p)%Z.
  intros a b p h1.
  destruct a.
  simpl.
  destruct b.
  omega.
  simpl.
  constructor.
  inversion h1.
  simpl.
  destruct b; simpl.
  inversion h1.
  apply Zlt_neg.
  do 2 rewrite Pos2Z.inj_mul.
  apply Zlt_mult_mono; auto.
  inversion h1.
  destruct b; simpl.
  constructor.
  constructor.
  do 2 rewrite Pos2Z.inj_mul.
  apply Zgt_mult_mono; auto.
  red in h1; red.
  simpl in h1; simpl.
  rewrite <- Pos.compare_antisym in h1.
  pose proof (f_equal CompOpp h1) as h2.
  rewrite <- Pos.compare_antisym in h2.
  simpl in h2; auto.
Qed.

Lemma Zlt_mult :
  forall (a b c:Z),
    (c < 0)%Z ->
    (a < b)%Z ->
    (a * c > b * c)%Z.
  intros a b c. revert a b.
  induction c.
  intros; omega.
  intros a b h2 h1.
  inversion h2.
  intros a b h1 h2.
  apply Zlt_mult_neg; auto.
Qed.


Lemma Z_neg_prod :
  forall (a b:Z),
    (a < 0)%Z ->
    (b < 0)%Z ->
    (a * b > 0)%Z.
  intro a.
  destruct a.
  intro b.
  destruct b.
  intros; omega.
  intros; omega.
  intros; omega.
  intro b.
  destruct b.
  intros; omega.
  intros h1 h2.
  destruct (Z_dec (Z.pos p * Z.pos p0) 0) as [h3 | h3].
  destruct h3 as [h3 | h3].
  inversion h3.
  inversion h2.
  inversion h2.
  intro h1.
  inversion h1.
  intro b.
  destruct b.
  intros; omega.
  intros ? h1.
  inversion h1.
  intros.
  simpl.
  constructor.
Qed.


Lemma Z_mixed_prod :
  forall (a b:Z),
    (a < 0)%Z ->
    (b > 0)%Z ->
    (a * b < 0)%Z.
  intro a. destruct a.
  intro b.
  destruct b.
  intros; omega.
  intros; omega.
  intros; omega.
  intros b h1.
  inversion h1.
  intro b.
  destruct b.
  intros; omega.
  intros; simpl.
  constructor.
  intros ? h1.
  inversion h1.
Qed.

Lemma Z_pos_prod :
  forall (a b:Z),
    (a > 0)%Z ->
    (b > 0)%Z ->
    (a * b > 0)%Z.
  intro a.
  destruct a.
  intros; omega.
  intro b.
  destruct b.
  intros; omega.
  intros; simpl.
  constructor.
  intros ? h1.
  inversion h1.
  intros ? h1.
  inversion h1.
Qed.

Lemma no_zero_divisors :
  forall (a b:Z),
    (a * b)%Z = 0%Z ->
    {a = 0%Z} + {b = 0%Z}.
  intros a b h1.
  destruct (Z_dec a 0) as [h2 | h2].
  destruct h2 as [h2 | h2].
  right.
  destruct (Z_dec b 0) as [h3 | h3].
  destruct h3 as [h3 | h3].
  pose proof (Z_neg_prod _ _ h2 h3) as h4.
  omega.
  pose proof (Z_mixed_prod _ _ h2 h3) as h4.
  omega.
  assumption.
  right.
  destruct (Z_dec b 0) as [h3 | h3].
  destruct h3 as [h3 | h3].
  pose proof (Z_mixed_prod _ _ h3 h2) as h4.
  rewrite Zmult_comm in h4.
  omega.
  pose proof (Z_pos_prod _ _ h2 h3) as h4.
  omega.
  assumption.
  left; auto.
Qed.

Lemma Zpos1 :
  forall (p:positive),
    (Zpos (p~1) =
     2 * Zpos p + 1)%Z.
  intro p; induction p; simpl; auto.
Qed.

Lemma Zpos0 :
  forall (p:positive),
    (Zpos (p~0) =
     2 * Zpos p)%Z.
  intro p; induction p; simpl; auto.
Qed.


Lemma Zneg1 :
  forall (p:positive),
    (Zneg (p~1) =
     2 * Zneg p - 1)%Z.
  intro p; induction p; simpl; auto.
Qed.

Lemma Zneg0 :
  forall (p:positive),
    (Zneg (p~0) =
     2 * Zneg p)%Z.
  intro p; induction p; simpl; auto.
Qed.

Lemma pos_to_nat_compat :
  forall (p:positive),
    Pos.to_nat p =
    Z.to_nat (Zpos p).
  intro p; induction p; simpl; auto.
Qed.

Lemma mult_to_nat :
  forall (n:nat) (a:Z),
    n * Z.to_nat a =
    Z.to_nat ((Z.of_nat n) * a).
  intro n.
  induction n as [|n ih]; simpl; auto.
  intro a.
  destruct a; simpl.
  apply mult_0_r.
  specialize (ih (Zpos p)). 
  simpl in ih.
  rewrite ih.
  rewrite Pos2Nat.inj_mul.
  rewrite SuccNat2Pos.id_succ.
  rewrite Z2Nat.inj_mul.
  rewrite Nat2Z.id.
  rewrite <- pos_to_nat_compat.
  rewrite <- S_compat.
  rewrite mult_plus_distr_r.
  rewrite mult_1_l.
  rewrite plus_comm; auto.
  omega.
  apply Zle_0_pos.
  omega.
Qed.

  
Lemma pos_to_nat1 :
  forall (p:positive),
    Pos.to_nat p~1 = 2 * Pos.to_nat p + 1. 
  intro p. 
  rewrite pos_to_nat_compat.
  rewrite Zpos1.
  rewrite pos_to_nat_compat.
  rewrite Z2Nat.inj_add.
  f_equal.
  simpl. 
  rewrite plus_0_r.
  rewrite <- mult2.
  do 2 rewrite pos_to_nat_compat.
  rewrite mult_to_nat.
  f_equal.
  pose proof (Zle_0_pos p) as h1.
  pose proof (Z.mul_0_r 2) as h2.
  rewrite <- h2.
  apply Z.mul_le_mono_nonneg; omega.
  omega.
Qed.

Lemma pred_double_Z : 
  forall p,
    Z.pos (Pos.pred_double p) =
    Z.pred_double (Zpos p).
  auto.
Qed.

  
Lemma Z_pred_inj :
  forall i j,
    Z.pred i = Z.pred j ->
    i = j.
  intro i.
  induction i; simpl.
  intros j h1.
  pose proof (f_equal Z.succ h1) as h2.
  simpl in h2.
  rewrite Z.succ_pred in h2; auto. 
  destruct p.
  intros j h1.
  pose proof (f_equal Z.succ h1) as h2.
  simpl in h2.
  rewrite h2, Z.succ_pred; auto.
  intros j h1.
  pose proof (f_equal Z.succ h1) as h2.
  rewrite Zpos0.
  rewrite pred_double_Z in h1.
  rewrite Z.pred_double_spec in h1.
  omega.
  intros; omega.
  intros j h1.
  rewrite <- Pos2Z.add_neg_neg in h1.
  omega.
Qed.

End Integers.

Section Rationals.

Require Import QArith.

Lemma Qminus_pos_Qlt :
  forall (b q:Q), (b > 0) ->
    (q - b < q).
intros b q h1.
rewrite Qlt_minus_iff. 
unfold Qminus. 
rewrite Qopp_plus at 1.
rewrite Qopp_opp.
rewrite Qplus_assoc.
assert (h3:Qeq (q - q) 0). ring.
rewrite h3.
rewrite Qplus_0_l at 1.
assumption.
Qed.

Lemma Qlt_plus_pos :
  forall (b q:Q), (b > 0) ->
    (q  < q + b).
intros b q h1.
rewrite Qlt_minus_iff. 
rewrite <- Qplus_assoc.
rewrite Qplus_comm.
rewrite <- Qplus_assoc.
rewrite (Qplus_comm (-q) q).
assert (h2:q - q == 0). ring.
rewrite h2.
rewrite Qplus_0_r.
assumption.
Qed.

Lemma Qlt_midpoint : 
  forall (p q:Q),
    p < q ->
    p < (p + (q - p) / (2 # 1)).
intros p q h1.
pose proof (Qplus_lt_l p q (-p)) as h2.
pose proof (iff2 h2 h1) as h3.
rewrite Qplus_opp_r in h3.
assert (h4: 0 < / (2 # 1)).
  constructor.
pose proof (Qmult_lt_l 0 (q + -p) (Qinv (2#1)) h4) as h5.
pose proof (iff2 h5 h3) as h6.
rewrite Qmult_0_r in h6.
rewrite Qmult_comm in h6.
pose proof (Qplus_lt_l 0 ( (q + - p) * / (2 # 1)) p) as h7.
rewrite Qplus_0_l in h7.
rewrite Qplus_comm.
rewrite h7 at 1.
assumption.
Qed.


Lemma midpoint_Qlt : 
  forall (p q:Q),
    p < q ->
    (p + (q - p) / (2 # 1)) < q.
intros p q h1.
assert (h7:p*(2 #1) + -p == p).  simpl.
  rewrite <- (Qplus_inj_r (p * (2 # 1) + - p) p p).
  rewrite <- Qplus_assoc.
  rewrite (Qplus_comm (-p) p).
  rewrite Qplus_opp_r.
  rewrite Qplus_0_r.
  rewrite <- (Qmult_1_r p) at 2 3.
  rewrite <- Qmult_plus_distr_r.
  assert (h7:2#1 = 1 + 1). constructor.
  rewrite h7.
  reflexivity.
assert (h7':q*(2 #1) + -q == q).  simpl.
  rewrite <- (Qplus_inj_r (q * (2 # 1) + - q) q q).
  rewrite <- Qplus_assoc.
  rewrite (Qplus_comm (-q) q).
  rewrite Qplus_opp_r.
  rewrite Qplus_0_r.
  rewrite <- (Qmult_1_r q) at 2 3.
  rewrite <- Qmult_plus_distr_r.
  assert (h7':2#1 = 1 + 1). constructor.
  rewrite h7'.
  reflexivity.
rewrite <- h7 in h1. rewrite <- h7' in h1.
pose proof (Qplus_lt_l (p * (2 # 1) + - p) ( q * (2 # 1) + - q) q) as h8.
pose proof (iff2 h8 h1) as h9. 
rewrite <- (Qplus_assoc  (q * (2 # 1)) (- q) q) in h9.
rewrite (Qplus_comm (-q) q) in h9.
rewrite Qplus_opp_r in h9.
rewrite Qplus_0_r in h9.
rewrite <- Qplus_assoc in h9.
rewrite (Qplus_comm (-p) q) in h9.
assert (h10:0 < (Qinv (2#1))).
  constructor.
pose proof (Qmult_lt_r (p * (2 # 1) +(q + - p)) (q * (2 # 1)) _ h10) as h11.
pose proof (iff2 h11 h9) as h12.
rewrite Qmult_plus_distr_l in h12.
rewrite <- Qmult_assoc in h12. rewrite <- Qmult_assoc in h12.

assert (h13: ~ 2 # 1 == 0).
  intro h13. inversion h13.
pose proof (Qmult_inv_r (2 # 1) h13) as h14.
rewrite h14 in h12.
do 2 rewrite Qmult_1_r in h12.
assumption.
Qed.


Lemma Qlt_impl_Qle : 
  forall (a b:Q),
    a < b -> a <= b.
intros a b h1. 
destruct a as [an ad], b as [bn bd].
apply Zlt_impl_Zle;  auto.
Qed.

Definition p_to_Z (p:positive) : Z :=
  Z.of_nat (Pos.to_nat p).

  
Lemma qmake0 :
  forall (p q:positive),
    Qmake Z0 p ==
    Qmake Z0 q.
  intro p.
  intros q.
  red; simpl; auto.
Qed.

Lemma Zpos_mul :
  forall (p q:positive),
    Zpos (Pos.mul p q) =
    (Zpos p * Zpos q)%Z.
  intros; simpl; auto.
Qed.


Lemma Q_inj_mult :
  forall q r s:Q,
    q == r ->
    s * q == s * r.
  intros q r s.
  revert q r.
  induction s as [n d].
  intros q r h1.
  red.
  red in h1.
  simpl.
  rewrite <- Zmult_assoc.
  symmetry.
  rewrite <- Zmult_assoc.
  f_equal. 
  rewrite Zpos_mul.
  rewrite Zmult_assoc.
  rewrite Zmult_comm.
  rewrite Zmult_assoc.
  rewrite (Zmult_comm _ (Qnum r)).
  rewrite <- h1. 
  rewrite <- Zmult_assoc.
  f_equal.
  simpl.
  rewrite Zpos_mul.
  simpl.
  f_equal.
  apply Pos.mul_comm.
Qed.


Lemma Z_cancel_r :
  forall (a:Z) (p:positive),
    a * ' p # p == a # 1.
  intros a p; red; simpl.
  rewrite Z.mul_1_r; auto.
Qed.

Lemma Z_cancel_l :
  forall (a:Z) (p:positive),
    ' p * a # p == a # 1.
  intros a ?; red; simpl;
  destruct a; simpl; auto; rewrite Pos.mul_1_r, Pos.mul_comm; auto.
Qed.
  
Lemma Q_neq_dec :
  forall (q r:Q),
    ~q == r ->
    {q < r} + {q > r}.
  intros q r ?; destruct (Q_dec q r) as [| h2]; auto; contradiction.
Qed.

Lemma Qnum_recip :
  forall (n:Z) (d:positive),
    n <> 0%Z ->
    (QArith_base.Qnum (/(n # d)) = Zsgn n * Zpos d)%Z.
  intros n; induction n; intros; try omega; simpl; auto.
Qed.


Definition Qmin (a b:Q) :=
  match (Q_dec a b) with 
    | inleft P =>
      if P then a else b
    | _ => b
  end.

Definition Qmax (a b:Q) :=
  match (Q_dec a b) with 
    | inleft P =>
      if P then b else a
    | _ => a
  end.


Lemma Qmin_le : 
  forall (a b:Q),
    Qmin a b <= a /\ Qmin a b <= b.
intros a b.
unfold Qmin. 
destruct Q_dec as [h1 | h1]. 
destruct h1 as [h1 | h1]. 
constructor.  apply Qle_refl.
apply Qlt_impl_Qle; auto.
split.
apply Qlt_impl_Qle; auto.
apply Qle_refl.
rewrite h1.
split; apply Qle_refl.
Qed.

Lemma le_Qmax : 
  forall (a b:Q),
    a <= Qmax a b /\ b <= Qmax a b.
intros a b.
unfold Qmax.
destruct Q_dec as [h1 | h1]. destruct h1 as [h1 | h1].
constructor. apply Qlt_impl_Qle. assumption.
apply Qle_refl.
split. apply Qle_refl. apply Qlt_impl_Qle; auto.
rewrite h1.
split; apply Qle_refl.
Qed.


End Rationals.


(*This section is used to develop lemmas that allow you to 
  destruct a [nat] discriminant in a match clause, without losing 
  information if that discriminant is a functional expression with 
  non-primitive actual paramaters.  
  Of course a different lemma will be needed for each form of the match 
  clause that is currently in use. No attempt is made to be 
  "comprehensive."*)
Section MatchDestruct.
      
Lemma match_destruct_nat_zerop :
  forall {T U V:Type}  (P:nat->U->V->T) (u:U) (v:V)(n:nat) (a:T),
    match n with
      | O => a
      | S m => P m u v
    end =
    let f := (fun m:nat => if (zerop m) then a else (P (pred m) u v)) in
    f n.
  intros T U V P u v n a.
  destruct n. simpl. reflexivity.
  simpl.
  reflexivity.
Qed.
  


End MatchDestruct.

