(* Copyright (C) 2014-2019, Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Coq Version 8.4 pl4.*)

Require Import LogicUtilities.
Require Import TypeUtilities.
Require Import EnsemblesImplicit.
Require Import Finite_sets_facts. 
Require Export List.
Require Import FunctionProperties.
Require Import Constructive_sets.
Require Import DecidableDec.
Require Import FunctionalExtensionality.
Require Import SetUtilities.
Require Import ArithUtilities.
Require Import NaryFunctions.
Require Import ProofIrrelevance.  
Require Import Description.
Require Import EqDec.
Require Import Program.Wf.
Require Import Equality.
Require Import Init.
Require Import Bool.
Require Import NPeano.
Require Import Basics.



Lemma length_subst :
  forall {T:Type} (l l':list T),
  l = l' ->
  length l = length l'.
  intros; subst; auto.
Qed.

Lemma list_eq_dec :
  forall {T:Type},
    type_eq_dec T -> type_eq_dec (list T).
  intros T h0.
  red in h0. red.
  intro l. revert h0.
  induction l as [|a l h1]; intros h0 l'; destruct l' as [|a' l'].
  left; auto.
  right.
  intro h1. discriminate h1.
  right.
  intro h2.
  discriminate h2.
  destruct (h0 a a') as [h2 | h2].
  subst.
  specialize (h1 h0 l').
  destruct h1 as [h1 | h1].
  subst. left; auto.
  right.
  intro h2. inversion h2; contradiction.
  right.
  intro h3. inversion h3; auto.
Qed.



Lemma list_nat_eq_dec : type_eq_dec (list nat).
  apply list_eq_dec; apply nat_eq_dec.
Qed.

Lemma list_nat_eq_dec_eq :
  forall (pf:type_eq_dec (list nat)),
    pf = list_nat_eq_dec.
  intro h1.
  apply functional_extensionality_dep.
  intro l.
  apply functional_extensionality_dep.
  intro l'.
  destruct (h1 l l'), (list_nat_eq_dec l l'); subst; try f_equal; try apply proof_irerlevance; auto.
  apply proof_irrelevance.
  contradict n; auto.
  contradict n; auto.
  apply proof_irrelevance.
Qed.


  
Lemma in_dec_eq :
  forall {T:Type} {l:list T} {a:T} (pfdt:type_eq_dec T)
         (pf:{In a l} + {~In a l}),
    in_dec pfdt a l = pf.
  intros T l a h1 h2.
  destruct h2 as [h2 | h2], in_dec as [h3 | h3]; f_equal; try apply proof_irrelevance; auto.
  contradict (h3 h2).
  contradict (h2 h3).
Qed.

         
Definition faml (T:Type) : Type := list (list T).

Lemma in_subst1 :
  forall {T x x'} {l:list T},
    x = x' ->
    In x l ->
    In x' l.
  intros; subst; auto.
Qed.
 

Lemma in_subst_l :
  forall {T:Type} (l l':list T) (x:T),
    l = l' ->
    In x l -> In x l'.
  intros; subst; auto.
Qed.

Lemma in_subst_r :
  forall {T:Type} (l l':list T) (x:T),
    l = l' ->
    In x l' -> In x l.
  intros; subst; auto.
Qed.

Lemma nin_subst1 :
  forall {T x x'} {l:list T},
    x = x' ->
    ~In x l ->
    ~In x' l.
  intros; subst; auto.
Qed.
 

Lemma nin_subst_l :
  forall {T:Type} (l l':list T) (x:T),
    l = l' ->
    ~In x l -> ~In x l'.
  intros; subst; auto.
Qed.

Lemma nin_subst_r :
  forall {T:Type} (l l':list T) (x:T),
    l = l' ->
    ~In x l' -> ~In x l.
  intros; subst; auto.
Qed.

Lemma in_subst1_pred :
  forall {T:Type} (l:list T) (x x':T),
    x = x' ->
    (fun l => In x l) =
    (fun l => In x' l).
  intros; subst; auto.
Qed.
  
Lemma in_subst_pred_l :
  forall {T:Type} (l l':list T) (x:T),
    l = l' ->
    (fun x => In x l) =
    (fun x => In x l').
  intros; subst; auto.
Qed.

Lemma in_subst_pred_r :
  forall {T:Type} (l l':list T) (x:T),
    l = l' ->
    (fun x => In x l') =
    (fun x => In x l).
  intros; subst; auto.
Qed.

Lemma nin_subst1_pred :
  forall {T} {x x':T},
    x = x' ->
    (fun l => ~In x  l) = 
    (fun l => ~In x' l).
  intros; subst; auto.
Qed.
 

Lemma nin_subst_pred_l :
  forall {T:Type} {l l':list T},
    l = l' ->
    (fun x => ~In x l) =
    (fun x => ~In x l').
  intros; subst; auto.
Qed.

Lemma nin_subst_pred_r :
  forall {T:Type} (l l':list T) (x:T),
    l = l' ->
    (fun x => ~In x l') =
    (fun x => ~In x l).
  intros; subst; auto.
Qed.

Lemma in_from_cons :
  forall {T} (l:list T) (a x:T),
    In a l ->
    In x (a::l) ->
    In x l.
  intros ? ? ? ? ? h1; destruct h1; subst; auto.
Qed.


Lemma in_not_nil : forall {T:Type} (l:list T) (x:T), 
                     In x l -> l <> nil.
intros; intro; subst; contradiction.
Qed.

Lemma not_nil_cons : forall {T:Type} (l:list T), 
                       l <> nil -> exists (a:T) (l':list T),
                                     l = a::l'.
intros T l h1.
destruct l as [|a l'].  contradict h1. auto.
exists a.  exists l'. auto.
Qed.

Lemma cons_neq_nil :
  forall {T:Type} (l:list T) (a:T),
    a :: l <> nil.
  intros ? ?  ? h1. discriminate h1.
Qed.





Lemma cons_inj : forall {T:Type} (a1 a2:T) (l1 l2:list T), a1::l1 = a2::l2 ->
  a1 = a2 /\ l1 = l2.
intros ? ? ? ? ? h1.
inversion h1. split; trivial.
Qed.

Lemma cons_eq : forall {T:Type} (a1 a2:T) (l1 l2:list T), 
  a1::l1 = a2::l2 <->  a1 = a2 /\ l1 = l2.
intros T a1 a2 l1 l2.
split.
intro h1. apply cons_inj; assumption.
intro h1. destruct h1; subst. reflexivity.
Qed.

  
Lemma ex_in_cons :
    forall {T:Type} (l:list T) (a:T),
      (exists x:T, In x l) ->
      (exists x:T, In x (a::l)).
  intros T l a h1.
  destruct h1 as [x h1].
  exists x. right; auto.
Qed.

Lemma nil_dec :
  forall {T:Type} (l:list T),
    {l = nil} + {exists a:T, In a l}.
  intros T l.
  induction l as [|a l ih].
  left; auto.
  destruct ih as [h1 | h1]. subst.
  right.
  exists a. left. reflexivity.
  right.
  apply ex_in_cons; auto. 
Qed.



Lemma ex_in_not_nil :
  forall {T:Type} (l:list T),
         (exists x:T, In x l) -> l <> nil.
  intros T l h1.
  destruct h1 as [x h1].
  intro. subst.
  contradiction.
Qed.

Lemma nil_dec' :
  forall {T:Type} (l:list T),
    {l = nil} + {l <> nil}.
  intros T l.
  destruct (nil_dec l) as [h1 | h1].
  left; auto.
  apply ex_in_not_nil in h1.
  right; auto.
Qed.

Lemma in_cons2 :
  forall {T} {l:list T} a b x,
    In x (a::l) ->
    In x (a::b::l).
  intros ? ? ? ? ? h1; destruct h1 as [|h1]; subst; auto;
  [left|right;right]; auto.
Qed.


Lemma in_cons_dec :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a x:T),
    In x (a::l) ->
    {x = a} + {In x l}.
  intros T h1 l a x h2.
  destruct (h1 x a) as [h3 | h3].
  left; auto.
  right.
  destruct h2 as [|h2]; subst.
  contradict h3; auto.
  assumption.
Qed.


(*When you can't use [simpl]!*)
Lemma length_nil :
  forall (T:Type),
    length (@nil T) = 0.
  auto.
Qed.


  

Lemma l_ex_dec :
  forall {T:Type} (pfdt:type_eq_dec T)
         (P:T->Prop) (l:list T),
    pred_dec P ->
    {exists x:T, In x l /\ P x} + {~exists x:T, In x l /\ P x}.
  intros T hdt P l.
  induction l as [|a l ih]; simpl. 
  intro hp; right. 
  intro h1.
  destruct h1 as [? h1]; destruct h1; auto.
  intro h1.
  specialize (ih h1).
  destruct ih as [ih | ih].
  left.
  destruct ih as [x ih].
  destruct ih.
  exists x.
  split; try right; auto.
  destruct (h1 a) as [h2 | h2].
  left.
  exists a.
  split; auto.
  right.
  intro h3.
  destruct h3 as [x h4].
  destruct h4 as [h4 h5].
  destruct h4 as [|h4]; subst.
  contradiction.
  contradict ih.
  exists x; auto.
Qed.


Lemma length_cons :
  forall {T:Type} (l:list T) (a:T),
    length (a::l) = S (length l).
  auto.
Qed.

Lemma l_lt_length_cons :
  forall {T:Type} (l:list T) (a:T) (n:nat),
    S n < length (a::l) -> n < length l.
intros T l a n h1.
simpl in h1.
omega.
Qed.

Lemma lt_cons_app_sing :
  forall {T:Type} (l:list T) (a b:T),
    S (length l) < length (a :: l ++ b :: nil).
  intros; simpl; rewrite app_length; simpl; rewrite S_compat; auto with arith.
Qed.

Lemma cons_app :
  forall {T} (l l':list T) (a:T),
    a::(l++l') =
    (a::l)++l'.
  auto.
Qed.


Lemma length_cons_eq :
  forall {T:Type} {l l':list T} {a:T},
    length (a::l) = length l' ->
    length l = length (tl l').
  intros T l l' a h2.
  simpl in h2.
  destruct l' as [|a' l']; simpl.
  simpl in h2.
  discriminate.
  simpl in h2.
  apply S_inj in h2; auto.
Qed.

Lemma length_cons_eq_iff :
  forall {T:Type} (l l':list T) (a:T),
    l' <> nil ->
    (length (a::l) = length l' <->
    length l = length (tl l')).
  intros T l l' a h1.
  split.
  apply length_cons_eq.
  intro h2.
  destruct l' as [|a' l']; simpl.
  contradict h1; auto.
  simpl in h2; f_equal; auto.
Qed.

  
Lemma lt_length :
  forall {T:Type} (l:list T) (n:nat),
    n < length l ->
    l <> nil.
  intros T l n h1 h2.
  subst.
  simpl in h1.
  apply lt_n_0 in h1.
  contradiction.
Qed.

Lemma length_eq_S :
  forall {T:Type} (l:list T) (n:nat),
    length l = S n ->
    l <> nil.
  intros; intro; subst; discriminate.
Qed.

Lemma length_eq0 :
  forall {T:Type} (l:list T),
    length l = 0 ->
    l = nil.
  intros T l.
  destruct l as [|a l]; auto.
  intro h1. simpl in h1.
  discriminate h1.
Qed.

Lemma not_nil_lt :
  forall {T:Type} (l:list T),
    l <> nil ->
    0 < length l.
  intros T l.
  destruct (zerop (length l)) as [h2 |]; auto.
  apply length_eq0 in h2.
  contradiction.
Qed.

Lemma not_nil_pred_lt :
  forall {T:Type} (l:list T),
    l <> nil ->
    pred (length l) < length l.
  intros T l h1.
  destruct (zerop (length l)) as [h2 |]; auto.
  apply length_eq0 in h2; subst.
  contradict h1; auto.
  apply (lt_pred_n _ _ l0).
Qed.


Lemma length_functional :
  forall {T:Type} (l l':list T),
    l = l' ->
    length l = length l'.
  intros; subst; auto.
Qed.

Lemma le_length_subst :
  forall {T:Type} (l l':list T)
         (m:nat),
    l = l' ->
    m <= length l ->
    m <= length l'.
  intros; subst; auto.
Qed.


Lemma lt_length_subst :
  forall {T:Type} (l l':list T)
         (m:nat),
    l = l' ->
    m < length l ->
    m < length l'.
  intros; subst; auto.
Qed.


Lemma remove_not_in' : forall {T:Type} (pfdt:type_eq_dec T)
                              (l:list T) (a:T),
                        ~ In a l -> l = remove pfdt a l.
intros T h0 l a h1.
induction l. simpl. auto.
simpl.
destruct (h0 a a0).
subst.
pose proof (in_eq a0 l). contradiction.
f_equal.
simpl in h1.
assert (h2: ~ In a l). tauto.
apply IHl; auto.
Qed.



Lemma remove_hd_no_dup : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (def:T), NoDup l -> 
    remove pfdt (hd def l) l = tl l.
intros T h0 l.
induction l as [|a l h1].
simpl. auto.
simpl.
intros def h2.  
destruct (h0 a a) as [h5 | h6]. 
inversion h2 as [|x l' h3 h4]. subst.
symmetry.
apply remove_not_in'; auto.
contradiction h6.
reflexivity.
Qed.





Lemma hd_indep :
  forall {T:Type} (l:list T) (d d':T),
    l <> nil ->
    hd d l = hd d' l.
  intros ? l; destruct l; auto; intros ? ? h1; try contradict h1; auto.
Qed.

Lemma hd_remove_eq : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (r:T),
    r <> hd r l ->
    hd r (remove pfdt r l) = hd r l.
intros T h0 l.
destruct l as [|a l]; simpl; auto.
intros r h1.
destruct (h0 r a) as [h5 | h6].
subst. 
contradict h1; auto.  
simpl. reflexivity.
Qed.


Lemma in_hd_not_nil : forall {T:Type} (pfdt:type_eq_dec T)
                             (l:list T) (def:T),
                l <> nil -> In (hd def l) l.
intros T pfd l.
induction l as [|a l' h1].
intros def h2.
contradict h2. reflexivity.
simpl.
intros def h2.
left. reflexivity.
Qed.


Lemma hd_map : 
  forall {T U:Type} (f:T->U) (def:T) (l:list T),
    hd (f def) (map f l) = f (hd def l).
intros T U f def l.
induction l as [|a l h1]; simpl; auto.
Qed.                    


Definition hd' {T:Type} (l:list T) : l <> nil -> T :=
  match l with
    | nil => (fun pf => False_rect T (pf eq_refl))
    | a::l' => (fun _ => a)
  end.


Lemma hd_compat' :
  forall {T:Type} (l:list T) (pf: l <> nil),
    let a := hd' l pf in
    l = a::tl l.
  intros T l.
  destruct l as [|a l]; auto.
  intro h1. contradict h1; auto.
Qed.

Lemma hd_compat :
  forall {T:Type} (l:list T) (pf: l <> nil),
    hd' l pf =
    hd (hd' l pf) l.
  intros T l.
  destruct l as [|a l].
  intro h1.
  contradict h1; auto.
  intros h1.
  simpl; auto.
Qed.

  
Lemma hd_functional :
  forall {T:Type} (l l':list T) (pf:l = l')
         (pf:l <> nil) (pf':l' <> nil),
    hd' l pf =
    hd' l' pf'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.


Lemma hd_functional' :
  forall {T:Type} (l:list T) (pf1 pf2:l <> nil),
    hd' l pf1 = hd' l pf2.
  intros; f_equal; apply proof_irrelevance.
Qed.

Lemma hd_hd_compat' :
  forall {T:Type} (l:list T) (pf:l <> nil) (def:T),
    hd def l = hd' l pf.
  intros T l.
  induction l as [|a l h0]; simpl.
  intro h1. contradict h1; auto.
  intros; auto.
Qed.


Lemma in_hd' :
  forall {T:Type} (l:list T) (pf:l <> nil),
    In (hd' l pf) l.
  intros T l h1.
  pose proof (hd_compat' l h1) as h2. simpl in h2.
  rewrite  h2 at 2.
  left.
  reflexivity.
Qed.


Lemma hd_cons' :
  forall {T:Type} (l:list T) (a:T) (pf:a::l <> nil),
    hd' (a::l) pf = a.
  intros T l a h1.
  pose proof (hd_compat' (a::l) h1) as h2. simpl in h2.
  apply cons_inj in h2.
  destruct h2; auto.
Qed.

Lemma hd_app' :
  forall {T:Type} (l l':list T) (pf:l <> nil) (pf':l ++ l' <> nil),
    hd' (l ++ l') pf' =
    hd' l pf.
  intros T l.
  destruct l as [|a l]; simpl.
  intros ? h1; contradict h1; auto.
  auto.
Qed.

Lemma removelast_not_nil :
  forall {T:Type} (l:list T) (pf:removelast l <> nil),
    l <> nil.
  intros ? ?  h1; intro; subst; simpl in h1; contradict h1; auto.
Qed.

Lemma hd_removelast' :
  forall {T:Type} (l:list T) (pf:removelast l <> nil),
    let pf' := removelast_not_nil _ pf in 
    hd' (removelast l) pf = hd' l pf'.
  intros T l h1; simpl.
  destruct l as [|c l]; simpl in h1.
  contradict h1; auto.
  simpl.
  destruct l as [|d l].
  contradict h1; auto.
  simpl; auto.
Qed.

           
Lemma impl_eq_cons :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T)
         (pf:l <> nil),
    tl l = l' ->
    l = hd' l pf ::l'.
  intros T h1 l.
  destruct l as [|a l]; simpl.
  intros ? h2.
  contradict h2; auto.
  intros l' h2 h3.
  subst; auto.
Qed.

Lemma impl_eq_cons' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T)
         (pf:l <> nil),
    hd' l pf = a ->
    tl l = l' ->
    l = a::l'.
  intros; subst; apply impl_eq_cons; auto.
Qed.

Lemma cons_eq_iff :
  forall {T:Type} (l l':list T) (a:T),
    l = a::l' <->
    exists pf:l <> nil,
      a = hd' l pf /\
      l' = tl l.
  intros T l l' a; split; intro h1; subst.
  exists (cons_neq_nil _ _).
  simpl; auto.
  destruct h1 as [h1 h2].
  destruct h2; subst.
  apply hd_compat'.
Qed.



Lemma in_or' :
  forall {T:Type} (l:list T) (x:T)
         (pf:In x l),
    x = hd' _ (in_not_nil _ _ pf) \/ In x (tl l).
  intros T l x h1.
  destruct l as [|a l].
  contradict h1.
  simpl in h1; simpl.
  destruct h1; subst; [left; auto | right; auto].
Qed.


Lemma remove_hd_no_dup' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (pf:l <> nil),
    NoDup l -> 
    remove pfdt (hd' l pf) l = tl l.
  intros T h1 l h2 h3.
  pose proof (not_nil_cons _ h2) as h4. 
  destruct h4 as [a h4]. destruct h4 as [l' h4].
  rewrite <- (hd_hd_compat' _ _ a).
  apply remove_hd_no_dup; auto.
Qed.


Lemma no_dup_cons : forall {T:Type} (l:list T) (a:T),
                      NoDup (a::l) -> NoDup l.
intros T l a h1.
rewrite <- app_nil_l in h1.
pose proof (NoDup_remove_1 _ _ _ h1) as h2.
rewrite app_nil_l in h2.
assumption.
Qed.

Lemma no_dup_cons2 : forall {T:Type} (l:list T) (a b:T),
                      NoDup (a::b::l) -> NoDup l.
intros T l a b h1.
rewrite <- app_nil_l in h1.
pose proof (NoDup_remove_1 _ _ _ h1) as h2.
pose proof (NoDup_remove_1 _ _ _ h2) as h3.
rewrite app_nil_l in h3.
assumption.
Qed.


Lemma no_dup_cons_nin : forall {T:Type} (l:list T) (a:T),
                          NoDup (a::l) -> ~In a l.
intros T l a h1.
rewrite <- app_nil_l in h1.
pose proof (NoDup_remove_2 _ _ _ h1) as h2.
rewrite app_nil_l in h2.
assumption.
Qed.

Lemma no_dup_cons2_nin : forall {T:Type} (l:list T) (a b:T),
                          NoDup (a::b::l) -> ~In a l.
intros T l a b h1.
rewrite <- app_nil_l in h1. 
pose proof (NoDup_remove_2 _ _ _ h1) as h2.
pose proof (NoDup_remove_1 _ _ _ h1) as h3.
pose proof (NoDup_remove_2 _ _ _ h3) as h4.
intro h5.
assert (h6: a <> b).
intro; subst.
rewrite app_nil_l in h2. 
contradict h2. left; auto.
contradict h2.
right; auto.
Qed.


Lemma no_dup_cons2' :
  forall {T:Type} (l:list T) (a b:T),
    NoDup (a::b::l) -> NoDup (a::l).
  intros T l a b h1.
  constructor.
  apply no_dup_cons_nin in h1; auto.
  contradict h1; right; auto.
  apply no_dup_cons2 in h1; auto.
Qed.


Lemma no_dup_cons2_neq :
  forall {T:Type} (l:list T) (a b:T),
    NoDup (a::b::l) -> a <> b.
  intros T l a b h1 h2; subst.
  apply no_dup_cons_nin in h1.
  contradict h1; left; auto.
Qed.
          
Lemma nin_cons1 :
  forall {T:Type}
         {l:list T} {a x:T},
    ~In x (a::l) ->
    x <>  a.
  intros; intro; subst; destruct H; auto; left; auto.
Qed.


Lemma nin_cons2 :
  forall {T:Type} 
         {l:list T} {a x:T},
    ~In x (a::l) ->
    ~In x l.
  intros; contradict H; right; auto.
Qed.

Lemma hd_firstn' :
  forall {T:Type} (l:list T) (m:nat)
         (pf:l <> nil) (pf':firstn m l <> nil),
    hd' l pf =
    hd' (firstn m l) pf'.
  intros T l.
  destruct l as [|a l]; simpl.
  intros ? h2; contradict h2; auto.
  intros m h2 h3.
  destruct m as [|m].
  simpl in h3.
  contradict h3; auto.
  simpl; auto.
Qed.


Lemma length_eq_1 :
  forall {T:Type} (l:list T)
         (pf:length l = 1),
    l = hd' l (lt_length _ _ (lt_eq_trans _ _ _ (lt_0_Sn 0) (eq_sym pf)))::nil.
  intros T l.
  destruct l as [|a l]; simpl; subst.
  intro; discriminate.
  intros h1.
  apply S_inj in h1; subst.
  apply length_eq0 in h1; subst; auto.
Qed.


Lemma sing_eq_iff :
  forall {T:Type} (l:list T) (a:T),
    l = a::nil <->
    exists pf:l <> nil,
      hd' l pf = a /\ tl l = nil.
  intros T l a; split; intro h1; subst.
  exists (cons_neq_nil _ _); simpl; split; auto.
  destruct h1 as [h1 [h2 h3]].
  rewrite (hd_compat' _ h1); f_equal; auto.
Qed.

Lemma tl_eq :
  forall {T:Type} (l:list T),
    tl l = skipn 1 l.
  intros T l.
  destruct l; auto.
Qed.

Lemma tl_eq_nil_iff :
  forall {T:Type} (l:list T),
    tl l = nil <->
    l = nil \/
    exists x:T,
      l = x :: nil.
  intros T l.
  destruct l as [|a l]; simpl.
  tauto.
  split.
  intro; subst.
  right.
  exists a; auto.
  intro h1.
  destruct h1 as [h1 | h1].
  discriminate.
  destruct h1 as [x h1].
  apply cons_inj in h1.
  destruct h1; auto.
Qed.
                                      

Lemma app_sing :
  forall {T:Type} (l:list T) (a:T),
    a :: nil ++ l = a::l.
  intros; auto.
Qed.

Lemma app_sing_app :
  forall {T:Type} (l l':list T) (a:T),
    (l ++ a::nil) ++ l' =
    l ++ a::l'.
  intros; rewrite <- app_assoc; simpl; auto.
Qed.

Lemma in_app_l :
  forall {T:Type} {l:list T} {x:T},
    In x l ->
    forall (l':list T),
      In x (l++l').
  intros; rewrite in_app_iff; left; auto.
Qed.

Lemma in_app_r :
  forall {T:Type} {l':list T} {x:T},
    In x l' ->
    forall (l:list T),
      In x (l++l').
    intros; rewrite in_app_iff; right; auto.
Qed.

Lemma app_not_nil_l :
  forall {T:Type} (l l':list T),
    l <> nil ->
    l ++ l' <> nil.
  intros T l l' h1.
  destruct l as [|a l].
  contradict h1; auto.
  rewrite <- app_comm_cons.
  intro h2.
  discriminate.
Qed.

Lemma app_not_nil_r :
  forall {T:Type} (l l':list T),
    l' <> nil ->
    l ++ l' <> nil.
  intros T l l' h1.
  destruct l as [|a l]; simpl; auto.
  intro; discriminate.
Qed.

Lemma lt_app :
  forall {T:Type} (l l':list T)
         (i:nat),
    i < length l ->
    i < length (l++l').
  intros T l l' i h1.
  rewrite app_length; auto with arith.
Qed.

         
Lemma sing_not_nil :
  forall {T:Type} (a:T),
    a :: nil <> nil.
  intros; discriminate.
Qed.


Lemma app_inj_eq_length :
  forall {T:Type} (l1 l2 l1' l2':list T),
    l1 ++ l2 = l1' ++ l2' ->
    length l1 = length l1' ->
    l1 = l1' /\ l2 = l2'.
  intros T l1.
  induction l1 as [|a1 l1 ih]; simpl.
  intros ? ? ? ? h1; subst.
  pose proof length_eq0 _ (eq_sym h1); subst.
  split; auto.
  intros l2 l1'. revert l2.
  destruct l1' as [|a1' l1']; simpl.
  intros l2 l2' h3. subst.
  intros; discriminate.
  intros l2 l2' h1.
  inversion h1 as [h3]; subst.
  clear h1.
  intro h1.
  apply S_inj in h1.
  specialize (ih _ _ _ H h1).
  destruct ih; subst.
  auto.
Qed.

Lemma app_nin_length :
  forall {T} (l0 l0' l1 l1':list T) b,
    ~In b l0 ->
    ~In b l0' ->
    l0  ++ b::l1 =
    l0' ++ b::l1' ->
    length l0 =
    length l0'.
  intros T l0.  
  induction l0 as [|a0 l0 ih]; simpl. 
  intro l0'. 
  destruct l0' as [|a0' l0']; simpl; auto.
  intros l1 l1' b h1 h2 h3.
  inversion h3; subst.
  contradict h3.
  contradict h2; left; auto.
  intro l0'.
  destruct l0' as [|a0' l0'].
  intros l1 l1' b h1 h2 h3.
  simpl in h3.
  inversion h3; subst.
  contradict h1; left; auto. 
  intros l1 l1' b h1 h3 h4; simpl in h4;  f_equal.
  specialize (ih l0' l1 l1' b).
  hfirst ih.
  contradict h1; right; auto.
  specialize (ih hf).
  hfirst ih.
  contradict h3; right; auto.
  specialize (ih hf0).
  inversion h4; subst.
  specialize (ih H1); simpl; f_equal; auto.
Qed.


  
  
Lemma app_nin_eq :
  forall {T} (l0 l0' l1 l1':list T) b,
    ~In b l0 ->
    ~In b l0' ->
    l0  ++ b::l1 =
    l0' ++ b::l1' ->
    l0 = l0' /\ l1 = l1'.
  intros T l0 l0' l1 l1' b h1 h2 h3.
  apply app_inj_eq_length in h3.
  destruct h3 as [? h3]; subst.
  inversion h3; tauto.
  apply app_nin_length in h3; auto.
Qed.
     


Lemma not_app_same_eq_r :
  forall {T:Type} (l l':list T),
    l' <> nil ->
    ~ l = l' ++ l.
  intros T l l' h1 h2.
  pose proof (f_equal (@length T) h2) as h3.
  rewrite app_length in h3.
  apply not_nil_lt in h1.
  omega.
Qed.

Lemma not_app_same_eq_l :
  forall {T:Type} (l l':list T),
    l' <> nil ->
    ~ l = l ++ l'.
  intros T l l' h1 h2.
  pose proof (f_equal (@length T) h2) as h3.
  rewrite app_length in h3.
  apply not_nil_lt in h1.
  omega.
Qed.
  

Lemma app_sing_eq_app_cons_cons :
  forall {T} (l l' l0:list T) (a b:T),
    ~In a (b::l0) ->
    l ++ a::nil = l' ++ a :: b :: l0 ->
    In a l.  
  intros T l; induction l as [|a l ih]; simpl; intros l' l0 b c h0 h1.
  destruct l' as [|d l']. 
  simpl in h1.
  inversion h1. 
  simpl in h1. 
  inversion h1; subst.
  pose proof (app_not_nil_r l' (d::c::l0) (cons_neq_nil _ _)).
  rewrite <- H1 in H.
  contradict H; auto.
  destruct l' as [|a' l'].
  simpl in h1.
  inversion h1; subst.
  left; auto.
  simpl in h1.
  inversion h1; subst.
  apply ih in H1.
  right; auto.
  assumption.
Qed.  


Lemma removelast_eq :
  forall {T:Type} (l:list T),
    removelast l =
    firstn (pred (length l)) l.
  intros T l.
  induction l as [|a l ih]; simpl; auto.
  destruct l as [|b l]; simpl; auto.
  f_equal.
  simpl in ih; auto.
Qed.


Lemma removelast_in :
  forall {T:Type} (l:list T) (x:T),
    In x (removelast l) ->
    In x l.
  intros T l x h1.
  erewrite (app_removelast_last x).
  apply in_app_l; auto.
  intro; subst.
  simpl in h1; auto.
Qed.


Lemma incl_removelast :
  forall {T:Type} (l:list T),
    incl (removelast l) l.
  intros T l.
  red.
  intros a h1.
  apply removelast_in; auto.
Qed.

Lemma length_removelast :
  forall {T:Type} (l:list T),
    length (removelast l) = pred (length l).
  intros ? l;
  induction l as [|? l ih]; simpl; try simpl in ih; auto; destruct l;simpl; auto.
Qed.

  
Lemma removelast_cons_cons : forall {T:Type} (l:list T)
                               (a a':T),
                          removelast (a::a'::l) = a::removelast (a'::l).
intros; simpl; auto.
Qed.

Lemma removelast_cons_cons_neq_nil :
  forall {T:Type} (l:list T) (a a':T),
    removelast (a::a'::l) <> nil.
  intros; eapply cons_neq_nil; apply removelast_cons_cons.
Qed.


Lemma firstn_removelast :
  forall {T:Type} (l:list T) (n:nat),
    n <= length (removelast l) ->
    firstn n (removelast l) =
    firstn n l.
  intros T l.
  induction l as [|a l ih]; simpl; auto.
  intros n h2.
  destruct l as [|b l].
  simpl in h2.
  apply le0 in h2; subst; simpl; auto.
  simpl in h2.
  destruct n as [|n].
  simpl; auto.
  apply le_S_n in h2.
  specialize (ih _ h2).
  simpl.
  f_equal.
  simpl in ih; auto.
Qed.

Lemma last_cons :
  forall {T:Type} (l:list T) (a:T),
    last (a::l) a = last l a.
  intros; simpl; destruct l; auto.
Qed.

Lemma last_cons' :
  forall {T:Type} (l:list T) (a d:T),
    l <> nil -> 
    last (a::l) d = last l d.
  intros; simpl; destruct l; auto.
  contradict H; auto.
Qed.


Lemma in_last :
  forall {T:Type} (l:list T) (d:T),
    l <> nil ->
    In (last l d) l.
  intros ? l; induction l as [|c l ih]; simpl; auto.
  intros d h1.
  destruct l as [|l].
  left; auto.
  right; apply ih; auto.
  apply cons_neq_nil.
Qed.


Lemma last_indep :
  forall {T:Type} (l:list T) (pf:l <> nil) (a b:T),
    last l a = last l b.
  intros T l; induction l as [|c l ih]; intro h1.
  contradict h1; auto.
  intros; simpl; auto.
  destruct l as [|d l]; auto.
  specialize (ih (cons_neq_nil _ _)).
  apply ih.
Qed.

Lemma last_app_sing :
  forall {T:Type} (l:list T) (a d:T),
    last (l++a::nil) d = a.
  intros T l. 
  induction l as [|c l ih]; simpl; auto.
  intros a d.
  pose proof (in_app_r (in_eq a nil) l) as h2.
  rewrite (hd_compat' _ (in_not_nil _ _ h2)).
  simpl.
  specialize (ih a d).
  destruct l; auto.
Qed.


Lemma removelast_app_sing :
  forall {T:Type} (l:list T) (a:T),
    removelast (l++a::nil) = l.
  intros T l. 
  induction l as [|a l ih]; simpl; auto.
  intro b.
  pose proof (in_app_r (in_eq b nil) l) as h2.
  rewrite (hd_compat' _ (in_not_nil _ _ h2)).
  f_equal.
  rewrite <- hd_compat'.
  apply ih.
Qed.


Lemma count_occ_nil :
  forall (T:Type) (pfdt:type_eq_dec T) (x:T),
    count_occ pfdt nil (A := T) x = 0.
  auto.
Qed.

Lemma count_occ_sing_eq :
  forall (T:Type) (pfdt:type_eq_dec T) (x:T),
    count_occ pfdt (x::nil) x = 1.
  simpl; intros; rewrite eq_dec_same at 1; auto.
Qed.

Lemma count_occ_sing_neq :
  forall (T:Type) (pfdt:type_eq_dec T) (x a:T),
    x <> a ->
    count_occ pfdt (a::nil) x = 0.
  simpl; intros;  lr_if_goal'; subst; auto;
  contradict H; auto.
Qed.


Lemma count_occ_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (m:nat),
    m < count_occ pfdt l x -> In x l.
  intros ? h1 ? ? ? ? ; rewrite (count_occ_In h1); omega.
Qed.


Lemma count_occ_eq_S_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (x:T) (n:nat),
    count_occ pfdt l x = S n ->
    In x l.
  intros.
  rewrite count_occ_In.
  rewrite H at 1.
  auto with arith.
Qed.

Lemma count_occ_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    ~ In x l -> count_occ pfdt l x = 0.
  intros T h1 l x h2.
  rewrite (count_occ_In h1) in h2; omega.
Qed.


Lemma count_occ_le :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    count_occ pfdt l x <= length l.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; auto with arith.
  intro x.
  destruct (h1 a x) as [| h3]; subst.  
  apply le_n_S.
  apply ih.
  specialize (ih x).
  auto with arith.
Qed.

Lemma count_occ_app :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T),
    count_occ pfdt (l++l') x =
    count_occ pfdt l x + count_occ pfdt l' x.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; auto.
  intros l' x.
  destruct (h1 a x) as [|h2]; subst.
  rewrite plus_comm.
  rewrite plus_S_shift_r.
  f_equal.
  rewrite plus_comm.
  apply ih.
  apply ih.
Qed.
 
Lemma weak_bound_count_occ_app :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (occ:nat),
    count_occ pfdt l x <= occ < count_occ pfdt (l++l') x ->
    occ - (count_occ pfdt l x) < count_occ pfdt l' x.
  intros T h1 l l' x occ h2.
  rewrite count_occ_app in h2.
  omega.
Qed.

Lemma count_occ_firstn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    count_occ pfdt (firstn n l) x =
    count_occ pfdt l x -
    count_occ pfdt (skipn n l) x.
  intros T h1 l x n.
  rewrite <- (firstn_skipn n l) at 2.
  rewrite count_occ_app; auto with arith.
  rewrite <- plus_minus_assoc; auto.
  rewrite minus_same; auto.
Qed.


Lemma count_occ_skipn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    count_occ pfdt (skipn n l) x =
    count_occ pfdt l x -
    count_occ pfdt (firstn n l) x.
  intros T h1 l x n.
  rewrite <- (firstn_skipn n l) at 2.
  rewrite count_occ_app; auto with arith.
Qed.

Lemma count_occ_last :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (d:T),
    l <> nil ->
    count_occ pfdt l (last l d) =
    S (count_occ pfdt (removelast l) (last l d)).
  intros T h1 l d h2.
  rewrite (app_removelast_last d h2) at 1.
  rewrite count_occ_app.
  simpl.
  destruct (h1 (last l d) (last l d)).
  rewrite S_compat; auto.
  contradict n; auto.
Qed.

Lemma count_occ_removelast_neq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    l <> nil ->
    x <> last l x ->
    count_occ pfdt (removelast l) x =
    count_occ pfdt l x.
  intros T h1 l x h2 h3.
  rewrite (app_removelast_last x h2) at 2.
  rewrite count_occ_app.
  rewrite plus_n_O at 1.
  f_equal.
  simpl.
  destruct (h1 (last l x) x) as [h4 | h4].
  rewrite h4 in h3.
  contradict h3; auto.
  reflexivity.
Qed.


Lemma count_occ_remove_neq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x:T),
    In x l ->
    a <> x ->
    count_occ pfdt (remove pfdt x l) a  =
    count_occ pfdt l a.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros b x h2 h3.
  destruct (h1 x a) as [h4 | h4]; subst.
  destruct (h1 a b) as [h4 | h4]; subst.
  contradict h3; auto.
  destruct (in_dec h1 a l) as [h5 | h5].
  auto.
  rewrite <- remove_not_in'; auto.
  destruct (h1 a b) as [|h5]; subst.
  simpl.
  destruct (h1 b b) as [h5 | h5].
  f_equal; apply ih.
  destruct h2; subst; auto.
  contradict h3; auto.
  assumption.
  contradict h5; auto.
  simpl.
  destruct (h1 a b); try contradiction.
  apply ih.
  destruct h2; subst; auto.
  contradict h4; auto.
  assumption.
Qed.

Lemma count_occ_remove_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    In a l ->
    count_occ pfdt (remove pfdt a l) a  = 0.
  intros T h1 l a h2.
  apply not_lt0_iff.
  intro h3.
  rewrite <- (count_occ_In h1 (remove h1 a l) a) in h3.
  apply remove_In in h3; auto.
Qed.

Lemma count_occs0 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    (forall x:T, count_occ pfdt l x = 0) ->
    l = nil.
  intros T h1 l h2.
  destruct l as [|a l]; auto.
  specialize (h2 a).
  simpl in h2.
  destruct (h1 a a) as [|h3].
  discriminate.
  contradict h3; auto.
Qed.

Lemma count_occ_gt0 :
  forall (l:list nat) (pf:forall n:nat, In n l -> n > 0),
    count_occ nat_eq_dec l 0 = 0.
  intro l.
  induction l as [|a l ih]; simpl.
  intros; auto.
  intros h1.
  destruct (nat_eq_dec a 0) as [|h2]; subst.
  specialize (h1 0 (or_introl eq_refl)).
  apply lt_irrefl in h1.
  contradiction.
  apply ih.
  intros n h3.
  apply h1.
  right; auto.
Qed.


(*Counts all the elements in [l] for which [pf] is true.*)
Fixpoint count_occ_prop
           {T:Type} 
           (P : T -> Prop)
           (pf : forall x, {P x} + {~ P x})
           (l:list T) : nat :=
  match l with
    | nil => 0
    | a :: l' => let n := count_occ_prop P pf l' in
                 if (pf a) then S n
                 else n
  end. 

         


Lemma skipn_nil :
  forall (T:Type) (n:nat),
    skipn n (@nil T) = nil.
  intros ? n.
  destruct n; auto.
Qed.

Lemma skipn0 :
  forall {T:Type} (l:list T),
    skipn 0 l = l.
  auto.
Qed.
  
Lemma skipnS_cons :
  forall {T:Type} (l:list T) (a:T) (n:nat),
    skipn (S n) (a::l)= skipn n l.
  intros; simpl; auto.
Qed.


Lemma skipn_eq_nil_iff :
  forall {T:Type} (l:list T) (n:nat),
    skipn n l = nil <->
    (l = nil \/ n >= length l).
  intros T l.
  induction l as [|a l ih]; simpl.
  intuition.
  apply skipn_nil.
  apply skipn_nil.
  intros n; split; intro h1.
  destruct n as [|n].
  simpl in h1.
  discriminate.
  rewrite skipnS_cons in h1.
  right.
  rewrite ih in h1.
  destruct h1; subst; auto with arith.
  destruct h1 as [|h1]; try discriminate.
  destruct n as [|n].
  omega.
  apply le_S_n in h1.
  rewrite skipnS_cons.
  rewrite ih.
  right; auto.
Qed.



Lemma skipn_length :
  forall {T:Type} (l:list T),
    skipn (length l) l = nil.
  intros ? l; induction l; simpl; auto.
Qed.

Lemma skipn_length_app :
  forall {T:Type} (l l':list T),
    skipn (length l) (l++l') =
    l'.
  intros ? l; induction l as [|l ih]; simpl; auto.
Qed.

Lemma skipn_length_app_plus :
  forall {T:Type} (l l':list T) (n:nat),
    skipn (length l + n) (l++l') =
    skipn n l'.
  intros ? l; induction l as [|l ih]; simpl; auto.
Qed.

Lemma skipn_tl :
  forall {T:Type} (l:list T) (n:nat),
    skipn n (tl l) =
    skipn (S n) l.
  intros T l.
  destruct l as [|a l]; simpl; auto.
  intro; rewrite skipn_nil; auto.
Qed.


Lemma firstn_nil :
  forall {T:Type} (n:nat),
    firstn n nil = nil (A:=T).
  intros T n.
  induction n as [|n h1]; auto.
Qed.

Lemma firstn_sing :
  forall {T:Type} (n:nat) (a:T),
    firstn (S n) (a::nil) = a::nil.
  simpl; intros; f_equal; rewrite firstn_nil; auto.
Qed.


Lemma firstn_eq_nil :
  forall {T:Type} (n:nat) (l:list T),
    firstn n l = nil ->
    {n = 0} + {l = nil}.
  intros T n.
  destruct n as [|n]; simpl.
  intros; left; auto.
  intros l h1.
  right.
  destruct l; auto.
  discriminate.
Qed.


Lemma firstn_S_neq_nil :
  forall {T:Type} (l:list T) (n:nat),
    S n <= length l ->
    firstn (S n) l <> nil.
  intros T l n h1; simpl.
  destruct l. simpl in h1. omega.
  intro h2.
  discriminate h2.
Qed.

  
Lemma firstn_cons :
  forall {T:Type} (l:list T) (n:nat) (a:T),
    0 < n ->
    firstn n (a::l) = a :: firstn (pred n) l.
  intros T l n. revert l.
  induction n as [|n h1]; simpl; intros; try omega; auto.
Qed.

Lemma skipn_cons :
  forall {T:Type} (l:list T) (n:nat) (a:T),
    0 < n ->
    skipn n (a::l) = skipn (pred n) l.
  intros T l n. revert l.
  induction n as [|n h1]; simpl; intros; try omega; auto.
Qed.


Lemma in_firstn :
  forall {T:Type} (l:list T) (m:nat) (x:T),
    In x (firstn m l) ->
    In x l.
  intros ? l m. revert l. 
  induction m; simpl.
  intros; contradiction. 
  intros l x h1. 
  destruct l; auto.
  destruct h1; subst; [left; auto | right; auto].
Qed.


Lemma in_skipn :
  forall {T:Type} (l:list T) (m:nat) (x:T),
    In x (skipn m l) ->
    In x l.
  intros ? l m. revert l.
  induction m; simpl; auto.
  intros; destruct l; auto.
  right; auto.
Qed.

Lemma nin_firstn :
  forall {T:Type} (l:list T) (m:nat) (x:T),
    ~In x l ->
    ~In x (firstn m l).
  intros; contradict H; eapply in_firstn; apply H.
Qed.
   
Lemma nin_skipn :
  forall {T:Type} (l:list T) (m:nat) (x:T),
    ~In x l ->
    ~In x (skipn m l).
  intros; contradict H; eapply in_skipn; apply H.
Qed.
  
  

  
Lemma firstn_app_le :
  forall {T:Type} (l l':list T) (m:nat),
    m <= length l ->
    firstn m (l ++ l') = firstn m l.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros l' m h1.
  apply le0 in h1; subst.
  simpl; auto.
  intros l' m h1.
  destruct m as [|m]; simpl; auto.
  apply le_S_n in h1.
  f_equal; auto.
Qed.


Lemma firstn_app_eq :
  forall {T:Type} (l l':list T),
    firstn (length l) (l++l') = l.
  intros T l.
  induction l as [|a l ih]; simpl; auto.
  intro l'.
  f_equal.
  apply ih; auto.
Qed.

Lemma firstn_app_ge :
  forall {T:Type} (l l':list T) m,
    m >= length l ->
    firstn m (l++l') = l ++ (firstn (m - length l) l').
  intros ? l; induction l as [|a l ih]; simpl; auto; intro l'; destruct l'; intros m ?; destruct m as [|m]; try rewrite <- minus_n_O; auto; try omega.
  simpl; f_equal; eapply ih; auto with arith.
  rewrite app_comm_cons.
  simpl.
  f_equal.
  eapply ih; auto with arith.
Qed.


Lemma firstn_length :
  forall {T:Type} (l:list T),
    firstn (length l) l = l.
  intros T l.
  rewrite <- (app_nil_r l) at 2.
  apply firstn_app_eq.
Qed.

Lemma firstn_length_eq :
  forall {T:Type} (l l':list T),
    firstn (length l) l' = l ->
    l' = l ++ (skipn (length l) l').
  intros T l l' h1.
  rewrite <- (firstn_skipn (length l) l').
  rewrite h1.
  f_equal.
  rewrite skipn_length_app; auto.
Qed.

Lemma nin_eq_firstn_length :
    forall {T} l l' (a:T),
      ~In a l ->
      l = firstn (length l) (a::l') ->
      l = nil.
  intros T l.
  destruct l as [|b l]; simpl; auto.
  intros l' a h1 h2; inversion h2; revert h2 h1 H1; subst.
  intros ? h1.
  contradict h1; left; auto.
Qed.


Lemma firstn_length_ge :
  forall {T:Type} (l:list T) (m:nat),
    length l <= m ->
    firstn m l = l.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros; rewrite firstn_nil; auto.
  intros m h1.
  rewrite firstn_cons.
  f_equal.
  apply ih.
  omega.
  omega.
Qed.

Lemma firstn_lt_nest :
  forall {T:Type} (l:list T) (m n:nat),
    m < n ->
    firstn m l = firstn m (firstn n l).
  intros T l m n. revert m l.
  induction n as [|n ih].
  intros; omega.
  intros m l h1.
  destruct l as [|a l].
  simpl; auto.
  destruct m as [|m]; simpl; auto.
  apply lt_S_n in h1.
  f_equal; auto.
Qed.


Lemma app_cons_assoc :
  forall {T} (l l':list T) (a:T),
    l ++ a :: l' = (l++a::nil) ++ l'.
  intros; rewrite <- app_assoc; auto.
Qed.


Lemma length_from_app_cons : 
  forall {T} (l l0 l' l0':list T) (a:T),
    length (l ++ a :: l') = length (l0 ++ a :: l') ->
    length (l ++ l') = length (l0 ++ l').
  intros ? ? ? ? ? ? h1;
  do 2 rewrite app_length in h1; simpl in h1;
  do 2 rewrite plus_S_shift_r in h1;
  do 2 rewrite <- app_length in h1; auto with arith.
Qed.



  
Lemma app_eq_iff :
  forall {T:Type} (l1 l2 l:list T),
    l1 ++ l2 = l <->
    l1 = (firstn (length l1) l) /\
    l2 = (skipn (length l1) l).
  intros T l1 l2 l.
  split; intro h1.
  pose proof h1 as h1'.
  rewrite <- (firstn_skipn (length l1) l) in h1'.
  apply app_inj_eq_length in h1'; auto.
  rewrite <- h1.
  rewrite firstn_app_eq; auto.
  destruct h1 as [h1 h2].
  rewrite h1, h2.
  apply firstn_skipn.
Qed.

Lemma app_eq_same_iff_l :
  forall {T:Type} (l l':list T),
    l = l ++ l' <->
    l' = nil.
  intros T l l'; split; intro h1.
  destruct (nil_dec' l') as [|h2]; auto.
  apply not_app_same_eq_l in h1; auto; contradiction.
  subst.
  rewrite app_nil_r; auto.
Qed.

Lemma app_eq_same_iff_r :
  forall {T:Type} (l l':list T),
    l = l' ++ l <->
    l' = nil.
  intros T l l'; split; intro h1.
  destruct (nil_dec' l') as [|h2]; auto.
  apply not_app_same_eq_r in h1; auto; contradiction.
  subst.
  rewrite app_nil_l; auto.
Qed.

Lemma app_eq_same_iff2 :
  forall {T:Type} (l l' l'':list T),
    l = l' ++ l ++ l'' <->
    l' = nil /\ l'' = nil.
  intros T l l' l''; split; intro h1.
  destruct (nil_dec' l') as [|h2].
  split; auto.
  rewrite e in h1.
  rewrite app_nil_l in h1.
  destruct (nil_dec' l'') as [|h2]; auto.
  apply not_app_same_eq_l in h1.
  contradiction.
  assumption.
  pose proof (f_equal (@length T) h1) as h3.
  do 2 rewrite app_length in h3.
  apply not_nil_lt in h2.
  omega.
  destruct h1; subst.
  rewrite app_nil_l, app_nil_r; auto.
Qed.

    
Lemma eq_in_cons :
  forall {T:Type} (l:list T) (x a:T),
    x = a -> In x (a::l).
  intros; subst; left; auto.
Qed.



Lemma neq_in_cons :
  forall {T:Type} (l:list T) (x a:T),
    x <> a -> In x (a::l) -> In x l.
  intros ? ? ? ? h0 h1; destruct h1; subst; auto.
  contradict h0; auto.
Qed.

  
Lemma in_list2_comm : forall {T:Type} (l:list T) (a x y:T),
                        In a (x::y::l) -> In a (y::x::l).
intros T l a x y h1.
destruct h1 as [h1a | h1b].
subst.
right. left. auto.
destruct h1b as [h2a | h2b].
subst. left. auto.
right. right. auto.
Qed.



Lemma nin_remove_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    ~In a l ->
    remove pfdt a l = l.
  intros T h0 l.
  induction l as [|a l h1]; simpl; auto.
  intros x h2.
  apply not_or_and in h2.
  destruct h2 as [h2 h3].
  destruct (h0 x a) as [h4 | h4]. subst. contradict h2; auto.
  f_equal.
  apply h1; auto.
Qed.

Lemma remove_cons_neq : forall {T:Type} (pfdt:type_eq_dec T)
                               (l:list T) (a b:T),
                          a <> b -> 
                          remove pfdt a (b::l) =
                          b :: remove pfdt a l.
intros T h0 l a b h1.
simpl.
destruct (h0 a b). contradiction. auto.
Qed.

Lemma remove_cons_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T),
    remove pfdt a (a::l) =
    remove pfdt a l.
  intros T h0 l a.
  simpl.
  destruct (h0 a a) as [h1 | h1]; auto.
  contradict h1; auto.
Qed.


Lemma remove_cons_eq_l' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x a:T),
    x = a ->
    remove pfdt x (a::l) =
    remove pfdt x l.
  intros; subst; apply remove_cons_eq; auto.
Qed.


Lemma remove_cons_eq_r' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x a:T),
    x = a ->
    remove pfdt x (a::l) =
    remove pfdt a l.
  intros; subst; apply remove_cons_eq; auto.
Qed.


  
Lemma remove_a_from_2cons_nil : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x y:T),
         remove pfdt a (x::y::l) = nil ->
         remove pfdt a (y::x::l) = nil.
  intros T h0 l a x y h1.
  simpl in h1. simpl.
  destruct (h0 a x); destruct (h0 a y); auto; try discriminate.
Qed.



Lemma remove2_eq : forall {T:Type} (pfdt:type_eq_dec T)
                          (l:list T) (a b:T),
                     remove pfdt a (remove pfdt b l) =
                     remove pfdt b (remove pfdt a l).
intros T h0 l a b.
induction l as [|c l h1].
simpl. auto.
simpl.
destruct (h0 b c) as [hc | hb]; destruct (h0 a c) as [hc' | ha].
subst; auto.
subst.
simpl.
destruct (h0 c c) as [h3 | h4]. auto.
contradict h4. auto.
subst. simpl.
destruct (h0 c c) as [h3 | h4]. auto.
contradict h4. auto.
simpl.
destruct (h0 a c) as [hc | ha']; destruct (h0 b c) as [hc' | hb']; subst; try contradiction; auto.
contradict ha. auto. contradict hb; auto.
f_equal. auto.
Qed.


Lemma in_remove_neq' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    In x l -> l <> remove pfdt x l.
intros T  hdt l x h1 h2.
rewrite h2 in h1.
pose proof (remove_In hdt l x).
contradiction.
Qed.


Lemma remove_a_in_eq : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x:T),
    In x l -> x <> a ->
    In x (remove pfdt a l).
intros T hd l a x h1 h3.
destruct (in_dec hd a l) as [h2 | h0].
induction l as [|b l h7].
contradiction.
simpl.
destruct (hd a b) as [h4 | h5]. subst.
destruct h1 as [h1a | h1b]. 
symmetry in h1a.
contradiction.
destruct (in_dec hd b l) as [h8 | h9].
apply h7; auto.
pose proof (remove_not_in' hd _ _ h9) as h10.
rewrite <- h10. auto.
destruct h1 as [h1a | h1b]; destruct h2 as [h2a | h2b]. 
subst. contradiction.
subst. left. auto.
subst.
contradict h5. auto.
pose proof (h7 h1b h2b) as h8.
right. auto.
rewrite <- remove_not_in'; assumption.
Qed.

Lemma in_remove_neq_in_l : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a b:T),
    In a (remove pfdt b l) ->
    a <> b -> In a l.
intros T h0 l a b h1 h2.
induction l as [|c l h3].
simpl in h1. auto.
simpl in h1.
destruct (h0 b c). subst.
specialize (h3 h1). right; auto.
destruct h1 as [h1a | h1b]. subst. left. auto.
specialize (h3 h1b). right. auto.
Qed.

Lemma in_remove_inv : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a b:T),
    In a (remove pfdt b l) ->
    In a l /\ a <> b.
intros T h0 l a b h1. 
pose proof remove_In.
destruct (h0 a b) as [h2 | h3]. subst.
pose proof (remove_In h0 l b). contradiction.
split; auto.
eapply in_remove_neq_in_l; auto. apply h1. assumption.
Qed.

Lemma in_remove_iff : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a b:T),
    In a (remove pfdt b l) <->
    In a l /\ a <> b.
intros T h0 l a b. split. intro h1.
apply (in_remove_inv h0); auto.
intro h1. destruct h1.
apply remove_a_in_eq; auto.
Qed.

Lemma in_remove_impl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a b:T),
    In a (remove pfdt b l) ->
    In a l.
  intros T h1 l a b h2.
  pose proof (in_remove_iff h1 l a b) as h3.
  rewrite h3 in h2.
  destruct h2; auto.
Qed.

  
Lemma nin_nin_remove :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x a:T),
    ~In x l -> ~In x (remove pfdt a l).
intros T h0 l x a h1.
induction l as [|b l h2]. simpl. auto.
simpl in h1. 
assert (h3:b <> x /\ ~ In x l). tauto.
destruct h3 as [h3a h3b].
pose proof (h2 h3b) as h4.
simpl.
destruct (h0 a b) as [h5 | h6]. subst. auto.
simpl. tauto.
Qed.


Lemma nin_remove_eq_nin : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a b:T),
    ~ In a (remove pfdt b l) ->
    a = b \/ ~ In a l.
intros T h0 l a b h1.
induction l as [|c l h2].
right. intro; auto.
simpl in h1.
destruct (h0 b c) as [h3 | h4]. subst.
specialize (h2 h1). simpl.
rewrite eq_sym_iff. rewrite eq_sym_iff.
tauto.
simpl in h1.
assert (h3:c <> a /\ ~ In a (remove h0 b l)). tauto.
destruct h3 as [h3l h3r].
specialize (h2 h3r).
destruct h2; [left; auto | idtac].
right. simpl. tauto.
Qed.

  

Lemma in_tl : forall {T:Type} (l:list T) (x:T), In x (tl l) -> In x l.
intros T l.
induction l; simpl; auto.
Qed.

Lemma in_tl_lt :
  forall {T:Type} (l:list T) (x:T),
    In x (tl l) ->
    1 < length l.
  intros T l.
  destruct l as [|a l]; simpl.
  intro; contradiction.
  intros ? h1 ; apply lt_n_S.
  apply not_nil_lt.
  eapply in_not_nil.
  apply h1.
Qed.

Lemma impl_in_tl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (m:nat),
    S m < count_occ pfdt l x ->
    In x (tl l).
  intros T h1 l.
  destruct l as [|a l]; simpl.
  intros; omega.
  intros x m h2.
  destruct (h1 a x) as [|h3]; subst.
  apply lt_S_n in h2.
  apply count_occ_in in h2; auto.
  apply count_occ_in in h2; auto.
Qed.
    

Lemma length_tl :
  forall {T:Type} (l:list T),
    length (tl l) = pred (length l).
  intros ? l.
  induction l; auto.
Qed.

Lemma S_length_tl :
  forall {T:Type} (l:list T),
    l <> nil ->
    S (length (tl l)) = length l.
  intros T l; induction l; auto.
  intro h1. contradict h1; auto.
Qed.

Lemma lt_length_tl :
  forall {T:Type} (l:list T) (m:nat),
    S m < length l ->
    m < length (tl l).
  intros T l m h1.
  pose proof (lt_length _ _ h1) as h2. 
  pose proof (hd_compat' _ h2) as h3. simpl in h3.
  rewrite h3 in h1. simpl in h1.
  apply lt_S_n in h1; auto.
Qed.

Lemma le_length_tl :
  forall {T:Type} (l:list T) (m:nat),
    S m <= length l ->
    m <= length (tl l).
  intros T l m h1.
  rewrite length_tl.
  omega.
Qed.



Lemma firstn_tl :
  forall {T:Type} (l:list T) (m:nat),
    S m <= length l -> 
    firstn m (tl l) =
    tl (firstn (S m) l).
  intros T l m. revert l. 
  induction m as [|a ih]; simpl.
  intros l h1.
  destruct l as [|a l]; auto.
  intros l h1.
  rewrite (hd_compat' _ (lt_length _ _ h1)).
  simpl; auto.
Qed.

Lemma tl_not_nil :
  forall {T:Type} (l:list T),
    tl l <> nil ->
    l <> nil.
  intros ? ? h1; intro; subst; simpl in h1; contradict h1; auto.
Qed.

Lemma tl_not_nil_lt :
  forall {T:Type} (l:list T),
    tl l <> nil ->
    1 < length l.
  intros T l h1; destruct l; simpl.
  contradict h1; auto.
  simpl in h1.
  destruct l as [|a l]; simpl; auto with arith.
  contradict h1; auto.
Qed.

Lemma tl1lt :
  forall {T:Type} (l:list T),
    1 < length l ->
    tl l <> nil.
  intros T l h1.
  destruct l as [|a l].
  simpl in h1.
  apply lt_n_0 in h1.
  contradiction.
  simpl in h1.
  apply lt_S_n in h1.
  simpl.
  apply lt_length in h1; auto.
Qed.


Lemma tl_not_nil_le :
  forall {T:Type} (l:list T),
    tl l <> nil ->
    2 <= length l.
  intros; apply tl_not_nil_lt; auto.
Qed.




Lemma tl_app :
  forall {T:Type} (l l':list T),
    l <> nil ->
    tl (l ++ l') =
    (tl l) ++ l'.
  intros T l.
  destruct l as [|a l]; simpl; auto.
  intros ? h1.
  contradict h1; auto.
Qed.

Lemma removelast_tl :
  forall {T:Type} (l:list T),
    removelast (tl l) = tl (removelast l).
  intros T l; destruct l as [|? l]; try destruct l; auto.
Qed.

  
Lemma tl_sing :
  forall {T} (l:list T) (pf:l <> nil),
    tl l = nil ->
    l = hd' l pf :: nil.
  intros T l h1; destruct l as [|? l].
  contradict h1; auto.
  simpl.
  intro; subst; auto.
Qed.

  

Lemma tl_seq :
  forall (m n:nat),
    tl (seq m n) =
    seq (S m) (pred n).
  intro m; induction m; simpl; auto; intro n; destruct n; auto.
Qed.



Lemma count_occ_tl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (pf: l <> nil) (x:T),
    x <> hd' l pf ->
    count_occ pfdt (tl l) x =
    count_occ pfdt l x.
  intros T h1 l h2 x h3.
  rewrite (hd_compat' _ h2) at 2.
  simpl.
  destruct (h1 (hd' l h2) x); subst; auto.
  contradict h3; auto.
Qed.

Lemma count_occ_tl_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    let c := count_occ pfdt (tl l) x in
    let c' := count_occ pfdt l x in
    {c = c'} + {c = pred c'}.
  intros T h1 l.
  destruct l as [|a l]; simpl.
  intros; left; auto.
  intro x.
  destruct (h1 a x); [right; auto | left; auto].
Qed.

Definition tl2 {T:Type} (l:list T) :=
  tl (tl l).


Definition hd2 {T:Type} (l:list T) (pf:1 < length l) : T :=
  hd' (tl l) (tl1lt l pf).

Lemma length2_not_nil :
  forall {T:Type} (l:list T),
    length l = 2 ->
    l <> nil.
  intros; intro; subst.
  simpl in H.
  discriminate.
Qed.

Lemma length2_lt :
  forall {T:Type} (l:list T),
    length l = 2 ->
    1 < length l.
  intros; omega.
Qed.

    
  
Lemma length2 :
  forall {T:Type} (l:list T)
         (pf:length l = 2),
    l = hd' l (length2_not_nil l pf) :: hd2 l (length2_lt _ pf) :: nil.
  intros T l.
  destruct l as [|a l]; simpl.
  intros; discriminate.
  intros h1.
  f_equal.
  unfold hd2.
  simpl.
  pose proof h1 as h2.
  apply S_inj in h2.
  rewrite (length_eq_1 _ h2) at 1.
  repeat f_equal.
  apply proof_irrelevance.
Qed.
  

Lemma skipn_plus :
  forall {T:Type} (l:list T) (m n:nat),
    skipn (m + n) l =
    skipn m (skipn n l).
  intros T l m n. revert m l. 
  induction n as [|n ih]; simpl; auto. 
  intro m.
  rewrite plus_0_r; auto.
  intros m l.
  rewrite plus_S_shift_r.
  destruct (nil_dec' l) as [|h1]; subst.
  do 2 rewrite skipn_nil; auto.
  rewrite (hd_compat' _ h1).
  simpl.
  apply ih.
Qed.


Lemma skipn_app :
  forall {T:Type} (l l':list T) (n:nat),
    n <= length l ->
    skipn n (l ++ l') =
    skipn n l ++ l'.
  intros T l l' n. 
  revert l l'.
  induction n as [|n ih]; simpl; auto.
  intros l l' h1.
  pose proof (lt_length _ _ h1) as h2.
  rewrite (hd_compat' _ h2).
  simpl.
  apply ih.
  rewrite  length_tl.
  omega.
Qed.


Lemma app_sing_inj :
  forall {T} (l l':list T) a b,
    l ++ a::nil =
    l' ++ b::nil ->
    l = l' /\ a = b.
  intros T l l' a b h1.
  pose proof (f_equal (fun m => last m a) h1) as h2; simpl in h2.
  do 2 rewrite last_app_sing in h2; subst.
  split; auto.
  pose proof (f_equal (firstn (length l)) h1) as h2.
  rewrite firstn_app_le in h2; auto with arith.
  rewrite firstn_length in h2.
  rewrite h2.
  pose proof (f_equal (@length _) h1) as h3.
  do 2 rewrite app_length in h3;
    simpl in h3; do 2 rewrite S_compat in h3;
    apply S_inj in h3.
  rewrite h3.
  rewrite firstn_app_le; auto.
  apply firstn_length.
Qed.


Lemma app_sings_eq_iff :
  forall {T:Type}
         (l l':list T) 
         (a a':T),
    l ++ a::nil = l' ++ a'::nil <->
    l = l' /\ a = a'.
  intros T l l' a a'; split; intro h1.
  pose proof (f_equal (@removelast _) h1) as h2. 
  rewrite removelast_app_sing in h2. 
  subst.
  rewrite removelast_app_sing in h1.
  rewrite removelast_app_sing.
  split; auto.
  pose proof (f_equal (fun l => (last l a)) h1) as h2.
  simpl in h2.
  do 2 rewrite last_app_sing in h2; auto.
  destruct h1; subst; auto.
Qed.

(*This can't be parameterized over type [Type] instead of [nat], 
because of universe inconsistencies.*)

Fixpoint list_sum (P:nat->Type) (lP:forall i, list (P i)) (ln:list nat)
: Type :=
  match ln with
    | nil => unit
    | a :: ln' => (P a + list_sum P lP ln')%type
  end.

 
Fixpoint list_eq {T:Type} (l:list T) : Prop :=
  match l with
    | nil => True
    | a::l' => match l' with
                 | nil => True
                 | b::l0 =>
                   a = b /\ list_eq l'
               end
  end.

Lemma list_eq_nil :
  forall (T:Type),
    list_eq (@nil T).
  simpl; auto.
Qed.

Lemma list_eq_rev_cons :
  forall {T:Type} (l:list T) (a:T)
         (pfn:l <> nil),
    list_eq (a::l) ->
    a = hd' l pfn /\ list_eq l.
  simpl; intros T l a h1 h2.
  rewrite (hd_compat' _ h1) in h2.
  destruct h2 as [? h2]; subst.
  rewrite <- hd_compat' in h2.
  split; auto.
Qed.


Lemma list_eq_cons :
  forall {T:Type} (l:list T) (a:T),
    list_eq l ->
    (exists x:T, In x l /\ x = a) ->
    list_eq (a::l).
  intros T l.
  induction l as [|b l ih]; simpl; auto.
  intros a h1 h2.
  destruct h2 as [x [h2 ?]]; subst.
  destruct h2 as [|h2]; subst.
  split; auto.
  split; auto.
  rewrite (hd_compat' _ (in_not_nil _ _ h2)) in h1.
  destruct h1 as [? h1]; subst.
  rewrite <- hd_compat' in h1.
  specialize (ih a h1).
  hfirst ih.
  exists a.
  split; auto.
  specialize (ih hf).
  simpl in ih.
  rewrite (hd_compat' _ (in_not_nil _ _ h2)) in ih.
  destruct ih; auto.
Qed.
  
Lemma list_eq_compat :
  forall {T:Type} (l:list T),
    list_eq l <->
    (forall x y:T,
      In x l ->
      In y l ->
      x = y).
  intros T l.
  split.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros ha x y hb hc.
  destruct hb as [|hb]; subst.
  destruct hc as [|hc]; auto. 
  rewrite (hd_compat' _ (in_not_nil _ _ hc)) in ha. 
  destruct ha as [? ha]; subst.
  rewrite <- hd_compat' in ha.
  apply (ih ha); auto.
  apply in_hd'. 
  destruct hc as [|hc]; subst.
  rewrite (hd_compat' _ (in_not_nil _ _ hb)) in ha.
  destruct ha as [? ha]; subst.
  apply ih.
  rewrite <- hd_compat' in ha; auto.
  assumption.
  apply in_hd'.
  rewrite (hd_compat' _ (in_not_nil _ _ hb)) in ha.
  destruct ha as [? ha]; subst.
  rewrite <- hd_compat' in ha.
  apply ih; auto.
  induction l as [|a l ih]; simpl; auto.
  intro h1.
  hfirst ih.
  intros; apply h1; right; auto.
  destruct (nil_dec' l) as [|h2]; subst; auto.
  rewrite (hd_compat' _ h2).
  destruct (in_dec eq_dec a l).
  split.
  apply hf; auto.
  apply in_hd'.
  rewrite <- hd_compat'.
  apply ih; auto.
  split.
  apply h1.
  left; auto.
  right.
  apply in_hd'.
  rewrite <- hd_compat'.
  apply ih; auto.
Qed.




Fixpoint conj_list (l:list Prop) :=
  match l with
    | nil => True
    | P::l' => P /\ conj_list l'
  end.

Fixpoint disj_list (l:list Prop) :=
  match l with
    | nil => False
    | P::l' => P \/ disj_list l'
  end.

Lemma conj_list_compat :
  forall (l:list Prop),
    conj_list l <->
    (forall (P:Prop) ,
       In P l ->
       P).
  intro l.
  induction l as [|Q l ih]; simpl.
  split; auto.
  intros; contradiction.
  split; intro h1.
  intros P h2.
  destruct h2 as [|h2]; subst.
  destruct h1; auto.
  destruct h1 as [h1 h3].
  rewrite ih in h3.
  apply h3; auto.
  hr ih.
  intros P h2.
  apply h1.
  right; auto.
  rewrite <- ih in hr.
  split; auto.
  apply h1. 
  left; auto.
Qed.

Lemma disj_list_compat :
  forall (l:list Prop),
    disj_list l <->
    (exists (P:Prop) ,
       In P l /\
       P).
  intro l.
  induction l as [|Q l ih]; simpl. 
  split; auto.
  intros; contradiction.
  intro h1.
  destruct h1 as [? [? ?]]; contradiction.
  split; intro h1.
  destruct h1 as [h1 | h1].
  exists Q.
  split; auto.
  rewrite ih in h1.
  destruct h1 as [P [h1 h2]].
  exists P. 
  split; auto.
  destruct h1 as [P [[h2 | h2] h3]]; subst.
  left; auto.
  right.
  rewrite ih.
  exists P.
  split; auto.
Qed.

Definition list_to_option {T} (l:list T) : list (option T) :=
  None :: (map (fun x => Some x) l).

Lemma length_list_to_option :
  forall {T} (l:list T),
    length (list_to_option l) = S (length l).
  unfold list_to_option; intros; simpl; rewrite map_length; auto.
Qed.


(*A list of all sequence lists obtained by choosing one element 
from each list in [l], in the same order as the lists in [l]*)

Fixpoint list_of_lists_seqs {T:Type} (l:faml T) : faml T :=
  match l with 
  | nil => cons nil nil
  | al::l' => map (fun x:(T * (list T)) => cons (fst x) (snd x)) 
    (list_prod al (list_of_lists_seqs l'))
  end.

Fixpoint list_to_set {T:Type} (l:list T) : Ensemble T :=
  match l with
  | nil => Empty_set _
  | a::l' => Add (list_to_set l') a
  end.

Fixpoint faml_to_fam {T:Type} (ll:faml T) : Family T :=
  match ll with
    | nil => Empty_set _
    | la::ll' => Add (faml_to_fam ll') (list_to_set la)
  end.

Lemma finite_faml_to_fam :
  forall {T:Type} (ll:faml T),
    Finite (faml_to_fam ll).
  intros T ll.
  induction ll as [|a ll h1]. 
  constructor.
  simpl.
  apply Add_preserves_Finite; auto.
Qed.


  
Lemma empty_set_nil : 
  forall {T:Type} (l:list T),
    list_to_set l = Empty_set _ -> l = nil.
intros T l h1.
induction l as [|a l]. reflexivity.
simpl in h1.
pose proof (Add_intro2 _ (list_to_set l) a) as h2.
rewrite h1 in h2.
contradiction.
Qed.

Lemma list_to_set_sing :
  forall {T:Type} (a:T),
    list_to_set (a::nil) =
    Singleton a.
  intros T a.
  simpl.
  symmetry.
  rewrite add_empty_sing.
  reflexivity.
Qed.

Lemma length_app_nprod_to_list :
  forall {T m n} (x:T^m) (y:T^n),
    length ((nprod_to_list _ _ x)
              ++
              (nprod_to_list _ _ y)) =
    m + n.
  intros T m.
  induction m as [|m ih]; simpl.
  intros; rewrite length_nprod_to_list; auto.
  intros n x y.
  destruct x as [t x]; simpl; f_equal; auto.
Qed.


Lemma nprod_of_list_inj :
  forall {T} (l l':list T)
         (pfe:length l = length l'), 
    nprod_of_list _ l = transfer_nprod_r pfe (nprod_of_list _ l') ->
    l = l'.
  intros T l.
  induction l as [|a l ih]; simpl .
  intros; subst.
  pose proof (eq_sym pfe) as h2.
  apply length_eq0 in h2; auto.
  intros l'.
  destruct l' as [|b l']; simpl.
  intros; discriminate.
  intros h1 h2.
  rewrite transfer_nprod_r_S in h2; simpl in h2.
  inversion h2; subst.
  rename H1 into h3.
  f_equal.
  specialize (ih l' (S_inj _ _ h1) h3); auto.
Qed.


Fixpoint lrep {T:Type} (x:T) (n:nat) :=
  match n with
    | O => nil
    | S n => x :: lrep x n
  end.

Lemma lrep_nil_iff :
  forall {T:Type} (a:T) (n:nat),
    lrep a n = nil <->
    n = 0.
  intros T a n.
  destruct n as [|n]; split; intros h1; auto.
  simpl in h1.
  discriminate.
  discriminate.
Qed.

Lemma lrep_eq_iff :
  forall {T:Type} (a:T) (n:nat) (l:list T),
    lrep a n = l <->
    ((forall x:T, In x l -> x = a) /\
     length l = n).
  intros T a n. revert a.
  induction n as [|n ih]; simpl.
  intros a l; split; intro h1; subst; simpl.
  split; auto.
  intros; contradiction.
  destruct h1 as [? h1].
  apply length_eq0 in h1; auto.
  intros a l; split; intro h2.
  destruct l as [|a' l].
  discriminate.
  inversion h2; subst.
  simpl.
  specialize (ih a' (lrep a' n)).
  pose proof (iff1 ih eq_refl) as h3.
  destruct h3 as [h3 h4].
  split; auto with arith.
  intros x h5.
  destruct h5; auto.
  destruct h2 as [h2 h3].
  destruct l as [|a' l].
  discriminate.
  simpl in h3.
  apply S_inj in h3.
  pose proof (h2 _ (in_eq _ _)); subst.
  f_equal.
  rewrite ih.
  split.
  intros; apply h2.
  right; auto.
  reflexivity.
Qed.


Lemma length_lrep :
  forall {T:Type} (a:T) (n:nat),
    length (lrep a n) = n.
  intros ? ? n; induction n as [|n ih]; simpl; auto.
Qed.

Lemma in_lrep :
  forall {T:Type} (a:T) (n:nat),
    lrep a n <> nil ->
    In a (lrep a n).
  intros T a n.
  destruct n; simpl.
  intro h1. contradict h1; auto.
  intros; left; auto.
Qed.

Lemma in_lrep' :
  forall {T:Type} (a b:T) (n:nat),
    In b (lrep a n) -> b = a.
  intros T a b n.
  induction n as [|n ih]; simpl.
  intros; contradiction.
  intro h1.
  destruct h1 as [|h1]; auto.
Qed.
 

Definition all_sing {T:Type} (l:list T) (a:T) :=
  forall x : T, In x l -> x = a.



Lemma all_sing_cons_iff :
  forall {T:Type} (l:list T) (a x:T),
    all_sing (a::l) x <->
    x = a /\ all_sing l a.
 intros T l a x; split; intro h1.
  red in h1.
  pose proof (h1 a (in_eq _ _)); subst.
  split; auto.
  red; intros c h2.
  apply h1.
  right; auto.
  destruct h1 as [? h1]; subst.
  red in h1; red.
  intros x h2.
  destruct h2; auto.
Qed.


Lemma all_sing_cons_app :
  forall {T:Type} (x:T)
         (l:list T),
    all_sing l x ->
    x :: l = l++x::nil.
  intros T x l. revert l.
  induction l as [|a l ih]; simpl; auto.
  intros h1.
  rewrite all_sing_cons_iff in h1.
  destruct h1 as [? h1]; subst.
  specialize (ih h1).
  rewrite ih; auto.
Qed.

Lemma all_sing_count_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (x:T)
         (l:list T),
    all_sing l x ->
    count_occ pfdt l x = length l.
  intros T h1 x l. revert x.
  induction l as [|a l ih]; simpl; auto.
  intros x h2.
  destruct (h1 a x) as [h3 | h3]; subst.
  f_equal; apply ih.
  red in h2; red.
  intros; apply h2; right; auto.
  red in h2.
  specialize (h2 _ (in_eq _ _)).
  contradiction.
Qed.



Lemma all_sing_lrep'' :
  forall {T:Type} (l:list T) (a:T),
    all_sing l a ->
    exists! n:nat,
      l = lrep a n.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros a h1.
  exists 0.
  red; simpl; split; auto.
  intros n h2.
  symmetry in h2.
  rewrite lrep_nil_iff in h2; auto.
  intros x h1.
  rewrite all_sing_cons_iff in h1.
  destruct h1 as [h1 h2]; subst.
  specialize (ih _ h2).
  destruct ih as [n h1]. red in h1. destruct h1 as [h1 h3]; subst.
  exists (S n).
  red. split.
  simpl; auto.
  intros c h4.
  destruct c as [|c].
  simpl in h4. discriminate.
  simpl in h4.
  inversion h4 as [h5].
  specialize (h3 _ h5); subst; auto.
Qed.

Lemma all_sing_lrep :
  forall {T:Type} (l:list T) (a:T),
    all_sing l a ->
    l = lrep a (length l).
  intros T l.
  induction l as [|a l ih]; simpl.
  intros; auto.
  intros x h1.
  rewrite all_sing_cons_iff in h1.
  destruct h1 as [? h1]; subst.
  f_equal.
  apply ih; auto.
Qed.


Lemma all_sing_lrep' :
  forall {T:Type} (a:T) (n:nat),
    all_sing (lrep a n) a.
  intros T a n. revert a.
  induction n as [|n ih]; simpl.
  intros; red. intros; contradiction.
  intros a b h1.
  destruct h1 as [|h1]; subst; auto.
  apply in_lrep' in h1; auto.
Qed.

 
  
Lemma tl_lrep :
  forall {T:Type} (a:T) (n:nat),
    tl (lrep a n) =
    lrep a (pred n).
  intros T a n.
  destruct n as [|n]; simpl; auto.
Qed.

Lemma remove_lrep :
  forall {T:Type} (pfdt:type_eq_dec T)
         (a:T) (n:nat),
    remove pfdt a (lrep a n) = nil.
  intros T h1 a n.
  induction n; simpl; auto.
  destruct (h1 a a) as [|h2]; simpl; auto.
  contradict h2; auto.
Qed.

Lemma all_sing_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    NoDup l ->
    l <> nil ->
    all_sing l a ->
    l = a::nil.
  intros T h1 l.
  induction l as [|b l ih].
  intros ? ? h2.
  contradict h2; auto.
  intros a h2 h3 h4.
  red in h4.
  pose proof (h4 _ (in_eq _ _)) as he; subst.
  f_equal.
  destruct (nil_dec' l) as [|h5]; auto.
  eapply ih in h5.
  rewrite h5 in h2.
  apply no_dup_cons_nin in h2.
  contradict h2; left; reflexivity.
  apply no_dup_cons in h2; auto.
  red.
  intros; apply h4.
  right; auto.
Qed.


Lemma all_sing_eq_iff :
  forall {T} (l l':list T) a,
    all_sing l a ->
    all_sing l' a ->
    (l = l' <->
     length l = length l').
  intros T l.
  induction l as [|b l]; simpl;
  intro l'; destruct l' as [|b' l']; simpl;
  intros; intuition; try discriminate.
  f_equal.
  inversion H1; subst; auto.
  apply S_inj in H1.
  pose proof H as h2.
  specialize (h2 _ (in_eq _ _)); subst.
  pose proof H0 as h1.
  specialize (h1 _ (in_eq _ _)); subst.
  f_equal. 
  rewrite <- IHl in H1; subst; auto.
  red.
  intros x h1.
  specialize (H _ (in_cons _ _ _ h1)).
  apply H.
  red.
  intros x h1.
  specialize (H0 _ (in_cons _ _ _ h1)); auto.
Qed.


Lemma all_sing_nprod_eq : 
  forall {T n} (x y:T^n) a,
    all_sing_nprod x a ->
    all_sing_nprod y a ->
    x = y.
  intros ? ? ? ? ? h1 h2; apply nprod_ext; intros i hi.
  red in h1, h2.
  specialize (h1 _ (in_nth_prod i hi x));
    specialize (h2 _ (in_nth_prod i hi y)); congruence.
Qed.


Lemma all_sing_length :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    all_sing l a ->
    length l = count_occ pfdt l a.
  intros T h1 l.
  induction l as [|a' l ih]; simpl; auto.
  intros a h2.
  rewrite all_sing_cons_iff in h2.
  destruct h2 as [? h2]; subst.
  destruct (h1 a' a') as [h3 | h3]; auto with arith.
  contradict h3; auto.
Qed.


Lemma all_sing_firstn :
  forall {T:Type} (l:list T) (a:T) (m:nat),
    all_sing l a ->
    all_sing (firstn m l) a.
  intros ? ? ? ? h1.
  red in h1; red.
  intros ? h2; apply h1.
  apply in_firstn in h2; auto.
Qed.

  
Lemma lrep1 :
  forall {T:Type} (a:T),
    a::nil = lrep a 1.
  auto.
Qed.

Lemma remove_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (a:T),
    remove pfdt a (a::nil) = nil.
  intros; rewrite lrep1; apply remove_lrep.
Qed.



Lemma pred_lt_length :
  forall {T:Type} (l:list T) (n:nat)
         (pf:n < S (length l))
         (pfn:l <> nil),
    pred n < length l.
  intros T l n.
  destruct n as [|n]; simpl.
  intros; apply not_nil_lt; auto.
  intros h1 h2.
  apply lt_S_n in h1; auto.
Qed.

Lemma in_all_sing :
  forall {T:Type} (l:list T) (a x:T),
    all_sing l a ->
    In x l ->
    x = a.
  intros ? ? ? ? h1; apply h1; auto.
Qed.

Lemma all_sing_in : forall {T:Type} (l:list T) (a:T),
                      (l <> nil) ->
                      all_sing l a ->
                      In a l.
intros T l a h1.
destruct l as [|b l']. contradict h1; auto.
intro h2.
pose proof (in_eq b l') as h3.
specialize (h2 _ h3). subst. auto.
Qed.

Lemma all_sing_cons_in :
  forall {T:Type} (l:list T) (a c:T),
    l <> nil ->
    all_sing (a::l) c ->
    In a l.
  intros T l a c h1 h2.
  pose proof (all_sing_lrep _ _ h2) as h3.
  simpl in h3.
  inversion h3 as [h4].
  apply in_lrep; auto.
  intro h5.
  pose proof (not_nil_lt _ h1) as h6.
  rewrite (S_pred _ _ h6) in h5.
  simpl in h5.
  discriminate.
Qed.


Lemma list_to_set_in_iff : forall {T:Type} (l:list T) (x:T),
  In x l <-> Inn (list_to_set l) x.
  intros T l; induction l as [|a l h2]; intro x; simpl; split;
  intro h1; try contradiction; 
  try destruct h1 as [|h1]; subst.
  right; constructor; auto. 
  left; rewrite <- h2; auto.
  right; rewrite h2; auto.
  destruct H; left; auto.
Qed.


Lemma impl_all_sing :
  forall {T:Type} (l:list T) (a:T),
    list_to_set l = Singleton a ->
    all_sing l a.
  intros T l a h1 x h2.
  induction l as [|a' l h3].

  simpl in h1.
  pose proof (In_singleton _ a) as h4.
  rewrite <- h1 in h4.
  contradiction.

  simpl in h1, h2.
  pose proof (Add_intro2 _ (list_to_set l) a') as h5. 
  rewrite h1 in h5. 
  destruct h5; subst. 
  destruct h2 as [|h2]; subst; auto.
  rewrite list_to_set_in_iff in h2.
  pose proof (Add_intro1 _ (list_to_set l) a _ h2) as h7.
  rewrite h1 in h7.
  destruct h7; auto.
Qed.

Lemma all_sing_remove :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x y:T),
      all_sing l x ->
      x <> y ->
      l = remove pfdt y l.
  intros T h1 l x y h2 h3.
  symmetry. apply nin_remove_eq.
  intro h4.
  red in h2.
  contradict h3.
  symmetry; auto.
Qed.  

Lemma all_sing_remove' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    all_sing l x ->
    remove pfdt x l = nil.
  intros T h1 l x h2.
  apply all_sing_lrep in h2.
  rewrite h2.
  apply remove_lrep.
Qed.


Lemma all_sing_tl :
  forall {T:Type} (l:list T) (a:T),
    all_sing l a ->
    all_sing (tl l) a.
  intros T l a h1.
  red in h1; red.
  intros ? h2; apply h1; auto.
  apply in_tl; auto.
Qed.

Lemma all_sing_nth :
  forall {T:Type} (l:list T) (a:T),
    all_sing l a ->
    forall m:nat,
      nth m l a = a.
  intros T l.
  induction l as [|a' l ih]; simpl.
  intros a h1 m.
  destruct m; auto.
  intros a h1 m.
  rewrite all_sing_cons_iff in h1.
  destruct h1 as [h1 h2].
  destruct m; auto.
  subst.
  apply ih; auto.
Qed.


Lemma all_sing_nil :
  forall {T:Type} (a:T),
    all_sing nil a.
  intros; red; intros; contradiction.
Qed.

Lemma app_eq_sing_iff :
  forall {T:Type} (l l':list T) (a:T),
    l ++ l' = a::nil <->
    (l = nil /\ l' = a::nil) \/
    (l' = nil /\ l = a::nil).
  intros T l l'.
  destruct l as [|a l], l' as [|b l']; simpl.
  intro a; split; intro h1.
  discriminate.
  destruct h1 as [h1|h1]; destruct h1; discriminate.
  intro a; split; intro h2.
  left; auto.
  destruct h2 as [h2|h2]; destruct h2; try discriminate; auto.
  intro b; split; intro h1; simpl in h1.
  rewrite app_nil_r in h1.
  right; split; auto.
  destruct h1 as [h1|h1].
  destruct h1; discriminate.
  rewrite app_nil_r; destruct h1; auto.
  intro c; split; intro h1.
  inversion h1; subst.
  pose proof (app_cons_not_nil l l' b) as h2.
  rewrite H1 in h2.
  contradict h2; auto.
  destruct h1 as [h1|h1]; destruct h1; discriminate.
Qed.


Fixpoint app_list {T:Type} (ll:faml T) :=
  match ll with
    | nil => nil
    | la::ll' => la ++ app_list ll'
  end.

Lemma app_list_eq_nil_iff :
  forall {T:Type} (ll:faml T),
    app_list ll = nil <->
    ll = nil \/ all_sing ll nil.
  intros T ll; induction ll as [|a ll ih]; simpl; try tauto.
  split; intro h1.
  right.
  apply app_eq_nil in h1.
  destruct h1; subst.
  rewrite ih in H0.
  destruct H0 as [|h1]; subst.
  red; intros l h1; destruct h1; auto.
  contradiction.
  red; intros l h2; destruct h2; auto.
  destruct h1 as [|h1].
  discriminate.
  rewrite all_sing_cons_iff in h1.
  destruct h1; subst.
  simpl.
  rewrite ih; right; auto.
Qed.

Lemma app_list_nprod_to_list_inj :
  forall {T n} (x y:(list T)^n),
    (forall i (pfi:i < n),
       length (nth_prod i pfi x) =
       length (nth_prod i pfi y)) ->
    app_list (nprod_to_list _ _ x) =
    app_list (nprod_to_list _ _ y) ->
    x = y.
  intros T n.
  induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intros l l' h1 h2.
  destruct l as [fl sl], l' as [fl' sl']; simpl.
  simpl in h1.
  specialize (ih sl sl').
  simpl in h2.
  apply app_inj_eq_length in h2.
  destruct h2 as [? h3]; subst; f_equal.
  hfirst ih.
  intros i hi.
  specialize (h1 (S i) (lt_n_S _ _ hi)); simpl in h1.
  erewrite f_equal_dep.
  rewrite h1 at 1; auto.
  reflexivity.
  specialize (ih hf h3).
  destruct ih; subst; auto; simpl.
  specialize (h1 0 (lt_0_Sn _)); simpl in h1; auto.
Qed.

Lemma app_list_all_sing :
  forall {T} (ll:faml T),
    all_sing ll nil ->
    app_list ll = nil.
  intros ? ll; induction ll as [|l ll ih]; simpl; auto;
  intro h1.
  hfirst ih.
  red in h1; red; intros; apply h1; left; auto.
  pose proof (h1 _ (in_eq _ _)); subst.
  symmetry; apply h1; right; auto.
  specialize (ih hf).
  rewrite ih.
  red in h1.
  specialize (h1 _ (in_eq _ _)); subst; auto.
Qed.
    
  
  
    
     
  


Lemma list_to_set_finite : forall {T:Type} (l:list T), 
  Finite (list_to_set l).
intros T l.
induction l as [|a l h1].
simpl. auto with sets.
simpl. apply Add_preserves_Finite; assumption.
Qed.


Lemma list_to_set_in_cons :
  forall {T:Type} (l:list T) (a:T),
    In a l -> list_to_set (a::l) =
              list_to_set l.
  intros T l a h1.
  simpl.
  apply in_add_eq.
  rewrite <- list_to_set_in_iff.
  assumption.
Qed.



Lemma list_to_set_list_eq :
  forall {T:Type} (l l':list T),
    list_to_set l =
    list_to_set l' ->
    list_eq l ->
    list_eq l'.
  intros T l l' h1 h2.
  rewrite list_eq_compat in h2.
  rewrite list_eq_compat.
  intros; apply h2. 
  rewrite list_to_set_in_iff.
  rewrite h1.
  rewrite <- list_to_set_in_iff; auto.
  rewrite list_to_set_in_iff.
  rewrite h1.
  rewrite <- list_to_set_in_iff; auto.
Qed.



Definition length_eq {T:Type} (ll:faml T) : Prop :=
  list_eq (map (@length T) ll).

Definition slength_eq {T:Set} (ll:faml T) : Prop :=
  list_eq (map (@length T) ll).






Section Matrix.

(*A [list] of [lists] all of the same length.  
Each [list] is thought of as a column,
which convention was used to emphasize the independent 
nature of the columns/[list]s.
And the index in the [faml] is considered the row number.*)
Inductive matrix {T:Type} : faml T -> Prop :=
  | matrix_intro : forall fl,
                     length_eq fl ->
                     matrix fl.

Definition nat_matrix m n : Type :=
  match m with
    | O => unit
    | S _ => 
      match n with
        | O => unit
        | S _ => list (list nat)
      end
  end.


End Matrix.


Lemma map_eq_nil_iff :
  forall {T U:Type} (f:T->U) (l:list T),
    map f l = nil <-> l = nil.
  intros ? ? ? l.
  destruct l; intuition; try discriminate.
Qed.


Lemma no_dup_map_inj :
  forall {T U:Type}  (l:list T) (f:T->U),
    NoDup l -> Injective f ->
    NoDup (map f l).
intros T U l.
induction l as [|a l h1].
simpl.
intros f h1 h2. constructor.
intros f h2 h3.
simpl.
inversion h2. subst.
specialize (h1 _ H2 h3).
constructor.
rewrite in_map_iff.
intro h4.
destruct h4 as [x h5].
destruct h5 as [h5l h5r].
red in h3.
specialize (h3 _ _ h5l).
subst.
contradiction.
assumption.
Qed.

Lemma hd_map' :
  forall {T U:Type} (f:T->U) (l:list T) (pf:map f l <> nil),
    let pf' := (iffn1 (map_eq_nil_iff f l) pf) in
    hd' (map f l) pf = f (hd' l pf').
  intros T U f l h1 h2.
  gen0.
  generalize h2.
  rewrite (hd_compat' _ h2).
  intros h4 h5.
  simpl; auto.
Qed.


Lemma length_eq_nil :
  forall (T:Type),
    length_eq (@nil (list T)).
  intro; unfold length_eq; apply list_eq_nil.
Qed.


Lemma length_eq_rev_cons :
  forall {T:Type} (ll:faml T) (la:list T)
         (pfn:ll <> nil),
    length_eq (la::ll) ->
    length la = length (hd' ll pfn) /\ length_eq ll.
  intros T ll la h1 h2; red in h2.
  pose proof (list_eq_rev_cons (map (@length T) ll) (length la)) as h3.
  hfirst h3.
  rewrite map_eq_nil_iff; auto.
  specialize (h3 hf).
  simpl in h2, h3.
  specialize (h3 h2).
  rewrite hd_map' in h3.
  destruct h3 as [h3 h4].
  split; auto.
  rewrite h3.
  repeat f_equal.
  apply proof_irrelevance.
Qed.

Lemma length_eq_cons :
  forall {T:Type} (ll:faml T) (la:list T),
    length_eq ll ->
    (exists l:list T, In l ll /\ length l = length la) ->
    length_eq (la::ll).
  intros T ll la h1 h2. 
  destruct h2 as [l [h2 h3]].
  apply list_eq_cons.
  apply h1.
  exists (length l).
  generalize dependent ll.
  intro ll.
  revert h3. revert l la.
  induction ll as [|lb ll ih].
  intros; contradiction.
  intros l la h1 h2 h3.
  simpl.
  split; auto.
  destruct h3 as [|h3]; subst.
  left; auto.
  right.
  apply length_eq_rev_cons in h2.
  specialize (ih l la h1 h2 h3).
  destruct ih as [ih]; auto.
  intro; subst; contradiction.
Qed.

Lemma length_eq_compat :
  forall {T:Type} (ll:faml T),
    length_eq ll <->
    (forall lx ly:list T,
      In lx ll ->
      In ly ll ->
      length lx = length ly).
  intros T ll.
  unfold length_eq.
  rewrite list_eq_compat.
  split; intros h1 x y h2 h3.
  apply h1; (apply in_map; auto).
  rewrite in_map_iff in h2, h3.
  destruct h2 as [lx [h4 h5]], h3 as [ly [h6 h7]]; subst.
  apply h1; auto.
Qed.

Definition length_eq_ind {T:Type} (ll:faml T)
           (pfle:length_eq ll)
           (la lb:list T)
           (pfinla:In la ll)
           (pfinlb:In lb ll)
           (i:nat)
           (pfi:i < length la)  : i < length lb :=
  lt_eq_trans
    _ _ _
    pfi
    (iff1 (length_eq_compat ll) pfle _ _
          pfinla pfinlb).



Lemma map_im_compat : forall {T U:Type} (f:T->U) (l:list T),
                        list_to_set (map f l) = Im (list_to_set l) f.
intros T U f l.
apply Extensionality_Ensembles.
red. split.
red.
intros y h1.
rewrite <- list_to_set_in_iff in h1.
rewrite in_map_iff in h1.
destruct h1 as [x h2].
destruct h2 as [h2l h2r].
apply Im_intro with x.
rewrite <- list_to_set_in_iff.
assumption. subst. reflexivity.
red.
intros x h1.
destruct h1 as [x h1]. subst.
rewrite <- list_to_set_in_iff.
rewrite <- list_to_set_in_iff in h1.
rewrite in_map_iff.
exists x. split; auto.
Qed.


Lemma not_nil_inhabited :
  forall {T:Type} (l:list T),
    l <> nil ->
    Inhabited (list_to_set l).
  intros T l h1.
  destruct l as [|a l]. contradict h1; auto.
  econstructor.
  rewrite <- list_to_set_in_iff.
  left. apply eq_refl.
Qed.

Lemma inh_not_nil : 
  forall {T:Type} (l:list T),
    Inhabited (list_to_set l) ->
    l <> nil.
  intros ? ? h1; intro; subst.
  apply not_inh_empty in h1; auto.
Qed.


Lemma in_inh :
  forall {T:Type} (l:list T) (a:T),
    In a l ->
    Inhabited (list_to_set l).
  intros T l a h1.
  apply Inhabited_intro with a.
  rewrite <- list_to_set_in_iff; auto.
Qed.


Lemma all_sing_list_to_set :
  forall {T:Type} (l:list T) (a:T),
    l <> nil ->
    all_sing l a ->
    list_to_set l = Singleton a.
  intros T l a h0 h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  red in h1. 
  rewrite <- list_to_set_in_iff in h2. 
  rewrite h1; auto. constructor; auto.
  destruct h2.
  rewrite <- list_to_set_in_iff.
  apply all_sing_in; auto.
Qed.

Lemma all_sing_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    all_sing l a <->
    card_fun1 (Add (list_to_set l) a) <= 1.
  intros T h0 l a.
  split.
  intro h1 .
  destruct (nil_dec' l) as [h2 | h2]; subst.
  simpl.
  rewrite add_empty_sing.
  rewrite card_fun1_sing; auto with arith.
  pose proof (all_sing_in _ _ h2 h1) as h3.
  rewrite in_add_eq.
  erewrite all_sing_list_to_set.
  rewrite card_fun1_sing; auto with arith.
  assumption.
  apply h1.
  rewrite <- list_to_set_in_iff; auto.
  intro h1.
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1 | h1].
 
  destruct (in_dec h0 a l) as [h2 | h2].
  rewrite in_add_eq in h1.
  assert (h3:card_fun1 (list_to_set l) = 0).
  omega.
  apply card_fun1_O in h3.
  destruct h3 as [h3 | h3].
  rewrite  list_to_set_in_iff in h2.
  rewrite h3 in h2.
  contradiction.
  contradict h3; apply list_to_set_finite.
  rewrite <- list_to_set_in_iff; auto.
  rewrite card_add_nin' in h1.
  apply lt_S_n in h1.
  omega.
  apply list_to_set_finite.
  rewrite <- list_to_set_in_iff; auto.
  destruct (in_dec h0 a l) as [h2 | h2].
  rewrite in_add_eq in h1.
  apply card_fun1_1 in h1.
  destruct h1 as [a' h1].
  rewrite  list_to_set_in_iff in h2.
  rewrite h1 in h2.
  destruct h2.
  red.
  intros x h2.
  rewrite list_to_set_in_iff in h2.
  rewrite h1 in h2.
  destruct h2; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite card_add_nin' in h1.
  apply S_inj in h1.
  apply card_fun1_O in h1.
  destruct h1 as [h1 | h1].
  apply empty_set_nil in h1; subst.
  apply all_sing_nil.
  contradict h1; apply list_to_set_finite.
  apply list_to_set_finite.
  rewrite <- list_to_set_in_iff; auto.
Qed.


Lemma all_sing_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    In a l ->
    {all_sing l a} + {~all_sing l a}.
  intros T h1 l a hi.
  destruct (list_eq_dec h1 l (lrep a (length l))) as [h2 | h2].
  left.
  rewrite h2. 
  apply all_sing_lrep'.
  right.
  contradict h2. 
  apply all_sing_lrep in h2; auto.
Qed.

Lemma remove_eq_nil_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    remove pfdt x l = nil <->
    (l = nil \/ all_sing l x).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  tauto.
  intro x; split; intro h2.
  destruct (h1 x a) as [|h3]; subst.
  rewrite ih in h2.
  right.
  destruct h2 as [|h2]; subst.
  apply impl_all_sing; simpl.
  rewrite add_empty_sing; auto.
  apply impl_all_sing.
  destruct (nil_dec' l) as [|hn]; subst.
  simpl.
  apply add_empty_sing.
  apply all_sing_list_to_set in h2; auto.
  simpl.
  rewrite h2.
  apply add_sing_same.
  discriminate.
  destruct (h1 x a) as [|h3]; subst.
  destruct h2 as [h2 | h2].
  discriminate.
  rewrite all_sing_cons_iff in h2.
  destruct h2 as [? h2].
  clear H.
  rewrite ih.
  right; auto.
  destruct h2 as [|h2]; try discriminate.
  rewrite all_sing_cons_iff in h2.
  destruct h2; contradiction.
Qed.



Lemma incl_nil :
  forall {T:Type} (l:list T),
    incl nil l.
  intros; red; intros; contradiction.
Qed.

Lemma incl_nil' :
  forall {T:Type} (l:list T),
    incl l nil ->
    l = nil.
  intros T l h1.
  destruct l as [|a l]; simpl; auto.
  red in h1.
  specialize (h1 a (in_eq _ _)).
  contradiction.
Qed.

Lemma incl_subst :
  forall {T:Type} {l l1 l2:list T},
    l1 = l2 ->
    incl l l1 ->
    incl l l2.
  intros; subst; auto.
Qed.



Lemma incl_sing_iff :
  forall {T:Type} (l:list T) (a:T),
    incl l (a::nil) <->
    l = nil \/ all_sing l a.
  intros T l a; split; intro h1.
  destruct (nil_dec' l) as [|h2]; subst.
  left; auto.
  right.
  apply impl_all_sing.
  red in h1.
  apply Extensionality_Ensembles; red; split; red; intros x h3.
  rewrite <- list_to_set_in_iff in h3. 
  apply h1 in h3.
  destruct h3; subst.
  constructor; auto.
  contradiction.
  destruct h3.
  destruct l as [|b l].
  contradict h2; auto.
  specialize (h1 _ (in_eq _ _)).
  destruct h1; subst.
  rewrite <- list_to_set_in_iff; left; auto.
  contradiction.
  destruct h1 as [|h1]; subst.
  apply incl_nil.
  red in h1; red.
  intros x h2.
  apply h1 in h2; subst.
  left; auto.
Qed.


Lemma incl_sing_no_dup_iff :
  forall {T:Type} (l:list T) (a:T),
    NoDup l ->
    (incl l (a::nil) <->
     l = nil \/ l = a::nil).
  intros T l a h1. 
  rewrite incl_sing_iff.
  split; intro h2; destruct h2 as [|h2]; subst.
  left; auto.
  destruct (nil_dec' l) as [h3 | h3].
  left; auto.
  right.
  destruct l as [|b l].
  contradict h3; auto.
  red in h2; pose proof (h2 _ (in_eq _ _)); subst; auto.
  f_equal.
  destruct l as [|b l]; auto.
  specialize (h2 b).
  hfirst h2.
  right; left; auto.
  apply h2 in hf.
  subst.
  apply no_dup_cons_nin in h1.
  contradict h1; left; auto.
  left; auto.
  right.
  red.
  intros x h2.
  destruct h2; auto.
  contradiction.
Qed.



Inductive sub_list {T:Type} (l:list T) : list T->Prop :=
| sub_list_intro : forall l1 l2:list T,
                     sub_list l (l1++l++l2).

Lemma sub_list_intro_eq :
  forall {T:Type} {l l1 l2:list T} (pf:sub_list l (l1 ++ l ++ l2)),
    pf = sub_list_intro l l1 l2.
  intros; apply proof_irrelevance.
Qed.


  
Definition sub_list_hd {T:Type} (l l':list T) : Prop :=
  firstn (length l) l' = l.

Definition sub_list_tl {T:Type} (l l':list T) : Prop :=
  skipn (length l) l' = l.


Lemma sub_list_nil1 :
  forall {T:Type} (l:list T),
    sub_list nil l.
  intros T l.
  do 2 rewrite <- app_nil_r.
  rewrite <- app_assoc.
  constructor.
Qed.

Lemma sub_list_nil2_iff :
  forall {T:Type} (l:list T),
    sub_list l nil <->
    l = nil.
  intros T l; split; intro h1.
  dependent induction h1.
  apply app_eq_nil in x.
  destruct x as [? h1]; subst.
  apply app_eq_nil in h1.
  destruct h1; auto.
  subst.
  rewrite <- app_nil_r.
  rewrite <- app_nil_l.
  constructor.
Qed.
         
Lemma length_sub_list :
  forall {T:Type} (l l':list T),
    sub_list l l' ->
    length l <= length l'.
  intros ? ? ? h1; inversion h1; do 2 rewrite app_length; omega.
Qed.

    
Lemma sub_list_refl :
  forall (T:Type),
    Reflexive _ (@sub_list T).
  intros T; red; intro l.
  rewrite <- app_nil_r.
  rewrite <- app_nil_l.
  constructor.
Qed.

Lemma sub_list_trans :
  forall (T:Type),
    Transitive _ (@sub_list T).
  intro T; red; intros l1 l2 l3 h4 h5.
  destruct h4 as [l1' l1''].
  dependent induction h5.
  rewrite app_assoc.
  rewrite app_assoc.
  rewrite <- app_assoc. 
  rewrite <- (app_assoc l1 l1'').
  constructor.
Qed.

Lemma sub_list_antisym :
  forall {T:Type} (l l':list T),
    sub_list l l' ->
    sub_list l' l ->
    l = l'.
  intros T l l' h1.
  destruct h1 as [l0 l1].
  intro h2.
  dependent induction h2.
  rewrite <- app_assoc in x. 
  rewrite <- (app_assoc l2 l3 l1) in x.
  rewrite (app_assoc l0 l4) in x.
  rewrite app_eq_same_iff2 in x.
  destruct x as [h1 h2].
  apply app_eq_nil in h1.
  destruct h1; subst.
  apply app_eq_nil in h2.
  destruct h2; subst.
  repeat rewrite app_nil_l, app_nil_r; auto.
Qed.


Lemma sub_list_incl :
  forall {T:Type} (l l':list T),
    sub_list l l' -> incl l l'.
  intros T l l' h1.
  destruct h1.
  red.
  intros a h1.
  apply in_app_r.
  apply in_app_l; auto.
Qed.


Lemma sub_list_in :
  forall {T:Type} (l l':list T) (a:T),
    In a l ->
    sub_list l l' ->
    In a l'.
  intros T l l'. revert l.
  induction l' as [|a' l' ih]; simpl.
  intros l a h1 h2.
  apply sub_list_nil2_iff in h2.
  subst.
  contradiction.
  intros l a h1 h2.
  dependent induction h2.  
  destruct l1 as [|a1 l1].
  rewrite app_nil_l in x.
  destruct l as [|b l].
  rewrite app_nil_l in x.
  destruct l2 as [|a2 l2].
  discriminate.
  inversion x.
  subst.
  contradiction.
  rewrite <- app_comm_cons in x.
  inversion x.
  subst.
  destruct h1 as [|h1]; subst.
  left; auto.
  right.
  apply in_app_l; auto.
  rewrite <- app_comm_cons in x.
  inversion x.
  subst.
  specialize (ih l a h1 (sub_list_intro _ _ _)).
  right; auto.
Qed.

Lemma sub_list_sing_dec :
  forall {T:Type} (l:list T) (a:T),
    sub_list l (a::nil) ->
    {l = nil} + {l = a::nil}.
  intros T l.
  destruct l as [|b l].
  intros; left; auto.
  intros a h1.
  right.
  inversion h1.
  rewrite app_eq_iff in H0.
  destruct H0 as [h2 h3].
  destruct l1 as [|c l1].
  simpl in h2, h3.
  inversion h3.
  apply app_eq_nil in H1.
  destruct H1; subst.
  repeat rewrite app_nil_l, app_nil_r; auto.
  simpl in h2, h3.
  rewrite firstn_nil in h2.
  rewrite skipn_nil in h3.
  inversion h2; subst.
  discriminate.
Qed.

Lemma not_sub_list_sing :
  forall {T:Type} (l:list T) (a:T),
    l <> nil ->
    l <> a::nil ->
    ~sub_list l (a::nil).
  intros T l a h1 h2 h3.
  apply sub_list_sing_dec in h3.
  tauto.
Qed.

  
Lemma sub_list_sing_sing :
  forall {T:Type} (a a':T),
    sub_list (a::nil) (a'::nil) -> a = a'.
  intros T a a' h1.
  inversion h1 as [l1 l2 h3].
  rewrite app_eq_iff in h3; simpl in h3.
  destruct h3 as [h3 h4].
  destruct l1 as [|a1 l1].
  simpl in h3, h4.
  inversion h4; auto.
  simpl in h3, h4.
  rewrite skipn_nil in h4.
  discriminate.
Qed.

Lemma sub_list_sing_cons :
  forall {T:Type} (l:list T) (a:T),
    sub_list (a::nil) (a::l).
  intros T l a; apply (sub_list_intro (a::nil) nil l).
Qed.
 
Lemma sub_list_sing_cons_dec :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a a':T),
    sub_list (a::nil) (a'::l) ->
    {a = a'} + {In a l}.
  intros T h1 l a a' h2.
  destruct (h1 a a') as [|h3]; subst.
  left; auto.
  right.
  inversion h2.
  destruct l1 as [|a1 l1].
  rewrite app_nil_l in H0.
  inversion H0; subst.
  contradict h3; auto.
  rewrite <- app_comm_cons in H0.
  inversion H0; subst.
  apply in_app_r.
  left; auto.
Qed.


Lemma sub_list_cons :
  forall {T:Type} (l l':list T) (a:T),
    sub_list l l' ->
    sub_list l (a::l').
  intros T l l' a h1.
  destruct h1; subst.
  rewrite app_comm_cons.
  constructor.
Qed.


  
Lemma sub_list_firstn :
  forall {T:Type} (l:list T) (n:nat),
    sub_list (firstn n l) l.
  intros T l n.
  rewrite <- (firstn_skipn n).
  rewrite <- app_nil_l.
  constructor.
Qed.

Lemma sub_list_skipn :
  forall {T:Type} (l:list T) (n:nat),
    sub_list (skipn n l) l.
  intros T l n.
  rewrite <- (firstn_skipn n).
  rewrite <- app_nil_r.
  rewrite <- app_assoc.
  constructor.
Qed.

Lemma sub_list_firstn_le :
  forall {T:Type} (l:list T) (m n:nat),
    m <= n ->
    sub_list (firstn m l) (firstn n l).
  intros T l m n h1.
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1 | h1].
  erewrite firstn_lt_nest.
  apply sub_list_firstn.
  assumption.
  subst.
  apply sub_list_refl.
Qed.



Lemma sub_list_hd_compat :
  forall {T:Type} (l l':list T),
    sub_list_hd l l' ->
    sub_list l l'.
  intros T l l' h1.
  red in h1.
  rewrite <- h1.
  apply sub_list_firstn.
Qed.

Lemma sub_list_tl_compat :
  forall {T:Type} (l l':list T),
    sub_list_tl l l' ->
    sub_list l l'.
  intros T l l' h1.
  red in h1.
  rewrite <- h1.
  apply sub_list_skipn.
Qed.



Lemma sub_list_rev_cons1 :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    sub_list (a::l) l' ->
    exists l1 l2:list T,
      l' = l1 ++ l ++ l2 /\
      last l1 a = a.
  intros T h1 l l' a h2.
  dependent induction h2.
  exists (l1 ++ a::nil), l2; simpl.
  rewrite <- app_assoc, last_app_sing; split; auto.
Qed.

Lemma sub_list_cons1 :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    sub_list (a::l) l' ->
    sub_list l l'.
  intros T h1 l l' a h2.
  inversion h2; subst.
  rewrite app_assoc.
  rewrite <- app_sing. 
  pose proof (app_assoc l1 (a::nil) l) as h3.
  simpl in h3.
  rewrite h3 at 1.
  rewrite <- app_assoc.
  constructor.
Qed.


Lemma sub_list_rev_2cons_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    sub_list (a::l) (a::l') ->
    {firstn (length l) l' = l} + {sub_list (a::l) l'}.
  intros T hdt l l' a h1.
  destruct (list_eq_dec hdt 
                        (a::l)
                        (firstn (S (length l)) (a::l'))) as [h2 | h2].
  left.
  simpl in h2.
  inversion h2 as [he].
  rewrite <- he; auto.
  right.
  simpl in h2.
  assert (h3:l <> firstn (length l) l').
  contradict h2; f_equal; auto.
  revert h3 h2.
  dependent induction h1.
  intros h3 h4.
  rewrite app_eq_iff in x.
  destruct x as [h5 h6].
  simpl in h5, h6.
  destruct l1 as [|b l1].
  simpl in h6.
  inversion h6; subst.
  rewrite firstn_app_eq in h3.
  contradict h3; auto.
  simpl in h5, h6.
  inversion h5 as [h7].
  rewrite <- (firstn_skipn (length l1)).
  rewrite <- h6.
  apply (sub_list_intro (a::l) (firstn (length l1) l') l2).
Qed.

  
Lemma sub_list_rev_2cons_neq :
  forall {T:Type} (l l':list T) (a a':T),
    a <> a' ->
    sub_list (a::l) (a'::l') ->
    sub_list l l'.
  intros T l l' a a' h2 h3. 
  dependent induction h3.
  destruct l1 as [|a1 l1].
  rewrite app_nil_l in x.
  rewrite <- app_comm_cons in x.
  inversion x; contradiction.
  rewrite <- app_comm_cons in x.
  inversion x; subst.
  rewrite  app_comm_cons. 
  rewrite <- (app_sing l a). 
  rewrite app_assoc.
  pose proof (app_assoc l1 (a::nil) l) as h3.
  simpl.
  simpl in h3.
  rewrite h3.
  rewrite <- app_assoc.
  constructor.
Qed.

Lemma neq_sub_list_cons :
  forall {T:Type} (l l':list T) (a':T),
    l <> firstn (length l) (a'::l') ->
    sub_list l (a'::l') ->
    sub_list l l'.
  intros T l l' a' h1 h2.
  revert h1.
  dependent induction h2.
  intro h3.
  destruct l1 as [|a l1].
  rewrite app_nil_l in x.
  contradict h3.
  rewrite <- x.
  rewrite firstn_app_eq; auto.
  rewrite <- app_comm_cons in x.
  inversion x; subst.
  constructor.
Qed.
  


Lemma sub_list_rev_cons_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a':T),
    sub_list l (a'::l') ->
    {sub_list l l'} +
    {exists pf: l <> nil,
       hd' l pf = a' /\ tl l = firstn (pred (length l)) l'}.
  intros T h1 l l' a h2.
  destruct l as [|b l]; simpl.
  left; apply sub_list_nil1.  
  destruct (h1 a b) as [|h3]; subst. 
  apply sub_list_rev_2cons_dec in h2.
  destruct h2 as [h2 | h2].
  right.
  exists (cons_neq_nil _ _ ).
  split; auto.
  left; auto.
  assumption. 
  apply neq_sym in h3.
  left.
  inversion h2 as [l1 l2 h4].
  destruct l1 as [|c l1].
  rewrite app_nil_l in h4.
  inversion h4.
  contradiction.
  rewrite <- app_comm_cons in h4.
  inversion h4; subst.
  rewrite app_comm_cons.
  constructor.
Qed.



Lemma sub_list_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    {sub_list l l'} + {~sub_list l l'}.
  intros T hdt l l'. revert l. 
  induction l' as [|a l' ih].
  intro l.
  destruct (nil_dec' l) as [|h1]; subst.
  left; apply sub_list_nil1.
  right.
  intro h2.
  rewrite sub_list_nil2_iff in h2.
  contradiction. 
  intro l.
  specialize (ih l).
  destruct ih as [h1 | h1].
  left.
  apply sub_list_cons; auto.    
  destruct (list_eq_dec hdt l (firstn (length l) (a::l'))) as [h2 | h2].
  left.
  rewrite h2.
  apply sub_list_firstn.
  right.
  intro h3.
  inversion h3 as [l1 l2 h4].
  destruct l1 as [|a1 l1].
  rewrite app_nil_l in h4.
  destruct l2 as [|a2 l2].
  rewrite app_nil_r in h4.
  subst.
  rewrite firstn_length in h2.
  contradict h2; auto.
  rewrite <- h4 in h2.
  rewrite firstn_app_eq in h2.
  contradict h2; auto.
  rewrite <- app_comm_cons in h4.
  inversion h4; subst.
  contradict h1.
  constructor.
Qed.

Lemma sub_list_hd_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    {sub_list_hd l l'} + {~sub_list_hd l l'}.
  intros T h1 l l'.
  unfold sub_list_hd.
  apply list_eq_dec.
  apply h1.
Qed.

Lemma sub_list_tl_dec : 
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    {sub_list_tl l l'} + {~sub_list_tl l l'}.
  intros T h1 l l'.
  unfold sub_list_tl.
  apply list_eq_dec.
  apply h1.
Qed.


  
Lemma le_count_occ_sub_list :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T)
         (x:T) (pfsl:sub_list l l'),
    count_occ pfdt l x <= count_occ pfdt l' x.
  intros T h1 l l'. revert l. 
  induction l' as [|a' l' ih]. 
  intros l x h2.  
  rewrite sub_list_nil2_iff in h2; subst; simpl; auto with arith. 
  intros l x h2. 
  destruct (in_dec h1 x l) as [hin | hin].
  pose proof (sub_list_in _ _ _ hin h2) as hi.
  pose proof h2 as h2'.
  apply sub_list_rev_cons_dec in h2; auto.  
  destruct h2 as [h2 | h2]. 
  pose proof (ih _ x h2) as ih'.
  simpl.    
  destruct (h1 a' x) as [|h3]; subst; auto.
  destruct h2 as [h2 h3].
  simpl.  
  destruct h3 as [? h4]; subst.
  destruct (h1 (hd' l h2) x) as [|h5]; subst.
  rewrite (hd_compat' _ h2) at 1.
  simpl.
  destruct (h1 (hd' l h2) (hd' l h2)) as [|h5].
  apply le_n_S.
  apply ih.
  rewrite h4.
  apply sub_list_firstn.
  contradict h5; auto.
  destruct hi as [|hi]; subst.
  contradict h5; auto.
  apply sub_list_rev_cons_dec in h2'.
  destruct h2' as [h6 | h6].
  apply ih; auto.
  destruct h6 as [h6 [h7 h8]].
  specialize (ih (tl l) x).
  hfirst ih.
  rewrite h8.
  apply sub_list_firstn.
  specialize (ih hf).
  erewrite count_occ_tl in ih; auto.
  apply neq_sym.
  apply h5.
  assumption.
  rewrite (count_occ_In h1) in hin; omega.
Qed.

Lemma le_count_occ_firstn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    le (count_occ pfdt (firstn n l) x)  (count_occ pfdt l x).
  intros; apply le_count_occ_sub_list; apply sub_list_firstn.
Qed.

Lemma le_count_occ_skipn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    le (count_occ pfdt (skipn n l) x)  (count_occ pfdt l x).
  intros; apply le_count_occ_sub_list; apply sub_list_skipn.
Qed.
  
  
Lemma count_occ_firstn_eq_whole_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    count_occ pfdt (firstn n l) x  = count_occ pfdt l x  <->
    ~In x (skipn n l).
  intros T h1 l x n. revert l x.
  induction n as [|n ih]; simpl.
  intros; split; intro h2; [rewrite (count_occ_In h1); omega | 
                            rewrite (count_occ_In h1) in h2; omega].
  intros l x.
  destruct l as [|a l].
  intros; split; auto.
  simpl.
  destruct (h1 a x) as [|h2]; subst.
  split; intro h2.
  apply S_inj in h2.
  rewrite ih in h2; auto.
  f_equal.
  rewrite ih; auto.
  apply ih.
Qed.


Lemma count_occ_skipn_eq_whole_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    count_occ pfdt (skipn n l) x  = count_occ pfdt l x  <->
    ~In x (firstn n l).
  intros T h1 l x n. revert l x.
  induction n as [|n ih]; simpl.
  tauto.
  intros l x; split; intro h0.
  destruct l as [|a l].
  tauto.
  simpl; simpl in h0.
  destruct (h1 a x) as [|h2]; subst.
  pose proof (le_count_occ_skipn h1 l x n); omega.
  intro h3.
  destruct h3 as [|h3].
  contradiction.
  rewrite ih in h0.
  contradiction.
  destruct l as [|a l]; auto.
  simpl.
  destruct (h1 a x) as [|h2]; subst.
  contradict h0; left; auto.
  rewrite ih.
  contradict h0.
  right; auto.
Qed.


Section NatAnalogues.
  (*In this section will be non-polymorphic [nat]-boolean-analogues to 
many of the [list] functions and propositions 
involving [type_eq_dec], so we can 
capitalize on [nat]'s "easy" computational abilities.*)
     
  Fixpoint In_nat (l:list nat) (m:nat) : bool :=
    match l with
      | nil => false
      | a :: l' =>
        if (beq_nat a m) then
          true
        else
          In_nat l' m
    end.
  
  
  Fixpoint remove_nat (l:list nat) (m:nat) : list nat :=
    match l with
      | nil => nil
      | a :: l' =>
        let ln := remove_nat l' m in
        let be := beq_nat a m in
        if (In_nat l' m) then
          if be then
            ln
          else
            a :: ln
        else
          if be then
            l'
          else
            a :: l'
    end.

End NatAnalogues.


Definition lS (l:list nat) :=
  map S l.

Definition lpred (l:list nat) :=
  map pred l.

Lemma lpred_lS :
  forall l:list nat,
    lpred (lS l) = l.
  unfold lpred, lS; intro.
  rewrite map_map.
  rewrite <- map_id.
  f_equal.
Qed.

Lemma length_lS :
  forall (l:list nat),
    length (lS l) = length l.
  unfold lS; intros; apply map_length.
Qed.

Lemma lS_eq_nil_iff :
  forall (l:list nat),
    lS l = nil <->
    l = nil.
  intro l.
  destruct l as [|a l]; simpl.
  tauto.
  split; intro h1; discriminate.
Qed.

  
Lemma count_occ_lS :
  forall (l:list nat) (m:nat),
    m > 0 ->
    count_occ nat_eq_dec (lS l) m =
    count_occ nat_eq_dec l (pred m).
  intros l m.
  induction l as [|a l ih]; simpl; auto.
  destruct (nat_eq_dec (S a) m) as [h1 | h1].
  subst.
  destruct (nat_eq_dec a (pred (S a))) as [h2 | h2].
  f_equal; auto.
  assert (h3:a = 0). omega.
  subst.
  simpl in h2.
  contradict h2; auto.
  destruct (nat_eq_dec a (pred m)) as [|h2]; subst; auto.
  assert (h2:m = 0). omega.
  subst.
  simpl in ih; simpl.
  intro; omega.
Qed.

Lemma firstn_lS :
  forall (l:list nat) (n:nat),
    firstn n (lS l) = lS (firstn n l).
  intros l n.
  revert l.
  induction n as [|n ih]; simpl; auto.
  intro l.
  destruct l as [|a l]; simpl; auto.
  f_equal; auto.
Qed.


Lemma skipn_lS :
  forall (l:list nat) (n:nat),
    skipn n (lS l) = lS (skipn n l).
  intros l n.
  revert l.
  induction n as [|n ih]; simpl; auto.
  intro l.
  destruct l as [|a l]; simpl; auto.
Qed.


Lemma firstn_lpred :
  forall (l:list nat) (n:nat),
    firstn n (lpred l) = lpred (firstn n l).
  intros l n.
  revert l.
  induction n as [|n ih]; simpl; auto.
  intro l.
  destruct l as [|a l]; simpl; auto.
  f_equal; auto.
Qed.


Lemma skipn_lpred :
  forall (l:list nat) (n:nat),
    skipn n (lpred l) = lpred (skipn n l).
  intros l n.
  revert l.
  induction n as [|n ih]; simpl; auto.
  intro l.
  destruct l as [|a l]; simpl; auto.
Qed.




Lemma length_lpred :
  forall (l:list nat),
    length (lpred l) = length l.
  unfold lpred; intros; apply map_length.
Qed.

Lemma neq_nil_lpred :
  forall (l:list nat),
    l <> nil ->
    lpred l <> nil.
  intros l h1.
  apply not_nil_cons in h1.
  destruct h1 as [a h2]; subst.
  destruct h2 as [l' h2]; subst.
  simpl.
  apply cons_neq_nil.
Qed.


Lemma lpred_neq_nil :
  forall (l:list nat),
    lpred l <> nil ->
    l <> nil.
  intros l h1 h2; subst.
  simpl in h1.
  contradict h1; auto.
Qed.


Lemma hd_lpred' :
  forall (l:list nat) (pf:lpred l <> nil),
         hd' (lpred l) pf =
         pred (hd' l (lpred_neq_nil _ pf)).
  intros l.
  destruct l as [|a l]; simpl.
  intro h1; contradict h1; auto.
  intros; auto.
Qed.

Lemma tl_lpred :
  forall (l:list nat),
    tl (lpred l) =
    lpred (tl l).
  intro l.
  destruct l; auto.
Qed.

Lemma in_lS_iff :
  forall (l:list nat) (m:nat),
    In m (lS l) <->
    exists m':nat,
      In m' l /\ m = S m'.
  intros l m.
  unfold lS.
  rewrite in_map_iff.
  split; intro h1; destruct h1 as [n hm]; destruct hm as [? h2]; subst; exists n; tauto.
Qed.

Lemma nin0_lS :
  forall (l:list nat),
    ~In 0 (lS l).
  intros l h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [? [? ?]].
  discriminate.
Qed.
  
Lemma in_lpred_iff :
  forall (l:list nat) (m:nat),
    In m (lpred l) <->
    exists m':nat,
      In m' l /\ m = pred m'.
  intros l m.
  unfold lpred.
  rewrite in_map_iff.
  split; intro h1; destruct h1 as [n hm]; destruct hm as [? h2]; subst; exists n; tauto.
Qed.

Lemma lS_lpred_iff :
  forall (l:list nat),
    ~ In 0 l <->
    l = lS (lpred l).
  intro l; split; intro h1.
  induction l as [|a l ih]; simpl; auto.
  destruct (zerop a) as [|h2]; subst.
  contradict h1; left; auto.
  rewrite <- (S_pred _ _ h2).
  f_equal; apply ih.
  contradict h1; right; auto.
  intro h2.
  rewrite h1 in h2.
  rewrite in_lS_iff in h2.
  destruct h2 as [? [h4 h5]].
  discriminate.
Qed.
 

Lemma no_dup_lS :
  forall (l:list nat),
    NoDup l ->
    NoDup (lS l).
  intros l h1.
  unfold lS.
  apply no_dup_map_inj; auto.
  red; apply S_inj.
Qed.

  
Lemma no_dup_lS_impl :
  forall (l:list nat),
    NoDup (lS l) ->
    NoDup l.
  intro l.
  induction l as [|a l ih]; simpl; auto.
  intro h1.
  pose proof (no_dup_cons _ _ h1) as h2.
  specialize (ih h2).
  pose proof (no_dup_cons_nin _ _ h1) as h3.
  constructor; auto.
  contradict h3.
  apply in_map; auto.
Qed.


Lemma no_dup_lpred :
  forall (l:list nat),
    (In 0 l -> ~In 1 l) ->
    (In 1 l -> ~In 0 l) ->
    NoDup l ->
    NoDup (lpred l).
  intro l.
  induction l as [|c l ih]; simpl. 
  intros; constructor.
  intros h0 h1 h2.
  hfirst ih.
  intuition.
  pose proof (no_dup_cons _ _ h2) as h4.
  pose proof (no_dup_cons_nin _ _ h2) as h5. 
  constructor. 
  contradict h5. 
  unfold lpred in h5. rewrite in_map_iff in h5.
  destruct h5 as [c' h5].
  destruct h5 as [h5 h7].
  destruct (lt_eq_lt_dec c 1) as [h8 | h8].
  destruct h8 as [h8 | h8]. 
  assert (h9:c = 0). omega.
  subst.
  simpl in h5.
  apply pred0_dec in h5.
  destruct h5; subst; auto.
  intuition.
  subst.
  simpl in h5. 
  apply pred0_dec in h5.
  destruct h5; subst.
  intuition.
  assumption.
  assert (h9:c' > 0). omega.
  apply pred_inj0 in h5; auto with arith.
  subst; auto.
  apply ih; auto.
  intuition.
Qed.


Lemma no_dup_lpred_impl :
  forall (l:list nat),
    NoDup (lpred l) ->
    NoDup l.
  intro l.
  induction l as [|a l ih]; simpl; auto.
  intro h1.
  pose proof (no_dup_cons _ _ h1) as h2.
  specialize (ih h2).
  pose proof (no_dup_cons_nin _ _ h1) as h3.
  constructor; auto.
  contradict h3.
  apply in_map; auto.
Qed.


(*Bit Vector*)
Definition bv : Type := list bool.

Definition trues (n:nat) : bv :=
  lrep true n.

Definition falses (n:nat) : bv :=
  lrep false n.


(*The indexes of [bv] which are [false]*)
Fixpoint bv0s_aux (v:bv) (acc:nat) : list nat:=
  match v with
    | nil => nil
    | a::l =>
      let n := bv0s_aux l (S acc) in 
      match a with
        | false => acc::n
        | true => n
      end
  end.


Definition bv0s (v:bv) :=
  bv0s_aux v 0.

Lemma length_bv0s_aux_le_S :
  forall (v:bv) (n:nat),
    length (bv0s_aux v n) <=
    length (bv0s_aux v (S n)).
  intro v.
  induction v as [|b v ih]; simpl; auto with arith.
  intro n.
  destruct b; simpl; simpl in ih; auto.
  apply le_n_S; auto.
Qed.


Lemma length_bv0s_le :
  forall (v:bv) (n:nat),
    length (bv0s v) <=
    length (bv0s_aux v n).
  intros v n. revert v.
  induction n as [|n ih]; simpl; auto.
  intro v.
  eapply le_trans.
  apply ih.
  apply length_bv0s_aux_le_S.
Qed.



Lemma length_bv0s_aux_le :
  forall (v:bv) (m n:nat),
    m <= n ->
    length (bv0s_aux v m) <=
    length (bv0s_aux v n).
  intro v.
  induction v as [|b v ih]; simpl; auto.
  intros m n h1. 
  destruct b; simpl.
  apply ih; auto with arith.
  apply le_n_S; auto.
  apply ih; auto with arith.
Qed.

Lemma bv0s_aux_S :
  forall (v:bv) (m:nat),
    bv0s_aux v (S m) =
    lS (bv0s_aux v m).
  intro v.
  induction v as [|b v ih]; simpl; auto with arith.
  intro m.
  destruct b.
  specialize (ih (S m)).
  simpl in ih; auto.
  simpl.
  f_equal.
  apply ih.
Qed.


Lemma length_bv0s_le' :
  forall (v:bv),
    length (bv0s v) <= length v.
  intro v.  
  unfold bv0s.   
  induction v as [|b v ih]; auto with arith.   
  simpl. 
  destruct b; simpl; rewrite bv0s_aux_S;
  rewrite length_lS; auto with arith.
Qed.

Lemma in_bv0s_lt :
  forall (v:bv) (i:nat),
    In i (bv0s v) ->
    i < length v.
  intro v.
  induction v as [|b v ih]; simpl. 
  intros; contradiction.
  intros i h1. 
  unfold bv0s in h1; simpl in h1.
  destruct b; auto with arith.
  rewrite bv0s_aux_S in h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [m [h2 h3]]; subst; auto with arith.
  destruct h1 as [|h1]; subst; auto with arith.
  rewrite bv0s_aux_S in h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [m [h2 h3]]; subst; auto with arith.
Qed.

Lemma no_dup_bv0s :
  forall (v:bv),
    NoDup (bv0s v).
  intro v.
  unfold bv0s.
  induction v as [|b v ih]; simpl.
  constructor.
  destruct b.
  rewrite bv0s_aux_S.
  apply no_dup_lS; auto.
  rewrite bv0s_aux_S.
  constructor.
  intro h2.
  rewrite in_lS_iff in h2.
  destruct h2 as [? [? ?]]; discriminate.
  apply no_dup_lS; auto.
Qed.
      

(*The indexes of [bv] which are [true]*)
Fixpoint bv1s_aux (v:bv) (acc:nat) : list nat:=
  match v with
    | nil => nil
    | a::l =>
      let n := bv1s_aux l (S acc) in 
      match a with
        | false => n
        | true => acc::n
      end
  end.

Definition bv1s (v:bv) :=
  bv1s_aux v 0.


Lemma length_bv1s_aux_le_S :
  forall (v:bv) (n:nat),
    length (bv1s_aux v n) <=
    length (bv1s_aux v (S n)).
  intro v.
  induction v as [|b v ih]; simpl; auto with arith.
  intro n.
  destruct b; simpl; simpl in ih; auto.
  apply le_n_S; auto.
Qed.



Lemma length_bv1s_le :
  forall (v:bv) (n:nat),
    length (bv1s v) <=
    length (bv1s_aux v n).
  intros v n. revert v.
  induction n as [|n ih]; simpl; auto.
  intro v.
  eapply le_trans.
  apply ih.
  apply length_bv1s_aux_le_S.
Qed.


Lemma length_bv1s_aux_le :
  forall (v:bv) (m n:nat),
    m <= n ->
    length (bv1s_aux v m) <=
    length (bv1s_aux v n).
  intro v.
  induction v as [|b v ih]; simpl; auto.
  intros m n h1.
  destruct b; simpl.
  apply le_n_S; auto.
  apply ih; auto with arith.
  apply ih; auto with arith.
Qed.


Lemma bv1s_aux_S :
  forall (v:bv) (m:nat),
    bv1s_aux v (S m) =
    lS (bv1s_aux v m).
  intro v. 
  induction v as [|b v ih]; simpl; auto with arith.
  intro m.
  destruct b.
  specialize (ih (S m)).
  simpl in ih; auto.
  simpl.
  f_equal; auto.
  apply ih.
Qed.



Lemma length_bv1s_le' :
  forall (v:bv),
    length (bv1s v) <= length v.
  intro v.  
  unfold bv1s.   
  induction v as [|b v ih]; auto with arith.   
  simpl. 
  destruct b; simpl; rewrite bv1s_aux_S;
  rewrite length_lS; auto with arith.
Qed.


Lemma in_bv1s_lt :
  forall (v:bv) (i:nat),
    In i (bv1s v) ->
    i < length v.
  intro v.
  induction v as [|b v ih]; simpl. 
  intros; contradiction.
  intros i h1. 
  unfold bv1s in h1; simpl in h1.
  destruct b; auto with arith.
  rewrite bv1s_aux_S in h1.
  destruct h1 as [|h1]; subst; auto with arith.
  rewrite in_lS_iff in h1.
  destruct h1 as [m [h2 h3]]; subst; auto with arith.
  rewrite bv1s_aux_S in h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [m [h2 h3]]; subst; auto with arith.
Qed.


Lemma no_dup_bv1s :
  forall (v:bv),
    NoDup (bv1s v).
  intro v.
  unfold bv1s.
  induction v as [|b v ih]; simpl.
  constructor.
  rewrite bv1s_aux_S.
  destruct b.
  constructor.
  intro h2.
  rewrite in_lS_iff in h2.
  destruct h2 as [? [? ?]]; discriminate.
  apply no_dup_lS; auto.
  apply no_dup_lS; auto.
Qed.


(*Typically, [l <> nil]*)
(*The bit vector of the [tl] in its given list*)
Definition bv_tl {T:Type}
           (l:list T) : bv :=
  match l with
    | nil => nil
    | _::l' => false :: lrep true (length l')
  end.

Lemma length_bv_tl :
  forall {T:Type} (l:list T),
    length (bv_tl l) = length l.
  intros T l.
  destruct l as [|a l]; simpl; auto.
  f_equal; rewrite length_lrep; auto.
Qed.

Lemma bv_count_occ :
  forall(v:bv),
    length v = count_occ bool_eq_dec v true +
               count_occ bool_eq_dec v false.
  intro v.
  induction v as [|a v]; simpl; auto.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; subst; lr_if_goal; fold hlr; destruct hlr as [hl' | hr']; subst; try discriminate.
  rewrite plus_S_shift_l.
  f_equal; auto.
  rewrite plus_S_shift_r.
  f_equal; auto.
  destruct a.
  contradict hr; auto.
  contradict hr'; auto.
Qed.


Lemma bv0s_cons_true :
  forall v,
    bv0s (true::v) =
    lS (bv0s v).
  intro v.
  unfold bv0s.
  induction v as [|b v ih]; simpl; auto.
  destruct b.
  simpl in ih.
  rewrite bv0s_aux_S in ih.
  rewrite bv0s_aux_S; auto.
  simpl.
  f_equal.
  rewrite bv0s_aux_S; auto.
Qed.

Lemma bv0s_cons_false :
    forall v,
    bv0s (false::v) =
    0 :: lS (bv0s v).
  intro v.
  unfold bv0s.
  induction v as [|b v ih]; simpl; auto.
  destruct b.
  simpl in ih.
  rewrite bv0s_aux_S in ih.
  rewrite bv0s_aux_S; auto.
  simpl.
  f_equal.
  rewrite bv0s_aux_S; auto.
Qed.

Lemma bv1s_cons_true :
  forall v,
    bv1s (true::v) =
    0:: lS (bv1s v).
  intro v.
  unfold bv1s.
  induction v as [|b v ih]; simpl; auto.
  destruct b.
  simpl in ih.
  rewrite bv1s_aux_S in ih.
  rewrite bv1s_aux_S; auto.
  simpl.
  f_equal.
  rewrite bv1s_aux_S; auto.
Qed.

Lemma bv1s_cons_false :
    forall v,
    bv1s (false::v) =
    lS (bv1s v).
  intro v.
  unfold bv1s. 
  induction v as [|b v ih]; simpl; auto.
  destruct b.
  simpl in ih.
  rewrite bv1s_aux_S in ih.
  rewrite bv1s_aux_S; auto.
  simpl.
  f_equal.
  rewrite bv1s_aux_S; auto.
Qed.

Lemma all_sing_true_iff :
  forall (v:bv),
    all_sing v true <->
    ~In false v.
  intro v; induction v as [|v]; simpl; intuition;
  try (red; intros; contradiction); subst;
  try (red in H1; specialize (H1 _ (in_eq _ _)); discriminate);
  try (red; intros; destruct H; subst); destruct x; auto.
Qed.

Lemma all_sing_false_iff :
  forall (v:bv),
    all_sing v false <->
    ~In true v.
  intro v; induction v as [|v]; simpl; intuition;
  try (red; intros; contradiction); subst;
  try (red in H1; specialize (H1 _ (in_eq _ _)); discriminate);
  try (red; intros; destruct H; subst); destruct x;
  try specialize (H2 (eq_refl _)); try contradiction; auto.
Qed.



Lemma nin_true_bv1s :
  forall (v:bv),
    ~In true v ->
    bv1s v = nil.
  intro v.
  induction v as [|b v ih]; simpl; auto.
  intro h1.
  destruct b.
  contradict h1; left; auto.
  rewrite bv1s_cons_false.
  rewrite ih; auto.
Qed.

Lemma nin_false_bv0s :
  forall (v:bv),
    ~In false v ->
    bv0s v = nil.
  intro v.
  induction v as [|b v ih]; simpl; auto. 
  intro h1.
  destruct b.
  rewrite bv0s_cons_true, ih; auto.
  contradict h1; left; auto.
Qed.

Lemma bv0s_eq_nil_iff :
  forall (v:bv),
    bv0s v = nil <->
    ~In false v.
  intro v; induction v as [|b v ih]; simpl.
  intuition.
  destruct b. 
  split; intro h1.
  rewrite bv0s_cons_true in h1.
  rewrite lS_eq_nil_iff in h1.
  rewrite ih in h1.
  contradict h1.
  destruct h1; try discriminate; auto.
  hr ih.
  contradict h1.
  right; auto.
  rewrite bv0s_cons_true.
  rewrite lS_eq_nil_iff.
  tauto.
  split; intro h1.
  rewrite bv0s_cons_false in h1.
  discriminate.
  contradict h1; left; auto.
Qed.

Lemma bv1s_eq_nil_iff :
  forall (v:bv),
    bv1s v = nil <->
    ~In true v.
  unfold bv1s; intro v; induction v as [|b v ih]; simpl.
  intuition.
  destruct b.
  split; intro h1; [discriminate|contradict h1; left; auto].
  rewrite bv1s_aux_S; split; intro h1.
  rewrite lS_eq_nil_iff in h1.
  rewrite ih in h1.
  contradict h1.
  destruct h1; try discriminate; auto.
  rewrite lS_eq_nil_iff.
  rewrite ih.
  contradict h1.
  right; auto.
Qed.

Fixpoint bv_lt_eq_len (v:bv) : forall w, length v = length w -> Prop :=
  match v
  with
    | nil => fun _ _  => False
    | b::v' =>
      fun w =>
        match w with
          | nil => fun pfe:length (b::v') = length nil =>
                     False_rect _ (disc_nat (length v') pfe)
          | c::w' => fun pfe =>
                       match (tso_dec_bool_lt b c) with
                         | inleft (left _) => True
                         | inleft (right _) =>
                           bv_lt_eq_len v' w' (S_inj _ _ pfe)
                         | inright _ => False
                       end
        end
  end.                                  


Definition bv_lt (v w:bv) : Prop :=
  match (lt_eq_lt_dec (length v) (length w))  with
    | inleft (left _) => True
    | inleft (right pfe) => bv_lt_eq_len v w pfe
    | inright _ => False
  end.


Lemma tso_dec_bv_lt_eq_len :
  forall (v w:bv) (pfe:length v = length w),
    {bv_lt_eq_len v w pfe} + {v = w} + {bv_lt_eq_len w v (eq_sym pfe)}.
  intro v.
  induction v as [|b v ih]; simpl.
  intros w h1.
  pose proof (length_eq0 _ (eq_sym h1)); subst. 
  left; right; auto.  
  intro w. 
  destruct w as [|c w]. 
  intro; discriminate. 
  simpl; intro h1.
  specialize (ih _ (S_inj _ _ h1)).
  destruct ih as [h2|h2].
  destruct h2 as [h2|h2].
  lr_match_sum_goal'. 
  destruct hl; subst. 
  left; auto.
  left; left.
  gterm0.
  fold c0.
  term0 h2.
  pose proof (proof_irrelevance _ c0 c1) as h3.
  rewrite h3; auto.
  lr_match_sum_goal'. 
  destruct hl as [h3|]; subst.
  right; auto.
  apply irrefl_bool_lt in hr; contradiction.
  pose proof (transitive_bool_lt _ _ _ hr hr0) as h3.
  apply irrefl_bool_lt in h3; contradiction.
  subst.
  lr_match_sum_goal'.
  destruct hl as [h2|]; subst.
  left; auto.
  lr_match_sum_goal'. 
  destruct hl.
  apply irrefl_bool_lt in b; contradiction.
  left; right; auto.
  apply irrefl_bool_lt in hr; contradiction.
  right.
  lr_match_sum_goal'. 
  destruct hl as [h2|]; subst; auto.
  apply irrefl_bool_lt in hr; contradiction.
  pose proof (transitive_bool_lt _ _ _ hr hr0) as h3.
  apply irrefl_bool_lt in h3; contradiction.
  lr_match_sum_goal'. 
  destruct hl as [h3|]; subst.
  left; auto.
  lr_match_sum_goal'. 
  destruct hl as [h3|]; subst.
  apply irrefl_bool_lt in h3; contradiction.
  right.
  gterm0.
  fold c0.
  term0 h2.
  pose proof (proof_irrelevance _ c0 c1) as h3.
  rewrite h3; auto.
  apply irrefl_bool_lt in hr; contradiction.
  right.
  lr_match_sum_goal'. 
  destruct hl as [h3|]; subst; auto.
  apply irrefl_bool_lt in hr; contradiction.
  pose proof (transitive_bool_lt _ _ _ hr hr0) as h3.
  apply irrefl_bool_lt in h3; contradiction.
Qed.

  


Lemma total_strict_order_bv_lt_eq_len :
  forall (v w:bv) (pfe:length v = length w),
    bv_lt_eq_len v w pfe \/ v = w \/ bv_lt_eq_len w v (eq_sym pfe).
  intro v.  
  induction v as [|b v ih]; simpl.
  intros w h1.
  pose proof (length_eq0 _ (eq_sym h1)); subst.
  right; left; auto.
  intro w.
  destruct w as [|c w].
  intro; discriminate.
  simpl; intro h1.
  lr_match_sum_goal'.
  destruct hl.
  left; auto.
  subst.
  specialize (ih _ (S_inj _ _ h1)).
  destruct ih as [h2|h2].
  left.
  gterm0.
  fold c0.
  term0 h2.
  pose proof (proof_irrelevance _ c0 c1) as h3.
  rewrite h3; auto.
  destruct h2 as [|h2]; subst.
  right; left; auto.
  lr_match_sum_goal'.
  destruct hl.
  apply irrefl_bool_lt in b; contradiction.
  right; right.
  gterm0.
  fold c0.
  term0 h2.
  pose proof (proof_irrelevance _ c0 c1) as h3.
  rewrite h3; auto.
  apply irrefl_bool_lt in hr; contradiction.
  right.
  lr_match_sum_goal'.
  destruct hl; subst.
  right; auto.
  apply irrefl_bool_lt in hr; contradiction.
  pose proof (transitive_bool_lt _ _ _ hr hr0) as h2.
  apply irrefl_bool_lt in h2; contradiction.
Qed.

Lemma irrefl_bv_lt_eq_len :
  forall (v:bv) (pfe:length v = length v),
    ~bv_lt_eq_len v v pfe.
  intro v.
  induction v as [|b v ih]; simpl; auto.
  intro.
  lr_match_sum_goal'.
  destruct hl as [hl|hl]; auto.
  apply irrefl_bool_lt in hl; contradiction.
  apply irrefl_bool_lt in hr; contradiction.
Qed.

Lemma transitive_bv_lt_eq_len :
  forall (v w x:bv)
         (pfe: length v = length w)
         (pfe':length w = length x),
    bv_lt_eq_len v w pfe  ->
    bv_lt_eq_len w x pfe' ->
    bv_lt_eq_len v x (eq_trans pfe pfe').
  intro v.
  induction v as [|b v ih]; simpl.
  intros; contradiction.
  intro w; destruct w as [|c w].
  intro; discriminate.
  intro x; destruct x as [|d x]; simpl.
  intro; discriminate.
  lr_match_sum_goal'.
  destruct hl as [hl|hl].
  lr_match_sum_goal'.
  destruct hl0 as [hl0|hl0].
  lr_match_sum_goal'. 
  destruct hl1 as [hl1|hl1]; auto.
  subst.
  pose proof (transitive_bool_lt _ _ _ hl hl0) as h1.
  apply irrefl_bool_lt in h1; contradiction.
  pose proof (transitive_bool_lt _ _ _ hl hl0) as h1.
  pose proof (transitive_bool_lt _ _ _ hr h1) as h2.
  apply irrefl_bool_lt in h2; contradiction.
  subst.
  lr_match_sum_goal'. 
  destruct hl0 as [hl0|]; subst; auto.
  apply irrefl_bool_lt in hl; contradiction.
  pose proof (transitive_bool_lt _ _ _ hl hr) as h1.
  apply irrefl_bool_lt in h1; contradiction.
  intros; contradiction.
  lr_match_sum_goal'. 
  destruct hl0 as [hl0|]; subst; auto.
  lr_match_sum_goal'. 
  destruct hl as [hl|]; subst; auto.
  apply irrefl_bool_lt in hl0; contradiction.
  pose proof (transitive_bool_lt _ _ _ hl0 hr) as h1.
  apply irrefl_bool_lt in h1; contradiction.
  lr_match_sum_goal'. 
  destruct hl as [hl|hl].
  apply irrefl_bool_lt in hl; contradiction.
  intros h1 h2 h3 h4.
  specialize (ih _ _ (S_inj _ _ h1) (S_inj _ _ h2) h3 h4).
  gterm0.
  fold c.
  term0 ih.
  fold c0 in ih.
  pose proof (proof_irrelevance _ c c0) as h5.
  rewrite h5.
  subst; auto.
  apply irrefl_bool_lt in hr; contradiction.
  intros; contradiction.
  intros; contradiction.
Qed.

Lemma tso_dec_bv_lt :
  tso_dec bv_lt.
  unfold bv_lt; red; intros v w.
  lr_match_sum_goal'.
  destruct hl as [hl|hl].
  left; auto.
  lr_match_sum_goal'.
  destruct hl0 as [h0|h0].
  omega.
  pose proof (proof_irrelevance _ h0 (eq_sym hl)); subst.
  apply tso_dec_bv_lt_eq_len.
  omega.
  lr_match_sum_goal'; try omega.
  destruct hl; try omega; auto.
Qed.

Lemma total_strict_order_bv_lt :
  total_strict_order bv_lt.
  unfold bv_lt; red; intros v w.
  lr_match_sum_goal'.
  destruct hl as [h2|h2].
  left; auto.
  lr_match_sum_goal'.
  destruct hl as [h3|h3].
  omega.
  pose proof (proof_irrelevance _ h3 (eq_sym h2)); subst.
  apply total_strict_order_bv_lt_eq_len.
  omega.
  lr_match_sum_goal'; try omega.
  destruct hl as [h2|h2]; try omega; auto.
Qed.

Lemma irrefl_bv_lt :
  irrefl bv_lt.
  unfold bv_lt; red; intro; lr_match_sum_goal'. 
  destruct hl as [hl|hl].
  omega.
  apply irrefl_bv_lt_eq_len.
  omega.
Qed.

Lemma transitive_bv_lt :
  Transitive _ bv_lt.
  unfold bv_lt; red.
  intros v w x.
  lr_match_sum_goal'.
  destruct hl as [hl|hl].
  lr_match_sum_goal'.
  destruct hl0 as [hl0|hl0].
  lr_match_sum_goal'.
  destruct hl1 as [hl1|hl1]; auto.
  omega.
  omega.
  lr_match_sum_goal'.
  destruct hl1 as [hl1|hl1]; auto.
  omega.
  omega.
  intros; contradiction.
  lr_match_sum_goal'.
  destruct hl0 as [hl0|hl0]; auto.
  lr_match_sum_goal'. 
  destruct hl1 as [hl1|hl1]; auto.
  omega.
  omega.
  lr_match_sum_goal'.
  destruct hl1 as [hl1|hl1]; auto.
  pose proof (proof_irrelevance _ hl1 (eq_trans hl hl0)); subst.
  apply transitive_bv_lt_eq_len.
  omega.
  intros; contradiction.
  intro; contradiction.
Qed.


Lemma sortdec_bv_lt :
  sort_dec bv_lt.
  constructor.
  apply tso_dec_bv_lt.
  apply transitive_bv_lt.
  apply irrefl_bv_lt.
Qed.

Lemma sortable_bv_lt :
  sortable bv_lt.
  constructor.
  apply total_strict_order_bv_lt.
  apply transitive_bv_lt.
  apply irrefl_bv_lt.
Qed.


(*Removes exactly one element from a list corresponding to a 
given index.*)
Fixpoint remove_ind {T} (l:list T) (i:nat) : list T :=
  match l with
    | nil => nil
    | a :: l' =>
      match i with
        | 0 => l'
        | S i' => a :: remove_ind l' i'
      end
  end.

Lemma in_remove_ind :
  forall {T} (l:list T) (x:T) (i:nat),
    In x (remove_ind l i) ->
    In x l.
  intros T l; induction l as [|? ? ih]; auto; intros; simpl;
    simpl in H; destruct i as [|i];
  [right|destruct H as [|?]; subst;
         [left|apply ih in H; right]]; auto.
Qed.

Lemma remove_ind_ge :
  forall {T} (l:list T) (i:nat),
    i >= length l ->
    remove_ind l i = l.
  intros T l; induction l as [|a l ih]; simpl; auto; intros i;
  destruct i as [|i]; try omega; intro h1; apply ge_S_n in h1; f_equal;
  auto.
Qed.


Lemma incl_iff :
  forall {T:Type} (l l':list T),
    incl l l' <-> Included (list_to_set l) (list_to_set l').
  intros T l l'. split.
  intro h1. red; intros x h2.
  rewrite <- list_to_set_in_iff in h2.
  apply h1 in h2.
  rewrite <- list_to_set_in_iff.
  assumption.
  intro h1. red; intros x h2.
  rewrite list_to_set_in_iff in h2.
  apply h1 in h2.
  rewrite list_to_set_in_iff; auto.
Qed.

Lemma list_to_set_incl :
  forall {T:Type} (l l':list T),
    list_to_set l = list_to_set l' ->
    incl l l'.
  intros; rewrite incl_iff, H.
  apply inclusion_reflexive.
Qed.

Lemma refl_incl :
  forall (T:Type),
    Reflexive _ (@incl T).
  intro; red; red; intros; auto.
Qed.


Lemma trans_incl :
  forall (T:Type),
    Transitive _ (@incl T).
  intro T.
  red.
  intros lx ly lz h1 h2.
  red in h1, h2; red.
  intros x h3.
  apply h2; apply h1; auto.
Qed.

Lemma incl_cons :
  forall {T:Type} (l:list T) (a:T),
    incl l (a::l).
  intros; red; intros; right; auto.
Qed.

Lemma incl_in_cons :
  forall {T} (l l':list T) (a:T),
    incl l l' ->
    In a l' ->
    incl (a::l) l'.
  intros ? ? ? ? h1 h2; red; intros x h3;
  destruct h3 as [|h3]; subst; auto.
Qed.

Lemma incl_from_cons :
  forall {T} (l l':list T) (a:T),
    incl l (a::l') ->
    In a l' ->
    incl l l'.
  intros ? ? ? ? h1 ?; red; intros ? h3.
  apply h1 in h3.
  destruct h3; subst; auto.
Qed.
  

Lemma incl_cons_nin :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T) (a:T),
    ~In a l ->
    incl l (a::l') ->
    incl (remove pfdt a l) l'.
  intros T h1 l l' a h2 h3. 
  red in h3; red.
  intros x h4.
  rewrite in_remove_iff in h4.
  destruct h4 as [h4 h5].
  apply h3 in h4.
  destruct h4; subst; auto; try contradict h5; auto.
Qed.

 
Lemma incl_cons2_same :
  forall {T:Type} (l l':list T) (a:T),
    ~In a l ->
    ~In a l' ->
    incl (a::l) (a::l') ->
    incl l l'.
  intros T l l' a h1 h2 h3.
  red in h3; red.
  intros x h4.
  specialize (h3 x (in_cons _ _ _ h4)).
  destruct h3 as [|h3]; subst.
  contradiction.
  assumption.
Qed.

Lemma incl_app_l :
  forall {T:Type} (l l':list T) ,
    incl l (l ++ l').
  intros; red; intros; apply in_app_l; auto.
Qed.

Lemma incl_app_r :
  forall {T:Type} (l l':list T) ,
    incl l' (l ++ l').
  intros; red; intros; apply in_app_r; auto.
Qed.

Lemma anti_sym_incl_list_to_set :
  forall {T:Type} (l l':list T),
    incl l l' ->
    incl l' l ->
    list_to_set l = list_to_set l'.
  intros T l l' h1 h2.
  apply Extensionality_Ensembles; red; split; red; intros x h3.
  rewrite <- list_to_set_in_iff in h3. rewrite <- list_to_set_in_iff.
  apply h1; auto.
  rewrite <- list_to_set_in_iff in h3. rewrite <- list_to_set_in_iff.
  apply h2; auto.
Qed.

Lemma list_to_set_eq_iff : 
  forall {T:Type} (l l':list T),
    list_to_set l = list_to_set l' <->
    incl l l' /\ incl l' l.
  intros T l l'; split; intro h1.
  split; red; intros x h2; rewrite list_to_set_in_iff in h2; rewrite list_to_set_in_iff.
  rewrite <- h1; auto.
  rewrite h1; auto.
  destruct h1; apply anti_sym_incl_list_to_set; auto.
Qed.


Definition disjointl {T:Type} (l l':list T) : Prop :=
  disjoint (list_to_set l) (list_to_set l').


Lemma sym_disjointl :
  forall {T:Type} (l l':list T),
    disjointl l l' ->
    disjointl l' l.
  intros T l l' h1.
  do 2 red in h1. do 2 red.
  rewrite Intersection_commutative; auto.
Qed.


Fixpoint list_minus {T:Type} (pfdt:type_eq_dec T)
         (l l':list T) : list T :=
  match l with
    | nil => nil
    | a::al => if (in_dec pfdt a l')
               then (list_minus pfdt al l')
               else a::(list_minus pfdt al l')
  end.

Lemma list_minus_nil2 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    list_minus pfdt l nil = l.
  intros T h0 l.
  induction l as [|a l h1]; simpl; auto.
  rewrite h1; auto.
Qed.



Lemma list_minus_nin_in_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    ~In a l -> In a l' ->
    list_minus pfdt l (remove pfdt a l') =
    list_minus pfdt l l'.
  intros T h0 l.
  induction l as [|b l h1]; simpl; auto.
  intros l' a h2 h4.
  apply not_or_and in h2. destruct h2 as [h2 h3].
  destruct in_dec as [h5 | h5].
  destruct in_dec as [h6 | h6].
  apply h1; auto.
  contradict h6.
  eapply in_remove_neq_in_l. apply h5. assumption. 
  destruct in_dec as [h6 | h6].
  contradict h5.
  rewrite in_remove_iff. split; auto.
  f_equal.
  apply h1; auto.
Qed.

Lemma list_minus_cons_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    list_minus pfdt l (a::l') =
    remove pfdt a (list_minus pfdt l l').
  intros T h1 l.
  induction l as [|c l ih]; simpl; auto.
  intros l' a.
  destruct (h1 a c) as [h2 | h2]; simpl; subst.
  destruct in_dec as [h2 | h2].
  apply ih.
  simpl.
  destruct (h1 c c) as [h3 | h3]. 
  apply ih.
  contradict h3; auto.
  destruct in_dec as [h3 | h3].
  apply ih.
  simpl.
  destruct (h1 a c) as [h4 | h4]; try contradiction.
  f_equal.
  apply ih.
Qed.


Lemma list_minus_cons_remove2 :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    In a l' ->
    list_minus pfdt (a::l) l' =
    list_minus pfdt l l'.
  intros T h1 l l' a h2; simpl.
  destruct in_dec as [h3 | h3]; auto.
  contradiction.
Qed.


Lemma list_minus_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    list_minus pfdt l (a::nil) =
    remove pfdt a l.
  intros T hdt l.
  induction l as [|al l ih]; simpl; auto.
  intro a.
  destruct (hdt a al) as [h1 | h1]; subst; simpl; auto.
  f_equal.
  apply ih.
Qed.


Lemma list_minus_same :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    list_minus pfdt l l = nil.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  destruct (h1 a a) as [h2 | h2]; auto.
  rewrite list_minus_cons_remove.
  rewrite ih.
  simpl.
  reflexivity.
  contradict h2; auto.
Qed.


Lemma impl_in_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T),
    In x l -> ~In x l' ->
    In x (list_minus pfdt l l').
  intros T h0 l.
  induction l as [|a l h1]; simpl.
  intros; contradiction.
  intros l' x h2 h3.
  destruct in_dec as [h4 | h4].
  destruct h2 as [h2 | h2]; subst.
  contradiction.
  apply h1; auto.
  destruct h2 as [h2 | h2]; subst.
  left; auto.
  right.
  apply h1; auto.
Qed.

Lemma in_list_minus_impl : 
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T),
    In x (list_minus pfdt l l') ->
    (In x l /\ ~In x l').
  intros T h0 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros l' x h1.
  destruct (in_dec h0 a l') as [h2 | h2].
  specialize (ih _ _ h1).
  destruct ih as [h3 h4].
  split. right; auto. assumption.
  simpl in h1. destruct h1 as [h1 | h1]; subst.
  split.
  left; auto.
  assumption.
  apply ih in h1.
  destruct h1 as [h1 h3].
  split.
  right; auto.
  assumption.
Qed.

Lemma in_list_minus_impl' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T),
    In x (list_minus pfdt l l') ->
    In x l.
  intros T h1 l l' x h2.
  pose proof (in_list_minus_impl h1 _ _ _ h2) as h3.
  destruct h3; auto.
Qed.

Lemma impl_list_minus_nil :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    incl l l' ->
    list_minus pfdt l l' = nil.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros l' h2.
  destruct in_dec as [h3 | h3].
  apply ih.
  red in h2; red.
  intros; apply h2.
  right; auto.
  contradict h3.
  red in h2.
  apply h2.
  left; auto.
Qed.


Lemma list_minus_eq_nil_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    list_minus pfdt l l' = nil <->
    incl l l'.
  intros T h1 l l'; split; intro h2.
  red.
  intros x h3.
  destruct (in_dec h1 x l') as [h4|h4]; auto.
  pose proof (impl_in_list_minus h1 l l' x h3 h4) as h5.
  rewrite h2 in h5.
  contradiction.
  apply impl_list_minus_nil; auto.
Qed.
  

Lemma list_to_set_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    list_to_set (list_minus pfdt l l') =
    Setminus (list_to_set l) (list_to_set l').
  intros T h0 l.  
  induction l as [|a l h1].
  intros. simpl. rewrite setminus_empty.
  reflexivity. 
  intro l'.  simpl. 
  destruct in_dec as [h2 | h2]. 
  specialize (h1 l').
  rewrite setminus_add.
  rewrite h1.
  symmetry. rewrite Union_commutative. rewrite <- inclusion_iff_union.
  red; intros x h3.
  destruct h3 as [h3 h4]. destruct h3.
  contradict h4.
  rewrite <- list_to_set_in_iff; auto.
  simpl.
  rewrite setminus_add.
  rewrite h1.
  unfold Add.
  f_equal.
  rewrite disjoint_setminus_compat; auto.
  red.
  apply eq_empty.
  red; intros x h3.
  destruct h3 as [x h4 h5].
  destruct h4.
  rewrite <- list_to_set_in_iff in h5.
  contradiction.
Qed.

Lemma no_dup_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
         NoDup l ->
         NoDup (list_minus pfdt l l').
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros l' h2.
  destruct (in_dec h1 a l') as [h3 | h3].
  apply ih.
  apply no_dup_cons in h2; auto.
  constructor.
  contradict h3.
  apply in_list_minus_impl in h3.
  destruct h3 as [h3].
  apply no_dup_cons_nin in h2; contradiction.
  apply ih.
  apply no_dup_cons in h2; auto.
Qed.

Lemma incl_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (x:T),
    incl (remove pfdt x l) l.
  intros T h1 l x.
  red; intros a h2.
  rewrite in_remove_iff in h2.
  destruct h2; auto.
Qed.
  
Lemma incl_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    incl (list_minus pfdt l l') l.
  intros T h0 l l'.
  rewrite incl_iff.
  rewrite list_to_set_list_minus.
  apply setminus_inc.
Qed.

Lemma in_incl :
  forall {T:Type} (l:list T) (x:T),
    In x l ->
    incl (x::nil) l.
  intros; red; intros x' h1; destruct h1; subst; try contradiction; auto.
Qed.


Lemma incl_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    {incl l l'} + {~incl l l'}.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intro; left; apply incl_nil.
  intros l'.
  specialize (ih l').
  destruct ih as [ih | ih].
  destruct (in_dec h1 a l') as [h2 | h2].
  left.
  red; intros x h3.
  destruct h3 as [|h3]; subst; auto.
  right.
  intro h3.
  red in h3.
  specialize (h3 _ (in_eq _ _)).
  contradiction. 
  right.
  contradict ih.
  eapply trans_incl.
  apply incl_cons.
  apply ih.
Qed.

Lemma lneq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    l <> nil ->
    l' <> nil ->
    (list_to_set l <> list_to_set l' <->
     exists x:T,
       In x l /\ ~In x l' \/
                 In x l'/\ ~In x l).
  intros T h0 l l' h1 h2; split; intro h3.
  rewrite list_to_set_eq_iff in h3.
  destruct (incl_dec h0 l l') as [h4 | h4].
  assert (h5:~ incl l' l).
    intro; contradict h3; split; auto.
  rewrite <- (list_minus_eq_nil_iff h0) in h5.
  exists (hd' _  h5). 
  right.
  split.
  eapply incl_list_minus. 
  apply in_hd'. 
  intro h6.
  pose proof (in_hd' (list_minus h0 l' l) h5) as h7.
  apply in_list_minus_impl in h7.
  destruct h7; contradiction.
  rewrite <- (list_minus_eq_nil_iff h0) in h4.
  exists (hd' _ h4).
  left.
  split.
  eapply incl_list_minus.
  apply in_hd'.
  intro h6.
  pose proof (in_hd' (list_minus h0 l l') h4) as h7.
  apply in_list_minus_impl in h7.
  destruct h7; contradiction.

  rewrite list_to_set_eq_iff.
  intro h4.
  destruct h4 as [h4 h5].
  destruct h3 as [x h6].
  destruct h6 as [h6 | h6].
  destruct h6 as [h6 h7].
  apply h4 in h6; contradiction.
  destruct h6 as [h6 h7].
  apply h5 in h6; contradiction.
Qed.


Lemma finite_mem_faml :
  forall {T:Type} (ll:faml T),
    finite_mem (faml_to_fam ll).
  intros T ll.
  induction ll as [|al lll h1]; simpl.
  do 2 red; intros; contradiction.
  intros A h2.
  destruct h2 as [A h2|A h2]; subst.
  apply h1; auto.
  inversion h2. subst.
  apply list_to_set_finite.
Qed.

  
Lemma in_faml_to_fam_impl :
  forall {T:Type} (ll:faml T)
         (A:Ensemble T) (pf:Inn (faml_to_fam ll) A),
  exists l:list T,
    In l ll /\ list_to_set l = A.
  intros T ll.
  induction ll as [|al ll h1]; simpl.
  intros; contradiction.
  intros A h2.
  destruct h2 as [A h2 | A h2].
  specialize (h1 _ h2). destruct h1 as [l h1].
  exists l. tauto.
  inversion h2. subst. clear h2.
  exists al. tauto.
Qed.  

Lemma in_faml_impl :
  forall {T:Type} (ll:faml T) (l:list T),
    In l ll ->
    Inn (faml_to_fam ll) (list_to_set l).
  intros T ll.
  induction ll as [|al ll h1]; simpl.
  intros; contradiction.
  intros l h2.
  destruct h2 as [|h2]; subst.
  right. constructor.
  specialize (h1 _ h2).
  left; auto.
Qed.


Lemma not_nil_inh_faml_to_fam :
  forall {T:Type} (ll:faml T),
    ll <> nil ->
    Inhabited (faml_to_fam ll).
  intros T ll h1.
  pose proof (not_nil_inhabited _ h1) as h2.
  destruct h2 as [l h2].
  apply Inhabited_intro with (list_to_set l).
  apply in_faml_impl.
  rewrite list_to_set_in_iff; auto.
Qed.



Lemma faml_to_fam_eq :
  forall {T:Type} (l:faml T),
    faml_to_fam l = list_to_set (map list_to_set l).
  intros T ll.
  apply Extensionality_Ensembles; red; split; red; intros A h1.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_iff.
  apply in_faml_to_fam_impl in h1.
  destruct h1 as [l h1]. exists l. tauto.
  rewrite <- list_to_set_in_iff in h1.
  rewrite in_map_iff in h1.
  destruct h1 as [l h1].
  destruct h1 as [h1 h2]. subst.
  apply in_faml_impl; auto.
Qed.


Lemma faml_to_fam_decompose :
  forall {T:Type} (pfdt:type_eq_dec (list T))
         (ll:faml T) (l:list T),
    In l ll ->
    faml_to_fam ll =
    Add (faml_to_fam (remove pfdt l ll)) (list_to_set l).
  intros T hdt ll l h1.
  apply Extensionality_Ensembles; red; split; red; intros lx h2.
  apply in_faml_to_fam_impl in h2.
  destruct h2 as [la [h2 h3]]; subst.
  destruct (hdt l la) as [h3 | h3]; subst.
  right; auto.
  constructor.
  left.
  apply in_faml_impl.
  rewrite in_remove_iff; auto.

  destruct h2 as [lx h2 | lx h2].
  apply in_faml_to_fam_impl in h2.
  destruct h2 as [la [h2 h3]]; subst.
  rewrite in_remove_iff in h2.
  destruct h2 as [h2 h3].
  apply in_faml_impl; auto.
  destruct h2.
  apply in_faml_impl; auto.
Qed.



Definition pairwise_disjointl {T:Type} (ll:faml T) : Prop :=
  forall l l':list T,
    (list_to_set l) <> (list_to_set l') ->
    In l  ll ->
    In l' ll ->
    disjointl l l'.

Lemma pairwise_disjotntl_nil :
  forall (T:Type),
    pairwise_disjointl (@nil (list T)).
  intro; red; simpl; intros; contradiction.
Qed.

Lemma pairwise_disjointl_compat :
  forall {T:Type} (ll:faml T),
    pairwise_disjointl ll <->
    pairwise_disjoint (faml_to_fam ll).
  intros T ll; split; intro h1; red in h1; red.
  intros A B h2 h3 h4.
  apply in_faml_to_fam_impl in h3.
  apply in_faml_to_fam_impl in h4.
  destruct h3 as [l [h3 h5]]; subst.
  destruct h4 as [l' [h4 h6]]; subst.
  apply (h1 _ _ h2 h3 h4).
  intros l l' h2 h3 h4.
  apply (h1 _ _ h2 (in_faml_impl _ _ h3) (in_faml_impl _ _ h4)).
Qed.


Lemma pairwise_disjointl_incl :
  forall {T:Type} (ll lm:faml T),
    pairwise_disjointl ll -> incl lm ll -> pairwise_disjointl lm.
  intros T ll lm h1 h2.
  red in h1. red.
  intros l l' h3 h4 h5.
  apply h2 in h4. apply h2 in  h5.
  apply h1; auto.
Qed.

Lemma pairwise_disjointl_cons :
  forall {T:Type} (ll:faml T),
    pairwise_disjointl ll ->
    forall l:list T,
           (forall l', In l' ll -> disjointl l l') ->
           pairwise_disjointl (l::ll).
  intros T ll h1 l' h2.
  red in h1. red.
  intros l1 l2 h3 h4 h5.
  destruct h4 as [|h4], h5 as [|h5]; subst; subst.
  contradict h3; auto.
  apply h2; auto.
  specialize (h2 _ h4).
  apply sym_disjoint; auto.
  apply h1; auto.
Qed.

Lemma pairwise_disjointl_cons_rev :
  forall {T:Type} (ll:faml T) (l:list T),
    pairwise_disjointl (l::ll) ->
    pairwise_disjointl ll.
  intros T ll l h1.
  red in h1. red.
  intros l1 l2 h3 h4 h5.
  apply h1; auto; right; auto.
Qed.

Definition partition_faml {T:Type} (ll:faml T) (l:list T) : Prop :=
  partition_fam (faml_to_fam ll) (list_to_set l).

Lemma partition_faml_nil :
  forall {T:Type} (l:list T),
    partition_faml nil l -> l = nil.
  intros T l h1.
  do 2 red in h1; simpl in h1.
  destruct h1 as [h1 h2].
  rewrite empty_family_union in h2.
  symmetry in h2. apply empty_set_nil in h2; auto.
Qed.

Lemma partition_faml_incl :
  forall {T:Type} (ll:faml T) (l lx:list T),
    partition_faml ll l ->
    In lx ll ->
    incl lx l.
  intros T ll l lx h1 h2.
  do 2 red in h1.
  destruct h1 as [h1 h3].
  red; intros x h4.
  rewrite  list_to_set_in_iff.
  rewrite <- h3.
  econstructor.
  apply in_faml_impl.
  apply h2.
  rewrite <- list_to_set_in_iff; auto.
Qed.

    
Lemma partition_faml_cons :
  forall {T:Type} (pfdt:type_eq_dec T) (ll:faml T) (l l':list T),
    ~ Inn (faml_to_fam ll) (list_to_set l) ->
    partition_faml (l::ll) l' ->
    partition_faml ll (list_minus pfdt l' l).
  intros T h1 ll l l' h2 h3. 
  do 2 red in h3. do 2 red.
  destruct h3 as [h3 h4].
  simpl in h3. simpl.
  split.
  apply pairwise_disjoint_add_rev in h3; auto.
  simpl in h4.
  rewrite family_union_add in h4.
  rewrite list_to_set_list_minus.
  rewrite <- h4.
  rewrite Union_commutative.
  rewrite setminus_union.
  rewrite disjoint_setminus_compat; auto.
  red.   
  apply pairwise_disjoint_add_family_union; auto.
Qed.
  
  
Lemma partition_faml_eq :
  forall {T:Type} (ll:faml T) (la lb lc:list T) (a:T),
    partition_faml ll la ->
    In lb ll -> In lc ll ->
    In a lb -> In a lc ->
    list_to_set lb = list_to_set lc.
  intros T ll la lb lc a h1 h2 h3 h4 h5.
  apply in_faml_impl in h2.
  apply in_faml_impl in h3.
  rewrite list_to_set_in_iff in h4, h5.
  eapply partition_fam_eq.
  apply h1. assumption. assumption.
  apply h4. assumption.
Qed.


  
Lemma partition_faml_dec :
  forall {T:Type} (ll:faml T) (la lb lu:list T),
    partition_faml ll lu ->
    In la ll -> In lb ll ->
    {list_to_set la = list_to_set lb} + {disjointl la lb}.
  intros T ll la lb lu h1 h2 h3.
  unfold disjointl.
  apply in_faml_impl in h2.
  apply in_faml_impl in h3.
  eapply partition_fam_dec.
  apply h1.
  assumption. assumption.
Qed.

  
Fixpoint plusl (l:list nat) :=
  match l with
    | nil => 0
    | a::l' => a + (plusl l')
  end.

Fixpoint timesl (l:list nat) :=
  match l with
    | nil => 1
    | a::l' => a * (timesl l')
  end.

Lemma plusl_app :
  forall l l',
    plusl (l ++ l') =
    plusl l + plusl l'.
  intro l.
  induction l as [|a l ih]; simpl; auto.
  intro l'.
  rewrite ih, plus_assoc; auto.
Qed.

Lemma timesl_app :
  forall l l',
    timesl (l ++ l') =
    timesl l * timesl l'.
  intro l.
  induction l as [|a l ih]; simpl; auto.
  intro l'.
  rewrite ih, mult_assoc; auto.
Qed.



Lemma plusl_eq0_iff :
  forall (l:list nat),
    plusl l = 0 <->
    all_sing l 0.
  intro l; induction l as [|a l ih]; simpl; split;intro h1; auto;
  try red; try (intros; contradiction).
  apply plus_is_O in h1; destruct h1; subst.
  rename H0 into h2.
  rewrite ih in h2; red in h2.
  intros ? h3; destruct h3; subst; auto.
  hr ih.
  red in h1; red; intros c hc; specialize (h1 _ (in_cons _ _ _ hc)); auto.
  rewrite <- ih in hr.
  rewrite hr, <- plus_n_O.
  red in h1; specialize (h1 _ (in_eq _ _)); auto.
Qed.

Lemma plusl_eq1_iff :
  forall (l:list nat),
    plusl l = 1 <->
    exists! l0 l1,
      l = l0 ++ 1::nil ++ l1 /\
      all_sing l0 0 /\
      all_sing l1 0. 
  intro l; induction l as [|a l ih]; simpl. 
  split; intro h1.
  discriminate.
  destruct h1 as [l0 [h1 h2]].
  destruct h1 as [l1 h1].
  red in h1.
  destruct h1 as [[h1 [h3 h5]] h4].
  symmetry in h1.
  apply app_eq_nil in h1.
  destruct h1; discriminate.
  split; intro h1.
  apply plus_is_one in h1.
  destruct h1 as [h1 | h1]; destruct h1 as [? h1]; subst.
  rewrite ih in h1.
  destruct h1 as [l0 [h1 h2]]; destruct h1 as [l1 h1].
  red in h1.
  destruct h1 as [[h1 [h3 h5]] h4]; subst; simpl in h4.  
  exists (0::l0); red; split.
  exists l1; red; split.
  split; simpl; auto.
  split; red; intros x h6.
  destruct h6 as [|h6]; subst; auto.
  specialize (h5 _ h6); auto.
  simpl.
  intros l h6.
  destruct h6 as [h6 h7].
  apply cons_inj in h6.
  destruct h6 as [h6 h8]; clear h6. 
  apply app_inj_eq_length in h8; auto. 
  destruct h8 as [h8 h9].  
  clear h8.
  inversion h9; auto.
  intros l' h6.
  specialize (h2 (tl l')).
  hfirst h2.
  destruct h6 as [l2 [[h6 [h7 h9]] h8]]; simpl in h6.
  exists l2; simpl.
  red.
  split. split.
  pose proof (f_equal (@tl _) h6) as h10; simpl in h10.
  rewrite h10.
  rewrite tl_app; auto.
  intro; subst.
  inversion h6.
  split; auto.
  apply all_sing_tl; auto.
  intros l h11.
  destruct h11 as [h11 [h12 h13]].
  apply h8.
  split.
  simpl.
  rewrite h11.
  assert (h14:l' <> nil).
  intro; subst; inversion h6.
  rewrite (hd_compat' _ h14); simpl.
  f_equal.
  red in h7.
  symmetry.
  apply h7.
  apply in_hd'.
  split; auto.
  specialize (h2 hf); subst.
  destruct h6 as [l h6].
  red in h6.
  destruct h6 as [h6 h7].
  destruct h6 as [h6 [h8 h9]].
  assert (h10:l' <> nil).
  intro; subst; inversion h6.
  rewrite (hd_compat' _ h10); simpl. 
  f_equal.
  symmetry.
  red in h8.
  apply h8.
  apply in_hd'. 
  rewrite plusl_eq0_iff in h1.
  exists nil.
  red.
  split; simpl.
  exists l.
  red; split.
  split; auto.
  split; auto.
  red.
  intros; contradiction.
  intros l' h2.
  destruct h2 as [h2 [h3 h4]].
  inversion h2; auto.
  intros l' h2.
  destruct h2 as [l0 h3].
  red in h3.
  destruct h3 as [h3 h4].
  destruct h3 as [h3 [h6 h7]].
  destruct l' as [|a' l']; auto.
  simpl in h3.
  inversion h3; subst.
  red in h6.
  specialize (h6 _ (in_eq _ _)). 
  discriminate. 
  destruct h1 as [l0 h1].
  red in h1.
  destruct h1 as [h1 h2].
  destruct h1 as [l1 h1].
  red in h1.
  destruct h1 as [[h1 [h3 h5]] h4].
  pose proof (f_equal plusl h1) as h6.
  simpl in h6.
  rewrite plusl_app in h6; simpl in h6.
  rewrite h6.
  rewrite plus_S_shift_r.
  f_equal.
  rewrite <- plusl_eq0_iff in h3, h5.
  rewrite h3, h5; auto with arith.
Qed.


Lemma timesl_eq0_iff :
  forall (l:list nat),
    timesl l = 0 <->
    In 0 l.
  intro l; induction l as [|a l ih]; simpl; split; intro h1;
  try discriminate; try contradiction.
  apply mult_is_O in h1; destruct h1 as [|h1]; subst.
  left; auto.
  rewrite ih in h1; right; auto.
  destruct h1 as [|h1]; subst.
  apply mult_0_l.
  rewrite <- ih in h1; rewrite h1, mult_0_r; auto.
Qed.


  
Lemma timesl_eq1_iff :
  forall (l:list nat),
    timesl l = 1 <->
    all_sing l 1.
  intro l; induction l as [|a l ih]; simpl; split;intro h1; auto;
  try red; try (intros; contradiction).
  apply mult_is_one in h1; destruct h1; subst.
  rename H0 into h2.
  rewrite ih in h2; red in h2.
  intros ? h3; destruct h3; subst; auto.
  hr ih.
  red in h1; red; intros c hc; specialize (h1 _ (in_cons _ _ _ hc)); auto.
  rewrite <- ih in hr.
  rewrite hr, mult_1_r.
  red in h1; specialize (h1 _ (in_eq _ _)); auto.
Qed.


  
Lemma plusl_nprod_eq :
  forall (l:list nat),
    plusl l =
    plus_nprod (nprod_of_list _  l).
  intro l; induction l; simpl; auto.
Qed.

Lemma timesl_nprod_eq :
  forall (l:list nat),
    timesl l =
    times_nprod (nprod_of_list _  l).
  intro l; induction l; simpl; auto.
Qed.

Lemma plus_nprod_eq :
  forall {n} (x:nat^n),
    plus_nprod x =
    plusl (nprod_to_list _ _ x).
  intro n; induction n; simpl; auto;
  intro x; destruct x; simpl; auto.
Qed.

Lemma times_nprod_l :
  forall {n} (x:nat^n),
    times_nprod x =
    timesl (nprod_to_list _ _ x).
  intro n; induction n; simpl; auto;
  intro x; destruct x; simpl; auto.
Qed.


Lemma length_app_list_nprod_to_list :
  forall {T n} (x:(list T)^n),
    length (app_list (nprod_to_list _ _ x)) =
    plus_nprod (map_nprod (fun i pf => length (nth_prod i pf x))).
  intros T n; induction n as [|n ih]; simpl; auto.
  intro x; destruct x as [l x]; simpl;
  specialize (ih x);
  rewrite app_length; f_equal; rewrite ih; repeat f_equal;
  apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro;
  unfold seg_fun_to_S; repeat f_equal; apply proof_irrelevance.
Qed.
 

Definition list_to_set_injable
           {T:Type} (P:list T->Prop) : Prop :=
  forall (l l':list T),
    P l ->
    P l' ->
    list_to_set l = list_to_set l' ->
    l = l'.

Lemma list_to_set_injable_in_cons :
  forall {T:Type} (ll:faml T) (la:list T),
    list_to_set_injable (fun lx => In lx (la::ll)) ->
    list_to_set_injable (fun lx => In lx ll).
  intros T ll la h1.
  red in h1; red.
  intros lb lc h2 h3.
  specialize (h1 _ _
                 (in_cons _ _ _ h2)
                 (in_cons _ _ _ h3)).
  intro h4.
  apply h1 in h4; auto.
Qed.

Lemma list_to_set_injable_in_remove :
  forall {T:Type} (pfdt:type_eq_dec (list T)) (ll:faml T) (la:list T),
    list_to_set_injable (fun lx => In lx ll) ->
    list_to_set_injable (fun lx => In lx (remove pfdt la ll)).
  intros T h1 ll la h2.
  red in h2; red; intros l l' h3 h4 h5.
  rewrite in_remove_iff in h3, h4.
  destruct h3 as [h3 ?], h4 as [h4 ?].
  apply h2; auto.
Qed.

              
Lemma partition_faml_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (ll:faml T) (l l':list T),
    In l' ll ->
    list_to_set_injable (fun xl => In xl ll) ->
    partition_faml ll l ->
    partition_faml (remove (list_eq_dec pfdt) l' ll)
                   (list_minus pfdt l l').
  intros T h1 ll l l' h2 h3 h4.
  pose proof (faml_to_fam_decompose (list_eq_dec h1) ll l' h2) as h6.
  destruct h4 as [h4 h5].
  rewrite h6 in h4, h5; clear h6.
  do 2 red in h4; do 2 red.
  split.
  apply pairwise_disjoint_add_rev in h4; auto.
  rewrite family_union_add in h5.
  rewrite list_to_set_list_minus.
  rewrite <- h5.
  rewrite Union_commutative.
  rewrite setminus_union.
  rewrite disjoint_setminus_compat; auto.
  red.
  apply pairwise_disjoint_add_family_union; auto.
  intro h6.
  apply in_faml_to_fam_impl in h6.
  destruct h6 as [la [h6 h7]].
  rewrite in_remove_iff in h6.
  destruct h6 as [h6 h8].
  apply h3 in h7; subst.
  contradict h8; auto.
  assumption.
  assumption.
Qed.



Lemma list_to_set_bij : forall {T:Type} (l:list T), 
  exists f:{x:T | In x l} -> {x:T | Inn (list_to_set l) x}, 
  bijective f.
intros T l.
pose proof (list_to_set_in_iff l) as h1.
assert (h2:forall x : T, In x l -> Inn (list_to_set l) x).
  intros. rewrite <- h1; assumption. 
exists (fun t:{x : T | In x l} => exist _ (proj1_sig t) (h2 _ (proj2_sig t))).

red. split.
red. intros ? ? h3.  apply exist_injective in h3. 
apply proj1_sig_injective. assumption.

red. intros y.
destruct (h1 (proj1_sig y)) as [? h1r].
specialize (h1r (proj2_sig y)).
exists (exist _ (proj1_sig y) h1r).
simpl. rewrite unfold_sig.
apply existTexist. apply subsetT_eq_compat. reflexivity.
Qed.

Lemma list_to_set_finitet : forall {T:Type} (l:list T), 
  FiniteT {x:T | In x l}.
intros T l.
pose proof (list_to_set_bij l) as h1.
pose proof (list_to_set_finite l) as h2.
pose proof (Finite_ens_type _ h2) as h3.
destruct h1 as [f h1].
pose proof (bijective_impl_invertible _ h1) as h4.
pose proof (invertible_impl_inv_invertible _ h4) as h5.
apply (bij_finite _ _ _ h3 h5).
Qed.

(*See [finite_set_list_no_dup] for a stronger version.*)
Lemma finite_set_list : forall {T:Type} (E:Ensemble T), Finite E ->
  exists (l:list T), E = list_to_set l.
intros T E h1.
induction h1 as [| E h2 h3 t h4].

exists nil.
unfold list_to_set. reflexivity.

destruct h3 as [l h5].
exists (t::l).
unfold list_to_set.
unfold list_to_set in h5.
rewrite <- h5.
reflexivity.
Qed.


Lemma list_to_set_app : 
  forall {T:Type} (l l':list T),
    list_to_set (l ++ l') = 
    Union (list_to_set l) (list_to_set l').
intros T l.
induction l as [|a l h1]; simpl.
intro l'. destruct l' as [|b l']; simpl.
rewrite empty_union.
reflexivity.
rewrite empty_union.
reflexivity.
intro l'.
rewrite h1.
unfold Add.
rewrite <- Union_associative.
rewrite (Union_commutative (list_to_set l') (Singleton a)).
rewrite Union_associative.
reflexivity.
Qed.



Lemma add_cons_compat : forall {T:Type} (l:list T) (a:T),
                          Add (list_to_set l) a =
                          list_to_set (a::l).
intros. simpl. auto.
Qed.


Lemma subtract_remove_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T), 
    Subtract (list_to_set l) x =
    list_to_set (remove pfdt x l).
intros T h0 l.
induction l; simpl.
apply  sub_empty.

intro x. simpl.
destruct (h0 x a) as [h2 | h3].
subst.
destruct (in_dec h0 a l) as [h4 | h5].
rewrite list_to_set_in_iff in h4.
pose proof (sub_add_same_in _ _ h4) as h6.
rewrite h6.
apply IHl.
rewrite <- (remove_not_in' h0 l a).
rewrite list_to_set_in_iff in h5.
apply sub_add_same_nin; auto. auto.
rewrite sub_add_comm.
rewrite IHl.
apply add_cons_compat.
apply neq_sym_iff; auto.
Qed.


Lemma list_to_set_map_remove :
  forall {T U:Type}
         (pfdt:type_eq_dec T)
         (pfdu:type_eq_dec U)
         (f:T->U) (l:list T) (x:T)
         (pf:In x l),
    Injective f ->
    list_to_set (map f (remove pfdt x l)) =
    Subtract (list_to_set (map f l)) (f x).
  intros T U hdt hdu f l x h2 hinj.
  rewrite (subtract_remove_compat hdu).
  apply Extensionality_Ensembles; red; split; red; intros c h3; rewrite <- list_to_set_in_iff in h3; rewrite <- list_to_set_in_iff.
  rewrite in_map_iff in h3.
  destruct h3 as [a [h3 h4]]; subst.
  rewrite in_remove_iff in h4.
  rewrite in_remove_iff.
  destruct h4 as [h4 h5].
  split.
  apply in_map; auto.
  contradict h5.
  apply hinj in h5; auto.
  rewrite in_remove_iff in h3.
  destruct h3 as [h3 h4].
  rewrite in_map_iff in h3.
  destruct h3 as [a [h3 h5]]; subst.
  apply in_map.
  rewrite in_remove_iff.
  split; auto.
  intro; subst.
  contradict h4; auto.
Qed.


Definition p_mem_l {T:Type} (P:T->Prop) (l:list T) : Prop :=
  forall x:T, In x l -> P x.

Lemma p_mem_l_iff :
  forall {T:Type} (P:T->Prop) (l:list T),
    p_mem_l P l <-> p_mem P (list_to_set l).
  intros T P l.
  split.
  intro h1.
  red in h1. red.
  intros x h2.
  apply h1.
  rewrite list_to_set_in_iff.
  assumption.
  intro h1.
  red in h1. red.
  intros x h2.
  apply h1.
  rewrite <- list_to_set_in_iff.
  assumption.
Qed.

Lemma p_mem_l_cons :
  forall {T:Type} (P:T->Prop) (l:list T) (a:T),
    p_mem_l P (a::l) ->
    p_mem_l P l.
  intros T P l a h1. 
  red in h1. red.
  intros x h2.
  apply h1; right; auto.
Qed.

Lemma p_mem_l_type_dep_eq_dec_impl :
  forall {T:Type} (P:T->Prop) (l:list T),
    p_mem_l P l ->
    type_dep_eq_dec P ->
    type_dep_eq_dec (fun x => P x /\ In x l).
  intros T P l.
  induction l as [|c l h1]; simpl.
  intros h1 h2.
  red.
  intros ? ? h3.
  destruct h3; contradiction.
  intros h2 h3.
  apply p_mem_l_cons in h2.
  specialize (h1 h2 h3).
  red in h1. red.
  intros a b h4 h5.
  destruct h4 as [h4 h6], h5 as [h5 h7].
  red in h3. specialize (h3 a b h4 h5).
  assumption.
Qed.

Definition no_dup_mem {T:Type} (ll:faml T) :=
  p_mem_l (@NoDup T) ll.

Lemma no_dup_mem_cons :
  forall {T:Type} (ll:faml T) (la:list T),
    no_dup_mem (la::ll) ->
    no_dup_mem ll.
  intros ? ? ? h1; do 2 red in h1; do 2 red; intros; apply h1; right; auto.
Qed.

Lemma no_dup_mem_remove :
  forall {T:Type} (pfdt:type_eq_dec (list T)) (ll:faml T) (la:list T),
    no_dup_mem ll ->
    no_dup_mem (remove pfdt la ll).
  intros ? ? ? ? h1; do 2 red in h1; do 2 red; intros ? h0; apply h1.
  rewrite in_remove_iff in h0.
  destruct h0; auto.
Qed.


(*Are indices of some list.*)
Inductive are_inds (T:Type) (ln:list nat) : Prop :=
| are_inds_intro :
    forall l:list T,
      p_mem_l (fun i => i < length l) ln ->
      are_inds T ln.

Definition lexcl_mid {T:Type} (l:list T) (P:T->Prop) : Type :=
  forall x:T, In x l -> {P x} + {~P x}.


Lemma in_dep_dec :
  forall {T:Type} {P:T->Prop},
    type_dep_eq_dec P ->
    forall (l:list T),
      p_mem_l P l ->
      forall (x:T) (pfx:P x),
        {In x l} + {~In x l}.
  intros T P h1 l.
  induction l as [|a l ih]; simpl.
  intros; right; intro; contradiction.
  intros h2 x h3.
  hfirst ih.
    red. red in h2.
    intros; apply h2. right; auto.
  specialize (ih hf).
  specialize (ih x h3).
  red in h2.
  specialize (h2 a (in_eq _ _)).
  red in h1. specialize (h1 _ _ h2 h3).
  destruct h1 as [h1 | h1]; subst.
  left; left; auto.
  destruct ih as [h4 | h4].
  left; right; auto.
  right.
  intro h5.
  destruct h5 as [h5 | h5]; subst.
  contradict h1; auto.
  contradiction.
Qed.


Lemma in_neq :
  forall {T} (l:list T) x a,
    In x l ->
    ~In a l ->
    x <> a.
  intros; intro; subst; auto.
Qed.

Lemma in_cons_neq :
  forall {T} (l:list T) x a,
    In x (a::l) ->
    x <> a ->
    In x l.
  intros ? ? ? ? h1; destruct h1; subst; auto; intro h1;
  contradict h1; auto.
Qed.




Fixpoint in_dep {T:Type} {P:T->Prop} (x:T) (pfx:P x) (l:list T)  {struct l} : Prop  :=
    match l with
      | nil => False
      | a::l' => P a /\ x = a \/ (in_dep x pfx l')
    end.

Lemma impl_in_dep :
  forall {T:Type} {P:T->Prop} (x:T) (pfx:P x) (l:list T),
    In x l -> in_dep x pfx l.
  intros T P x h1 l. revert h1; revert x.  
  induction l as [|a l ih]; simpl; auto.
  intros ? ? h2. 
  destruct h2;
    [left | right]; auto; subst; auto.
Qed.

Lemma in_dep_iff :
  forall {T:Type} {P:T->Prop} (x:T) (pfx:P x) (l:list T),
    in_dep x pfx l <-> (In x l /\ P x).
  intros T P x h1 l.
  split.
  induction l as [|l a ih]; simpl; auto.
  intro h3.
  destruct h3 as [h3 | h3]; subst. 
  split; try left; auto.
  destruct h3; auto.
  specialize (ih h3).
  destruct ih as [h2 h5].
  split; auto.
  intro h2; destruct h2 as [h2 h3]; apply impl_in_dep; auto.
Qed.


Fixpoint remove_dep {T:Type} (P:T->Prop)
         (pfe:type_dep_eq_dec P)
         (l:list T)  (x:T) (pfp:P x) {struct l} :
  p_mem_l P l -> list T :=
  match l with
    | nil => (fun _   => nil)
    | a::l' =>
      (fun pfpm =>
         let r := remove_dep P pfe l' x pfp (p_mem_l_cons _ _ _ pfpm) in
         if (pfe x a pfp (pfpm a (in_eq _ _))) then
           r
         else a::r)
  end.

Lemma remove_dep_not_in :
  forall {T:Type} (P:T->Prop) (pfe:type_dep_eq_dec P) (l:list T) (x:T) (pfp:P x) (pfpm:p_mem_l P l),
    ~In x l ->
    remove_dep P pfe l x pfp pfpm = l.
  intros T P h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros x h2 h3 h4.
  destruct_head_match as [h5 | h5]; subst.
  contradict h4. left; auto.
  f_equal.
  rewrite ih; auto.
Qed.




Lemma add_list_to_set_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    In a l ->
    Add (list_to_set (remove pfdt a l)) a = list_to_set l.
  intros T h0 l a h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2 | x h2].
  rewrite <- subtract_remove_compat in h2.
  destruct h2; auto.
  destruct h2.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- subtract_remove_compat.
  rewrite add_sub_compat_in; auto.
  rewrite <- list_to_set_in_iff; auto.
Qed.






Definition seg_list (n:nat) := seq 0 n.

Lemma seg_list0 :
  seg_list 0 = nil.
  auto.
Qed.


Lemma length_seg_list :
  forall (n:nat),
    length (seg_list n) = n.
  intro; unfold seg_list; rewrite seq_length; auto.
Qed.

Lemma seg_list_eq_nil :
  forall (n:nat),
    seg_list n = nil ->
    n = 0.
  intro n.
  destruct n; auto.
  intro h1.
  unfold seg_list, seq in h1.
  discriminate.
Qed.


Lemma seq_eq_seq_0 :
  forall (m n:nat),
    seq m n = map (plus m) (seg_list n).
  intros m n. revert m.
  induction n as [|n h1]; simpl; auto.
  intro m.
  rewrite plus_comm.
  simpl.
  f_equal. 
  rewrite <- (seq_shift n 0) at 1.
  rewrite map_map.
  rewrite h1.
  f_equal.
  apply functional_extensionality.
  intros; omega.
Qed.


Lemma seg_list_seq_cons :
  forall m:nat,
    0 < m ->
    seg_list m = 0::seq 1 (pred m).
  intro m. 
  destruct m as [|m].
  intro h1; apply lt_irrefl in h1.
  contradiction. 
  intro h1.
  destruct (zerop m) as [|h2]; subst.
  unfold seg_list; simpl; auto.
  unfold seg_list.
  simpl.
  auto.
Qed.



Lemma list_to_set_seg_list_eq :
  forall (n:nat),
    list_to_set (seg_list n) = seg n.
  intro n.
  induction n as [|n h1]; simpl.
  rewrite seg_0_eq.
  reflexivity.
  rewrite seq_eq_seq_0.
  rewrite map_im_compat at 1.
  rewrite h1.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2 | x h2].
  destruct h2 as [x h2]. subst.
  destruct h2 as [h2].
  constructor. omega.
  destruct h2.
  constructor. omega.
  destruct h2 as [h2].
  destruct (zerop x) as [h3 | h3].
  subst. right. constructor.
  left.
  apply Im_intro with (pred x).
  constructor. omega.
  omega.
Qed.

Lemma list_to_set_eq_seg_in :
  forall (l:list nat) (m n:nat),
    list_to_set l = seg n ->
    m < n ->
    In m l.
  intros l m n h1 h2.
  pose proof (in_seg _ _ h2) as h3.
  rewrite <- h1 in h3.
  rewrite <- list_to_set_in_iff in h3; auto.
Qed.

Lemma list_to_set_eq_seg_lt :
  forall (l:list nat) (m n:nat),
    list_to_set l = seg n ->
    In m l ->
    m < n.
  intros l m n h1 h2.
  rewrite list_to_set_in_iff in h2.
  rewrite h1 in h2.
  destruct h2; auto.
Qed.

Lemma in_seg_list_iff :
  forall (m n:nat),
    In m (seg_list n) <-> m < n.
  intros m n.
  rewrite list_to_set_in_iff.
  rewrite list_to_set_seg_list_eq.
  split.
  intro h1. destruct h1; auto.
  intros; constructor; auto.
Qed.



Lemma in_seq_iff :
  forall i m n,
    In i (seq m n) <->
     m <= i < m + n.
  intros i m n.
  revert i m.
  induction n as [|n ih]; simpl; intros i m; split; intro h1.
  contradiction.
  omega.
  destruct h1 as [|h1]; subst.
  omega.
  specialize (ih i (S m)).
  rewrite ih in h1; omega.
  destruct h1 as [h1 h2].
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1|]; subst.
  right.
  rewrite ih.
  omega.
  left; auto.
Qed.

Lemma hd_seg_list' :
  forall (n:nat) (pf:seg_list n <> nil),
    hd' (seg_list n) pf = 0.
  intro n.
  destruct n as [|n]; simpl; auto.
  intro h1; contradict h1; auto.
Qed.

Lemma tl_seg_list :
  forall (n:nat),
    tl (seg_list n) = seq 1 (pred n).
  intro n; destruct n; simpl; auto.
Qed.


Lemma seq_S :
  forall (m n:nat),
    0 < m  ->
    seq n m = n::seq (S n) (pred m).
  intro m.
  induction m as [|m ih]; simpl.
  intros; omega.
  intros; simpl; auto.
Qed.


Lemma seq_nil :
  forall (m n:nat),
    seq m n = nil ->
    n = 0.
  intros m n.
  revert m.
  destruct n as [|n]; auto.
  intros m h1.
  rewrite seq_S in h1.
  discriminate.
  auto with arith.
Qed.

Lemma seq_app :
  forall (m n:nat),
    0 < m ->
    seq n m = seq n (pred m) ++ n+m-1::nil.
  intro m.
  induction m as [|m ih]; simpl.
  intros; omega.
  intros n h1; simpl.
  destruct m as [|m]; simpl.
  f_equal. omega.
  f_equal.
  red in h1.
  apply le_S_n in h1.
  apply le_lt_eq_dec in h1.
  destruct h1 as [h1 | h1].
  simpl in ih.
  specialize (ih (S n) h1).
  rewrite ih.
  f_equal. f_equal.
  omega.
  discriminate.
Qed.


Lemma seg_list_seq_app :
  forall m n:nat,
    m <= n ->
    seg_list n = seg_list m ++ seq m (n - m).
  intro m.
  unfold seg_list.
  induction m as [|m ih]. 
  simpl.
  intros; f_equal; auto with arith. 
  intros n h1.
  destruct (zerop m) as [|h2]; subst.
  simpl.
  rewrite seg_list_seq_cons.
  f_equal.
  simpl.
  f_equal; omega.
  omega. 
  rewrite ih; try omega.
  simpl.
  rewrite (S_pred _ _ h2) at 1.
  simpl.
  f_equal.
  pose proof @seq_S.
  rewrite (seq_S (n - m) m); try omega.
  rewrite (seq_app _ _ h2).
  simpl.
  rewrite <- app_assoc.
  rewrite <- minus_n_O.
  simpl.
  repeat f_equal; try omega.
Qed.

Lemma seg_list_S :
  forall n, 
    seg_list (S n) = seg_list n ++ (n :: nil).
  unfold seg_list; intro; simpl.
  destruct (zerop n) as [|h1]; subst; simpl; auto.
  symmetry. 
  rewrite seq_S; simpl.
  f_equal.
  symmetry.
  rewrite seq_app; simpl.
  f_equal.
  f_equal; auto with arith.
  assumption.
  assumption.
Qed.


Lemma firstn_seq :
  forall (m n r:nat),
    firstn m (seq n r) =
    seq n (min m r).
  intros m n r.
  destruct (le_lt_dec m r) as [h1 | h1].
  rewrite min_l; auto.
  revert h1 n. revert r.
  induction m as [|m ih]; simpl; auto.
  intros r h1 n.
  specialize (ih (pred r)).
  hfirst ih.
  apply le_S_n.
  erewrite <- S_pred; auto.
  assert (h2:0 < r). omega. 
  apply h2.
  specialize (ih hf).
  rewrite seq_S.
  f_equal.
  apply ih.
  omega.
  rewrite min_r; auto with arith.
  pose proof (seq_length r n) as h2.
  rewrite <- h2 in h1.
  apply firstn_length_ge.
  omega.
Qed.


Lemma seg_list_app :
  forall (n:nat),
    0 < n ->
    seg_list n = seg_list (pred n) ++ n-1::nil.
  intros; unfold seg_list; rewrite seq_app; auto.
Qed.

Lemma seg_list_not_nil :
  forall (n:nat),
    0 < n ->
    seg_list n <> nil.
  intros; rewrite seg_list_app; auto.
  apply neq_sym.
  apply app_cons_not_nil.
Qed.

Lemma intersection_seg_sing :
  forall (n:nat) (m:nat),
    m >= n ->
    Intersection (list_to_set (seg_list n)) (Singleton m) = Empty_set _.
  intros m n h1; apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2 h3].
  destruct h3.
  rewrite <- list_to_set_in_iff in h2.
  rewrite in_seg_list_iff in h2.
  omega.
  destruct h2; contradiction.
Qed.


Lemma seg_list_eq_sing :
  forall (n a:nat),
    seg_list n = a::nil ->
    n = 1 /\ a = 0.
  intros n a.
  destruct n as [|n]; simpl.
  rewrite seg_list0.
  intro; discriminate.
  intro h1.
  rewrite seg_list_S in h1.
  rewrite app_eq_iff in h1.
  destruct h1 as [h1 h2].
  rewrite length_seg_list in h1, h2.
  destruct n as [|n].
  simpl in h1, h2.
  inversion h2; auto.
  simpl in h1, h2.
  rewrite skipn_nil in h2.
  discriminate.
Qed.

Lemma seg_list_eq_couple :
  forall (n a b:nat),
    seg_list n = a::b::nil ->
    n = 2 /\ a = 0 /\ b = 1.
  intros n a b.
  destruct n as [|n]; simpl.
  rewrite seg_list0.
  intro; discriminate.
  destruct n as [|n]; simpl.
  rewrite seg_list_S.
  rewrite seg_list0.
  simpl.
  intro h1.
  inversion h1; discriminate.
  do 2 rewrite seg_list_S.
  intro h1.
  destruct a as [|a], b as [|b]; auto with arith.
  simpl in h1.
  rewrite app_eq_iff in h1.
  destruct h1 as [h1 h2].
  rewrite app_length in h2.
  rewrite length_seg_list in h2.
  simpl in h2.
  rewrite S_compat in h2.
  simpl in h2.
  destruct n as [|n].
  simpl in h2.
  inversion h2.
  simpl in h2.
  rewrite skipn_nil in h2.
  discriminate.
  pose proof (f_equal (fun l => last l 0) h1) as h2.
  simpl in h2.
  rewrite last_app_sing in h2.
  apply S_inj in h2; subst.
  pose proof (f_equal (@removelast _) h1) as h2.
  simpl in h2.
  rewrite removelast_app_sing in h2.
  destruct b as [|b]; auto.
  rewrite app_eq_iff in h2.
  destruct h2 as [h2 h3].
  rewrite length_seg_list in h3.
  simpl in h3.
  rewrite skipn_nil in h3.
  discriminate.
  rewrite app_eq_iff in h1.
  destruct h1 as [h1 h2].
  rewrite app_length in h2.
  rewrite length_seg_list in h2.
  simpl in h2.
  rewrite S_compat in h2.
  simpl in h2.
  destruct n as [|n].
  simpl in h2.
  inversion h2.
  simpl in h2.
  rewrite skipn_nil in h2.
  discriminate.
  rewrite app_eq_iff in h1.
  destruct h1 as [h1 h2].
  rewrite app_length, length_seg_list in h2; simpl in h2.
  rewrite S_compat in h2; simpl in h2.
  destruct n as [|n].
  simpl in h2.
  inversion h2.
  subst.
  simpl in h1.
  inversion h1.
  simpl in h2.
  rewrite skipn_nil in h2.
  discriminate.
Qed.


Lemma seg_list_eq_cons :
  forall (l:list nat) (n a:nat),
    seg_list n = a::l ->
    a = 0 /\ l = seq 1 (pred n).
  intros l n. revert l. 
  induction n as [|n ih].
  intros l a h1.
  rewrite seg_list0 in h1.
  discriminate.
  intros l a h1. 
  rewrite seg_list_S in h1. 
  destruct (nil_dec' l) as [|h2]; subst. 
  rewrite app_eq_iff in h1.
  destruct h1 as [h1 h2].
  rewrite length_seg_list in h1.
  rewrite length_seg_list in h2.
  symmetry in h2.
  destruct (zerop n) as [|h3]; subst.
  simpl in h2; simpl.
  inversion h2; auto.
  rewrite (S_pred _ _ h3) in h2.
  simpl in h2.
  rewrite skipn_nil in h2.
  discriminate.
  simpl.
  rewrite (app_removelast_last a h2) in h1 at 1. 
  rewrite app_comm_cons in h1.
  rewrite app_sings_eq_iff in h1.
  destruct h1 as [h1 ?]; subst.
  rewrite seq_app.
  simpl.
  rewrite <- minus_n_O.
  apply ih in h1.
  destruct h1 as [? h1]; subst.
  split; auto.
  rewrite <- h1.
  apply app_removelast_last; auto.
  destruct (zerop (last l a)) as [h3 | h3].
  rewrite h3 in h1.
  rewrite seg_list0 in h1.
  discriminate.
  assumption.
Qed.


Fixpoint inv_iml {T U:Type} (pfdu:type_eq_dec U)
           (f:T->U) (l:list T) (u:U) : list T :=
  match l with
    | nil => nil
    | a::l' => let il := inv_iml pfdu f l' u in
               if (pfdu u (f a)) then
                    a :: il
                 else
                   il
  end.

Lemma inv_iml_functional :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (f f':T->U) (l l':list T) (u u':U),
    f = f' ->
    l = l' ->
    u = u' ->
    inv_iml pfdu f l u =
    inv_iml pfdu f' l' u'.
  intros; subst; auto.
Qed.

Lemma in_inv_iml_iff :
  forall {T U:Type} (pfdu:type_eq_dec U)
           (l:list T) (u:U) (f:T->U) (x:T),
    In x (inv_iml pfdu f l u) <->
    In x l /\ f x  = u.
  intros T U h1 l.
  induction l as [|a l ih]; simpl.
  intros; split.
  intro; contradiction.
  intro h2. destruct h2; auto.
  intros u f x.
  lr_if_goal.
  fold hlr.
  destruct hlr as [h2 | h2]; subst.
  split.
  intro h2.
  destruct h2 as [|h2]; subst.
  intuition.
  specialize (ih (f a) f x).
  rewrite ih in h2.
  destruct h2 as [h2 h3].
  intuition.
  intro h2.
  destruct h2 as [h2 h3].
  destruct h2 as [|h2]; subst.
  left; auto.
  right.
  rewrite ih.
  split; auto.
  split.
  intros h3.
  rewrite ih in h3.
  destruct h3 as [h3 h4]; subst.
  intuition.
  intro h3.
  destruct h3 as [h3]; subst.
  destruct h3 as [|h3]; subst.
  contradict h2; auto.
  rewrite ih.
  intuition.
Qed.

Lemma incl_inv_iml :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (f:T->U) (l:list T) (u:U),
    incl (inv_iml pfdu f l u) l.
  intros; red; intros.
  rewrite in_inv_iml_iff in H.
  destruct H; auto.
Qed.

Lemma inv_iml_eq_nil_iff :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:T->U) (l:list T) (b:U),
    inv_iml pfdu f l b = nil <->
    (l = nil \/ (forall a, In a l -> f a <> b)).
  intros T U h1 h1' f l.
  induction l as [|a l ih]; simpl.
  intuition.
  intros b.
  destruct (h1' b (f a)) as [|h2]; subst.
  split; intro h2.
  discriminate.
  destruct h2 as [|h2].
  discriminate.
  specialize (h2 a (or_introl eq_refl)).
  contradict h2; f_equal; apply proof_irrelevance.
  specialize (ih b).
  rewrite ih.
  split; intro h3.
  destruct h3 as [|h3]; subst.
  right.
  intros a' h3.
  destruct h3; subst; try contradiction.
  contradict h2.
  rewrite <- h2.
  f_equal; apply proof_irrelevance.
  right.
  intros a' h4.
  destruct (h1 a a') as [|h5]; subst.
  contradict h2; rewrite <- h2; f_equal; apply proof_irrelevance.
  destruct h4 as [|h4]; subst.
  contradict h5; auto.
  apply h3; auto.
  destruct h3 as [|h3].
  discriminate.
  destruct (nil_dec' l) as [h4 | h4].
  left; auto.
  right.
  intros a' h5.
  apply h3; auto.
Qed.


Lemma inv_iml_sing_eq :
  forall {T U:Type} {l:list T} {a:T} (pfdu:type_eq_dec U)
         (f:T->U)
         (u:U),
    u = f a  ->
    inv_iml pfdu f (a::nil) u = a::nil.
  intros T U l a h1 f u h2.
  subst.
  simpl.
  lr_if_goal.
  fold hlr.
  destruct hlr as [h2 | h2]; auto.
  contradict h2; auto.
Qed.


Lemma count_occ_map :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:T->U) (l:list T) (y:U),
    count_occ pfdu (map f l) y =
    length (inv_iml pfdu f l y).
  intros T U h1 h2 f l.
  induction l as [|a l ih]; simpl; auto.
  intro y.
  rewrite switch_eq_dec_eq at 1.
  destruct (h2 y (f a)) as [h4 | h4]; subst.
  unfold switch_eq_dec; simpl. 
  f_equal; apply ih.
  unfold switch_eq_dec; auto.
Qed.

(*See way below for [lS_app]*)

  
Definition famlS (ll:faml nat) :=
  map lS ll.

Definition faml_lpred (ll:faml nat) :=
  map lpred ll.


Lemma famlS_nil : famlS nil = nil.
  auto.
Qed.

Lemma famlS_eq_nil_iff :
  forall (ll:faml nat),
    famlS ll = nil <-> ll = nil.
  intro ll; split; destruct ll; simpl; intros; try discriminate; auto.
Qed.
  
Lemma famlS_cons :
  forall (ll:faml nat) (al:list nat),
    famlS (al::ll) =
    (lS al) :: famlS ll.
  auto.
Qed.

Lemma length_famlS :
  forall (ll:faml nat),
    length (famlS ll) = length ll.
  intro ll; induction ll; simpl; auto.
Qed.

Lemma in_famlS_iff :
  forall (ll:faml nat) (ln:list nat),
    In ln (famlS ll) <->
    exists ln':list nat,
      In ln' ll /\ ln = lS ln'.
  intros ll ln.
  unfold famlS.
  rewrite in_map_iff.
split; intro h1; destruct h1 as [n hm]; destruct hm as [? h2]; subst; exists n; tauto.
Qed.

Lemma in_famlS_iff' :
  forall (ll:faml nat) (ln:list nat),
    In ln (famlS ll) <->
    (In (lpred ln) ll /\
     ~In 0 ln).
  unfold famlS; intros ll ln.
  rewrite in_map_iff.
  split; intro h1.
  destruct h1 as [ln' [h1 h2]]; subst.
  split.
  rewrite lpred_lS; auto.
  apply nin0_lS.
  destruct h1 as [h1 h2].
  exists (lpred ln).
  split.
  rewrite lS_lpred_iff in h2.
  rewrite h2.
  f_equal.
  rewrite lpred_lS; auto.
  assumption.
Qed.

  
Lemma length_faml_lpred :
  forall (ll:faml nat),
    length (faml_lpred ll) = length ll.
  intro ll; induction ll; simpl; auto.
Qed.


Lemma in_faml_lpred_iff :
  forall (ll:faml nat) (lm:list nat),
    In lm (faml_lpred ll) <->
    exists ln:list nat,
      In ln ll /\ lm = lpred ln.
  unfold faml_lpred.
  intros ll lm; split; intro h2.
  rewrite in_map_iff in h2.
  destruct h2 as [ln [h2 h3]]; subst.
  exists ln; auto.
  destruct h2 as [ln [h3 h4]]; subst.
  rewrite in_map_iff.
  exists ln; auto.
Qed.


Lemma map_inj :
  forall {T U:Type} (f:T->U),
    Injective f ->
    forall (l l':list T),
      map f l = map f l' ->
      l = l'.
  intros T U f h1 l.
  induction l as [|a l ih]; simpl.
  intros ? h2.
  symmetry in h2; apply map_eq_nil in h2; auto.
  intros l' h2.
  destruct l' as [|a' l'].
  simpl in h2; discriminate.
  simpl in h2.
  inversion h2.
  apply h1 in H0; subst.
  f_equal; auto.
Qed.


Lemma inj_lS :
  Injective lS.
  red; intros l l' h1.
  unfold lS in h1.
  apply map_inj in h1; auto.
  red; apply S_inj.
Qed.


Lemma inj_famlS :
  forall (ll ll':faml nat),
    famlS ll = famlS ll' ->
    ll = ll'.
  intros ll ll' h1.
  unfold famlS in h1.
  apply map_inj in h1; auto.
  apply inj_lS.
Qed.




Lemma map_dep_sig_functional :
  forall {T U:Type} {l l':list T}
         (pf:l = l'),
    let V := (fun al => list {a | In a al}) in
    forall (df:dep_fun V)
           (f:{a | In a l}->U),
  map f (df l) = 
  map (transfer_dep (U:=fun al:list T=> {a|In a al}->U) pf f) (df l').
  intros; subst. rewrite transfer_dep_eq_refl; auto.
Qed.

Definition fun_in_dep {T:Type} (l:list T) (U:Type) : Type :=
  prop_dep_fun (fun x => In x l) U.

Definition fun_in_dep_prop {T:Type} (l:list T) (P:T->Prop) (U:Type) : Type :=
  prop_dep_fun2 (fun x => In x l) P U.



Definition fun_in_dep_nil (T U:Type) : fun_in_dep (@nil T) U :=
  fun x pf => False_rect _ pf.

Lemma inj_dep_nil :
  forall (T U:Type),
    inj_dep (fun_in_dep_nil T U).
  red; intros; contradiction.
Qed.

Lemma surj_dep_nil :
  forall (T:Type),
    surj_dep (fun_in_dep_nil T False).
  red; intros; contradiction.
Qed.

Lemma bij_dep_nil :
  forall (T:Type),
    bij_dep (fun_in_dep_nil T False).
  red; intros; split; [apply inj_dep_nil | apply surj_dep_nil].
Qed.

Lemma fun_in_dep_ext :
  forall {T:Type} (l:list T) (U:Type)
         (f g:fun_in_dep l U),
    (forall (x:T) (pf:In x l),
       f x pf = g x pf) -> f = g.
  intros; apply prop_dep_fun_ext; auto.
Qed.

Lemma fun_in_dep_functional :
  forall {T U:Type} {l:list T}
         (f:fun_in_dep l U)
         (x x':T)
         (pfx:In x l) (pfx':In x' l),
    x = x' ->
    f x pfx = f x' pfx'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.
  
Lemma  fun_in_dep_prop_ext :
  forall {T:Type} (l:list T) (P:T->Prop) (U:Type)
         (f g:fun_in_dep_prop l P U),
    (forall (x:T) (pf:In x l) (pfp:P x),
       f x pf pfp = g x pf pfp) -> f = g.
  intros; apply prop_dep_fun2_ext; auto.
Qed.


Definition fun_in_dep_cons
           {T U} {l:list T} (pfdt:type_eq_dec T)
           (f:fun_in_dep l U) (a:T) (b:U) : 
fun_in_dep (a::l) U :=
  fun x (pf:In x (a::l)) =>
    match (in_cons_dec pfdt _ _ _ pf) with
      | left pfe => b
      | right pfin => f x pfin
    end.


Definition seg_fun_in_dep {U:Type} {n:nat} (f:seg_fun n U) :
  fun_in_dep (seg_list n) U.
  red in f; do 2 red.
  intros m h1; eapply f.
  apply (iff1 (in_seg_list_iff _ _) h1).
Defined.

Definition remove_ind_in_dep {T U} {l:list T} m
           (f:fun_in_dep l U) : fun_in_dep (remove_ind l m) U :=
  fun x pf => f x (in_remove_ind _ _ _ pf).

Lemma inj_remove_ind_in_dep :
  forall {T U} {l:list T} m 
         (f:fun_in_dep l U),
    inj_dep f ->
    inj_dep (remove_ind_in_dep m f).
  intros ? ? ? ? ? h1; red in h1; red; intros ? ? ? ? h2;
  unfold remove_ind_in_dep in h2; apply h1 in h2; auto.
Qed.

Definition prop_in_dep {T:Type} (l:list T) : Type :=
  prop_dep_prop (fun x => In x l).

Definition prop_in_dep_u {T:Type} (l:list T) (U:Type) : Type :=
  prop_dep_prop_u (fun x => In x l) U.

Definition set_in_dep {T:Type} (U:T->Set) (l:list T) : Type :=
  prop_dep_set_dep (fun x => In x l) U.


Definition transfer_fun_in_dep {T U:Type} {l l':list T} (f:fun_in_dep l U)
           (pf:l = l') :
  fun_in_dep l' U :=
  transfer_dep pf f (U:= (fun xl => fun_in_dep xl U)).  


Definition transfer_fun_in_dep_prop {T U:Type} {l l':list T}
           {P:T->Prop} (f:fun_in_dep_prop l P U)
           (pf:l = l') :
  fun_in_dep_prop l' P U :=
  transfer_dep pf f (U:= (fun xl => fun_in_dep_prop xl P U)).  

Definition transfer_fun_in_dep_ran
           {T U V:Type} {l:list T}
           (pfe:U = V) (f:fun_in_dep l U) :
  fun_in_dep l V :=
  fun x pf => transfer pfe (f x pf).


Definition transfer_prop_in_dep {T:Type} {l l':list T} (f:prop_in_dep l)
           (pf:l = l') :
  prop_in_dep l' :=
  transfer_dep pf f.


Definition transfer_prop_in_dep_u {T U:Type} {l l':list T}
           (f:prop_in_dep_u l U) (pf:l = l') :
  prop_in_dep_u l' U :=
  transfer_dep pf f (U:= (fun xl => prop_in_dep_u xl U)).


Definition transfer_set_in_dep {T:Type} {U:T->Set} 
           {l l':list T} (f:set_in_dep U l)
           (pf:l = l') :
  set_in_dep U l' :=
  transfer_dep (U:=fun xl =>set_in_dep U xl) pf f.


Lemma transfer_fun_in_dep_eq_refl :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U),
    transfer_fun_in_dep f eq_refl = f.
  intros; unfold transfer_fun_in_dep; rewrite transfer_dep_eq_refl; auto.
Qed.

Lemma transfer_fun_in_dep_prop_eq_refl :
  forall {T U:Type} {l:list T} {P:T->Prop} (f:fun_in_dep_prop l P U),
    transfer_fun_in_dep_prop f eq_refl = f.
  intros; unfold transfer_fun_in_dep_prop; rewrite transfer_dep_eq_refl; auto.
Qed.

Lemma transfer_prop_in_dep_eq_refl :
  forall {T:Type} {l:list T} (f:prop_in_dep l),
    transfer_prop_in_dep f eq_refl = f.
  intros; unfold transfer_prop_in_dep; rewrite transfer_dep_eq_refl; auto.
Qed.


Lemma transfer_prop_in_dep_u_eq_refl :
  forall {T U:Type} {l:list T} (f:prop_in_dep_u l U),
    transfer_prop_in_dep_u f eq_refl = f.
  intros; unfold transfer_prop_in_dep_u; rewrite transfer_dep_eq_refl; auto.
Qed.

Lemma transfer_set_in_dep_eq_refl :
  forall {T:Type} {l:list T} {U:T->Set} (f:set_in_dep U l),
    transfer_set_in_dep f eq_refl = f.
  intros; unfold transfer_set_in_dep; rewrite transfer_dep_eq_refl; auto.
Qed.

Lemma transfer_fun_in_dep_compat :
  forall {T U:Type} {l l':list T} (f:fun_in_dep l U) (x:T) (pfin:In x l')
         (pfeq:l = l'),
    transfer_fun_in_dep f pfeq x pfin  =
    f x (transfer_dep_r (U:=fun xl => In x xl) pfeq pfin).
  intros; subst; rewrite transfer_fun_in_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.

Lemma transfer_fun_in_dep_prop_compat :
  forall {T U:Type} {l l':list T} {P:T->Prop}
         (f:fun_in_dep_prop l P U) (x:T) (pfin:In x l') (pfp:P x)
         (pfeq:l = l'),
    transfer_fun_in_dep_prop f pfeq x pfin pfp  =
    f x (transfer_dep_r (U:=fun xl => In x xl) pfeq pfin) pfp.
  intros; subst; rewrite transfer_fun_in_dep_prop_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.


Lemma transfer_prop_in_dep_compat :
  forall {T:Type} {l l':list T} (f:prop_in_dep l)
         (x:T) (pfin:In x l') (pfeq:l = l'),
    transfer_prop_in_dep f pfeq x pfin  =
    f x (transfer_dep_r (U:=fun xl => In x xl) pfeq pfin).
  intros; subst; rewrite transfer_prop_in_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.

Lemma transfer_prop_in_dep_u_compat :
  forall {T U:Type} {l l':list T} (f:prop_in_dep_u l U)
         (x:T) (pfin:In x l') (pfeq:l = l'),
    transfer_prop_in_dep_u f pfeq x pfin  =
    f x (transfer_dep_r (U:=fun xl => In x xl) pfeq pfin).
  intros; subst; rewrite transfer_prop_in_dep_u_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.


Lemma transfer_set_in_dep_compat :
  forall {T:Type}  {U:T->Set} {l l':list T} (f:set_in_dep U l)
         (x:T) (pfin:In x l') (pfeq:l = l'),
    transfer_set_in_dep f pfeq x pfin  =
    f x (transfer_dep_r (U:=fun xl => In x xl) pfeq pfin).
  intros; subst; rewrite transfer_set_in_dep_eq_refl, transfer_dep_r_eq_refl; auto.
Qed.


Definition compose_in_dep_prop
           {T U} {l:list U} (f:fun_in_dep l T) (R:T->Prop) :
  prop_in_dep l :=
  fun_in_dep_fun_compose f R.

Definition compose_sig_prop
           {T U} {l:list T} (f:U->{x:T|In x l})
           (R:{x:T| In x l}->Prop) :
  U -> Prop :=
  compose R f.

Definition compose_sig_prop_in_dep 
           {T U} {l:list U} (f:T->{x:U | In x l})
           (R:prop_in_dep l) :
  T->Prop := fun x => R (proj1_sig (f x)) (proj2_sig _).

           
Definition unq_ex_ex_to_fun_in_dep 
           {T U:Type} {l:list T} {f:list T->list U} {P:prop_in_dep_u l U}
                      (pf:forall y:U, In y (f l) ->
                         exists! x:T,
                           exists pfin:In x l,
                             P x pfin y) : fun_in_dep (f l) T.
  do 2 red. intros y h1. specialize (pf y h1). 
  refine (proj1_sig (constructive_definite_description _ pf)). 
Defined.


Lemma unq_ex_ex_to_fun_in_dep_compat :
  forall {T U:Type} {l:list T} {f:list T->list U} {P:prop_in_dep_u l U}
         (pf:forall (y:U) (pf:In y (f l)),
             exists! x:T,
               exists pfin:In x l,
                 P x pfin y)
         (y:U) (pfiny:In y (f l)),
    let f := unq_ex_ex_to_fun_in_dep pf  in
    let x := f y pfiny in
    exists pfin:In x l,
      P  x pfin y.
  unfold unq_ex_ex_to_fun_in_dep; intros;
  destruct constructive_definite_description; auto.
Qed.

Lemma unq_ex_ex_to_fun_in_dep_functional :
  forall {T U:Type} {l:list T} {f:list T->list U} {P:prop_in_dep_u l U}
         (pf:forall (y:U) (pf:In y (f l)),
             exists! x:T,
               exists pfin:In x l,
                 P x pfin y)
         (y y':U) (pfiny:In y (f l)) (pfiny':In y' (f l)),
    y = y' ->
    unq_ex_ex_to_fun_in_dep pf y pfiny =
    unq_ex_ex_to_fun_in_dep pf y' pfiny'.
  intros; subst.
  pose proof (proof_irrelevance _ pfiny' pfiny); subst; auto.
Qed.



Definition id_in_dep {T:Type} (l:list T) : fun_in_dep l T
  := id_prop_dep (fun x => In x l).

Lemma id_in_dep_compat :
  forall {T:Type} (l:list T) (x:T) (pf:In x l),
    id_in_dep l x pf = x.
  intros; unfold id_in_dep, id_prop_dep; auto.
Qed.

Lemma inj_id_in_dep :
  forall {T} (l:list T),
    inj_dep (id_in_dep l).
  intros; red; intros; auto.
Qed.
  


Definition id_in_dep_sig {T:Type} (l:list T) :
  {x:T|In x l} -> T :=
  id_prop_dep_sig (fun x => In x l).

Definition id_in_dep_sig' {T:Type} (l:list T) :
  fun_in_dep l {x:T|In x l} :=
  id_prop_dep_sig' (fun x => In x l).

Definition sigl_from_cons
           {T U l a}
           (f:{x:T | In x (a::l)} -> U) :
  {x:T|In x l} -> U :=
  fun x => f (exist
                (fun c => In c (a::l)) _ (in_cons _ _ _ (proj2_sig x))).

Definition sigl_with_cons
           {T U l a}
           (f:{x:T | In x (a::l)} -> U) :
  {x:T|In x (removelast (a::l)) } -> U :=
  fun x => f (exist
                (fun c => In c (a::l)) _ (removelast_in _ _ (proj2_sig x))).


Definition const_in_dep {T U:Type} (l:list T) (val:U) : fun_in_dep l U :=
  fun _ _ => val.


Definition fun_from_cons {T U:Type} {l:list T} {a:T}
           (f:fun_in_dep (a::l) U) : fun_in_dep l U :=
  fun x (pf:In x l) => f x (in_cons _ _ _ pf).


Lemma fun_from_cons_functional :
  forall {T U:Type} {l:list T} {a:T}
         (f g:fun_in_dep (a::l) U) (x:T),
    f x = g x->
    fun_from_cons f x = fun_from_cons g x.
  intros T U l a f g x h1.
  apply functional_extensionality. intro h2.
  unfold fun_from_cons.
  rewrite h1; auto.
Qed.


Lemma fun_from_cons_compat :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U)
         (x:T) (pf:In x l),
    let f' := fun_from_cons f in
    f' x pf = f x (in_cons _ _ _ pf).
  intros; unfold f', fun_from_cons; auto; f_equal; apply proof_irrelevance.
Qed.

Definition fun_from_cons2 {T U} {l:list T} {a b:T}
           (f:fun_in_dep (a::b::l) U) :
  fun_in_dep (a::l) U :=
  fun x pfx => f x (in_cons2 a b x pfx).



Lemma inj_from_cons :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U),
    inj_dep f ->
    inj_dep (fun_from_cons f).
  intros T U l a f h1.
  red in h1; red.
  intros x y hx hy h2.   
  apply (h1 x y (in_cons _ _ _ hx) (in_cons _ _ _  hy)).
  unfold fun_from_cons in h2. 
  erewrite fun_in_dep_functional.
  symmetry.
  erewrite fun_in_dep_functional.
  symmetry.
  apply h2.
  reflexivity.
  reflexivity.
Qed.

Lemma surj_from_cons_in :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U),
    In a l ->
    surj_dep f ->
    surj_dep (fun_from_cons f).
  intros T U l a f h1 h2; red in h2; red; intro y.
  specialize (h2 y); destruct h2 as [x [h2 h3]]; subst.
  destruct h2 as [|h2]; subst.
  exists x, h1; unfold fun_from_cons; f_equal; apply proof_irrelevance.
  exists x, h2; unfold fun_from_cons; f_equal; apply proof_irrelevance.
Qed.

Lemma bij_from_cons_in :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U),
    In a l ->
    bij_dep f ->
    bij_dep (fun_from_cons f).
  intros ? ? ? ? ? ? h1; red in h1; red; destruct h1; split;
  [apply inj_from_cons |
   apply surj_from_cons_in]; auto.
Qed.


Lemma inj_from_cons2 :
  forall {T U} {l:list T} {a b:T}
         (f:fun_in_dep (a::b::l) U),
    inj_dep f ->
    inj_dep (fun_from_cons2 f).
  intros T U l a b f h1.
  red in h1; red.
  intros x y hx hy h2.   
  apply (h1 x y (in_cons2 _ _ _ hx) (in_cons2 _ _ _  hy)).
  unfold fun_from_cons in h2. 
  erewrite fun_in_dep_functional.
  symmetry.
  erewrite fun_in_dep_functional.
  symmetry.
  apply h2.
  reflexivity.
  reflexivity.
Qed.

Lemma surj_from_cons2_in :
  forall {T U} {l:list T} {a b:T}
         (f:fun_in_dep (a::b::l) U),
    In b l ->
    surj_dep f ->
    surj_dep (fun_from_cons2 f).
  unfold fun_from_cons2; intros ? ? ? ? ? ? hi h1; red in h1;
  red; intro y;
  specialize (h1 y); destruct h1 as [x [h1 h3]]; subst;
  destruct h1 as [|h1]; subst;
  [exists x, (in_eq _ _); f_equal; apply proof_irrelevance |];
  destruct h1 as [|h1]; subst;
  [exists x, (in_cons _ _ _ hi); f_equal; apply proof_irrelevance |
   exists x, (in_cons _ _ _ h1); f_equal; apply proof_irrelevance].
Qed.

Lemma bij_from_cons2_in :
  forall {T U:Type} {l:list T} {a b:T} (f:fun_in_dep (a::b::l) U),
    In b l ->
    bij_dep f ->
    bij_dep (fun_from_cons2 f).
  intros ? ? ? ? ? ? ? h1; red in h1; red; destruct h1; split;
  [apply inj_from_cons2 |
   apply surj_from_cons2_in]; auto.
Qed.

  
Definition fun_from_cons_ran {T U:Type} {l:list T} {a:T}
           (f:fun_in_dep (a::l) U)
           (pfinj:inj_dep f)
           (pfn : ~In a l) : 
  fun_in_dep l {y:U | y <> f a (in_eq _ _)} :=
  fun x (pfx:In x l) =>
    let pfx' := in_cons _ _ _ pfx in
    let pfneq := inj_dep_impl_contra f pfinj _ _ pfx' (in_eq _ _) (in_neq _ _ _ pfx pfn) in
    exist (fun c => c <> f a (in_eq _ _)) (f x pfx') pfneq.

    
Lemma fun_from_cons_ran_functional :
    forall {T U:Type} {l:list T} {a:T}
           (f g:fun_in_dep (a::l) U)
           (pfnin:~In a l)
           (pfinjf:inj_dep f)
           (pfinjg:inj_dep g)
           (x:T)
           (pfx:In x l)
           (pfx':In x (a::l)),
      f x pfx' = g x pfx' ->
      proj1_sig (fun_from_cons_ran f pfinjf pfnin x pfx) = 
      proj1_sig (fun_from_cons_ran g pfinjg pfnin x pfx).
  unfold fun_from_cons_ran; intros; simpl; erewrite f_equal_dep; auto;
  rewrite H at 1; f_equal; apply proof_irrelevance.
Qed.

Lemma fun_from_cons_ran_compat :
    forall {T U:Type} {l:list T} {a:T}
           (f:fun_in_dep (a::l) U)
           (pfnin:~In a l)
           (pfinjf:inj_dep f)
           (x:T)
           (pfx:In x l),
      let pfx' := in_cons _ _ _ pfx in
      proj1_sig (fun_from_cons_ran f pfinjf pfnin x pfx) = 
      f x pfx'.
  unfold fun_from_cons_ran; intros; simpl; auto.
Qed.

Lemma inj_from_cons_ran :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U)
         (pfinj:inj_dep f)
         (pfnin : ~In a l),
    inj_dep (fun_from_cons_ran f pfinj pfnin).
  intros T U l a f h1 h2; red in h1; red; intros x y ha ha' h3.
  unfold fun_from_cons_ran in h3; apply exist_injective in h3;
  apply h1 in h3; auto.
Qed.

Lemma inj_surj_from_cons_ran :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U)
         (pfinj:inj_dep f)
         (pfnin : ~In a l),
    surj_dep f ->
    surj_dep (fun_from_cons_ran f pfinj pfnin).
  intros T U l a f hinj hnin h1; red in h1; red; intro y;
  destruct y as [y h2].
  specialize (h1 y).
  destruct h1 as [x [h1 ?]]; subst.
  destruct h1 as [|h1]; subst.
  contradict h2; f_equal; apply proof_irrelevance.
  exists x, h1.
  unfold fun_from_cons_ran.
  apply f_equal_dep; f_equal; apply proof_irrelevance.
Qed.

Lemma inj_surj_bij_from_cons_ran :
  forall {T U:Type} {l:list T} {a:T} (f:fun_in_dep (a::l) U)
         (pfinj:inj_dep f)
         (pfnin : ~In a l),
    surj_dep f ->
    bij_dep (fun_from_cons_ran f pfinj pfnin).
  unfold fun_from_cons_ran; intros; red; split; red;
  [intros x y hx hy h1; 
   eapply exist_injective in h1; eapply pfinj in h1|
   intro y; destruct y as [y hy]; red in H; specialize (H y);
   destruct H as [c [hc ?]]; subst;
   destruct hc as [|hc];
   [subst; contradict hy; f_equal; apply proof_irrelevance|
    exists c, hc; do 2 apply f_equal_dep]
  ]; auto.
Qed.


Definition tl_in_dep {T U:Type} {l:list T}
           (f:fun_in_dep l U) :
  fun_in_dep (tl l) U :=
  fun x pf => f x (in_tl _ _ pf).

Lemma tl_in_dep_cons :
  forall {T U:Type} {l:list T} {a}
         (pfdt:type_eq_dec T)
         (f:fun_in_dep (a::l) U),
    tl_in_dep f = fun_from_cons f.
  unfold tl_in_dep, fun_from_cons; intros; simpl;
  apply functional_extensionality_dep; intro;
  apply functional_extensionality; intro; f_equal;
  apply proof_irrelevance.
Qed.

Lemma inj_tl :
  forall {T U} {l:list T}
         (f:fun_in_dep l U),
    inj_dep f ->
    inj_dep (tl_in_dep f).
  intros ? ? ? ? h1; red in h1; red; intros x y hx hy h2.
  unfold tl_in_dep in h2.
  apply h1 in h2; auto.
Qed.

Lemma surj_tl :
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pfn:l <> nil)
         (pfin:In (hd' l pfn) (tl l)),
    surj_dep f ->
    surj_dep (tl_in_dep f).
  intros ? ? ? ? h1 h2 h3; red in h3; red; intro y.
  specialize (h3 y). 
  destruct h3 as [x [h3 ?]]; subst.
  pose proof h3 as h4.
  rewrite  (hd_compat' _ h1) in h4.
  destruct h4 as [|h4]; subst.
  exists _, h2; unfold tl_in_dep; f_equal; apply proof_irrelevance.
  exists _, h4.
  unfold tl_in_dep; f_equal; apply proof_irrelevance.
Qed.


Lemma bij_tl :
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pfn:l <> nil)
         (pfin:In (hd' l pfn) (tl l)),
    bij_dep f ->
    bij_dep (tl_in_dep f).
  intros ? ? ? ? ? ? h1; red in h1. red. split;
    destruct h1 as [h1 h2]; 
  [apply inj_tl; auto |
   eapply surj_tl; auto; apply pfin].
Qed.

Definition tl_in_dep_ran {T U} {l:list T}
           (f:fun_in_dep l U)
           (pfn:l <> nil)
           (pfinj:inj_dep f)
           (pfnin : ~In (hd' l pfn) (tl l)) : 
  fun_in_dep (tl l) {y:U | y <> f (hd' l pfn) (in_hd' _ _)} :=
  fun x (pfx:In x (tl l)) =>
    let pfx' := in_tl _ _ pfx in
    let pfneq := inj_dep_impl_contra
                   f pfinj _ _ pfx'
                   (in_hd' _ _) (in_neq _ _ _ pfx pfnin) in
    exist (fun c => c <> f (hd' l pfn) (in_hd' _ _))
          (f x pfx') pfneq.

Lemma proj1_sig_tl_in_dep_ran  :
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pfn:l <> nil)
         (pfinj:inj_dep f)
         (pfnin : ~In (hd' l pfn) (tl l)),
    proj1_sig_fun_dep (tl_in_dep_ran f pfn pfinj pfnin) =
    tl_in_dep f.
  unfold proj1_sig_fun_dep, tl_in_dep_ran, tl_in_dep,
  fun_in_dep_fun_compose; intros;
  apply functional_extensionality_dep; intro;
  apply functional_extensionality; intros; simpl; auto.
Qed.

Lemma inj_tl_ran:
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pfn:l <> nil)
         (pfinj:inj_dep f)
         (pfnin : ~In (hd' l pfn) (tl l)),
    inj_dep (tl_in_dep_ran f pfn pfinj pfnin).
  unfold tl_in_dep_ran; intros ? ? ? ? ? hinj ?; red.
  intros ? ? ? ? h1.
  apply exist_injective in h1.
  apply hinj in h1; auto.
Qed.

Lemma surj_tl_ran :
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pfn:l <> nil)
         (pfinj:inj_dep f)
         (pfnin : ~In (hd' l pfn) (tl l)),
    surj_dep f ->
    surj_dep (tl_in_dep_ran f pfn pfinj pfnin).
  unfold tl_in_dep_ran; intros ? ? ? ? hn hinj ? hs; red; intro y;
  destruct y as [y hy].
  specialize (hs y).
  destruct hs as [x [hx ?]]; subst.
  exists x.
  ex_goal.
  pose proof hx as hx'.
  rewrite (hd_compat' _ hn) in hx'.
  destruct hx' as [|h1]; subst; auto.
  contradict hy.
  f_equal; apply proof_irrelevance.
  exists hex.
  apply proj1_sig_injective; simpl; f_equal; apply proof_irrelevance.
Qed.

Lemma bij_tl_ran :
  forall {T U} {l:list T}
         (f:fun_in_dep l U)
         (pfn:l <> nil)
         (pfinj:inj_dep f)
         (pfnin : ~In (hd' l pfn) (tl l)),
    bij_dep f ->
    bij_dep (tl_in_dep_ran f pfn pfinj pfnin).
  intros ? ? ? ? ? ? ? h1; red in h1; red.
  destruct h1 as [h1 h2]; split.
  apply inj_tl_ran; auto.
  apply surj_tl_ran; auto.
Qed.
  


 
(*This function removes a point in the domain list of f, yet 
DOESN'T remove other equal values at different points.*)
(*Contrast with [fun_in_dep_remove_in], which does remove 
points at equal values.*)

Definition fun_in_dep_remove {T U:Type} {l:list T} (pfdt:type_eq_dec T)
           (f:fun_in_dep l U) (x:T) :
  fun_in_dep (remove pfdt x l) U.
  do 2 red; intros ? h1; eapply f.
  apply in_remove_impl in h1.
  apply h1.
Defined.


Lemma fun_in_dep_remove_functional :
  forall {T U:Type}{l:list T}
         (pfdt:type_eq_dec T)
         (f g:fun_in_dep l U) (a x:T),
    f x = g x->
    fun_in_dep_remove pfdt f a x = fun_in_dep_remove pfdt g a x.
  intros T U l h1 f g a x h2.
  apply functional_extensionality.
  intro h3.
  unfold fun_in_dep_remove.
  rewrite h2; auto.
Qed.


Lemma fun_in_dep_remove_compat :
  forall {T U:Type}{l:list T}
         (pfdt:type_eq_dec T)
         (f:fun_in_dep l U) (a x:T)
         (pf:In x (remove pfdt a l)),
    let pf' := in_remove_impl pfdt _ _ _ pf in
    fun_in_dep_remove pfdt f a x pf = f x pf'.
  simpl; intros.
  unfold fun_in_dep_remove; auto.
Qed.

Lemma inj_dep_remove :
  forall {T U:Type} {l:list T} (pfdt:type_eq_dec T)
         (f:fun_in_dep l U) (x:T),
    inj_dep f ->
    inj_dep (fun_in_dep_remove pfdt f x).
  intros T U l h1 f x h2.
  red in h2; red.
  intros a b h3 h4.
  unfold fun_in_dep_remove.
  intro h5.
  apply h2 in h5; auto.
Qed.



Definition fun_in_dep_comm_cons {T U:Type}
           {l:list T} {a b:T}
           (f:fun_in_dep (a::b::l) U) : fun_in_dep (b::a::l) U.
  do 2 red in f; do 2 red.
  intros x h1.
  apply in_list2_comm in h1.
  refine (f _ h1).
Defined.

Lemma fun_in_dep_comm_cons_compat :
  forall {T U:Type} {l:list T} {a b:T}
         (f:fun_in_dep (a::b::l) U)
         (x:T) (pf:In x (a::b::l)),
    let f' := fun_in_dep_comm_cons f in
    let pf' := in_list2_comm _ _ _ _ pf in
    f x pf = f' x pf'.
simpl; unfold fun_in_dep_comm_cons; intros; f_equal; apply proof_irrelevance.
Qed.

Lemma fun_from_cons2_fun_in_dep_comm_cons :
  forall {T U:Type} {l:list T} {a b:T} (f:fun_in_dep (a::b::l) U),
    fun_from_cons (fun_from_cons f) =
    fun_from_cons (fun_from_cons (fun_in_dep_comm_cons f)).
  intros; apply functional_extensionality_dep; intro; apply functional_extensionality; intro.
  do 4 rewrite fun_from_cons_compat.
  rewrite fun_in_dep_comm_cons_compat at 1.
  f_equal.
  apply proof_irrelevance.
Qed.


Fixpoint inv_im_in_dep {T U:Type} (pfdu:type_eq_dec U)
         (l:list T)  (u:U) : fun_in_dep l U -> list T :=
  match l with
    | nil => fun _ => nil
    | a::l' => fun f =>
                 let il := inv_im_in_dep pfdu l' u (fun_from_cons f) in
                 if (pfdu u (f a (in_eq _ _))) then
                 a :: il
                 else
                   il
  end.

Lemma inv_im_in_dep_functional :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l:list T) (u u':U) (f f':fun_in_dep l U),
    u = u' ->
    f = f' ->
    inv_im_in_dep pfdu l u f =
    inv_im_in_dep pfdu l u' f'.
  intros; subst; auto.
Qed.

Lemma inv_im_in_dep_transfer_functional :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l l':list T) (u u':U) (f:fun_in_dep l U)
         (pf:l = l'),
    u = u' ->
    let f' := transfer_fun_in_dep f pf in
    inv_im_in_dep pfdu l u f =
    inv_im_in_dep pfdu l' u' f'.
  simpl; intros; subst; rewrite transfer_fun_in_dep_eq_refl; auto.
Qed.



Lemma in_inv_im_in_dep_iff :
  forall {T U:Type} (pfdu:type_eq_dec U)
           (l:list T) (u:U) (f:fun_in_dep l U) (x:T),
    In x (inv_im_in_dep pfdu l u f) <->
    exists pf:In x l,
      f x pf = u.
  intros T U h1 l.
  induction l as [|a l ih]; simpl.
  intros; split.
  intro; contradiction.
  intro h2. destruct h2; auto.
  intros u f x.
  lr_if_goal.
  fold hlr.
  destruct hlr as [h2 | h2]; subst.
  split.
  intro h2.
  destruct h2 as [|h2]; subst.
  exists (or_introl eq_refl).
  f_equal. apply proof_irrelevance.
  specialize (ih (f a (in_eq a l)) (fun_from_cons f) x).
  rewrite ih in h2.
  destruct h2 as [h2 h3].
  exists (or_intror h2).
  rewrite fun_from_cons_compat in h3.
  rewrite <- h3.
  f_equal. apply proof_irrelevance.
  intro h2.
  destruct h2 as [h2 h3].
  destruct h2 as [|h2]; subst.
  left; auto.
  right.
  rewrite ih.
  exists h2.
  rewrite <- h3.
  rewrite fun_from_cons_compat.
  f_equal. apply proof_irrelevance.
  split.
  intros h3.
  rewrite ih in h3.
  destruct h3 as [h3 h4]; subst.
  exists (or_intror h3).
  rewrite fun_from_cons_compat.
  f_equal. apply proof_irrelevance.
  intro h3.
  destruct h3 as [h3]; subst.
  destruct h3 as [|h3]; subst.
  contradict h2.
  f_equal. apply proof_irrelevance.
  rewrite ih.
  exists h3.
  rewrite fun_from_cons_compat.
  f_equal. apply proof_irrelevance.
Qed.


  
Lemma incl_inv_im_in_dep :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l:list T)  (u:U) (f:fun_in_dep l U),
    incl (inv_im_in_dep pfdu l u f) l.
  intros T U h1 l u f.
  red.
  intros x h2.
  rewrite in_inv_im_in_dep_iff in h2.
  destruct h2; auto.
Qed.

Lemma inv_im_in_dep_eq_nil_iff :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (b:U) (f:fun_in_dep l U),
    inv_im_in_dep pfdu l b f = nil <->
    (l = nil \/ (forall (a:T) (pf:In a l), f a pf <> b)).
  intros T U h1 h1' l.
  induction l as [|a l ih]; simpl.
  intuition.
  intros b f.
  destruct (h1' b (f a (in_eq a l))) as [|h2]; subst.
  split; intro h2.
  discriminate.
  destruct h2 as [|h2].
  discriminate.
  specialize (h2 a (or_introl eq_refl)).
  contradict h2; f_equal; apply proof_irrelevance.
  specialize (ih b (fun_from_cons f)).
  rewrite ih.
  split; intro h3.
  destruct h3 as [|h3]; subst.
  right.
  intros a' h3.
  destruct h3; subst; try contradiction.
  contradict h2.
  rewrite <- h2.
  f_equal; apply proof_irrelevance.
  right.
  intros a' h4.
  destruct (h1 a a') as [|h5]; subst.
  contradict h2; rewrite <- h2; f_equal; apply proof_irrelevance.
  destruct h4 as [|h4]; subst.
  contradict h5; auto.
  specialize (h3 _ h4).
  unfold fun_from_cons in h3.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
  destruct h3 as [|h3].
  discriminate.
  destruct (nil_dec' l) as [h4 | h4].
  left; auto.
  right.
  intros a' h5.
  specialize (h3 _ (in_cons _ _ _ h5)).
  contradict h3.
  rewrite <- h3.
  unfold fun_from_cons.
  f_equal.
Qed.



Lemma inv_im_in_dep_eq_nil :
  forall {T U:Type} (pfdu:type_eq_dec U) (l:list T) (b:U) (f:fun_in_dep l U),
    inv_im_in_dep pfdu l b f = nil ->
    l = nil \/ (exists (a:T) (pf:In a l), f a pf <> b).
  intros T U h1 l.
  induction l as [|a l ih]; simpl.
  intros; left; auto.
  intros b f h2.
  right.
  destruct (h1 b (f a (in_eq a l))) as [h3 | h3].
  discriminate.
  specialize (ih _ _ h2).
  destruct ih as [ih | ih]; subst.
  exists a, (or_introl eq_refl).
  contradict h3; subst.
  f_equal. apply proof_irrelevance.
  destruct ih as [a' [h4 h5]].
  exists a', (in_cons _ _ _ h4).
  contradict h5; subst.
  unfold fun_from_cons.
  f_equal.
Qed.


    
Lemma inv_im_in_dep_sing_eq :
  forall {T U:Type} {l:list T} {a:T} (pfdu:type_eq_dec U)
         (f:fun_in_dep (a::nil) U)
         (u:U),
    u = f a (in_eq _ _) ->
    inv_im_in_dep pfdu _ u f = a::nil.
  intros T U l a h1 f u h2.
  subst.
  simpl.
  lr_if_goal.
  fold hlr.
  destruct hlr as [h2 | h2]; auto.
  contradict h2; auto.
Qed.

Definition list_pred {T} (P:T->Prop) (l:list T) : Type :=
  forall x,
    In x l ->
    P x.

Lemma list_pred_irrelevance :
  forall {T} {P:T->Prop}
         (l:list T)
         (pfl pfl':list_pred P l),
    pfl = pfl'.
  intros.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality_dep.
  intro h1.
  apply proof_irrelevance.
Qed.



Definition list_pred_type {T} (l:list T) : Type :=
  forall x,
    In x l ->
    Prop.

Definition list_pred_type_transfer {T} {l l':list T}
           (pfe:l = l') (P:list_pred_type l) : list_pred_type l' :=
  transfer_dep pfe P.

Lemma list_pred_type_transfer_eq_refl :
  forall {T} {l:list T}
         (P:list_pred_type l),
    list_pred_type_transfer eq_refl P = P.
  unfold list_pred_type_transfer; intros; rewrite transfer_dep_eq_refl; auto.
Qed.


Definition list_pred_dec
           {T} {l:list T}
           (P:list_pred_type l) : Type :=
  forall x (pfin:In x l),
    { P x pfin} +
    {~P x pfin}.

Lemma list_pred_dec_irrelevance : 
  forall {T} {l:list T}
         {P:list_pred_type l}
         (pfl pfl':list_pred_dec P),
    pfl = pfl'.
  intros T l P hl hl'.
  apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality_dep.
  intro j.
  destruct (hl i j), (hl' i j); f_equal; try apply proof_irrelevance; try contradiction.
Qed.

 
Lemma list_pred_dec_type_transfer :
  forall {T} {l l':list T}
         (pfe:l = l')
         (P:list_pred_type l),
    list_pred_dec P ->
    list_pred_dec (list_pred_type_transfer pfe P).
  intros; subst; rewrite list_pred_type_transfer_eq_refl; auto.
Qed.



Definition list_pred_incl 
           {T} {P:T->Prop} {l l':list T} {a:T} 
           (P':list_pred P l')
           (pf:incl l l') : list_pred P l :=
  fun x (pfin:In x l) =>
    P' x (pf _ pfin).


Definition list_pred_cons
           {T} {P:T->Prop} {l:list T} {a:T}
           (P':list_pred P (a::l)) :
  list_pred P l :=
  fun x (pfin:In x l) =>
    P' x (in_cons _ _ _ pfin).

Definition list_pred_app_l
           {T} {P:T->Prop} {l l':list T} 
           (P':list_pred P (l++l')) : 
  list_pred P l :=
  fun x (pfin:In x l) =>
    P' x (in_app_l pfin l').

Definition list_pred_app_r
           {T} {P:T->Prop} {l l':list T} 
           (P':list_pred P (l++l')) : 
  list_pred P l' :=
  fun x (pfin:In x l') =>
    P' x (in_app_r pfin l).


Definition list_pred_type_incl 
           {T} {l l':list T} {a:T} 
           (P:list_pred_type l')
           (pf:incl l l') : list_pred_type l :=
  fun x (pfin:In x l) =>
    P x (pf _ pfin).
                                           
  
Definition list_pred_type_cons
           {T} {l:list T} {a:T} 
           (P:list_pred_type (a::l)) :
  list_pred_type l :=
  fun x (pfin:In x l) =>
    P x (in_cons _ _ _ pfin).

  
Definition list_pred_type_app_sing
           {T} {l:list T} {a:T} 
           (P:list_pred_type (l++a::nil)) :
  list_pred_type l :=
  fun x pf => P x (in_app_l pf _).

Definition list_pred_type_app_l
           {T} {l l':list T}
           (P:list_pred_type (l++l')) :
  list_pred_type l :=
  fun x pf => P x (in_app_l pf _).

Definition list_pred_type_app_r
           {T} {l l':list T}
           (P:list_pred_type (l++l')) :
  list_pred_type l' :=
  fun x pf => P x (in_app_r pf _).

Lemma list_pred_dec_cons :
  forall {T} {l:list T} {a:T}
         {P:list_pred_type (a::l)},
    list_pred_dec P ->
    list_pred_dec (list_pred_type_cons P).
  intros T l a h1 h2.
  red in h2; red.
  intros x h3.
  specialize (h2 x (in_cons _ _ _ h3)).
  unfold list_pred_type_cons; auto.
Qed.

Lemma list_pred_dec_app_sing :
  forall {T} {l:list T} {a:T}
         {P:list_pred_type (l++a::nil)},
    list_pred_dec P ->
    list_pred_dec (list_pred_type_app_sing P).
  intros T l a h1 h2.
  red in h2; red.
  intros x h3.
  specialize (h2 _ (in_app_l h3 _)).
  unfold list_pred_type_app_sing; auto.
Qed.

Lemma list_pred_dec_app_l :
  forall {T} {l l':list T}
         {P:list_pred_type (l++l')},
    list_pred_dec P ->
    list_pred_dec (list_pred_type_app_l P).
  intros T l a h1 h2.
  red in h2; red.
  intros x h3.
  specialize (h2 _ (in_app_l h3 _)).
  unfold list_pred_type_app_l; auto.
Qed.

Lemma list_pred_dec_app_r :
  forall {T} {l l':list T}
         {P:list_pred_type (l++l')},
    list_pred_dec P ->
    list_pred_dec (list_pred_type_app_r P).
  intros T l a h1 h2.
  red in h2; red.
  intros x h3.
  specialize (h2 _ (in_app_r h3 _)).
  unfold list_pred_type_app_r; auto.
Qed.



Definition seg_pred_list_pred
           {n}
           (P:nat->Prop) 
           (pfsp:seg_pred n P) : 
  list_pred P (seg_list n) :=
  fun x (pf:In x (seg_list n)) =>
    let pflt := iff1 (in_seg_list_iff _ _) pf in    
    pfsp x pflt.


Definition seg_pred_type_list_pred 
           {n}
           (pfsp:seg_pred_type n) : 
  list_pred_type (seg_list n) :=
  fun x (pf:In x (seg_list n)) =>
    let pflt := iff1 (in_seg_list_iff _ _) pf in    
    pfsp x pflt.

Definition seg_pred_dec_list_pred
           {n}
           (pfsp:seg_pred_type n)
           (pfd:seg_pred_dec pfsp) :
  list_pred_dec (seg_pred_type_list_pred pfsp) :=
  fun x (pfi:In x (seg_list n)) =>
    let pflt := iff1 (in_seg_list_iff _ _) pfi in    
    pfd x pflt.
 

Definition list_pred_seg_pred
           {n}
           {P:nat->Prop}
           (pflp:list_pred P (seg_list n)) : 
  seg_pred n P :=
  fun i (pf:i < n) =>
    let pflt := iff2 (in_seg_list_iff _ _) pf in    
    pflp i pflt.


Definition list_pred_type_seg_pred 
           {n}
           (pflp:list_pred_type (seg_list n)) : 
  seg_pred_type n :=
  fun i (pf:i < n) =>
    let pflt := iff2 (in_seg_list_iff _ _) pf in    
    pflp i pflt.


Definition list_pred_dec_seg_pred
           {n}
           (pflp:list_pred_type (seg_list n))
           (pfd:list_pred_dec pflp) :
  seg_pred_dec (list_pred_type_seg_pred pflp) :=
  fun i (pfi:i < n) =>
    let pflt := iff2 (in_seg_list_iff _ _) pfi in    
    pfd i pflt.



(*Returns first index for which [pflpt] is true, and [length l] 
if it's never true.*)
Fixpoint first_l_pred 
         {T} {l:list T} :
  forall {P:list_pred_type l}
         (pfd:list_pred_dec P),
    nat :=
  match l with
    | nil => fun _ _ => 0
    | a::l' => fun P pfd =>
                 match (pfd a (in_eq _ _)) with
                   | left pfe => 0
                   | right pfne =>
                     let P' := list_pred_type_cons P in
                     let pfld' := list_pred_dec_cons pfd in
                     S (first_l_pred pfld')
                 end
  end.

Lemma first_l_pred_functional :
  forall {T} {l l':list T}
         {P:list_pred_type l}
         {P':list_pred_type l'}
         (pfd:list_pred_dec P)
         (pfd':list_pred_dec P')
         (pfe:l = l'),
    P' = transfer_dep pfe P ->
    first_l_pred pfd = first_l_pred pfd'.
  intros; subst.
  revert pfd pfd'.
  rewrite transfer_dep_eq_refl.
  intros; f_equal.
  apply list_pred_dec_irrelevance.
Qed.


Lemma first_l_pred_dec_type_transfer :
  forall {T} {l l':list T} {P:list_pred_type l}
         (pfe:l = l')
         (pfd:list_pred_dec P),
    first_l_pred (list_pred_dec_type_transfer pfe P pfd) =
    first_l_pred pfd.
  intros; subst;
  gterml.
  redterm0 x.
  fold c.
  apply (first_l_pred_functional
           c pfd (eq_refl _)
           (transfer_dep_eq_refl _ _ (eq_refl _) P)).
Qed.

         
Lemma le_first_l_pred :
  forall {T} {l:list T}
         {P:list_pred_type l}
         (pfd:list_pred_dec P),
    first_l_pred pfd <= length l.
  intros T l.
  induction l as [|a l ih]; simpl; auto with arith.
  intros h1 h2.
  lr_if_goal; fold hlr; destruct hlr; auto with arith.
Qed.


Lemma first_l_pred_eq_length_iff :
  forall {T} {l:list T}
         {P:list_pred_type l}
         (pfd:list_pred_dec P),
    first_l_pred pfd = length l <->
    forall (x:T) (pf:In x l),
      ~P x pf.
  intros T l P hd; split.
  revert hd. revert  P.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros P hd h1 x h2.
  lr_if h1; fold hlr in h1; destruct hlr as [hl | hr].
  discriminate.
  destruct h2 as [|h2]; subst.
  contradict hr.
  erewrite f_equal_dep; auto.
  apply hr.
  apply S_inj in h1.

  specialize (ih (list_pred_type_cons P)
                 (list_pred_dec_cons hd) h1 x h2).
  unfold list_pred_type_cons in ih.
  contradict ih.
  erewrite f_equal_dep; auto.
  apply ih.
  revert hd. revert P. 
  induction l as [|a l ih].
  intros; simpl; auto.
  intros P hd h1.
  simpl. 
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  specialize (h1 a (in_eq _ _)).
  contradiction.
  f_equal.
  apply ih.
  unfold list_pred_type_cons.
  intros; apply h1.
Qed.



Lemma lt_first_l_pred :
  forall {T} {l:list T}
         {P:list_pred_type l}
         (pfd:list_pred_dec P),
    first_l_pred pfd < length l ->
    exists y (pf:In y l),
      P y pf.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros ? ? h1.
  apply lt_irrefl in h1.
  contradiction.
  intros P hd h1. 
  pose proof (le_first_l_pred (list_pred_dec_cons hd)) as h2.
  apply le_lt_eq_dec in h2. 
  lr_if h1; fold hlr in h1; destruct hlr as [hl | hr].
  specialize (ih (list_pred_type_cons P)
                 (list_pred_dec_cons hd)).
  destruct h2 as [h2 | h2].
  specialize (ih h2).
  destruct ih as [y [h3 h4]].
  exists y.
  ex_goal.
  right; auto.
  exists hex.
  unfold list_pred_type_cons in h4.
  erewrite f_equal_dep; auto.
  apply h4.
  pose proof (first_l_pred_eq_length_iff (list_pred_dec_cons hd)) as h3.
  pose proof h2 as h4.
  rewrite h3 in h4.
  exists a.
  ex_goal.
  left; auto.
  exists hex.
  erewrite f_equal_dep; auto.
  apply hl.
  apply lt_S_n in h1.
  specialize (ih (list_pred_type_cons P) (list_pred_dec_cons hd) h1).
  destruct ih as [x [hx h3]].
  exists x.
  ex_goal.
  right; auto.
  exists hex.
  unfold list_pred_type_cons in h3.
  erewrite f_equal_dep; auto.
  apply h3.
Qed.


Lemma first_l_pred_cons_P : 
  forall {T} {l:list T} {a:T}
         {P:list_pred_type (a::l)}
         (pfd:list_pred_dec P)
         (pfd':list_pred_dec (list_pred_type_cons P)),
    P a (in_eq _ _) ->
    first_l_pred pfd = 0.
  intros T l a P hd hd' h1.
  unfold first_l_pred.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; auto.
  contradiction.
Qed.


Lemma first_l_pred_cons_nP : 
  forall {T} {l:list T} {a:T}
         {P:list_pred_type (a::l)}
         (pfd:list_pred_dec P)
         (pfd':list_pred_dec (list_pred_type_cons P)),
    ~ P a (in_eq _ _) ->
    first_l_pred pfd = S (first_l_pred pfd').
  intros T l a P hd hd' h1.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  contradiction.
  f_equal.
  f_equal.
  apply list_pred_dec_irrelevance.
Qed.


(*Returns the first number < n such that [pf] is true, and n 
if it's not true*)
Definition first_seg_pred
           {n}
           {pf:seg_pred_type n}
           (pfs:seg_pred_dec pf) :=
  let pf'  := seg_pred_type_list_pred pf in
  let pfs' := seg_pred_dec_list_pred pf pfs in
  first_l_pred pfs'.


Lemma le_first_seg_pred : 
  forall {n}
         {P:seg_pred_type n}
         (pfd:seg_pred_dec P),
    first_seg_pred pfd <= n.
  unfold first_seg_pred.
  intros.
  eapply le_trans.
  apply le_first_l_pred.
  rewrite length_seg_list; auto with arith.
Qed.


Lemma eq_first_seg_pred :
  forall {n}
         {P:seg_pred_type n}
         (pfd:seg_pred_dec P),
    first_seg_pred pfd = n ->
    forall i (pf:i < n),
      ~P i pf.
  intros n P hd h1.
  pose proof @first_l_pred_eq_length_iff.
  unfold first_seg_pred in h1.
  rewrite <- (length_seg_list n) in h1.
  rewrite first_l_pred_eq_length_iff in h1.
  intros i hi.
  specialize (h1 i).
  hfirst h1.
  rewrite in_seg_list_iff; auto.
  specialize (h1 hf).
  contradict h1.
  unfold seg_pred_type_list_pred.
  erewrite f_equal_dep.
  apply h1.
  reflexivity.
Qed.


Lemma lt_first_seg_pred :
  forall {n}
         {P:seg_pred_type n}
         (pfd:seg_pred_dec P),
    first_seg_pred pfd < n ->
    exists i (pf:i < n),
      P i pf.
  intros n P hd h1. 
  pose proof (lt_first_l_pred (seg_pred_dec_list_pred _ hd)) as h2.
  hfirst h2.
  rewrite length_seg_list; auto.
  specialize (h2 hf).
  destruct h2 as [i h2].
  destruct h2 as [h2 h3].
  pose proof h2 as h4.
  rewrite in_seg_list_iff in h4.
  exists _, h4.
  unfold seg_pred_type_list_pred in h3.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
Qed.


(*This is the analogue of [count_occ_prop] for the [seg] structure, 
whereby this function counts all elements less than [n] for which
P is true.*)
Fixpoint count_occ_prop_seg
         {n:nat} :
  forall {P:seg_pred_type n}
         (pfd:seg_pred_dec P),
    nat :=
  match n
  with 
    | 0 => fun _ _  => 0 
    | S n' => fun (P:seg_pred_type (S n'))
                  (pfd:seg_pred_dec P)
              =>
                let P' := seg_pred_type_from_S P in
                let pfd' := seg_pred_dec_from_S pfd in
                let k := count_occ_prop_seg pfd' in
                if (pfd n' (lt_n_Sn _)) then
                  S k
                else
                  k
  end.


(*This function removes a point in the domain list of f, and also 
removes other points of the same value.*)
Definition fun_in_dep_remove_in
           {T U:Type} {l:list T}
           (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
           (f:fun_in_dep l U) (x:T) (pf:In x l) :
  fun_in_dep (list_minus pfdt l (inv_im_in_dep pfdu l (f x pf) f)) U.
  do 2 red in f; do 2 red.
  intros x' h2.
  pose proof (in_list_minus_impl' pfdt  _ _ _ h2) as h3.
  refine (f _ h3).
Defined.

Lemma fun_in_dep_remove_in_compat :
  forall {T U:Type} {l:list T}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U) (a x:T) (pfa:In a l)
         (pfx':In x (list_minus pfdt l (inv_im_in_dep pfdu l (f a pfa) f))),
    let pfx := in_list_minus_impl' pfdt _ _ _ pfx' in
    let f' := fun_in_dep_remove_in pfdt pfdu f a pfa in
    f x pfx = f' x pfx'.
  unfold fun_in_dep_remove_in; intros; auto.
Qed.

Lemma fun_in_dep_remove_in_compat' :
  forall {T U:Type} {l:list T}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U) (a x:T) (pfa:In a l)
         (pfx':In x (list_minus pfdt l (inv_im_in_dep pfdu l (f a pfa) f)))
         (pfx:In x l),
         let f' := fun_in_dep_remove_in pfdt pfdu f a pfa in
    f x pfx = f' x pfx'.
  unfold fun_in_dep_remove_in; intros; auto.
  f_equal. apply proof_irrelevance.
Qed.


(*Assumes constant l to preserve functional equality.*)
Lemma fun_in_dep_remove_in_functional :
  forall {T U:Type} {l:list T}
         (pfdt pfdt':type_eq_dec T)
         (pfdu pfdu':type_eq_dec U)
         (f f':fun_in_dep l U)
         (a x a' x':T)
         (pfa:In a l) (pfa':In a' l)
         (pfx: In x  (list_minus pfdt l (inv_im_in_dep pfdu l (f a pfa) f)))
         (pfx':In x' (list_minus pfdt' l (inv_im_in_dep pfdu' l (f' a' pfa') f'))),
    f = f' ->
    a = a' ->
    x = x' ->
    fun_in_dep_remove_in pfdt pfdu   f  a  pfa  x  pfx =
    fun_in_dep_remove_in pfdt' pfdu' f' a' pfa' x' pfx'.
  intros T U l h1 h1' h2 h2' f f' a x a' x' h3 h3' h4 h4' h5 h6 h7.
  subst.
  pose proof (type_eq_dec_eq h1 h1').
  pose proof (type_eq_dec_eq h2 h2').
  pose proof (proof_irrelevance _ h3 h3'); subst.
  pose proof (proof_irrelevance _ h4 h4'); subst.
  reflexivity.
Qed.

(*Doesn't assume [NoDup lu], so there will be duplicates in the 
function value, if there are duplicates in [lu].*)
Fixpoint inv_im_in_dep_l {T U:Type} (pfdu:type_eq_dec U)
         (l:list T) (f:fun_in_dep l U) (lu:list U) : list T :=
  match lu with
    | nil => nil
    | b::l' =>
      (inv_im_in_dep pfdu l b f) ++ (inv_im_in_dep_l pfdu l f l')
  end.

Lemma inv_im_in_dep_l_functional :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l:list T) (f f':fun_in_dep l U) (lu lu':list U),
    f = f' ->
    lu = lu' ->
    inv_im_in_dep_l pfdu l f lu =
    inv_im_in_dep_l pfdu l f' lu'.
  intros; subst; auto.
Qed.


Lemma inv_im_in_dep_l_transfer_functional :
  forall {T U:Type} (pfdu:type_eq_dec U) (l l':list T)
         (f:fun_in_dep l U) (lu lu':list U)
         (pf:l = l'),
    lu = lu' ->
    let f' := transfer_fun_in_dep f pf in
    inv_im_in_dep_l pfdu l f lu =
    inv_im_in_dep_l pfdu l' f' lu'.
  simpl; intros; subst; rewrite transfer_fun_in_dep_eq_refl; auto.
Qed.
 

Lemma in_inv_im_in_dep_l_iff :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l:list T) (f:fun_in_dep l U) (lu: list U) (x:T),
    In x (inv_im_in_dep_l pfdu l f lu) <->
    exists pf:In x l,
      In (f x pf) lu.
  intros T U h1 l f lu x; split; intro h2.
  revert h2. revert x f. revert l.
  induction lu as [|u lu ih]; simpl.
  intros; contradiction.
  intros l x f h2.
  apply in_app_or in h2.
  destruct h2 as [h2 | h2].
  rewrite in_inv_im_in_dep_iff in h2.
  destruct h2 as [h2]; subst.
  exists h2. left; auto.
  specialize (ih _ _ _ h2).
  destruct ih as [h3 h4].
  exists h3. right; auto.
  revert h2. revert x. revert  f. revert l.
  induction lu as [|u lu ih]; simpl.
  intros l h2 h3 h4.
  destruct h4; auto.
  intros l f x h2.
  destruct h2 as [h2 h3].  
  destruct h3 as [h3 | h3].
  symmetry in h3.
  pose proof (ex_intro (fun c => f x c = u) h2 h3) as h4.
  rewrite <- (in_inv_im_in_dep_iff h1) in h4.
  apply in_app_l; auto.
  specialize (ih l f x).
  hfirst ih.
  exists h2; auto.
  specialize (ih hf).
  apply in_app_r; auto.
Qed.

 
Lemma inv_im_in_dep_l_incl :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l:list T) (lu:list U) (f:fun_in_dep l U),
    incl (inv_im_in_dep_l pfdu l f lu) l.
  intros T U h1 l lu f.
  red; intros x h2.
  rewrite in_inv_im_in_dep_l_iff in h2.
  destruct h2; auto.
Qed.
  
(*See past [map_in_dep] below, for a function 
[fun_in_dep_list_minus_incl], analogous to [fun_in_dep_remove_in]*)


Definition fun_prop_from_cons
           {T U:Type} {l:list T} {a:T} {P:T->Prop}
           (f:fun_in_dep_prop (a::l) P U) : fun_in_dep_prop l P U.
  do 2 red; intros ? h1; eapply f; auto; right; apply h1.
Defined.

Lemma fun_prop_from_cons_compat :
  forall {T U:Type} {l:list T} {a:T} {P:T->Prop}
         (f:fun_in_dep_prop (a::l) P U) (x:T) (pf:In x l) (pfp:P x),
    let f' := fun_prop_from_cons f in
    f' x pf pfp = f x (in_cons _ _ _ pf) pfp.
  intros; unfold f', fun_prop_from_cons; auto; f_equal; apply proof_irrelevance.
Qed.

Definition set_from_cons
           {T:Type} {U:T->Set} {l:list T} {a:T}
           (f:set_in_dep U (a::l)) : set_in_dep U l.
  do 2 red; intros ? h1; eapply f; auto; right; apply h1.
Defined.


Lemma set_from_cons_compat : 
  forall {T:Type} {U:T->Set} {l:list T} {a:T}
         (f:set_in_dep U (a::l)) (x:T) (pf:In x l),
    let f' := set_from_cons f in
    f' x pf = f x (in_cons _ _ _ pf).
  intros; unfold f', set_from_cons; auto; f_equal; apply proof_irrelevance.
Defined.


Definition fun_from_app1 {T U:Type} {l l':list T}
           (f:fun_in_dep (l++l') U) : fun_in_dep l U.
  do 2 red; intros ? h1. eapply f. rewrite in_app_iff.
  left. apply h1.
Defined.

Definition fun_from_app2 {T U:Type} {l l':list T}
           (f:fun_in_dep (l++l') U) : fun_in_dep l' U.
  do 2 red; intros ? h1. eapply f. rewrite in_app_iff.
  right. apply h1.
Defined.

Lemma fun_from_app1_compat :
  forall {T U:Type} {l l':list T}
         (f:fun_in_dep (l++l') U)
         (x:T) (pf:In x l),
    let f' := fun_from_app1 f in
    f' x pf = f x (in_app_l pf l').
  intros; unfold f', fun_from_app1; f_equal; apply proof_irrelevance.
Qed.

Lemma fun_from_app2_compat :
  forall {T U:Type} {l l':list T}
         (f:fun_in_dep (l++l') U)
         (x:T) (pf:In x l'),
    let f' := fun_from_app2 f in
    f' x pf = f x (in_app_r pf l).
  intros; unfold f', fun_from_app2; f_equal; apply proof_irrelevance.
Qed.

Definition fun_in_dep_incl
           {T U:Type} (l l':list T) (f:fun_in_dep l U)
           (pfi:incl l' l) :
  fun_in_dep l' U. 
  do 2 red. intros x h1. do 2 red in f.
  apply pfi in h1. specialize (f _ h1).
  refine f.
Defined.


Lemma fun_in_dep_incl_compat :
  forall {T U:Type} (l l':list T) (f:fun_in_dep l U)
         (pfi:incl l' l),
    let f' := fun_in_dep_incl l l' f pfi in
    forall (x:T) (pf:In x l'),
      f' x pf = f x (pfi x pf).
  intros; unfold fun_in_dep_incl; auto.
Qed.


Lemma inj_dep_incl :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U)
         (l':list T) (pf:incl l' l),
    let f' := fun_in_dep_incl _ _ f pf in
    inj_dep f ->
    inj_dep f'.
  intros T U l f l' h1 f' h2.
  red in h2; red.
  intros ? ? h3 h4.
  apply (h2 _ _ (h1 _ h3) (h1 _ h4)).
Qed.


Definition fun_in_dep_map_to_sig_set_im 
           {T U V:Type} (g:T->U) (l:list T)
           (f:fun_in_dep (map g l) V) :
  sig_set (Im (list_to_set l) g)->V :=
  fun x:sig_set (Im (list_to_set l) g) =>
        let pf := proj2_sig x in
        let pf' := in_eq_set _ _ (eq_sym (map_im_compat g l)) (proj1_sig x) pf in
        let pf0 := iff2 (list_to_set_in_iff _ _) pf' in
        f _ pf0.


Definition sig_set_im_to_fun_in_dep
           {T U V:Type} (g:T->U) (l:list T)
           (f:sig_set (Im (list_to_set l) g)->V) : 
  fun_in_dep (map g l) V :=
  fun (x:U) (pf:In x (map g l)) =>
    let pf' := iff1 (list_to_set_in_iff _ _) pf in
    let pf0 := in_eq_set _ _ (map_im_compat g l) x  pf' in
    f (exist _ _ pf0).

Lemma fun_in_dep_map_to_seg_set_im_undoes :
  forall {T U V:Type} (g:T->U) (l:list T)
         (f:sig_set (Im (list_to_set l) g)->V),
    fun_in_dep_map_to_sig_set_im
      g l
      (sig_set_im_to_fun_in_dep g l f) =
    f.
  intros T U V g l f.
  unfold fun_in_dep_map_to_sig_set_im,
  sig_set_im_to_fun_in_dep.
  apply functional_extensionality.
  intro x.
  f_equal.
  apply proj1_sig_injective; auto.
Qed.

Lemma sig_set_im_to_fun_in_dep_undoes :
  forall {T U V:Type} (g:T->U) (l:list T)
         (f:fun_in_dep (map g l) V),
    (sig_set_im_to_fun_in_dep
       g l 
       (fun_in_dep_map_to_sig_set_im
          g l f)) = 
    f.
  intros T U V g l f.
  unfold fun_in_dep_map_to_sig_set_im,
  sig_set_im_to_fun_in_dep.
  simpl.
  apply functional_extensionality_dep.
  intro y.
  apply functional_extensionality.
  intro h1.
  f_equal.
  apply proof_irrelevance.
Qed.


  
Fixpoint map_seg {U:Type} {n:nat} : seg_fun n U -> list U :=
  match n with
    | 0 => fun _ => nil
    | S n => fun f => (map_seg (seg_fun_from_S f)) ++ (f n (lt_n_Sn _)) ::nil
  end.


Lemma map_seg_S :
  forall {U:Type} (n:nat) (pf:n < S n) (f:seg_fun (S n) U),
    map_seg (seg_fun_from_S f) ++ (f n pf) :: nil =
    map_seg f.
  intros; simpl; repeat f_equal; apply proof_irrelevance.
Qed.



Lemma map_seg_functional :
  forall {U:Type} {n n':nat} (pf:n = n') (f:seg_fun n U),
    map_seg f = map_seg (transfer_seg_fun pf f).
  intros U n n' h1; subst.
  intro f.
  rewrite transfer_seg_fun_eq_refl; auto.
Qed.

Lemma length_map_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    length (map_seg f) = n.
  intros ? n.
  induction n as [|n ih]; simpl; auto.
  intro; rewrite app_length; rewrite ih; f_equal; simpl;  omega.
Qed.

Lemma map_seg_length_eq :
  forall {U V:Type} {n:nat} (f:seg_fun n U) (g:seg_fun n V),
    length (map_seg f) =
    length (map_seg g).
  intros; do 2 rewrite length_map_seg; auto.
Qed.


Lemma map_seg_eq_nil :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    map_seg f = nil ->
    n = 0.
  intros U n.
  destruct n; simpl; auto.
  intros f h1.
  apply app_eq_nil in h1.
  destruct h1; discriminate.
Qed.

Lemma map_seg_eq_nil_iff :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    map_seg f = nil <->
    n = 0.
  intros; split; try apply map_seg_eq_nil; intro; subst; simpl; auto.
Qed.


Lemma in_map_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U)
         (i:nat)
         (pf:i < n),
    In (f i pf) (map_seg f).
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros f i h1.
  pose proof h1 as h1'.
  apply le_lt_eq_dec in h1'.
  destruct h1' as [h2 | h2].
  apply lt_S_n in h2.
  specialize (ih (seg_fun_from_S f) i h2).
  apply in_app_l.
  unfold seg_fun_from_S in ih.
  erewrite f_equal_dep.
  apply ih.
  reflexivity.
  apply S_inj in h2; subst.
  apply in_app_r.
  constructor.
  f_equal; apply proof_irrelevance.
Qed.


Lemma in_map_seg_iff :
  forall {U:Type} {n:nat} (f:seg_fun n U) (b:U),
    In b (map_seg f) <->
    exists (m:nat) (pf:m < n),
        f m pf = b.
  intros U n.
  induction n as [|n ih]; simpl.
  intros f b.
  split; intro h1; try contradiction.
  destruct h1 as [m h1].
  destruct h1; omega.
  intros f b.
  split; intro h1.
  apply in_app_or in h1.
  destruct h1 as [h1 | h1].
  rewrite ih in h1.
  destruct h1 as [m h1]. destruct h1 as [h1 h2].
  subst.
  exists m. 
  ex_goal.
  auto with arith.  
  exists hex.
  unfold seg_fun_from_S. f_equal. apply proof_irrelevance.
  simpl in h1.
  destruct h1 as [h1|]; try contradiction; subst.
  exists n.
  exists (lt_n_Sn _); auto.
  destruct h1 as [m [h1 ?]]; subst.
  pose proof h1 as h3.
  apply le_S_n in h3.
  apply le_lt_eq_dec in h3.
  destruct h3 as [h5 | h5]; subst.
  apply in_app_l.
  rewrite ih.
  exists m, h5.
  unfold seg_fun_from_S; f_equal; apply proof_irrelevance.
  apply in_app_r; simpl.
  left.
  f_equal; apply proof_irrelevance.
Qed.



Lemma map_seg_ext :
  forall {U:Type} {n:nat} (f f':seg_fun n U),
    (forall (m:nat) (pf:m < n), f m pf = f' m pf) ->
    map_seg f =
    map_seg f'.
  intros U n.
  induction n as [|n ih]; simpl; auto.
  intros f f' h1.
  f_equal.
  apply ih.
  intros; apply h1.
  rewrite h1; auto.
Qed.

Lemma map_seg_lt :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    map_seg f <> nil ->
    pred n < n.
  intros U n f h1.
  destruct n as [|n].
  simpl in h1; contradict h1; auto.
  simpl; auto.
Qed.

Lemma map_seg_lt0 :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    map_seg f <> nil ->
    0 < n.
  intros U n f h1.
  destruct n as [|n].
  simpl in h1; contradict h1; auto.
  auto with arith.
Qed.



Fixpoint map_seq {U:Type} {n w:nat} : seq_fun n w U -> list U :=
  match w with
    | 0 => fun _ => nil 
    | S w' => fun f => map_seq (seq_fun_from_S f) ++ f (n+w') (seq_S_w n w') :: nil
  end.

Lemma length_map_seq :
  forall {U:Type} {n w:nat} (f:seq_fun n w U),
    length (map_seq f) = w.
  intros U n w. revert n.
  induction w as [|w ih]; simpl; auto.
  intros n h1.
  rewrite app_length.
  simpl.
  rewrite S_compat.
  f_equal.
  apply ih; auto.
Qed.
  
Lemma map_seg0_decompose :
  forall {U:Type} {n:nat} (f:seg_fun (S n) U),
    map_seg f = (f 0 (lt_0_Sn _)) :: map_seq (seg_S_tl f).
  intros U n.
  induction n as [|n ih]; simpl.
  intro. f_equal. f_equal. apply proof_irrelevance.
  intro f.
  specialize (ih (seg_fun_from_S f)).
  simpl in ih. simpl.
  rewrite ih.
  rewrite <- app_comm_cons.
  unfold seg_fun_from_S.
  f_equal.
  f_equal. apply proof_irrelevance.
  unfold seg_S_tl.
  f_equal.
  unfold seq_fun_from_S.
  f_equal.
  apply functional_extensionality_dep.
  intro m.
  apply functional_extensionality.
  intro. f_equal. apply proof_irrelevance.
  f_equal. f_equal. apply proof_irrelevance.
Qed.

Lemma list_to_set_map_seg :
  forall {U:Type} {n:nat}
         (f:seg_fun n U),
    list_to_set (map_seg f) =
    im_in_ens (seg_fun_to_ens f).
  intros U n f; apply Extensionality_Ensembles; red; split; red; intros x h1.
  rewrite <- list_to_set_in_iff in h1.
  rewrite in_map_seg_iff in h1.
  destruct h1 as [m [h1 h2]]; subst.
  apply im_in_ens_intro with m (in_seg _ _ h1).
  rewrite seg_fun_to_ens_compat.
  f_equal.
  apply proof_irrelevance.
  rewrite <- list_to_set_in_iff.
  destruct h1 as [i h1]; subst.
  destruct h1.
  rewrite seg_fun_to_ens_compat.
  apply in_map_seg.
Qed.


Lemma list_to_set_map_seg_ :
  forall {U:Type} (f:nat->U) (n:nat), 
    list_to_set (map_seg (fun_to_seg f n)) =
    Im (seg n) f.
  intros U f n; apply Extensionality_Ensembles; red; split; red; intros x h1.
  rewrite <- list_to_set_in_iff in h1.
  rewrite in_map_seg_iff in h1.
  destruct h1 as [m [h1 h2]]; subst.
  unfold fun_to_seg.
  apply Im_intro with m; auto.
  constructor; auto.
  destruct h1 as [x h1]; subst.
  destruct h1 as [h1].
  rewrite <- list_to_set_in_iff.
  rewrite in_map_seg_iff.
  exists x, h1.
  unfold fun_to_seg; auto.
Qed.


Lemma in_im_in_ens_nat :
  forall {U:Type} {n:nat}
         (f:seg_fun n U)
         (y:U),
    Inn (im_in_ens f) y ->
    In y (map_seg f).
  intros U n f y h1.
  destruct h1; subst.
  red in pf.
  apply in_map_seg.
Qed.

Lemma in_im_seg :
  forall {U:Type} {n:nat}
         (f:nat->U)
         (y:U),
    Inn (Im (seg n) f) y ->
    In y (map_seg (fun i (pf:i < n) => f i)).
  intros U n f y h1.
  destruct h1; subst.
  destruct H.
  rewrite in_map_seg_iff.
  exists x, H; auto.
Qed.



(*The image of a [seg_fun] in natural [list] form.*)
Fixpoint im_seg_fun {n} : seg_fun n nat -> list nat :=
  match n with
    | 0 => fun _ => nil
    | S n' => fun f =>
                let f' := seg_fun_from_S f in
                (im_seg_fun f') ++ (f n' (lt_n_Sn _)) :: nil
  end.


Lemma in_map_seq_iff :
  forall {U:Type} {n w:nat}
         (f:seq_fun n w U)
         (y:U),
    In y (map_seq f) <->
    exists i (pf:n <= i < n + w),
      f i pf = y.
  intros U n w. revert n.
  induction w as [|w ih]; simpl. 
  intros n f y; split; intro h1.
  contradiction.
  destruct h1 as [i [h1]].
  omega.
  intros n f y; split; intro h1.
  apply in_app_or in h1.
  destruct h1 as [h1 | h1].
  rewrite ih in h1.
  destruct h1 as [i [h1 h2]].
  subst.
  destruct h1 as [h1 h2].
  pose proof (lt_n_Sn (n + w)) as h3.
  pose proof (lt_trans _ _ _ h2 h3) as h4.
  rewrite <- plus_S_shift_r in h4.
  exists i.
  ex_goal.
  split; auto.
  exists hex.
  rewrite seq_fun_from_S_compat; f_equal; apply proof_irrelevance.
  simpl in h1.
  destruct h1 as [h1 | h1].
  exists (n + w).
  ex_goal.
  omega.
  exists hex.
  rewrite <- h1.
  f_equal.
  apply proof_irrelevance.
  contradiction.
  destruct h1 as [i [h1 h2]].
  subst.
  pose proof h1 as h2.
  destruct h2 as [h2 h3].
  apply lt_dec in h3.
  destruct h3 as [h3 | h3].
  rewrite plus_S_shift_r in h3.
  simpl in h3.
  apply in_app_l.
  rewrite ih.
  exists _, (conj h2 h3).
  rewrite seq_fun_from_S_compat.
  f_equal.
  apply proof_irrelevance.
  subst.
  apply in_app_r.
  constructor.
  apply f_equal_dep.
  omega.
Qed.

Fixpoint map_in_dep {T U:Type} {l:list T} : fun_in_dep l U ->list U :=
  match l with
    | nil => fun _ => nil
    | a::l' => fun f => f a (in_eq a l') :: map_in_dep (fun_from_cons f)
  end.  


Lemma map_in_dep_ext :
  forall {T U:Type} {l:list T}
         (f g:fun_in_dep l U),
    (forall x:T, In x l -> f x = g x) ->
    map_in_dep f = map_in_dep g.
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros; auto.
  intros f g h1.
  rewrite h1.
  f_equal.
  apply ih.
  intros x h2. 
  specialize (h1 x (or_intror _ h2)).
  apply fun_from_cons_functional; auto.  
  left; auto.
Qed.


(*Search for [map_in_dep_cons] for the more standard 
properties of the above.*)
Fixpoint map_in_dep_prop_l {T U:Type} {P:T->Prop} {l:list T} 
: (forall x:T, In x l -> {P x} + {~ P x}) ->
  fun_in_dep_prop l P U -> list U :=
  match l with
    | nil => fun _  _=> nil
    | a::l' => fun pf f =>
                 match (pf a (in_eq a l')) with
                   | left pfp =>
                     f a (in_eq a l') pfp ::
                       (map_in_dep_prop_l (set_from_cons pf)
                                          (fun_prop_from_cons f))
                   | right _ =>
                     map_in_dep_prop_l (set_from_cons pf)
                                       (fun_prop_from_cons f)
                 end
  end.


Fixpoint map_in_dep_prop_r {T U:Type} {P:T->Prop} {l:list T} 
: (forall x:T, In x l -> {P x} + {~P x}) ->
  fun_in_dep_prop l (not_pred P) U -> list U :=
  match l with
    | nil => fun _  _=> nil
    | a::l' => fun pf f =>
                 match (pf a (in_eq a l')) with
                   | left _ =>
                     map_in_dep_prop_r (set_from_cons pf)
                                       (fun_prop_from_cons f)
                   | right pfq =>
                     f a (in_eq a l') pfq ::
                       (map_in_dep_prop_r (set_from_cons pf)
                                          (fun_prop_from_cons f))
                 end
  end.


(*Use this when all elements in [l] satisfy [P]*)
Definition fun_in_dep_prop_all
           {T U:Type} {P:T->Prop} {l:list T}
           (f:fun_in_dep_prop l P U)
           (pfl:forall x:T, In x l -> P x) : fun_in_dep l U :=
  fun (x:T) (pf:In x l) => f x pf (pfl x pf).


Lemma map_in_dep_prop_l_ext :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pf pf':lexcl_mid l P)
         (f g:fun_in_dep_prop l P U),
    (forall x:T, In x l -> P x -> f x = g x) ->
    map_in_dep_prop_l pf f = map_in_dep_prop_l pf' g.
  intros T U P l h0.
  induction l as [|a l ih]; simpl; auto.
  intros h1 f g h2.
  destruct (h0 a (in_eq a l)) as [h3 | h3], (h1 a (in_eq a l)) as [h3' | h3']; try contradiction.
  rewrite h2.
  f_equal. f_equal. apply proof_irrelevance.
  f_equal. f_equal.
  simpl in h0.
  apply prop_dep_set_dep_ext.
  apply unit_set_pred_fexcl_mid.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality.
  intro h4.
  apply functional_extensionality.
  intro h5.
  do 2 rewrite fun_prop_from_cons_compat.
  rewrite h2.
  f_equal.
  right; auto.
  assumption.
  left; auto.
  assumption.
  erewrite ih.
  reflexivity.
  intros x h4 h5.
  apply functional_extensionality.
  intro h6.
  apply functional_extensionality.
  intro h7.
  do 2 rewrite fun_prop_from_cons_compat.
  rewrite h2; auto.
Qed.



Lemma map_in_dep_prop_l_all :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pf:forall x:T, In x l -> {P x} + {~ P x})
         (f:fun_in_dep_prop l P U)
         (pfl:forall x:T, In x l -> P x),
    let f' := fun_in_dep_prop_all f pfl in
    map_in_dep_prop_l pf f = 
    map_in_dep f'.
  intros T U P l. simpl.
  induction l as [|a l ih]; simpl; auto.
  intros h1 f h2.
  lr_match_goal. 
  fold hlr.
  destruct hlr as [h3 | h3].
  unfold fun_in_dep_prop_all.
  f_equal. f_equal. apply proof_irrelevance. 
  hfirst ih.
  intros x h4. apply h1. right; auto. 
  specialize (ih hf (fun_prop_from_cons f)).
  hfirst ih.  
  intros; apply h2. right; auto.
  specialize (ih hf0).
  unfold set_from_cons.
  gterml.
  terml ih. 
  assert (h4:x = x0).
    unfold x, x0.
    apply map_in_dep_prop_l_ext.
    intros; auto.
  fold x.
  rewrite h4.
  unfold x0.
  rewrite ih. 
  apply map_in_dep_ext. 
  intros c h5.
  unfold fun_in_dep_prop_all, fun_from_cons.
  apply functional_extensionality.
  intros h6.
  unfold fun_prop_from_cons. 
  f_equal; apply proof_irrelevance.
  pose proof (h2 a (in_eq _ _)); contradiction.
Qed.

Lemma map_in_dep_prop_l_none :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pf:forall x:T, In x l -> {P x} + {~ P x})
         (f:fun_in_dep_prop l P U),
    (forall x:T, In x l -> ~ P x) ->
    map_in_dep_prop_l pf f = nil.
  intros T U P l.
  induction l as [|a l ih]; simpl; auto.
  intros h1 f h2.
  destruct (h1 a (in_eq a l)) as [h3 | h3].
  contradict h3.
  apply h2; auto.
  apply ih; auto.
Qed.


Lemma map_in_dep_prop_r_ext :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pf pf':lexcl_mid l P)
         (f g:fun_in_dep_prop l (not_pred P) U),
    (forall x:T, In x l -> ~ P x -> f x = g x) ->
    map_in_dep_prop_r pf f = map_in_dep_prop_r pf' g.
  intros T U P l h0.
  induction l as [|a l ih]; simpl; auto.
  intros h1 f g h2.
  destruct (h0 a (in_eq a l)) as [h3 | h3], (h1 a (in_eq a l)) as [h3' | h3']; try contradiction.
  f_equal. f_equal.
  apply prop_dep_set_dep_ext.
  apply unit_set_pred_fexcl_mid.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality.
  intro h4.
  apply functional_extensionality.
  intro h5.
  simpl in h5.
  do 2 rewrite fun_prop_from_cons_compat.
  rewrite h2.
  f_equal.
  right; auto.
  assumption.
  rewrite h2.
  f_equal. f_equal.
  apply functional_extensionality_dep.
  intro.
  contradiction.
  erewrite ih.
  reflexivity.
  intros x h4 h5.
  apply functional_extensionality.
  intro h6.
  apply functional_extensionality.
  intro h7.
  do 2 rewrite fun_prop_from_cons_compat.
  rewrite h2; auto.
  left; auto.
  assumption.
Qed.



Lemma map_in_dep_prop_r_all :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pf:forall x:T, In x l -> {P x} + {~ P x})
         (f:fun_in_dep_prop l (not_pred P) U)
         (pfl:forall x:T, In x l -> ~ P x),
    let f' := fun_in_dep_prop_all f pfl in
    map_in_dep_prop_r pf f = 
    map_in_dep f'.
  intros T U P l. simpl.
  induction l as [|a l ih]; simpl; auto.
  intros h1 f h2.
  lr_match_goal. 
  fold hlr.
  destruct hlr as [h3 | h3].
  contradict h3.
  apply h2. left; auto.
  unfold fun_in_dep_prop_all.
  f_equal. f_equal. apply proof_irrelevance. 
  hfirst ih.
  intros x h4. apply h1. right; auto. 
  specialize (ih hf (fun_prop_from_cons f)).
  hfirst ih. 
  intros; apply h2. right; auto.
  specialize (ih hf0).
  unfold set_from_cons.
  gterml.
  terml ih.
  assert (h4:x = x0).
    unfold x, x0.
    apply map_in_dep_prop_r_ext.
    intros; auto.
  fold x.
  rewrite h4.
  unfold x0.
  rewrite ih.
  apply map_in_dep_ext. 
  intros c h5.
  unfold fun_in_dep_prop_all, fun_from_cons.
  apply functional_extensionality.
  intros h6.
  unfold fun_prop_from_cons.
  f_equal; apply proof_irrelevance.
Qed.


Lemma map_in_dep_prop_r_none :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pf:forall x:T, In x l -> {P x} + {~ P x})
         (f:fun_in_dep_prop l (not_pred P) U),
    (forall x:T, In x l -> P x) ->
    map_in_dep_prop_r pf f = nil.
  intros T U P l.
  induction l as [|a l ih]; simpl; auto.
  intros h1 f h2.
  destruct (h1 a (in_eq a l)) as [h3 | h3].
  apply ih; auto.
  contradict h3.
  apply h2; auto.
Qed.



Lemma in_map_in_dep_prop_l :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pfd:(forall x:T, In x l -> {P x} + {~P x}))
         (f:fun_in_dep_prop l P U) (x:T)
         (pfin:In x l) (pfp:P x),
    In (f x pfin pfp) (map_in_dep_prop_l pfd f).
  intros T U P l.
  induction l as [|a l ih]; simpl.  
  intros; contradiction.
  intros h1 f x h3 h4. 
  destruct h3 as [h3 | h3]; subst.
  destruct (h1 x (in_eq x l)) as [h5 | h5]. simpl.
  left. f_equal. apply proof_irrelevance. apply proof_irrelevance.
  contradiction.
  specialize (ih (set_from_cons h1) (fun_prop_from_cons f)).
  destruct (h1 a (in_eq a l)) as [h5 | h5].
  right.
  apply ih.
  apply ih.
Qed.



Lemma in_map_in_dep_prop_r :
  forall {T U:Type} {P:T->Prop} {l:list T}
         (pfd:(forall x:T, In x l -> {P x} + {~P x}))
         (f:fun_in_dep_prop l (fun x => ~P x) U) (x:T)
         (pfin:In x l) (pfp:~P x),
    In (f x pfin pfp) (map_in_dep_prop_r pfd f).
  intros T U P l.
  induction l as [|a l ih]; simpl.  
  intros; contradiction.
  intros h1 f x h3 h4. 
  destruct h3 as [h3 | h3]; subst. 
  destruct (h1 x (in_eq x l)) as [h5 | h5]. simpl.
  contradiction.
  left. f_equal. apply proof_irrelevance. apply proof_irrelevance.
  specialize (ih (set_from_cons h1) (fun_prop_from_cons f)).
  destruct (h1 a (in_eq a l)) as [h5 | h5].
  apply ih.
  right.
  apply ih.
Qed.


Lemma in_map_in_dep_prop_l_iff :
  forall {T U:Type} {P:T->Prop} {l:list T} 
         (pfd:(forall x:T, In x l -> {P x} + {~P x}))
         (f:fun_in_dep_prop l P U) (y:U),
       In y (map_in_dep_prop_l pfd f) <->
    (exists (x:T) (pf:In x l) (pfp:P x),
      f x pf pfp = y).
  intros T U P l.
  induction l as [|a l h1]; simpl.
  intros. split. intro; contradiction.
  intro h1. destruct h1 as  [? [ ]]. contradiction.
  split.
  intro h2. 
  destruct (pfd a (in_eq a l)) as [h3 | h3].
  destruct h2 as [h2 | h2]; subst.
  exists a. exists (or_introl eq_refl). exists h3. f_equal. apply proof_irrelevance.
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) y).
  rewrite h1 in h2.
  destruct h2 as [x h2]. destruct h2 as [h2 h4]. destruct h4 as [h4 h5]; subst.
  unfold fun_prop_from_cons.
  exists x. exists (or_intror h2). exists h4. f_equal.
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) y).
  rewrite h1 in h2.
  destruct h2 as [x h2]. destruct h2 as [h2 h4]. destruct h4 as [h4 h5]; subst.
  unfold fun_prop_from_cons.
  exists x. exists (or_intror h2). exists h4. f_equal.
  simpl.
  intro h2. 
  destruct h2 as [x h2]. destruct h2 as [h2 h3]. destruct h3 as [h3 h4]; subst.
  destruct h2 as [h2 | h2]; subst.
  destruct (pfd x (in_eq x l)) as [h5 | h5]. 
  simpl. left. f_equal. apply proof_irrelevance. apply proof_irrelevance.
  contradiction.
  destruct (pfd a (in_eq a l)) as [h4 | h4].
  right.
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) (f x (or_intror h2) h3)).
  rewrite h1.
  exists x, h2, h3.
  unfold fun_prop_from_cons. f_equal.
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) (f x (or_intror h2) h3)).
  rewrite h1.
  exists x, h2, h3.
  unfold fun_prop_from_cons. f_equal.
Qed.

Lemma in_map_in_dep_prop_r_iff :
  forall {T U:Type} {P:T->Prop} {l:list T} 
         (pfd:(forall x:T, In x l -> {P x} + {~P x}))
         (f:fun_in_dep_prop l (not_pred P) U) (y:U),
       In y (map_in_dep_prop_r pfd f) <->
    (exists (x:T) (pf:In x l) (pfp:~ P x),
      f x pf pfp = y).
  intros T U P l.
  induction l as [|a l h1]; simpl.
  intros. split. intro; contradiction.
  intro h1. destruct h1 as  [? [ ]]. contradiction. 
  split. 
  intro h2.  
  destruct (pfd a (in_eq a l)) as [h3 | h3].
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) y).
  rewrite h1 in h2.
  destruct h2 as [x h2]. destruct h2 as [h2 h4]. destruct h4 as [h4 h5]; subst.
  unfold fun_prop_from_cons.
  exists x. exists (or_intror h2). exists h4. f_equal.
  destruct h2 as [h2 | h2]; subst.
  exists a, (or_introl eq_refl), h3. f_equal. apply proof_irrelevance.
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) y).
  rewrite h1 in h2.
  destruct h2 as [x [h2 [h4 h5]]].
  exists x, (or_intror h2), h4.
  rewrite <- h5 at 1.
  unfold fun_prop_from_cons. f_equal.
  intro h2.
  destruct h2 as [x h2]. destruct h2 as [h2 h3]. destruct h3 as [h3 h4]; subst.
  destruct h2 as [h2 | h2]; subst.
  destruct (pfd x (in_eq x l)) as [h4 | h4].
  contradiction.
  left. f_equal. apply proof_irrelevance. apply proof_irrelevance.
  destruct (pfd a (in_eq a l)) as [h4 | h4].
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) (f x (or_intror h2) h3)).
  rewrite h1.
  exists x, h2, h3.
  unfold fun_prop_from_cons.
  f_equal.
  right.
  specialize (h1 (set_from_cons pfd) (fun_prop_from_cons f) (f x (or_intror h2) h3)).
  rewrite h1.
  exists x, h2, h3.
  unfold fun_prop_from_cons. f_equal.
Qed.



Lemma map_in_dep_cons :
  forall {T U:Type} {l:list T} (a:T) (f:fun_in_dep (a::l) U),
    map_in_dep f = (f a (in_eq a l)) :: map_in_dep (fun_from_cons f).
  intros; auto.
Qed.

Lemma map_in_dep_functional :
  forall {T U:Type} {l l':list T} (f:fun_in_dep l U)
         (pf: l = l'),
    map_in_dep f = map_in_dep (transfer_fun_in_dep f pf).
  intros; subst; rewrite transfer_fun_in_dep_eq_refl; auto.
Qed.

Lemma map_in_dep_sing :
  forall {T U:Type} {a:T} (f:fun_in_dep (a::nil) U),
    map_in_dep f = f a (in_eq _ _) :: nil.
  intros; auto.
Qed.


Lemma in_map_in_dep :
  forall {T U:Type} {l:list T} (f:forall x:T, In x l->U) (x:T)
         (pf:In x l),
    In (f x pf) (map_in_dep f).
  intros T U l.
  induction l as [|a l h1].
  intros; contradiction.
  intros f x h2. 
  destruct h2 as [h2 | h2]; subst.
  unfold map_in_dep. simpl.
  left. f_equal. apply proof_irrelevance.
  specialize (h1 (fun_from_cons f) x h2).
  rewrite fun_from_cons_compat in h1.
  simpl. right.
  assert (h3:in_cons a x l h2 = or_intror h2).
    apply proof_irrelevance.
  rewrite <- h3; auto.
Qed.

Lemma in_map_in_dep_iff :
  forall {T U:Type} {l:list T} (f:forall x:T, In x l->U) (y:U),
    In y (map_in_dep f) <->
    exists (x:T) (pf:In x l),
      f x pf = y.
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros. split. intro; contradiction.
  intro h1.  destruct h1 as [x h1]. destruct h1; contradiction.
  intros f y.
  specialize (ih (fun_from_cons f) y).
  split.
  intros h1.
  destruct h1 as [h1 | h1]. subst.
  exists a. exists (or_introl eq_refl). f_equal. apply proof_irrelevance. 
  rewrite ih in h1.  clear ih.
  destruct h1 as [x [h1 h2]]. subst.
  exists x. exists (or_intror h1).
  rewrite fun_from_cons_compat. f_equal. apply proof_irrelevance.
  intro h1.
  destruct h1 as [x h1]. destruct h1 as [h1 h2].
  subst.
  destruct h1 as [h1|h1]. subst.
  left. f_equal. apply proof_irrelevance.
  destruct ih as [h2 h3].
  hfirst h3.
  exists x, h1. rewrite fun_from_cons_compat.
  f_equal. apply proof_irrelevance.
  specialize (h3 hf). clear hf.
  right.
  assumption.
Qed.

Lemma length_map_in_dep :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U),
    length (map_in_dep f) = length l.
  intros  ? ? l; induction l; simpl; auto.
Qed.

Lemma map_in_dep_eq_nil :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U),
    map_in_dep f = nil -> l = nil.
  intros T U l.
  destruct l as [|a l]; auto.
  intros f h1.
  pose proof (in_map_in_dep f a (in_eq a l)) as h2.
  rewrite h1 in h2.
  contradiction.
Qed.

Lemma map_in_dep_eq_nil_iff :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U),
    map_in_dep f = nil <->
    l = nil.
  intros; split; intro h1; try apply map_in_dep_eq_nil in h1;
  auto; subst; auto.
Qed.
  

Lemma map_in_dep_id :
  forall {T:Type} (l:list T),
    map_in_dep (id_in_dep l) = l.
  intros T l; induction l; simpl; auto.
  rewrite id_in_dep_compat.
  f_equal; auto.
Qed.

Lemma map_const_in_dep :
  forall {T U:Type} (l:list T) (val:U),
    map_in_dep (const_in_dep l val) =
    lrep val (length l).
  intros T U l.
  induction l as [|a l ih]; simpl; auto.
  intro y.
  unfold const_in_dep.
  f_equal.
  apply ih.
Qed.

Lemma map_map_in_dep_ext :
  forall {T U:Type} (l:list T)
         (f:T->U) (g:fun_in_dep l U),
    (forall (x:T) (pf:In x l), f x = g x pf) ->
    map f l = map_in_dep g.
  intros T U l.
  induction l as [|a l ih]; simpl; auto.
  intros f g h1.
  pose proof (h1 a (or_introl _ eq_refl)) as h2.
  rewrite h2.
  f_equal. f_equal. apply proof_irrelevance.
  apply ih. intros; apply h1.
Qed.

(*See past defintiion of [nth_lt] for a similar lemma [map_in_dep_eq_iff]*)
Lemma map_in_dep_app :
  forall {T U:Type} {l l':list T} (f:fun_in_dep (l++l') U),
    map_in_dep f =
    (map_in_dep (fun_from_app1 f)) ++
                                   (map_in_dep (fun_from_app2 f)). 
  intros T U.
  induction l as [|a l h1]; simpl.
  intros l' f.
  unfold fun_from_app2.
  f_equal. apply fun_in_dep_ext.
  intros. f_equal. apply proof_irrelevance.
  intros l' f.
  rewrite fun_from_app1_compat.
  f_equal. f_equal. apply proof_irrelevance.
  rewrite h1.
  f_equal. f_equal. apply fun_in_dep_ext.
  intros.
  unfold fun_from_cons.
  do 2 rewrite fun_from_app1_compat.
  f_equal. apply proof_irrelevance.
  specialize (h1 _ (fun_from_cons f)).
  unfold fun_from_app2. f_equal.
  apply fun_in_dep_ext.
  intros.  unfold fun_from_cons. f_equal. apply proof_irrelevance.
Qed.


Lemma map_in_dep_map :
  forall {T U V:Type} (la:list T) (f:T->U)
         (g:fun_in_dep (map f la) V),
    map_in_dep (fun (b:U) (pfb:In b (map f la)) => g b pfb) =
    map_in_dep (fun (a:T) (pfa:In a la) => g (f a) (in_map _ _ _ pfa)).
  intros T U V la.
  induction la as [|a la ih]; simpl; auto.
  intros f g.
  f_equal. f_equal. apply proof_irrelevance.
  specialize (ih f (fun_from_cons g)).
  let P :=
      type of ih
  in match P with
         ?x = _ => match goal with |- ?y = _ => assert (h1:x = y) end end.
  apply map_in_dep_ext.
  intros b h1.
  apply functional_extensionality. 
  intros. unfold fun_from_cons. f_equal. 
  rewrite h1 in ih. rewrite ih.
  apply map_in_dep_ext.
  intros x h3.
  apply functional_extensionality.
  intro h4.
  unfold fun_from_cons.  f_equal. apply proof_irrelevance.
Qed.

  
Lemma map_map_in_dep :
  forall {T U V:Type} (la:list T)
         (f:fun_in_dep la U)
         (g:U->V),
    map g (map_in_dep f) =
    map_in_dep (fun (a:T) (pfa:In a la) => g (f a pfa)).
  intros T U V la.
  induction la as [|a la ih]; simpl; auto. 
  intros f g.
  f_equal. 
  specialize (ih (fun_from_cons f) g).
  rewrite ih.
  unfold fun_from_cons; auto.
Qed.


Lemma map_in_dep_map_in_dep :
  forall {T U V:Type} (la:list T)
         (f:fun_in_dep la U)
         (g:fun_in_dep (map_in_dep f) V),
    map_in_dep (fun (b:U) (pfb:In b (map_in_dep f)) => g b pfb) =
    map_in_dep (fun (a:T) (pfa:In a la) => g (f a pfa) (in_map_in_dep _ _  pfa)).
  intros T U V la.
  induction la as [|a la ih]; simpl; auto.
  intros f g. 
  f_equal. f_equal. apply proof_irrelevance.
  specialize (ih (fun_from_cons f) (fun_from_cons g)).
  unfold fun_from_cons. unfold fun_from_cons in ih.
  rewrite ih.
  apply map_in_dep_ext.
  intros b h1.
  apply functional_extensionality.
  intros; f_equal.
  apply proof_irrelevance.
Qed.


Lemma map_eq_map_in_dep :
  forall {T U:Type} (l:list T) (f:T->U),
    map f l = map_in_dep (fun (x:T) (_:In x l) => f x).
  intros ? ? l.
  induction l as [|a l h1]; simpl; auto. 
  intro; f_equal. auto.
  unfold fun_from_cons; auto.
Qed.


Lemma hd_map_in_dep' :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U)
         (pf:map_in_dep f <> nil),
    hd' (map_in_dep f) pf =
    f (hd' l (iffn1 (map_in_dep_eq_nil_iff f) pf)) (in_hd' _ _).
  intros T U l.
  destruct l as [|a l]; simpl.
  intros ? h1; contradict h1; auto.
  intros; f_equal; apply proof_irrelevance.
Qed.


Lemma list_to_set_map_in_dep :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U),
    list_to_set (map_in_dep f) =
    im_in_ens (fun i (pf:Inn (list_to_set l) i) =>
                 let pf' := iff2 (list_to_set_in_iff _ _) pf in
                 f i pf').
  intros T U l f.
  apply Extensionality_Ensembles; red; split; red; intros y h1.
  rewrite <- list_to_set_in_iff in h1.
  rewrite in_map_in_dep_iff in h1.
  destruct h1 as [x [h1 h2]]; subst.
  pose proof h1 as h1'.
  rewrite list_to_set_in_iff in h1'.
  apply im_in_ens_intro with x h1'.
  f_equal.
  apply proof_irrelevance.
  destruct h1 as [y h1]; subst.
  rewrite <- list_to_set_in_iff.
  apply in_map_in_dep.
Qed.



Definition fun_in_dep_decompose1  
           {T U:Type} (pfdt:type_eq_dec T) (xl l:list T)
           (pfincl:incl xl l)
           (f:fun_in_dep l U) : fun_in_dep xl U :=
  fun x (pfi:In x xl) => f x (pfincl _ pfi).

Definition fun_in_dep_decompose2
           {T U:Type} (pfdt:type_eq_dec T) (xl l:list T)
           (pfincl:incl xl l)
           (f:fun_in_dep l U) :
  fun_in_dep (list_minus pfdt l xl) U :=
  fun x (pfi:In x (list_minus pfdt l xl)) =>
      let (pfi', _) := in_list_minus_impl _ _ _ _ pfi in
      f x pfi'.

(*These use [sub_list] instead of [incl]*)
Definition fun_in_dep_decompose1'
           {T U:Type} (pfdt:type_eq_dec T) (xl l:list T)
           (pf:sub_list xl l)
           (f:fun_in_dep l U) :
  fun_in_dep xl U :=
  fun_in_dep_decompose1 pfdt xl l (sub_list_incl _ _ pf) f.


Definition fun_in_dep_decompose2'
           {T U:Type} (pfdt:type_eq_dec T) (xl l:list T)
           (pf:sub_list xl l)
           (f:fun_in_dep l U) :
  fun_in_dep (list_minus pfdt l xl) U :=
  fun_in_dep_decompose2 pfdt xl l (sub_list_incl _ _ pf) f.



Lemma list_to_set_map_in_dep_decompose :
  forall {T U:Type} (pfdt:type_eq_dec T)
         (l l':list T) (f:fun_in_dep l' U)
         (pf:incl l l'),
    let f1 := fun_in_dep_decompose1 pfdt l l' pf f in
    let f2 := fun_in_dep_decompose2 pfdt l l' pf f in
    list_to_set (map_in_dep f) =
    Union (list_to_set (map_in_dep f1)) (list_to_set (map_in_dep f2)).
  intros T U h0 l l' f h1. simpl.
  apply Extensionality_Ensembles; red; split; red; intros y h2. 
  rewrite <- list_to_set_in_iff  in h2.
  rewrite in_map_in_dep_iff in h2.
  destruct h2 as [x [h2 h3]]; subst.
  destruct (in_dec h0 x l) as [h3 | h3].
  left.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  exists x, h3.
  unfold fun_in_dep_decompose1; f_equal; apply proof_irrelevance. 
  right.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff. 
  exists x, (impl_in_list_minus _ _ _ _ h2 h3).
  unfold fun_in_dep_decompose2; simpl.
  destruct (in_list_minus_impl h0 l' l x (impl_in_list_minus _ _ _ _ h2 h3)).
  f_equal.
  apply proof_irrelevance.

  destruct h2 as [y h2 | y h2].
  rewrite <- list_to_set_in_iff in h2.
  rewrite in_map_in_dep_iff in h2.
  destruct h2 as [x [h2 h3]]. subst.
  rewrite <- list_to_set_in_iff.
  unfold fun_in_dep_decompose1.
  rewrite in_map_in_dep_iff.
  exists x, (h1 x h2); auto.
  rewrite <- list_to_set_in_iff in h2.
  rewrite in_map_in_dep_iff in h2.
  destruct h2 as [x [h2 h3]]. subst.
  rewrite <- list_to_set_in_iff.
  unfold fun_in_dep_decompose2.
  destruct (in_list_minus_impl _ _ _ _ h2).
  apply in_map_in_dep_iff.
  exists x, i; auto. 
Qed.



(*This is kind of weak, but the next lemma strengthens it.*)
Lemma list_to_set_map_in_dep_remove_inj :
  forall {T U:Type} (pfdt:type_eq_dec T) (l:list T)
         (f:fun_in_dep l U)
         (x:T) (pf:In x l),
    inj_dep f ->
    list_to_set (map_in_dep (fun_in_dep_remove pfdt f x)) =
    Subtract (list_to_set (map_in_dep f)) (f x pf).
  intros T U h1 l f x h2 hi.
  apply Extensionality_Ensembles; red; split; red; intros y h3.
  rewrite <- list_to_set_in_iff in h3.
  rewrite in_map_in_dep_iff in h3.
  destruct h3 as [a [h3 h4]]; subst.
  pose proof h3 as h3'. rewrite in_remove_iff in h3'.
  destruct h3' as [h4 h5].
  constructor.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  exists a, h4.
  rewrite fun_in_dep_remove_compat.
  f_equal. apply proof_irrelevance.
  rewrite in_sing_iff.
  intro h6.
  rewrite fun_in_dep_remove_compat in h6.
  red in hi.
  specialize (hi x a h2 (in_remove_impl h1 l a x h3) h6).
  subst.
  contradict h5; auto.
  destruct h3 as [h3 h4].
  rewrite in_sing_iff in h4.
  rewrite <- list_to_set_in_iff in h3.
  rewrite in_map_in_dep_iff in h3.
  destruct h3 as [a [h3 h5]]; subst.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  exists a.
  ex_goal.
  rewrite in_remove_iff.
  split; auto.
  intro; subst.
  contradict h4.
  f_equal. apply proof_irrelevance.
  exists hex.
  rewrite fun_in_dep_remove_compat.
  f_equal. apply proof_irrelevance.
Qed.

Lemma list_to_set_map_in_dep_remove_in :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (f:fun_in_dep l U)
         (x:T) (pf:In x l),
    list_to_set (map_in_dep (fun_in_dep_remove_in pfdt pfdu f x pf)) =
    Subtract (list_to_set (map_in_dep f)) (f x pf).
  intros T U h1 h2 l f a h3.
  apply Extensionality_Ensembles; red; split; red; intros y h4.
  rewrite <- list_to_set_in_iff in h4.
  rewrite in_map_in_dep_iff in h4.
  destruct h4 as [x [h4 h5]].
  pose proof (in_list_minus_impl h1 _ _ _ h4) as h6.
  destruct h6 as [h6 h7].
  constructor.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  exists x, h6.
  subst.
  rewrite <- fun_in_dep_remove_in_compat.
  f_equal.
  apply proof_irrelevance.
  contradict h7.
  rewrite in_sing_iff in h7; subst.
  erewrite <- fun_in_dep_remove_in_compat in h7.
  rewrite in_inv_im_in_dep_iff.
  exists h6.
  rewrite h7.
  f_equal. apply proof_irrelevance.

  destruct h4 as [h4 h5].
  rewrite <- list_to_set_in_iff in h4.
  rewrite in_map_in_dep_iff in h4.
  destruct h4 as [x [h4 h6]].
  subst.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  exists x.
  ex_goal.
  eapply impl_in_list_minus; auto.
  contradict h5.
  rewrite in_sing_iff.
  rewrite in_inv_im_in_dep_iff in h5.
  destruct h5 as [h5 h6].
  pose proof (proof_irrelevance _ h4 h5); subst.
  rewrite h6; auto.
  exists hex.
  
  erewrite <- fun_in_dep_remove_in_compat'.
  reflexivity.
Qed.

  
Lemma list_to_set_map_in_dep_decompose_sumbool :
  forall {T U:Type} (l:list T) (Q:T->Prop)
         (pfs:forall x:T, In x l -> {Q x} + {~Q x})
         (fq:fun_in_dep_prop l Q U)
         (fr:fun_in_dep_prop l (fun x => ~ Q x) U), 
    let f := fun (x:T) (pf:In x l) =>
               match (pfs x pf) with
                 | left pfq => fq x pf pfq
                 | right pfr => fr x pf pfr
               end in
    list_to_set (map_in_dep f) =
    Union (list_to_set (map_in_dep_prop_l pfs fq))
          (list_to_set (map_in_dep_prop_r pfs fr)).
  intros T U l Q h1 fq fr. simpl.
  apply Extensionality_Ensembles; red; split; red; intros y h2. 
  rewrite <- list_to_set_in_iff in h2.
  rewrite in_map_in_dep_iff in h2. 
  destruct h2 as [x h2]. destruct h2 as [h2 h3]; subst.
  destruct (h1 x h2) as [h4 | h4]; subst.
  left.
  rewrite <- list_to_set_in_iff.
  apply in_map_in_dep_prop_l.
  right.
  rewrite <- list_to_set_in_iff.
  apply in_map_in_dep_prop_r.

  destruct h2 as [y h2 | y h2].
  rewrite <- list_to_set_in_iff in h2.
  rewrite in_map_in_dep_prop_l_iff in h2.
  destruct h2 as [x [h2 [h3 h4]]]; subst.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  exists x, h2.
  destruct (h1 x h2) as [h4 | h4].
  f_equal. apply proof_irrelevance.
  contradiction.
  rewrite <- list_to_set_in_iff in h2.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_prop_r_iff in h2.
  destruct h2 as [x [h2 [h3 h4]]]; subst.
  rewrite in_map_in_dep_iff.
  exists x, h2.
  destruct (h1 x h2) as [h4 | h4].
  contradiction.
  f_equal. apply proof_irrelevance.
Qed.


Lemma lS_map_in_dep :
  forall {T:Type} (l:list T) (f:fun_in_dep l nat),
    lS (map_in_dep f) =
    map_in_dep (fun c pf =>  S (f c pf)).
  intros T l.
  induction l as [|a l ih]; simpl; auto.
  intro; f_equal; apply ih.
Qed.
                
    

(*Returns a list of all elemnts in [l] which satisfy [P]*)
Fixpoint map_filter {T U:Type} (P:U -> Prop) (f:T->U)
         (pfp:forall u:U, {P u} + {~P u}) 
         (l:list T) : list T :=
  match l with
    | nil => nil
    | a::l' => let ml := map_filter P f pfp l' in 
               if (pfp (f a) ) then
                 a :: ml
               else
                 ml
  end.

Fixpoint map_filter_not {T U:Type} (P:U -> Prop) (f:T->U)
         (pfr:forall u:U, {P u} + {~P u}) 
         (l:list T) : list T :=
  match l with
    | nil => nil
    | a::l' => let ml := map_filter P f pfr l' in 
               if (pfr (f a) ) then
                 ml               
               else
                 a :: ml
  end.

(*Filters out all elements whose [f] values don't match [u]*)
Definition map_filter_eq {T U:Type} (pfdu:type_eq_dec U) (f:T->U)
           (u:U) (l:list T) :=
  map_filter (fun c => u = c) f (pfdu u) l.


Definition map_filter_lt {T U:Type}
           (pfdu:type_eq_dec U)
           (R:Relation U)
           (pfrlem:Rlem_dec R)
           (f:T->U)
           (u:U) (l:list T) :=
  map_filter (fun c => R c u) f (fun c => (pfrlem c u)) l.


Definition map_filter_gt {T U:Type}
           (pfdu:type_eq_dec U)
           (R:Relation U)
           (pfrlem:Rlem_dec R)
           (f:T->U)
           (u:U) (l:list T) :=
  map_filter (fun c => R u c) f (pfrlem u) l.

Definition map_filter_le {T U:Type}
           (pfdu:type_eq_dec U)
           (R:Relation U)
           (pfrlem:Rlem_dec R)
           (f:T->U) 
           (u:U) (l:list T) :=
  map_filter (fun c => R c u) f (fun c => (pfrlem c u)) l.


Definition map_filter_ge {T U:Type}
           (pfdu:type_eq_dec U)
           (R:Relation U)
           (pfrwlem:Rwlem_dec R)
           (f:T->U)
           (u:U) (l:list T) :=
  map_filter (fun c => Rw R u c) f (pfrwlem u) l.

  
Lemma map_filter_eq_nil_iff :
  forall {T U:Type} (P:U -> Prop) (f:T->U)
         (pfp:forall u:U, {P u} + {~P u}) 
         (l:list T),
    map_filter P f pfp l = nil <->
    l = nil \/ forall x:T, In x l -> ~P (f x).
  intros T U P f h1 l.
  revert h1. revert f P.
  induction l as [|c l ih]; simpl.
  intuition.
  intros f P h1; split; intro h2.
  destruct (h1 (f c)) as [h6 | h6].
  discriminate.
  right.
  intros x h4.
  destruct h4 as [|h4]; subst; auto.
  specialize (ih f P h1).
  rewrite ih in h2.
  destruct h2 as [| h2]; subst.
  contradiction.
  apply h2; auto.
  destruct h2 as [h2 | h2].
  discriminate.
  destruct (h1 (f c)) as [h6 | h6].
  specialize (h2 c (or_introl eq_refl)).
  contradiction.
  rewrite ih.
  right; auto.
Qed.


  
Lemma inv_im_in_dep_l_cons_eq_nil_impl :
  forall {T U:Type} (pfdu:type_eq_dec U) (l:list T) (a:T)
         (f:fun_in_dep (a::l) U) (lu:list U),
    inv_im_in_dep_l pfdu (a::l) f lu = nil ->
    inv_im_in_dep_l pfdu l (fun_from_cons f) lu = nil.
  intros T U h1 l a f lu.
  revert f. revert a l h1.
  induction lu as [|b lu ih]; simpl.
  intros; auto.
  intros a l h1 v h2.
  destruct (h1 b (v a (in_eq _ _))) as [h3 | h3]; subst.
  apply app_eq_nil in h2.
  destruct h2 as [h2 h3].
  discriminate.
  apply app_eq_nil in h2.
  destruct h2 as [h2 h4]. 
  rewrite h2, ih; auto.
Qed.


Lemma inv_im_in_dep_l_eq_nil :
  forall {T U:Type} (pfdu:type_eq_dec U) (l:list T) (f:fun_in_dep l U)
         (lu:list U),
    inv_im_in_dep_l pfdu l f lu = nil ->
    l = nil \/ lu = nil \/ ~incl (map_in_dep f) lu.
  intros T U h1 l f lu. revert f. revert l h1.
  induction lu as [|b lu ih]; simpl.
  intros; right; left; auto.
  intros l h1 f h2.
  apply app_eq_nil in h2.
  destruct h2 as [h2 h3].
  apply inv_im_in_dep_eq_nil in h2. 
  destruct h2 as [h2 | h2]. 
  left; auto. 
  destruct h2 as [a [h4 h5]].
  right; right.
  specialize (ih _ _ _ h3).
  destruct ih as [ih | [ih | ih]].
  pose proof h4 as h4'. rewrite ih in h4'.
  contradiction.
  subst.
  simpl in h3.
  intro h6.
  red in h6. 
  specialize (h6 _ (in_map_in_dep _ _ h4)).
  simpl in h6.
  destruct h6; subst.
  pose proof (h5 eq_refl); auto.
  contradiction.
  intro h7.
  red in h7.
  specialize (h7 _ (in_map_in_dep _ _ h4)).
  inversion h7 as [h10 | h10]; subst.
  contradict h5; auto.
  pose proof (in_inv_im_in_dep_l_iff h1 l f lu a) as h8.
  let P := type of h8 in match P with _ <-> ?pf => assert (h11:pf) end.
  exists h4; auto.
  rewrite <- h8 in h11.
  rewrite h3 in h11.
  contradiction.
Qed.


Definition fun_in_dep_list_minus_incl
           {T U:Type} {l:list T}
           (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
           (f:fun_in_dep l U) (lx: list T) (pf:incl lx l) :
  fun_in_dep (list_minus pfdt l (inv_im_in_dep_l pfdu l f (map_in_dep (fun_in_dep_incl _ _ f pf)))) U.
  do 2 red; do 2 red in f; intros x h1.
  pose proof (in_list_minus_impl' pfdt  _ _ _ h1) as h2.
  refine (f _ h2).
Defined.



Lemma fun_in_dep_list_minus_incl_compat :
  forall {T U:Type} {l:list T}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U) (la: list T) (pfa:incl la l) (x:T) 
         (pfx':In x  (list_minus pfdt l (inv_im_in_dep_l pfdu l f (map_in_dep (fun_in_dep_incl _ _ f pfa)))))
         (pfx:In x l),
         let f' := fun_in_dep_list_minus_incl pfdt pfdu f la pfa in
    f x pfx = f' x pfx'.
  intros T U l h1 h2 f la h3 x h4 h5; simpl.
  unfold fun_in_dep_list_minus_incl.
  f_equal. apply proof_irrelevance.
Qed.


Lemma incl_inv_im_in_dep_l :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l:list T) (la:list T) (f:fun_in_dep l U)
         (pfa:incl la l),
    incl la (inv_im_in_dep_l pfdu l f (map_in_dep f)).
  intros T U h1 l la f h2; red; intros x h3.
  rewrite in_inv_im_in_dep_l_iff.
  exists (h2 _ h3).
  rewrite in_map_in_dep_iff.
  exists x, (h2 _ h3); auto.
Qed.


Lemma list_to_set_map_in_dep_list_minus_incl :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (f:fun_in_dep l U)
         (lx:list T) (pf:incl lx l),
    let pf' := inv_im_in_dep_l_incl pfdu _ (map_in_dep (fun_in_dep_incl _ _ f pf)) f in
    let f' := fun_in_dep_incl _ _ f pf' in
    list_to_set (map_in_dep (fun_in_dep_list_minus_incl pfdt pfdu f lx pf)) =
    Setminus (list_to_set (map_in_dep f)) (list_to_set (map_in_dep f')).
  intros T U h1 h2 l f lx h3.
  simpl. 
  apply Extensionality_Ensembles; red; split; red; intros y h4.
  rewrite <- list_to_set_in_iff in h4.
  rewrite in_map_in_dep_iff in h4.
  destruct h4 as [x [h4 h5]]; subst.
  pose proof (in_list_minus_impl h1 _ _ _ h4) as h5.
  destruct h5 as [h5 h6].
  constructor.
  rewrite <- list_to_set_in_iff.
  unfold fun_in_dep_list_minus_incl.
  apply in_map_in_dep.
  intro h7.
  rewrite <- list_to_set_in_iff in h7.
  rewrite in_map_in_dep_iff in h7.
  destruct h7 as [x' [h7 h8]].
  unfold fun_in_dep_incl, fun_in_dep_list_minus_incl in h8.
  contradict h6.
  rewrite in_inv_im_in_dep_l_iff.
  exists h5.
  rewrite in_map_in_dep_iff.
  pose proof h7 as h7'.
  rewrite in_inv_im_in_dep_l_iff in h7'.
  destruct h7' as [h9 h10].
  rewrite in_map_in_dep_iff in h10.
  destruct h10 as [x'' [h11 h12]].
  exists x'', h11.
  rewrite h12.
  let P := type of h8 in match P with f x' ?pf = _ => assert (h13:pf = h9) end.
  apply proof_irrelevance.
  subst.
  rewrite h8.
  f_equal. apply proof_irrelevance.

  destruct h4 as [h4 h5].
  rewrite <- list_to_set_in_iff in h4.
  rewrite <- list_to_set_in_iff in h5.
  rewrite in_map_in_dep_iff in h5.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff in h4.
  destruct h4 as [x [h4 h6]]; subst.
  rewrite in_map_in_dep_iff.
  exists x.
  ex_goal.
  eapply impl_in_list_minus; auto.
  intro h6.
  contradict h5.
  exists x, h6.
  unfold fun_in_dep_incl.
  f_equal.
  apply proof_irrelevance.
  exists hex.
  unfold fun_in_dep_list_minus_incl.
  f_equal.
  apply proof_irrelevance.
Qed.
  


Lemma incl_map_in_dep_incl :
  forall {T U:Type} (l' l:list T) (f:fun_in_dep l U)
         (pf:incl l' l),
    incl (map_in_dep (fun_in_dep_incl l l' f pf))
          (map_in_dep f).
  intros T U l' l f h1; red; intros x h2.
  rewrite in_map_in_dep_iff in h2.
  rewrite in_map_in_dep_iff.
  destruct h2 as [x' [h2 h3]]; subst.
  exists x', (h1 _ h2); auto.
Qed.


Lemma in_match_in_dec_iff :
  forall {T U:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (A:In a l ->Ensemble U)
         (B: ~In a l->Ensemble U)
         (x:U),
    Inn match in_dec pfdt a l with
          | left pfa => A pfa
          | right pfb => B pfb
        end x <->
    (exists pfin : In a l, Inn (A pfin) x) \/
    (exists pfnin : ~In a l, Inn (B pfnin) x).
  intros T' U h1 l a A B  x.
  destruct (in_dec h1 a l) as [h2 | h2]; split; intros h3.
  left. exists h2; auto.
  destruct h3 as [h3 | h3]. destruct h3 as [h3 h4].
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
  destruct h3; contradiction.
  right. exists h2; auto.
  destruct h3 as [h3 | h3]. destruct h3; contradiction.
  destruct h3 as [h3 h4].
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
Qed.


    


Lemma list_decompose : 
  forall {T:Type} (l:list T) (def:T), 
    l <> nil -> l = (hd def l) :: (tl l).
intros T l.
induction l; auto.
intros def h1. contradict h1; auto.
Qed.

   
(*The last element (like [last]), and unlike [last], 
it requires no default value,
just a guarantee the list is non-empty.*)
Definition lst {T:Type} (l:list T) : l <> nil -> T :=
  match l with
    | nil  => fun pf => False_rect T (pf eq_refl)
    | a::l'  => fun pf => last l (hd a l)
  end.

Lemma lst_functional :
  forall {T:Type} (l l':list T) (pf:l <> nil) (pf':l' <> nil),
    l = l' ->
    lst l  pf =
    lst l' pf'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

    
Lemma lst_indep :
  forall {T:Type} (l:list T) (pf pf':l <> nil),
    lst l pf =
    lst l pf'.
  intros; f_equal; apply proof_irrelevance.
Qed.

  

Lemma lst_compat :
  forall {T:Type} (l:list T)
         (pf:l <> nil),
    let a := lst l pf in
    l = (removelast l)++(a::nil).
  intros T l.
  induction l as [|a l ih].
  intro h1; contradict h1; auto.
  intros h1. simpl.
  destruct l as [|c l].
  rewrite app_nil_l; auto.
  simpl. 
  f_equal.
  simpl in ih.
  specialize (ih (cons_neq_nil _ _)).
  rewrite ih.
  f_equal.
  destruct l; auto.
  erewrite last_indep.
  reflexivity.
  auto with datatypes.
Qed.

Lemma lst_compat' :
  forall {T:Type} (l:list T)
         (pf:l <> nil),
    lst l pf =
    last l (hd' l pf).
  intros ? l; destruct l; auto.
Qed.

Lemma last_compat :
  forall {T:Type} (l:list T) (def:T)
         (pf:l <> nil),
    last l def =
    lst l pf.
  intros; erewrite lst_functional.
  erewrite lst_compat' at 1.
  apply last_indep; auto.
  reflexivity.

  Grab Existential Variables.
  assumption.
Qed.
  
Lemma in_lst :
  forall {T:Type} (l:list T)
         (pf:l <> nil),
    In (lst l pf) l.
  intros T l.
  destruct l as [|a l].
  intro h1. contradict h1; auto.
  intro; apply in_last; auto.
Qed.

Lemma lst_tl :
  forall {T:Type} (l:list T)
         (pf: tl l <> nil),
    let pf' := tl_not_nil _ pf in
    lst (tl l) pf = lst l pf'.
  intros ? l; destruct l; simpl.
  intro h1; contradict h1; auto.
  intro.
  rewrite (hd_compat' _ pf) at 2.
  rewrite lst_compat'.
  apply last_indep; auto.
Qed.


(*The analogue of the standard library*)
Lemma app_removelast_lst :
  forall {T:Type} (l:list T) (pf:l <> nil),
    l = removelast l ++ (lst l pf) :: nil.
  intros T l.
  destruct l as [|a l]; simpl.
  intro h1.
  contradict h1; auto.
  intro h1.
  apply (app_removelast_last a (cons_neq_nil l a)).
Qed.


Lemma removelast_not_nil' :
  forall {T:Type} (l:list T),
    removelast l <> nil ->
    pred (pred (length l)) < length l.
  intros T l h1.
  rewrite <- length_removelast.
  pose proof (not_nil_lt _ h1) as h2.
  destruct l as [|a l].
  simpl in h1; auto.
  erewrite length_removelast.
  simpl.
  omega.
Qed.

          
Lemma removelast_eq_iff :
  forall {T:Type} (l l':list T)
         (pf:l <> nil),
    removelast l = l' <->
    l = l' ++ lst l pf::nil.
  intros T l l' h1; split; intro h2.
  rewrite <- h2.
  apply app_removelast_lst.
  rewrite h2.
  rewrite removelast_app_sing; auto.
Qed.


Definition map_proj1_sig {T} {P:T->Prop}
           (l:list {x:T | P x}) : list T :=
  map (@proj1_sig _ P) l.

Lemma in_map_proj1_sig_iff :
  forall {T} {P:T->Prop}
         (l:list {x:T | P x}) (x:T),
    In x (map_proj1_sig l) <->
    exists (pfx:P x),
      In (exist _ x pfx) l.
  unfold map_proj1_sig; intros T P l x; rewrite in_map_iff; split;
  intro h1.
  destruct h1 as [y [? hy]]; subst.
  destruct y as [y hp].
  simpl; exists hp; auto.
  destruct h1 as [h2 h3].
  exists (exist _ _ h2); simpl.
  split; auto.
Qed.


Lemma map_proj1_sig_incl :
  forall {T} (l':list T) (l:list {x:T | In x l'}),
    incl (map_proj1_sig l) l'.
  intros; red; intros x h1.
  rewrite (in_map_proj1_sig_iff (P := fun c => In c l')) in h1.
  destruct h1; auto.
Qed.


Definition proj1_sig_leq {T} {P Q:T->Prop}
           (l: list {x:T | P x}) 
           (l':list {x:T | Q x}) : Prop :=
  map_proj1_sig l =
  map_proj1_sig l'.

Lemma proj1_sig_leq_nil1 :
  forall {T} (P Q:T->Prop) 
         (l': list {x:T | P x}),
    proj1_sig_leq nil l' (P:=P) ->
    l' = nil.
  unfold proj1_sig_leq; intros ? ? ? l.
  destruct l; auto; simpl.
  intro; discriminate.
Qed.

Lemma proj1_sig_leq_nil2 :
  forall {T} (P Q:T->Prop) 
         (l: list {x:T | P x}),
    proj1_sig_leq l nil (Q:=Q)->
    l = nil.
  unfold proj1_sig_leq; intros ? ? ? l.
  destruct l; auto; simpl.
  intro; discriminate.
Qed.

Lemma proj1_sig_leq_cons :
  forall {T} {P Q:T->Prop}
         (l: list {x:T | P x}) 
         (l':list {x:T | Q x})
         (a:T)
         (pfp:P a)
         (pfq:Q a),
    proj1_sig_leq l l' ->
    proj1_sig_leq
      (exist _ _ pfp :: l)
      (exist _ _ pfq :: l').
  unfold proj1_sig_leq; intros; simpl; f_equal; auto.
Qed.

Lemma proj1_sig_leq_cons' :
  forall {T} {P Q:T->Prop}
         (l: list {x:T | P x}) 
         (l':list {x:T | Q x})
         (a:{x:T | P x}) 
         (b:{x:T | Q x}),
    proj1_sig_leq
      (a :: l)
      (b :: l') ->
    proj1_sig a = proj1_sig b /\
    proj1_sig_leq l l'.
  unfold proj1_sig_leq; intros ? ? ? ? ? ? ? h1; simpl in h1.
  inversion h1; auto.
Qed.



Definition map_sig {T:Type} (l:list T) : list {x:T | In x l} :=
  map_in_dep (fun x (pfx:In x l) => (exist (fun c => In c l) _ pfx)).

Lemma map_sig_nil :
  forall T,
    @map_sig T nil = nil.
  unfold map_sig; intros; apply map_in_dep_eq_nil_iff; auto.
Qed.

Lemma map_sig_in : forall {T:Type} (l:list T) (i:{x:T | In x l}), 
                     In i (map_sig l).
  unfold map_sig; intros T l i;
  rewrite in_map_in_dep_iff;
  exists (proj1_sig i), (proj2_sig i);
  destruct i; auto.
Qed.

Lemma length_map_sig :
  forall {T} (l:list T),
    length (map_sig l) = length l.
  unfold map_sig; intros; apply length_map_in_dep.
Qed.

Lemma in_map_sig : forall {T:Type} (l:list T) (x:{t:T | In t l}),
                     In x (map_sig l) -> In (proj1_sig x) l.
intros; apply proj2_sig.
Qed.

Lemma in_map_sig_iff :
  forall {T:Type} (l:list T) (x:{t:T | In t l}),
    In x (map_sig l) <-> In (proj1_sig x) l.
  intros; split; intro h1.
  apply in_map_sig; auto.
  unfold map_sig. 
  rewrite in_map_in_dep_iff.
  exists (proj1_sig x), (proj2_sig x).
  apply proj1_sig_injective; auto.
Qed.

Definition map_sig_incl {T} {l l':list T} (pfinc:incl l l') :
  list {x:T | In x l'} :=
  map_in_dep
    (fun x (pfin:In x l) =>
       exist (fun c => In c l') x (pfinc _ pfin)).

Lemma length_map_sig_incl :
  forall {T} {l l':list T} (pfinc:incl l l'),
    length (map_sig_incl pfinc) = length l.
  unfold map_sig_incl; intros; rewrite length_map_in_dep; auto.
Qed.


  
Definition map_sig' {T} {l:list T} {P:T->Prop}
           (Q:forall x, In x l -> P x) :
  list {x:T | P x} :=
  map_in_dep (fun x (pf:In x l) =>
                exist P x (Q x pf)).

Lemma length_map_sig' :
  forall {T} {l:list T} {P:T->Prop}
         (Q:forall x, In x l -> P x),
    length (map_sig' Q) =
    length l.
  unfold map_sig'; intros; rewrite length_map_in_dep; auto.
Qed.

Lemma in_map_sig_iff' :
  forall {T} {l:list T} {P:T->Prop}
         (Q:forall x, In x l -> P x)
         (y:{x:T | P x}),
    In y (map_sig' Q) <->
    In (proj1_sig y) l.
  unfold map_sig'; intros T l P Q y; split; intro h1.
  rewrite in_map_in_dep_iff in h1.
  destruct h1 as [? [? ?]]; subst; simpl.
  auto.
  rewrite in_map_in_dep_iff.
  exists (proj1_sig y), h1.
  apply proj1_sig_injective; auto.
Qed.


Definition map_sig_cons {T} (l:list T) a :
  list {x:T | In x (a::l)} :=
  map_sig' (fun x (pf:In x l) => in_cons _ _ _ pf).

Lemma in_map_sig_cons_iff :
  forall {T} (l:list T) a x,
    In x (map_sig_cons l a) <->
    In (proj1_sig x) l.
  unfold map_sig_cons; intros; rewrite in_map_sig_iff'; tauto.
Qed.

Lemma length_map_sig_cons :
  forall {T} (l:list T) a,
    length (map_sig_cons l a) =
    length l.
  unfold map_sig'; intros; apply length_map_sig'; auto.
Qed.

Definition map_sig_from_in_cons {T} {l:list T} {a}
           (pfa:In a l) 
           (l':list {x:T | In x (a::l)}) :
  list {x:T | In x l} :=
  map (fun x =>
         exist _ (proj1_sig x) (in_from_cons _ _ _ pfa (proj2_sig x))) l'.

Lemma in_map_sig_from_in_cons_iff :
  forall {T} {l:list T} {a}
         (pfa:In a l) 
         (l':list {x:T | In x (a::l)})
         (y:{x:T | In x l}),
    In y (map_sig_from_in_cons pfa l') <->
    In (exist (fun c => In c (a::l))
              (proj1_sig y) (in_cons _ _ _ (proj2_sig y)))
       l'.
  unfold map_sig_from_in_cons; intros; rewrite in_map_iff; split; intro h1.
  destruct h1 as [x [? h1]]; subst; simpl.
  eapply in_subst1.
  apply proj1_sig_injective; simpl.
  reflexivity.
  assumption.
  exists (exist (fun c => In c (a::l)) _ (in_cons _ _ _ (proj2_sig y))); simpl.
  split.
  apply proj1_sig_injective; auto.
  assumption.
Qed.

Lemma in_map_sig_from_in_cons_iff' :
  forall {T} {l:list T} {a}
         (pfa:In a l) 
         (l':list {x:T | In x (a::l)})
         (y:{x:T | In x l}),
    In y (map_sig_from_in_cons pfa l') <->
    In (proj1_sig y) (map_proj1_sig l').
  intros; rewrite in_map_sig_from_in_cons_iff,
          (in_map_proj1_sig_iff (P := fun c => In c (a::l)));
  destruct y; simpl; split; intro h1.
  exists (in_cons _ _ _ i); auto.
  destruct h1 as [? h2].
  erewrite f_equal_dep.
  apply h2.
  reflexivity.
Qed.

Definition map_sig_from_neq_cons {T} {l:list T} {a}
           (l':list {x:T | In x (a::l)})
           (pfn:forall y, In y l' -> proj1_sig y <> a) :
  list {x:T | In x l} :=
  map_in_dep
    (fun y (pfy:In y l') =>
       exist (fun c => In c l)
             (proj1_sig y)
             (in_cons_neq _ _ _ (proj2_sig y) 
                          (pfn y pfy))).


Lemma in_map_sig_from_neq_cons :
  forall {T} {l:list T} {a}
         (l':list {x:T | In x (a::l)})
         (pfn:forall y, In y l' -> proj1_sig y <> a)
         (y:{x:T | In x l}),
    In y (map_sig_from_neq_cons l' pfn) <->
    In (exist (fun c => In c (a::l)) (proj1_sig y)
              (in_cons _ _ _ (proj2_sig y))) l'.
  unfold map_sig_from_neq_cons; intros; rewrite in_map_in_dep_iff;
  split; intro h1.
  destruct h1 as [x [h1 h2]]; subst; simpl.
  eapply in_subst1.
  apply proj1_sig_injective; simpl; reflexivity.
  assumption.
  destruct y as [y hy]; simpl in h1.
  exists (exist (fun c => In c (a::l)) _ (in_cons _ _ _ hy)), h1;
    apply proj1_sig_injective; simpl; auto.
Qed.


Lemma in_map_sig_from_neq_cons':
  forall {T} {l:list T} {a}
         (l':list {x:T | In x (a::l)})
         (pfn:forall y, In y l' -> proj1_sig y <> a)
         (y:{x:T | In x l}),
    In y (map_sig_from_neq_cons l' pfn) <->
    In (proj1_sig y) (map_proj1_sig l').
  intros; rewrite in_map_sig_from_neq_cons,
          (in_map_proj1_sig_iff (P := fun c => In c (a::l)));
  destruct y; 
  split; intro h1; simpl in h1; simpl.
  exists (in_cons _ _ _ i); auto.
  destruct h1 as [h1 h2].
  erewrite f_equal_dep.
  apply h2.
  reflexivity.
Qed.

Lemma in_from_sig_cons_remove :
  forall {T} {l:list T} {a}
         (pfdt:type_eq_dec {x:T | In x (a::l)})
         (l':list {x:T | In x (a::l)})
         (x:{t:T | In t (a::l)}),
    In x (remove pfdt (exist (fun c => In c (a::l)) a (in_eq _ _)) l') ->
    In (proj1_sig x) l.
  intros T l a hdt l' x h2.
  rewrite in_remove_iff in h2.
  destruct h2 as [h2 h3].
  destruct x as [x hx].
  simpl.
  destruct hx as [|hx]; subst; auto.
  contradict h3.
  f_equal; apply proof_irrelevance.
Qed.


Definition list_from_sig_cons_remove {T} {l:list T} {a}
           (pfdt:type_eq_dec T) 
           (l':list {x:T | In x (a::l)}) :
  list {x:T | In x l} :=
  map_in_dep
    (fun c (pfc:In c (remove (sig_eq_dec pfdt (fun c => In c (a::l)))
                            (exist (fun c => In c (a::l)) a (in_eq _ _))
                            l')) =>
                   exist _
                         (proj1_sig c) 
                         (in_from_sig_cons_remove _ l' c pfc)).

Lemma in_list_from_sig_cons_remove_iff :
  forall {T} {l:list T} {a}
         (pfdt:type_eq_dec T) 
         (l':list {x:T | In x (a::l)})
         (y:{x:T | In x l}),
    In y (list_from_sig_cons_remove pfdt l') <->
    In (exist (fun c => (In c (a::l))) (proj1_sig y)
              (in_cons _ _ _ (proj2_sig y))) l' /\
    (proj1_sig y <> a).
  unfold list_from_sig_cons_remove; intros;
  rewrite in_map_in_dep_iff; split; intro h1.
  destruct h1 as [x [h1 ?]]; subst; simpl.
  pose proof h1 as h1'.
  rewrite in_remove_iff in h1'.
  destruct h1' as [h2 h3].
  split.
  eapply in_subst1.
  apply proj1_sig_injective; simpl; reflexivity.
  assumption.
  destruct x; simpl; contradict h3; subst; f_equal;
  apply proof_irrelevance.
  destruct h1 as [h1 h2].
  exists (exist (fun c => (In c (a::l))) (proj1_sig y)
                (in_cons _ _ _ (proj2_sig y))).
  ex_goal.
  rewrite in_remove_iff.
  split; auto.
  intro h3.
  apply exist_injective in h3; subst.
  contradict h2; auto.
  exists hex.
  simpl.
  destruct y; simpl; f_equal; apply proof_irrelevance.
Qed.


(*This is an induction principle to tie in the left-recursive/
right-recursive incompatibility of induction on [nat] versus 
induction on [list]*)
Lemma list_nat_ind : 
  forall {T} (P:list T->nat->Prop) (l:list T) n,
    length l = n ->
    P nil 0 ->
    (forall a l' n',
       length l' = n' ->
       P l' n' ->
       P (l' ++ a::nil) (S n')) ->
    P l n. 
  intros T P l n. revert l P.
  induction n as [|n ih]; simpl.
  intros l P h1.
  apply length_eq0 in h1; subst; auto.
  intros l P h1 h2 h3.
  pose proof (lt_eq_trans _ _ _ (lt_0_Sn n) (eq_sym h1)) as h4.
  apply lt_length in h4.
  rewrite (app_removelast_lst _ h4) in h1 at 1.
  rewrite app_length, S_compat in h1.
  apply S_inj in h1; subst. 
  rewrite length_removelast.
  rewrite <- (S_pred _ _ (not_nil_lt _ h4)).
  rewrite (app_removelast_lst _ h4), app_length; simpl; rewrite S_compat. 
  apply h3; auto.
  apply ih; auto.
Qed.


  
Lemma im_full_set_segT_eq :
  forall {T:Type} {n:nat} (f:n ---> T) (def:T),
         Im (Full_set (n @)) f =
         list_to_set (map (segT_fun_to_nat_fun f def) (seg_list n)).
  intros T n f def.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_iff.
  exists (proj1_sig x).
  split.
  erewrite segT_fun_to_nat_fun_compat.
  f_equal.
  pose proof (unfold_sig _ x) as h2.
  symmetry in h2.
  apply h2.
  rewrite list_to_set_in_iff.
  rewrite list_to_set_seg_list_eq.
  constructor.
  apply proj2_sig.
  rewrite <- list_to_set_in_iff in h1.
  rewrite in_map_iff in h1.
  destruct h1 as [m h1].
  destruct h1 as [h1 h2].
  rewrite in_seg_list_iff in h2.
  apply Im_intro with (exist (fun a => a < n) _ h2).
  constructor.
  subst.
  rewrite (segT_fun_to_nat_fun_compat _ _ _ h2).
  reflexivity.
Qed. 


Lemma in_list_power1 : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1 * T2)), In pl (list_power l1 l2) -> forall (pr:T1*T2),
  In pr pl -> In (fst pr) l1.
intros T1 T2 l1 l2. 
induction l1 as [|b l1 h4]. 
intros pl h1. 
simpl in h1.
destruct h1 as [h1l | h1r].
subst. contradiction. contradiction.
intros pl h1.
simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [pl' h5].
destruct h5 as [h5l h5r].
rewrite in_map_iff in h5r.
destruct h5r as [t2 h6].
destruct h6 as [h6l h6r].
specialize (h4 pl' h5l).
intros pr h7.
subst.
destruct h7 as [h7l | h7r].
subst. simpl. left. reflexivity.
specialize (h4 pr h7r).
right.
assumption.
Qed.

Lemma in_list_power2 : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1 * T2)), In pl (list_power l1 l2) -> forall (pr:T1*T2),
  In pr pl -> In (snd pr) l2.
intros T1 T2 l1 l2.
induction l1 as [|b l1 h4]. 
intros pl h1. 
simpl in h1.
intros pr h2.
destruct h1 as [h1l | h1r].
subst. contradiction. contradiction.
intros pl h2.
simpl in h2.
rewrite in_flat_map in h2.
destruct h2 as [pl' h3].
destruct h3 as [h3l h3r].
rewrite in_map_iff in h3r.
destruct h3r as [t2 h5].
destruct h5 as [h5l h5r].
subst.
intros pr h6.
destruct h6 as [h6l | h6r].
subst. simpl. assumption.
specialize (h4 _ h3l _ h6r).
assumption.
Qed.



Lemma no_dup_fun_well_defined_fst : 
  forall {T U:Type} (S:Ensemble (T*U)) (l:list (T*U)),
    S = list_to_set l -> NoDup l ->fun_well_defined S -> NoDup (map (@fst _ _) l).
intros T U S l h0 h1 h2. generalize dependent S. revert h1.
induction l as [|pr l h1].
simpl. intros; constructor.
intros h2 S h3 h4.
simpl.
constructor.
assert (h5:Inn S pr). rewrite h3. rewrite <- list_to_set_in_iff. left. reflexivity.
intro h6. rewrite in_map_iff in h6. destruct h6 as [pr' h6].
destruct h6 as [h6l h6r].
assert (h6:Inn S pr'). rewrite h3. rewrite <- list_to_set_in_iff. right. assumption.
rewrite surjective_pairing in h5. rewrite surjective_pairing in h6. rewrite h6l in h6.
pose proof (fp_functional h4 _ _ _ h5 h6) as h7.
assert (h8:pr = pr'). apply injective_projections; auto.
subst.
pose proof (no_dup_cons_nin _ _ h2).
contradiction.
simpl in h3.
pose proof (f_equal (fun S=>Subtract S pr) h3) as h5.
simpl in h5.
rewrite sub_add_same_nin in h5.
pose proof (no_dup_cons _ _ h2) as h6.
pose proof (incl_subtract  S pr) as h7.
pose proof (fun_well_defined_incl _ _ h7 h4) as h8.
specialize (h1 h6 _ h5 h8).
assumption.
intro h6.
rewrite <- list_to_set_in_iff in h6.
pose proof (no_dup_cons_nin _ _ h2).
contradiction.
Qed.



Inductive functionally_paired_l {T U:Type} (la:list T)
          (lb:list U) (lp:list (T*U)) : Prop :=
  functional_pairs_intro :
    (forall (x:T),
      In x la -> (exists! y:U,
                  In y lb /\ In (x, y) lp)) ->
    (forall (pr:T*U), In pr lp -> In (fst pr) la /\
    In (snd pr) lb) ->
                  functionally_paired_l la lb lp.

Lemma fpl_in_dom :
  forall {T U:Type} {la:list T}
         {lb:list U} {lp:list (T*U)},
    functionally_paired_l la lb lp ->
    forall pr:T*U, In pr lp -> In (fst pr) la.
intros T U la lb lp h1 pr h2.
destruct h1 as [h1a h1b].
specialize (h1b _ h2).
destruct h1b; assumption.
Qed.

Lemma fpl_in_ran :
  forall {T U:Type} {la:list T}
         {lb:list U} {lp:list (T*U)},
    functionally_paired_l la lb lp ->
    forall pr:T*U, In pr lp -> In (snd pr) lb.
intros T U la lb lp h1 pr h2.
destruct h1 as [h1a h1b].
specialize (h1b _ h2).
destruct h1b; assumption.
Qed.

Lemma fpl_functional :
  forall {T U:Type} {la:list T} {lb:list U}
         {lp:list (T*U)},
    functionally_paired_l la lb lp ->
    forall (x:T) (y1 y2:U), In (x, y1) lp ->
                            In (x, y2) lp ->
                            y1 = y2.
intros T U la lb lp h1 x y1 y2 h2 h3.
inversion h1 as [h4 h5].
pose proof (h5 _ h2) as h6.
pose proof (h5 _ h3) as h7.
simpl in h6. simpl in h7.
destruct h6 as [h6a h6b].
destruct h7 as [h7a h7b].
pose proof (h4 _ h6a) as h8.
destruct h8 as [y h9].
red in h9.
destruct h9 as [h9a h9b].
pose proof (h9b _ (conj h6b h2)).
pose proof (h9b _ (conj h7b h3)).
subst.
assumption.
Qed.

Lemma fpl_empty1 : forall (T U:Type) (lb:list U),
                        functionally_paired_l (@nil T) lb nil. 
intros; constructor; intros; contradiction.
Qed.

Lemma fpl_empty1_s : forall {T U:Type} (lb:list U) (lp:list (T*U)),
                      functionally_paired_l (@nil T) lb lp ->
                      lp = nil.
intros T U lb lp h1.
destruct h1 as [h2 h3].
destruct lp as [|pr lp].
reflexivity.
pose proof (in_eq pr lp) as h4.
specialize (h3 _ h4).
destruct h3; contradiction.
Qed.

Lemma no_fpl_empty2 : forall {T U:Type} (la:list T)
                            (lp:list (T*U)),
                        functionally_paired_l la (@nil U) lp ->
                        la = nil.
intros T U la.
destruct la as [|a la].
intros; reflexivity.
intros lp h2.
destruct h2 as [h2 h3].
pose proof (in_eq a la) as h4.
specialize (h2 _ h4).
destruct h2 as [y h2].
red in h2.
destruct h2 as [[]]; contradiction.
Qed.

Lemma no_fpl_empty2' : forall {T U:Type} (la:list T)
                            (lp:list (T*U)),
                        functionally_paired_l la (@nil U) lp ->
                        lp = nil.
intros T U la lp.
destruct lp as [|pr lp].
intros; reflexivity.
intro h1.
pose proof (no_fpl_empty2 _ _ h1).
subst.
pose proof (fpl_empty1_s _ _ h1).
discriminate.
Qed.



Lemma fp_fpl_compat : forall {T U:Type} (A:Ensemble T) (B:Ensemble U)
                             (S:Ensemble (T*U)) 
                             (la:list T) (lb:list U) (lp:list (T*U)),
                        list_to_set la = A ->
                        list_to_set lb = B ->
                        list_to_set lp = S ->
                        ((functionally_paired A B S) <->
                        (functionally_paired_l la lb lp)).
intros T U A B S la lb lp h1 h2 h3.
split.
intro h4.
destruct h4 as [h5 h6].
constructor. 
intros x h7.
rewrite list_to_set_in_iff in h7.
rewrite h1 in h7.
specialize (h5 _ h7).
destruct h5 as [y h5].
exists y.
red in h5. red.
destruct h5 as [h5l h5r].
rewrite <- h2 in h5l. rewrite <- h3 in h5l.
do 2 rewrite <- list_to_set_in_iff in h5l.
split; try assumption.
intros b h8.
do 2 rewrite list_to_set_in_iff in h8.
rewrite h2 in h8. rewrite h3 in h8.
specialize (h5r _ h8). assumption.
intros pr h7.
rewrite list_to_set_in_iff in h7.
rewrite h3 in h7.
specialize (h6 _ h7).
rewrite  <- h1 in h6.  rewrite <- h2 in h6.
do 2 rewrite <- list_to_set_in_iff in h6.
assumption.
intro h4.
destruct h4 as [h4 h5].
constructor.
intros x h6.
rewrite <- h1 in h6.
rewrite <- list_to_set_in_iff in h6.
specialize (h4 _ h6).
destruct h4 as [y h4].
exists y.
red in h4. red.
destruct h4 as [h4l h4r].
split.
do 2 rewrite list_to_set_in_iff in h4l.
rewrite  h2 in h4l. rewrite h3 in h4l.
assumption.
intros b h7.
rewrite <- h2 in h7. rewrite <- h3 in h7.
do 2 rewrite <- list_to_set_in_iff in h7.
apply h4r; assumption.
intros pr h6.
rewrite <- h3 in h6.
rewrite <- list_to_set_in_iff in h6.
specialize (h5 _ h6).
do 2 rewrite list_to_set_in_iff in h5.
rewrite h1 in h5. rewrite h2 in h5.
assumption.
Qed.



Lemma fpl_to_fun_in_dep_ex :
  forall {T U:Type} {la:list T} {lb:list U}
         {lp:list (T*U)},
    functionally_paired_l la lb lp ->
    exists! f:fun_in_dep la U,
      forall (a:T) (pf:In a la),
        In (a, f a pf) lp.
  intros T U la lb lp h1.
  inversion h1 as [h0 h2].
  pose (fun (x:T) (pf:In x la) => proj1_sig (constructive_definite_description _ (h0 x pf))) as f.
  exists f.
  unfold f.
  red. split.
  intros a h3.  
  destruct constructive_definite_description as [u h4].
  destruct h4 as [h4 h5].
  simpl; auto.
  intros f' h3.
  apply fun_in_dep_ext.
  intros a h4.
  destruct constructive_definite_description as [u h5].
  simpl.
  specialize (h3 _ h4).
  destruct h5 as [h5 h6]. 
  pose proof h0 as h0'.
  specialize (h0' _ h4). 
  destruct h0' as [u' h7].
  red in h7.
  destruct h7 as [h7 h8]. destruct h7 as [h7 h9].
  pose proof (h8 u (conj h5 h6)) as h10.
  subst.
  apply h8.
  split; auto.
  apply (fpl_in_ran  h1 _ h3).
Qed.

Definition fpl_to_fun_in_dep {T U:Type}
           {la:list T} {lb:list U} {lp:list (T*U)}
           (pf:functionally_paired_l la lb lp) : fun_in_dep la U :=
  proj1_sig
    (constructive_definite_description
       _ (fpl_to_fun_in_dep_ex pf)).

Lemma fpl_to_fun_in_dep_compat :
  forall {T U:Type}
         {la:list T} {lb:list U} {lp:list (T*U)}
         (pf:functionally_paired_l la lb lp),
    let f := fpl_to_fun_in_dep pf in
    forall (a:T) (pf:In a la),
        In (a, f a pf) lp.
  unfold fpl_to_fun_in_dep; intros; destruct constructive_definite_description; auto.
Qed.  

Lemma fpl_to_fun_in_dep_in_eq :
  forall {T U:Type} (la:list T) (lb:list U) (lp:list (T*U))
         (pffp:functionally_paired_l la lb lp)
         (pr:T*U) (pfin:In pr lp),
    let f := fpl_to_fun_in_dep pffp in
    let pfina := fpl_in_dom pffp pr pfin in
      snd pr = f (fst pr) pfina.
  intros T U la lb lp h1 pr h2.
  simpl.
  symmetry. gen0. intro h3.
  pose proof (fpl_to_fun_in_dep_compat h1 _ h3) as h4.
  rewrite (surjective_pairing pr) in h2.
  eapply fpl_functional. apply h1. apply h4. assumption.
Qed.
  
Definition im1l 
           {T U V:Type} {la:list T}
           {lb:list U} {lc:list V} {lp:list (T*U*V)}
           (pf:functionally_paired_l (list_prod la lb) lc lp) 
           (x:T) (pfx:In x la) : list V :=
  let f := fpl_to_fun_in_dep pf in
  map_in_dep (fun (y:U) (pfy:In y lb) => f (x, y) (in_prod _ _ _  _  pfx pfy)).


Definition im2l
           {T U V:Type} {la:list T}
           {lb:list U} {lc:list V} {lp:list (T*U*V)}
           (pf:functionally_paired_l (list_prod la lb) lc lp) 
           (y:U) (pfy:In y lb) : list V :=
  let f := fpl_to_fun_in_dep pf in
  map_in_dep (fun (x:T) (pfx:In x la) => f (x, y) (in_prod _ _ _  _  pfx pfy)).



Lemma no_dup_sing : 
  forall {T:Type} (a:T),
    NoDup  (a::nil).
  intros; constructor.
  intro; contradiction.
  constructor.
Qed.

  
Lemma no_dup_comm_cons :
  forall {T:Type} (l:list T) (a b:T),
    NoDup (a::b::l) ->NoDup (b::a::l).
  intros T l a b h1.
  pose proof (no_dup_cons _ _ h1) as h2. 
  pose proof (no_dup_cons _ _ h2) as h3.
  pose proof (no_dup_cons_nin _ _ h1) as h4.
  constructor.
  contradict h4.
  destruct h4 as [|h4]; subst.
  left; auto.
  pose proof (no_dup_cons_nin _ _ h2); contradiction.
  constructor.
  contradict h4.
  right; auto.
  assumption.
Qed.


Lemma no_dup_remove_singleton : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    NoDup l ->
    remove pfdt a l = nil ->
    l = nil \/ l = a::nil.
intros T h0 l.
induction l as [|a l h1]; simpl; auto.
intros x h2 h3.
right.
destruct (h0 x a) as [h4 | h5].
subst.
pose proof (no_dup_cons _ _ h2) as h4.
specialize (h1 _ h4 h3).
destruct h1 as [h1 | h1].
subst; auto.
rewrite h1 in h2.
inversion h2.
contradict H1. left. reflexivity.
discriminate h3.
Qed.


Lemma no_dup_remove : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    NoDup l -> NoDup (remove pfdt x l).
intros T h0 l.
induction l as [|a l h1]; simpl; auto.
intros x h2.
destruct (h0 x a) as [h3 | h4]. subst.
apply no_dup_cons in h2.
specialize (h1 a h2). assumption.
pose proof (no_dup_cons _ _ h2) as h5. 
pose proof (no_dup_cons_nin _ _ h2) as h6. clear h2.
specialize (h1 x h5).
constructor; auto.
apply nin_nin_remove.
assumption.
Qed.


Lemma no_dup_tl : forall {T:Type} (l:list T),
                      NoDup l -> NoDup (tl l).
intros T l.
induction l as [|a l h1].
simpl. auto.
intro h2.
simpl.
apply no_dup_cons with a.
assumption.
Qed.

  
Lemma count_occ_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup l ->
    forall x:T,
      In x l ->
      count_occ pfdt l x = 1.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros h2 x h3.
  destruct h3 as [|h3]; subst.
  destruct (h1 x x) as [h3 | h3].
  f_equal.
  pose proof (no_dup_cons_nin _ _ h2) as h4.
  rewrite (count_occ_In h1) in h4; omega.
  contradict h3; auto.
  destruct (h1 a x) as [h4 | h4]; subst.
  f_equal.
  pose proof (no_dup_cons_nin _ _ h2) as h4.
  rewrite (count_occ_In h1) in h4; omega.
  apply ih; auto.
  eapply no_dup_cons.
  apply h2.
Qed.


Lemma count_occ_map_inj :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:T->U) (l:list T) (x:T)
         (pfx:In x l),
    Injective f ->
    count_occ pfdu (map f l) (f x) =
    count_occ pfdt l x.
  intros T U h1 h2 f l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h3 h4.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  apply h4 in hl; subst.
  erewrite eq_dec_same at 1. 
  f_equal.
  destruct (in_dec h1 x l) as [h5 | h5].
  apply ih; auto.
  pose proof (count_occ_In h1 l x) as h6.
  rewrite h6 in h5.
  pose proof (count_occ_In h2 (map f l) (f x)) as h7.
  apply ngt_nlt in h5.
  rewrite not_lt0_iff in h5.
  rewrite h5.
  rewrite <- not_lt0_iff.
  apply ngt_nlt.
  rewrite <- (count_occ_In h2).
  intro h8.
  rewrite in_map_iff in h8.
  destruct h8 as [x' [h8 h9]].
  apply h4 in h8; subst.
  rewrite h6 in h9.
  omega.
  assumption.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr']; subst.
  contradict (hr eq_refl).
  destruct h3 as [|h3]; try contradiction.
  apply ih; auto.
Qed.


Lemma nin_hd_tl_no_dup :
  forall {T:Type} (l:list T) (def:T),
    NoDup l -> 
    ~In (hd def l) (tl l).
  intros T l. 
  induction l as [|a l h1]; simpl; auto.
  intros ? h2 h3.
  pose proof (no_dup_cons_nin _ _ h2) as h5.
  contradiction.
Qed.

Lemma nin_hd_tl_no_dup' :
  forall {T:Type} (l:list T) (pf:l <> nil),
    NoDup l -> 
    ~In (hd' l pf) (tl l).
  intros T l h1.
  pose proof (not_nil_cons l h1) as h2.
  destruct h2 as [a [l' h2]]; subst.
  intro h2.
  simpl.
  apply no_dup_cons_nin; auto.
Qed.

Definition no_dup_faml {T:Type} (ll:faml T) : Prop :=
  forall l:list T,
    In l ll ->
    NoDup l.


Inductive dup {T:Type} : list T -> Prop :=
| dup_intro_in : forall (x:T) (l:list T), In x l -> dup (x::l)
| dup_intro_cons: forall (a:T) (l:list T), dup l -> dup (a::l).

Lemma not_dup_nil (T:Type):
  ~ (dup (@nil T)).
  intro h1; inversion h1.
Qed.

Lemma not_dup_sing :
  forall {T:Type} (a:T),
    ~dup (a::nil).
  intros T a h1.
  inversion h1 as [x l h2 h3 | x l h2 h3]; subst.
  red in h2; auto.
  inversion h2.
Qed.

Lemma not_dup_cons :
  forall {T:Type} (l:list T) (a:T),
    ~dup (a::l) -> ~ dup l.
  intros T l a h1 h2.
  contradict h1.
  eapply dup_intro_cons; auto.
Qed.

Lemma not_dup_cons_nin :
  forall {T:Type} (l:list T) (a:T),
    ~dup (a::l) -> ~In a l.
  intros T l a h1 h2.
  contradict h1.
  constructor; auto.
Qed.



Lemma no_dup_iff :
  forall {T:Type} (l:list T),
    NoDup l <-> ~ dup l.
  intros T l.
  induction l as [|a l ih].
  split.
  intros h1 h2. inversion h2.
  intro; constructor.
  split.
  intros h1 h2.
  inversion h2 as [x l' h3 h4|]; subst.
  pose proof (no_dup_cons_nin _ _ h1) as h4.
  contradiction.
  pose proof (no_dup_cons _ _ h1) as h3.
  rewrite ih in h3.
  contradiction.
  intro h1. 
  pose proof (not_dup_cons _ _ h1) as h2.
  pose proof (not_dup_cons_nin _ _ h1) as h3.
  constructor; auto.
  rewrite ih; auto.
Qed.

Lemma dup_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    {dup l} + {NoDup l}.
  intros T h0 l.
  induction l as [|a l ih].
  right. constructor.
  destruct ih as [ih | ih]. 
  left.
  apply dup_intro_cons; auto.
  destruct (in_dec h0 a l) as [h1 | h1].
  left.
  constructor; auto.
  right.
  constructor; auto.
Qed.



Lemma dup_iff :
  forall {T:Type} (pf:type_eq_dec T) (l:list T),
    dup l <-> ~ NoDup l.
  intros T h0 l. 
  induction l as [|a l ih]; simpl.
  split.
  intro h1. inversion h1.
  intro h1. contradict h1. constructor.
  split.
  intro h1. 
  inversion h1 as [x l' h2 h3| xl' h2 h3]; subst.
  intro h3.  
  pose proof (no_dup_cons_nin _ _ h3) as h4.
  contradiction.
  rewrite ih in h3.
  intro h4. contradict h3.
  eapply no_dup_cons. apply h4. 
  intro h1.
  destruct (dup_dec h0 (a::l)); auto.
  contradiction.
Qed.


  
Lemma dup_cons_iff :
  forall {T:Type} (l:list T) (a:T),
    dup (a::l) <->
    In a l \/ dup l.
  intros T l h1. split.
  intro h2.
  inversion h2; subst.
  left; auto.
  right; auto.
  intro h2.
  destruct h2 as [h2 | h2]; simpl.
  left; auto.
  right; auto.
Qed.

Lemma dup_cons :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T),
    dup (a::l) ->
    {In a l} + {dup l}.
  intros T h0 l a h1.
  destruct (dup_dec h0 l) as [h2 | h2].
  right; auto.
  left.
  rewrite dup_cons_iff in h1.
  destruct h1 as [|h1]; auto.
  rewrite dup_iff in h1. contradiction.
  assumption.
Qed.

   

Lemma dup_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    dup (remove pfdt a l) ->
    dup l.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros a' h2.
  destruct (h1 a' a) as [h3 | h3]; subst.
  rewrite dup_cons_iff.
  apply ih in h2. right; auto.
  rewrite dup_cons_iff in h2.
  destruct h2 as [h2 | h2].
  left.
  rewrite in_remove_iff in h2.
  destruct h2; auto.
  apply ih in h2.
  right; auto.
Qed.

Lemma dup_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (l':list T),
    dup (list_minus pfdt l l') ->
    dup l.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros a' h2.
  destruct (in_dec h1 a a') as [h3 | h3].
  right.
  eapply ih. apply h2.
  rewrite dup_cons_iff in h2.
  destruct h2 as [h2 | h2].
  apply in_list_minus_impl in h2.
  destruct h2 as [h2 h4].
  left. assumption.
  right.
  eapply ih.
  apply h2.
Qed.


Lemma list_to_set_eq_sing :
  forall {T:Type} (l:list T) (a:T),
    NoDup l ->
    list_to_set l = Singleton a ->
    l = a::nil.
  intros T l.
  induction l as [|c l h1]; simpl.
  intros a h1 h2.
  symmetry in h2.
  apply sing_not_empty in h2. contradiction.
  intros a h2 h3. symmetry in h3.
  apply sing_eq_add in h3.
  destruct h3 as [h3 h4]. subst.
  destruct h4 as [h4 | h4].
  apply empty_set_nil in h4. subst; auto.
  pose proof (no_dup_cons _ _ h2) as h3.
  specialize (h1 _ h3 h4). subst.
  inversion h2.
  contradict H1. left; auto.
Qed.


Lemma list_to_set_eq_couple :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a b:T),
    NoDup l ->
    list_to_set l = Couple a b ->
    a = b \/ l = a::b::nil \/ l = b::a::nil.
  intros T hdt l a b h1 h2.
  destruct l as [|c l].
  simpl in h2.
  symmetry in h2.
  pose proof (inh_couple a b) as h3.
  rewrite h2 in h3.
  destruct h3; contradiction.
  pose proof (Couple_l _ a b) as h3.
  rewrite <- h2 in h3.
  rewrite <- list_to_set_in_iff in h3.
  destruct h3 as [|h3]; subst.
  pose proof (Couple_r _ a b) as h3.
  rewrite <- h2 in h3.
  rewrite <- list_to_set_in_iff in h3.
  destruct h3 as [|h3]; subst.
  left; auto.
  right.
  left.
  f_equal.
  pose proof (no_dup_cons_nin _ _ h1) as h4.
  simpl in h2.
  rewrite couple_comm, couple_add_sing in h2.
  apply add_nin_inj_eq in h2.
  apply list_to_set_eq_sing in h2; auto.
  apply no_dup_cons in h1; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite in_sing_iff; intro; subst; contradiction.
  pose proof (Couple_r _ a b) as h4. 
  simpl in h2. 
  rewrite <- h2 in h4. 
  destruct h4 as [b h4 | b h4].
  rewrite <- list_to_set_in_iff in h4.
  assert (h5:c <> a).
    intro; subst.
    apply no_dup_cons_nin in h1; contradiction.
  assert (h6:c <> b).
    intro; subst.
    apply no_dup_cons_nin in h1; contradiction.
  pose proof (Add_intro2 _ (list_to_set l) c) as h7.
  rewrite h2 in h7.
  destruct h7.
  contradict h5; auto.
  contradict h6; auto.
  destruct h4.
  pose proof (no_dup_cons_nin _ _ h1) as h4.
  rewrite couple_add_sing in h2.
  apply add_nin_inj_eq in h2.
  apply list_to_set_eq_sing in h2; subst.
  right; right; auto.
  apply no_dup_cons in h1.
  assumption.
  rewrite <- list_to_set_in_iff; auto.
  rewrite in_sing_iff; intro; subst; contradiction.
Qed.


Lemma lt_length_removelast :
  forall {T:Type} (l:list T) (m:nat),
    S m < length l ->
    m < length (removelast l).
  intros T l m h1.
  pose proof (lt_length _ _ h1) as h2. 
  pose proof (lst_compat _ h2) as h3. simpl in h3.
  rewrite h3 in h1. simpl in h1.
  rewrite app_length in h1.
  simpl in h1.
  omega.
Qed.


Lemma le_length_removelast :
  forall {T:Type} (l:list T) (m:nat),
    S m <= length l ->
    m <= length (removelast l).
  intros T l m h1.
  pose proof (lt_length _ _ h1) as h2. 
  pose proof (lst_compat _ h2) as h3. simpl in h3.
  rewrite h3 in h1. simpl in h1.
  rewrite app_length in h1.
  simpl in h1.
  omega.
Qed.

Lemma S_length_remove_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    NoDup l -> In x l ->
    S (length (remove pfdt x l)) =
    length l.
  intros T h0 l.
  induction l as [|a l h1]. intros; contradiction.
  intros x h2 h3.
  pose proof (no_dup_cons_nin _ _ h2) as h4.
  simpl in h3. 
  destruct h3 as [h3 | h3].  subst.
  simpl.
  destruct (h0 x x) as [h3 | h3].
  f_equal.
  rewrite <- remove_not_in'; auto.
  contradict h3; auto.
  pose proof (no_dup_cons _ _ h2) as h5.
  specialize (h1 _ h5 h3).
  simpl.
  f_equal.
  destruct (h0 x a) as [h6 | h6]. subst.
  contradiction.
  simpl.
  assumption.
Qed.

 

Lemma length_remove_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    NoDup l -> In a l ->
    length (remove pfdt a l) = pred (length l).
  intros T h0 l a h1 h2.
  assert (h3:l <> nil).
    intro; subst. simpl in h2. contradiction.
  pose proof (list_decompose _ a h3) as h4.
  rewrite h4.
  simpl.
  destruct (h0 a (hd a l)) as [h5 | h5].
  rewrite h5 at 1.
  pose proof (nin_hd_tl_no_dup _ a h1) as h6.
  rewrite <- (remove_not_in' h0 (tl l) (hd a l)); auto.
  simpl.
  assert (h6:In a (tl l)).
    rewrite h4 in h2.
    simpl in h2. destruct h2 as [h2 | h2].
    symmetry in h2. contradiction.
    assumption.
  apply S_length_remove_no_dup; auto.
  rewrite h4 in h1.
  apply (no_dup_cons _ _ h1).
Qed.


Lemma incl_no_dup_le :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    NoDup l ->
    NoDup l' ->
    incl l l' ->
    length l <= length l'.
  intros T hdt l.
  induction l as [|a l ih]; simpl; auto with arith.
  intros l' h1 h2 h3.
  destruct l' as [|b l'].
  apply incl_nil' in h3.
  discriminate. 
  simpl.
  apply le_n_S.
  destruct (hdt a b) as [|h4]; subst.
  apply incl_cons2_same in h3.
  apply ih.
  apply no_dup_cons in h1.
  apply no_dup_cons in h2.
  assumption. 
  apply no_dup_cons in h2; auto.
  assumption. 
  apply no_dup_cons_nin in h1; auto.
  apply no_dup_cons_nin in h2; auto.
  specialize (ih (remove hdt a (b::l')) (no_dup_cons _ _ h1)).
  hfirst ih.
  apply no_dup_remove; auto.
  specialize (ih hf).
  hfirst ih.
  red in h3; red.
  intros x h6.
  specialize (h3 _ (in_cons _ _ _ h6)).
  destruct h3 as [|h3]; subst.
  rewrite in_remove_iff.
  split.
  left; auto.
  auto.
  rewrite in_remove_iff.
  split.
  right; auto.
  intro; subst.
  apply no_dup_cons_nin in h1.
  contradiction.
  specialize (ih hf0).
  rewrite length_remove_no_dup in ih.
  simpl in ih; auto.
  assumption.
  right.
  red in h3.
  specialize (h3 _ (in_eq _ _)).
  destruct h3; subst; auto.
  contradict h4; auto.
Qed. 


Lemma incl_cons_length :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T) (a:T),
    incl (a::l) l' ->
    incl l' l ->
    NoDup (a::l) ->
    NoDup l' ->
    length l' - length l = 1.
  intros T hdt l l' a h2 h3 h4 h5.
  pose proof (incl_no_dup_le hdt _ _ h4 h5 h2) as h6.
  pose proof (incl_no_dup_le hdt _ _ h5 (no_dup_cons _ _ h4) h3) as h7.
  simpl in h6.
  omega.
Qed.


Lemma partition_faml_le :
  forall {T:Type} (pfdt:type_eq_dec T)
         (ll:faml T) (l lx:list T),
    no_dup_mem ll ->
    NoDup l ->
    partition_faml ll l ->
    In lx ll ->
    length lx <= length l.
  intros T hdt ll l lx hnd hnd' h1 h2.
  pose proof (partition_faml_incl _ _ _ h1 h2) as h3.
  apply incl_no_dup_le in h3; auto.
Qed.


Lemma no_dup_list_to_set_eq_length :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    NoDup l ->
    NoDup l' ->
    list_to_set l =
    list_to_set l' ->
    length l = length l'.
  intros T hdt l.
  induction l as [|a l ih]; simpl.
  intros ? ? ? h1.
  symmetry in h1.
  apply empty_set_nil in h1; subst; auto.
  intro l'.
  destruct l' as [|a' l'].
  intros ? ? h1.
  simpl in h1.
  apply add_not_empty in h1; contradiction.
  intros h1 h2 h3.
  simpl in h3; simpl.
  destruct (hdt a a') as [|h4]; subst.
  pose proof (f_equal (fun S => Subtract S a') h3) as h4.
  simpl in h4.
  do 2 rewrite sub_add_same_nin in h4.
  f_equal.
  apply ih.
  apply no_dup_cons in h1; auto.
  apply no_dup_cons in h2; auto.
  assumption.
  pose proof (no_dup_cons_nin _ _ h2) as h5.
  rewrite <- list_to_set_in_iff; auto.
  pose proof (no_dup_cons_nin _ _ h1) as h5.
  rewrite <- list_to_set_in_iff; auto.
  pose proof (no_dup_cons_nin _ _ h1) as h5.
  rewrite <- list_to_set_in_iff; auto. 
  f_equal.
  pose proof (f_equal (fun S => Subtract S a) h3) as h5.
  simpl in h5.
  rewrite sub_add_same_nin in h5.
  specialize (ih (remove hdt a (a'::l')) (no_dup_cons _ _ h1)).
  hfirst ih.
  apply no_dup_remove; auto.
  specialize (ih hf).
  simpl in ih.
  destruct (hdt a a') as [h6 | h6].
  contradiction.
  simpl in ih.
  rewrite sub_add_comm in h5.
  rewrite (subtract_remove_compat hdt) in h5.
  specialize (ih h5).
  rewrite ih.
  rewrite length_remove_no_dup.
  pose proof (Add_intro2 _ (list_to_set l) a) as h7.
  rewrite h3 in h7.
  destruct h7 as [a h7 | a h7].
  rewrite <- list_to_set_in_iff in h7.
  apply in_not_nil in h7.
  apply not_nil_lt in h7.
  omega.
  destruct h7.
  contradict h4; auto.
  apply no_dup_cons in h2; auto.
  pose proof (Add_intro2 _ (list_to_set l) a) as h7.
  rewrite h3 in h7.
  destruct h7 as [a h7 | a h7].
  rewrite <- list_to_set_in_iff in h7; auto.
  destruct h7; contradict h4; auto.
  auto.
  apply no_dup_cons_nin in h1.
  rewrite <- list_to_set_in_iff; auto.
Qed.


Lemma list_minus_remove :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T) (a':T),
    NoDup l ->
    In a' l' ->
    ~In a' l ->
    list_minus pfdt l (remove pfdt a' l') =
    list_minus pfdt l l'.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros l' a' hnd h2 hnin.
  destruct in_dec as [h3 | h3].
  destruct in_dec as [h4 | h4].
  apply ih; auto.
  apply no_dup_cons in hnd; auto.
  rewrite in_remove_iff in h3.
  destruct h3; contradiction. 
  destruct in_dec as [h4 | h4]. 
  destruct (h1 a a') as [|h5]; subst.
  contradict hnin; left; auto.
  contradict h3.
  rewrite in_remove_iff.
  split; auto.
  f_equal.
  apply ih.
  apply no_dup_cons in hnd; auto.
  assumption.
  contradict hnin.
  right; auto.
Qed.  




Lemma no_dup_length_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    NoDup l  ->
    NoDup l' ->
    incl l' l ->
    length (list_minus pfdt l l') =
    length l - length l'.
  intros T hdt l.
  induction l as [|a l ih]; simpl; auto. 
  intros l' h1 h2 h3.
  destruct l' as [|a' l'].
  destruct in_dec.
  contradiction. 
  simpl.
  f_equal. 
  rewrite ih; simpl; auto with arith.
  apply no_dup_cons in h1; auto.
  apply incl_nil.
  destruct (hdt a a') as [|h4]; subst.
  destruct in_dec as [h4 | h4].
  rewrite list_minus_cons_remove.
  rewrite <- (remove_not_in' hdt (list_minus hdt l l') a').
  simpl.
  apply ih. 
  apply no_dup_cons in h1; auto.
  apply no_dup_cons in h2; auto.
  pose proof (no_dup_cons_nin _ _ h1) as h5.
  apply incl_cons2_same in h3; auto.
  apply no_dup_cons_nin in h2; auto.
  apply no_dup_cons_nin in h1.
  contradict h1.
  apply in_list_minus_impl in h1.
  destruct h1; auto.
  contradict h4.
  left; auto.
  destruct in_dec as [h5 | h5].
  destruct h5 as [|h5]; subst.
  contradict h4; auto.
  pose proof (h3 _ (in_eq _ _)) as hi.
  destruct hi as [|hi]; subst.
  contradict h4; auto.
  simpl. 
  rewrite list_minus_cons_remove.
  rewrite length_remove_no_dup.
  specialize (ih (remove hdt a (a'::l')) (no_dup_cons _ _ h1) (no_dup_remove hdt _ _ h2)).
  rewrite remove_cons_neq in ih.
  simpl in ih.
  rewrite list_minus_cons_remove in ih.
  rewrite length_remove_no_dup in ih.
  simpl.
  rewrite length_remove_no_dup in ih.
  pose proof (in_not_nil _ _ h5) as h7.
  apply not_nil_lt in h7.
  rewrite <- (S_pred _ _ h7) in ih; auto.
  rewrite <- ih.
  f_equal.
  f_equal.
  symmetry. 
  apply list_minus_remove; auto. 
  apply no_dup_cons in h1; auto.
  apply no_dup_cons_nin in h1; auto.
  apply no_dup_cons in h2; auto.
  red in h3; red.
  intros x ha.
  destruct ha as [|ha]; subst; auto.
  rewrite in_remove_iff in ha.
  destruct ha as [ha hb].
  specialize (h3 _ (in_cons _ _ _ ha)).
  destruct h3 as [|h3]; subst.
  contradict hb; auto.
  assumption.
  apply no_dup_cons in h2; auto.
  assumption.
  apply no_dup_list_minus.
  apply no_dup_cons in h1; auto.
  apply impl_in_list_minus; auto.
  intro ha.
  rewrite in_remove_iff in ha.
  destruct ha as [ha hb].
  apply no_dup_cons_nin in h2.
  contradiction.
  assumption.
  apply no_dup_list_minus.
  apply no_dup_cons in h1; auto.
  apply impl_in_list_minus; auto.
  apply no_dup_cons_nin in h2; auto.
  simpl. 
  rewrite list_minus_cons_remove. 
  destruct (in_dec hdt a' l) as [hi | hi].
  
  specialize (ih l' (no_dup_cons _ _ h1) (no_dup_cons _ _ h2)).
  hfirst ih.
  red.
  intros x ha.
  destruct (hdt x a) as [|hb]; subst.
  contradict h5; right; auto.
  red in h3.
  specialize (h3 _ (in_cons _ _ _ ha)).
  destruct h3 as [|h3]; subst.
  contradict hb; auto.
  assumption.
  specialize (ih hf). 
  rewrite length_remove_no_dup.
  pose proof (impl_in_list_minus hdt l l' _ hi (no_dup_cons_nin _ _ h2)) as h6.
  apply in_not_nil in h6.
  apply not_nil_lt in h6.
  rewrite <- (S_pred _ _ h6); auto.
  apply no_dup_list_minus.
  apply no_dup_cons in h1; auto. 
  apply impl_in_list_minus; auto.
  apply no_dup_cons_nin in h2; auto.
  contradict hi.
  red in h3.
  specialize (h3 _ (in_eq _ _)).
  destruct h3 as [|h3]; subst.
  contradict h4; auto.
  assumption.
Qed.


Lemma no_dup_partition_faml_nat_list_to_set_eq :
  forall (ll ll':faml nat) (lm ln:list nat),
    NoDup ll ->
    NoDup ll' ->
    no_dup_mem ll ->
    no_dup_mem ll' ->
    list_to_set_injable (fun lx => In lx ll) ->
    list_to_set_injable (fun lx => In lx ll') ->
    list_to_set ll = list_to_set ll' ->
    NoDup lm ->
    NoDup ln ->
    partition_faml ll  lm ->
    partition_faml ll' ln ->
    length lm = length ln.
  intro ll.
  induction ll as [|la ll ih]; simpl.
  intros ? ? ? ? ? ? ? ? ? h1.
  symmetry in h1.
  apply empty_set_nil in h1; subst.
  intros ? ? h1 h2.
  apply partition_faml_nil in h1.
  apply partition_faml_nil in h2.
  subst; auto.
  intro ll'.
  destruct ll' as [|la' ll'].
  intros ? ? ? ? ? ? ? ? h1 ? ?.
  simpl in h1.
  apply add_not_empty in h1.
  contradiction.
  intros lm ln h1 h2 hndpmm hndpmn hli hli' h3 hnm hnn.
  simpl in h3.
  destruct (list_nat_eq_dec la la') as [|h4]; subst. 
  pose proof (f_equal (fun S => Subtract S la') h3) as h4.
  simpl in h4.
  do 2 rewrite sub_add_same_nin in h4.
  intros h5 h6.
  pose proof h5 as h5'.
  pose proof h6 as h6'.
  apply (partition_faml_cons nat_eq_dec  _ la') in h5.
  apply (partition_faml_cons nat_eq_dec  _ la') in h6. 
  specialize (ih ll'
                 (list_minus nat_eq_dec lm la')
                 (list_minus nat_eq_dec ln la')
                 (no_dup_cons _ _ h1) (no_dup_cons _ _ h2)
                 (no_dup_mem_cons _ _ hndpmm)
                 (no_dup_mem_cons _ _ hndpmn)
                 (list_to_set_injable_in_cons _ _ hli)
                 (list_to_set_injable_in_cons _ _ hli')
                 h4
                 (no_dup_list_minus nat_eq_dec _ _ hnm)
                 (no_dup_list_minus nat_eq_dec _ _ hnn) h5 h6).
  pose proof (hndpmm _ (in_eq _ _)) as h7.
  pose proof (hndpmn _ (in_eq _ _)) as h8.
  do 2 rewrite no_dup_length_list_minus in ih; auto.
  pose proof (partition_faml_incl _ _ _ h5' (in_eq _ _)) as h9.
  pose proof (partition_faml_incl _ _ _ h6' (in_eq _ _)) as h10.
  pose proof h9 as h9'.
  pose proof h10 as h10'.
  apply incl_no_dup_le in h9'; auto.
  apply incl_no_dup_le in h10'; auto.
  omega.
  apply nat_eq_dec.
  apply nat_eq_dec.
  eapply partition_faml_incl in h6'.
  apply h6'.
  left; auto.
  eapply partition_faml_incl in h5'.
  apply h5'.
  left; auto.
  eapply partition_faml_incl in h5'.
  apply h5'.
  left; auto.
  eapply partition_faml_incl in h5'.
  apply h5'.
  left; auto.
  eapply partition_faml_incl in h5'.
  apply h5'.
  left; auto.
  intro h7. 
  apply in_faml_to_fam_impl in h7.
  destruct h7 as [l [h7 h8]].
  pose proof (no_dup_cons_nin _ _ h2) as h9.
  red in hli'.
  pose proof (hli' _ _ (in_cons _ _ _ h7) (in_eq _ _)) as h10.
  apply h10 in h8; subst.
  contradiction.
  intro h7. 
  apply in_faml_to_fam_impl in h7.
  destruct h7 as [l [h7 h8]].
  pose proof (no_dup_cons_nin _ _ h1) as h9.
  pose proof (hli _ _ (in_cons _ _ _ h7) (in_eq _ _)) as h10.
  apply h10 in h8; subst.
  contradiction.
  apply no_dup_cons_nin in h2.
  rewrite <- list_to_set_in_iff; auto.
  apply no_dup_cons_nin in h1.
  rewrite <- list_to_set_in_iff; auto.
  apply no_dup_cons_nin in h1.
  rewrite <- list_to_set_in_iff; auto.
  intros h5 h6.
  specialize (ih (remove list_nat_eq_dec la (la'::ll'))
                 (list_minus nat_eq_dec lm la)
                 (list_minus nat_eq_dec ln la)
                 (no_dup_cons _ _ h1)
                 (no_dup_remove _ _ _ h2)
                 (no_dup_mem_cons _ _ hndpmm)
                 (no_dup_mem_remove _ _ _ hndpmn)
                 (list_to_set_injable_in_cons _ _ hli)
                 (list_to_set_injable_in_remove _ _ _ hli')).
  hfirst ih.
  rewrite <- subtract_remove_compat.
  simpl.
  rewrite <- h3.
  rewrite sub_add_same_nin; auto.
  apply no_dup_cons_nin in h1.
  rewrite <- list_to_set_in_iff; auto.
  specialize (ih
                hf
                (no_dup_list_minus _ _ _ hnm)
                (no_dup_list_minus _ _ _ hnn)).
  pose proof h5 as h5'.
  pose proof h6 as h6'.
  apply (partition_faml_cons nat_eq_dec  _ la) in h5.
  pose proof (partition_faml_remove nat_eq_dec (la' :: ll') ln la) as h7.
  hfirst h7.
  right.
  pose proof (Add_intro2 _ (list_to_set ll) (la)) as h8.
  rewrite h3 in h8.
  destruct h8 as [la h8 | la h8].
  rewrite list_to_set_in_iff; auto.
  destruct h8.
  contradict h4; auto.
  pose proof (partition_faml_incl _ _ _ h5' (in_eq _ _)) as h9.
  pose proof (partition_faml_incl _ _ _ h6' hf0) as h10.
  pose proof h9 as h9'.
  pose proof h10 as h10'.
  apply incl_no_dup_le in h9; auto.
  apply incl_no_dup_le in h10; auto.
  specialize (h7 hf0 hli' h6').
  pose proof (list_nat_eq_dec_eq (list_eq_dec nat_eq_dec)) as h8.
  rewrite h8 in h7.
  specialize (ih h5 h7).
  do 2 rewrite no_dup_length_list_minus in ih; auto.
  omega.
  apply nat_eq_dec.
  apply nat_eq_dec.
  pose proof (no_dup_cons_nin _ _ h1) as h7.
  contradict h7.
  apply in_faml_to_fam_impl in h7.
  destruct h7 as [l [h7 h8]].
  apply hli in h8; subst; auto.
Qed.


Lemma partition_faml_length :
  forall {T:Type} (pfdt:type_eq_dec T)
         (ll:faml T) (l:list T),
    NoDup l ->
    NoDup ll ->
    no_dup_mem ll ->
    list_to_set_injable (fun lx => In lx ll) ->
    partition_faml ll l ->
    plusl (map (@length _) ll) =
    length l.
  intros T hdt ll.
  induction ll as [|al ll ih]; simpl.
  intros l hnd0 hnd hndm h0 h1.
  apply partition_faml_nil in h1; subst; auto.
  intros l hnd0 hnd hndm h1.
  specialize (ih (list_minus hdt l al) (no_dup_list_minus _ _ _ hnd0)
                 (no_dup_cons _ _ hnd)
                 (no_dup_mem_cons _ _ hndm)
                 (list_to_set_injable_in_cons _ _ h1)).
  pose proof (partition_faml_remove hdt (al::ll) l _ (in_eq _ _) h1) as h2.
  intro h3.
  specialize (h2 h3).
  rewrite remove_cons_eq in h2.
  erewrite <- remove_not_in' in h2.
  apply ih in h2.
  rewrite h2.
  rewrite no_dup_length_list_minus; auto with arith.
  pose proof (partition_faml_le hdt (al::ll) _ _ hndm hnd0 h3 (in_eq _ _)) as h4.
  omega.
  eapply hndm.
  left; auto.
  red.
  intros a h4.
  do 2 red in h3.
  simpl in h3.
  destruct h3 as [h3 h5].
  rewrite family_union_add in h5.
  rewrite list_to_set_in_iff.
  rewrite <- h5.
  left.
  rewrite <- list_to_set_in_iff; auto.
  apply no_dup_cons_nin in hnd; auto.
Qed.


Lemma partition_faml_same_length_eq :
  forall (ll ll':faml nat)
         (ln:list nat),
    NoDup ll ->
    NoDup ll' ->
    no_dup_mem ll ->
    no_dup_mem ll' ->
    list_to_set_injable (fun lx => In lx ll) ->
    list_to_set_injable (fun lx => In lx ll') ->
    list_to_set ll = list_to_set ll' ->
    NoDup ln ->
    partition_faml ll  ln ->
    partition_faml ll' ln ->
    length ll = length ll'.
  intro ll.
  induction ll as [|al ll ih]; simpl.
  intros ? ? ? ? ? ? ? ? h1.
  symmetry in h1.
  apply empty_set_nil in h1; subst; auto.
  intro ll'.
  destruct ll' as [|al' ll'].
  intros ? ? ? ? ? ? ? h1.  
  simpl in h1.
  apply add_not_empty in h1; contradiction.  
  intros ln h1 h2 h3 h4 h5 h6 h7 h8 h9 h10.
  pose proof (no_dup_cons_nin _ _ h1) as hnin.
  pose proof (no_dup_cons_nin _ _ h2) as hnin'.
  pose proof (no_dup_cons _ _ h1) as h1'.
  pose proof (no_dup_cons _ _ h2) as h2'.
  pose proof (no_dup_mem_cons _ _ h3) as h11.
  pose proof (no_dup_mem_cons _ _ h4) as h12.
  pose proof (list_to_set_injable_in_cons ll al h5) as h13.
  pose proof (list_to_set_injable_in_cons ll' al' h6) as h14.
  pose proof (partition_faml_cons nat_eq_dec ll al ln) as h15. 
  hfirst h15. 
  intro h16.
  apply in_faml_to_fam_impl in h16.
  destruct h16 as [l [h16 h17]].
  red in h5.
  pose proof (h5 l al (in_cons _ _ _ h16) (in_eq _ _) h17); subst.
  contradiction.
  specialize (h15 hf h9).
  simpl; f_equal. 
  destruct (list_nat_eq_dec al al') as [h16 | h16]; subst.
  simpl in h7.
  pose proof (f_equal (fun S => Subtract S al') h7) as h17.
  simpl in h17.   
  do 2 rewrite sub_add_same_nin in h17; auto.

  pose proof (partition_faml_cons nat_eq_dec ll' al' ln) as h15'.
  hfirst h15'.
  intro h16.
  apply in_faml_to_fam_impl in h16.
  destruct h16 as [l [h16 h17']].
  red in h6.
  pose proof (h6 l al' (in_cons _ _ _ h16) (in_eq _ _) h17'); subst.
  contradiction.
  specialize (h15' hf0 h10).
  specialize
    (ih
       ll' (list_minus nat_eq_dec ln al')
       h1' h2' h11 h12 h13 h14 h17
       (no_dup_list_minus nat_eq_dec _ _ h8) h15 h15'); auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.

  specialize
    (ih (remove list_nat_eq_dec al (al'::ll'))
        (list_minus nat_eq_dec ln al) h1'
        (no_dup_remove _ _ _ h2) h11
        (no_dup_mem_remove _ _ _ h4)
        h13
        (list_to_set_injable_in_remove _ _ _ h6)).
  hfirst ih.
  rewrite <- subtract_remove_compat; simpl.
  simpl in h7; rewrite <- h7.
  rewrite sub_add_same_nin; auto.
  rewrite <- list_to_set_in_iff; auto.
  specialize
    (ih
       hf0 (no_dup_list_minus _ _ _ h8) h15).
  pose proof (partition_faml_remove nat_eq_dec (al' :: ll') ln al) as h17.
  hfirst h17.
  right.
  pose proof (Add_intro2 _ (list_to_set ll) al) as h18.
  rewrite h7 in h18.
  destruct h18 as [al h18 | al h18].
  rewrite list_to_set_in_iff; auto.
  destruct h18.
  contradict h4; auto.
  specialize (h17 hf1 h6 h10).
  rewrite (list_nat_eq_dec_eq (list_eq_dec nat_eq_dec)) in h17.
  specialize (ih h17).
  rewrite length_remove_no_dup in ih; auto.
Qed.


Lemma le_length_app_l :
  forall {T:Type} (l l':list T),
    length l <= length (l ++ l').
  intros; rewrite app_length; auto with arith.
Qed.

Lemma le_length_app_r :
  forall {T:Type} (l l':list T),
    length l' <= length (l ++ l').
  intros; rewrite app_length; auto with arith.
Qed.


Lemma lt_length_app :
  forall {T:Type} (l l':list T) (a:T),
    length l < length (l++(a::l')).
  intros; rewrite app_length; simpl; omega.
Qed.


Lemma lt_length_app_sing_dec :
  forall {T:Type} (l:list T) (a:T) (n:nat),
    n < length (l++(a::nil)) ->
    {n < length l} + {n = length l}.
  intros T l a n h1.
  rewrite app_length in h1. simpl in h1.
  rewrite S_compat in h1.
  apply lt_S_dec in h1; auto.
Qed.


Lemma lt_length_app_sing_S :
  forall {T:Type} (l:list T) (a:T) (n:nat),
    S n < length (l++(a::nil)) ->
    n < length l.
  intros T l a n h1.
  rewrite app_length in h1.
  simpl in h1.
  rewrite S_compat in h1.
  apply lt_S_n in h1; auto.
Qed.

Lemma le_length_app_sing_S :
  forall {T:Type} (l:list T) (a:T) (n:nat),
    S n <= length (l++(a::nil)) ->
    n <= length l.
  intros T l a n h1.
  rewrite app_length in h1.
  simpl in h1.
  rewrite S_compat in h1.
  apply le_S_n in h1; auto.
Qed.



Lemma lt_length_app_ge_impl :
  forall {T:Type} (l l':list T) (n:nat),
    n >= length l ->
    n < length (l++l') ->
    n - length l < length l'.
  intros ? ? ? ? ? h1; rewrite app_length in h1; omega.
Qed.

Lemma lt_length_app_dec : 
  forall {T:Type} (l l' :list T) (n:nat),
         n < length (l ++ l') ->
         {n < length l} + {n - length l < length l'}.
  intros T l l' n h1.
  destruct (lt_eq_lt_dec n (length l)) as [h2 | h2].
  destruct h2 as [h2 | h2]; subst.
  left; auto.
  right.
  rewrite minus_same.
  destruct (nil_dec' l') as [h2 | h2]; subst.
  rewrite app_nil_r in h1.
  apply lt_irrefl in h1; contradiction.
  apply not_nil_lt; auto.
  right.
  apply lt_length_app_ge_impl; auto with arith.
Qed.



Fixpoint list_singularize_aux {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (acc:list T) :=
  match l with
    | nil => nil
    | a::l' => if (in_dec pfdt a l') then
                 (list_singularize_aux pfdt l' acc)
               else (a::(list_singularize_aux pfdt l' (a::acc)))
  end.

(*This returns the list without the duplicate entries, 
  in the same order as each new element appearing starting 
  backwards, i.e. [list_singularize 
      (3::4::2::4::2::3::3::3::nil) = 4::2::3].*)
Definition list_singularize {T:Type} (pfdt:type_eq_dec T)
           (l:list T) :=
  list_singularize_aux pfdt l nil.

Lemma list_singularize_not_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (acc:list T),
    (forall (x:T), In x acc -> ~In x l) ->
    list_singularize_aux pfdt l acc =
    list_singularize_aux pfdt l nil.
intros T h0 l.
induction l as [|a l h1].
intros. simpl. auto.
intros acc h2.
simpl.
destruct (in_dec h0 a l) as [h3 | h4].
assert (h4:forall x:T, In x acc -> ~ In x l).
  intros x h4.
  specialize (h2 x h4). simpl in h2. tauto.
apply h1; auto.
assert (h5: forall x:T, In x (a::acc) -> ~In x l).
  intros x h6.
  destruct h6 as [|h6]. subst. auto.
  specialize (h2 _ h6).
  simpl in h2. tauto.
pose proof (h1 _ h5) as h7.
rewrite h7.
assert (h8:forall x:T, In x (a::nil) -> ~In x l).
  intros x h9.
  destruct h9. subst. auto. contradiction.
rewrite (h1 _ h8). reflexivity.
Qed.

Lemma list_singularize_nil :
  forall {T:Type} (pfdt:type_eq_dec T),
    list_singularize pfdt nil = nil.
  intros; simpl; auto.
Qed.
  
  
Lemma list_singularize_nil_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    list_singularize pfdt l = nil <->
    l = nil.
  unfold list_singularize; intros T h1 l; split; intro h2.
  revert h2.
  induction l as [|a l ih]; simpl; auto.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; simpl.
  intro h2.
  apply ih in h2; subst.
  contradict hl.
  intro; discriminate.
  subst; auto.
Qed.


(* Returns the list [l] with [x] at the head of it,
   and all other occurrences of [x] removed.*)
Definition new_head {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
: list T := x::remove pfdt x l.

Lemma list_singularize_new_head_not_in_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    ~In x l ->
    list_singularize pfdt l =
    remove pfdt x (list_singularize pfdt l).
unfold list_singularize; intros T h0 l.  
induction l as [|a l h1].
intros x h1.
simpl. auto.
intros x h2.
simpl.
simpl in h2.
destruct (in_dec h0 a l) as [h4 | h5].
apply h1; auto; simpl. 
pose proof (remove_cons_neq h0 (list_singularize_aux h0 l (a::nil)) x a) as h6.
hfirst h6.
contradict h2; left; auto.
specialize (h6 hf).
rewrite h6.
f_equal. 
rewrite list_singularize_not_in.
specialize (h1 x).
hfirst h1.
contradict h2; right; auto.
specialize (h1 hf0).
rewrite h1; auto.
f_equal; auto.
intros y hy; destruct hy; subst; auto.
Qed.

Lemma subtract_new_head_compat : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    Subtract (list_to_set l) x = list_to_set (remove pfdt x l).
intros T h0 l. induction l as [|a l h1].

intros x. simpl.
apply Extensionality_Ensembles.
red. split. red. intros y h1.
destruct h1; contradiction.
auto with sets.

intro x. 
simpl.
destruct (h0 x a) as [h2 | h3]. subst.
pose proof (in_dec h0 a l) as h4.
destruct h4 as [h4l | h4r].
rewrite list_to_set_in_iff in h4l.
rewrite sub_add_same_in. apply h1. auto.
rewrite list_to_set_in_iff in h4r.
rewrite sub_add_same_nin.
rewrite nin_remove_eq; auto.
rewrite list_to_set_in_iff. auto. auto.
simpl.
rewrite sub_add_comm.
rewrite h1. auto. auto.
Qed.



Lemma add_list_to_set_to_remove : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (A:Ensemble T) (x:T),
    ~ Inn A x ->
    Add A x = list_to_set (list_singularize pfdt l) ->
    A = list_to_set (remove pfdt x (list_singularize pfdt l)).
unfold list_singularize; intros T h0 l. induction l as [|a l h1].
intros A x h1 h2. simpl in h2.
pose proof (add_not_empty A x). contradiction.
intros A x h2 h3.
simpl in h3. 
destruct (in_dec h0 a l) as [h4 | h5].
rewrite (h1 _ x).
simpl.
destruct (in_dec h0 a l) as [h5|?]. reflexivity.
contradiction. assumption. assumption.
simpl. 
destruct (in_dec h0 a l) as [h6|h7].
contradiction. simpl.
destruct (h0 x a) as [h8 | h9].
subst.
assert (h6:forall c:T, In c (a::nil) -> ~In c l).
  intros y h8.
  destruct h8; [subst | contradiction]; auto.
simpl in h3.
pose proof (add_nin_sub_compat _ _ _ h2 h3) as h8.
rewrite (subtract_new_head_compat h0) in h8. auto.
assert (h6:forall c:T, In c (a::nil) -> ~In c l).
  intros y h8.
  destruct h8; [subst | contradiction]; auto.
rewrite list_singularize_not_in in h3.
rewrite list_singularize_not_in.
assert (h10:Subtract (Add A x) x = Subtract (list_to_set (a::list_singularize_aux h0 l nil)) x).
  f_equal. auto.
rewrite sub_add_same_nin in h10.
rewrite h10.
simpl.
rewrite sub_add_comm.
rewrite (subtract_new_head_compat h0).
auto. auto. auto.
intros y h8.
destruct h8; [subst | contradiction]; auto.
intros y h8.
destruct h8; [subst | contradiction]; auto.
Qed.


Lemma remove_list_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    remove pfdt x (list_singularize pfdt l) =
    list_singularize pfdt (remove pfdt x l).
unfold list_singularize; intros T h0 l. 
induction l as [|a l h1].
intros x. simpl. auto.
intros x. simpl.
destruct (in_dec h0 a l) as [h2 | h3];
  destruct (h0 x a) as [h4 | h5].  
subst. auto.
apply neq_sym in h5.
simpl.  
destruct (in_dec h0 a (remove h0 x l)) as [h6 | h7]; auto.
pose proof (nin_remove_eq_nin h0 l a x h7) as h8. 
destruct h8; contradiction.
subst. 
simpl.
rewrite eq_dec_eq at 1; auto.
rewrite list_singularize_not_in; auto.
intros x hx; destruct hx; subst; auto.
simpl. 
rewrite eq_dec_neq_r at 1; auto.
destruct (in_dec h0 a (remove h0 x l)).
rewrite in_remove_iff in i.
destruct i; contradiction.
rewrite list_singularize_not_in.
rewrite (list_singularize_not_in h0 (remove h0 x l) (a::nil)); auto.
f_equal; auto.
intros c hc; destruct hc; subst; auto.
intros c hc; destruct hc; subst; auto.

Grab Existential Variables.
auto.
Qed.


Lemma list_singularize_in_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T), 
    In x l <-> In x (list_singularize pfdt l).
unfold list_singularize; intros T h0 l.
induction l as [|a l h1].
(* nil *)
split.
(* -> *)
intro; contradiction.
simpl. tauto.
(* <- *)
(* split *)
intro x.
split.
(* -> *)
intros h2.
destruct h2 as [h2 | h3]. subst.
simpl.
destruct (in_dec h0 x l) as [h2 | h3].
apply h1. assumption.
constructor. reflexivity.
simpl.
destruct (in_dec h0 a l) as [h4 | h5].
apply h1. assumption.
rewrite list_singularize_not_in.
right.
apply h1. assumption.
intros y h6.
destruct h6; try subst.
assumption. contradiction.
intro h2.
simpl in h2.
destruct (in_dec h0 a l) as [h3 | h4].
right.
apply h1. assumption.
rewrite list_singularize_not_in in h2.
destruct h2 as [h2a | h2b].
subst.
constructor. reflexivity.
right. apply h1. assumption.
intros y h3.
destruct h3; subst. assumption.
contradiction.
Qed.


Lemma list_singularize_lrep :
  forall {T:Type} (pfdt:type_eq_dec T)
         (x:T) (n:nat),
    list_singularize pfdt (lrep x (S n)) =
    x::nil.
  unfold list_singularize; intros T h1 x n. revert x.
  induction n as [|n ih]; simpl; auto.
  intro x.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  specialize (ih x).
  simpl in ih.
  lr_if_goal; fold hlr; fold hlr in ih; destruct hlr as [hl' | hr']; auto.
  f_equal.
  contradict hr.
  left; auto.
Qed.

Lemma list_singularize_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    list_to_set l = Singleton x ->
    list_singularize pfdt l = x::nil.
  unfold list_singularize; intros T h1 l x h2.
  pose proof h2 as h2'.
  apply impl_all_sing in h2.
  apply all_sing_lrep in h2.
  rewrite h2.
  pose proof (In_singleton _  x) as h3.
  rewrite <- h2' in h3.
  rewrite <- list_to_set_in_iff in h3.
  pose proof (in_not_nil _ _ h3) as h4.
  apply S_length_tl in h4.
  rewrite <- h4.
  apply list_singularize_lrep.
Qed.

Lemma remove_app :
  forall {T:Type} (pfdt:type_eq_dec T)
         (a:T) (l l':list T),
    remove pfdt a l ++ (remove pfdt a l') =
    remove pfdt a (l ++ l').
intros T h0 a l. revert a.
induction l as [|b l h1].
simpl.
intros; auto.
intros a l'.
simpl.
destruct (h0 a b) as [h3 | h4].
subst.
apply h1.
simpl.
rewrite h1.
reflexivity.
Qed.


Lemma list_to_set_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T), list_to_set l =
                     list_to_set (list_singularize pfdt l).
unfold list_singularize; intros T h0 l.
induction l as [|a l h1]. simpl. reflexivity.
simpl.
destruct (in_dec h0 a l) as [h2 | h3].
rewrite list_to_set_in_iff in h2.
rewrite h1 in h2. rewrite h1.
rewrite in_add_eq. reflexivity. assumption.
simpl. rewrite h1.
assert (h9:forall c:T, In c (a::nil) -> ~In c l).
  intros y h10.
  destruct h10; [subst | contradiction]; auto.
rewrite (list_singularize_not_in _ _ _  h9).
reflexivity.
Qed.


Lemma list_singularize_rev_cons : 
  forall {T:Type} (pfdt:type_eq_dec T) (a:T) (l:list T), 
    list_singularize pfdt(l++(a::nil)) =
    (remove pfdt a (list_singularize pfdt l))++(a::nil).
unfold list_singularize; intros T hdt a l. revert a.
induction l as [|b l h0].
intro a.
simpl. reflexivity.
simpl.
intro a.
destruct (in_dec hdt b (l++a::nil)) as [h1|h2];
  destruct (in_dec hdt b l) as [h3|h4].
rewrite h0.
reflexivity.
pose proof (in_app_or _ _ _ h1) as h5.
destruct h5 as [h5a | h5b].
contradiction.
destruct h5b; subst.
assert (h6:forall x:T, In x (b::nil) -> ~In x l).
  intros x h7.
  destruct h7; subst.
  assumption. contradiction.
pose proof (list_singularize_not_in hdt _ _ h6) as h7.
rewrite h7.
simpl.
destruct (hdt b b) as [h8 | h9].
apply h0.
contradict h9. reflexivity.
contradiction.
contradict h2.
apply in_or_app. left; assumption.
assert (h5:a <> b).
  intro h5.
  subst.
  contradict h2.
  apply in_or_app. right; constructor.  reflexivity.
rewrite list_singularize_not_in.
pose proof (list_singularize_not_in hdt l (b::nil)) as h6.
assert (h7:(forall x:T, In x (b::nil) -> ~In x l)).
  intros x h7. destruct h7; subst. assumption. contradiction.
specialize (h6 h7); clear h7.
rewrite h6.
pose proof (app_comm_cons (list_singularize_aux hdt l nil) (a::nil) b) as h7.
rewrite remove_cons_neq.
rewrite <- app_comm_cons.
rewrite <- h0.
reflexivity.
assumption.
intros x h6.
destruct h6; subst. assumption. contradiction.
Qed.


Lemma list_singularize_in_cons :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    In a l ->
    list_singularize pfdt (a::l) = list_singularize pfdt l.
  unfold list_singularize; intros T h0 l.
  induction l as [|b l h1]; simpl.
  intros; contradiction.
  intros a h2.
  destruct h2 as [h2  |h2]; subst.
  destruct (h0 a a) as [h2 | h2].
  reflexivity.
  contradict h2. reflexivity.
  destruct (h0 b a) as [h3  |h3].
  reflexivity.
  destruct (in_dec h0 a l) as [h4 | h4].
  reflexivity.
  contradiction.
Qed.


Lemma list_singularize_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T) (pf: ~In a l),
    list_singularize pfdt (a::l) = a::list_singularize pfdt l.
  
  unfold list_singularize; intros T h0 l a h1; simpl.
  destruct in_dec as [h2 | h2].
  contradiction.  
  rewrite list_singularize_not_in; auto.
  intros x h3.
  destruct h3; subst; auto.
Qed.

Lemma no_dup_list_singularize : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup (list_singularize pfdt l).
  intros T h0 l.
  induction l as [|a l h1].
  simpl. constructor.
  simpl.
  destruct (in_dec h0 a l) as [h2 | h3].
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_nin.
  constructor.
  rewrite <- list_singularize_in_iff.
  assumption. assumption. assumption.
Qed.

Lemma list_singularize_no_dup : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup l <-> list_singularize pfdt l = l.
intros T h0 l.
split.
(* -> *)
intro h1.
induction h1 as [|x l h2 h3 h4]. simpl. reflexivity.
simpl.
destruct (in_dec h0 x l) as [h6 | h7].
contradiction.
rewrite list_singularize_nin.
rewrite h4. reflexivity.
assumption.
intro h1.
rewrite <- h1.
apply no_dup_list_singularize.
Qed.


Lemma incl_list_singularize :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    incl (list_singularize pfdt l) l.
  intros T h1 l; red; intros x h2.
  rewrite <- list_singularize_in_iff in h2; auto.
Qed.

Lemma incl_list_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    incl l l' ->
    incl (list_singularize pfdt l)
          (list_singularize pfdt l').
  intros T h1 l l' h2; red; intros x h3.
  rewrite <- list_singularize_in_iff in h3.
  rewrite <- list_singularize_in_iff; auto.
Qed.



Lemma finite_set_list_no_dup : 
  forall {T:Type} (E:Ensemble T),
    Finite E -> 
    exists (l:list T),
      E = list_to_set l /\ NoDup l.
  intros T E h1.
  induction h1 as [|E h2 h3 a h5].
  exists nil; simpl.
  split; auto; try constructor.
  destruct h3 as [l h3].
  destruct h3 as [h3 h4].
  exists (a::l); simpl.
  subst.
  split; auto.
  constructor.
  rewrite list_to_set_in_iff; auto.
  assumption.
Qed.




Lemma length_list_singularize_cons_gt_0 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    0 < length (list_singularize pfdt (a::l)).
  intros T h0 l.
  induction l as [|b l h1]; simpl; auto with arith.
  intro a.
  destruct (h0 b a) as [h2 | h2]; subst.
  destruct (in_dec h0 a l) as [h2 | h2].
  rewrite list_singularize_in_cons.
  rewrite list_singularize_in_cons.
  rewrite list_singularize_in_iff in h2.
  apply in_not_nil in h2.
  apply not_nil_lt in h2; auto.
  apply h2.
  assumption. 
  left; auto. 
  rewrite list_singularize_in_cons; simpl.
  rewrite list_singularize_nin; simpl; auto with arith.
  left; auto.
  destruct (in_dec h0 a l) as [h3 | h3].
  destruct (in_dec h0 b l) as [h4 | h4].
  rewrite list_singularize_in_cons.
  rewrite list_singularize_in_cons.
  rewrite (list_singularize_in_iff h0) in h3.
  apply in_not_nil, not_nil_lt in h3; auto.
  assumption.
  right; auto.
  rewrite list_singularize_in_cons. 
  rewrite (list_singularize_in_iff h0) in h3.
  apply in_not_nil, not_nil_lt in h3; auto.
  right; auto.
  rewrite list_singularize_nin.
  simpl; auto with arith.
  intro h4.
  destruct h4; subst.
  contradict h2; auto.
  contradiction.
Qed.


Lemma list_singularize_nin_app_nil :
  forall {T:Type} (pfdt:type_eq_dec T) (x:T) (l:list T),
    ~In x l ->
    list_singularize pfdt (l ++ x::nil) =
    list_singularize pfdt l ++ x :: nil.
  intros T h1 x l.
  revert x.
  induction l as [|a l ih]; simpl.
  intros; auto.
  intros x h2.
  
  destruct (in_dec h1 a (l++x::nil)) as [hl | hr].
  destruct (in_dec h1 a l) as [hl' | hr'].
  destruct (in_dec h1 x l) as [h3 | h3].
  contradict h2.
  right; auto.
  rewrite list_singularize_in_cons.
  rewrite list_singularize_in_cons.
  
  apply ih; auto.
  assumption.
  apply in_app_l; auto.
  apply in_app_or in hl.
  destruct hl as [|hl].
  contradiction.
  destruct hl; subst.
  contradict h2.
  left; auto.
  contradiction.

  rewrite list_singularize_nin; auto.
  destruct (in_dec h1 a l) as [hl' | hr'].
  
  contradict hr.
  apply in_app_l; auto.
  rewrite list_singularize_nin; auto.
  rewrite <- app_comm_cons.
  f_equal; auto.
Qed.


Lemma list_singularize_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l),
    list_singularize pfdt (remove pfdt x l) = 
    remove pfdt x (list_singularize pfdt l).
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros x h2.
  destruct (h1 x a) as [|h3]; subst.
  destruct (in_dec h1 a l) as [h3 | h3].
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_nin; auto.
  rewrite <- remove_not_in'; auto.
  rewrite remove_cons_eq.
  rewrite <- remove_not_in'; auto.
  rewrite <- list_singularize_in_iff; auto.
  destruct h2 as [|h2].
  subst; contradict h3; auto.
  destruct (in_dec h1 a (remove h1 x l)) as [h4 | h4].
  rewrite list_singularize_in_cons.
  pose proof h4 as h4'.
  rewrite in_remove_iff in h4'.
  destruct h4' as [h4' h5].
  rewrite list_singularize_in_cons; auto.
  assumption.
  rewrite list_singularize_nin; auto.
  assert (h5:~In a l).
    contradict h4.
    rewrite in_remove_iff; auto.
  rewrite list_singularize_nin; auto.
  rewrite remove_cons_neq; auto.
  f_equal; auto.
Qed.


Lemma list_singularize_in_app_nil :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    In a l ->
    list_singularize pfdt (l++a::nil) =
    list_singularize pfdt (remove pfdt a l) ++ a::nil.
  intros T h1 l. 
  induction l as [|b l ih]; simpl.
  intros; contradiction.
  intros a h2.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; subst.
  rewrite list_singularize_in_cons.
  destruct (in_dec h1 b l) as [h3 | h3].
  apply ih; auto.
  rewrite <- remove_not_in'; auto.
  rewrite list_singularize_nin_app_nil; auto.
  apply in_app_r; constructor; auto.
  destruct h2 as [h2 | h2].
  subst.
  contradict hr; auto.
  simpl.
  destruct (in_dec h1 b l) as [hl' | hr'].
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_in_cons; auto.
  rewrite in_remove_iff.
  split; auto.
  apply in_app_l; auto.
  rewrite list_singularize_nin.
  rewrite list_singularize_nin.
  simpl.
  f_equal; auto.
  intro h3.
  rewrite in_remove_iff in h3.
  destruct h3 as [h3 h4].
  contradiction.
  intro h3.
  apply in_app_or in h3.
  destruct h3 as [|h3]; subst.
  contradiction.
  destruct h3; subst.
  contradict hr; auto.
  contradiction.
Qed.



Lemma list_to_sets_eq_length :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    list_to_set l = list_to_set l' ->
    length (list_singularize pfdt l) =
    length (list_singularize pfdt l').
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros ? h2.
  symmetry in h2.
  apply empty_set_nil in h2; subst; simpl; auto.
  intros l'.
  destruct l' as [|a' l'].
  intro h2.
  apply add_not_empty in h2.
  contradiction.
  simpl.
  intro h2.
  destruct (h1 a a') as [|h3]; subst. 
  destruct (in_dec h1 a' l) as [h4 | h4],
                               (in_dec h1 a' l') as [h5 | h5].
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_in_cons; auto.
  apply ih.
  do 2 rewrite in_add_eq in h2; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite in_add_eq in h2.
  simpl.
  pose proof (ih (a'::l') h2) as h6. 
  simpl in h6.
  rewrite list_singularize_in_cons; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_in_cons; auto.
  simpl. 
  rewrite (in_add_eq (list_to_set l')) in h2.
  pose proof (f_equal (fun S => Subtract S a') h2) as h6.
  simpl in h6.
  rewrite sub_add_same_nin in h6.
  specialize (ih (remove h1 a' l')).
  hfirst ih.
  rewrite <- subtract_remove_compat; auto.
  specialize (ih hf).
  rewrite list_singularize_remove,
  length_remove_no_dup in ih.
  pose proof h5 as h5'.
  rewrite (list_singularize_in_iff h1) in h5'.
  apply in_not_nil in h5'.
  apply not_nil_lt in h5'.
  omega.
  apply no_dup_list_singularize.
  rewrite <- list_singularize_in_iff; auto.
  assumption.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_nin; auto; simpl.
  f_equal; apply ih.
  pose proof (f_equal (fun S => Subtract S a') h2) as h6.
  simpl in h6.
  rewrite sub_add_same_nin in h6.
  rewrite sub_add_same_nin in h6; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto. 
  destruct (in_dec h1 a l) as [h4 | h4],
                              (in_dec h1 a' l') as [h5 | h5].
  do 2 rewrite in_add_eq in h2; auto.
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_in_cons; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_nin; auto.
  simpl.
  rewrite in_add_eq in h2.
  specialize (ih (a'::l')).
  apply ih in h2.
  rewrite list_singularize_nin in h2; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_in_cons; auto.
  simpl. 
  rewrite (in_add_eq (list_to_set l')) in h2.
  pose proof (f_equal (fun S => Subtract S a) h2) as h6.
  simpl in h6.
  rewrite sub_add_same_nin in h6.
  specialize (ih (remove h1 a l')).
  hfirst ih.
  rewrite <- subtract_remove_compat; auto.
  specialize (ih hf).
  rewrite ih.
  rewrite list_singularize_remove.
  rewrite length_remove_no_dup.
  rewrite (list_singularize_in_iff h1) in h5.
  apply in_not_nil in h5.
  apply not_nil_lt in h5.
  omega. 
  apply no_dup_list_singularize.
  rewrite <- list_singularize_in_iff; auto.
  pose proof (Add_intro2 _ (list_to_set l) a) as h7.
  rewrite h2 in h7.
  rewrite list_to_set_in_iff; auto.
  pose proof (Add_intro2 _ (list_to_set l) a) as h7.
  rewrite h2 in h7. 
  rewrite list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- list_to_set_in_iff; auto.
  rewrite list_singularize_nin; auto.
  rewrite list_singularize_nin; auto. 
  simpl; f_equal.
  
  pose proof (Add_intro2 _ (list_to_set l) a) as h6.
  rewrite h2 in h6.
  inversion h6; subst.
  rewrite <- list_to_set_in_iff in H.
  pose proof (Add_intro2 _ (list_to_set l') a') as h6'.
  rewrite <- h2 in h6'.
  inversion h6; subst.
  rewrite <- list_to_set_in_iff in H0.
  simpl.
  pose proof (f_equal (fun S => Subtract S a) h2) as h7.
  simpl in h7.
  rewrite sub_add_same_nin in h7.
  
  specialize (ih (remove h1 a (a'::l'))).
  hfirst ih.
  rewrite <- subtract_remove_compat.
  simpl; auto.
  specialize (ih hf).
  rewrite list_singularize_remove in ih.
  rewrite list_singularize_nin in ih; auto.
  rewrite length_remove_no_dup in ih.
  simpl in ih; auto.
  constructor.
  rewrite <- list_singularize_in_iff; auto.
  apply no_dup_list_singularize.
  right.
  rewrite <- list_singularize_in_iff; auto.
  right; auto.
  rewrite <- list_to_set_in_iff; auto.
  destruct H0.
  contradiction.
  destruct H.
  contradict h3; auto.
Qed.


Lemma all_sing_list_singularize :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    l <> nil ->
    all_sing l a ->
    list_singularize pfdt l = a::nil.
  intros T h0 l.
  induction l as [|c l ih]; simpl.
  intros ? h1. contradict h1; auto.
  intros a h1 h2.
  destruct (nil_dec' l) as [hn | hn]; subst.
  rewrite (list_singularize_sing h0 _ c).
  f_equal.
  red in h2.
  apply h2.
  left; auto.
  simpl; auto.
  rewrite add_empty_sing; auto.
  destruct (in_dec h0 c l) as [h3 | h3].
  rewrite list_singularize_in_cons; auto.
  apply ih; auto.
  red in h2; red; intros; apply h2; auto.
  right; auto.
  rewrite list_singularize_nin; auto.
  pose proof h2 as h2'.
  red in h2'.
  specialize (h2' c (in_eq _ _)); subst.
  apply all_sing_cons_in in h2; auto.
  contradiction.
Qed.

Definition singular_fun_in_dep
           {T U:Type} {l:list T} (pfdt:type_eq_dec T)
           (f:fun_in_dep l U) :
  fun_in_dep (list_singularize pfdt l) U :=
  fun x pf => f x (iff2 (list_singularize_in_iff pfdt l x) pf).


Lemma singular_fun_in_dep_compat :
  forall {T U:Type} {l:list T} (pfdt:type_eq_dec T)
         (f:fun_in_dep l U) (x:T) (pf:In x (list_singularize pfdt l)),
    singular_fun_in_dep pfdt f x pf =
    f x (iff2 (list_singularize_in_iff pfdt l x) pf).
  unfold singular_fun_in_dep; intros; auto.
Qed.

Lemma singular_fun_in_dep_compat' :
  forall {T U:Type} {l:list T} (pfdt:type_eq_dec T)
         (f:fun_in_dep l U) (x:T) (pf:In x l),
    f x pf = 
    singular_fun_in_dep pfdt f x (iff1 (list_singularize_in_iff pfdt l x) pf).
  unfold singular_fun_in_dep; intros; f_equal; apply proof_irrelevance.
Qed.

Definition fun_in_dep_in_ens
           {T U:Type} {l:list T}
           (f:fun_in_dep l U) :
  fun_in_ens (list_to_set l) U :=
  fun x pf =>
    f x (iff2 (list_to_set_in_iff l x) pf).


Definition fun_in_dep_to_sig
           {T U:Type} {l:list T}
           (f:fun_in_dep l U) :
  sig_set (list_to_set l) -> U :=
  fun x =>
    f (proj1_sig x) (iff2 (list_to_set_in_iff l (proj1_sig x)) (proj2_sig x)).

Lemma fun_in_dep_in_ens_compat : 
  forall {T U:Type} {l:list T} (pfdt:type_eq_dec T)
         (f:fun_in_dep l U) (x:T) (pf:In x l),
    fun_in_dep_in_ens f x (iff1 (list_to_set_in_iff l x) pf) = 
    f x pf.
  unfold fun_in_dep_in_ens; intros; f_equal; apply proof_irrelevance.
Qed.
  

Lemma fun_in_dep_to_sig_compat : 
  forall {T U:Type} {l:list T} (pfdt:type_eq_dec T)
         (f:fun_in_dep l U) (x:T) (pf:In x l),
    fun_in_dep_to_sig f (exist _ _ (iff1 (list_to_set_in_iff l x) pf)) = 
    f x pf.
  unfold fun_in_dep_to_sig; intros; f_equal; apply proof_irrelevance.
Qed.

                                                    
Definition sing_unq {T:Type} (ll:faml T) : Prop :=
  forall l l',
    In l ll ->
    In l' ll ->
    list_to_set l =
    list_to_set l' ->
    l = l'.


Lemma list_power_no_dup : 
  forall {T U:Type} (la:list T) (lb:list U) (lp:list (T*U)),
    NoDup la -> NoDup lb ->
    (In lp (list_power la lb)) ->
    NoDup lp.
intros T U la.
induction la as [|a la h1].
simpl.
intros lb lp h1 h2 h3.
destruct h3 as [h3l | h3r].
rewrite <- h3l. constructor. contradiction.
intros lb lp h2 h3 h4.
rewrite <- app_nil_l in h2.
pose proof (NoDup_remove_1 _ _ _ h2) as h5.
rewrite app_nil_l in h5.
simpl in h4.
rewrite in_flat_map in h4.
destruct h4 as [lp' h6].
destruct h6 as [h6l h6r].
specialize (h1 _ _ h5 h3 h6l).
rewrite in_map_iff in h6r.
destruct h6r as [b h7].
destruct h7 as [h7l h7r].
rewrite <- h7l.
constructor.
intro h8.
rewrite app_nil_l in h2.
inversion h2.
pose proof (in_list_power1 _ h6l _ h8) as h9.
simpl in h9.
contradiction.
assumption.
Qed.


Lemma no_dup_map_in_dep_inj :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U),
    NoDup l ->
    inj_dep f ->
    NoDup (map_in_dep f).
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros; constructor.
  intros f h1 h2.
  pose proof (no_dup_cons _ _ h1) as h3.
  pose proof (no_dup_cons_nin _ _ h1) as h4.
  specialize (ih (fun_from_cons f) h3 (inj_from_cons _ h2)).
  constructor.
  contradict h4.
  rewrite in_map_in_dep_iff in h4.
  destruct h4 as [a' [h4 h5]].
  unfold fun_from_cons in h5.
  apply h2 in h5; subst; auto.
  assumption.
Qed.

Lemma length_list_singularize_map_in_dep_inj :
  forall {T U:Type} {l:list T} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U),
    inj_dep f ->
    length (list_singularize pfdu (map_in_dep f)) =
    length (list_singularize pfdt l).
  intros T U l h1 h2.
  induction l as [|a l ih]; simpl; auto.
  intros f h3.
  destruct (in_dec h2 (f a (in_eq _ _)) (map_in_dep (fun_from_cons f))) as [hl | hr].
  rewrite list_singularize_in_cons; auto.
  destruct (in_dec h1 a l) as [hl' | hr'].
  rewrite list_singularize_in_cons; auto.
  rewrite in_map_in_dep_iff in hl.
  destruct hl as [a' h4]. 
  destruct h4 as [h4 h5].
  apply ih.
  apply inj_from_cons; auto.
  red in h3.
  rewrite in_map_in_dep_iff in hl.
  destruct hl as [a' hl].
  destruct hl as [hl h4].
  unfold fun_from_cons in h4.
  apply h3 in h4; subst; contradiction.
  rewrite list_singularize_nin; auto.
  destruct (in_dec h1 a l) as [hl' | hr'].
  rewrite list_singularize_in_cons; auto.
  contradict hr.
  rewrite in_map_in_dep_iff.
  exists a, hl'.
  unfold fun_from_cons; f_equal; apply proof_irrelevance.
  rewrite list_singularize_nin; auto.
  simpl; f_equal; apply ih.
  apply inj_from_cons; auto.
Qed.


  
Lemma no_dup_app : 
  forall {T:Type} (l l':list T),
    NoDup l -> NoDup l' ->
    Intersection (list_to_set l) (list_to_set l') =
    Empty_set _ -> NoDup (l ++ l').
intros T l1.
induction l1 as  [|a l1 h1].
simpl.
intros. assumption.
intros l2 h2 h3 h4.
rewrite <- app_comm_cons.
constructor.
intro h5.
pose proof (in_app_or _ _ _ h5) as h6.
destruct h6 as [h6l | h6r].
inversion h2. contradiction.
rewrite list_to_set_in_iff in h6r.
pose proof (in_eq a l1) as h7.
rewrite list_to_set_in_iff in h7.
pose proof (Intersection_intro _ _ _ _ h7 h6r) as h8.
rewrite h4 in h8.
contradiction.
assert (h5:Included (list_to_set l1) (list_to_set (a::l1))).
  red. simpl.
  intros x h5.
  left. assumption.
pose proof (intersection_preserves_inclusion _ _ (list_to_set l2) h5) as h6.
rewrite Intersection_commutative in h6.
rewrite (Intersection_commutative _ (list_to_set l2) (list_to_set (a::l1))) in h6.
rewrite h4 in h6.
assert (h7:(Intersection (list_to_set l1) (list_to_set l2)) =
         (Empty_set T)).
  apply Extensionality_Ensembles.
  red; split; auto with sets.
inversion h2.
apply h1; assumption.
Qed.

Lemma no_dup_app_sing_nin :
  forall {T:Type} (l:list T) (a:T),
    NoDup (l++a::nil) ->
    ~In a l.
  intros T l.
  induction l as [|b l ih]; simpl.
  intros; auto.
  intros a h1 h2.
  pose proof (no_dup_cons _ _ h1) as h3.
  pose proof (no_dup_cons_nin _ _ h1) as h4.
  specialize (ih _ h3).
  destruct h2 as [|h2]; subst.
  contradict h4.
  apply in_app_r; left; auto.
  contradiction.
Qed.


Lemma no_dup_seg_list :
  forall (n:nat),
    NoDup (seg_list n).
  intro n.
  induction n as [|n ih]; simpl.
  rewrite seg_list0.
  constructor.
  rewrite seg_list_S.
  apply no_dup_app; auto.
  apply no_dup_sing.
  simpl.
  rewrite add_empty_sing.
  apply intersection_seg_sing; auto with arith.
Qed.

Lemma no_dup_seq :
  forall (m n:nat),
    NoDup (seq m n).
  intros m n. revert m.
  induction n as [|n ih]; simpl.
  intros; constructor.
  intro m.
  constructor.
  intro h1.
  rewrite in_seq_iff in h1.
  omega.
  apply ih.
Qed.


Lemma no_dup_map_seg_inj :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    inj_seg f ->
    NoDup (map_seg f).
  intros U n.
  induction n as [|n ih]; simpl.
  intros; constructor.
  intros f h1.
  apply no_dup_app.
  apply ih.
  apply inj_from_S; auto.
  constructor.
  intro; contradiction.
  constructor.
  simpl.
  rewrite add_empty_sing.
  rewrite <- nin_iff.
  intro h2.
  rewrite <- list_to_set_in_iff in h2.
  rewrite in_map_seg_iff in h2.
  destruct h2 as [m [h2 h3]].
  rewrite seg_fun_from_S_compat in h3.
  apply h1 in h3; subst.
  apply lt_irrefl in h2; auto.
Qed.



Lemma list_to_set_eq_seg :
  forall (l:list nat) (n:nat),
    NoDup l ->
    list_to_set l = seg n ->
    n = length l.
  intros l n. revert l.
  induction n as [|n ih]; simpl.
  intros ? ? h1.
  rewrite seg_0_eq in h1.
  apply empty_set_nil in h1; subst; auto.
  intros l h1 h2.
  destruct (nil_dec' l) as [|h3]; subst.
  simpl in h2.
  pose proof (in_seg _ _ (lt_0_Sn n)) as h3.
  rewrite <- h2 in h3.
  contradiction.
  pose proof (lt_n_Sn n) as h4.
  apply in_seg in h4.
  rewrite <- h2 in h4.
  rewrite <- list_to_set_in_iff in h4.
  specialize (ih (remove nat_eq_dec n l)).
  hfirst ih.
  apply no_dup_remove; auto.
  specialize (ih hf).
  hfirst ih.
  rewrite <- subtract_remove_compat.
  rewrite h2.
  rewrite seg_Sn.
  rewrite sub_add_same_nin; auto.
  apply nin_seg.
  specialize (ih hf0).
  rewrite ih.
  rewrite length_remove_no_dup; auto.
  apply not_nil_lt in h3.
  rewrite <- (S_pred _ _ h3); auto.
Qed.


Lemma list_to_set_eq_seg_lt' :
  forall (l:list nat) (m n:nat),
    NoDup l ->
    list_to_set l = seg n ->
    In m l ->
    m < length l.
  intros l m n h1 h2 h3.
  pose proof (list_to_set_eq_seg_lt l m n h2 h3) as hl.
  rewrite <- list_to_set_seg_list_eq in h2.
  pose proof (incl_no_dup_le nat_eq_dec (seg_list n) l (no_dup_seg_list n) h1) as h4.
  hfirst h4.
  red; intros a h5.
  rewrite list_to_set_in_iff in h5.
  rewrite <- h2 in h5.
  rewrite <- list_to_set_in_iff in h5; auto.
  apply h4 in hf.
  rewrite length_seg_list in hf.
  eapply lt_le_trans.
  apply hl.
  assumption.
Qed.

  
Lemma no_dup_partition_faml_nat_seg_list_to_set_eq :
  forall (ll ll':faml nat) (m n:nat),
    NoDup ll ->
    NoDup ll' ->
    no_dup_mem ll ->
    no_dup_mem ll' ->
    list_to_set_injable (fun lx => In lx ll) ->
    list_to_set_injable (fun lx => In lx ll') ->
    list_to_set ll = list_to_set ll' ->
    partition_faml ll  (seg_list m) ->
    partition_faml ll' (seg_list n) ->
    m = n.
  intros ll ll' m n h1 h2 h3 h4 h5 h6 h7 h8 h9.
  pose proof (no_dup_partition_faml_nat_list_to_set_eq
                ll ll'
                (seg_list m) (seg_list n)
                h1 h2 h3 h4 h5 h6 h7 (no_dup_seg_list _) (no_dup_seg_list _) h8 h9) as h10.
  do 2 rewrite length_seg_list in h10; auto.
Qed.

  
Lemma no_dup_list_prod : 
  forall {T U:Type} (la:list T) (lb:list U),
    NoDup la -> NoDup lb -> NoDup (list_prod la lb).
intros T U la.
induction la as [|a la h1].
simpl. intros lb h1 h2. constructor.
simpl.
intros lb h2 h3.
inversion h2. subst.
specialize (h1 _ H2 h3).
assert (h4: Injective (fun y:U => (a, y))).
  red.
  intros x1 x2 h4.
  inversion h4. subst. reflexivity.
pose proof (no_dup_map_inj _ _ h3 h4) as h5.
apply no_dup_app; try assumption.
apply Extensionality_Ensembles.
red. split.
red.
intros x h6.
destruct h6 as [pr h6l h6r].
rewrite <- list_to_set_in_iff in h6l.
rewrite <- list_to_set_in_iff in h6r.
rewrite in_map_iff in h6l.
destruct h6l as [x h7].
destruct h7; subst.
rewrite in_prod_iff in h6r.
destruct h6r; contradiction.
auto with sets.
Qed.



Lemma list_of_lists_seqs_no_dup :
  forall {T:Type} (ll:faml T),
    no_dup_faml ll ->
    NoDup (list_of_lists_seqs ll).
  intros T ll.
  induction ll as [|al ll ih]; simpl.
  intro; apply no_dup_sing.
  intro h2.
  hfirst ih.
  red in h2; red.
  intros; apply h2; right; auto.
  specialize (ih hf).
  apply no_dup_map_inj.
  apply no_dup_list_prod.
  red in h2.
  apply h2; left; auto.
  assumption.
  red.
  intros p1 p2 p3.
  inversion p3.
  rewrite surjective_pairing.
  rewrite <- H0, <- H1.
  rewrite <- surjective_pairing; auto.
Qed.

Lemma list_of_lists_seqs_length :
  forall {T:Type} (ll:faml T) (l:list T),
         In l (list_of_lists_seqs ll) ->
         length l = length ll.
  intros T ll.
  induction ll as [|al ll ih]; simpl.
  intros; subst; auto.
  destruct H; subst; auto; try contradiction.
  intros l h1.
  rewrite in_map_iff in h1.
  destruct h1 as [p h1].
  destruct h1 as [h1 h2].
  subst.
  simpl.
  f_equal.
  apply ih.
  destruct p as [x y].
  rewrite in_prod_iff in h2.
  simpl.
  destruct h2; auto.
Qed.


Definition ens_list_to_fam {T:Type} (E:Ensemble (list T)) :
  Family T :=
  Im E (@list_to_set _).


         
Definition list_reps {T:Type} (E:Ensemble T) : Ensemble (list T) :=
  [l:list T | E = list_to_set l /\ NoDup l].

Lemma in_list_reps_eq :
  forall {T:Type} (E:Ensemble T) (l:list T),
    Inn (list_reps E) l ->
    list_to_set l = E.
  intros T E l h1.
  destruct h1 as [h1].
  destruct h1; auto.
Qed.

Lemma list_reps_empty :
  forall (T:Type),
    list_reps (Empty_set T) = Singleton nil.
  intro T.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [h1]. destruct h1 as [h1 h2].
  symmetry in h1. apply empty_set_nil in h1.
  subst. constructor.
  destruct h1.
  constructor; simpl.  split; try constructor; auto.
Qed.


Lemma list_reps_add :
  forall {T:Type} (E:Ensemble T) (a:T),
    ~Inn E a ->
    list_reps (Add E a) =
    FamilyUnion (
        Im (list_reps E)
           (fun l => [m:list T | NoDup m /\ In a m /\ remove eq_dec a m = l])).
  intros T E a h1.
  apply Extensionality_Ensembles; red; split; red; intros l h3.
  destruct h3 as [h3]. destruct h3 as [h3 h4].
  assert (h5:Inn (list_reps E) (remove eq_dec a l)).
    constructor. split.
    rewrite <- subtract_remove_compat.
    rewrite <- h3.
    rewrite sub_add_same_nin; auto.
    apply no_dup_remove; auto.
  apply family_union_intro with ((fun l0 : list T =>
                                    [m : list T | NoDup m /\ In a m /\ remove eq_dec a m = l0]) (remove eq_dec a l)).
  apply Im_intro with (remove eq_dec a l); auto.
  constructor.
  repeat split; auto.
  pose proof (Add_intro2 _ E a) as h6.
  rewrite  h3 in h6.
  rewrite list_to_set_in_iff; auto.

  destruct h3 as [S l h3 h4].
  destruct h3 as [S h3]. subst.
  destruct h4 as [h4].
  destruct h4 as [h4 [h5 h6]]. subst.
  destruct h3 as [h3]. destruct h3 as [h3 h6].
  subst.
  constructor.
  split; auto.
  rewrite <- subtract_remove_compat.
  rewrite add_sub_compat_in; auto.
  rewrite <- list_to_set_in_iff; auto.
Qed.

Definition interp {T:Type} (l:list T) (x:T) (n:nat) :=
  (firstn n l) ++ (x::nil) ++ (skipn n l).
 

Lemma inh_list_reps :
  forall {T:Type},
    forall (E:Ensemble T),
      Finite E ->
      Inhabited (list_reps E).
  intros T E h1.
  pose proof (finite_set_list_no_dup E h1) as h2.
  destruct h2  as [l h2].
  apply Inhabited_intro with l; constructor; auto.
Qed.

Lemma inh_not_nil_in_list_reps :
  forall {T:Type} (E:Ensemble T) (l:list T),
    Inhabited E -> Inn (list_reps E) l ->
    l <> nil.
  intros T E l h1 h2 h3.
  subst.
  destruct h2 as [h2]. destruct h2 as [h2 h3].
  subst.
  simpl in h1.
  destruct h1 as [x h1].
  contradiction.
Qed.

Lemma in_list_reps :
  forall {T:Type} (E:Ensemble T) (l:list T),
    Inn (list_reps E) l ->
    forall x:T,
      In x l ->
      Inn E x.
  intros T E l h1 x h2.
  destruct h1 as [h1]. destruct h1 as [h1 h3].
  subst.
  rewrite <- list_to_set_in_iff.
  assumption.
Qed.


(*See way below for finite_list_reps*)
Lemma list_reps_sing :
  forall {T:Type} (x:T),
    list_reps (Singleton x) = Singleton (x::nil).
  intros T a.
  apply Extensionality_Ensembles; red; split; red; intros l h1.
  destruct h1 as [h1 h2].
  destruct h1 as [h1 h2]. symmetry in h1.
  apply list_to_set_eq_sing in h1. subst.
  constructor. assumption.
  destruct h1.
  constructor. simpl. split.
  rewrite add_empty_sing; auto.
  constructor; simpl. intro h1. contradiction.
  constructor.
Qed. 


Lemma fun_list_to_set_eq :
  forall {T:Type} {U:Ensemble T->T->Set}
         (A:Ensemble T),
    Finite A ->
    (forall (x:T) (u u' : U A x), u = u') ->
    (forall (x:T) (l:list T), list_to_set l = A -> U A x) ->
    (forall x:T, U A x).
  intros T U A h1 h0.
  intros h2 x.
  pose proof (finite_inh_or_empty_dec A h1) as h5.
  specialize (h2 x). 
  pose proof (inh_list_reps _ h1) as h3.
  eapply fun_in_set_inh_const.
  refine (fun (l:list T) (pf:Inn (list_reps A) l) =>
            (h2 _ (in_list_reps_eq _ _ pf))).
  apply (h0 x).
  assumption.
Qed.


Lemma fun_dep_list_to_set_eq :
  forall {T:Type} {P:T->Prop} {U:forall (C:Ensemble T) (x:T), P x -> Set}
         (A:Ensemble T),
    Finite A ->
    (forall (x:T) (pfx:P x) (u u' : U A x pfx), u = u') ->
    (forall (x:T) (pfx:P x) (l:list T), list_to_set l = A -> U A x pfx) ->
    (forall (x:T) (pfx:P x), U A x pfx).
  intros T P U A h1 h0.
  intros h2 x hx.
  pose proof (finite_inh_or_empty_dec A h1) as h5.
  specialize (h2 x hx). 
  pose proof (inh_list_reps _ h1) as h3.
  eapply fun_in_set_inh_const.
  refine (fun (l:list T) (pf:Inn (list_reps A) l) =>
            (h2 _ (in_list_reps_eq _ _ pf))).
  apply (h0 x).
  assumption.
Qed.


Lemma fun_finite_list_to_set_eq : 
  forall
    {T:Type},
    let P := (@Finite T) in
    forall {U:forall (A B:Ensemble T), P A -> P B -> Set}
           (A:Ensemble T) (pfa:P A),
    (forall (B:Ensemble T) (pfb:P B) (u u' : U A B pfa pfb), u = u') ->
    (forall (B:Ensemble T) (pfb:P B) (l:list T), list_to_set l = A -> 
                                                 U A B pfa pfb) ->
    (forall (B:Ensemble T) (pfb:P B), U A B pfa pfb).
  intros T. simpl.
  intros U A h1 h0 h2 B h4.
  specialize (h2 B h4).
  pose proof (inh_list_reps _ h1) as h6.
  cut (forall (l:list T) (pf:Inn (list_reps A) l), U A B h1 h4); auto.
  intro h7.
  eapply fun_in_set_inh_const.
  refine (fun (l:list T) (pf:Inn (list_reps A) l) =>
            (h2 _ (in_list_reps_eq _ _ pf))).
  apply h0.
  assumption.
  intros l h7.
  eapply h2.
  destruct h7 as [h7]. destruct h7 as [h7 h8].
  symmetry.
  apply h7.
Qed.




Lemma list_prod_nil :
  forall {T U:Type} (l:list T),
         list_prod l (@nil U) =
         nil.
  intros ? ? l.
  induction l; auto.
Qed.


Lemma list_prod_eq_nil :
  forall {T U:Type} (l:list T) (l':list U),
    list_prod l l' = nil ->
    l = nil \/ l' = nil.
  intros T U l.
  destruct l as [|a l]; simpl.
  intros; left; auto.
  intros l' h1.
  right.
  apply app_eq_nil in h1.
  destruct h1 as [h1].
  apply (map_eq_nil _ _ h1).
Qed.

 

Lemma list_prod_cart_prod_compat :
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U)
         (la:list T) (lb:list U),
    list_to_set la = A ->
    list_to_set lb = B ->
    list_to_set (list_prod la lb) = cart_prod A B.
intros T U A B la.
induction la as [|a la h1].
simpl.
intros lb h1 h2.
rewrite <- h1.
rewrite cart_prod_empty.
reflexivity.
simpl.
intros lb h2 h3.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros pr h4.
rewrite <- list_to_set_in_iff in h4.
pose proof (in_app_or _ _ _ h4) as h5.
destruct h5 as [h5l | h5r].
rewrite in_map_iff  in h5l.
destruct h5l as [b h6].
destruct h6 as [h6l h6r]. subst.
constructor.
split.
simpl.
apply Add_intro2.
simpl.
rewrite <- list_to_set_in_iff. assumption.
rewrite (surjective_pairing pr) in h5r.
rewrite in_prod_iff in h5r.
constructor.
do 2 rewrite list_to_set_in_iff in h5r.
destruct h5r as [h5a h5b].
pose proof (Add_intro1 _ _ a _ h5a) as h6.
rewrite h2 in h6.
rewrite h3 in h5b.
split; assumption.
(* >= *)
red.
intros pr h4.
rewrite <- list_to_set_in_iff.

apply in_or_app.
rewrite <- h2 in h4.
rewrite <- h3 in h4.
destruct h4 as [h4].
destruct h4 as [h4l h4r].
inversion h4l as [? h5|? h6]. 
subst.
right.
rewrite (surjective_pairing pr).
rewrite <- list_to_set_in_iff in h4r.
rewrite <- list_to_set_in_iff in h5.
apply in_prod; assumption.
subst.
inversion h6.
rewrite (surjective_pairing pr).
left.
simpl.
rewrite in_map_iff.
exists (snd pr). subst.
split; auto.
rewrite list_to_set_in_iff.
assumption.
Qed.





Lemma nth_0 :
  forall {T:Type} (l:list T) (def:T),
    nth 0 l def =
    hd def l.
  intros ? l.
  induction l; simpl; auto.
Qed.

Lemma nth_indep' :
  forall {T:Type} (l:list T) (n:nat) (pf:n < length l) (def:T),
    nth n l def = nth n l (hd' l (lt_length _ _ pf)).
  intros; apply nth_indep; auto.
Qed.


Definition nth_lt {T:Type} (l:list T) (n:nat) : n < length l -> T :=
  match l  with
    | nil =>  (fun pf => (False_rect _ (lt_n_0 _ pf)))
    | a::l' => (fun _ => nth n (a::l') a)
  end.



Definition indl {T} (l:list T) i : option T :=
match lt_le_dec i (length l) with
| left pflt => Some (nth_lt l i pflt)
| right _ => None
end.

Notation "l '\_' i" := (indl l i) (at level 30, right associativity).


Lemma in_nth_lt :
  forall {T:Type} (l:list T) (n:nat) (pf:n < length l),
    In (nth_lt l n pf) l.
  intros T l.
  destruct l as [|a l]; simpl.
  intros; omega.
  intros n h1.
  destruct n as [|n].
  left; auto.
  apply lt_S_n in h1.
  right.
  apply nth_In; auto.
Qed.


Lemma in_infl :
  forall {T:Type} (l:list T) (n:nat),
    n < length l ->
    match (l \_ n) with
      | Some x => In x l
      | None => True
    end.
  intros T l n h1.
  unfold indl.
  lr_match_goal; fold hlr; destruct hlr  as [hl | hr].
  apply in_nth_lt.
  auto.
Qed.

Lemma nth_lt_compat : 
  forall {T:Type} (l:list T)
         (n:nat) (pflt:n < length l),
    let x' := nth_lt l n pflt in
    nth_lt l n pflt = nth n l x'.
  intros T l n h1; unfold nth_lt.
  simpl.
  destruct l as [|a l]. simpl in h1. pose proof (lt_n_0 _ h1). contradiction.
  destruct n as [|n]; simpl; auto.
  apply nth_indep.
  apply lt_S_n in h1; auto.
Qed.

Lemma nth_lt_cons :
  forall {T:Type} (l:list T) (a:T) (n:nat)
         (pflt:S n < S (length l)),
    let pflt' := lt_S_n _ _ pflt in
    nth_lt (a::l) (S n) pflt =
    nth_lt l n pflt'.
  intros T l a n h1; simpl; unfold nth_lt; simpl.
  destruct l as [|l c]; simpl.
  destruct n; destruct lt_n_0.
  destruct n; auto.
  apply nth_indep.
  simpl in h1.
  do 2 apply lt_S_n in h1; auto.
Qed.



Lemma nth_inj : 
  forall {T:Type} (l:list T),
    NoDup l ->
    forall (m n:nat) (d1 d2:T),
      m < length l -> n < length l ->
      nth m l d1 = nth n l d2 ->
      m = n. 
intros T l.
induction l as [|a l h1]; simpl.
intros; omega. 
intros h2 m n d1 d2 h3 h4 h5.
pose proof (no_dup_cons_nin _ _ h2) as hnin.
pose proof (no_dup_cons _ _ h2) as hdup.
destruct (le_lt_eq_dec _ _ h3) as [h6 | h7]; 
  destruct (le_lt_eq_dec _ _ h4) as [h8 | h9].
apply lt_S_n in h6. apply lt_S_n in h8.
apply (h1 hdup _ _ d1 d2 h6 h8).
destruct m; destruct n; auto.
apply nth_indep. assumption.
apply lt_S_n in h4.
pose proof (nth_In  _ d1 h4) as h11.
subst. 
pose proof (nth_indep). 
pose proof (nth_indep _  d1 d2 h4) as hi.
rewrite hi in h11. contradiction. subst.
apply lt_S_n in h3.
pose proof (nth_In _ d2 h3) as h12.
pose proof (nth_indep _ d1 d2 h3) as hi.
rewrite <- hi in h12.
contradiction.
apply h1 in h5. subst.
apply nth_indep; auto. 
assumption. apply lt_S_n. assumption. apply lt_S_n. assumption.  
apply S_inj in h9.
apply lt_S_n in h6. subst.
destruct (zerop (length l)) as [h8 | h9]. omega.
pose proof (S_pred _ _ h9) as h10.
rewrite h10 in h5.   
destruct m as [|m].
assert (h11:pred (length l) < length l). omega. 
pose proof (nth_In _ d2 h11) as h12.
rewrite <- h5 in h12.
contradiction.
assert (h11:pred (length l) < length l). omega.
apply lt_S_n in h3.
pose proof (h1 hdup _ _ d1 d2 h3 h11 h5). omega.
apply lt_S_n in h8. apply S_inj in h7. subst.
destruct (zerop (length l)) as [h9 | h10]. omega.
rewrite (S_pred _ _ h10) in h5.
destruct n as [|n].
assert (h11:pred (length l) < length l). omega.
pose proof (nth_In _ d1 h11) as h12. rewrite h5 in h12. contradiction.
apply lt_S_n in h4.
assert (h11:pred (length l) < length l). omega.
pose proof (h1 hdup _ _ d1 d2 h11 h4 h5) as h12.
omega.
apply S_inj in h7. apply S_inj in h9. congruence.
Qed.


Lemma nth_lt_inj : 
  forall {T:Type} (l:list T) (pfnd:NoDup l) (m n:nat) 
         (pfm: m < length l) (pfn: n < length l),
    nth_lt l m pfm = nth_lt l n pfn ->
    m = n.
  intros T l h1 m n h2 h3 h4.
  pose proof (nth_lt_compat _ _ h2) as h5.
  pose proof (nth_lt_compat _ _ h3) as h6.
  simpl in h5, h6.
  rewrite h6, h5 in h4.
  eapply nth_inj; auto. apply h1. apply h2. apply h3. apply h4.
Qed.

Lemma inj_seg_nth_lt :
  forall {T:Type} (l:list T),
    NoDup l ->
    inj_seg (nth_lt l).
  intros T l h1.
  red; intros x y h2 h3 h4.
  apply nth_lt_inj in h4; auto.
Qed.


Lemma nth_lt_0 :
  forall {T:Type} (l:list T) (a:T),
    NoDup (a::l) ->
    forall (n:nat) (pfl:n < length (a::l)),
      nth_lt (a::l) n pfl = a ->
      n = 0.
  intros T l a h1 n h2 h3.
  destruct n; auto.
  rewrite nth_lt_cons in h3.
  pose proof (nth_lt_compat l  n (lt_S_n n (length l) h2)) as h4.
  simpl in h4. 
  simpl in h2.
  subst.
  pose proof (in_nth_lt l n (lt_S_n _ _ h2)) as h3.
  pose proof (lt_S_n _ _ h2) as h5.
  pose proof (no_dup_cons_nin _ _ h1) as h8.
  contradiction.
Qed.

Lemma nth_lt0 :
  forall {T:Type} (l:list T) (pf:l <> nil),
    let pf' := not_nil_lt l pf in
    nth_lt l 0 pf' = hd' l pf.
  intros T l.
  destruct l as [|a l]; simpl; auto.
  intro h1; contradict h1; auto.
Qed.

Lemma nth_lt0' :
  forall {T:Type} (l:list T) (pf:0 < length l),
    let pf' := lt_length _ _ pf in
    nth_lt l 0 pf = hd' l pf'.
  intros T l.
  destruct l as [|a l].
  simpl; intros h1; contradict (lt_irrefl _ h1).
  intros; simpl; auto.
Qed.


Lemma nth_lt_functional1 :
  forall {T:Type} (l l':list T) (n:nat) (pfeq: l = l')
         (pf: n < length l) (pf' : n < length l'),
    nth_lt l n pf = nth_lt l' n pf'.
  intros T l l' n  h0 h1 h2. subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.


Lemma nth_lt_functional2 :
  forall {T:Type} (l:list T) (n n':nat) (pfeq: n = n')
         (pf: n < length l) (pf' : n' < length l),
    nth_lt l n pf = nth_lt l n' pf'.
  intros T l n n' h0 h1 h2. subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.

Lemma nth_lt_functional3 :
  forall {T:Type} (l:list T) (n:nat) (pf1 pf2: n < length l),
    nth_lt l n pf1 = nth_lt l n pf2.
  intros T l n h1 h2.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.

Lemma nth_lt_0' :
  forall {T:Type} (l:list T) (a:T) (pf:0 < length (a::l)),
    nth_lt (a::l) 0 pf = a.
  intros; simpl; auto.
Qed.

Lemma nth_lt_0_hd' :
  forall {T:Type} (l:list T) (pf:0 < length l),
    nth_lt l 0 pf  = hd' l (lt_length _ _ pf).
  intros T l h1.
  destruct l as [|a l]; simpl.
  f_equal. apply proof_irrelevance.
  reflexivity.
Qed.


  

Lemma nth_lt_cons' :
  forall {T:Type} (a:T) (l:list T) (n:nat) (pf:n < S (length l))
         (pflt : 0 < n),
    nth_lt (a::l) n pf =
    nth_lt l (pred n) (mono_pred_lt _ _ pf pflt).
  intros T a l n h1 h2.
  simpl.
  destruct n as [|n]; simpl.
  pose proof (lt_irrefl _  h2). contradiction.
  symmetry.
  gen0.
  intro h3. simpl in h3.
  rewrite  nth_lt_compat.
  apply nth_indep; auto.
Qed.


Lemma nth_lt_S :
  forall {T:Type}  (l:list T) (n:nat) (pf:S n < length l),
    nth_lt l (S n) pf =
    nth_lt (tl l) n (lt_length_tl _ _ pf).
  intros T l n h1.
  pose proof (hd_compat' l  (lt_length _ _ h1)) as h3.
  simpl in h3.
  generalize h1. rewrite h3.
  intro. rewrite (nth_lt_cons' _ _ _  _ (lt_0_Sn _)) at 1.
  simpl.
  f_equal. apply proof_irrelevance.
Qed.


Lemma nth_lt_cons_in :
  forall {T:Type} (l:list T) (a:T) (n:nat) (pf0: 0 < n)
         (pfl:n < S (length l)),
   nth_lt (a::l) n pfl = a -> In a l.
  intros T l a n h1 h2 h3.
  pose proof (nth_lt_cons' a l n h2 h1) as h4. 
  rewrite h3 in h4.
  pose proof (nth_lt_compat l (pred n) (mono_pred_lt n (S (length l)) h2 h1)) as h5.
  simpl in h5.
  destruct h5 as [h5 h6].
  subst.
  apply in_nth_lt.
Qed.
 

Lemma nth_lt_1 :
  forall {T:Type} (l:list T) (pf: 1 < length l)
         (pftl:tl l <> nil),
    nth_lt l 1 pf =
    hd' (tl l) pftl.
  intros T l.
  induction l as [|a l h1]; simpl.
  intros; omega.
  intros h2 h3.
  rewrite nth_0.
  erewrite hd_hd_compat'.
  reflexivity.
Qed.


Lemma nth_lt_sing :
  forall {T:Type} (a:T) (m:nat) (pf:m < length (a::nil)),
    nth_lt (a::nil) m pf = a.
  intros T a m h1. simpl in h1.
  assert (h2:m = 0). omega. subst.
  rewrite nth_lt_0_hd'.
  rewrite hd_cons'; auto.
Qed.


Lemma nth_lt_dup : 
  forall {T:Type} (l:list T) (m n:(length l) @),
    m <> n ->
    nth_lt _ _ (proj2_sig m) =
    nth_lt _ _ (proj2_sig n) ->
    ~NoDup l.
  intros T l m n h1 h2 h3.
  pose proof (nth_lt_compat _ _ (proj2_sig m)) as h4. simpl in h4.
  pose proof (nth_lt_compat _ _ (proj2_sig n)) as h5. simpl in h5.
  rewrite h2 in h4.
  rewrite h4 in h5 at 1.
  pose proof (nth_inj  _ h3 _ _ _ _ (proj2_sig m) (proj2_sig n) h5) as h6.
  apply proj1_sig_injective in h6.
  contradict h1; auto.
Qed.


Lemma nth_lt_dup' : 
  forall {T:Type} (l:list T) (m n:nat)
         (pfm : m < length l)
         (pfn : n < length l),
    m <> n ->
    nth_lt _ _ pfm =
    nth_lt _ _ pfn ->
    ~NoDup l.
  intros T l m n pfm pfn h1 h2 h3.
  pose proof (nth_lt_compat _ _ pfm) as h4. simpl in h4.
  pose proof (nth_lt_compat _ _ pfn) as h5. simpl in h5.
  rewrite h2 in h4.
  rewrite h4 in h5 at 1.
  pose proof (nth_inj  _ h3 _ _ _ _ pfm pfn h5); subst.
  contradict h1; auto.
Qed.


          


Lemma nth_lt_S_neq :
  forall {T:Type} (l:list T) (n:nat) (pfnil: l <> nil)
         (pf: n < length l),
    nth_lt l n pf <> hd' l pfnil ->
    In (nth_lt l n pf) (tl l).
  intros T l.
  induction l as [|a l h1]; simpl; intros n h2 h3 h6; try omega.
  revert h6. revert h2 h1 h3. revert a l.
  induction n as [|n h5].
  intros a l h2 h0 h3 hn.
  contradict hn; auto.
  intros; apply nth_In.
  apply lt_S_n in h3; auto.
Qed.

Lemma nth_lt_map :
  forall {T U:Type} (l:list T) (f:T->U) (n:nat) (pf:n < length l)
         (pf': n < length (map f l)),
    nth_lt (map f l) n pf' =
    f (nth_lt l n pf).
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros ? ? h1. pose proof (lt_n_0 _  h1). contradiction.
  intros f n h1 h2.
  destruct n as [|n]; simpl; auto.
  apply lt_S_n in h1.  apply lt_S_n in h2.
  specialize (ih f n h1 h2). 
  rewrite nth_lt_compat in ih.
  erewrite nth_indep.
  rewrite ih at 1.
  f_equal.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  assumption.
Qed.


Lemma nth_lt_map' :
  forall {T U:Type} (l:list T) (f:T->U) (n:nat)
         (pf': n < length (map f l)),
    let pf := lt_eq_trans _ _ _ pf' (map_length f l) in
    nth_lt (map f l) n pf' =
    f (nth_lt l n pf).
  intros T U l f n h1 h2.
  erewrite nth_lt_map.
  reflexivity.
Qed.

Lemma map_eq_iff :
  forall {T U:Type} (l:list T) (f:T->U) (l':list U),
    map f l = l' <->
    (exists pfeq:length l = length l', 
       (forall (n:nat) (pflt:n < length l),
          f (nth_lt l n pflt) =
          nth_lt l' n (lt_eq_trans _ _ _ pflt pfeq))).
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros.
  split; intro h1.
  subst; simpl.
  exists eq_refl.
  intros; omega.
  destruct h1 as [h1 h2].
  pose proof (length_eq0 _ (eq_sym h1)); auto.
  intros f l'.
  split; intro h1.
  pose proof (f_equal (@length _) h1) as h2.
  simpl in h2.
  rewrite map_length in h2.
  exists h2.
  intros n h3.
  destruct n as [|n]; simpl.
  rewrite nth_lt_0_hd'.
  erewrite hd_compat' in h1.
  inversion h1 as [h4].
  rewrite h4 at 1.
  reflexivity.
  pose proof (lt_S_n _ _ h3) as h4.
  symmetry.
  gen0.
  rewrite <- h1.
  intro h5; simpl.
  rewrite map_nth; auto.
  destruct h1 as [h1 h2].
  destruct l' as [|a' l'].
  simpl in h1; discriminate.
  simpl in h1.
  pose proof (S_inj _ _ h1) as h3.
  pose proof (h2 0 (lt_0_Sn _)) as h4.
  simpl in h4; subst.
  f_equal.
  specialize (ih f l').
  rewrite ih.
  exists h3.
  intros n h4.
  specialize (h2 (S n)).
  hfirst h2.
  apply lt_n_S; auto.
  specialize (h2 hf).
  simpl in h2.
  rewrite nth_lt_compat.
  erewrite nth_indep.
  rewrite  h2 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  omega.
  assumption.
Qed.

Lemma maps_eq_iff :
  forall {T U V:Type} (l:list T) (l':list U) (f:T->V) (g:U->V),
    map f l = map g l' <->
    exists (pfe:length l = length l'),
      forall i (pflt:i < length l),
         let pflt' := lt_eq_trans _ _ _ pflt pfe in
         f (nth_lt l  i pflt) =
         g (nth_lt l' i pflt').
  intros T U V l l' f g.
  rewrite map_eq_iff.
  split; intro h1; destruct h1 as [h1 h2].
  pose proof h1 as h1'.
  rewrite map_length in h1'.
  exists h1'.
  intros i h3 h4.
  rewrite h2.
  rewrite (nth_lt_map _ _ _ h4); auto.
  ex_goal.
  rewrite map_length; auto.
  exists hex.
  intros n h3.
  rewrite (nth_lt_map _ _ _ (lt_eq_trans _ _ _ h3 h1)).
  rewrite h2; auto.
Qed.


Lemma nth_lt_app1 :
  forall {T:Type} (l l':list T) (n:nat)
         (pfn:n < length l) (pfn' : n < (length (l++l'))),
    nth_lt (l++l') n pfn' =
    nth_lt l n pfn.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros l' n h1. pose proof (lt_n_0 _ h1). contradiction.
  intros l' n h1 h2.
  destruct n as [|n]; simpl; auto.
  apply lt_S_n in h1.
  apply lt_S_n in h2.
  specialize (ih _ _ h1 h2).
  rewrite nth_lt_compat in ih.
  erewrite  nth_indep.
  rewrite ih at 1.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  assumption.
Qed.




Lemma nth_lt_app2 :
  forall {T:Type} (l l':list T) (n:nat)
         (pfn:n >= length l) (pfn' : n < (length (l++l'))),
    nth_lt (l++l') n pfn' =
    nth_lt l' (n - length l) (lt_length_app_ge_impl _ _ _ pfn pfn').
  intros T l.
  induction l as [|a l ih]; simpl.
  intros l' n h1 h2.
  symmetry.
  gen0. simpl.
  pose proof (minus_n_O n) as h3.
  rewrite <- h3.
  intro h4.
  f_equal. apply proof_irrelevance. 
  intros l' n h1 h2.
  destruct n as [|n]; simpl.
  omega.
  pose proof (lt_S_n _ _ h2) as h2'.
  pose proof (le_S_n _ _ h1) as h1'.
  specialize (ih _ _ h1' h2').
  rewrite nth_lt_compat in ih.
  erewrite nth_indep.
  rewrite ih at 1.
  f_equal. apply proof_irrelevance.
  apply lt_S_n in h2.
  assumption.
Qed.


Lemma nth_lt_pred_length :
  forall {T:Type} (l:list T) (pf:pred (length l) < length l),
    nth_lt l (pred (length l)) pf =
    lst l (lt_length _ _ pf).
  intros T l h1.
  pose proof (lst_compat l (lt_length _ _ h1)) as h2.
  simpl in h2.
  pose proof (nth_lt_functional1 _ _ (pred (length l)) h2 h1) as h3.
  hfirst h3.
  rewrite <- h2; auto.
  specialize (h3 hf).
  rewrite h3.
  erewrite nth_lt_app2.
  simpl.
  rewrite length_removelast.
  rewrite minus_same; auto.

  Grab Existential Variables.
  rewrite length_removelast; auto.
Qed.

Lemma lst_eq : 
  forall {T:Type} (l:list T) (pfn:l <> nil),
    lst l pfn = 
    nth_lt l (pred (length l)) (not_nil_pred_lt _ pfn).
  intros; rewrite nth_lt_pred_length; f_equal; apply proof_irrelevance.
Qed.


  
Lemma all_sing_nth_lt :
  forall {T:Type} (l:list T) (a:T),
    all_sing l a ->
    forall (m:nat)
           (pfm:m < length l),
      nth_lt l m pfm = a.
  intros T l.
  destruct l as [|a l]; simpl.
  intros; omega.
  intros a' h1 m h2.
  rewrite all_sing_cons_iff in h1.
  destruct h1 as [? h1]; subst. 
  destruct m as [|m]; auto.
  apply all_sing_nth; auto.
Qed.



Lemma nth_lt_seg_list :
  forall (m n:nat) (pf:m < length (seg_list n)),
    nth_lt (seg_list n) m pf = m.
  intro m.
  induction m as [|m ih]; simpl; auto; intros n h1.
  revert h1.
  induction n as [|n]; simpl; auto.
  intros; omega.
  rewrite nth_lt_S.
  unfold seg_list.
  gen0. rewrite tl_seq.
  intro h2.
  pose proof h2 as h2'. rewrite seq_length in h2'.
  specialize (ih (pred n)).
  hfirst ih.
  rewrite length_seg_list; auto.
  specialize (ih hf).
  gen0.
  rewrite seq_eq_seq_0.
  intros h3; simpl.
  erewrite nth_lt_map.
  rewrite ih at 1.
  reflexivity.
Qed.

Lemma seg_list_eq_n_seg_list :
  forall m n,
    seg_list m = n :: seg_list n ->
    m = 1.
  intros m.
  destruct m as [|m]; simpl.
  intros n h1.
  pose proof (in_eq n (seg_list n)) as h2.
  rewrite <- h1, seg_list0 in h2.
  simpl in h2; contradiction.
  intros n h1.
  f_equal.
  pose proof (lt_0_Sn m) as h2.
  pose proof (nth_lt_functional1 _ _ 0 h1) as h3.
  hfirst h3.
  rewrite length_seg_list; auto with arith.
  specialize (h3 hf (lt_0_Sn _)).
  simpl in h3; subst. 
  rewrite seg_list0 in h1.
  pose proof (lt_n_Sn m) as h3.
  apply in_seg in h3.
  rewrite <- list_to_set_seg_list_eq in h3.
  rewrite h1 in h3.
  simpl in h3.
  rewrite add_empty_sing in h3.
  rewrite in_sing_iff in h3; auto.
Qed.


Lemma nth_lt_seq :
  forall (m n r:nat) (pf:r < length (seq m n)),
    nth_lt (seq m n) r pf = m + r.
  intros m n.
  revert m.
  induction n as [|n ih].
  simpl; intros; omega.
  intros m r h1.
  pose proof h1 as h1'.
  rewrite seq_length in h1'.
  gen0.
  rewrite seq_S; auto with arith.
  intros h2.
  destruct r as [|r].
  erewrite nth_lt_functional3, nth_lt0.
  simpl; auto.
  simpl.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite ih at 1; omega.
  rewrite seq_length.
  omega.

  Grab Existential Variables.
  rewrite seq_length; omega.
  intro; discriminate.
Qed.

 
Lemma nth_lt_map_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U) (m:nat)
         (pf:m < n),
    let pf' := lt_eq_trans _ _ _ pf (eq_sym (length_map_seg f)) in
    nth_lt (map_seg f) m pf' =
    f m pf.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros f m h1.
  simpl in ih.
  red in h1.
  pose proof h1 as h1'. apply le_S_n in h1'. 
  destruct (le_lt_eq_dec _ _ h1') as [h2 | h2].
  specialize (ih (seg_fun_from_S f) m h2).
  erewrite nth_lt_app1.
  unfold seg_fun_from_S in ih.
  rewrite ih at 1.
  f_equal. apply proof_irrelevance.
  subst.
  erewrite nth_lt_app2.
  simpl.
  rewrite length_map_seg.
  rewrite minus_same.
  f_equal.
  apply proof_irrelevance.
  Grab Existential Variables.
  rewrite length_map_seg; auto with arith.
Qed.


Lemma nth_lt_map_seg' :
  forall {U:Type} {n:nat} (f:seg_fun n U) (m:nat)
         (pf:m < length (map_seg f)),
    let pf' := lt_eq_trans _ _ _ pf (length_map_seg f) in
    nth_lt (map_seg f) m pf =
    f m pf'.
  intros U n f m h1.
  simpl.
  erewrite nth_lt_functional3.
  erewrite nth_lt_map_seg.
  reflexivity.
Qed.


Lemma lst_map_seg :
  forall {U:Type} {n:nat} {f:seg_fun n U}
         (pfn:map_seg f <> nil),
    lst (map_seg f) pfn =
    f (pred n) (map_seg_lt _ pfn).
  intros U n f h1.
  rewrite lst_eq.
  rewrite nth_lt_map_seg'.
  apply f_equal_dep.
  rewrite length_map_seg; auto.
Qed.


Lemma last_map_seg :
  forall {U:Type} {n:nat} (def:U)
         (f:seg_fun n U)
         (pf:0 < n),
    last (map_seg f) def =
    f (pred n) (lt_pred_n _ _ pf).
  intros; erewrite last_compat, lst_eq, nth_lt_map_seg'; apply f_equal_dep.
  rewrite length_map_seg; auto.

  Grab Existential Variables.
  intro h1.
  apply map_seg_eq_nil in h1.
  subst.
  apply lt_irrefl in pf; contradiction.
Qed.


Lemma map_to_map_seg :
  forall {T U:Type} (f:T->U) (l:list T),
    map f l =
    map_seg (fun i (pf:i < length l) =>
               f (nth_lt _ _ pf)).
  intros T U f l.
  rewrite map_eq_iff.
  ex_goal.
  rewrite length_map_seg; auto.
  exists hex.
  intros n h1.
  symmetry.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  reflexivity.
Qed.


Lemma nth_lt_map_in_dep : 
  forall {T U:Type} {l:list T} (f:fun_in_dep l U)
         (n:nat) (pf:n < length l) (pf':n < length (map_in_dep f)),
    nth_lt (map_in_dep f) n pf' =
    f (nth_lt l n pf) (in_nth_lt _ _ _).
  intros T U l.
  induction l as [|a l h1]; simpl.
  intros; omega.
  intros f n.
  destruct n as [|n]; simpl.
  intros h2 h3.
  f_equal. apply proof_irrelevance.
  intros h2 h3.
  pose proof h2 as h2'.
  pose proof (lt_S_n _ _ h2) as h4.
  pose proof (lt_S_n _ _ h3) as h5.
  specialize (h1 (fun_from_cons f) _ h4 h5).
  rewrite nth_lt_compat in h1 at 1.
  erewrite nth_indep.
  rewrite h1 at 1.
  rewrite fun_from_cons_compat.
  apply f_equal_prop_dep.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  assumption.
Qed.

         

Lemma nth_lt_map_in_dep' : 
  forall {T U:Type} {l:list T} (f:fun_in_dep l U)
         (n:nat)(pf':n < length (map_in_dep f)),
    let pf := lt_eq_trans _ _ _ pf' (length_map_in_dep f) in
    nth_lt (map_in_dep f) n pf' =
    f (nth_lt l n pf) (in_nth_lt _ _ _).
  intros; erewrite nth_lt_map_in_dep.
  reflexivity.
Qed.


Lemma map_seg_list :
  forall {U:Type} (f:nat->U) (n:nat),
    map f (seg_list n) =
    map_seg (fun_to_seg f n).
  intros U f n.
  rewrite map_eq_iff.
  ex_goal.
  rewrite length_seg_list.
  rewrite length_map_seg; auto.
  exists hex.
  intros m h1.
  symmetry.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg at 1.
  unfold fun_to_seg.
  f_equal.
  rewrite nth_lt_seg_list; auto.

  Grab Existential Variables.
  rewrite length_seg_list in h1; auto.
Qed.


Lemma map_in_dep_eq_iff :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U) (l':list U),
    map_in_dep f = l' <->
    (exists pfeq:length l = length l',
      forall (n:nat) (pflt:n < length l),
        f (nth_lt l n pflt) (in_nth_lt _ _ _) =
        nth_lt l' n (lt_eq_trans _ _ _ pflt pfeq)).
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros f l'.
  split; intro h1.
  subst.
  simpl.
  exists eq_refl.
  intros; omega.
  destruct h1 as [h1].
  pose proof (length_eq0 _ (eq_sym h1)); auto.
  intros f l'. 
  split; intro h1.
  ex_goal.
  pose proof (f_equal (@length _) h1) as h2.
  simpl in h2.
  rewrite length_map_in_dep in h2; auto.
  exists hex.
  intros n h2.
  destruct n as [|n]; simpl.
  rewrite nth_lt_0_hd'.
  symmetry.
  gen0. rewrite <- h1.
  intro h3.
  simpl.
  f_equal. apply proof_irrelevance.
  pose proof h2 as h2'. apply lt_S_n in h2'.
  destruct l' as [|a' l'].
  simpl in hex; discriminate.
  simpl in hex.
  pose proof hex as hex'. apply S_inj in hex'.
  simpl.
  inversion h1 as [h3]; subst.
  erewrite nth_indep.
  erewrite <- nth_lt_compat at 1.
  rewrite nth_lt_map_in_dep at 1.
  unfold fun_from_cons.
  apply f_equal_prop_dep.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  rewrite length_map_in_dep; auto.
  destruct h1 as [h1 h2].
  pose proof (h2 0 (lt_0_Sn  _)) as h2'.
  simpl in h2'.
  rewrite nth_lt_0_hd' in h2'.
  erewrite hd_compat'.
  rewrite <- h2' at 1.
  f_equal.
  f_equal. apply proof_irrelevance.
  specialize (ih (fun_from_cons f) (tl l')).
  rewrite ih.
  destruct l' as [|a' l']; simpl.
  simpl in h1; discriminate.
  simpl in h1.
  pose proof h1 as h1'. apply S_inj in h1'.
  exists h1'.
  intros n h3.
  specialize (h2 (S n) (lt_n_S _ _ h3)).
  simpl in h2.
  unfold fun_from_cons.
  rewrite nth_lt_compat.
  erewrite nth_indep.
  rewrite <- h2 at 1.
  apply f_equal_prop_dep.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  eapply lt_eq_trans. apply h3.
  assumption.
  Grab Existential Variables.
  assumption.
  rewrite length_map_in_dep; auto.
Qed.

Lemma map_in_dep_to_map_seg :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U),
    map_in_dep f =
    map_seg (fun i (pf:i < length l) =>
               f _ (in_nth_lt _ _ pf)).
  intros T U l f.
  rewrite map_in_dep_eq_iff.
  ex_goal.
  rewrite length_map_seg; auto.
  exists hex.
  intros n h1.
  symmetry.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  reflexivity.
Qed.


Lemma map_in_deps_eq_iff :
  forall {T U V:Type} (l:list T) (l':list U)
         (f:fun_in_dep l V) (g:fun_in_dep l' V),
    map_in_dep f = map_in_dep g  <->
    exists (pfe:length l = length l'),
      forall i (pflt:i < length l),
         let pflt' := lt_eq_trans _ _ _ pflt pfe in
         f (nth_lt l  i pflt) (in_nth_lt _ _ _) =
         g (nth_lt l' i pflt') (in_nth_lt _ _ _).
  intros T U V l l' f g.
  rewrite map_in_dep_eq_iff.
  split; intro h1; destruct h1 as [h1 h2].
  ex_goal.
  pose proof h1 as h1'.
  rewrite length_map_in_dep in h1'; auto.
  exists hex.
  intros i h3 h4.
  rewrite h2.
  rewrite nth_lt_map_in_dep'.
  apply f_equal_dep.
  apply nth_lt_functional3.
  ex_goal.
  rewrite length_map_in_dep; auto.
  exists hex.
  intros n h3.
  rewrite nth_lt_map_in_dep'.
  rewrite h2.
  apply f_equal_dep.
  apply nth_lt_functional3.
Qed.



Lemma nth_lt_map_sig :
  forall {T} (l:list T) i (pf:i < length (map_sig l)),
    nth_lt (map_sig l) i pf  =
    exist (fun c => In c l)
           (nth_lt l i (lt_eq_trans _ _ _
                                   pf
                                   (length_map_sig _)))
           (in_nth_lt _ _ _).
  unfold map_sig; intros; rewrite nth_lt_map_in_dep';
  apply proj1_sig_injective; simpl; f_equal; apply proof_irrelevance.
Qed.


Lemma map_sig_compat :
  forall {T:Type} (l:list T),
    map (@proj1_sig _ _) (map_sig l) = l. 
  intros T l; rewrite map_eq_iff;
  exists (length_map_sig _).
  intros; rewrite nth_lt_map_sig; simpl; auto.
Qed.

Lemma map_map_sig_functional :
  forall {T U:Type} (l l':list T) (f:{t:T | In t l}->U)
         (pf:l = l'),
    map f (map_sig l) = map (transfer_dep (U:= (fun xl:list T => {t:T | In t xl}->U)) pf f) (map_sig l').
  intros; subst; rewrite transfer_dep_eq_refl; auto.
Qed.

 
Lemma list_of_lists_seqs_not_nil :
  forall {T:Type} (ll:faml T),
    p_mem_l (fun l => l <> nil) ll ->
    list_of_lists_seqs ll <> nil.
  intros T ll.
  induction ll as [|la ll ih]; simpl.
  intros; intro; discriminate.
  intros h1.
  hfirst ih.
  eapply p_mem_l_cons.
  apply h1.
  specialize (ih hf).
  intro h2.
  apply map_eq_nil_iff in h2.
  apply list_prod_eq_nil in h2.
  destruct h2 as [|h2]; subst.
  red in h1.
  specialize (h1 nil (in_eq _ _)).
  contradict h1; auto.
  contradiction.
Qed.



(*Elementary binary vector of width [n] and support [m].*)
Definition ebv (m n:nat) : bv :=
  map_seg (fun i (pf:i < n) => if (nat_eq_dec i m) then true else false).

(*Complement of a binary vector.*)
Definition nbv (v:bv) : bv :=
  map negb v.

Definition tbv {v v':bv} (pfeq:length v = length v') : bv :=
  map_seg (fun i (pfi:i < length v) =>
             let pfi' := lt_eq_trans _ _ _ pfi pfeq in
             (andb (nth_lt v i pfi)
                   (nth_lt v' i pfi'))).

Definition pbv {v v':bv} (pfeq:length v = length v') : bv :=
  map_seg (fun i (pfi:i < length v) =>
             let pfi' := lt_eq_trans _ _ _ pfi pfeq in
             (orb (nth_lt v i pfi)
                  (nth_lt v' i pfi'))).

Definition xbv {v v':bv} (pfeq:length v = length v') : bv :=
  map_seg (fun i (pfi:i < length v) =>
             let pfi' := lt_eq_trans _ _ _ pfi pfeq in
             (xorb (nth_lt v i pfi)
                   (nth_lt v' i pfi'))).
                       
(*Complement of an elementary binary vector.*)
Definition cebv (m n:nat) : bv :=
  nbv (ebv m n).

Definition cons_mask (n:nat) : bv :=
  cebv 0 n.

Definition inds_to_bv {T:Type} (l:list T) (ln:list nat):
  bv :=
  map_seg (fun i (pfi:i < length l) =>
             if (in_dec nat_eq_dec i ln)
             then true
             else false).


  
Lemma in_bv0s :
  forall (v:bv) (m:nat)
         (pfm:In m (bv0s v)),
    let pfv := in_bv0s_lt _ _ pfm in
    nth_lt v _ pfv = false.
  intro v.
  induction v as [|b v ih]; simpl.
  intros; contradiction.
  intros m h1.
  unfold bv0s in h1; simpl in h1.
  destruct b.
  rewrite bv0s_aux_S in h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [? [h1 ?]]; subst.
  specialize (ih _ h1).
  simpl in ih.
  rewrite <- ih.
  rewrite nth_lt_compat.
  apply nth_indep.
  pose proof h1 as h1'.
  apply in_bv0s_lt in h1'; auto.
  destruct m as [|m]; auto.
  destruct h1 as [|h1].
  discriminate.
  rewrite bv0s_aux_S in h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [? [h1 h2]].
  apply S_inj in h2; subst.
  specialize (ih _ h1).
  simpl in ih.
  rewrite <- ih.
  rewrite nth_lt_compat.
  apply nth_indep.
  pose proof h1 as h1'.
  apply in_bv0s_lt in h1'; auto.
Qed.



Lemma in_bv1s :
  forall (v:bv) (m:nat)
         (pfm:In m (bv1s v)),
    let pfv := in_bv1s_lt _ _ pfm in
    nth_lt v _ pfv = true.
  intro v.
  induction v as [|b v ih]; simpl.
  intros; contradiction.
  intros m h1.
  unfold bv1s in h1; simpl in h1.
  destruct b.
  
  destruct m as [|m]; auto.
  destruct h1 as [|h1].
  discriminate.

  rewrite bv1s_aux_S in h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [? [h1 h2]].
  apply S_inj in h2; subst.
  specialize (ih _ h1).
  simpl in ih.
  rewrite <- ih.
  rewrite nth_lt_compat.
  apply nth_indep.
  pose proof h1 as h1'.
  apply in_bv1s_lt in h1'; auto.
  
  rewrite bv1s_aux_S in h1.
  rewrite in_lS_iff in h1.
  destruct h1 as [? [h1 ?]]; subst.
  specialize (ih _ h1).
  simpl in ih.
  rewrite <- ih.
  rewrite nth_lt_compat.
  apply nth_indep.
  pose proof h1 as h1'.
  apply in_bv1s_lt in h1'; auto.
Qed.


(*Think of [coarse_list] as kind of a "coarsening" so that 
for instance [coarse_list [1, 2, 5, 0, 8] [3, 1, 4, 2, 5, 6, 0, 8, 9 10]],
i.e. the first argument is contained in the second argument, 
albeit interspersed by other elements.*)
Inductive coarse_list {T:Type} : list T -> list T -> Prop :=
  | coarse_list_nil : forall l:list T, coarse_list nil l
  | coarse_list_cons1 : forall (l:list T) (l':list T) (a:T),
                       coarse_list l l' -> coarse_list l (a::l')
  | coarse_list_cons2 : forall (l:list T) (l':list T) (a:T),
                       coarse_list l l' -> coarse_list (a::l) (a::l').



Lemma coarse_list_refl :
  forall {T:Type} (l:list T),
    coarse_list l l.
  intros T l.
  induction l as [|a l ih]; simpl. 
  constructor.
  apply coarse_list_cons2; auto.
Qed.
  
    
Lemma length_coarse_list :
  forall {T:Type} (l l':list T),
    coarse_list l l' ->
    length l <= length l'.
  intros ? ? ? h1; induction h1; simpl; auto with arith.
Qed.

Lemma coarse_list_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    coarse_list (remove pfdt a l) l.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; constructor.
  intro a'.
  destruct (h1 a' a) as [|h2]; subst.
  apply coarse_list_cons1.
  apply ih.
  apply coarse_list_cons2.
  apply ih.
Qed.

Lemma le_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    length (remove pfdt a l) <= length l.
  intros; apply length_coarse_list; apply coarse_list_remove.
Qed.


Lemma lt_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    In x l ->
    length (remove pfdt x l) < length l.
  intros T h1 l.
  induction l as [|a l ih].
  intros; contradiction.
  intros x h2.
  simpl.
  simpl in h2.
  destruct (h1 x a) as [|h3]; subst.
  pose proof (le_remove h1 l a) as h4.
  omega.
  destruct h2 as [|h2]; subst.
  contradict h3; auto.
  simpl.
  apply lt_n_S.
  apply ih; auto.
Qed.



Lemma coarse_list_nil2_iff :
  forall {T:Type} (l:list T),
    coarse_list l nil <->
    l = nil.
  intros ? l; destruct l; split; intro h2; try inversion h2; subst; auto; try constructor.
Qed.

Lemma not_coarse_list_cons_nil :
  forall {T:Type} (l:list T) (a:T),
    ~coarse_list (a::l) nil.
  intros ? ? ? h1; inversion h1.
Qed.


Lemma coarse_list_sing_cons :
  forall {T:Type} (l:list T) (a:T),
    coarse_list (a::nil) (a::l).
  intros T l a.
  apply coarse_list_cons2.
  constructor.
Qed.


Lemma coarse_list_sing1_iff :
  forall {T:Type} (l:list T) (a:T),
    coarse_list (a::nil) l <->
    In a l.
  intros T l; induction l as [|b l ih]; simpl; split; intro h2.
  rewrite coarse_list_nil2_iff in h2.
  discriminate.
  contradiction.
  inversion h2; subst.
  right.
  rewrite <- ih; auto.
  left; auto.
  destruct h2 as [|h2]; subst.
  apply coarse_list_sing_cons.
  rewrite <- ih in h2.
  apply coarse_list_cons1; auto.
Qed.


Lemma coarse_list_rev_cons1 :
  forall {T:Type} (l l':list T) (a:T),
    coarse_list (a::l) l' ->
    coarse_list l l'.
  intros ? ? ? ?  h1.
  dependent induction h1. 
  apply coarse_list_cons1; auto.
  apply coarse_list_cons1; auto.
Qed.



Lemma coarse_list_rev_cons_impl :
  forall {T:Type} (l l':list T) (a:T),
    coarse_list (a::l) (a::l') ->
    coarse_list (a::l) l' \/
    coarse_list l l'.
  intros T l l' a h1.
  inversion h1; subst.
  left; auto.
  right; auto.
Qed.

Lemma coarse_list_tl :
  forall {T:Type} (l:list T),
    coarse_list (tl l) l.
  destruct l as [|a l]; simpl.
  constructor.
  apply coarse_list_cons1.
  apply coarse_list_refl.
Qed.

Lemma coarse_list_cons1_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    coarse_list (a::l) l' <->
    exists pf':l' <> nil,
      (a = hd' l' pf' /\ coarse_list l (tl l')) \/
      (a <> hd' l' pf' /\ coarse_list (a::l) (tl l')).
  intros T h1 l l' a; split; intro h2.
  dependent induction h2; subst; simpl.
  exists (cons_neq_nil _ _).
  destruct (h1 a a0) as [|h3]; subst.
  left.
  split; auto.
  eapply coarse_list_rev_cons1.
  apply h2.
  right.
  split; auto.
  exists (cons_neq_nil _ _).
  left.
  split; auto.
  destruct h2 as [h2 h3].
  rewrite (hd_compat' _ h2).
  destruct h3 as [h3 | h3].
  destruct h3 as [h3 h4].
  rewrite <- h3.
  apply coarse_list_cons2; auto.
  destruct h3 as [h3 h4].
  apply coarse_list_cons1; auto.
Qed.


Lemma not_cons_coarse_list_same :
  forall {T:Type} (l:list T) (a:T),
    ~ coarse_list (a::l) l.
  intros T l a h1.
  pose proof (length_coarse_list _  _ h1) as h2.
  simpl in h2.
  omega.
Qed.


Lemma coarse_list_antisym :
  forall {T:Type} (l l':list T),
    coarse_list l l' ->
    coarse_list l' l ->
    l = l'.
  intros T l l' h1.
  induction h1.
  intro h2.
  rewrite coarse_list_nil2_iff in h2; auto.
  intro h2.
  dependent induction h2.
  pose proof (coarse_list_rev_cons1 _ _ _ h1) as h3.
  pose proof (coarse_list_rev_cons1 _ _ _ h2) as h4.
  specialize (IHh2 h3).
  hfirst IHh2.
  intro h5.
  hfirst IHh1.
  constructor; auto.
  specialize (IHh1 hf).
  subst.
  apply not_cons_coarse_list_same in h5.
  contradiction.
  specialize (hf h4); subst.
  hfirst IHh1.
  constructor; auto.
  specialize (IHh1 hf).
  pose proof (f_equal (@length _) IHh1) as h5.
  simpl in h5.
  omega.
  f_equal.
  hfirst IHh1.
  constructor; auto.
  specialize (IHh1 hf).
  subst.
  apply not_cons_coarse_list_same in h2; contradiction.
  intro h2.
  f_equal.
  apply IHh1.
  inversion h2; subst.
  apply length_coarse_list in h1.
  apply length_coarse_list in H1.
  simpl in H1.
  omega.
  assumption.
Qed.  


Lemma in_coarse_list :
  forall {T:Type} (l l':list T) (a:T),
    In a l ->
    coarse_list l l' ->
    In a l'.
  intros T l l' a h1 h2. revert h1. revert a.
  induction h2 as [|l0 l1 a0 h5 ih|l0 l1 a0 h5 ih]; subst.
  intros; contradiction.
  intros; right; apply ih; auto.
  intros a h6.
  destruct h6 as [|h6]; subst.
  left; auto.
  right.
  apply ih; auto.
Qed.

Lemma coarse_list_cons_in :
  forall {T:Type} (l l':list T) (a:T),
    coarse_list (a::l) l' ->
    In a l'.
  intros; eapply in_coarse_list.
  apply in_eq.
  apply H.
Qed.


Lemma coarse_list_incl :
  forall {T:Type} (l l':list T),
    coarse_list l l' ->
    incl l l'.
  intros T l l' h1.  
  induction h1 as [|l l' a ih h2|l l' a ih h2].
  apply incl_nil.
  red in h2; red.
  intros x h3.
  specialize (h2 _ h3).
  right; auto.
  red in h2; red.
  intros x h3.
  destruct h3 as [|h3]; subst.
  left; auto.
  specialize (h2 _ h3).
  right; auto.
Qed.

Lemma coarse_list_cons1_not_nil :
  forall {T:Type} (l l':list T) (a:T),
    coarse_list (a::l) l' ->
    l' <> nil.
  intros ? ? ? ? h1; inversion h1; subst; apply cons_neq_nil.
Qed.

(*Scroll past definition of [lind_occ] for [coarse_list_rev_cons1_iff]*)
Lemma coarse_list_rev_cons2_iff :
  forall {T:Type} (l l':list T) (a':T),
    coarse_list l (a'::l') <->
    (l = nil \/
    (exists (pf:l <> nil),
       hd' l pf = a' /\ coarse_list (tl l) l') \/
    coarse_list l l').
  intros T l l' a'; split; intro h1.
  dependent induction h1.
  left; auto.
  right; right; auto. 
  right.
  left.
  exists (cons_neq_nil _ _).
  simpl.
  split; auto.
  destruct h1 as [|h1]; subst.
  constructor.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 [h2 h3]].
  subst.
  rewrite (hd_compat' _ h1) at 1.
  apply coarse_list_cons2; auto.
  apply coarse_list_cons1; auto.
Qed.






Lemma coarse_list_trans :
  forall (T:Type) (pfdt:type_eq_dec T),
    Transitive _ (@coarse_list T).
  intros T hdt; red; intros l1 l2 l3 h1. revert l3.
  induction h1.
  intros; constructor.
  intros ? h7; apply IHh1.
  apply coarse_list_rev_cons1 in h7; auto.
  intros l0 h2.
  generalize dependent l0.
  intro l0.
  revert IHh1 h1. revert a l' l.
  induction l0 as [|a0 l0 ih].
  intros.
  apply not_coarse_list_cons_nil in h2.
  contradiction.
  intros a l' l h1 h2 h3.
  destruct (hdt a a0) as [|h4]; subst.
  apply coarse_list_cons2.
  pose proof h3 as h3'.
  apply coarse_list_rev_cons_impl in h3'.
  destruct h3' as [h3' | h3'].
  specialize (ih a0 l'  l h1 h2 h3').
  apply coarse_list_rev_cons1 in ih; auto.
  apply h1; auto.
  specialize (ih a l' l h1 h2).
  rewrite coarse_list_rev_cons2_iff in h3.
  destruct h3 as [h3 | h3].
  discriminate.
  destruct h3 as [h3 | h3].
  destruct h3 as [h3 h5].
  simpl in h5.
  destruct h5; contradiction.
  specialize (ih h3).
  apply coarse_list_cons1; auto.
Qed.


Lemma coarse_list_rev_cons2_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a':T),
    coarse_list l (a'::l') ->
    {l = nil} + 
    {exists (pf:l <> nil), hd' l pf = a' /\ coarse_list (tl l) l'} +
    {coarse_list l l'}.
  intros T hdt l l' a h1.  
  destruct (nil_dec' l) as [|h2]; subst.
  left; auto.
  destruct (hdt (hd' l h2) a) as [|h3]; subst.   
  left; right.
  exists h2.
  split; auto.
  rewrite coarse_list_rev_cons2_iff in h1.
  destruct h1 as [|[|h1]]; subst.
  contradict h2; auto.
  destruct H as [? [? ?]]; auto.
  eapply coarse_list_trans; auto.
  apply coarse_list_tl.
  assumption.
  right.
  rewrite coarse_list_rev_cons2_iff in h1.
  destruct h1 as [|[|h1]]; subst.
  contradict (h2 (eq_refl _)).
  destruct H as [? [? ?]]; auto.
  pose proof (proof_irrelevance _ x h2); subst.
  contradict h3; auto.
  assumption.
Qed.


  

Lemma coarse_list_app_l :
  forall {T:Type} (l l1 l2:list T),
    coarse_list l l1 ->
    coarse_list l (l1 ++ l2).
  intros T l l1 l2 h1. revert l2.
  induction h1 as [l|l1 l2 a h2 ih|l1 l2 a h2 ih].
  intros; constructor.
  intros l'; simpl.
  apply coarse_list_cons1; auto.
  intro l'.
  rewrite <- app_comm_cons.
  apply coarse_list_cons2; auto.
Qed.


Lemma coarse_list_app_r :
  forall {T:Type} (l l1 l2:list T),
    coarse_list l l2 ->
    coarse_list l (l1 ++ l2).
  intros T l l1. revert l.
  induction l1 as [|a1 l1 ih].
  intros; rewrite app_nil_l; auto.
  intros l l2 h2.
  specialize (ih _ _ h2).
  rewrite <- app_comm_cons.
  apply coarse_list_cons1; auto.
Qed.


Lemma coarse_list_app_interp_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (l l' l0 l1:list T),
    coarse_list l (l0++l') ->
    coarse_list l (l0++l1++l').
  intros T h0 l l' l0 l1 h1. revert l1.
  dependent induction h1.
  intros; constructor.
  destruct l0 as [|a0 l0].
  rewrite app_nil_l in x. 
  subst.
  intro l1.
  specialize (IHh1 l'0 nil).
  rewrite app_nil_l in IHh1.
  specialize (IHh1 eq_refl (l1++a::nil)).
  rewrite app_nil_l.
  rewrite app_nil_l in IHh1; simpl in IHh1.
  rewrite <- app_assoc in IHh1.
  simpl in IHh1; auto.

  intros l1.
  rewrite <- app_comm_cons in x.
  inversion x; subst. 
  specialize (IHh1 l' l0 eq_refl l1).
  rewrite <- app_comm_cons.
  apply coarse_list_cons1; auto.

  intro l1.
  destruct l0 as [|a0 l0].
  rewrite app_nil_l in x; subst.
  rewrite app_nil_l.
  rewrite coarse_list_cons1_iff. 
  ex_goal.
  rewrite app_eq_iff; simpl.
  intro h2.
  destruct h2 as [h2 h3].
  rewrite skipn_nil in h3.
  discriminate.
  exists hex. 
  destruct l1 as [|b l1].
  left.
  generalize hex.
  rewrite app_nil_l.
  intro hex'.
  simpl.
  split; auto.
  simpl.
  destruct (h0 a b) as [|h2]; subst.
  left.
  split; auto.
  apply coarse_list_app_r.
  apply coarse_list_cons1; auto.
  right; split; auto.
  apply coarse_list_app_r.
  apply coarse_list_cons2; auto.
  assumption.
  rewrite <- app_comm_cons in x.
  inversion x; subst.
  rewrite <- app_comm_cons.
  apply coarse_list_cons2; auto.
Qed.
 
Lemma coarse_list_app2 :
  forall {T:Type} (la la' lb lb':list T),
    coarse_list la  lb ->
    coarse_list la' lb' ->
    coarse_list (la ++ la') (lb ++ lb').
  intros T la la' lb lb' h1.
  induction h1.  
  intros; apply coarse_list_app_r; auto.
  intro h2; specialize (IHh1 h2). 
  rewrite <- app_comm_cons.
  apply coarse_list_cons1; auto.
  intro h2; specialize (IHh1 h2).
  do 2 rewrite <- app_comm_cons.
  apply coarse_list_cons2; auto.
Qed.


(*This removes the coarse sub-list from the super-list.*)
Fixpoint remove_coarse_list
         {T} {l l':list T} (pfdt:type_eq_dec T) :
  coarse_list l l' -> list T :=
  match l' with
    | nil => fun _ => nil
    | a'::l' => fun (pf:coarse_list l (a'::l')) =>
                  match (coarse_list_rev_cons2_dec pfdt _ _ _ pf) with
                    | inleft (left _) => nil
                    | inleft (right pfe) =>
                      let (_, pfc') := constructive_definite_proof pfe in
                      let (_, pfc) := pfc' in
                      @remove_coarse_list _ (tl l) l' pfdt pfc
                    | inright pfc =>
                      @remove_coarse_list _ l l' pfdt pfc
                  end
  end.


Lemma in_remove_coarse_list : 
  forall {T} {l l':list T} (pfdt:type_eq_dec T)
         (pfc:coarse_list l l') x, 
    In x (remove_coarse_list pfdt pfc) ->
    In x l'.
  intros T l l'. revert l.
  induction l' as [|a' l' ih]; simpl.
  intros; contradiction.
  intros l hdt h1 x h2.
  lr_match_sum' h2.
  destruct hl as [|hl]; subst.
  contradiction.
  destruct hl as [h3 [h4 h5]]; subst.
  destruct constructive_definite_description; simpl in h2.
  destruct a as [h4 h6]; subst.
  specialize (ih _ hdt h6 x h2).
  right; auto.
  specialize (ih _ hdt hr x h2).
  right; auto.
Qed.


Definition coarse_lists {T:Type} (l:list T) :=
  [l':list T | coarse_list l' l].

Lemma inh_coarse_lists :
  forall {T:Type} (l:list T),
    Inhabited (coarse_lists l).
  intros; econstructor; constructor; apply coarse_list_nil.
Qed.

Lemma coarse_lists_nil :
  forall T:Type,
    (coarse_lists (@nil T)) =
    Singleton nil.
  intro T.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [h1].
  apply coarse_list_nil2_iff in h1.
  subst.
  constructor; auto.
  destruct h1.
  constructor.
  apply coarse_list_nil.
Qed.

Lemma coarse_lists_cons :
  forall {T:Type} (l:list T) (a:T),
    coarse_lists (a::l) =
    Union (coarse_lists l)
          (Im (coarse_lists l) (cons a)).
  intros T l a; apply Extensionality_Ensembles; red; split; red; intros lx h1.
  destruct h1 as [h1].
  rewrite coarse_list_rev_cons2_iff in h1.
  destruct h1 as [|h1]; subst.
  left.
  constructor.
  apply coarse_list_nil.
  destruct h1 as [h1 | h1].
  destruct h1 as [h1 [h2 h3]]; subst.
  right.
  apply Im_intro with (tl lx).
  constructor; auto.
  apply hd_compat'.
  left.
  constructor; auto.

  destruct h1 as [lx h1 | lx h1].
  destruct h1 as [h1].
  constructor.
  apply coarse_list_cons1; auto.
  destruct h1 as [lx h1]; subst.
  destruct h1 as [h1].
  constructor.
  apply coarse_list_cons2; auto.
Qed.


Lemma finite_coarse_lists :
  forall {T:Type} (l:list T),
    Finite (coarse_lists l).
  intros T l.
  induction l as [|a l ih]; simpl.
  rewrite coarse_lists_nil.
  apply Singleton_is_finite.
  rewrite coarse_lists_cons.
  apply Union_preserves_Finite; auto.
  apply finite_image; auto.
Qed.


Definition coarse_seg_list (l:list nat) n : Prop :=
  coarse_list l (seg_list n).

Lemma coarse_list_app_seg_list_n :
  forall {n l},
    coarse_seg_list l n ->
    coarse_seg_list (l++n::nil) (S n).
  intros n l h1; red in h1; red.
  rewrite seg_list_app; auto with arith.
  simpl.
  rewrite <- minus_n_O.
  apply coarse_list_app2; auto.
  apply coarse_list_refl.
Qed.



Lemma coarse_list_inv_im_in_dep :
  forall {T U:Type} {l:list T} (pfdu:type_eq_dec U)
         (f:fun_in_dep l U) (u:U),
    coarse_list (inv_im_in_dep pfdu l u f) l.
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros; constructor.
  intros h1 f u.
  destruct (h1 u (f a (in_eq a l))) as [|h2]; subst.
  apply coarse_list_cons2; auto.
  apply coarse_list_cons1; auto.
Qed.

Lemma coarse_list_inv_iml :
  forall {T U:Type} {l:list T} (pfdu:type_eq_dec U)
         (f:T->U) (u:U),
    coarse_list (inv_iml pfdu f l u) l.
  intros T U l.
  induction l as [|a l ih]; simpl.
  intros; constructor.
  intros h1 f u.
  destruct (h1 u (f a)) as [|h2]; subst.
  apply coarse_list_cons2; auto.
  apply coarse_list_cons1; auto.
Qed.



Lemma coarse_list_removelast :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    coarse_list (removelast l) l.
  intros T hdt l. 
  destruct (nil_dec' l) as [|h1]; subst; simpl.
  constructor.
  rewrite (app_removelast_lst _ h1) at 2.
  apply coarse_list_app_l.
  apply coarse_list_refl.
Qed.


Lemma no_dup_coarse_list :
  forall {T:Type} (l l':list T),
    coarse_list l l' ->
    NoDup l' ->
    NoDup l.
  intros T l l' h1.
  induction h1.
  intro; constructor.
  intro; apply IHh1.
  apply no_dup_cons in H; auto.
  intro.  
  constructor.
  intro h2.
  pose proof (in_coarse_list _ _ _ h2 h1) as h3.
  pose proof (no_dup_cons_nin _ _ H); contradiction.
  apply IHh1.
  apply no_dup_cons in H; auto.
Qed.


Lemma no_dup_removelast :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T),
    NoDup l ->
    NoDup (removelast l).
  intros; eapply no_dup_coarse_list; auto.
  apply coarse_list_removelast; auto.
  assumption.
Qed.

  
Lemma no_dup_inv_im_in_dep :
  forall {T U:Type} (pfdu:type_eq_dec U)
         (l:list T) (u:U) (f:fun_in_dep l U),
    NoDup l ->
    NoDup (inv_im_in_dep pfdu l u f).
  intros; eapply no_dup_coarse_list.
  apply coarse_list_inv_im_in_dep.
  assumption.
Qed.



Lemma no_dup_app' : 
  forall {T:Type} (l l':list T),
    NoDup (l ++ l') ->
    NoDup l /\ NoDup l'.
  intros T l l'.
  revert l.
  induction l' as [|a' l' ih]; simpl.
  intros ? h1.
  rewrite app_nil_r in h1.
  split; auto.
  constructor.
  intros l h1.
  specialize (ih (l++a'::nil)).
  hfirst ih.
  rewrite <- app_assoc.
  rewrite <- app_comm_cons.
  simpl; auto.
  specialize (ih hf).
  destruct ih as [h2 h3].
  pose proof (no_dup_coarse_list l (l++a'::nil)) as h4.
  hfirst h4.
  apply coarse_list_app_l.
  apply coarse_list_refl.
  specialize (h4 hf0).
  specialize (h4 h2).
  split; auto.
  pose proof (no_dup_coarse_list (a'::l') (l++a'::l')) as h5.
  hfirst h5.
  apply coarse_list_app_r.
  apply coarse_list_refl.
  specialize (h5 hf1 h1); auto.
Qed.




Lemma no_dup_app_intersection : 
  forall {T:Type} (l l':list T),
    NoDup (l ++ l') ->
    Intersection (list_to_set l) (list_to_set l') = Empty_set _.
  intros T l l'. revert l. 
  induction l' as [|a' l' ih]; simpl.
  intros.
  rewrite Intersection_commutative.
  apply empty_intersection.
  intros l h1.
  pose proof (ih (l++a'::nil)) as ih'.
  hfirst ih'.
  rewrite <- app_assoc, <- app_comm_cons; simpl; auto.
  specialize (ih' hf). 
  rewrite list_to_set_app in ih'.
  simpl in ih'.
  rewrite add_empty_sing in ih'.
  rewrite Intersection_commutative in ih'.
  rewrite <- ih'.
  unfold Add. 
  rewrite Distributivity.
  rewrite Distributivity. 
  rewrite Intersection_commutative.
  f_equal. 
  pose proof (no_dup_app' _ _ h1) as h1'.
  destruct h1' as [h2 h3]. 
  pose proof (ih (a'::nil) h3) as ih0.
  simpl in ih0.
  rewrite add_empty_sing in ih0.
  rewrite Intersection_commutative in ih0.
  rewrite ih0.
  apply no_dup_app' in hf.
  destruct hf as [h4 h5].
  rewrite <- nin_iff.
  rewrite <- list_to_set_in_iff.
  apply no_dup_app_sing_nin; auto.
Qed.



Lemma no_dup_app_nin : 
  forall {T:Type} (l l':list T),
    NoDup (l ++ l') ->
    forall x:T,
      ~(In x l /\ In x l').
  intros T l l' h1 x h2.
  pose proof (no_dup_app_intersection l l' h1) as h3.
  destruct h2 as [h2 h4].
  rewrite list_to_set_in_iff in h2, h4.
  pose proof (Intersection_intro _ _ _ _ h2 h4) as h5.
  rewrite h3 in h5.
  contradiction.
Qed.


Lemma no_dup_map_seg_inj' :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    NoDup (map_seg f) ->
    inj_seg f.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; red; intros; omega.
  intros f h1.
  specialize (ih (seg_fun_from_S f)).
  pose proof (no_dup_app' _ _ h1) as h2.
  destruct h2 as [h2 h3].
  specialize (ih h2).
  red in ih; red.
  intros i j h5 h6.
  pose proof h5 as h5'.
  pose proof h6 as h6'.
  apply le_lt_eq_dec in h5'.
  apply le_lt_eq_dec in h6'.
  destruct h5' as [h7 | h7], h6' as [h8 | h8].
  intros; eapply ih.
  unfold seg_fun_from_S.
  erewrite f_equal_dep.
  rewrite H at 1.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  apply S_inj in h8; subst.
  intro h8.
  pose proof (no_dup_app_sing_nin _ _ h1) as h9.
  contradict h9.
  erewrite f_equal_dep.
  rewrite <- h8 at 1.
  rewrite in_map_seg_iff.
  exists i, (lt_S_n _ _ h7).
  unfold seg_fun_from_S; f_equal; apply proof_irrelevance.
  reflexivity.
  apply S_inj in h7; subst.
  intro h9.
  pose proof (no_dup_app_sing_nin _ _ h1) as h10.
  contradict h10.
  erewrite f_equal_dep.
  rewrite h9 at 1.
  rewrite in_map_seg_iff.
  exists j, (lt_S_n _ _ h8).
  unfold seg_fun_from_S.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  omega.

  Grab Existential Variables.
  apply lt_S_n in h8; auto.
  apply lt_S_n in h7; auto.
Qed.


Definition mask_to_coarse_list
           {T:Type} {l:list T} {v:bv}
           (pfeq:length l = length v) : list T :=
  map_in_dep (fun (i:nat) (pfi:In i (bv1s v)) =>
                let pfi' := lt_eq_trans _ _ _ (in_bv1s_lt _ _ pfi) 
                                        (eq_sym pfeq) in
                nth_lt l i pfi').

Lemma length_mask_to_coarse_list :
  forall {T:Type} {l:list T} {v:bv}
         (pfeq:length l = length v),
    length (mask_to_coarse_list pfeq) = length (bv1s v).
  intros; unfold mask_to_coarse_list; rewrite length_map_in_dep; auto.
Qed.


Lemma mask_to_coarse_list_eq_nil_iff :
  forall {T:Type} {l:list T} {v:bv}
         (pfeq:length l = length v),
    mask_to_coarse_list pfeq = nil <->
    bv1s v = nil.
  unfold mask_to_coarse_list; intros;
  rewrite map_in_dep_eq_nil_iff,
  bv1s_eq_nil_iff; tauto.
Qed.

(*Is True if [l] can be gotten from [l'] via [v], 
when [coarse_list l l'] is constructed via the "mask" [v].*)
(*There are two variants.*)
(*The first version is for easy construction, given essential 
components.  The more verbose version (primed) is easier to 
prove in a goal, when you are applying the [constructor] tactic.
Think of the difference between the 
two as being analogous to [Inductive Im], as it's 
done in the library, and without the redundant parameteters.  
The latter is easier when constructing [In (Im A f) (f x)],
the former (library version) is easier when using
[econstructor] in a goal.*)
Inductive coarse_list_mask {T} :
  forall (l l':list T), coarse_list l l' ->
                 forall (v:bv),
                   length l' = length v -> Prop :=
| coarse_list_mask_nil :
    forall l,
      coarse_list_mask nil l
                     (coarse_list_nil l)
                     (lrep false (length l))
                     (eq_sym (length_lrep false (length l)))
| coarse_list_mask_cons1 :
    forall l l' a (pfc:coarse_list l l')
           v (pfe:length l' = length v),
      coarse_list_mask l l' pfc v pfe ->
      coarse_list_mask l (a::l')
                      (coarse_list_cons1 l l' a pfc)
                      (false::v)
                      (f_equal S pfe)
| coarse_list_mask_cons2 :
    forall l l' a (pfc:coarse_list l l')
           v (pfe:length l' = length v),
      coarse_list_mask l l' pfc v pfe ->
      coarse_list_mask  (a::l)
                      (a::l')
                      (coarse_list_cons2 l l' a pfc)
                      (true::v)
                      (f_equal S pfe).


Inductive coarse_list_mask' {T} :
  forall (l l':list T),
    coarse_list l l' ->
    forall (v:bv),
      length l' = length v -> Prop :=
| coarse_list_mask_nil' :
    forall l v
           (pfc:coarse_list nil l)
           (pfe:length l = length v),
      all_sing v false ->
      coarse_list_mask' nil l
                      pfc
                      v
                      pfe
| coarse_list_mask_cons1'  :
    forall l l' a
           (pfc:coarse_list l l')
           (pfc':coarse_list l (a::l'))
           v 
           (pfe:length l' = length v)
           (pfe':S (length l') = S (length v))
           (b:bool),
      b = false ->
      coarse_list_mask' l l' pfc v pfe ->
      coarse_list_mask' l (a::l')
                      pfc'
                      (b::v)
                      pfe'
| coarse_list_mask_cons2':
    forall l l' a
           (pfc:coarse_list l l')
           (pfc':coarse_list (a::l) (a::l'))
           v
           (pfe:length l' = length v)
           (pfe':S (length l') = S (length v))
           (b:bool),
      b = true ->
      coarse_list_mask' l l' pfc v pfe ->
      coarse_list_mask' (a::l)
                      (a::l')
                      pfc'
                      (b::v)
                      pfe'.


Lemma coarse_list_mask_functional :
  forall {T} (l l' l0 l0':list T) v v'
         (pfle:  l = l0)
         (pfle': l' = l0')
         (pfc:coarse_list l l')
         (pfc':coarse_list l0 l0')
         (pfv:v = v')
         (pfe:length l' = length v)
         (pfe':length l0' = length v'),
    coarse_list_mask l l' pfc v pfe ->
    coarse_list_mask l0 l0' pfc' v' pfe'.
  intros; subst;
  pose proof (proof_irrelevance _ pfc pfc');
  pose proof (proof_irrelevance _ pfe pfe');
  subst; auto.
Qed.

Lemma coarse_list_mask_functional' :
  forall {T} (l l' l0 l0':list T) v v'
         (pfle:  l = l0)
         (pfle': l' = l0')
         (pfc:coarse_list l l')
         (pfc':coarse_list l0 l0')
         (pfv:v = v')
         (pfe:length l' = length v)
         (pfe':length l0' = length v'),
    coarse_list_mask' l l' pfc v pfe ->
    coarse_list_mask' l0 l0' pfc' v' pfe'.
  intros; subst;
  pose proof (proof_irrelevance _ pfc pfc');
  pose proof (proof_irrelevance _ pfe pfe');
  subst; auto.
Qed.



Lemma coarse_list_mask_iff' :
  forall {T} {l l':list T} {v:bv}
         (pfc:coarse_list l l')
         (pfe:length l' = length v),
    coarse_list_mask l l' pfc v pfe <->
    coarse_list_mask' l l' pfc v pfe.
  intros; split; intro h1; induction h1; subst.
  constructor.
  red; intros x h1; apply in_lrep' in h1; auto.
  eapply coarse_list_mask_cons1'; auto; apply IHh1; auto.
  eapply coarse_list_mask_cons2'; auto; apply IHh1; auto. 
  pose proof (coarse_list_mask_nil l) as h2.
  pose proof (all_sing_lrep _ _ H) as h1.
  pose proof (proof_irrelevance _ pfc (coarse_list_nil l)) as h4.
  gen0.
  rewrite h1.
  rewrite <- pfe.
  rewrite h4.
  intro pfe'.
  pose proof (proof_irrelevance _ pfe'
                                (eq_sym (length_lrep false (length l)))) as h5.
  rewrite h5; auto.
  pose proof (coarse_list_mask_cons1 _ _ a pfc v pfe IHh1) as h2.
  pose proof (proof_irrelevance _ pfc' (coarse_list_cons1 l l' a pfc)).
  pose proof (proof_irrelevance _ pfe' (f_equal S pfe));
    subst; constructor; auto.
  pose proof (coarse_list_mask_cons2 _ _ a pfc v pfe IHh1) as h2.
  pose proof (proof_irrelevance _ pfc' (coarse_list_cons2 l l' a pfc)).
  pose proof (proof_irrelevance _ pfe' (f_equal S pfe));
  subst; constructor; auto.
Qed.


Lemma coarse_list_mask_nil_iff :
forall {T} {l':list T} {v:bv}
       (pfc:coarse_list nil l')
       (pfe:length l' = length v),
  coarse_list_mask nil l' pfc v pfe <->
  v = lrep false (length l').
  intros; split; intro h3; dependent induction h3; subst; auto.
  rewrite coarse_list_mask_iff'; constructor.
  red.
  intros x h1.
  apply in_lrep' in h1; auto.
Qed.


Lemma coarse_list_mask_cons_true_not_nil :
  forall {T} {l l':list T} {v:bv} {a'}
         (pfc:coarse_list l (a'::l'))
         (pfe:length (a'::l') = length (true::v)),
    coarse_list_mask l (a'::l') pfc (true::v) pfe ->
    l <> nil.
  intros; intro; subst; inversion H.
Qed.

Lemma coarse_list_mask_cons_true_iff :
  forall {T} {l l':list T} {v:bv} {a'}
         (pfc:coarse_list l (a'::l'))
         (pfe:length (a'::l') = length (true::v)),
    coarse_list_mask l (a'::l') pfc (true::v) pfe <->
    exists (pfn:l <> nil),
      a' = hd' l pfn /\
      exists pfc':coarse_list (tl l) l',
        coarse_list_mask (tl l) l' pfc' v (S_inj _ _ pfe).
  intros T l l' v a' hc h1; split; intro h2. 
  exists (coarse_list_mask_cons_true_not_nil _ _ h2).
  inversion h2; subst; simpl; auto. 
  split; auto.
  exists pfc.
  eapply coarse_list_mask_functional in H1; auto.
  apply H1.
  destruct h2 as [h2 [h3 h4]].
  revert h4.
  subst.
  intro h4.
  destruct h4 as [h4 h5].
  pose proof (coarse_list_mask_cons2 (tl l) l' (hd' l h2) h4 v (S_inj _ _ h1) h5) as h6.
  eapply coarse_list_mask_functional in h6.
  apply h6.
  rewrite (hd_compat' _ h2); auto.
  reflexivity.
  reflexivity.
Qed.


Lemma coarse_list_mask_same_nil :
  forall {T} (a c:T) b
         (pfeq':length (a::nil) = length (b::nil))
         (pfc':coarse_list (a::nil) (c::nil)),
    coarse_list_mask (a :: nil) (c :: nil) pfc' (b::nil) pfeq' ->
    a = c /\
    b = true.
  intros T a c b h1 h2 h3.
  inversion h2; subst.
  rewrite coarse_list_nil2_iff in H1.
  discriminate.
  split; auto.
  inversion h3; subst; auto.
  pose proof (coarse_list_nil2_iff (c::nil)) as h5.
  pose proof pfc as h6.
  rewrite h5 in h6.
  discriminate.
Qed.


Lemma coarse_list_mask_cons_false_iff :
  forall {T} {l l':list T} {v:bv} {a'}
         (pfc:coarse_list l (a'::l'))
         (pfe:length (a'::l') = length (false::v)),
    coarse_list_mask l (a'::l') pfc (false::v) pfe <->
      exists pfc':coarse_list l l', 
        coarse_list_mask l l' pfc' v (S_inj _ _ pfe).
  intros T l l'.
  destruct l' as [|a' l']. 
  simpl.
  intros v a h1 h2.  
  pose proof (S_inj _ _ h2) as h3. 
  symmetry in h3.
  apply length_eq0 in h3; subst.
  inversion h1; subst.
  split; intro h3.
  exists (coarse_list_nil nil).
  rewrite coarse_list_mask_iff'. 
  constructor.  
  red; intros; contradiction.
  rewrite coarse_list_mask_iff'. 
  constructor.  
  red; intros x hx; destruct hx; auto.
  contradiction.
  rewrite coarse_list_nil2_iff in H1; subst.
  split; intro h3.
  exists (coarse_list_nil nil).
  rewrite coarse_list_mask_iff'.
  constructor.
  red; intros; contradiction.
  rewrite coarse_list_mask_iff'. 
  constructor.  
  red; intros x hx; destruct hx; auto.
  contradiction.
  rewrite coarse_list_nil2_iff in H1; subst.
  split; intro h3.
  apply coarse_list_mask_same_nil in h3.
  destruct h3; discriminate.
  destruct h3 as [h3 h4].
  pose proof h3 as h5.
  rewrite coarse_list_nil2_iff in h5; subst.
  discriminate.

  intros v a hc he.
  split; intro h1.
  inversion h1; subst.
  exists (coarse_list_nil (a'::l')).
  rewrite coarse_list_mask_iff'.
  constructor.
  red; intros x h2.
  destruct h2 as [|h2]; auto.
  apply in_lrep' in h2; auto.
  exists pfc.
  gterm0.
  pose proof (proof_irrelevance _ c pfe) as h3.
  fold c.
  rewrite h3; auto.
  destruct h1 as [h1 h2].
  erewrite coarse_list_mask_iff'.
  econstructor.
  reflexivity.
  rewrite <- coarse_list_mask_iff'.
  apply h2.
Qed.

(*The set of all bit-masks that can construct a given coarse-sub-list.*)
Definition coarse_list_masks {T} {l l':list T}
           (pfc:coarse_list l l') : Ensemble bv :=
  [v:bv | exists pfe:length l' = length v,
                     coarse_list_mask l l' pfc v pfe].


Lemma finite_bv_eq_length :
  forall {T} (l:list T),
    Finite [v:bv | length l = length v].
  intros T l.
  induction l as [|a l ih]; simpl.
  gterm0.
  fold c.
  assert (h1:c = Singleton nil).
    apply Extensionality_Ensembles; red; split; red; intros x h1.
    destruct h1 as [h1].
    symmetry in h1.
    apply length_eq0 in h1; subst.
    constructor.
    destruct h1.
    constructor.
    auto.
  rewrite h1.
  apply Singleton_is_finite.
  gterm0.
  fold c.
  term0 ih.
  fold c0 in ih.
  assert (h1:c = Union (Im c0 (fun l => true::l))
                       (Im c0 (fun l => false::l))).
    apply Extensionality_Ensembles; red; split; red; intros x h1.
    destruct h1 as [h1].
    destruct x as [|b v]; try discriminate.
    simpl in h1.
    apply S_inj in h1.
    destruct b.
    left.
    apply Im_intro with v; auto.
    constructor; auto.
    right.
    apply Im_intro with v; auto.
    constructor; auto.
    destruct h1 as [x h1 | x h1]; destruct h1 as [x h1]; subst;
    destruct h1 as [h1]; constructor; simpl; auto with arith.
    rewrite h1.
    apply Union_preserves_Finite; apply finite_image; auto.
Qed.

Lemma finite_coarse_lists_masks :
  forall {T} {l l':list T}
         (pfc:coarse_list l l'),
    Finite (coarse_list_masks pfc).
  intros; eapply Finite_downward_closed.
  apply (finite_bv_eq_length l').
  red; intros x h1.
  destruct h1 as [h1].
  destruct h1 as [h1 h2].
  constructor; auto.
Qed.


Lemma inh_coarse_lists_masks :
  forall {T} {l l':list T}
         (pfc:coarse_list l l'),
    Inhabited (coarse_list_masks pfc).
  intros T l l'; revert l.
  induction l' as [|a l' ih].
  intros.
  pose proof pfc as h1.
  rewrite coarse_list_nil2_iff in h1; subst.
  apply Inhabited_intro with nil.
  constructor.
  exists (eq_refl _).
  rewrite coarse_list_mask_iff'.
  constructor.
  red; intros x h1.
  contradiction.
  intros l h1.
  inversion h1; subst.
  apply Inhabited_intro with (lrep false (S (length l'))).
  constructor.
  exists (eq_sym (length_lrep _ _)).
  simpl.
  rewrite coarse_list_mask_iff'.
  constructor.
  red; intros x h2.
  destruct h2 as [|h2]; auto.
  apply in_lrep' in h2; auto.
  specialize (ih l H1).
  destruct ih as [v h3].
  destruct h3 as [h3].
  destruct h3 as [h3 h4].
  apply Inhabited_intro with (false::v).
  constructor.
  exists (f_equal S h3).
  pose proof (coarse_list_mask_cons1 l l' a  _ _ h3 h4) as h5.
  pose proof (proof_irrelevance _ h1
                                (coarse_list_cons1 l l' a H1)); subst; auto.
  specialize (ih _ H1).
  destruct ih as [v h2].
  destruct h2 as [h2].
  destruct h2 as [h2 h3].
  pose proof (coarse_list_mask_cons2 l0 l' a H1 v h2 h3) as h4.
  apply Inhabited_intro with (true::v).
  constructor.
  exists (f_equal S h2).
  pose proof (proof_irrelevance _ h1
                                (coarse_list_cons2 l0 l' a H1)); subst; auto.
Qed.


Definition coarse_list_to_mask {T} {l l':list T}
           (pfc:coarse_list l l') : bv :=
  min_set_sortable (coarse_list_masks pfc)
                   (finite_coarse_lists_masks pfc)
                   (inh_coarse_lists_masks pfc)
                   bv_lt
                   sortable_bv_lt.

Lemma coarse_list_to_mask_compat :
  forall {T} {l l':list T} (pfc:coarse_list l l'),
    exists pfe:length l' = length (coarse_list_to_mask pfc),
      coarse_list_mask l l' pfc (coarse_list_to_mask pfc) pfe /\
      forall (v:bv) (pfe':length l' = length v), 
        v <> coarse_list_to_mask pfc ->
        coarse_list_mask l l' pfc v pfe' ->
        bv_lt (coarse_list_to_mask pfc) v.
  intros T l l' h1.
  pose proof (min_set_sortable_compat (coarse_list_masks h1)
                   (finite_coarse_lists_masks h1)
                   (inh_coarse_lists_masks h1)
                   bv_lt
                   sortable_bv_lt) as h2.
  simpl in h2.
  destruct h2 as [h2 h3].
  destruct h2 as [h2].
  destruct h2 as [h2 h4].
  exists h2.
  split; auto.
  intros v he h5 h6.
  specialize (h3 v).
  apply h3.
  constructor.
  constructor.
  exists he; auto.
  contradict h5.
  inversion h5; auto.
Qed.



Lemma length_cons_count_occ_false :
  forall {T:Type} (l:list T) (a:T) (v:bv)
         (pfleq:length (a::l) = count_occ bool_eq_dec (false::v) false),
    length l = count_occ bool_eq_dec v false.
  intros T l a v h1.
  simpl in h1.
  lr_if h1; fold hlr in h1; destruct hlr; auto with arith.
  contradict n; auto.
Qed.

Lemma length_count_occ_true_false :
  forall {T:Type} (l:list T) (v:bv)
         (pfleq:length l = count_occ bool_eq_dec (true::v) false),
    length l = count_occ bool_eq_dec v false.
  intros T l v h1.
  simpl in h1.
  lr_if h1; fold hlr in h1; destruct hlr; auto with arith.
  discriminate.
Qed.


(*Inserts an [x] in between the elements of [l] so that 
the final indices of [x] in the returned list correspond to the [trues]
in [v]. Think of it as an inverse of [remove].
Assumes [~In x l], althought it doesn't use it in the computation,
only in the compatibility*)
Fixpoint interp_bv_aux 
         {T:Type} (x:T)
         (acc : list T)
         (v:bv) :
  forall (l:list T),
    length l = count_occ bool_eq_dec v false -> list T :=
  match v
  with
    | nil => fun _ _ => acc 
    | b::v' =>
      match b with
        | true =>
          fun l
               (pfleq:length l = count_occ bool_eq_dec (true::v') false)
           =>
           let acc' := acc ++ x::nil in
           let pfleq' := length_count_occ_true_false _ _  pfleq in
           interp_bv_aux x acc' v' l pfleq'
        | false =>
          fun l =>
             match l
             with
               | nil => (fun _ => acc)
               | a::l' => (fun
                               (pfleq:length (a::l') = count_occ bool_eq_dec (false::v') false) =>
                             let acc' := acc ++ a::nil in
                             let pfleq' := length_cons_count_occ_false _ _ _ pfleq in
                             interp_bv_aux x acc' v' l' pfleq')
             end
      end
  end.

Lemma interp_bv_aux_functional :
  forall {T:Type} (x x':T)
         (acc acc': list T)
         (v v':bv) (l l':list T)
         (pf:length l = count_occ bool_eq_dec v false)
         (pf':length l' = count_occ bool_eq_dec v' false),
    x = x' ->
    acc = acc' ->
    v = v' ->
    l = l' ->
    interp_bv_aux x acc v l pf =
    interp_bv_aux x acc' v' l' pf'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

Definition interp_bv 
           {T:Type} (x:T)
           (v:bv) (l:list T)
           (pfe:length l = count_occ bool_eq_dec v false) :=
  interp_bv_aux x nil v l pfe.

Lemma interp_bv_functional :
  forall {T:Type} (x x':T)
         (v v':bv) (l l':list T)
         (pf:length l = count_occ bool_eq_dec v false)
         (pf':length l' = count_occ bool_eq_dec v' false),
    x = x' ->
    v = v' ->
    l = l' ->
    interp_bv x v l pf =
    interp_bv x v' l' pf'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.



Lemma length_interv_bv_aux :
  forall {T:Type} (x:T) (v:bv) (l:list T)
         (pfe:length l = count_occ bool_eq_dec v false)
         (acc:list T),
    length (interp_bv_aux x acc v l pfe) =
    length v + length acc.
  intros T x v. revert x.
  induction v as [|b v ih]; simpl; auto.
  intros x l h1.
  pose proof h1 as h1'.
  lr_if h1'; fold hlr in h1'; destruct hlr as [hl | hr]; subst.
  intro acc.
  destruct l as [|a l].
  simpl in h1'.
  discriminate.
  specialize (ih x l (length_cons_count_occ_false l a v h1) (acc ++ a::nil)).
  rewrite ih, app_length, S_compat, <- plus_S_shift_r; auto.
  intro acc.
  destruct b.
  rewrite ih, app_length, S_compat, <- plus_S_shift_r; auto.
  contradict hr; auto.
Qed.



Lemma length_interp_bv :
  forall {T:Type} (x:T)
         (v:bv) (l:list T)
         (pfe:length l = count_occ bool_eq_dec v false),
    length (interp_bv x v l pfe) =
    length v.
  unfold interp_bv; intros; rewrite length_interv_bv_aux; simpl; auto.
Qed.

Lemma length_count_occ_false_tl :
  forall {T:Type} (x:T) (l:list T) (v:bv),
    length l = count_occ bool_eq_dec (false::v) false ->
    length (tl l) = count_occ bool_eq_dec v false.
  intros T x l v h1.
  destruct l as [|a l]; simpl.
  simpl in h1.
  lr_if h1; fold hlr in h1; destruct hlr as [h3 | h3]; subst.
  discriminate.
  contradict h3; auto.
  simpl in h1.
  lr_if h1; fold hlr in h1; destruct hlr as [h3 | h3]; subst; auto with arith.
  contradict h3; auto.
Qed.

Lemma length_count_occ_true :
  forall {T:Type} (x:T) (l:list T) (v:bv),
    length l = count_occ bool_eq_dec (true::v) false ->
    length l = count_occ bool_eq_dec v false.
  intros.
  rewrite count_occ_cons_neq in H; auto.
  intro; discriminate.
Qed.

Lemma length_count_occ_true_true :
  forall {T:Type} (x:T) (l:list T) (v:bv),
    length l = count_occ bool_eq_dec (true::(true::v)) false ->
    length l = count_occ bool_eq_dec v false.
  intros. 
  rewrite count_occ_cons_neq in H; auto.
  rewrite count_occ_cons_neq in H; auto.
  intro; discriminate.
  intro; discriminate.
Qed.



Lemma interp_bv_aux_false :
  forall {T:Type} (x:T)
         (v:bv) (l acc:list T)
         (pfle:length l = count_occ bool_eq_dec (false::v) false)
         (pfn:l <> nil),
    let pfle' := length_count_occ_false_tl x _ _ pfle in
    interp_bv_aux x acc (false :: v) l pfle =
    acc ++ hd' l pfn :: interp_bv_aux x nil v (tl l) pfle'.
  intros T x v.
  revert x.
  induction v as [|b v ih].
  intros ? l acc h1 h2. 
  destruct l as [|a l].
  contradict h2; auto.   
  simpl; auto.
  destruct b.
  intros x l acc h1 h2 h3.
  simpl.
  destruct l as [|a l].
  contradict h2; auto.
  pose proof ih as ih'.
  specialize (ih x (x::l) (acc++a::nil)).
  pose proof h1 as h1'.
  simpl in h1'.
  lr_if h1'; fold hlr in h1'; destruct hlr as [hl | hr].
  lr_if h1'; fold hlr in h1'; destruct hlr as [hl' | hr'].
  discriminate.
  apply S_inj in h1'.
  hfirst ih.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl0 | hr0].
  f_equal; auto.
  contradict hr0; auto.
  specialize (ih hf (cons_neq_nil _ _)).
  simpl in ih.
  erewrite f_equal_dep.
  rewrite ih at 1.
  rewrite <- app_assoc.
  simpl.
  do 2 f_equal.
  specialize (ih' x (x::l) nil).
  simpl in ih'.
  hfirst ih'.
  lr_if_goal; fold hlr; destruct hlr as [hl0 | hr0].
  f_equal; auto.
  contradict hr0; auto.
  specialize (ih' hf0 (cons_neq_nil _ _)).
  symmetry.
  erewrite f_equal_dep.
  rewrite ih' at 1.
  repeat f_equal.
  apply proof_irrelevance.
  reflexivity.
  reflexivity.
  contradict hr; auto.
  
  intros x l acc h1 h2 h3.
  pose proof ih as ih'.
  pose proof h1 as h1'.
  simpl in h1'. 
  lr_if h1'; fold hlr in h1'; destruct hlr as [hl | hr].
  specialize (ih x (tl l) (acc++hd' l h2::nil)).
  hfirst ih.
  omega.
  specialize (ih hf).
  hfirst ih.
  destruct l as [|a l].
  simpl in h1'.
  discriminate.
  simpl.
  destruct l as [|l].
  simpl in h1'.
  apply S_inj in h1'.
  discriminate.
  apply cons_neq_nil.
  specialize (ih hf0).
  simpl.
  simpl in ih.
  destruct l as [|b l].
  simpl in h1'.
  discriminate.
  destruct l as [|c l].
  simpl; auto.
  simpl.
  simpl in ih.
  erewrite f_equal_dep.
  rewrite ih at 1.
  rewrite <- app_assoc.
  simpl.
  do 2 f_equal.
  specialize (ih' x (c::l) nil). 
  hfirst ih'.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr'].
  simpl in h1'.
  apply S_inj in h1'; auto.
  contradict hr'; auto.
  specialize (ih' hf1 (cons_neq_nil _ _)).
  simpl in ih'.
  symmetry.
  erewrite f_equal_dep.
  rewrite ih' at 1.
  repeat f_equal.
  apply proof_irrelevance.
  reflexivity.
  reflexivity.
  contradict hr; auto.
Qed.

Lemma interp_bv_false_nil :
  forall {T:Type} (x:T)
         (v:bv)
         (pfle:0 = count_occ bool_eq_dec (false::v) false),
    interp_bv x (false::v) nil pfle =
    nil.
  unfold interp_bv; intros; simpl; auto.
Qed.

Lemma interp_bv_false :
  forall {T:Type} (x:T)
         (v:bv) (l:list T)
         (pfle:length l = count_occ bool_eq_dec (false::v) false)
         (pfn:l <> nil),
    let pfle' := length_count_occ_false_tl x _ _ pfle in
    interp_bv x (false::v) l pfle =
    (hd' _ pfn)::interp_bv x v (tl l) pfle'.
  intros T x v l h1 h2 h3.
  unfold interp_bv.
  erewrite interp_bv_aux_false.
  simpl.
  reflexivity.
Qed.

 
Lemma interp_bv_aux_true_true :
  forall {T:Type} (x:T)
         (v:bv) (l acc:list T)
         (pfle:length l = count_occ bool_eq_dec (true::true::v) false)
         (pfn:l <> nil),
    (forall (x : T) (l acc : list T)
         (pfle : length l = count_occ bool_eq_dec (true :: v) false),
       l <> nil ->
       let pfle' := length_count_occ_true x l v pfle in
       interp_bv_aux x acc (true :: v) l pfle =
       acc ++ x :: interp_bv_aux x nil v l pfle') ->
    let pfle' := length_count_occ_true x _ _ pfle in
    interp_bv_aux x acc (true :: true :: v) l pfle =
    acc ++ x :: interp_bv_aux x nil (true :: v) l pfle'.
  intros T x v l acc h1 h2 ih h4. 
  pose proof h1 as h1'.
  simpl in h1'.
  lr_if h1'; fold hlr in h1'; destruct hlr as [hl | hr].
  discriminate.
  pose proof ih as ih'.   
  specialize (ih x l (acc++x::nil)). 
  hfirst ih.
  rewrite count_occ_cons_neq; auto.
  specialize (ih hf h2).
  simpl in ih.
  simpl.
  erewrite f_equal_dep.
  rewrite ih at 1; simpl.  
  rewrite <- app_assoc.
  simpl.
  do 2 f_equal.
  specialize (ih' x l nil).
  hfirst ih'.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr'].
  discriminate.
  simpl in h1'; auto.
  specialize (ih' hf0 h2).
  simpl in ih'.
  symmetry.
  erewrite f_equal_dep.
  rewrite ih' at 1.
  repeat f_equal.
  apply proof_irrelevance.
  reflexivity.
  reflexivity.
Qed.


Lemma interp_bv_aux_nil :
  forall {T:Type} (x:T)
         (v:bv) (acc:list T)
         (pf:count_occ bool_eq_dec v false = 0),
    interp_bv_aux x acc v nil (eq_sym pf) =
    acc ++ lrep x (length v).
  intros T x v. revert x. 
  induction v as [|b v ih]; simpl.
  intros. rewrite app_nil_r; auto.
  intros x acc h1. 
  pose proof h1 as h1'.
  lr_if h1'; fold hlr in h1'; destruct hlr as [hl | hr].
  subst.
  discriminate.
  destruct b.
  specialize (ih x (acc++x::nil) h1').
  pose proof (proof_irrelevance
                _
                (eq_sym h1')
                (length_count_occ_true_false (@nil T) v (eq_sym h1))) as h8.
  rewrite <- h8.
  rewrite <- app_assoc in ih; auto.
  contradict hr; auto.
Qed.

Lemma interp_bv_aux_true :
  forall {T:Type} (x:T)
         (v:bv) (l acc:list T)
         (pfle:length l = count_occ bool_eq_dec (true::v) false)
         (pfn:l <> nil),
    let pfle' := length_count_occ_true x _ _ pfle in
    interp_bv_aux x acc (true :: v) l pfle =
    acc ++ x :: interp_bv_aux x nil v l pfle'.
  intros T x v. 
  revert x.
  induction v as [|b v ih].  
  intros; auto.
  destruct b.
  intros x l acc h1 h2 h3.
  rewrite interp_bv_aux_true_true; simpl.
  repeat f_equal.
  assumption.
  assumption.
  
  intros x l acc h1 h2 h3.
  destruct l as [|a l]; auto.
  pose proof h1 as h1'.
  rewrite count_occ_cons_neq in h1'.
  assert (h4:interp_bv_aux x (acc++x::nil) (false::v) (a::l) h1' =
             interp_bv_aux x acc (true :: false :: v) (a :: l) h1).
    simpl.
    repeat f_equal.
    apply proof_irrelevance.
  rewrite <- h4.
  do 2 erewrite interp_bv_aux_false.
  simpl.
  rewrite <- app_assoc.
  repeat f_equal.
  simpl.
  repeat f_equal.
  apply proof_irrelevance.
  intro; discriminate.

  Grab Existential Variables.
  apply cons_neq_nil.
  apply cons_neq_nil.
Qed.

Lemma interp_bv_true :
  forall {T:Type} (x:T)
         (v:bv) (l:list T)
         (pfle:length l = count_occ bool_eq_dec (true::v) false),
    let pfle' := eq_trans pfle (count_occ_cons_neq bool_eq_dec v diff_true_false) in
    interp_bv x (true::v) l pfle =
    x::interp_bv x v l pfle'.
  unfold interp_bv; intros T x v l h1.
  destruct (nil_dec' l) as [|h2].
  subst.
  pose proof (interp_bv_aux_nil x (true::v) nil  (eq_sym h1)) as h2.
  pose proof (proof_irrelevance _ h1 (eq_sym (eq_sym h1))) as h3.
  rewrite h3 at 1.
  simpl; simpl in h2.
  rewrite h2.
  f_equal.
  pose proof (interp_bv_aux_nil x v nil  (eq_sym (eq_trans h1 (count_occ_cons_neq bool_eq_dec v diff_true_false)))) as h4.
  simpl in h4.
  simpl.
  pose proof (proof_irrelevance
                _
                (eq_trans h1 (count_occ_cons_neq bool_eq_dec v diff_true_false))
                (eq_sym (eq_sym (eq_trans h1 (count_occ_cons_neq bool_eq_dec v diff_true_false))))) as h5.
  rewrite h5 at 1.
  rewrite h4 at 1.
  reflexivity.
  rewrite interp_bv_aux_true; simpl; auto.
  repeat f_equal.
  apply proof_irrelevance.
Qed.


Lemma in_interp_bv :
  forall {T:Type} (x:T)
         (v:bv) (l:list T)
         (pf:length l = count_occ bool_eq_dec v false),
    In true v ->
    In x (interp_bv x v l pf).
  intros T x v.
  revert x.
  induction v as [|b v ih].
  intros; contradiction.
  intros x l h1 h2.
  destruct h2 as [|h2]; subst.
  rewrite interp_bv_true.
  left; auto.
  destruct b.
  rewrite interp_bv_true.
  left; auto.
  pose proof h1 as h1'.
  rewrite count_occ_cons_eq in h1'.
  pose proof h1' as h3.
  apply length_eq_S in h3.
  rewrite (interp_bv_false _ _ _ _ h3).
  right.
  apply ih; auto.
  reflexivity.
Qed.

Lemma nin_false_interp_bv : 
  forall {T:Type} (x:T) (v:bv) (l:list T)
         (pf:length l = count_occ bool_eq_dec v false),
    ~ In false v ->
    interp_bv x v l pf = lrep x (length v).
  intros T x v.
  revert x.
  induction v as [|b v ih].
  intros x l h1.
  pose proof h1 as h1'.
  apply length_eq0 in h1'.
  subst.
  intro.
  unfold interp_bv; auto. 
  intros x l h1 h2.
  destruct b.
  pose proof h1 as h1'.   
  rewrite count_occ_cons_neq in h1'; auto.
  rewrite interp_bv_true.
  simpl.
  f_equal; apply ih.
  contradict h2; right; auto.
  intro; discriminate.
  contradict h2; left; auto.
Qed.


Lemma nin_true_interp_bv : 
  forall {T:Type} (x:T) (v:bv) (l:list T)
         (pf:length l = count_occ bool_eq_dec v false),
    ~ In true v ->
    interp_bv x v l pf = l.
  intros T x v l h1 h2.
  pose proof (bv_count_occ v) as h3.
  pose proof h2 as hnin.
  rewrite (count_occ_In bool_eq_dec) in h2.
  apply ngt_nlt in h2.
  rewrite not_lt0_iff in h2.
  rewrite h2 in h3.
  rewrite plus_0_l in h3.
  rewrite <- h1 in h3.
  revert h3 h1. revert l x.
  induction v as [|b v ih].
  intros ? ? ? h1.
  symmetry in h3.
  apply length_eq0 in h3; subst.
  unfold interp_bv; simpl; auto.
  intros l x h1 h4. 
  destruct b.
  contradict hnin; left; auto.
  erewrite interp_bv_false.
  symmetry.
  erewrite hd_compat' at 1.
  f_equal.
  symmetry.
  apply ih.
  rewrite count_occ_cons_neq in h2; auto.
  contradict hnin; right; auto.
  rewrite length_tl.
  simpl in h1.
  omega.

  Grab Existential Variables.
  simpl in h1.
  symmetry in h1.
  apply length_eq_S in h1; auto.
Qed.


Lemma remove_interp_bv :
  forall {T:Type} (pfdt:type_eq_dec T) (x:T)
         (v:bv) (l:list T)
         (pf:length l = count_occ bool_eq_dec v false),
    ~In x l ->
    remove pfdt x (interp_bv x v l pf) = l.
  intros T h1 x v.
  revert x.
  induction v as [|b v ih]; simpl.
  intros ? ? h2.
  apply length_eq0 in h2; auto.
  intros x l h3 h4.
  destruct b.
  rewrite interp_bv_true.
  rewrite remove_cons_eq.
  apply ih; auto.
  pose proof h3 as h3'.
  lr_if h3'; fold hlr in h3'; destruct hlr as [hl | hr].
  pose proof h3' as h5.
  apply length_eq_S in h5.
  rewrite (interp_bv_false _ _ _ _ h5).
  rewrite remove_cons_neq.
  rewrite ih.
  symmetry.
  apply hd_compat'.
  contradict h4.
  rewrite (hd_compat' _ h5).
  right; auto.
  intro; subst.
  contradict h4.
  apply in_hd'.
  contradict hr; auto.
Qed.

Lemma count_occ_interp_bv :
  forall {T:Type} (pfdt:type_eq_dec T) (x:T)
         (v:bv) (l:list T)
         (pfe:length l = count_occ bool_eq_dec v false),
    ~In x l ->
    count_occ pfdt (interp_bv x v l pfe) x =
    length (bv1s v).
  intros T h1 x v. revert x. 
  induction v as [|b v ih]; auto. 
  intros x l h2.
  destruct b.
  pose proof h2 as h2'.
  rewrite count_occ_cons_neq in h2'.
  intro h3.
  rewrite bv1s_cons_true.
  rewrite interp_bv_true.
  rewrite count_occ_cons_eq; auto.
  simpl.
  f_equal.
  rewrite length_lS.
  rewrite ih; auto.
  intro; discriminate.
  intro  h3.
  rewrite bv1s_cons_false.
  destruct l as [|a l].
  pose proof h2 as h2'.
  rewrite count_occ_cons_eq in h2'.
  discriminate.
  reflexivity.
  rewrite (interp_bv_false _ _ _ _ (cons_neq_nil _ _)).
  simpl.
  destruct (h1 a x) as [|h4]; subst.
  contradict h3; left; auto.
  rewrite ih.
  rewrite length_lS; auto.
  contradict h3; right; auto.
Qed.


Lemma nth_lt_interp_bv :
  forall {T:Type} (pfdt:type_eq_dec T)
         (x:T) (l:list T)
         (v:list bool)
         (n:nat)
         (pfe:length l = count_occ bool_eq_dec v false)
         (pf:In x (interp_bv x v l pfe))
         (pfn:n < length (interp_bv x v l pfe))
         (pfinn:In n (bv1s v)),
    ~In x l ->
    nth_lt (interp_bv x v l pfe) n pfn = x.
  intros T h1 x l v.
  revert l x. 
  induction v as [|b v ih]; simpl.
  intros; contradiction.
  intros l x n h2 h3 h4 h5 h6. 
  pose proof h2 as h2'.
  lr_if h2'; fold hlr in h2'; destruct hlr as [hl | hr]; subst.
  rewrite bv1s_cons_false in h5.
  rewrite in_lS_iff in h5.
  destruct h5 as [m' [hi h5]]; subst.
  gen0.
  erewrite interp_bv_false.
  intro h7.
  rewrite nth_lt_cons.
  rewrite ih; auto.
  pose proof h2' as h8.
  apply length_eq_S in h8.
  pose proof (interp_bv_false x v l) as h9.
  specialize (h9 h2 h8).
  simpl in h9.
  pose proof h3 as h3'.
  rewrite h9 in h3'.
  destruct h3' as [h10 | h10]; subst.
  apply in_interp_bv.
  pose proof hi as h10.
  pose proof (in_bv1s _ _ h10) as h11.
  simpl in h11.
  rewrite <- h11.
  apply in_nth_lt.
  assumption.
  contradict h6.
  apply in_tl in h6; auto.
  destruct b.
  gen0.
  rewrite interp_bv_true.
  intro h7.
  rewrite bv1s_cons_true in h5.
  destruct h5 as [|h5]; subst.
  rewrite nth_lt_0_hd'; auto.
  rewrite in_lS_iff in h5.
  destruct h5 as [m [h8 h9]]; subst. 
  rewrite nth_lt_cons.
  pose proof h7 as h9.
  simpl in h9.
  apply lt_S_n in h9.
  pose proof h8 as h10.
  pose proof (in_bv1s _ _ h10) as h12.
  simpl in h12.
  pose proof (in_nth_lt v m (in_bv1s_lt v m h10)) as h13.
  rewrite h12 in h13.
  apply ih; auto.
  apply in_interp_bv; auto.
  contradict hr; auto.

  Grab Existential Variables.
  apply length_eq_S in h2'; auto.
Qed.

Lemma in_firstn_iff :
  forall {T:Type} (l:list T) (x:T) (n:nat) (pfn:n <= length l),
    In x (firstn n l) <->
    exists (m:nat) (pfmn:m < n),
      nth_lt l m (lt_le_trans _ _ _ pfmn pfn)  = x.
  intros T l x n.
  revert l x.
  induction n as [|n ih]; simpl.
  intuition.
  destruct H as [? [? ?]]; omega.
  intros l x h1; split; intro h2.
  destruct l as [|a l].
  contradiction.
  simpl in h1, h2.
  pose proof h1 as h1'.
  apply le_S_n in h1'.
  destruct h2 as [|h2]; subst.
  exists 0, (lt_0_Sn _).
  simpl; auto.
  rewrite (ih _ _ h1') in h2.
  destruct h2 as [m [h2 h3]].
  exists (S m). 
  exists (lt_n_S _ _ h2).
  simpl.
  subst.
  rewrite nth_lt_compat.
  apply nth_indep. 
  eapply lt_le_trans.
  apply h2.
  assumption. 
  destruct l as [|a l].
  simpl in h1; omega.
  simpl in h1.
  pose proof h1 as h1'.
  apply le_S_n in h1'.
  destruct h2 as [m [h2 h3]].
  simpl in h3.
  destruct m as [|m]; subst.
  left; auto.
  simpl. 
  apply lt_S_n in h2.
  specialize (ih l (nth m l a) h1'). 
  hr ih. 
  exists m, h2.
  rewrite nth_lt_compat; apply nth_indep.
  eapply lt_le_trans.
  apply h2.
  assumption.
  rewrite <- ih in hr.
  right; auto.
Qed.

Lemma firstn_S_eq :
  forall {T:Type} (l:list T) (m:nat) (pflt:m < length l),
    firstn (S m) l = firstn m l ++ ((nth_lt l m pflt)::nil).
  intros T l.
  induction l as [| a l h1]; simpl. 
  intros; omega.
  intros m h2.
  destruct m as [|m]; simpl; auto.
  f_equal.
  apply lt_S_n in h2.
  specialize (h1 _ h2).
  rewrite nth_lt_compat in h1.
  erewrite nth_indep.
  rewrite <- h1 at 1.
  reflexivity.
  assumption.
Qed.


Lemma skipn_S_eq :
  forall {T:Type} (l:list T) (m:nat),
    skipn (S m) l = tl (skipn m l).
  intros ? l; simpl.
  induction l; simpl.
  intros; rewrite skipn_nil; auto.
  intro m; destruct m; simpl; auto.
Qed.

Lemma tl_skipn :
  forall {T:Type} (l:list T) (m:nat),
    tl (skipn m l) =
    skipn m (tl l).
  intros T l m.
  revert l.
  induction m as [|m ih]; simpl; auto.
  intro l.
  destruct l as [|a l]; simpl; auto.
  rewrite <- skipn_S_eq.
  simpl; auto.
Qed.

Lemma length_firstn_S_eq :
  forall {T:Type} (l:list T) (m:nat) (pflt:m < length l),
    length (firstn (S m) l) = S (length (firstn m l)).
  intros T l m h1.
  rewrite (firstn_S_eq _ _ h1).
  rewrite app_length; simpl.
  rewrite S_compat; auto.
Qed.
 
Lemma incl_firstn :
  forall {T:Type} (l:list T) (n:nat),
    n <= length l ->
    incl (firstn n l) l.
  intros; red; intros x h1.
  rewrite in_firstn_iff in h1.
  destruct h1 as [h1 [h2 h3]].
  rewrite <- h3.
  apply in_nth_lt.

  Grab Existential Variables.
  assumption.
Qed.


Lemma incl_firstn_le :
  forall {T:Type} (l:list T) (m n:nat),
    m <= n ->
    incl (firstn m l) (firstn n l).
  intros T l m n h0; red; intros ? h1.
  destruct (le_lt_dec n (length l)) as [h2 | h2]; subst.
  rewrite (in_firstn_iff _ _ _ (le_trans _ _ _ h0 h2)) in h1.
  rewrite (in_firstn_iff _ _ _ h2).
  destruct h1 as [r h3].
  destruct h3 as [h3 h4]; subst.
  exists r.
  ex_goal.
  eapply lt_le_trans.
  apply h3.
  assumption.
  exists hex.
  apply nth_lt_functional3.
  rewrite firstn_length_ge.
  apply in_firstn in h1; auto with arith.
  auto with arith.
Qed.


Lemma skipn_sing :
  forall {T:Type} (l:list T) (x:T) (n:nat),
    x::nil = skipn n l <->
    exists pf:length l = S n,
      let pf' := lt_eq_trans _ _ _ (lt_n_Sn _) (eq_sym pf) in
      nth_lt l n pf' = x.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros ? ?. split; intro h1.  rewrite skipn_nil in h1.
  discriminate.
  destruct h1; discriminate.
  intros x n.
  split; intro h1.
  destruct n as [|n].
  simpl in h1.
  inversion h1; subst.
  ex_goal; auto.
  exists hex; auto.
  simpl in h1.
  apply ih in h1; auto.
  destruct h1 as [h1 h2].
  simpl in h2.
  ex_goal.
  f_equal; auto.
  exists hex.
  rewrite nth_lt_compat in h2.
  erewrite nth_indep.
  rewrite h2 at 1; auto.
  rewrite h1.
  apply lt_n_Sn.
  destruct h1 as [h1 h2].
  apply S_inj in h1.
  destruct n as [|n]; simpl.
  subst; auto.
  apply length_eq0 in h1.
  subst; auto.
  subst.
  rewrite ih.
  exists h1.
  simpl.
  rewrite nth_lt_compat.
  apply nth_indep.
  rewrite h1; auto with arith.
Qed.


Lemma firstn_map :
  forall {T U:Type} (f:T->U) (l:list T) (n:nat),
    n <= length l ->
    firstn n (map f l) =
    map f (firstn n l).
  intros T U f l n.
  revert f l.
  induction n as [|n ih]; auto.
  intros f l h0.
  erewrite firstn_S_eq.
  erewrite firstn_S_eq.
  rewrite map_app.
  rewrite ih.
  simpl.
  rewrite nth_lt_map'.
  reflexivity.
  auto with arith.
  Grab Existential Variables.
  rewrite map_length; auto with arith.
Qed.

Lemma firstn_map_seg :
  forall {U:Type} {n:nat} (f:seg_fun n U) (m:nat)
         (pfm:m <= n),
    firstn m (map_seg f) =
    map_seg (fun i pf => f i (lt_le_trans _ _ _ pf pfm)).
  intros U n.
  induction n as [|n ih]; simpl.
  intros ? ? h1.
  pose proof h1 as h1'.
  apply le0 in h1'.
  subst.
  simpl; auto.
  intros f m h1.
  pose proof h1 as h1'.
  apply le_lt_eq_dec in h1'.
  destruct h1' as [h2|]; subst.
  apply lt_n_Sm_le in h2.
  specialize (ih (seg_fun_from_S f) m h2).
  rewrite firstn_app_le.
  rewrite ih at 1.
  f_equal.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality.
  intro y.
  unfold seg_fun_from_S.
  f_equal.
  apply proof_irrelevance.
  rewrite length_map_seg; auto with arith.
  rewrite firstn_length_ge.
  simpl.
  f_equal.
  f_equal.
  apply functional_extensionality_dep.
  intro x.
  unfold seg_fun_from_S.
  apply functional_extensionality.
  intro; f_equal; apply proof_irrelevance.
  repeat f_equal.
  apply proof_irrelevance.
  rewrite app_length.
  rewrite length_map_seg.
  simpl; rewrite S_compat; auto with arith.
Qed.


Lemma firstn_eq :
  forall {T:Type} (l:list T) (n:nat)
         (pflt:pred n < (length l))
         (pfnd:NoDup l),
    0 < n ->
    firstn n l =
    firstn (pred n) l ++ ((nth_lt _ _ pflt) :: nil).
  intros T l n. revert l.
  induction n as [|n h1]; simpl; auto.
  intros; omega.
  intros l h2 h3 h4.
  destruct (zerop n) as [h5 | h5]; subst.
  simpl. 
  destruct l.
  simpl in h2. omega.
  pose proof (nth_lt_compat (t :: l) 0 h2) as h5.
  simpl in h5.
  simpl.
  reflexivity.
  destruct l as [|a l].
  destruct n. simpl.
  pose proof (nth_lt_compat  nil 0 h2) as h6.
  simpl in h6. omega.
  simpl in h2. omega.
  simpl in h2.
  assert (h6:pred n < length l). omega.
  pose proof (no_dup_cons _ _ h3) as h7.
  specialize (h1 l h6 h7 h5).
  rewrite h1.
  rewrite firstn_cons.
  simpl.
  f_equal. f_equal.
  assert (h8:S (pred n) < S (length l)). omega.
  pose proof (nth_lt_cons l a (pred n) h8) as h9.
  simpl in h9.
  assert (h10: n = S (pred n)). omega.
  pose proof (subsetT_eq_compat _ (fun m => m < (S (length l))) _ _ h2 h8 h10) as h11.
  assert (h12:nth_lt (a::l) (S (pred n)) h8 =
              nth_lt (a::l) n h2).
    f_equal.
    dependent rewrite <-h11.
    auto.
  simpl in h12. rewrite <- h12.
  rewrite h9.
  f_equal. f_equal. apply proof_irrelevance.
  assumption.
Qed.


Lemma firstn_eq_firstn_cons_iff :
  forall {T:Type} (m:nat) (l:list T) (a:T), 
    m <= length l ->
    (firstn m l = firstn m (a::l) <->
     all_sing (firstn m l) a).
  intros T m.
  induction m as [|m ih]; simpl.
  intros; split; intro; auto. 
  apply all_sing_nil. 
  intros l a hl; split; intro h1.
  destruct l as [|b l].
  discriminate.
  inversion h1 as [h2]; subst.
  rewrite all_sing_cons_iff.
  split; auto. 
  simpl in hl.
  rewrite <- ih; auto with arith.
  destruct l as [|b l].
  simpl in hl; omega.
  simpl in hl.
  apply le_S_n in hl; auto.
  rewrite all_sing_cons_iff in h1.
  destruct h1 as [? h1]; subst.
  f_equal.
  rewrite ih; auto.
Qed.

Lemma length_firstn :
  forall {T:Type} (l:list T) (m:nat),
    m <= length l ->
    length (firstn m l) = m.
  intros T l m. revert l.
  induction m as [|m ih]; simpl; auto.
  intros l h1.
  destruct l as [|a l]; simpl.
  simpl in h1; omega.
  simpl in h1.
  apply le_S_n in h1.
  f_equal.
  apply ih; auto.
Qed.


Lemma length_firstn_S :
  forall {T:Type} (l:list T) (m:nat),
    m < length l ->
    length (firstn (S m) l) = S m.
  intros T l m. revert l.
  induction m as [|m ih]; simpl; auto.  
  intros l h1. 
  destruct l as [|a l]; simpl.
  simpl in h1; omega.
  reflexivity.
  intros l h1.
  destruct l as [|a l].
  simpl in h1; omega.
  simpl in h1.
  apply lt_S_n in h1.
  simpl.
  f_equal; apply ih; auto.
Qed.

                                
Lemma length_firstn_le :
  forall {T:Type} (l:list T) (m:nat),
    length (firstn m l) <= length l.
  intros T l.
  induction l as [|a l h1]; simpl.
  intro m.
  destruct m; simpl; auto with arith.
  intro m.
  destruct m; simpl; auto with arith.
Qed.

  
Lemma length_eq_firstn :
  forall {T U:Type} (l:list T) (l':list U)
         (m:nat),
      length l = length l' ->
      length (firstn m l) = length (firstn m l').
  intros T U l l' m h1.
  destruct (le_dec m (length l)) as [h2 | h2].
  rewrite length_firstn; auto.
  rewrite length_firstn; auto.
  eapply le_eq_trans.
  apply h2.
  assumption.
  erewrite firstn_length_ge.
  erewrite firstn_length_ge; auto.
  omega.
  omega.
Qed.

Lemma eq_firstn_iff :
  forall {T} (l l' l0:list T),
    length l = length l' ->
    (l = firstn (length l) (l' ++ l0) <->
     l = l').
  intros T l; induction l as [|a l ih]; simpl.
  intros l l' h1.
  symmetry in h1.
  apply length_eq0 in h1; subst.
  tauto.
  intro l'.
  destruct l' as [|a' l']. 
  simpl.
  intro; discriminate.
  intro l0; destruct l0 as [|a0 l0]; simpl.
  intro h1.
  apply S_inj in h1.
  pose proof h1 as h2.
  eapply ih in h2.
  split; intro h3.
  inversion h3.
  rewrite h2 in H1; auto; subst; auto.
  inversion h3; subst.
  f_equal.
  rewrite h2; auto.
  intro h1.
  apply S_inj in h1.
  split; intro h2.
  inversion h2.
  revert h1 h2; revert ih H1; revert l; subst.
  intros l h2 ih h3 h4.
  f_equal.
  pose proof ih as h5.
  rewrite h2 in ih; auto; subst.
  symmetry; rewrite h2; auto.
  inversion h2; auto; subst.
  f_equal.
  rewrite ih; auto.
Qed.

Lemma lt_length_firstn :
  forall {T:Type} (l:list T) (m n:nat),
    m < n ->
    m < length l ->
    m < length (firstn n l).
  intros T l m n. revert m l.
  induction n as [|n ih]; simpl; auto.
  intros m ? ? h2.
  destruct l; auto.
  simpl in h2; simpl.
  destruct m; auto with arith.
Qed.

  


Lemma firstn_idem :
  forall {T:Type} (l:list T) (m:nat),
    firstn m (firstn m l) = firstn m l.
  intros T l m.
  destruct (le_dec m (length l)) as [h1 | h1].
  apply firstn_length_ge.
  rewrite length_firstn; auto with arith.
  apply not_le in h1.
  rewrite firstn_length_ge; auto with arith.
  rewrite firstn_length_ge; auto with arith.
Qed.


Lemma firstn_le_nest :
  forall {T:Type} (l:list T) (m n:nat),
    m <= n ->
    firstn m l = firstn m (firstn n l).
  intros T l m n h1.
  apply le_lt_eq_dec in h1.
  destruct h1; subst.
  apply firstn_lt_nest; auto.
  symmetry.
  apply firstn_idem.
Qed.

  
Lemma length_skipn :
  forall {T:Type} (l:list T) (m:nat),
    length (skipn m l) = length l - m.
  intros ? l m; revert l.
  induction m as [|m ih]; simpl; auto with arith.
  intro l; destruct l; simpl; auto.
Qed.

Lemma skipn_firstn_nil :
  forall {T:Type} (l:list T) (n:nat),
    skipn n (firstn n l) = nil.
  intros T l n. revert l.
  induction n; simpl; auto.
  intro l; destruct l as [|a l]; simpl; auto.
Qed.

Lemma eq_firstn_iff' :
  forall {T} (l l' l0:list T),
    l = firstn (length l) (l' ++ l0) <->
    (length l < length l' -> (l = firstn (length l) l' /\
                              skipn (length l) l' <> nil)) /\
    (length l = length l' -> l = l') /\
    (length l > length l' -> l = l' ++ firstn (length l - length l') l0).
  intros T l l' l0; split; intros h1; try split; try intro h2. 
  pose proof (firstn_length_eq _ _ (eq_sym h1)) as h3.  
  split.
  rewrite firstn_app_le in h1; auto with arith.
  intro h4.
  rewrite skipn_eq_nil_iff in h4.
  destruct h4; revert h1; subst.
  simpl in h2; contradict (lt_n_O _ h2).
  intro h4.
  omega. 
  split.
  intro h2.
  apply eq_firstn_iff in h1; auto. 
  intros h2; revert h1; subst; intro h3.
  rewrite firstn_app_ge in h3; auto with arith.
  destruct h1 as [h1 [h2 h3]].
  destruct (lt_eq_lt_dec (length l) (length l')) as [[h4|h4]|h4].
  specialize (h1 h4).
  destruct h1 as [h1 h5].
  rewrite firstn_app_le; auto with arith.
  apply h2 in h4; subst.
  rewrite firstn_app_eq; auto.
  specialize (h3 h4). 
  rewrite firstn_app_ge; auto with arith.
Qed.


Lemma in_nth_lt_firstn :
  forall {T:Type} (l:list T) (m n:nat) (pfn:n <= length l)
         (pfm:m < n),
    In (nth_lt (firstn n l) m (lt_eq_trans _ _ _  pfm (eq_sym (length_firstn _ _ pfn)))) (firstn n l).
  intros; apply in_nth_lt.
Qed.



Lemma nth_lt_firstn_eq :
  forall {T:Type} (l:list T) (m n:nat) (pf:m < length (firstn n l)),
    nth_lt (firstn n l) m pf =
    nth_lt l m ( lt_le_trans _ _ _ pf (length_firstn_le l n)).
  intros T l m n h1. simpl.
  pose proof (firstn_skipn n l) as h2.
  symmetry.
  gen0. generalize h1.
  rewrite <- h2; auto.
  intros h3 h4.
  rewrite (nth_lt_app1 _ _ _ h1).
  generalize h3.
  rewrite h2;
  intros h5.
  f_equal.
  apply proof_irrelevance. 
Qed.


Lemma nth_lt_skipn_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (m n:nat)
         (pfm:m < length (skipn n l))
         (pfmn:m + n < length l),
    nth_lt (skipn n l) m pfm =
    nth_lt l (m + n) pfmn.
  intros T hdt l m n.
  revert l m.
  induction n as [|n ih].
  simpl; intros; apply nth_lt_functional2; auto with arith.
  intros l m h1 h2.
  gen0.
  rewrite skipn_S_eq.
  intro h3.
  symmetry.
  gen0.
  rewrite plus_S_shift_r.
  intro.
  rewrite nth_lt_S.
  erewrite <- ih.
  apply nth_lt_functional1.
  rewrite tl_skipn; auto.

  Grab Existential Variables.
  rewrite length_skipn.
  rewrite length_tl in h3.
  rewrite length_skipn in h3.
  rewrite length_tl; auto.
  rewrite length_skipn in h1.
  rewrite pred_minus in h3; auto with arith. 
  omega.
Qed.


Lemma all_sing_app_eq :
  forall {T} (l0 l0' l1 l1':list T) a b,
    a <> b ->
    all_sing l0 a ->
    all_sing l0' a ->
    l0 ++ b :: l1 =
    l0' ++ b :: l1' ->
    l0 = l0' /\ l1 = l1'.
  intros T l0 l0' l1 l1' a b h1 h2 h3 h4.
  red in h2, h3.
  apply neq_sym in h1.
  apply app_nin_eq in h4; auto.
Qed.


Lemma in_nth_lt_firstn' :
  forall {T:Type} (l:list T) (m n:nat) (pfn:n <= length l)
         (pfm:m < n),
    In (nth_lt l m (lt_le_trans _ _ _  pfm pfn)) (firstn n l).
  intros.
  erewrite nth_lt_functional3.
  rewrite <- nth_lt_firstn_eq at 1.
  apply in_nth_lt_firstn.
  Grab Existential Variables.
  assumption.
  assumption.
Qed.


Lemma nth_lt_removelast :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat)
         (pf':n < length (removelast l))
         (pf:n < length l),
    nth_lt (removelast l) n pf' =
    nth_lt l n pf.
  intros; gen0; rewrite removelast_eq; intro; rewrite nth_lt_firstn_eq.
  apply nth_lt_functional3.
Qed.


Lemma lst_removelast :
  forall {T:Type} {hdt:type_eq_dec T}
         (l:list T) (pf:removelast l <> nil),
let pflt := removelast_not_nil' _ pf in
lst (removelast l) pf =
nth_lt l (pred (pred (length l))) pflt.
  intros; erewrite lst_eq, nth_lt_removelast; auto.
  gen0.
  rewrite length_removelast; intros; apply nth_lt_functional3; apply proof_irrelevance.

  Grab Existential Variables.
  rewrite length_removelast; auto.
Qed.
  


Lemma firstn_map_in_dep_firstn :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U)
         (m n:nat)
         (pfm:m <= n)
         (pfn:n <= length l),
    firstn m (map_in_dep (fun x (pf:In x (firstn n l)) =>
                            f x (incl_firstn  _ _ pfn _ pf))) =
    firstn m (map_in_dep f). 
  intros T U l f m. revert f.  revert l.
  induction m as [|m ih]; auto.
  intros l f n h1 h2.
  erewrite firstn_S_eq.
  erewrite firstn_S_eq.
  rewrite nth_lt_map_in_dep'.
  rewrite nth_lt_map_in_dep'.
  f_equal.
  apply ih.
  auto with arith.
  f_equal.
  gen0.
  intro h3.
  apply f_equal_dep.
  rewrite nth_lt_firstn_eq.
  apply nth_lt_functional3.

  Grab Existential Variables.
  rewrite length_map_in_dep.
  omega.
  rewrite length_map_in_dep, length_firstn; auto.
Qed.




Lemma in_firstn_S_dec :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (m:nat)
         (pf:S m <= length l) (x:T),
    In x (firstn (S m) l) ->
    {In x (firstn m l)} + {x = nth_lt l m pf}.
  intros T hdt l m h1 x h2.
  destruct (hdt x (nth_lt l m h1)); subst.
  right; auto.
  left.
  rewrite (in_firstn_iff _ _ _ h1) in h2.
  destruct h2 as [r [h2 h3]].
  rewrite <- h3.
  rewrite in_firstn_iff.
  exists r. 
  ex_goal.
  red in h2.
  pose proof h2 as h2'.
  apply le_S_n in h2'.
  apply le_lt_eq_dec in h2'.
  destruct h2'; subst; auto.
  contradict n.
  apply nth_lt_functional3.
  exists hex.
  apply nth_lt_functional3.

  Grab Existential Variables.
  auto with arith.
Qed.


Lemma impl_lt_length_firstn :
  forall {T:Type} (l:list T) (m n:nat),
    n <= length l ->
    m < n ->
    m < (length (firstn n l)).
  intros; rewrite length_firstn; auto.
Qed.

Lemma lt_length_firstn_impl :
  forall {T:Type} (l:list T) (m n:nat),
    n <= length l  ->
    m < (length (firstn n l)) ->
    m < n.
  intros ? ? ? ? ? h1; rewrite length_firstn in h1; auto.
Qed.

Lemma nth_firstn :
  forall {T:Type} (l:list T) (m n:nat)
         (def:T),
    m < n ->
    nth m (firstn n l) def = nth m l def. 
  intros T l.
  induction l as [|a l h1]; simpl.
  intros m n def h3. 
  destruct n as [|n]. 
  omega. 
  rewrite firstn_nil; auto.
  intros m n def h2.
  destruct n as [|n]; simpl.
  omega. 
  red in h2.
  apply le_S_n in h2.
  apply le_lt_eq_dec in h2.
  destruct m as [|m]; auto.
  destruct h2 as [h2 | h2].
  assert (h3:m < n). omega.
  specialize (h1 _ _ def h3); auto.
  subst.
  assert (h2: m < S m). auto with arith.
  specialize (h1 _ _ def h2).
  assumption.
Qed.


Lemma nth_firstn' :
  forall {T:Type} (l:list T) (m n:nat)
         (d1 d2:T),
    m < length l ->
    m < n ->
    nth m (firstn n l) d1 = nth m l d2.
  intros T l m n d1 d2 h0 h1.
  rewrite nth_firstn; auto. 
  apply nth_indep; auto.
Qed.


  
Lemma nth_lt_map_seq :
  forall {U:Type} {n w:nat} (f:seq_fun n w U) (m:nat)
         (pf:m < w),
    let pf' := lt_eq_trans _ _ _ pf (eq_sym (length_map_seq f)) in
    let pf'' := lt_seq _ _ _  pf in
    nth_lt (map_seq f) m pf' =
    f _ pf''.
  intros U n w. revert n.
  induction w as [|w ih]; simpl.
  intros. omega.
  intros  n f m  h1.
  simpl in ih.
  red in h1.
  pose proof h1 as h1'. apply le_S_n in h1'. 
  destruct (le_lt_eq_dec _ _ h1') as [h2 | h2].
  specialize (ih n (seq_fun_from_S f) m h2).
  erewrite nth_lt_app1.
  unfold seq_fun_from_S in ih.
  rewrite ih at 1. 
  f_equal. apply proof_irrelevance.
  subst.
  erewrite nth_lt_app2.
  simpl.
  rewrite length_map_seq.
  rewrite minus_same.
  f_equal.
  apply proof_irrelevance.
  Grab Existential Variables.
  rewrite length_map_seq; auto with arith.
Qed.

Lemma nth_lt_lrep :
  forall {T:Type} (a:T) (m n:nat)
         (pf:m < length (lrep a n)),
    nth_lt (lrep a n) m pf = a.
  intros T a m n. revert m a.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros m a h1.
  destruct m as [|m]; auto.
  apply lt_S_n in h1.
  specialize (ih _ _ h1).
  rewrite <- ih at 3.
  rewrite nth_lt_compat at 1.
  apply nth_indep; auto.
Qed.


Lemma count_occ_lrep :
  forall {T:Type} (pfdt:type_eq_dec T) (a:T) (n:nat),
    count_occ pfdt (lrep a n) a = n.
  intros T h1 a n. revert a.
  induction n as [|n ih]; simpl; auto.
  intro a.
  erewrite eq_dec_same at 1.
  f_equal.
  apply ih.
  assumption.
Qed.
  

Lemma count_occ_lrep' :
  forall {T:Type} (pfdt:type_eq_dec T) (a x:T) (n:nat),
    In x (lrep a n) ->
    count_occ pfdt (lrep a n) x = n.
  intros T h1 a x n h2.
  apply in_lrep' in h2; subst.
  apply count_occ_lrep.
Qed.


Lemma sub_list_firstn_removelast_le :
  forall {T:Type} (l:list T) (n:nat),
    sub_list (firstn n (removelast l)) (firstn n l).
  intros T l n.
  destruct (nil_dec' l) as [|h1]; subst.
  simpl.
  apply sub_list_refl.
  pose proof (app_removelast_last (hd'  _ h1) h1) as h2.
  rewrite h2 at 2.
  destruct (le_dec n (length (removelast l))) as [h3 | h3].
  rewrite firstn_app_le; auto with arith.
  apply sub_list_refl.
  apply not_le in h3.
  rewrite firstn_length_ge; auto with arith.
  rewrite firstn_length_ge; auto with arith.
  rewrite <- (app_nil_l (removelast l ++ last l (hd' l h1) :: nil)).
  constructor.
  rewrite app_length.
  simpl.
  omega.
Qed.
  

Lemma count_occ_nth_lt0 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (pf:0 < length l),
    count_occ pfdt l (nth_lt l 0 pf) =
    S (count_occ pfdt (tl l) (nth_lt l 0 pf)).
  intros T h1 l.   
  destruct l as [|a l]; simpl.
  intro h2; apply lt_irrefl in h2; contradiction.
  intro h2.
  destruct (h1 a a) as [|h4]; auto.
  contradict h4; auto.
Qed.


Lemma length_list_singularize_map_inj :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (f:T->U),
    Injective f ->
    length (list_singularize pfdu (map f l)) =
    length (list_singularize pfdt l).
  intros T U h1 h2 l.
  induction l as [|a l ih]; simpl; auto.
  intros f h3.
  destruct (in_dec h2 (f a) (map f l)) as [hl | hr].
  rewrite list_singularize_in_cons; auto.
  destruct (in_dec h1 a l) as [hl' | hr'].
  rewrite list_singularize_in_cons; auto.
  rewrite in_map_iff in hl.
  destruct hl as [a' h4].
  destruct h4 as [h4 h5].
  apply h3 in h4; subst; contradiction.
  rewrite list_singularize_nin; auto.
  destruct (in_dec h1 a l) as [hl' | hr'].
  rewrite list_singularize_in_cons; auto.
  contradict hr.
  rewrite in_map_iff.
  exists a; tauto.
  rewrite list_singularize_nin; auto.
  simpl; auto.
Qed.

Lemma length_list_singularize_no_dup :
  forall {T} (pfdt:type_eq_dec T) l,
    NoDup l ->
    length (list_singularize pfdt l) =
    length l.
  intros T hdt l h1.
  rewrite list_singularize_no_dup in h1.
  rewrite h1; auto.
Qed.


Lemma nth_lt_list_singularize_no_dup :
  forall {T} (pfdt:type_eq_dec T) (l:list T)
         (pfndp:NoDup l) i (pfi:i < length (list_singularize pfdt l)),
    let pfi' := lt_eq_trans
                  _ _ _ pfi
                  (length_list_singularize_no_dup pfdt l pfndp) in
    nth_lt (list_singularize pfdt l) i pfi =
    nth_lt l i pfi'.
  intros; apply nth_lt_functional1; apply list_singularize_no_dup; auto.
Qed.

    
Lemma nth_lt_list_singularize_map :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (f:T->U) (i:nat)
         (pfinj:Injective f)
         (pflt:i < length (list_singularize pfdt l)),
    let pflt' := lt_eq_trans
                 _ _ _
                 pflt
                 (eq_sym (length_list_singularize_map_inj pfdt pfdu l f pfinj)) in
      nth_lt (list_singularize pfdu (map f l)) i pflt' =
      f (nth_lt (list_singularize pfdt l) i pflt).
  intros T U h1 h2 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros f i h3 h4.
  destruct (in_dec h2 (f a) (map f l)) as [hl | hr].
  pose proof hl as hl'.
  rewrite in_map_iff in hl'.
  destruct hl' as [a' h5].
  destruct h5 as [h5 h5'].
  apply h3 in h5; subst.
  gen0.
  generalize h4.  
  rewrite list_singularize_in_cons.
  rewrite list_singularize_in_cons; auto.
  intros h5 h6.
  specialize (ih f i h3 h5).
  simpl in ih.
  rewrite <- ih.
  apply nth_lt_functional3.
  assumption.
  gen0.
  rewrite list_singularize_nin; auto.
  intro h5.
  assert (h6:~In a l).
    contradict hr.
    rewrite in_map_iff.
    exists a; tauto.
  generalize h4.
  rewrite list_singularize_nin; auto.
  intro h7.
  simpl.
  destruct i as [|i]; auto.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  apply ih.
  simpl in h7; auto with arith.
  simpl in h7; auto with arith.

  Grab Existential Variables.
  assumption.
  simpl in h7; auto with arith.
Qed.


Lemma nth_lt_list_singularize_map_in_dep :
  forall {T U:Type} {l:list T} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U) (i:nat)
         (pfinj:inj_dep f)
         (pflt:i < length (list_singularize pfdt l)),
    let pflt' := lt_eq_trans
                 _ _ _
                 pflt
                 (eq_sym (length_list_singularize_map_in_dep_inj pfdt pfdu f pfinj)) in
      nth_lt (list_singularize pfdu (map_in_dep f)) i pflt' =
      f _ ((iff2 (list_singularize_in_iff pfdt l (nth_lt (list_singularize pfdt l) i pflt))) (in_nth_lt (list_singularize pfdt l) i pflt)).
  intros T U l h1 h2.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros f i h3 h4. 
  destruct (in_dec h2 (f a (in_eq _ _)) (map_in_dep (fun_from_cons f))) as [hl | hr].
  pose proof hl as hl'.
  rewrite in_map_in_dep_iff in hl'.
  destruct hl' as [a' h5].
  destruct h5 as [h5 h5'].
  pose proof h3 as h3'.
  red in h3'.
  unfold fun_from_cons in h5'.
  apply h3' in h5'; subst.
  gen0.
  generalize h4. 
  rewrite (list_singularize_in_cons h2 _ _ hl); auto.
  intros h5' h6.
  specialize (ih (fun_from_cons f) i).
  hfirst ih.
  apply inj_from_cons; auto.
  specialize (ih hf).
  hfirst ih.
  rewrite list_singularize_in_cons in h4; auto.
  specialize (ih hf0).
  simpl in ih.
  erewrite nth_lt_functional3.
  rewrite ih at 1.
  unfold fun_from_cons; apply f_equal_dep.
  symmetry.
  gen0.
  rewrite list_singularize_in_cons.
  intro h7.
  apply nth_lt_functional3; auto.
  assumption.
  gen0.
  rewrite list_singularize_nin; auto.
  intro h5. 
  assert (h6:~In a l).
    contradict hr.
    rewrite in_map_in_dep_iff.
    exists a, hr.
    unfold fun_from_cons; f_equal; apply proof_irrelevance. 
  symmetry.
  gen0.
  generalize h4.
  rewrite (list_singularize_nin h1 l a); auto.
  intros h7 h10.
  simpl.
  destruct i as [|i]; auto.
  f_equal; apply proof_irrelevance.
  symmetry.
  specialize (ih (fun_from_cons f) i).
  hfirst ih.
  apply inj_from_cons; auto.
  specialize (ih hf).
  hfirst ih.
  rewrite (list_singularize_nin h1 _ _ h6) in h4.
  simpl in h4; auto with arith.
  specialize (ih hf0).
  simpl in ih.
  rewrite nth_lt_compat in ih.
  erewrite nth_indep.
  rewrite ih at 1.
  unfold fun_from_cons.
  apply f_equal_dep.
  erewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  simpl in h5; auto with arith.
Qed.



Lemma list_singularize_map_inj :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (f:T->U),
    Injective f ->
    list_singularize pfdu (map f l) =
    map f (list_singularize pfdt l).
  intros T U h1 h2 l f h3. 
  symmetry.
  rewrite map_eq_iff.
  exists (eq_sym (length_list_singularize_map_inj _ _ _ _ h3)).
  intros n h4.
  symmetry; apply nth_lt_list_singularize_map; auto.
Qed.

Lemma list_singularize_map_in_dep_inj :
  forall {T U:Type} {l:list T}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U),
    inj_dep f ->
    list_singularize pfdu (map_in_dep f) =
    map_in_dep (fun x (pf:In x (list_singularize pfdt l)) =>
                  f x (iff2 (list_singularize_in_iff pfdt l x) pf)).
  intros T U l h1 h2 f h3.
  symmetry.
  rewrite map_in_dep_eq_iff.
  exists (eq_sym (length_list_singularize_map_in_dep_inj _ _ _  h3)).
  intros n h4.
  symmetry; apply nth_lt_list_singularize_map_in_dep; auto.
Qed.


Lemma list_singularize_firstn_S :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (m:nat) (pfsm: S m <= length l),
    list_singularize pfdt (firstn (S m) l) =
   remove pfdt (nth_lt l m pfsm) (list_singularize pfdt (firstn m l)) ++
                     nth_lt l m pfsm :: nil.
  intros T h1 l m h2.
  erewrite firstn_S_eq.
  apply list_singularize_rev_cons.
Qed.


Lemma incl_list_singularize_firstn_S_app_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (n:nat),
    S n <= length (l++a::nil) ->
    incl (list_singularize pfdt (firstn n l))
          (list_singularize pfdt (firstn (S n) (l++a::nil))).
  intros T hdt l a n hn; apply incl_list_singularize_compat; red; intros x h1.
  rewrite app_length in hn; rewrite S_compat in hn.
  apply le_S_n in hn.
  rewrite (in_firstn_iff _ _ _ hn) in h1.
  destruct h1 as [m [h1 h2]]; subst.
  erewrite in_firstn_iff.
  exists m. 
  exists (lt_trans _ _ _ h1 (lt_n_Sn _)).
  erewrite nth_lt_app1.
  apply nth_lt_functional3.
  Grab Existential Variables.
  eapply lt_le_trans.
  apply h1.
  assumption.
  rewrite app_length; rewrite S_compat; auto with arith.
Qed.


Lemma count_occ_map_in_dep :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (f:fun_in_dep l U)
         (y:U),
    count_occ pfdu (map_in_dep f) y =
    length (inv_im_in_dep pfdu l y f).
  intros T U h1 h2 l.
  induction l as [|a l ih]; simpl; auto.
  intros f y.
  rewrite switch_eq_dec_eq at 1.
  destruct (h2 y (f a (in_eq a l))) as [h4 | h4]; subst.
  unfold switch_eq_dec; simpl.
  f_equal; apply ih; auto.
  unfold switch_eq_dec.
  apply ih; apply inj_from_cons; auto.
Qed.

Lemma count_occ_map_in_dep_inj :
  forall {T U:Type} {l:list T}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U) (x:T)
         (pfx:In x l),
    inj_dep f ->
    count_occ pfdu (map_in_dep f) (f x pfx) =
    count_occ pfdt l x.
  intros T U l h1 h2.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros f x h3 h4. 
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  apply h4 in hl; subst.
  erewrite eq_dec_same at 1.
  f_equal.
  destruct (in_dec h1 x l) as [h5 | h5].
  specialize (ih (fun_from_cons f) x h5).
  hfirst ih.
  apply inj_from_cons; auto.
  specialize (ih hf).
  unfold fun_from_cons in ih at 2.
  pose proof (proof_irrelevance _ h3 (or_intror h5)); subst.
  rewrite <- ih; repeat f_equal; apply proof_irrelevance.
  pose proof (count_occ_In h1 l x) as h6.
  rewrite h6 in h5.
  pose proof (count_occ_In h2 (map_in_dep (fun_from_cons f)) (f x h3)) as h7.
  apply ngt_nlt in h5.
  rewrite not_lt0_iff in h5.
  rewrite h5.
  rewrite <- not_lt0_iff.
  apply ngt_nlt.
  rewrite <- (count_occ_In h2).
  intro h8.
  rewrite in_map_in_dep_iff in h8.
  destruct h8 as [x' [h8 h9]].
  unfold fun_from_cons in h9.
  apply h4 in h9; subst.
  rewrite h6 in h8.
  omega.
  assumption.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr']; subst.
  contradict hr.
  f_equal; apply proof_irrelevance.
  destruct h3 as [|h3]; try contradiction.
  specialize (ih (fun_from_cons f) x h3).
  hfirst ih.
  apply inj_from_cons; auto.
  specialize (ih hf).
  unfold fun_from_cons in ih at 2; auto.
  rewrite <- ih.
  repeat f_equal; apply proof_irrelevance.
Qed.


Lemma map_seg_eq_in_dep :
  forall {U:Type} {n:nat} (f:seg_fun n U),
    map_seg f =
    map_in_dep (seg_fun_in_dep f).
  intros U n f.
  symmetry. rewrite map_in_dep_eq_iff.
  ex_goal.
  rewrite length_seg_list, length_map_seg; auto.
  exists hex.
  intros m h1.
  unfold seg_fun_in_dep.
  gen0.
  rewrite nth_lt_seg_list.
  intro h2.
  erewrite nth_lt_functional3. rewrite nth_lt_map_seg.
  reflexivity.
Qed.

Lemma count_occs_eq_list_to_set :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    (forall x:T, count_occ pfdt l x = count_occ pfdt l' x) ->
    list_to_set l = list_to_set l'.
  intros T h1 l l' h2.
  apply Extensionality_Ensembles; red; split; red; intros x h3; rewrite <- list_to_set_in_iff in h3; rewrite <- list_to_set_in_iff.
  specialize (h2 x).
  apply (count_occ_In h1) in h3.
  rewrite h2 in h3.
  apply (count_occ_In h1); auto.
  specialize (h2 x).
  apply (count_occ_In h1) in h3.
  rewrite <- h2 in h3.
  apply (count_occ_In h1); auto.
Qed.

Lemma count_occs_eq_impl :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    (forall x:T, count_occ pfdt l x = count_occ pfdt l' x) ->
    length (list_singularize pfdt l) =
    length (list_singularize pfdt l').  
  intros T h1 l l' h2.
  apply count_occs_eq_list_to_set in h2.
  apply list_to_sets_eq_length; auto.
Qed.



Lemma map_seg_eq_iff :
  forall {U:Type} (n:nat) (f:seg_fun n U) (l:list U),
    map_seg f = l <->
    (exists pfeq:n = length l,
       forall (m:nat) (pflt:m < n),
         let pflt' := lt_eq_trans _ _ _ pflt pfeq in
         f m pflt = 
         nth_lt l m pflt').
  intros U n.
  induction n as [|n ih]; simpl.
  intros.
  split; intro h1.
  subst.
  simpl.
  exists eq_refl.
  intros; omega.
  destruct h1 as [h1 h2]. 
  symmetry.
  apply length_eq0; auto.
  intros f l.
  rewrite app_eq_iff.
  rewrite length_map_seg.
  split; intro h1.
  destruct h1 as [h1 h2]. 
  apply skipn_sing in h2.
  destruct h2 as [h2 h2'].
  simpl in h2'.
  exists (eq_sym h2).
  intros m h4.
  pose proof h4 as h4'.
  apply le_S_n in h4'.
  apply le_lt_eq_dec in h4'.
  destruct h4' as [h5 | h5].
  rewrite ih in h1.
  destruct h1 as [h1 h6].
  specialize (h6 _ h5).
  simpl in h6.
  unfold seg_fun_from_S in h6.
  rewrite nth_lt_firstn_eq in h6.
  erewrite nth_lt_functional3.
  rewrite <- h6 at 1.
  f_equal. apply proof_irrelevance.
  subst.
  erewrite nth_lt_functional3.
  rewrite  h2' at 1.
  f_equal.
  apply proof_irrelevance.
  destruct h1 as [h1 h2].
  split.
  rewrite ih.
  ex_goal.
  rewrite length_firstn; auto.
  rewrite <- h1; auto with arith.
  exists hex.
  intros m h3; simpl.
  unfold seg_fun_from_S.
  rewrite h2.
  rewrite nth_lt_firstn_eq. apply nth_lt_functional3.
  rewrite skipn_sing.
  exists (eq_sym h1).
  simpl.
  rewrite h2.
  apply nth_lt_functional3.
Qed.

Lemma map_segs_eq_iff :
  forall {U:Type} (n:nat) (f g:seg_fun n U),
    map_seg f = map_seg g  <->
    forall i (pflt:i < n),
      f i pflt =
      g i pflt.
  intros U n f g.
  rewrite map_seg_eq_iff.
  split; intro h1.
  destruct h1 as [h1 h2].
  simpl in h2.
  intros i h3.
  specialize (h2 i h3).
  rewrite h2.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  reflexivity.
  ex_goal.
  rewrite length_map_seg; auto.
  exists hex.
  intros i h2; simpl.
  specialize (h1 _ h2).
  rewrite h1.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  reflexivity.
Qed.


Lemma map_seg_id_seg :
  forall n,
    map_seg (id_seg n) = seg_list n.
  intro n.
  unfold id_seg.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_seg_list; auto.
  exists hex.
  intros m h1 h2.
  rewrite nth_lt_seg_list; auto.
Qed.

  

Lemma map_seg_nth_lt :
  forall {T:Type} (l:list T),
    map_seg (nth_lt l) = l.
  intros; rewrite map_seg_eq_iff.
  exists eq_refl.
  intros; f_equal; apply proof_irrelevance.
Qed.


Lemma removelast_map_seg :
  forall {U:Type} {n:nat}
         (pfdu:type_eq_dec U) (f:seg_fun n U)
         (pf:0 < n),
    removelast (map_seg f) =
    map_seg (seg_fun_rest f (lt_pred_n _ _ pf)).
  intros U hdu n f h1.
  symmetry.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_removelast, length_map_seg; auto.
  exists hex.
  intros m h2 h3.
  erewrite nth_lt_removelast.
  rewrite seg_fun_rest_compat, nth_lt_map_seg.
  reflexivity.
  assumption.
Qed.


Lemma map_seg_last :
  forall {U:Type} {n:nat} (f:seg_fun (S n) U),
    map_seg f =
    map_seg (seg_fun_from_S f) ++ (f n (lt_n_Sn _)) ::nil.
  intros U n f.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite app_length; simpl.
  rewrite S_compat.
  f_equal.
  rewrite length_map_seg; auto.
  exists hex.
  intros m h1 h2.
  unfold seg_fun_from_S. 
  erewrite nth_lt_functional3.
  pose proof h1 as h1'.
  apply lt_S_dec in h1'.
  destruct h1' as [h3 | h3].
  erewrite nth_lt_app1 at 1.
  erewrite nth_lt_map_seg.
  f_equal; apply proof_irrelevance.
  subst.
  erewrite nth_lt_app2 at 1.
  symmetry.
  gen0.
  rewrite length_map_seg.
  intro h3.
  simpl.
  rewrite minus_same.
  f_equal; apply proof_irrelevance.
  Grab Existential Variables.
  rewrite length_map_seg; auto with arith.
  rewrite app_length; simpl.
  rewrite S_compat.
  rewrite length_map_seg; auto with arith.
  assumption.
Qed.



Lemma map_seq_fun_to_seg :
  forall {U:Type} {n w:nat} (f:seq_fun n w U),
    map_seq (fun m pf =>
               let pf' := proj2 (lt_seq' m n w pf) in
               seq_fun_to_seg f (m - n) pf') = 
    map_seg (seq_fun_to_seg f).
  intros U n w f. 
  symmetry.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_map_seq; auto.
  exists hex.
  intros m h1.
  simpl.
  unfold seq_fun_to_seg.
  erewrite nth_lt_functional3.
  rewrite (nth_lt_map_seq _ _ h1) at 1.
  apply f_equal_prop_dep.
  omega.
Qed.


Lemma map_seq_eq_iff :
  forall {U:Type} {n w:nat} (f:seq_fun n w U) (l:list U),
    map_seq f = l <->
    (exists pfeq:w = length l,
       forall (m:nat) (pflt: m < w),
         let pflt' := lt_seq _ _ _ pflt in
         let pflt'' := lt_eq_trans _ _ _ pflt pfeq in
         f _ pflt' = 
         nth_lt l m pflt'').
  intros U n w f l.
  split; intro h1. 
  rewrite (seq_fun_eq_seg f) in h1.
  rewrite map_seq_fun_to_seg in h1.
  rewrite map_seg_eq_iff in h1.
  simpl in h1; simpl.
  destruct h1 as [h1 h2].
  exists h1.
  intros m h3.
  specialize (h2 m h3).
  unfold seq_fun_to_seg in h2.
  assumption. 
  rewrite (seq_fun_eq_seg f).
  rewrite map_seq_fun_to_seg.
  rewrite map_seg_eq_iff.
  simpl in h1; simpl.
  destruct h1 as [h1 h2].
  exists h1.
  intros m h3.
  specialize (h2 _ h3).
  unfold seq_fun_to_seg.
  assumption.
Qed.



Lemma map_seg_decompose :
  forall {U:Type} {m n:nat} (f:seg_fun n U)
         (pfm:m < n),
    map_seg f = (map_seg (seg_fun_rest f pfm))
                  ++ map_seq (seg_fun_rest_seq f pfm).
  intros U m n f h1. 
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite app_length.
  rewrite length_map_seg, length_map_seq; auto with arith.
  exists hex.
  intros m' h2 h3.
  destruct (lt_le_dec m' m) as [h4 | h4].  
  erewrite nth_lt_app1.
  unfold h3; clear h3.
  symmetry.
  gen0.
  intro h3. 
  erewrite nth_lt_functional3.
  rewrite length_map_seg in h3.
  pose proof (nth_lt_map_seg (seg_fun_rest f h1) _  h3) as  h5.
  simpl in h5.
  erewrite nth_lt_functional3 in h5.
  rewrite h5 at 1.
  unfold seg_fun_rest.
  f_equal. apply proof_irrelevance.
 
  unfold h3; clear h3.
  symmetry.
  gen0. 
  intro h3.
  pose proof h3 as h3'.
  rewrite app_length in h3'.
  rewrite length_map_seg, length_map_seq in h3'.  
  erewrite nth_lt_app2.
  gen0.  
  rewrite length_map_seg.
  intro h10.
  pose proof h10 as h10'.
  rewrite length_map_seq in h10'.
  erewrite nth_lt_functional3.
  pose (seg_fun_rest_seq f h1).
  pose proof (nth_lt_map_seq (seg_fun_rest_seq f h1) (m' - m) h10') as h11.
  simpl in h11.
  rewrite h11 at 1.
  unfold seg_fun_rest_seq.
  apply f_equal_dep.
  omega.
  Grab Existential Variables.
  rewrite length_map_seg; auto.
  rewrite length_map_seg; auto.
  rewrite length_map_seg; auto.
Qed.

Definition seg_pred_f_eq_dec
           {U n} (pfdu:type_eq_dec U)
           (f:seg_fun n U)
           i
           (pfi:i < n) : 
  seg_pred_dec (fun j (pfj: j < n) => f i pfi = f j pfj) :=
  fun j pfj => pfdu (f i pfi) (f j pfj).


Definition seg_pred_f_eq_dec_S_eq 
           {U n} 
           {f:seg_fun (S n) U}
           {i}
           {pfi':i < S n}
           (pfdu:type_eq_dec U)
           (pfi: i < n)
           (pffs:seg_pred_dec (fun j (pfj:j < S n) => f i pfi' = f j pfj)) :
  seg_pred_dec (fun j (pfj:j < n) =>
                  (seg_fun_from_S f) i pfi = 
                  (seg_fun_from_S f) j pfj).
  unfold seg_pred_dec in pffs; unfold seg_pred_dec. 
  intros j h2.
  unfold seg_fun_from_S.
  specialize (pffs j (lt_S _ _ h2)).
  destruct pffs as [h3 | h3].
  left.
  erewrite f_equal_dep.
  rewrite h3 at 1; auto.
  reflexivity.
  right.
  erewrite f_equal_dep.
  apply h3.
  reflexivity.
Defined.
  
  
Definition map_ind {T:Type} (l:list T) : list (nat * T) :=
  map_seg (fun (m:nat) (pf:m < length l) => (m, nth_lt l m pf)).


Definition list_eq_prop {T:Type} (l l':list T) (pfeq:length l = length l')
: list Prop :=
  map_seg (fun (n:nat) (pf:n < length l) =>
             nth_lt l n pf = nth_lt l' n (lt_eq_trans _ _ _ pf pfeq)).



(*This is a binary analogue of [remove].*)
Fixpoint keep_list_pr
  {T U:Type} (pfdu:type_eq_dec U)
             (l:list (T * U)) (u:U) :
      list T :=
  match l with
    | nil => nil
    | ap::l' => let lk := keep_list_pr pfdu l' u in
                if (pfdu (snd ap) u) then
                  (fst ap) :: lk
                else
                  lk
  end.





Lemma bv_tl_compat :
  forall {T:Type} (l:list T),
    l <> nil ->
    mask_to_coarse_list (eq_sym (length_bv_tl l)) = tl l.
  intros T l h1.
  induction l as [|a l ih].
  contradict h1; auto.  
  simpl.
  unfold mask_to_coarse_list.
  destruct l as [|b l]; simpl; auto.
  simpl in ih.
  f_equal.
  specialize (ih (cons_neq_nil _ _)).
  rewrite <- ih at 3.
  unfold fun_from_cons, mask_to_coarse_list; simpl.
  simpl. 
  unfold bv1s.
  simpl.
  rewrite map_in_dep_eq_iff.
  ex_goal.
  rewrite length_map_in_dep.
  rewrite bv1s_aux_S.
  rewrite length_lS; auto.
  exists hex.
  intros n h2.
  pose proof h2 as h2'.
  rewrite bv1s_aux_S in h2'.
  rewrite length_lS in h2'.
  rewrite (nth_lt_map_in_dep _ _ h2').
  pose proof (bv1s_aux_S (lrep true (length l)) 1) as h3.
  pose proof (nth_lt_functional1 _ _ n h3 h2) as h4.
  hfirst h4.
  rewrite <- bv1s_aux_S; auto.
  specialize (h4 hf).
  rewrite h4.
  unfold lS.
  rewrite (nth_lt_map _ _  _ h2').
  pose proof (in_nth_lt (bv1s_aux (lrep true (length l)) 1) n h2') as h5.
  destruct ( nth_lt (bv1s_aux (lrep true (length l)) 1) n h2'); auto.
  apply nth_indep.
  rewrite bv1s_aux_S in h5.
  rewrite in_lS_iff in h5.
  destruct h5 as [m [h5 h6]].
  apply S_inj in h6.
  rewrite h6.
  apply in_bv1s_lt in h5.
  rewrite length_lrep in h5; auto.
Qed.  


Lemma mask_to_coarse_list_cons_false :
  forall {T:Type} {l:list T} {a:T} {v:bv}
         (pfeq:length (a::l) = length (false::v)),
    mask_to_coarse_list pfeq = mask_to_coarse_list (S_inj _ _ pfeq).
  intros T l a v h1. 
  pose proof h1 as h1'.
  simpl in h1'.
  apply S_inj in h1'.
  unfold mask_to_coarse_list; simpl.
  unfold bv1s.
  simpl.
  rewrite map_in_dep_eq_iff.
  ex_goal.
  rewrite length_map_in_dep.
  rewrite bv1s_aux_S.
  rewrite length_lS; auto.
  exists hex.
  intros n h2.
  unfold fun_from_cons.
  erewrite nth_lt_map_in_dep.
  pose proof (bv1s_aux_S v 0) as h3.
  pose proof (nth_lt_functional1 _ _ _ h3 h2) as h4.
  hfirst h4.
  rewrite <- h3; auto.
  specialize (h4 hf).
  rewrite h4.
  unfold lS.
  erewrite nth_lt_map.
  symmetry.
  rewrite nth_lt_compat.
  apply nth_indep.
  pose proof (in_nth_lt (bv1s v) n) as h5.
  hfirst h5.
  pose proof h2 as h2'. rewrite bv1s_aux_S in h2'.
  rewrite length_lS in h2'; auto.
  specialize (h5 hf0).
  apply in_bv1s_lt in h5.
  rewrite h1'.
  erewrite nth_lt_functional3.
  apply h5.

  Grab Existential Variables.
  rewrite bv1s_aux_S in h2.
  rewrite length_lS in h2; auto.
Qed.


Lemma mask_to_coarse_list_cons_true :
  forall {T:Type} {l:list T} {a:T} {v:bv}
         (pfeq:length (a::l) = length (true::v)),
    mask_to_coarse_list pfeq = a :: mask_to_coarse_list (S_inj _ _ pfeq).
  intros T l a v h1.
  unfold mask_to_coarse_list; simpl.
  f_equal.
  rewrite map_in_dep_eq_iff.
  ex_goal.
  rewrite length_map_in_dep.
  rewrite bv1s_aux_S.
  rewrite length_lS.
  reflexivity. 
  exists hex.
  intros n h2. 
  unfold fun_from_cons.
  erewrite nth_lt_map_in_dep.
  pose proof (bv1s_aux_S v 0) as h3.
  pose proof (nth_lt_functional1 _ _ _ h3 h2) as h4.
  hfirst h4.
  rewrite <- h3; auto.
  specialize (h4 hf).
  rewrite h4.
  unfold lS.
  erewrite nth_lt_map.
  symmetry.
  rewrite nth_lt_compat.
  apply nth_indep.
  pose proof (in_nth_lt (bv1s v) n) as h5.
  hfirst h5.
  pose proof h2 as h2'. rewrite bv1s_aux_S in h2'.
  rewrite length_lS in h2'; auto.
  specialize (h5 hf0).
  apply in_bv1s_lt in h5.
  pose proof h1 as h1'. simpl in h1'. apply S_inj in h1'.
  rewrite h1'.
  erewrite nth_lt_functional3.
  apply h5.
    
  Grab Existential Variables.
  erewrite <- length_lS.
  rewrite <- bv1s_aux_S; auto.
Qed.


Lemma mask_to_coarse_list_compat :
  forall {T:Type} {l:list T} {v:bv}
         (pfdt:type_eq_dec T) (pfeq:length l = length v),
    coarse_list (mask_to_coarse_list pfeq) l.
  intros T l. 
  induction l as [|a l ih].
  intros v hdt h1.
  pose proof h1 as h1'.
  symmetry in h1'.
  apply length_eq0 in h1'; subst.
  simpl in h1.
  unfold mask_to_coarse_list.
  constructor.
  intros v hdt h1.
  destruct v as [|b v].
  simpl in h1; discriminate.
  pose proof h1 as h1'.
  apply S_inj in h1'.
  specialize (ih _ hdt h1').
  destruct b. 
  pose proof (mask_to_coarse_list_cons_true h1) as h2.
  rewrite h2.
  apply coarse_list_cons2.
  match goal with |- coarse_list ?l'' _ => pose l'' as l' end.
  assert (h3:l' = mask_to_coarse_list h1').
    unfold l', mask_to_coarse_list.
    apply map_in_dep_ext.
    intros i h3.
    apply functional_extensionality.
    intro h3'.
    apply nth_lt_functional3.
  subst.
  rewrite h3; auto.
  pose proof (mask_to_coarse_list_cons_false h1) as h2. 
  rewrite h2.
  match goal with |- coarse_list ?l'' _ => pose l'' as l' end.
  assert (h3:l' = mask_to_coarse_list h1').
    unfold l', mask_to_coarse_list.
    apply map_in_dep_ext.
    intros i h3.
    apply functional_extensionality.
    intro h3'.
    apply nth_lt_functional3.
  subst.
  rewrite h3; auto.
  eapply coarse_list_trans.
  assumption.
  apply ih.
  apply coarse_list_cons1.
  apply coarse_list_refl.
Qed.




Lemma coarse_list_mask_to_coarse_list :
  forall {T} {l:list T} {v:bv}
         (pfdt:type_eq_dec T)
         (pfe:length l = length v),
    let pfc := mask_to_coarse_list_compat pfdt pfe in 
    coarse_list_mask (mask_to_coarse_list pfe) l pfc v pfe.
  intros T l. 
  induction l as [|a l ih].
  intros v hdt he.
  pose proof (length_eq0 _ (eq_sym he)); subst.
  simpl.
  rewrite coarse_list_mask_iff'.
  constructor.
  red. 
  intros; contradiction. 
  intro v.
  destruct v as [|b v].
  intros hdt he hc.
  simpl in he.
  discriminate.
  intros hdt he hc.
  pose proof he as he'.
  simpl in he'. 
  specialize (ih v hdt (S_inj _ _ he')).
  simpl in ih.
  term2 ih.
  destruct b.
  pose proof (coarse_list_mask_cons2 _ _ a c _ (S_inj _ _ he') ih) as h4.
  eapply coarse_list_mask_functional in h4.
  apply h4.
  rewrite mask_to_coarse_list_cons_true.
  repeat f_equal; apply proof_irrelevance.
  reflexivity.
  reflexivity.
  pose proof (coarse_list_mask_cons1 _ _ a c _ (S_inj _ _ he') ih) as h4.
  eapply coarse_list_mask_functional in h4.
  apply h4.
  rewrite mask_to_coarse_list_cons_false.
  repeat f_equal; apply proof_irrelevance.
  reflexivity.
  reflexivity.
Qed.



(*Finds the index of the [S occ]'s occureence of [x] in [l]. 
if [occ >= pred (count_occ)] then the value is the same as 
[pred(count_occ)]'s *)
Fixpoint lind_occ
         {T:Type} (pfdt:type_eq_dec T) (l:list T) (occ:nat) (x:T) : 
  In x l -> nat :=
  match l with
    | nil => fun pf => False_rect nat pf
    | a::l' => fun pf =>
                 match (pfdt x a) with
                   | left _ => 
                     match occ with
                       | 0 => 0
                       | S n =>
                         match in_dec pfdt x l' with
                           | left pfs => S (lind_occ pfdt l' n x pfs)
                           | right _ => O
                         end
                     end
                   | right pfn =>
                     let pf' := neq_in_cons _ _ _ pfn pf in
                       S (lind_occ pfdt l' occ x pf')
                 end
  end.



(*Minimum list index of a given element in a list.*)
Definition lind {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l):=
  lind_occ pfdt l 0 x pf.


Lemma lind_functional :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x y:T)
         (pfx:In x l) (pfy:In y l),
    x = y ->
    lind pfdt l x pfx = lind pfdt l y pfy.
  intros T h0 l x y h2 h3 h4.
  subst.
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
Qed.

Lemma lind_functional' : 
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x y:T)
         (pfx:In x l) (pfy:In y l'),
    l = l'  ->
    x = y ->
    lind pfdt l x pfx = lind pfdt l' y pfy.
  intros T h0 l l' x y h2 h3 h4 h5. subst.
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
Qed.

Lemma lind_functional3 : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfx pfx':In x l),
    lind pfdt l x pfx = lind pfdt l x pfx'.
  intros; f_equal; apply proof_irrelevance.
Qed.



Lemma lind_occ_functional :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (occ occ':nat)
         (x x':T) (pf:In x l) (pf':In x' l'),
    l = l' ->
    occ = occ' ->
    x = x' ->
    lind_occ pfdt l  occ x pf =
    lind_occ pfdt l' occ' x' pf'.
  intros T h1 l l' occ occ' x x' h2 h3 h4 h5 h6.
  subst.
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
Qed.


Lemma lind_occ_functional3 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (occ:nat)
         (x:T) (pf pf':In x l),
    lind_occ pfdt l occ x pf =
    lind_occ pfdt l occ x pf'.
  intros; f_equal; apply proof_irrelevance.
Qed.


Lemma lt_lind_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (occ:nat) (x:T)
         (pf:In x l),
    lind_occ pfdt l occ x pf < length l.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros occ x h3.
  destruct (h1 x a) as [h4 | h4]; subst.
  destruct occ as [|occ]; simpl; auto with arith.
  destruct (in_dec h1 a l) as [h4 | h4].
  apply lt_n_S.
  apply ih.
  auto with arith.
  apply lt_n_S.
  apply ih.
Qed.


Lemma lind_occ_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (occ:nat) (x:T)
         (pf:In x l),
    nth_lt l (lind_occ pfdt l occ x pf) (lt_lind_occ _ _ _ _ _) = x.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros occ x h2.
  destruct (h1 x a) as [h3 | h3]; simpl; subst.
  destruct occ as [|occ]; simpl; auto.
  destruct (in_dec h1 a l) as [h3 | h3]; simpl.
  specialize (ih occ a h3).
  simpl in ih.
  rewrite <- ih.
  erewrite nth_lt_functional3.
  rewrite nth_lt_compat at 1.
  apply nth_indep.
  apply lt_lind_occ.
  reflexivity.
  specialize (ih occ x (neq_in_cons _ _ _ h3 h2)).
  simpl in ih.
  rewrite <- ih.
  erewrite nth_lt_functional3.
  rewrite nth_lt_compat at 1.
  apply nth_indep.
  apply lt_lind_occ.
  Grab Existential Variables.
  apply lt_lind_occ.
  apply lt_lind_occ.
Qed.


Lemma ex_lind_occ_nth_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (n:nat)
         (pf:n < length l),
  exists (occ:nat),
    occ < count_occ pfdt l (nth_lt l n pf) /\
    lind_occ pfdt l occ (nth_lt l n pf) (in_nth_lt _ _ _) = n.
  intros T h1 l.
  induction l as [|a l ih].
  intros ? h2; simpl in h2; omega.
  intros n h2.
  simpl in h2.
  destruct n as [|n].
  exists 0.
  simpl.
  destruct (h1 a a) as [h3 | h3].
  split; auto with arith.
  contradict h3; auto.
  pose proof (lt_S_n _ _ h2) as h2'.
  specialize (ih n h2').
  destruct ih as [occ [h3 h4]].
  simpl.
  rewrite switch_eq_dec_eq.
  destruct (h1 (nth n l a) a) as [h5 | h5].
  unfold switch_eq_dec.
  destruct in_dec as [h6 | h6].
  exists (S occ).
  split.
  apply lt_n_S.
  rewrite nth_lt_compat in h3.
  erewrite nth_indep.
  apply h3.
  assumption.
  f_equal.
  rewrite <- h4.
  apply lind_occ_functional; auto.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  contradict h6.
  apply nth_In; auto.
  unfold switch_eq_dec.
  exists occ.
  split; auto.
  rewrite nth_lt_compat in h3.
  erewrite nth_indep in h3.
  apply h3.
  assumption.
  f_equal.
  rewrite <- h4.
  apply lind_occ_functional; auto.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
Qed.


Lemma lt_lind :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfin:In x l),
    lind pfdt l x pfin < length l.
  unfold lind; intros; apply lt_lind_occ.
Qed.
  


Lemma lind_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T) (pf:In a l),
    nth_lt l (lind pfdt l a pf) (lt_lind pfdt l a pf) =
    a.
  unfold lind; intros; erewrite nth_lt_functional3; rewrite lind_occ_compat at 1; auto.
Qed.

(*See [lind_cons_same] for [occ = 0]*)
Lemma lind_occ_cons_in_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         occ
         (pf:In x l)
         (pf':In x (x::l)),
     0 < occ ->
     lind_occ pfdt (x::l) occ x pf' =
     S (lind_occ pfdt l (pred occ) x pf).
  intros T h1 l x occ h2 h3 h4.
  simpl.
  rewrite eq_dec_eq at 1; auto.
  rewrite (S_pred _ _ h4).
  lr_match_goal'.
  f_equal. 
  apply lind_occ_functional; auto.
  contradiction.
Qed.


Lemma lind_occ_cons_in_neq :
   forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x:T)
          (pf:In x l) (occ:nat),
     x <> a ->
     lind_occ pfdt (a::l) occ x (in_cons _ _ _ pf) =
     S (lind_occ pfdt l occ x pf).
  intros T h1 l a x h2 occ h3.
  simpl.
  destruct (h1 x a); subst.
  contradict h3; auto.
  do 2 f_equal.
  apply proof_irrelevance.
Qed.




Lemma nth_lt_lind_eq : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfin:In x l),
    let pf := lt_lind pfdt l x pfin in
    nth_lt l (lind pfdt l x pfin) pf  = x.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h3.
  unfold lind. simpl.
  destruct (h1 x a) as [h4 | h4]; auto.
  simpl in ih.
  pose proof (neq_in_cons _ _ _ h4 h3) as h5.
  specialize (ih _ h5).
  rewrite <- ih.
  pose proof (nth_lt_compat l (lind h1 l x h5) (lt_lind h1 l x h5)) as h6.
  simpl in h6.
  rewrite h6.
  match goal with |- nth ?x _ _ = nth ?y _ _ => assert (h7:x = y) end.
  unfold lind. f_equal.
  apply proof_irrelevance.
  rewrite h7.
  apply nth_indep.
  apply lt_lind.
Qed.


Lemma lind_occ_inj :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (x y:T)
         (pfx:In x l) (pfy:In y l)
         (m m':nat),
    lind_occ pfdt l m x pfx = lind_occ pfdt l m' y pfy ->
    x = y.
  intros T h0 l x y h2 h3 m m' h4.
  pose proof (nth_lt_functional2 _ _ _ h4 (lt_lind_occ _ _ _ _ _) (lt_lind_occ _ _ _ _ _)) as h5.
  do 2 rewrite lind_occ_compat in h5; auto.
Qed.

Lemma lind_occ_inj' :
  forall {T:Type} (pfdt:type_eq_dec T) 
         (l:list T) 
         (x:T) 
         (pfx pfx':In x l) 
         (m m':nat),
    m < count_occ pfdt l x ->
    m' < count_occ pfdt l x ->
    lind_occ pfdt l m x pfx = lind_occ pfdt l m' x pfx' ->
    m = m'.
  intros T hdt l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros x hx hx' m m'. 
  destruct (hdt a x) as [|h7]; subst.
  rewrite eq_dec_same at 1.
  rewrite eq_dec_same at 1.
  destruct m as [|m], m' as [|m']; auto.
  intros h1 h2 h3.
  apply lt_S_n in h2.
  pose proof (count_occ_in hdt _ _ _ h2) as h4.
  destruct in_dec as [h5 | h5].
  discriminate.
  contradiction.
  intro h1.
  apply lt_S_n in h1.
  pose proof (count_occ_in hdt _ _ _ h1) as h4.
  destruct in_dec as [h5 | h5].
  discriminate.
  contradiction.
  intros h1 h2.
  apply lt_S_n in h1.
  apply lt_S_n in h2.
  pose proof (count_occ_in hdt _ _ _ h1) as h4.
  pose proof (count_occ_in hdt _ _ _ h2) as h5.
  destruct in_dec as [h6 | h6].
  intro h7.
  apply S_inj in h7.
  f_equal; auto.
  eapply ih.
  apply h1.
  assumption.
  apply h7.
  contradiction.
  assumption.
  assumption.
  intros h8 h9.
  destruct hx as [|h10]; subst.
  contradiction (h7 eq_refl).
  destruct (hdt x a) as [h11 | h11]; subst.
  contradict h7; auto.
  intro h12.
  apply S_inj in h12.
  eapply ih.
  apply h8.
  assumption.
  apply h12.
Qed.
  

Lemma lind_inj_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (x y:T)
         (pfx:In x l) (pfy:In y l),
       lind pfdt l x pfx = lind pfdt l y pfy <->
       x = y.
  intros T h0 l x y h2 h3; split; intro h4.
  pose proof (nth_lt_lind_eq h0 l x h2) as h5.
  pose proof (nth_lt_lind_eq h0 l y h3) as h6.
  simpl in h5, h6.
  rewrite <- h5, <- h6.
  gen0.
  rewrite h4.
  intro; f_equal.
  apply proof_irrelevance.
  subst.
  apply lind_functional; auto.
Qed.


Lemma inj_seg_nth_lt' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    inj_seg (nth_lt l) ->
    NoDup l.
  intros T hdt l.
  induction l as [|a l ih]; simpl.
  intros; constructor.
  intro h1.
  constructor.
  intro h2.
  red in h1.
  specialize (h1 0 (S (lind hdt _ _ h2)) (lt_0_Sn _)).
  hfirst h1.
  apply lt_n_S.
  apply lt_lind.
  specialize (h1 hf).
  simpl in h1.
  hfirst h1.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  apply h1 in hf0.
  discriminate.
  apply ih.
  red in h1; red.
  intros i j h2 h3 h4.
  specialize (h1 (S i) (S j)).
  hfirst h1.
  apply lt_n_S; auto.
  specialize (h1 hf).
  hfirst h1.
  apply lt_n_S; auto.
  specialize (h1 hf0).
  simpl in h1.
  hfirst h1.
  erewrite nth_indep.
  symmetry.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  rewrite <- h4 at 1.
  apply nth_lt_compat.
  assumption.
  assumption.
  apply h1 in hf1.
  apply S_inj in hf1; auto.
Qed.


Lemma lind_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (a x:T) (pf:In x (a::nil)),
    lind pfdt (a::nil) x pf = 0.
  intros T hd a x h0.
  destruct h0 as [h0 | h0]. subst.
  unfold lind. simpl.
  destruct (hd x x) as [hx | hx]. auto.
  contradict hx; auto.
  simpl in h0. contradiction.
Qed.


  
Lemma lind_firstn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (m:nat)
         (pf:In x (firstn m l)),
      let pf' := in_firstn _ _ _ pf in
      lind pfdt (firstn m l) x pf =
      lind pfdt l x pf'.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros ? ? h2. pose proof h2 as h2'. rewrite firstn_nil in h2'.
  contradiction.
  intros x m h2.
  unfold lind; simpl.
  destruct (h1 x a) as [h4 | h4]; subst.
  destruct (zerop m) as [h4 | h4]; subst.
  simpl in h2. contradiction.
  gen0. rewrite (firstn_cons l m a h4).
  intro h5. simpl.
  destruct (h1 a a) as [h6 | h6]; auto.
  contradict h6; auto.
  simpl in ih.
  destruct (zerop m) as [h5 | h5]; subst.
  simpl in h2.
  contradiction.
  gen0.
  rewrite (S_pred _ _ h5).
  intro h0.
  simpl.
  simpl in h0.
  destruct h0 as [|h0]; subst.
  contradict h4; auto.
  destruct (h1 x a) as [|ha]; subst.
  contradict h4; auto.
  f_equal.
  gen0.
  intro h3.
  specialize (ih _ _ h3).
  unfold lind in ih.
  rewrite ih.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma lind_cons_same :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T)
         (pf:In a (a::l)),
    lind pfdt (a::l) a pf = 0.
  intros T h1 l a.
  unfold lind.
  simpl.
  destruct (h1 a a) as [h2 | h2]; auto.
  contradict h2.
  reflexivity.
Qed.



Lemma lind_hd' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (pf:l <> nil),
    lind pfdt l (hd' _ pf) (in_hd' _ _) = 0.
  intros T h1 l.
  destruct l as [|a l].
  intro h2; contradict h2; auto.
  intro h2.
  simpl.
  erewrite lind_functional. rewrite lind_cons_same; auto.
  reflexivity.
  Grab Existential Variables.
  left; auto.
Qed.

Lemma lind_occ_hd :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (occ:nat) (pfin:In x (x::l))
         (pfo:occ < count_occ pfdt l x),
    let pfin' := count_occ_in _ _ _ _ pfo   in
    lind_occ pfdt (x :: l) (S occ) x pfin =
    S (lind_occ pfdt l occ x pfin').
  intros T h1 l x occ h2 h3 h4; simpl.
  destruct (h1 x x) as [h6|h6].
  destruct in_dec as [h7|h7].
  repeat f_equal; try apply proof_irrelevance.
  contradiction.
  contradict h6; auto.
Qed.


Lemma lind_occ_hd' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (pf:l <> nil) (occ:nat) 
         (pfo:S occ < count_occ pfdt l (hd' _ pf)),
    let pfin' := impl_in_tl pfdt _ _ _ pfo in
    lind_occ pfdt l (S occ) (hd' l pf) (in_hd' _ _) =
    S (lind_occ pfdt (tl l) occ _ pfin').
  intros T h1 l h2 occ h3 h4.
  destruct l as [|a l].
  pose proof (h2 eq_refl); contradiction.
  simpl.
  rewrite (eq_dec_same h1) at 1.
  destruct in_dec as [h5 | h5].
  repeat f_equal.
  apply proof_irrelevance.
  simpl in h3.
  pose proof h3 as h3'. rewrite (eq_dec_same h1) in h3' at 1.
  simpl in h3'.
  apply lt_S_n in h3'.
  contradict h5.
  apply count_occ_in in h3'; auto.
Qed.


Lemma lind_eq0_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) 
         (pfin:In x l),
    lind pfdt l x pfin = 0 <->
    hd' l (in_not_nil _ _ pfin) = x.
  intros T h1 l x h2; split; intro h3.
  destruct l as [|a l]; simpl.
  contradiction.
  unfold lind in h3.
  simpl in h3.
  destruct (h1 x a) as [h4 | h4]; auto.
  discriminate.
  generalize h2.
  rewrite <- h3.
  intro h2'.
  erewrite lind_functional.
  apply lind_hd'.
  reflexivity.
Qed.


Lemma lind_occ_eq0_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (occ:nat)
         (pfin:In x l),
    occ < count_occ pfdt l x ->
    (lind_occ pfdt l occ x pfin = 0 <->
     (occ = 0 /\ hd' l (in_not_nil _ _ pfin) = x)).
  intros T h1 l.
  destruct l as [|a l]; simpl.
  intros; contradiction. 
  intros x occ h2 hlt.
  destruct h2 as [|h2]; subst.
  destruct occ as [|occ].
  destruct (h1 x x) as [h2 | h2].
  tauto.
  contradict h2; auto.
  destruct (h1 x x) as [h2 | h2].
  apply lt_S_n in hlt.
  destruct (in_dec h1 x l) as [h3 | h3].
  intuition. 
  contradict h3.
  eapply count_occ_in. apply hlt.
  contradict h2; auto.
  rewrite switch_eq_dec_eq in hlt.
  destruct (h1 x a) as [|h3]; subst.
  unfold switch_eq_dec in hlt.
  destruct occ as [|occ].
  tauto.
  apply lt_S_n in hlt.
  destruct in_dec as [h3 | h3].
  intuition.
  contradict h3.
  eapply count_occ_in.
  apply hlt.
  unfold switch_eq_dec in hlt.
  intuition.
Qed.

Lemma lind_occ_eq0_iff' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (occ:nat)
         (pfin:In x l),
    (lind_occ pfdt l occ x pfin = 0 <->
     ((occ = 0 \/ count_occ pfdt l x = 1) /\
      hd' l (in_not_nil _ _ pfin) = x)).
  intros T h1 l.
  destruct l as [|a l]; simpl.
  intros; contradiction. 
  intros x occ h2.
  destruct h2 as [|h2]; subst.
  destruct occ as [|occ].
  destruct (h1 x x) as [h2 | h2].
  tauto.
  contradict h2; auto.
  destruct (h1 x x) as [h2 | h2].
  destruct (in_dec h1 x l) as [h3 | h3].
  intuition.
  apply S_inj in H.
  pose proof (count_occ_In h1 l x) as h4.
  pose proof h3 as h3'. rewrite h4 in h3'.
  omega.
  rewrite (count_occ_In h1 l x) in h3.
  intuition.
  contradict h2; auto.
  rewrite switch_eq_dec_eq.
  destruct (h1 a x) as [|h3]; subst.
  unfold switch_eq_dec.
  destruct occ as [|occ].
  tauto.
  destruct in_dec as [h3 | h3].
  intuition.
  apply S_inj in H.
  rewrite (count_occ_In h1 l x) in h2.
  omega.
  contradiction.
  unfold switch_eq_dec.
  intuition.
Qed.




Lemma lind_cons_neq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a x:T) 
         (pfix:In x l) (pfn:x <> a),
    lind pfdt (a::l) x (or_intror pfix) = 
    S (lind pfdt l x pfix).
  intros T h1 l a x h3.
  unfold lind.
  simpl.
  destruct (h1 x a) as [|h4]; subst.
  intro h4; contradict h4; auto.
  intro h5.
  f_equal.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma nth_lt_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat) (pflt:n < length l) (x:T),
    nth_lt l n pflt = x <->
    exists (pfin:In x l) (occ:nat),
      occ < count_occ pfdt l x /\
      lind_occ pfdt l occ x pfin = n.
  intros T h1 l n h2 ?; split; intro ha.
  subst.
  exists (in_nth_lt _ _ _).
  revert h2. revert n.
  induction l as [|a l ih]; simpl. 
  intros; omega. 
  intros n h2.
  destruct n as [|n].
  destruct (h1 a a) as [|h3].
  exists 0.
  split; auto with arith.
  contradict h3; auto.
  pose proof h2 as h2'. apply lt_S_n in h2'.
  specialize (ih n h2').
  destruct ih as [occ ih].
  destruct ih as [h3 h4].
  rewrite (switch_eq_dec_eq h1 a (nth n l a)).
  destruct (h1 (nth n l a) a) as [h5 | h5].
  unfold switch_eq_dec.
  destruct in_dec as [h6 | h6].
  exists (S occ).
  rewrite nth_lt_compat in h3.
  split.
  apply lt_n_S.
  erewrite nth_indep.
  apply h3.
  apply lt_S_n in h2; auto.
  f_equal.
  generalize h6. erewrite nth_indep.
  rewrite <- nth_lt_compat.
  intro h6'.
  erewrite lind_occ_functional.
  apply h4.
  reflexivity.
  reflexivity.
  reflexivity.
  apply lt_S_n in h2; auto.
  contradict h6.
  apply nth_In.
  apply lt_S_n in h2; auto.
  unfold switch_eq_dec.
  exists occ. split.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  apply h3.
  apply lt_S_n in h2; auto.
  f_equal.
  erewrite lind_occ_functional.
  apply h4.
  reflexivity.
  reflexivity.
  rewrite nth_lt_compat.
  apply nth_indep.
  apply lt_S_n in h2; auto.

  destruct ha as [h3 [occ [h4 h5]]].
  subst.
  erewrite nth_lt_functional3. rewrite lind_occ_compat; auto.
Qed.



Lemma nth_lts_eq_length_list_singularize :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U) (pfe:length l = length l'),
    (forall (m n : nat) (pfm : m < length l) (pfn : n < length l),
       let pfm' := lt_eq_trans m (length l) (length l') pfm pfe in
       let pfn' := lt_eq_trans n (length l) (length l') pfn pfe in
       nth_lt l m pfm = nth_lt l n pfn <->
       nth_lt l' m pfm' = nth_lt l' n pfn') ->
    length (list_singularize pfdt l) =
    length (list_singularize pfdu l').
  intros T U hdt hdu l.
  induction l as [|a l ih].
  intros l' h1.
  pose proof h1 as h1'.
  symmetry in h1'.
  apply length_eq0 in h1'; subst; simpl; auto.
  intro  l'.
  destruct l' as [|a' l'].
  simpl; intros; discriminate.
  intros h1 h2.
  pose proof h1 as h1'.
  simpl in h1'.
  apply S_inj in h1'.
  destruct (in_dec hdt a l) as [h3 | h3],
                               (in_dec hdu a' l') as [h4 | h4].
  rewrite list_singularize_in_cons; auto.
  rewrite list_singularize_in_cons; auto.
  eapply ih.
  intros m n h5 h6 h7 h8.
  specialize (h2 (S m) (S n)).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf0).
  simpl in h2. 
  erewrite nth_lt_compat.
  erewrite (nth_lt_compat l n).
  erewrite (nth_lt_compat l' m).
  erewrite (nth_lt_compat l' n). 
  erewrite nth_indep; auto.
  erewrite (nth_indep l _ _  h6); auto. 
  erewrite (nth_indep l'); auto.
  erewrite (nth_indep l' _ _  h8); auto. 
  apply h2.

  specialize (h2 (S (lind hdt _ _ h3)) 0).
  hfirst h2.
  simpl. 
  eapply lt_n_S.
  apply lt_lind.
  specialize (h2 hf).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf0).
  simpl in h2.
  erewrite (nth_indep l) in h2.
  erewrite <- nth_lt_compat in h2.
  rewrite lind_compat in h2.
  pose proof (iff1 h2 eq_refl) as h5.
  pose proof (@nth_In _ (lind hdt l a h3) l' a') as h6.
  hfirst h6.
  eapply lt_eq_trans.
  apply lt_lind.
  auto.
  specialize (h6 hf1).
  rewrite h5 in h6.
  contradiction.
  apply lt_lind.

  specialize (h2 (S (lind hdu _ _ h4)) 0).
  hfirst h2.
  simpl. 
  eapply lt_n_S.
  rewrite h1'.
  apply lt_lind.
  specialize (h2 hf).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf0).
  simpl in h2.
  erewrite (nth_indep l') in h2.
  erewrite <- nth_lt_compat in h2.
  rewrite lind_compat in h2.
  pose proof (iff2 h2 eq_refl) as h5.
  pose proof (@nth_In _ (lind hdu l' a' h4) l a) as h6.
  hfirst h6.
  eapply lt_eq_trans.
  apply lt_lind.
  auto.
  specialize (h6 hf1).
  rewrite h5 in h6.
  contradiction.
  apply lt_lind.

  rewrite list_singularize_nin; auto.
  rewrite list_singularize_nin; auto.
  simpl; f_equal.
  eapply ih.
  intros m n h5 h6 h7 h8.
  specialize (h2 (S m) (S n)).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf0).
  simpl in h2.
  
  erewrite nth_lt_compat.
  erewrite (nth_lt_compat l n).
  erewrite (nth_lt_compat l' m).
  erewrite (nth_lt_compat l' n). 
  erewrite nth_indep; auto.
  erewrite (nth_indep l _ _  h6); auto. 
  erewrite (nth_indep l'); auto.
  erewrite (nth_indep l' _ _  h8); auto. 
  apply h2.

  Grab Existential Variables.
  assumption.
  assumption.
Qed.




Lemma leq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    l = l' <->
    exists (pfe:length l = length l'),
      forall (i:nat) (pfi:i < length l),
        let pfi' := lt_eq_trans _ _ _ pfi pfe in
        nth_lt l i pfi = 
        nth_lt l' i pfi'.
  intros T h0 l l'; split; intro h1.
  subst.
  exists eq_refl.
  intros i h1 h2.
  apply nth_lt_functional3.
  destruct h1 as [h1 h2].
  generalize dependent l'.
  induction l as [|a l ih]; simpl.
  intros l' h1.
  pose proof h1 as h1'.
  symmetry in h1'.
  apply length_eq0 in h1'; subst; auto.
  intros l' h1.
  destruct l' as [|a' l'].
  simpl in h1.
  discriminate.
  simpl in h1.
  pose proof h1 as h1'.
  apply S_inj in h1'.
  intro h2.
  destruct (h0 a a') as [|h3]; subst.
  f_equal.
  eapply ih.
  intros i h3.
  specialize (h2 (S i)).
  hfirst h2.
  apply lt_n_S; auto.
  specialize (h2 hf).
  simpl in h2.
  simpl.
  rewrite nth_lt_compat.
  symmetry.
  rewrite nth_lt_compat.
  symmetry.
  erewrite nth_indep.
  rewrite h2 at 1.
  apply nth_indep.
  rewrite <- h1'; auto.
  assumption.
  specialize (ih _ h1'); simpl in ih.
  specialize (h2 0 (lt_0_Sn _)).
  simpl in h2.
  contradiction.

  Grab Existential Variables.
  assumption.
Qed.


Lemma app_cons_eq_sing_iff :
  forall {T} (l l':list T) a,
    l ++ a :: l' = a :: nil <->
    l = nil /\ l' = nil.
  intros T l l' a; split; intro h1; destruct l, l'; auto; simpl in h1; inversion h1; subst; try discriminate.
  apply app_not_nil_l in H1; contradict H1.
  subst.
  rewrite app_nil_l.
  intro; discriminate.
  apply app_not_nil_l in H1; contradict H1; subst.
  simpl.
  intro; discriminate.
Qed.

Lemma app_cons_nil_eq_iff : 
  forall {T} {l l':list T} a a',
    ~In a l ->
    ~In a l' -> 
     (l ++ a :: nil = l' ++ a' :: nil <-> 
      (a = a' /\ l = l')).
  intros T l.
  induction l as [|b l ih]; subst.
  intro l'; destruct l' as [|b l']; simpl; intros; split; intro h1.
  inversion h1; auto.
  destruct h1; subst; auto.
  symmetry in h1.
  inversion h1; subst.
  contradict H0; left; auto.
  destruct h1; discriminate.
  intro l'.
  destruct l' as [|b' l'].
  intros; simpl; split; intro h1.
  inversion h1; subst. 
  apply app_not_nil_r in H3; contradict H3.
  discriminate.
  destruct h1; discriminate.
  intros a c h1 h2; split; intro h3.
  inversion h3; subst.
  inversion h3.
  rewrite ih in H0.
  destruct H0; subst; auto.
  contradict h1; right; auto.
  contradict h2; right; auto.
  destruct h3 as [? h3]; subst.
  inversion h3; subst; auto.
Qed.


Lemma app_cons_nil_eq_iff' :
  forall {T} {l l0 l':list T} a,
    ~In a l ->
    ~In a l' ->
    ~In a l0 -> 
     (l ++ a :: nil = l' ++ a :: l0 <-> 
      l0 = nil /\ l = l').
  intros T l l0 l' a h2 h3; split; intro h4.
  destruct l0; simpl in h4.
  rewrite app_cons_nil_eq_iff in h4; tauto.
  rewrite app_eq_iff in h4.
  destruct h4 as [h4 h5].
  symmetry in h5.
  rewrite sing_eq_iff in h5.
  destruct h5 as [h5 [h6 h7]].
  destruct (nil_dec' l') as [h8|h8], (nil_dec' l) as [h9|h9];
    revert h7 h6 h4; subst; simpl.
  intro; discriminate.
  intros h6 h7 h8.
  pose proof (nin_eq_firstn_length _ (t::l0) a h2 h8); contradiction.
  intros h9 h10.
  erewrite hd_app' in h10.
  rewrite <- h10 in h3.
  contradict h3.
  apply in_hd'.
  intros h10 h11. 
  pose proof (tl_sing _ h5 h10) as h12.
  rewrite h11 in h12.
  intro h13.
  pose proof (firstn_skipn (length l) (l'++a::t::l0)) as h14.
  rewrite <- h13 in h14.
  rewrite h12 in h14.
  apply app_sing_eq_app_cons_cons in h14.
  contradiction.
  assumption.
  destruct h4; subst; auto.

  Grab Existential Variables.
  assumption.
Qed.

Lemma app_cons_eq_length :
  forall {T} (l l0 l' l0':list T) (a:T),
    ~In a l ->
    ~In a l0 ->
    ~In a l' ->
    ~In a l0' ->
    l ++ a :: l0 = l' ++ a :: l0' ->
    length l = length l'.
  intros T l.
  induction l as [|b l ih]; simpl.
  intros l0 l'.
  destruct l' as [|b' l']; simpl; auto.
  intros l0' a h1 h2 h3 h4 h5.
  inversion h5; subst.
  contradict h3; left; auto.
  intros l0 l'; destruct l' as [|b' l'].
  intros l' a h1 h2 h3 h4 h5.
  simpl in h5; inversion h5; subst; contradict h1; left; auto.
  intros l0' a h1 h2 h3 h4 h5.
  simpl in h5.
  inversion h5; subst.
  inversion h5.
  simpl; f_equal.
  eapply ih.
  contradict h1.
  right; apply h1.
  apply h2.
  contradict h3; right; auto.
  apply h4.
  assumption.
Qed.


Lemma app_cons_eq_iff :
  forall {T} (l l0 l' l0':list T) (a:T),
    ~In a l ->
    ~In a l0 ->
    ~In a l' ->
    ~In a l0' ->
    (l ++ a :: l0 = l' ++ a :: l0' <->
     l = l' /\ l0 = l0').  
  intros T l l0 l' l0' a h1 h2 h3 h4; split; intro h5.
  pose proof (f_equal (firstn (length l)) h5) as h6. 
  rewrite firstn_app_le, firstn_length in h6.
  pose proof (f_equal (@length _) h5) as h7.
  do 2 rewrite app_length in h7; simpl in h7.
  do 2 rewrite plus_S_shift_r in h7.
  apply S_inj in h7.
  pose proof (f_equal (firstn (length l')) (eq_sym h5)) as h6'.
  rewrite firstn_app_le, firstn_length in h6'; auto.
  rewrite h5 in h6'.
  erewrite app_cons_eq_length in h6.
  split.
  rewrite h6; rewrite h6' at 2.
  reflexivity.
  pose proof (f_equal (skipn (length l)) h5) as h16.
  rewrite skipn_length_app in h16.
  erewrite app_cons_eq_length in h16.
  rewrite skipn_length_app in h16.
  clear h6 h6'.
  inversion h16; subst; auto.
  apply h1.
  apply h2.
  apply h3.
  apply h4.
  assumption.
  apply h1.
  apply h2.
  assumption.
  apply h4.
  assumption.
  apply le_refl.
  destruct h5; subst; auto.
Qed.


  
Lemma le_lind_nth_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (n:nat)
         (pflt:n < length l),
    lind pfdt l (nth_lt l n pflt) (in_nth_lt _ _ _) <= n.
  intros T h1 l n. revert l.
  induction n as [|n ih]; simpl.
  intros l h2.
  pose proof (nth_lt0 l (lt_length _ _ h2)) as h3.
  pose proof (proof_irrelevance
                _ h2
                (not_nil_lt _ (lt_length l 0 h2))) as h4.
  simpl in h3.
  rewrite <- h4 in h3.
  generalize (in_nth_lt l 0 h2).
  rewrite h3.
  intro h5.
  erewrite lind_functional. rewrite lind_hd'; auto with arith.
  reflexivity.
  intros l h2.
  pose proof h2 as h2'.
  rewrite (S_pred _ _ h2') in h2'.
  apply lt_S_n in h2'.
  specialize (ih (tl l)).
  hfirst ih.
  rewrite length_tl; auto.
  specialize (ih hf).
  pose proof (lt_length _ _ h2) as h3.
  pose proof (hd_compat' _ h3) as h4.
  simpl in h4.
  pose proof (nth_lt_functional1 _ _ (S n) h4 h2) as h5.
  simpl in h5.
  hfirst h5.
  apply lt_n_S. rewrite length_tl; auto.
  specialize (h5 hf0).
  pose proof (@lind_functional).
  erewrite nth_indep in h5. rewrite <- (nth_lt_compat _ _ hf) in h5 at 1.
  pose proof (in_nth_lt (tl l) n hf) as h6.
  pose proof (in_cons (hd' l h3) _ _ h6) as h7.
  pose proof (lind_functional' h1 _ _ _ _   (in_nth_lt l (S n) h2) h7 h4 h5) as h8.
  rewrite h8.
  simpl.
  destruct (h1 (hd' l h3) (nth_lt (tl l) n hf)) as [h9 | h9].
  generalize h7.
  rewrite <- h9.
  intro h7'.
  erewrite lind_functional. rewrite lind_cons_same; auto with arith.
  reflexivity.
  apply neq_sym in h9.
  pose proof (lind_cons_neq h1 _ _ _ h6 h9) as h10.
  pose proof (proof_irrelevance _ h7
                                (or_intror h6)) as h11.
  rewrite <- h11 in h10.
  rewrite h10.
  apply le_n_S.
  erewrite lind_functional.
  apply ih.
  reflexivity.
  assumption.

  Grab Existential Variables.
  left; auto.
Qed.


Lemma lind_app_in1 :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T)
           (pf:In a l),
    let pf' := in_app_l pf l' in
    lind pfdt (l++l') a pf' =
    lind pfdt l a pf.
  intros T h1 l.
  induction l  as [|b l ih]; simpl.
  intros; contradiction.
  intros l' a h2.
  destruct h2 as [|h2]; subst.
  erewrite lind_functional.
  rewrite lind_cons_same; auto.
  erewrite lind_functional.
  rewrite lind_cons_same; auto.
  reflexivity.
  reflexivity.
  specialize (ih l' a h2).
  simpl in ih.
  destruct (h1 a b) as [|h3]; subst.
  erewrite lind_functional. rewrite lind_cons_same.
  erewrite lind_functional. rewrite lind_cons_same.
  reflexivity. reflexivity. reflexivity.
  pose proof (lind_cons_neq h1 (l++l') b a) as h4.
  hfirst h4.
  apply in_app_l; auto.
  specialize (h4 hf h3).
  rewrite lind_cons_neq; auto.
  erewrite lind_functional.
  rewrite h4 at 1.
  f_equal.
  erewrite lind_functional.
  rewrite ih at 1; auto.
  reflexivity.
  reflexivity.

  Grab Existential Variables.
  left; auto.
  left; auto.
  left; auto.
  left; auto.
Qed.
 
Lemma lind_app_in2 :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a':T)
         (pfin':In a' l') (pfnin: ~In a' l),
    let pfin := in_app_r pfin' l in
    lind pfdt (l++l') a' pfin =
    length l + lind pfdt l' a' pfin'.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; f_equal; apply proof_irrelevance.
  intros l' a' h2 h3.
  unfold lind.
  simpl.
  destruct (h1 a' a) as [|h4]; subst.
  contradict h3; left; auto.
  f_equal.
  specialize (ih l' a' h2).
  hfirst ih.
  contradict h3; right; auto.
  specialize (ih hf).
  simpl in ih.
  unfold lind in ih.
  rewrite <- ih.
  f_equal.
  apply proof_irrelevance.
Qed.

   
Lemma lind_occ_app_lt_in1 :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T) 
         (occ:nat) (pflt:occ < count_occ pfdt l x),
    let pfin := count_occ_in pfdt _ _ _ pflt in
    let pfin' := in_app_l pfin l' in
    lind_occ pfdt (l++l') occ x pfin' =
    lind_occ pfdt l occ x pfin.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros l' x occ h3.
  pose proof h3 as h3'.
  rewrite switch_eq_dec_eq in h3'.
  destruct (h1 x a) as [|h4]; subst.
  unfold switch_eq_dec in h3'.
  destruct occ as [|occ] ;auto.
  apply lt_S_n in h3'.
  destruct (in_dec h1 a l) as [h4 | h4].
  destruct in_dec as [h5 | h5]. 
  f_equal.
  erewrite lind_occ_functional.
  rewrite ih at 1.
  f_equal. apply proof_irrelevance.
  reflexivity.
  reflexivity.
  reflexivity.
  contradict h5.
  apply in_app_l; auto.
  contradict h4.
  rewrite (count_occ_In h1); omega. 
  unfold switch_eq_dec in h3'.
  f_equal.
  erewrite lind_occ_functional.
  rewrite ih; auto.
  f_equal. apply proof_irrelevance.
  reflexivity.
  reflexivity.
  reflexivity.
  Grab Existential Variables.
  assumption.
  assumption.
Qed.


Lemma lind_occ_app_ge_in2 :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (pf':In x l') (occ:nat),
    count_occ pfdt l x <= occ ->
    let pfa := in_app_r pf' l in
    lind_occ pfdt (l++l') occ x pfa =
    length l + lind_occ pfdt l' (occ - count_occ pfdt l x) x pf'.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; f_equal; auto with arith; apply proof_irrelevance.
  intros l' x h2 occ h3.
  rewrite (switch_eq_dec_eq h1 x a).
  destruct (h1 a x) as [|h4]; subst. 
  unfold switch_eq_dec.
  assert (h4:occ = S (pred occ)).  omega.
  rewrite h4.
  destruct in_dec as [h5 | h5].
  f_equal.
  rewrite <- h4.
  erewrite lind_occ_functional. rewrite ih at 1.
  f_equal.
  f_equal.
  assert (h6:pred occ - count_occ h1 l x = occ - S (count_occ h1 l x)).
    omega.
  rewrite h6 at 1.
  omega.
  omega.
  reflexivity.
  reflexivity.
  reflexivity.
  contradict h5; apply in_app_r; auto.
  unfold switch_eq_dec.
  f_equal.
  erewrite lind_occ_functional.
  rewrite ih.
  f_equal.
  assumption.
  reflexivity.
  reflexivity.
  reflexivity.
Qed.


Lemma lind_app_cons_nin_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T),
    ~In a l ->
    l <> nil ->
    let pfin := in_app_r (in_eq _ _) l in
    lind pfdt (l++a::l') a pfin = length l.
  intros T h1 l.
  induction l as [|c l ih].
  intros l' a h2 h3.
  contradict h3; auto.
  intros l' a h2 h3.
  unfold lind.
  simpl.
  destruct (h1 a c) as [h4 | h4].
  subst.
  contradict h2; left; auto.
  f_equal.
  assert (h5:~In a l).
  contradict h2.
  right; auto.
  specialize (ih l' _ h5).
  destruct (nil_dec' l) as [|h6]; subst.
  gen0.
  rewrite app_nil_l.
  intro h6.
  simpl.
  destruct (h1 a a) as [|h7]; auto.
  contradict h7; auto.
  specialize (ih h6).
  simpl in ih.
  erewrite lind_functional in ih.
  unfold lind in ih.
  rewrite ih at 1.
  reflexivity.
  reflexivity.
Qed.

  
  
Lemma lind_app_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T)
         (pf: ~In a l) (pf': In a l'),
    let pfin := in_app_r pf' l in
    lind pfdt (l++l') a pfin =
    length l + lind pfdt l' a pf'.
  intros T h1 l l' a h2 h3. simpl.
  destruct (nil_dec' l) as [|ha]; subst.
  gen0.
  rewrite app_nil_l.
  intros; simpl; f_equal; apply proof_irrelevance.
  revert ha h2. revert  l. revert h3. revert a.
  induction l' as [|a' l' ih].
  intros; contradiction.
  intros a h3 l ha h2.
  simpl.
  destruct h3 as [|h3]; subst. 
  unfold lind at 2.
  simpl. 
  destruct (h1 a a) as [|h3]. 
  rewrite plus_0_r.
  gen0.
  rewrite <- (app_sing l' a).
  pose proof (app_assoc l (a::nil) l') as h4.
  simpl in h4.
  simpl.
  rewrite h4.
  intro h5.
  erewrite lind_functional.  
  rewrite lind_app_in1 at 1. 
  rewrite lind_app_cons_nin_eq at 1; auto.
  reflexivity.
  contradict h3; auto.
  unfold lind at 2. simpl. 
  destruct (h1 a a') as [|h4]; subst.
  rewrite plus_0_r. 
  erewrite lind_functional.  
  rewrite lind_app_cons_nin_eq at 1; auto.
  reflexivity.
  specialize (ih a h3 (l++a'::nil)).
  hfirst ih.
  intro h5.
  apply app_eq_nil in h5.
  destruct h5; discriminate.
  specialize (ih hf).
  hfirst ih.
  intro h6.
  rewrite in_app_iff in h6.
  destruct h6 as [| h6]; try contradiction.
  destruct h6; subst.
  contradict h4; auto.
  contradiction.
  specialize (ih hf0). 
  revert ih.
  Gen0.
  rewrite <- app_assoc.
  simpl.
  intros h6 ih.
  erewrite lind_functional.
  rewrite ih at 1.
  rewrite app_length. 
  simpl.
  unfold lind.
  rewrite <- plus_assoc.
  rewrite (plus_comm 1).
  rewrite S_compat.
  f_equal.
  f_equal.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
Qed.


Lemma skipn_lind_app_cons :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T)
         (pf:In a (l++a::l')),
    ~In a l ->
    skipn (lind pfdt (l++a::l') a pf) (l++a::l') = a::l'.
  intros T h1 l l' a h2 h3.
  erewrite lind_functional. rewrite lind_app_nin at 1. 
  rewrite skipn_plus.
  rewrite lind_cons_same.
  rewrite skipn0.
  rewrite skipn_length_app; auto.
  assumption.
  reflexivity.

  Grab Existential Variables.
  left; auto.
Qed.

 
Lemma skipn_lind_sub_list_not_nil :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a:T)
         (pf:sub_list (a::l) l'),
    let pfin := sub_list_in (a::l) l' a (in_eq _ _) pf in
    let n := lind pfdt  _ _ pfin in
    skipn n l' <> nil.
  intros T h1 l l' a h2.
  simpl.
  inversion h2 as [l1 l2]; subst.
  destruct (in_dec h1 a l1) as [h3 | h3].
  pose proof (lind_app_in1 h1 l1 ((a::l) ++ l2) a h3) as h4.
  simpl in h4.
  erewrite lind_functional.
  rewrite h4 at 1.
  rewrite skipn_app.
  intro h5.
  apply app_eq_nil in h5.
  destruct h5 as [? h5].
  apply app_eq_nil in h5.
  destruct h5; discriminate.
  apply lt_le_weak.
  apply lt_lind.
  reflexivity.
  destruct (nil_dec' l2) as [| h4]; subst.
  rewrite app_nil_r at 3.
  match goal with
      |- skipn (lind _ _  _ ?pf) _ <> _ =>  generalize pf end.
  rewrite app_nil_r.
  intro h4.
  rewrite skipn_lind_app_cons.
  intro; discriminate.
  assumption.
  pose proof (lind_app_nin h1 l1 ((a::l) ++ l2)  a h3) as h5.
  hfirst h5.
  apply in_app_l; left; auto.
  specialize (h5 hf).
  simpl in h5.
  erewrite lind_functional.
  rewrite h5 at 1.
  rewrite skipn_length_app_plus.
  erewrite lind_functional. rewrite lind_cons_same.
  rewrite skipn0.
  intro h6.
  apply app_eq_nil in h6.
  destruct h6; discriminate.
  reflexivity.
  reflexivity.

  Grab Existential Variables.
  left; auto.
Qed.

  

Lemma count_occ_firstn_eq0_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    count_occ pfdt (firstn n l) x = 0  <->
    (forall pf:In x l,
       n <= lind pfdt l x pf).
  intros T h1 l x n. revert l x.
  induction n as [|n ih]; simpl. 
  intros l x; split; intros; auto with arith.
  intros l x; split; intro h2.
  intro hi.
  destruct l as [|a l].
  contradiction.
  simpl in h2.
  destruct (h1 a x) as [|h3]; subst.
  discriminate.
  rewrite ih in h2.
  unfold lind; simpl.
  apply neq_sym in h3.
  destruct (h1 x a) as [|h5].
  contradiction.
  apply le_n_S.
  destruct hi as [|hi]; subst.
  contradict h5; auto.
  specialize (h2 hi).
  unfold lind in h2.
  erewrite lind_occ_functional.
  apply h2.
  reflexivity.
  reflexivity.
  reflexivity.
  destruct l as [|a l]; simpl; auto.
  destruct (h1 a x) as [|h3]; subst.
  specialize (h2 (in_eq _ _)).
  unfold lind in h2.
  simpl in h2.
  destruct (h1 x x) as [|h4].
  omega.
  pose proof (h4 eq_refl); contradiction.
  rewrite ih.
  intro hi.
  specialize (h2 (in_cons _ _ _ hi)).
  unfold lind in h2.
  simpl in h2.
  destruct (h1 x a) as [|h5]; subst.
  contradict h3; auto.
  apply le_S_n in h2.
  unfold lind.
  erewrite lind_occ_functional.
  apply h2.
  reflexivity.
  reflexivity.
  reflexivity.
Qed.

  

Lemma count_occ_skipn_eq0_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat),
    count_occ pfdt (skipn n l) x = 0  <->
    (forall (pf:In x l) (occ:nat),
       occ < count_occ pfdt l x ->
       lind_occ pfdt l occ x pf < n).
  intros T h1 l x n. revert x l.
  induction n as [|n ih]; simpl.
  intros x l; split; intro h2.
  intros h3 occ h4.
  pose proof h3 as h3'. rewrite (count_occ_In h1) in h3'.
  omega.
  destruct (in_dec h1 x l) as [h3 | h3].
  specialize (h2 h3 0).
  hfirst h2.
  rewrite <- (count_occ_In h1 l x); auto.
  specialize (h2 hf).
  omega.
  rewrite (count_occ_In h1) in h3.
  omega.
  intros x l; split; intro h2.
  intros h3 occ h4.
  destruct l as [|a l].
  simpl in h4.
  omega.
  simpl in h4, h3. simpl.
  rewrite (switch_eq_dec_eq h1 a x) in h4.
  destruct (h1 x a) as [|h5]; subst.
  unfold switch_eq_dec in h4.
  destruct occ as [|occ]; auto with arith.
  destruct (in_dec h1 a l) as [h5 | h5].
  apply lt_n_S.
  apply ih; auto with arith.
  auto with arith.
  unfold switch_eq_dec in h4.
  destruct h3 as [|h3]; subst.
  contradict h5; auto.
  apply lt_n_S.
  apply ih; auto.
  destruct l as [|a l]; simpl; auto.
  rewrite ih.
  intros h3 occ h4.
  destruct (h1 x a) as [|h5]; subst.
  specialize (h2 (in_eq _ _) (S occ)).
  simpl in h2.
  destruct (h1 a a) as [|h5].
  hfirst h2; auto with arith.
  specialize (h2 hf).
  destruct in_dec as [|h6]; auto with arith.
  pose proof (proof_irrelevance _ i h3); subst; auto with arith.
  contradiction.
  pose proof (h5 eq_refl); contradiction.
  specialize (h2 (in_cons _ _ _ h3)).
  specialize (h2 occ).
  simpl in h2.
  destruct (h1 a x) as [|h6]; subst.
  contradict h5; auto.
  specialize (h2 h4).
  destruct (h1 x a) as [|h7]; subst.
  contradict h5; auto.
  apply lt_S_n in h2.
  erewrite lind_occ_functional.
  apply h2.
  reflexivity.
  reflexivity.
  reflexivity.
Qed.

 
Lemma count_occ_skipn_S_lind :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l),
    count_occ pfdt (skipn (S (lind pfdt l x pf)) l) x  =
    pred (count_occ pfdt l x).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2.
  destruct h2 as [|h2]; subst.
  unfold lind; simpl.
  rewrite eq_dec_same at 1 2; auto.
  destruct (h1 a x) as [|h3]; subst.
  erewrite lind_functional.
  rewrite lind_hd' at 1; simpl; auto.
  reflexivity.
  erewrite <- ih.
  repeat f_equal.
  unfold lind; simpl.
  destruct (h1 x a) as [|h4]; subst.
  contradict h3; auto.
  f_equal. f_equal.
  apply proof_irrelevance.
  Grab Existential Variables.
  assumption.
  apply cons_neq_nil.
Qed.


Lemma lt_lind_S_occ :
  forall {T:Type} (h1:type_eq_dec T) (l:list T) (x:T) (pf:In x l) (occ:nat),
    lind h1 l x pf < lind_occ h1 l (S occ) x pf \/ 1 = count_occ h1 l x.
  intros T h1 l. 
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 occ.
  destruct (h1 x a) as [|h3]; subst.
  destruct (h1 a a) as [|h4].
  destruct in_dec as [h3 | h3].
  unfold lind.
  simpl.
  destruct (h1 a a) as [|h4].
  left; auto with arith.
  contradict h4; auto. 
  right.
  simpl.
  rewrite (count_occ_In h1) in h3.
  omega.
  contradict h4; auto.
  unfold lind.
  simpl. 
  destruct h2 as [|h2]; subst.
  contradict h3; auto.
  specialize (ih _ h2).
  destruct (h1 a x) as [|he]; subst.
  contradict h3; auto.
  destruct (h1 x a) as [|he']; subst.
  left; auto with arith.
  match goal with |- S (lind_occ _ _ _ _ ?pf) < _ \/ _  => assert (h5: h2 = pf) end.
  apply proof_irrelevance.
  rewrite <- h5; auto.
  fold (lind h1 l x h2). 
  specialize (ih occ).
  destruct ih as [h4 | h4].
  left.
  apply lt_n_S.
  pose proof (proof_irrelevance _ he' h3) as h6.
  rewrite <- h6.
  rewrite <- h5; auto.
  right; auto.
Qed.

Lemma lt_lind_occ_S :
  forall {T:Type} (h1:type_eq_dec T) (l:list T) (x:T) (pf:In x l) (occ:nat),
    S occ < count_occ h1 l x ->
    lind_occ h1 l occ x pf < lind_occ h1 l (S occ) x pf.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 occ h3.
  destruct h2 as [|h2]; subst.
  destruct (h1 x x) as [h4 | h4].
  apply lt_S_n in h3.
  pose proof (count_occ_in h1 _ _ _ h3) as h5.
  destruct occ as [|occ].
  destruct in_dec as [| h6]; auto with arith.
  contradiction.
  destruct in_dec as [h6 | h6].
  apply lt_n_S.
  apply ih; auto.
  contradiction.
  apply lt_n_S.
  apply ih; auto.
  rewrite switch_eq_dec_eq in h3.
  destruct (h1 x a) as [|h4]; subst.
  unfold switch_eq_dec in h3.
  apply lt_S_n in h3.
  destruct in_dec as [|h4].
  destruct occ as [|occ]; auto with arith.
  contradiction.
  unfold switch_eq_dec in h3.
  apply lt_n_S.
  apply ih; auto.
Qed.

Lemma lt_lind_occ_lt :
  forall {T:Type} (h1:type_eq_dec T) (l:list T) (x:T) (pf:In x l) (occ occ':nat),
    occ < occ' ->
    occ' < count_occ h1 l x ->
    lind_occ h1 l occ x pf < lind_occ h1 l occ' x pf.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 occ occ' h3 h4.
  destruct h2 as [|h2]; subst.
  destruct (h1 x x) as [h5 | h5].
  destruct occ' as [|occ'].
  apply lt_n_0 in h3; contradiction.
  apply lt_S_n in h4.
  destruct in_dec as [h6 | h76].
  destruct occ as [|occ]; auto with arith.
  apply count_occ_in in h4; contradiction. 
  contradict h5; auto.
  rewrite switch_eq_dec_eq in h4.
  destruct (h1 x a) as [|h5]; subst.
  simpl in h4.
  destruct occ' as [|occ'].
  apply lt_n_0 in h3.
  contradiction.
  destruct occ as [|occ].
  destruct in_dec as [h5 | h5]; auto with arith.
  contradiction.
  apply lt_S_n in h3.
  apply lt_S_n in h4.
  destruct in_dec as [h5 | h5].
  apply lt_n_S.
  apply ih; auto.
  contradiction.
  simpl in h4.
  apply lt_n_S.
  apply ih; auto.
Qed.


Lemma in_list_of_lists_seqs_iff :
  forall  {T:Type} (ll:faml T) (l:list T),
    In l (list_of_lists_seqs ll) <->
    (length l = length ll /\
     (forall (i:nat) (pfi:i < length l),
      exists pfi':i < length ll,
        In (nth_lt l i pfi) (nth_lt ll i pfi'))).
  intros T ll. 
  induction ll as [|al ll ih]; simpl.
  intros l.
  split; intro h1.
  destruct h1 as [|h1]; subst; try contradiction.
  simpl.
  split; auto.
  intros ? h1.
  pose proof (lt_n_0 _ h1).
  contradiction.
  destruct l as [|a l].
  left; auto.
  destruct h1 as [h1].
  simpl in h1.
  discriminate.
  intro l.
  split.
  intros h1.
  rewrite in_map_iff in h1.
  destruct h1 as [p [h1 h2]].
  subst.
  simpl.
  destruct p as [x l].
  simpl.
  rewrite in_prod_iff in h2.
  destruct h2 as [h2 h3].
  rewrite ih in h3.
  destruct h3 as [h3 h4].
  split. 
  f_equal; auto.
  intros i h5.
  rewrite h3 in h5. 
  exists h5.
  destruct i as [|i].
  assumption.
  apply lt_S_n in h5.
  rewrite <- h3 in h5.
  specialize (h4 _ h5).
  destruct h4 as [h4 h6].
  rewrite nth_lt_compat in h6.
  rewrite (nth_lt_compat ll) in h6.
  erewrite nth_indep, (nth_indep ll).
  apply h6.
  assumption.
  assumption.

  intro h1.
  destruct h1 as [h1 h2].
  destruct l as [|a l].
  simpl in h1.
  discriminate.
  simpl in h1.
  apply S_inj in h1.
  specialize (ih l).
  hr ih.
  split; auto.
  intros i h3.
  specialize (h2 (S i)).
  hfirst h2.
  simpl.
  apply lt_n_S; auto.
  specialize (h2 hf).
  destruct h2 as [h2 h4].
  simpl in h2.
  apply lt_S_n in h2.
  exists h2.
  simpl in h4.
  erewrite nth_lt_compat, (nth_lt_compat ll).
  erewrite nth_indep, (nth_indep ll).
  apply h4.
  assumption.
  assumption.
  rewrite <- ih in hr.
  rewrite in_map_iff.
  exists (a, l).
  simpl.
  split; auto.
  rewrite in_prod_iff.
  split; auto.
  specialize (h2 0 (lt_0_Sn _)).
  simpl in h2.
  destruct h2; auto.
Qed.

Lemma in_list_of_lists_seqs_length :
  forall  {T:Type} (ll:faml T) (l:list T),
    In l (list_of_lists_seqs ll) ->
    length l = length ll.
  intros T ll l h1.
  rewrite in_list_of_lists_seqs_iff in h1.
  destruct h1; auto.
Qed.


Lemma in_list_of_lists_seqs_lt :
  forall  {T:Type} (ll:faml T) (l:list T)
          (i:nat),
    In l (list_of_lists_seqs ll) ->
    i < length l ->
    i < length ll.
  intros; rewrite <- (in_list_of_lists_seqs_length ll l); auto.
Qed.


Lemma coarse_list_in_list_of_lists_seqs :
  forall {T:Type} (ll:faml T) (l:list T)
         (v:bv) (pfeq:length l = length v)
         (pfi:In l (list_of_lists_seqs ll)),
    let pfeq' := eq_trans (eq_sym (in_list_of_lists_seqs_length _ _ pfi)) pfeq in
    let fl := map in 
    In (mask_to_coarse_list pfeq) (list_of_lists_seqs (mask_to_coarse_list pfeq')).
  intros T ll l v h1 h2; simpl.
  rewrite in_list_of_lists_seqs_iff.
  split.
  do 2 rewrite length_mask_to_coarse_list; auto.
  intros i h3.
  ex_goal.
  rewrite length_mask_to_coarse_list.
  rewrite length_mask_to_coarse_list in h3; auto.
  exists hex.
  pose proof h2 as h2'.
  rewrite in_list_of_lists_seqs_iff in h2'.
  destruct h2' as [h4 h5].
  pose proof h3 as h3'.
  rewrite length_mask_to_coarse_list in h3'.
  specialize (h5 (nth_lt _ _ h3')).
  hfirst h5.
  rewrite h1.
  apply in_bv1s_lt.
  apply in_nth_lt.
  specialize (h5 hf).
  destruct h5 as [h5 h6].
  unfold mask_to_coarse_list.
  erewrite nth_lt_map_in_dep.
  erewrite nth_lt_map_in_dep.
  erewrite nth_lt_functional3, (nth_lt_functional3 ll).
  apply h6.
Qed.


    


Lemma eq_hd_lind_occ_tl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x (tl l))
         (occ:nat),
    x = hd x l  ->
    let pf' := in_tl _ _ pf in
    lind_occ pfdt (tl l) occ x pf = pred (lind_occ pfdt l (S occ) x pf').
  intros T h1 l.
  destruct l as [|a l]; simpl.
  intros; contradiction. 
  intros x h2 occ ?; subst.
  destruct (h1 a a) as [h3 | h3].
  destruct in_dec as [h4 | h4].
  simpl.
  unfold lind; f_equal; apply proof_irrelevance.
  contradiction.
  contradict h3; auto.
Qed.


Lemma neq_hd_lind_occ_tl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x (tl l))
         (occ:nat),
    hd x l <> x ->
    let pf' := in_tl _ _ pf in
    lind_occ pfdt (tl l) occ x pf = pred (lind_occ pfdt l occ x pf').
  intros T h1 l.
  destruct l as [|a l]; simpl.
  intros; contradiction.
  intros x h2 h3 hn.
  destruct (h1 x a) as [|h4]; subst.
  contradict h3; auto.
  simpl.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma eq_hd_lind_tl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x (tl l)),
    x = hd x l  ->
    let pf' := in_tl _ _ pf in
    lind pfdt (tl l) x pf = pred (lind_occ pfdt l 1 x pf').
  unfold lind; intros; apply eq_hd_lind_occ_tl; auto.
Qed.

Lemma neq_hd_lind_tl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x (tl l)),
    hd x l <> x ->
    let pf' := in_tl _ _ pf in
    lind pfdt (tl l) x pf = pred (lind pfdt l x pf').
  unfold lind; intros; apply neq_hd_lind_occ_tl; auto.
Qed.



Lemma in_tl_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
      In x (tl l) <->
      (exists pf:In x l,
         x = hd' l (in_not_nil _ _ pf) -> count_occ pfdt l x > 1).
  intros T h1 l x; split; intro h2.
  exists (in_tl _ _ h2).
  intro h3.
  destruct l as [|a l].
  contradict h2.
  simpl in h2, h3.
  subst.
  simpl.
  destruct (h1 a a) as [h3 | h3].
  apply lt_n_S.
  rewrite <- (count_occ_In h1 l a); auto.
  contradict h3; auto.
  destruct h2 as [h2 h3].
  pose proof h2 as h2'.
  rewrite (hd_compat' _ (in_not_nil _ _ h2)) in h2'.
  destruct h2' as [h4 | h4].
  symmetry in h4.
  specialize (h3 h4).
  destruct l as [|a l].
  simpl in h3.
  contradiction.
  simpl in h2, h3, h4.
  simpl.
  destruct (h1 a x) as [h5 | h5].
  apply lt_S_n in h3.
  rewrite (count_occ_In h1); auto.
  subst.
  contradict h5; auto.
  assumption.
Qed.

   
Lemma in_tl_iff' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    In x (tl l) <->
      (exists (n:nat) (pf:S n < length l),
         nth_lt l (S n) pf = x).
  intros T h1 l x; split; intro h2. 
  pose proof (lt_lind h1 _ _ h2) as h3.
  exists (lind h1 (tl l) x h2).
  ex_goal.
  rewrite length_tl in h3; omega.
  exists hex.
  rewrite nth_lt_S.
  erewrite nth_lt_functional3. rewrite lind_compat; auto.
  destruct h2 as [n [h2 h3]]; subst.
  rewrite nth_lt_S.
  apply in_nth_lt.
Qed.



Lemma length_inv_im_in_dep_inj :
  forall {T U:Type}
         (pfdu:type_eq_dec T)
         (pfdu:type_eq_dec U)
         (l:list T),
    NoDup l ->
    forall (u:U) (f:fun_in_dep l U),
      inj_dep f ->
      let l' := inv_im_in_dep pfdu l u f in
      l' <> nil ->
      length l' = 1.
  intros T U h1 h1' l hnd u f h2 l' h3.
  destruct (nil_dec' (tl l')) as [h4 | h4].
  rewrite (hd_compat' _ h3).
  rewrite h4; auto.
  pose proof (in_hd' _ h3) as h5.
  pose proof (in_hd' _ h4) as h6.
  rewrite (in_tl_iff h1) in h6.
  destruct h6 as [h6 h7].
  unfold l' in h5, h6.
  pose proof h6 as h6'.
  rewrite in_inv_im_in_dep_iff in h5, h6'.
  destruct h5 as [h5 h5'], h6' as [h8 h8']. 
  rewrite <- h5' in h8'.
  apply h2 in h8'.
  fold l' in h8'.
  pose proof (proof_irrelevance _ h3
                                (in_not_nil l' (hd' (tl l') h4) h6)) as h9.
  rewrite <- h9 in h7.
  specialize (h7 h8').
  rewrite count_occ_no_dup in h7; auto.
  omega.
  unfold l'.
  apply no_dup_inv_im_in_dep; auto.
Qed.

  
Lemma in_firstn_iff' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (m:nat),
    In x (firstn m l) <->
    (exists pf:In x l,
       lind pfdt l x pf < m).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; rewrite firstn_nil.
  simpl.
  split; intro; auto; try contradiction.
  destruct H; auto.
  intros x m; split; intro h2.
  destruct m as [|m].
  simpl in h2.
  contradiction.
  simpl in h2.
  destruct h2 as [|h2]; subst.
  exists (or_introl eq_refl).
  unfold lind.
  simpl.
  destruct (h1 x x); auto with arith.
  contradict n; auto.
  rewrite ih in h2.
  destruct h2 as [h2 h3].
  exists (or_intror h2).
  unfold lind; simpl.
  destruct (h1 x a) as [|h4]; auto with arith.
  apply lt_n_S.
  erewrite lind_occ_functional.
  unfold lind in h3.
  apply h3.
  reflexivity.
  reflexivity.
  reflexivity.
  destruct h2 as [h2 h3].
  destruct h2 as [|h2]; subst.
  unfold lind in h3.
  simpl in h3.
  destruct (h1 x x) as [|h4].
  rewrite (S_pred _ _ h3).
  simpl.
  left; auto.
  pose proof (h4 eq_refl); contradiction.
  unfold lind in h3.
  simpl in h3.
  destruct (h1 x a) as [|h4]; subst.
  rewrite (S_pred _ _ h3).
  simpl.
  left; auto.
  rewrite (S_pred _ _ h3).
  simpl.
  right.
  rewrite ih.
  exists h2.
  let P := type of h3 in match P with S (lind_occ _ _ _ _ ?pf) < _ => pose pf as hyp end.
  fold hyp in h3.
  fold (lind h1 l x hyp) in h3.
  pose proof (proof_irrelevance _ h2 hyp) as h5.
  rewrite h5.
  omega.
Qed.

Lemma in_nth_lt_skipn :
  forall {T:Type} (l:list T) (n:nat)
         (pflt:S n <= length l),
      In (nth_lt l n pflt) (skipn n l).
  intros T l n.
  revert l.
  induction n as [|n ih]; simpl.
  intros; apply in_nth_lt.
  intros l h1.
  destruct l as [|a l].
  simpl in h1.
  omega.
  simpl in h1.
  pose proof h1 as h1'. apply le_S_n in h1'.
  simpl.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  apply ih.
  omega.
  Grab Existential Variables.
  omega.
Qed.





Lemma in_skipn_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (m:nat),
    In x (skipn m l) <->
    (exists (pf:In x l) (occ:nat),
       occ < count_occ pfdt l x /\
       lind_occ pfdt l occ x pf >= m).
  intros T h1 l x m; split; intro h2.
  pose proof (in_skipn _ _ _ h2) as h3.
  exists h3.
  revert h3 h2. revert x l. 
  induction m as [|m ih]; simpl.
  intros; auto with arith.
  exists 0; auto with arith.
  split; auto with arith.
  rewrite  (count_occ_In h1) in h2; auto with arith.
  intros x l h2.
  destruct l as [|a l].
  contradiction.
  intro h3. 
  pose proof (in_skipn _ _ _ h3) as h4.
  specialize (ih _ _ h4 h3).
  destruct ih as [occ ih]. 
  destruct h2 as [|h2]; subst.
  exists (S occ).
  simpl.
  destruct (h1 x x) as [h5 | h5].
  destruct in_dec as [h6 | h6].
  pose proof (proof_irrelevance _ h4 h6); subst; auto with arith.
  destruct ih as [ih ih'].
  split; auto with arith.
  contradiction.
  contradict h5; auto. 
  destruct ih as [ih ih'].
  destruct (h1 a x) as [|h5]; subst.
  exists (S occ).
  simpl.
  destruct (h1 x x) as [h1x | h1x].
  destruct in_dec as [h5 | h5].
  pose proof (proof_irrelevance _ h4 h5); subst.
  split; auto with arith.
  contradiction.
  contradict h1x; auto.
  exists occ; simpl. 
  rewrite (switch_eq_dec_eq h1 a x).
  destruct (h1 x a) as [|h6]; subst.
  contradict h5; auto.
  unfold switch_eq_dec.
  pose proof (proof_irrelevance _ h4
                                (neq_in_cons l x a h6 (or_intror h2)));
    subst.
  split; auto with arith.

  destruct h2 as [h2 [occ [h3 h4]]].
  revert h2 occ h3 h4. revert x l.
  induction m as [|m ih]; simpl.
  intros; eapply count_occ_in.
  apply h3.
  intros x l h2 occ h3 h4.
  destruct l as [|a l].
  contradiction. 
  destruct h2 as [|h2]; subst.
  rewrite count_occ_cons_eq in h3; auto.
  destruct occ as [|occ].
  simpl in h4. 
  destruct (h1 x x) as [h5 | h5].
  omega.
  pose proof (h5 eq_refl); contradiction.
  apply lt_S_n in h3.
  simpl in h4.
  destruct (h1 x x) as [h5 | h5].
  destruct in_dec as [h6 | h6]. 
  eapply ih.
  apply h3.
  apply le_S_n in h4.
  apply h4.
  omega.
  pose proof (h5 eq_refl); contradiction.
  simpl in h3, h4.
  rewrite switch_eq_dec_eq in h3.
  destruct (h1 x a) as [|h5]; subst.
  unfold switch_eq_dec in h3.
  destruct occ as [|occ].
  omega.
  destruct in_dec as [h5 | h5].
  apply lt_S_n in h3.
  apply le_S_n in h4.
  eapply ih.
  apply h3. apply h4.
  contradiction.
  unfold switch_eq_dec in h3.
  apply le_S_n in h4.
  eapply ih.
  apply h3.
  apply h4.
Qed.

Lemma in_skipn_le :
  forall {T:Type} (m n:nat) (l:list T) (x:T),
    m <= n ->
    In x (skipn n l) ->
    In x (skipn m l).
  intros T m n. revert m.
  induction n as [|n ih]; simpl.
  intros m l x h1.
  apply le0 in h1; subst; simpl; auto.
  intros m l x h1 h2.
  destruct l as [|a l]; simpl.
  contradiction.
  destruct m as [|m]; simpl.
  right.
  apply in_skipn in h2; auto.
  apply le_S_n in h1; auto.
Qed.

Lemma in_firstn_le :
  forall {T:Type} (m n:nat) (l:list T) (x:T),
    m <= n ->
    In x (firstn m l) ->
    In x (firstn n l).
  intros T m n. revert m.
  induction n as [|n ih]; simpl.
  intros m l x h1.
  apply le0 in h1; subst; simpl; auto.
  intros m l x h1 h2.
  destruct l as [|a l]; simpl.
  rewrite firstn_nil in h2.
  contradiction.
  destruct m as [|m]; simpl.
  simpl in h2.
  contradiction.
  simpl in h2; auto.
  destruct h2 as [|h2].
  left; auto.
  right.
  apply le_S_n in h1.
  specialize (ih _ _ _ h1 h2); auto.
Qed.


Lemma nin_firstn_le_lind :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (m:nat)
         (pf:In x l),
    ~In x (firstn m l) ->
    m <= lind pfdt l x pf.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x m h2.
  destruct h2 as [|h2]; subst. 
  intro h2.
  destruct m as [|m]; auto with arith.
  unfold lind.
  simpl.
  destruct (h1 x x) as [|h3].
  contradict h2.
  simpl.
  left; auto.
  contradict h3; auto.
  intro h3.
  unfold lind.
  simpl.
  destruct (h1 x a) as [|h4]; subst.
  destruct m as [|m]; auto.
  contradict h3.
  simpl.
  left; auto. 
  specialize (ih x (pred m) h2).
  hfirst ih.
  contradict h3.
  destruct m as [|m].
  simpl in h3.
  contradiction. 
  simpl.
  simpl in h3.
  right; auto.
  specialize (ih hf).
  pose proof (proof_irrelevance
                _ h2
                (neq_in_cons _ _ a h4 (or_intror h2))) as h6.
  rewrite <- h6.
  unfold lind in ih; omega.
Qed.

 
Lemma in_firstn_nin_lt :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (i j:nat),
    i <= length l ->
    In x (firstn i l) ->
    ~In x (firstn j l) ->
    j < i.
  intros T hdt l x i j hle h1 h2. 
  rewrite in_firstn_iff in h1.
  destruct h1 as [k [h3 h4]].
  rewrite <- h4 in h2.
  eapply (nin_firstn_le_lind hdt _   _ _ (in_nth_lt _ _ _)) in h2.
  pose proof (le_lind_nth_lt hdt l k (lt_le_trans _ _ _ h3 hle)) as h5.
  eapply le_lt_trans.
  eapply le_trans.
  apply h2.
  apply h5.
  assumption.
Qed.

  
Lemma lind_occ_firstn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (m:nat)
         (pf:In x (firstn m l)) (occ:nat),
    occ < count_occ pfdt (firstn m l) x ->
    lind_occ pfdt (firstn m l) occ x pf =
    lind_occ pfdt l occ x (in_firstn _ _ _ pf).
  intros T h1 l.
  induction l as [|a l ih]; simpl. 
  intros ? ? h2. pose proof h2 as h2'. rewrite firstn_nil in h2'.
  contradiction.
  intros x m h2 occ hlt.
  destruct m as [|m].
  simpl in h2. contradiction.
  simpl.
  simpl in h2, hlt.
  rewrite switch_eq_dec_eq in hlt.
  destruct (h1 x a) as [h4 | h4]; subst.
  unfold switch_eq_dec in hlt. 
  destruct occ as [|occ]; auto.
  apply lt_S_n in hlt; auto.
  destruct in_dec as [h3 | h3].
  destruct in_dec as [h4 | h4].
  f_equal. 
  rewrite ih; auto; f_equal; apply proof_irrelevance.
  contradict h4.
  apply in_firstn in h3; auto.
  contradict h3.
  rewrite (count_occ_In h1); omega.
  unfold switch_eq_dec in hlt.
  destruct h2 as [|h2]; subst.
  contradict h4; auto.
  f_equal.
  rewrite ih; auto.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma ex_lind_occ_nth_lt_firstn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (m n:nat)
         (pfle:m <= length l)
         (pflt:n < length l)
         (pfnm:n < m)
         (pfin:In (nth_lt l n pflt) (firstn m l)),
  exists (occ:nat),
    occ < count_occ pfdt (firstn m l) (nth_lt l n pflt) /\
    lind_occ pfdt (firstn m l) occ (nth_lt l n pflt) pfin = n.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega. 
  intros m n h2 h3 hnm h4.
  destruct n as [|n]; simpl.
  destruct m as [|m]; simpl.
  simpl in h4.
  contradiction.
  apply  le_S_n in h2.
  simpl in h4.
  destruct (h1 a a) as [h5 | h5].
  destruct in_dec as [h6 | h6].
  exists 0.
  split; auto with arith.
  exists 0.
  split; auto with arith.
  contradict h5; auto.
  apply lt_S_n in h3. 
  destruct m as [|m].
  simpl in h4.
  contradiction.
  apply le_S_n in h2.
  simpl in h4.
  simpl.
  rewrite switch_eq_dec_eq.
  destruct (h1 (nth n l a) a) as [h5 | h5].
  unfold switch_eq_dec.
  destruct in_dec as [h6 | h6].
  apply le_S_n in hnm.
  specialize (ih _ _ h2 h3 hnm). 
  hfirst ih. 
  rewrite nth_lt_compat.
  erewrite nth_indep.
  apply h6.
  assumption.  
  specialize (ih hf).
  destruct ih as [occ [h7 h8]].
  exists (S occ).
  split.
  apply lt_n_S.
  erewrite nth_lt_compat in h7.
  erewrite nth_indep in h7.
  apply h7.
  assumption. 
  f_equal.
  rewrite <- h8.
  apply lind_occ_functional; auto.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  apply lt_S_n in hnm.
  contradict h6.
  rewrite in_firstn_iff.
  exists n, hnm.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  apply lt_S_n in hnm.
  unfold switch_eq_dec.
  destruct h4 as [h4 | h4].
  pose proof h5 as h5'.
  rewrite h4 in h5' at 2. contradict h5'; auto.
  specialize (ih m n h2 h3 hnm).
  hfirst ih.
  rewrite nth_lt_compat.
  erewrite nth_indep.
  apply h4.
  omega.
  specialize (ih hf).
  destruct ih as [occ [h6 h7]].
  exists occ.
  split.
  rewrite nth_lt_compat in h6.
  erewrite nth_indep in h6.
  apply h6.
  omega.
  f_equal.
  rewrite <- h7.
  apply lind_occ_functional; auto.
  erewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  Grab Existential Variables.
  assumption.
Qed.


Lemma count_occ_eq1_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    count_occ pfdt l x = 1 <->
    exists pf:In x l,
      ~ In x (skipn (S (lind pfdt l x pf)) l).
  intros T h1 l x; split; intro h2.
  pose proof (count_occ_In h1 l x) as h3.
  rewrite h2 in h3.
  let P := type of h3 in match P with _ <-> ?hyp => assert (h4:hyp) end.
  auto with arith.
  rewrite <- h3 in h4.
  exists h4.
  intro h5.
  rewrite (in_skipn_iff h1 l x) in h5.
  destruct h5 as [h5 [occ [h6 h7]]].
  rewrite h2 in h6.
  assert (h8:occ = 0). omega.
  subst.
  unfold lind in h7.
  pose proof (proof_irrelevance _ h4 h5); subst.
  omega.
  destruct h2 as [h2 h3].
  rewrite (in_skipn_iff h1 l x) in h3.
  pose proof (count_occ_In h1 l x) as h4.
  pose proof h2 as h2'.
  rewrite h4 in h2'.
  rewrite (S_pred _ _ h2').
  f_equal.
  rewrite <- not_not0_iff.
  contradict h3.
  exists h2.
  exists 1.
  split.
  omega.
  pose proof (lt_lind_S_occ h1  l x h2 0) as h5.
  destruct h5 as [h5 | h5].
  omega.
  rewrite <- h5 in h3.
  simpl in h3.
  contradict h3; auto.
Qed.

  
Lemma nin_skipn_S_lind_occ_pred_count_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l),
    ~ In x (skipn (S (lind_occ pfdt l (pred (count_occ pfdt l x)) x pf)) l).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2.
  destruct h2 as [|h2]; subst.
  destruct (h1 x x) as [h3 | h3].
  simpl.
  destruct (in_dec h1 x l) as [h4 | h4].
  pose proof h4 as h4'.
  rewrite (count_occ_In h1) in h4'.
  rewrite (S_pred _ _ h4').
  apply ih.
  rewrite (count_occ_In h1) in h4.
  rewrite (not_lt0_iff (count_occ h1 l x)) in h4 at 1.
  rewrite h4.
  simpl.
  intro h5.
  rewrite (count_occ_In h1) in h5.
  omega.
  contradict h3; auto.
  destruct (h1 x a) as [|h3]; subst.
  destruct (h1 a a) as [h3 | h3].
  simpl.
  rewrite (count_occ_In h1) in h2.
  rewrite (S_pred _ _ h2). 
  destruct in_dec as [h4 | h4].
  apply ih.
  simpl; auto.
  contradict h3; auto.
  destruct (h1 a x) as [|h4]; subst.
  contradict h3; auto.
  apply ih.
Qed.
                                      
Lemma lind_occ_S :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l)
           (occ:nat),
    S occ < count_occ pfdt l x ->
    let l' := skipn (S  (lind pfdt l x pf)) l in
    forall (pf':In x l'),
      let lo := lind_occ pfdt l' occ x pf' in
      lind_occ pfdt l (S occ) x pf =
      lind pfdt l x pf + S lo.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 occ h3.
  destruct h2 as [|h2]; subst.
  destruct (h1 x x) as [|h4]; subst.
  apply lt_S_n in h3.
  destruct in_dec as [h4 | h4].
  erewrite lind_functional.
  rewrite lind_hd'.
  simpl.
  f_equal.
  intros; repeat f_equal; apply proof_irrelevance.
  simpl; auto.
  contradict h4.
  apply (count_occ_in h1) in h3; auto.
  contradict h4; auto.
  intros h4. 
  destruct (h1 x a) as [|h5]; subst.
  destruct (h1 a a) as [h10 | h10].
  apply lt_S_n in h3.
  destruct in_dec as [h5 | h5].
  generalize h4.
  erewrite lind_functional.
  rewrite lind_hd'.
  simpl.
  intro h6.
  repeat f_equal; apply proof_irrelevance.
  simpl.
  reflexivity.
  contradiction.
  contradict h10; auto.
  unfold lind.
  simpl.
  unfold lind in h4.
  simpl in h4.
  destruct (h1 a x) as [|h6]; subst.
  contradict h5; auto.
  assert (h7:h1 x a = right (x = a) h5).
    destruct (h1 x a). contradiction. f_equal; apply proof_irrelevance.
  rewrite h7 at 1.
  pose proof h4 as h4'.
  rewrite h7 in h4' at 1.
  simpl in h4'.
  rewrite plus_S_shift_r.
  f_equal.
  specialize (ih _ _ _ h3 h4').
  rewrite ih at 1.   
  rewrite plus_S_shift_r, plus_S_shift_l.
  f_equal.
  f_equal.
  generalize h4. rewrite h7.
  intro h8.
  f_equal.
  apply proof_irrelevance.
  Grab Existential Variables.
  apply cons_neq_nil.
  apply cons_neq_nil.
Qed.

Lemma in_firstn_S_lind :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l),
    In x (firstn (S (lind pfdt l x pf)) l).
  intros T h1 l x h2.
  erewrite firstn_S_eq.
  apply in_app_r.
  simpl.
  left.
  apply lind_compat.
Qed.

Lemma lind_occ_lt_mono :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfin:In x l)
         (occ occ':nat),
    occ < occ' ->
    occ' < count_occ pfdt l x ->
    lind_occ pfdt l occ x pfin < lind_occ pfdt l occ' x pfin.
  intros T h1 l.
  induction l  as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 occ occ' h3 h4.
  destruct h2 as [|h2]; subst.
  destruct (h1 x x) as [h5 | h5].
  destruct occ' as [|occ'].
  omega.
  destruct occ as [|occ].
  destruct in_dec as [h6 | h6]; auto with arith.
  contradict h6.
  apply lt_S_n in h4. apply count_occ_in in h4; auto.
  destruct in_dec as [h6 | h6]; auto.
  apply lt_n_S.
  apply ih; auto with arith.
  contradict h6.
  apply lt_S_n in h4. apply count_occ_in in h4; auto.
  contradict h5; auto.
  rewrite switch_eq_dec_eq in h4.
  destruct (h1 x a) as [|h5]; subst.
  simpl in h4.
  destruct occ' as [|occ'].
  omega.
  destruct occ as [|occ].
  destruct in_dec as [h5 | h5]; auto with arith.
  contradiction.
  destruct in_dec as [h5 | h5].
  apply lt_n_S.
  apply ih; auto with arith.
  contradiction.
  simpl in h4.
  apply lt_n_S.
  apply ih; auto.
Qed.


Lemma lind_occ_plus_occ_in_skip :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfin:In x l) (occ occ':nat)
         (pflt'':occ + (S occ') < count_occ pfdt l x),
    let lo  := lind_occ pfdt l occ x pfin in
    forall (pfins:In x (skipn (S lo) l)),
      let lo' := lind_occ pfdt (skipn (S lo) l) occ' x pfins in
      lind_occ pfdt l (occ + (S occ')) x pfin = lo  + (S lo').
  intros T h1 l x h2 occ.  revert h2. revert l x.
  induction occ as [|occ ih]; simpl.
  intros l x h2 occ' h3 hs.
  destruct l as [|a l]. 
  contradiction. 
  simpl. 
  simpl in h2, h3, hs.
  destruct h2 as [|h2]; subst.
  rewrite (eq_dec_same h1) at 1 2. 
  destruct (h1 x x) as [h5 | h5].
  destruct in_dec as [h6 | h6].
  simpl. 
  repeat f_equal; apply proof_irrelevance.
  simpl in hs. contradiction.
  pose proof (h5 eq_refl); contradiction.
  rewrite (switch_eq_dec_eq h1) in h3.
  destruct (h1 x a) as [h4 | h4]; subst.
  simpl in h3, hs.
  destruct in_dec as [h8 | h8]; simpl.
  repeat f_equal; apply proof_irrelevance.
  contradiction.
  simpl in h3.
  rewrite plus_S_shift_r.
  f_equal.
  simpl.
  rewrite <- plus_S_shift_r.
  fold (lind h1 l x (neq_in_cons l x a h4 (or_intror h2))).
  erewrite lind_occ_functional. rewrite lind_occ_S at 1.
  f_equal. f_equal.
  assumption.
  reflexivity.
  reflexivity.
  reflexivity.

  intros l x h2 occ' h3 h5.
  destruct l as [|a l].
  contradiction.
  simpl in h2, h3, h5.
  simpl. 
  destruct h2 as [|h2]; subst.
  destruct (h1 x x) as [h6 | h6].
  apply lt_S_n in h3.
  destruct in_dec as [h7 | h7].
  specialize (ih l x h7 occ' h3).
  specialize (ih h5).
  simpl in ih.
  rewrite ih.  
  rewrite plus_S_shift_l.
  reflexivity. 
  simpl in h5. contradiction.
  pose proof (h6 eq_refl); contradiction.
  rewrite switch_eq_dec_eq in h3.
  destruct (h1 x a) as [h6 |]; subst.
  simpl in h3.
  apply lt_S_n in h3.
  destruct in_dec as [h6 | h6].
  pose proof (proof_irrelevance _ h2 h6); subst.
  specialize (ih l a h6 _ h3 h5).
  simpl in ih.
  rewrite ih.
  rewrite plus_S_shift_l.
  reflexivity.
  simpl in h5.
  contradiction.
  simpl in h3.
  rewrite plus_S_shift_l.
  f_equal.
  pose proof (lind_occ_S h1 l x (neq_in_cons l x a n (or_intror h2)) (occ + S occ') h3) as h6.
  pose proof (lind_occ_S h1 l x (neq_in_cons l x a n (or_intror h2)) occ) as h7.
  hfirst h7.   omega. 
  specialize (h7 hf).
  
  pose proof (in_skipn_le
                (S (lind h1 l x (neq_in_cons l x a n (or_intror h2))))
                (S (lind_occ h1 l (S occ) x (neq_in_cons l x a n (or_intror h2)))) l x) as h9.
  hfirst h9. 
  apply le_S.
  pose proof (lt_lind_S_occ h1 l x (neq_in_cons l x a n (or_intror h2)) occ) as h10.
  destruct h10 as [h10 | h10].
  apply h10.
  rewrite <- h10 in h3.
  omega.
  specialize (h9 hf0 h5).
  specialize (h6 h9). specialize (h7 h9). 
  rewrite h6 at 1. rewrite h7 at 1.
  rewrite <- plus_assoc.
  f_equal. 
  rewrite plus_S_shift_l.
  f_equal. 
  pose proof (ih _ _ h9 occ). 
  pose proof (count_occ_skipn_S_lind h1 l x (neq_in_cons l x a n (or_intror h2))) as h8.
  pose proof (ih  (skipn (S (lind h1 l x (neq_in_cons l x a n (or_intror h2)))) l) x h9 occ') as ih'.
  hfirst ih'.
  rewrite h8.
  omega.  
  specialize (ih' hf1).
  assert (h10: In x
          (skipn (S (lind_occ h1
            (skipn (S (lind h1 l x (neq_in_cons l x a n (or_intror h2)))) l)
            occ x h9))
                  (skipn (S (lind h1 l x (neq_in_cons l x a n (or_intror h2)))) l))). 
  rewrite (in_skipn_iff h1).
  exists h9.
  exists (occ + S occ').
  split; auto.
  apply lind_occ_lt_mono; omega.
  specialize (ih' h10).
  rewrite ih'.
  f_equal.
  f_equal.
  generalize h10. rewrite <- skipn_plus.
  rewrite plus_comm.
  intro hf2'. 
  generalize h5.
  erewrite lind_occ_S.
  intro h5'.
  rewrite h8 in hf1.
  pose proof (proof_irrelevance
                _ 
                h5' hf2'); subst; auto.
  assumption.
Qed.


Lemma lind_occ_plus_occ_nin_skip :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfin:In x l) (occ occ':nat),
    let lo  := lind_occ pfdt l occ x pfin in
    forall (pfins: ~In x (skipn (S lo) l)),
      lind_occ pfdt l (occ + (S occ')) x pfin = lo.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 occ occ' h4.
  destruct h2 as [|h2]; subst.
  destruct (h1 x x) as [h5 | h5]. 
  destruct occ as [|occ].
  simpl in h4.
  simpl.
  destruct in_dec as [h6 | h6].
  contradiction.
  reflexivity.
  rewrite plus_S_shift_l.
  destruct in_dec as [h6 | h6].
  f_equal.
  specialize (ih x h6 occ occ' h4); auto.
  simpl in h4.
  reflexivity.
  pose proof (h5 eq_refl); contradiction.
  destruct (h1 x a) as [|h5]; subst.
  destruct occ as [|occ].
  simpl in h4.
  contradiction.
  rewrite plus_S_shift_l.
  destruct in_dec as [h5 | h5].
  f_equal.
  specialize (ih _ h5 occ occ' h4); auto.
  reflexivity.
  f_equal.
  specialize (ih x  (neq_in_cons l x a h5 (or_intror h2)) occ occ' h4); auto.
Qed.


Lemma lind_occ_terminates :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
           (pf:In x l) (occ:nat),
    occ >= pred (count_occ pfdt l x) ->
    lind_occ pfdt l occ x pf =
    lind_occ pfdt l (pred (count_occ pfdt l x)) x pf.
  intros T h1 l x h2 occ h3. 
  apply le_lt_eq_dec in h3.
  destruct h3 as [h3 | h3].
  apply lt_decompose in h3.
  destruct h3; subst.
  rewrite lind_occ_plus_occ_nin_skip; auto.
  apply nin_skipn_S_lind_occ_pred_count_occ.
  rewrite h3; auto.
Qed.


Lemma le_lind_occ :
  forall {T:Type} (h1:type_eq_dec T) (l:list T) (x:T) (pf:In x l) (occ:nat),
    lind h1 l x pf <= lind_occ h1 l occ x pf.
  intros T h1 l x h2 occ.
  destruct occ as [|occ].
  unfold lind; auto. 
  pose proof (lt_lind_S_occ h1 l x h2 occ) as h3. 
  destruct h3 as [| h3]; auto with arith.
  symmetry in h3.
  rewrite count_occ_eq1_iff in h3.
  destruct h3 as [h3 h4].
  rewrite (in_skipn_iff h1 l x) in h4.
  apply not_lt. 
  contradict h4.
  pose proof (lt_lind_S_occ h1 l x h2 occ) as h5.
  destruct h5 as [|h5]. omega.
  assert (h6:S occ >= count_occ h1 l x). omega.
  rewrite lind_occ_terminates in h4.
  rewrite <- h5 in h4.
  simpl in h4.
  unfold lind in h4.
  apply lt_irrefl in h4; contradiction.
  omega.
Qed.  

   
Lemma in_removelast_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    In x (removelast l) <->
    ((x <> last l x -> In x l) /\
     (x = last l x ->
     1 < count_occ pfdt l x)).
  intros T h1 l x.
  destruct (nil_dec' l) as [|h2]; subst.
  simpl.
  split; intro h2.
  contradiction.
  destruct h2 as [h2 h3].
  specialize (h3 eq_refl).
  omega.
  split; intro h3.
  split.
  intro h4. 
  rewrite (app_removelast_last x).
  apply in_app_l; auto.
  assumption.
  intro h4.
  rewrite h4.
  rewrite count_occ_last.
  apply lt_n_S.
  rewrite (count_occ_In h1) in h3.
  rewrite <- h4; auto with arith.
  assumption.
  destruct h3 as [h3 h4].
  destruct (h1 x (last l x)) as [h5 | h5]. 
  specialize (h4 h5). 
  rewrite (app_removelast_last x h2) in h4.
  rewrite count_occ_app in h4.
  simpl in h4.
  destruct (h1 (last l x) x) as [h6 | h6].
  rewrite S_compat in h4.
  apply lt_S_n in h4.
  rewrite (count_occ_In h1); auto. 
  rewrite h5 in h6 at 2.
  contradict h6; auto.
  specialize (h3 h5).
  rewrite (app_removelast_last x h2) in h3.
  apply in_app_or in h3.
  destruct h3 as [|h3]; auto.
  destruct h3 as [h3|].
  rewrite <- h3 in h5 at 1.
  contradict h5; auto.
  contradiction.
Qed.


  
Lemma lind_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l) (n:nat)
         (pflt:n < length l),
    lind pfdt l x pf = n <->
     (nth_lt l n pflt = x /\
     ~In x (firstn n l)).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros x h2 n h3.
  unfold lind.
  simpl.
  destruct (h1 x a) as [|h4]; subst.
  split; intro h4; subst; simpl.
  split; auto.
  destruct n as [|n]; auto.
  destruct h4 as [h4 h5].
  contradict h5.
  simpl.
  left; auto.
  split; intro h5.
  destruct n as [|n].
  discriminate.
  apply S_inj in h5.
  subst.
  split.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  apply lt_S_n in h3.
  destruct h2 as [|h2]; subst.
  pose proof (h4 eq_refl). contradiction.
  pose (lt_lind h1 _ _ h2) as h5.
  pose proof (lind_compat h1 l x h2) as h6.
  unfold lind in h6.
  pose proof (proof_irrelevance
                _ h2
                (neq_in_cons _ _ a h4 (or_intror h2))) as h7.
  rewrite h7 in h6.
  rewrite h6 at 1; auto.
  apply lt_S_n in h3; auto. 
  intro h5. 
  simpl in h5.
  destruct h5 as [|h5]; subst.
  pose proof (h4 eq_refl); contradiction.
  destruct h2 as [|h2]; subst.
  pose proof (h4 eq_refl); contradiction.
  apply lt_S_n in h3.
  specialize (ih _ h2 _ h3).
  unfold lind in ih.
  pose proof (proof_irrelevance _ h2
                                (neq_in_cons _ _ a h4 (or_intror h2))) as h7.
  rewrite <- h7 in ih at 1.
  pose proof (eq_refl (lind_occ h1 l 0 x h2)) as h8.
  rewrite ih in h8.
  destruct h8 as [h8 h9].
  contradiction.
  destruct h2 as [|h2]; subst.
  contradict h4; auto.
  destruct n as [|n]; subst.
  destruct h5 as [? h5]; subst.
  contradict h4; auto.
  apply lt_S_n in h3.
  destruct h5 as [h5 h6].
  f_equal.
  specialize (ih _ h2 n h3).
  let P := type of ih in match P with _ <-> ?hyp => assert (h7:hyp) end.
  rewrite nth_lt_compat.
  split.
  erewrite nth_indep.
  rewrite h5 at 1; auto.
  assumption.
  contradict h6.
  simpl.
  right; auto.
  rewrite <- ih in h7.
  rewrite <- h7.
  unfold lind.
  f_equal.
  apply proof_irrelevance.
Qed.



  
Definition lind_singularize {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (n:nat) (pf:n < length l) : nat :=
  let pf' := iff1 (list_singularize_in_iff pfdt _ _) (in_nth_lt l n pf) in
  lind pfdt (list_singularize pfdt l) (nth_lt l n pf) pf'.

Lemma lt_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat) (pf:n < length l),
    lind_singularize pfdt l n pf < length (list_singularize pfdt l).
  intros; apply lt_lind.
Qed.

Lemma lind_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (n:nat) (pf:n < length l),
    nth_lt l n pf =
    nth_lt (list_singularize pfdt l) (lind_singularize pfdt l n pf) (lt_lind_singularize pfdt _ _ _).
  intros T h1 l n h2.
  unfold lind_singularize.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
Qed.

Lemma lind_singularize_eq_nth_lt :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (i j:nat)
         (pfi:i < length l)
         (pfj:j < length l),
    lind_singularize pfdt l i pfi =
    lind_singularize pfdt l j pfj ->
    nth_lt l i pfi =
    nth_lt l j pfj.
  intros T hdt l i j h1 h2 h3.
  unfold lind_singularize in h3.
  apply lind_inj_iff in h3; auto.
Qed.


Lemma in_nth_lt_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (i:nat) (pflt:i < length l),
    Inn (list_to_set l)
    (nth_lt (list_singularize pfdt l)
       (lind_singularize pfdt l i pflt)
       (lt_lind_singularize pfdt l i pflt)).
  intros T h1 l i h2.
  rewrite (list_to_set_singularize_compat h1).
  rewrite <- list_to_set_in_iff.
  apply in_nth_lt.
Qed.



Lemma inj_ind_fun :
  forall {T:Type} (pfdt:type_eq_dec T)
         (m:nat) (l:list T) 
         (pflt:pred m < length l),
    Injective
      (fun i : (S (lind_singularize pfdt l (pred m) pflt)) @ =>
         let pflt' :=
             le_lt_trans (proj1_sig i)
                         (lind_singularize pfdt l (pred m) pflt)  _ (lt_n_Sm_le _ _ (proj2_sig i)) (lt_lind_singularize _ _ _ _) in
         nth_lt (list_singularize pfdt l) _ pflt').
  intros T h1 m l h2; red; intros x y h4.
  apply nth_lt_inj in h4.
  apply proj1_sig_injective in h4; auto.
  apply no_dup_list_singularize.
Qed.



Lemma lind_singularize_cons_in0 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (a:T) (pflt: 0 < S (length l))
         (pfi:In a l),
    let pfi' := iff1 (list_singularize_in_iff pfdt _ _) pfi in
    lind_singularize pfdt (a::l) 0 pflt =
    lind pfdt _ _  pfi'.
  intros T hdt l a h1 h2.
  unfold lind_singularize.
  gen0.
  rewrite nth_lt_0_hd'; simpl.
  rewrite list_singularize_in_cons; auto.
  intros; apply lind_functional; auto.
Qed.


Lemma lind_singularize_cons_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (i:nat)
         (a:T) (pf:i < S (length l))
         (pflt:i > 0),
    In a l ->
    let pf' := mono_pred_lt _ _  pf pflt  in
    lind_singularize pfdt (a::l) i pf =
    lind_singularize pfdt l (pred i) pf'.
  intros T hdt l i a h1 h2 h3 h4.
  unfold lind_singularize.
  gen0.
  rewrite list_singularize_in_cons; auto.
  erewrite nth_lt_cons'.
  intro h5.
  apply lind_functional.
  apply nth_lt_functional3.

  Grab Existential Variables. 
  assumption.
Qed.


Lemma lind_singularize_cons_nin0 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (a:T) (pflt: 0 < S (length l)),
    ~In a l ->
    lind_singularize pfdt (a::l) 0 pflt =
    0.
  intros T hdt l a h1 h2.
  unfold lind_singularize.
  gen0.
  rewrite list_singularize_nin; auto.
  rewrite nth_lt_0_hd'; simpl.
  intros h3.
  erewrite lind_functional3.
  rewrite lind_cons_same; auto.

  Grab Existential Variables.
  left; auto.
Qed.


Lemma lind_singularize_cons_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (i:nat)
         (a:T) (pf:i < S (length l))
         (pflt:i > 0)
         (pfni: ~In a l),
    let pf' := mono_pred_lt _ _  pf pflt  in
    lind_singularize pfdt (a::l) i pf =
    S (lind_singularize pfdt l _ pf').
  intros T hdt l i a h1 h2 h3.
  unfold lind_singularize.
  gen0.
  erewrite nth_lt_cons'.
  rewrite list_singularize_nin; auto.
  intro h4.
  pose proof h4 as h4'.
  destruct h4' as [h5 | h5].
  contradict h3.
  rewrite h5.
  apply in_nth_lt.
  erewrite lind_functional3.
  pose proof @lind_cons_neq.
  pose proof (lind_cons_neq hdt (list_singularize hdt l) a _ h5) as h6.
  hfirst h6.
  intro h7.
  pose proof h5 as h5'.
  rewrite h7 in h5'.
  rewrite <- list_singularize_in_iff in h5'; contradiction.
  specialize (h6 hf).
  rewrite h6.
  repeat f_equal.
  apply proof_irrelevance.
Qed.


Definition im_lind_singularize
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (n:nat)
           (pf:n <= length l) :
  Ensemble nat :=
  im_in_ens
    (fun i (pfi:Inn (seg n) i) =>
       let pfilt := seg_in _ _ pfi in
       let pfilt' := lt_le_trans _ _ _ pfilt pf in
       lind_singularize pfdt l i pfilt').


Lemma lt_im_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (i n:nat)
         (pf:n <= length l),
    Inn (im_lind_singularize pfdt l n pf) i ->
    i < length (list_singularize pfdt l).
  intros T hdt l i n h1 h2.
  destruct h2 as [y h2]; subst.
  simpl.
  destruct h2 as [h2].
  apply lt_lind_singularize.
Qed.

Lemma finite_im_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat)
         (pf:n <= length l),
    Finite (im_lind_singularize pfdt l n pf).
  intros T hdt l n h1.
  unfold im_lind_singularize.
  apply finite_im_in_ens.
  apply finite_seg.
Qed.

Definition map_seg_lind_singularize
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (n:nat)
           (pf:n <= length l) :
  list nat :=
  map_seg 
    (fun i (pfi:i < n) =>
       let pfilt' := lt_le_trans _ _ _ pfi pf in
       lind_singularize pfdt l i pfilt').

Lemma lt_map_seg_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (i n:nat)
         (pf:n <= length l),
    In i (map_seg_lind_singularize pfdt l n pf) ->
    i < length (list_singularize pfdt l).
  intros T hdt l i n h1 h2.
  unfold map_seg_lind_singularize in h2.
  rewrite in_map_seg_iff in h2.
  destruct h2 as [m [h2 h3]]; subst.
  apply lt_lind_singularize.
Qed.

Lemma length_map_seg_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat)
         (pf:n <= length l),
    length (map_seg_lind_singularize pfdt l n pf) = n.
  unfold map_seg_lind_singularize; intros; rewrite length_map_seg; auto.
Qed.

  
Definition inv_map_seg_lind_singularize 
      {T:Type} (pfdt:type_eq_dec T)
      (l:list T)
      (i n:nat)
      (pfle:n <= length l)
      (pfi:In i (map_seg_lind_singularize pfdt l n pfle)) :
  nat :=
  lind nat_eq_dec  _ _ pfi.

Lemma inv_map_seg_lind_singularize_functional :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (i i' n:nat)
         (pfle:n <= length l)
         (pfi: In i  (map_seg_lind_singularize pfdt l n pfle))
         (pfi':In i' (map_seg_lind_singularize pfdt l n pfle)),
    i = i' ->
    inv_map_seg_lind_singularize pfdt l i n pfle pfi =
    inv_map_seg_lind_singularize pfdt l i' n pfle pfi'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

   
Lemma lt_inv_map_seg_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (i n:nat)
         (pfle:n <= length l)
         (pfi:In i (map_seg_lind_singularize pfdt l n pfle)),
    inv_map_seg_lind_singularize pfdt l i n pfle pfi < n.
  intros T hdt l i n h1 h2.
  unfold inv_map_seg_lind_singularize, map_seg_lind_singularize.
  eapply lt_eq_trans.
  apply lt_lind.
  apply length_map_seg.
Qed.


Lemma inv_map_seg_lind_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (i n:nat)
         (pfle:n <= length l)
         (pfi:In i (map_seg_lind_singularize pfdt l n pfle)),
    let pflt := lt_inv_map_seg_lind_singularize pfdt l i n pfle pfi in
    let pflt' := lt_le_trans _ _ _ pflt pfle in 
    lind_singularize pfdt l
                     (inv_map_seg_lind_singularize pfdt l i n pfle pfi) pflt' =
    i. 
  intros T hdt l i n h1 h2 h3 h4.
  unfold lind_singularize, inv_map_seg_lind_singularize, map_seg_lind_singularize.
  pose proof h2 as h2'.
  unfold map_seg_lind_singularize in h2'.
  rewrite in_map_seg_iff in h2'.
  destruct h2' as [m [h5 h6]].
  rewrite <- h6.
  unfold lind_singularize. 
  apply lind_functional.
  rewrite (nth_lt_eq_iff hdt).
  exists (in_nth_lt _ _ _).
  exists 0.
  split.
  apply gt_lt.
  rewrite <- (count_occ_In hdt).
  apply in_nth_lt.
  symmetry.
  pose proof @lind_eq_iff.
  pose proof (lind_eq_iff nat_eq_dec (map_seg_lind_singularize hdt l n h1) i h2 (lind hdt l _ (in_nth_lt _ _ (lt_le_trans _ _ _ h5 h1)))) as hl.
  hfirst hl.
  rewrite length_map_seg_lind_singularize; auto.
  eapply le_lt_trans.
  eapply le_lind_nth_lt.
  assumption.
  specialize (hl hf).
  unfold map_seg_lind_singularize, lind_singularize in hl.
  rewrite hl.
  split.
  rewrite nth_lt_map_seg'.
  gen0.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intros h7.
  rewrite <- h6.
  unfold lind_singularize.
  apply lind_functional; auto.
  intro h7.
  rewrite in_firstn_iff in h7.
  destruct h7 as [r [h7 h8]].
  rewrite nth_lt_map_seg' in h8.
  rewrite <- h6 in h8.
  unfold lind_singularize in h8.
  rewrite lind_inj_iff in h8.
  revert h8.
  gen0. 
  intros h9 h10.
  pose proof (le_lind_nth_lt hdt l r h9) as h11.
  revert h11.
  generalize (in_nth_lt _ _  h9).
  rewrite h10.
  intros h11 h12.
  pose proof (proof_irrelevance
                _ h11 
                (in_nth_lt _ _ _)); subst.
  omega.

  Grab Existential Variables.
  rewrite length_map_seg.
  pose proof (le_lind_nth_lt hdt l m (lt_le_trans _ _ _ h5 h1)) as h7.
  unfold lind in h7.   
  apply lt_impl_le. 
  eapply le_lt_trans.
  apply h7.
  assumption.
Qed.


Lemma list_to_set_map_seg_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat)
         (pf:n <= length l),
    list_to_set (map_seg_lind_singularize pfdt l n pf) =
    im_lind_singularize pfdt l n pf.
  intros T pfdt l n h1.
  unfold im_lind_singularize, map_seg_lind_singularize.
  rewrite list_to_set_map_seg.
  reflexivity.
Qed.

Lemma in_im_lind_singularize_map_seg : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (i n:nat)
         (pf:n <= length l),
    Inn (im_lind_singularize pfdt l n pf) i ->
    In i (map_seg_lind_singularize pfdt l n pf).
  intros; rewrite list_to_set_in_iff, list_to_set_map_seg_lind_singularize; auto.
Qed.

Definition inv_im_lind_singularize 
      {T:Type} (pfdt:type_eq_dec T)
      (l:list T)
      (i n:nat)
      (pfle:n <= length l)
      (pfi:Inn (im_lind_singularize pfdt l n pfle) i) :=
  let pfi' := in_im_lind_singularize_map_seg pfdt l i n pfle pfi in
  inv_map_seg_lind_singularize pfdt l i n pfle pfi'.

Lemma lt_inv_im_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (i n:nat)
         (pfle:n <= length l)
         (pfi:Inn (im_lind_singularize pfdt l n pfle) i),
  let pfi' := in_im_lind_singularize_map_seg pfdt l i n pfle pfi in
  inv_map_seg_lind_singularize pfdt l i n pfle pfi' < n.
  unfold inv_map_seg_lind_singularize; intros.
  eapply lt_eq_trans.
  apply lt_lind.
  apply length_map_seg.
Qed.


Lemma inv_im_lind_singularize_compat : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (i n:nat)
         (pfle:n <= length l)
         (pfi:Inn (im_lind_singularize pfdt l n pfle) i),
    let pflt :=
        lt_le_trans _ _ _ (lt_inv_im_lind_singularize pfdt l i n pfle pfi) pfle in
    lind_singularize pfdt l (inv_im_lind_singularize pfdt l i n pfle pfi) pflt
    = i.
  intros T hdt l i n h1 h2 h3.
  unfold inv_im_lind_singularize.
  pose proof (inv_map_seg_lind_singularize_compat
                hdt l i n h1
                (in_im_lind_singularize_map_seg hdt l i n h1 h2)) as h4.
  simpl in h4.
  rewrite <- h4.
  apply f_equal_dep; auto.
Qed.


Definition im_im_lind_singularize
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T)
           (n:nat)
           (pfle:n <= length l) :
  Ensemble T :=
  im_in_ens
    (fun (i : nat) (pf : Inn (im_lind_singularize pfdt l n pfle) i) =>
       let pf' := lt_im_lind_singularize pfdt l i n pfle pf in
       nth_lt (list_singularize pfdt l) i pf').

Lemma finite_im_im_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l),
    Finite (im_im_lind_singularize pfdt l n pfle).
  unfold im_im_lind_singularize; intros.
  apply finite_im_in_ens.
  apply finite_im_lind_singularize.
Qed.



Definition map_in_dep_map_seg_lind_singularize
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T)
           (n:nat)
           (pfle:n <= length l) :
  list T :=
  map_in_dep
    (fun (i : nat) (pf : In i (map_seg_lind_singularize pfdt l n pfle)) =>
       let pf' := lt_map_seg_lind_singularize pfdt l i n pfle pf in
       nth_lt (list_singularize pfdt l) i pf').

Lemma length_map_in_dep_map_seg_lind_singularize : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l),
    length (map_in_dep_map_seg_lind_singularize pfdt l n pfle) = n.
  unfold map_in_dep_map_seg_lind_singularize.
  intros; rewrite length_map_in_dep.
  apply length_map_seg_lind_singularize.
Qed.

Lemma in_map_in_dep_map_seg_lind_singularize' : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T),
    In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle) ->
    In x l.
  intros T hdt l n h1 x h2.
  unfold map_in_dep_map_seg_lind_singularize in h2.
  rewrite in_map_in_dep_iff in h2.
  destruct h2 as [c [h2 h3]]; subst.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
Qed.
    
    

Lemma list_to_set_map_in_dep_map_seg_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l),
    list_to_set (map_in_dep_map_seg_lind_singularize pfdt l n pfle) =
    im_im_lind_singularize pfdt l n pfle.
  intros T hdt l n h1.
  unfold map_in_dep_map_seg_lind_singularize, im_im_lind_singularize.
  rewrite list_to_set_map_in_dep.
  pose proof (list_to_set_map_seg_lind_singularize hdt l n h1) as h2.
  rewrite (im_in_ens_transfer _ h2).
  f_equal.
  apply functional_extensionality_dep.
  intro i.
  apply functional_extensionality.
  intro h3.
  rewrite fun_in_ens_transfer_compat.
  simpl.
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma in_map_in_dep_map_seg_lind_singularize : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T),
    Inn (im_im_lind_singularize pfdt l n pfle) x ->
    In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle).
  intros T hdt l n pfle x h1.
  rewrite <- list_to_set_map_in_dep_map_seg_lind_singularize in h1.
  rewrite <- list_to_set_in_iff in h1; auto.
Qed.


Definition inv_map_in_dep_map_seg_lind_singularize
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T)
           (n:nat)
           (pfle:n <= length l)
           (x:T)
           (pf:In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle)) :
  nat :=
  let pflt := lt_eq_trans
                _ _ _ 
                (lt_eq_trans _ _ _ (lt_lind pfdt _ _ _)
                             (length_map_in_dep_map_seg_lind_singularize
                                pfdt l n pfle))
                (eq_sym (length_map_seg_lind_singularize pfdt l n pfle)) in
  nth_lt (map_seg_lind_singularize pfdt l n pfle)
         (lind pfdt _ x pf) pflt.

Lemma lt_inv_map_in_dep_map_seg_lind_singularize : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T)
         (pf:In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle)),
    inv_map_in_dep_map_seg_lind_singularize pfdt l n pfle x pf <
    length (list_singularize pfdt l).
  intros T hdt l n h1 x h2.
  unfold inv_map_in_dep_map_seg_lind_singularize.
  unfold map_seg_lind_singularize.
  rewrite nth_lt_map_seg'.
  unfold lind_singularize.
  apply lt_lind.
Qed.


Lemma inv_map_in_dep_map_seg_lind_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T)
         (pf:In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle))
         (pflt:inv_map_in_dep_map_seg_lind_singularize pfdt l n pfle x pf <
               length (list_singularize pfdt l)),
    nth_lt _ _ pflt = x.
  intros T hdt l n hle x h1 h2.
  rewrite nth_lt_eq_iff.
  ex_goal.
  rewrite <- list_singularize_in_iff.
  eapply in_map_in_dep_map_seg_lind_singularize'.
  apply h1.
  exists hex.
  exists 0.
  split.
  apply gt_lt.
  apply (count_occ_In hdt).
  assumption.
  symmetry.
  unfold inv_map_in_dep_map_seg_lind_singularize, map_seg_lind_singularize. 
  rewrite nth_lt_map_seg'.
  unfold lind_singularize.
  fold (lind hdt _ _ hex).
  apply lind_functional.
  rewrite nth_lt_eq_iff.
  ex_goal.
  rewrite <- list_singularize_in_iff in hex; auto.
  exists hex0.
  exists 0.
  split.
  apply gt_lt.
  apply (count_occ_In hdt); auto.
  unfold lind.
  unfold map_in_dep_map_seg_lind_singularize.
  symmetry.
  change
    (lind  hdt (map_in_dep
        (fun (i : nat) (pf : In i (map_seg_lind_singularize hdt l n hle)) =>
         nth_lt (list_singularize hdt l) i
           (lt_map_seg_lind_singularize hdt l i n hle pf))) x h1 =
     lind hdt l x hex0).
  rewrite lind_eq_iff.
  split.
  rewrite nth_lt_map_in_dep'.
  rewrite nth_lt_eq_iff.
  exists hex, 0.
  split.
  apply gt_lt.
  apply (count_occ_In hdt); auto.
  unfold map_seg_lind_singularize.
  rewrite nth_lt_map_seg'.
  unfold lind_singularize.
  apply lind_occ_functional; auto.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
  intro h3.
  rewrite in_firstn_iff in h3.
  destruct h3 as [m [h3 h4]]. 
  rewrite nth_lt_map_in_dep' in h4.
  rewrite nth_lt_eq_iff in h4.
  destruct h4 as [h4 [occ [h5 h6]]].
  unfold map_seg_lind_singularize in h6.
  rewrite nth_lt_map_seg' in h6.
  unfold lind_singularize in h6.
  apply lind_occ_inj in h6.
  symmetry in h6.
  revert h6.
  gen0.
  intros h6 h7; subst.
  pose proof (le_lind_nth_lt hdt l m h6) as h7.
  pose proof (proof_irrelevance
                _ hex0
                (in_nth_lt _ _ _ )); subst.
  omega.
 
  Grab Existential Variables.
  rewrite length_map_in_dep, length_map_seg_lind_singularize.
  unfold map_in_dep_map_seg_lind_singularize in h1.
  pose proof h1 as h1'.
  rewrite in_map_in_dep_iff in h1'.
  destruct h1' as [c [h3 h4]]; subst.
  unfold map_seg_lind_singularize in h3.
  pose proof h3 as h3'.
  rewrite in_map_seg_iff in h3'.
  destruct h3' as [m [h4 h6]]; subst.
  unfold lind_singularize.
  generalize hex0.
  unfold lind_singularize.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro h7.
  eapply lt_impl_le.
  eapply le_lt_trans.
  erewrite lind_functional3.
  apply le_lind_nth_lt.
  assumption.
  rewrite length_map_in_dep, length_map_seg_lind_singularize.

  unfold map_in_dep_map_seg_lind_singularize in h1.
  pose proof h1 as h1'.
  rewrite in_map_in_dep_iff in h1'.
  destruct h1' as [c [h3 h4]]; subst.
  unfold map_seg_lind_singularize in h3.
  pose proof h3 as h3'.
  rewrite in_map_seg_iff in h3'.
  destruct h3' as [m [h4 h6]]; subst.
  unfold lind_singularize.
  generalize hex0.
  unfold lind_singularize.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro h7.
  eapply le_lt_trans.
  erewrite lind_functional3.
  apply le_lind_nth_lt.
  assumption.
Qed.



Lemma in_inv_map_in_dep_map_seg_lind_singularize : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T)
         (pf:In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle)),
    In (inv_map_in_dep_map_seg_lind_singularize pfdt l n pfle x pf)
       (map_seg_lind_singularize pfdt l n pfle).
  intros T hdt l n h1 x h2.
  unfold map_seg_lind_singularize.
  rewrite in_map_seg_iff.
  pose proof h2 as h2'.
  unfold map_in_dep_map_seg_lind_singularize in h2'.
  rewrite in_map_in_dep_iff in h2'.
  destruct h2' as [i [h3 h4]].
  unfold map_seg_lind_singularize in h3.
  pose proof h3 as h5.
  rewrite in_map_seg_iff in h5; subst.
  destruct h5 as [m [h5 h6]].
  exists m, h5.
  rewrite h6.
  subst.
  unfold lind_singularize.
  eapply nth_lt_inj.
  eapply no_dup_list_singularize.
  rewrite inv_map_in_dep_map_seg_lind_singularize_compat.
  apply nth_lt_functional2; auto.

  Grab Existential Variables.
  apply lt_inv_map_in_dep_map_seg_lind_singularize.
  apply lt_lind.
Qed.


Definition inv_inv_map_in_dep_map_seg_lind_singularize
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T)
           (n:nat)
           (pfle:n <= length l)
           (x:T)
           (pf:In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle)) :
  nat :=
  let pf' := in_inv_map_in_dep_map_seg_lind_singularize pfdt l n pfle x pf in
  inv_map_seg_lind_singularize pfdt l _ n pfle pf'.

Lemma lt_inv_inv_map_in_dep_map_seg_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T)
         (pf:In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle)),
    inv_inv_map_in_dep_map_seg_lind_singularize pfdt l n pfle x pf <
    n.
  unfold inv_inv_map_in_dep_map_seg_lind_singularize; intros.
  apply lt_inv_map_seg_lind_singularize.
Qed.

Definition inv_inv_im_im_lind_singularize
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T)
           (n:nat)
           (pfle:n <= length l)
           (x:T)
           (pf:Inn (im_im_lind_singularize pfdt l n pfle) x)
: nat :=
      let pf' := in_map_in_dep_map_seg_lind_singularize pfdt l n pfle  x pf in
      inv_inv_map_in_dep_map_seg_lind_singularize pfdt l n pfle x pf'.

Lemma lt_inv_inv_im_im_lind_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T)
         (pf:Inn (im_im_lind_singularize pfdt l n pfle) x),
    inv_inv_im_im_lind_singularize pfdt l n pfle x pf <
    n.
  unfold inv_inv_im_im_lind_singularize; intros;
  apply lt_inv_inv_map_in_dep_map_seg_lind_singularize.
Qed.


Lemma in_im_im_lind_singularize_inv :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (i n:nat)
         (pfle:n <= length l) (pfin:i < n)
         (pfi:Inn (im_im_lind_singularize pfdt l n pfle)
                    (nth_lt l i (lt_le_trans i n (length l) pfin pfle))),
    Inn (im_lind_singularize pfdt l n pfle)
        (lind_singularize pfdt l i (lt_le_trans i n (length l) pfin pfle)).
  intros T hdt l i n hle hin hi.
  unfold im_lind_singularize.
  econstructor.
  apply f_equal_dep.
  reflexivity.

  Grab Existential Variables.
  constructor; auto.
Qed.


(*Search below for [inv_inv_map_in_dep_map_seg_lind_singularize_compat]*)


Lemma nth_lt_tl :
  forall {T} (l:list T) i (pfi:i < length (tl l))
         (pfn:l <> nil),
    let pfi0 := lt_eq_trans _ _ _ pfi (length_tl l) in
    let pfi' := lt_eq_trans
                  _ _ _
                  (lt_n_S _ _ pfi0)
                  (eq_sym (S_pred _ _ (not_nil_lt l pfn))) in
    nth_lt (tl l) i pfi =
    nth_lt l (S i) pfi'.
  intros T l i hi hn; simpl.
  destruct l as [|a l]; simpl.
  contradict hn; auto.
  rewrite nth_lt_compat.
  apply nth_indep.
  simpl in hi; auto.
Qed.


Lemma tl_map_seg :
  forall {U:Type} {n:nat} (pfdu:type_eq_dec U) (f:seg_fun n U),
    tl (map_seg f) =
    map_seq (seg_fun_tl f).
  intros U n hdu f.  
  symmetry.
  rewrite map_seq_eq_iff.
  ex_goal.
  rewrite length_tl, length_map_seg; auto.
  exists hex.
  intro m; simpl; intro h1.
  unfold seg_fun_tl, seq_fun_tl.
  rewrite seg_fun_to_seq_compat.
  symmetry. 
  rewrite (nth_lt_eq_iff hdu).
  ex_goal.
  rewrite in_tl_iff'.
  exists m.
  ex_goal.
  rewrite length_map_seg; omega.
  exists hex0.
  rewrite nth_lt_map_seg'.
  f_equal; apply proof_irrelevance.
  auto.
  exists hex0.
  pose proof h1 as h1'.
  rewrite hex in h1'. 
  pose proof (ex_lind_occ_nth_lt hdu _ _ h1') as h2.
  destruct h2 as [occ [h2 h3]].
  exists occ.
  split.
  pose proof (lt_length _ _ h1') as h4.
  pose proof (nth_lt_tl _ _ h1') as h5.
  hfirst h5.
  rewrite map_seg_eq_nil_iff.
  omega.
  rewrite (nth_lt_tl _ _ h1' hf) , nth_lt_map_seg' in h2 at 1.
  specialize (h5 hf); simpl in h5.
  erewrite f_equal_dep.
  apply h2.
  reflexivity.
  rewrite <- h3.
  apply lind_occ_functional; auto.
  rewrite nth_lt_tl, nth_lt_map_seg' at 1.
  f_equal; apply proof_irrelevance.

  Grab Existential Variables.
  rewrite map_seg_eq_nil_iff.
  omega.
Qed.


Lemma nth_lt_lind_occ_same :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l)
         (occ occ':nat),
    let i := lind_occ pfdt l  occ  x pf in
    let i' := lind_occ pfdt l occ' x pf in 
    nth_lt l i  (lt_lind_occ _ _ _ _ _) =
    nth_lt l i' (lt_lind_occ _ _ _ _ _).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 occ occ'. 
  destruct (h1 a x), (h1 x a); try contradiction; simpl.
  clear e. subst.
  destruct occ as [|occ], occ' as [|occ']; simpl; auto.
  
  destruct (h1 a a) as [ha | ha].
  destruct (in_dec h1 a l) as [h5 | h5].
  erewrite nth_indep.
  rewrite <- (nth_lt_compat _ _ (lt_lind_occ _ _ _ _ _)) at 1.
  rewrite lind_occ_compat; auto.
  apply lt_lind_occ.
  reflexivity.
  destruct (in_dec h1 a l) as [h5 | h5].
  contradict ha; auto.
  reflexivity.
  
  destruct (in_dec h1 a l) as [h5 | h5].
  erewrite nth_indep.
  rewrite <- (nth_lt_compat _ _ (lt_lind_occ _ _ _ _ _)) at 1.
  rewrite lind_occ_compat; auto.
  apply lt_lind_occ.
  reflexivity.
  destruct (in_dec h1 a l) as [h5 | h5]; simpl.
  erewrite nth_indep.
  rewrite <- (nth_lt_compat _ _ (lt_lind_occ _ _ _ _ _)) at 1.
  symmetry.
  erewrite nth_indep.
  rewrite <- (nth_lt_compat _ _ (lt_lind_occ _ _ _ _ _)) at 1.
  apply ih; auto.
  apply lt_lind_occ.
  apply lt_lind_occ.
  reflexivity.
  subst.
  contradict n; auto.
  subst.
  contradict n; auto.
  pose proof (neq_in_cons _ _ _ n0 h2) as h5.
  erewrite nth_indep.
  rewrite <- (nth_lt_compat _ _ (lt_lind_occ _ _ _ _ _)) at 1.
  symmetry.
  erewrite nth_indep.
  rewrite <- (nth_lt_compat _ _ (lt_lind_occ _ _ _ _ _)) at 1.
  apply ih; auto.
  apply lt_lind_occ.
  apply lt_lind_occ.
Qed.




Lemma nth_lt_lind_lind_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (occ:nat) (x:T) (pf:In x l),
    nth_lt l (lind_occ pfdt l occ  x pf) (lt_lind_occ _ _ _ _ _) =
    nth_lt l (lind pfdt l x pf) (lt_lind _ _ _ _).
  unfold lind; intros; erewrite nth_lt_lind_occ_same; apply nth_lt_functional3.
Qed.

Lemma nth_lt_lind_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (occ:nat) (x:T) (pf:In x l),
    nth_lt l (lind_occ pfdt l occ  x pf) (lt_lind_occ _ _ _ _ _) =
    x.
  intros; rewrite nth_lt_lind_lind_occ, lind_compat; auto.
Qed.

Lemma lind_occ_map_inj :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:T->U) 
         (l:list T) (x:T) (pfx:In x l)
         (pffx:In (f x) (map f l))
         (occ:nat),
    Injective f ->
    lind_occ pfdu (map f l) occ (f x) pffx =
    lind_occ pfdt l occ x pfx.
  intros T U h1 h2 f l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h3 h4 occ h5.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  apply h5 in hl; subst.
  erewrite eq_dec_same at 1.
  destruct occ as [|occ]; auto.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  f_equal.
  apply ih; auto.
  pose proof hl as hl'.
  rewrite in_map_iff in hl'.
  destruct hl' as [a' [h6 h7]].
  apply h5 in h6; subst.
  contradiction.
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  contradict hr; apply in_map; auto.
  reflexivity.
  assumption.
  destruct h4 as [h4 | h4].
  pose proof h4 as h4'.
  symmetry in h4'; contradiction.
  destruct h3 as [|h3]; subst.
  contradict hr; auto.
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  subst.
  contradict hr; auto.
  f_equal.
  apply ih; auto.
Qed.


Lemma lind_occ_map_in_dep_inj :
  forall {T U:Type} {l:list T}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (f:fun_in_dep l U) 
         (x:T) (pfx:In x l)
         (pffx:In (f x pfx) (map_in_dep f))
         (occ:nat),
    inj_dep f ->
    lind_occ pfdu (map_in_dep f) occ (f x pfx) pffx =
    lind_occ pfdt l occ x pfx.
  intros T U l h1 h2.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros f x h3 h4 occ h5.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  apply h5 in hl; subst.
  erewrite eq_dec_same at 1.
  destruct occ as [|occ]; auto.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  f_equal.
  specialize (ih (fun_from_cons f) a hl').
  hfirst ih.
  apply in_map_in_dep.
  specialize (ih hf occ).
  hfirst ih.
  apply inj_from_cons; auto.
  specialize (ih hf0).
  unfold fun_from_cons in ih at 2.
  rewrite <- ih; apply f_equal_dep; f_equal; apply proof_irrelevance.
  pose proof hl as hl'.
  rewrite in_map_in_dep_iff in hl'.
  destruct hl' as [a' [h6 h7]].
  unfold fun_from_cons in h7.
  red in h5.
  apply h5 in h7; subst; contradiction. 
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  contradict hr.
  rewrite in_map_in_dep_iff.
  exists a, hl'.
  unfold fun_from_cons; f_equal; apply proof_irrelevance.
  reflexivity.
  assumption.
  destruct h4 as [h4 | h4].
  pose proof h4 as h4'.
  symmetry in h4'; contradiction.
  destruct h3 as [|h3]; subst. 
  contradict hr; f_equal; apply proof_irrelevance.
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  subst.
  contradict hr; f_equal; apply proof_irrelevance.
  f_equal.
  specialize (ih (fun_from_cons f) x h3 (in_map_in_dep _ _ _) occ).
  hfirst ih. 
  apply inj_from_cons; auto. 
  specialize (ih hf).
  unfold fun_from_cons in ih at 2; auto.
  symmetry.
  erewrite f_equal_dep; auto.
  rewrite <- ih at 1.
  apply f_equal_dep; f_equal; apply proof_irrelevance.
Qed.

  
Lemma lind_map_seg_fun0 :
  forall {U:Type} {n:nat} (pfdu:type_eq_dec U)
         (f:seg_fun n U)
         (pflt: 0 < n)
         (pfi:In (f 0 pflt) (map_seg f)),
    lind pfdu (map_seg f) (f 0 pflt) pfi = 0.
  intros U n hdu f h1 h2.
  rewrite lind_eq_iff.
  split.
  rewrite nth_lt_map_seg'.
  f_equal.
  apply proof_irrelevance.
  simpl.
  intro; contradiction.

  Grab Existential Variables.
  rewrite length_map_seg; auto.
Qed.

  
Lemma nat_fun_lind_map_seg :
  forall {U:Type} {n:nat} (pfdu:type_eq_dec U)
         (f:seg_fun n U)
         (i:nat)
         (pfi:i < n)
         (pffi:In (f i pfi) (map_seg f))
         (pflt:lind pfdu (map_seg f) (f i pfi) pffi < n),
    f (lind pfdu (map_seg f) (f i pfi) pffi) pflt = f i pfi.
  intros U n.
  induction n as [|n ih]; simpl.
  intros; contradiction.
  intros hdu f i h2 h3 h4. 
  destruct i as [|i]; simpl.
  pose proof h3 as h3'.
  apply in_app_or in h3'.
  destruct h3' as [h5 | h5].
  generalize h4.
  erewrite lind_functional3.
  erewrite lind_app_in1.
  intro h6.
  destruct (zerop n) as [|h7]; subst.
  pose proof h6 as h6'.
  apply lt1 in h6'.
  gen0.
  rewrite h6'; intros; f_equal; apply proof_irrelevance.
  pose proof (lind_map_seg_fun0 hdu (seg_fun_from_S f) h7) as h8. 
  hfirst h8.
  apply in_map_seg.
  specialize (h8 hf).
  unfold seg_fun_from_S in h8.
  unfold seg_fun_from_S.
  pose proof (f_equal_dep f _ 0 h6 h2) as h9.
  hfirst h9.
  rewrite <- h8.
  apply lind_functional.
  f_equal.
  apply proof_irrelevance.
  specialize (h9 hf0).
  rewrite h9 at 1.
  reflexivity.
  simpl in h5. 
  destruct h5 as [h5 | h5].
  destruct (zerop n) as [|h6]; subst.
  gen0. 
  erewrite lind_functional.
  simpl.
  rewrite lind_sing.
  intros; f_equal; apply proof_irrelevance.
  reflexivity.
  gen0.
  erewrite lind_functional.
  erewrite lind_app_in1.
  erewrite lind_map_seg_fun0.
  intros; f_equal; apply proof_irrelevance.
  unfold seg_fun_from_S.
  f_equal.
  apply proof_irrelevance.
  contradiction.
    
  pose proof h2 as h2'.
  apply lt_S_n in h2'.
  pose proof h3 as h3'.
  apply in_app_or in h3'. 
  pose proof h2 as hn.
  red in hn. 
  apply le_S_n in hn. 
  apply le_lt_eq_dec in hn.
  destruct hn as [hn | hn].
  destruct h3' as [h5 | h5]. 
  gen0.
  erewrite lind_functional. 
  erewrite lind_app_in1.
  intro h6.
  specialize (ih hdu (seg_fun_from_S f) (S i) hn).
  hfirst ih. 
  rewrite seg_fun_from_S_compat; auto.
  rewrite in_map_seg_iff.
  exists (S i), hn.
  rewrite seg_fun_from_S_compat; auto.
  specialize (ih hf).
  hfirst ih.
  eapply lt_eq_trans.
  apply lt_lind.
  rewrite length_map_seg; auto.
  specialize (ih hf0).
  do 2 rewrite seg_fun_from_S_compat in ih.
  erewrite f_equal_dep in ih.
  rewrite ih at 1.
  f_equal.
  apply proof_irrelevance.
  apply lind_functional.
  reflexivity.
  rewrite seg_fun_from_S_compat.
  f_equal.
  apply proof_irrelevance.
  destruct h5 as [h5|].
  rewrite <- h5.
  gen0.
  generalize h3.
  rewrite <- h5. 
  intros h6.
  erewrite lind_functional.
  rewrite lind_app_in1.
  intro h7.
  rewrite  h5.
  specialize (ih hdu (seg_fun_from_S f) (S i) hn).
  hfirst ih.
  apply in_map_seg.
  specialize (ih hf).
  hfirst ih.
  eapply lt_eq_trans.
  apply lt_lind.
  apply length_map_seg.
  specialize (ih hf0).
  do 2 rewrite seg_fun_from_S_compat in ih.
  symmetry.
  erewrite f_equal_dep.
  rewrite <- ih at 1.
  apply f_equal_dep.
  apply lind_functional.
  reflexivity.
  reflexivity.
  rewrite seg_fun_from_S_compat.
  rewrite h5.
  f_equal.
  apply proof_irrelevance.
  contradiction. 
  subst.
  destruct (in_dec hdu (f (S i) h2) (map_seg (seg_fun_from_S f))) as [hi | hi].
  pose proof hi as hj.
  rewrite in_map_seg_iff in hj.
  destruct hj as [j [hj h6]].
  specialize (ih hdu (seg_fun_from_S f) j hj).
  hfirst ih.
  apply in_map_seg.
  specialize (ih hf).
  hfirst ih.
  eapply lt_eq_trans.
  apply lt_lind.
  apply length_map_seg.
  specialize (ih hf0).
  rewrite seg_fun_from_S_compat in ih.
  revert ih.
  Gen0.
  generalize hf.
  rewrite h6. 
  intros h7 h8 h9.
  rewrite <- h9.
  
  apply f_equal_dep.
  erewrite lind_functional.
  rewrite lind_app_in1.
  reflexivity.
  reflexivity.

  gen0.
  erewrite lind_functional.
  rewrite lind_app_in2, lind_sing, plus_0_r, length_map_seg.
  intros; f_equal; apply proof_irrelevance.
  apply hi.
  reflexivity.
  
  Grab Existential Variables.
  constructor.
  f_equal; apply proof_irrelevance.
  apply in_map_seg.
  apply in_map_seg.
  assumption.
  apply in_map_seg.
  apply in_app_r.
  constructor.
  assumption.
  assumption.
Qed.


Lemma nth_lt_lS :
  forall (l:list nat)
         i (pfi:i < (length (lS l))),
    nth_lt (lS l) i pfi =
    S (nth_lt l i (lt_eq_trans _ _ _ pfi (length_lS _))).
  intros l i hi.
  gtermr.
  redterm0 y.
  rewrite (nth_lt_eq_iff nat_eq_dec).
  ex_goal.
  rewrite in_lS_iff.
  exists c; split.
  unfold c; apply in_nth_lt.
  reflexivity. 
  exists hex.
  pose proof hi as hi'.
  rewrite length_lS in hi'.
  pose proof (ex_lind_occ_nth_lt nat_eq_dec _ _ hi') as h1.
  destruct h1 as [occ [h1 h2]].
  exists occ; split; auto.
  rewrite count_occ_lS; simpl.
  erewrite nth_lt_functional3.
  apply h1.
  auto with arith.
  rewrite <- h2.
  unfold lS.
  erewrite lind_occ_map_inj.
  apply lind_occ_functional; auto.
  apply nth_lt_functional3.
  red; intros ? ? h3. apply S_inj in h3; auto.

  Grab Existential Variables.
  apply in_nth_lt.
Qed.


  

Lemma nat_fun_lind_map_seg' :
  forall {U:Type} (pfdu:type_eq_dec U)
         (f:nat->U)
         (i n:nat)
         (pfi:i < n)
         (pffi:In (f i) (map_seg (fun_to_seg f n))),
    f (lind pfdu (map_seg (fun_to_seg f n)) (f i) pffi) = f i.
  intros U hdu f i n h1 h2.
  pose proof
       (nat_fun_lind_map_seg
          hdu (fun_to_seg f n) i h1
       (in_map_seg _ _ h1)) as h3.
  hfirst h3.
  eapply lt_eq_trans.
  apply lt_lind.
  apply length_map_seg.
  specialize (h3 hf).
  do 2 rewrite fun_to_seg_compat in h3.
  rewrite <- h3.
  f_equal.
  apply lind_functional.
  rewrite fun_to_seg_compat; auto.
Qed.


Lemma lt_lind_count_occ_S :
    forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
           (pf:count_occ pfdt l x = 1)
           (n:nat),
      let pfin := count_occ_eq_S_in pfdt l x  _ pf  in
      lind pfdt l x pfin =
      lind_occ pfdt l (S n) x pfin.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros x h2 n. 
  destruct (h1 x a) as [|h3]; subst.
  pose proof h2 as h2'. 
  terml h2'.
  assert (h4:x = S (count_occ h1 l a)).
  unfold x.
  destruct (h1 a a); auto.
  contradict n0; auto.
  destruct in_dec as [h3 | h3].
  
  fold x in h2'.
  rewrite h2' in h4.
  apply S_inj in h4.
  pose proof h3 as h3'. rewrite  (count_occ_In h1) in h3'.
  omega.
  unfold lind.
  simpl.
  assert (h5:h1 a a = left eq_refl).
    destruct (h1 a a); auto.
    f_equal. apply proof_irrelevance.
    contradict n0; auto.
  rewrite h5; auto.
  pose proof (neq_sym _ _ h3) as h3'.
  assert (h5:h1 a x = right h3').
    destruct (h1 a x).
    contradiction.
    f_equal; apply proof_irrelevance.
  pose proof h2 as h2'. rewrite h5 in h2'.
  specialize ( ih x h2').
  simpl in ih.
  unfold lind.
  simpl.
  assert (h5':h1 x a = right h3).
    destruct (h1 x a).
    contradiction.
    f_equal; apply proof_irrelevance.
  rewrite h5'.
  f_equal.
  specialize (ih n).
  match goal with |- lind_occ _ _ _ _ ?pf = _ => generalize pf  end.
  intro h6.
  fold (lind h1 l x h6).
  erewrite lind_functional.
  rewrite ih at 1.
  f_equal. apply proof_irrelevance.
  reflexivity.
Qed.


Lemma lind_occ_pred_count_occ_lt_length_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfin:In x l),
    lind_occ pfdt l (pred (count_occ pfdt l x)) x pfin < length l.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2.
  rewrite (switch_eq_dec_eq h1 a x).
  destruct (h1 x a) as [|h3]; subst.
  unfold switch_eq_dec.
  simpl.
  destruct (in_dec h1 a l) as [h3 | h3].
  pose proof h3 as h3'.
  rewrite (count_occ_In h1) in h3'.
  rewrite (S_pred _ _ h3').
  apply lt_n_S; auto.
  rewrite (count_occ_In h1) in h3.
  rewrite (not_lt0_iff (count_occ h1 l a)) in h3.
  rewrite h3; auto with arith.
  apply lt_n_S.
  unfold switch_eq_dec.
  apply ih.
Qed.

Lemma lind_occ_pred_nth_lt_pred :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (pflt:pred (length l) < length l)
           (pfi:In (nth_lt l _ pflt) l),
    lind_occ pfdt l (pred (count_occ pfdt l (nth_lt l (pred (length l))
                                                    pflt)))
             (nth_lt _ _ pflt) pfi =
    pred (length l).
  intros T hdt l.
  induction l as [|a l ih].
  intro; contradiction.
  destruct (nil_dec' l) as [|h2]; subst.
  intros ? ?. 
  simpl; do 2 rewrite (eq_dec_same hdt a) at 1; simpl; auto. 
  intros h1 h3. 
  pose proof h1 as h4.
  simpl in h4.
  pose proof (mono_pred_lt _ _ h4 (not_nil_lt _ h2)) as h5.
  simpl in h5.
  pose proof h3 as h3'.
  simpl in h3'.
  rewrite (S_pred _ _ (not_nil_lt _ h2)) in h3'. 
  erewrite nth_indep in h3'.
  erewrite <- nth_lt_compat in h3'. 
  destruct h3' as [h6 | h6].
  simpl.
  specialize (ih h5 (in_nth_lt _ _ _)).
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  rewrite (S_pred _ _ (not_nil_lt _ h2)) in hl.
  rewrite (S_pred _ _ (not_nil_lt _ h2)).
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr']; simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl0 | hr0]; simpl.
  pose proof hl0 as h7.
  rewrite (count_occ_In hdt) in h7.
  rewrite (S_pred _ _ h7).
  f_equal.
  rewrite <- ih.
  apply lind_occ_functional; auto.
  repeat f_equal.
  rewrite nth_lt_compat.
  apply nth_indep.
  eapply lt_pred_n.
  apply (not_nil_lt _ h2).
  rewrite nth_lt_compat.
  apply nth_indep.
  eapply lt_pred_n.
  apply (not_nil_lt _ h2).
  contradict hr0.
  erewrite nth_indep.
  erewrite <- nth_lt_compat at 1.
  apply in_nth_lt.
  eapply lt_pred_n.
  apply (not_nil_lt _ h2).
  contradict hr'.
  rewrite h6 at 1.
  erewrite nth_indep.
  erewrite <- nth_lt_compat at 1.
  reflexivity.
  eapply lt_pred_n.
  apply (not_nil_lt _ h2).
  pose proof hr as h7.
  rewrite (S_pred _ _ (not_nil_lt _ h2)) in h7.
  contradict h7.
  rewrite h6 at 2.
  rewrite nth_lt_compat.
  apply nth_indep.
  eapply lt_pred_n.
  apply (not_nil_lt _ h2).
  simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  rewrite (S_pred _ _ (not_nil_lt _ h2)) in hl.
  rewrite (S_pred _ _ (not_nil_lt _ h2)).
  rewrite hl.
  rewrite (eq_dec_same hdt a) at 1.
  simpl. 
  pose proof h6 as h7.
  rewrite nth_lt_compat in h7.
  erewrite nth_indep in h7.
  rewrite hl in h7 at 1.
  pose proof h7 as h8.
  rewrite (count_occ_In hdt) in h8.
  rewrite (S_pred _ _ h8).

  lr_match_goal; fold hlr; destruct hlr as [hl0 | hr0]; simpl.
  f_equal.
  erewrite <- ih.
  apply lind_occ_functional; auto.
  do 2 f_equal.
  rewrite <- hl at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  eapply lt_pred_n.
  eapply not_nil_lt.
  assumption.
  rewrite <- hl at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  eapply lt_pred_n.
  eapply not_nil_lt.
  assumption.
  contradiction.
  eapply lt_pred_n.
  eapply not_nil_lt.
  assumption.

  pose proof hr as h7.
  rewrite (S_pred _ _ (not_nil_lt _ h2)) in h7.
  generalize (neq_in_cons _ _ _ hr h3).
  rewrite (S_pred _ _ (not_nil_lt _ h2)).
  intro h8.
  lr_if_goal; fold hlr; destruct hlr as [hl0 | hr0]; simpl.
  contradict h7; auto.
  f_equal.
  erewrite <- ih.
  apply lind_occ_functional; auto.
  repeat f_equal.
  rewrite nth_lt_compat.
  apply nth_indep.
  eapply lt_pred_n.
  apply not_nil_lt.
  assumption.
  rewrite nth_lt_compat.
  apply nth_indep.
  eapply lt_pred_n.
  apply not_nil_lt.
  assumption.
  eapply lt_pred_n.
  apply not_nil_lt.
  assumption.

  Grab Existential Variables.
  apply in_nth_lt.
  eapply lt_pred_n.
  apply not_nil_lt; auto.
  apply in_nth_lt.
  eapply lt_pred_n.
  apply not_nil_lt; auto.
  eapply lt_pred_n.
  apply not_nil_lt; auto.
  eapply lt_pred_n.
  apply not_nil_lt; auto.
Qed.


Fixpoint dup_witness {T:Type}  {l:list T} (pfdt:type_eq_dec T) :
  dup l -> nat*nat :=
  match l with
    | nil => fun pf => (False_rect _ (not_dup_nil _ pf))
    | a :: l' => fun (pf:dup (a::l')) =>
                  match (in_dec pfdt a l') with
                    | left pfia => (0, (S (lind pfdt _ _  pfia)))
                    | right pfnia => let pfdc := dup_cons pfdt _ _ pf in
                                     let pfd := dec_not_l pfdc pfnia in
                                     let p := dup_witness pfdt pfd in
                                     (S (fst p), S (snd p))
                  end
  end.


Lemma lt_dup_witness1 :
  forall {T:Type}  {l:list T} (pfdt:type_eq_dec T)
         (pfd:dup l),
    fst (dup_witness pfdt pfd) < length l.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros ? h1.
  pose proof h1 as h1'.
  apply not_dup_nil in h1'.
  contradiction.
  intros h1 h2.
  destruct in_dec as [h3 | h3]; simpl; auto with arith.
Qed.


Lemma lt_dup_witness2 :
  forall {T:Type}  {l:list T} (pfdt:type_eq_dec T)
         (pfd:dup l),
    snd (dup_witness pfdt pfd) < length l.
  intros T l. 
  induction l as [|a l ih]; simpl. 
  intros ? h1.
  pose proof h1 as h1'.
  apply not_dup_nil in h1'.
  contradiction.
  intros h1 h2. 
  destruct in_dec as [h3 | h3]; simpl; auto with arith.
  apply lt_n_S.
  apply lt_lind.
Qed.

Lemma dup_witness2_pos :
  forall {T:Type}  {l:list T} (pfdt:type_eq_dec T)
         (pfd:dup l),
    snd (dup_witness pfdt pfd) > 0.
  intros T l.
  destruct l as [|a l].
  intros ? h1.
  contradict (not_dup_nil _ h1).
  intros h1 h2.
  simpl.
  destruct in_dec; simpl; auto with arith.
Qed.

Lemma dup_witness_neq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (pfd:dup l),
    fst (dup_witness pfdt pfd) <> snd (dup_witness pfdt pfd).
  intros T h1 l.
  induction l as [|a l ih].
  intro h2.
  contradict (not_dup_nil _ h2).
  intros h2.
  pose proof (dup_cons h1 _ _ h2) as h3.
  destruct h3 as [h3 | h3].
  simpl.
  destruct in_dec as [h4 | h4]; simpl.
  intro; discriminate.
  contradiction.
  specialize (ih h3).

  
  simpl.
  destruct in_dec as [h4 | h4].
  simpl; intro; discriminate.
  simpl.
  intro h5.
  apply S_inj in h5.
  pose proof (proof_irrelevance
                _ h3
                (dec_not_l (dup_cons h1 l a h2) h4) ); 
    subst; contradiction.
Qed.

Lemma dup_witness_compat :
  forall {T:Type}  {l:list T} (pfdt:type_eq_dec T)
         (pfd:dup l),
    let pf := dup_witness pfdt pfd in
    nth_lt l _ (lt_dup_witness1 pfdt pfd) =
    nth_lt l _ (lt_dup_witness2 pfdt pfd).
Proof.
  intros T l.
  induction l as [|a l ih]; simpl.
  intros ? h1.
  contradict (not_dup_nil _ h1).
  intros h1 h2.
  destruct in_dec as [h3 | h3]; simpl.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  symmetry.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  rewrite ih.
  reflexivity.
  apply lt_dup_witness2.
  apply lt_dup_witness1.
Qed.

Lemma dup_iff' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    dup l <->
    exists m n
           (pfm:m < length l)
           (pfn:n < length l),
      m <> n /\ 
      nth_lt _ _ pfm =
      nth_lt _ _ pfn.
  intros T h1 l; split; intro h2.
  exists
    _, _,
  (lt_dup_witness1 h1 h2),
  (lt_dup_witness2 h1 h2).
  split.
  apply dup_witness_neq.
  apply dup_witness_compat.

  revert h2.
  induction l as [|a l ih]; simpl.
  intro h2.
  destruct h2 as [? [? [h2]]].
  contradict (lt_n_0 _ h2).
  intro h2. 
  destruct h2 as [m [n [h3 [h4 [h5 h6]]]]].
  destruct m as [|m], n as [|n]. 
  contradict h5; auto.
  apply lt_S_n in h4. 
  erewrite nth_indep in h6.
  erewrite <- (nth_lt_compat _ _ h4) in h6 at 1.
  subst.
  constructor.
  apply in_nth_lt.
  erewrite nth_indep in h6.
  erewrite <- (nth_lt_compat _ _ h4) in h6 at 1.
  subst.
  assumption.
  assumption.
  apply lt_S_n in h3.
  erewrite nth_indep in h6.
  erewrite <- (nth_lt_compat _ _ h3) in h6 at 1.
  subst.
  constructor.
  apply in_nth_lt.
  assumption.
  apply lt_S_n in h3.
  apply lt_S_n in h4.
  apply dup_intro_cons.
  apply ih.
  exists _, _, h3, h4.
  split.
  intro; subst.
  contradict h5; auto.
  erewrite nth_lt_compat.
  erewrite nth_indep; auto.
  rewrite h6 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
Qed.


Lemma lind_nth_lt_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (n:nat)
         (pf:n < length l)
         (pfi:In (nth_lt _ _ pf) l),
    NoDup l ->
    lind pfdt l (nth_lt l n pf) pfi = n.
  intros T h1 l n h2 h3 h4.
  rewrite lind_eq_iff.
  split. 
  apply nth_lt_functional3.
  intro h5.
  rewrite in_firstn_iff in h5.
  destruct h5 as [m [h5 h6]].
  rewrite no_dup_iff in h4.
  rewrite dup_iff' in h4.
  contradict h4.
  exists m, n, (lt_trans _ _ _ h5 h2), h2.
  split.
  intro; subst; contradict (lt_irrefl _ h5).
  rewrite <- h6.
  apply nth_lt_functional3.
  assumption.

  Grab Existential Variables.
  auto with arith.
  assumption.
Qed.




Lemma lind_seg_list :
  forall i n (pf:In i (seg_list n)),
    lind nat_eq_dec (seg_list n) i pf = i.
  intros i n h1.
  eapply nth_lt_inj.
  apply no_dup_seg_list.
  rewrite lind_compat, nth_lt_seg_list; auto.

  Grab Existential Variables.
  rewrite in_seg_list_iff in h1.
  rewrite length_seg_list; auto.
Qed.


Lemma lind_seq :
  forall i m n (pf:In i (seq m n)),
    lind nat_eq_dec (seq m n) i pf = i - m.
  intros i m n h1.
  pose proof h1 as h1'.
  rewrite in_seq_iff in h1'.
  eapply nth_lt_inj.
  apply no_dup_seq. 
  rewrite lind_compat, nth_lt_seq; auto.
  omega.

  Grab Existential Variables.
  rewrite seq_length.
  omega.
Qed.

Definition seq_lind {T:Type} (l:list T) (m n:nat)
           (pf:m < length l) :=
  firstn n (skipn m l).

Lemma length_seq_lind :
  forall {T:Type} (l:list T) (m n:nat)
         (pf:m < length l),
    n <= length l - m ->
    length (seq_lind l m n pf) = n.
  intros T l m n h1 h2.
  unfold seq_lind.
  rewrite length_firstn; auto.
  rewrite length_skipn; auto.
Qed.


Lemma nth_lt_seq_lind :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (m n r:nat)
         (pfm:m < length l)
         (pfr:r < length (seq_lind l m n pfm))
         (pfmr:m + r < length l),
    nth_lt (seq_lind l m n pfm) r pfr =
    nth_lt l (m + r) pfmr.
  intros T hdt l m n r h1 h2 h3. 
  unfold seq_lind.  
  rewrite nth_lt_firstn_eq.
  erewrite nth_lt_skipn_eq.
  apply nth_lt_functional2; auto with arith.
  assumption.

  Grab Existential Variables.
  omega.
Qed.

Lemma removelast_seq_lind :
  forall {T:Type} (l:list T) (m n:nat)
         (pfm:m < length l),
    m + n <= length l ->
    removelast (seq_lind l m n pfm) =
    seq_lind l m (pred n) pfm.
  intros T l m n h1 hm.
  unfold seq_lind.
  destruct n as [|n]; auto.
  rewrite removelast_firstn.
  simpl; auto.
  rewrite length_skipn.
  omega.
Qed.
  
  
Lemma seq_lind1_pred :
  forall {T:Type} (l:list T)
         (pf:1 < length l),
    seq_lind l 1 (pred (length l)) pf = tl l.
  intros T l.
  destruct l as [|a l].
  simpl; intros; omega.
  intros; simpl.
  unfold seq_lind.
  simpl.
  rewrite firstn_length; auto.
Qed.


Lemma seq_lind1_pred_pred :
  forall {T:Type} (l:list T)
         (pf1:1 < length l)
         (pfn:l <> nil),
    seq_lind l 1 (pred (pred (length l))) pf1 =
    removelast (tl l).
  intros T l h1 h2.
  pose proof (seq_lind1_pred l h1) as h3.
  rewrite <- h3.
  rewrite <- removelast_seq_lind; auto.
  omega.
Qed.


(*Substitutes a single element (indicated by a given index [k]) 
for a given arbitrary type-element [y] in a given list [l]
(substitutes the same whether or not [y] is in [l]).*)
Definition list_subst
           {T:Type} (l:list T)
           (k:nat) (pfk:k < length l)
           (x:T) : list T :=
  map_seg (seg_fun_subst (nth_lt l) x k pfk).

  
Lemma length_list_subst : 
  forall {T:Type} (l:list T)
         (k:nat) (pfk:k < length l)
         (x:T),
    length (list_subst l k pfk x) = length l.
  unfold list_subst; intros; rewrite length_map_seg; auto.
Qed.


Lemma list_subst_cons0 :
  forall {T:Type} (l:list T) (a b:T)
         (pf : 0 < length (a::l)),
    list_subst (a::l) 0 pf b =
    b::l.
  intros T l a b h1.
  unfold list_subst.
  rewrite map_seg_eq_iff.
  simpl.
  exists eq_refl. 
  intros m h2. 
  destruct m; simpl; auto.
  apply nth_indep; auto with arith.
Qed.

Lemma list_subst_cons_S :
  forall {T:Type} (l:list T) (a b:T) (i:nat)
           (pf:S i < S (length l)),
    list_subst (b::l) (S i) pf a =
    b::list_subst l i (lt_S_n _ _ pf) a.
  intros T l a b i h1.
  unfold list_subst.
  rewrite map_seg_eq_iff. 
  ex_goal.
  simpl; rewrite length_map_seg; auto.
  exists hex.
  intros m h2.
  intro h3. 
  destruct m as [|m]; subst; auto.
  rewrite nth_lt_cons, nth_lt_map_seg'.
  simpl.
  unfold seg_fun_to_S.
  simpl.
  f_equal.
  apply functional_extensionality_dep.
  intro c.
  apply functional_extensionality.
  intro hc.
  symmetry.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  apply proof_irrelevance.
Qed.



Lemma nth_lt_list_subst_same  :
  forall {U:Type} (l:list U) (i:nat) (a:U)
         (pfi:i < length l)
         (pfi':i < length (list_subst l i pfi a)),
    nth_lt (list_subst l i pfi a) i pfi' =
    a.
  unfold list_subst; intros U l.
  induction l as [|a l ih].
  simpl.
  intros; omega.
  intros i y h2 h3.
  simpl in h2.
  pose proof h2 as h4. 
  destruct i as [|i].
  rewrite nth_lt_map_seg'.
  simpl; auto.
  rewrite nth_lt_map_seg'.
  apply lt_S_n in h4.
  specialize (ih i y h4).
  hfirst ih.
  rewrite length_map_seg; auto.
  specialize (ih hf).
  rewrite nth_lt_map_seg' in ih.
  simpl.
  unfold seg_fun_to_S.
  simpl.
  rewrite <- ih.
  f_equal.
  apply functional_extensionality_dep; intro; apply functional_extensionality; intro.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  apply proof_irrelevance.
  apply proof_irrelevance.
Qed.
  

Lemma nth_lt_list_subst_neq :
  forall {U:Type} (l:list U) (i m:nat) (y:U)
         (pfi:i < length l)
         (pfm:m < length (list_subst l i pfi y)),
    i <> m ->
    nth_lt (list_subst l i pfi y) m pfm =
    nth_lt l m (lt_eq_trans _ _ _ pfm (length_list_subst _ _ _ _)).
  unfold list_subst; intros U l.
  induction l as [|a l ih].
  simpl; intros; omega.
  intros i m y hi.
  simpl in hi.
  pose proof hi as h1.
  destruct i as [|i].
  intros h2 h3.
  apply neq_sym in h3.
  rewrite neq0_iff in h3.
  symmetry.
  gen0.
  generalize h2.
  rewrite (S_pred _ _ h3).
  intros h4 h5.
  rewrite nth_lt_cons, nth_lt_map_seg'.
  simpl in h5.
  simpl.
  rewrite nth_lt_compat.
  apply nth_indep.
  apply lt_S_n in h5; auto.
  intros h2 h3.
  pose proof h2 as h4. 
  rewrite length_map_seg in h4.
  rewrite nth_lt_map_seg'.
  simpl in h4.
  destruct m as [|m].
  simpl; auto.
  simpl.
  apply lt_S_n in h4.
  apply lt_S_n in h1.
  specialize (ih i m y h1).
  hfirst ih.
  rewrite length_map_seg; auto.
  specialize (ih hf).
  hfirst ih. 
  contradict h3; auto.
  specialize (ih hf0).
  rewrite nth_lt_map_seg' in ih.
  rewrite nth_lt_compat in ih.
  erewrite nth_indep in ih.
  rewrite <- ih at 1.
  unfold seg_fun_to_S; simpl.
  f_equal.
  apply functional_extensionality_dep; intro; apply functional_extensionality; intro.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  apply proof_irrelevance.
  apply proof_irrelevance.
  assumption.
Qed.


Lemma list_subst_seq_pred :
  forall (s n:nat)
         (pfl:pred (length (seq s n)) < length (seq s n)),
    list_subst (seq s n) (pred (length (seq s n))) pfl (pred s) =
    seq s (pred n) ++ pred s :: nil.
  unfold list_subst; intros s n h1. 
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite seq_length, app_length, seq_length; simpl.
  destruct n as [|n].
  simpl in h1. 
  apply lt_irrefl in h1; contradiction.
  simpl.
  rewrite S_compat; auto.
  exists hex.
  intros m h2 h3.
  pose proof (seg_fun_subst_compat (nth_lt (seq s n)) (pred s) (pred (length (seq s n))) h1 m h2) as h4.
  simpl in h4.
  destruct h4 as [h4 h5].
  pose proof h2 as h6.
  rewrite seq_length in h6.
  symmetry. 
  destruct n as [|n].
  contradict (lt_n_0 _ h6).
  apply lt_dec in h6.
  simpl in h6.
  destruct h6 as [h6 | h6].
  hfirst h5.
  rewrite seq_length; simpl; intro; subst.
  contradict (lt_irrefl _ h6).
  specialize (h5 hf).
  rewrite h5.
  simpl.
  destruct m as [|m].
  erewrite nth_lt_app1. 
  rewrite nth_lt_seq, plus_0_r; auto. 
  erewrite nth_lt_app1.
  rewrite nth_lt_seq.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite nth_lt_seq; omega.
  rewrite seq_length; auto with arith.
  unfold h3.
  clear h3.
  subst.
  hfirst h4.
  rewrite seq_length; simpl; auto.
  specialize (h4 hf).
  rewrite h4.
  erewrite nth_lt_app2.
  gen0.
  rewrite seq_length.
  intro h6.
  simpl.
  rewrite minus_same; auto.

  Grab Existential Variables.
  rewrite seq_length; simpl; auto with arith.
  rewrite seq_length; auto with arith.
  rewrite seq_length; auto.
  rewrite seq_length; auto.
Qed.



Lemma inv_inv_map_in_dep_map_seg_lind_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T)
         (pf:In x (map_in_dep_map_seg_lind_singularize pfdt l n pfle)),
    let pflt :=
        lt_le_trans
          _ _ _
          (lt_inv_inv_map_in_dep_map_seg_lind_singularize
             pfdt l n pfle x pf) pfle in
    nth_lt l _ pflt = x.  
  intros T hdt l n hle x h1 h2. 
  rewrite nth_lt_eq_iff.
  pose proof h1 as h3.
  unfold map_in_dep_map_seg_lind_singularize in h3.
  rewrite in_map_in_dep_iff in h3.
  destruct h3 as [i [h3 h4]]; subst.
  ex_goal.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  exists hex.
  pose proof h3 as h5.
  unfold map_seg_lind_singularize in h5.
  rewrite in_map_seg_iff in h5.
  exists 0.
  split.
  apply gt_lt.
  rewrite <- (count_occ_In hdt).
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  unfold inv_inv_map_in_dep_map_seg_lind_singularize, 
  inv_map_in_dep_map_seg_lind_singularize,
  inv_map_seg_lind_singularize.
  symmetry.
  gen0.
  unfold inv_map_in_dep_map_seg_lind_singularize, map_seg_lind_singularize.
  rewrite nth_lt_map_seg'.
  intros h10.
  pose proof (lind_eq_iff nat_eq_dec   (map_seg
        (fun (i0 : nat) (pfi : i0 < n) =>
           lind_singularize hdt l i0 (lt_le_trans i0 n (length l) pfi hle)))) as hc.
  match goal with |- lind _ _ ?x' _ = _ => pose x' as x end.
  gtermr.
  specialize (hc x h10 y). 
  hfirst hc.
  unfold y.
  destruct h5 as [r [hd he]].
  generalize hex.
  generalize (lt_map_seg_lind_singularize hdt l i n hle h3).
  rewrite <- he.
  unfold lind_singularize.
  intro hg.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro hh.
  fold (lind hdt l (nth_lt l r (lt_le_trans r n (length l ) hd hle)) hh).
  eapply le_lt_trans.
  erewrite lind_functional3.
  apply le_lind_nth_lt.
  rewrite length_map_seg.
  assumption.
  specialize (hc hf).
  fold x y.
  rewrite hc at 1.
  split.
  rewrite nth_lt_map_seg'.
  apply f_equal_dep.
  unfold lind.
  symmetry.
  unfold y.
  change
    (lind hdt (map_in_dep_map_seg_lind_singularize hdt l n hle)
              (nth_lt (list_singularize hdt l) i
                      (lt_map_seg_lind_singularize hdt l i n hle h3)) h1 =
     lind hdt l
              (nth_lt (list_singularize hdt l) i
                      (lt_map_seg_lind_singularize hdt l i n hle h3)) hex).
  rewrite lind_eq_iff.
  split.
  unfold map_in_dep_map_seg_lind_singularize.
  rewrite nth_lt_map_in_dep'.
  apply nth_lt_functional2.
  unfold map_seg_lind_singularize.
  rewrite nth_lt_map_seg'.
  unfold lind_singularize.
  gen0.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro h11.
  rewrite lind_nth_lt_no_dup; auto.
  apply no_dup_list_singularize.
  intro h11.
  rewrite in_firstn_iff in h11.
  destruct h11 as [r [h14 h12]].
  unfold map_in_dep_map_seg_lind_singularize in h12.
  rewrite nth_lt_map_in_dep' in h12.
  apply nth_lt_inj in h12.
  unfold map_seg_lind_singularize in h12.
  rewrite nth_lt_map_seg' in h12.
  unfold lind_singularize in h12.
  revert h12.
  match goal with
      |- lind _ _ (nth_lt _ _ ?pf) _ = _ -> _ => generalize pf end.
  intros h15 h16.
  revert h14.
  generalize hex.
  generalize (lt_map_seg_lind_singularize hdt _ _ _ _ h3).
  rewrite <- h16.
  intros h17.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intros h18 h19.
  pose proof (le_lind_nth_lt hdt l r h15) as h20.
  pose proof (proof_irrelevance
                _
                h18
                (in_nth_lt _ _ _)) as h21.
  rewrite <- h21 in h20.
  omega.
  apply no_dup_list_singularize.
  intro h22.
  rewrite in_firstn_iff in h22.
  destruct h22 as [r [h23 h24]].
  rewrite nth_lt_map_seg' in h24.
  apply lind_singularize_eq_nth_lt in h24.
  revert h24.
  gen0.
  intros h24 h25.
  symmetry in h25.
  revert h25.
  gen0.
  intros h25 h26.
  rewrite (nth_lt_eq_iff hdt) in h26.  
  destruct h26 as [h27 [occ [h28 h29]]].
  symmetry in h29.
  pose proof (lind_eq_iff hdt  (map_in_dep_map_seg_lind_singularize hdt l n hle)
                          (nth_lt (list_singularize hdt l) i
                                  (lt_map_seg_lind_singularize hdt l i n hle h3)) h1 (lind_occ hdt l occ (nth_lt l r h24) h27)) as ha.
  hfirst ha.
  rewrite <- h29.
  apply lt_lind.
  specialize (ha hf0).
  rewrite ha in h29 at 1.
  destruct h29 as [h29 h30].
  unfold map_in_dep_map_seg_lind_singularize in h29.
  rewrite nth_lt_map_in_dep' in h29.
  apply nth_lt_inj in h29.
  unfold map_seg_lind_singularize in h29.
  rewrite nth_lt_map_seg' in h29.
  revert h29.
  gen0.
  intros h31 h32.
  unfold lind_singularize in h32.
  pose proof (lind_eq_iff hdt (list_singularize hdt l)
                            (nth_lt l (lind_occ hdt l occ (nth_lt l r h24) h27) h31)
          (iff1
             (list_singularize_in_iff hdt l
                (nth_lt l (lind_occ hdt l occ (nth_lt l r h24) h27) h31))
             (in_nth_lt l (lind_occ hdt l occ (nth_lt l r h24) h27) h31)) i  (lt_map_seg_lind_singularize hdt l i n hle h3)) as hb.
  rewrite hb in h32 at 1.
  destruct h32 as [h32 h33].
  symmetry in h32.
  erewrite nth_lt_functional3 in h32.
  rewrite lind_occ_compat in h32.
  erewrite nth_lt_functional3 in h33.
  rewrite lind_occ_compat in h33.
  rewrite h32 in h33.
  revert h23.
  change
    (r < lind hdt l (nth_lt _ _ (lt_map_seg_lind_singularize hdt l i n hle h3)) hex
     -> False). 
  generalize hex.
  rewrite <- h32.
  intros h34 h35.
  pose proof (le_lind_nth_lt hdt l r h24) as h36.
  pose proof (proof_irrelevance _ h34
                                (in_nth_lt _ _ _)) as h37.
  rewrite <- h37 in h36.
  omega.
  apply no_dup_list_singularize.
 
  Grab Existential Variables.
  unfold y.
  rewrite length_map_seg.

  destruct h5 as [r [h5 h6]].
  generalize hex.
  generalize (lt_map_seg_lind_singularize hdt l i n hle h3).
  rewrite <- h6.
  intro he.
  unfold lind_singularize.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro hg.
  fold (lind hdt l (nth_lt l r (lt_le_trans r n (length l) h5 hle)) hg).
  eapply le_trans.
  erewrite lind_functional3.
  eapply le_lind_nth_lt.
  apply lt_impl_le; auto.
  rewrite length_map_in_dep_map_seg_lind_singularize.

  destruct h5 as [r [h5 h6]].
  generalize hex.
  generalize (lt_map_seg_lind_singularize hdt l i n hle h3).
  rewrite <- h6.
  intro he.
  unfold lind_singularize.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro hg.
  fold (lind hdt l (nth_lt l r (lt_le_trans r n (length l) h5 hle)) hg).
  eapply le_trans.
  erewrite lind_functional3.
  eapply le_lind_nth_lt.
  apply lt_impl_le; auto.
  rewrite length_map_in_dep_map_seg_lind_singularize.

  destruct h5 as [r [h5 h6]].
  generalize hex.
  generalize (lt_map_seg_lind_singularize hdt l i n hle h3).
  rewrite <- h6.
  intro he.
  unfold lind_singularize.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro hg.
  fold (lind hdt l (nth_lt l r (lt_le_trans r n (length l) h5 hle)) hg).
  eapply le_lt_trans.
  erewrite lind_functional3.
  eapply le_lind_nth_lt.
  assumption.
Qed.


Lemma inv_inv_im_im_lind_singularize_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfle:n <= length l)
         (x:T)
         (pf:Inn (im_im_lind_singularize pfdt l n pfle) x),
    let pflt :=
        lt_le_trans
          _ _ _
          (lt_inv_inv_im_im_lind_singularize
             pfdt l n pfle x pf) pfle in
    nth_lt l _ pflt = x.
  unfold inv_inv_im_im_lind_singularize; intros.
  erewrite nth_lt_functional3.
  apply inv_inv_map_in_dep_map_seg_lind_singularize_compat.
Qed.  

   
Lemma im_list_to_set_seg_list :
  forall {U:Type} (pfdu:type_eq_dec U) (n:nat) (f:nat->U),
    Im (list_to_set (seg_list n)) f =
    list_to_set (map_seg (fun_to_seg f n)).
  intros; rewrite list_to_set_map_seg_, list_to_set_seg_list_eq; auto.
Qed.

(*This is the minimal "firstn sub_list" of [l] that contains 
  all [x]s in [l].*)
Definition firstn_all_min {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (x:T) (pf:In x l) :=
  firstn (S (lind_occ pfdt l (pred (count_occ pfdt l x)) x pf)) l.

(*This is the minimal "skipn sublist" that contains all [x]s in [l].*)
Definition skipn_all_min  {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (x:T) (pf:In x l) :=
  skipn (lind pfdt l x pf) l.

 
Lemma last_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    last (firstn_all_min pfdt l x pf) x = x.
  intros T h1 l x h2.
  unfold firstn_all_min.
  erewrite firstn_S_eq.
  rewrite last_app_sing.
  rewrite lind_occ_compat; auto.
Qed.

  
Lemma in_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l),
      In x (firstn_all_min pfdt l x pf).
  intros T h1 l x h2.
  unfold firstn_all_min.
  rewrite (in_firstn_iff' h1).
  exists h2.
  pose proof h2 as h2'. 
  rewrite (count_occ_In h1) in h2' . 
  rewrite (S_pred _ _ h2').
  simpl. 
  pose proof (lt_lind_S_occ h1 l x h2 (pred (pred (count_occ h1 l x)))) as h3.
  destruct h3 as [h3|h3]; auto.
  destruct (zerop (pred (count_occ h1 l x))) as [h4 | h4].
  rewrite h4 in h3.
  simpl in h3.
  rewrite h4.
  unfold lind.
  apply lt_n_Sn.
  rewrite (S_pred _ _ h4).
  simpl.
  omega.
  rewrite <- h3.
  simpl.
  unfold lind.
  apply lt_n_Sn.
Qed.

Lemma in_skipn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l),
    In x (skipn_all_min pfdt l x pf).
  intros T h1 l x h2.
  unfold skipn_all_min.
  rewrite (in_skipn_iff h1).
  exists h2.
  exists 0.
  unfold lind.
  pose proof h2 as h2'.
  rewrite (count_occ_In h1) in h2'.
  split; auto with arith.
Qed.


Lemma hd_skipn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    hd x (skipn_all_min pfdt l x pf) = x.
  intros T h1 l x h2.
  pose proof (in_skipn_all_min h1 _ _ h2) as h3.
  unfold skipn_all_min in h3. unfold skipn_all_min.
  induction l as [|a l ih].
  contradiction.
  unfold lind in h3. unfold lind.
  simpl in h3.
  simpl.
  destruct (h1 x a) as [|h4]; subst.
  simpl; auto.
  simpl in h3.
  simpl.
  destruct h2 as [|h2]; subst.
  pose proof (h4 eq_refl); contradiction.
  specialize (ih h2).
  hfirst ih.
  unfold lind.
  erewrite lind_occ_functional.
  apply h3.
  reflexivity.
  reflexivity.
  reflexivity.
  specialize (ih hf).
  unfold lind in ih.
  erewrite lind_occ_functional.
  rewrite ih at 1.
  reflexivity.
  reflexivity.
  reflexivity.
  reflexivity.
Qed.


Lemma count_occ_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    count_occ pfdt (firstn_all_min pfdt l x pf) x =
    count_occ pfdt l x.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros x h2.
  rewrite (switch_eq_dec_eq h1 x a).
  destruct (h1 a x) as [|h3]; subst.
  f_equal.
  unfold switch_eq_dec.
  simpl. 
  destruct (in_dec h1 x l) as [h3 | h3].
  pose proof h3 as h3'.
  rewrite (count_occ_In h1) in h3'.
  rewrite (S_pred _ _ h3').
  simpl.
  rewrite <- (S_pred _ _ h3').
  apply ih.
  rewrite (count_occ_In h1) in h3.
  rewrite (not_lt0_iff (count_occ h1 l x)) in h3.
  rewrite h3.
  simpl.
  reflexivity.
  destruct h2 as [|h2]; subst.
  contradict h3; auto.
  apply ih.
Qed.

Lemma count_occ_skipn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    count_occ pfdt (skipn_all_min pfdt l x pf) x =
    count_occ pfdt l x.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros; contradiction.
  intros x h2.
  destruct h2 as [|h2]; subst.
  unfold skipn_all_min, lind; simpl. 
  destruct (h1 x x) as [h3 | h3].
  rewrite skipn0.
  simpl.
  destruct (h1 x x) as [h3' | h3']; auto.
  contradiction.
  contradict h3; auto.
  specialize (ih _ h2).
  destruct (h1 a x) as [|h3]; subst.
  rewrite <- ih.
  unfold skipn_all_min, lind.
  simpl.
  destruct (h1 x x) as [h3 | h3].
  simpl.
  destruct (h1 x x) as [h4 | h4].
  f_equal; auto.
  contradiction.
  contradict h3; auto.
  unfold skipn_all_min, lind.
  simpl.
  destruct (h1 x a) as [|h4]; subst.
  contradict h3; auto.
  simpl.
  unfold skipn_all_min, lind in ih.
  rewrite <- ih.
  repeat f_equal; try apply proof_irrelevance.
Qed.


Lemma firstn_all_min_cons_in_hd :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf':In x (x::l))
         (pf:In x l),
    firstn_all_min pfdt (x::l) x pf' = x::firstn_all_min pfdt l x pf.
  intros T h1 l x h2 h3; unfold firstn_all_min; simpl.
  f_equal.
  destruct (h1 x x) as [|h4]; simpl.
  pose proof h3 as h3'.
  rewrite (count_occ_In h1) in h3'.
  rewrite (S_pred _ _ h3').
  destruct in_dec as [h4 | h4].
  simpl.
  rewrite (hd_compat' _ (in_not_nil _ _ h4)) at 1.
  rewrite (hd_compat' _ (in_not_nil _ _ h4)) at 6.
  repeat f_equal; try apply proof_irrelevance.
  contradiction.
  contradict h4; auto.
Qed.

  
Lemma firstn_all_min_cons_nin_hd :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x (x::l)),
    ~In x l ->
    firstn_all_min pfdt (x::l) x pf = x::nil.
  intros T h1 l x h2 h3; unfold firstn_all_min; simpl.
  f_equal.
  destruct (h1 x x) as [|h4]; simpl.
  rewrite (count_occ_In h1) in h3.
  rewrite (not_lt0_iff (count_occ h1 l x)) in h3 at 1.
  rewrite h3; simpl; auto.
  contradict h4; auto.
Qed.


Lemma skipn_all_min_cons_hd :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf':In x (x::l)),
    skipn_all_min pfdt (x::l) x pf' = x::l.
  intros T h1 l x h2.
  unfold skipn_all_min.
  erewrite lind_functional. rewrite lind_hd'.
  simpl; auto.
  simpl; auto.
  Grab Existential Variables.
  apply cons_neq_nil.
Qed.


Lemma skipn_all_min_cons_in_last :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf':In x (l ++ x::nil))
         (pf:In x l),
    skipn_all_min pfdt (l ++ x::nil) x pf' =
    (skipn_all_min pfdt l x pf) ++ x::nil.
  intros T h1 l x h2 h3.
  unfold skipn_all_min, lind.
  rewrite skipn_app.
  f_equal.
  erewrite lind_occ_functional. rewrite lind_occ_app_lt_in1 at 1.
  f_equal.  f_equal. apply proof_irrelevance.
  reflexivity.
  reflexivity.
  reflexivity.
  erewrite lind_occ_functional.
  rewrite lind_occ_app_lt_in1.
  apply lt_le_weak.
  apply lt_lind_occ.
  reflexivity.
  reflexivity.
  reflexivity.
  Grab Existential Variables.
  rewrite (count_occ_In h1) in h3; auto with arith.
  rewrite (count_occ_In h1) in h3; auto with arith.
Qed.
  

Lemma skipn_all_min_cons_nin_last :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf':In x (l ++ x::nil))
         (pf: ~In x l),
    skipn_all_min pfdt (l ++ x::nil) x pf' =
    x::nil.
  intros T h1 l x h2 h3.
  pose proof h2 as h2'. apply in_app_or in h2'.
  destruct h2' as [|h4].
  contradiction.
  unfold skipn_all_min.
  erewrite lind_functional.
  rewrite lind_app_nin.
  rewrite skipn_length_app_plus.
  rewrite lind_cons_same.
  simpl; auto.
  assumption.
  reflexivity.

  Grab Existential Variables.
  left; auto.
Qed.


Lemma lind_occ_hd_in :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (pfn:l <> nil) (occ:nat),
    let h := hd' l pfn  in
    let pfh := in_hd' l pfn in
    forall (pfh':In h (tl l)),
      lind_occ pfdt l (S occ) h pfh =
      S (lind_occ pfdt (tl l) occ h pfh').
  intros T h1 l h2 occ h h3 h4.
  destruct l as [|a l]; simpl.
  contradiction.
  simpl in h, h3, h4.
  destruct in_dec as [h7 | h7].
  unfold h. unfold h in h4. 
  pose (left (a <> a) (eq_refl a)) as h8.
  pose proof (classic_eq (h1 a a) h8) as h9.
  rewrite h9.
  unfold h8.
  repeat f_equal; try apply proof_irrelevance.
  contradiction.
Qed.


Lemma lind_occ_skipn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (m:nat)
         (pf:In x (skipn m l)) (occ:nat),
    m <= length l ->
    occ < count_occ pfdt (skipn m l) x ->
    let pf' := in_skipn _ _ _ pf in
    lind_occ pfdt (skipn m l) occ x pf =
    lind_occ pfdt l (occ + count_occ pfdt (firstn m l) x) x pf' - m.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros ? ? hi.
  pose proof hi as hi'.
  rewrite skipn_nil in hi'.
  contradiction.
  intros x m h2 occ h3 h4.
  destruct m as [|m].
  simpl in h2, h4. simpl.
  destruct h2 as [|h2]; subst.
  rewrite eq_dec_same at 1 2.
  rewrite <- plus_n_O.
  rewrite <- minus_n_O.
  reflexivity.
  assumption.
  rewrite switch_eq_dec_eq in h4.
  destruct (h1 x a) as [|h5]; subst.
  simpl in h4.
  rewrite <- minus_n_O.
  rewrite <- plus_n_O.
  reflexivity.
  simpl in h4.
  rewrite <- minus_n_O.
  f_equal.
  rewrite <- plus_n_O.
  f_equal.
  apply proof_irrelevance.
  simpl in h2, h4. simpl.
  apply le_S_n in h3.
  rewrite (switch_eq_dec_eq h1 a x).
  destruct (h1 x a) as [|h5]; subst.
  simpl.
  rewrite plus_S_shift_r.
  destruct in_dec as [h5 | h5].
  rewrite ih at 1; auto. 
  assert (h5 = in_skipn l m a h2).
    apply proof_irrelevance.
  subst; omega.
  contradict h5.
  eapply in_skipn; apply h2.
  simpl.
  rewrite ih at 1; auto.
  f_equal.
  f_equal.
  apply proof_irrelevance.
Qed.
  


  
Lemma count_occ_removelast_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    count_occ pfdt (removelast (firstn_all_min pfdt l x pf)) x =
    pred (count_occ pfdt l x).
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros x h2.
  rewrite (switch_eq_dec_eq h1 x a).
  destruct (h1 a x) as [|h3]; subst.
  unfold switch_eq_dec.
  simpl.
  destruct (in_dec h1 x l) as [h3 | h3].
  pose proof h3 as h3'.
  rewrite (count_occ_In h1) in h3'.
  rewrite (S_pred _ _ h3').
  simpl.
  specialize (ih x h3).
  rewrite (hd_compat' _ (in_not_nil _ _ h3)) at 1.
  simpl.
  destruct (h1 x x) as [|h4].
  f_equal.
  apply ih.
  contradict h4; auto.
  rewrite (count_occ_In h1) in h3.
  rewrite (not_lt0_iff (count_occ h1 l x)) in h3.
  rewrite h3.
  simpl.
  reflexivity.
  destruct h2 as [|h2]; subst.
  contradict h3; auto.
  unfold switch_eq_dec.
  pose proof h2 as h2'.
  pose proof (in_firstn_all_min h1 _ _ h2') as h4.
  rewrite (count_occ_In h1) in h4.
  simpl.
  rewrite (hd_compat' _ (in_not_nil _ _ h2)) at 1.
  simpl.
  destruct (h1 a x) as [|h5]; subst.
  contradict h3; auto.
  apply ih.
Qed.

  
Lemma count_occ_tl_skipn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    count_occ pfdt (tl (skipn_all_min pfdt l x pf)) x =
    pred (count_occ pfdt l x).
  intros T h1 l.
  unfold skipn_all_min, lind.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2.
  destruct h2 as [|h2]; subst. 
  destruct (h1 x x) as [|h2]; simpl; auto.
  rewrite (switch_eq_dec_eq h1 a x).
  destruct (h1 a x) as [|h3]; subst.
  unfold switch_eq_dec.
  destruct (h1 x x) as [|h3].
  simpl; auto.
  contradict h3; auto.
  destruct (h1 x a) as [|h4]; subst.
  contradict h3; auto.
  unfold switch_eq_dec.
  simpl.
  erewrite <- ih.
  f_equal.
Qed.


  
Lemma pred_count_occ_hd :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    (pred (count_occ pfdt (x :: l) x)) =
    count_occ pfdt l x.
  intros T h1 l x.
  simpl.
  destruct (h1 x x) as [|h2]; subst; simpl; auto.
  contradict h2; auto.
Qed.


  
Lemma lind_occ_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l)
         (occ:nat),
    occ < count_occ pfdt l x ->
    let pf' := in_firstn_all_min pfdt _ _ pf in
    lind_occ pfdt (firstn_all_min pfdt l x pf) occ x pf' =
    lind_occ pfdt l occ x pf.
  intros T h1 l x h2 occ h3 h4.
  unfold firstn_all_min.
  rewrite lind_occ_firstn.
  f_equal.
  apply proof_irrelevance.
  fold (firstn_all_min h1 l x h2).
  rewrite count_occ_firstn_all_min; auto.
Qed.

Lemma in_removelast_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    1 < count_occ pfdt l x ->
    In x (removelast (firstn_all_min pfdt l x pf)).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2.
  rewrite (switch_eq_dec_eq h1 x a).
  destruct (h1 a x) as [h3 | h3]; subst.
  unfold switch_eq_dec. 
  simpl.
  intro h3.
  apply lt_S_n in h3.
  rewrite (S_pred _ _ h3).
  rewrite <- (count_occ_In h1 l x) in h3.
  destruct in_dec as [h4 | h4].
  simpl.
  rewrite (hd_compat' _ (in_not_nil _ _ h4)) at 1.
  left; auto.
  contradiction.
  intro h4.
  destruct h2 as [|h2].
  contradiction.
  unfold switch_eq_dec.
  simpl.
  rewrite (hd_compat' _ (in_not_nil _ _ h2)) at 1.
  right.
  apply ih; auto.
Qed.


  
Lemma firstn_all_min_cons_in_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
           (pfin:In x l),
    firstn_all_min pfdt (x::l) x (in_eq _ _) =
    x::firstn_all_min pfdt l x pfin.
  intros T h1 l x h2.
  unfold firstn_all_min.
  simpl.
  f_equal.
  destruct (h1 x x) as [|h3].
  simpl.
  pose proof h2 as h2'.
  rewrite (count_occ_In h1) in h2'.
  rewrite (S_pred _ _ h2').
  destruct in_dec as [h3 | h3].
  simpl.
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
  contradiction.
  contradict h3; auto.
Qed.

                                  
Lemma lind_occ_lt_length_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfin:In x l) (occ:nat),
    occ < count_occ pfdt l x ->
    lind_occ pfdt l occ x pfin < length (firstn_all_min pfdt l x pfin).
  intros T h1 l x h2 occ h3.
  pose proof (in_firstn_all_min h1 _ _ h2) as h4.
  rewrite <- lind_occ_firstn_all_min.
  apply lt_lind_occ.
  assumption.
Qed.


Lemma lind_removelast_in :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf':In x (removelast l)),
    let pf := removelast_in _ _ pf' in
    lind pfdt (removelast l) x pf' =
    lind pfdt l x pf.
  intros T h1 l x h2 h5.
  pose proof (app_removelast_last x (in_not_nil _ _ (removelast_in l x h2))) as h3.
  pose proof h5 as h5'.
  rewrite h3 in h5'.
  pose proof (lind_functional' h1 _ _ _ _ h5 h5' h3 eq_refl) as h6.
  symmetry.
  rewrite h6.
  erewrite lind_functional.
  rewrite lind_app_in1.
  reflexivity.
  reflexivity.
Qed.

Lemma lind_occ_removelast_in :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf':In x (removelast l))
         (occ:nat),
    occ < count_occ pfdt (removelast l) x ->
    let pf := removelast_in _ _ pf' in
    lind_occ pfdt (removelast l) occ x pf'  =
    lind_occ pfdt l occ _ pf.
  intros T h1 l x h2 occ h3 h4.  
  pose proof (app_removelast_last x (in_not_nil _ _ (removelast_in l x h2))) as h5.
  generalize h4, h2.
  rewrite h5, removelast_app_sing. 
  intros h4' h2'.
  pose proof (lind_occ_app_lt_in1 h1 (removelast l) (last l x::nil) _ _ h3) as h6.
  simpl in h6.
  symmetry.
  erewrite lind_occ_functional.
  rewrite h6 at 1.
  f_equal. apply proof_irrelevance.
  reflexivity.
  reflexivity.
  reflexivity.
Qed.  


Lemma lind_occ_lt_length_removelast_firstn_all_min :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfin:In x l) (occ:nat),
    S occ < count_occ pfdt l x ->
    lind_occ pfdt l occ x pfin < length (removelast(firstn_all_min pfdt l x pfin)).
  intros T h1 l x h2 occ h3.
  pose proof (in_removelast_firstn_all_min h1 l x h2) as h4. 
  hfirst h4.
  omega.
  specialize (h4 hf).
  pose proof (lt_lind_occ h1 _ occ _ h4) as h5.
  rewrite lind_occ_removelast_in in h5.
  erewrite lind_occ_functional in h5.
  rewrite lind_occ_firstn_all_min in h5 at 1.
  apply h5.
  omega.
  reflexivity.
  reflexivity.
  reflexivity.
  rewrite count_occ_removelast_firstn_all_min.
  omega.
Qed.

Lemma count_occ_firstn_removelast :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (n:nat),
    let lr := removelast l in
    n <= length lr ->
    count_occ pfdt (firstn n lr) x =
    count_occ pfdt (firstn n l) x.
  simpl; intros; rewrite firstn_removelast; auto.
Qed.

Lemma count_occ_firstn_lind_occ_S :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l) (occ:nat),
    S occ < count_occ pfdt l x ->
    count_occ pfdt (firstn (lind_occ pfdt l (S occ) x pf) l) x =
    S
      (count_occ pfdt
                 (firstn (lind_occ pfdt l occ x pf) (firstn_all_min pfdt l x pf)) x).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros x h2 occ h3.
  rewrite switch_eq_dec_eq in h3.
  destruct (h1 x a) as [|h4]; subst.
  unfold switch_eq_dec in h3.
  apply lt_S_n in h3.
  destruct in_dec as [h4 | h4].
  destruct occ as [|occ]; simpl.
  destruct (h1 a a) as [|h5].
  simpl.
  f_equal.
  fold (lind h1 l a h4).
  rewrite count_occ_firstn_eq0_iff.
  intros h4'.
  pose proof (proof_irrelevance _ h4 h4'); subst; auto.
  contradict h5; auto.
  destruct (h1 a a) as [|h5].
  f_equal.
  rewrite ih; auto.
  f_equal. f_equal.
  simpl.
  rewrite (S_pred _ _ h3).
  destruct in_dec as [h5 | h5].
  pose proof (proof_irrelevance _ h4 h5); subst.
  reflexivity.
  contradiction.
  contradict h5; auto.
  contradict h4.
  rewrite (count_occ_In h1).
  omega.
  unfold switch_eq_dec in h3.
  destruct h2 as [|h2]; subst.
  contradict h4; auto.
  specialize (ih _ h2 _ h3).
  pose proof (lt_lind_occ h1 l (S occ) x (neq_in_cons l x a h4 (or_intror h2))) as h5.
  simpl.
  rewrite (switch_eq_dec_eq h1 x a). 
  destruct (h1 a x) as [|h6]; subst.
  pose proof (h4 eq_refl); contradiction.
  unfold switch_eq_dec.
  pose proof (proof_irrelevance _ h2
                                (neq_in_cons l x a h4 (or_intror h2)))
       as h7.
  rewrite <- h7.
  rewrite ih.
  f_equal.
  f_equal.
  unfold firstn_all_min.
  f_equal.
  f_equal.
  f_equal.
  f_equal.
  apply proof_irrelevance.
Qed. 


Lemma lind_occ_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (occ n:nat)
         (pfin:In x l) (pflt:n < length l),
    occ < count_occ pfdt l x ->
    (lind_occ pfdt l occ x pfin = n <->
     nth_lt l n pflt = x /\
     count_occ pfdt (firstn (S n) l) x =  S occ /\
     count_occ pfdt (skipn (S n) l) x = count_occ pfdt l x - (S occ)).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x occ n h2 h3 h4.
  rewrite switch_eq_dec_eq in h4.
  destruct h2 as [|h2]; subst.
  rewrite (eq_dec_same h1) at 1 2 3.
  rewrite (eq_dec_same h1) in h4 at 1.
  simpl in h4.
  destruct occ as [|occ]. 
  split; intro h5; subst; simpl.
  split; auto with arith.
  simpl in h5.
  destruct n as [|n]; auto.
  destruct h5 as [h5 [h6 h7]].
  simpl in h6, h7.
  destruct l as [|a l].
  simpl in h3.
  omega.
  simpl in h3.
  apply lt_S_n in h3.
  simpl in h4, h5, h6, h7.
  destruct (h1 a x) as [|h8].
  apply S_inj in h6.
  discriminate.
  apply S_inj in h6.
  destruct n as [|n].
  contradiction.
  apply lt_S_n in h3. 
  pose proof (lt_length _ _ h3) as h9.
  erewrite nth_indep in h5. rewrite <- (nth_lt_compat _ _ h3) in h5 at 1.
  erewrite nth_lt_functional3 in h5. 
  pose proof (nth_lt_firstn_eq l n (S n)) as h10.
  hfirst h10.
  apply lt_length_firstn; auto with arith.
  specialize (h10 hf).
  pose proof (in_nth_lt (firstn (S n) l) n hf) as h11. 
  rewrite (count_occ_In h1) in h11.
  rewrite h10 in h11.
  erewrite nth_lt_functional3 in h5 at 1. rewrite h5 in h11 at 1.
  omega.
  assumption.
  apply lt_S_n in h4.
  destruct in_dec as [h5 | h5].
  destruct n as [|n].
  simpl.
  intuition.
  apply lt_S_n in h3.
  specialize (ih x occ n h5 h3 h4).
  split; intro h6.
  apply S_inj in h6.
  rewrite ih in h6.
  destruct h6 as [h6 [h7 h8]].
  split.
  erewrite nth_indep.
  rewrite <- nth_lt_compat at 1.
  apply h6.
  assumption.
  split.
  f_equal; auto.
  omega.
  f_equal.
  rewrite ih. 
  destruct h6 as [h6 [h7 h8]].
  split; auto.
  erewrite nth_indep in h6 at 1.
  rewrite <- nth_lt_compat in h6 at 1.
  apply h6.
  assumption.
  apply count_occ_in in h4.
  contradiction.
     
  destruct n as [|n].
  destruct (h1 x a) as [|h5]; subst.
  simpl in h4.
  simpl.
  rewrite (eq_dec_same h1) at 1 2; simpl.
  destruct occ as [|occ].
  intuition.
  apply lt_S_n in h4.
  destruct in_dec as [h5 | h5].
  intuition. 
  contradiction.
  simpl in h4.
  assert (he:h1 a x = switch_eq_dec _ _ (right (x = a) h5)).
  unfold switch_eq_dec.
  destruct (h1 a x); subst.
  contradict h5; auto. f_equal; apply proof_irrelevance.
  rewrite he.
  unfold switch_eq_dec.
  simpl.
  intuition.
  apply lt_S_n in h3.
  rewrite (switch_eq_dec_eq h1 a x).
  destruct (h1 x a) as [|h5]; subst.
  simpl in h4.
  rewrite (eq_dec_same h1) at 1 2.
  destruct occ as [|occ].
  split; intro h5.
  discriminate.
  destruct h5 as [h5 [h6 h7]]. 
  specialize (ih a 0 n h2 h3).
  hfirst ih.
  rewrite <- (count_occ_In h1 l a); auto. 
  specialize (ih hf).
  apply S_inj in h6.
  let P := type of ih in match P with ?hyp <-> _ => assert (h10:~ hyp) end.
  intro h10.
  rewrite ih in h10.
  destruct h10; omega.
  revert h10.
  generalize h2.
  rewrite <- h5.
  intros h10 h11.
  pose proof (le_lind_nth_lt h1 _ _ h3) as h12.
  unfold lind in h12.
  revert h12.
  generalize (in_nth_lt l n h3). erewrite nth_lt_compat.
  intros h13 h14.
  assert (h15: lind_occ h1 l 0 (nth n l a) h10 =
               lind_occ h1 l 0 (nth n l (nth_lt l n h3)) h13).
    apply lind_occ_functional; auto.
    apply nth_indep.
    assumption.
  rewrite h15 in h11.
  assert (h16:lind_occ h1 l 0 (nth n l (nth_lt l n h3)) h13 < n).
    omega.
  apply not_lt0_iff in h6.
  contradict h6.  
  rewrite <- (count_occ_In h1 (firstn (S n) l) a).
  rewrite in_firstn_iff'.
  exists h2.
  unfold lind.
  eapply lt_trans.
  erewrite lind_occ_functional. apply h16.
  reflexivity. reflexivity. rewrite <- h5. apply nth_indep.
  assumption.
  auto with arith.
  apply lt_S_n in h4.
  destruct in_dec as [h5 | h5].
  specialize (ih a occ n h2 h3 h4). 
  split; intro h6.
  apply S_inj in h6.
  pose proof (proof_irrelevance _ h2 h5) as h7.
  rewrite <- h7 in h6.
  rewrite ih in h6.
  destruct h6 as [h6 [h8 h9]].
  subst.
  split.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  rewrite h8, h9.
  split; auto.
  f_equal.
  pose proof (proof_irrelevance _ h2 h5) as h7.
  rewrite <- h7.
  rewrite ih.
  destruct h6 as [h6 [h8 h9]].
  split.
  rewrite <- h6. rewrite nth_lt_compat. apply nth_indep.
  assumption.
  apply S_inj in h8.
  rewrite h8, h9.
  intuition.
  contradiction.
  simpl in h4. simpl.
  specialize (ih x occ n h2 h3 h4).
  split; intro h6.
  apply S_inj in h6.
  pose proof (proof_irrelevance _ h2
                                (neq_in_cons l x a h5 (or_intror h2))) as h8.
  rewrite h8 in ih.
  rewrite ih in h6.
  destruct h6 as [h6 [h9 h10]].
  split.
  rewrite <- h6. rewrite nth_lt_compat. apply nth_indep.
  assumption.
  destruct l; tauto.
  f_equal.
  pose proof (proof_irrelevance _ h2
                                (neq_in_cons l x a h5 (or_intror h2))) as h8.
  rewrite <- h8.
  rewrite ih.
  destruct h6 as [h6 [h9 h10]].
  split.
  rewrite nth_lt_compat. rewrite <- h6. apply nth_indep.
  assumption.
  destruct l as [|c l].
  tauto.
  simpl; simpl in h9, h10.
  destruct (h1 c x) as [h11 | h11].
  tauto.
  tauto.
  Grab Existential Variables.
  assumption.
Qed.



Lemma lind_occ_hd'' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (occ:nat)
         (pfn:l <> nil),
    let pfin := in_hd' _ pfn in
    forall (pfin':In (hd' _   pfn) (tl l)),
      let pfin'' := in_not_nil _ _ pfin' in
      occ < count_occ pfdt l (hd'  _ pfin'') ->
      occ = 0 \/ 
      lind_occ pfdt l occ _ pfin =
      S (lind_occ pfdt (tl l) (pred occ) _ pfin').
  intros T h1 l occ. revert l.
  destruct occ as [|occ]; simpl.
  intros; left; auto.
  intros l h2 h3 h4.
  right.
  pose proof (hd_compat' _ h2) as h5; simpl in h5.
  pose proof (lind_occ_functional h1 l _ (S occ)  (S occ) _ _ (in_hd' _ h2) (in_eq (hd' l h2) (tl l)) h5 eq_refl eq_refl) as h6.
  rewrite h6; simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  lr_match_goal; fold hlr; destruct hlr as [hl0 | hr0].
  f_equal. f_equal. apply proof_irrelevance.
  contradiction.
  contradict hr; auto.
Qed.


Lemma in_firstn_S_lind_occ :
  forall {T:Type} (h1:type_eq_dec T) (l:list T) (x:T) (occ:nat)
         (pf:In x l),
    In x (firstn (S (lind_occ h1 l occ x pf)) l).
  intros T h1 l x occ h2. 
  rewrite (in_firstn_iff' h1). 
  exists h2.
  pose proof (le_lind_occ h1 l x h2 occ) as h4.
  omega.
Qed.

Lemma lind_occ_lrep :
  forall {T:Type} (pfdt:type_eq_dec T) (a x:T) (n:nat) (occ:nat)
         (pf:In x (lrep a n)),
    occ < n ->
    lind_occ pfdt (lrep a n) occ x pf = occ.
  intros T h1 a x n. revert a x.
  induction n as [|n ih]; simpl.
  intros; contradiction.
  intros a x occ h2 h3.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr]; subst.
  destruct occ as [|occ]; auto.
  apply lt_S_n in h3.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr]; subst.
  f_equal; auto.
  contradict hr.
  apply in_lrep.
  intro h4.
  apply lrep_nil_iff in h4; subst.
  omega.
  destruct h2 as [|h2]; subst. 
  contradict hr; auto.
  pose proof h2 as h2'.
  apply in_lrep' in h2'.
  contradiction.
Qed.


Lemma lind_occ_firstn_pred_count_occ_nth_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (m:nat)
         (pfsm:S m <= length l)
         (pfin:In (nth_lt l m pfsm) (firstn (S m) l)),
    lind_occ pfdt (firstn (S m) l)
    (pred (count_occ pfdt (firstn (S m) l) (nth_lt l m pfsm)))
    (nth_lt l m pfsm) pfin = m.
  intros T h1 l m h2 h3.
  rewrite lind_occ_eq_iff.
  split.
  rewrite nth_lt_firstn_eq.
  apply nth_lt_functional3.
  split.
  rewrite (count_occ_In h1) in h3.
  rewrite <- (S_pred _ _ h3).
  f_equal.
  apply firstn_idem.
  rewrite (count_occ_In h1) in h3.
  rewrite <- (S_pred _ _ h3).
  rewrite minus_same.
  rewrite count_occ_skipn.
  rewrite firstn_idem.
  apply minus_same.
  rewrite (count_occ_In h1) in h3.
  rewrite (S_pred _ _ h3); auto with arith.
  Grab Existential Variables.
  rewrite length_firstn; auto.
Qed.


Lemma lS_app :
  forall (l l':list nat),
    lS (l++l') =
    lS l ++ lS l'.
  intros l l'.
  unfold lS.
  rewrite map_eq_iff.
  ex_goal.
  do 2 rewrite app_length.
  do 2 rewrite map_length; auto.
  exists hex.
  intros n h1.
  pose proof h1 as h1'. apply lt_length_app_dec in h1'.
  destruct h1' as [h2 | h2].
  rewrite (nth_lt_app1 _ _ _ h2).
  erewrite nth_lt_app1.
  erewrite nth_lt_map.
  reflexivity.
  destruct (nil_dec' l') as [|h3]; subst.
  simpl in h2. omega.
  pose proof h1 as h1'.
  rewrite app_length in h1'.
  destruct (lt_le_dec n (length l)) as [h4 | h4].
  rewrite (nth_lt_app1 _ _ _ h4).
  erewrite nth_lt_app1.
  erewrite nth_lt_map.
  reflexivity.
  rewrite (nth_lt_app2 _ _ _ h4).
  erewrite nth_lt_app2.
  erewrite nth_lt_map.
  f_equal.
  apply nth_lt_functional2.
  rewrite map_length; auto.
  Grab Existential Variables.
  rewrite map_length; auto.
  rewrite map_length; auto.
  rewrite map_length; auto.
  rewrite map_length; auto.
Qed.

Lemma lS_map_seg :
  forall {T:Type} (n:nat) (f:seg_fun n nat),
    lS (map_seg f) =
    map_seg (fun c pf =>  S (f c pf)).
  intros T n.
  induction n as [|? ih]; simpl; auto.
  intro.
  rewrite lS_app, ih; auto.
Qed.

Lemma faml_lpred_famlS :
  forall (ll:faml nat),
    faml_lpred (famlS ll) = ll.
  unfold faml_lpred, famlS; intros; rewrite map_map.
  rewrite map_eq_iff.
  exists eq_refl.
  intros n h1.
  rewrite lpred_lS.
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma nin_famlS :
  forall (ll:faml nat) (ln:list nat),
    In 0 ln ->
    ~In ln (famlS ll).
  intros ll ln h1 h2.
  rewrite in_famlS_iff in h2. 
  destruct h2 as [ln' [h2 h3]]; subst.
  rewrite in_lS_iff in h1.
  destruct h1 as [? [? ?]]; discriminate.
Qed.


Lemma famlS_eq_iff :
  forall (ll ll':faml nat),
    famlS ll = ll' <->
     ((forall ln,
        In ln ll' ->
        ~In 0 ln) /\
     ll = faml_lpred ll').
  intros ll ll'. split; intro h1; subst.
  split.
  intros ln h1.
  rewrite in_famlS_iff in h1.
  destruct h1 as [ln' [h2 h3]]; subst.
  intro h3.
  rewrite in_lS_iff in h3.
  destruct h3 as [? [? ?]]; discriminate.
  rewrite faml_lpred_famlS; auto.
  destruct h1 as [h1 h2]; subst.
  unfold famlS, faml_lpred.
  rewrite map_map.
  rewrite map_eq_iff.
  exists eq_refl.
  intros n h2.
  specialize (h1 _ (in_nth_lt ll' n h2)).
  pose proof (lS_lpred_iff (nth_lt ll' n h2)) as h3.
  rewrite h3 in h1.
  rewrite <- h1.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma cardinal_length_compat : forall {T:Type} (l:list T), NoDup l -> cardinal _ (list_to_set l) (length l).
intros T l.
induction l as [|a l h1].
intros; simpl. constructor.
simpl; intro h2.
inversion h2 as [|x l' h4 h5]. clear h2. subst.
constructor.
apply h1; auto.
rewrite <- list_to_set_in_iff.
assumption.
Qed.

Lemma le_card_fun1_length : 
  forall {T:Type} (l:list T),
    card_fun1 (list_to_set l) <= length l.
  intros T l.
  induction l as [|a l h1]; simpl.
  rewrite card_fun1_empty.
  auto with arith.
  destruct (classic_dec (In a l)) as [h2 | h2].
  rewrite list_to_set_in_iff in h2.
  rewrite in_add_eq; auto.
  rewrite card_add_nin'.
  apply le_n_S.
  assumption.
  apply list_to_set_finite.
  rewrite <- list_to_set_in_iff.
  assumption.
Qed.

Lemma card_fun1_length_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    card_fun1 (list_to_set l) = length (list_singularize pfdt l).
  intros T hdt l.
  induction l as [|a l h1]; simpl.
  rewrite card_fun1_empty. reflexivity.
  destruct (in_dec hdt a l) as [h2 | h2].
  rewrite in_add_eq; auto.
  rewrite list_singularize_in_cons; auto.
  rewrite <- list_to_set_in_iff; assumption.
  rewrite list_singularize_nin; auto.
  rewrite card_add_nin'; simpl.
  rewrite h1; auto.
  apply list_to_set_finite.
  rewrite <- list_to_set_in_iff; auto.
Qed.


Lemma list_to_set_no_dup_length : forall {T:Type} (l:list T) (E:Ensemble T),
                                    list_to_set l = E -> NoDup l ->
                                    exists !n, length l = n.
intros T l.
induction l as [|a l h1].
simpl. intros. exists 0. red. split; auto.
simpl.
intros E h2 h3.
inversion h3 as [|x l' h4 h5].
subst. clear h3.
specialize (h1 _ (eq_refl _ )).
specialize (h1 h5).
destruct h1 as [n h1]. red in h1. destruct h1 as [h1 h2].
exists (S n). red. split.
f_equal. assumption.
intros x h6.
specialize (h2 _ (eq_refl _)).
rewrite <- h2 in h6.
assumption.
Qed.


Lemma list_to_set_maps_eq_length :
  forall {T U V:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (f:T->V) (g:U->V)
         (l:list T) (l':list U),
         NoDup l ->
         NoDup l' ->
         Injective f ->
         Injective g ->
         list_to_set (map f l) =
         list_to_set (map g l') ->
         length l =
         length l'.
  intros T U V hdt hdu hdv f g l.
  induction l as [|a l ih]; simpl.
  intros ? ? ? ? ?  h1.
  symmetry in h1.
  apply empty_set_nil in h1.
  rewrite map_eq_nil_iff in h1; subst; auto.
  intro l'. 
  destruct l' as [|a' l'].
  intros ? ? ? ?  h1.
  simpl in h1.
  apply add_not_empty in h1.
  contradiction. 
  simpl.
  intros h1 h2 h3 h4 h5.
  destruct (hdv (f a) (g a')) as [h6 | h6].
  rewrite h6 in h5.
  apply add_nin_inj_eq in h5.
  f_equal.
  apply ih; auto.
  apply no_dup_cons in h1; auto.
  apply no_dup_cons in h2; auto.
  intro h7.
  rewrite <- h6 in h7.
  rewrite <- list_to_set_in_iff in h7.
  rewrite in_map_iff in h7.
  destruct h7 as [x [h7 h8]].
  apply h3 in h7; subst.
  apply no_dup_cons_nin in h1.
  contradiction.
  intro h7. 
  rewrite <- list_to_set_in_iff in h7.
  rewrite in_map_iff in h7.
  destruct h7 as [x [h7 h8]].
  apply h4 in h7; subst.
  apply no_dup_cons_nin in h2.
  contradiction.
  
  pose proof (f_equal (fun S => Subtract S (f a)) h5) as h7.
  simpl in h7.
  rewrite sub_add_same_nin in h7.
  pose proof (Add_intro2 _ (list_to_set (map f l)) (f a)) as h8.
  rewrite h5 in h8. 
  inversion h8 as [? h9 | ? h9]; subst.
  pose proof h9 as h10.
  rewrite <- list_to_set_in_iff in h10.
  rewrite in_map_iff in h10.
  destruct h10 as [x [h11 h12]].    
  specialize (ih (remove hdu x (a'::l')) (no_dup_cons _ _ h1)).
  hfirst ih.
  apply no_dup_remove.
  assumption.
  specialize (ih hf h3 h4).
  hfirst ih.
  rewrite h7.
  rewrite <- h11. 
  symmetry. 
  rewrite list_to_set_map_remove.
  reflexivity.
  assumption.
  right; auto.
  assumption.
  apply ih in hf0.
  rewrite length_remove_no_dup in hf0.
  simpl in hf0; auto with arith.
  assumption.
  right; auto.
  destruct h9.
  contradict h6; auto.
  intro h8.
  rewrite <- list_to_set_in_iff in h8.
  rewrite in_map_iff in h8.
  destruct h8 as [c [h8 h9]]; subst.
  apply h3 in h8; subst.
  apply no_dup_cons_nin in h1.
  contradiction.
Qed.


Lemma list_to_set_map_in_deps_eq_length :
  forall {T U V:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (l:list T) (l':list U)
         (f:fun_in_dep l V) (g:fun_in_dep l' V),
         NoDup l ->
         NoDup l' ->
         inj_dep f ->
         inj_dep g ->
         list_to_set (map_in_dep f) =
         list_to_set (map_in_dep g) ->
         length l =
         length l'.
  intros T U V hdt hdu hdv l.
  induction l as [|a l ih]; simpl.
  intros ? ? ? ? ? ? ? h1.
  symmetry in h1.
  apply empty_set_nil in h1.
  rewrite map_in_dep_eq_nil_iff in h1; subst; auto.
  intro l'. 
  destruct l' as [|a' l'].
  intros ? ? ? ? ? ? h1.
  simpl in h1.
  apply add_not_empty in h1.
  contradiction. 
  simpl.
  intros f g h1 h2 h3 h4 h5.
  destruct (hdv (f a (in_eq _ _)) (g a' (in_eq _ _))) as [h6 | h6].
  rewrite h6 in h5.
  apply add_nin_inj_eq in h5.
  f_equal.
  eapply ih; auto.
  apply no_dup_cons in h1; auto.
  apply no_dup_cons in h2; auto.
  eapply inj_from_cons.
  apply h3.
  eapply inj_from_cons.
  apply h4.
  assumption.
  intro h7.
  rewrite <- h6 in h7.
  rewrite <- list_to_set_in_iff in h7.
  rewrite in_map_in_dep_iff in h7.
  destruct h7 as [x [hi h7]].
  apply h3 in h7; subst.
  apply no_dup_cons_nin in h1.
  contradiction.
  intro h7. 
  rewrite <- list_to_set_in_iff in h7.
  rewrite in_map_in_dep_iff in h7.
  destruct h7 as [x [hi h7]].
  apply h4 in h7; subst.
  apply no_dup_cons_nin in h2.
  contradiction.
  
  pose proof (f_equal (fun S => Subtract S (f a (in_eq _ _))) h5) as h7.
  simpl in h7.
  rewrite sub_add_same_nin in h7.
  pose proof (Add_intro2 _ (list_to_set (map_in_dep (fun_from_cons f))) (f a (in_eq _ _))) as h8.
  rewrite h5 in h8. 
  inversion h8 as [? h9 | ? h9]; subst.
  pose proof h9 as h10.
  rewrite <- list_to_set_in_iff in h10.
  rewrite in_map_in_dep_iff in h10.
  destruct h10 as [x [hi h11]].    
  specialize (ih (remove hdu x (a'::l')) (fun_from_cons f)
  (fun_in_dep_remove hdu g x) (no_dup_cons _ _ h1)).
  hfirst ih.
  apply no_dup_remove.
  assumption.
  specialize (ih hf (inj_from_cons _ h3) (inj_dep_remove hdu g x h4)).
  hfirst ih.
  rewrite h7.
  rewrite <- h11. 
  symmetry. 
  erewrite list_to_set_map_in_dep_remove_inj.
  reflexivity.
  assumption.
  apply ih in hf0.
  rewrite length_remove_no_dup in hf0.
  simpl in hf0; auto with arith.
  assumption.
  right; auto.
  destruct h9.
  contradict h6; auto.
  intro h8.
  rewrite <- list_to_set_in_iff in h8.
  rewrite in_map_in_dep_iff in h8.
  destruct h8 as [c [hi h8]]; subst.
  apply h3 in h8; subst.
  apply no_dup_cons_nin in h1.
  contradiction.
Qed.




Definition list_fun (T U:Type) : Type :=
  list T -> list U.

Definition list_fun1 (T:Type) : Type := list_fun T T.


(*Gets an incresaing list of all the indices of a given element in 
  the given list.*)
Definition linds_occ' {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l) :=
    map_seg (fun (n:nat) (pflt:n < count_occ pfdt l x) =>
               lind_occ pfdt l n x pf).
 
  
Definition linds_occ {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) :
  list nat :=
  match (in_dec pfdt x l) with
    | left pf =>
      linds_occ' pfdt l x pf
    | right _ => nil
  end.

Lemma linds_occ_functional :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x x':T),
    x = x' ->
    l = l' ->
    linds_occ pfdt l  x  =
    linds_occ pfdt l' x'.
  intros; subst; auto.
Qed.

Lemma linds_occ_functional' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x x':T)
         (pfx:In x l) (pfx':In x' l'),
    x = x' ->
    l = l' ->
    linds_occ' pfdt l  x pfx =
    linds_occ' pfdt l' x' pfx'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

Lemma linds_occ_compat' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x :T)
         (pf:In x l),
    linds_occ pfdt l x = linds_occ' pfdt l x pf.
  intros T h1 l x h2.
  unfold linds_occ, linds_occ'.
  destruct in_dec as [h3 | h3].
  f_equal. apply functional_extensionality_dep.
  intro a.
  apply functional_extensionality.
  intro ha.
  f_equal.
  apply proof_irrelevance.
  contradiction.
Qed.
      
Lemma in_linds_occ_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (n:nat),
    In n (linds_occ pfdt l x) <->
    exists (m:nat) (pf:In x l),
      m < count_occ pfdt l x /\
      lind_occ pfdt l m x pf = n. 
  unfold linds_occ; intros T hdt l x occ. 
  lr_match_goal; fold hlr; destruct hlr as [hl | hr]; simpl.
  unfold linds_occ'.
  rewrite in_map_seg_iff.
  split; intros h1.
  destruct h1 as [m [h3 h2]]; subst.
  exists m, hl; auto.
  destruct h1 as [m h1].
  destruct h1 as [h1 [h2 h3]].
  pose proof (proof_irrelevance _ h1 hl); subst.
  exists m, h2; auto.
  split; intro h1; try destruct h1 as [? [? ?]]; contradiction.
Qed.

Lemma in_linds_occ_iff' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (n:nat) (pf:In x l),
    (In n (linds_occ' pfdt l x pf) <->
    exists (m:nat),
      m < count_occ pfdt l x /\
      lind_occ pfdt l m x pf = n).
  intros T h1 l x n h2; split; intro h3.
  rewrite <- linds_occ_compat' in h3.
  rewrite in_linds_occ_iff in h3.
  destruct h3 as [m [h3 [h4 h5]]]; subst.
  exists m.
  split; auto.
  f_equal; apply proof_irrelevance.
  rewrite <- linds_occ_compat'.
  destruct h3 as [m [h3 h4]]; subst.
  rewrite in_linds_occ_iff.
  exists m, h2.
  split; auto.
Qed.


Lemma in_lind_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l),
    In (lind pfdt l x pf) (linds_occ pfdt l x).
  intros T h1 l x h2.
  rewrite in_linds_occ_iff.
  exists 0, h2.
  split.
  rewrite (count_occ_In h1) in h2.
  apply gt_lt; auto.
  reflexivity.
Qed.

Lemma in_lind_occ_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l) (m:nat),
    m < count_occ pfdt l x ->
    In (lind_occ pfdt l m x pf) (linds_occ pfdt l x).
  intros T h1 l x h2 m hm.
  rewrite in_linds_occ_iff.
  exists m, h2.
  split; auto.
Qed.

  
Lemma in_linds_occ_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (n:nat),
    In n (linds_occ pfdt l x) ->
    n < length l.
  intros T h1 l x n h2.
  rewrite in_linds_occ_iff in h2.
  destruct h2 as [m [h2 [h3 h4]]].
  subst.
  apply lt_lind_occ.
Qed.


Lemma linds_occ_nil :
  forall {T:Type} (pfdt:type_eq_dec T) (x:T),
    linds_occ pfdt nil x = nil.
  unfold linds_occ; intros; destruct in_dec; auto.
Qed.


Lemma linds_occ_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    ~ In a l ->
    linds_occ pfdt l a = nil.
  intros T h1 l a h2.
  unfold linds_occ.
  destruct in_dec as [|h3].
  contradiction.
  reflexivity.
Qed.

Lemma linds_occ_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    In a l ->
    linds_occ pfdt l a <> nil.
  intros T h1 l a h2.
  unfold linds_occ.
  destruct in_dec.
  unfold linds_occ'.
  intro h3.
  rewrite map_seg_eq_iff in h3.
  destruct h3 as [h3 h4].
  simpl in h3.
  rewrite (count_occ_In h1) in h2.
  omega.
  contradiction.
Qed.

 
Lemma linds_occ_cons_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    linds_occ pfdt (a::l) a = 0::nil ++ lS (linds_occ pfdt l a).
  intros T h1 l a; unfold linds_occ, linds_occ'; simpl.  
  rewrite eq_dec_eq at 1; auto.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite eq_dec_eq at 1; auto; simpl.
  f_equal.
  rewrite length_lS.
  lr_match_goal'; simpl.
  rewrite length_map_seg; auto.
  pose proof (count_occ_In h1 l a) as h2.
  rewrite h2 in hr.
  apply not_lt_le in hr; apply le0 in hr; auto.
  exists hex.
  intros m h2; simpl.
  rewrite eq_dec_eq at 1; auto.
  destruct m as [|m]; auto.
  rewrite eq_dec_eq in h2 at 1; simpl in h2; auto.
  apply lt_S_n in h2.
  lr_match_goal'.
  gtermr.
  redterm1 y.
  pose proof (nth_lt_compat c m) as h3.
  hfirst h3.
  unfold c; rewrite length_lS, length_map_seg; auto.
  specialize (h3 hf); simpl in h3.
  fold c.
  erewrite nth_indep.
  rewrite <- h3 at 1.
  unfold c.
  rewrite nth_lt_lS, nth_lt_map_seg'; auto.
  assumption.
  pose proof (count_occ_In h1 l a) as h3.
  rewrite h3 in hr.
  apply not_lt_le in hr.
  apply le0 in hr.
  omega.
Qed.


Lemma linds_occ_cons_eq' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pf:In a l),
    linds_occ' pfdt (a::l) a (in_eq _ _) = 0::nil ++ lS (linds_occ' pfdt l a pf).
  intros T h1 l a h2.
  pose proof (linds_occ_cons_eq h1 l a) as h3.
  unfold linds_occ in h3.
  destruct in_dec as [h4 | h4], in_dec as [h5 | h5].
  simpl in h3.
  simpl.
  pose proof (proof_irrelevance _ h2 h5); subst.
  rewrite <- h3.
  f_equal. apply proof_irrelevance.
  contradiction.
  discriminate.
  discriminate.
Qed.

Lemma linds_occ_cons_eq_nin' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pf:~In a l),
    linds_occ' pfdt (a::l) a (in_eq _ _) = 0::nil.
  intros T h1 l.
  induction l as [|b l ih]; simpl.
  intros.
  unfold linds_occ'.
  rewrite map_seg_eq_iff.
  simpl.
  destruct (h1 a a) as [h2 | h2].
  exists eq_refl.
  intros m h3.
  assert (m = 0). omega. subst.
  reflexivity.
  contradict h2; auto.
  intros a h2.
  specialize (ih a).
  unfold linds_occ'.
  rewrite map_seg_eq_iff.
  ex_goal.
  simpl.
  destruct (h1 a a) as [h3 | h3].
  destruct (h1 b a) as [h4 | h4]; subst.
  contradict h2.
  left; auto.
  f_equal.
  assert (h5:~In a l).
  contradict h2.
  right; auto.
  rewrite (count_occ_In h1) in h5; omega.
  contradict h3; auto.
  exists hex.
  intros m h3 h4.
  pose proof h3 as h3'. rewrite hex in h3'.
  simpl in h3'.
  assert (h5:m = 0). omega.
  generalize h4. rewrite h5.
  intros h6.
  simpl.
  destruct (h1 a a) as [h7 | h7]; auto.
  contradict h7; auto.
Qed.


  
Lemma linds_occ_cons_neq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x:T),
    x <> a ->
    linds_occ pfdt (a::l) x = lS (linds_occ pfdt l x).
  intros T h1 l a x h2; unfold linds_occ, linds_occ'; simpl.
  destruct (h1 a x) as [h3 | h3]; subst.
  contradict h2; auto. 
  destruct (in_dec h1 x l) as [h4 | h4].
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_lS, length_map_seg; auto.
  exists hex.
  intros m h5; simpl.
  destruct (h1 x a) as [h6 | h6]; subst.
  contradict h2; auto.
  unfold lS.
  erewrite nth_lt_map.
  f_equal.
  rewrite nth_lt_map_seg.
  apply lind_occ_functional; auto.
  auto.
  Grab Existential Variables.
  assumption.
Qed.

Lemma linds_occ_cons_neq' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x:T)
         (pfx:In x l),
    let pfx' := in_cons _ _ _ pfx in
    x <> a ->
    linds_occ' pfdt (a::l) x pfx' = lS (linds_occ' pfdt l x pfx).
  intros T h1 l a x h2 h3 h4.
  pose proof (linds_occ_cons_neq h1 l a x h4) as h5.
  unfold linds_occ in h5.
  destruct in_dec as [h6 | h6], in_dec as [h7 | h7].
  pose proof (proof_irrelevance _ h2 h7).
  unfold h3. clear h3. subst.
  rewrite <- h5.
  f_equal. apply proof_irrelevance.
  contradiction.
  contradict h6; right; auto.
  contradict h6; right; auto.
Qed.


Lemma linds_occ_app_nin_l' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (pfxn:~In x l) (pfxin:In x l'),
    let pfx := in_app_r pfxin l in
    linds_occ' pfdt (l++l') x pfx =
    map (plus (length l)) (linds_occ' pfdt l' x pfxin).
  intros T h1 l l' x h2 h3 h4.
  unfold linds_occ'.
  rewrite (count_occ_In h1) in h2.
  rewrite (not_lt0_iff (count_occ h1 l x)) in h2.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite count_occ_app.
  rewrite map_length.
  rewrite length_map_seg.
  rewrite h2; auto with arith.
  exists hex.
  intros m h5 h6.
  pose proof h5 as h5'.
  rewrite count_occ_app in h5'.
  rewrite h2 in h5'; simpl in h5'.
  erewrite nth_lt_functional3.
  erewrite nth_lt_map.
  rewrite nth_lt_map_seg.
  symmetry.
  pose proof (lind_occ_app_ge_in2 h1 l l' x h3 m) as h7.
  hfirst h7.
  rewrite h2; auto with arith.
  specialize (h7 hf).
  simpl in h7.
  rewrite h2, <- minus_n_O in h7.
  pose proof (proof_irrelevance _ h4 (in_app_r h3 l));
    subst; auto.
  Grab Existential Variables.
  assumption.
  erewrite map_length.
  rewrite length_map_seg.
  assumption.
Qed.

Lemma linds_occ_app_nin_l :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (pfxn:~In x l) (pfxin:In x l'),
    let pfx := in_app_r pfxin l in
    linds_occ pfdt (l++l') x     =
    map (plus (length l)) (linds_occ pfdt l' x).
  intros; do 2 erewrite linds_occ_compat'; apply linds_occ_app_nin_l'; auto.
  Grab Existential Variables.
  assumption.
Qed.
  
Lemma linds_occ_app_nin_r' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (pfxin:In x l) (pfxn:~In x l'),
    let pfx := in_app_l pfxin l' in
    linds_occ' pfdt (l++l') x pfx =
    linds_occ' pfdt l x pfxin.
  intros T h1 l l' x h2 h3 h4.
  unfold linds_occ'.
  rewrite (count_occ_In h1) in h3.
  rewrite (not_lt0_iff (count_occ h1 l' x)) in h3.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite count_occ_app.
  rewrite h3, <- plus_n_O.
  rewrite length_map_seg; auto.
  exists hex.
  intros m h5; simpl.
  pose proof h5 as h5'. rewrite count_occ_app in h5'.
  rewrite h3, <- plus_n_O in h5'.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  pose proof (lind_occ_app_lt_in1 h1 l l' x _ h5') as h6.
  simpl in h6.
  pose proof (proof_irrelevance _ h4
                                (in_app_l (count_occ_in h1 l x m h5') l')) as h7.
  pose proof (proof_irrelevance _ h2 (count_occ_in h1 l x m h5')) as h8.
  rewrite h7, h8; auto.
  Grab Existential Variables.  
  assumption.
Qed.

Lemma linds_occ_app_nin_r :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (pfxin:In x l) (pfxn:~In x l'),
    let pfx := in_app_l pfxin l' in
    linds_occ pfdt (l++l') x =
    linds_occ pfdt l x.
  intros; do 2 erewrite linds_occ_compat'; apply linds_occ_app_nin_r'; auto.
  Grab Existential Variables.
  assumption.
Qed.



Lemma linds_occ_app_in' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (pfx:In x l) (pfx':In x l'),
    let pfxa := in_app_l pfx l' in
    linds_occ' pfdt (l++l') x pfxa =
    linds_occ' pfdt l x pfx ++ map (plus (length l)) (linds_occ' pfdt l' x pfx').
  intros T h1 l l' x h2 h3 h4.
  unfold linds_occ'.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite count_occ_app.
  rewrite app_length.
  rewrite map_length.
  do 2 rewrite length_map_seg; auto.
  exists hex.
  intros m h5 h6. 
  pose proof h6 as h6'.
  rewrite app_length in h6'.
  rewrite map_length in h6'.
  do 2 rewrite length_map_seg in h6'.
  destruct (lt_le_dec m (count_occ h1 l x)) as [h7 | h7].
  erewrite nth_lt_app1.
  erewrite nth_lt_functional3.
  erewrite nth_lt_map_seg.
  erewrite lind_occ_functional.
  rewrite lind_occ_app_lt_in1.
  apply lind_occ_functional; auto.
  reflexivity.
  reflexivity.   
  reflexivity. 

  erewrite nth_lt_app2.
  erewrite nth_lt_map.
  rewrite nth_lt_map_seg.
  rewrite length_map_seg.
  erewrite lind_occ_functional.
  erewrite lind_occ_app_ge_in2.
  reflexivity.
  assumption.
  reflexivity.
  reflexivity.
  reflexivity.

  Grab Existential Variables.
  rewrite length_map_seg.
  omega.
  rewrite length_map_seg.
  assumption.
  assumption.
  assumption.
  rewrite length_map_seg; auto.
Qed.


Lemma linds_occ_app_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (x:T)
         (pfx:In x l) (pfx':In x l'),
    let pfxa := in_app_l pfx l' in
    linds_occ pfdt (l++l') x =
    linds_occ pfdt l x ++ map (plus (length l)) (linds_occ pfdt l' x).
  intros; do 3 erewrite linds_occ_compat'; apply linds_occ_app_in'; auto.
  Grab Existential Variables.
  assumption.
  assumption.
Qed.


Lemma length_linds_occ' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfx:In x l),
    length (linds_occ' pfdt l x pfx) =
    count_occ pfdt l x.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros x h2.
  destruct (h1 a x) as [|h3]; subst.
  destruct (in_dec h1 x l) as [ha | ha]. 
  erewrite linds_occ_functional'. rewrite linds_occ_cons_eq' at 1.
  simpl.
  f_equal.
  rewrite length_lS.
  apply ih.
  reflexivity.
  reflexivity.

  erewrite linds_occ_functional'. rewrite linds_occ_cons_eq_nin'.
  simpl.
  f_equal.
  rewrite (count_occ_In h1) in ha.
  omega.
  apply ha.
  reflexivity.
  reflexivity.
  destruct h2 as [|h2]; subst.
  contradict h3; auto.
  specialize (ih _ h2).
  erewrite linds_occ_functional'. rewrite linds_occ_cons_neq'.
  rewrite length_lS.
  apply ih.
  contradict h3; rewrite h3; auto.
  reflexivity.
  reflexivity.
  Grab Existential Variables.
  assumption.
Qed.

Lemma length_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    length (linds_occ pfdt l x) =
    count_occ pfdt l x.
  intros T h1 l x.
  unfold linds_occ.
  destruct in_dec as [h2 | h2].
  apply length_linds_occ'.
  simpl.
  rewrite (count_occ_In h1) in h2.
  omega.
Qed.


Lemma linds_occ_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l) (ln:list nat),
    linds_occ pfdt l x = ln <->
    (exists pfeq : count_occ pfdt l x = length ln,
       forall (m : nat) (pflt : m < count_occ pfdt l x),
         lind_occ pfdt l m x pf =
         nth_lt ln m (lt_eq_trans m (count_occ pfdt l x) (length ln) pflt pfeq)).
  intros T h1 l x h2 ln.
  unfold linds_occ.
  destruct in_dec as [h3 | h3].
  unfold linds_occ'.
  rewrite map_seg_eq_iff; simpl.
  pose proof (proof_irrelevance _ h2 h3); subst.
  tauto.
  contradiction.
Qed.

Lemma nin_linds_occ_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:~In x l) (ln:list nat),
    linds_occ pfdt l x = ln <->
    ln = nil.
  intros T h1 l x h2 ln.
  apply (linds_occ_nin h1) in h2.
  rewrite h2.
  intuition.
Qed.


Lemma in_linds_occ_nth_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (r:nat)
         (pf:r < length l),
    In r (linds_occ pfdt l (nth_lt l r pf)).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros r h2.
  destruct r as [|r]; simpl.
  rewrite linds_occ_cons_eq.
  left; auto.
  apply lt_S_n in h2.
  destruct (h1 a (nth r l a)) as [h3 | h3].
  rewrite <- h3.
  rewrite linds_occ_cons_eq.
  right.
  rewrite app_nil_l. 
  apply in_map. 
  specialize (ih r h2).
  rewrite nth_lt_compat in ih.
  erewrite nth_indep in ih.
  rewrite <- h3 in ih at 1; auto.
  assumption.
  specialize (ih r h2).
  rewrite linds_occ_cons_neq; auto.
  apply in_map.
  erewrite nth_lt_compat in ih. 
  erewrite nth_indep in ih.
  apply ih.
  assumption.
Qed.

Lemma in_linds_occ_nth_lt_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    In x l ->
    forall (i j:nat)
           (pfii:In i (linds_occ pfdt l x))
           (pfij:In j (linds_occ pfdt l x)),
      let pflti := in_linds_occ_lt pfdt _ _ _ pfii in
      let pfltj := in_linds_occ_lt pfdt _ _ _ pfij in
      nth_lt l i pflti =
      nth_lt l j pfltj.
  intros T h1 l x h2 i j h3 h4 h5 h6.
  pose proof h3 as h3'.
  pose proof h4 as h4'.
  rewrite in_linds_occ_iff in h3'.
  rewrite in_linds_occ_iff in h4'.
  destruct h3' as [m [h9 [h7 h8]]].
  destruct h4' as [m' [h9' [h7' h8']]].
  gen0.
  rewrite <- h8.
  intro h10.
  symmetry.
  gen0.
  rewrite <- h8'.
  intro h11.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat; auto.
Qed.

Lemma in_linds_occ_nth_lt_eq' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    In x l ->
    forall (i j:nat)
           (pfii:In i (linds_occ pfdt l x))
           (pfij:In j (linds_occ pfdt l x))
           (pflti:i < length l)
           (pfltj:j < length l),
      nth_lt l i pflti =
      nth_lt l j pfltj.
  intros T h1 l x h2 i j h3 h4 h5 h6.
  erewrite nth_lt_functional3.
  rewrite in_linds_occ_nth_lt_eq at 1.
  apply nth_lt_functional3.
  apply h2.
  Grab Existential Variables.
  apply h4.
  assumption.
Qed.

  
Lemma in_linds_occ_nth_lt' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (r:nat)
         (pf:r < length l),
    In r (linds_occ' pfdt l (nth_lt l r pf) (in_nth_lt _ _ _)).
  intros T h1 l r h2.
  pose proof (in_linds_occ_nth_lt h1 l r h2) as h3.
  unfold linds_occ in h3.
  destruct in_dec as [h4 | h4].
  pose proof (proof_irrelevance _ h4 (in_nth_lt _ _ _)).
  subst.
  assumption.
  red in h3.
  contradiction.
Qed.

Lemma nth_lt_linds_occ' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l)
         (m:nat) (pfm:m < length (linds_occ' pfdt l x pf)),
    nth_lt (linds_occ' pfdt l x pf) m pfm = lind_occ pfdt l m x pf.
  intros T h1 l x h2 m h3.
  unfold linds_occ'.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg; auto.
  Grab Existential Variables.
  rewrite length_linds_occ' in h3; auto.
Qed.

Lemma nth_lt_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l)
         (m:nat) (pfm:m < length (linds_occ pfdt l x)),
    nth_lt (linds_occ pfdt l x) m pfm = lind_occ pfdt l m x pf.
  intros.
  gen0.
  erewrite linds_occ_compat'.
  intro h1.
  rewrite nth_lt_linds_occ'.
  reflexivity.
Qed.

Lemma hd_prime_linds_occ' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfin:In x l)
         (pfn:linds_occ' pfdt l x pfin <> nil),
    hd' (linds_occ' pfdt l x pfin) pfn =
    lind pfdt l x pfin.
  intros T h1 l x h2 h3.
  rewrite <- nth_lt0.
  rewrite nth_lt_linds_occ'.
  unfold lind; auto.
Qed.


Lemma hd_linds_occ' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfin:In x l)
         (pfn:linds_occ' pfdt l x pfin <> nil) (d:nat),
    hd d (linds_occ' pfdt l x pfin) =
    lind pfdt l x pfin.
  intros; rewrite (hd_hd_compat' _  pfn).
  rewrite hd_prime_linds_occ'; auto.
Qed.

Lemma hd_prime_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfin:In x l)
         (pfn:linds_occ pfdt l x  <> nil),
    hd' (linds_occ pfdt l x) pfn =
    lind pfdt l x pfin.
  intros T h1 l x h2.
  rewrite (linds_occ_compat' h1 l x h2).
  intro h3.
  apply hd_prime_linds_occ'.
Qed.

Lemma hd_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pfin:In x l)
         (pfn:linds_occ pfdt l x <> nil) (d:nat),
    hd d (linds_occ pfdt l x) =
    lind pfdt l x pfin.
  intros; rewrite (hd_hd_compat' _ pfn).
  erewrite hd_prime_linds_occ; auto.
Qed.
  
    


Lemma no_dup_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    NoDup (linds_occ pfdt l x).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intro x.
  rewrite linds_occ_nil.
  constructor.
  intro x.
  destruct (h1 a x) as [|h2]; subst.
  rewrite linds_occ_cons_eq.
  simpl.
  constructor.
  intro h2.
  rewrite in_lS_iff in h2.
  destruct h2 as [? [? ?]]; discriminate.
  apply no_dup_lS; auto.
  rewrite linds_occ_cons_neq.
  apply no_dup_lS; auto.
  contradict h2; auto.
Qed.

Lemma no_dup_linds_occ':
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l),
    NoDup (linds_occ' pfdt l x pf).
  intros. rewrite <- linds_occ_compat'.
  apply no_dup_linds_occ.
Qed.

Lemma impl_count_occ_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x y:T),
    list_to_set (linds_occ pfdt l x) = 
    list_to_set (linds_occ pfdt l y) ->
    count_occ pfdt l x = count_occ pfdt l y.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros x y h2.
  destruct (h1 a x) as [|h3]; subst.
  destruct (h1 x y) as [|h4]; subst.
  reflexivity.
  rewrite linds_occ_cons_eq, linds_occ_cons_neq in h2.
  simpl in h2.
  pose proof (Add_intro2 _ (list_to_set (lS (linds_occ h1 l x))) 0) as h5.
  rewrite h2 in h5.
  rewrite <- list_to_set_in_iff in h5.
  rewrite in_lS_iff in h5.
  destruct h5 as [? [? ?]]; discriminate.
  apply neq_sym; auto.
  destruct (h1 a y) as [|h4]; subst.
  rewrite linds_occ_cons_eq, linds_occ_cons_neq in h2.
  simpl in h2.
  pose proof (Add_intro2 _ (list_to_set (lS (linds_occ h1 l y))) 0) as h5.
  rewrite <- h2 in h5.
  rewrite <- list_to_set_in_iff in h5.
  rewrite in_lS_iff in h5.
  destruct h5 as [? [? ?]]; discriminate.
  apply neq_sym; auto.
  do 2 rewrite linds_occ_cons_neq in h2; auto.
  apply ih.
  unfold lS in h2.
  do 2 rewrite map_im_compat in h2.
  apply im_inj_inj in h2; auto.
  red; intros; apply S_inj; auto.
Qed.


Lemma all_lind_occ_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x x':T)
           (pfx:In x l) (pfx':In x' l),
    (count_occ pfdt l x = count_occ pfdt l x' /\
     (forall (n:nat),
        n < count_occ pfdt l x ->
        lind_occ pfdt l n x  pfx =
        lind_occ pfdt l n x' pfx')) <->
    x = x'.
  intros T h1 l.
  induction l as [|a l ih]; simpl.  
  intros; contradiction.
  split; intro h2.
  destruct h2 as [h2 h2'].
  destruct (h1 a x) as [h3 | h3], (h1 a x') as [h4 | h4]; simpl; subst; auto.

  destruct (h1 x x) as [h5 | h5], (h1 x' x) as [h6 | h6]; simpl; auto.
  specialize (h2' 0 (lt_0_Sn _)).
  simpl in h2'.
  discriminate.
  pose proof (h5 eq_refl); contradiction.
  destruct (h1 x x') as [h4 | h4]; subst; auto.
  destruct (h1 x' x') as [h5 | h5].
  specialize (h2' 0).
  pose proof (neq_in_cons _ _ _ h4 pfx) as h6.
  hfirst h2'.
  pose proof (count_occ_In h1 l x) as h7.
  rewrite h7 in h6; auto with arith.
  specialize (h2' hf).
  simpl in h2'.
  discriminate.
  pose proof (h5 eq_refl); contradiction.
  apply neq_sym in h3.
  apply neq_sym in h4.
  pose proof (neq_in_cons _ _ _ h3 pfx) as h5.
  pose proof (neq_in_cons _ _ _ h4 pfx') as h6.
  destruct (h1 x a) as [h7 | h7], (h1 x' a) as [h8 | h8]; subst.
  contradict h3; auto.
  contradict h3; auto.
  contradict h4; auto.
  rewrite <- ih.
  split; auto.
  intros n h9.
  specialize (h2' _ h9).
  apply S_inj in h2'; auto.
  apply h2'.
  subst.
  destruct (h1 a x') as [h2 | h2], (h1 x' a) as [h3 | h3]; simpl; auto.
  subst.
  contradict h3; auto. 
  split; auto.
  intros; auto.
  f_equal. f_equal. f_equal. apply proof_irrelevance.
Qed.


Lemma linds_occ_firstn :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (n:nat)
         (pfin:In x l) (pfin':In x (firstn n l)),
    linds_occ pfdt (firstn n l) x =
    firstn (count_occ pfdt (firstn n l) x) (linds_occ pfdt l x).
  intros T h1 l x n h2 h3.
  unfold linds_occ.
  destruct in_dec as [h4 | h4].
  destruct in_dec as [h5 | h5].
  unfold linds_occ'.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_firstn; auto.
  rewrite length_map_seg.
  apply le_count_occ_sub_list.
  apply sub_list_firstn.
  exists hex.
  intros m h6 h7.  
  rewrite nth_lt_firstn_eq.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg at 1.
  rewrite lind_occ_firstn.
  f_equal.
  apply proof_irrelevance.
  assumption.
  contradiction.
  contradiction.
  Grab Existential Variables.
  eapply lt_le_trans.
  apply h6. 
  apply le_count_occ_sub_list.
  apply sub_list_firstn.
Qed.

Lemma linds_occ_firstn_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (m n:nat),
    m < n ->
    forall (pfinm:In x (firstn m l)) (pfinn:In x (firstn n l)),
      linds_occ pfdt (firstn m l) x =
      firstn (count_occ pfdt (firstn m l) x) (linds_occ pfdt (firstn n l) x). 
  intros T h1 l x m n h2 h3 h4.  
  rewrite (firstn_lt_nest _ _ _ h2) at 1; auto. 
  rewrite linds_occ_firstn at 1; auto.
  f_equal. f_equal.
  rewrite <- firstn_lt_nest; auto.
  rewrite <- firstn_lt_nest; auto.
Qed.

Lemma in_linds_occ_firstn_le :
  forall (T:Type) (pfdt:type_eq_dec T) (l:list T) 
         (x:T) (r m n:nat) (pfsm:S m <= length l),
    m <= n ->
    r < m ->
    In r (linds_occ pfdt (firstn m l) x) ->
    In r (linds_occ pfdt (firstn n l) x).
  intros T h1 l x r m n h2 hle h3 h4.
  rewrite in_linds_occ_iff in h4.
  destruct h4 as [o [h4 [h5 h6]]]; subst.
  rewrite in_linds_occ_iff.
  exists o.
  ex_goal.
  eapply in_firstn_le.
  apply hle.
  assumption.
  exists hex.
  split.
  eapply lt_le_trans.
  apply h5.
  apply le_count_occ_sub_list.
  apply sub_list_firstn_le; auto with arith.
  rewrite lind_occ_firstn; auto with arith.
  rewrite lind_occ_firstn; auto with arith.
  apply lind_occ_functional; auto.
  eapply lt_le_trans.
  apply h5.
  apply le_count_occ_sub_list.
  apply sub_list_firstn_le; auto with arith.
Qed.



Lemma in_nin_linds_occ_firstn_S :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (r m:nat) (pfsm:S m <= length l),
    r < m ->
    In x (firstn m l) ->
    ~In (nth_lt l m pfsm) (firstn m l) ->
    In r (linds_occ pfdt (firstn (S m) l) x) ->
    In r (linds_occ pfdt (firstn m l) x).
  intros T h1 l x r m h2 h3 h4 h5 h6.
  rewrite in_linds_occ_iff in h6.
  destruct h6 as [o [h6 [h7 h8]]].
  rewrite in_linds_occ_iff.
  subst.
  rewrite lind_occ_firstn in h3; auto with arith.
  pose proof h6 as h6'.
  rewrite (firstn_S_eq _ _ h2) in h6', h7.
  rewrite count_occ_app in h7.
  simpl in h7.
  destruct (h1 (nth_lt l m h2) x) as [h8 | h8]; subst.
  contradiction.
  rewrite <- plus_n_O in h7.
  apply in_app_or in h6'.
  destruct h6' as  [h9 | h9].
  exists o, h9.
  split; auto with arith.
  rewrite lind_occ_firstn.
  rewrite lind_occ_firstn.
  apply lind_occ_functional; auto.
  eapply lt_le_trans.
  apply h7.
  apply le_count_occ_sub_list.
  apply sub_list_firstn_le; auto with arith.
  assumption.
  destruct h9; subst.
  contradiction.
  simpl in H.
  contradiction.
Qed.

 
Lemma in_nin_linds_occ_firstn_S' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) 
         (m r:nat) (pfsm:S m <= length l),
    r < S m ->
    x <> nth_lt l m pfsm ->
    In r (linds_occ pfdt (firstn (S m) l) x) ->
    In r (linds_occ pfdt (firstn m l) x).
  intros T h1 l x m r h2 h3 h4 h5.
  rewrite in_linds_occ_iff in h5.
  rewrite in_linds_occ_iff.
  destruct h5 as [o [h6 [h7 h8]]].
  pose proof h6 as h6'.
  apply (in_firstn_S_dec h1  _ _ h2)  in h6'.
  destruct h6' as [h9 | h9].
  exists o.
  exists h9.
  
  rewrite (firstn_S_eq _ _ h2) in h7.
  rewrite count_occ_app in h7.
  simpl in h7.
  lr_if h7; fold hlr in h7; destruct hlr as [hl | hr]; subst.
  contradict h4; auto.
  rewrite <- plus_n_O in h7; auto.
  rewrite lind_occ_firstn; auto.
  split; auto.
  rewrite lind_occ_firstn; auto.
  apply lind_occ_functional; auto.
  eapply lt_le_trans.
  apply h7.
  apply le_count_occ_sub_list.
  apply sub_list_firstn_le; auto with arith.
  subst.
  contradict h4; auto.
Qed.


Lemma nin_lind_nth_lt_firstn_S :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (m:nat)
         (pfsm:S m <= length l)
         (pfi:In (nth_lt l m pfsm) (firstn (S m) l)),
      ~In (nth_lt l m pfsm) (firstn m l) ->
      lind pfdt (firstn (S m) l) (nth_lt l m pfsm) pfi = m.
  intros T h1 l m h2 h3 h4. 
  rewrite lind_eq_iff.
  split. 
  rewrite nth_lt_firstn_eq.
  apply nth_lt_functional3.
  erewrite <- firstn_lt_nest; auto with arith.
  Grab Existential Variables.
  rewrite length_firstn; auto with arith.
Qed.

Lemma map_linds_occ_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    map (linds_occ pfdt l) l =
    map_in_dep (linds_occ' pfdt l).
  intros T h1 l.
  rewrite map_eq_iff.
  ex_goal.
  rewrite length_map_in_dep; auto.
  exists hex.
  intros m h2.
  erewrite nth_lt_map_in_dep, linds_occ_compat'.
  reflexivity.
Qed.

Lemma map_linds_occ_compat' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T)
         (pfi:incl l' l),
    map (linds_occ pfdt l) l' =
    map_in_dep (fun (x:T) (pfix:In x l') => (linds_occ' pfdt _ _ (pfi _  pfix))).
  intros T h1 l l' h2.
  rewrite map_eq_iff.
  ex_goal.
  rewrite length_map_in_dep; auto.
  exists hex.
  intros m h3.
  erewrite nth_lt_map_in_dep, linds_occ_compat'.
  reflexivity.
Qed.

Definition remove_sig {T} (l:list T) (a:T) : Type :=
  {x:{t:T | t <> a} | In (proj1_sig x) l}.


Definition remove_sig_list {T l a} (pfdt:type_eq_dec T) (pfn:~In a l) :
  list {x:T | x <> a} :=
  map_in_dep (fun x (pfx:In x l) =>
                match (pfdt x a) with
                  | left pfe =>
                    let pfx := in_subst1 pfe pfx in
                    False_rect _ (pfn pfx)
                  | right pfn => exist (fun c => c <> a) _ pfn
                end).

                  


Definition remove_mask {T:Type} (pfdt:type_eq_dec T) (l:list T)
           (x:T) : bv :=
  inds_to_bv l (linds_occ pfdt l x).


Lemma length_remove_mask :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    length (remove_mask pfdt l x) = length l.
  intros T h1 l x.
  unfold remove_mask, inds_to_bv.
  apply length_map_seg.
Qed.



Lemma lind_interp_bv_ge_hd :
  forall {T:Type} (pfdt:type_eq_dec T)
         (x:T) (l:list T)
         (v:list bool)
         (pfe:length l = count_occ bool_eq_dec v false)
         (pfxnin: ~In x l)
         (pfxin:In x (interp_bv x v l pfe))
         (pfn:bv1s v <> nil),
    lind pfdt (interp_bv x v l pfe) x pfxin >=
    hd' (bv1s v) pfn.
  intros T h1 x l v.
  revert l x.
  induction v as [|b v ih].
  intros ? ? ? ? ? h5.
  unfold bv1s in h5; simpl in h5.
  contradict h5; auto.
  intros l x h2 h3 h4 h5.
  destruct b.
  generalize h5.
  rewrite bv1s_cons_true; simpl.
  intros; auto with arith.
  generalize h5.
  rewrite bv1s_cons_false; simpl.
  intro h6.
  unfold lS.
  rewrite hd_map'.
  generalize h4.
  pose proof h2 as h2'.
  rewrite count_occ_cons_eq in h2'; auto.
  apply length_eq_S in h2'.
  rewrite (interp_bv_false _ _ _ _ h2').
  intro h7. 
  destruct h7 as [h7 | h7]; subst.
  contradict h3.
  apply in_hd'.
  rewrite lind_cons_neq.
  apply le_n_S.
  apply ih.
  contradict h3.
  apply in_tl in h3; auto.
  intro; subst.
  contradict h3.
  apply in_hd'.
Qed.

  
Lemma lind_interp_bv :
  forall {T:Type} (pfdt:type_eq_dec T)
         (x:T) (l:list T)
         (v:list bool)
         (pfe:length l = count_occ bool_eq_dec v false)
         (pfxnin: ~In x l)
         (pfxin:In x (interp_bv x v l pfe))
         (pfn:bv1s v <> nil),
    lind pfdt (interp_bv x v l pfe) x pfxin =
    hd' (bv1s v) pfn.
  intros T h1 x l v h2 h3 h4 h5. 
  pose proof (in_bv1s_lt v _ (in_hd' _ h5)) as h6.
  pose proof (lind_eq_iff h1 _ _ h4  (hd' _ h5)) as h7.
  hfirst h7.
  rewrite length_interp_bv; auto.
  specialize (h7 hf). 
  rewrite h7.
  split. 
  apply nth_lt_interp_bv; auto.
  apply in_hd'.  
  intro h8.
  rewrite (in_firstn_iff' h1) in h8.
  destruct h8 as [h8 h9].
  pose proof (proof_irrelevance _ h8 h4); subst.
  nhl h7.
  intro h10.
  rewrite h10 in h9; apply lt_irrefl in h9; contradiction. 
  rewrite h7 in hl. 
  contradict hl.
  pose proof (lind_interp_bv_ge_hd h1 x l v h2 h3 h4 h5).
  omega.
Qed.


Lemma lind_occ_interp_bv :
  forall {T:Type} (pfdt:type_eq_dec T)
         (x:T)
         (v:bv) (l:list T)
         (occ:nat)
         (pfeq:length l = count_occ bool_eq_dec v false)
         (pfnin: ~In x l)
         (pfin: In x (interp_bv x v l pfeq))
         (pfo:occ < length (bv1s v)),
    lind_occ pfdt (interp_bv x v l pfeq) occ x pfin =
    nth_lt (bv1s v) occ pfo.
  intros T pfdt x v. revert x.
  induction v as [|b v ih].
  intros; contradiction.
  intros x l occ h1 hn h2 h3.
  destruct b.
  gen0.
  rewrite interp_bv_true.
  intro h5.
  pose proof h1 as h1'.
  rewrite count_occ_cons_neq in h1'.  
  destruct (in_dec bool_eq_dec true v) as [h6|h6]. 
  pose proof (in_interp_bv x _ _ h1' h6) as h7.
  destruct occ as [|occ].
  symmetry.
  gen0.  
  rewrite bv1s_cons_true. 
  intro ha.
  simpl.
  rewrite eq_dec_eq at 1; auto.
  symmetry.
  gen0.
  rewrite bv1s_cons_true.
  intro h8.
  rewrite nth_lt_S. 
  pose proof h7 as h7'.

  term0 h5.
  redterm0 c.
  redterm0 c0.
  pose proof (proof_irrelevance _ c1 h1') as he.
  unfold c1 in he.
  rewrite <- he in h7'.
  rewrite (lind_occ_cons_in_eq _ _ _ _ h7' h5).
  simpl.
  rewrite nth_lt_lS.
  f_equal.
  simpl in h8.
  pose proof h8 as h8'.
  apply lt_S_n in h8'.
  rewrite length_lS in h8'. 
  specialize (ih x l occ _ hn h7' h8').
  rewrite ih at 1.
  apply nth_lt_functional3.
  auto with arith.
  destruct (zerop occ) as [|hz]; subst. 
  gterml.
  redterm3 x0.
  redterm0 c.
  fold c0. 
  pose proof (lind_cons_same pfdt c0 x h5) as h7. 
  rewrite h7 at 1.  
  simpl; auto.  
  
  pose proof (nin_true_bv1s _ h6) as h9.
  symmetry.
  gen0.
  rewrite bv1s_cons_true, h9.
  intro h10.
  simpl in h10.
  assert (h11:occ = 0).
    omega.
  gen0.
  rewrite h11.
  intro h12.
  simpl.
  assert (h13:pfdt x x = left (eq_refl x)).
    destruct (pfdt x x); f_equal; try apply proof_irrelevance.
    contradict n; auto.
  rewrite h13; auto.
  intro; discriminate.
  gen0.
  pose proof h1 as h1'.
  rewrite count_occ_cons_eq in h1'; auto.
  pose proof h1' as h4.
  apply length_eq_S in h4.
  rewrite (interp_bv_false _ _ _ h1 h4).
  intro h5.
  symmetry.
  assert (hneq: x <> hd' l h4).
    intro h6; subst; contradict hn; apply in_hd'.
  gen0.
  rewrite bv1s_cons_false.
  intro h6.
  unfold lS.
  pose proof h6 as h6'.
  rewrite length_lS in h6'.
  rewrite (nth_lt_map _ _ _ h6').
  destruct h5 as [h5 | h5]; subst.
  contradict hneq; auto.
  destruct occ as [|occ]; simpl. 
  assert (he:pfdt x (hd' l h4) = right hneq).
    destruct (pfdt x (hd' l h4)) as [h8 | h8].
    contradiction.
    f_equal; apply proof_irrelevance.
  rewrite he.
  f_equal.
  pose proof (lind_interp_bv pfdt x (tl l) v (length_count_occ_false_tl x l v h1)) as h8.
  hfirst h8.
  contradict hn.
  apply in_tl in hn; auto.
  specialize (h8 hf h5).
  hfirst h8.
  apply lt_length in h6'; auto.
  specialize (h8 hf0).
  unfold lind in h8.
  erewrite lind_occ_functional.
  rewrite h8 at 1.
  rewrite nth_lt_0_hd'.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  reflexivity.
  reflexivity.
  assert (h7:pfdt x (hd' l h4) = right hneq).
    destruct (pfdt x (hd' l h4)).
    contradiction.
    f_equal; apply proof_irrelevance.
  rewrite h7.
  f_equal.
  symmetry.
  apply ih.
  contradict hn.
  apply in_tl; auto.
Qed.


Lemma interp_bv_compat :
  forall {T:Type} (pfdt:type_eq_dec T) (x:T)
         (v:bv) (l:list T)
         (pfe:length l = count_occ bool_eq_dec v false)
         (i:nat) (pf:i < length v),
    ~In x l ->
    bv1s v = linds_occ pfdt (interp_bv x v l pfe) x.
  intros T h1 x v l h2 i h3 h4.
  symmetry.
  destruct (in_dec bool_eq_dec true v) as [h5|h5].
  pose proof (in_interp_bv x v l h2 h5) as h6.
  rewrite (linds_occ_eq_iff _ _ _ h6).
  ex_goal.
  apply count_occ_interp_bv; auto.
  exists hex.
  intros occ h7.
  pose proof h7 as h8.
  rewrite hex in h8.
  apply lind_occ_interp_bv; auto. 
  rewrite nin_true_interp_bv, nin_true_bv1s; auto.
  rewrite nin_linds_occ_eq_iff; auto.
Qed.

(*Returns an analogous list to [l], which has a [true]
  at values at x, and a [false] at all others.*)
Fixpoint list_to_bv {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) : bv :=
  match l with
    | nil => nil
    | a::l' =>
      let v := list_to_bv pfdt l' x in
      if (pfdt x a) then
        (true :: v) else
        (false :: v)
  end.

Lemma count_occ_list_to_bv :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    count_occ bool_eq_dec (list_to_bv pfdt l x) false =
    length (remove pfdt x l).
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intro x.
  destruct (h1 x a) as [|h2]; subst.
  rewrite count_occ_cons_neq.
  apply ih.
  intro; discriminate.
  rewrite count_occ_cons_eq; auto.
  simpl.
  f_equal; auto.
Qed.

Lemma list_to_bv_cons_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    list_to_bv pfdt (x::l) x =
    true :: list_to_bv pfdt l x.
  intros T h1 l x.
  simpl.
  destruct (h1 x x) as [h3 | h3]; auto.
  contradict h3; auto.
Qed.

Lemma list_to_bv_cons_neq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x a:T),
    x <> a ->
    list_to_bv pfdt (a::l) x =
    false :: list_to_bv pfdt l x.
  intros T h1 l x a h2.
  simpl.
  destruct (h1 x a) as [h3 | h3].
  contradiction.
  reflexivity.
Qed.


Lemma interp_bv_list_to_bv_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (x:T)
         (l:list T),
    let pfeq := eq_sym (count_occ_list_to_bv pfdt l x) in
         interp_bv x (list_to_bv pfdt l x) (remove pfdt x l) pfeq = l.
  intros T h1 x l.
  revert x.
  induction l as [|a l ih].
  intros x.
  unfold interp_bv; simpl; auto.
  intros x h2.
  unfold h2.
  clear h2.
  destruct (h1 x a) as [|h3]; subst.
  gen0.
  rewrite list_to_bv_cons_eq.
  rewrite remove_cons_eq.
  intro h2.
  rewrite interp_bv_true.
  f_equal.
  specialize (ih a).
  simpl in ih.
  erewrite f_equal_dep.
  rewrite ih at 1; auto.
  reflexivity.
  gen0.
  rewrite list_to_bv_cons_neq, remove_cons_neq; auto.
  intro h4.
  erewrite interp_bv_false.
  simpl.
  f_equal.
  specialize (ih x).
  simpl in ih.
  erewrite f_equal_dep.
  rewrite ih at 1; auto.
  reflexivity.

  Grab Existential Variables.
  apply cons_neq_nil.
Qed.


     
Lemma remove_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T)
         (x:T),
    let v := list_to_bv pfdt l x in
    remove pfdt x l = l' <->
    exists (pf:~In x l') (pfe:length l' = count_occ bool_eq_dec v false),
      l = interp_bv x v l' pfe.
Proof.
  intros T h1 l l' x v; split; intro h3; subst.  
  ex_goal.
  apply remove_In.
  exists hex.
  unfold v.
  ex_goal.
  rewrite count_occ_list_to_bv; auto.
  exists hex0.
  symmetry.
  erewrite f_equal_dep.
  apply interp_bv_list_to_bv_remove; auto.
  reflexivity.
  unfold v in h3.
  clear v.
  destruct h3 as [h3 h4].
  destruct h4 as [h4 h5].
  rewrite h5.
  apply remove_interp_bv; auto.
Qed.


(*"List Equivalence class" or "List Equal class".*)
Definition lec {T:Type} {l:list T} {i:nat}
           (pfdt:type_eq_dec T)
           (pf:i < length l) :
  list nat :=
  linds_occ pfdt l (nth_lt l i pf).

(*This increments each element of the list equivalence class
  of a given representative index.*)
Definition lec_S {T:Type} {l:list T} {i:nat}
           (pfdt:type_eq_dec T)(pf:i < length l) :
  list nat :=
  lS (lec pfdt pf).

Definition lec_cons {T:Type} {l:list T}  {i:nat}
           (pfdt:type_eq_dec T)
           (pf:i < length l)
           (a:T) :
  list nat :=
  let pf' := lt_n_S _ _  pf in
  lec pfdt pf' (l := a::l).

Lemma length_lec :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (i:nat) (pf:i < length l),
    length (lec pfdt pf) = count_occ pfdt l (nth_lt _ _ pf).
  intros T h1 i h2 h3.
  unfold lec.
  rewrite length_linds_occ; auto.
Qed.



Lemma lecs_eq :
  forall {T:Type} (pfdt:type_eq_dec T) 
         (l l':list T) (m:nat) (pfm:m < length l)
         (pf:l = l'),
    let pfe := length_subst l l' pf in
    let pfm' := lt_eq_trans _ _ _ pfm pfe in
    lec pfdt pfm = lec pfdt pfm'.
  simpl; intros; subst; f_equal; apply proof_irrelevance.
Qed.


Lemma linds_occ_nth_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (i:nat) (pfi:i < length l),
    linds_occ pfdt l (nth_lt _ _ pfi) =
    lec pfdt pfi.
  unfold lec; auto.
Qed.

Lemma no_dup_lec :
  forall {T:Type} {l:list T} {i:nat} (pfdt:type_eq_dec T) 
         (pfi:i < length l),
    NoDup (lec pfdt pfi).
  unfold lec; intros; apply no_dup_linds_occ.
Qed.




(*See past [sort] below for a sorted version of [linds]
  called [sorted_linds].*)
(*These functions return a family of lists of the indices of each 
  element in [l].*)
Definition linds {T:Type} (pfdt:type_eq_dec T) (l:list T) : faml nat :=
  map (linds_occ pfdt l) (list_singularize pfdt l).

Definition linds' {T:Type} (pfdt:type_eq_dec T) (l:list T) : faml nat :=
  let f := fun_in_dep_incl _ _ (linds_occ' pfdt l) (incl_list_singularize pfdt _)  in
  map_in_dep f.

Lemma linds_functional :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    l = l' ->
    linds pfdt l =
    linds pfdt l'.
  intros; subst; auto.
Qed.

Lemma linds_functional' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T),
    l = l' ->
    linds' pfdt l =
    linds' pfdt l'.
  intros; subst; auto.
Qed.

  
Lemma linds_compat' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    linds pfdt l =
    linds' pfdt l.
  intros T h1 l.
  unfold linds, linds'.
  rewrite (map_eq_iff (list_singularize h1 l) (linds_occ h1 l)).
  ex_goal.
  rewrite length_map_in_dep; auto.
  exists hex.
  intros n h2.
  erewrite nth_lt_map_in_dep.
  unfold fun_in_dep_incl.
  unfold linds_occ.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  f_equal. apply proof_irrelevance.
  contradict hr.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
Qed.

Lemma linds_nil :
  forall {T:Type} (pfdt:type_eq_dec T),
    linds pfdt nil = nil.
  intros; unfold linds; simpl; auto.
Qed.

Lemma linds_nil' :
  forall {T:Type} (pfdt:type_eq_dec T),
    linds' pfdt nil = nil.
  intros; unfold linds'; simpl; auto.
Qed.

Lemma linds_eq_nil_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    linds pfdt l = nil <-> l = nil.
  intros T h1 l.
  destruct l as [|a l].
  rewrite linds_nil; tauto.
  unfold linds; simpl.
  destruct (in_dec h1 a l).
  split; intro h2.
  apply map_eq_nil in h2.
  apply list_singularize_nil_iff in h2; subst; auto.
  discriminate.
  rewrite list_singularize_nin; auto.
  simpl.
  rewrite linds_occ_cons_eq.
  split; intro; discriminate.
Qed.

Lemma linds_eq_nil_iff' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    linds' pfdt l = nil <-> l = nil.
  intros; rewrite <- linds_compat'; apply linds_eq_nil_iff; auto.
Qed.


Lemma nin_nil_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T), 
    ~In nil (linds pfdt l).
  intros T h1 l h2.
  unfold linds in h2.
  rewrite in_map_iff in h2.
  destruct h2 as [x h2].
  destruct h2 as [h2 h3].
  unfold linds_occ in h2.
  destruct in_dec as [h4 | h4].
  unfold linds_occ' in h2.
  apply map_seg_eq_nil in h2.
  rewrite (count_occ_In h1) in h4.
  rewrite <- h2 in h4.
  apply lt_irrefl in h4; auto.
  contradict h4.
  rewrite <- list_singularize_in_iff in h3; auto.
Qed.

Lemma linds_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (a:T),
    linds pfdt (a::nil) = (0::nil)::nil.
  unfold linds; intros; simpl.
  rewrite linds_occ_cons_eq; simpl; auto.
Qed.




Lemma inj_dep_linds_occ' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    inj_dep (linds_occ'  pfdt l).
  intros T h1 l.
  red; intros x y hx hy h2. 
  unfold linds_occ' in h2.
  rewrite map_seg_eq_iff in h2.
  destruct h2 as [h2 h5]. 
  pose proof h2 as h2'. rewrite length_map_seg in h2'.
  erewrite <- all_lind_occ_eq_iff.
  split.
  apply h2'.
  intros n h6.
  specialize (h5 _ h6).
  simpl in h5.
  erewrite nth_lt_functional3 in h5. rewrite nth_lt_map_seg in h5. 
  apply h5.
  Grab Existential Variables.
  rewrite <- h2'; auto.
Qed.

Lemma inj_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (x x':T),
    In x l ->
    In x' l ->
    linds_occ pfdt l x = linds_occ pfdt l x' ->
    x = x'.
  intros T h1 l x x' h2 h3 h5.
  pose proof (inj_dep_linds_occ' h1 l) as h4.
  red in h4.
  specialize (h4 _ _ h2 h3).
  unfold linds_occ in h5.
  lr_match h5; fold hlr in h5; destruct hlr as [hl | hr]; 
  lr_match h5; fold hlr in h5; destruct hlr as [hl' | hr']; try contradiction.
  pose proof (proof_irrelevance _ h2 hl).
  pose proof (proof_irrelevance _ h3 hl');
    subst; auto.
Qed.


Lemma in_linds_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds pfdt l) <->
    exists! (x:T),
      In x l /\
      ln = linds_occ pfdt l x.
  intros T h1 l ln.
  unfold linds, linds_occ, linds_occ'.
  rewrite in_map_iff.   
  split; intro h2.
  destruct h2 as [x [h2 hin]].
  rewrite <- list_singularize_in_iff in hin.
  exists x.
  red; split.
  split; auto.
  intros x' h3.
  destruct h3 as [h3 h4]; subst. 
  destruct (in_dec h1 x l) as [h5 | h5], (in_dec h1 x' l) as [h6 | h6].
  rewrite map_seg_eq_iff in h4.
  destruct h4 as [h4 h7].
  pose proof h4 as h4'. rewrite length_map_seg in h4'.
  pose proof (all_lind_occ_eq_iff h1 l x x' h5 h6) as h8.
  rewrite <- h8. 
  split; auto.
  intros n h9.
  specialize (h7 n h9).
  simpl in h7.
  erewrite nth_lt_functional3 in h7.
  rewrite nth_lt_map_seg in h7; auto.
  contradiction.
  contradiction.
  contradiction.
  destruct h2 as [x [h2 hin]].
  destruct h2 as [h2 h3]; simpl; subst.
  exists x.
  split; auto.
  rewrite <- list_singularize_in_iff; auto.
  Grab Existential Variables.
  rewrite <- h4'.
  assumption.
Qed.

Lemma in_linds_iff' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds' pfdt l) <->
    exists! (x:T),
      exists (pf:In x l),
      ln = linds_occ' pfdt l x pf.
  intros T h1 l ln.
  rewrite <- linds_compat'.
  rewrite in_linds_iff.
  split; intro h2.
  destruct h2 as [x [[h3 h4] h5]]; subst.
  exists x.
  red.
  split.
  exists h3.
  rewrite <- linds_occ_compat'; auto.
  intros x' h4.
  destruct h4 as [h4 h6].
  apply h5.
  split; auto.
  rewrite <- linds_occ_compat' in h6; auto.
  destruct h2 as [x [[h3 h4] h5]]; subst.
  exists x.
  red; split.
  split; auto.
  rewrite <- linds_occ_compat'; auto.
  intros x' h4.
  destruct h4 as [h4 h6].
  apply h5.
  exists h4.
  do 2 rewrite <- linds_occ_compat'.
  rewrite <- linds_occ_compat' in h6; auto.
Qed.  

Lemma in_linds_lec_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds pfdt l) <->
    exists (i:nat) (pf:i < length l),
      ln = lec pfdt pf.
  intros T h1 l ln.
  rewrite in_linds_iff.
  split; intro h2.
  destruct h2 as [x h2].
  red in h2.
  destruct h2 as [h2 h3].
  destruct h2 as [h2]; subst.
  unfold lec.
  exists (lind h1 _ _ h2).
  exists (lt_lind _ _ _ _).
  rewrite lind_compat; auto.

  destruct h2 as [i [h2 h3]]; subst.
  unfold lec.
  exists (nth_lt l i h2).
  red; split.
  split.
  apply in_nth_lt.
  reflexivity.
  intros x h3.
  destruct h3 as [h3 h4].
  apply inj_linds_occ in h4; auto.
  apply in_nth_lt.
Qed.


Lemma in_linds_subst :
  forall {T:Type} {l l':list T} {ln:list nat} (pfdt:type_eq_dec T),
    In ln (linds pfdt l) ->
    l = l' ->
    In ln (linds pfdt l').
  intros; subst; auto.
Qed.
    
Lemma in_linds_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat) (i:nat),
    In i ln ->
    In ln (linds pfdt l) ->
    i < length l.
  intros T h1 l ln i h2 h3.
  rewrite in_linds_iff in h3.
  destruct h3 as [x [[h3 h4] h5]]; subst.
  rewrite in_linds_occ_iff in h2.
  destruct h2 as [m [h6 [h7 h8]]]; subst.
  apply lt_lind_occ.
Qed.


Lemma in_linds_not_nil :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds pfdt l) ->
    ln <> nil.
  intros T h1 l ln h2 h3; subst. 
  rewrite in_linds_iff in h2.
  destruct h2 as [x [[h3 h4] h5]].
  symmetry in h4.
  pose proof (linds_occ_eq_iff h1 _ _ h3 nil) as h6.
  rewrite h6 in h4.
  destruct h4 as [h4].
  simpl in h4.
  pose proof h3 as h3'.
  pose proof (count_occ_In h1 l x) as h7.
  rewrite h7 in h3'.
  rewrite h4 in h3'.
  apply lt_irrefl in h3'.
  contradiction.
Qed.

Lemma in_linds_occ_linds' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l),
    In (linds_occ' pfdt l x pf) (linds' pfdt l).
  intros T h1 l x h2.
  unfold linds'.
  rewrite in_map_in_dep_iff.
  exists x.
  ex_goal.
  rewrite <- list_singularize_in_iff; auto.
  exists hex.
  unfold fun_in_dep_incl.
  f_equal.
  apply proof_irrelevance.
Qed.

    
Lemma in_linds_occ_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    In x l ->
    In (linds_occ pfdt l x) (linds pfdt l).
  intros; erewrite linds_compat', linds_occ_compat'.
  apply in_linds_occ_linds'.
  Grab Existential Variables.
  assumption.
Qed.

  

Lemma linds_map_inj : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (f:T->U),
    Injective f ->
    linds pfdu (map f l) = linds pfdt l.
  intros T U h1 h2 l f h3. symmetry.
  unfold linds.
  rewrite (map_eq_iff (list_singularize h1 l) (linds_occ h1 l)).
  ex_goal.
  rewrite map_length.
  erewrite list_singularize_map_inj.
  rewrite map_length.
  reflexivity.
  assumption.
  exists hex.
  intros n h4.
  erewrite nth_lt_map.
  rewrite linds_occ_eq_iff.
  ex_goal.
  rewrite length_linds_occ.
  rewrite nth_lt_list_singularize_map.
  erewrite count_occ_map_inj.
  reflexivity.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  assumption.
  exists hex0.
  intros m h5.
  erewrite nth_lt_linds_occ.
  symmetry.
  gen0.
  erewrite nth_lt_list_singularize_map.
  intro h6.
  erewrite lind_occ_map_inj.
  reflexivity.
  assumption.

  Grab Existential Variables.
  rewrite nth_lt_list_singularize_map.
  apply in_map.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  assumption.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
Qed.
  
(*This prints the value given by the indices [ln] in [l]*)
Definition in_linds_val {T:Type} (pfdt:type_eq_dec T) (l:list T)
           (ln:list nat) (pfiln:In ln (linds pfdt l)) : T :=
  let pfn := in_linds_not_nil _ _ _ pfiln in
  let h := hd' _ pfn in
  let pfh := in_hd' _ pfn in
  let pflt := in_linds_lt pfdt _ _ _  pfh pfiln in
  nth_lt l h pflt.

Lemma in_linds_val_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (ln:list nat) (pfiln:In ln (linds pfdt l)),
    In (in_linds_val pfdt l ln pfiln) l.
  unfold in_linds_val; intros; apply in_nth_lt.
Qed.
  
Lemma nth_lt_in_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l) (i:nat)
         (pfln:In i (linds_occ pfdt l x)),
    nth_lt l i (in_linds_lt _ _ _ i pfln (in_linds_occ_linds pfdt l x pf)) = x.
  intros T h1 l x h2 i h3.
  pose proof h3 as h3'.
  rewrite in_linds_occ_iff in h3'.
  destruct h3' as [m [h4 [h5 h6]]]; subst.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat; auto.
Qed.


  
Lemma in_linds_length :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (pfi:In ln (linds pfdt l)),
    length ln = count_occ pfdt l (in_linds_val _ _ _ pfi).
  intros T h1 l ln h2.
  pose proof h2 as h2'.
  rewrite in_linds_iff in h2'.
  destruct h2' as [x [[h3 h4] h5]]; subst.
  rewrite length_linds_occ.
  specialize (h5 (in_linds_val h1 l (linds_occ h1 l x) h2)).
  hfirst h5.
  split; auto.
  apply in_linds_val_in.
  unfold in_linds_val.
  pose proof (nth_lt_in_linds_occ h1 l x h3 _ (in_hd' (linds_occ h1 l x)
                                                      (in_linds_not_nil h1 l (linds_occ h1 l x) h2))) as h6.
  rewrite <- h6 at 1.
  repeat f_equal.
  apply proof_irrelevance.
  specialize (h5 hf).
  rewrite h5 at 1; auto.
Qed.

Lemma count_occ_in_linds_val :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (pfi:In ln (linds pfdt l)),
    count_occ pfdt l (in_linds_val _ _ _ pfi) = length ln.
  intros; symmetry; apply in_linds_length.
Qed.


Lemma nth_lt_eq_in_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln : list nat)
         (m r:nat) (pfm:m < length l) (pfr:r < length l),
    In m ln ->
    nth_lt l m pfm = nth_lt l r pfr ->
    In ln (linds pfdt l) ->
    In r ln.
  intros T h1 l ln m r h2 h3 h4 h5 h6. 
  rewrite in_linds_iff in h6.
  destruct h6 as [x h6]. red in h6. destruct h6 as [[h6 h7] h8]; subst. 
  rewrite in_linds_occ_iff in h4.
  destruct h4 as [m' h7].
  destruct h7 as [h7 [h9 h10]]; subst.
  erewrite nth_lt_functional3 in h5. rewrite nth_lt_lind_occ in h5; subst.
  apply in_linds_occ_nth_lt.
Qed.



Lemma length_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    length (linds pfdt l) =
    length (list_singularize pfdt l).
  intros; unfold linds; rewrite map_length; auto.
Qed.

Lemma length_linds' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    length (linds' pfdt l) =
    length (list_singularize pfdt l).
intros; unfold linds'; rewrite length_map_in_dep; auto.
Qed.

Lemma linds_eq_length_singular :
  forall {T U:Type} {pfdt:type_eq_dec T} {pfdu:type_eq_dec U}
         (l:list T) (l':list U),
    linds pfdt l =
    linds pfdu l' ->
    length (list_singularize pfdt l) =
    length (list_singularize pfdu l').
  intros; do 2 rewrite <- length_linds.
  rewrite H; auto.
Qed.

Lemma linds_eq_length_singular' :
  forall {T U:Type} {pfdt:type_eq_dec T} {pfdu:type_eq_dec U}
         (l:list T) (l':list U),
    linds' pfdt l =
    linds' pfdu l' ->
    length (list_singularize pfdt l) =
    length (list_singularize pfdu l').
  intros; do 2 rewrite <- length_linds'.
  rewrite H; auto.
Qed.
         
Lemma no_dup_linds' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup (linds' pfdt l).
  intros T h1 l.
  unfold linds'.
  apply no_dup_map_in_dep_inj. 
  apply no_dup_list_singularize.
  apply inj_dep_incl.
  apply inj_dep_linds_occ'.
Qed.

Lemma no_dup_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup (linds pfdt l).
  intros; rewrite linds_compat'.
  apply no_dup_linds'.
Qed.

Lemma no_dup_mem_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    no_dup_mem (linds pfdt l).
  intros T h1 l; red; intros ln h2.
  rewrite in_linds_iff in h2.
  destruct h2 as [x [[h2 h3] h4]]; subst.
  apply no_dup_linds_occ.
Qed.


Lemma all_sing_in_linds :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (ln:list nat) (m:nat),
    all_sing ln m ->
    In ln (linds pfdt l) ->
    ln = m::nil.
  intros T h1 l ln m h2 h3.
  rewrite in_linds_iff in h3.
  destruct h3 as [x [[h3 h4] h5]]; subst.
  pose proof h3 as h3'.
  pose proof (no_dup_linds_occ h1 l x) as h6.
  rewrite (list_singularize_no_dup nat_eq_dec) in h6.
  rewrite <- h6.
  erewrite  all_sing_list_singularize.
  reflexivity.
  intro h7.
  rewrite linds_occ_eq_iff in h7.
  destruct h7 as [h7].
  rewrite (count_occ_In h1) in h3.
  simpl in h7; omega.
  assumption.

  Grab Existential Variables.
  assumption.
Qed.


Lemma nth_lts_eq_length_linds : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U) (pfe:length l = length l'),
    (forall (m n : nat) (pfm : m < length l) (pfn : n < length l),
       let pfm' := lt_eq_trans m (length l) (length l') pfm pfe in
       let pfn' := lt_eq_trans n (length l) (length l') pfn pfe in
       nth_lt l m pfm = nth_lt l n pfn <->
       nth_lt l' m pfm' = nth_lt l' n pfn') ->
    length (linds pfdt l) =
    length (linds pfdu l').
  intros; do 2 rewrite length_linds; eapply nth_lts_eq_length_list_singularize.
  apply H.
Qed.


Lemma in_lec :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (i:nat)
         (pf: i < length l),
    In (lec pfdt pf) (linds pfdt l).
  intros.
  apply in_linds_occ_linds.
  apply in_nth_lt.
Qed.



Lemma in_lec_lt :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (m r:nat) (pflt:m < length l),
    In r (lec pfdt pflt) ->
    r < length l.
  intros T h1 l m r h2 h3.
  unfold lec in h3.
  rewrite in_linds_occ_iff in h3.
  destruct h3 as [? [? [h3 ?]]]; subst.
  apply lt_lind_occ.
Qed.
  

Lemma in_lec_S :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (i:nat)
         (pf: i < length l),
    In (lec_S pfdt pf) (famlS (linds pfdt l)).
  intros T h1 l i h2.
  unfold lec_S, famlS.
  apply in_map.
  apply in_lec.
Qed.

Lemma nin0_lec_S :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat) (pflt:n < length l),
    ~In 0 (lec_S pfdt pflt).
  intros T h1 l n h2 h3.
  unfold lec_S in h3.
  rewrite in_lS_iff in h3.
  destruct h3 as [? [? ?]]; discriminate.
Qed.

Lemma lec_cons_nth_lt_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (m:nat) (pfm:m < length l),
    let a := nth_lt l m pfm in
      lec_cons pfdt pfm a =
      0 :: lec_S pfdt pfm.
  intros T h1 l m h2; simpl.
  unfold lec_cons, lec_S, lec.
  rewrite linds_occ_eq_iff.
  ex_goal.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  rewrite length_lS, length_linds_occ.
  rewrite nth_lt_compat.
  do 2 f_equal.
  apply nth_indep; auto.
  contradict hr.
  rewrite nth_lt_compat.
  apply nth_indep; auto.
  exists hex.
  intros occ h3.
  simpl; simpl in h3.
  rewrite switch_eq_dec_eq in h3. 
  lr_match_goal; fold hlr in h3; fold hlr; destruct hlr as [hl | hr]; unfold switch_eq_dec in h3.
  destruct occ as [|occ]; auto.
  destruct in_dec as [h4 | h4].
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  unfold lS.
  erewrite nth_lt_map.
  f_equal.
  symmetry.
  erewrite nth_lt_functional3.
  erewrite nth_lt_linds_occ at 1.
  apply lind_occ_functional; auto.
  rewrite length_lS, length_linds_occ.
  apply lt_S_n in h3.
  erewrite nth_lt_compat.
  erewrite nth_indep.
  apply h3.
  assumption.
  contradict h4.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  apply in_nth_lt. 
  assumption.
  contradict hr.
  rewrite nth_lt_compat.
  f_equal.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.

  Grab Existential Variables.
  assumption.
  apply in_nth_lt.
  rewrite length_linds_occ.
  apply lt_S_n in h3.
  erewrite nth_lt_compat; auto.
  apply lt_S_n in h3.
  rewrite length_linds_occ, nth_lt_compat; auto.
  apply  lt_S_n in h3.
  rewrite length_lS, length_linds_occ, nth_lt_compat; auto.

  left.
  simpl.
  apply nth_lt_compat.
Qed.


Lemma in_lec_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (m r:nat)
         (pfm:m < length l)
         (pfr:r < length l),
    In r (lec pfdt pfm) <->
    nth_lt l m pfm =
    nth_lt l r pfr.
  intros T h1 l m r h2 h3; split; intro h4.
  unfold lec in h4.
  rewrite in_linds_occ_iff in h4.
  destruct h4 as [c [h6 [h4 h5]]].
  subst.
  symmetry.
  gen0.
  erewrite lind_occ_functional.
  intro h5.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat at 1.
  reflexivity.
  reflexivity.
  reflexivity.
  reflexivity.

  unfold lec.
  rewrite in_linds_occ_iff.
  pose proof (ex_lind_occ_nth_lt h1 _ _ h3) as h5.
  destruct h5 as [occ [h5 h6]].
  rewrite <- h6.
  exists occ.
  exists (in_nth_lt _ _ _).
  split.
  rewrite h4; auto.
  gen0.
  rewrite h4.
  intro; apply lind_occ_functional; auto.

  Grab Existential Variables.
  apply in_nth_lt.
Qed.

Lemma in_lec_iff' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (m r:nat)
         (pfm:m < length l),
    In r (lec pfdt pfm) <->
    exists pfr:r < length l, 
      nth_lt l m pfm =
      nth_lt l r pfr.
  intros T h1 l m r h2; split; intro h3.
  exists (in_lec_lt _ _ _ _ _ h3).
  pose proof h3 as h4.
  rewrite in_lec_iff in h4.
  apply h4.
  destruct h3 as [h3 h4].
  rewrite in_lec_iff.
  apply h4.
Qed.

Lemma in_lec_S_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (m r:nat)
         (pfm:m < length l),
    In r (lec_S pfdt pfm) <->
    0 < r /\
    exists pfpr:pred r < length l, 
      nth_lt l m pfm =
      nth_lt l (pred r) pfpr.
  intros T h1 l m r h2. 
  unfold lec_S.
  rewrite in_lS_iff. 
  split; intro h3. 
  destruct h3 as [m' [h3 h4]]; subst.
  simpl.
  split; auto with arith.
  exists (in_lec_lt _ _ _ _ _ h3).
  pose proof h3 as h3'.
  rewrite in_lec_iff' in h3'. 
  destruct h3' as [h4 h5].
  rewrite h5.
  apply nth_lt_functional3.
  destruct h3 as [hr [h3 h4]].
  exists (pred r).
  split.
  rewrite in_lec_iff.
  apply h4.
  apply (S_pred _ _ hr).
Qed.

  
Lemma rep_in_lec :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (m:nat) (pf:m < length l),
    In m (lec pfdt pf).
  intros T h1 l m h2.
  rewrite in_lec_iff.
  reflexivity.
Qed.

    

Lemma lec_not_nil :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (m:nat) (pf:m < length l),
    lec pfdt pf <> nil.
  intros T h1 l m h2 h3.
  unfold lec in h3.
  rewrite linds_occ_eq_iff in h3.
  destruct h3 as [h3 h4].
  simpl in h3.
  pose proof (in_nth_lt l m h2) as h5.
  rewrite (count_occ_In h1) in h5.
  rewrite h3 in h5.
  apply lt_irrefl in h5; auto.

  Grab Existential Variables.
  apply in_nth_lt.
Qed.

Lemma lecs_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) 
         (m:nat) (pfm:m < length l) 
         (n:nat) (pfn:n < length l),
    lec pfdt pfm =
    lec pfdt pfn <->
    nth_lt _ _ pfm =
    nth_lt _ _ pfn.
  intros T h1 l m h2 n h3; split; intro h4.
  unfold lec in h4.
  apply inj_linds_occ in h4; try apply in_nth_lt; auto.
  unfold lec.
  rewrite h4; auto.
Qed.


Lemma lec_Ss_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) 
         (m:nat) (pfm:m < length l) 
         (n:nat) (pfn:n < length l),
    lec_S pfdt pfm =
    lec_S pfdt pfn <->
    nth_lt _ _ pfm =
    nth_lt _ _ pfn.
  intros; unfold lec_S; split; intro.
  apply inj_lS in H.
  rewrite lecs_eq_iff in H; auto.
  f_equal.
  rewrite lecs_eq_iff; auto.
Qed.


Lemma lecs_eq_count_occ :
  forall {T U:Type}
         (pfdt:type_eq_dec T)
         (pfdu:type_eq_dec U)
         (l:list T)
         (l':list U)
         (m n:nat)
         (pfm:m < length l) 
         (pfn:n < length l'),
    lec pfdt pfm =
    lec pfdu pfn ->
    count_occ pfdt l (nth_lt _ _ pfm) =
    count_occ pfdu l' (nth_lt _ _ pfn).
  intros T U h1 h2 l l' m n h3 h4 h5.
  pose proof (f_equal (@length nat) h5) as h6.
  do 2 rewrite length_lec in h6; auto.
Qed.

Lemma impl_in_lec_app :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T) (m:nat)
         (pfm:m < length l) (pfm':m < length (l++l'))
         (r:nat),
    In r (lec pfdt pfm) ->
    In r (lec pfdt pfm').
  intros T h1 l l' m h2 h3 r h4.
  pose proof (in_lec_lt h1 _ _ _ h2 h4) as h5.
  rewrite in_lec_iff in h4.
  rewrite in_lec_iff.
  rewrite (nth_lt_app1 _ _ _ h5).
  rewrite (nth_lt_app1 _ _ _ h2).
  apply h4.

  Grab Existential Variables.
  rewrite app_length; auto with arith.
Qed.

Lemma lec_lrep :
  forall {T:Type} (pfdt:type_eq_dec T) (a:T) (m n:nat)
         (pfm:m < length (lrep a n)),
    lec pfdt pfm = seg_list n.
  intros T h1 a m n h2.
  unfold lec.
  pose proof h2 as h2'.
  rewrite length_lrep in h2'.
  rewrite linds_occ_eq_iff.
  ex_goal.
  rewrite count_occ_lrep', length_seg_list; auto.
  apply in_nth_lt.
  exists hex.
  intros occ h3.
  rewrite nth_lt_seg_list.
  rewrite count_occ_lrep' in h3.
  apply lind_occ_lrep; auto.
  apply in_nth_lt.

  Grab Existential Variables.
  apply in_nth_lt.
Qed.

Lemma all_sing_lec :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (m:nat) (pfm:m < length l),
    all_sing l a ->
    lec pfdt pfm = seg_list (length l).
  intros T h1 l a m h2 h3.
  pose proof h3 as h3'. 
  apply all_sing_lrep in h3'. 
  gen0.
  rewrite h3'.
  intro h0.
  pose proof (length_lrep a (length l)) as h4. 
  pose proof (lec_lrep h1 a m _ h0) as h5.
  rewrite h5.
  rewrite h3'.
  rewrite length_lrep; f_equal; auto.
Qed.



Lemma all_sing_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pfa:In a l),
    all_sing l a ->
    linds pfdt l =
    seg_list (length l) :: nil.
  intros T h1 l a h2 h3.

  unfold linds.
  rewrite (map_eq_iff (list_singularize h1 l) (linds_occ h1 l)).
  ex_goal.
  simpl.
  erewrite all_sing_list_singularize.
  simpl; auto.
  eapply in_not_nil.
  apply h2.
  apply h3.
  apply all_sing_lrep in h3.
  exists hex.
  intros n h4.
  simpl.
  rewrite linds_occ_eq_iff.
  ex_goal.
  rewrite h3 at 1.
  rewrite count_occ_lrep'; auto. 
  do 3 (destruct n; repeat try rewrite length_seg_list; auto).
  rewrite <- h3.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  exists hex0.
  intros m h5.
  destruct n as [|n].
  rewrite nth_lt_seg_list.
  gen0.
  generalize h4.
  rewrite h3.
  intros h4' h7.
  apply lind_occ_lrep.
  rewrite h3 in h5 at 1.
  rewrite count_occ_lrep' in h5; auto.
  rewrite <- h3.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  gen0.
  generalize dependent h4.
  rewrite h3.
  intros h4 h5 h6 i.
  rewrite lind_occ_lrep.
  destruct n as [|n]; try rewrite nth_lt_seg_list; auto.
  rewrite count_occ_lrep' in h6; auto.

  Grab Existential Variables.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
Qed.

Lemma neq_nin_0_lec_Sn :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T)
         (m m':nat)
         (pfm:m < length l)
         (pfsm:S m < length (a::l)),
    a <> nth_lt l m pfm ->
    ~In 0 (lec pfdt pfsm).
  intros T h1 l a m m' h2 h3 h4 h5.
  rewrite in_lec_iff in h5.
  simpl in h5.
  contradict h4.
  rewrite nth_lt_compat.
  symmetry.
  erewrite nth_indep.
  apply h5.
  assumption.

  Grab Existential Variables.
  simpl.
  apply lt_0_Sn.
Qed.


Lemma lec_0_cons_nin_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    ~In a l ->
    lec pfdt (lt_0_Sn (length l)) (l:=a::l) =
    0::nil.
  intros T h0 l a h1.
  unfold lec.
  simpl.
  rewrite linds_occ_eq_iff.
  ex_goal.
  simpl.
  erewrite eq_dec_same at 1.
  rewrite (count_occ_In h0) in h1.
  f_equal.
  omega.
  assumption.
  exists hex.
  intros m h2.
  simpl. 
  erewrite eq_dec_same at 1.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  contradiction.
  do 2 (destruct m; auto).
  assumption.

  Grab Existential Variables.
  left; auto.
Qed.

Definition linds_remove {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (a:T) :=
  remove list_nat_eq_dec (linds_occ pfdt l a) (linds pfdt l).


Lemma nin_linds_remove :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T),
    In a l ->
    ~In (linds_occ pfdt l a) (linds_remove pfdt l a).
  intros T h1 l a h2 h3.
  unfold linds_remove in h3.
  rewrite in_remove_iff in h3.
  destruct h3 as [? h3]; contradict h3; auto.
Qed.


Lemma in_linds_nth_lt_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (m n:nat)
         (pfln:In ln (linds pfdt l))
         (pfm:In m ln)
         (pfn:In n ln),
    let pfm' := in_linds_lt  _ _ _ _ pfm pfln in
    let pfn' := in_linds_lt  _ _ _ _ pfn pfln in
    nth_lt l m pfm' =
    nth_lt l n pfn'.
  intros T h1 l ln m n h2 h3 h4; simpl.
  pose proof h2 as h2'.
  rewrite in_linds_iff in h2'.
  destruct h2' as [x [[h7 h8] h9]]; subst.
  erewrite nth_lt_functional3.
  symmetry.
  erewrite nth_lt_functional3.
  apply in_linds_occ_nth_lt_eq.
  apply h7.
  
  Grab Existential Variables.
  apply h4.
  assumption.
Qed.

  
Lemma in_linds_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat) (m:nat)
         (pfm:In m ln)
         (pfln: In ln (linds pfdt l)),
    let pf := in_linds_lt pfdt l ln _ pfm pfln  in
    In (nth_lt _ _ pf) l.
  intros; apply in_nth_lt.
Qed.



Lemma no_dup_faml_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    no_dup_faml (linds pfdt l).
  intros T h1 l.
  red.
  intros l' h2.
  rewrite in_linds_iff in h2.
  destruct h2 as [x h3]; red in h3.
  destruct h3 as [[h3 h4] h5]; subst.
  apply no_dup_linds_occ.
Qed.



Lemma disj_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    pairwise_disjointl (linds pfdt l).
  intros T h1 l.
  red; intros lm ln h2 h3 h4; do 2 red.
  apply eq_empty.
  red; intros r h5.
  inversion h5 as [? h8 h9]; subst.
  rewrite <- list_to_set_in_iff in h8, h9.
  rewrite in_linds_iff in h3, h4.
  destruct h3 as [x [h3 h10]],
                 h4 as [y [h4 h11]].
  destruct h3 as [h3],
                 h4 as [h4]; subst.
  rewrite in_linds_occ_iff in h8.
  destruct h8 as [m [h8 [h12 h13]]]; subst.
  rewrite in_linds_occ_iff in h9. 
  destruct h9 as [m' [h9 [h13 h14]]]; subst.
  apply lind_occ_inj in h14; subst.
  contradict h2; auto.
Qed.  


Lemma union_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    FamilyUnion (faml_to_fam (linds pfdt l)) =
    list_to_set (seg_list (length l)).
  intros T h1 l.
  apply Extensionality_Ensembles; red; split; red; intros r h2.
  inversion h2 as [S h3 h4 h5]; subst.
  pose proof (in_faml_to_fam_impl _ _ h4) as h6.
  destruct h6 as [ln [h6 h7]]; subst.
  rewrite <- list_to_set_in_iff.
  rewrite in_seg_list_iff.
  rewrite <- list_to_set_in_iff in h5.
  rewrite in_linds_iff in h6.
  destruct h6 as [x [[h7 h8] h9]]; subst.
  rewrite in_linds_occ_iff in h5.
  destruct h5 as [m [h10 [h11 h12]]]; subst.
  apply lt_lind_occ.

  rewrite <- list_to_set_in_iff in h2.
  rewrite in_seg_list_iff in h2.
  pose proof (family_union_intro _ (faml_to_fam (linds h1 l)) (list_to_set (linds_occ h1 l (nth_lt _ _ h2))) r) as h3.
  hfirst h3.
  apply in_faml_impl.
  apply in_linds_occ_linds.
  apply in_nth_lt.
  specialize (h3 hf).
  hfirst h3.
  rewrite <- list_to_set_in_iff.
  rewrite in_linds_occ_iff. 
  pose proof (ex_lind_occ_nth_lt h1 _ _ h2) as h4.
  destruct h4 as [occ h4].
  exists occ, (in_nth_lt _ _ _); auto.
  auto.
Qed.


Lemma partition_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    partition_faml (linds pfdt l) (seg_list (length l)).
  intros T h1 l.
  red; red.
  split.
  rewrite <- pairwise_disjointl_compat.
  apply disj_linds.
  apply union_linds.
Qed.


Lemma list_to_set_linds_eq_length_list_singularize :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l') ->
    length (list_singularize pfdt l) = length (list_singularize pfdu l').
  intros T U h1 h2 l h3 h4.
  do 2 rewrite linds_compat' in h4.
  unfold linds' in h4.
  eapply list_to_set_map_in_deps_eq_length in h4.
  assumption.
  assumption.
  assumption.
  apply list_nat_eq_dec.
  apply no_dup_list_singularize.
  apply no_dup_list_singularize.
  apply inj_dep_incl.
  apply inj_dep_linds_occ'.
  apply inj_dep_incl.
  apply inj_dep_linds_occ'.
Qed.

Lemma count_occ_linds_occ' :
  forall {T:Type} (pfdt:type_eq_dec T) 
         (l:list T) (x:T)
         (pfi:In x l),
    count_occ list_nat_eq_dec (linds' pfdt l) (linds_occ' pfdt l x pfi) = 1.
  intros T h1 l x h2; apply count_occ_no_dup.
  apply no_dup_linds'.
  rewrite in_linds_iff'.
  exists x.
  red; split.
  exists h2; auto.
  intros x' h3.
  destruct h3 as [h3 h4].
  apply inj_dep_linds_occ' in h4; auto.
Qed.

Lemma count_occ_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) 
         (l:list T) (x:T)
         (pfi:In x l),
    count_occ list_nat_eq_dec (linds pfdt l)
              (linds_occ pfdt l x) = 1.
  intros T h1 l x h2.
  rewrite linds_compat'.
  rewrite (linds_occ_compat' h1 l _ h2).
  apply count_occ_linds_occ'.
Qed.

Lemma count_occ_linds :
  forall {T:Type} (pfdt:type_eq_dec T) 
         (l:list T) (ln:list nat),
    In ln (linds pfdt l) ->
    count_occ list_nat_eq_dec (linds pfdt l) ln = 1.
  intros T h1 l ln h2.
  apply count_occ_no_dup; auto.
  apply no_dup_linds.
Qed.

Lemma count_occ_linds_cons_nin' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (ln:list nat),
    l <> nil ->
    ~In a l ->
    ~In 0 ln ->
    count_occ list_nat_eq_dec (linds' pfdt (a::l)) ln =
    count_occ list_nat_eq_dec (linds' pfdt l) (lpred ln).
  intros T h1 l a ln hn h2 h3.    
  destruct (in_dec list_nat_eq_dec (lpred ln) (linds' h1 l)) as [h4 | h4]. 
  pose proof h4 as h4'. 
  rewrite in_linds_iff' in h4.
  destruct h4 as [x [[h4 h6] h5]]. 
  rewrite count_occ_no_dup.
  rewrite h6.
  rewrite count_occ_linds_occ'; auto. 
  apply no_dup_linds'.
  rewrite in_linds_iff'. 
  exists x.
  red; split.
  exists (in_cons _ _ _ h4).
  rewrite linds_occ_cons_neq'.
  rewrite lS_lpred_iff in h3.
  rewrite h3.
  f_equal; auto. 
  intro; subst; contradiction.  
  intros x' h7.  
  destruct h7 as [h7 h8]; subst.
  destruct (h1 x' a) as [|hf]; subst.
  contradict h3.
  rewrite in_linds_occ_iff'.
  exists 0.
  simpl.
  destruct (h1 a a) as [he | he].
  split; auto with arith.
  contradict he.
  reflexivity. 
  destruct h7 as [|h7]; subst. contradict hf; auto.
  pose proof (proof_irrelevance 
                _ (or_intror h7)
                (in_cons a x' l h7)) as he.
  rewrite he in h6.
  rewrite linds_occ_cons_neq' in h6; auto.
  rewrite lpred_lS in h6.  
  apply inj_dep_linds_occ' in h6; auto.

  pose proof h4 as h4'.
  rewrite (count_occ_In list_nat_eq_dec) in h4.
  change (~0 <  count_occ list_nat_eq_dec (linds' h1 l) (lpred ln)) in h4.
  rewrite not_lt0_iff in h4.
  rewrite h4.
  rewrite <- not_lt0_iff.
  change (~count_occ list_nat_eq_dec (linds' h1 (a::l)) ln > 0). 
  rewrite <- (count_occ_In list_nat_eq_dec).
  contradict h4'.
  rewrite in_linds_iff' in h4'.
  destruct h4' as [x [[h5 h6] h7]]; subst.
  destruct (h1 x a) as [|h8]; subst. 
  contradict h3.
  rewrite in_linds_occ_iff'.
  exists 0.
  simpl.
  destruct (h1 a a) as [he | he].
  split; auto with arith.
  contradict he.
  reflexivity.
  destruct h5 as [|h5]; subst.
  contradict h8; auto.
  rewrite in_linds_iff'.
  exists x.
  red; split.
  exists h5.
  pose proof (proof_irrelevance _ 
                                (or_intror h5) (in_cons a x l h5)) as he.
  rewrite he.
  rewrite linds_occ_cons_neq'.
  rewrite lpred_lS.
  reflexivity.
  assumption.
  intros x' h9.
  destruct h9 as [h9 h10].
  apply h7; auto.
  exists (in_cons _ _ _ h9).
  assert (h11:x' <> a).
  intro; subst; contradiction.

  pose proof (proof_irrelevance
                _ 
                (or_intror h5)
                (in_cons a x l h5)) as h12.
  rewrite h12.
  rewrite linds_occ_cons_neq'; auto.
  rewrite linds_occ_cons_neq'; auto.
  f_equal.
  rewrite h12 in h10.
  rewrite linds_occ_cons_neq' in h10; auto.
  rewrite lpred_lS in h10; auto.
Qed.


Definition coarse_list_in_list_of_lists_seqs_fun_list_nat
           {T:Type} (pfdt:type_eq_dec T) (l:list T)
           (ln ln':list nat) (pfcl:coarse_list ln' ln)
           (pfi:In ln (list_of_lists_seqs (linds pfdt l))) : 
  list T :=
  map_seg
     (fun i (pfi':i < length ln') =>
        let pfi0 := lt_le_trans _ _ _ pfi' (length_coarse_list _ _ pfcl) in
        let pfi1 := in_list_of_lists_seqs_lt (linds pfdt l) ln i pfi pfi0 in
        proj1_sig (constructive_definite_description
                     _
                     (iff1 (in_linds_iff pfdt l _)
                           (in_nth_lt (linds pfdt l) i pfi1)))).


Lemma in_list_of_lists_seqs_mask_to_coarse_list :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (v:bv) (pfeq:length ln = length v),
    let ll := linds pfdt l in
    forall (pfi:In ln (list_of_lists_seqs ll)),
      let ln' := mask_to_coarse_list pfeq in
      let pfeq' := in_list_of_lists_seqs_length _ _ pfi in
      let pfeq0 := eq_trans (eq_sym pfeq') pfeq in 
      let ll' := mask_to_coarse_list pfeq0 in
      In ln' (list_of_lists_seqs ll').
  intros; apply coarse_list_in_list_of_lists_seqs.
Qed.
 
(*Returns a list with [true]s corresponding to elements whose 
  [count_occ]s are equal to [n], [false] otherwise.*)
Definition occ_count_bv {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (n:nat) : bv :=
  map (fun x => if (nat_eq_dec (count_occ pfdt l x) n) then
                  true else false)
      (list_singularize pfdt l).

Lemma occ_count_bv0 :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T),
    occ_count_bv pfdt l 0 = falses (length (list_singularize pfdt l)).
  intros T h1 l.
  unfold occ_count_bv, falses.
  rewrite (map_eq_iff (list_singularize h1 l) (fun x : T => if nat_eq_dec (count_occ h1 l x) 0 then true else false)).
  ex_goal.
  rewrite length_lrep; auto.
  exists hex.
  intros n h2.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  pose proof (in_nth_lt _ _ h2) as h3.
  rewrite <- list_singularize_in_iff in h3.
  rewrite (count_occ_In h1) in h3.
  omega.
  rewrite nth_lt_lrep; auto.
Qed.


Lemma occ_count_bv_length :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat),
    length (occ_count_bv pfdt l n) = length (list_singularize pfdt l).
  intros T h1 l n.
  unfold occ_count_bv.
  rewrite map_length; auto.
Qed.

(*A list of all elements that occur [n] times, in the order 
  that they appear in [list_singuliarze l].*)
Definition occ_count {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (n:nat) : list T :=
  mask_to_coarse_list (eq_sym (occ_count_bv_length pfdt l n)).

Lemma count_occ_map_count_occ : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T),
    count_occ nat_eq_dec (map (count_occ pfdt l) (list_singularize pfdt l))
              (count_occ pfdt l x) =
    length (inv_iml nat_eq_dec (count_occ pfdt l) (list_singularize pfdt l) (count_occ pfdt l x)). 
  intros; apply count_occ_map.
  assumption.
Qed.
  

Definition linds_prod {T:Type} (pfdt:type_eq_dec T) (l:list T) : faml nat :=
  list_of_lists_seqs (linds pfdt l).

Lemma linds_prod_not_nil :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    p_mem_l (fun lx : list nat => lx <> nil) (linds pfdt l).
  intros T h1 l; red; intros lx h2.
  rewrite in_linds_iff in h2. 
  destruct h2 as [x h2].
  red in h2.
  destruct h2 as [h2 h3].
  destruct h2 as [h2]; subst.
  intro h4.
  rewrite (linds_occ_eq_iff _ _ _ h2) in h4.
  destruct h4 as [h4].
  simpl in h4.
  pose proof h2 as h2'. rewrite (count_occ_In h1) in h2'.
  omega.
Qed.


Lemma linds_prod_not_nil' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    l <> nil ->
    linds_prod pfdt l <> nil.
  intros T h1 l h2.
  unfold linds_prod.
  apply list_of_lists_seqs_not_nil.
  apply linds_prod_not_nil.
Qed.




Lemma no_dup_linds_prod :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup (linds_prod pfdt l).
  intros T h1 l.
  unfold linds_prod.
  apply list_of_lists_seqs_no_dup.
  apply no_dup_faml_linds.
Qed.

                                                   
Lemma in_linds_prod_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds_prod pfdt l) <->
    (length ln = length (list_singularize pfdt l) /\
    (forall (i:nat) (pfi:i < length ln),
     exists pfi':i < length (linds pfdt l),
        In (nth_lt ln i pfi) (nth_lt _ i pfi'))).
  intros T h1 l ln.
  unfold linds_prod.
  rewrite in_list_of_lists_seqs_iff.
  rewrite length_linds at 1; tauto.
Qed.


Lemma in_linds_prod_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (i:nat),
    In ln (linds_prod pfdt l) ->
    In i ln ->
    i < length l.
  intros T h1 l ln i h2 h3.
  rewrite in_linds_prod_iff in h2.
  destruct h2 as [h2 h4].
  specialize (h4 (lind nat_eq_dec _ _ h3) (lt_lind _ _ _ _)).
  destruct h4 as [h4 h5].
  rewrite lind_compat in h5.
  unfold linds in h5.
  erewrite nth_lt_map in h5.
  eapply in_linds_occ_lt.
  apply h5.

  Grab Existential Variables.
  eapply lt_eq_trans.
  apply lt_lind.
  assumption.
Qed.

Lemma in_linds_prod_lind_lt :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (pfilnlp:In ln (linds_prod pfdt l))
         (i:nat) (pfiln:In i ln),
    lind nat_eq_dec ln i pfiln < length (linds pfdt l).
  intros T h1 l ln h2 i h3.
  rewrite in_linds_prod_iff in h2.
  destruct h2 as [h2 h4].
  specialize (h4 _ (lt_lind nat_eq_dec ln i h3)).
  destruct h4; auto.
Qed.



Lemma in_linds_prod_in_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (i:nat) 
         (pfilnlp:In ln (linds_prod pfdt l))
         (pfiln:In i ln),
    let pflt := in_linds_prod_lt pfdt l ln i pfilnlp pfiln in
    In (linds_occ pfdt l (nth_lt _ _ pflt)) (linds pfdt l).
  intros; apply in_linds_occ_linds; apply in_nth_lt.
Qed.




Lemma in_linds_prod_length :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds_prod pfdt l) ->
    length ln = length (list_singularize pfdt l).
  intros T h1 l n h2.
  rewrite in_linds_prod_iff in h2.
  destruct h2; auto.
Qed.

Lemma nil_in_linds_prod :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    In nil (linds_prod pfdt l) ->
    l = nil.
  intros T h1 l h2.
  rewrite in_linds_prod_iff in h2.
  simpl in h2.
  destruct h2 as [h2].
  symmetry in h2.
  apply length_eq0 in h2.
  rewrite list_singularize_nil_iff in h2; auto.
Qed.


Lemma in_cons_linds_prod :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (i:nat),
    In (i::ln) (linds_prod pfdt l) ->
    In ln (list_of_lists_seqs (tl (linds pfdt l))).
  intros T h1 l ln i h3.
  rewrite in_linds_prod_iff in h3.
  simpl in h3.
  destruct h3 as [h3 h4].
  rewrite in_list_of_lists_seqs_iff.
  split.
  rewrite <- (length_linds h1 l) in h3.
  pose proof (lt_length (linds h1 l) 0) as h5.
  rewrite <- h3 in h5.
  specialize (h5 (lt_0_Sn _)).
  rewrite (hd_compat' _ h5) in h3.
  simpl in h3.
  apply S_inj in h3; auto. 
  intros j h5.
  specialize (h4 (S j) (lt_n_S _ _ h5)).
  simpl in h4.
  destruct h4 as [h4 h6].
  pose proof (lt_length (linds h1 l) (S j) h4) as h7.
  ex_goal.
  pose proof h4 as h4'. rewrite (hd_compat' _ h7) in h4'.
  simpl in h4'.
  apply lt_S_n in h4'; auto. 
  exists hex. 
  rewrite nth_lt_compat in h6.
  erewrite <- (nth_indep (linds h1 l) ln) in h6; auto.
  rewrite (hd_compat' _ h7) in h6.
  simpl in h6.
  rewrite nth_lt_compat, (nth_lt_compat (tl (linds h1 l))).
  erewrite nth_indep, (nth_indep (tl (linds h1 l))); auto.
  apply h6.
Qed.

Definition dups {T:Type} (pfdt:type_eq_dec T)
           (l:list T) : list T :=
  list_singularize pfdt (map_filter_gt nat_eq_dec lt Rlem_dec_lt_nat (count_occ pfdt l) 1 l).



Lemma no_dup_dups :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup l ->
    dups pfdt l = nil.
  intros T h1 l.
  unfold dups.
  rewrite list_singularize_nil_iff.
  unfold map_filter_gt.
  rewrite map_filter_eq_nil_iff.
  destruct (nil_dec' l) as [|h2]; subst.
  intros; left; auto.
  intro h3.
  right.
  pose proof (count_occ_no_dup h1 l h3) as h4.
  intros x h5.
  specialize (h4 x h5).
  omega.
Qed.


Lemma nth_lt_in_or_0 :
  forall {T:Type} (l:list T) (n:nat) (pf:n < length l),
    n = 0 \/ (n > 0 /\ In (nth_lt l n pf) l).
  intros T l n.
  revert l.
  induction n as [|n h1]; try tauto.
  intros l h2.
  right.
  split. auto with arith.
  pose proof (nth_lt_compat l (S n) h2) as h3. simpl in h3.
  destruct h3; auto.
  apply in_nth_lt.
Qed.


Lemma nth_lt_lind_hd :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (pfn:l <> nil) (pfin:In (hd' l pfn) l)
         (pflt:lind pfdt l (hd' l pfn) pfin  < length l),
    nth_lt l (lind pfdt l (hd' l pfn) pfin) pflt =
    hd' l pfn.
  intros T h0 l h1 h2 h3.
  pose proof (lind_compat h0 l (hd' l h1) h2) as h4.
  erewrite nth_lt_functional2.
  rewrite h4 at 1; auto.
  reflexivity.
Qed.

Definition l_to_inds_aux {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (x:T) (pfin:In x l) : nat :=
  let pfin := iff1 (list_singularize_in_iff pfdt l x) pfin in
  lind pfdt _ _ pfin.

(*Returns a list with each element of [l] replaced by its 
  index in [list_singularize].*)
Definition l_to_inds {T:Type} (pfdt:type_eq_dec T)
           (l:list T) : list nat :=
  map_in_dep (l_to_inds_aux pfdt l).

Lemma l_to_inds_aux_functional3 :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pfin pfin':In x l),
    l_to_inds_aux pfdt l x pfin =
    l_to_inds_aux pfdt l x pfin'.
  intros; f_equal; apply proof_irrelevance.
Qed.


Lemma length_l_to_inds :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T),
    length (l_to_inds pfdt l) = length l.
  unfold l_to_inds; intros. apply length_map_in_dep.
Qed.

Lemma in_l_to_inds_aux :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    In (l_to_inds_aux pfdt l x pf) (l_to_inds pfdt l).
  unfold l_to_inds_aux, l_to_inds; intros.
  rewrite in_map_in_dep_iff.
  exists x, pf.
  unfold l_to_inds_aux; auto.
Qed.

  
Lemma l_to_inds_eq_length :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T)
         (l':list U),
    l_to_inds pfdt l = l_to_inds pfdu l' ->
    length l = length l'.
  unfold l_to_inds; intros T U hdt hdu l l' h1.
  rewrite map_in_dep_eq_iff in h1.
  destruct h1 as [h1 h2].
  clear h2.
  rewrite length_map_in_dep in h1; auto.
Qed.

  
Lemma l_to_inds_aux_cons_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pfi:In a l) (pfi':In a (a::l)),
    l_to_inds_aux pfdt (a::l) a pfi' =
    l_to_inds_aux pfdt l a pfi.
  intros T h1 l a h2 h3.
  unfold l_to_inds_aux.
  simpl.
  gen0.
  rewrite list_singularize_in_cons; auto.
  intro h5.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma l_to_inds_aux_cons_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pfi:In a (a::l)),
    ~In a l ->
    l_to_inds_aux pfdt (a::l) a pfi = 0.
  intros T h1 l a h2 h3.
  unfold l_to_inds_aux.
  simpl.
  gen0.
  rewrite list_singularize_nin; auto.
  intro h5.
  erewrite lind_functional.
  rewrite lind_cons_same; auto.
  reflexivity.

  Grab Existential Variables.
  left; auto.
Qed.

Lemma l_to_inds_aux_cons_in_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x:T)
         (pfn:x <> a) (pfi:In a l) (pfi:In x (a::l)),
      l_to_inds_aux pfdt (a::l) x pfi =
      l_to_inds_aux pfdt l x (neq_in_cons _ _ _ pfn pfi).
  intros T h1 l a x h2 h3 h4.
  destruct h4 as [|h4]; subst.
  contradict h2; auto.
  unfold l_to_inds_aux.
  gen0.
  rewrite list_singularize_in_cons.
  intro h5.
  f_equal; apply proof_irrelevance.
  assumption.
Qed.

Lemma l_to_inds_aux_cons_nin_in:
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a x:T)
         (pfn:x <> a) (pfi:~In a l) (pfi':In x (a::l)),
      l_to_inds_aux pfdt (a::l) x pfi' =
      S (l_to_inds_aux pfdt l x (neq_in_cons _ _ _ pfn pfi')).
  intros T h1 l a x h2 h3 h4.
  destruct h4 as [|h4]; subst.
  contradict h2; auto.
  unfold l_to_inds_aux.
  gen0.
  rewrite list_singularize_nin; auto.
  intro h5.

  pose proof (lind_cons_neq h1 (list_singularize h1 l) a x) as h6.
  hfirst h6.
  rewrite <- list_singularize_in_iff; auto.
  specialize (h6 hf h2).
  pose proof (proof_irrelevance _ h5 (or_intror hf)).
  subst.
  rewrite h6.
  repeat f_equal; apply proof_irrelevance.
Qed.


Lemma fun_from_cons_l_to_inds_aux_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    In a l ->
    fun_from_cons (l_to_inds_aux pfdt (a::l)) =
    l_to_inds_aux pfdt l.
  intros T h1 l a hi.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality.
  intro h2.
  unfold fun_from_cons.
  destruct (h1 x a) as [|h3]; subst.
  apply l_to_inds_aux_cons_in.
  rewrite (l_to_inds_aux_cons_in_in h1 _ _ _ h3 hi (in_cons _ _ _ h2)).
  f_equal; apply proof_irrelevance.
Qed.


Lemma fun_from_cons_l_to_inds_aux_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    ~In a l ->
    fun_from_cons (l_to_inds_aux pfdt (a::l)) =
    fun x pf => S (l_to_inds_aux pfdt l x pf).
  intros T h1 l a hni.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality.
  intro h2.
  assert (h3:x <> a).
    intro; subst; contradiction.
  pose proof (l_to_inds_aux_cons_nin_in h1 l a x h3 hni (in_cons a _ _ h2)) as h4.
  unfold fun_from_cons.
  erewrite l_to_inds_aux_functional3.
  rewrite h4 at 1.
  repeat f_equal.
  apply proof_irrelevance.
Qed.


Lemma l_to_inds_cons_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pf:In a l),
    l_to_inds pfdt (a::l) =
    (l_to_inds_aux pfdt l a pf) :: l_to_inds pfdt l.
  intros T h1 l a h2.
  unfold l_to_inds; simpl.
  pose proof (l_to_inds_aux_cons_in h1 l a h2 (in_eq _ _)) as h3.
  rewrite h3, fun_from_cons_l_to_inds_aux_in; auto.
Qed.

Lemma l_to_inds_cons_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pf:~In a l),
    l_to_inds pfdt (a::l) =
    0 :: lS (l_to_inds pfdt l).
  intros T h1 l a h2.
  unfold l_to_inds; simpl.
  pose proof (l_to_inds_aux_cons_nin h1 l a (in_eq _ _) h2) as h3.
  rewrite h3, fun_from_cons_l_to_inds_aux_nin; auto.
  repeat f_equal.
  rewrite lS_map_in_dep; auto.
Qed.






Lemma l_to_inds_aux_inj_dep :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    inj_dep (l_to_inds_aux pfdt l).
  intros T h1 l; red.
  intros x y h2 h3 h4.
  unfold l_to_inds_aux in h4.
  rewrite (lind_eq_iff _ _ _ _ _ (lt_lind _ _ _ _)) in h4.
  destruct h4; subst.
  rewrite (nth_lt_eq_iff h1).
  ex_goal.
  rewrite <- list_singularize_in_iff; auto.
  exists hex.
  exists 0.
  split.
  apply gt_lt.
  rewrite <- (count_occ_In h1).
  rewrite <- list_singularize_in_iff; auto.
  unfold lind; f_equal; apply proof_irrelevance.
Qed.


Lemma linds_l_to_inds : 
  forall {T:Type}
         (pfdt:type_eq_dec T)
         (l:list T),
         linds nat_eq_dec (l_to_inds pfdt l) =
         linds pfdt l.
  intros T h1 l.
  unfold linds, l_to_inds.
  rewrite (map_eq_iff (list_singularize nat_eq_dec (map_in_dep (l_to_inds_aux h1 l))) (linds_occ nat_eq_dec (map_in_dep (l_to_inds_aux h1 l)))).
  ex_goal.
  rewrite map_length.
  apply (length_list_singularize_map_in_dep_inj h1).
  apply l_to_inds_aux_inj_dep.
  exists hex.
  intros n h2.
  pose proof h2 as h2'.
  rewrite (length_list_singularize_map_in_dep_inj h1) in h2'.
  rewrite nth_lt_map'.
  rewrite linds_occ_eq_iff.
  ex_goal.
  rewrite length_linds_occ.
  pose proof (nth_lt_list_singularize_map_in_dep h1 nat_eq_dec (l_to_inds_aux h1 l) n (l_to_inds_aux_inj_dep h1 l) h2') as h3.
  simpl in h3. 
  erewrite nth_lt_functional3.
  rewrite h3 at 1.
  rewrite (count_occ_map_in_dep_inj h1).
  f_equal.
  apply nth_lt_functional3.
  apply l_to_inds_aux_inj_dep.
  exists hex0.
  intros m h4.
  pose proof (nth_lt_linds_occ h1 l (nth_lt (list_singularize h1 l) n
           (lt_eq_trans n
              (length (map (linds_occ h1 l) (list_singularize h1 l)))
              (length (list_singularize h1 l))
              (lt_eq_trans n
                 (length
                    (list_singularize nat_eq_dec
                       (map_in_dep (l_to_inds_aux h1 l))))
                 (length (map (linds_occ h1 l) (list_singularize h1 l))) h2
                 hex) (map_length (linds_occ h1 l) (list_singularize h1 l))))) as h5.
  hfirst h5.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  specialize (h5 hf m).
  hfirst h5.
  rewrite length_linds_occ.
  pose proof h4 as h4'.
  erewrite nth_lt_functional3 in h4'.
  rewrite nth_lt_list_singularize_map_in_dep in h4' at 1.
  rewrite count_occ_map_in_dep_inj in h4' at 1.
  apply h4'.
  apply l_to_inds_aux_inj_dep.
  specialize (h5 hf0).
  symmetry.
  erewrite nth_lt_functional3.
  rewrite h5 at 1.
  pose proof (lind_occ_map_in_dep_inj h1 nat_eq_dec (l_to_inds_aux h1 l) (nth_lt  _ _ h2')) as h6.
  hfirst h6.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  specialize (h6 hf1).
  hfirst h6.
  apply in_map_in_dep.
  specialize (h6 hf2 m).
  hfirst h6.
  apply l_to_inds_aux_inj_dep; auto.
  specialize (h6 hf3). 
  symmetry. 
  gen0.
  erewrite nth_lt_functional3.
  rewrite nth_lt_list_singularize_map_in_dep.

  intro  h7.
  pose proof (proof_irrelevance
                _ hf1
                (iff2
                   (list_singularize_in_iff h1 l
                                            (nth_lt (list_singularize h1 l) n h2'))
                   (in_nth_lt (list_singularize h1 l) n h2'))).
  subst.
  pose proof (proof_irrelevance _ hf2 h7); subst.
  rewrite h6 at 1.
  apply lind_occ_functional; auto.
  apply nth_lt_functional3.
  apply l_to_inds_aux_inj_dep.

  Grab Existential Variables.
  apply l_to_inds_aux_inj_dep.
  apply l_to_inds_aux_inj_dep.
  rewrite in_map_in_dep_iff.
  exists (nth_lt _ _ h2').
  ex_goal.
  rewrite list_singularize_in_iff.
  apply in_nth_lt.
  exists hex0.
  erewrite nth_lt_functional3.
  rewrite nth_lt_list_singularize_map_in_dep.
  apply l_to_inds_aux_functional3.

  Grab Existential Variables.
  apply l_to_inds_aux_inj_dep.
Qed.



Lemma linds_list_to_set_injable :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    list_to_set_injable (fun ll => ll = linds pfdt l).
  intros T h1 l; red; intros; subst; auto.
Qed.





Fixpoint count_occ_l {T:Type} (pfdt:type_eq_dec T) (l l':list T) : nat :=
  let pfdt' := list_eq_dec pfdt in
  match l' with
    | nil => if (pfdt' l nil) then 1 else 0
    | a::la => let co := count_occ_l pfdt l la in
               if (pfdt' l (firstn (length l) l')) then
                 S co 
               else co
  end.


(*Finds the index of the [S occ]th occureence of [l] in [l']
  as a sub_list.
  If [occ >= pred (count_occ_l)] then the value is the same as
  [pred(count_occ_l)]'s.*)
Fixpoint slind_occ
         {T:Type} (pfdt:type_eq_dec T) (l l':list T) (occ:nat) 
         (pfln:l <> nil) : l' <> nil -> sub_list l l' -> nat :=
  let pfdt' := list_eq_dec pfdt in
  match l' with 
    | nil => fun pfn _ => False_rect _ (pfn eq_refl) 
    | a::la => fun _  pfsl =>
                 match (nil_dec' la) with
                   | left _ => 0
                   | right pfnla =>
                     match (pfdt' l (firstn (length l) (a::la)))
                     with
                       | left _ =>
                         match occ with
                           | 0 => 0
                           | S n =>
                             match (sub_list_dec pfdt l la) with
                               | left pfsl' =>
                                 S (slind_occ pfdt l la n pfln pfnla pfsl')
                               | right _ => 0
                             end
                         end
                       | right pfn =>
                         let pf' := neq_sub_list_cons l la a pfn pfsl in
                         S (slind_occ pfdt l la  occ pfln pfnla pf')
                     end
                 end
  end.


    
(*This is the "right index" (corresponding to the last element in [l])
  of the [S occ]s occurence of [l] in [l'].*)
Definition slind_occ'
           {T:Type} (pfdt:type_eq_dec T)
           (l l':list T) (occ:nat)
           (pfn:l <> nil) (pfn':l' <> nil) 
           (pfsl:sub_list l l') : nat :=
  slind_occ pfdt l l' occ pfn pfn' pfsl     + (length l).


Definition slind {T:Type} (pfdt:type_eq_dec T) (l l':list T)
           (pfn:l <> nil) (pfn':l' <> nil) 
           (pfsl:sub_list l l') :=
  slind_occ pfdt l l' 0  pfn pfn' pfsl.


Definition slind'
           {T:Type} (pfdt:type_eq_dec T)
           (l l':list T)
           (pfn:l <> nil) (pfn':l' <> nil) 
           (pfsl:sub_list l l') : nat :=
  slind_occ' pfdt l l' 0 pfn pfn' pfsl.



Lemma slind_occ_functional :
  forall {T:Type} (pfdt:type_eq_dec T) (l l' l0 l0':list T) (occ occ':nat)
         (pfn:l <> nil) (pfn':l' <> nil)
         (pfn0:l0 <> nil) (pfn0':l0' <> nil) 
         (pf:sub_list l l0) (pf':sub_list l' l0'),
         l = l' ->
         l0 = l0' ->
         occ = occ' ->
         slind_occ pfdt l  l0  occ pfn pfn0  pf =
         slind_occ pfdt l' l0' occ' pfn' pfn0' pf'.
  intros; subst;
  pose proof (proof_irrelevance _ pf pf');
  pose proof (proof_irrelevance _ pfn pfn');
  pose proof (proof_irrelevance _ pfn0 pfn0');
  subst; auto.
Qed.


Lemma slind_occ'_functional' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l' l0 l0':list T) (occ occ':nat)
         (pfn:l <> nil) (pfn':l' <> nil)
         (pfn0:l0 <> nil) (pfn0':l0' <> nil) 
         (pf:sub_list l l0) (pf':sub_list l' l0'),
         l = l' ->
         l0 = l0' ->
         occ = occ' ->
    slind_occ' pfdt l  l0  occ pfn pfn0  pf =
    slind_occ' pfdt l' l0' occ' pfn' pfn0' pf'.
  intros; subst;
  pose proof (proof_irrelevance _ pf pf');
  pose proof (proof_irrelevance _ pfn pfn');
  pose proof (proof_irrelevance _ pfn0 pfn0');
  subst; auto.
Qed.


Lemma slind_functional :
  forall {T:Type} (pfdt:type_eq_dec T) (l l' l0 l0':list T)
         (pfn:l <> nil) (pfn':l' <> nil)
         (pfn0:l0 <> nil) (pfn0':l0' <> nil) 
         (pf:sub_list l l0) (pf':sub_list l' l0'),
         l = l' ->
         l0 = l0' ->
         slind pfdt l  l0 pfn pfn0  pf =
         slind pfdt l' l0' pfn' pfn0'  pf'.
  intros; subst;
  pose proof (proof_irrelevance _ pf pf');
  pose proof (proof_irrelevance _ pfn pfn');
  pose proof (proof_irrelevance _ pfn0 pfn0');
  subst; auto.
Qed.


Lemma slind_functional' :
  forall {T:Type} (pfdt:type_eq_dec T) (l l' l0 l0':list T)
         (pfn:l <> nil) (pfn':l' <> nil)
         (pfn0:l0 <> nil) (pfn0':l0' <> nil) 
         (pf:sub_list l l0) (pf':sub_list l' l0'),
         l = l' ->
         l0 = l0' ->
         slind' pfdt l  l0 pfn pfn0 pf =
         slind' pfdt l' l0' pfn' pfn0'  pf'.
  intros; subst;
  pose proof (proof_irrelevance _ pf pf');
  pose proof (proof_irrelevance _ pfn pfn');
  pose proof (proof_irrelevance _ pfn0 pfn0');
  subst; auto.
Qed.

  
Lemma slind_firstn_length :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T)
         (pf:l <> nil)
         (pf':l' <> nil)
         (pfsl:sub_list l l'),
    firstn (length l) l' = l ->
    slind pfdt l l' pf pf'  pfsl = 0.
  intros T h1 l l'. revert l.
  induction l' as [|a' l' ih]; simpl.
  intros ? ? h2.
  contradict h2; auto.
  intros l h2 h3 h4 h5; unfold slind; simpl.
  destruct (nil_dec' l') as [h6 | h6]; auto.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr]; auto.
  pose proof hr as hr'.
  rewrite h5 in hr'.
  contradict hr'; auto.
Qed.

Lemma slind_occ_lt :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T)
         (pf:l <> nil)
         (pf':l' <> nil)
         (occ:nat) (pfsl:sub_list l l'),
    slind_occ pfdt l l' occ pf pf' pfsl < length l'.
  intros T h1 l l'. revert l.
  induction l' as [|a' l' ih]; simpl.
  intros ? ? h2; contradict h2; auto.
  intros l h2 h3 occ h4.    
  destruct (nil_dec' l') as [|h5]; auto with arith.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  destruct occ as [|occ]; auto with arith.
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  apply lt_n_S.
  apply ih; auto with arith.
  auto with arith.
  apply lt_n_S, ih; auto.
Qed.


(*Returns the pair of surrounding lists of the first occurence of 
  [l] in [l'].*)
Definition sub_list_partition0 {T:Type} (pfdt:type_eq_dec T)
           (l l':list T) (pfn:l <> nil) (pfn':l' <> nil)
           (pfsl:sub_list l l') : list T * list T :=
  let n := slind pfdt l l' pfn pfn' pfsl in
  let n' := slind' pfdt l l' pfn pfn' pfsl in
  (firstn n l', skipn n' l').

 
Lemma sub_list_partition0_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T)
         (pfn:l <> nil) (pfn':l' <> nil)
         (pfsl:sub_list l l'),
    let lp := sub_list_partition0 pfdt l l' pfn pfn' pfsl in
    let lfp := fst lp in
    let lsp := snd lp in
    l' = lfp ++ l ++ lsp.
  intros T h1 l l'. revert l. 
  induction l' as [|a' l' ih]; simpl.
  intros ? ? h2.
  contradict h2; auto.
  intros l h2 h3 h4.
  unfold slind, slind', slind_occ'.
  simpl.
  destruct (nil_dec' l') as [h5 | h5].
  subst.
  simpl.
  unfold slind', slind_occ'.
  simpl.
  apply sub_list_sing_dec in h4.
  destruct h4; subst.
  contradict h2; auto. 
  simpl; auto.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr]. 
  simpl.
  rewrite hl.
  rewrite length_firstn.
  rewrite firstn_skipn; auto.
  apply length_sub_list; auto.
  simpl.
  f_equal.
  apply ih.
Qed.


Lemma map_in_dep_decompose :
  forall {T U:Type} (pfdt:type_eq_dec T)
         (l l':list T)
         (pfn: l  <> nil)
         (pfn':l' <> nil)
         (f:fun_in_dep l' U)
         (pfsl:sub_list l l'),
    let (lx, ly) := sub_list_partition0 pfdt l l' pfn pfn' pfsl in
    let n := slind pfdt l l' pfn pfn' pfsl in
    let n' := slind' pfdt l l' pfn pfn' pfsl in
    let pfslx := sub_list_firstn _ n in
    let pfsly := sub_list_skipn  _ n' in
    let fl := fun_in_dep_decompose1' pfdt l l' pfsl f in
    let flx := fun_in_dep_decompose1' pfdt _ l' pfslx f in
    let fly := fun_in_dep_decompose1' pfdt _ l' pfsly f in
    map_in_dep f = map_in_dep flx ++ map_in_dep fl ++ map_in_dep fly.
  unfold fun_in_dep_decompose1', fun_in_dep_decompose2',
  fun_in_dep_decompose1, fun_in_dep_decompose2;
  intros T U h1 l l' h2 h3 f h4.
  simpl.
  pose proof (sub_list_partition0_compat h1 l l' h2 h3 h4) as h6.
  simpl in h6.
  rewrite (map_in_dep_functional _  h6).
  do 2 rewrite map_in_dep_app.
  f_equal.
  unfold fun_from_app1.
  apply map_in_dep_ext.
  intros x h7.
  apply functional_extensionality.
  intro h8.
  rewrite transfer_fun_in_dep_compat.
  f_equal. apply proof_irrelevance.
  f_equal.
  apply map_in_dep_ext.
  intros x h7.
  unfold fun_from_app2, fun_from_app1.
  apply functional_extensionality.
  intro h8.
  rewrite transfer_fun_in_dep_compat.
  f_equal.
  apply proof_irrelevance.
  unfold fun_from_app2.
  apply map_in_dep_ext.
  intros x h7.
  apply functional_extensionality.
  intro h8.
  rewrite transfer_fun_in_dep_compat.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma map_in_dep_firstn_S :
  forall {T U:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (n:nat)
         (pfsn : S n <= length l)
         (f:fun_in_dep (firstn (S n) l) U),
    let pfsl := sub_list_firstn_le _ _ _ (le_n_Sn _) in
    let pflt := lt_le_trans _ _ _ (lt_n_Sn _) pfsn in
    let f' := fun_in_dep_decompose1' pfdt _ _ pfsl f in
    map_in_dep f = map_in_dep f' ++ f (nth_lt l n  pflt) (in_nth_lt_firstn' _  _  _ pfsn (lt_n_Sn _)) :: nil.
  intros T U h1 l n h2 f h3 h4.
  destruct n as [|n].
  destruct l as [|a l].
  simpl in h2.
  omega.
  simpl.
  repeat f_equal.
  apply proof_irrelevance.
  pose proof (map_in_dep_decompose h1 (firstn (S n) l) (firstn (S (S n)) l) (firstn_S_neq_nil _ _ (le_trans _ _ _ (le_n_Sn _) h2)) (firstn_S_neq_nil _ _ h2) f h3) as h5.
  simpl.
  simpl. simpl in h5.
  rewrite h5.
  unfold fun_in_dep_decompose1', fun_in_dep_decompose1.
  match goal with |- map_in_dep ?g' ++ _ ++ _ = _ => pose g' as g end.
  let P := type of g in
  match P with
      forall x:T, In x ?l' -> _  => pose l' as l0 end.
  assert (h8:l0 = nil).
  unfold l0.
  rewrite slind_firstn_length.
  simpl; auto.
  change (firstn (length (firstn (S n) l)) (firstn (S (S n)) l) =
          firstn (S n) l).
  rewrite length_firstn.
  rewrite <- firstn_le_nest; auto.
  auto with arith.
  fold l0 in g.
  fold l0 g.
  rewrite (map_in_dep_functional g h8).
  simpl.
  f_equal.
  match goal with |- map_in_dep ?g' = _ => pose g' as h end.
  let P := type of h in
  match P with
      forall x:T, In x ?l' -> _  => pose l' as l1 end.
  assert (h9:l1 = nth_lt l (S n) h4 :: nil).
    unfold l1.
    unfold slind', slind_occ'.
    match goal with
      |- skipn (slind_occ h1 ?x2 ?x3 0 ?x4 ?x5 ?x6 + ?x7) _  = _ =>  
      change (skipn (slind h1 x2 x3 x4 x5 x6 + x7) (firstn (S (S n)) l)  = nth_lt l (S n) h4 ::nil ) end.
    rewrite slind_firstn_length.
    simpl.
    change (skipn (length (firstn (S n) l)) (firstn (S (S n)) l) = nth_lt l (S n) h4::nil). 
  rewrite length_firstn_S.
  rewrite (firstn_S_eq _ _ h2). 
  rewrite skipn_app.
  rewrite skipn_firstn_nil.
  rewrite app_nil_l.
  repeat f_equal.
  apply proof_irrelevance.
  rewrite length_firstn; auto with arith.
  omega.
  change (firstn (length (firstn (S n) l)) (firstn (S (S n)) l) =
          firstn (S n) l).
  rewrite length_firstn.
  rewrite <- firstn_le_nest; auto.
  auto with arith.
  fold l1 in h.
  fold l1 h.
  rewrite (map_in_dep_functional h h9).
  simpl.
  rewrite transfer_fun_in_dep_compat.
  unfold h.
  repeat f_equal.
  apply proof_irrelevance.
Qed.


Lemma firstn_map_in_dep_removelast_S :
  forall {T U:Type} {l:list T} (f:fun_in_dep l U)
         (n:nat),
    S n <= length l ->
    firstn n (map_in_dep (fun x pf => f x (incl_removelast l _ pf))) =
    firstn n (map_in_dep f). 
  intros T U l f n h1.
  pose proof (removelast_eq l) as h2.
  pose proof (map_in_dep_functional  (fun (x : T) (pf : (fun x0 : T => In x0 (removelast l)) x) =>
         f x (incl_removelast l x pf)) h2) as h3.
  rewrite h3 at 1.
  pose proof (firstn_map_in_dep_firstn f n (pred (length l))) as h4.
  hfirst h4.
  omega.
  specialize (h4 hf).
  hfirst h4.
  auto with arith.
  specialize (h4 hf0).
  rewrite <- h4.
  repeat f_equal.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality.
  intro h5.
  rewrite transfer_fun_in_dep_compat.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma firstn_map_in_dep :
  forall {T U:Type} (pfdt:type_eq_dec T)
         (l:list T) (f:fun_in_dep l U) (n:nat)
         (pfle:n <= length l),
    firstn n (map_in_dep f) =
    map_in_dep (fun x (pf:In x (firstn n l)) =>
                  f x
                    (incl_firstn _ _ pfle _ pf)).
  intros T U hdt l f n.
  revert f. revert l. 
  induction n as [|n ih]; auto.
  intros l f h0.
  specialize (ih (removelast l) (fun x pf => f x (incl_removelast l _ pf))). 
  hfirst ih.
  rewrite length_removelast; omega.
  specialize (ih hf).
  erewrite firstn_S_eq.
  rewrite firstn_map_in_dep_removelast_S in ih at 1.
  rewrite ih. 
  symmetry. 
  erewrite map_in_dep_firstn_S.
  pose proof (@map_in_dep_functional). 
  pose proof (firstn_removelast l n) as h1.
  hfirst h1. 
  rewrite length_removelast; omega. 
  specialize (h1 hf0).
  pose proof (map_in_dep_functional
                 (fun (x : T) (pf : In x (firstn n (removelast l))) =>
                    f x (incl_removelast l x (incl_firstn (removelast l) n hf x pf)))
                 h1) as h2.
  rewrite h2 at 1.
  f_equal. 
  f_equal.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality.
  intro h3. 
  rewrite transfer_fun_in_dep_compat.
  unfold fun_in_dep_decompose1', fun_in_dep_decompose1.
  f_equal.
  apply proof_irrelevance.
  rewrite nth_lt_map_in_dep'.
  f_equal. 
  apply f_equal_dep. 
  apply nth_lt_functional3.
  assumption.
 
  Grab Existential Variables.
  assumption.
  assumption.
  rewrite length_map_in_dep; auto with arith.
Qed.



Lemma nth_lt_lind_occ_l_to_inds :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (occ:nat) (x:T) (pfi:In x l)
         (pf:lind_occ nat_eq_dec (l_to_inds pfdt l) occ
                      (l_to_inds_aux pfdt l x pfi) (in_l_to_inds_aux pfdt l x pfi) < length l),
    (nth_lt l
            (lind_occ nat_eq_dec (l_to_inds pfdt l) occ
                      (l_to_inds_aux pfdt l x pfi) (in_l_to_inds_aux pfdt l x pfi))
            pf)  = x.
  intros T h1 l occ x h2 h3.
  pose proof (lind_occ_compat nat_eq_dec (l_to_inds h1 l) occ (l_to_inds_aux h1 l x h2)
                              (in_l_to_inds_aux h1 l x h2)) as h4.
  unfold l_to_inds in h4.
  rewrite nth_lt_map_in_dep' in h4.
  apply l_to_inds_aux_inj_dep in h4.
  rewrite <- h4.
  apply nth_lt_functional3.
Qed.


  
Lemma count_occ_l_to_inds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l),
    count_occ nat_eq_dec (l_to_inds pfdt l) (l_to_inds_aux pfdt l x pf) =
    count_occ pfdt l x.
  intros T h1 l. 
  induction l as [|a l ih]; simpl; auto.
  intros x h2.
  destruct h2 as [|h2]; subst.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  rewrite eq_dec_same at 1.
  f_equal.
  destruct (in_dec h1 x l) as [h2 | h2].
  rewrite fun_from_cons_l_to_inds_aux_in.
  rewrite (l_to_inds_aux_cons_in _ _ _ h2).
  apply ih.
  assumption.
  pose proof h2 as h2'.
  rewrite (count_occ_In h1) in h2'.
  apply not_lt0_iff in h2'.
  rewrite h2'.
  rewrite fun_from_cons_l_to_inds_aux_nin.
  rewrite (l_to_inds_aux_cons_nin _ _ _ _ h2).
  apply count_occ_gt0.
  intros n h3.
  rewrite in_map_in_dep_iff in h3.
  destruct h3 as [y [h3 h4]]; subst; auto with arith.
  assumption.
  assumption.
  contradict hr.
  f_equal; apply proof_irrelevance.
 
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  destruct (in_dec h1 a l) as [h3 | h3].
  rewrite (l_to_inds_aux_cons_in _ _ _ h3) in hl.
  destruct (h1 a x) as [h4 | h4]; subst.
  f_equal. 
  rewrite fun_from_cons_l_to_inds_aux_in; auto.
  rewrite (l_to_inds_aux_cons_in _ _ _ h2).
  apply ih.
  apply neq_sym in h4.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h4) in hl.
  apply l_to_inds_aux_inj_dep in hl; subst.
  contradict h4; auto.
  assumption.
  rewrite l_to_inds_aux_cons_nin in hl; auto.
  destruct (h1 a x) as [|h4]; subst.
  contradiction.
  apply neq_sym in h4.
  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h4) in hl; auto.
  discriminate.
  destruct (h1 a x) as [|h3]; subst.
  contradict hr.
  repeat f_equal.
  apply proof_irrelevance.
  apply neq_sym in h3.
  destruct (in_dec h1 a l) as [h4 | h4].
  rewrite fun_from_cons_l_to_inds_aux_in; auto.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h3); auto.
  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h3); auto.
  rewrite fun_from_cons_l_to_inds_aux_nin; auto.
  rewrite <- (ih _ h2).
  rewrite <- lS_map_in_dep.
  rewrite count_occ_lS.
  simpl.
  repeat f_equal.
  apply proof_irrelevance.
  auto with arith.
Qed.


  
Lemma count_occ_firstn_lind_occ_l_to_inds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l)
         (occ:nat),
    occ < count_occ pfdt l x ->
    count_occ pfdt
     (firstn
        (lind_occ nat_eq_dec (l_to_inds pfdt l) occ (l_to_inds_aux pfdt l x pf)
           (in_l_to_inds_aux pfdt l x pf)) l) x = occ.
  intros T h1 l.
  induction l as [|a l ih].
  intros; contradiction.
  intros x h2 occ h3.
  destruct (h1 a x) as [|h4]; subst.
  destruct (in_dec h1 x l) as [h4 | h4].
  generalize (in_l_to_inds_aux h1 (x::l) x h2).
  rewrite (l_to_inds_aux_cons_in _ _ _ h4).
  rewrite (l_to_inds_cons_in _ _ _ h4).
  destruct occ as [|occ].
  simpl.
  intro h5.
  rewrite eq_dec_same at 1.
  simpl.
  reflexivity.
  apply nat_eq_dec.
  rewrite count_occ_cons_eq in h3; auto.
  apply lt_S_n in h3.
  intro h5.
  erewrite lind_occ_functional3.
  rewrite lind_occ_hd at 1.
  simpl.
  rewrite eq_dec_same at 1.
  f_equal.
  specialize (ih _ h4 occ h3).
  erewrite lind_occ_functional3.
  rewrite ih at 1; auto.
  assumption.
  rewrite count_occ_cons_eq in h3; auto.
  pose proof h4 as h4'.
  rewrite (count_occ_In h1) in h4'. 
  apply ngt_nlt in h4'.
  rewrite not_lt0_iff in h4'.
  rewrite h4' in h3.
  assert (occ = 0). omega.
  subst.
  simpl.
  pose proof (proof_irrelevance _ h2 (in_eq _ _)); subst.
  rewrite eq_dec_same at 1.
  simpl; auto.
  apply nat_eq_dec.
  rewrite count_occ_cons_neq in h3; auto.
  destruct h2 as [|h2]; subst.
  contradict h4; auto.
  destruct (in_dec h1 a l) as [h5 | h5].
  generalize (in_l_to_inds_aux h1 (a::l) x (or_intror h2)).
  apply neq_sym in h4.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h4).
  rewrite (l_to_inds_cons_in _ _ _ h5).
  intro h6.
  simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  apply l_to_inds_aux_inj_dep in hl.
  contradiction.
  simpl.
  destruct (h1 a x) as [|h7]; subst.
  contradict (h4 eq_refl).
  specialize (ih _ h2 occ h3).
  erewrite lind_occ_functional3.
  pose proof (proof_irrelevance
                _ h2
                (neq_in_cons l x a h4 (or_intror h2))) as h8.
  rewrite h8 in ih.
  apply ih.
  assumption.
  generalize (in_l_to_inds_aux h1 (a::l) x (or_intror h2)).
  rewrite l_to_inds_cons_nin; auto.
  apply neq_sym in h4.
  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h4); auto. 
  intro h6.
  pose proof h6 as h6'.
  inversion h6'.
  discriminate.
  rewrite in_lS_iff in H.
  destruct H as [m [h7 h8]].
  apply S_inj in h8.
  subst. 
    
  erewrite lind_occ_functional3.
  rewrite lind_occ_cons_in_neq.

  simpl.
  destruct (h1 a x) as [|h8].
  subst.
  contradict (h4 eq_refl).
  unfold lS.
  rewrite lind_occ_map_inj at 1.
  apply ih.
  assumption.
  red; intros; apply S_inj; auto.
  auto with arith.

  Grab Existential Variables.
  rewrite in_lS_iff.
  exists (l_to_inds_aux h1 l x (neq_in_cons l x a h4 (or_intror h2))).
  split; auto.
  rewrite count_occ_l_to_inds; auto.
  left; auto.
Qed.


Lemma lind_occ_l_to_inds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (occ:nat) (x:T)
         (pf:In x l),
    occ < count_occ pfdt l x ->
    lind_occ pfdt l occ x pf =
    lind_occ nat_eq_dec (l_to_inds pfdt l) occ (l_to_inds_aux pfdt l x pf) (in_l_to_inds_aux _ _ _ _).
  intros T h1 l occ x h2 holt.  
  rewrite lind_occ_eq_iff.
  split.
  gen0. 
  rewrite (lind_occ_map_in_dep_inj h1 nat_eq_dec (l_to_inds_aux h1 l)).
  intro h3.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat; auto.
  apply l_to_inds_aux_inj_dep.
  split.
  erewrite firstn_S_eq.
  rewrite count_occ_app.
  simpl.
  rewrite nth_lt_lind_occ_l_to_inds.
  rewrite eq_dec_same at 1; auto.
  rewrite S_compat.
  f_equal.
  
  pose proof (firstn_S_eq _ _  (lt_lind_occ nat_eq_dec (l_to_inds h1 l) occ (l_to_inds_aux h1 l x h2) (in_l_to_inds_aux _ _ _ _))) as h3.
  
  unfold l_to_inds in h3 at 1.
  unfold l_to_inds in h3.
  rewrite firstn_map_in_dep in h3 at 1.
  apply count_occ_firstn_lind_occ_l_to_inds.
  assumption.
  assumption.

  rewrite count_occ_skipn.
  f_equal.
  erewrite firstn_S_eq.
  rewrite count_occ_app.
  simpl.
  rewrite nth_lt_lind_occ_l_to_inds.
  rewrite eq_dec_same at 1.
  rewrite S_compat.
  f_equal.
  apply count_occ_firstn_lind_occ_l_to_inds; auto.
  assumption.
  assumption.

  Grab Existential Variables.
  eapply lt_eq_trans.
  apply lt_lind_occ.
  rewrite length_l_to_inds; auto.
  eapply lt_eq_trans.
  apply lt_lind_occ.
  rewrite length_map_in_dep; auto.
  eapply lt_eq_trans.
  apply lt_lind_occ.
  apply length_l_to_inds.
  eapply lt_eq_trans.
  apply lt_lind_occ.
  apply length_l_to_inds.
Qed.
 

Lemma linds_occ_l_to_inds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pf:In x l),
    linds_occ pfdt l x =
    linds_occ nat_eq_dec (l_to_inds pfdt l) (l_to_inds_aux pfdt l x pf).
  intros T h1 l x h2.
  rewrite linds_occ_eq_iff.
  ex_goal.
  rewrite length_linds_occ.
  rewrite count_occ_l_to_inds; auto.
  exists hex.
  intros m h3.
  erewrite nth_lt_linds_occ.
  apply lind_occ_l_to_inds; auto.
Qed.


Lemma nth_lts_eq_cons_l_to_inds_aux :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (a:T) (a':U)
         (pfe:length (a::l) = length (a'::l'))
         (pfi:In a l) (pfi':In a' l'),
    (forall (m n : nat) (pfm : m < length (a :: l))
            (pfn : n < length (a :: l)),
       let pfm' :=
           lt_eq_trans m (length (a :: l)) (length (a' :: l')) pfm pfe in
       let pfn' :=
           lt_eq_trans n (length (a :: l)) (length (a' :: l')) pfn pfe in
       nth_lt (a :: l) m pfm = nth_lt (a :: l) n pfn <->
       nth_lt (a' :: l') m pfm' = nth_lt (a' :: l') n pfn') ->
    l_to_inds_aux pfdt l a pfi =
    l_to_inds_aux pfdu l' a' pfi'.
  intros T U hdt hdu l.
  induction l as [|a l ih].
  intros; contradiction.
  intro l'.
  destruct l' as [|a' l'].
  intros; contradiction.
  intros b b' h2 h3 h4 h5.
  destruct h3 as [|h3], h4 as [|h4]; subst.
  destruct (in_dec hdt b l) as [h6 | h6],
                               (in_dec hdu b' l') as [h7 | h7].
  rewrite (l_to_inds_aux_cons_in _ _ _ h6).
  rewrite (l_to_inds_aux_cons_in _ _ _ h7).
  eapply ih.
  intros m n h8 h9 h10 h11.
  specialize (h5 (S m) (S n)).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  simpl.
  destruct m as [|m], n as [|n]; auto; try tauto. 

  specialize (h5 (S (S (lind hdt _ _ h6))) 0).
  hfirst h5. 
  simpl; do 2 apply lt_n_S.
  apply lt_lind.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  erewrite nth_indep in h5.
  erewrite <- nth_lt_compat in h5.
  rewrite lind_compat in h5.
  hl h5; auto. 
  rewrite h5 in hl.
  pose proof (@nth_In _ (lind hdt l b h6) l' b') as h8.
  hfirst h8.
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite <- h2.
  apply lt_lind.
  specialize (h8 hf1).
  rewrite hl in h8.
  contradiction.
  apply lt_lind.

  specialize (h5 (S (S (lind hdu _ _ h7))) 0).
  hfirst h5. 
  simpl; do 2 apply lt_n_S.
  simpl in h2.
  pose proof h2 as h2'.
  do 2 apply S_inj in h2'.
  rewrite h2'.
  apply lt_lind.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  symmetry in h5.
  erewrite nth_indep in h5.
  erewrite <- nth_lt_compat in h5.
  rewrite lind_compat in h5.
  hl h5; auto. 
  rewrite h5 in hl.
  pose proof (@nth_In _ (lind hdu l' b' h7) l b) as h8.
  hfirst h8.
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite h2.
  apply lt_lind.
  specialize (h8 hf1).
  rewrite hl in h8.
  contradiction.
  apply lt_lind.

  rewrite l_to_inds_aux_cons_nin; auto.
  rewrite l_to_inds_aux_cons_nin; auto.

  destruct (hdu b' a') as [|h6]; subst.
  destruct (in_dec hdt b l) as [h6 | h6].
  rewrite (l_to_inds_aux_cons_in _ _ _ h6); auto.
  rewrite (l_to_inds_aux_cons_in _ _ _ h4); auto.
  eapply ih.
  intros m n h7 h8 h9 h10.
  simpl in h5; simpl.
  destruct m as [|m], n as [|n].
  tauto.

  specialize (h5 1 (S (S n))).
  hfirst h5.
  apply lt_n_S; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  do 2 apply lt_n_S.
  simpl in h8; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  assumption.

  specialize (h5 (S (S m)) 1).
  hfirst h5.
  apply lt_n_S; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  assumption.

  simpl in h7, h8.
  specialize (h5 (S (S m)) (S (S n))); auto with arith.

  specialize (h5 0 (S (S (lind hdu _ _ h4)))).
  simpl in h5.
  hfirst h5; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  do 2 apply lt_n_S.
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite h2.
  apply lt_lind.
  hfirst h5.
  do 2 apply lt_n_S.
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite h2.
  apply lt_lind.
  hfirst h5.
  specialize (h5 hf1).
  hr h5.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  rewrite lind_compat.
  reflexivity.
  apply lt_lind.
  rewrite <- h5 in hr.
  pose proof (@nth_In _ (lind hdu l' a' h4) l b) as h7.
  hfirst h7. 
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite h2.
  apply lt_lind.
  do 2 apply lt_n_S.
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite h2.
  apply lt_lind.
  specialize (h5 hf2).
  hr h5.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  rewrite lind_compat.
  reflexivity.
  apply lt_lind.
  rewrite <- h5 in hr. 
  pose proof (@nth_In _ (lind hdu l' a' h4) l b) as h7.
  hfirst h7.
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite h2.
  apply lt_lind.
  specialize (h7 hf3).
  rewrite <- hr in h7.
  contradiction. 
  destruct (in_dec hdt b l) as [h7 | h7].
  rewrite (l_to_inds_aux_cons_in _ _ _ h7).
  destruct (in_dec hdu a' l') as [hi | hi].
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h6); auto.
  eapply ih.
  intros m n h9 h10 h11 h12.
  destruct m as [|m], n as [|n]; simpl.
  tauto.
 
  specialize (h5 0 1).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  hl h5; auto.
  rewrite h5 in hl.
  contradiction.

  specialize (h5 0 1).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  hl h5; auto.
  rewrite h5 in hl.
  contradiction.
 
  specialize (h5 (S (S m)) (S (S n))).
  simpl in h5.
  hfirst h5; auto with arith.

  specialize (h5 0 1).
  simpl in h5.
  hfirst h5; auto with arith.
  specialize (h5 hf).
  hfirst h5; auto with arith.
  specialize (h5 hf0).
  hl h5; auto.
  rewrite h5 in hl.
  subst; auto.
  contradict h6; auto.
  specialize (h5 0 1).
  simpl in h5.
  hfirst h5; auto with arith.
  specialize (h5 hf).
  hfirst h5; auto with arith.
  specialize (h5 hf0).
  hl h5; auto.
  rewrite h5 in hl; contradiction.

  destruct (hdt a b) as [h6 | h6]; subst.
  destruct (in_dec hdu b' l') as [h7 | h7].
  rewrite (l_to_inds_aux_cons_in _ _ _ h3).
  rewrite (l_to_inds_aux_cons_in _ _ _ h7).
  eapply ih.
  intros m n h6 h8 h9.
  specialize (h5 (S m) (S n)).
  simpl in h5; simpl.
  auto with arith.

  specialize (h5 0 (S (S (lind hdt _ _  h3)))).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; do 2 apply lt_n_S. 
  apply lt_lind.
  specialize (h5 hf0).
  simpl in h5.
  hl h5.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  rewrite h5 in hl.
  pose proof (@nth_In _ (lind hdt l b h3) l' b') as h8.
  hfirst h8.
  simpl in h2.
  do 2 apply S_inj in h2.
  rewrite <- h2.
  apply lt_lind.
  specialize (h8 hf1).
  rewrite <- hl in h8; contradiction.

  destruct (in_dec hdt a l) as [h7 | h7],
                               (in_dec hdu b' l') as [h8 | h8].
  apply neq_sym in h6.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h6).
  rewrite (l_to_inds_aux_cons_in _ _ _  h8).
  eapply ih.
  intros m n h9 h10 h11 h12.
  simpl in h5.
  simpl.
  destruct m as [|m], n as [|n].
  tauto. 
  specialize (h5 0 (S (S n))).
  auto with arith.
  simpl in h9.
  pose proof h9 as h9'.
  apply lt_S_n in h9'. 
  specialize (h5 (S (S m)) 0); auto with arith.   
  specialize (h5 (S (S m)) (S (S n))); simpl in h5; auto with arith.
  assumption.

  specialize (h5 0 (S (S (lind hdt _ _ h3)))).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5; simpl.
  simpl in h2.
  do 2 apply lt_n_S.
  apply lt_lind.
  specialize (h5 hf0).
  simpl in h5.
  hl h5.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  rewrite h5 in hl.
  pose proof (@nth_In _ (lind hdt l b h3) l' b') as h9.
  hfirst h9.
  simpl in h2.
  pose proof h2 as h2'.
  do 2 apply S_inj in h2'.
  rewrite <- h2'.
  apply lt_lind.
  specialize (h9 hf1).
  rewrite <- hl in h9.
  contradiction.

  specialize (h5 0 1).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  apply neq_sym in h6.
  rewrite h5 in h6.
  contradict h6; auto.

  specialize (h5 0 1).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  apply neq_sym in h6.
  rewrite h5 in h6.
  contradict h6; auto.

  destruct (hdt a b) as [h7 | h7], (hdu a' b') as [h6 | h6]; subst.
  rewrite (l_to_inds_aux_cons_in _ _ _ h3).
  rewrite (l_to_inds_aux_cons_in _ _ _ h4).
  eapply ih.
  intros m n h6 h7 h8 h9.
  specialize (h5 (S m) (S n)).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf). 
  hfirst h5; simpl; auto with arith. 
  specialize (h5 hf0); auto.
  specialize (h5 0 1).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  hl h5; auto.
  rewrite h5 in hl; subst.
  contradict h6; auto.
  specialize (h5 0 1).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  hr h5; auto.
  rewrite <- h5 in hr; subst.
  contradict h7; auto. 
  apply neq_sym in h7.
  apply neq_sym in h6.

  destruct (in_dec hdt a l) as [hi | hi],
                           (in_dec hdu a' l') as [hi' | hi'].
  rewrite (l_to_inds_aux_cons_in_in _ _ _  _ h7); auto with arith.
  rewrite (l_to_inds_aux_cons_in_in _ _ _  _ h6); auto with arith.
  eapply ih. 
  intros m n h8 h9 h10 h11.
  simpl; destruct m as [|m], n as [|n].
  tauto.
  specialize (h5 0 (S (S n))).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf). 
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf0); simpl in h5; auto with arith.
 
  specialize (h5 (S (S m)) 0).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf). 
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf0); simpl in h5; auto with arith.

  specialize (h5 (S (S m)) (S (S n))).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf). 
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf0); simpl in h5; auto with arith.
 
  specialize (h5 1 (S (S (lind hdt _ _ hi)))).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5; simpl.
  do 2 apply lt_n_S.
  apply lt_lind.
  specialize (h5 hf0).
  simpl in h5.
  hl h5.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  rewrite h5 in hl.
  pose proof (@nth_In _ (lind hdt l a hi) l' b') as h8.
  hfirst h8.
  simpl in h2, hf0.
  do 2 apply S_inj in h2.
  rewrite <- h2.
  apply lt_lind.
  specialize (h8 hf1).
  rewrite <- hl in h8.
  contradiction.

  specialize (h5 (S (S (lind hdu _ _ hi'))) 1).
  hfirst h5; simpl; auto with arith.
  do 2 apply lt_n_S.
  simpl in h2.
  pose proof h2 as h2'.
  do 2 apply S_inj in h2'.
  rewrite h2'.
  apply lt_lind.
  specialize (h5 hf).
  hfirst h5; simpl; auto with arith.
  specialize (h5 hf0).
  simpl in h5.
  hr h5.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  rewrite <- h5 in hr.
  pose proof (@nth_In _ (lind hdu l' a' hi') l b) as h8.
  hfirst h8.
  simpl in h2, hf0.
  do 2 apply S_inj in h2.
  rewrite h2.
  apply lt_lind.
  specialize (h8 hf1).
  rewrite hr in h8.
  contradiction.

  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h7); auto.
  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h6); auto.
 
  f_equal.
  eapply ih.
  intros m n h8 h9 h10 h11.
  simpl.
  destruct m as [|m], n as [|n].
  tauto.
  
  specialize (h5 0 (S (S n))).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith. 
  specialize (h5 hf0); auto.


  specialize (h5 (S (S m)) 0).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith. 
  specialize (h5 hf0); auto.

  specialize (h5 (S (S m)) (S (S n))).
  hfirst h5.
  simpl; auto with arith.
  specialize (h5 hf).
  hfirst h5.
  simpl; auto with arith. 
  specialize (h5 hf0); auto.

  Grab Existential Variables.
  simpl in h2; simpl; auto with arith.
  simpl in h2; simpl; auto with arith.
  simpl in h2; simpl; auto with arith.
  simpl in h2; simpl; auto with arith.
  simpl in h2; simpl; auto with arith.
  simpl in h2; simpl; auto with arith.
  simpl in h2; simpl; auto with arith.
  simpl in h2; simpl; auto with arith.
Qed.


Lemma nth_lts_eq_l_to_inds :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U) (pfe:length l = length l'),
         (forall (m n:nat)
                      (pfm:m < length l)
                      (pfn:n < length l),
                 let pfm' := lt_eq_trans _ _ _ pfm pfe in
                 let pfn' := lt_eq_trans _ _ _ pfn pfe in
                 nth_lt l m pfm =
                 nth_lt l n pfn <->
                 nth_lt l' m pfm' =
                 nth_lt l' n pfn') ->
         l_to_inds pfdt l =
         l_to_inds pfdu l'.
  intros T U hdt hdu l.
  induction l as [|a l ih].
  intros l' h1.
  pose proof h1 as h1'.
  symmetry in h1'.
  apply length_eq0 in h1'; subst; auto.
  intro l'.
  destruct l' as [|a' l'].
  simpl.
  intro; discriminate.
  intros h1 h2.
  pose proof h1 as h1'.
  simpl in h1'.
  apply S_inj in h1'.
  destruct (in_dec hdt a l) as [h3 | h3],
                               (in_dec hdu a' l') as [h4 | h4].
  rewrite (l_to_inds_cons_in _ _ _ h3).
  rewrite (l_to_inds_cons_in _ _ _ h4).
  f_equal.
  eapply nth_lts_eq_cons_l_to_inds_aux.
  apply h2.
  eapply ih.
  intros m n h5 h6 h7 h8.
  specialize (h2 (S m) (S n)).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf0).
  simpl in h2; simpl; auto.
  rewrite nth_lt_compat.
  rewrite (nth_lt_compat l n).
  rewrite (nth_lt_compat l').
  rewrite (nth_lt_compat l' n).

  erewrite nth_indep.
  erewrite (@nth_indep _ l n).
  erewrite (@nth_indep _ l').
  erewrite (@nth_indep _ l' n).
  apply h2.
  assumption.
  assumption.
  assumption.
  assumption.

  specialize (h2 0 (S (lind hdt _ _ h3))).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf).
  hfirst h2.
  simpl; apply lt_n_S.
  apply lt_lind.
  specialize (h2 hf0).
  simpl in h2.
  hl h2.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  rewrite h2 in hl.
  pose proof (@nth_In _ (lind hdt l a h3) l' a') as h5.
  hfirst h5.
  rewrite <- h1'.
  apply lt_lind.
  specialize (h5 hf1).
  rewrite <- hl in h5. 
  contradiction.

  specialize (h2 (S (lind hdu _ _ h4)) 0).
  hfirst h2.
  simpl; auto with arith.
  simpl in h1.
  rewrite h1.
  apply lt_n_S.
  apply lt_lind.
  specialize (h2 hf).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf0).
  simpl in h2.
  hr h2.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite lind_compat; auto.
  apply lt_lind.
  rewrite <- h2 in hr.
  pose proof (@nth_In _ (lind hdu l' a' h4) l a) as h5.
  hfirst h5.
  rewrite h1'.
  apply lt_lind.
  specialize (h5 hf1).
  rewrite hr in h5.
  contradiction.

  rewrite (l_to_inds_cons_nin _ _ _ h3).
  rewrite (l_to_inds_cons_nin _ _ _ h4).
  f_equal.
  f_equal.
  eapply ih.
  intros m n h5 h6 h7 h8.
  specialize (h2 (S m) (S n)).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf).
  hfirst h2.
  simpl; auto with arith.
  specialize (h2 hf0).
  simpl in h2.

  rewrite nth_lt_compat.
  rewrite (nth_lt_compat l n).
  rewrite (nth_lt_compat l').
  rewrite (nth_lt_compat l' n).

  erewrite nth_indep.
  erewrite (@nth_indep _ l n).
  erewrite (@nth_indep _ l').
  erewrite (@nth_indep _ l' n).
  apply h2.
  assumption.
  assumption.
  assumption.
  assumption.

  Grab Existential Variables.
  assumption.
  assumption.
Qed.
  

Fixpoint count_occl_aux {T:Type} (pfdt:type_eq_dec T) (l l':list T) :
  sub_list l l' -> nat :=
  let pfdt' := list_eq_dec pfdt in
  match l' with
    | nil => fun _ => 0
    | a :: la => fun pf =>
                   match pfdt' l (firstn (length l) (a::la)) with
                     | left _ =>
                       match sub_list_dec pfdt l la with
                         | left pfsl => S (count_occl_aux pfdt l la pfsl)
                         | right _ => 1
                       end
                     | right pfn =>
                       let pf' := neq_sub_list_cons l la a pfn pf in
                       count_occl_aux pfdt l la pf'
                   end
  end.

Definition count_occl {T:Type} (pfdt:type_eq_dec T) (l l':list T) :=
  let pfdt' := list_eq_dec pfdt in
  match (sub_list_dec pfdt l l') with
    | left pfsl =>
      count_occl_aux pfdt l l' pfsl
    | right _ => 0
  end.

  
Lemma count_occl_aux_nil1 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (pf:sub_list nil l),
    count_occl_aux pfdt nil l pf = length l.
  intros T h1 l.
  induction l as [|a l ih]; simpl; auto.
  intro h2.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr']; auto with arith.
  contradict hr'.
  apply sub_list_nil1.
  contradict hr; auto.
Qed.
 
  
Lemma count_occl_nil1 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    count_occl pfdt nil l = length l.
  intros T h1 l.
  unfold count_occl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  revert hl.
  intros; apply count_occl_aux_nil1.
  contradict hr.
  apply sub_list_nil1.
Qed.


Lemma count_occl_aux_nil2 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (pf:sub_list l nil),
    count_occl_aux pfdt l nil pf  = 0.
  auto.
Qed.

Lemma count_occl_nil2 :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    count_occl pfdt l nil = 0.
  intros T h1 l.
  unfold count_occl.
  lr_match_goal; fold hlr; destruct hlr; auto.
Qed.

Lemma count_occl_cons2_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a':T),
    l = firstn (length l) (a' :: l') ->
    count_occl pfdt l (a'::l') =
    S (count_occl pfdt l l').
  intros T h1 l l' a' h3.
  unfold count_occl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl0 | hr0].
  lr_match_goal; fold hlr; destruct hlr as [hl1 | hr1].
  repeat f_equal; apply proof_irrelevance.
  contradiction.
  contradiction.
  simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl0 | hr0].
  lr_match_goal; fold hlr; destruct hlr as [hl1 | hr1].
  contradiction.
  reflexivity.
  contradiction.
  contradict hr.
  rewrite h3.
  apply sub_list_firstn.
Qed.

Lemma count_occl_cons2_neq :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T) (a':T),
    l <> firstn (length l) (a' :: l') ->
    count_occl pfdt l (a'::l') =
    count_occl pfdt l l'.
  intros T h1 l l' a' h2.
  unfold count_occl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl0 | hr0].
  contradiction.
  repeat f_equal; apply proof_irrelevance.
  pose proof (neq_sub_list_cons _ _ _ h2 hl).
  contradiction.
  lr_match_goal; fold hlr; destruct hlr as [hl1 | hr1].
  contradict hr.
  apply sub_list_cons; auto.
  reflexivity.
Qed.

  
Lemma count_occl_sub_list_eq_le :
  forall {T:Type} (pfdt:type_eq_dec T) (l1 l2 l:list T),
    l1 <> nil ->
    l2 <> nil ->
    l1 = firstn (length l1) l2 ->
    count_occl pfdt l2 l <= count_occl pfdt l1 l. 
  intros T h1 l1 l2 l. revert l1 l2. 
  induction l as [|a l ih]; simpl.
  intros; do 2 rewrite count_occl_nil2; auto.
  intros l1 l2 hn1 hn2 he.
  pose proof (length_sub_list l1 l2) as h5.
  hfirst h5.
  rewrite he.
  apply sub_list_firstn.
  specialize (h5 hf).
  destruct (list_eq_dec h1 l2 (firstn (length l2) (a :: l))) as [h3 | h3].
  rewrite count_occl_cons2_eq; auto.
  destruct (list_eq_dec h1 l1 (firstn (length l1) (a :: l))) as [h4 | h4].  rewrite count_occl_cons2_eq; auto with arith. 
  apply le_lt_eq_dec in h5.
  destruct h5 as [h5 | h5].
  rewrite he in h4 at 1.
  rewrite h3 in h4.
  rewrite <- firstn_lt_nest in h4; auto.
  contradict h4; auto.
  rewrite h5 in he.
  rewrite firstn_length in he.
  rewrite <- he in h3.
  contradiction.
  destruct (list_eq_dec h1 l1 (firstn (length l1) (a :: l))) as [h4 | h4].
  rewrite count_occl_cons2_neq.
  rewrite count_occl_cons2_eq.
  eapply le_trans.
  apply ih.
  apply hn1.
  assumption.
  assumption.
  auto with arith.
  assumption.
  assumption.
  rewrite count_occl_cons2_neq; auto.
  rewrite count_occl_cons2_neq; auto. 
Qed.


                
Lemma sub_list_count_occl :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    l <> nil -> l' <> nil ->
    sub_list l l' ->
    0 < count_occl pfdt l l'.
  intros T h1 l l'. revert l.
  induction l' as [|a' l' ih]; simpl. 
  intros ? ? h2. contradict h2; auto.
  intros l h2 h3 h4.
  unfold count_occl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  lr_match_goal; fold hlr; destruct hlr as [hl0 | hr0]; auto with arith. 
  destruct (nil_dec' l') as [|h5]; subst.
  apply sub_list_sing_dec in h4.
  destruct h4 as [|h4]; try contradiction.
  subst.
  simpl in hr'.
  contradict hr'; auto.
  pose proof (neq_sub_list_cons _ _ _ hr' h4) as h6.
  specialize (ih _ h2 h5 h6).
  unfold count_occl in ih.
  lr_match ih; fold hlr in ih; destruct hlr as [hl0 | hr0].
  pose proof (proof_irrelevance
                _ hl0
                (neq_sub_list_cons l l' a' hr' hl)); subst; auto.
  omega.
  contradiction.
Qed.

  
Lemma not_sub_list_count_occl :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    ~sub_list l l' ->
    count_occl pfdt l l' = 0.
  intros T h1 l l' h2.
  unfold count_occl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr]; auto.
  contradiction.
Qed.

Lemma count_occl_aux_same :
  forall {T:Type} (pfdt:type_eq_dec T) 
         (l:list T)
         (pf:l <> nil)
         (pfsl:sub_list l l),
    count_occl_aux pfdt l l pfsl = 1.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intro h2; contradict h2; auto.
  intros h2 h3. 
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  lr_match_goal; fold hlr; destruct hlr as [hl' | hr'].
  pose proof (length_sub_list _ _ hl') as h4.
  simpl in h4.
  omega.
  reflexivity.
  contradict hr.
  f_equal.
  rewrite firstn_length; auto.
Qed.

Lemma count_occl_same :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T),
    l <> nil ->
    count_occl pfdt l l = 1.
  intros T h1 l h2.
  unfold count_occl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  apply count_occl_aux_same; auto.
  contradict hr.
  apply sub_list_refl.
Qed.


Fixpoint remove1 {T U:Type} (pfdt:type_eq_dec T) (l:list (T*U)) (x:T) :=
  match l with 
    | nil => nil
    | pr::l' => if (pfdt (fst pr) x) then 
                  (remove1 pfdt l' x) else
                  pr::(remove1 pfdt l' x)
  end.

Fixpoint remove2 {T U:Type} (pfdt:type_eq_dec U) (l:list (T*U)) (y:U) :=
  match l with 
    | nil => nil
    | pr::l' => if (pfdt (snd pr) y) then 
                  (remove2 pfdt l' y) else
                  pr::(remove2 pfdt l' y)
  end.

Lemma remove1_nin : forall {T U:Type} (pfdt:type_eq_dec T)
                           (l:list (T*U)) (x:T) (y:U),
                      ~In (x, y) (remove1 pfdt l x).
intros T U h0 l.
induction l as [|l pr h1].
simpl. auto.
simpl.
intros x y.
destruct (h0 (fst l) x) as [h2 | h3].
apply h1.
intro h4.
destruct h4 as [h4 | h5].
contradict h3.
rewrite h4. simpl. reflexivity.
pose proof (h1 x y).
contradiction.
Qed.

Lemma remove2_nin : forall {T U:Type} (pfdt:type_eq_dec U)
                           (l:list (T*U)) (x:T) (y:U),
                      ~In (x, y) (remove2 pfdt l y).
intros T U h0 l.
induction l as [|l pr h1].
simpl. auto.
simpl.
intros x y.
destruct (h0 (snd l) y) as [h2 | h3].
apply h1.
intro h4.
destruct h4 as [h4 | h5].
contradict h3.
rewrite h4. simpl. reflexivity.
pose proof (h1 x y).
contradiction.
Qed.


Lemma remove1_in :
  forall {T U:Type} (pfdt:type_eq_dec T) (l:list (T*U)) (pr:T*U) (x:T),
         fst pr <> x -> In pr l -> In pr (remove1 pfdt l x).
intros T U h0 l.
induction l as [|a l h1].
simpl.
auto.
simpl.
intros pr x h2 h3.
destruct (h0 (fst a) x) as [h4 | h5].
subst.
destruct h3 as [h3l | h3r].
subst.
contradict h2. reflexivity.
apply h1; assumption.
destruct h3 as [h3l | h3r].
subst. left. reflexivity.
right.
apply h1; assumption.
Qed.

Lemma remove2_in :
  forall {T U:Type} (pfdt:type_eq_dec U) (l:list (T*U)) (pr:T*U) (y:U),
         snd pr <> y -> In pr l -> In pr (remove2 pfdt l y).
intros T U h0 l.
induction l as [|a l h1].
simpl.
auto.
simpl.
intros pr y h2 h3.
destruct (h0 (snd a) y) as [h4 | h5].
subst.
destruct h3 as [h3l | h3r].
subst.
contradict h2. reflexivity.
apply h1; assumption.
destruct h3 as [h3l | h3r].
subst. left. reflexivity.
right.
apply h1; assumption.
Qed.


Lemma list_to_set_remove1_inc : 
  forall {T U:Type} (pfdt:type_eq_dec T) (lp:list (T*U)) (x:T),
         Included (list_to_set (remove1 pfdt lp x)) (list_to_set lp). 
intros T U h0 lp.
induction lp as [|pr lp h1].
simpl. auto with sets.
intros x.
red.
simpl.
intros pr' h2.
destruct (h0 (fst pr) x) as [h3 | h4].
subst.
specialize (h1 (fst pr)).
left.
auto with sets.
simpl in h2.
specialize (h1 x).
destruct h2 as [pr'' h5 | pr'' h6].
left.
auto with sets.
destruct h6.
right. constructor.
Qed.

Lemma list_to_set_remove2_inc : 
  forall {T U:Type} (pfdt:type_eq_dec U) (lp:list (T*U)) (y:U),
         Included (list_to_set (remove2 pfdt lp y)) (list_to_set lp). 
intros T U h0 lp.
induction lp as [|pr lp h1].
simpl. auto with sets.
intros y.
red.
simpl.
intros pr' h2.
destruct (h0 (snd pr) y) as [h3 | h4].
subst.
specialize (h1 (snd pr)).
left.
auto with sets.
simpl in h2.
specialize (h1 y).
destruct h2 as [pr'' h5 | pr'' h6].
left.
auto with sets.
destruct h6.
right. constructor.
Qed.

Lemma remove1_inc : 
  forall {T U:Type} (pfdt:type_eq_dec T) (lp:list (T*U)) (x:T) (pr:T*U),
    In pr (remove1 pfdt lp x) -> In pr lp.
intros T U h0 lp x pr h1.
rewrite list_to_set_in_iff in h1.
rewrite list_to_set_in_iff.
pose proof (list_to_set_remove1_inc h0 lp x) as h2.
auto with sets.
Qed.


Lemma remove2_inc : 
  forall {T U:Type} (pfdt:type_eq_dec U) (lp:list (T*U)) (y:U) (pr:T*U),
    In pr (remove2 pfdt lp y) -> In pr lp.
intros T U h0 lp y pr h1.
rewrite list_to_set_in_iff in h1.
rewrite list_to_set_in_iff.
pose proof (list_to_set_remove2_inc h0 lp y) as h2.
auto with sets.
Qed.
 

Lemma remove1_list_to_set_union :
  forall {T U:Type} (pfdt:type_eq_dec T) (lab:list (T*U)) (x:T)
  (A:Ensemble T) (B:Ensemble U),
    ~ Inn A x -> 
    list_to_set lab = 
    Union (cart_prod (Singleton x) B) (cart_prod A B) -> 
    list_to_set (remove1 pfdt lab x) = cart_prod A B. 
intros T U hdt lab x A B h0  h2.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros pr h4.
pose proof (list_to_set_remove1_inc hdt lab x) as h5.
rewrite h2 in h5.
assert (h6:Inn  (Union (cart_prod (Singleton x) B) (cart_prod A B)) pr). auto with sets.
destruct h6 as [pr h7 | pr h8].
destruct h7 as [h7].
destruct h7 as [h7l h7r].
rewrite surjective_pairing in h4.
destruct h7l. 
rewrite <- list_to_set_in_iff in h4.
pose proof (remove1_nin hdt lab x (snd pr)).
contradiction.
assumption.
(* >= *)
red.
intros pr h3.
rewrite <- list_to_set_in_iff.
assert (h4:fst pr <> x).
  intro h4.
  destruct h3 as [h3].
  destruct h3 as [h3l h3r].
  subst.
  contradiction.
apply remove1_in; try assumption.
rewrite list_to_set_in_iff.
rewrite h2.
right. assumption.
Qed.

Lemma remove1_no_dup :
  forall {T U:Type} (pfdt:type_eq_dec T)
         (l:list (T*U)) (x:T),
    NoDup l -> NoDup (remove1 pfdt l x).
intros T U h0 l.
induction l as [|a l h1].
simpl. auto.
simpl.
intros x h2.
pose proof (no_dup_cons _ _ h2) as h3.
specialize (h1 x h3).
destruct (h0 (fst a) x) as [h4 | h5].
assumption.
constructor.
intro h6.
pose proof (remove1_inc _ _ _ _ h6).
inversion h2.
contradiction.
assumption.
Qed.

Lemma remove2_no_dup :
  forall {T U:Type} (pfu:type_eq_dec U)
         (l:list (T*U)) (y:U),
    NoDup l -> NoDup (remove2 pfu l y).
intros T U h0 l.
induction l as [|a l h1].
simpl. auto.
simpl.
intros y h2.
pose proof (no_dup_cons _ _ h2) as h3.
specialize (h1 y h3).
destruct (h0 (snd a) y) as [h4 | h5].
assumption.
constructor.
intro h6.
pose proof (remove2_inc _ _ _ _ h6).
inversion h2.
contradiction.
assumption.
Qed.

Lemma in_firstn_hd :
  forall {T:Type} (l:list T) (x:T) (n:nat),
    0 < n ->
    In x (firstn n (x::l)).
  intros T l x n. revert x l.
  induction n as [|n h1]; simpl; auto.
  intros; omega.
Qed.


Lemma in_skipn_impl :
  forall {T:Type} (l:list T) (x:T) (n:nat),
    In x (skipn n l) ->
    In x l.
  intros T l x n. revert x l.
  induction n as [|n h2]; simpl; auto.
  intros x l.
  destruct l as [|a l]; auto.
  intro h1.
  specialize (h2 _ _ h1).
  right; auto.
Qed.


Lemma no_dup_firstn :
  forall {T:Type} (l:list T) (n:nat),
    NoDup l ->
    NoDup (firstn n l).
  intros T l n h1. revert h1. revert l.
  induction n as [|n h1]; simpl.
  intros. constructor.
  intros l h2. 
  destruct l. constructor.
  pose proof (no_dup_cons_nin _ _ h2) as h6.
  pose proof (no_dup_cons _ _ h2) as h4.
  specialize (h1 _ h4).
  constructor.
  contradict h6.
  eapply in_firstn. apply h6.
  assumption.
Qed.

Lemma no_dup_firstn_S :
  forall {T:Type} (l:list T) (n:nat)
         (pfsn:S n <= length l),
    NoDup (firstn n l) -> ~In (nth_lt l n pfsn) (firstn n l) ->
    NoDup (firstn (S n) l).
  intros T l n h1 h2 h3.
  rewrite (firstn_S_eq _ _ h1).
  apply no_dup_app; auto.
  constructor.
  intro; contradiction.
  constructor.
  simpl.
  rewrite add_empty_sing.
  rewrite list_to_set_in_iff in h3.
  rewrite nin_iff in h3; auto.
Qed.


  
Lemma in_linds_val_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (pfx:In x l),
    forall (pf:In (linds_occ pfdt l x) (linds pfdt l)),
      in_linds_val pfdt l (linds_occ pfdt l x) pf = x.
  intros T h1 l x h2 h3.
  unfold in_linds_val.
  pose proof (nth_lt_in_linds_occ h1 l x h2 _ (in_hd' (linds_occ h1 l x) (in_linds_not_nil h1 l (linds_occ h1 l x) h3))) as h4.
  rewrite <- h4.
  apply nth_lt_functional3.
Qed.


Lemma nth_lt_lind_in_linds_val :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat) (m:nat)
         (pfi:In ln (linds pfdt l))
         (pfm:In m ln),
    let pfm' := in_linds_lt pfdt _ _ _ pfm pfi in
    nth_lt l _ (lt_lind pfdt _ _ (in_linds_val_in _ _ _ pfi)) =
    nth_lt l m pfm'.
  intros T h1 l ln m h2 h3 h4.
  rewrite lind_compat.
  unfold in_linds_val.
  apply in_linds_nth_lt_eq.
Qed.

Lemma in_lind_in_linds_val :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (pf:In ln (linds pfdt l)),
    In (lind pfdt _ _ (in_linds_val_in _ _ _ pf)) ln.
  intros T h1 l ln h2.
  pose proof h2 as h2'.
  rewrite in_linds_iff in h2'.
  destruct h2' as [x [[h3 h4] h5]]; subst.
  rewrite in_linds_occ_iff.
  exists 0, h3. split.
  apply lt_gt.
  rewrite <- (count_occ_In h1 l x); auto.
  unfold lind.
  apply lind_occ_functional; auto.
  rewrite in_linds_val_linds_occ; auto.
Qed.


Lemma in_lind_occ_in_linds_val :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat)
         (pfiln:In ln (linds pfdt l))
         (occ:nat)
         (pfo:occ < count_occ pfdt l (in_linds_val pfdt l ln pfiln)),
    In (lind_occ pfdt l occ (in_linds_val pfdt l ln pfiln) (in_linds_val_in pfdt l ln pfiln)) ln.
  intros T h1 l ln h2 occ h3. 
  pose proof h2 as h2'.
  rewrite in_linds_iff in h2'.
  destruct h2' as [x [[h4 h5] h6]]; subst.
  rewrite in_linds_val_linds_occ in h3; auto.
  rewrite in_linds_occ_iff.
  exists occ, h4.
  split; auto.
  symmetry.
  gen0.
  rewrite in_linds_val_linds_occ.
  intro h7.
  apply lind_occ_functional; auto.
  assumption.
Qed.


Lemma linds_occ_in_linds_val : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (ln:list nat)
         (pfln:In ln (linds pfdt l)),
    linds_occ pfdt l
              (in_linds_val pfdt l ln pfln) = ln.
  intros T h1 l ln h2.
  pose proof h2 as h2'.
  rewrite in_linds_iff in h2'.
  destruct h2' as [x [[h3 h4] h5]]; subst.
  apply linds_occ_functional; auto.
  rewrite in_linds_val_linds_occ; auto.
Qed.

Lemma lind_occ_in_linds_val :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (i:nat)
         (pflt:i < length (linds pfdt l)) (ln:list nat)
         (pfiln:In ln (linds pfdt l))
         (occ:nat) (pfo:occ < length ln),
    lind_occ pfdt l occ (in_linds_val pfdt l ln pfiln) (in_linds_val_in _ _ _ _) =
    nth_lt ln occ pfo.
  intros T h1 l i h2 ln h3 occ h4.
  pose proof h3 as h3'.
  rewrite in_linds_iff in h3'.
  destruct h3' as [x [[h5 h6] h7]]; subst.
  gen0.
  rewrite in_linds_val_linds_occ.
  intro h8.
  erewrite nth_lt_linds_occ.
  reflexivity.
  assumption.
Qed.

  
Lemma nth_lt_linds_occ_in_linds_val :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (i:nat)
         (pflt:i < length (linds pfdt l)),
    nth_lt (linds pfdt l) i pflt =
    linds_occ pfdt l
              (in_linds_val pfdt l (nth_lt (linds pfdt l) i pflt)
                            (in_nth_lt (linds pfdt l) i pflt)). 
  intros; rewrite linds_occ_in_linds_val; auto.
Qed.

Lemma in_lind_occ_nth_lt_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (i:nat)
         (occ:nat)
         (pflt:i < length (linds pfdt l)),
    occ < count_occ pfdt l
                    (in_linds_val pfdt l (nth_lt (linds pfdt l) i pflt)
                                  (in_nth_lt (linds pfdt l) i pflt)) ->
    In
      (lind_occ pfdt l occ
                (in_linds_val pfdt l (nth_lt (linds pfdt l) i pflt)
                              (in_nth_lt (linds pfdt l) i pflt))
                (in_linds_val_in pfdt l (nth_lt (linds pfdt l) i pflt)
                                 (in_nth_lt (linds pfdt l) i pflt))) (nth_lt (linds pfdt l) i pflt).
  intros; apply in_lind_occ_in_linds_val; auto.
Qed.

Lemma in_linds_val_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pfi:In x l)
         (ln:list nat) (pfiln:In ln (linds pfdt l)),
    In (lind pfdt l x pfi) ln ->
    in_linds_val pfdt l ln pfiln = x.
  intros T h1 l x h2 ln h3 h4.
  unfold in_linds_val.
  pose proof h3 as h3'.
  rewrite in_linds_iff in h3'.
  destruct h3' as [x' [[h7 h5] h6]]; subst.
  pose proof (nth_lt_in_linds_occ h1 _ _ h7 _ h4) as h8.
  erewrite nth_lt_functional3 in h8.
  rewrite lind_compat in h8.
  subst.
  pose proof (in_linds_val_linds_occ h1 _ _ h7 h3) as h5.
  rewrite <- h5.
  unfold in_linds_val.
  reflexivity.
Qed.

  
  
Lemma nth_lt_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (i:nat)
         (pfi:i < length (linds pfdt l)),
    let pfin := in_nth_lt _ _ pfi in
    let x := in_linds_val pfdt _ _ pfin in
    nth_lt (linds pfdt l) i pfi =
    linds_occ pfdt l x. 
  intros; rewrite  nth_lt_linds_occ_in_linds_val; auto.
Qed.


Lemma in_remove_ind_or :
  forall {T} (l:list T) (x:T) (i:nat) (pfi:i < length l),
    In x l ->
    x = nth_lt l i pfi \/ In x (remove_ind l i).
  intros T l; induction l as [|a l ih]; simpl.
  intro; contradiction.
  intros x i; destruct i as [|i].
  intros; intuition.
  intro h1; apply lt_S_n in h1; intros h2;
  destruct h2 as [|h2]; subst.
  right; left; auto.
  specialize (ih x i h1 h2).
  destruct ih as [|h3]; subst.
  left.
  rewrite nth_lt_compat; apply nth_indep; auto.
  right; right; auto.
Qed.  

Definition rem_nth_lt
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (i:nat) (pfi:i < length l) :
  list T :=
  remove pfdt (nth_lt _ _ pfi) l.


Lemma interp_remove_nth_lind :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfn:NoDup l) (pfi:In x l),
    let n := lind pfdt l x pfi in
    l = interp (remove pfdt x l) (nth n l x) n.
  intros T h0 l.
  induction l as [|a l h1]; simpl; auto.
  intros; contradiction.
  intros x h2 h3. 
  pose proof (no_dup_cons _ _ h2) as h4.
  pose proof (no_dup_cons_nin _ _ h2) as h5. 
  destruct (h0 x a) as [h6 | h6].
  subst.
  pose proof (lind_cons_same h0 l a h3) as h7.
  rewrite h7.
  rewrite <- (remove_not_in' _  _ _ h5).
  unfold interp.
  simpl.
  reflexivity. 
  destruct h3 as [h3 | h3]. subst.
  contradict h6; auto. 
  unfold interp; simpl.
  rewrite lind_cons_neq.
  specialize (h1 _ h4 h3). simpl in h1. 
  simpl. unfold interp in h1.
  f_equal.
  rewrite h1 at 1.
  reflexivity.
  assumption.
Qed.


(*This one allows for duplicates.*)
Definition interp_in {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (x:T) (pf:In x l) :=
  let n := lind pfdt l x pf in
  let l' := firstn n l in
  let l'' := skipn (S n) l in
  l' ++ (x::nil) ++ l''.

Definition interp_nin 
  {T:Type} (pfdt:type_eq_dec T)
  (l:list T) (x:T) (pf:In x l) : Prop :=
  let n := lind pfdt l x pf in
  let l' := firstn n l in
  let l'' := skipn (S n) l in
  ~In x l' /\ ~In x l''.


Lemma no_dup_interp_nin :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T) (pf:In x l),
    NoDup l -> interp_nin pfdt l x pf.
  intros T h1 l.
  induction l as [|c l ih]; simpl.
  intros; contradiction.
  intros x h2 h3.
  pose proof (no_dup_cons_nin _ _  h3) as h4.
  pose proof (no_dup_cons _ _  h3) as h5.
  red; simpl.
  unfold lind; simpl.
  destruct (h1 x c) as [h6 | h6]; subst; simpl.
  split; auto.
  specialize (ih _ (neq_in_cons _ _ _ h6 h2) h5).
  red in ih.
  split.
  contradict ih.
  destruct ih as [|ih]; subst.
  contradict h6; auto.
  intro h7.
  destruct h7; contradiction.
  destruct ih as [h11 h12].
  destruct l as [|d l]; simpl; auto.
Qed.


Lemma interp_in_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T)
         (pfi:In x l),
    l = interp_in pfdt l x pfi.
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2; simpl.
  unfold interp_in, lind; simpl.
  destruct (h1 x a) as [h3 | h3]; subst; simpl; auto.
  f_equal.
  destruct h2 as [h2 | h2]; simpl; subst.
  contradict h3; auto.
  specialize (ih _ h2).
  rewrite ih at 1.
  unfold interp_in.
  f_equal.
  unfold lind.
  f_equal. f_equal. apply proof_irrelevance.
  simpl.
  f_equal.
  destruct l as [|c l]; auto.
  unfold lind.
  f_equal. f_equal. apply proof_irrelevance.
Qed.



  
Lemma remove_interp :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T) (n:nat),
    ~In a l -> NoDup l ->
    remove pfdt a (interp l a n) = l.
  intros T h0 l a n. revert l a.
  induction n as [|n h1]; simpl; auto. 
  intros l a h2 h3.
  destruct (h0 a a) as [h1 |h1]. 
  rewrite  <- (remove_not_in' _ _ _ h2); auto.
  contradict h1; auto.
  intro l.
  destruct l as [| a l].
  intros x h2 h3; simpl.
  destruct (h0 x x) as [h4 | h4]; auto.
  contradict h4; auto.
  intros x h3 h4.
  simpl in h3.
  apply not_or_and in h3.
  destruct h3 as [h3 h5].
  pose proof (no_dup_cons _ _ h4) as h6.
  specialize (h1 _ _ h5 h6).
  simpl.
  destruct (h0 x a) as [h7 | h7]. clear h1. subst.
  contradict h3; auto.
  f_equal.
  rewrite <- h1 at 3.
  reflexivity.
Qed.


Lemma no_dup_interp :
  forall {T:Type} (l:list T) (x:T) (n:nat),
    ~In x l -> NoDup l -> n <= length l ->
    NoDup (interp l x n).
  intros T l x n. revert x l.
  unfold interp.
  induction n as [|n h1]; simpl; auto.
  intros x l h1 h2 h3.
  constructor; auto.
  intros x l h2 h3 h4.
  destruct l as [|a l]; simpl.
  constructor; auto.
  simpl in h2.
  apply not_or_and in h2.
  destruct h2 as [h2a h2b].
  pose proof (no_dup_cons _ _ h3) as h5.
  specialize (h1 x l h2b h5).
  simpl in h4.
  assert (h7:n <= length l). omega.
  specialize (h1 h7).
  simpl in h1.
  constructor; auto.
  intro h8.
  apply in_app_or in h8.
  pose proof (no_dup_cons_nin _ _ h3) as h9. 
  destruct h8 as [h8 | h8].
  contradict h9.
  eapply in_firstn. apply h8.
  simpl in h8. 
  destruct h8 as [h8 | h8]; subst.
  contradict h2a; auto.
  contradict h9.
  eapply in_skipn_impl.
  apply h8.
Qed.

  
Lemma list_in_no_dup_remove_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T),
    ~In a l ->
    NoDup l ->
    [l':list T | NoDup l' /\ In a l' /\ remove pfdt a l' = l] =
    Im (seg_weak (length l)) (interp l a).
  intros T hdt l a hnin h0.
  apply Extensionality_Ensembles; red; split; red; intros l' h1. 
  destruct h1 as [h1]. destruct h1 as [h1 [h2 h3]].
  subst.
  pose (lind  hdt l' a h2) as n.  
  assert (h3: n <= length (remove hdt a l')).
    unfold n.
    rewrite length_remove_no_dup; auto.
    unfold n.
    pose proof (lt_lind hdt l' a h2) as h3.
    omega.
  apply Im_intro with n.
  constructor; auto.
  pose proof (nth_lt_lind_eq hdt l'  a h2) as h4. simpl in h4.
  rewrite <- h4 at 2.
  pose proof (nth_lt_compat l' (lind hdt l' a h2) (lt_lind hdt l' a h2)) as h5.
  simpl in h5.
  rewrite h5.
  rewrite h4.
  unfold n.
  apply (interp_remove_nth_lind hdt _ _ h1 h2).
  inversion h1 as [n h2 ? h3]. subst.
  destruct h2 as [h2].
  constructor. split.
  apply no_dup_interp; auto.
  split.
  unfold interp.
  simpl.
  rewrite in_app_iff.
  right. left; auto.
  apply remove_interp; auto.
Qed.


Lemma finite_list_reps :
  forall {T:Type} (E:Ensemble T),
    Finite E ->
    Finite (list_reps E).
  intros T E h1.
  induction h1 as [|S h2 h3 a h4]; simpl.
  rewrite list_reps_empty.
  apply Singleton_is_finite.
  rewrite list_reps_add; auto.
  apply Finite_Finite_Union; auto.
  intros U h5. 
  destruct h5 as [l h5]. subst.
  destruct h5 as [h5]. destruct h5 as [h5 h6]. subst.
  assert (h7:~In a l).
    rewrite <- list_to_set_in_iff in h4.
    assumption.
  rewrite (list_in_no_dup_remove_eq _ _ _ h7 h6).
  apply finite_image.
  apply finite_seg_weak.
  apply finite_image.
  assumption.
Qed.



 
Lemma destruct_list_cart_prod : 
  forall {T U:Type}
         (A:Ensemble T) (B:Ensemble U) 
         (lab:list (T * U)),
    lab <> nil ->
    NoDup lab -> list_to_set lab = cart_prod A B ->
    exists (la:list T) (lb:list U),
      list_to_set la = A /\ list_to_set lb = B /\            
      NoDup la /\ NoDup lb /\
      list_to_set lab = list_to_set (list_prod la lb).
intros T U  A B lab h1 h2 h3.
pose proof (list_to_set_finite lab) as h4.
assert (h5:list_to_set lab <> Empty_set _).
  intro h5.
  pose proof (empty_set_nil _ h5).
  contradiction.
rewrite h3 in h5.
rewrite h3 in h4.
assert (h6: A <> Empty_set _).
  intro h6.
  rewrite h6 in h5.
  rewrite cart_prod_empty in h5.
  contradict h5. reflexivity.
assert (h7: B <> Empty_set _).
  intro h7.
  rewrite h7 in h5.
  rewrite cart_prod_empty' in h5.
  contradict h5. reflexivity. 
pose proof (not_empty_Inhabited _ _ h6) as h8.
pose proof (not_empty_Inhabited _ _ h7) as h9.
pose proof (cart_prod_fin_rev1 _ _ h4 h9) as h10.
pose proof (cart_prod_fin_rev2 _ _ h4 h8) as h11. 
pose proof (finite_set_list_no_dup  _ h10) as h12.
pose proof (finite_set_list_no_dup  _ h11) as h13.
destruct h12 as [la h12].
destruct h13 as [lb h13].
exists la, lb.
destruct h12 as [h12l h12r].
destruct h13 as [h13l h13r]. 
symmetry in h12l. symmetry in h13l.
repeat split; try assumption.
rewrite h3.
symmetry.
apply list_prod_cart_prod_compat; assumption.
Qed.


Lemma in_list_power_add :
  forall {T U:Type} (la:list T) (lb:list U)
         (lp:list (T*U)) (a:T) (b:U),
    In b lb ->
    In lp (list_power la lb) ->
    In ((a, b)::lp) (list_power (a::la) lb).
intros T U la lb lp a b h2 h3.
simpl.
rewrite in_flat_map.
exists lp. split; auto.
rewrite in_map_iff.
exists b. split; auto.
Qed.
 

Lemma fpl_cons :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfu:type_eq_dec U)
         (la:list T) (lb:list U)
         (lp:list (T*U)) (pr:T*U),
    In pr lp ->
    NoDup ((fst pr)::la) ->
    functionally_paired_l ((fst pr)::la) lb lp ->
    functionally_paired_l la lb (remove (impl_type_eq_dec_prod pfdt pfu) pr lp). 
intros T U hdt hdu la lb lp pr h0 h1 h4.
constructor.
inversion h4 as [h5 h6].
intros x h7.
specialize (h5 x (in_cons (fst pr) x la h7)).
destruct h5 as [y h5].
exists y.
red.  red in h5.
pose proof NoDup_remove_2.
rewrite <- app_nil_l in h1.
pose proof (NoDup_remove_2 _ _ _ h1) as h8.
rewrite app_nil_l in h8.
assert (h9:(fst pr) <> x).
  intro; subst; contradiction.
split.
destruct h5 as [h5l h5r].
destruct h5l as [h5a h5b].
split; try assumption.
assert (h10:(x,y) <> pr).
  intro h11.
  rewrite surjective_pairing in h11.
  inversion h11.
  subst.
  contradict h9.
  reflexivity.
apply remove_a_in_eq; assumption.
intros u h10.
destruct h5 as [h5l h5r].
destruct h10 as [h10l h10r].
assert (h11:(x,u) <> pr).
  intro h12.
  rewrite surjective_pairing in h12.
  inversion h12.
  subst.
  contradict h9.
  reflexivity.
pose proof (in_remove_neq_in_l _ _ _ _ h10r h11) as h12.
apply (h5r _ (conj h10l h12)).

intros pr' h5.
destruct (eq_dec pr pr') as [h6 | h7].
subst.
pose proof (remove_In (impl_type_eq_dec_prod hdt hdu) lp pr').
contradiction.
apply neq_sym_iff in h7.
pose proof (in_remove_neq_in_l _ _ _  _ h5 h7) as h8.
pose proof (fpl_in_dom  h4 _ h8) as h9.
pose proof (fpl_in_ran  h4 _ h8) as h10.
split.
destruct h9 as [h9l | h9r].
rewrite (surjective_pairing pr) in h0.
rewrite h9l in h0.
rewrite (surjective_pairing pr') in h8.
pose proof (fpl_functional h4 _ _ _ h0 h8) as h11.
contradict h7.
rewrite (surjective_pairing pr). rewrite (surjective_pairing pr').
rewrite h9l. rewrite h11.
reflexivity.
assumption.
assumption.
Qed.


Lemma fpl_cons_nin : 
  forall {T U:Type} (la:list T) (lb:list U)
         (lp:list (T*U)) (a:T) (b:U), 
    ~In a la -> In b lb ->  
    functionally_paired_l la lb lp ->
    functionally_paired_l (a::la) lb ((a, b) :: lp).
intros T U la lb lp a b h1 h2 h3.
constructor.
intros x h4.
destruct h4 as [h4l | h4r]. subst.
exists b. red.
split. split; auto. left. reflexivity.
intros b' h4.
destruct h4 as [h4l h4r].
destruct h4r as [h5 | h6].
inversion h5; subst. reflexivity.
destruct h3 as [h3l h3r].
specialize (h3r _ h6).
simpl in h3r.
destruct h3r; contradiction. 
inversion h3 as [h3l h3r].
specialize (h3l _ h4r).
destruct h3l as [y h3l].
red in h3l.
exists y. red.
destruct h3l as [h5 h6].
destruct h5 as [h5l h5r].
split.
split; auto. right. auto.
intros y' h7. 
destruct h7 as [h7l h7r].
destruct h7r as [h8 | h9].
inversion h8.  subst.
contradiction.
specialize (h3r _ h9).
specialize (h6 _ (conj h7l h9)).
assumption.
intros pr h4.
destruct h4 as [h4l | h4r].
rewrite (surjective_pairing pr) in h4l.
inversion h4l. subst.
split. left. reflexivity. assumption.
destruct h3 as [h3l h3r].
specialize (h3r _ h4r).
destruct h3r as [h5  h6].
split. right. auto. assumption.
Qed.



Inductive synced {T U:Type} (la:list T) (lp:list (T*U)) :=
  | synced_intro : map (fun pr => (fst pr)) lp = la ->
                   synced la lp.

Lemma synced_hd : forall {T U:Type} (def:T*U) (la:list T) 
                           (lp:list (T*U)),
                      synced la lp ->
                      fst (hd def lp) = hd (fst def) la.
intros T U def la lp.
induction la as [|a la h1]; induction lp as [|pr lp h2].
simpl. auto.
intro h3.
destruct h3 as [h3].
simpl in h3.
discriminate.
simpl.
intro h2.
destruct h2 as [h2].
simpl in h2.
discriminate.
intro h3.
simpl.
destruct h3 as [h3].
simpl in h3.
inversion h3.
reflexivity.
Qed.

Lemma synced_nil1 : forall {T U:Type} (lp:list (T*U)),
                      synced nil lp -> lp = nil.
intros T U lp.
induction lp as [|pr lp h1].
auto.
intro h2.
destruct h2 as [h2].
simpl in h2.
discriminate.
Qed.

Lemma synced_nil2 : forall {T U:Type} (la:list T),
                      synced la (@nil (T*U)) -> la = nil.
intros T U la.
induction la as [|a la h1].
auto.
intro h2.
destruct h2 as [h2].
simpl in h2.
discriminate.
Qed.


Lemma synced_cons : forall {T U:Type}  (la:list T)
                           (lp:list (T*U)) (a:T) (pr:(T*U)),
                      synced (a::la) (pr::lp) ->
                      synced la lp.
intros T U la lp a pr h1.
constructor.
destruct h1 as [h1].
simpl in h1.
pose proof (cons_inj _ _ _ _ h1) as h2.
destruct h2; assumption.
Qed.


Lemma synced_cons' : forall {T U:Type}  (la:list T)
                           (lp:list (T*U)) (a:T),
                      synced (a::la) lp ->
                      synced la (tl lp).
intros T U la lp.
revert la.
induction lp as [|pr lp h1].
simpl.
intros la a h1.
destruct h1 as [h1].
simpl in h1.
discriminate.
intros la a h2. simpl.
apply (synced_cons _ _ a pr).
assumption.
Qed.


Lemma synced_cons'' : 
  forall {T U:Type} (la:list T) 
         (lp:list (T*U)) (a:T) (b:U), 
    synced la lp -> synced (a::la) ((a, b) :: lp).
intros T U la lp a b h1.
constructor.
destruct h1. simpl.
f_equal. assumption.
Qed.



Lemma remove_synced_cons :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U)
         (lp:list (T*U)) (pr:(T*U)),
    synced ((fst pr)::la) lp ->
    functionally_paired_l ((fst pr)::la) lb lp ->
    In pr lp -> NoDup lp ->
    remove (impl_type_eq_dec_prod pfdt pfdu) pr lp = tl lp.
intros T U hdt hdu la lb lp. revert la.
induction lp as [|pr' lp h1].
simpl. auto.
intros la pr h2 h2' h3 h4.
simpl.
destruct (impl_type_eq_dec_prod hdt hdu pr pr') as [h5 | h6].
subst.
pose proof (no_dup_cons_nin _ _ h4) as h5.
symmetry.
apply remove_not_in'.
assumption.
destruct h3 as [h7 | h8].
symmetry in h7.
contradiction.
pose proof (synced_hd pr _ _ h2) as h7.
simpl in h7.
pose proof (in_cons pr' pr lp h8) as h9.
pose proof (in_eq pr' lp) as h10.
rewrite (surjective_pairing pr) in h9.
rewrite (surjective_pairing pr') in h10 at 1.
rewrite h7 in h10.
pose proof (fpl_functional  h2' _ _ _ 
                            h9 h10) as h11.
assert (h12:pr = pr').
  rewrite (surjective_pairing pr).
  rewrite (surjective_pairing pr').
  rewrite h7, h11.
  reflexivity.
contradiction.
Qed.


Lemma fpl_cons_synced : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U)
         (lp:list (T*U)) (a:T),
    synced (a::la) lp ->
    NoDup (a::la) ->
    NoDup lp ->
    functionally_paired_l (a::la) lb lp ->
    functionally_paired_l la lb (tl lp).
intros T U hdt hdu la lb lp a h1 h2 h2' h3.
inversion h3 as [h4 h5].
specialize (h4 a (in_eq a la)).
destruct h4 as [b h6].
red in h6.
destruct h6 as [h6 h7].
destruct h6 as [h6a h6b].
pose proof (fpl_cons hdt hdu  _  _  _(a, b) h6b h2 h3) as h8.
rewrite (remove_synced_cons hdt hdu la lb) in h8.
assumption.
simpl. assumption.
simpl. assumption. assumption.
assumption.
Qed.


Lemma in_tl_list_power_synced :
  forall {T U:Type} (la:list T) (lb:list U) (lp:list (T*U))
         (a:T),
    In (tl lp) (list_power la lb) ->
    synced (a :: la) lp ->
    functionally_paired_l (a :: la) lb lp ->
    In lp (list_power (a::la) lb).
intros T U la lb lp a h1 h2 h2'.
revert h1 h2 h2'. revert a lb la.
induction lp as [|pr lp h3].
simpl.
intros a lb la h1 h2.
destruct h2 as [h2].
simpl in h2. discriminate.
simpl.
intros a lb la h1 h2 h4.
rewrite in_flat_map. exists lp.
split. assumption.
rewrite in_map_iff.
exists (snd pr).
destruct h2 as [h2].
simpl in h2.
pose proof (cons_inj _ _ _ _ h2) as h5.
destruct h5 as [h5l h5r].
rewrite <- h5l.
rewrite <- surjective_pairing.
split. reflexivity.
destruct h4 as [h4 h5].
specialize (h5 _ (in_eq pr lp)).
destruct h5; assumption.
Qed.

Lemma in_list_power_synced_tl:
  forall {T U:Type} (la:list T) (lb:list U) (lp:list (T*U))
         (a:T),
    synced (a :: la) lp ->
    functionally_paired_l (a :: la) lb lp ->
    In lp (list_power (a::la) lb) ->
    In (tl lp) (list_power la lb).
intros T U la lb lp a h1 h2 h2'.
revert h1 h2 h2'. revert a lb la.
destruct lp as [|pr lp].
simpl.
intros a lb la h1.
destruct h1 as [h1].
simpl in h1. discriminate.
simpl.
intros a lb la h1 h2 h4.
rewrite in_flat_map in h4.
destruct h4 as [lp' h4].
destruct h4 as [h4 h5].
rewrite in_map_iff in h5.
destruct h5 as [b h6].
destruct h6 as [h6l h6r].
inversion h6l.
subst.
assumption.
Qed.

Lemma in_tl_list_power_synced_iff :
  forall {T U:Type} (la:list T) (lb:list U) (lp:list (T*U))
         (a:T),
    synced (a :: la) lp ->
    functionally_paired_l (a :: la) lb lp ->
    (In (tl lp) (list_power la lb) <->
    In lp (list_power (a::la) lb)).
intros T U la lb lp a h1 h2.
split.
intro; apply in_tl_list_power_synced; assumption.
intro; apply in_list_power_synced_tl with a; assumption.
Qed.


Lemma synced_no_dup : 
  forall {T U:Type} (l:list T) (lp:list (T*U)),
    NoDup l -> synced l lp -> NoDup lp.
intros T U l lp h1 h2.
revert lp h2. 
induction h1 as [|a l h3 h4 h5]. 
intros lp h1.
pose proof (synced_nil1 _ h1). subst.
constructor.
intros lp. induction lp as [|pr lp h2].
intros; constructor.
intro h6.
pose proof (synced_cons' _ _ _ h6) as h7.
simpl in h7.
specialize (h5 _ h7).
constructor.
pose proof (synced_hd pr _ _  h6) as h8.
simpl in h8.
intro h9.
inversion h7 as [h10].
pose proof (in_map (fun pr0:T*U => fst pr0) lp pr h9) as h11.
rewrite h10 in h11.
rewrite h8 in h11.
contradiction.
assumption.
Qed.

Lemma in_list_power_synced : 
  forall {T U:Type} (la:list T) (lb:list U) (lp:list (T*U)),
    In lp (list_power la lb) ->
    synced la lp.
intros T U la.
induction la as [|a la h1].
intros  lb lp h1.
simpl in h1.
destruct h1; try contradiction. subst.
constructor.
simpl. reflexivity.
intros lb lp h2.
simpl in h2.
rewrite in_flat_map in h2.
destruct h2 as [lp' h2].
destruct h2 as [h2l h2r].
constructor.
rewrite in_map_iff in h2r.
destruct h2r as [b h2r].
destruct h2r as [h2a h2b].
specialize (h1 _ _ h2l).
destruct h1 as [h1].
rewrite <- h2a.
simpl.
rewrite h1.
reflexivity.
Qed.

Lemma synced_length : 
  forall {T U:Type} (la:list T) (lp:list (T*U)),
    synced la lp -> length la = length lp.
intros T U la lp h1.
destruct h1 as [h1].
subst.
apply map_length.
Qed.
 

Lemma fpl_in_list_power : 
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U) (lp:list (T*U)),
    synced la lp ->
    NoDup la ->  NoDup lp ->
    (functionally_paired_l la lb lp <->
     In lp (list_power la lb)).
intros T U hdt hdu  la lb lp h0 h1 h3.  split.
(* -> *)
revert h0 h1  h3. revert lb lp.
induction la as [|a la h1].
intros lb lp h0 h1  h3.
simpl. left. symmetry.
apply fpl_empty1_s with lb. assumption.
intros lb lp h0  h3 h4 h5.
inversion h5 as [h6 h7].
specialize (h6 a (in_eq a la)).
destruct h6 as [b h6].
red in h6.
destruct h6 as [h6a h6b]. 
destruct h6a as [h8 h9].
pose proof (fpl_cons_synced hdt hdu _ _ _ _ h0 h3 h4 h5) as h10.
pose proof (synced_cons' _ _ _ h0) as h11.
pose proof (no_dup_cons _ _ h3) as h12.
pose proof (no_dup_tl _ h4) as h13.
specialize (h1 _ _ h11 h12 h13 h10).
apply in_tl_list_power_synced; assumption.
(* <- *) 
intro h4.
revert lb lp h0 h1 h3 h4.
induction la as [|a la h1].
(* nil *)
simpl.
intros lb lp h1 h2 h3 h4.
destruct h4 as [h4 | h5].
subst.
apply fpl_empty1. contradiction.
(* cons *)
intros lb lp h2 h3 h4 h5.
simpl in h5.
rewrite in_flat_map in h5.
destruct h5 as [lp' h6].
destruct h6 as [h6l h6r].
rewrite in_map_iff in h6r.
destruct h6r as [b h7].
destruct h7 as [h7a h7b].
constructor.
intros x h8.
destruct h8 as [h8 | h9].
subst.
exists b. red.
split.
split; try constructor; auto.
intros b' h8.
destruct h8 as [h8l h8r].
pose proof (no_dup_cons_nin _ _ h3) as h9.
destruct h8r as [h10 | h11].
inversion h10. reflexivity.
pose proof (in_list_power1 _ h6l _ h11) as h12.
simpl in h12.
contradiction.
rewrite <- h7a in h2.
pose proof (synced_cons _ _ _ _ h2) as h10.
rewrite <- h7a in h4.
pose proof (no_dup_cons _ _ h3) as h11.
pose proof (no_dup_cons _ _ h4) as h12.
specialize (h1 _ _ h10 h11 h12 h6l).
destruct h1 as [h13 h14].
specialize (h13 _ h9).
destruct h13 as [y h13].
exists y.
red. red in h13.
destruct h13 as [h13a h13b].
split. split. destruct h13a as [h13aa h13ab].
assumption.
destruct h13a as [h13aa h13ab].
pose proof (in_cons (a, b) (x, y) lp' h13ab) as h15.
rewrite h7a in h15. assumption.
intros b' h15.
destruct h15 as [h15l h15r].
rewrite <- h7a in h15r.
destruct h15r as [h16 | h17].
inversion h16. subst.
pose proof (no_dup_cons_nin _ _ h3).
contradiction.
apply h13b.
split; assumption.
intros pr h8. split.
rewrite <- h7a in h8.
destruct h8 as [h8 | h9].
left. rewrite surjective_pairing in h8.
inversion h8. reflexivity.
pose  proof (in_list_power1  _ h6l _ h9).
right. assumption.
rewrite <- h7a in h8.
destruct h8 as [h8l | h8r]. rewrite surjective_pairing in h8l.
inversion h8l. subst. assumption.
pose proof (in_list_power2 _ h6l _ h8r). assumption.
Qed.


Lemma in_list_power_fpl : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (pfdu: type_eq_dec U)
         (la:list T) (lb:list U) (lp:list (T*U)),
    NoDup la ->  NoDup lp ->
    In lp (list_power la lb) ->
    functionally_paired_l la lb lp.
intros T U hdt hdu la lb lp h1 h2 h3.
rewrite fpl_in_list_power; auto.
apply in_list_power_synced with lb; assumption.
Qed.


Lemma in_list_of_lists_seqs_map_map_pair :
  forall {T U V:Type} (p:T->U->V) (li:list T) (lj:list U) (l:list V),
    In l (list_of_lists_seqs (map (fun i : T => map (p i) lj) li)) ->
    exists 
      lp:list (T*U), In lp (list_of_lists_seqs
                              (map (fun i:T => map (pair i) lj) li)) /\ l = map (fun pr:(T*U) => p (fst pr) (snd pr)) lp.
intros T U V p li lj.
induction li as [|i li h1]; induction lj as [|j lj h2].
simpl.
intros l h3. 
exists nil. split. left. reflexivity.
simpl. destruct h3 as [h3l | h3r]. subst. reflexivity.
contradiction.
simpl.
intros l h3.
exists nil. split.
left. reflexivity. simpl. destruct h3; auto. contradiction.
simpl. intros; contradiction.
intros l h5.
simpl in h5.
rewrite in_map_iff in h5.
destruct h5 as [vl h5].
destruct h5 as [h5l h5r].
destruct (in_app_or _ _ _ h5r) as [h6 | h7].
clear h5r.
rewrite in_map_iff in h6.
destruct h6 as [lv h6].
destruct h6 as [h6l h6r].
specialize (h1  _ h6r).
destruct h1 as [lp h1].
exists ((i, j) :: lp).
simpl.
rewrite in_map_iff.
split.
exists (i, j, lp).
simpl. split. reflexivity.
apply in_or_app.
left.
rewrite in_map_iff. exists lp.
split. reflexivity.
destruct h1 as [h1l h1r].
assumption.  
destruct h1 as [h1l h1r].
subst. simpl. reflexivity.
rewrite (surjective_pairing vl) in h7.
rewrite in_prod_iff in h7. 
destruct h7 as [h7l h7r].
specialize (h1  _ h7r).
destruct h1 as [lp h1]. 
rewrite in_map_iff in h7l.
destruct h7l as [u h8].
destruct h8 as [h8l h8r].
exists ((i, u)::lp).
simpl.
rewrite in_map_iff.
split.
exists ((i, u), lp).
simpl. split.
reflexivity.
apply in_or_app.
right.
rewrite in_prod_iff.
split.
rewrite in_map_iff.
exists u. split.
reflexivity.
assumption.
destruct h1 as [h1l h1r].
assumption.
rewrite <- h5l.
f_equal. rewrite h8l. reflexivity.
destruct h1 as [h1l h1r].
assumption.
Qed.

 

Lemma in_list_of_lists_seqs_map_map_pair_in_list_power :
  forall {T U V:Type} (p:T->U->V) (li:list T) (lj:list U) (lp:list (T*U)),
    In lp (list_of_lists_seqs (map (fun i:T => map (pair i) lj) li)) ->
    In lp (list_power li lj).
intros T U V p li lj.
induction li as [|i li h1]; induction lj as [|j lj h2].
simpl. auto.
simpl. auto.
simpl. intros; contradiction.
simpl. intros lp h3.
rewrite in_flat_map.
rewrite in_map_iff in h3.
destruct h3 as [tr h4].
destruct h4 as [h4l h4r].
destruct (in_app_or _ _ _ h4r) as [h5 | h6].
clear h4r.
rewrite in_map_iff in h5.
destruct h5 as [lp' h6].
destruct h6 as [h6l h6r].
specialize (h1 _ h6r).
exists lp'. split; auto.
simpl.
subst. simpl. left.
reflexivity.
subst.
rewrite (surjective_pairing tr) in h6.
pose proof (in_prod_iff).
rewrite in_prod_iff in h6.
destruct h6 as [h6l h6r].
specialize (h1 _ h6r).
exists (snd tr).
split; auto.
simpl. right.
rewrite in_map_iff.
rewrite in_map_iff in h6l.
destruct h6l as [u h6l].
destruct h6l as [h6a h6b].
exists u. rewrite h6a.
split; auto. 
Qed.

Lemma map_ext_in : forall {T U:Type} (f g:T->U) (l:list T),
                     (forall x:T, In x l -> f x = g x) ->
                          map f l = map g l.
intros T U f g l.
induction l as [|a l h1].
simpl. auto.
simpl. intro h2.
pose proof (h2 a) as h3.
rewrite h3.
f_equal. apply h1.
intros x h4.
apply h2. right. assumption.
left. reflexivity.
Qed.

Lemma fpl_impl_map : 
  forall {T U:Type} (li:list T) (lj:list U) (pfi:NoDup li)
         (lp:list (T*U))
         (pffp:functionally_paired_l li lj lp),
    synced li lp ->
    let f := fpl_to_fun_in_dep pffp in
    lp = map_in_dep (fun (i:T) (pfi:In i li) => (i, f i pfi)).
  intros T U li lj h1 lp h2 h5. simpl.
  destruct h5 as [h5].
  subst.
  rewrite map_in_dep_map.
  rewrite <- map_in_dep_id at 1.
  apply map_in_dep_ext.
  intros pr h3.
  apply functional_extensionality.
  intro h4.
  rewrite id_in_dep_compat.
  rewrite  (surjective_pairing pr) at 1.
  f_equal.
  symmetry.
  erewrite (fpl_to_fun_in_dep_in_eq _ _ _ _ _ h3). 
  f_equal. 
  apply proof_irrelevance.
Qed.


Lemma in_list_of_lists_seqs_map_map : 
  forall {T U V:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (p:T->U->V) (li:list T) (lj:list U) (pfi:NoDup li)
         (pfj:NoDup lj) (l:list V),
    In l (list_of_lists_seqs (map (fun i : T => map (p i) lj) li)) ->
    exists (lp:list (T*U)) (pfin:In lp (list_power li lj)),
      let pffp :=
          in_list_power_fpl
            pfdt pfdu li lj lp pfi
            (list_power_no_dup li lj lp pfi pfj pfin) pfin in
      let f := fpl_to_fun_in_dep pffp in
      l = map_in_dep (fun (i:T) (pfi:In i li) => p i (f i pfi)).
  intros T U V h1 h2 p li lj h3 h4 l h5.
  simpl.
  pose proof (in_list_of_lists_seqs_map_map_pair p li lj l h5) as h7.
  destruct h7 as [lp h7].
  destruct h7 as [h7l h7r].
  subst.
  pose proof (in_list_of_lists_seqs_map_map_pair_in_list_power p li lj lp h7l) as h8.
  exists lp, h8.
  pose proof (in_list_power_synced _ _ _ h8) as h9.
  pose proof (list_power_no_dup _ _ _  h3 h4 h8) as h10.
  pose proof (in_list_power_fpl h1 h2 _ _ _ h3 h10 h8) as h11.
  pose proof (fpl_impl_map  _ _ h3 _ h11 h9) as h12.
  simpl in h12.
  rewrite h12 at 1.
  rewrite map_map_in_dep.
  apply map_in_dep_ext.
  intros; apply functional_extensionality; intro.
  f_equal; simpl; f_equal; apply proof_irrelevance.
Qed.

Lemma in_list_of_lists_seqs_map_map_pair_rev :
  forall {T U V:Type} (p:T->U->V) (li:list T) (lj:list U) (lp:list (T*U)),
    In lp (list_of_lists_seqs
             (map (fun i:T => map (pair i) lj) li)) ->
    exists l:list V,
    In l (list_of_lists_seqs (map (fun i : T => map (p i) lj) li)) /\ l = map (fun pr:(T*U) => p (fst pr) (snd pr)) lp.
intros T U V p li lj.
induction li as [|i li h1]; induction lj as [|j lj h2].
simpl.
intros lp h1.
destruct h1 as [h1l | h1r].
subst.
simpl.
exists nil. split; auto. contradiction.
simpl.
intros lp h3.
destruct h3 as [h3 | h4].
subst. simpl. exists nil. split; auto. contradiction.
simpl. intros; contradiction.
intros lp h3.
simpl. simpl in h3.
rewrite in_map_iff in h3.
destruct h3 as [tr h3].
destruct h3 as [h3l h3r]. 

destruct (in_app_or _ _ _  h3r) as [h4 | h5].
clear h3r.
rewrite in_map_iff in h4.
destruct h4 as [lp' h4].
destruct h4 as [h4l h4r].
rewrite surjective_pairing in h4l.
inversion h4l.
specialize (h1 _ h4r).
destruct h1 as [l h1].
exists ((p i j) :: l).
split.
rewrite in_map_iff.
exists (p i j, l).
simpl.
split; auto.
apply (in_or_app).
left.
rewrite in_map_iff.
exists l.
split; auto.
destruct h1 as [h1l h1r].
simpl in h1l.
assumption. 
rewrite <- h3l.
simpl. f_equal. 
f_equal. 
inversion H0. auto. inversion H0. auto.
destruct h1 as [h1l h1r].
rewrite H1 in h1r.
assumption.
rewrite (surjective_pairing tr) in h5.
rewrite in_prod_iff in h5.
destruct h5 as [h5l h5r].
rewrite in_map_iff in h5l.
destruct h5l as [u h5l].
destruct h5l as [h5a h5b].
specialize (h1 _ h5r).
destruct h1 as [l h1].
destruct h1 as [h1l h1r].
exists ((p i u) :: l).
split.
rewrite in_map_iff.
exists (p i u, l).
simpl.
split; auto.
apply in_or_app.
right.
rewrite in_prod_iff.
split. 
rewrite in_map_iff.
exists u. split; auto.
assumption.
rewrite <- h3l.
simpl.
f_equal.
f_equal.
inversion h5a. simpl. auto.
inversion h5a. simpl. auto.
assumption.
Qed.


Lemma in_list_power_in_list_of_lists_seqs_map_map_pair :
  forall {T U:Type} (li:list T) (lj:list U) (lp:list (T*U)),
    In lp (list_power li lj) ->
    In lp (list_of_lists_seqs (map (fun i:T => map (pair i) lj) li)).
intros T U li lj.
induction li as [|i li h1]; induction lj as [|j lj h2].
simpl. auto.
simpl. auto.
simpl. simpl in h1.
intros lp h2.
rewrite in_flat_map in h2.
destruct h2 as [? h2].
destruct h2; contradiction.
simpl.
intros lp h3.
rewrite in_map_iff.
rewrite in_flat_map in h3.
destruct h3 as [lp' h3].
destruct h3 as [h3l h3r].
simpl in h3r.
destruct h3r as [h3a | h3b].
exists (i, j, lp'). simpl.
split; auto.
apply in_or_app.
left.
rewrite in_map_iff.
exists lp'. split; auto.
rewrite in_map_iff in h3b.
destruct h3b as [u h3b].
destruct h3b as [h4 h5].
exists (i, u, lp'). simpl. split; auto.
apply in_or_app.
right.
rewrite in_prod_iff.
split.
rewrite in_map_iff. exists u.
split; auto.
auto.
Qed.

Lemma in_map_list_of_lists_seqs_map_in_dep :
  forall {T U V:Type} (pft:type_eq_dec T) (pfu:type_eq_dec U)
         (p:T->U->V) (li:list T) (lj:list U)
         (ls:list (T*U)) 
         (pfi:NoDup li) (pfj:NoDup lj)
         (pfin:In ls (list_power li lj)),
    let pffp := in_list_power_fpl pft pfu li lj ls pfi
                    (list_power_no_dup li lj ls pfi pfj
                                       pfin) pfin in
    let f := fpl_to_fun_in_dep pffp in
    exists l: list V,
      In l (list_of_lists_seqs (map (fun i : T => map (p i) lj) li)) /\
      l =  map_in_dep
             (fun (i : T) (pfi:In i li) =>
                p i (f i pfi)).
  intros T U V hdt hdu p li lj ls h1 h2 h3. simpl.
  pose proof (in_list_power_in_list_of_lists_seqs_map_map_pair _ _ _ h3) as h4.
  pose proof (in_list_of_lists_seqs_map_map_pair_rev p _ _ _  h4) as h5.
  destruct h5 as [l h5].
  destruct h5 as [h5l h5r].
  exists l.
  subst.
  split; auto.
  pose proof (in_list_power_synced _ _ _ h3) as h6.
  pose proof (list_power_no_dup _ _ _  h1 h2 h3) as h7.
  pose proof (in_list_power_fpl hdt hdu _ _ _  h1 h7 h3 ) as h8.
  pose proof (fpl_impl_map _ _ h1 _ h8  h6) as h9.
  rewrite h9 at 1.
  rewrite map_map_in_dep.
  apply map_in_dep_ext.
  simpl.
  intros x h10.
  apply functional_extensionality. intro.
  f_equal. f_equal.
  apply proof_irrelevance.
Qed.
  


Lemma in_list_of_lists_seqs_map_map_pair_in_list_power' :
  forall {T U:Type}  (li:list T) (lj:list U) (lp:list (U*T)),
    In lp (list_of_lists_seqs
          (map (fun j : U => map (fun i : T => (j, i)) li) lj)) ->
    In lp (list_power lj li).
intros T U li lj. 
induction lj as [|j lj h1]; simpl; auto.
intros lp h2.
rewrite in_map_iff in h2.
destruct h2 as [tr h2].
destruct h2 as [h2a h2b].
rewrite (surjective_pairing tr) in h2b.
rewrite in_prod_iff in h2b.
destruct h2b as [h2l h2r].
specialize (h1 _ h2r).
rewrite in_flat_map.
exists (snd tr). split; auto.
rewrite in_map_iff.
rewrite in_map_iff in h2l.
destruct h2l as [x h2l]. destruct h2l as [h3 h4].
exists x. split. rewrite <- h2a.
f_equal. assumption.
assumption.
Qed.

Lemma in_list_power_in_list_of_lists_seqs_map_map_pair' :
  forall {T U:Type}  (li:list T) (lj:list U) (lp:list (U*T)),
    In lp (list_power lj li) ->
    In lp (list_of_lists_seqs
          (map (fun j : U => map (fun i : T => (j, i)) li) lj)).
intros T U li lj.
induction lj as [|j lj h1]; simpl; auto.
intros lp h2.
rewrite in_flat_map in h2.
destruct h2 as [lp' h2].
destruct h2 as [h2l h2r].
rewrite in_map_iff.
specialize (h1 _ h2l).
rewrite in_map_iff in h2r.
destruct h2r as [i h2r].
destruct h2r as [h2a h2b].
exists (j,i, lp').
simpl. split. assumption.
rewrite in_prod_iff.
split.
rewrite in_map_iff.
exists i. split; auto.
assumption.
Qed.


Lemma in_list_of_lists_seqs_map_map_pair' : 
  forall {T U V:Type} (p:T->U->V) (li:list T) (lj:list U)
         (l:list V),
    In l
       (list_of_lists_seqs
          (map (fun j : U => map (fun i : T => p i j) li) lj)) ->
    exists lp:list (U*T),
       In lp (list_of_lists_seqs
          (map (fun j : U => map (fun i : T => (j, i)) li) lj))
       /\ l = map (fun pr:(U*T) => p (snd pr) (fst pr)) lp.
intros T U V p li lj. revert li.
induction lj as [|j lj h1].
simpl.
intros li l h2.
exists nil. simpl. split. left; auto. destruct h2; [subst; auto | contradiction]. 
intros li l h5.
simpl in h5.
rewrite in_map_iff in h5.
destruct h5 as [pr  h5].
destruct h5 as [h5l h5r].
rewrite (surjective_pairing pr) in h5r.
rewrite in_prod_iff in h5r.
destruct h5r as [h5a h5b].
specialize (h1 _ _  h5b).
destruct h1 as [lp h1].
destruct h1 as [h1a h1b].
pose proof (in_list_of_lists_seqs_map_map_pair_in_list_power' _ _  _ h1a) as h7.
rewrite in_map_iff in h5a.
destruct h5a as [i h5a]. destruct h5a as [h9 h10].
pose proof (in_list_power_add _ _ _ j i h10 h7) as h11.
pose proof (in_list_power_in_list_of_lists_seqs_map_map_pair' _ _ _  h11) as h12.
exists ((j, i)::lp).
split; auto.
simpl. subst.
f_equal. rewrite h9. reflexivity.
assumption.
Qed.

Lemma in_list_of_lists_seqs_map_map' : 
  forall {T U V:Type} (hdt:type_eq_dec T) (hdu:type_eq_dec U)
         (p:T->U->V) (li:list T) (lj:list U)
         (pfi:NoDup li) (pfj:NoDup lj) (l:list V) (def:T),
    In l
       (list_of_lists_seqs
          (map (fun j : U => map (fun i : T => p i j) li) lj)) ->
    exists (lp:list (U*T)) (pfin:In lp (list_power lj li)),
      let pffp :=
          in_list_power_fpl
            hdu hdt lj li lp pfj
            (list_power_no_dup lj li lp pfj pfi pfin) pfin in
      let f := fpl_to_fun_in_dep pffp in
            l =  map_in_dep
                   (fun (j:U) (pfinj:In j lj) =>
                      p (f j pfinj) j).
  intros T U V hdt hdu p li lj h1 h2 l def h3.
  pose proof (in_list_of_lists_seqs_map_map_pair _ _ _ _ h3) as h4.
  destruct h4 as [lp h4].
  destruct h4 as [h4l h4r].
  pose proof (in_list_of_lists_seqs_map_map_pair_in_list_power' li lj lp h4l) as h5.
  exists lp, h5.
  pose proof (in_list_power_synced _ _ _ h5) as h6.
  pose proof (list_power_no_dup _ _ _  h2 h1 h5) as h7.
  pose proof (in_list_power_fpl hdu hdt _ _ _ h2 h7 h5) as h8.
  pose proof (fpl_impl_map _ _ h2 _ h8 h6) as h9.
  rewrite h4r.
  simpl. simpl in h9.
  rewrite h9 at 1.
  rewrite map_map_in_dep.
  apply map_in_dep_ext.
  intros. apply functional_extensionality. intro.
  f_equal; f_equal; simpl.
  f_equal; apply proof_irrelevance.
Qed.


Lemma in_list_of_lists_seqs_map_map_pair_rev' :
  forall {T U V:Type} (p:T->U->V) (li:list T) (lj:list U) (lp:list (U*T)),
    In lp (list_of_lists_seqs
             (map (fun j : U => map (fun i : T => (j, i)) li) lj)) ->
    exists l:list V,
      In l (list_of_lists_seqs
          (map (fun j : U => map (fun i : T => p i j) li) lj)) /\
         l = map (fun pr:(U*T) => p (snd pr) (fst pr)) lp.
intros T U V p li lj.
induction lj as [|j lj h1]; destruct li as [|i li]; simpl.
intros lp h1. exists nil. split. left. reflexivity.
destruct h1 as [h1l | h1r]. subst. simpl. reflexivity.
contradiction.
intros lp h1. exists nil. split. left. reflexivity.
destruct h1 as [h1l | h1r]. subst. simpl. reflexivity.
contradiction.
intros; contradiction.
intros lp h2.
rewrite in_map_iff in h2.
destruct h2 as [tr [h2 h3]].
destruct (in_app_or _ _ _ h3) as [h4 | h5]. clear h3.
rewrite in_map_iff in h4. 
destruct h4 as [lp' [h4 h5]].
specialize (h1 _ h5).
destruct h1 as [l h1].
destruct h1 as [h1l h1r].
exists ((p i j)::l).
rewrite in_map_iff.
split.
exists ((p i j), l).
simpl.
split. reflexivity.
apply in_or_app.
left.
rewrite in_map_iff.
exists l. split; auto.  rewrite <- h2.
simpl. rewrite <- h4.
simpl. f_equal.
assumption.
rewrite (surjective_pairing tr) in h5. clear h3.
rewrite in_prod_iff in h5.
destruct h5 as [h5l h5r].
specialize (h1 _ h5r).
destruct h1 as [l [h1l h1r]]. 
rewrite in_map_iff in h5l.
destruct h5l as [a h5l]. destruct h5l as [h5a h5b].
exists ((p a j)::l). split.
rewrite in_map_iff.
exists ((p a j), l). split.
simpl. reflexivity.
apply in_or_app. right.
rewrite in_prod_iff. split.
rewrite in_map_iff. exists a. split; auto.
assumption. 
rewrite <- h2.
simpl. rewrite <- h5a. simpl. f_equal.
assumption.
Qed.


Lemma map_in_dep_fpl_to_fun_in_dep :
  forall {T U:Type} {la:list T} {lb:list U}
         {lp:list (T*U)}
         (pf:functionally_paired_l la lb lp),
    synced la lp ->
    let f := fpl_to_fun_in_dep pf in
    map_in_dep f = map (snd (B:=U)) lp.
  intros T U la lb lp h1 h2. simpl.
  destruct h2 as [h2].
  subst.
  rewrite map_in_dep_map.
  rewrite map_eq_map_in_dep.
  apply map_in_dep_ext.
  intros pl h3.
  apply functional_extensionality.
  intro.
  rewrite (fpl_to_fun_in_dep_in_eq _ _ _ h1 _ h3) at 1; auto.
  f_equal. apply proof_irrelevance.
Qed.



Lemma map_in_dep_fpl_to_fun_cons :
  forall {T U:Type} {la:list T} {lb:list U} {lp:list (T*U)} {a:T}
         (pf:functionally_paired_l la lb (tl lp))
         (pf':functionally_paired_l (a::la) lb lp),
    NoDup lp ->
    let f := fpl_to_fun_in_dep pf in
    let f' := fpl_to_fun_in_dep pf' in
    map_in_dep f = map_in_dep (fun_from_cons f').
  intros T U la lb lp a h1 h2 h3.
  simpl.
  apply map_in_dep_ext.
  intros x h4.
  apply functional_extensionality.
  intro h5.
  rewrite fun_from_cons_compat; auto.
  pose proof (fpl_to_fun_in_dep_compat h1 x h5) as h6.
  pose proof (fpl_to_fun_in_dep_compat h2 x (in_cons a x la h5)) as h7.
  pose proof (in_tl _ _ h6) as h9.
  apply (fpl_functional h2 _ _ _ h9 h7).
Qed.


Lemma in_map_in_dep_list_of_lists_seqs_map :
  forall {T U V:Type} (hdt:type_eq_dec T) (hdu:type_eq_dec U)
         (p:T->U->V) (li:list T) (lj:list U)
         (ls: list (U*T))
         (pfi:NoDup li) (pfj:NoDup lj)
         (pfin:In ls (list_power lj li)),
    let pffp := in_list_power_fpl hdu hdt lj li ls pfj
                                  (list_power_no_dup lj li ls pfj pfi
                                                     pfin) pfin in
    let f := fpl_to_fun_in_dep pffp in
    exists l:list V,
      In l (list_of_lists_seqs
          (map (fun j : U => map (fun i : T => p i j) li) lj)) /\
      l = (map_in_dep (fun (j : U) (pfinj:In j lj) =>
                         p (f j pfinj) j)).
  intros T U V hdt hdu p li lj ls h1 h2 h3.
  pose proof (in_list_power_in_list_of_lists_seqs_map_map_pair' _ _ _ h3) as h4.
  pose proof (in_list_of_lists_seqs_map_map_pair_rev' p _ _ _  h4) as h5.
  destruct h5 as [l h5].
  destruct h5 as [h5l h5r].
  exists l.
  subst.
  split; auto.
  pose proof (in_list_power_synced _ _ _ h3) as h6.
  pose proof (list_power_no_dup _ _ _  h2 h1 h3) as h7.
  pose proof (in_list_power_fpl hdu hdt _ _ _ h2 h7 h3) as h8.
  pose proof (fpl_impl_map _ _ h2 _ h8 h6) as h9.
  rewrite h9 at 1.
  rewrite map_map_in_dep.
  apply map_in_dep_ext.
  intros. apply functional_extensionality. intro.
  f_equal. simpl. f_equal. apply proof_irrelevance.
Qed.
  

Lemma in_map_in_dep_list_of_lists_seqs_map' :
  forall {T U V:Type} (pft:type_eq_dec T) (pfu:type_eq_dec U)
         (p:T->U->V) (li:list T) (lj:list U)
         (ls:list (U*T))
         (pfi:NoDup li) (pfj:NoDup lj)
         (pfin:In ls (list_power lj li)),
    let pffp := in_list_power_fpl pfu pft lj li ls pfj
                    (list_power_no_dup lj li ls pfj pfi
                                       pfin) pfin in
    let f := fpl_to_fun_in_dep pffp in
    exists l: list V,
      In l (list_of_lists_seqs (map (fun j : U => map (fun (i:T) => p i j) li) lj)) /\
      l =  map_in_dep
             (fun (j : U) (pfj:In j lj) =>
                p (f j pfj) j).
  intros T U V hdt hdu p li lj ls h1 h2 h3. simpl. 
  pose proof (in_list_power_in_list_of_lists_seqs_map_map_pair _ _ _ h3) as h4.
  pose proof (in_list_power_in_list_of_lists_seqs_map_map_pair' _ _ _ h3) as h4'.
  pose proof (in_list_of_lists_seqs_map_map_pair_rev' p _ _ _  h4') as h5.
  destruct h5 as [l h5].
  destruct h5 as [h5l h5r]. 
  exists l.
  subst.  
  split; auto.
  pose proof (in_list_power_synced _ _ _ h3) as h6.
  pose proof (list_power_no_dup _ _ _  h2 h1 h3) as h7.
  pose proof (in_list_power_fpl hdu hdt _ _ _  h2 h7 h3 ) as h8.
  pose proof (fpl_impl_map _ _ h2 _ h8  h6) as h9. simpl in h9.
  rewrite h9 at 1.
  rewrite map_map_in_dep.
  apply map_in_dep_ext.
  intros y h10.
  apply functional_extensionality.
  intro h11. simpl.
  f_equal. f_equal. apply proof_irrelevance.
Qed.



Lemma sync_fpl : forall {T U:Type} {la:list T} {lb:list U}
                        {lp:list (T*U)},
                   functionally_paired_l la lb lp ->
                   exists lp':list (T*U),
                     functionally_paired_l la lb lp' /\
                     synced la lp' /\
                     list_to_set lp = list_to_set lp'.
intros T U la lb lp h1.
inversion h1 as [h2 h3].
pose (fun (x:{t:T | In t la}) => 
        proj1_sig (constructive_definite_description 
                     _ (h2 (proj1_sig x) (proj2_sig x)))) as f.
pose (fun (x:{t:T | In t la}) => (proj1_sig x, f x)) as g.
pose (map g (map_sig la)) as lp'.
exists lp'.
repeat split.
intros x h4.
exists (f (exist _ x h4)).
red.
repeat split.
unfold f.
destruct constructive_definite_description as [b h6].
simpl in h6. simpl. destruct h6 as [h6l h6r].
assumption.
unfold lp'.
rewrite in_map_iff. unfold g.
exists (exist _ x h4). simpl. split.
reflexivity.
apply map_sig_in.
intros b h5.
unfold f.
destruct constructive_definite_description as [b' h6].
simpl in h6. simpl.
unfold lp' in h5.
destruct h5 as [h5l h5r].
rewrite in_map_iff in h5r.
destruct h5r as [x' h7].
destruct h7 as [h7l h7r].
unfold g in h7l.
inversion h7l as [h8].
unfold f in H.
unfold f.
destruct constructive_definite_description as [a h9].
simpl. simpl in H. subst.
destruct h9 as [h9l h9r]. destruct h6 as [h6l h6r].
apply (fpl_functional h1 _ _ _ h6r h9r).
unfold lp' in H.
unfold g in H.
rewrite in_map_iff in H.
destruct H as [x h4].
destruct h4 as [h4l h4r].
rewrite surjective_pairing in h4l.
inversion h4l.
apply proj2_sig.
unfold lp' in H. unfold g in H.
rewrite in_map_iff in H.
destruct H as [x h4].
destruct h4 as [h4l h4r].
rewrite surjective_pairing in h4l.
inversion h4l.
unfold f.
destruct constructive_definite_description as [x' h6].
simpl. destruct h6; assumption.
unfold lp'.
rewrite map_map. simpl.
apply map_sig_compat.
apply Extensionality_Ensembles.
red. split.
red.
(* <= *)
intros pr h4.
rewrite <- list_to_set_in_iff.
rewrite <- list_to_set_in_iff in h4.
unfold lp'.
unfold g.
rewrite in_map_iff.
rewrite (surjective_pairing pr) in h4.
pose proof (fpl_in_dom  h1 _ h4) as h5.
simpl in h5.
exists (exist _ (fst pr) h5). 
simpl. split.
rewrite surjective_pairing.
unfold f.
destruct constructive_definite_description as [x h6]. simpl. simpl in h6.
destruct h6 as [h6l h6r].
pose proof (fpl_functional  h1 _ _ _ h4 h6r). subst.
reflexivity.
apply map_sig_in.
red.
intros pr h4.
rewrite <- list_to_set_in_iff.
rewrite <- list_to_set_in_iff in h4.
unfold lp' in h4.
unfold g in h4. rewrite in_map_iff in h4.
destruct h4 as [x h4].
destruct h4 as [h4l h4r].
rewrite <- h4l.
unfold f.
destruct constructive_definite_description as [b h5].
simpl.
destruct h5; assumption.
Qed.


Section NprodList.

  Lemma in_nprod_iff' :
    forall {T n} y (x:T^n),
      in_nprod y x <->
      In y (nprod_to_list _ _ x).
    intros T n; induction n as [|n ih]; simpl.
    tauto.
    intros y x.
    destruct x as [f s]; simpl.
    split; intro h1; destruct h1 as [|h1]; auto; right.
    rewrite ih in h1; auto.
    rewrite ih; auto.
  Qed.

    
  Lemma nprod_to_list_nprod_of_list_eq : 
    forall {T:Type} (l:list T), 
      nprod_to_list _ (length l) (nprod_of_list _ l) = l.
    intros T l.
    induction l as [|a l h1]; simpl.
    reflexivity.
    f_equal; auto.
  Qed.

  
  Lemma nprod_of_list_nprod_to_list_eq : 
    forall {T n} (x:T^n), 
      nprod_of_list _ (nprod_to_list _ _ x) =
      transfer_nprod_r (length_nprod_to_list _) x.
    intros T n.
    induction n as [|n ih]; simpl.
    intro; apply unit_eq.
    intro x; destruct x as [t x]; simpl.
    rewrite ih, transfer_nprod_r_S; repeat f_equal;
    apply proof_irrelevance.
  Qed.
  

  Lemma all_sing_nprod_to_list_iff : 
    forall {T n} (x:T^n) a,
      all_sing (nprod_to_list _ _ x) a <->
      all_sing_nprod x a.
    intros T n x; split; intro h1;
  red in h1; red;
  intros b h2;
  [rewrite in_nprod_iff' in h2 |
   rewrite <- in_nprod_iff' in h2]; auto.
  Qed.
  
  Lemma all_sing_nprod_of_list_iff : 
    forall {T} (l:list T) a,
      all_sing_nprod (nprod_of_list _ l) a <->
      all_sing l a.
    intros T n x; split; intro h1;
    red in h1; red;
    intros b h2.
    specialize (h1 b).
    hfirst h1.
    rewrite in_nprod_iff'.
    rewrite nprod_to_list_nprod_of_list_eq; auto.
    auto.
    rewrite in_nprod_iff' in h2.
    rewrite nprod_to_list_nprod_of_list_eq in h2; auto.
  Qed.
     
  


  
  Lemma app_list_nprod_to_list_eq_sing_iff :
  forall {T n} (x:(list T)^n) (a:T),
    (forall c, in_nprod c x -> c <> nil) ->
    (app_list (nprod_to_list _ _ x) = a::nil <->
     exists pfe:n = 1,
       hd_nprod (lt_eq_trans _ _ _ (lt_0_Sn 0) (eq_sym pfe)) x = a::nil).
  intros T n; induction n as [|n ih];  
  intros x a h1; split; intro h2.
  simpl in h2.
  discriminate.
  destruct h2; discriminate.
  destruct x as [c x]; simpl; simpl in h2, h1.
  
  destruct c as [|c]; simpl.
  simpl in h2.
  destruct n as [|n]; simpl in h2.
  discriminate.
  destruct x as [c x]; simpl in h2.
  rewrite app_eq_sing_iff in h2.
  destruct h2 as [h2|h2].
  destruct h2 as [? h2]; subst.
  simpl in h1;
  specialize (h1 nil). 
  hfirst h1; tauto.
  destruct h2 as [h2]; subst.
  specialize (h1 nil). 
  hfirst h1; tauto.
  simpl in h2.  
  inversion h2; subst.
  inversion h2. 
  apply app_eq_nil in H0. 
  destruct H0; subst.
  rewrite app_list_eq_nil_iff in H0.
  destruct H0 as [h3|h3].
  rewrite (nprod_to_list_eq_nil_iff x) in h3; subst.
  exists (eq_refl _).
  f_equal; auto.
  simpl in H1. 
  rewrite app_list_eq_nil_iff in H1.
  destruct H1 as [h4|h4].
  rewrite (nprod_to_list_eq_nil_iff x) in h4; subst.
  exists (eq_refl _).
  f_equal; auto.
  inversion h2; clear h2.
  rewrite app_list_eq_nil_iff in H0.
  destruct H0 as [h5|h5].
  rewrite (nprod_to_list_eq_nil_iff x) in h5; subst.
  exists (eq_refl _). 
  f_equal; auto.
  red in h3. 
  specialize (ih x a).
  destruct (zerop n) as [|h6]; subst.
  exists (eq_refl _),.
  specialize (h1 (hd_nprod h6 x)).
  hfirst h1.
  right.
  rewrite hd_nprod_eq.
  apply in_nth_prod.
  apply h1 in hf.
  red in h4.
  specialize (h4 (hd_nprod h6 x)).
  hfirst h4. 
  rewrite hd_nprod_eq.
  rewrite <- in_nprod_iff'.
  apply in_nth_prod.
  specialize (h4 hf0).
  contradiction.
  destruct h2 as [h2 h3].
  pose proof (S_inj _ _ h2); subst.
  destruct x as [? x]; simpl; simpl in h3.
  rewrite app_nil_r; auto.
  Qed.

  
Fixpoint map_tuple_prod_to
         {T U n} {A:T->Type} (f:forall t:T, A t -> U) :
  forall (x:T^n) (y:tuple_prod A x), list U :=
  match n with
    | 0 => fun _ _ => nil
    | S n' => fun (x:T^S n') (y:tuple_prod A x) =>
                f (fst x) (fst y) :: map_tuple_prod_to f (snd x) (snd y)
  end.

Lemma in_map_tuple_prod_to :
  forall {T U n} {A:T->Type} (f:forall t:T, A t -> U)
         (x:T^n) (y:tuple_prod A x)
         i (pfi:i < n),
    In (f (nth_prod i pfi x) (nth_tuple_prod y i pfi))
       (map_tuple_prod_to f x y).
  intros T U n; induction n as [|n ih]; simpl.
  intros; omega.
  intros A f x y i hi.
  destruct x as [xf xs]; simpl in y.
  destruct y as [yf ys]; simpl.
  destruct i as [|i]; simpl.
  left; auto.
  specialize (ih A f xs ys i (lt_S_n _ _ hi)).
  right; auto.
Qed.

Lemma in_map_tuple_prod_to_iff :
  forall {T U n} {A:T->Type} (f:forall t:T, A t -> U)
         (x:T^n) (y:tuple_prod A x)
         (u:U),
    In u (map_tuple_prod_to f x y) <->
    exists i (pfi:i < n),
      u = f (nth_prod i pfi x) (nth_tuple_prod y i pfi).
  intros T U n.
  induction n as [|n ih]; simpl.
  intros; split; intro h1.
  contradiction.
  destruct h1 as [? [? ?]]; omega.
  intros A f x y u.
  destruct x as [xf xs]; simpl in y.
  destruct y as [yf ys]; simpl.
  split; intro h1. 
  destruct h1 as [|h1]; subst.
  exists 0, (lt_0_Sn _); simpl; auto.
  rewrite ih in h1.
  destruct h1 as [i [hi ?]]; subst.
  exists (S i), (lt_n_S _ _ hi); simpl.
  gtermr.
  redterm0 y.
  redterm1 y.
  redterm0 c.
  fold c1.
  pose proof (proof_irrelevance _ c1 hi) as h2.
  rewrite h2; auto.
  destruct h1 as [i [hi ?]]; subst.
  destruct i as [|i]; simpl.
  left; auto.
  right.
  specialize (ih A f xs ys).
  gterm1.
  fold c.
  specialize (ih c).
  rewrite ih.
  unfold c.
  exists i, (lt_S_n _ _ hi); auto.
Qed.


Lemma nth_prod_to_list :
  forall {T:Type} {i n:nat}
         (pflt:i < n) (x:T^n),
    let pflt' := lt_eq_trans _ _ _ pflt (eq_sym (length_nprod_to_list _)) in
    nth_prod i pflt x =
    nth_lt (nprod_to_list _ _ x) i pflt'.
  intros T i n. revert i.
  induction n as [|n ih]; simpl.
  intros; omega.
  intros i hlt x.  
  destruct x as [x y], i as [|i]; simpl; auto.
  rewrite ih.
  erewrite nth_lt_compat, nth_lt_functional3.
  apply nth_indep.
  rewrite length_nprod_to_list; auto with arith.

  Grab Existential Variables.
  rewrite length_nprod_to_list; auto with arith.
Qed.

Lemma nth_prod_of_list :
  forall {T:Type} (l:list T)
         (i:nat)
         (pflt:i < length l),
    nth_prod i pflt (nprod_of_list _ l)  = 
    nth_lt l i pflt.
  intros T l.
  induction l as [|a l ih].
  simpl; intros; omega.
  intros i hlt.
  destruct i as [|i].
  rewrite nth_lt_0'.
  simpl; auto.
  simpl in hlt.
  pose proof hlt as h1.
  apply lt_S_n in h1.
  rewrite nth_lt_S; simpl.
  rewrite ih.
  apply nth_lt_functional3.
Qed.

 
Lemma nprod_inj : 
  forall {T:Type} {n:nat} (x1 x2:T^n) (def:T),
    x1 = x2 -> forall m:nat, m < n ->
    nth m (nprod_to_list _ _ x1) def = 
    nth m (nprod_to_list _ _ x2) def.
intros; subst; auto.
Qed.

Fixpoint in_n {T:Type} {n:nat} (l:list T) : T^n -> Prop :=
  match n return (T^n-> Prop) with
    | O => (fun x => True)
    | S n' => (fun x => In (fst x) l /\ (in_n l (snd x)))
  end.

Lemma in_n_in_nprod_to_list_compat : 
  forall {T:Type} {n:nat} (l:list T) (t:T^n),
    in_n l t -> (forall x:T, In x (nprod_to_list _ _ t) -> In x l).
intros T n.
induction n as [|n h1]; simpl.
intros l tt. 
tauto.
intros l t.
intros h2 ft' h3.
destruct t as [ft st].
destruct h3 as [h3l | h3r].
subst.
simpl in h2. 
destruct h2; auto.
simpl in h2.
destruct h2 as [h2l h2r].
specialize (h1 _ _  h2r _ h3r).
assumption.
Qed.

Lemma in_nprod_to_list_in_n_compat : 
  forall {T:Type} {n:nat} (l:list T) (t:T^n),
    (forall x:T, In x (nprod_to_list _ _ t) -> In x l) -> in_n l t.
intros T n.
induction n as [|n h1]; simpl.
intros. tauto.  
intros l t h2.
destruct t as [ft st]. simpl.
assert (h3:forall x:T, In x (nprod_to_list T n st) -> In x l).
  intros x h3.
  pose proof (in_cons ft x _ h3) as h4.
apply h2; auto.
specialize (h1 _ _ h3). 
specialize (h2 ft (in_eq _ _)).
split; auto.
Qed.

Lemma in_n_in_nprod_to_list_compat_iff : 
  forall {T:Type} {n:nat} (l:list T) (t:T^n),
    in_n l t <-> (forall x:T, In x (nprod_to_list _ _ t) -> In x l).
intros; split; [apply in_n_in_nprod_to_list_compat | apply in_nprod_to_list_in_n_compat].
Qed.


Lemma in_n_cons : 
  forall {T:Type} {n:nat} (l:list T) (a:T) (t:T^n),
    in_n l t -> in_n (a::l) t.
intros T n l a t h1.
rewrite in_n_in_nprod_to_list_compat_iff in h1.
rewrite in_n_in_nprod_to_list_compat_iff.
intros x h2.
specialize (h1 _ h2).
right.
assumption.
Qed.


Lemma in_n_nprod_to_list : 
  forall {T:Type} {n:nat} (t:T^n), in_n (nprod_to_list _ _ t) t.
intros T n.
induction n as [|n h1]; simpl. auto.
intro t. split.
destruct t; simpl. left. auto.
destruct t; simpl. apply in_n_cons. apply h1.
Qed.


Lemma nprod_to_fpl_ex : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U),
    NoDup la ->
    let n := length la in
    forall (t:U^n), 
      in_n lb t ->
      exists! lp:list (T*U), NoDup lp /\ 
        synced la lp /\
        exists (pf:functionally_paired_l la lb lp),
          let f := fpl_to_fun_in_dep pf in
          map_in_dep f = nprod_to_list _ _ t.
intros T U hdt hdu la.
induction la as [|a la h1]; simpl.
intros.
exists (list_prod nil lb). 
red. simpl.
split. split. constructor. split; try constructor. simpl. reflexivity.
constructor. intros; contradiction. intros; contradiction.
reflexivity. 
intros pl h1.
destruct h1 as [h1l [h1c h1r]].
destruct h1r as [h1r].
symmetry.
apply (fpl_empty1_s _ _ h1r).
intros lb h0 t h2.
destruct h2 as [h2l h2r].
inversion h0 as [|? ? h0l h0r].  subst.
specialize (h1 lb h0r _ h2r).
destruct h1 as [lp h1].
red in h1. destruct h1 as [h1l h1r].
destruct h1l as [h3 h4].
destruct h4 as [h4 h5]. 
exists ((a, (fst t))::lp).
red.
split. split.
constructor.
intro hi. destruct h5 as [h5l h5r].
pose proof (fpl_in_dom h5l _ hi) as h0'.
simpl in h0'. contradiction. assumption. 
split.
apply synced_cons''. assumption.
destruct h5 as [h5 h6].
pose proof (fpl_cons_nin _ _ _ _ _ h0l h2l h5) as h8.
exists h8.
assert (h10:  map_in_dep
                (fun (x : T) (h1 : In x la) => fpl_to_fun_in_dep h8 x (in_cons _ _ _ h1)) =
              map_in_dep (fpl_to_fun_in_dep h5)).
  apply map_in_dep_ext.
  intros x h11. 
  apply functional_extensionality.
  intro hin.
  pose proof (proof_irrelevance _ hin h11); subst.
  pose proof (fpl_to_fun_in_dep_compat h8 x (in_cons _ _ _ h11)) as h8'.
  pose proof (fpl_to_fun_in_dep_compat h5 x h11) as h5'.
  assert (h12:x <> a). intro; subst; contradiction.
  inversion h8' as [hinv | hinv]. inversion hinv. subst.
  contradict h12; auto.
  eapply fpl_functional.
  apply h5. apply hinv. assumption.
unfold nprod. unfold nprod in h10.
unfold fun_from_cons. rewrite h10.
rewrite h6.
pose proof (fpl_to_fun_in_dep_compat h8 a (in_eq a la)) as h11.
destruct h11 as [h11l | h11r].
inversion h11l.
destruct t. simpl. reflexivity.  
pose proof (fpl_in_dom  h5 _ h11r) as h12.
simpl in h12. contradiction.

intros lp' h6.
destruct h6 as [h6l [h6c h6r]].
destruct h6r as [h7 h8].
pose proof (synced_cons' _ _ _ h6c) as h9.
specialize (h1r (tl lp')).
assert (h13:hd  (a, fpl_to_fun_in_dep h7 a (in_eq a la))  lp' =   (a, fpl_to_fun_in_dep h7 a (in_eq a la)) ).
  inversion h6c as [h12]. 
  apply injective_projections. simpl. 
  pose proof (synced_hd  (a, fpl_to_fun_in_dep h7 a (in_eq a la))  _ _ h6c) as h13.
  simpl in h13.
  assumption.
  simpl.
  assert (h13:lp' <> nil). intro h14. subst.
  apply synced_nil2 in h6c. discriminate h6c. 
  pose proof (in_hd_not_nil (impl_type_eq_dec_prod hdt hdu)   _   (a, fpl_to_fun_in_dep h7 a (in_eq a la)) h13) as h14.
  pose proof (fpl_to_fun_in_dep_in_eq _ _ _ h7 _ h14) as h15.  
  rewrite h15.
  f_equal. 
  assert (h13': (hd a (a::la)) = a). simpl. reflexivity. 
  rewrite <- h12 in h13'.
  assert (h16:a = fst (a, fpl_to_fun_in_dep h7 a (in_eq a la))). auto.
  rewrite h16 in h13' at 1.
  rewrite hd_map in h13'.
  gen0.
  rewrite h13'. 
  intro. f_equal. apply proof_irrelevance.
hfirst h1r.
split.
apply no_dup_tl; auto.
  split; auto.
  pose proof (fpl_to_fun_in_dep_compat h7 a (in_eq a la)) as h10.
  pose proof (fpl_cons hdt hdu _ _ _  _  h10 h0 h7) as h11.
  assert (h12:  (remove (impl_type_eq_dec_prod hdt hdu)
                        (a, fpl_to_fun_in_dep h7 a (in_eq a la)) lp') = tl lp').
  rewrite <- h13.
  rewrite remove_hd_no_dup; auto.
  rewrite h12 in h11.
  exists h11.
  destruct t.
  inversion h8 as [h13']. simpl.
  erewrite map_in_dep_fpl_to_fun_cons.
  apply H. assumption.

specialize (h1r hf). subst.
destruct t. simpl.
inversion h8.
assert (h12:lp' <> nil).
  intro h13'.
  subst.
  simpl in h9. destruct h6c as [h6c]. simpl in h6c. discriminate.
pose proof (list_decompose _ (a, u) h12) as h13'.
rewrite h13'.
f_equal.
rewrite surjective_pairing.
f_equal. 
rewrite (synced_hd _ _ _ h6c).
simpl. reflexivity.
pose proof (in_hd_not_nil  (impl_type_eq_dec_prod hdt hdu) _ (a, u) h12 ) as h14. 
pose proof (fpl_to_fun_in_dep_in_eq _ _ _ h7 _ h14) as h15.
rewrite h15. 
inversion h6c as [h16].
pose proof (in_eq a la) as h17.
rewrite <- h16 in h17.
pose proof (hd_map (fun pr:T*U=>fst pr) (a, u) lp') as h18.
rewrite h16 in h18.
simpl in h18.
symmetry.  gen0. rewrite <- h18.
intro.
f_equal.
apply proof_irrelevance.
Qed.


Definition nprod_to_fpl {T U:Type}
           (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
           (la:list T) (lb:list U)
           (pfla:NoDup la) (t:U^(length la)) (pfin:in_n lb t) := 
  proj1_sig (constructive_definite_description _ (nprod_to_fpl_ex pfdt pfdu la lb pfla t pfin)).


Lemma nprod_to_fpl_compat :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U)
         (pfla:NoDup la),
         let n := length la in 
         forall (t:U^n) (pfin:in_n lb t),
           let lp := nprod_to_fpl pfdt pfdu la lb pfla t pfin  in
           NoDup lp /\  synced la lp /\
        exists (pf:functionally_paired_l la lb lp),
          let f := fpl_to_fun_in_dep pf in
          map_in_dep f = nprod_to_list _ _ t.
  unfold nprod_to_fpl; intros; destruct constructive_definite_description; auto.
Qed.


Lemma nprod_to_list_inj : 
  forall {T:Type} (n:nat) (t t':T^n),
    nprod_to_list _ n t = nprod_to_list _ n t' ->
    t = t'.
intros T n.
induction n as [|n h0].
simpl. intros t t' h1. destruct t; destruct t'. reflexivity.
simpl.
intros t t' h1.
destruct t as [ft st]; destruct t' as [ft' st'].
inversion h1.  subst. clear h1.
specialize (h0 st st' H1).
apply injective_projections; auto.
Qed.


Lemma nprod_to_fpl_inj : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U) 
         (pfla:NoDup la) (t t':U^(length la)) (pfin:in_n lb t) 
         (pfin':in_n lb t'),
    nprod_to_fpl pfdt pfdu la lb pfla t pfin =
    nprod_to_fpl pfdt pfdu la lb pfla t' pfin' -> t = t'.
intros T U hdt hdu la lb h1 t t' h2 h3 h4.
pose proof (nprod_to_fpl_compat hdt hdu la lb h1 t h2) as h5.
pose proof (nprod_to_fpl_compat hdt hdu la lb h1 t' h3) as h6.
destruct h5 as [h5a [h5b h5c]].
destruct h6 as [h6a [h6b h6c]].
destruct h5c as [h5c h5d].
destruct h6c as [h6c h6d].
generalize dependent h6c.
rewrite <- h4.
intro h6c.
pose proof (proof_irrelevance _ h5c h6c); subst.
intro h7. simpl in h7.
rewrite h5d in h7.
apply (nprod_to_list_inj _ _ _ h7).
Qed.


Lemma nprod_to_fpl_nil : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (lb:list U) (pfla:NoDup (nil (A:=T))) (t:unit) (pfin:True), 
    nprod_to_fpl pfdt pfdu nil lb pfla t pfin = nil.
intros T U hdt hdu lb h1 t h2.
pose proof (nprod_to_fpl_compat hdt hdu nil lb h1 t h2) as h3.
destruct h3 as [h3a [h3b h3c]].
apply synced_nil1. assumption.
Qed.




Lemma fpl_in_map_snd_impl_in_lb : 
  forall {T U:Type} (la:list T) (lb:list U) 
         (lp:list (T*U)),
    functionally_paired_l la lb lp ->
  forall (y:U),
    In y (map (snd (B:=U)) lp) -> In y lb.
intros T U la lb lp h1 y.
intro h2.
rewrite in_map_iff in h2.
destruct h2 as [pr [h2l h2r]].
destruct h1 as [h1l h1r].
specialize (h1r _ h2r).
destruct h1r as [h1a h1b].
subst.
assumption.
Qed.


Lemma map_in_dep_fpl_to_prod :
  forall {T U:Type} {la:list T} {lb:list U}
         {lp:list (T*U)}
         (pffp:functionally_paired_l la lb lp),
    type_eq_dec T ->
    type_eq_dec U ->
    NoDup la -> NoDup lp ->
    synced la lp ->
    forall (f:fun_in_dep la U),
      map_in_dep f =
      map_in_dep (fun (pr:T*U) (pf:In pr lp) =>
                    f (fst pr) (fpl_in_dom pffp pr pf)).
  intros T U la.
  induction la as [|a la ih]; simpl.
  intros ? lp h1  hdt hdu  ? f. 
  pose proof (fpl_empty1_s  _ _  h1) as h1'. subst.
  simpl; auto. 
  intros lb lp h1 hdt hdu h7 hnd hs f. 
  pose proof (fpl_to_fun_in_dep_compat h1 _ (in_eq _ _)) as h5. 
  pose proof (fpl_cons hdt hdu la lb lp _  h5 h7 h1) as h6.
  simpl in h6.
  pose proof (no_dup_cons_nin _ _ h7) as h8.
  pose proof (no_dup_cons _ _ h7) as h9.
  pose proof (no_dup_remove (impl_type_eq_dec_prod hdt hdu) lp (a, fpl_to_fun_in_dep h1 a (in_eq a la)) hnd) as h10.
  specialize (ih _ _ h6 hdt hdu h9 h10).
  pose proof (remove_synced_cons hdt hdu la lb lp (a, fpl_to_fun_in_dep h1 a (in_eq a la)) hs h1 h5 hnd) as h11. 
  generalize dependent h6.
  rewrite h11.
  intros h6 h12.
  pose proof (in_not_nil _ _ h5) as h13.
  pose proof (list_decompose  lp (a, fpl_to_fun_in_dep h1 a (in_eq a la)) h13) as h14.
  pose proof (synced_hd (a, fpl_to_fun_in_dep h1 a (in_eq a la)) _ _ hs) as h15.
  simpl in h15.
  rewrite <- h15 in hs.
  pose proof (remove_synced_cons hdt hdu la lb lp _ hs) as h16.
  rewrite h15 in h16.
  specialize (h16 h1).
  hfirst h16.
  apply in_hd_not_nil; auto.
  apply impl_type_eq_dec_prod; auto.
  specialize (h16 hf hnd).
  rewrite <- h16 in h14.
  match goal with |- _ = map_in_dep ?g' => pose g' as g end.
  pose proof (map_in_dep_functional g h14) as h17.
  fold g. rewrite h17.
  rewrite map_in_dep_cons.
  rewrite transfer_fun_in_dep_compat.
  unfold g.
  f_equal.
  symmetry.
  gen0.
  rewrite h15.
  intro. f_equal. apply proof_irrelevance.
  rewrite h15 in hs.
  pose proof (fpl_impl_map _ lb  h7 lp h1 hs) as h18.
  simpl in h18.
  match goal with |- _ = map_in_dep ?q => pose q as q' end.
  rewrite h14 in h18 at 1.
  inversion h18 as [h19]. clear h18.
  match goal with |- _ = map_in_dep ?k => pose proof (map_in_dep_functional k H) as h22 end.
  rewrite h22.
  rewrite map_in_dep_map_in_dep.
  apply map_in_dep_ext.
  intros x h23. apply functional_extensionality.
  intro y.
  rewrite transfer_fun_in_dep_compat.
  unfold fun_from_cons.
  rewrite transfer_fun_in_dep_compat.
  f_equal.
  apply proof_irrelevance.
Qed.


Lemma list_fst_snd_eq : 
  forall {T U:Type} (lp lp':list (T*U)),
    map (fst (B:=U)) lp = map (fst (B:=U)) lp' ->
    map (snd (B:=U)) lp = map (snd (B:=U)) lp' ->
    lp = lp'.
intros T U lp. 
induction lp as [|a lp h1]; auto; simpl.
intros lp' h2 h3. 
symmetry in h3.
pose proof (map_eq_nil _ _ h3).
subst. reflexivity. 
intro lp'.
induction lp' as [|a' lp' h2]; simpl.
intros h2 h3. discriminate.
intros h3 h4.
inversion h3; inversion h4.
specialize (h1 _ H1 H3).
f_equal.
apply injective_projections; auto.
assumption.
Qed.


Lemma fpl_to_fun_in_dep_eq :
  forall {T U:Type} {la:list T} {lb:list U}
         {lp lp':list (T*U)}
         (pf: functionally_paired_l la lb lp)
         (pf':functionally_paired_l la lb lp'),
    lp = lp' ->
    fpl_to_fun_in_dep pf  = fpl_to_fun_in_dep pf'.
  intros ? ? ? ? ? ? h1 h2 ?; subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.


Lemma fpl_to_fun_in_dep_eq' :
  forall {T U:Type} {la:list T} {lb:list U}
         {lp lp':list (T*U)}
         (pffp: functionally_paired_l la lb lp)
         (pffp':functionally_paired_l la lb lp'),
    lp = lp' ->
    forall (x:T) (pfa:In x la),
      fpl_to_fun_in_dep pffp x pfa = fpl_to_fun_in_dep pffp' x pfa.
  intros ? ? ? ? ? ? h1 h2 ?; subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.
 
Fixpoint nprod_of_list_n 
         {T:Type} (l:list T) (n:nat) (def:T) {struct n} : 
  T ^ n :=
  match n return T^n with 
    | O => tt
    | S n' => ((hd def l), (nprod_of_list_n (tl l) n' def))
  end.


      
Lemma nprod_to_fpl_surj : 
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U) (lp:list (T*U))
    (pf:functionally_paired_l la lb lp) (pfla:NoDup la), 
    NoDup lp -> synced la lp ->
    let f := fpl_to_fun_in_dep pf in
    let n := length la in 
    exists (t:U^n) (pfin:in_n lb t),
      map_in_dep f = nprod_to_list _ n t /\
      lp = (nprod_to_fpl pfdt pfdu la lb pfla t pfin).
intros T U hdt hdu la lb lp h1 h2 h3 h4 f n.
pose proof (synced_length _ _ h4) as h5.  
pose (nprod_of_list _ (map (@snd _ _) lp)) as t.
 assert (h6':length (map (snd (B:=U)) lp) = n).
  unfold n. rewrite h5. apply map_length.
subst.
pose (transfer_dep h6' t) as t'.
exists t'.
assert (h8:in_n lb t).
  rewrite in_n_in_nprod_to_list_compat_iff.
  intros x h8.
  unfold t in h8.
  rewrite nprod_to_list_nprod_of_list_eq in h8.
  eapply fpl_in_map_snd_impl_in_lb; auto.
  apply h1; auto. assumption.
pose (fun n':(sigT (fun n => (nprod U n))) => (in_n (n:=(projT1 n')) lb (projT2 n'))) as P.
pose proof (transfer_dep_prop _ _ h6' t P) as h9.
unfold P in h9. simpl in h9.
unfold t'.
rewrite h9 in h8.
exists h8.
split. 
pose proof (fpl_impl_map  _ _ h2 _ h1  h4) as h10.
pose proof (f_equal (map (@snd T U)) h10) as h11.
rewrite map_map_in_dep in h11. simpl in h11.
let P := type of h11 in match P with _ = ?x => assert (h12:x = map_in_dep f) end.
  apply map_in_dep_ext. intros.
  apply functional_extensionality.
  intro. unfold f; auto.
rewrite <- h12.
unfold t. 
pose proof (transfer_dep_existT _ _ h6' (nprod_of_list U (map (snd(B:=U)) lp))) as h13.
dependent rewrite <- h13.
rewrite nprod_to_list_nprod_of_list_eq.
rewrite (map_in_dep_fpl_to_prod h1); auto.
rewrite map_eq_map_in_dep.
eapply map_in_dep_ext.
intros pr h14.
apply functional_extensionality.
intro.
erewrite fpl_to_fun_in_dep_in_eq.
f_equal.
pose proof (nprod_to_fpl_compat hdt hdu _ _  h2 _ h8) as h10. 
destruct h10 as [h10a [h10b h10c]].
destruct h10c as [h10 h11].
unfold t in h11.
pose proof (transfer_dep_existT _ _ h6' (nprod_of_list U (map (snd (B:=U)) lp))) as h13.
rewrite <- nprod_to_list_nprod_of_list_eq in h11 at 1.
pose proof (fpl_impl_map _ _ h2 _ h10 h10b) as h14.
pose proof (f_equal (map (@snd T U)) h14) as h15. 
rewrite h14.
pose proof (fpl_impl_map _ _ h2 _ h1 h4) as h10'.
pose proof (f_equal (map (@snd T U)) h10') as h11'.
rewrite map_map_in_dep in h11'. simpl in h11'.
rewrite h10' at 1.
f_equal.
apply fun_in_dep_ext.
intros; auto.
f_equal. 

assert (h16:lp = (nprod_to_fpl hdt hdu la lb  h2
       (transfer_dep  h6' t) h8)).   
  unfold t.  
  unfold nprod_to_fpl. 
  destruct constructive_definite_description as [lp' h18]. simpl.
  destruct h18 as [h18a [h18b h18c]].
  destruct h18c as [h19 h20].
  dependent rewrite <- h13 in h20.
  rewrite nprod_to_list_nprod_of_list_eq in h20. 
  rewrite (map_in_dep_fpl_to_prod h19) in h20; auto.
  apply list_fst_snd_eq.
  destruct h18b as [h18b].
  pose proof h4 as h4'. clear h4.
  destruct h4' as [h4'].
  rewrite <- h18b in h4'.
  assert (h21:(fun pr:T*U=>fst pr) = (fst (B:=U))).
    apply functional_extensionality; auto.
  rewrite h21 in h4'.
  assumption.
  symmetry. 
  rewrite <- h20.
  rewrite map_eq_map_in_dep.
  apply map_in_dep_ext.
  intros pr h21.
  apply functional_extensionality.
  intro.
  symmetry.
  erewrite fpl_to_fun_in_dep_in_eq.
  f_equal.
  apply fpl_to_fun_in_dep_eq'; auto.
Qed.

Lemma in_n_Sn : forall {T:Type} {n:nat} (l:list T) (t:T^n) (x:T),
                  in_n (n:=S n) l (x, t) -> In x l /\ in_n l t.
intros T n l t x h1.
rewrite in_n_in_nprod_to_list_compat_iff in h1.
split.
apply h1. simpl. left. reflexivity.
assert (h2:forall x0 : T, In x0 (nprod_to_list T n t) -> In x0 l).
  intros x' h2.
  specialize (h1 x').
  simpl in h1.
  apply h1. right. assumption.
rewrite <- in_n_in_nprod_to_list_compat_iff in h2.
assumption.
Qed.


Lemma im_nprod_new_proj2_ex : 
  forall {T:Type} {l:list T} {n:nat} (t':{t:T^n | in_n l t}),
         exists! t:{x:T|In x (nprod_to_list _ _ (proj1_sig t'))}^n,
           proj1_sig t' = im_nprod (@proj1_sig _ _) t.
intros T l n t'.           
induction n as [|n ret]. 
(* 0 *)
simpl. exists tt.  red.  simpl in t'.  destruct t'.  simpl.  destruct x.  split. reflexivity. intros tt' ?. destruct tt'. assumption.
(* S n*)
destruct t' as [t' h2].  destruct t' as [ft' st'];  simpl. 
apply in_n_Sn in h2.
destruct h2 as [h2l h2r].
specialize (ret (exist _ _ h2r)). simpl in ret. 
pose (exist (fun x=>ft'=x \/ In x (nprod_to_list T n st')) ft' (or_introl (In ft' (nprod_to_list T n st')) (eq_refl _))) as ft.
destruct ret as [ret h3].
red in h3. 
pose  (im_nprod (U:={x:T|ft' = x \/ In x (nprod_to_list T n st')})
                (fun x=>(exist _ (proj1_sig x) (or_intror  (ft' = (proj1_sig x)) (proj2_sig x))))
                ret) as st.
exists (ft, st).
red. simpl. split.
unfold st. 
apply injective_projections; simpl. reflexivity. 
rewrite im_nprod_im_nprod. simpl.
destruct h3 as [h3l h3r].
rewrite h3l at 1.
f_equal.
intros x h4.
destruct x as [fx sx].
simpl in h4.
inversion h4 as [h5].
destruct fx as [fx h6].
simpl in h4.
simpl in h5.
unfold ft. unfold st.
apply injective_projections.
simpl.
destruct h6.
apply proj1_sig_injective.
simpl.
assumption.
apply proj1_sig_injective.
simpl.
assumption.
simpl.  
destruct h3 as [h3l h3r]. 
rewrite h3l in H at 1. 
apply im_nprod_inj with (f:=(proj1_sig (P:=fun x:T=>ft' = x \/ In x (nprod_to_list T n st')))).
red.
apply proj1_sig_injective.
rewrite im_nprod_im_nprod.
simpl.
assumption.
Qed.


Definition im_nprod_new_proj2 {T:Type} {l:list T} {n:nat} 
           (t':{t:T^n | in_n l t}) : {x:T|In x (nprod_to_list _ _ (proj1_sig t'))}^n := 
  (proj1_sig (constructive_definite_description _ (im_nprod_new_proj2_ex t'))).

Lemma im_nprod_new_proj2_compat : 
  forall {T:Type} {l:list T} {n:nat} 
           (t':{t:T^n | in_n l t}),
    proj1_sig t' = im_nprod (@proj1_sig _ _) (im_nprod_new_proj2 t').
intros T l n t'.
unfold im_nprod_new_proj2.
destruct constructive_definite_description.
simpl.
assumption.
Qed.


Lemma nprod_in_n_to_nprod_in_l_ex : 
  forall {T:Type} {l:list T} {n:nat}
         (t':{t:T^n|in_n l t}), exists! ret:{x:T|In x l}^n,
    proj1_sig t' = im_nprod (@proj1_sig _ _) ret.
intros T l n t'.
pose (im_nprod_new_proj2 t') as t''.
pose proof (im_nprod_new_proj2_compat t') as h0.
destruct t' as [t' h1].
simpl in h0. 
pose proof h1 as h1'.
rewrite in_n_in_nprod_to_list_compat_iff in h1'.
simpl in t''.
simpl.
exists (im_nprod (fun x=>(exist _ (proj1_sig x) (h1' _ (proj2_sig x)))) t'').
red.
split.
rewrite im_nprod_im_nprod.
simpl.
unfold t''.
assumption.
intros x' h2.
unfold t''.
apply im_nprod_inj with (f:=(proj1_sig (P:=fun x:T => In x l))).
red.
apply proj1_sig_injective.
rewrite im_nprod_im_nprod.
simpl.
rewrite <- h2.
symmetry. assumption.
Qed.

Definition nprod_in_n_to_nprod_in_l {T:Type} (l:list T) (n:nat) 
           (t':{t:T^n | in_n l t}) : {x:T|In x l}^n :=
  (proj1_sig (constructive_definite_description _ (nprod_in_n_to_nprod_in_l_ex t'))).

Lemma nprod_in_n_to_nprod_in_l_compat : 
  forall {T:Type} {l:list T} {n:nat}
         (t':{t:T^n|in_n l t}),
    proj1_sig t' = im_nprod (@proj1_sig _ _) (nprod_in_n_to_nprod_in_l _ _ t') .
intros T l n t'.
unfold nprod_in_n_to_nprod_in_l.
destruct constructive_definite_description.
simpl.
assumption.
Qed.

Lemma nprod_in_n_to_nprod_in_l_bij : 
  forall {T:Type} (l:list T) (n:nat), 
    bijective (nprod_in_n_to_nprod_in_l l n).
intros T l n.
red.
split.
red.
intros x1 x2 h1.
pose proof (f_equal (fun x => im_nprod (@proj1_sig _ _) x) h1) as h2.
simpl in h2.
do 2 rewrite <- nprod_in_n_to_nprod_in_l_compat in h2.
apply proj1_sig_injective. assumption.
red. 
intro t'.
induction n as [|n h1]. 
simpl in t'. simpl.
exists (exist _ tt I).
unfold nprod_in_n_to_nprod_in_l.
destruct constructive_definite_description.
simpl.
simpl in x. destruct t'; destruct x.
reflexivity.
simpl in t'.
destruct t' as [ft' st'].
specialize (h1 st').
destruct h1 as [t h2].
simpl.
destruct t as [t h1].
destruct ft' as [x h3].
exists (exist _ (x, t) (conj h3 h1)).
apply injective_projections.
simpl.
apply proj1_sig_injective.
simpl.
unfold nprod_in_n_to_nprod_in_l.
destruct constructive_definite_description as [y h4].
simpl. simpl in h4.
inversion h4.
reflexivity.
simpl.
unfold nprod_in_n_to_nprod_in_l.
destruct constructive_definite_description as [y h4].
simpl. simpl in h4. 
subst.
inversion h4.
apply im_nprod_inj with (f:=proj1_sig (P:=fun x0:T => In x0 l)).
red.
apply proj1_sig_injective.
subst.
rewrite <- nprod_in_n_to_nprod_in_l_compat.
simpl.
reflexivity.
Qed.

Lemma in_sig_fun_ex : 
  forall {T:Type} {l:list T}
         (x:{t:T|In t l}), 
            exists! y:{t:T|Inn (list_to_set l) t},
  proj1_sig x = proj1_sig y.
intros T l x.
destruct x as [x h1].
pose proof h1 as h1'.
rewrite list_to_set_in_iff in h1'.
exists (exist _ _ h1').
red.
simpl. split; auto.
intros y h2.
destruct y as [y h3].
simpl in h2. subst.
pose proof (proof_irrelevance _ h3 h1'); subst; f_equal.
Qed.

Definition in_sig_fun {T:Type} {l:list T}
  (x:{t:T|In t l}):= 
  proj1_sig (constructive_definite_description _ (in_sig_fun_ex x)).

Lemma in_sig_fun_compat : forall {T:Type} {l:list T}
      (x:{t:T|In t l}), proj1_sig x = proj1_sig (in_sig_fun x).
intros T l x.
unfold in_sig_fun. destruct constructive_definite_description as [x' h1].
simpl. assumption.
Qed.

Lemma in_sig_fun_bij : forall {T:Type} (l:list T),
                         bijective (@in_sig_fun _ l).
intros T l.
red. split.
red.
intros x1 x2 h1.
pose proof (f_equal (@proj1_sig _ _) h1) as h2.
do 2 rewrite <- in_sig_fun_compat in h2.
apply proj1_sig_injective; auto.
red.
intro y.
destruct y as [y h1].
pose proof h1 as h1'.
rewrite <- list_to_set_in_iff in h1'.
exists (exist _ y h1').
apply proj1_sig_injective.
simpl.
rewrite <- in_sig_fun_compat.
simpl.
reflexivity.
Qed.

Lemma finite_list_sig : forall {T:Type} (l:list T),
                           FiniteT {x:T | In x l}.
intros T l.
pose proof (in_sig_fun_bij l) as h1.
apply bijective_impl_invertible in h1.

pose proof (list_to_set_finite l) as h2.
pose proof (Finite_ens_type _ h2) as h3.
apply (bij_finite _ _ 
                  (proj1_sig (function_inverse (@in_sig_fun _ l) h1))
                  h3 (invertible_impl_inv_invertible _ h1)).
Qed.

Lemma cardinal_full_set_in_l : 
  forall {T:Type} (l:list T),
    NoDup l ->
    cardinal _ (Full_set {x:T | In x l}) (length l).
intros T l h0.
pose (list_to_set l) as A.
induction l as [|a l h1].
simpl.
rewrite full_false_empty.
constructor.
simpl.
inversion h0. subst.
specialize (h1 H2).
pose (fun x':{x:T|In x l} => (exist (fun x=>a=x \/ In x l) (proj1_sig x') (or_intror (a=(proj1_sig x')) (proj2_sig x')))) as f.
assert (h3:Injective f).
  red. unfold f.
  intros x1 x2 h4.
  destruct x1; destruct x2.
  simpl in h4.
  apply proj1_sig_injective.
  simpl.
  pose proof (f_equal (@proj1_sig _ _) h4) as h5. simpl in h5.
  assumption.
pose proof (injection_preserves_cardinal _ _ f _ _ h1 h3) as h4.
assert (h5:Im (Full_set {x:T | In x l}) f = [x':{x:T | a = x \/ In x l} | In (proj1_sig x') l]).
  apply Extensionality_Ensembles.
  red.  split.
  red.
  intros y h5.
  destruct h5 as [x h5 y]. subst.
  constructor.
  unfold f.
  simpl. apply proj2_sig.
  red.
  intros y h5.
  destruct h5 as [h5].
  destruct y as [y h6].
  simpl in h5.
  apply Im_intro with (exist _ y h5).
  constructor.
  unfold f.
  apply proj1_sig_injective.
  simpl. reflexivity.
rewrite h5 in h4.
assert (h6:Full_set {x:T | a = x \/ In x l} = 
           Add [x':{x:T | a = x \/ In x l} | In (proj1_sig x') l] (exist _ a (or_introl (In a l) (eq_refl _)))).
  apply Extensionality_Ensembles.
  red. split.
  red.
  intros x h6.
  destruct x as [x h7]. 
  destruct h7 as [h7l | h7r].
  right. rewrite h7l.
  constructor.
  constructor.
  constructor. simpl.
  assumption.
  red. constructor.
rewrite h6.
apply card_add.
assumption.
intro h7.
destruct h7 as [h7].
simpl in h7.
contradiction.
Qed.

Lemma cardinal_in_l : 
  forall {T:Type} (l:list T), 
    NoDup l -> 
    FiniteT_card {x:T | In x l} (finite_list_sig l) =
    length l.
intros T l h1.
apply FiniteT_card_cond.
apply cardinal_full_set_in_l; auto.
Qed.


Definition l_exp_seg_to_nprod_in_l 
           {T:Type} (l:list T) (n:nat) 
           (f:{m:nat | m < n} -> {x:T|In x l}) : {x:T | In x l} ^ n.
induction n as [|n h1].
simpl. refine tt.
simpl.
pose (fun m':{m:nat | m < n} => f (exist _ (proj1_sig m') (lt_S _ _ (proj2_sig m')))) as f'.
refine ((f (exist _ n (lt_n_Sn n)), (h1 f'))).
Defined.


Lemma l_exp_seg_to_nprod_in_l_compat : 
  forall {T:Type} (l:list T) (n:nat) 
         (f:{m:nat | m < n} -> {x:T|In x l})
         (def:{x:T|In x l})
         (x:nat) (pf:x < n),
    nth (n - x -1) (nprod_to_list _ _ (l_exp_seg_to_nprod_in_l _ _ f)) def =
    f (exist _ x pf).
intros T l n f def x.
revert f def.
induction n as [|n h1]; induction x as [|x h2].
simpl; intros; try omega.
intros; omega.
intros f def h2.
destruct (zerop n) as [h3 | h4].
subst; simpl.
pose proof (proof_irrelevance _ h2 (lt_n_Sn _)); subst; auto.
simpl.
assert (h5:n-0 = n). auto with arith.
rewrite h5.
destruct n as [|n]. omega.
specialize (h1 (fun m' : {m : nat | m < S n} =>
            f
              (exist (fun m : nat => m < S (S n)) (proj1_sig m')
                 (lt_S (proj1_sig m') (S n) (proj2_sig m')))) def h4).
assert (h6:S n - 0 -1 = n). omega.
rewrite h6 in h1.
rewrite h1.
simpl.
pose proof (proof_irrelevance _ h2 
                              (lt_S 0 (S n) h4)); f_equal; f_equal; auto.
intros f def h3.
assert (h4:x < n). omega.
assert (h5:S n - S x - 1 = n - x -1). omega. 
rewrite h5.
destruct (zerop (n - x -1)) as [h6 | h7].
rewrite h6. 
simpl.
assert (h7:n = S x). omega.
subst.
pose proof (proof_irrelevance _ h3
                              (lt_n_Sn (S x))); subst; auto.
destruct (lt_eq_lt_dec x (n-1)) as [[h10' | h11] | h12].
pose proof (S_pred _ _ h7) as h8.
rewrite h8.
simpl. 
assert (h9:pred (n - x -1) = n - S x - 1).  
  apply S_inj.  
  rewrite <- (S_pred _ _ h7).
  assert (h10:1 <= n - S x). 
  pose proof (lt_le_S _ _ h7) as h9.
  rewrite (minus_plus_simpl_l_reverse (n-x) 1 x) in h9.
  assert (h11:x <= n).  
    pose proof (lt_S _ _ h4) as h11.
    unfold lt in h11.
    apply le_S_n. assumption.
  rewrite <- (le_plus_minus _ _ h11) in h9.
  rewrite <- (S_compat x).
  assumption.
  rewrite (minus_Sn_m _ _ h10).
  do 2 rewrite <- pred_of_minus.
  f_equal.
  unfold lt in h4.
  rewrite (minus_Sn_m _ _ h4).
  auto with arith.
rewrite <- h9 in h1.

assert (h13 : S x < n). unfold lt in h10'.
  pose proof (le_lt_n_Sm _ _ h10') as h14.
  rewrite <- pred_of_minus in h14.
  rewrite <- (S_pred _ _ h4) in h14.
  assumption.
specialize (h1 (fun m' : {m : nat | m < n} =>
            f
              (exist (fun m : nat => m < S n) (proj1_sig m')
                 (lt_S (proj1_sig m') n (proj2_sig m')))) def h13).
rewrite h1.
f_equal.
f_equal.
apply proof_irrelevance. 
subst. 
destruct (zerop (n-1)) as [h8 | h9].
rewrite h8 in h7.
rewrite <- (minus_n_O n) in h7.
rewrite h8 in h7.
contradict h7. auto with arith.
rewrite <- pred_of_minus in h9.
rewrite <- (pred_of_minus n) in h7.
rewrite (n_minus_pred_n _ h9) in h7.
assert (h10:1 - 1 = 0). auto with arith.
rewrite h10 in h7.
contradict h7. auto with arith.
unfold lt in h12. rewrite <- pred_of_minus in h12.
rewrite <- (S_pred _ _ h4) in h12.
clear h5 h7.
omega.
Qed.


Lemma l_exp_seg_to_nprod_in_l_bij : 
  forall {T:Type} (l:list T) (n:nat),
    (bijective (l_exp_seg_to_nprod_in_l l n)).
intros T l n.
induction n as [|n h1].
(* 0 *)
red. split.
red.
intros x1 x2 h1.
simpl in h1.
apply functional_extensionality.
intro x. destruct x. 
pose proof (lt_n_0 x). contradiction.
red.
intro y.
exists (fun x:{m:nat | m < 0} => False_rect {x:T | In x l} (lt_n_0 _ (proj2_sig x))).
simpl. simpl in y. destruct y. reflexivity.
(* S n *)
red. split. 
(* injective *)
red.
intros f1 f2 h2.
simpl in h2. 
inversion h2 as [h0]. 
apply functional_extensionality.
intro x.
destruct x as [x h3]. 
destruct (lt_eq_lt_dec x n) as [[h4 | h5] | h6].
pose proof (f_equal (@nprod_to_list {x:T|In x l} n) H) as h5.
pose proof (f_equal (fun l => nth (n-x-1) l (f1 (exist _ x h3))) h5) as h6.
simpl in h6. 
do 2 rewrite (l_exp_seg_to_nprod_in_l_compat _ _ _ _ _ h4) in h6. 
simpl in h6.
pose proof (proof_irrelevance _ h3 (lt_S _ _ h4)) as h7; subst; auto.
subst.
pose proof (proof_irrelevance _ h3 (lt_n_Sn n)); subst; auto.
omega.
(* surjective *)
red.
simpl.
intro y.
red in h1.
destruct h1 as [h1l h1r].
red in h1r.
destruct y as [fy sy].
specialize (h1r sy).
destruct h1r as [f h2].
pose (fun m':{m:nat | m <= n} => 
        let pf := (le_lt_eq_dec (proj1_sig m') n (proj2_sig m')) in
        match pf with
          | left pf' => (f (exist (fun m=>m<n) (proj1_sig m') pf'))
          | right _ => fy
        end) as f'.
pose (fun m':{m:nat | m < S n} => f' (exist _ _ (le_S_n _ _ (proj2_sig m')))) as f''.
exists f''.
apply injective_projections.
simpl.
unfold f''.
simpl.
unfold f'.
simpl.
destruct le_lt_eq_dec as [h3 | h4].
contradict h3. auto with arith.
reflexivity.
simpl.
unfold f''.
simpl.
unfold f'.
simpl.
assert (h3:(fun m' : {m : nat | m < n} =>
      match
        le_lt_eq_dec (proj1_sig m') n
          (le_S_n (proj1_sig m') n (lt_S (proj1_sig m') n (proj2_sig m')))
      with
      | left pf' => f (exist (fun m : nat => m < n) (proj1_sig m') pf')
      | right _ => fy
      end) = f).
  apply functional_extensionality.
  intro x.
  destruct x as [x h3].
  simpl.
  destruct le_lt_eq_dec as [h4 | h5].
  pose proof (proof_irrelevance _ h3 h4); subst; auto.
  omega.
rewrite h3.
exact h2.
Qed.

Lemma finite_l_exp_seg : 
  forall {T:Type} (l:list T) (n:nat),
    FiniteT ({m:nat | m < n} -> {x:T|In x l}).
intros T l n.
pose proof (finite_nat_initial_segment n) as h1.
pose proof (list_to_set_finitet l) as h2.
apply finite_exp; auto.
Qed.


Lemma finite_nprod_in_l : 
  forall {T:Type} (l:list T) (n:nat),
    FiniteT ({x:T|In x l}^n).
intros T l n.
pose proof (finite_nat_initial_segment n) as h1.
pose proof (finite_list_sig l) as h2.
pose proof (finite_exp _ _ h1 h2) as h3.
pose proof (l_exp_seg_to_nprod_in_l_bij l n) as h4.
apply bijective_impl_invertible in h4. 
apply (bij_finite _ _ _ h3 h4).
Qed.

End NprodList.


Section ListSort.
  Fixpoint maxlR {T:Type} (R:Relation T) (pftso:tso_dec R)
           (l:list T) : l <> nil -> T :=
  match l with
    | nil => fun pf => False_rect T (pf eq_refl)
    | a::l' => fun pf =>
                 match (nil_dec' l') with
                   | left _ => a
                   | right pf' =>
                     maxR R pftso a (maxlR R pftso l' pf')
                 end
  end.

  Lemma in_maxlR :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R)
         (l:list T) (pf:l <> nil),
    In (maxlR R pftso l pf) l.
  intros T R h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros h2.
  destruct (nil_dec' l) as [| h3]; subst.
  left; auto.
  specialize (ih h3).
  unfold maxR.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]. 
  right; auto.
  left; auto.
  Qed.  

  Lemma maxlR_compat :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R)
         (pftr:Transitive _ R)
         (l:list T) (x:T)
         (pfin:In x l),
    let pfn := in_not_nil _ _ pfin in
    Rw R x (maxlR R pftso l pfn).
  intros T R h1 ht l.
  induction l as [|a l hi]; simpl. 
  intros; contradiction.
  intros n h2.
  red.
  destruct (nil_dec' l) as [h3 | h3]; subst.
  right; destruct h2; try contradiction; auto.
  destruct h2 as [h2l | h2r]. subst.
  unfold maxR.
  lr_if_goal. fold hlr. destruct hlr as [hl | hr].
  destruct hl; [left; auto | right; auto]. 
  right; auto.
  unfold maxR.
  lr_if_goal. fold hlr. destruct hlr as [hl | hr].
  specialize (hi _ h2r).
  simpl in hi.
  pose proof (proof_irrelevance _ h3 (in_not_nil _ _ h2r)); subst; auto.
  specialize (hi _ h2r).
  simpl in hi.
  red in hi. 
  destruct hi as [hi | hi].
  left. 
  red in ht.
  eapply ht.
  apply hi.
  pose proof (proof_irrelevance _ h3 
                                (in_not_nil l n h2r)); subst; auto.
  rewrite hi.
  left.
  pose proof (proof_irrelevance _ h3
                                (in_not_nil l n h2r)).
  clear hi. subst; auto.
Qed.
  
Lemma maxlR_functional :
  forall {T:Type} (R:Relation T) 
         (pft:tso_dec R)
         (l l':list T)
         (pfl:l <> nil) (pfl':l' <> nil),
    l = l' ->
    maxlR R pft l  pfl =
    maxlR R pft l' pfl'.
  intros T R h1 l l' h2 h3 h4; subst.
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
Qed.

Lemma maxlR_functional' :
  forall {T:Type} (R R':Relation T) (pfso:sort_dec R) (pfso':sort_dec R')
         (pft:tso_dec R)
         (l l':list T)
         (pfl:l <> nil) (pfl':l' <> nil),
    R = R' ->
    l = l' ->
    maxlR R pft  l  pfl =
    maxlR R' (sort_dec_tso_dec _ pfso') l' pfl'.
  intros T R R' h1 h1' ht l l' h2 h2' h3 ?.
  subst.
  pose proof (proof_irrelevance _ h2 h2'); subst.
  f_equal.
  symmetry.
  apply sort_dec_tso_dec_eq.
Qed.


Lemma maxlR_eq_iff : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (m:T) (pfn: l <> nil),
    let pftso := sort_dec_tso_dec _ pfso in
    (maxlR R pftso l pfn = m) <->
    (In m l /\ forall x:T, In x l -> Rw R x m).
  intros T R h1 l m h3 htso.
  split; intro h4.
  subst.
  split.
  apply in_maxlR.
  intros x h4.
  erewrite maxlR_functional.
  apply maxlR_compat; auto.
  apply sort_dec_trans; auto.
  reflexivity.
  destruct h4 as [h4 h5].
  specialize (h5 (maxlR R htso l h3) (in_maxlR _ _ _ _)).
  pose proof (maxlR_compat R htso (sort_dec_trans _ h1) l m h4) as h6.
  simpl in h6.
  pose proof (proof_irrelevance _ h3
                                (in_not_nil l m h4)); subst.
  apply (antisym_Rw R (sort_dec_so _ h1) _ _ h5 h6).
  Grab Existential Variables.
  assumption.
Qed.

  
Lemma maxlR_lrep :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (x:T) (n:nat),
      let pftso := sort_dec_tso_dec _ pfso in
      maxlR R pftso (lrep x (S n)) (cons_neq_nil _ _)  = x.
  intros T R hso x n. revert x.
  induction n as [|n ih]; simpl.
  destruct (nil_dec' nil) as [|h3]; auto.
  contradict h3; auto.
  intro x. 
  destruct (nil_dec' (x::lrep x n)) as [h3 | h3]; auto.
  specialize (ih x).
  erewrite <- ih at 3.
  simpl. 
  destruct (nil_dec' (lrep x n)) as [h4 | h4]; auto.
  apply maxR_idem. 
  unfold  maxR.
  destruct (sort_dec_tso_dec R hso x
                             (maxlR R (sort_dec_tso_dec R hso) (lrep x n) h4)) as [h5 | h5].
  destruct h5 as [h5 | h5].
  destruct (sort_dec_tso_dec R hso x x) as [h6 | h6]; auto. 
  destruct (sort_dec_tso_dec R hso x  (maxlR R (sort_dec_tso_dec R hso) (lrep x n) h4)) as [h7 | h7]; auto.
  pose proof (sort_dec_trans _ hso _ _ _ h5 h7) as h8.
  apply (sort_dec_irrefl _ hso) in h8; contradiction.
  apply (sort_dec_irrefl _ hso) in h6; contradiction.
  destruct (sort_dec_tso_dec R hso x
                             (maxlR R (sort_dec_tso_dec R hso) (lrep x n) h4)) as [h6 | h6]; auto.
  destruct (sort_dec_tso_dec R hso x x); auto. 
Qed.


Lemma maxlR_sing :
  forall {T:Type} (R:Relation T) (pftso:sort_dec R)
         (l:list T) (x:T)
         (pfe:Singleton x = list_to_set l),
    let pft := sort_dec_tso_dec _ pftso in
    let pf := inh_sing x in
    let pfn := P_subst pf pfe in
    let pfn' := inh_not_nil _ pfn in
    x = maxlR R pft l pfn'.
  intros T R h1 l x h2; simpl.
  pose proof h2 as h2'.
  symmetry in h2'.
  apply impl_all_sing in h2'.
  apply all_sing_lrep in h2'.
  symmetry. 
  gen0.
  rewrite h2'.
  intro h3.
  erewrite maxlR_functional.
  erewrite <- maxlR_lrep.
  apply maxlR_functional.
  reflexivity.
  simpl. 
  rewrite lrep_nil_iff in h3.
  rewrite neq0_iff in h3.
  rewrite (S_pred _ _ h3).
  simpl.
  reflexivity.
  Grab Existential Variables.
  simpl.
  apply cons_neq_nil.
Qed.

Lemma maxlR_list_singularize : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) 
         (pfn:l <> nil),
    let pftso := sort_dec_tso_dec _ pfso in
    let pfdt := sort_dec_eq_dec _ pfso in
    let pfn' := iffn2 (list_singularize_nil_iff pfdt l) pfn in
    maxlR R pftso l pfn =
    maxlR R pftso (list_singularize pfdt l) pfn'.
  intros T R h1 l h2 htso hdt hn.
  unfold htso.
  rewrite maxlR_eq_iff.
  split.
  rewrite list_singularize_in_iff.
  apply in_maxlR.
  intros x h3.
  rewrite (list_singularize_in_iff hdt) in h3.
  pose proof (maxlR_compat R (sort_dec_tso_dec R h1) (sort_dec_trans _ h1) (list_singularize hdt l) x h3) as h4.
  simpl in h4.
  pose proof (proof_irrelevance
                _ hn
                (in_not_nil _ _ h3)) as h5.
  rewrite h5; auto.
Qed.


Fixpoint minlR {T:Type} (R:Relation T) (pftso:tso_dec R)
         (l:list T) : l <> nil -> T :=
  match l with
    | nil => fun pf => False_rect T (pf eq_refl)
    | a::l' => fun pf =>
                 match (nil_dec' l') with
                   | left _ => a
                   | right pf' =>
                     minR R pftso a (minlR R pftso l' pf')
                 end
  end.

Lemma in_minlR :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R)
         (l:list T) (pf:l <> nil),
    In (minlR R pftso l pf) l.
  intros T R h1 l.
  induction l as [|a l ih]; simpl; auto.
  intros h2.
  destruct (nil_dec' l) as [| h3]; subst.
  left; auto.
  specialize (ih h3).
  unfold minR.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  left; auto.
  right; auto.
Qed.  
  
Lemma minlR_compat :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R)
         (pftr:Transitive _ R)
         (l:list T) (x:T)
         (pfin:In x l),
    let pfn := in_not_nil _ _ pfin in
    Rw R (minlR R pftso l pfn) x.
  intros T R h0 h1 l.
  induction l as [|a l ih]; simpl. 
  intros; contradiction.
  intros n h2.
  red.
  destruct (nil_dec' l) as [h3 | h3]; subst.
  right; destruct h2; try contradiction; auto.
  destruct h2 as [h2l | h2r]. subst.
  unfold minR. 
  lr_if_goal. fold hlr. destruct hlr as [hl | hr].
  right; auto.
  left; auto.
  unfold minR. 
  lr_if_goal. fold hlr. destruct hlr as [hl | hr].
  destruct hl as [hl | hl].
  specialize (ih _ h2r). 
  simpl in ih. 
  red in ih.
  destruct ih as [ih | ih].
  left. 
  red in h1.
  eapply h1.
  apply hl.
  pose proof (proof_irrelevance _ h3 (in_not_nil l n h2r)); subst; auto.
  left.
  pose proof (proof_irrelevance _ h3 (in_not_nil l n h2r)) as h4.
  rewrite <- h4 in ih.
  clear h4. subst; auto.
  subst.
  simpl in ih.
  specialize (ih _ h2r).
  pose proof (proof_irrelevance _ h3
                                (in_not_nil l n h2r)); 
  subst.
  red in ih; auto.
  red in ih. 
  specialize (ih _ h2r).
  pose proof (proof_irrelevance
                _ h3
                (in_not_nil l n h2r));
    subst; auto.
Qed.




Lemma minlR_functional :
  forall {T:Type} (R:Relation T) 
         (pft:tso_dec R)
         (l l':list T)
         (pfl:l <> nil) (pfl':l' <> nil),
    l = l' ->
    minlR R pft l  pfl =
    minlR R pft l' pfl'.
  intros T R h1 l l' h2 h3 h4; subst.
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
Qed.

Lemma minlR_functional' :
  forall {T:Type} (R R':Relation T) (pfso:sort_dec R) (pfso':sort_dec R')
         (pft:tso_dec R)
         (l l':list T)
         (pfl:l <> nil) (pfl':l' <> nil),
    R = R' ->
    l = l' ->
    minlR R pft  l  pfl =
    minlR R' (sort_dec_tso_dec _ pfso') l' pfl'.
  intros T R R' h1 h1' ht l l' h2 h2' h3 ?.
  subst. 
  pose proof (proof_irrelevance _ h2 h2'); subst.
  f_equal.
  symmetry.
  apply sort_dec_tso_dec_eq.
Qed.


Lemma minlR_eq_iff : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (m:T) (pfn: l <> nil),
    let pftso := sort_dec_tso_dec _ pfso in
    (minlR R pftso l pfn = m) <->
    (In m l /\ forall x:T, In x l -> Rw R m x).
  intros T R h1 l m h3 htso.
  split; intro h4; subst.
  split.
  apply in_minlR.
  intros x h4.
  erewrite minlR_functional.
  apply minlR_compat; auto.
  apply sort_dec_trans; auto.
  reflexivity.
  destruct h4 as [h4 h5].
  specialize (h5 (minlR R htso l h3) (in_minlR _ _ _ _)).
  pose proof (minlR_compat R htso (sort_dec_trans _ h1) l m h4) as h6.
  simpl in h6.
  pose proof (proof_irrelevance _ h3
                                (in_not_nil l m h4)); subst.
  symmetry.
  apply (antisym_Rw R (sort_dec_so _ h1) _ _ h5 h6).
  Grab Existential Variables.
  assumption.
Qed.


Lemma minlR_lrep :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (x:T) (n:nat),
      let pftso := sort_dec_tso_dec _ pfso in
      minlR R pftso (lrep x (S n)) (cons_neq_nil _ _)  = x.
  intros T R hso x n. revert x.
  induction n as [|n ih]; simpl.
  destruct (nil_dec' nil) as [|h3]; auto.
  contradict h3; auto.
  intro x. 
  destruct (nil_dec' (x::lrep x n)) as [h3 | h3]; auto.
  specialize (ih x).
  erewrite <- ih at 3.
  simpl.
  destruct (nil_dec' (lrep x n)) as [h4 | h4]; auto.
  apply minR_idem. 
  unfold  minR.
  destruct (sort_dec_tso_dec R hso x
                             (minlR R (sort_dec_tso_dec R hso) (lrep x n) h4)) as [h5 | h5].
  destruct (sort_dec_tso_dec R hso x x) as [h6 | h6]; auto. 
  destruct (sort_dec_tso_dec R hso x  (minlR R (sort_dec_tso_dec R hso) (lrep x n) h4)) as [h6 | h6]; auto.
  destruct h6 as [h6 | h6].
  pose proof (sort_dec_trans _ hso _ _ _ h5 h6) as h7.
  apply (sort_dec_irrefl _ hso) in h7.
  contradiction.
  assumption.
Qed.

Lemma minlR_sing :
  forall {T:Type} (R:Relation T) (pftso:sort_dec R)
         (l:list T) (x:T)
         (pfe:Singleton x = list_to_set l),
    let pft := sort_dec_tso_dec _ pftso in
    let pf := inh_sing x in
    let pfn := P_subst pf pfe in
    let pfn' := inh_not_nil _ pfn in
    x = minlR R pft l pfn'.
  intros T R h1 l x h2; simpl.
  pose proof h2 as h2'.
  symmetry in h2'.
  apply impl_all_sing in h2'.
  apply all_sing_lrep in h2'.
  symmetry. 
  gen0.
  rewrite h2'.
  intro h3.
  erewrite minlR_functional.
  erewrite <- minlR_lrep.
  apply minlR_functional.
  reflexivity.
  simpl. 
  rewrite lrep_nil_iff in h3.
  rewrite neq0_iff in h3.
  rewrite (S_pred _ _ h3).
  simpl.
  reflexivity.
  Grab Existential Variables.
  simpl.
  apply cons_neq_nil.
Qed.

Lemma minlR_list_singularize : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) 
         (pfn:l <> nil),
    let pftso := sort_dec_tso_dec _ pfso in
    let pfdt := sort_dec_eq_dec _ pfso in
    let pfn' := iffn2 (list_singularize_nil_iff pfdt l) pfn in
    minlR R pftso l pfn =
    minlR R pftso (list_singularize pfdt l) pfn'.
  intros T R h1 l h2 htso hdt hn.
  unfold htso.
  rewrite minlR_eq_iff.
  split.
  rewrite list_singularize_in_iff.
  apply in_minlR.
  intros x h3.
  rewrite (list_singularize_in_iff hdt) in h3.
  pose proof (minlR_compat R (sort_dec_tso_dec R h1) (sort_dec_trans _ h1) (list_singularize hdt l) x h3) as h4.
  simpl in h4.
  pose proof (proof_irrelevance _ hn
                                (in_not_nil _ _ h3)) as h5.
  rewrite h5; auto.
Qed.

Lemma minlR_not_lt:
  forall {T:Type} (R:Relation T) (pftso:sort_dec R)
         (l:list T) (pf:l <> nil)
         (x:T),
    In x l ->
    let pftso := sort_dec_tso_dec _ pftso in
    ~ R x (minlR R pftso l pf).
  intros T R h1 l.
  induction l as [|a l ih].
  intros; contradiction.
  intros h2 x h3 hts h4. 
  destruct h3 as [|h3]; subst.
  destruct (nil_dec' l) as [| h5]; subst.
  simpl in h4.
  lr_match h4; fold hlr in h4; destruct hlr as [hl | hr].
  apply (sort_dec_irrefl _ h1) in h4; auto.
  pose proof (hr eq_refl); auto.
  simpl in h4.
  lr_match h4; fold hlr in h4; destruct hlr as [hl | hr].
  apply (sort_dec_irrefl _ h1) in h4; auto.
  unfold minR in h4.
  lr_if h4; fold hlr in h4; destruct hlr as [hl' | hr'].
  apply (sort_dec_irrefl _ h1) in h4; auto.
  pose proof (sort_dec_trans _ h1 _ _ _ hr' h4) as h6.
  apply (sort_dec_irrefl _ h1) in h6; auto.
  simpl in h4.
  lr_match h4; fold hlr in h4; destruct hlr as [hl | hr].
  subst.
  contradict h3.
  specialize (ih hr x h3).
  simpl in ih.
  unfold hts in h4. 
  apply lt_minR in h4.
  destruct h4; contradiction.
  apply sort_dec_trans; auto.
Qed.

Lemma minlR_remove :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T) (pfn:l <> nil),
    In a l ->
    let pfdt := sort_dec_eq_dec R pfso in
    forall (pfn':remove pfdt a l <> nil),
      let pftso := sort_dec_tso_dec R pfso in
        minR R pftso a
             (minlR R pftso (remove pfdt a l) pfn') =
        minlR R (sort_dec_tso_dec R pfso) l pfn.
  intros T R h1 l a h2 hnin h3 h4 h5.
  unfold minR.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  destruct hl as [hl | hl].
  pose proof (minlR_compat R h5 (sort_dec_trans _ h1) (remove h3 a l)) as h6.
  simpl in h6.
  symmetry.
  rewrite minlR_eq_iff.
  split; auto.
  intros x h7.
  destruct (sort_dec_eq_dec _ h1 x a) as [| h8]; subst.
  right; auto.
  specialize (h6 x).
  hfirst h6.
  rewrite in_remove_iff.
  split; auto.
  specialize (h6 hf).
  left.
  eapply trans_R_Rw.
  apply sort_dec_trans; auto.
  apply hl.
  erewrite minlR_functional.
  apply h6; auto.
  reflexivity.
  pose proof (in_minlR R h5 (remove h3 a l) h4) as h6.
  rewrite <- hl in h6.
  apply remove_In in h6.
  contradiction.
  symmetry.
  rewrite minlR_eq_iff.
  split.
  pose proof (in_minlR R h5 (remove h3 a l) h4) as h6.
  rewrite in_remove_iff in h6.
  destruct h6; auto.
  intros x h6.
  pose proof (minlR_compat R h5 (sort_dec_trans _ h1) (remove h3  a l)) as h8.
  destruct (sort_dec_eq_dec _ h1 x a) as [|h10]; subst.
  left; auto.
  specialize (h8 x).
  hfirst h8.
  rewrite in_remove_iff.
  split; auto.
  specialize (h8 hf); simpl in h8.
  erewrite minlR_functional.
  apply h8; auto.
  reflexivity.
Qed.


  
Lemma minlR_list_to_set_eq : 
  forall {T:Type} 
         (R:Relation T) (pfso:sort_dec R) 
         (l l':list T) 
         (pf:l <> nil) (pf':l' <> nil),
    list_to_set l = list_to_set l' ->
    let pfdt := sort_dec_tso_dec _ pfso in
    let pftso := sort_dec_tso_dec _ pfso in
    minlR R pftso l  pf  =
    minlR R pftso l' pf'.
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros ? h3.
  contradict h3; auto.
  intros l' h3 h4 h5.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  subst.
  simpl in h5. rewrite add_empty_sing in h5.
  simpl in ih.
  symmetry in h5.
  apply impl_all_sing in h5.
  apply all_sing_lrep in h5.
  symmetry.
  gen0.
  rewrite h5.
  intro h6.
  pose proof h6 as h6'.
  rewrite lrep_nil_iff in h6'. 
  rewrite neq0_iff in h6'. 
  gen0.
  rewrite (S_pred _ _ h6').
  intro h7.
  erewrite (minlR_functional R (sort_dec_tso_dec _ h1) (lrep a (S (pred (length l')))) (lrep a (S (pred (length l'))))).
  rewrite minlR_lrep at 1; auto.
  reflexivity.
  destruct (in_dec (sort_dec_eq_dec _ h1) a l) as [h6 | h6].
  rewrite in_add_eq in h5.
  pose proof (f_equal (fun S => Subtract S a) h5) as h7.
  simpl in h7.
  specialize (ih l' hr h4 h5).
  simpl in ih.
  rewrite <- ih.
  unfold minR.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr'].
  destruct hl' as [h8|]; auto.
  apply minlR_not_lt in h8; auto.
  contradiction.
  reflexivity.
  rewrite <- list_to_set_in_iff; auto.
  pose proof (f_equal (fun S => Subtract S a) h5) as h7.
  simpl in h7.
  rewrite sub_add_same_nin in h7.
  rewrite (subtract_remove_compat (sort_dec_eq_dec _ h1)) in h7.
  destruct (nil_dec' (remove (sort_dec_eq_dec _ h1) a l')) as [h8 | h8].
  rewrite remove_eq_nil_iff in h8.
  destruct h8 as [|h8]; try contradiction.
  contradict hr.
  apply all_sing_lrep in h8.
  rewrite h8 in h7.
  rewrite remove_lrep in h7.
  simpl in h7.
  apply empty_set_nil in h7; auto.
  specialize (ih _ hr h8 h7).
  simpl in ih.
  rewrite ih.
  erewrite minlR_functional.
  apply minlR_remove.
  rewrite list_to_set_in_iff.
  rewrite <- h5.
  right; constructor; auto.
  reflexivity.
  rewrite <- list_to_set_in_iff; auto.
  Grab Existential Variables.
  assumption.
Qed.

Lemma minlR_incl :
  forall {T:Type} 
         (R:Relation T) (pfso:sort_dec R) 
         (l l':list T) 
         (pf:l <> nil) (pf':l' <> nil),
    incl l l' ->
    let pftso := sort_dec_tso_dec _ pfso in
    Rw R (minlR R pftso l' pf') (minlR R pftso l  pf).
  intros T R h1 l l' h2 h3 h4 h6.
  pose proof (in_minlR R h6 l h2) as h8.
  apply h4 in h8.
  pose proof (minlR_compat R (sort_dec_tso_dec _ h1)
                           (sort_dec_trans _ h1) l' _ h8) as h9. 
  simpl in h9.
  pose proof (proof_irrelevance _ h3
                                (in_not_nil l' _ h8)); subst.
  unfold h6. unfold h6 in h9.
  assumption.
Qed.

Lemma lt_minlR_remove :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (l:list T) (pfn:l <> nil),
    let pfdt := sort_dec_eq_dec _ pfso in
    let pftso := sort_dec_tso_dec _ pfso in
    let l' := remove pfdt (minlR R pftso l pfn) l in
    forall (pfn':l' <> nil), 
    R (minlR R pftso l pfn)
      (minlR R pftso l' pfn').
  intros T R h1 l h2 h3 h4 l' h5.
  pose proof (in_minlR R h4 l' h5) as h7.
  unfold l' in h7.
  apply incl_remove in h7.
  pose proof (minlR_compat R h4 (sort_dec_trans _ h1) l _ h7) as h8.
  simpl in h8.
  pose proof (proof_irrelevance _ h2
                                (in_not_nil _ _ h7)) as h9.
  rewrite h9.
  red in h8.
  destruct h8 as [| h8]; auto.
  rewrite <- h9 in h8.
  pose proof (in_minlR R h4 (remove h3 (minlR R h4 l h2) l) h5) as h10.
  rewrite <- h8 in h10.
  apply remove_In in h10.
  contradiction.
Qed.


Fixpoint le_listR {T:Type} (R:Relation T) (pf:tso_dec R)
         (l m:list T) : Prop :=
  match l with
    | nil => True
    | a::l' =>
      match m with
        | nil => False
        | b::m' =>
          match (pf a b) with
            | inleft (left _) => True
            | inleft (right _) => le_listR R pf l' m'
            | inright P => False
          end
      end
  end.



Definition lt_listR {T:Type} (R:Relation T) (pf:tso_dec R)
           (l m:list T) : Prop :=
  le_listR R pf l m /\ l <> m.



Lemma lt_list_impl_leR :
  forall {T:Type} (R:Relation T) (pf:tso_dec R) (l m:list T),
    lt_listR R pf l m -> le_listR R pf l m.
  intros T R h1 l m h2.
  red in h2.
  destruct h2; auto.
Qed.

Lemma not_lt_listR_nil :
  forall {T:Type} (R:Relation T) (pf:tso_dec R)  (l:list T),
    ~ lt_listR R pf l nil.
  intros T R h1 l h2.
  destruct h2 as [h2 h3].
  rewrite (hd_compat' _ h3) in h2.
  simpl in h2; auto.
Qed.

  
Lemma le_listR_nil :
  forall {T:Type} (R:Relation T) (pf:tso_dec R)  (l:list T),
    le_listR R pf nil  l.
  intros; simpl; auto.
Qed.


Lemma le_listR_dec :
  forall {T:Type} (R:Relation T) (pfdt:tso_dec R)
         (l m:list T), {le_listR R pfdt l m} + {~le_listR R pfdt l m}.
  intros T R h0 l.
  induction l as [|a l h1].
  intro m. left. simpl. auto.
  intros m. simpl.
  destruct m as [|b m]. right. intro; contradiction.
  destruct (h0 a b) as [h3 | h3].
  destruct h3 as [h3 | h3]. left. auto.
  subst.
  apply h1.
  right. intro; contradiction.
Qed.



Lemma refl_le_listR :
  forall {T:Type} (R:Relation T) (pft:tso_dec R) (pfi:irrefl R)
         (l:list T),
    le_listR R pft l l.
  intros T R ht hi l.
  induction l as [|a l h1]; simpl.
  auto. 
  destruct (ht a a) as [h2 | h2]. 
  destruct h2 as [h2 | h2]; auto.
  apply hi in h2; auto.
Qed.


Lemma antisym_le_listR :
  forall {T:Type} (R:Relation T) (pfsd:sort_dec R)
         (l m:list T),
    let pft := sort_dec_tso_dec _ pfsd in
    le_listR R pft l m ->
    le_listR R pft m l ->
    l = m.
  intros T R h0 l.
  destruct h0 as [hts htr hir].
  induction l as [|a l h1]; simpl. 
  intro m. destruct m; auto.
  intros ? h1. simpl in h1. contradiction.
  intros m h2.
  destruct m as [|b m]. contradiction.
  unfold sort_dec_tso_dec in h2.
  destruct (hts a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  intro h4. simpl in h4.
  destruct (hts b a) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  pose proof (htr  _ _ _ h3 h5) as h6.
  apply hir in h6; contradiction.
  apply hir in h3; contradiction.
  contradiction.
  subst.
  intro h3.
  simpl in h3.
  destruct (hts b b) as [h4 | h4].
  destruct h4 as [h4 | h4].
  apply hir in h4. contradiction.
  specialize (h1 _ h2 h3). subst.
  reflexivity.
  contradiction.
  contradiction.
Qed.
  

Lemma antisym_not_le_listR :
  forall {T:Type} (R:Relation T) (pfdt:sort_dec R)
         (l m:list T),
    let pft := sort_dec_tso_dec _ pfdt in
    ~ le_listR R pft l m ->
    ~ le_listR R pft m l ->
    l = m.
  intros T R h0 l.
  destruct h0 as [hts htr hir].
  induction l as [|a l h1]; simpl; try tauto.
  intros m h2 h3.
  destruct m as [|b m].
  contradict h3. simpl. auto.
  destruct (hts a b) as [h4 | h4].
  destruct h4 as [h4 | h4].
  contradict h2. auto. subst.
  simpl in h3.
  destruct (hts b b) as [h4 | h4].
  destruct h4 as [h4 | h4].
  apply hir in h4. contradiction.
  specialize (h1 _ h2 h3). subst.
  reflexivity.
  apply hir in h4.
  contradiction.
  contradict h3. simpl.
  destruct (hts b a) as [h3 | h3].
  destruct h3 as [h3 | h3]; auto.
  subst.
  apply hir in h4. contradiction.
  pose proof (htr _ _ _ h4 h3) as h5.
  apply hir in h5. contradiction.
Qed.


Lemma lt_listR_iff :
  forall {T:Type} (R:Relation T) (pf:sort_dec R) 
         (l m:list T),
    let pfts := sort_dec_tso_dec _ pf  in
      lt_listR R pfts l m <->
    ~ le_listR R pfts m l.
  intros T R h0 l m.
  destruct h0 as [hts htr hir].
  split; intro h1.
  red in h1.
  destruct h1 as [h1 h2].
  intro h3.
  pose proof (antisym_le_listR R (sort_dec_intro _ hts htr hir) _ _ h1 h3).
  subst.
  contradict h2; auto.
  red; split.
  destruct (le_listR_dec R hts l m) as [h2 | h2]; auto.
  pose proof (antisym_not_le_listR R (sort_dec_intro _ hts htr hir) _ _ h1 h2). subst.
  apply refl_le_listR; auto.
  intro. subst.
  contradict h1.
  apply refl_le_listR; auto.
Qed.

Lemma lt_listR_cons :
  forall {T:Type} (R:Relation T) (pf:sort_dec R) 
         (l m:list T) (a:T),
    let pfts := sort_dec_tso_dec _ pf  in
    lt_listR R pfts (a::l) (a::m) ->
    lt_listR R pfts l m.
  intros T R h1 l m a hts h2.
  red in h2.
  destruct h2 as [h2 h3]. 
  red in h2.
  destruct (hts a a) as [h4 | h4].
  destruct h4 as [h4 | h5].
  apply (sort_dec_irrefl _ h1)  in h4.
  contradiction.
  red. split; auto.
  contradict h3.
  subst; auto.
  contradiction.
Qed.

Lemma lt_listR_cons_iff :
  forall {T:Type} (R:Relation T) (pf:sort_dec R) 
         (l l':list T) (a:T) (pfn:l <> nil),
    let pfts := sort_dec_tso_dec _ pf  in
    lt_listR R pfts l (a::l') <->
    R (hd' l pfn) a \/
    (hd' l pfn = a /\ lt_listR R pfts (tl l) l'). 
  intros T R h1 l l' a h2; simpl; split; intro h3.
  rewrite (hd_compat' _ h2) in h3.
  red in h3.
  simpl in h3.
  destruct h3 as [h3 h4].
  destruct (sort_dec_tso_dec R h1 (hd' l h2) a) as [h5 | h5].
  destruct h5 as [h5 | h5].
  left; auto.
  subst.
  right. 
  split; auto.
  red.
  split; auto.
  intro; subst.
  contradict h4; auto.
  contradiction.

  destruct h3 as [h3 | h3].
  red.  split.
  rewrite (hd_compat' _ h2).
  simpl.
  destruct (sort_dec_tso_dec R h1 (hd' l h2) a) as [h5 | h5].
  destruct h5 as [h5 | h5]; auto.
  subst.
  apply (sort_dec_irrefl _ h1) in h3.
  contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h3 h5) as h6.
  apply (sort_dec_irrefl _ h1) in h6; auto.
  intro; subst.
  simpl in h3.
  apply (sort_dec_irrefl _ h1) in h3; auto.
  destruct h3 as [? h3]; subst.
  rewrite (hd_compat' _ h2) at 1; simpl.
  red.
  split.
  simpl.
  destruct (sort_dec_tso_dec R h1 (hd' l h2) (hd' l h2)) as [h4 | h4].
  destruct h4 as [h4 | h4]; auto.
  red in h3.
  destruct h3; auto.
  apply (sort_dec_irrefl _ h1) in h4; auto.
  red in h3.
  destruct h3 as [h3 h4].
  contradict h4.
  inversion h4; auto.
Qed.
  


Lemma le_listR_dec' :
  forall {T:Type} (R:Relation T) (pf:sort_dec R) (l m:list T),
    let pfts := sort_dec_tso_dec _ pf in
    {le_listR R pfts l m} + {lt_listR R pfts m l}.
  intros T R h0 l m.
  destruct h0 as [hts htr hir].
  destruct (le_listR_dec R hts l m) as [h1 | h1].
  left; auto.
  rewrite <- (lt_listR_iff R  (sort_dec_intro _ hts htr hir)) in h1 at 1.
  right; auto.
Qed.

Lemma tso_dec_lt_listR :
  forall {T:Type} (R:Relation T) (pf:sort_dec R),
    let pfts := sort_dec_tso_dec _ pf in
    tso_dec (lt_listR R pfts).
  intros T R h1.
  destruct h1 as [hts htr hir].
  red; intros l m.
  destruct (le_listR_dec' R (sort_dec_intro _ hts htr hir) l m) as [h1 | h1].
  destruct (le_listR_dec' R (sort_dec_intro _ hts htr hir) m l) as [h2 | h2].
  pose proof (antisym_le_listR R (sort_dec_intro _ hts htr hir) _ _ h1 h2). subst; auto.
  left; left; auto.
  right; auto.
Qed.

Lemma trans_le_listR :
  forall {T:Type} (R:Relation T) (pf:sort_dec R),
    let pfts := sort_dec_tso_dec _ pf in
    Transitive _ (le_listR R pfts).
  intros T R h0.
  destruct h0 as [hts htr hir].
  red.
  intro la.
  induction la as [|a la h1]; simpl.
  intros lb lc. destruct lb as [|b lb], lc as [|c lc]; auto.
  intros lb lc. destruct lb as [|b lb], lc as [|c lc]; auto.
  intro. contradiction. 
  destruct (hts a b) as [h2 | h2].
  destruct h2 as [h2 | h2].
  destruct (hts a c) as [h3 | h3].
  destruct h3 as [h3 | h3].
  auto.
  subst.
  intros ? h3.
  simpl in h3.
  destruct (hts b c)  as [h4 | h4].
  destruct h4 as [h4 | h4].
  pose proof (htr _ _ _ h4 h2) as ht.
  apply hir in ht; contradiction.
  subst.
  apply hir in h2; contradiction.
  contradiction.
  intros ? h4.
  simpl in h4.
  destruct (hts b c) as [h5 | h5]. destruct h5 as [h5 | h5].
  pose proof (htr _ _ _ h2 h5) as h6.
  pose proof (htr _ _ _ h6 h3) as h7.
  apply hir in h7; contradiction.
  subst.
  pose proof (htr _ _ _ h3 h2) as h5.
  apply hir in h5; auto.
  assumption.
  subst.
  intros h2 h3.
  simpl in h3.
  destruct (hts b c) as [h4 | h4].
  destruct h4 as [h4 | h4]. auto. subst.
  specialize (h1 _ _ h2 h3). assumption.
  contradiction.
  intro. contradiction.
Qed.


Lemma trans_le_lt_listR :
  forall {T:Type} (R:Relation T) (pf:sort_dec R) (la lb lc:list T),
    let pftso := sort_dec_tso_dec _ pf in
    le_listR R pftso la lb ->
    lt_listR R pftso lb lc ->
    lt_listR R pftso la lc.
  intros T R h1.
  destruct h1 as [hts htr hir].
  intros la lb lc h0 h1 h2.  
  red in h2. destruct h2 as [h2 h3].
  pose proof (trans_le_listR R (sort_dec_intro _ hts htr hir)  _ _ _ h1 h2) as h4.
  red. split; auto.
  intro h5. subst.
  pose proof (antisym_le_listR R (sort_dec_intro _ hts htr hir) _ _ h2 h1). subst.
  contradict h3; auto.
Qed.

Lemma trans_lt_le_listR :
  forall {T:Type} (R:Relation T) (pf:sort_dec R) (la lb lc:list T),
    let pftso := sort_dec_tso_dec _ pf in
    lt_listR R pftso la lb ->
    le_listR R pftso lb lc ->
    lt_listR R pftso la lc.
  intros T R h1.
  destruct h1 as [hts htr hir].
  intros la lb lc h0 h1 h2.
  red in h1. destruct h1 as [h1 h3].
  pose proof (trans_le_listR R (sort_dec_intro _ hts htr hir)  _ _ _ h1 h2) as h4.
  split; auto.
  intro h5. subst.
  pose proof (antisym_le_listR R (sort_dec_intro _ hts htr hir) _ _ h2 h1).
  subst.
  contradict h3; auto.
Qed.

Lemma trans_lt_listR :
  forall {T:Type} (R:Relation T) (pf:sort_dec R),
    let pftso := sort_dec_tso_dec _ pf in
    Transitive _ (lt_listR R pftso).
  intros T R h1.
  destruct h1 as [hts htr hir]. 
  red. intros la lb lc h1 h2.
  destruct h1 as [h1 h1'], h2 as [h2 h2'].
  split.
  apply (trans_le_listR R (sort_dec_intro _ hts htr hir) _ _ _ h1 h2).
  intro. subst.
  pose proof (antisym_le_listR R (sort_dec_intro _ hts htr hir) _ _ h2 h1).
  subst.
  contradict h1'; auto.
Qed.


Lemma irrefl_lt_listR :
  forall {T:Type} (R:Relation T) (pf:tso_dec R) (l:list T),
    ~ lt_listR R pf l l.
  intros T R h0 l h1.
  red in h1. destruct h1 as [? h1].
  contradict h1; auto.
Qed.


Lemma so_lt_listR :
  forall {T:Type} (R:Relation T) (pf:sort_dec R),
    let pftso := sort_dec_tso_dec _ pf in
  Strict_Order (lt_listR R pftso).
  intros T R h1.
  destruct h1 as [hts htr hir]. constructor.
  apply (trans_lt_listR R (sort_dec_intro _ hts htr hir)).
  red. apply irrefl_lt_listR.
Qed.


Lemma le_listR_lt_eq_dec :
  forall {T:Type} (R:Relation T) (pf:sort_dec R)
         (l m:list T),
    let pftso := sort_dec_tso_dec _ pf in
    le_listR R pftso l m ->
    {lt_listR R pftso l m} + {l = m}.
  intros T R h0.
  destruct h0 as [hts htr hir].
  intro l.
  induction l as [|a l h1]; destruct m as [|b m]; simpl.
  intros; right; auto.
  intros; left. red. split.
  apply le_listR_nil.
  intro h1. discriminate h1.
  intros; contradiction.
  intro h2.
  destruct (hts a b) as [h3 | h3].
  destruct h3 as [h3|]; subst.
  left.
  red. simpl.
  destruct (hts a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst. split; auto.
  intro h5.
  inversion h5; subst. clear h5.
  apply hir in h4; auto.
  apply hir in h3. contradiction.
  pose proof (htr _ _ _ h3 h4) as h5.
  apply hir in h5; contradiction.
  specialize (h1 _ h2).
  destruct h1 as [h1 | h1].
  left.  red. split.
  simpl.
  destruct (hts b b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  apply hir in h3; contradiction.
  assumption.
  apply hir in h3; auto.
  intro h3.
  inversion h3 as [h4]. clear h3. subst.
  apply irrefl_lt_listR in h1; auto. subst.
  right; auto.
  contradiction.
Qed.

 
Lemma lt_listR_tso_dec :
  forall {T:Type} (R:Relation T) (pf:sort_dec R)
         (l m:list T),
    let pftso := sort_dec_tso_dec _ pf in
    {lt_listR R pftso l m} +
    {l = m} + 
    {lt_listR R pftso m l}.
  intros T R h0.
  destruct h0 as [hts htr hir].
  intros l m.
  destruct (le_listR_dec' R (sort_dec_intro _ hts htr hir) l m) as [h0 | h0].
  destruct (le_listR_lt_eq_dec R (sort_dec_intro _ hts htr hir)  _ _ h0) as [h1 | h1]. 
  left; auto.
  subst.
  left. right; auto.
  right; auto.
Qed.


Lemma lt_listR_sort_dec :
  forall {T:Type} (R:Relation T) (pf:sort_dec R),
    let pftso := sort_dec_tso_dec _ pf in
    sort_dec (lt_listR R pftso).
  intros T R h0.
  destruct h0 as [hts htr hir].
  constructor.
  eapply tso_dec_lt_listR.
  apply trans_lt_listR.
  red; apply irrefl_lt_listR.
Defined.

Lemma lt_listR_sortable :
  forall {T:Type} (R:Relation T) (pf:sort_dec R),
    let pftso := sort_dec_tso_dec _ pf in
    sortable (lt_listR R pftso).
  intros T R h1; simpl.
  apply sort_dec_sortable.
  apply lt_listR_sort_dec.
Qed.

Lemma minlR_in :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (ll:faml T) (l:list T) (pfin:In l ll),
    let pft' := sort_dec_tso_dec _ pfso in
    let pftl := tso_dec_lt_listR _ pfso in
    minlR (lt_listR R pft') pftl ll (in_not_nil _ _ pfin) =
    minlR (lt_listR R pft') pftl (l::ll) (cons_neq_nil ll l).
  intros T R h1 ll l h2 h3 h4.
  pose proof (minlR_list_singularize (lt_listR R h3) (lt_listR_sort_dec _ _) ll (in_not_nil ll l h2)) as h5. 
  simpl in h5.
  unfold h4, h3.
  unfold h3 in h5.
  pose proof (minlR_list_singularize (lt_listR R h3) (lt_listR_sort_dec _ _) (l::ll) (cons_neq_nil ll l)) as h5'.
  simpl in h5'.
  simpl.
  lr_match_goal; fold hlr; fold hlr in h5'; destruct hlr as [hl | hr].
  clear h5'; subst; contradict h2.
  destruct h1 as [h1 h6 h7].
  simpl.
  simpl in h5.
  rewrite h5.
  erewrite sort_dec_tso_dec_eq in h5'.
  assert (h8:h1 = h3). unfold h3. symmetry. apply sort_dec_tso_dec_eq.
  rewrite h8.
  rewrite h5' at 1.
  simpl; auto.
  apply minlR_functional.
  rewrite list_singularize_in_cons; auto.
Qed.

Lemma maxlR_in :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (ll:faml T) (l:list T) (pfin:In l ll),
    let pft' := sort_dec_tso_dec _ pfso in
    let pftl := tso_dec_lt_listR _ pfso in
    maxlR (lt_listR R pft') pftl ll (in_not_nil _ _ pfin) =
    maxlR (lt_listR R pft') pftl (l::ll) (cons_neq_nil ll l).
  intros T R h1 ll l h2 h3 h4.
  pose proof (maxlR_list_singularize (lt_listR R h3) (lt_listR_sort_dec _ _) ll (in_not_nil ll l h2)) as h5. 
  simpl in h5.
  unfold h4, h3.
  unfold h3 in h5.
  pose proof (maxlR_list_singularize (lt_listR R h3) (lt_listR_sort_dec _ _) (l::ll) (cons_neq_nil ll l)) as h5'.
  simpl in h5'.
  simpl.
  lr_match_goal; fold hlr; fold hlr in h5'; destruct hlr as [hl | hr].
  clear h5'; subst; contradict h2.
  destruct h1 as [h1 h6 h7].
  simpl.
  simpl in h5.
  rewrite h5.
  erewrite sort_dec_tso_dec_eq in h5'.
  assert (h8:h1 = h3). unfold h3. symmetry. apply sort_dec_tso_dec_eq.
  rewrite h8.
  rewrite h5' at 1.
  apply maxlR_functional.
  rewrite list_singularize_in_cons; auto.
Qed.

(*upper bound filter for list*)
Fixpoint ub_list {T:Type} (R:Relation T) (pftso:tso_dec R)
      (l:list T) (b:T) : list T :=
  match l with
    | nil => nil
    | a :: l' => match (pftso a b) with
                   | inleft (left _)  => a::(ub_list R pftso l' b)
                   | inleft (right _) => ub_list R pftso l' b
                   | inright _ => ub_list R pftso l' b
                 end
  end.

(*weak upper bound filter for list*)
Fixpoint wub_list {T:Type} (R:Relation T) (pftso:tso_dec R)
      (l:list T) (b:T) : list T :=
  match l with
    | nil => nil
    | a :: l' => match (pftso a b) with
                   | inleft _ => a::(wub_list R pftso l' b)
                   | inright _ =>  wub_list R pftso l' b
                 end
  end.

(*lower bound filter for list.*)
Fixpoint lb_list {T:Type} (R:Relation T) (pftso:tso_dec R)
      (l:list T) (b:T) : list T :=
  match l with
    | nil => nil
    | a :: l' => match (pftso a b) with
                   | inleft _  => lb_list R pftso l' b
                   | inright _ => a::lb_list R pftso l' b
                 end
  end.


(*weak lower bound filter for list.*)
Fixpoint wlb_list {T:Type} (R:Relation T) (pftso:tso_dec R)
      (l:list T) (b:T) : list T :=
  match l with
    | nil => nil
    | a :: l' => match (pftso a b) with
                   | inleft (left _)  => wlb_list R pftso l' b
                   | inleft (right _) =>  a::(wlb_list R pftso l' b)
                   | inright _ =>  a::wlb_list R pftso l' b
                 end
  end.

Lemma ub_list_nil :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (b:T),
    ub_list R pftso nil b = nil.
  intros; simpl; auto.
Qed.
         
Lemma wub_list_nil :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (b:T),
    wub_list R pftso nil b  = nil.
  intros; simpl; auto.
Qed.
  
Lemma lb_list_nil :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (b:T),
    lb_list R pftso nil b  = nil.
  intros; simpl; auto.
Qed.

Lemma wlb_list_nil :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (b:T),
    wlb_list R pftso nil b  = nil.
  intros; simpl; auto.
Qed.



Lemma ub_list_cons_same :
forall {T:Type} (R:Relation T) (pfir:irrefl R) (pftso:tso_dec R)
       (l:list T) (b:T),
  ub_list R pftso (b::l) b =
  ub_list R pftso l b.
  intros T R hir h1 l b.
  simpl.
  destruct (h1 b b) as [h2 | h2].
  destruct h2 as [h2 | h2]; auto.
  apply hir in h2.
  contradiction.
  reflexivity.
Qed.

Lemma wub_list_cons_same :
forall {T:Type} (R:Relation T) (pfir:irrefl R) (pftso:tso_dec R)
       (l:list T) (b:T),
  wub_list R pftso (b::l) b =
  b :: wub_list R pftso l b.
  intros T R hir h1 l b; simpl.
  destruct (h1 b b) as [h2 | h2].
  destruct h2 as [h2 | h2].
  apply hir in h2; contradiction.
  reflexivity.
  apply hir in h2; contradiction.
Qed.

Lemma lb_list_cons_same :
forall {T:Type} (R:Relation T) (pfir:irrefl R) (pftso:tso_dec R)
       (l:list T) (b:T),
  lb_list R pftso (b::l) b =
  lb_list R pftso l b.
  intros T R hir h1 l b.
  simpl.
  destruct (h1 b b) as [h2 | h2].
  destruct h2 as [h2 | h2]; auto.
  apply hir in h2.
  contradiction.
Qed.

Lemma wlb_list_cons_same :
forall {T:Type} (R:Relation T) (pfir:irrefl R) (pftso:tso_dec R)
       (l:list T) (b:T),
  wlb_list R pftso (b::l) b =
  b :: wlb_list R pftso l b.
  intros T R hir h1 l b; simpl.
  destruct (h1 b b) as [h2 | h2].
  destruct h2 as [h2 | h2].
  apply hir in h2; contradiction.
  reflexivity.
  apply hir in h2; contradiction.
Qed.




Lemma ub_list_cons_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R x b ->
    let pftso := sort_dec_tso_dec _ pfso in
    ub_list R pftso (b::l) x =
    ub_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl.
  destruct (htso b x) as [h2 | h2]; subst; auto.
  destruct h2 as [h2|]; subst.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
  apply (sort_dec_irrefl _ hso) in h0; contradiction.
Qed.



Lemma wub_list_cons_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R x b ->
    let pftso := sort_dec_tso_dec _ pfso in
    wub_list R pftso (b::l) x =
    wub_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl.
  destruct (htso b x) as [h2 | h2]; subst; auto.
  destruct h2 as [h2|]; subst.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
  apply (sort_dec_irrefl _ hso) in h0; contradiction.
Qed.

Lemma lb_list_cons_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R x b ->
    let pftso := sort_dec_tso_dec _ pfso in
    lb_list R pftso (b::l) x =
    b :: lb_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl. 
  destruct (htso b x) as [h2 | h2]; subst; auto.
  destruct h2 as [h2|]; subst.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
  apply (sort_dec_irrefl _ hso) in h0; contradiction.
Qed.

Lemma wlb_list_cons_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R x b ->
    let pftso := sort_dec_tso_dec _ pfso in
    wlb_list R pftso (b::l) x =
    b :: wlb_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl. 
  destruct (htso b x) as [h2 | h2]; subst; auto.
  destruct h2 as [h2|]; subst.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
  apply (sort_dec_irrefl _ hso) in h0; contradiction.
Qed.

Lemma ub_list_cons_gt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R b x ->
    let pftso := sort_dec_tso_dec _ pfso in
    ub_list R pftso (b::l) x =
    b :: ub_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl. 
  destruct (htso b x) as [h2 | h2]; subst; auto.
  destruct h2 as [h2|]; subst; auto.
  apply (sort_dec_irrefl _ hso) in h0; contradiction.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
Qed.
  
Lemma wub_list_cons_gt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R b x ->
    let pftso := sort_dec_tso_dec _ pfso in
    wub_list R pftso (b::l) x =
    b :: wub_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl. 
  destruct (htso b x) as [h2 | h2]; subst; auto.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
Qed.

Lemma lb_list_cons_gt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R b x ->
    let pftso := sort_dec_tso_dec _ pfso in
    lb_list R pftso (b::l) x =
    lb_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl. 
  destruct (htso b x) as [h2 | h2]; subst; auto.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
Qed.


Lemma wlb_list_cons_gt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    R b x ->
    let pftso := sort_dec_tso_dec _ pfso in
    wlb_list R pftso (b::l) x =
    wlb_list R pftso l x.
  intros T R hso l b x h0 htso.
  simpl. 
  destruct (htso b x) as [h2 | h2]; subst; auto.
  destruct h2 as [h2|]; subst; auto.
  apply (sort_dec_irrefl _ hso) in h0; contradiction.
  pose proof (sort_dec_trans _ hso _ _ _ h0 h2) as h4.
  apply (sort_dec_irrefl _ hso) in h4; contradiction.
Qed.

Lemma ub_list_trivial :
   forall {T:Type} (R:Relation T) (pfso:sort_dec R)
          (l:list T) (b:T),
     (forall x:T, In x l -> R x b) -> 
    let pftso := sort_dec_tso_dec _ pfso in
    ub_list R pftso l b = l.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros b h2 htso.
  simpl.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  f_equal; apply ih.
  intros; apply h2; right; auto.
  subst.
  specialize (h2 b (in_eq _ _)).
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  specialize (h2 a (in_eq _ _)).
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
Qed.

Lemma wub_list_trivial :
   forall {T:Type} (R:Relation T) (pfso:sort_dec R)
          (l:list T) (b:T),
     (forall x:T, In x l -> Rw R x b) -> 
    let pftso := sort_dec_tso_dec _ pfso in
    wub_list R pftso l b = l.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros b h2 htso.
  simpl.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3]. 
  f_equal; apply ih.
  intros; apply h2; right; auto.
  subst.
  f_equal; apply ih.
  intros; apply h2; right; auto.
  specialize (h2 a (in_eq _ _)).
  destruct h2 as [h2|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  apply (sort_dec_irrefl _ h1) in h3; contradiction.
Qed.

  

Lemma lb_list_trivial :
   forall {T:Type} (R:Relation T) (pfso:sort_dec R)
          (l:list T) (b:T),
     (forall x:T, In x l -> R b x) -> 
    let pftso := sort_dec_tso_dec _ pfso in
    lb_list R pftso l b = l.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros b h2 htso.
  simpl.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  specialize (h2 a (in_eq _ _)).
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  subst.
  specialize (h2 b (in_eq _ _)).
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  f_equal; apply ih.
  intros; apply h2; right; auto.
Qed.

Lemma wlb_list_trivial :
   forall {T:Type} (R:Relation T) (pfso:sort_dec R)
          (l:list T) (b:T),
     (forall x:T, In x l -> Rw R b x) -> 
     let pftso := sort_dec_tso_dec _ pfso in
     wlb_list R pftso l b = l.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros b h2 htso.
  simpl.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  specialize (h2 a (in_eq _ _)).
  destruct h2 as [h2|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  apply (sort_dec_irrefl _ h1) in h3; contradiction.
  subst.
  f_equal; apply ih.
  intros; apply h2; right; auto.
  f_equal; apply ih.
  intros; apply h2; right; auto.
Qed.



Lemma ub_list_app :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l l':list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    ub_list R pftso (l ++ l') b =
    ub_list R pftso l b ++ ub_list R pftso l' b.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros l' b hts.
  simpl.
  destruct (hts a b) as [h2 | h2].
  destruct h2 as [h2|]; subst; auto.
  rewrite <- app_comm_cons.
  f_equal; auto.
  auto.
Qed.

Lemma wub_list_app :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l l':list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    wub_list R pftso (l ++ l') b =
    wub_list R pftso l b ++ wub_list R pftso l' b.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros l' b hts.
  simpl. 
  destruct (hts a b) as [h2 | h2].
  rewrite <- app_comm_cons.
  f_equal; auto.
  auto.
Qed.

Lemma lb_list_app :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l l':list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    lb_list R pftso (l ++ l') b =
    lb_list R pftso l b ++ lb_list R pftso l' b.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros l' b hts.
  simpl.
  destruct (hts a b) as [h2 | h2]; auto.
  rewrite <- app_comm_cons.
  f_equal; auto.
Qed.

Lemma wlb_list_app :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l l':list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    wlb_list R pftso (l ++ l') b =
    wlb_list R pftso l b ++ wlb_list R pftso l' b.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros l' b hts.
  simpl.
  destruct (hts a b) as [h2 | h2].
  destruct h2 as [h2|]; subst; auto.
  rewrite <- app_comm_cons.
  f_equal; auto.
  rewrite <- app_comm_cons.
  f_equal; auto.
Qed.

        
  
Lemma in_ub_list_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (ub_list R pftso l b) <->
    In x l /\ R x b.
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros; tauto.
  intros b x.
  destruct (sort_dec_tso_dec R h1 a b) as [h2 | h2].
  destruct h2 as [h2 | h2].
  split; intro h3. 
  split.
  destruct h3 as [|h3].
  left; auto.
  right.
  rewrite ih in h3.
  destruct h3; auto.
  destruct h3 as [|h3]; subst; auto.
  rewrite ih in h3.
  destruct h3; auto.
  destruct h3 as [h3 h4].
  destruct h3 as [|h3]; subst.
  left; auto.
  pose proof (conj h3 h4) as h5.
  rewrite <- ih in h5.
  right; auto.
  subst.
  split; intro h2.
  rewrite ih in h2.
  destruct h2 as [h2 h3].
  split; auto.
  destruct h2 as [h2 h3].
  destruct h2 as [|h2]; subst.
  apply (sort_dec_irrefl _ h1) in h3; contradiction.
  rewrite ih.
  split; auto.
  rewrite ih.
  split; intro h3.
  destruct h3 as [h3 h4].
  split; auto.
  destruct h3 as [h3 h4].
  destruct h3 as [|h3]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h4) as h5.
  apply (sort_dec_irrefl _ h1) in h5.
  contradiction.
  split; auto.
Qed.

Lemma in_wub_list_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (wub_list R pftso l b) <->
    In x l /\ Rw R x b.
  intros T R h1 l.
  induction l as [|a l ih]; intros b x h2; split; intro h3. 
  simpl in h3; contradiction.
  destruct h3; contradiction. 
  fold h2 in ih.
  simpl in h3; simpl.
  destruct (h2 a b) as [h4 | h4].
  destruct h4 as [h4|]; subst.
  destruct h3 as [|h3]; subst.
  split; left; auto.
  rewrite ih in h3.
  destruct h3 as [h3 h5].
  split; [right|]; auto.
  destruct h3 as [|h3]; subst.
  split; [left | right]; auto.
  rewrite ih in h3.
  tauto.
  rewrite ih in h3.
  tauto.
  fold h2 in ih.
  destruct h3 as [h3 h4].
  simpl.
  destruct h3 as [|h3]; subst.
  destruct (h2 x b) as [h5 | h5].
  left; auto.
  pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ h1) in h6.
  contradiction.
  destruct (h2 a b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  pose proof (conj h3 h4) as h6.
  rewrite <- ih in h6.
  right; auto.
  pose proof (conj h3 h4) as h5.
  rewrite <- ih in h5.
  right; auto.
  pose proof (conj h3 h4) as h6.
  rewrite <- ih in h6; auto.
Qed.

Lemma in_lb_list_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (lb_list R pftso l b) <->
    In x l /\ R b x.
  intros T R h1 l; induction l as [|a l ih]; intros b x h2; split; intro h3.

  simpl in h3; contradiction.
  simpl in h3. destruct h3; contradiction.
  fold h2 in ih.
  simpl in h3.
  destruct (h2 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst.
  rewrite ih in h3.
  destruct h3.
  split; try right; auto.
  rewrite ih in h3.
  destruct h3 as [h3 h4].
  split; try right; auto.
  destruct h3 as [|h3]; subst.
  split; try left; auto.
  rewrite ih in h3.
  destruct h3; split; try right; auto.
  simpl.
  destruct h3 as [h3 h4].
  destruct h3 as [|h3]; destruct (h2 x b) as [h5 | h5]; simpl; try destruct h5 as [h5 | h5]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ h1) in h6; contradiction.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  destruct (h2 x b) as [h6 | h6].
  destruct h6 as [h6 |]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h5 h6) as h7.
  apply (sort_dec_irrefl _ h1) in h7; contradiction.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
  left; auto.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  apply (sort_dec_irrefl _ h1) in h7; contradiction.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  pose proof (conj h3 h4) as h6.
  fold h2 in ih.
  rewrite <- ih in h6. 
  destruct (h2 a b); try right; auto.
Qed.


Lemma in_wlb_list_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (b x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (wlb_list R pftso l b) <->
    In x l /\ Rw R b x.
  intros T R h1 l.
  induction l as [|a l ih]; intros b x h2; split; intro h3; simpl in h3; try destruct h3; try contradiction.
  fold h2 in ih.
  destruct (h2 a b) as [h4 | h4].
  destruct h4 as [|h4]; subst.
  rewrite ih in h3.
  destruct h3; split; auto; try right; auto. 
  destruct h3 as [|h3]; subst; split;
  [left; auto | right; auto | idtac | idtac].
  rewrite ih in h3.
  destruct h3; right; auto.
  rewrite ih in h3. 
  destruct h3; auto.
  destruct h3 as [|h3]; subst.
  split; left; auto.
  rewrite ih in h3. 
  destruct h3 as [h3 h5]; split; [right | idtac]; auto.
  destruct H as [|H]; subst.
  simpl.  
  destruct (h2 x b) as [h3 | h3]. 
  destruct h3 as [h3 | h3].
  pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ H0 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  subst.
  left; auto.
  left; auto.
  simpl.
  pose proof (conj H H0) as h7.
  rewrite <- ih in h7.
  fold h2 in h7.
  destruct (h2 a b) as [h3 | h3]. 
  destruct h3 as [|h3]; auto; subst.
  right; auto.
  right; auto.
Qed.


Lemma ub_list_eq_nil_iff :
  forall {T:Type} (R:Relation T) (pftso:sort_dec R)
         (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pftso in
    ub_list R pftso l b = nil <->
    l = nil \/ (l <> nil /\ (forall x:T, In x l -> Rw R b x)).
  intros T R h1 l.
  destruct l as [|a l]; simpl.
  tauto.
  intro b; split; intro h2. 
  destruct (sort_dec_tso_dec R h1 a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  discriminate.
  subst.
  right.
  split.
  intro; discriminate.
  intros x h3.
  destruct h3 as [|h3]; subst.
  right; auto.
  destruct (sort_dec_tso_dec R h1 b x) as [h4 | h4].
  destruct h4 as [h4 | h4].
  left; auto.
  subst.
  right; auto.
  pose proof (conj h3 h4) as h5.
  rewrite <- (in_ub_list_iff R h1) in h5.
  rewrite h2 in h5.
  contradict h5.
  right.
  split. intro; discriminate.
  intros x h4.
  destruct h4 as [|h4]; subst.
  left; auto.
  destruct (sort_dec_tso_dec R h1 b x) as [h5 | h5].
  destruct h5 as [h5 | h5].
  left; auto.
  subst.
  right; auto.
  pose proof (conj h4 h5) as h6.
  rewrite <- (in_ub_list_iff R h1) in h6.
  rewrite h2 in h6.
  contradict h6.
  destruct h2 as [| h2]; try discriminate.
  destruct (sort_dec_tso_dec R h1 a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  destruct h2 as [hn h2].
  specialize (h2  a (or_introl eq_refl)).
  pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  subst.
  revert h2. revert b.
  induction l as [|a l ih]; simpl.
  intros; auto.
  intros b h2.
  destruct (sort_dec_tso_dec R h1 a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  destruct h2 as [hn h2].
  specialize (h2 a (or_intror (or_introl eq_refl))).
  pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  subst.
  destruct h2 as [hn h2].
  apply ih; split; auto. intro; discriminate.
  apply ih.
  split. intro; discriminate.
  intros x h4.
  destruct h4 as [|h4]; subst.
  right; auto.
  apply h2.
  right. right; auto.
  revert h3 h2.
  induction l as [|c l ih]; simpl.
  intros; auto.
  intros h2 h3.
  destruct (sort_dec_tso_dec R h1 c b) as [h4 | h4].
  destruct h4 as [h4 | h4].
  destruct h3 as [hn h3].
  specialize (h3 c (or_intror (or_introl eq_refl))).
  pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h3 h4) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
  subst.
  apply ih; auto.
  split. intro; discriminate.
  intros x h4.
  apply h3.
  destruct h4.
  left; auto.
  right; right; auto.
  apply ih; auto.
  split. intro; discriminate.
  intros x h5.
  apply h3.
  destruct h5.
  left; auto.
  right; right; auto.
Qed.


Lemma wub_list_eq_nil_iff :
  forall {T:Type} (R:Relation T) (pftso:sort_dec R)
         (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pftso in
    wub_list R pftso l b = nil <->
    l = nil \/ (l <> nil /\ (forall x:T, In x l -> R b x)).
  intros T R h1 l; induction l as [|a l ih]; intros b h2; split; intro h3;  simpl; try fold h2 in ih.   
  left; auto.
  reflexivity.
  right. 
  simpl in h3. 
  destruct (h2 a b) as [h4 | h4].
  discriminate.
  rewrite ih in h3.
  destruct h3 as [|h3]; subst.
  split. intro; discriminate.
  intros x h5. 
  destruct h5 as [|h5]; subst; auto.
  contradict h5.
  split. intro; discriminate.
  intros x h5. 
  destruct h5 as [|h5]; subst; auto.
  destruct h3 as [hn h3].
  apply h3; auto.
  destruct h3 as [h3 | h3]. discriminate.
  destruct h3 as [hn h3].
  pose proof (h3 _ (in_eq _ _)) as h3'.
  assert (hnin : ~In b l).
    intro h4.
    specialize (h3 _ (in_cons _ _ _ h4)).
    apply (sort_dec_irrefl _ h1) in h3.
    contradiction.
  destruct (h2 a b) as [h4 | h4].
  destruct h4 as [h4|]; subst.
  specialize (h3 _ (in_eq _ _)).
  pose proof ((sort_dec_trans _ h1) _ _ _ h3 h4) as h5.
  apply (sort_dec_irrefl _ h1) in h5.
  contradiction.
  apply (sort_dec_irrefl _ h1) in h3'.
  contradiction.
  destruct (nil_dec' l) as [|h5]; subst; auto.
  rewrite ih.
  right; split; auto.
  intros; apply h3; right; auto.
Qed.

Lemma lb_list_eq_nil_iff :
   forall {T:Type} (R:Relation T) (pftso:sort_dec R)
         (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pftso in
    lb_list R pftso l b = nil <->
    l = nil \/ (l <> nil /\ (forall x:T, In x l -> Rw R x b)).
  intros T R h1 l; induction l as [|a l ih]; intros b h2; split; intro h3; simpl; try fold h2 in ih; auto.
  simpl in h3.
  destruct (h2 a b) as [h4 | h4].
  destruct h4 as [h4|]; subst.  
  right.
  split. intro; discriminate.
  intros x h5.
  destruct h5 as [|h5]; subst.
  left; auto.
  rewrite ih in h3.
  destruct h3 as [|h3]; subst.
  contradict h5.
  apply h3; auto.
  right.
  split. intro; discriminate.
  intros x h4.
  destruct h4 as [|h4]; subst.
  right; auto.
  rewrite ih in h3.
  destruct h3 as [|h3]; subst.
  contradict h4.
  apply h3; auto.
  discriminate.
  destruct h3 as [|h3].
  discriminate.
  destruct (h2 a b) as [h4 | h4].
  destruct h4 as [h4|]; subst.
  destruct (nil_dec' l) as [|h5]; subst; auto. 
  rewrite ih.
  right. 
  split; intros; try apply h3; try right; auto.
  destruct (nil_dec' l) as [|h5]; subst; auto. 
  rewrite ih.
  right; intros; try apply h3; try right; auto.
  split; auto.
  destruct h3 as [hn h3].
  intros; apply h3; auto. right; auto.
  destruct h3 as [hn h3].
  specialize (h3 _ (in_eq _ _)).
  pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h3 h4) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
Qed.


Lemma wlb_list_eq_nil_iff :
  forall {T:Type} (R:Relation T) (pftso:sort_dec R)
         (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pftso in
    wlb_list R pftso l b = nil <->
    l = nil \/ (l <> nil /\ (forall x:T, In x l -> R x b)).
  intros T R h1 l; induction l as [|a l ih]; intros b h2; split; intro h3; try fold h2 in ih; auto.
  simpl in h3.
  destruct (h2 a b) as [h4 | h4]; try destruct h4 as [h4|]; subst.
  rewrite ih in h3.
  right.
  destruct h3 as [|h3]; subst.
  split. intros; discriminate.
  intros x h5.
  destruct h5 as [|h5]; subst; auto.
  contradict h5.
  split. intro; discriminate.
  intros x h5.
  destruct h3 as [hn h3].
  destruct h5 as [|h5]; subst; auto.
  discriminate.
  discriminate.
  destruct h3 as [|h3]. discriminate.
  simpl.
  destruct (h2 a b) as [h4 | h4].
  destruct h4 as [h4|]; subst.
  destruct (nil_dec' l) as [|hnin]; subst; auto.
  rewrite ih.
  right.
  split; intros; try apply h3; auto.
  right; auto.
  destruct h3 as [hn h3].
  specialize (h3 _ (in_eq _ _)).
  apply (sort_dec_irrefl _ h1) in h3.
  contradiction.
  destruct h3 as [hn h3].
  specialize (h3 _ (in_eq _ _)).
  pose proof (sort_dec_trans _ h1 _ _ _ h3 h4) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
Qed.



Lemma ub_list_lt_length :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    length (ub_list R pftso l b) < S (length l).
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros; auto with arith.
  intro b.
  destruct (h1 a b) as [h2 | h2]; simpl.
  destruct h2 as [h2 |]; subst; simpl.
  apply lt_n_S.
  apply ih.
  eapply lt_trans.
  apply ih.
  auto with arith.
  eapply lt_trans.
  apply ih.
  auto with arith.
Qed.


Lemma wub_list_lt_length :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    length (wub_list R pftso l b) < S (length l).
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros; auto with arith.
  intro b.
  destruct (h1 a b) as [h2 | h2]; simpl.
  destruct h2 as [h2 |]; subst; simpl. 
  apply lt_n_S.
  apply ih.
  apply lt_n_S.
  apply ih.
  eapply lt_trans.
  apply ih.
  auto with arith.
Qed.


Lemma lb_list_lt_length :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    length (lb_list R pftso l b) < S (length l).
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros; auto with arith.
  intro b.
  destruct (h1 a b) as [h2 | h2]; simpl.
  destruct h2 as [h2 |]; subst; simpl.
  eapply lt_trans.
  apply ih.
  auto with arith.
  eapply lt_trans.
  apply ih.
  auto with arith.
  apply lt_n_S.
  apply ih.
Qed.


Lemma wlb_list_lt_length :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    length (wlb_list R pftso l b) < S (length l).
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros; auto with arith.
  intro b.
  destruct (h1 a b) as [h2 | h2]; simpl.
  destruct h2 as [h2 |]; subst; simpl.
  eapply lt_trans.
  apply ih.
  auto with arith.
  apply lt_n_S.
  apply ih.
  apply lt_n_S.
  apply ih.
Qed.

Lemma ub_list_le :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in 
    length (ub_list R pftso l a) <=
    length (ub_list R pftso l b).
  intros T R hs l.
  induction l as [|c l ih]; auto with arith.
  intros a b h2 ht.
  simpl.
  destruct (ht c a) as [h3 | h3]; simpl.
  destruct h3 as [h3 | h3]; simpl.
  destruct (ht c b) as [h4 | h4]; simpl.
  destruct h4 as [h4 | h5]; simpl.
  simpl.
  apply le_n_S.
  apply ih; auto.
  subst.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ hs) in h4.
  contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h4) as h5.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6.
  contradiction.
  subst.
  destruct (ht a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  simpl.
  eapply le_trans.
  apply ih.
  apply h3.
  fold ht.
  auto with arith.
  apply (sort_dec_irrefl _ hs) in h2.
  contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ hs) in h4.
  contradiction.
  destruct (ht c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  simpl.
  eapply le_trans.
  apply ih. apply h2.
  fold ht.
  auto with arith.
  apply ih; auto.
  apply ih; auto.
Qed.


Lemma wub_list_le :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in 
    length (wub_list R pftso l a) <=
    length (wub_list R pftso l b).
  intros T R hs l.
  induction l as [|c l ih]; auto with arith.
  intros a b h2 ht.
  simpl.
  destruct (ht c a) as [h3 | h3]; simpl.
  destruct h3 as [h3 | h3]; simpl.
  destruct (ht c b) as [h4 | h4]; simpl.
  destruct h4 as [h4 | h5]; simpl.
  simpl.
  apply le_n_S.
  apply ih; auto.
  subst.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ hs) in h4.
  contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h4) as h5.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6.
  contradiction.
  subst.
  destruct (ht a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  simpl. 
  auto with arith.
  simpl.
  auto with arith.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ hs) in h4.
  contradiction.
  destruct (ht c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  simpl.
  eapply le_trans.
  apply ih. apply h2.
  fold ht.
  auto with arith.
  simpl. auto with arith.
  apply ih; auto.
Qed.


Lemma lb_list_le :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in 
    length (lb_list R pftso l b) <=
    length (lb_list R pftso l a).
  intros T R hs l.
  induction l as [|c l ih]; auto with arith.
  intros a b h2 ht.
  simpl.  
  destruct (ht c a) as [h3 | h3]; simpl.
  destruct h3 as [h3 | h3]; simpl.
  destruct (ht c b) as [h4 | h4]; simpl.
  destruct h4 as [h4 | h5]; simpl.
  simpl. 
  apply ih; auto.
  subst.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h5.
  apply (sort_dec_irrefl _ hs) in h5.
  contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h4) as h5.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6.
  contradiction.
  subst.
  destruct (ht a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  simpl. 
  apply ih; auto.
  apply (sort_dec_irrefl _ hs) in h2.
  contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ hs) in h4.
  contradiction. 
  destruct (ht c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  simpl.
  eapply le_trans.
  apply ih. apply h2.
  fold ht.
  auto with arith.
  eapply le_trans.
  apply ih. apply h3.
  fold ht. auto with arith.
  simpl.
  apply le_n_S.
  apply ih; auto.
Qed.


Lemma wlb_list_le :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in 
    length (wlb_list R pftso l b) <=
    length (wlb_list R pftso l a).
  intros T R hs l.
  induction l as [|c l ih]; auto with arith.
  intros a b h2 ht.
  simpl.  
  destruct (ht c a) as [h3 | h3]; simpl.
  destruct h3 as [h3 | h3]; simpl.
  destruct (ht c b) as [h4 | h4]; simpl.
  destruct h4 as [h4 | h5]; simpl.
  simpl. 
  apply ih; auto.
  subst.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h5.
  apply (sort_dec_irrefl _ hs) in h5.
  contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h4) as h5.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6.
  contradiction.
  subst.
  destruct (ht a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  simpl.
  eapply le_trans.
  apply ih. apply h3.
  fold ht. auto with arith.
  apply (sort_dec_irrefl _ hs) in h2.
  contradiction.
  simpl.
  apply le_n_S.
  apply ih.
  assumption.
  destruct (ht c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst. 
  auto with arith.
  simpl.
  apply le_n_S.
  apply ih.
  assumption.
  simpl.
  apply le_n_S.
  apply ih; auto.
Qed.


Lemma coarse_list_ub_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    coarse_list (ub_list R pftso l b) l.
  intros T R h1 l.
  induction l as [|a l h2]; simpl.
  intros; constructor.
  intro b.
  destruct (h1 a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  apply coarse_list_cons2.
  apply h2.
  apply coarse_list_cons1.
  apply h2.
  apply coarse_list_cons1.
  apply h2.
Qed.


Lemma no_dup_ub_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (a:T),
    NoDup l ->
    NoDup (ub_list R pftso l a).
  intros T R h1 l a h2.
  eapply no_dup_coarse_list.
  apply coarse_list_ub_list.
  assumption.
Qed.


Lemma coarse_list_ub_list_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in
    coarse_list (ub_list R pftso l a) (ub_list R pftso l b).
  intros T R hs l. 
  induction l as [|c l ih].
  intros; constructor.
  intros a b h2 htso.
  simpl.
  destruct (htso c a) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h2) as h4.
  destruct (htso c b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  apply coarse_list_cons2; auto.
  apply (sort_dec_irrefl _ hs) in h4; contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  apply coarse_list_cons1; auto.
  subst.
  apply (sort_dec_irrefl _ hs) in h2; contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  apply coarse_list_cons1; auto.
  auto.
  auto.
Qed.

Lemma coarse_list_wub_list_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in
    coarse_list (wub_list R pftso l a) (wub_list R pftso l b).
  intros T R hs l. 
  induction l as [|c l ih].
  intros; constructor.
  intros a b h2 htso.
  simpl.
  destruct (htso c a) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h2) as h4.
  destruct (htso c b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  apply coarse_list_cons2; auto.
  apply (sort_dec_irrefl _ hs) in h4; contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3]. 
  apply coarse_list_cons2; auto.
  subst.
  apply (sort_dec_irrefl _ hs) in h2; contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  apply coarse_list_cons1; auto.
  apply coarse_list_cons1; auto.
  auto.
Qed.


Lemma coarse_list_lb_list_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in
    coarse_list (lb_list R pftso l b) (lb_list R pftso l a).
  intros T R hs l. 
  induction l as [|c l ih].
  intros; constructor.
  intros a b h2 htso.
  simpl.
  destruct (htso c a) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h2) as h4.
  destruct (htso c b) as [h5 | h5].
  destruct h5 as [h5|]; subst; auto.
  pose proof (sort_dec_trans _ hs _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; auto.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  apply coarse_list_cons1; auto.
  apply coarse_list_cons1; auto.
  apply coarse_list_cons2; auto.
Qed.



Lemma coarse_list_wlb_list_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (l:list T) (a b:T),
    R a b ->
    let pftso := sort_dec_tso_dec _ pfso in
    coarse_list (wlb_list R pftso l b) (wlb_list R pftso l a).
intros T R hs l. 
  induction l as [|c l ih].
  intros; constructor.
  intros a b h2 htso.
  simpl.
  destruct (htso c a) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  pose proof (sort_dec_trans _ hs _ _ _ h3 h2) as h4.
  destruct (htso c b) as [h5 | h5].
  destruct h5 as [h5|]; subst; auto.
  apply (sort_dec_irrefl _ hs) in h4; contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; auto.
  apply coarse_list_cons1; auto.
  subst; auto.
  apply (sort_dec_irrefl _ hs) in h2; contradiction.
  pose proof (sort_dec_trans _ hs _ _ _ h2 h3) as h6.
  apply (sort_dec_irrefl _ hs) in h6; contradiction.
  destruct (htso c b) as [h4 | h4].
  destruct h4 as [h4 | h4]; subst.
  apply coarse_list_cons1; auto. 
  apply coarse_list_cons2; auto.
  apply coarse_list_cons2; auto.
Qed.


Lemma in_ub_wlb_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (l:list T) (a x y:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (ub_list R pftso l a) ->
    In y (wlb_list R pftso l a) ->
    R x y.
  intros T R h1 l a x y h2 h3 h4.
  unfold h2 in h3, h4.
  rewrite in_ub_list_iff in h3; destruct h3 as [h3 h5].
  rewrite in_wlb_list_iff in h4; destruct h4 as [h4 h6].
  eapply trans_R_Rw.
  eapply sort_dec_trans.
  assumption.
  apply h5.
  assumption.
Qed.

Lemma in_lb_wub_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (l:list T) (a x y:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (wub_list R pftso l a) ->
    In y (lb_list R pftso l a) ->
    R x y.
  intros T R h1 l a x y h2 h3 h4.
  unfold h2 in h3, h4.
  rewrite in_wub_list_iff in h3; destruct h3 as [h3 h5].
  rewrite in_lb_list_iff in h4; destruct h4 as [h4 h6].
  eapply trans_Rw_R.
  eapply sort_dec_trans.
  assumption.
  apply h5.
  assumption.
Qed.






Lemma ub_list_nest :
  forall {T:Type} (R:Relation T) (pfs:sort_dec R) (l:list T) (a b:T),
    let pftso := sort_dec_tso_dec R pfs in
    ub_list R pftso (ub_list R pftso l a) b =
    ub_list R pftso l (minR R pftso a b).
  intros T R h1 l a b h3.
  unfold minR. 
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst.
  revert h4. revert a b.
  induction l as [|c l ih]; simpl; auto. 
  intros a b h2.
  destruct (h3 c a) as [h4 | h4].
  destruct h4 as [h4|]; subst; simpl.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h2) as h5.
  destruct (h3 c b) as [h6 | h6].
  destruct h6 as [h6 | h6].
  f_equal; apply ih; auto.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h5); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h6 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  auto.
  auto.
  
  induction l as [|a l ih]; simpl; auto.
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 | h4]; simpl.
  destruct (h3 a b) as [h5 | h5].
  destruct h5 as [h5 | h5].
  f_equal; auto.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  subst; auto.
  apply ih.
  
  induction l as [|c l ih]; simpl; auto. 
  destruct (h3 c a) as [h5 | h5].
  destruct h5 as [h5|]; subst; auto.
  simpl.
  destruct (h3 c b) as [h5' | h5'].
  destruct h5' as [h5' | h5'].
  f_equal; auto. 
  subst; auto.
  auto.
  destruct (h3 a b) as [h5|h5]; subst.
  destruct h5 as [h5 | h5].
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  subst; auto.
  auto.
  destruct (h3 c b) as [h6 | h6].
  destruct h6 as [h6 | h6].
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_trans _ h1 _ _ _ h7 h6) as h8.
  pose proof (sort_dec_irrefl _ h1  _  h8); contradiction.
  subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h8.
  pose proof (sort_dec_irrefl _ h1  _  h8); contradiction.
  auto.
Qed.


Lemma wub_list_nest :
  forall {T:Type} (R:Relation T) (pfs:sort_dec R) (l:list T) (a b:T),
    let pftso := sort_dec_tso_dec R pfs in
    wub_list R pftso (wub_list R pftso l a) b =
    wub_list R pftso l (minR R pftso a b).
  intros T R h1 l a b h3.
  unfold minR. 
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst.
  revert h4. revert a b.
  induction l as [|c l ih]; simpl; auto. 
  intros a b h2.
  destruct (h3 c a) as [h4 | h4].
  destruct h4 as [h4|]; subst; simpl.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h2) as h5.
  destruct (h3 c b) as [h6 | h6].
  destruct h6 as [h6 | h6].
  f_equal; apply ih; auto.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h5); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h6 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst.
  f_equal; auto.
  pose proof (sort_dec_irrefl _ h1  _  h2); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h4) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  auto.
  
  induction l as [|a l ih]; simpl; auto.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  destruct hl as [hl | hl]; simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr'].
  f_equal; auto.
  pose proof (sort_dec_trans _ h1 _ _ _ hl hr') as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  subst.
  lr_if_goal; fold hlr; destruct hlr as [hl' | hr'].
  f_equal; auto.
  pose proof (sort_dec_irrefl _ h1  _  hr'); contradiction.
  auto.
  
  induction l as [|c l ih]; simpl; auto. 
  destruct (h3 c a) as [h5 | h5].
  destruct h5 as [h5|]; subst; auto.
  simpl.
  destruct (h3 c b) as [h5' | h5'].
  f_equal; auto.
  auto.
  destruct (h3 a b) as [h5|h5]; subst.
  destruct h5 as [h5 | h5].
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  simpl.
  destruct (h3 a b) as [h6 | h6].
  destruct h6 as [h6 | h6].
  pose proof (sort_dec_trans _ h1 _ _ _ h5 h6) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h5); contradiction.
  auto.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h8.
  destruct (h3 c b) as [h9 | h9].
  destruct h9 as [h9|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h8 h9) as h10.
  pose proof (sort_dec_irrefl _ h1  _  h10); contradiction.
  pose proof (sort_dec_irrefl _ h1  _  h8); contradiction.
  assumption.
Qed.


Lemma lb_list_nest :
  forall {T:Type} (R:Relation T) (pfs:sort_dec R) (l:list T) (a b:T),
    let pftso := sort_dec_tso_dec R pfs in
    lb_list R pftso (lb_list R pftso l a) b =
    lb_list R pftso l (maxR R pftso a b).
  intros T R h1 l a b h3.
  unfold maxR.
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst. 
  revert h4. revert a b.
  induction l as [|c l ih]; simpl; auto. 
  intros a b h2.
  destruct (h3 c a) as [h4 | h4].
  destruct h4 as [h4|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h2) as h5.
  destruct (h3 c b) as [h6 | h6].
  destruct h6 as [h6 | h6].
  apply ih; auto.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h5); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h6 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst.
  simpl.
  destruct (h3 a b) as [h5 | h5].
  destruct h5 as [h5 | h5].
  apply ih; auto.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  pose proof (sort_dec_irrefl _ h1  _  h2); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h4) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  simpl.
  destruct (h3 c b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  apply ih; auto.
  apply ih; auto.
  f_equal; apply ih; auto.

  induction l as [|a l ih]; simpl; auto.
  destruct (h3 a b) as [h4 | h4]; subst.
  destruct h4 as [h4|]; subst.
  apply ih; auto.
  simpl.
  destruct (h3 b b) as [h4 |].
  destruct h4 as [h4 | h5].
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  apply ih.
  apply ih.
  simpl.
  destruct (h3 a b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  f_equal; apply ih; auto.

  induction l as [|c l ih]; simpl; auto. 
  destruct (h3 c a) as [h5 | h5].
  destruct h5 as [h5|]; subst; auto.
  simpl.
  destruct (h3 c b) as [h5' | h5'].
  destruct h5' as [h5'|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_trans _ h1 _ _ _ h7 h5') as h8.
  pose proof (sort_dec_irrefl _ h1  _  h8); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  f_equal; auto.
Qed.
  
Lemma wlb_list_nest :
  forall {T:Type} (R:Relation T) (pfs:sort_dec R) (l:list T) (a b:T),
    let pftso := sort_dec_tso_dec R pfs in
    wlb_list R pftso (wlb_list R pftso l a) b =
    wlb_list R pftso l (maxR R pftso a b).
  intros T R h1 l a b h3.
  unfold maxR.
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst. 
  revert h4. revert a b.
  induction l as [|c l ih]; simpl; auto. 
  intros a b h2.
  destruct (h3 c a) as [h4 | h4].
  destruct h4 as [h4|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h2) as h5.
  destruct (h3 c b) as [h6 | h6].
  destruct h6 as [h6 | h6].
  apply ih; auto.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h5); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h6 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  destruct (h3 a b) as [h4 | h4].
  destruct h4 as [h4 |]; subst.
  simpl.
  destruct (h3 a b) as [h5 | h5].
  destruct h5 as [h5 | h5].
  apply ih; auto.
  subst.
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  pose proof (sort_dec_irrefl _ h1  _  h2); contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h4) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  simpl.
  destruct (h3 c b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  apply ih; auto.
  f_equal.
  apply ih; auto.
  f_equal.
  apply ih; auto.

  induction l as [|a l ih]; simpl; auto.
  destruct (h3 a b) as [h4 | h4]; subst.
  destruct h4 as [h4|]; subst.
  apply ih; auto.
  simpl.
  destruct (h3 b b) as [h4 |].
  destruct h4 as [h4 | h5].
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  f_equal; apply ih; auto.
  f_equal; apply ih; auto.
  simpl.
  destruct (h3 a b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  pose proof (sort_dec_irrefl _ h1  _  h4); contradiction.
  f_equal; apply ih; auto.

  induction l as [|c l ih]; simpl; auto.
  destruct (h3 c a) as [h5 | h5].
  destruct h5 as [h5|]; subst; auto.
  simpl.
  destruct (h3 a b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_irrefl _ h1  _  h7); contradiction.
  f_equal; auto.
  f_equal; auto.
  simpl.
  destruct (h3 c b) as [h6 | h6].
  destruct h6 as [h6|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  pose proof (sort_dec_trans _ h1 _ _ _ h6 h7) as h8.
  pose proof (sort_dec_irrefl _ h1  _  h8); contradiction.
  f_equal.
  apply ih.
  f_equal.
  apply ih.
Qed.

Lemma ub_list_wlb_list_comm :
    forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a b:T),
      R a b ->
      let pftso := sort_dec_tso_dec _ pfso in 
      ub_list R pftso (wlb_list R pftso l a) b =
      wlb_list R pftso (ub_list R pftso l b) a.
  intros T R h1 l.
  induction l as [|c l ih]; auto.
  intros a b h2 hts.
  simpl. 
  destruct (hts c b) as [h4 | h4].
  destruct h4 as [h4|]; subst.
  simpl.
  destruct (hts c a) as [h3 | h3].
  destruct h3 as [h3|]; subst; auto.
  simpl.
  destruct (hts a b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  f_equal; auto.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ h1) in h6; contradiction.
  simpl.
  destruct (hts c b) as [h6 | h6].
  destruct h6 as [h7 | h7].
  f_equal; auto.
  subst.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h6) as h7.
  apply (sort_dec_irrefl _ h1) in h7; contradiction.
  destruct (hts b a) as [h4 | h4].
  destruct h4 as [h4|]; subst; auto.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  simpl.
  destruct (hts b b) as [h5 | h5].
  destruct h5 as [h5 | h5].
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
  auto.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
  destruct (hts c a) as [h3 | h3].
  destruct h3 as [h3|]; subst; auto.
  simpl.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h4) as h7.
  apply (sort_dec_irrefl _ h1) in h7; contradiction.
  simpl.
  destruct (hts c b) as [h5 | h5].
  destruct h5 as [h5 | h5].
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h7.
  apply (sort_dec_irrefl _ h1) in h7; contradiction.
  subst.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  auto.
Qed.

Lemma in_or_ub_wlb :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T),
    In a l ->
    let pftso := sort_dec_tso_dec _ pfso in
    In a (ub_list R pftso l a) \/
    In a (wlb_list R pftso l a).
  intros T R h1 l.
  destruct l as [|b l];  auto.
  intros a h2 hts.
  destruct h2 as [|h2]; subst.
  destruct (hts a a) as [h2 | h2].
  destruct h2 as [h2 | h2].
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  rewrite wlb_list_cons_same.
  right; left; auto.
  apply sort_dec_irrefl; auto.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  simpl.
  destruct (hts b a) as [h3 | h3].
  destruct h3 as [h3 | h3].
  right.
  unfold hts. rewrite in_wlb_list_iff.
  split; auto; right; auto.
  subst.
  right; left; auto.
  right.
  simpl.
  right.
  unfold hts. rewrite in_wlb_list_iff.
  split; auto; right; auto.
Qed.


Lemma in_or_lb_wub :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T),
    In a l ->
    let pftso := sort_dec_tso_dec _ pfso in
    In a (lb_list R pftso l a) \/
    In a (wub_list R pftso l a).
  intros T R h1 l.
  destruct l as [|b l];  auto.
  intros a h2 hts.
  destruct h2 as [|h2]; subst.
  destruct (hts a a) as [h2 | h2].
  destruct h2 as [h2 | h2].
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  rewrite wub_list_cons_same.
  right; left; auto.
  apply sort_dec_irrefl; auto.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  simpl.
  destruct (hts b a) as [h3 | h3].
  destruct h3 as [h3 | h3].
  right.
  unfold hts. right. rewrite in_wub_list_iff.
  split; auto; right; auto.
  subst.
  right; left; auto.
  right.
  simpl.
  unfold hts. rewrite in_wub_list_iff.
  split; auto; right; auto.
Qed.
 
Lemma in_ub_list_iffn :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a x:T),
    In x l ->
    let pftso := sort_dec_tso_dec _ pfso in
    In x (ub_list R pftso l a) <->
    ~In x (wlb_list R pftso l a).
  intros T R h1 l a x h2 h3; split; intro h4.
  intro h5.
  unfold h3 in h4, h5.
  rewrite in_ub_list_iff in h4.
  rewrite in_wlb_list_iff in h5.
  destruct h4 as [? h4], h5 as [? h5].
  rewrite R_iffn in h4; auto. 
  unfold h3 in h4. rewrite in_wlb_list_iff in h4.
  unfold h3. rewrite in_ub_list_iff.
  assert (h5:~ Rw R a x). tauto.
  split; auto.
  rewrite R_iffn; auto.
Qed.


Lemma in_wub_list_iffn :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a x:T),
    In x l ->
    let pftso := sort_dec_tso_dec _ pfso in
    In x (wub_list R pftso l a) <->
    ~In x (lb_list R pftso l a).
  intros T R h1 l a x h2 h3; split; intro h4.
  intro h5.
  unfold h3 in h4, h5.
  rewrite in_wub_list_iff in h4.
  rewrite in_lb_list_iff in h5.
  destruct h4 as [? h4], h5 as [? h5]. 
  rewrite R_iffn in h5; auto. 
  unfold h3 in h4. rewrite in_lb_list_iff in h4.
  unfold h3. rewrite in_wub_list_iff.
  assert (h5:~ R a x). tauto.
  split; auto.
  rewrite Rw_iffn; auto.
Qed.


Lemma in_lb_list_iffn :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a x:T),
    In x l ->
    let pftso := sort_dec_tso_dec _ pfso in
    In x (lb_list R pftso l a) <->
    ~In x (wub_list R pftso l a).
  intros T R h1 l a x h2 h3; split; intro h4.
  intro h5.
  unfold h3 in h4, h5.
  rewrite in_lb_list_iff in h4.
  rewrite in_wub_list_iff in h5.
  destruct h4 as [? h4], h5 as [? h5].
  rewrite R_iffn in h4; auto. 
  unfold h3 in h4. rewrite in_wub_list_iff in h4.
  unfold h3. rewrite in_lb_list_iff.
  assert (h5:~ Rw R x a). tauto.
  split; auto.
  rewrite R_iffn; auto.
Qed.


Lemma in_wlb_list_iffn :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a x:T),
    In x l ->
    let pftso := sort_dec_tso_dec _ pfso in
    In x (wlb_list R pftso l a) <->
    ~In x (ub_list R pftso l a).
  intros T R h1 l a x h2 h3; split; intro h4.
  intro h5.
  unfold h3 in h4, h5.
  rewrite in_wlb_list_iff in h4.
  rewrite in_ub_list_iff in h5.
  destruct h4 as [? h4], h5 as [? h5].
  rewrite R_iffn in h5; auto. 
  unfold h3 in h4. rewrite in_ub_list_iff in h4.
  unfold h3. rewrite in_wlb_list_iff.
  assert (h5:~ R x a). tauto.
  split; auto. 
  rewrite Rw_iffn; auto.
Qed.


Lemma length_decompose_ub_wlb :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (a:T),
    length l =
    length (ub_list R pftso l a) +
    length (wlb_list R pftso l a).
  intros T R h1 l.
  induction l as [|c l ih]; simpl; auto.
  intro a.
  destruct (h1 c a) as [h2 | h2]; simpl.
  destruct h2 as [h2 |]; subst; simpl.
  f_equal; auto.
  rewrite plus_S_shift_r.
  f_equal; auto.
  rewrite plus_S_shift_r.
  f_equal; auto.
Qed.

Lemma length_decompose_lb_wub :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (a:T),
    length l =
    length (lb_list R pftso l a) +
    length (wub_list R pftso l a).
  intros T R h1 l.
  induction l as [|c l ih]; simpl; auto.
  intro a. 
  destruct (h1 c a) as [h2 | h2]; simpl.
  destruct h2 as [h2 |]; subst; simpl.
  rewrite plus_S_shift_r.
  f_equal; auto.
  rewrite plus_S_shift_r. 
  f_equal; auto.
  f_equal; auto.
Qed.
  

Lemma length_wlb_list_decompose :
    forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a b:T),
      R a b ->
      let pftso := sort_dec_tso_dec _ pfso in 
      length (wlb_list R pftso l a) =
      length (ub_list R pftso (wlb_list R pftso l a) b) + 
      length (wlb_list R pftso l b).
  intros T R h1 l.
  induction l as [|c l ih]; auto.
  intros a b h2 htso.
  simpl.
  destruct (htso c a) as [h3 | h3].
  destruct h3 as [h3|]; subst.
  pose proof (sort_dec_trans _ h1 _ _ _ h3 h2) as h4.
  destruct (htso c b) as [h5 | h5].
  destruct h5 as [h5|]; subst; auto.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h4 h5) as h6.
  apply (sort_dec_irrefl _ h1) in h6; contradiction.
  simpl.
  destruct (htso a b) as [h3 | h3].
  destruct h3 as [h3|]; subst.
  simpl; f_equal; auto.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h6.
  apply (sort_dec_irrefl _ h1) in h6; contradiction.
  simpl.
  destruct (htso c b) as [h5 | h5].
  destruct h5 as [h5|]; subst.
  simpl; f_equal; auto.
  simpl.
  rewrite plus_S_shift_r.
  f_equal; auto.
  simpl.
  rewrite plus_S_shift_r.
  f_equal; auto.
Qed.

  
Lemma count_occ_ub_wlb_decompose :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T) (pftso:tso_dec R) (l:list T) (a x:T),
    count_occ pfdt l x = 
    count_occ pfdt (ub_list R pftso l a) x +
    count_occ pfdt (wlb_list R pftso l a) x.
  intros T h1 R h2 l.
  induction l as [|c l ih]; simpl; auto.
  intros a x.
  destruct (h1 c x) as [|h3]; subst.
  destruct (h2 x a) as [h4|h4]; subst.
  destruct h4 as [h4 | h4].
  simpl.
  destruct (h1 x x) as [h5 | h5].
  rewrite plus_S_shift_l.
  f_equal; auto.
  contradict h5; auto.
  subst.
  simpl.
  destruct (h1 a a) as [h3 | h3].
  rewrite plus_S_shift_r.
  f_equal; auto.
  contradict h3; auto.
  simpl.
  destruct (h1 x x) as [h5 | h5].
  rewrite plus_S_shift_r.
  f_equal; auto.
  contradict h5; auto.
  destruct (h2 c a) as [h4 | h4].
  destruct h4 as [h4 | h4].
  simpl.
  destruct (h1 c x) as [h5 | h5].
  contradiction.
  apply ih.
  subst.
  simpl.
  destruct (h1 a x) as [h4 | h4].
  contradiction.
  apply ih.
  simpl.
  destruct (h1 c x) as [h5 | h5].
  contradiction.
  apply ih.
Qed.


Lemma coarse_list_wub_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    coarse_list (wub_list R pftso l b) l.
  intros T R h1 l.
  induction l as [|a l h2]; simpl.
  intros; constructor.
  intro b. 
  destruct (h1 a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  apply coarse_list_cons2.
  apply h2.
  apply coarse_list_cons2.
  apply h2.
  apply coarse_list_cons1.
  apply h2.
Qed.


Lemma no_dup_wub_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (a:T),
    NoDup l ->
    NoDup (wub_list R pftso l a).
  intros T R h1 l a h2.
  eapply no_dup_coarse_list.
  apply coarse_list_wub_list.
  assumption.
Qed.

Lemma coarse_list_lb_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    coarse_list (lb_list R pftso l b) l.
  intros T R h1 l.
  induction l as [|a l h2]; simpl.
  intros; constructor.
  intro b. 
  destruct (h1 a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  apply coarse_list_cons1.
  apply h2.
  apply coarse_list_cons1.
  apply h2.
  apply coarse_list_cons2.
  apply h2.
Qed.


Lemma no_dup_lb_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (a:T),
    NoDup l ->
    NoDup (lb_list R pftso l a).
  intros T R h1 l a h2.
  eapply no_dup_coarse_list.
  apply coarse_list_lb_list.
  assumption.
Qed.


Lemma coarse_list_wlb_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (b:T),
    coarse_list (wlb_list R pftso l b) l.
  intros T R h1 l.
  induction l as [|a l h2]; simpl.
  intros; constructor.
  intro b. 
  destruct (h1 a b) as [h3 | h3].
  destruct h3 as [h3 | h3]; subst.
  apply coarse_list_cons1.
  apply h2.
  apply coarse_list_cons2.
  apply h2.
  apply coarse_list_cons2.
  apply h2.
Qed.


Lemma no_dup_wlb_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (a:T),
    NoDup l ->
    NoDup (wlb_list R pftso l a).
  intros T R h1 l a h2.
  eapply no_dup_coarse_list.
  apply coarse_list_wlb_list.
  assumption.
Qed.



Lemma ub_list_lt_length_step :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    match ub_list R pftso l b with
      | nil => True
      | a :: l' => length (ub_list R pftso l' a) < S (length l')
    end.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  simpl; auto.
  intros b ht.
  simpl.
  destruct (ht a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  pose proof (coarse_list_ub_list R ht (ub_list R ht l b) a) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S.
  assumption.
  subst.
  destruct (nil_dec' (ub_list R ht l b)) as [h4 | h4].
  rewrite h4; simpl.
  auto with arith.
  rewrite (hd_compat' _ h4). 
  pose proof (coarse_list_ub_list R ht (tl (ub_list R ht l b)) (hd' (ub_list R ht l b) h4)) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S; auto.
  destruct (nil_dec' (ub_list R ht l b)) as [h4 | h4].
  rewrite h4; simpl; auto.
  rewrite (hd_compat' _ h4).
  pose proof (coarse_list_ub_list R ht (tl (ub_list R ht l b)) (hd' (ub_list R ht l b) h4)) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S; auto.
Qed.


Lemma wub_list_lt_length_step :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    match wub_list R pftso l b with
      | nil => True
      | a :: l' => length (wub_list R pftso l' a) < S (length l')
    end.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  simpl; auto.
  intros b ht.
  simpl.
  destruct (ht a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  pose proof (coarse_list_wub_list R ht (wub_list R ht l b) a) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S; auto.
  subst.
  pose proof (coarse_list_wub_list R ht (wub_list R ht l b) b) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S; auto.
  destruct (nil_dec' (wub_list R ht l b)) as [h4 | h4]. 
  rewrite h4; auto.
  rewrite (hd_compat' _ h4).
  pose proof (coarse_list_wub_list R ht (tl (wub_list R ht l b))  (hd' (wub_list R ht l b) h4)) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  apply le_lt_n_Sm; auto.
Qed.


Lemma lb_list_lt_length_step :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    match lb_list R pftso l b with
      | nil => True
      | a :: l' => length (lb_list R pftso l' a) < S (length l')
    end.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros b ht.
  simpl; auto.
  intros b ht.
  simpl.
  destruct (ht a b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  apply ih.
  subst. 
  destruct (nil_dec' (lb_list R ht l b)) as [h4 | h4].
  rewrite h4; simpl; auto.
  rewrite (hd_compat' _ h4).
  pose proof (coarse_list_lb_list R ht (tl (lb_list R ht l b)) (hd' (lb_list R ht l b) h4))  as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S; auto.
  pose proof (lb_list_le R h1 (lb_list R ht l b) b a h3) as h4.
  simpl in h4.
  fold ht in h4.
  apply le_lt_n_Sm; auto.
  eapply le_trans.
  apply h4.
  pose proof (coarse_list_lb_list R ht (lb_list R ht l b) b) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  assumption.
Qed.


Lemma wlb_list_lt_length_step :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    match wlb_list R pftso l b with
      | nil => True
      | a :: l' => length (wlb_list R pftso l' a) < S (length l')
    end.
  intros T R h1 l.
  induction l as [|a l ih]; auto.
  intros b ht.
  simpl; auto.
  intros b ht.
  simpl. 
  destruct (ht a b) as [h3 | h3]. 
  destruct h3 as [h3 | h3]. 
  destruct (nil_dec' (wlb_list R ht l b)) as [h4 | h4].
  rewrite h4; simpl; auto.
  rewrite (hd_compat' _ h4).
  pose proof (coarse_list_wlb_list R ht (tl (wlb_list R ht l b)) (hd' (wlb_list R ht l b) h4))  as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S; auto.
  subst.
  pose proof (coarse_list_wlb_list R ht (wlb_list R ht l b) b) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  red.
  apply le_n_S; auto.
  pose proof (coarse_list_wlb_list R ht (wlb_list R ht l b) a) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  apply le_lt_n_Sm; auto.
Qed.


Lemma ub_wlb_list_lt_length_step :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    match ub_list R pftso l b with
      | nil => True
      | a :: l' =>
        length (wlb_list R (sort_dec_tso_dec R pfso) l' a) < S (length l')
    end.
  intros T R h1 l.
  induction l as [|a l ih]; simpl; auto. 
  intro b. 
  destruct (sort_dec_tso_dec R h1 a b) as [h4 | h4].
  destruct h4 as [h4 | h4]; simpl.
  pose proof (coarse_list_wlb_list R (sort_dec_tso_dec R h1) (ub_list R (sort_dec_tso_dec R h1) l b) a) as h5.
  pose proof (length_coarse_list _ _ h5) as h6.
  apply le_lt_n_Sm; auto.
  subst.
  apply ih.
  apply ih.
Qed.

Lemma wlb_ub_list_lt_length_step :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (b:T),
    let pftso := sort_dec_tso_dec _ pfso in
    match wlb_list R pftso l b with
      | nil => True
      | a :: l' =>
        length (ub_list R (sort_dec_tso_dec R pfso) l' a) < S (length l')
    end.
  intros T R h1 l.
  induction l as [|a l ih]; simpl; auto. 
  intro b.  
  destruct (sort_dec_tso_dec R h1 a b) as [h4 | h4].
  destruct h4 as [h4 | h4]; simpl.
  destruct (nil_dec' (wlb_list R (sort_dec_tso_dec R h1) l b)) as [h5 | h5].
  rewrite h5; auto.
  rewrite (hd_compat' _ h5).
  pose proof (coarse_list_ub_list R (sort_dec_tso_dec R h1) (tl (wlb_list R (sort_dec_tso_dec R h1) l b)) (hd' (wlb_list R (sort_dec_tso_dec R h1) l b) h5) ) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  apply le_lt_n_Sm; auto.
  subst.
  pose proof (coarse_list_ub_list R (sort_dec_tso_dec R h1) (wlb_list R (sort_dec_tso_dec R h1) l b) b) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  apply le_lt_n_Sm; auto.
  pose proof (coarse_list_ub_list R (sort_dec_tso_dec R h1) (wlb_list R (sort_dec_tso_dec R h1) l b) a) as h6.
  pose proof (length_coarse_list _ _ h6) as h7.
  apply le_lt_n_Sm; auto.
Qed.

Fixpoint increasingR {T:Type}
         (R:Relation T) (l:list T) : Prop :=
  match l with
    | nil => True
    | a::l' => match l' with
                 | nil => True
                 | a'::l'' =>  R a a' /\ increasingR R l'
               end
  end.

Fixpoint weak_increasingR {T:Type}
         (R:Relation T) (l:list T) : Prop :=
  match l with
    | nil => True
    | a::l' => match l' with
                 | nil => True
                 | a'::l'' =>  Rw R a a' /\ weak_increasingR R l'
               end
  end.

Lemma increasingR_cons :
  forall {T:Type} (R:Relation T) (l:list T) (a:T),
    increasingR R (a::l) ->
    increasingR R l.
  intros T R l a h1.
  simpl in h1.
  revert a h1.
  destruct l as [|a l']; simpl; auto.
  intros a' h1.
  destruct h1; auto.
Qed.

Lemma weak_increasingR_cons :
  forall {T:Type} (R:Relation T) (l:list T) (a:T),
    weak_increasingR R (a::l) ->
    weak_increasingR R l.
  intros T R l a h1.
  simpl in h1.
  revert a h1.
  destruct l as [|a l']; simpl; auto.
  intros a' h1.
  destruct h1; auto.
Qed.


Lemma increasingR_impl_weak :
  forall {T:Type} (R:Relation T) (l:list T),
    increasingR R l ->
    weak_increasingR R l.
  intros T R l.
  induction l as [|a l ih]; simpl; auto.
  intro h1.
  destruct l as [|c l]; simpl; auto.
  destruct h1 as [h1 h2].
  specialize (ih h2).
  split; auto; red. left; auto.
Qed.

Lemma lt_increasingR_cons :
  forall {T:Type} (R:Relation T) (l:list T) (a m:T),
    Transitive _ R ->
    In a l ->
    increasingR R (m::l)->
    R m a.
  intros T R l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros a' m ht h1 h2.
  destruct h2 as [h2 h3].
  destruct h1 as [h1 | h1]; subst; auto.
  specialize (ih _ _ ht h1 h3).
  eapply ht.
  apply h2. assumption.
Qed.

Lemma le_weak_increasingR_cons :
  forall {T:Type} (R:Relation T) (l:list T) (a m:T),
    Transitive _ R ->
    In a l ->
    weak_increasingR R (m::l)->
    Rw R m a.
  intros T R l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros a' m ht h1 h2.
  destruct h2 as [h2 h3].
  destruct h1 as [h1 | h1]; subst; auto.
  specialize (ih _ _ ht h1 h3).
  eapply trans_Rw; auto.
  apply h2. assumption.
Qed.


Lemma increasingR_cons_impl :
  forall {T:Type} (R:Relation T) (l:list T) (a:T),
    increasingR R (a::l) ->
    increasingR R l.
  intros T R l a; destruct l; simpl.
  intros; auto.
  intros h1.
  destruct h1; auto.
Qed.

Lemma weak_increasingR_cons_impl :
  forall {T:Type} (R:Relation T) (l:list T) (a:T),
    weak_increasingR R (a::l) ->
    weak_increasingR R l.
  intros T R l; destruct l; simpl.
  intros; auto.
  intros ? h1.
  destruct h1; auto.
Qed.

Lemma increasingR_cons_impl_cons :
  forall {T:Type} (R:Relation T) (l:list T) (a d:T),
    increasingR R (a::l) ->
    R d a ->
    increasingR R (d::a::l).
  intros l  a d h1 h2.
  simpl.
  split; auto.
Qed.


Lemma weak_increasingR_cons_impl_cons :
  forall {T:Type} (R:Relation T) (l:list T) (a d:T),
    weak_increasingR R (a::l) ->
    Rw R d a ->
    weak_increasingR R (d::a::l).
  intros T R l a d h1 h2.
  simpl.
  split; auto.
Qed.

Lemma increasingR_impl_cons :
  forall {T:Type} (R:Relation T) (l:list T)
    (pf:l <> nil) (d:T),
    increasingR R l ->
    R d (hd' l pf) ->
    increasingR R (d::l).
  intros T R l.
  destruct l as [|h l]; simpl.
  intros; auto.
  intros h1 d h2 h3.
  split; auto.
Qed.

Lemma weak_increasingR_impl_cons :
  forall {T:Type} (R:Relation T) (l:list T)
    (pf:l <> nil) (d:T),
    weak_increasingR R l ->
    Rw R d (hd' l pf) ->
    weak_increasingR R (d::l).
  intros T R l.
  destruct l as [|h l]; simpl.
  intros; auto.
  intros h1 d h2 h3.
  split; auto.
Qed.

Lemma increasingR_impl_cons' :
  forall {T:Type} (R:Relation T) (l:list T) (d:T),
    increasingR R l ->
    (forall c:T, In c l -> R d c) ->
    increasingR R (d::l).
  intros T R l.
  induction l as [|c l ih]; simpl.
  intros; auto.
  intros d h1 h2.
  split; auto.
Qed.

Lemma weak_increasingR_impl_cons' :
  forall {T:Type} (R:Relation T)
         (l:list T) (d:T),
    weak_increasingR R l ->
    (forall c:T, In c l -> Rw R d c) ->
    weak_increasingR R (d::l).
  intros T R l.
  induction l as [|c l ih]; simpl.
  intros; auto.
  intros d h1 h2.
  split; auto.
Qed.


Lemma increasingR_tl :
  forall {T:Type} (R:Relation T) (l:list T),
    increasingR R l ->
    increasingR R (tl l).
  intros T R l.
  destruct l as [|a l]; simpl; auto.
  intro h1.
  eapply increasingR_cons_impl.
  apply h1.
Qed.

Lemma weak_increasingR_tl :
  forall {T:Type} (R:Relation T) (l:list T),
    weak_increasingR R l ->
    weak_increasingR R (tl l).
  intros T R l.
  destruct l as [|a l]; simpl; auto.
  intro h1.
  eapply weak_increasingR_cons_impl.
  apply h1.
Qed.

Lemma increasingR_remove :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T) (l:list T) (a:T),
    Transitive _ R ->
    increasingR R l ->
    increasingR R (remove pfdt a l).
  intros T hdt R l.
  induction l as [|c l ih]; simpl; auto.
  intros a ht h1.
  destruct (hdt a c) as [h2 | h2]; subst.
  apply ih; auto.
  eapply increasingR_cons_impl; auto.
  apply h1.
  simpl.
  pose proof (increasingR_cons_impl R _ _ h1)as h3.
  apply increasingR_impl_cons'.
  apply ih; auto.
  intros d h4.
  pose proof (remove_In hdt l a) as h5.
  assert (h6:d <> a).
    intro; subst; contradiction.
  pose proof (in_remove_neq_in_l hdt l d a  h4 h6) as h7.
  eapply lt_increasingR_cons.
  assumption.
  apply h7.
  assumption.
Qed.

Lemma weak_increasingR_remove :
  forall {T:Type} (pfdt:type_eq_dec T)  (R:Relation T)
         (l:list T) (a:T),
    Transitive _ R ->
    weak_increasingR R l ->
    weak_increasingR R (remove pfdt a l).
  intros T hdt R l.
  induction l as [|c l ih]; simpl; auto.
  intros a ht h1.
  destruct (hdt a c) as [h2 | h2]; subst.
  apply ih; auto.
  eapply weak_increasingR_cons_impl; auto.
  apply h1.
  simpl.
  pose proof (weak_increasingR_cons_impl R _ _ h1)as h3.
  apply weak_increasingR_impl_cons'.
  apply ih; auto.
  intros d h4.
  pose proof (remove_In hdt l a) as h5.
  assert (h6:d <> a).
    intro; subst; contradiction.
  pose proof (in_remove_neq_in_l hdt l d a  h4 h6) as h7.
  eapply le_weak_increasingR_cons.
  assumption.
  apply h7.
  assumption.
Qed.

Lemma incrR_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T) (R:Relation T)
         (l l':list T),
    Transitive _ R ->
    increasingR R l ->
    let lm := list_minus pfdt l l' in
    increasingR R lm.
  intros T hdt R l l'. revert l.
  induction l' as [|a l ih]; simpl; auto.
  intros. rewrite list_minus_nil2; auto.
  intros l' ht h2.
  rewrite list_minus_cons_remove.
  apply increasingR_remove.
  assumption.
  apply ih; auto.
Qed.

Lemma weak_incrR_list_minus :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T) (l l':list T),
    Transitive _ R ->
    weak_increasingR R l ->
    let lm := list_minus pfdt l l' in
    weak_increasingR R lm.
  intros T hdt R l l'. revert l.
  induction l' as [|a l ih]; simpl; auto.
  intros. rewrite list_minus_nil2; auto.
  intros l' ht h2.
  rewrite list_minus_cons_remove.
  apply weak_increasingR_remove.
  assumption.
  apply ih; auto.
Qed.

Lemma increasingR_cons_lt_subst :
  forall {T:Type} (R:Relation T) (l:list T) (a b:T),
    Transitive _ R ->
    increasingR R  (a::l) ->
    R b a ->
    increasingR R (b::l).
  intros T R l a b ht h1 h2.
  destruct l as [|c l]; simpl; auto.
  simpl in h1.
  destruct h1 as [h1 h3].
  split; auto.
  eapply ht.
  apply h2. assumption.
Qed.

Lemma weak_increasingR_cons_le_subst :
  forall {T:Type} (R:Relation T) (l:list T) (a b:T),
    Transitive _ R ->
    weak_increasingR R (a::l) ->
    Rw R b a ->
    weak_increasingR R (b::l).
  intros T R l a b ht h1 h2.
  destruct l as [|c l]; simpl; auto.
  simpl in h1.
  destruct h1 as [h1 h3].
  split; auto.
  eapply trans_Rw; auto.
  apply h2. assumption.
Qed.


Lemma incrR_dec :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T),
    {increasingR R l} + {~increasingR R l}.
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  left; auto.
  destruct ih as [ih | ih].
  destruct l as [|b l].
  left; auto.
  destruct (sort_dec_tso_dec _ h1 a b) as [h2 | h2].
  destruct h2 as [h2 |]; subst.
  left; split; auto.
  right; intro h2.
  destruct h2 as [h2].
  apply (sort_dec_irrefl _ h1) in h2; auto.
  right.
  intro h3.
  destruct h3 as [h3 h4].
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h5.
  apply (sort_dec_irrefl _ h1) in h5; auto.
  destruct l as [|b l].
  left; auto.
  right.
  intro h2.
  destruct h2; contradiction.
Qed.


Lemma weak_incrR_dec :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T),
    {weak_increasingR R l} + {~weak_increasingR R l}.
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  left; auto.
  destruct ih as [ih | ih].
  destruct l as [|b l].
  left; auto.
  destruct (sort_dec_tso_dec _ h1 a b) as [h2 | h2].
  destruct h2 as [h2 |]; subst.
  left; split; auto.
  left; auto.
  left; split; auto.
  right; auto.
  right.
  intro h3.
  destruct h3 as [h3 h4].
  pose proof (trans_Rw_R R (sort_dec_trans _ h1) _ _ _ h3 h2) as h5.
  apply (sort_dec_irrefl _ h1) in h5; auto.
  destruct l as [|b l].
  left; auto.
  right.
  intro h2.
  destruct h2; contradiction.
Qed.



Lemma increasingR_cons_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T) (l:list T) (x:T),
    Strict_Order R ->
    increasingR R (x::l) ->
    NoDup l.
  intros T hdt R l x  hso. revert x. 
  inversion hso as [ht hi].
  induction l as [|a l ih]; simpl.
  intros; constructor.
  intros x  h1.

  destruct h1 as [h1 h2].
  specialize (ih _  h2).
  constructor; auto.
  intro h3.
  pose proof (interp_in_eq hdt  _ _ h3) as h5.
  unfold interp_in in h5; simpl in h5.
  destruct l as [|c l].
  simpl in h3; auto.
  simpl in h5.
  unfold lind in h5; simpl in h5.
  destruct (hdt a c) as [h6 | h6]; subst.
  simpl in h5.
  destruct h2 as [h2 h6].
  apply hi in h2; auto.
  simpl in h5.
  inversion h5 as [h8].
  assert (h9:In a l).
    rewrite h8.
    apply in_app_r.
    left; auto.
  simpl in h2.
  destruct h2 as [h2 h4].
  destruct h3 as [h3 | h3].
  rewrite h3 in h2. apply hi in h2; auto.
  pose proof (lt_increasingR_cons R _ _ _  ht h9 h4) as h10.
  pose proof (ht _ _ _ h2 h10) as h11.
  apply hi in h11; auto.
Qed.


Lemma increasingR_cons_nin :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T) (l:list T) (a:T),
    Strict_Order R ->
    increasingR R (a::l) ->
    ~In a l.
  intros T hdt R l a hso. revert a.
  inversion hso as [ht hi].
  induction l as [|d l ih]; simpl.
  intros; auto.
  intros a h1.
  destruct h1 as [h1 h3].
  intro h2.
  destruct h2 as [h2 | h2]; subst.
  apply hi in h1; auto.
  pose proof (interp_in_eq hdt  _ _ h2) as h4.
  unfold interp_in in h4; simpl in h4.
  destruct l as [|c l].
  simpl in h2; auto.
  simpl in h4.
  unfold lind in h4; simpl in h4.
  destruct (hdt a c) as [h6 | h6]; subst.
  simpl in h4.
  destruct h3 as [h3 h6].
  pose proof (ht _ _ _ h1 h3) as h7.
  apply hi in h7; auto.
  simpl in h4.
  inversion h4 as [h8]. clear h4.
  assert (h9:In a l).
    rewrite h8.
    apply in_app_r.
    left; auto.
  simpl in h2, h3.
  destruct h3 as [h3 h4].
  destruct h2 as [h2 | h2].
  clear h8. subst. contradict h6; auto. 
  pose proof (lt_increasingR_cons R _ _ _  ht h9 h4) as h10.
  pose proof (ht _ _ _ h3 h10) as h11.
  pose proof (ht _ _ _ h1 h11) as h12.
  apply hi in h12; auto.
Qed.


Lemma increasingR_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T) (R:Relation T)
         (l:list T),
    Strict_Order R ->
    increasingR R l ->
    NoDup l.
  intros T hdt R l hso.
  destruct l as [|a l]; simpl.
  intros; constructor.
  intro h1.
  constructor.
  eapply increasingR_cons_nin; auto.
  apply hso.
  apply h1.
  eapply increasingR_cons_no_dup; auto.
  apply hso.
  apply h1.
Qed.


Lemma increasingR_firstn :
  forall {T:Type} (pfdt:type_eq_dec T) (R:Relation T)
         (l:list T) (m:nat),
    Transitive _ R ->
    increasingR R l ->
    increasingR R (firstn m l).
  intros T hdt R l.
  induction l as [|c l ih]; simpl.
  intros.
  rewrite firstn_nil.
  red; auto.
  intros m ht h1; simpl.
  destruct m as [|m]; auto.
  simpl; auto.
  pose proof (increasingR_cons_impl R _ _ h1) as h2.
  specialize (ih m ht h2).
  simpl.
  apply increasingR_impl_cons'; auto.
  intros a h3.
  pose proof (in_firstn _ _ _ h3) as h4.
  eapply lt_increasingR_cons; auto.
  apply h4.
  assumption.
Qed.


Lemma weak_increasingR_firstn :
  forall {T:Type} (R:Relation T) (l:list T) (m:nat),
    Transitive _ R ->
    weak_increasingR R l ->
    weak_increasingR R (firstn m l).
  intros T R l m ht. revert m.
  induction l as [|c l ih]; simpl.
  intros.
  rewrite firstn_nil.
  red; auto.
  intros m h1; simpl.
  destruct m as [|m]; auto.
  simpl; auto.
  pose proof (weak_increasingR_cons_impl R _ _ h1) as h2.
  specialize (ih m h2).
  simpl.
  apply weak_increasingR_impl_cons'; auto.
  intros a h3.
  pose proof (in_firstn _ _ _ h3) as h4.
  eapply le_weak_increasingR_cons.
  assumption.
  apply h4.
  assumption.
Qed.

Lemma increasingR_skipn :
  forall {T:Type} (R:Relation T) (l:list T) (m:nat),
    increasingR R l ->
    increasingR R (skipn m l).
  intros T R l.
  induction l as [|c l ih]; simpl.
  intros.
  rewrite skipn_nil.
  red; auto.
  intros m h1; simpl.
  destruct m as [|m]; auto.
  simpl; auto.
  pose proof (increasingR_cons_impl R _ _ h1) as h2.
  specialize (ih m h2).
  simpl.
  assumption.
Qed.


Lemma weak_increasingR_skipn :
  forall {T:Type} (R:Relation T) (l:list T) (m:nat),
    weak_increasingR R l ->
    weak_increasingR R (skipn m l).
  intros T R l.
  induction l as [|c l ih]; simpl.
  intros.
  rewrite skipn_nil.
  red; auto.
  intros m h1; simpl.
  destruct m as [|m]; auto.
  simpl; auto.
  pose proof (weak_increasingR_cons_impl R _ _ h1) as h2.
  specialize (ih m h2).
  simpl.
  assumption.
Qed.

Lemma increasingR_app_iff :
  forall {T:Type} (R:Relation T),
    Transitive _ R ->
    forall (l l':list T)
           (pfl:l <> nil)
           (pfl':l' <> nil),
      (increasingR R l /\
       increasingR R l' /\
       R (lst l pfl) (hd' l' pfl')) <->
      increasingR R (l++l').
  intros T R ht l.
  induction l as [|a l ih].
  intros l h1. contradict h1; auto.
  intros l' h1 h2.
  split.
  intros h345.
  destruct h345 as [h3 [h4 h5]].
  destruct l as [|b l]; simpl. 
  simpl in h5.
  destruct l' as [|c l']; simpl; auto.
  simpl in h3, h5. split.
  destruct h3; auto.
  destruct h3 as [h3 h6].
  specialize (ih l' (cons_neq_nil _ _) h2).
  let P := type of ih in match P with ?pf <-> _ => assert (h7:pf) end.
  repeat split; auto.
  simpl.
  destruct l as [|c l]; auto.
  erewrite last_indep in h5.
  apply h5.
  apply cons_neq_nil.
  rewrite ih in h7.
  destruct l as [|c l]; simpl.
  destruct l' as [|d l']; simpl; auto.
  split.
  destruct h6; auto.
  rewrite <- app_comm_cons in h7.
  apply increasingR_cons in h7.
  assumption. 
  intro h3. 
  pose proof (increasingR_cons _ _ _ h3) as h4.
  destruct l as [|b l]; simpl.
  simpl in ih, h3.
  destruct l' as [|c l'].
  contradict h2; auto.
  split; auto.
  split; auto.
  simpl.
  destruct h3; auto.
  specialize (ih l' (cons_neq_nil _ _) h2).
  rewrite <- ih in h4.
  destruct h4 as [h4 [h5 h6]].
  simpl in h6.
  split; auto. split.
  pose proof (lt_increasingR_cons _ _ _ _ ht (in_eq _ _) h3); auto.
  assumption.
  split; auto.
  destruct l as [|c l]; auto.
  erewrite last_indep.
  apply h6.
  apply cons_neq_nil.
Qed.


Lemma increasingR_last :
  forall {T:Type} (R:Relation T) (l:list T),
    Transitive _ R ->
    increasingR R l ->
    forall x:T,
      In x l ->
      Rw R x (last l x).
  intros T R l.  
  induction l as [|a l ih]. 
  intros; contradiction.
  intros htr h1 x h2. 
  erewrite last_indep; try apply cons_neq_nil.
  rewrite last_cons.
  pose proof (increasingR_cons R _ _ h1) as h3. 
  specialize (ih htr h3).
  destruct h2 as [|h2]; subst.
  simpl in h1.
  destruct l as [|c l].
  simpl.
  right; auto.
  destruct h1 as [h1 h2].
  left.
  eapply trans_R_Rw; auto.
  apply h1.
  erewrite last_indep.
  apply ih.
  left; auto.
  apply cons_neq_nil.
  erewrite last_indep.
  apply ih; auto.
  eapply in_not_nil.
  apply h2.
Qed.

Lemma increasingR_last' :
  forall {T:Type}  (R:Relation T) (pftr:Transitive _ R) (l:list T),
    increasingR R l ->
    forall (x:T),
      In x l ->
      x <> last l x ->
      R x (last l x).
  intros T R htr l h2 x h3 h4.
  pose proof (increasingR_last R l htr h2 _ h3) as h5.
  destruct h5 as [h5 | h5]; auto; try contradiction.
Qed.
 

Lemma weak_increasingR_last :
  forall {T:Type} (R:Relation T) (l:list T),
    Transitive _ R ->
    weak_increasingR R l ->
    forall (x:T) (pf:In x l), 
      Rw R x (last l x).
  intros T R l.  
  induction l as [|a l ih]. 
  intros; contradiction.
  intros htr h1 x h2. 
  erewrite last_indep; try apply cons_neq_nil.
  rewrite last_cons.
  destruct h2 as [|h2]; subst. 
  pose proof (weak_increasingR_cons R _ _ h1) as h3.
  specialize (ih htr h3).
  simpl in h1.
  destruct l as [|c l].
  simpl.
  right; auto. 
  destruct h1 as [h1 h2].
  destruct h1 as [h1 | h1].
  specialize (ih _ (in_eq _ _)).
  left.
  erewrite last_indep.
  apply (trans_R_Rw _ htr _ _  _ h1 ih).
  intro; discriminate.
  subst.
  apply ih; left; auto.
  erewrite last_indep.  apply ih; auto.
  eapply weak_increasingR_cons.
  apply h1.
  eapply in_not_nil. apply h2.
Qed.


Lemma weak_increasingR_app_iff :
  forall {T:Type} (R:Relation T),
    Transitive _ R ->
    forall (l l':list T)
           (pfl:l <> nil)
           (pfl':l' <> nil),
    (weak_increasingR R l /\
     weak_increasingR R l' /\
     Rw R (lst l pfl)  (hd' l' pfl')) <->
    weak_increasingR R (l++l').
  intros T R h0 l.
  induction l as [|a l ih].
  intros l h1. contradict h1; auto.
  intros l' h1 h2.
  split; intro h3.
  destruct h3 as [h3 [h4 h5]].
  destruct l as [|b l]; simpl.
  simpl in h5.
  destruct l' as [|c l']; simpl; auto.
  simpl in h3, h5. split.
  destruct h3; auto.
  destruct h3 as [h3 h6].
  specialize (ih l' (cons_neq_nil _ _) h2).
  let P := type of ih in match P with ?pf <-> _ => assert (h7:pf) end.
  simpl; repeat split; auto.
  destruct l as [|l c]; auto.
  erewrite last_indep. apply h5.
  apply cons_neq_nil.
  rewrite ih in h7.
  simpl in h7.
  assumption. 
  pose proof (weak_increasingR_cons _ _ _ h3) as h4. 
  destruct l as [|b l].
  simpl; split; auto.
  split; auto.
  simpl in h3.
  pose proof (hd_compat' _ h2) as h5. simpl in h5.
  rewrite h5 in h3.
  destruct h3; auto. 
  specialize (ih l' (cons_neq_nil _ _) h2).
  rewrite <- ih in h4.
  destruct h4 as [h4 [h5 h6]].
  pose proof (le_weak_increasingR_cons _ _ _ _ h0 (in_eq _ _) h3) as h7.
  simpl.
  simpl in h6.
  repeat split; auto.
  erewrite (last_indep (b::l) (cons_neq_nil _ _) a) at 1.
  apply h6.
Qed.

Lemma weak_incrR_no_dup_impl :
  forall {T:Type} (R:Relation T) (l:list T),
    weak_increasingR R l ->
    NoDup l ->
    increasingR R l.
  intros T R l.
  induction l as [|a l ih]; simpl; auto.
  destruct l as [|b l]; simpl; auto.
  intro h2.
  simpl in ih.
  destruct h2 as [h2 h3].
  specialize (ih h3); intro h4; split; auto.
  red in h2.
  destruct h2; subst; auto.
  apply no_dup_cons_nin in h4.
  contradict h4; left; auto.
  apply ih.
  apply no_dup_cons in h4; auto.
Qed.


Lemma increasingR_map :
  forall {T U:Type} (R:Relation T) (R':Relation U) (f:T->U) (l:list T),
    incrR_fun R R' f ->
    increasingR R l ->
    increasingR R' (map f l).
  intros T U R R' f l. revert f.
  induction l as [|a l ih]; simpl; auto.
  intros f h1 h2.
  specialize (ih f h1).
  hfirst ih.
  eapply increasingR_cons_impl. apply h2.
  specialize (ih hf).
  destruct l as [|c l].
  simpl; auto.
  destruct h2 as [h2 h3].
  simpl. split.
  apply h1; auto.
  simpl in ih; auto.
Qed.


Lemma weak_increasingR_map :
  forall {T U:Type} (R:Relation T) (R':Relation U) (f:T->U) (l:list T),
    weak_incrR_fun R R' f ->
    weak_increasingR R l ->
    weak_increasingR R' (map f l).
  intros T U R R' f l. revert f.
  induction l as [|a l ih]; simpl; auto.
  intros f h1 h2.
  specialize (ih f h1).
  hfirst ih.
  eapply weak_increasingR_cons_impl. apply h2.
  specialize (ih hf).
  destruct l as [|c l].
  simpl; auto.
  destruct h2 as [h2 h3].
  red in h2.
  destruct h2 as [h2 | h2]; subst.
  simpl. split.
  apply h1; auto.
  simpl in ih; auto.
  simpl.
  split; auto.
  right; auto.
Qed.

Lemma weak_incrR_lrep :
  forall {T:Type} (R:Relation T) (a:T) (n:nat),
    weak_increasingR R (lrep a n).
  intros ? ? ? n.
  induction n as [|n ih]; simpl; auto.
  destruct n as [|n]; simpl; auto.
  split.
  right; auto.
  assumption.
Qed.

  
Lemma incrR_all_sing :
  forall {T:Type} (pfdt:type_eq_dec T) (R:Relation T) (pf:Strict_Order R)
         (l:list T) (x:T),
    all_sing l x ->
    increasingR R l ->
    l = nil \/ l = x::nil.
  intros T hdt R hso l.
  induction l as [|c l ih]; simpl.
  intros; left; auto.
  intros m h1 h2.
  right.
  pose proof (h1 _ (in_eq _ _)); subst.
  f_equal.
  destruct (nil_dec l) as [|h3]; auto.
  destruct h3 as [a h3].
  specialize (h1 _ (in_cons _ _ _ h3)); subst.
  apply increasingR_cons_nin in h2; auto.
  contradiction.
Qed.

Lemma incrR_no_dup_minimal :
  forall {T:Type} (pfdt:type_eq_dec T) (R:Relation T) (pft:sort_dec R) 
         (l l':list T),
    increasingR R l ->
    NoDup l' ->
    list_to_set l = list_to_set l' ->
    le_listR R (sort_dec_tso_dec _ pft) l' l ->
    l = l'.
  intros T hdt R ht l l'. revert l.
  pose proof (sort_dec_so _ ht) as hso.
  destruct ht as [hts htr hir].
  induction l' as [|a l' ih]; simpl.
  intros ? ? ? h1. apply empty_set_nil in h1; auto.
  intros l h1 h0 h2 h3.
  destruct l as [|c l].
  contradiction. 
  pose proof (increasingR_no_dup hdt _ _ hso h1) as hn.
  pose proof (no_dup_cons_nin _ _ hn) as hnin.
  pose proof (no_dup_cons_nin _ _ h0) as hnin'.
  pose proof (no_dup_cons _ _  h0) as hnd.  
  unfold sort_dec_tso_dec in h3; simpl in h3. 
  destruct (hts a c) as [h4 | h4].
  destruct h4 as [h4 | h4].
  assert (h5:In a l).
    pose proof (Add_intro2 _ (list_to_set l') a) as h5.
    rewrite <- h2 in h5.
    rewrite <- list_to_set_in_iff in h5.
    destruct h5 as [h5 | h5]; subst.
    apply hir in h4; auto.
    contradiction.
    assumption.
  pose proof (lt_increasingR_cons R _ _ _ htr h5 h1) as h6.
  pose proof (htr _ _ _ h6 h4) as hi.
  apply hir in hi; contradiction.
  subst.
  simpl in h2.
  apply add_nin_inj_eq in h2.
  f_equal. apply ih; auto.
  eapply increasingR_cons.
  apply h1.
  contradict hnin.
  rewrite  list_to_set_in_iff; auto.
  contradict hnin'.
  rewrite  list_to_set_in_iff; auto.
  contradiction.
Qed.


Lemma weak_incrR_length_count_occ_eq :
  forall {T:Type} (pfdt:type_eq_dec T) (R:Relation T) (pfso:sort_dec R)
         (l l':list T),
    weak_increasingR R l ->
    weak_increasingR R l' ->
    length l = length l' ->
    (forall x:T, count_occ pfdt l x = count_occ pfdt l' x) ->
    l = l'.
  intros T h1 R h2 l.
  induction l as [|a l ih]; simpl. 
  intros l ? h3 h4 h5.
  symmetry in h4.
  apply length_eq0 in h4; auto.
  intros l' h3 h4 h5 h6.
  destruct l' as [|a' l'].
  simpl in h5. discriminate.
  simpl in h5.
  apply S_inj in h5.
  destruct (h1 a a') as [h7 | h7].
  subst.
  f_equal.
  apply ih.
  eapply weak_increasingR_cons.
  apply h3.
  eapply weak_increasingR_cons.
  apply h4.
  assumption.
  intro x.
  specialize (h6 x).
  destruct (h1 a' x) as [h7 | h7]; subst.
  simpl in h6.
  destruct (h1 x x) as [h7 | h7]; auto with arith.
  contradict h7; auto.
  simpl in h6.
  destruct (h1 a' x) as [h8 | h8].
  contradiction.
  assumption.
  destruct (in_dec h1 a' l) as [h8 | h8].
  pose proof (le_weak_increasingR_cons _ _ _ _ (sort_dec_trans _ h2) h8 h3) as h9.
  destruct (in_dec h1 a l') as [h10 | h10].
  pose proof (le_weak_increasingR_cons _ _ _ _ (sort_dec_trans _ h2) h10 h4) as h11.
  pose proof (antisym_Rw R (sort_dec_so _ h2) _ _  h9 h11) as h12.
  contradiction.
  specialize (h6 a).
  destruct (h1 a a) as [h11 | h11].
  simpl in h6.
  destruct (h1 a' a) as [h12 | h12].
  subst; contradict h7; auto.
  rewrite (count_occ_In h1) in h10.
  omega.
  contradict h11; auto.
  specialize (h6 a').
  simpl in h6.
  destruct (h1 a a') as [h9 | h9].
  contradiction.
  destruct (h1 a' a') as [h10 | h10].
  rewrite (count_occ_In h1) in h8.
  omega.
  contradict h10; auto.
Qed.



  
Lemma weak_incrR_coarse_list_cons_in :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T) (pft:Transitive _ R) (l l':list T) (a':T),
    weak_increasingR R (a'::l') ->
    coarse_list l (a'::l') ->
    forall (x:T),
      In x l ->
      Rw R a' x. 
  intros T hdt R ht l l' a' h1 h2. revert h1. 
  dependent induction h2; simpl; auto. 
  intros; contradiction. 
  intros h3 x h4.
  destruct (hdt x a') as [|he]; subst.
  right; auto.
  destruct l' as [|b l'].
  rewrite coarse_list_nil2_iff in h2; subst.
  simpl in h4. contradiction.
  destruct h3 as [h3 h5].
  destruct (hdt x b) as [|he']; subst.
  assumption.
  pose proof (in_coarse_list  l (b::l') _ h4 h2) as h7.
  destruct h7 as [|h7]; subst.
  contradict he'; auto.
  eapply le_weak_increasingR_cons in h5.
  eapply trans_Rw; auto.
  apply h3.
  apply h5.
  assumption.
  assumption.
  intros h1 x h3.
  destruct h3 as [|h3]; subst.
  right; auto.
  pose proof (in_coarse_list l l' _ h3 h2) as h5.
  pose proof h5 as h5'.
  rewrite (hd_compat' _ (in_not_nil _ _ h5)) in h1, h5'.
  destruct h1 as [h1 h4].
  destruct (hdt x (hd' l' (in_not_nil _ _ h5))) as [he | he].
  rewrite he; auto.
  destruct h5' as [h8 | h8].
  rewrite  h8 in he.
  contradict h3; auto.
  pose proof (le_weak_increasingR_cons R _ _ _ ht h8 h4) as h6.
  eapply trans_Rw.
  apply ht.
  apply h1.
  assumption.
Qed.

  
  
Lemma weak_incrR_coarse_list :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T) (pft:Transitive _ R) (l l':list T),
    weak_increasingR R l' ->
    coarse_list l l' ->
    weak_increasingR R l.
  intros T hdt R ht l l' h1 h2.
  revert h1.
  dependent induction h2; simpl; auto.  
  intro h3; apply IHh2.
  destruct l' as [|a' l']; simpl; auto. 
  destruct h3; auto.
  intro h3.
  destruct l' as [|a' l']; simpl; auto.
  destruct l as [|b l]; auto.
  apply not_coarse_list_cons_nil in h2; contradiction.
  destruct h3 as [h3 h4].
  destruct (nil_dec' l) as [|h5]; subst; auto.
  rewrite (hd_compat' _ h5).
  split.
  eapply trans_Rw; auto.
  apply h3.
  eapply weak_incrR_coarse_list_cons_in; auto.
  apply h4.
  apply h2.
  apply in_hd'.
  rewrite <- hd_compat'.
  apply IHh2; auto.
Qed.



Lemma incrR_le_nth_lt :
  forall {T:Type} (R:Relation T) (pftr:Transitive _ R)
         (l:list T),
    increasingR R l ->
    forall (i j:nat) (pfij:i < j)
           (pfi:i < length l) (pfj:j < length l),
      R (nth_lt l i pfi) (nth_lt l j pfj).
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros h0 i j h2 h3 h4.
  destruct (nil_dec' l) as [|h5]; subst.
  simpl in h3, h4.
  omega.
  rewrite (hd_compat' _ h5) in h0.
  destruct h0 as [h0 h6].
  rewrite <- (hd_compat' _ h5) in h6.
  destruct i as [|i], j as [|j].
  omega.
  apply lt_S_n in h4.
  destruct j as [|j].
  erewrite nth_indep. erewrite <- nth_lt_compat. rewrite nth_lt_0_hd'.
  erewrite hd_functional'.
  apply h0.
  assumption.
  specialize (ih h6 0 (S j) (lt_0_Sn _)).
  hfirst ih. omega.
  specialize (ih hf h4).
  
  rewrite nth_lt_0_hd' in ih.
  erewrite hd_functional' in h0.
  rewrite nth_lt_compat in ih.
  erewrite nth_indep in ih.
  eapply h1.
  apply h0.
  apply ih.
  assumption.
  omega.

  apply lt_S_n in h2.
  apply lt_S_n in h3.
  apply lt_S_n in h4.
  erewrite (nth_indep l a _ h3).
  erewrite (nth_indep l a _ h4).
  erewrite <- nth_lt_compat. erewrite <- nth_lt_compat.
  apply ih; auto with arith.
  Grab Existential Variables.
  assumption.
  assumption.
  assumption.
Qed.


Lemma weak_incrR_le_nth_lt :
  forall {T:Type} (R:Relation T) (pftr:Transitive _ R)
         (l:list T),
    weak_increasingR R l ->
    forall (i j:nat) (pfij:i < j)
           (pfi:i < length l) (pfj:j < length l),
      Rw R (nth_lt l i pfi) (nth_lt l j pfj).
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros; omega.
  intros h0 i j h2 h3 h4.
  destruct (nil_dec' l) as [|h5]; subst.
  simpl in h3, h4.
  omega.
  rewrite (hd_compat' _ h5) in h0.
  destruct h0 as [h0 h6].
  rewrite <- (hd_compat' _ h5) in h6.
  destruct i as [|i], j as [|j].
  right; auto.
  destruct h0 as [h0 | h0].
  apply lt_S_n in h4.
  destruct j as [|j].
  left.
  erewrite nth_indep. erewrite <- nth_lt_compat.  rewrite nth_lt_0_hd'.
  erewrite hd_functional'.  apply h0.
  assumption.
  specialize (ih h6 0 _ (lt_0_Sn _ ) (not_nil_lt _ h5) h4).
  rewrite nth_lt_0_hd' in ih.
  erewrite hd_functional' in h0.
  pose proof (trans_R_Rw _ h1  _ _ _ h0 ih) as h7.
  left; auto.
  rewrite nth_lt_compat in h7.
  erewrite nth_indep.
  apply h7.
  assumption.
  apply lt_S_n in h4.
  subst.
  destruct j as [|j].
  right; auto.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite nth_lt_0_hd'.
  apply hd_functional'. 
  assumption.
  erewrite nth_indep. erewrite <- nth_lt_compat .
  erewrite hd_functional'.  
  rewrite <- (nth_lt_0_hd' l).
  apply ih; auto with arith.
  assumption.
  omega.
  apply lt_S_n in h3. apply lt_S_n in h4.
  erewrite (nth_indep l a _ h3).
  erewrite (nth_indep l a _ h4).
  erewrite <- nth_lt_compat. erewrite <- nth_lt_compat.
  apply ih; auto with arith.
  Grab Existential Variables.
  assumption.
  assumption.
  apply not_nil_lt; auto.
  assumption.
  assumption.
  assumption.
Qed.


Lemma weak_incrR_ub_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    weak_increasingR R l ->
    let pftso := sort_dec_tso_dec _ pfso in
    weak_increasingR R (ub_list R pftso l a).
  intros T R h1 l a h2 h3; eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec. apply h1.
  apply sort_dec_trans; auto.
  apply h2.
  apply coarse_list_ub_list.
Qed.

Lemma weak_incrR_wub_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    weak_increasingR R l ->
    let pftso := sort_dec_tso_dec _ pfso in
    weak_increasingR R (wub_list R pftso l a).
  intros T R h1 l a h2 h3; eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec. apply h1.
  apply sort_dec_trans; auto.
  apply h2.
  apply coarse_list_wub_list.
Qed.

Lemma weak_incrR_lb_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    weak_increasingR R l ->
    let pftso := sort_dec_tso_dec _ pfso in
    weak_increasingR R (lb_list R pftso l a).
  intros T R h1 l a h2 h3; eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec. apply h1.
  apply sort_dec_trans; auto.
  apply h2.
  apply coarse_list_lb_list.
Qed.

Lemma weak_incrR_wlb_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    weak_increasingR R l ->
    let pftso := sort_dec_tso_dec _ pfso in
    weak_increasingR R (wlb_list R pftso l a).
  intros T R h1 l a h2 h3; eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec. apply h1.
  apply sort_dec_trans; auto.
  apply h2.
  apply coarse_list_wlb_list.
Qed.


Lemma weak_incrR_app_sing_ub_list_eq_nin :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T),
    ~In a l ->
      let pftso := sort_dec_tso_dec _ pfso in
      weak_increasingR R (l++a::nil) ->
      ub_list R pftso l a = l.
  intros T R h1 l.
  induction l as [|c l ih]; auto. 
  intros a hnin htso h2.
  simpl.
  destruct (htso c a) as [h3 | h3].
  destruct h3 as [h3 |]; subst; auto.
  f_equal. apply ih.
  contradict hnin; right; auto.
  rewrite <- app_comm_cons in h2.
  apply weak_increasingR_cons_impl in h2; auto.
  contradict hnin; left; auto.
  rewrite <- app_comm_cons in h2.
  pose proof (@le_weak_increasingR_cons).
  pose proof (le_weak_increasingR_cons R (l++a::nil) a c (sort_dec_trans _ h1) (in_app_r (in_eq _ _) _) h2) as h4.
  pose proof (trans_Rw_R _ (sort_dec_trans _ h1) _ _ _ h4 h3) as h5.
  apply (sort_dec_irrefl _ h1) in h5; contradiction.
Qed.


Lemma weak_incrR_app_sing_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T),
         let pfdt := sort_dec_eq_dec _ pfso in 
      weak_increasingR R (l++a::nil) <->
      (weak_increasingR  R l /\
       (forall x:T, In x l -> Rw R x a)).
  intros T R h1 l a hdt.
  destruct l as [|b l]; simpl.
  intuition. 
  split; intro h2.
  split.  
  destruct l as [|c l]; auto.
  rewrite <- app_comm_cons in h2.
  destruct h2 as [h2 h3].
  split; auto.
  rewrite  app_comm_cons in h3.
  rewrite <- weak_increasingR_app_iff in h3.
  destruct h3; auto.
  apply sort_dec_trans; auto.
  intros x h3.   
  generalize dependent l. intro l. revert x a b.
  induction l as [|c l ih]; simpl.
  intuition.
  subst; auto.
  intros x a b h3 h2.
  destruct h3 as [h3 h4].
  pose proof (ih x a c h4) as ih'. 
  destruct h2 as [|h2]; subst.
  pose proof (ih c a c) as h5.
  eapply (trans_Rw _ (sort_dec_trans _ h1)).
  apply h3.
  apply h5; auto.
  apply ih'; auto.
 
  destruct h2 as [h2 h3].
  destruct l as [|c l]; simpl.
  split; auto.
  destruct h2 as [h2 h4].
  split; auto .
  change (weak_increasingR R (c::l++a::nil)).
  rewrite app_comm_cons.
  rewrite <- weak_increasingR_app_iff. 
  simpl; repeat split; auto.
  apply h3.
  right.
  destruct l as [|d l].
  left; auto.
  right.
  apply in_last.
  apply cons_neq_nil.
  apply sort_dec_trans; auto.
  Grab Existential Variables.
  apply cons_neq_nil.
  apply cons_neq_nil.
  apply cons_neq_nil.
  apply cons_neq_nil.
Qed.


Lemma weak_incrR_app_sing_iff_in :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T)
         (pfin:In a l),
      let pfdt := sort_dec_eq_dec _ pfso in 
      weak_increasingR R (l++a::nil) <->
      (weak_increasingR  R (firstn (lind pfdt l a pfin) l) /\
       (let i := lind pfdt l a pfin in 
        i > 0 -> Rw R (last (firstn i l) a) a) /\
       skipn_all_min pfdt l a pfin = lrep a (count_occ pfdt l a)).
  intros T R h1 l a h2 h3.
  rewrite weak_incrR_app_sing_iff; auto.
  split; intro h4.
  destruct h4 as [h4 h5].
  split.
  apply weak_increasingR_firstn; auto.
  apply sort_dec_trans; auto.
  split.
  intros i h6.
  apply h5.
  pose proof (in_last (firstn i l) a) as h7.
  hfirst h7.
  intro h8.
  apply firstn_eq_nil in h8.
  destruct h8 as [h8 | h8]; try omega.
  pose proof h2 as h2'. rewrite h8 in h2'; contradiction.
  specialize (h7 hf).
  rewrite in_firstn_iff' in h7.
  destruct h7; auto.
  symmetry.  
  rewrite lrep_eq_iff. 
  match goal with |- ?hyp1 /\ ?hyp2 => assert (h6:hyp1) end.
  intros x h6.
  unfold skipn_all_min in h6.
  rewrite (in_skipn_iff (sort_dec_eq_dec _ h1)) in h6.
  destruct h6 as [h6 [occ h7]].
  destruct h7 as [h7 h8].
  apply le_lt_eq_dec in h8.
  destruct h8 as [h8 | h8].
  pose proof (weak_incrR_le_nth_lt R (sort_dec_trans _ h1) _ h4 _ _ h8 (lt_lind _ _ _ _) (lt_lind_occ _ _ _ _ _)) as h9.
  rewrite lind_compat, lind_occ_compat in h9.
  pose proof (h5 _ h6).
  eapply antisym_Rw.
  apply sort_dec_so.
  apply h1. 
  assumption.
  assumption.  
  rewrite (lind_eq_iff h3 _ _ h2 _ (lt_lind_occ _ _ _ _ _)) in h8.
  destruct h8 as [h8 h9].
  rewrite lind_occ_compat in h8; auto.
  split; auto.
  pose proof (all_sing_length (sort_dec_eq_dec _ h1) _ _ h6) as h7.
  rewrite count_occ_skipn_all_min in h7; auto.
  
  destruct h4 as [h4 [h5 h6]].
  split; auto.
  rewrite <- (firstn_skipn (lind h3 l a h2)).
  destruct (zerop (lind h3 l a h2)) as [h7 | h7].
  rewrite h7.
  simpl.
  unfold skipn_all_min in h6. rewrite h7 in h6.
  simpl in h6.
  rewrite h6.
  apply weak_incrR_lrep.
  specialize (h5 h7).
  rewrite <- weak_increasingR_app_iff.
  split; auto.
  split.
  unfold skipn_all_min in h6. rewrite h6.
  apply weak_incrR_lrep.
  fold (skipn_all_min h3 l a h2).
  erewrite <- hd_hd_compat'. rewrite hd_skipn_all_min at 1.
  rewrite lst_compat'.
  erewrite last_indep.
  apply h5.
  intro h8.
  apply firstn_eq_nil in h8.
  destruct h8 as [h8 | h8].
  omega.
  subst.
  contradiction.
  apply sort_dec_trans; auto.
  intros x h7.
  destruct (zerop (lind h3 l a h2)) as [h8 | h8].
  unfold skipn_all_min in h6.
  rewrite h8 in h6.
  simpl in h6.
  rewrite h6 in h7.
  apply in_lrep' in h7.
  rewrite h7.
  right; auto.
  specialize (h5 h8).
  rewrite <- (firstn_skipn (lind h3 l a h2)) in h7.
  apply in_app_or in h7.
  destruct h7 as [h7 | h7].
  pose proof (weak_increasingR_last _ _ (sort_dec_trans _ h1) h4 _ h7) as h9.
  eapply trans_Rw.
  apply sort_dec_trans; auto.
  apply h9.
  erewrite last_indep.
  apply h5.
  intro h10.
  apply firstn_eq_nil in h10.
  destruct h10 as [h10 | h10].
  omega.
  subst.
  contradiction.
  unfold skipn_all_min in h6.
  rewrite h6 in h7.
  apply in_lrep' in h7; subst; right; auto.
  Grab Existential Variables.
  intro h8.
  rewrite skipn_eq_nil_iff in h8.
  destruct h8 as [|h8]; subst.
  contradiction.
  pose proof (lt_lind h3 l a h2).
  omega.
  intro h8.
  apply firstn_eq_nil in h8.
  destruct h8 as [h8|]; subst.
  omega.
  contradiction.
  eapply sort_dec_eq_dec.
  apply h1.
Qed.


Definition weak_incrR_add {T:Type} (R:Relation T) (pftso:tso_dec R)
           (l:list T) (a:T) : list T :=
  ub_list R pftso l a ++ a :: wlb_list R pftso l a.

Lemma weak_incrR_add_nil :
  forall {T:Type} (R:Relation T)(pftso:tso_dec R) (a:T),
    weak_incrR_add R pftso nil a = a :: nil.
  intros; unfold weak_incrR_add.
  rewrite ub_list_nil, wlb_list_nil; auto.
Qed.

Lemma weak_incrR_add_sing :
  forall {T:Type} (R:Relation T)(pftso:sort_dec R) (a b:T),
    let pftso := sort_dec_tso_dec _ pftso in
    weak_incrR_add R pftso (a::nil) b   =
    (minR R pftso a b) :: (maxR R pftso a b) :: nil.
  intros T R h1 a b htso.
  unfold weak_incrR_add, minR, maxR.
  destruct (htso a b) as [h2 | h2].
  destruct h2 as [h2|]; subst.
  unfold htso.
  rewrite ub_list_cons_gt, ub_list_nil; auto.
  repeat rewrite app_nil_l; simpl.
  f_equal. f_equal.
  fold htso.
  destruct (htso a b) as [h3 |h3]; auto.
  destruct h3 as [h3 | h3]; auto.
  subst.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  simpl.
  destruct (htso b b) as [h2 | h2].
  destruct h2 as [h2 | h2]; auto.
  apply (sort_dec_irrefl _ h1) in h2; contradiction.
  unfold htso; rewrite ub_list_cons_lt, ub_list_nil, app_nil_l; simpl; auto.
  fold htso.
  destruct (htso a b) as [h3 |h3]; auto.
  f_equal.
  destruct h3 as [h3 | h3].
  pose proof (sort_dec_trans _ h1 _ _ _ h2 h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4; contradiction.
  reflexivity.
Qed.

Lemma in_weak_incrR_add_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (weak_incrR_add R pftso l a) <->
    (x = a \/ In x l).
  unfold weak_incrR_add; simpl; intros T R h1 l a x; split; intro h2.
  apply in_app_or in h2.
  destruct h2 as [h2 | h2].
  rewrite in_ub_list_iff in h2.
  destruct h2 as [h2 h3].
  right; auto.
  destruct h2 as [|h2]; subst.
  left; auto.
  rewrite in_wlb_list_iff in h2.
  destruct h2; right; auto.
  destruct h2 as [|h2]; subst.
  apply in_app_r.
  left; auto.
  destruct (sort_dec_tso_dec _ h1 x a) as [h3 | h3].
  destruct h3 as [h3 | h3].
  apply in_app_l.
  rewrite in_ub_list_iff.
  split; auto.
  subst.
  apply in_app_r.
  left; auto.
  apply in_app_r.
  right.
  rewrite in_wlb_list_iff.
  split; auto.
  left; auto.
Qed.

  
Lemma weak_incrR_add_cons_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a c:T),
    R a c ->
    let pftso := sort_dec_tso_dec _ pfso in
    weak_incrR_add R pftso (a::l) c = a :: weak_incrR_add R pftso l c.
  intros T R h1 l a c hac h2.
  unfold weak_incrR_add, minR.
  simpl.
  destruct (h2 a c) as [h3 | h3].
  destruct h3 as [h3 | h3].
  rewrite <- app_comm_cons; auto.
  subst.
  apply (sort_dec_irrefl _ h1) in hac.
  contradiction.
  pose proof (sort_dec_trans _ h1 _ _ _ hac h3) as h4.
  apply (sort_dec_irrefl _ h1) in h4.
  contradiction.
Qed.



Lemma weak_incrR_add_cons_same :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    let pftso := sort_dec_tso_dec _ pfso in
    weak_incrR_add R pftso (a::l) a =
    ub_list R pftso l a ++ a::a::wlb_list R pftso l a.
  intros T R h1 l a h2.
  unfold weak_incrR_add.
  rewrite ub_list_cons_same, wlb_list_cons_same; auto.
  apply sort_dec_irrefl; auto.
  apply sort_dec_irrefl; auto.
Qed.

Lemma weak_incrR_add_cons_gt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a c:T),
    R c a ->
    let pftso := sort_dec_tso_dec _ pfso in
    weak_incrR_add R pftso (a::l) c = ub_list R pftso l c ++ (c :: a :: wlb_list R pftso l c).
  intros T R h1 l a c h2 h3.
  unfold weak_incrR_add.
  unfold h3. rewrite ub_list_cons_lt, wlb_list_cons_lt; auto.
Qed.



Lemma weak_incrR_add_in :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In a (weak_incrR_add R pftso l a).
  intros; unfold weak_incrR_add.
  apply in_app_r.
  left; auto.
Qed.

Lemma weak_incrR_add_in_l :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a x:T),
    In x l ->
    let pftso := sort_dec_tso_dec _ pfso in
    In x (weak_incrR_add R pftso l a).
  intros T R hso l a x h2 h3.
  unfold weak_incrR_add.
  destruct (h3 a x) as [h4 | h4].
  destruct h4 as [h4 | h4].
  apply in_app_r.
  right.
  unfold h3.
  rewrite in_wlb_list_iff.
  split; auto.
  left; auto.
  subst.
  apply in_app_r.
  left; auto.
  apply in_app_l.
  unfold h3.
  rewrite in_ub_list_iff.
  split; auto.
Qed.

Lemma length_weak_incrR_add :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    let pftso := sort_dec_tso_dec _ pfso in
    length (weak_incrR_add R pftso l a) = S (length l).
  intros; unfold weak_incrR_add; rewrite app_length.
  simpl.
  rewrite plus_S_shift_r.
  f_equal.
  symmetry.
  apply length_decompose_ub_wlb.
Qed.

Lemma count_occ_weak_incrR_add_eq :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    let pfdt := sort_dec_eq_dec _ pfso in 
    let pftso := sort_dec_tso_dec _ pfso in
    count_occ pfdt (weak_incrR_add R pftso l a) a = S (count_occ pfdt l a).
  intros T R h1 l a h2 h3.
  unfold weak_incrR_add.
  rewrite count_occ_app.
  simpl.
  destruct (h2 a a) as [h4 | h4].
  rewrite plus_S_shift_r.
  f_equal.
  symmetry; apply count_occ_ub_wlb_decompose.
  contradict h4; auto.
Qed.

Lemma count_occ_weak_incrR_add_neq :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a x:T),
    x <> a ->
    let pfdt := sort_dec_eq_dec _ pfso in 
    let pftso := sort_dec_tso_dec _ pfso in
    count_occ pfdt (weak_incrR_add R pftso l a) x = count_occ pfdt l x.
  intros T R h1 l a x h2 h3 h4.
  unfold weak_incrR_add.
  rewrite count_occ_app.
  simpl.
  destruct (h3 a x) as [h5 | h5].
  subst; contradict h2; auto.
  symmetry; apply count_occ_ub_wlb_decompose.
Qed.



Lemma coarse_list_weak_incrR_add_cons :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a c:T),
    let pftso := sort_dec_tso_dec _ pfso in 
    coarse_list (weak_incrR_add R pftso l c)
                (weak_incrR_add R pftso (a::l) c).
  intros T R h1 l a c h2.
  unfold weak_incrR_add.
  simpl.
  destruct (h2 a c) as [h3 | h3].
  destruct h3 as [h3 | h3].
  rewrite <- app_comm_cons.
  apply coarse_list_cons1.
  apply coarse_list_refl.
  subst. 
  rewrite <- (app_sing (c::wlb_list R h2 l c)).
  apply (coarse_list_app_interp_compat (sort_dec_eq_dec _ h1) (ub_list R h2 l c ++ c :: wlb_list R h2 l c) (c::wlb_list R h2 l c) (ub_list R h2 l c) (c::nil)).
  apply coarse_list_refl.  
  rewrite <- (app_sing (a::wlb_list R h2 l c)).
  pose proof (app_assoc (ub_list R h2 l c) (c::nil) (wlb_list R h2 l c)) as h4.
  simpl in h4; simpl.
  rewrite h4.
  rewrite <- (app_sing (a::wlb_list R h2 l c)) at 1.
  pose proof (app_assoc (ub_list R h2 l c) (c::nil) (a::wlb_list R h2 l c)) as h5.
  simpl in h5; simpl.
  rewrite h5.
  rewrite <- (app_sing (wlb_list R h2 l c)) at 1.
  apply (coarse_list_app_interp_compat (sort_dec_eq_dec _ h1)
                                       ((ub_list R h2 l c ++ c :: nil) ++ wlb_list R h2 l c)
                                       (wlb_list R h2 l c)
                                       (ub_list R h2 l c ++ c::nil)
                                       (a::nil)).
  apply coarse_list_refl.
Qed.



  

Lemma weak_incrR_add_compat : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T) (a:T),
    weak_increasingR R l ->
    let pftso := sort_dec_tso_dec _ pfso in
    weak_increasingR R (weak_incrR_add R pftso l a).
  intros T R h1 l a h2 h3.
  unfold weak_incrR_add.
  destruct (nil_dec' (ub_list R h3 l a)) as [h4 | h4].
  rewrite h4.
  rewrite app_nil_l.
  destruct (nil_dec' (wlb_list R h3 l a)) as [h5 | h5].
  rewrite h5; simpl; auto.
  apply (weak_increasingR_impl_cons _ _ h5).
  eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec. apply h1.
  apply sort_dec_trans; auto.
  apply h2.
  apply coarse_list_wlb_list.
  pose proof (in_hd' (wlb_list R h3 l a) h5) as h6.
  unfold h3 in h6. rewrite in_wlb_list_iff in h6. 
  unfold h3.
  destruct h6; auto.
  destruct (nil_dec' (wlb_list R h3 l a)) as [h5 | h5].
  rewrite h5.
  rewrite  <- (weak_increasingR_app_iff  R (sort_dec_trans _ h1) _ _  h4 (cons_neq_nil _ _)).
  repeat split.
  apply weak_incrR_ub_list; auto.
  simpl.
  pose proof (in_lst (ub_list R h3 l a) h4) as h6.
  unfold h3 in h6; rewrite in_ub_list_iff in h6.
  left.
  unfold h3.
  destruct h6; auto.

  rewrite  <- (weak_increasingR_app_iff  R (sort_dec_trans _ h1) _ _  h4 (cons_neq_nil _ _)).
  repeat split.
  apply weak_incrR_ub_list; auto.
  simpl.
  rewrite (hd_compat' _ h5).
  split.
  pose proof (in_hd' (wlb_list R h3 l a) h5) as h6.
  unfold h3 in h6; rewrite in_wlb_list_iff in h6.
  unfold h3.
  destruct h6; auto.
  rewrite <- hd_compat'.
  apply weak_incrR_wlb_list; auto.
  simpl.

  pose proof (in_lst (ub_list R h3 l a) h4) as h6.
  unfold h3 in h6; rewrite in_ub_list_iff in h6.
  left.
  unfold h3.
  destruct h6; auto.
Qed.


Fixpoint sort_list_aux {T:Type} (R:Relation T) (pftso:tso_dec R)
         (acc l:list T) : list T :=
  match l with
    | nil => acc
    | a::l' => sort_list_aux R pftso (weak_incrR_add R pftso acc a) l'
  end.

Definition sort_list  {T:Type} (R:Relation T) (pftso:tso_dec R)
           (l:list T) : list T :=
  sort_list_aux R pftso nil l.

Lemma sort_list_functional :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (pfso:sort_dec R)
         (l l':list T) (pfeq:l = l'),
    sort_list R pftso  l  =
    sort_list R (sort_dec_tso_dec _ pfso) l'.
  intros; subst.
  f_equal.
  symmetry; apply sort_dec_tso_dec_eq.
Qed.


Lemma sort_list_aux_nil :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (acc:list T),
    sort_list_aux R pftso acc nil = acc.
  auto.
Qed.

Lemma sort_list_nil :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R),
    sort_list R pftso nil = nil.
  auto.
Qed.

(*See less trivial primed version.*)
Lemma sort_list_cons :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T) (a:T),
    sort_list R pftso (a::l) =
    sort_list_aux R pftso (a::nil) l.
  auto.
Qed.

Lemma sort_list_sing :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (a:T),
    sort_list R pftso (a::nil) = a::nil.
  intros; unfold sort_list; simpl.
  rewrite weak_incrR_add_nil; auto.
Qed.


  
Lemma in_acc_sort_list_aux :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l:list T) (x:T),
    In x acc ->
    let pftso := sort_dec_tso_dec _ pfso in
    In x (sort_list_aux R pftso acc l).
  intros T R h1 acc l. revert acc.
  induction l as [|b l ih]; simpl; auto.
  intros acc x h2.
  apply ih.
  eapply weak_incrR_add_in_l; auto.
Qed.

Lemma in_sort_list_aux_app_l : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l l':list T)
         (x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x l ->
    In x (sort_list_aux R pftso acc (l++l')).
  intros T R h1 acc l. revert acc.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros acc l' x h2.
  destruct h2 as [|h2]; subst.
  apply in_acc_sort_list_aux.
  apply weak_incrR_add_in.
  apply ih; auto.
Qed.

Lemma in_sort_list_aux : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l:list T)
         (x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x l ->
    In x (sort_list_aux R pftso acc l).
  intros; rewrite <- (app_nil_r l).
  apply in_sort_list_aux_app_l; auto.
Qed.

Lemma in_sort_list_aux_app_r : 
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l l':list T)
         (x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x l' ->
    In x (sort_list_aux R pftso acc (l++l')). 
  intros T R h1 acc l. revert acc.
  induction l as [|a l ih]; simpl.
  intros; apply in_sort_list_aux; auto.
  intros; auto.
Qed.


Lemma in_sort_list_aux_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l:list T) (x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (sort_list_aux R pftso acc l) <->
    In x l \/ In x acc.
  intros T R h1 acc l x; split; intro h2.
  revert h2. revert x acc.
  induction l as [|a l ih]; simpl.
  intros; right; auto.
  intros x acc h2.
  apply ih in h2.
  destruct h2 as [h2 | h2].
  left; right; auto.
  unfold weak_incrR_add in h2.
  apply in_app_or in h2.
  destruct h2 as [h2 | h2]. 
  right.
  rewrite in_ub_list_iff in h2.
  destruct h2; auto.
  destruct h2 as [|h2]; subst.
  left; left; auto.
  right.
  rewrite in_wlb_list_iff in h2.
  destruct h2; auto.
  destruct h2 as [h2 | h2].
  apply in_sort_list_aux; auto.
  apply in_acc_sort_list_aux; auto.
Qed.


Lemma in_sort_list_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (x:T),
    let pftso := sort_dec_tso_dec _ pfso in
    In x (sort_list R pftso l) <->
    In x l.
  intros; unfold sort_list, pftso; rewrite in_sort_list_aux_iff; intuition.
Qed.


Lemma sort_list_aux_eq_nil_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l:list T),
    let pftso := sort_dec_tso_dec _ pfso in 
    sort_list_aux R pftso acc l = nil <-> l = nil /\ acc = nil.
  intros T R h1 acc l; split; intro h2.
  destruct (nil_dec' l) as [|h3]; subst.
  rewrite sort_list_aux_nil in h2.
  split; auto.
  destruct l as [|a l]; auto.
  rewrite (hd_compat' _ h3) in h2.
  simpl in h2.
  pose proof (weak_incrR_add_in R h1 acc a) as h4.
  simpl in h4.
  pose proof (in_acc_sort_list_aux R h1 _ l _  h4) as h5.
  simpl in h5.
  rewrite h2 in h5.
  contradiction.
  destruct h2; subst; auto.
Qed.


Lemma sort_list_eq_nil_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T),
    let pftso := sort_dec_tso_dec _ pfso in 
    sort_list R pftso l = nil <-> l = nil.
  simpl; intros; unfold sort_list.
  rewrite sort_list_aux_eq_nil_iff at 1.
  split; auto.
  intro h1; destruct h1; auto.
Qed.

  
Lemma weak_incrR_acc_sort_list_aux :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l:list T),
    weak_increasingR R acc ->
    let pftso := sort_dec_tso_dec _ pfso in 
    weak_increasingR R (sort_list_aux R pftso acc l).
  intros T R h1 acc l.
  revert acc.
  induction l as [|a l ih]; auto.
  intros acc h2 htso.
  apply ih.
  specialize (ih _ h2).
  unfold weak_incrR_add.
  pose proof (weak_incrR_coarse_list (sort_dec_eq_dec _ h1) _ (sort_dec_trans _ h1) _ _ h2 (coarse_list_ub_list R htso acc a)) as h3.
  pose proof (weak_incrR_coarse_list (sort_dec_eq_dec _ h1) _ (sort_dec_trans _ h1) _ _ h2 (coarse_list_wlb_list R htso acc a)) as h4.
  destruct (nil_dec' (ub_list R htso acc a)) as [h5 | h5].
  rewrite h5.
  rewrite app_nil_l.
  unfold htso in h5; rewrite ub_list_eq_nil_iff in h5. 
  destruct h5 as [|h5]; subst.
  simpl; auto.
  destruct h5 as [hn h5].
  destruct (nil_dec' (wlb_list R htso acc a)) as [h6 | h6].
  rewrite h6; simpl; auto.
  simpl.
  rewrite (hd_compat' _ h6).
  pose proof (in_hd' (wlb_list R htso acc a) h6) as h7.
  unfold htso in h7. rewrite in_wlb_list_iff in h7.
  unfold htso; destruct h7; auto.
  split; auto.
  rewrite <- hd_compat'.
  fold htso; auto.
  
  destruct (nil_dec' (wlb_list R htso acc a)) as [h6 | h6].
  rewrite h6; simpl; auto.
  unfold htso in h6; rewrite wlb_list_eq_nil_iff in h6.
  destruct h6 as [|h6]; subst.
  simpl in h5. contradict h5; auto. 
  destruct h6 as [hn h6]. 
  erewrite <- (weak_increasingR_app_iff _ _ _ _ h5 (cons_neq_nil _ _)); simpl.
  split; auto.
  split; auto.
  left.
  apply h6.
  pose proof (in_lst (ub_list R htso acc a) h5) as h7.
  unfold htso in h7. rewrite in_ub_list_iff in h7.
  unfold htso; destruct h7; auto.
  rewrite <- weak_increasingR_app_iff.
  repeat split; auto.
  pose proof (in_hd' _ h6) as h7.
  unfold htso in h7. rewrite in_wlb_list_iff in h7.
  destruct h7 as [h7 h8].
  simpl.
  rewrite (hd_compat' _ h6); simpl.  
  split; auto.
  rewrite (hd_compat' _ h6) in h4.
  simpl in h4; auto.
  simpl.
  pose proof (in_lst _ h5) as h7.
  unfold htso in h7. rewrite in_ub_list_iff in h7.
  destruct h7 as [h7 h8].
  left.
  apply h8.
  apply sort_dec_trans; auto.
  Grab Existential Variables.
  intro; discriminate.
  apply sort_dec_trans; auto.
Qed.

Lemma weak_incrR_sort_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T),
    let pftso := sort_dec_tso_dec _ pfso in 
    weak_increasingR R (sort_list R pftso l).
  intros T R h1 l.
  induction l as [|a l ih]; auto. 
  unfold sort_list; simpl; auto.
  simpl; simpl in ih.
  apply weak_incrR_acc_sort_list_aux.
  red; auto.
Qed.


Lemma length_sort_list_aux :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (acc l:list T),
    length (sort_list_aux R pftso acc l) =
    length acc + length l.
  intros T R htso acc l. revert acc.
  induction l as [|a l ih]; simpl; auto with arith.
  intro acc.
  rewrite ih.
  unfold weak_incrR_add.
  rewrite app_length; simpl.
  rewrite (length_decompose_ub_wlb _ htso acc a).
  omega.
Qed.


Lemma length_sort_list :
  forall {T:Type} (R:Relation T) (pftso:tso_dec R) (l:list T),
    length (sort_list R pftso l) = length l.
  unfold sort_list; intros; rewrite length_sort_list_aux; auto.
Qed.
  

Lemma list_to_set_sort_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T),
    let pftso := sort_dec_tso_dec _ pfso in
    list_to_set l = list_to_set (sort_list R pftso l).
  intros T R h1 l; apply Extensionality_Ensembles; red; split; red; intros x h2.
  rewrite <- list_to_set_in_iff in h2; rewrite <- list_to_set_in_iff.
  rewrite in_sort_list_iff; auto.
  rewrite <- list_to_set_in_iff in h2; rewrite <- list_to_set_in_iff.
  rewrite in_sort_list_iff in h2; auto.
Qed.


Lemma faml_to_fam_sort_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:faml T),
    let pfso' := lt_listR_sort_dec _ pfso in
    let pftso' := sort_dec_tso_dec _ pfso' in
    faml_to_fam (sort_list _ pftso' l) = 
    faml_to_fam l.
  intros T R h1 ll; apply Extensionality_Ensembles; red; split; red; intros x h2; apply in_faml_to_fam_impl in h2; destruct h2 as [l' [h2 h3]]; subst.
  apply in_faml_impl.
  rewrite in_sort_list_iff in h2; auto.
  apply in_faml_impl.
  rewrite in_sort_list_iff; auto.
Qed.

  

Lemma count_occ_sort_list_aux :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (acc l:list T) (x:T),
    let pfdt := sort_dec_eq_dec _ pfso in
    let pftso := sort_dec_tso_dec _ pfso in
    count_occ pfdt (sort_list_aux R pftso acc l) x = 
    count_occ pfdt l x + count_occ pfdt acc x.
  intros T R hso acc l. revert acc.
  induction l as [|a l ih].
  intros; simpl; auto.
  intros acc x hdt htso. 
  simpl. 
  destruct (hdt a x) as [|h2]; subst.
  rewrite ih.
  unfold weak_incrR_add.
  rewrite count_occ_app.
  simpl.
  fold hdt.
  destruct (hdt x x) as [h2 | h2].
  do 2 rewrite plus_S_shift_r.
  f_equal. f_equal.
  symmetry.
  apply count_occ_ub_wlb_decompose.
  contradict h2; auto.
  rewrite ih.
  unfold weak_incrR_add.
  rewrite count_occ_app; simpl.
  fold hdt.
  destruct (hdt a x) as [|h3].
  contradiction.
  f_equal.
  symmetry.
  apply count_occ_ub_wlb_decompose.
Qed.

Lemma count_occ_sort_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (x:T),
    let pfdt := sort_dec_eq_dec _ pfso in
    let pftso := sort_dec_tso_dec _ pfso in
    count_occ pfdt (sort_list R pftso l) x = 
    count_occ pfdt l x.
  simpl; intros; simpl; unfold sort_list;
  rewrite count_occ_sort_list_aux; simpl; auto with arith.
Qed.

Lemma weak_incrR_sort_list_eq :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T),
    let pftso := sort_dec_tso_dec _ pfso in
    weak_increasingR R l ->
    sort_list R pftso l = l.
  intros T R h1 l htso h2.
  eapply weak_incrR_length_count_occ_eq.
  apply h1.
  apply  weak_incrR_sort_list.
  assumption.
  apply length_sort_list.
  apply count_occ_sort_list.
Qed.

(*And the much more useful version!*)
Lemma sort_list_cons' :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T),
    let pftso := sort_dec_tso_dec _ pfso in
    sort_list R pftso (a::l) =
    weak_incrR_add R pftso (sort_list R pftso l) a.
  intros T R h1 l a htso.
  eapply (weak_incrR_length_count_occ_eq (sort_dec_eq_dec _ h1)).
  apply h1.
  apply weak_incrR_sort_list.
  apply weak_incrR_add_compat.
  apply weak_incrR_sort_list.
  rewrite length_sort_list.
  unfold htso.
  rewrite length_weak_incrR_add.
  rewrite length_sort_list; auto.
  simpl.
  intro x. 
  unfold htso. rewrite count_occ_sort_list.
  destruct (sort_dec_eq_dec _ h1 x a) as [|h2]; subst.
  rewrite count_occ_weak_incrR_add_eq.
  rewrite count_occ_sort_list; simpl.
  destruct (sort_dec_eq_dec R h1 a a) as [|h2]; auto.
  contradict h2; auto.
  rewrite count_occ_weak_incrR_add_neq; auto.
  rewrite count_occ_sort_list; simpl.
  destruct (sort_dec_eq_dec R h1 a x) as [|h3]; auto.
  subst.
  contradict h2; auto.
Qed.

Lemma sort_list_decompose_ub_wlb :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T),
    let pftso := sort_dec_tso_dec _ pfso in
    sort_list R pftso l = (ub_list R pftso (sort_list R pftso l) a)
                            ++
                            (wlb_list R pftso (sort_list R pftso l) a).
  intros T R h1 l a hts.
  eapply (weak_incrR_length_count_occ_eq (sort_dec_eq_dec _ h1)).
  apply h1.
  apply weak_incrR_sort_list.
  destruct (nil_dec' (ub_list R hts (sort_list R hts l) a)) as [h2 | h2].
  rewrite h2.
  rewrite app_nil_l.
  eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec.
  apply h1.
  apply sort_dec_trans; auto.
  apply weak_incrR_sort_list.
  apply coarse_list_wlb_list.
  destruct (nil_dec' (wlb_list R hts (sort_list R hts l) a)) as [h3 | h3].
  rewrite h3.
  rewrite app_nil_r.
  eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec.
  apply h1.
  apply sort_dec_trans; auto.
  apply weak_incrR_sort_list.
  apply coarse_list_ub_list.
  rewrite <- (weak_increasingR_app_iff _ (sort_dec_trans _ h1) _ _  h2 h3).
  split.
  eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec.
  apply h1.
  apply sort_dec_trans; auto.
  apply weak_incrR_sort_list.
  apply coarse_list_ub_list.
  split.
  eapply weak_incrR_coarse_list.
  eapply sort_dec_eq_dec.
  apply h1.
  apply sort_dec_trans; auto.
  apply weak_incrR_sort_list.
  apply coarse_list_wlb_list.
  left.
  eapply in_ub_wlb_lt.
  apply in_lst.
  apply in_hd'.
  rewrite app_length.
  apply length_decompose_ub_wlb.
  intro x.
  rewrite count_occ_app.
  apply count_occ_ub_wlb_decompose.
Qed.

Lemma no_dup_sort_list :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T),
    NoDup l ->
    let pftso := sort_dec_tso_dec _ pfso in
    NoDup (sort_list R pftso l).
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  rewrite sort_list_nil; auto.
  intro h2.
  pose proof (no_dup_cons_nin _ _ h2) as h3.
  pose proof (no_dup_cons _ _ h2) as h4.
  rewrite sort_list_cons'.
  unfold weak_incrR_add.
  apply no_dup_app.
  apply no_dup_ub_list.
  auto.
  constructor.
  contradict h3.
  rewrite in_wlb_list_iff in h3.
  destruct h3 as [h3].
  rewrite in_sort_list_iff in h3; auto.
  apply no_dup_wlb_list.
  auto.
  apply not_inhabited_empty.
  intro h5.
  destruct h5 as [x h5].
  destruct h5 as [x h5 h6].
  rewrite <- list_to_set_in_iff in h5, h6.
  pose proof h5 as h5'.
  rewrite in_ub_list_iffn in h5'.
  destruct h6 as [h6 | h6].
  subst.
  rewrite in_ub_list_iff in h5.
  destruct h5 as [? h5].
  apply (sort_dec_irrefl _ h1) in h5; auto.
  contradiction.
  rewrite in_ub_list_iff in h5.
  destruct h5; auto.
Qed.
  



Lemma lt_length_sort_list_cons :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (a:T)
         (i:nat)
         (pf:i < length (sort_list R (sort_dec_tso_dec _ pfso) (a::l))),
    a <> nth_lt _ _ pf ->
    pred i < length (sort_list R (sort_dec_tso_dec _ pfso) l).
  intros T R h1 l a i h2 h3.
  destruct i as [|i].
  revert h3. 
  generalize h2.
  rewrite sort_list_cons'. 
  intros h2' h3.
  destruct (nil_dec' l) as [|h4]; subst.
  simpl in h3.
  contradict h3; auto.
  simpl.
  rewrite length_sort_list.
  apply not_nil_lt in h4; auto.
  simpl.
  pose proof h2 as h2'.
  rewrite sort_list_cons' in h2'.
  rewrite length_weak_incrR_add in h2'.
  apply lt_S_n in h2'; auto.
Qed.


Fixpoint sort_list_perm {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (i:nat) :
  i < length (sort_list R (sort_dec_tso_dec _ pfso) l) -> nat :=
  match l with
    | nil => fun pf =>
               let pfn := length_functional _ _  (sort_list_nil R (sort_dec_tso_dec _ pfso)) in
               let pflt := lt_eq_trans _ _ _ pf pfn in
               False_rect nat (lt_n_0 _ pflt)
    | a::l' => fun pf:i < length (sort_list R (sort_dec_tso_dec _ pfso) (a::l')) =>
                 match (sort_dec_tso_dec _ pfso a (nth_lt _ _ pf)) with
                   | inleft (left pflt) =>
                     let pfn := lt_irrefl_neq _ _ _  (sort_dec_irrefl _ pfso) pflt in
                     let pflt' := lt_length_sort_list_cons R pfso l' a i pf pfn in
                     S (sort_list_perm R pfso l' (pred i) pflt')
                   | inleft (right pfe) => 0
                   | inright pfgt =>
                     let pfn := gt_irrefl_neq _ _ _  (sort_dec_irrefl _ pfso) pfgt in
                     let pfgt' := lt_length_sort_list_cons R pfso l' a i pf pfn in              
                     (sort_list_perm R pfso l' (pred i) pfgt')
                 end
  end.


Lemma sort_list_perm_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l:list T) (i:nat)
         (pf:i < length (sort_list R (sort_dec_tso_dec _ pfso) l)),
    sort_list_perm R pfso l i pf < length l.
  intros T R h1 l.
  induction l as [|a l ih]; simpl.
  intros ? h2.
  contradict (lt_n_0 _ h2).
  intros i h2.
  destruct (sort_dec_tso_dec _ h1 a (nth_lt _ _ h2)) as [h3|]; auto with arith.
  destruct h3 as [h3 | h3]; auto with arith.
Qed.


(*Returns a permutation from [sort_list l] to [l]
Assumes [NoDup], but not in the definition -- only behavior*)
Definition sort_list_perm_no_dup 
           {T:Type} (R:Relation T) (pfso:sort_dec R)
           (l:list T)
           (i:nat) (pflt:i < length (sort_list R (sort_dec_tso_dec _ pfso) l)) : nat :=
  let pfin := in_nth_lt _ _ pflt in
  let pfin' := iff1 (in_sort_list_iff _ pfso _ _) pfin in
  lind (sort_dec_eq_dec _ pfso) _ _ pfin'.

(*Assumes [NoDup] semnatically, but not overtly.*)
Definition list_perm_sort_no_dup 
           {T:Type} (R:Relation T) (pfso:sort_dec R)
           (l:list T)
           (i:nat) (pflt:i < length l) : nat :=
  let pfin := in_nth_lt _ _ pflt in
  let pfin' := iff2 (in_sort_list_iff _ pfso _ _) pfin in
  let pftso := sort_dec_tso_dec _ pfso in
  lind (sort_dec_eq_dec _ pfso) _ _ pfin'.

(*Assumes [NoDup] semnatically, but not overtly.*)
Lemma sort_list_perm_no_dup_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T)
         (i:nat) (pflt:i < length (sort_list R (sort_dec_tso_dec _ pfso) l)),
    sort_list_perm_no_dup R pfso l i pflt < length l.
  unfold sort_list_perm_no_dup; intros; eapply lt_eq_trans.
  apply lt_lind.
  reflexivity.
Qed.

(*Assumes [NoDup] semantically, but not overtly.*)
Lemma list_perm_sort_no_dup_lt :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T)
         (i:nat) (pflt:i < length l),
  list_perm_sort_no_dup R pfso l i pflt < length (sort_list R (sort_dec_tso_dec _ pfso) l).
  unfold list_perm_sort_no_dup; intros; eapply lt_eq_trans.
  apply lt_lind.
  reflexivity.
Qed.

(*Assumes [NoDup] semantically, but not overtly.*)
Lemma sort_list_perm_no_dup_compat :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T)
         (i:nat) (pflt:i < length (sort_list R (sort_dec_tso_dec _ pfso) l)),
    let pf' := sort_list_perm_no_dup_lt R pfso l i pflt in
    nth_lt (sort_list R (sort_dec_tso_dec _ pfso) l) i pflt =
    nth_lt l (sort_list_perm_no_dup R pfso l i pflt)
           (sort_list_perm_no_dup_lt _ _ _ _ _).
  unfold sort_list_perm_no_dup; intros.
  symmetry.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
Qed.

(*Assumes [NoDup] semantically, but not overtly.*)
Lemma list_perm_sort_no_dup_compat :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T)
         (i:nat) (pflt:i < length l),
    let pf' := list_perm_sort_no_dup_lt R pfso l i pflt in
    nth_lt l i pflt =
    nth_lt (sort_list R (sort_dec_tso_dec _ pfso) l)
           (list_perm_sort_no_dup R pfso l i pflt)
           (list_perm_sort_no_dup_lt _ _ _ _ _).
  unfold list_perm_sort_no_dup; intros.
  symmetry.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
Qed.


Lemma list_to_set_injable_sort_list_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T),
    let pftso := sort_dec_tso_dec _ pfso in
    list_to_set_injable (fun lx => lx = l) <->
    list_to_set_injable (fun lx => lx = sort_list _ pftso l).
  intros; split; intro h3; red in h3; red; intros; subst; auto.
Qed.


Definition sort_set {T:Type} (R:Relation T) (pfso:sort_dec R) 
           (E:Ensemble T) (pf:Finite E) : list T :=
  let pfso' := lt_listR_sortable _ pfso in
  let pfts := sort_dec_tso_dec _ pfso in
  min_set_sortable (list_reps E)
                   (finite_list_reps E pf)
                   (inh_list_reps E pf)
                   _ pfso'.


Definition is_sorted_set {T:Type} (l:list T) : Prop :=
  exists (R:Relation T) (pfso:sort_dec R), 
    l = sort_set R pfso (list_to_set l) (list_to_set_finite l).

 

Lemma sort_set_functional :
  forall {T:Type} (R:Relation T) (pfso pfso':sort_dec R) 
         (E E':Ensemble T) (pf:Finite E) (pf':Finite E'),
    E = E' ->
    sort_set R pfso E pf = sort_set R pfso' E' pf'.
  intros; subst.
  f_equal. apply sort_dec_eq. apply proof_irrelevance.
Qed.

Lemma in_list_reps_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (E:Ensemble T) (pf:Finite E),
    Inn (list_reps E) (sort_set R pfso E pf).
  intros T R h0 E h1. unfold sort_set.
  match goal with |- Inn _ (min_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5) => 
                  pose proof (min_set_sortable_compat x1 x2 x3 x4 x5) as h2 end.
  simpl in h2.
  destruct h2; auto.
Qed.


Lemma in_sort_set_iff :
  forall  {T:Type} (R:Relation T) (pfso:sort_dec R) 
          (E:Ensemble T) (pf:Finite E) (x:T),
    Inn E x <->
    In x (sort_set R pfso E pf).
  intros T R hs0 E h0 x.
  split.
  intro h2.
  unfold sort_set.
  match goal with |- In _ (min_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5) => 
                  pose proof (min_set_sortable_compat x1 x2 x3 x4 x5) as h1 end.
  destruct h1 as [h1 h3].
  destruct h1 as [h1]. destruct h1 as [h1 h4].
  rewrite h1 in h2.
  rewrite <- list_to_set_in_iff in h2.
  assumption.
  intro h2.
  unfold sort_set in h2.
  let P := type of h2 in match P with In _ (min_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5) => 
                  pose proof (min_set_sortable_compat x1 x2 x3 x4 x5) as h1 end.
   simpl in h1.
  destruct h1 as [h1 h3].
  eapply in_list_reps.
  apply h1. assumption.
Qed.

Lemma inj_sort_set :
  forall  {T:Type} (R:Relation T) (pfso:sort_dec R) 
          (A B:Ensemble T) (pfa:Finite A) (pfb:Finite B),
    sort_set R pfso A pfa =
    sort_set R pfso B pfb ->
    A = B.
  intros T R hs A B h1 h2 h3.
  apply Extensionality_Ensembles; red; split; red; intros x h4.
  rewrite in_sort_set_iff in h4.
  rewrite h3 in h4 at 1.
  rewrite <- in_sort_set_iff in h4; auto.
  rewrite in_sort_set_iff in h4.
  rewrite <- h3 in h4 at 1.
  rewrite <- in_sort_set_iff in h4; auto.
Qed.

(*Search below for [inj_sort_set']*)



Lemma no_dup_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) 
         (E:Ensemble T) (pf:Finite E),
    NoDup (sort_set R pfso E pf).
  intros T R hs E h1.
  unfold sort_set.
  match goal with |- NoDup (min_set_sortable ?pf1 ?pf2 ?pf3 ?pf4 ?pf5) => pose proof (min_set_sortable_compat pf1 pf2 pf3 pf4 pf5) as h2 end.
  simpl in h2.
  destruct h2 as [h2 h3].
  inversion h2 as [h4].
  destruct h4; auto.
Qed.


Lemma sort_set_compat :
  forall  {T:Type} (R:Relation T) (pfso:sort_dec R) 
          (E:Ensemble T) (pf:Finite E),
    list_to_set (sort_set R pfso E pf) = E.
  intros T R hs E h0.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  rewrite <- list_to_set_in_iff in h2.
  eapply in_list_reps.
  apply in_list_reps_sort_set. apply h2.
  rewrite <- list_to_set_in_iff.
  rewrite <- in_sort_set_iff; auto.
Qed.


Lemma sort_set_empty :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R),
    sort_set R pfso (Empty_set _) (Empty_is_finite _ ) =
    nil.
  unfold sort_set.
  intros T R hs.
  match goal with |- min_set_sortable ?pf1 ?pf2 ?pf3 ?pf4 ?pf5  = _  => pose proof (min_set_sortable_compat pf1 pf2 pf3 pf4 pf5) as h2 end.
  simpl in h2.
  destruct h2 as [h2].
  rewrite list_reps_empty in h2 at 1.
  inversion h2; auto.
Qed.

Lemma sort_set_nil :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
    (A:Ensemble T) (pf:Finite A),
    sort_set R pfso A pf = nil ->
    A = Empty_set _.
  intros T R hs A h1 h2.
  apply Extensionality_Ensembles; red; split; auto with sets.
  red.
  intros x h3.
  erewrite in_sort_set_iff in h3.
  rewrite h2 in h3 at 1.
  red in h3; contradiction.
Qed.


Lemma inh_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (pf:Finite A),
    Inhabited A ->
    sort_set R pfso A pf <> nil.
  intros T R hs A h1 h2.
  destruct h2 as [a h2].
  rewrite (in_sort_set_iff R hs A h1) in h2.
  eapply in_not_nil; apply h2.
Qed.



Lemma sort_set_sing :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (x:T) (pf:Finite (Singleton x)),
    sort_set R pfso (Singleton x) pf = x::nil.
  intros T R hs n h1.
  unfold sort_set.
  match goal with |- min_set_sortable _ ?pf1 ?pf2 _ _ = _ => generalize pf1 pf2 end.
  rewrite list_reps_sing.
  intros h2 h3. rewrite min_set_sortable_sing at 1; auto.
Qed.

Lemma nin_hd_tl_sort_set' :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (pff:Finite A)
         (pfia:Inhabited A),
    let l := sort_set R pfso A pff in
    let pfic := inh_sort_set R pfso A pff pfia in
    ~In (hd' l pfic) (tl l).
  intros T R hs A h1 h2.
  simpl.
  apply nin_hd_tl_no_dup'.
  apply no_dup_sort_set.
Qed.


Lemma hd_sort_set' :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (pff:Finite A)
         (pfi:Inhabited A),
    let pfso' := sort_dec_sortable _ pfso in
    hd' (sort_set R pfso A pff)
        (inh_sort_set R pfso A pff pfi) =
    min_set_sortable A pff pfi R pfso'.
  intros T R hs A h1 h2; simpl.
  unfold sort_set, min_set_sortable.
  pose proof h2 as h2'.
  destruct h2' as [a h2'].
  rewrite <- (hd_hd_compat' _ _ a).
  unfold sort_set, min_set_sortable.
  destruct constructive_definite_description as [l h3]. simpl.
  destruct h3 as [h3 h4].
  destruct constructive_definite_description as [a' h5]; simpl.
  destruct h5 as [h5 h6]. 
  pose proof (h6 (hd a l)) as h6'.
  pose proof h3 as h3'.
  destruct h3' as [h3']. destruct h3' as [? h3']; subst.
  destruct (sort_dec_tso_dec _ hs  (hd a l) a') as [h7 | h7]; auto.
  destruct h7 as [h7 | h7].
  hfirst h6'.
  constructor.
  pose proof (inh_not_nil_in_list_reps _ _ h2 h3) as h8.
  rewrite (hd_hd_compat' _ h8).
  rewrite <- list_to_set_in_iff. 
  apply in_hd'.  
  rewrite in_sing_iff.  
  intro; subst.
  apply (sort_dec_irrefl _ hs) in h7; auto. 
  specialize (h6' hf).
  pose proof (sort_dec_trans _ hs _ _ _ h6' h7) as hi.
  apply (sort_dec_irrefl _ hs) in hi.
  contradiction.
  assumption.
  specialize (h4 (a'::remove (sort_dec_eq_dec _ hs)  a' l)).
  hfirst h4.
  constructor.
  constructor. split.
  simpl.
  rewrite <- subtract_remove_compat.
  rewrite add_sub_compat_in.
  destruct h3 as [h3]. destruct h3; auto.
  assumption.
  constructor. 
  rewrite in_remove_iff.
  intro h8.
  destruct h8 as [h8 h9].
  contradict h9; auto.
  apply no_dup_remove. 
  assumption. 
  rewrite in_sing_iff. 
  intro h8.
  rewrite (hd_compat' l (inh_not_nil_in_list_reps _ _ h2 h3)) in h8.
  inversion h8.
  rewrite <- (hd_hd_compat' l _ a) in h8.
  inversion h8.
  lr_if  H3. fold hlr in H3. destruct hlr as [hl | hr].
  subst.
  rewrite <- hd_hd_compat' in h7 at 1.
  apply (sort_dec_irrefl _ hs) in h7; auto.
  contradict hr; auto.
  specialize (h4 hf).
  rewrite (hd_compat' l (inh_not_nil_in_list_reps _ _ h2 h3)) in h4 at 1.
  red in h4.
  destruct h4 as [h4 h9].
  simpl in h4.
  destruct ( sort_dec_tso_dec R hs
           (hd' l (inh_not_nil_in_list_reps (list_to_set l) l h2 h3)) a') as [h10 | h10].
  destruct h10 as [h10 | h10].
  pose proof (sort_dec_trans _ hs _ _ _ h10 h7) as h11.
  rewrite <- hd_hd_compat' in h11 at 1.
  apply (sort_dec_irrefl _ hs) in h11.
  contradiction.
  subst.
  apply hd_hd_compat'.
  contradiction.
Qed.

Lemma in_min :
    forall {T:Type} (R:Relation T) (pfso:sort_dec R)
           (A:Ensemble T) (pff:Finite A) (pfi:Inhabited A) (x:T),
      Inn A x ->
      let pfso' := sort_dec_sortable _ pfso in
      x = min_set_sortable A pff pfi R pfso' \/
      In x (tl (sort_set R pfso A pff)).
  intros T R hso A h1 h2 x h3.
  pose (inh_sort_set R hso A h1 h2) as h4.
  rewrite (in_sort_set_iff R hso A h1) in h3.
  simpl.
  rewrite (hd_compat' _ h4) in h3.
  unfold h4 in h3.
  rewrite hd_sort_set' in h3.
  destruct h3; [left | right]; auto.
Qed.


Lemma tl_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (pff:Finite A) (pfi:Inhabited A),
    let pfso' := sort_dec_sortable _ pfso in
    let pft := sort_dec_eq_dec _ pfso in 
    tl (sort_set R pfso  A pff) =
    remove pft (min_set_sortable A pff pfi R pfso') (sort_set R pfso A pff).
  intros T R h0 A hf hi.
  simpl.
  pose proof (finite_set_list_no_dup _ hf) as h2.
  destruct h2 as [l h2]. destruct h2 as [h2 h3]; subst.
  rewrite <- hd_sort_set'.
  rewrite remove_hd_no_dup'; auto.
  apply no_dup_sort_set.
Qed.

Lemma incrR_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (C:Ensemble T) (pf:Finite C),
    increasingR R (sort_set R pfso C pf).
  intros T R hso C h1.
  unfold sort_set, min_set_sortable.
  destruct constructive_definite_description as [l h2].
  simpl.
  destruct h2 as [h2 h3].
  revert h3 h2 h1. revert  C.
  induction l as [|a l ih]; simpl.
  intros ? ? h1. 
  intros; auto.
  intros C h1 h2 h3.
  destruct h2 as [h2]. destruct h2 as [h2 hn]; subst.
  pose proof (no_dup_cons _ _ hn) as hn'.
  pose proof (no_dup_cons_nin _ _ hn) as h7.
  specialize (ih (Subtract (Add (list_to_set l) a) a)).
  hfirst ih.
  intros l' h4.
  destruct h4 as [h4 h5].
  destruct h4 as [h4].
  destruct h4 as [h4 h6]; subst.
  simpl in h4.
  rewrite sub_add_same_nin in h4.
  specialize (h1 (a::l')).
  rewrite in_sing_iff in h5.
  hfirst h1.
  constructor. constructor. split.
  simpl. f_equal; auto.
  constructor.
  intro h8.
  rewrite list_to_set_in_iff in h8.
  rewrite <- h4 in h8.
  rewrite <- list_to_set_in_iff in h8.
  contradiction.
  assumption.
  intro h8.
  rewrite in_sing_iff in h8.
  inversion h8; subst.
  contradict h5; auto.
  specialize (h1 hf).
  red in h1.
  destruct h1 as [h1 h9].
  simpl in h1.
  destruct (sort_dec_tso_dec R hso  a a) as [h10 | h10].
  destruct h10 as [h10 | h10].
  apply (sort_dec_irrefl _ hso) in h10. contradiction.
  red. split; auto.
  contradiction.
  contradict h7.
  rewrite <- list_to_set_in_iff in h7.
  assumption.
  specialize (ih hf).
  hfirst ih.
  constructor. split.
  rewrite sub_add_same_nin; auto.
  contradict h7.
  rewrite list_to_set_in_iff; auto.
   assumption.
  specialize (ih hf0 (subtract_preserves_finite _ _ h3)).
  destruct l as [|a' l].
  simpl; auto.
  simpl.
  simpl in ih.
  split; auto.
  assert (h8:a <> a').
    intro; subst.
    contradict h7; left; auto.
  specialize (h1 (a'::a::l)).
  hfirst h1.
  constructor.
  constructor. split.
  simpl.
  apply add_comm.
  constructor.
  intro h9.
  destruct h9 as [h9 | h9].
  contradiction.
  pose proof (no_dup_cons_nin _ _ hn'). contradiction.
  constructor.
  contradict h7.
  right; auto.
  pose proof (no_dup_cons _ _ hn'); auto.
  rewrite in_sing_iff.
  intro h9.
  inversion h9; subst.
  contradict h8; auto.
  specialize (h1 hf1).
  red in h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  destruct (sort_dec_tso_dec R hso a a') as [h9 | h9].
  destruct h9 as [h9 | h9]; auto.
  subst.
  contradict h8; auto.
  contradiction.
Qed.


Lemma incrR_impl_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T),
    increasingR R l ->
    l = sort_set R pfso (list_to_set l) (list_to_set_finite _).
  intros T R hso l h1.
  pose proof (increasingR_no_dup (sort_dec_eq_dec _ hso) R l (sort_dec_so _ hso )h1) as hn.
  unfold sort_set, min_set_sortable.
  destruct constructive_definite_description as [l' h2]; simpl.
  destruct h2 as [h2 h3].
  destruct h2 as [h2].
  destruct h2 as [h2 h4].
  destruct (list_eq_dec (sort_dec_eq_dec _ hso) l l') as [h5 | h5]; auto.
  specialize (h3 l).
  hfirst h3.
  constructor. constructor. split; auto.
  rewrite in_sing_iff; auto.
  specialize (h3 hf).
  eapply incrR_no_dup_minimal; auto.
  red in h3.
  destruct h3; auto.
  apply (sort_dec_eq_dec _ hso).
  apply h1.
  apply lt_list_impl_leR.
  apply h3.
Qed.

Lemma is_sorted_set_cons :
  forall {T:Type} (l:list T) (a:T),
    is_sorted_set (a::l)  ->
    is_sorted_set l.
  intros T l m h1.
  red in h1. red.
  destruct h1 as [R [hso h2]].
  pose proof (incrR_sort_set R hso (list_to_set (m::l)) (list_to_set_finite _ )) as h3.
  rewrite <- h2 in h3.
  pose proof (increasingR_cons R l _  h3) as h4.
  exists R, hso.
  apply incrR_impl_sort_set; auto.
Qed.


Lemma sort_set_cons :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (pf:Finite A) (ln:list T) (a:T),
    sort_set R pfso A pf = a::ln ->
    ln = sort_set R pfso (Subtract A a) (subtract_preserves_finite _ _ pf).
  intros T R hso A h1 ln a h2.
  pose proof (f_equal list_to_set h2) as h3.
  rewrite sort_set_compat in h3; subst.
  simpl.
  pose proof (incrR_sort_set R hso  (list_to_set (a::ln)) h1) as h3.
  rewrite h2 in h3.
  pose proof (increasingR_cons_nin (sort_dec_eq_dec _ hso) R _ _ (sort_dec_so _ hso) h3) as h4.
  symmetry.
  gen0.
  rewrite sub_add_same_nin.
  intro h5.
  symmetry.
  erewrite sort_set_functional.
  apply incrR_impl_sort_set.
  eapply increasingR_cons.
  apply h3.
  reflexivity.
  rewrite <- list_to_set_in_iff; auto.
  Grab Existential Variables.
  assumption.
Qed.

Lemma lst_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T)
         (pf:Finite A) (pfi:Inhabited A),
    let pfso' := sort_dec_sortable _ pfso in
    let pfn := inh_sort_set R pfso _ _ pfi in
    lst (sort_set R pfso A pf) pfn =
    max_set_sortable A pf pfi R pfso'.
  intros T R h1 A h2 h3.
  pose proof (incrR_sort_set R h1 A h2) as h4.
  simpl.
  match goal with |- _ = max_set_sortable ?x1 ?x2 ?x3 ?x4 ?x5  =>
                  pose proof (max_set_sortable_compat x1 x2 x3 x4 x5) as h6 end.
  simpl in h6.
  destruct h6 as [h6 h7].
  rewrite (in_sort_set_iff R h1 A h2) in h6.
  rewrite lst_compat'.
  pose proof (increasingR_last _ _ (sort_dec_trans _ h1) h4 (max_set_sortable A h2 h3 R (sort_dec_sortable R h1)) h6) as h5.
  let P := type of h5 in match P with Rw R ?x ?y => destruct (sort_dec_eq_dec _ h1 x y) as [h8|h8] end.
  rewrite h8.
  apply last_indep.
  apply inh_sort_set; auto.
  let P := type of h5 in match P with Rw R _ ?x => specialize (h7 x) end.
  hfirst h7.
  constructor.
  rewrite in_sort_set_iff.
  apply in_last.
  apply inh_sort_set; auto.
  rewrite in_sing_iff; auto.
  specialize (h7 hf).
  pose proof (trans_R_Rw R (sort_dec_trans _ h1) _ _ _ h7 h5) as h9.
  apply (sort_dec_irrefl _ h1) in h9.
  contradiction.
Qed.

Lemma max_set_sortable_add_lst_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (x:T)
         (pf:Finite A) (pfi:Inhabited A),
    ~Inn A x ->
    let pf' := Add_preserves_Finite _ A x pf in
    let pfi' := inh_add A x in
    let pfx := inh_sort_set R pfso _ _ pfi in
    let pfso' := sort_dec_sortable _ pfso in
    x = max_set_sortable (Add A x) pf'  pfi' R pfso' ->
    R (lst (sort_set R pfso A pf) pfx) x.
  intros T R hso A x h0 h1 h2 h3 h4 h5  h6 h20.
  rewrite h20.
  pose proof (max_set_sortable_compat (Add A x) h3 h4 R h6) as h7.
  simpl in h7.
  destruct h7 as [h7 h8].
  apply h8.
  constructor.
  left.
  rewrite in_sort_set_iff.
  apply in_lst.
  rewrite in_sing_iff.
  intro h9.
  unfold h6 in h9, h20.
  rewrite <- h20 in h9 at 1.
  contradict h2.
  rewrite in_sort_set_iff.
  rewrite h9.
  apply in_lst.
Qed.

Lemma sort_set_app_add :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A A':Ensemble T)
         (pfa:Finite A) (pfa':Finite A') (pfi:Inhabited A') (a:T),
    ~Inn A a ->
    A' = Add A a ->
    a = max_set_sortable  A' pfa' pfi R (sort_dec_sortable _ pfso) ->
    sort_set R pfso A' pfa' =
    sort_set R pfso A pfa ++ a::nil.
  intros T R hso A A' h1 h2 h3 a h0 h4 h5.
  pose proof (finite_inh_or_empty _ h1) as hf.
  destruct hf as [hf |hf].
  pose proof (finite_set_list_no_dup _ h1) as h6.
  destruct h6 as [l h6]. destruct h6 as [h6 h7].
  rewrite h6 in h4.
  gen0.
  rewrite h4.
  intro h8.
  pose proof (list_to_set_app l (a::nil)) as h9.
  simpl in h9.
  rewrite add_empty_sing in h9.
  gen0.
  unfold Add. rewrite <- h9.
  intro h10. 
  erewrite sort_set_functional.
  symmetry.
  eapply incrR_impl_sort_set. 
  erewrite <- increasingR_app_iff.
  split.
  apply incrR_sort_set.
  split.
  red; auto.
  apply max_set_sortable_add_lst_sort_set.
  assumption.
  rewrite h5.
  apply max_set_sortable_functional; auto.
  rewrite <- h5.
  rewrite h6.
  assumption.
  apply sort_dec_trans; auto.
  rewrite h9.
  rewrite list_to_set_app.
  rewrite sort_set_compat.
  simpl.
  f_equal; auto.
  rewrite <- add_empty_sing; auto.
  rewrite hf in h4.
  rewrite add_empty_sing in h4.
  clear h5. subst.
  rewrite sort_set_sing.
  erewrite sort_set_functional.
  rewrite sort_set_empty at 1.
  rewrite app_nil_l.
  reflexivity.
  reflexivity.
  Grab Existential Variables.
  assumption.
  assumption.
  apply cons_neq_nil.
  assumption.
Qed. 


Lemma tl_sort_set_add :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (c:T)
         (pf:Finite (Add A c)),
    list_to_set (tl (sort_set R pfso (Add A c) pf)) =
    Subtract (Add A c) (min_set_sortable  _ pf (inh_add _ _) R (sort_dec_sortable _ pfso)).
  intros T R hso A c h1.
  pose proof (no_dup_sort_set R hso (Add A c) h1) as hn.
  apply Extensionality_Ensembles; red; split; red; intros m h3.
  rewrite <- list_to_set_in_iff in h3.
  pose proof (min_set_sortable_compat (Add A c) h1 (inh_add _ _) R (sort_dec_sortable _ hso)) as h4.
  simpl in h4.
  destruct h4 as [h4 h5].
  pose proof h4 as h4'.
  rewrite (in_sort_set_iff R hso _ h1 _) in h4'.
  pose proof (nin_hd_tl_sort_set' R hso _ h1 (inh_add _ _)) as h6.
  simpl in h6.
  rewrite hd_sort_set' in h6.
  pose proof (in_tl _ _ h3) as h8.
  rewrite list_to_set_in_iff in h8.
  rewrite sort_set_compat in h8.
  constructor; auto.
  rewrite in_sing_iff.
  intro; subst.
  contradiction.

  destruct h3 as [h3 h4].
  rewrite in_sing_iff in h4.
  rewrite <- list_to_set_in_iff.
  pose proof (hd_compat' (sort_set R hso (Add A c) h1) (inh_sort_set R hso _ _ (inh_add _ _))) as h5.
  simpl in h5.
  rewrite hd_sort_set' in h5.
  rewrite (in_sort_set_iff R hso _ h1) in h3.
  rewrite h5 in h3.
  destruct h3; try contradiction; auto.
Qed.


Lemma in_tl_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (pf:Finite A) (x y:T),
    Inn A x ->
    Inn A y ->
    R x y ->
    In y (tl (sort_set R pfso A pf)).
  intros T R hso A h1.
  pose proof (finite_set_list_no_dup _  h1) as h2.
  destruct h2 as [l [h2 h3]]; subst.
  revert h1 h3.
  induction l as [|c l ih]; simpl.
  intros ? ? ? ? h1.
  contradict h1.
  intros h1 h2 x y h3 h4 h5. 
  pose proof (no_dup_cons_nin _ _ h2) as h10.
  hfirst ih.
  apply list_to_set_finite.
  pose proof (no_dup_cons _ _ h2) as h6.
  specialize (ih hf h6).
  rewrite list_to_set_in_iff.
  rewrite tl_sort_set_add.
  destruct h4 as [y h4 | y h4], h3 as [x h3 | x h3]; subst.
  assert (hx:c <> x).
    intro; subst.
    contradict h10. rewrite list_to_set_in_iff; auto.
  assert (hy:c <> y).
    intro; subst.
    contradict h10. rewrite list_to_set_in_iff; auto. 
  specialize (ih x y h3 h4 h5).
  constructor.
  left; auto.
  rewrite in_sing_iff.  
  intro; subst.
  pose proof (min_set_sortable_compat (Add (list_to_set l) c) h1 (inh_add (list_to_set l) c) R (sort_dec_sortable _ hso)) as h7.
  simpl in h7.
  destruct h7 as [h7 h8]. 
  specialize (h8 x).
  hfirst h8.
  constructor. left; auto.
  rewrite in_sing_iff.
  intro; subst.
  apply (sort_dec_irrefl _ hso) in h5.
  contradiction.
  specialize (h8 hf0).
  pose proof (sort_dec_trans _ hso _ _ _ h8 h5) as hi.
  apply (sort_dec_irrefl _ hso) in hi; contradiction.
  rewrite in_sing_iff in h3; subst.
  constructor. left; auto.
  rewrite in_sing_iff.
  intro; subst.
  pose proof (min_set_sortable_compat (Add (list_to_set l) x) h1 (inh_add (list_to_set l) x) R (sort_dec_sortable _ hso)) as h7.
  simpl in h7.
  destruct h7 as [h7 h8].
  specialize (h8 x).
  hfirst h8.
  constructor. right. constructor.
  rewrite in_sing_iff.
  intro he. rewrite he in h5.
  apply (sort_dec_sortable _ hso) in h5; contradiction.
  specialize (h8 hf0).
  pose proof (sort_dec_trans _ hso _ _ _ h8 h5) as hi.
  apply (sort_dec_irrefl _ hso) in hi.
  contradiction.
  destruct h4.
  constructor.
  right. constructor.
  rewrite in_sing_iff.
  intro h7.
  pose proof (min_set_sortable_compat (Add (list_to_set l) c) h1 (inh_add (list_to_set l) c) R (sort_dec_sortable _ hso)) as h11.
  simpl in h11.
  destruct h11 as [h11 h12].
  specialize (h12 x).
  hfirst h12.
  constructor.
  left; auto.
  rewrite in_sing_iff.
  intro h13.
  rewrite <- h13 in h3.
  rewrite  h7 in h3 at 1.
  contradict h10.
  rewrite list_to_set_in_iff; auto.
  specialize (h12 hf0).
  rewrite h7 in h12.
  pose proof (sort_dec_trans _ hso _ _ _ h12 h5) as hi.
  apply (sort_dec_irrefl _ hso) in hi.
  contradiction.
  destruct h3, h4.
  apply (sort_dec_irrefl _ hso) in h5.
  contradiction.
Qed.


Lemma list_minus_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A B:Ensemble T) (pfa:Finite A) (pfb:Finite B)
         (pfab:Finite (Setminus A B)),
    list_minus (sort_dec_eq_dec _ pfso)
               (sort_set R pfso A pfa)
               (sort_set R pfso B pfb) =
    sort_set R pfso (Setminus A B) pfab.
  intros T R hso A B h1 h2 h3.
  pose proof (list_to_set_list_minus (sort_dec_eq_dec _ hso) (sort_set R hso A h1) (sort_set R hso B h2)) as h4.
  do 2 rewrite sort_set_compat in h4.
  symmetry.
  gen0.
  rewrite <- h4.
  intro h5.
  symmetry.
  pose proof (proof_irrelevance _ h5
                                (list_to_set_finite _)); subst.
  apply incrR_impl_sort_set.
  apply incrR_list_minus.
  apply sort_dec_trans; auto.
  apply incrR_sort_set.
Qed.


Definition fam_to_ens_sort_set {T:Type} (R:Relation T)
           (pfso:sort_dec R)
           (F:Family T) (pf:finite_mem F) :
  Ensemble (list T) :=
  Im (full_sig F) (fun S => sort_set R pfso _ (pf _ (proj2_sig S))).


Lemma fam_to_ens_sort_set_empty_eq :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (pf:finite_mem (Empty_set (Ensemble T))),
    fam_to_ens_sort_set R pfso _ pf = Empty_set _.
  intros.
  unfold fam_to_ens_sort_set.
  rewrite full_empty_sig_empty.
  apply image_empty.
Qed.

Lemma fam_to_ens_sort_set_sing_eq :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A:Ensemble T) (pf:Finite A) (pffm:finite_mem (Singleton A)),
    fam_to_ens_sort_set R pfso (Singleton A) pffm =
    Singleton (sort_set R pfso A pf).
  intros.
  unfold fam_to_ens_sort_set.
  rewrite full_sig_sing.
  rewrite im_singleton. simpl.
  f_equal. f_equal. apply proof_irrelevance.
Qed.


Lemma finite_fam_to_ens_sort_set :
  forall  {T:Type} (R:Relation T) (pfso:sort_dec R)
          (F:Family T) (pf:finite_mem F),
    Finite F ->
    Finite (fam_to_ens_sort_set R pfso F pf).
  intros T R hso F h1 h2.
  unfold fam_to_ens_sort_set.
  apply finite_image.
  rewrite <- finite_full_sig_iff; auto.
Qed.

Lemma inh_fam_to_ens_sort_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (F:Family T) (pf:finite_mem F),
    Inhabited F ->
    Inhabited (fam_to_ens_sort_set R pfso F pf).
  intros T R hso F h1 h2.
  destruct h2 as [x h2].
  apply Inhabited_intro with (sort_set R hso _ (h1 _ h2)).
  apply Im_intro with (exist _ _ h2).
  constructor. simpl.
  reflexivity.
Qed.


Lemma fam_to_ens_sort_set_functional :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (F F':Family T) (pf:finite_mem F) (pf':finite_mem F'),
    F = F' ->
    fam_to_ens_sort_set R pfso F  pf  =
    fam_to_ens_sort_set R pfso F' pf'.
  intros T R hso F F' h1 h2 h3.
  subst.
  pose proof (proof_irrelevance _ h1 h2); subst; auto.
Qed.


Lemma fam_to_ens_sort_set_subtract :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (F:Family T) (pff:Finite F) (pffm:finite_mem F) (l:list T),
    Inn F (list_to_set l) ->
    increasingR R l ->
    Subtract (fam_to_ens_sort_set R pfso F pffm) l =
    fam_to_ens_sort_set R pfso (Subtract F (list_to_set l))
                            (finite_mem_subtract _ _ pffm).
  intros T R hso F h1 h0 l h2 hi.
  unfold fam_to_ens_sort_set.
  match goal with |- Subtract (Im _ ?f) _  = _ =>
                  pose proof (im_full_sig_sub_decompose_inj (@ens_eq_dec T) F (list_to_set l) f h2) as h3 end.
  simpl in h3.
  rewrite h3.
  pose proof (incrR_impl_sort_set R hso l hi) as h4.
  rewrite h4 at 7.
  pose proof (proof_irrelevance _ (h0 (list_to_set l) h2)
                                (list_to_set_finite _)) as h6.
  rewrite h6.
  rewrite sub_add_eq at 1.
  rewrite subtract_nin.
  apply im_ext_in.
  intros x h7.
  unfold restriction_sig. simpl.
  apply sort_set_functional; auto.
  intro h7.
  clear h4.
  inversion h7 as [x h8 q h9]. subst.
  pose proof (incrR_impl_sort_set R hso _  hi) as h4.
  rewrite <- h4 in h9.
  unfold restriction_sig in h9.
  simpl in h9.
  destruct x as [x h10].
  simpl in h9.
  destruct h10 as [h10 h11].
  pose proof h11 as h11'. rewrite in_sing_iff in h11'.
  pose proof (f_equal (@list_to_set _) h9) as h12.
  rewrite sort_set_compat in h12.
  contradiction.
Qed.

Lemma im_fam_to_ens_sort_set_list_to_set :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (F:Family T) (pf:finite_mem F),
    Im (fam_to_ens_sort_set R pfso F pf) list_to_set = F.
  intros T R hso F h1.
  unfold fam_to_ens_sort_set.
  rewrite im_im.
  rewrite im_full_sig_proj1_sig.
  apply im_ext_in.
  intros x h2. clear h2.
  destruct x as [S h2]. simpl.
  apply sort_set_compat.
Qed.


Lemma in_fam_to_ens_sort_set_impl :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (F:Family T) (pf:finite_mem F) (l:list T),
    Inn (fam_to_ens_sort_set R pfso F pf) l ->
    Inn F (list_to_set l).
  intros T R hso F h1 l h2.
  destruct h2 as [l h2]. subst. clear h2.
  rewrite sort_set_compat.
  apply proj2_sig.
Qed.

Lemma in_fam_impl :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (F:Family T) (pffm:finite_mem F) (A:Ensemble T)
         (pfin:Inn F A),
    Inn (fam_to_ens_sort_set R pfso F pffm)
                 (sort_set R pfso A (pffm _ pfin)).
  intros T R hso F h1 A h2.
  unfold fam_to_ens_sort_set.
  apply Im_intro with (exist _ _ h2). constructor. simpl.
  reflexivity.
Qed.

Lemma in_fam_to_ens_sort_set_incr :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (F:Family T) (pffm:finite_mem F) (l:list T),
    Inn (fam_to_ens_sort_set R pfso F pffm) l ->
    increasingR R l.
  intros T R hso F h1 l h2.
  destruct h2 as [l h2]; subst. clear h2.
  destruct l as [l h2].
  simpl.
  apply incrR_sort_set.
Qed.


Lemma setminus_sort_set_fam :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (A B:Family T)
         (pfa:finite_mem A)
         (pfb:finite_mem B),
    let f := fun S : sig_set A =>
             sort_set R pfso (proj1_sig S)
                                 (pfa (proj1_sig S) (proj2_sig S)) in
    let g := fun S : sig_set B =>
             sort_set R pfso (proj1_sig S)
                                 (pfb (proj1_sig S) (proj2_sig S)) in
    let f' := restriction_sig f _ (setminus_inc _ _) in
    Setminus (Im (full_sig A) f)
             (Im (full_sig B) g) =
    Im (full_sig (Setminus A B)) f'.
  intros T R hso A B h1 h2 f g f'.
apply Extensionality_Ensembles; red; split; red; intros C h3.
destruct h3 as [h3 h4].
destruct h3 as [C h3]; subst. clear h3.
destruct C as [C h5].
unfold f'.
assert (h6:~Inn  B C).
contradict h4.
unfold f, g.
apply Im_intro with (exist _ _ h4). constructor.
simpl.
f_equal. apply proof_irrelevance.
apply Im_intro with (exist (fun x => Inn (Setminus A B) x) C (conj h5 h6)).
constructor.
unfold restriction_sig.
f_equal.
apply proj1_sig_injective; auto.

destruct h3 as [C h3]; subst. clear h3.
destruct C as [C h4]. destruct h4 as [h4 h5].
constructor.
econstructor. constructor. unfold f', restriction_sig. reflexivity.
intro h6.
inversion h6 as [x h8 q h9 h10]; subst.
unfold f', g, restriction_sig, f in h9.
simpl in h9.
apply inj_sort_set in h9.
subst.
destruct x as [x h9].
simpl in h5.
contradiction.
Qed.


Definition min_listR {T:Type} (R:Relation T) (pft:tso_dec R)
           (l m:list T) : list T :=
  if (le_listR_dec R pft l m) then l else m.

Definition max_listR {T:Type} (R:Relation T) (pft:tso_dec R)
           (l m:list T) : list T :=
  if (le_listR_dec R pft l m) then m else l.


Lemma min_listR_compat1 :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l m:list T),
    le_listR R
             (sort_dec_tso_dec _ pfso)
             (min_listR R (sort_dec_tso_dec _ pfso) l m)  l.
  intros T R hso l m. unfold min_listR.
  destruct (le_listR_dec R (sort_dec_tso_dec _ hso) l m) as [h1 | h1].
  apply refl_le_listR; auto.
  apply (sort_dec_irrefl _ hso).
  rewrite <- lt_listR_iff in h1.
  destruct h1; auto.
Qed.


Lemma min_listR_compat2 :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l m:list T),
    le_listR R (sort_dec_tso_dec _ pfso)
             (min_listR R (sort_dec_tso_dec _ pfso) l m)  m.
  intros T R hso l m. unfold min_listR.
  lr_if_goal; fold hlr; destruct hlr as [h1 | h1]; auto.
  rewrite <- lt_listR_iff in h1.
  destruct h1; auto.
  apply refl_le_listR.
  apply sort_dec_irrefl; auto.
Qed.


Lemma min_listR_eq_compat1 :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (la lb lc:list T),
    let pft := sort_dec_tso_dec _ pfso in
    le_listR R pft lb lc ->
    le_listR R pft
             (min_listR R pft la lb)
             (min_listR R pft la lc).
  intros T R hso la lb lc ht h1.
  unfold min_listR.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2].
  lr_if_goal; fold hlr; destruct hlr as [h3 | h3]; auto.
  apply refl_le_listR. 
  apply sort_dec_irrefl.
  assumption.
  pose proof (trans_le_listR R hso  _ _ _ h2 h1). contradiction.
  lr_if_goal; fold hlr; destruct hlr as [h3 | h3]; auto.
  unfold ht in h2.
  rewrite <- lt_listR_iff in h2.
  destruct h2; auto.
Qed.


Lemma min_listR_eq_compat2 :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (la lb lc:list T),
    let pft := sort_dec_tso_dec _ pfso in
    le_listR R pft la lb ->
    le_listR R pft
             (min_listR R pft la lc)
             (min_listR R pft lb lc).
  intros T R hso la lb lc ht h1.
  unfold min_listR.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2]; auto.
  lr_if_goal; fold hlr; destruct hlr as [h3 | h3]; auto.
  lr_if_goal; fold hlr; destruct hlr as [h3 | h3]; auto.
  pose proof (trans_le_listR R hso _ _ _ h1 h3). contradiction.
  apply refl_le_listR.
  apply sort_dec_irrefl; auto.
Qed.

Lemma min_listR_nil1 :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T),
    let pft := sort_dec_tso_dec _ pfso in
    min_listR R pft nil l = nil.
  intros.
  unfold min_listR.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2]; auto.
  contradict h2.
  apply le_listR_nil.
Qed.


Lemma min_listR_nil2 :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T),
    let pft := sort_dec_tso_dec _ pfso in
    min_listR R pft l nil = nil.
  intros.
  unfold min_listR.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2]; auto.
  destruct l; auto.
  contradict h2.
Qed.


Lemma min_listR_eq_cons :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l m:list T) (b:T),
    let pft := sort_dec_tso_dec _ pfso in
    min_listR R pft (b :: l) (b :: m) = b:: min_listR R pft l m.
  intros T R hso l m b.
  unfold min_listR.
  lr_if_goal; fold hlr; destruct hlr as [h1 | h1]; auto.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2]; auto.
  do 2 red in h1.
  destruct hso as [hts htr hir].
  destruct (hts b b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  apply hir in h3. contradiction.
  contradiction.
  contradiction.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2]; auto.
  unfold le_listR in h1. simpl in h1.
  destruct (sort_dec_tso_dec R hso b b) as [h3 | h3].
  destruct h3 as [h3 | h3].
  apply (sort_dec_irrefl _ hso) in h3. contradiction.
  contradiction.
  apply (sort_dec_irrefl _ hso) in h3. contradiction.
Qed.  

Lemma min_listR_comm :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l m:list T),
    let pft := sort_dec_tso_dec _ pfso in
    min_listR R pft l m =
    min_listR R pft m l.
  intros T R hso l m ht.
  unfold min_listR.
  lr_if_goal; fold hlr; destruct hlr as [h1 | h1]; auto.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2]; auto.
  eapply antisym_le_listR.
  apply h1. assumption.
  lr_if_goal; fold hlr; destruct hlr as [h2 | h2]; auto.
  eapply antisym_not_le_listR; auto.
  apply h2. assumption.
Qed.


Lemma min_listR_assoc :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (la lb lc:list T),
    let pft := sort_dec_tso_dec _ pfso in
    min_listR R pft
              la
              (min_listR R pft lb lc) =
    min_listR R pft
              (min_listR R pft la lb)
              lc.
  intros T R hso la lb lc.
  unfold min_listR.
  destruct (le_listR_dec R) as [h1 | h1]; auto.
  destruct (le_listR_dec R) as [h2 | h2]; auto.
  destruct (le_listR_dec R) as [h3 | h3]; auto.
  destruct (le_listR_dec R) as [h4 | h4]; auto.
  contradiction.
  destruct (le_listR_dec R) as [h4 | h4]; auto.
  pose proof (trans_le_listR R hso _ _ _ h4 h2).
  contradiction. contradiction.
  destruct (le_listR_dec R) as [h3 | h3]; auto.
  destruct (le_listR_dec R) as [h4 | h4]; auto.
  contradiction.
  destruct (le_listR_dec R) as [h4 | h4]; auto.
  contradiction.
  rewrite <- lt_listR_iff in h4, h3.
  pose proof (trans_lt_listR R hso _ _ _ h3 h4) as h5.
  destruct h5 as [h5 h6].
  pose proof (antisym_le_listR R hso _ _ h1 h5). subst.
  contradict h6; auto.
  destruct (le_listR_dec R) as [h2 | h2]; auto.
  rewrite <- lt_listR_iff in h1.
  destruct (le_listR_dec R) as [h3 | h3]; auto.
  destruct (le_listR_dec R) as [h4 | h4]; auto.
  pose proof (trans_le_lt_listR R hso _ _ _ h4 h1) as h5.
  destruct h5 as [h5 h6].
  contradict h6; auto.
  destruct (le_listR_dec R) as [h4 | h4]; auto.
  rewrite <- lt_listR_iff in h3.
  pose proof (trans_lt_le_listR R hso  _ _ _ h3 h4) as h5.
  destruct h5 as [h5 h6].
  pose proof (antisym_le_listR R hso _ _ h5 h2). subst.
  reflexivity.
  contradiction.
  destruct (le_listR_dec R) as [h3 | h3]; auto.
  destruct (le_listR_dec R) as [h4 | h4]; auto.
  contradiction. contradiction.
Qed.


Lemma min_listR_idem :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l:list T),
    let pft := sort_dec_tso_dec _ pfso in
    min_listR R pft l l = l.
  intros ? R. intros.
  unfold min_listR.
  destruct (le_listR_dec R); auto.
Qed.


Lemma min_listR_mono :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R)
         (l la lb:list T),
    let pft := sort_dec_tso_dec _ pfso in
    le_listR R pft la lb ->
    le_listR R pft
             (min_listR R (sort_dec_tso_dec _ pfso) l la)
             (min_listR R (sort_dec_tso_dec _ pfso) l lb).
  intros T R hso l la lb ht h1.
  unfold min_listR.
  destruct (le_listR_dec R) as [h2 | h2]; auto.
  destruct (le_listR_dec R) as [h3 | h3]; auto. 
  apply refl_le_listR.
  apply sort_dec_irrefl; auto.
  pose proof (trans_le_listR R hso _ _ _ h2 h1). contradiction.
  destruct (le_listR_dec R) as [h3 | h3]; auto.
  rewrite <- lt_listR_iff in h2.
  destruct h2; auto.
Qed.

Lemma Rw_lt_listR_iff :
  forall {T:Type} (R:Relation T) (pfs:sort_dec R)
         (l l':list T),
    let pft := sort_dec_tso_dec _ pfs in
    Rw (lt_listR R pft) l l' <->
    le_listR R pft l l'.
  intros T R h1 l l'; split; intro h2.
  red in h2. destruct h2 as [h2 |]; subst.
  apply lt_list_impl_leR; auto. 
  apply refl_le_listR; auto.
  apply sort_dec_irrefl; auto.
  apply le_listR_lt_eq_dec in h2. 
  destruct h2 as [h2 | h2].
  left; auto.
  right; auto.
Qed.


Lemma lt_listR_map :
    forall {T U:Type} (R:Relation T) (R':Relation U)
           (pftso:tso_dec R) (pfir:irrefl R)
           (pfso':sort_dec R')
           (f:T->U),
      incrR_fun R R' f ->
      forall (l l':list T),
        let pftso' := sort_dec_tso_dec _ pfso' in
        lt_listR R pftso l l' ->
        lt_listR R' pftso' (map f l) (map f l').
  intros T U R R' htso hir hso' f h1 l l' htso'.
  pose proof (incrR_fun_inj R R' f htso (sort_dec_irrefl _ hso')  h1) as h1'.
  revert h1' h1. revert l f.
  induction l' as [|a' l' ih]; simpl.
  intros ? ? h1' h1 h2.
  apply not_lt_listR_nil in h2; contradiction. 
  intros l f h1' h1 h2. 
  destruct l as [|a l].
  simpl.
  red; simpl; split; auto.
  intro; discriminate. 
  simpl; simpl in h2.
  red in h2; red.
  destruct h2 as [h2 h3].
  simpl in h2. 
  simpl.  
  destruct (htso a a') as [h4 | h4].
  destruct h4 as [h4 | h4].
  destruct (htso' (f a) (f a'))
    as [h5 | h5].
  destruct h5 as [h5 | h5]; auto.
  split; auto.
  intro h6.
  inversion h6 as [h6'].
  apply map_inj in H ; subst; contradict h3; auto. 
  apply h1' in h6'; subst; auto. 
  apply map_inj in H ; subst; auto.
  rewrite h6' in h5.
  apply (sort_dec_irrefl _ hso')  in h5; contradiction.
  apply h1' in h5; subst; auto. 
  contradict h4; auto.
  
  apply h1 in h4; subst; contradict h3; auto.
  pose proof (sort_dec_trans _ hso' _ _ _ h4 h5) as ht.
  apply (sort_dec_irrefl _ hso') in ht.
  contradiction.
  split.
  destruct (htso' (f a) (f a')) as [h6 | h6].
  destruct h6; auto. 
  apply h1' in e; subst.
  apply lt_list_impl_leR.
  apply ih; auto.
  split; auto.
  intro; subst; contradict h3; auto.
  subst.
  apply (sort_dec_irrefl _ hso') in h6; auto.
  intro h5.
  inversion h5.
  apply h1' in H0; subst.
  apply map_inj in H1; subst; contradict h3; auto.
  apply map_inj in H1; subst; auto.
  contradiction.
Qed.


Definition incrR_list_fun
           {T U:Type} (R:Relation T) (R':Relation U)
           (pftso:tso_dec R) (pftso':tso_dec R')
           (f:list_fun T U) : Prop :=
  incrR_fun (lt_listR R pftso) (lt_listR R' pftso') f.

Definition weak_incrR_list_fun
           {T U:Type} (R:Relation T) (R':Relation U)
           (pftso:tso_dec R) (pftso':tso_dec R')
           (f:list_fun T U) : Prop :=
  weak_incrR_fun (lt_listR R pftso) (lt_listR R' pftso') f.


Lemma incrR_list_fun_impl_weak :
  forall {T U:Type} (R:Relation T) (R':Relation U)
         (pftso:tso_dec R) (pftso':tso_dec R')
         (f:list_fun T U),
    incrR_list_fun R R' pftso pftso' f ->
    weak_incrR_list_fun R R' pftso pftso' f.
  unfold incrR_list_fun, weak_incrR_list_fun; intros; eapply incrR_fun_impl_weak; auto.
Qed.

  
End ListSort.
Section ListNat.

  Definition increasing (l:list nat) : Prop :=
    increasingR lt l.
  
  Definition weak_increasing (l:list nat) : Prop :=
    weak_increasingR lt l.
  
  Lemma increasing_cons :
  forall (l:list nat) (a:nat),
    increasing (a::l) ->
    increasing l.
    unfold increasing; apply increasingR_cons.
  Qed.
  
  Lemma weak_increasing_cons :
    forall (l:list nat) (a:nat),
      weak_increasing (a::l) ->
      weak_increasing l.
    unfold weak_increasing; apply weak_increasingR_cons.
  Qed.
  
  Lemma increasing_impl_weak :
  forall (l:list nat),
    increasing l ->
    weak_increasing l.
    unfold increasing, weak_increasing; apply increasingR_impl_weak.
  Qed.
  
  Lemma lt_increasing_cons : 
  forall (l:list nat) (a m:nat),
    In a l ->
    increasing (m::l)->
    m < a.
    unfold increasing;
    intros l a m h1 h2;
    apply (lt_increasingR_cons lt l a m lt_trans h1 h2).
  Qed.
  
  Lemma le_weak_increasing_cons : 
  forall (l:list nat) (a m:nat),
    In a l ->
    weak_increasing (m::l)->
    m <= a.
    unfold weak_increasing;
    intros l a m h1 h2;
    rewrite <- Rw_le_iff;
    apply (le_weak_increasingR_cons lt l a m lt_trans h1 h2).
  Qed.

  Lemma increasing_cons_impl :
  forall (l:list nat) (a:nat),
    increasing (a::l) ->
    increasing l.
    unfold increasing; apply increasingR_cons_impl.
  Qed.
  
  Lemma weak_increasing_cons_impl :
    forall (l:list nat) (a:nat),
      weak_increasing (a::l) ->
      weak_increasing l.
    unfold weak_increasing; apply weak_increasingR_cons_impl.
  Qed.
  
  Lemma increasing_cons_impl_cons :
    forall (l:list nat) (a d:nat),
    increasing (a::l) ->
    d < a ->
    increasing (d::a::l).
    unfold increasing; apply increasingR_cons_impl_cons.
  Qed.
  
  Lemma weak_increasing_cons_impl_cons :
  forall (l:list nat) (a d:nat),
    weak_increasing (a::l) ->
    d <= a ->
    weak_increasing (d::a::l).
    unfold weak_increasing.
    intros l a d h1 h2.
    rewrite <- Rw_le_iff in h2.
    apply (weak_increasingR_cons_impl_cons lt l a d h1 h2).
  Qed.

  Lemma increasing_impl_cons :
    forall (l:list nat) 
    (pf:l <> nil) (d:nat),
      increasing l ->
      d < hd' l pf ->
      increasing (d::l).
    unfold increasing; apply increasingR_impl_cons.
  Qed.
  
  Lemma weak_increasing_impl_cons :
    forall (l:list nat) 
    (pf:l <> nil) (d:nat),
      weak_increasing l ->
      d <= hd' l pf ->
      weak_increasing (d::l).
    unfold weak_increasing; intros ? ? ? ? h1; rewrite <- Rw_le_iff in h1; eapply weak_increasingR_impl_cons; auto.
    apply h1.
  Qed.
  
Lemma increasing_impl_cons' :
  forall (l:list nat) (d:nat),
    increasing l ->
    (forall c:nat, In c l -> d < c) ->
    increasing (d::l).
  unfold increasing; apply increasingR_impl_cons'.
Qed.

Lemma weak_increasing_impl_cons' :
  forall (l:list nat) (d:nat),
    weak_increasing l ->
    (forall c:nat, In c l -> d <= c) ->
    weak_increasing (d::l).
  unfold weak_increasing.
  intros l d h1 h2.
  apply (weak_increasingR_impl_cons' lt l d h1).
  intros; rewrite Rw_le_iff; auto.
Qed.

Lemma increasing_tl :
  forall (l:list nat),
    increasing l ->
    increasing (tl l).
  unfold increasing; apply increasingR_tl.
Qed.

Lemma weak_increasing_tl :
  forall (l:list nat),
    weak_increasing l ->
    weak_increasing (tl l).
  unfold weak_increasing; apply weak_increasingR_tl.
Qed.

Lemma increasing_remove :
  forall (l:list nat) (a:nat),
    increasing l ->
    increasing (remove nat_eq_dec a l).
  unfold increasing.
  intros; apply (increasingR_remove nat_eq_dec lt); auto.
  red; intro; eapply lt_trans.
Qed.
 
Lemma weak_increasing_remove :
  forall (l:list nat) (a:nat),
    weak_increasing l ->
    weak_increasing (remove nat_eq_dec a l). 
  unfold weak_increasing. intros l a h1.
  apply (weak_increasingR_remove nat_eq_dec lt); auto.
  red; intro; eapply lt_trans.
Qed.


Lemma incr_list_minus :
  forall (l l':list nat),
    increasing l ->
    let lm := list_minus nat_eq_dec l l' in
    increasing lm.
  unfold increasing; intros l l'.
  apply (incrR_list_minus nat_eq_dec lt  l l' lt_trans); auto.
Qed.

Lemma weak_incr_list_minus :
  forall (l l':list nat),
    weak_increasing l ->
    let lm := list_minus nat_eq_dec l l' in
    weak_increasing lm.
  unfold weak_increasing; intros l l'.
  apply (weak_incrR_list_minus nat_eq_dec lt  l l' lt_trans); auto.
Qed.


Lemma increasing_cons_lt_subst :
  forall (l:list nat) (a b:nat),
    increasing (a::l) ->
    b < a ->
    increasing (b::l).
  unfold increasing; intros l a b h1 h2;
  apply (increasingR_cons_lt_subst lt l a b lt_trans h1 h2).
Qed.

Lemma weak_increasing_cons_le_subst :
  forall (l:list nat) (a b:nat),
    weak_increasing (a::l) ->
    b <= a ->
    weak_increasing (b::l).
  unfold weak_increasing; intros l a b h1 h2;
  apply (weak_increasingR_cons_le_subst lt l a b lt_trans h1).
  rewrite Rw_le_iff; auto.
Qed.
 
Lemma increasing_cons_no_dup :
  forall (l:list nat) (n:nat),
    increasing (n::l) ->
    NoDup l.
  unfold increasing; intros l n; apply (increasingR_cons_no_dup nat_eq_dec lt l n  so_lt_nat).
Qed.

Lemma increasing_cons_nin :
  forall (l:list nat) (a:nat),
    increasing (a::l) ->
    ~In a l.
  unfold increasing; intros l a; apply (increasingR_cons_nin nat_eq_dec lt l a so_lt_nat).
Qed.

Lemma increasing_no_dup :
  forall (l:list nat) ,
    increasing l ->
    NoDup l.
  unfold increasing; intro l; apply (increasingR_no_dup nat_eq_dec lt l so_lt_nat).
Qed.

Lemma increasing_firstn :
  forall (l:list nat) (m:nat),
    increasing l ->
    increasing (firstn m l).
  unfold increasing; intros l m; apply (increasingR_firstn nat_eq_dec lt l m).
  red; intro; apply lt_trans.
Qed.

Lemma weak_increasing_firstn :
  forall (l:list nat) (m:nat),
    weak_increasing l ->
    weak_increasing (firstn m l).
  unfold weak_increasing; intros l m h1; apply (weak_increasingR_firstn lt l m lt_trans h1).
Qed.

Lemma increasing_skipn :
  forall (l:list nat) (m:nat),
    increasing l ->
    increasing (skipn m l).
  unfold increasing; intros l m h1; apply (increasingR_skipn lt l m h1).
Qed.

Lemma weak_increasing_skipn :
  forall (l:list nat) (m:nat),
    weak_increasing l ->
    weak_increasing (skipn m l).
  unfold weak_increasing; intros l m h1; apply (weak_increasingR_skipn lt l m h1).
Qed.
 

Lemma increasing_app :
  forall (l l':list nat)
         (pfl:l <> nil)
         (pfl':l' <> nil),
    increasing l ->
    increasing l' ->
    lst l pfl <  hd' l' pfl' ->
    increasing (l++l'). 
  unfold increasing; intros; rewrite <- increasingR_app_iff; repeat split; auto.
  apply H1.
  apply lt_trans'.
Qed.

Lemma weak_increasing_app :
  forall (l l':list nat)
         (pfl:l <> nil)
         (pfl':l' <> nil),
    weak_increasing l ->
    weak_increasing l' ->
    lst l pfl <=  hd' l' pfl' ->
    weak_increasing (l++l').
  unfold weak_increasing; intros ? ? ? ? ? ? h1; rewrite <- Rw_le_iff in h1; erewrite <- weak_increasingR_app_iff; repeat split; auto.
  apply h1.
  apply lt_trans'.
Qed.

Lemma weak_incr_no_dup_impl :
  forall (l:list nat),
    weak_increasing l ->
    NoDup l ->
    increasing l.
  unfold weak_increasing, increasing; intros; apply weak_incrR_no_dup_impl; auto.
Qed.


Lemma increasing_map :
  forall (f:nat->nat) (l:list nat),
    incr_fun f ->
    increasing l ->
    increasing (map f l).
  intros ? ? h1; rewrite incr_fun_iff in h1; unfold incr_fun' in h1; unfold increasing; apply increasingR_map; auto.
Qed.

Lemma weak_increasing_map :
  forall (f:nat->nat) (l:list nat),
    weak_incr_fun f ->
    weak_increasing l ->
    weak_increasing (map f l).
    intros ? ? h1; rewrite weak_incr_fun_iff in h1; unfold incr_fun' in h1; unfold weak_increasing; apply weak_increasingR_map; auto.
Qed.

Lemma incr_seg_list :
  forall (n:nat),
    increasing (seg_list n).
  intro n.
  induction  n as [|n ih].
  unfold seg_list, seq; red; simpl; auto.
  rewrite seg_list_app; auto with arith.
  destruct (zerop n) as [hz | hz]; subst.
  simpl; auto.
  simpl.
  pose proof (in_lst _ (seg_list_not_nil _ hz)) as h1.
  rewrite in_seg_list_iff in h1.
  rewrite <- minus_n_O.
  eapply increasing_app; auto.
  do 2 red; auto.
  simpl.
  apply h1.
  Grab Existential Variables.
  apply cons_neq_nil.
Qed.


Lemma incr_seq :
  forall (m n:nat),
    increasing (seq m n).
  intros m n.
  rewrite seq_eq_seq_0.
  apply increasing_map.
  red.
  intros; omega.
  apply incr_seg_list.
Qed.


Lemma incr_all_sing :
  forall(l:list nat) (m:nat),
    all_sing l m ->
    increasing l ->
    l = nil \/ l = m::nil.
  unfold increasing; apply incrR_all_sing.
  apply nat_eq_dec.
  apply so_lt_nat.
Qed.

Lemma incr_lS :
  forall (l:list nat),
    increasing l ->
    increasing (lS l).
  unfold lS; intros; apply increasing_map; auto.
  apply incr_fun_S.
Qed.

Lemma weak_incr_lS :
  forall (l:list nat),
    weak_increasing l ->
    weak_increasing (lS l).
  unfold lS; intros; apply weak_increasing_map; auto.
  apply weak_incr_fun_S.
Qed.



Lemma incr_lpred_nin0 :
  forall (l:list nat) ,
    ~ In 0 l ->
    increasing l ->
    increasing (lpred l).
  intro l.
  induction l as [|a l ih]; simpl. 
  intros; auto.
  intros h1 h2.
  pose proof (increasing_cons_nin l a h2) as h3. 
  assert (h4:a <> 0).
  intro; subst.
  contradict h1.
  left; auto.  
  assert (h5:~In 0 l).
    intro h5.
    contradict h1.
    right; auto.
  specialize (ih h5).
  hfirst ih.  
  eapply increasing_cons_impl. apply h2.
  specialize (ih hf). 
  destruct l as [|c l]; simpl; auto.
  pose proof (lt_increasing_cons _  _ _ (in_eq _ _) h2) as h76.
  split.
  apply mono_pred_lt; auto.
  omega.
  assumption.
Qed.


Lemma incr_lpred_remove_in0_in1 :
  forall (l:list nat),
    In 0 l ->
    In 1 l ->
    increasing l ->
    increasing (lpred (remove nat_eq_dec 0 l)).
  intro l.
  destruct l as [|a l].
  intros; auto. 
  intros h1 h2 h3. 
  pose proof (increasing_cons_nin l a h3) as h4.
  assert (h5:a = 0).
  destruct h1 as [|h1]; auto.
  pose proof (lt_increasing_cons _ _ _ h1 h3) as h5.
  omega.
  subst.
  destruct h2 as [|h2].
  discriminate.
  rewrite remove_cons_eq.
  rewrite nin_remove_eq; auto.
  red in h3.
  pose proof (increasing_cons_impl _ _ h3) as h5.
  apply incr_lpred_nin0; auto.
Qed.


Lemma incr_lpred_in0_nin1 :
  forall (l:list nat) ,
    In 0 l  ->
    ~ In 1 l ->
    increasing l ->
    increasing (lpred l).
  intro l. 
  induction l as [|a l ih]; simpl.
  intros; contradiction. 
  intros h1 h2 h3.
  pose proof (increasing_cons_nin _ _ h3) as h6.
  destruct a as [|a].
  simpl.
  destruct l as [|c l]; simpl; auto.
  pose proof (lt_increasing_cons _ _ _ (in_eq _ _) h3) as h7. 
  assert (h8: 1 < c).
    destruct c as [|c]. omega.
    assert (1 <> S c).
      contradict h2. right. left; auto.
    omega.
  split.
  omega.
  pose proof (increasing_cons_impl (c::l) _  h3) as h9.
  pose proof (incr_lpred_nin0 _ h6 h9) as h10.
  assumption.
  simpl.
  destruct h1 as [h1 | h1]; try discriminate.
  assert (h7:a <> 0).
    intro. subst. contradict h2; left; auto.
  specialize (ih h1).
  hfirst ih.
  contradict h2. right; auto.
  specialize (ih hf).
  hfirst ih.
  eapply increasing_cons_impl. apply h3.
  specialize (ih hf0).
  eapply increasing_impl_cons; auto.
  destruct l as [|c l].
  simpl.
  contradict h1.
  simpl.
  simpl in h3.
  destruct h3; omega.
  Grab Existential Variables.
  apply neq_nil_lpred.
  intro; subst.
  contradict h1.
Qed.

Lemma weak_incr_lpred :
  forall (l:list nat),
    weak_increasing l ->
    weak_increasing (lpred l).
  intro l.
  induction l as [|c l ih]; simpl; auto.
  destruct l as [|a l]; simpl; auto. 
  intro h1.
  red in h1.
  simpl in h1.
  destruct h1 as [h1 h2].
  rewrite Rw_le_iff in h1.
  red. split.
  rewrite Rw_le_iff. omega.
  apply ih; auto.
Qed.


  

Lemma increasing_linds_occ :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T),
    increasing (linds_occ pfdt l x).
  intros T h1 l.
  induction l as [|a l ih]; simpl.
  intros.
  rewrite linds_occ_nil; constructor.
  intro x.
  destruct (h1 x a) as [h3 | h3]; subst.
  rewrite linds_occ_cons_eq.
  simpl.
  destruct (nil_dec' (linds_occ h1 l a)) as [h3 | h3].
  rewrite h3; simpl; auto.
  do 2 red; auto.
  apply not_nil_cons in h3.
  destruct h3 as [n [ln h3]].
  pose proof (in_eq n ln) as h4.
  rewrite <- h3 in h4.
  rewrite in_linds_occ_iff in h4.
  destruct h4 as [m [h4 h5]].
  destruct h5 as [h5 h6]; subst. 
  rewrite h3.
  simpl.
  split.
  apply lt_0_Sn.
  specialize (ih a).
  rewrite h3 in ih. 
  apply (incr_lS (lind_occ h1 l m a h4 :: ln)); auto.
  specialize (ih x).
  rewrite linds_occ_cons_neq; auto.
  apply incr_lS; auto.
Qed.
 

Lemma increasing_linds_occ' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (x:T) (pf:In x l),
    increasing (linds_occ' pfdt l x pf).
  intros; rewrite <- linds_occ_compat'; apply increasing_linds_occ; auto.
Qed.

Lemma in_linds_incr :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds pfdt l) ->
    increasing ln.
  intros T h1 l ln h2.
  rewrite in_linds_iff in h2.
  destruct h2 as [x [[h2 h3] h4]]; subst.
  apply increasing_linds_occ.
Qed.


Lemma list_to_set_injable_increasingR :
  forall {T:Type} (R:Relation T),
    sort_dec R ->
    list_to_set_injable (increasingR R).
  intros T R hso.
  pose proof (sort_dec_eq_dec _ hso) as hdt.
  pose proof (sort_dec_trans _ hso) as htr.
  pose proof (sort_dec_irrefl _ hso) as hir.
  red; intro l.
  induction l as [|a l ih]; simpl.
  intros ? ? ? h1.
  symmetry in h1; apply empty_set_nil in h1; subst; auto.
  intros l' h1 h2 h3.
  destruct l' as [|a' l'].
  simpl in h3.
  apply add_not_empty in h3.
  contradiction.
  simpl in h3.
  simpl in h2.
  destruct l as [|b l]; simpl.
  simpl in h3.
  rewrite add_empty_sing in h3.
  apply sing_eq_add in h3.
  destruct h3 as [? h3].
  subst.
  destruct h3 as [h3 | h3].
  apply empty_set_nil in h3; subst; auto.
  f_equal.
  destruct l' as [|a']; auto.
  simpl in h3.
  symmetry in h3.
  apply sing_eq_add in h3.
  destruct h3 as [h3]; subst.
  destruct h2 as [h2].
  apply hir in h2; contradiction.  
  destruct l' as [|b' l'].
  simpl in h3.
  rewrite add_empty_sing in h3.
  symmetry in h3.
  apply sing_eq_add in h3.
  destruct h3 as [h3 h4].
  subst.
  destruct h4 as [h4 | h4].
  apply add_not_empty in h4; contradiction.
  symmetry in h4.
  apply sing_eq_add in h4.
  destruct h4 as [? h4]; subst.
  destruct h1 as [h1].
  apply hir in h1; contradiction.
  simpl in h3.
  destruct h1 as [h1 h4], h2 as [h2 h5].

  destruct (hdt a a') as [|h6]; subst. 
  f_equal.
  apply ih; auto.
  pose proof (add_nin_inj_eq (Add (list_to_set l) b) (Add (list_to_set l') b') a') as h6.
  hfirst h6.

  intro h8. 
  destruct h8 as [a' h8|a' h8]; subst. 
  rewrite  <- list_to_set_in_iff in h8. 
  pose proof (lt_increasingR_cons R l _ _ htr h8 h4) as h9.
  pose proof (htr _ _ _ h1 h9) as h10.
  apply hir in h10; contradiction.
  destruct h8; subst.
  apply hir in h1; contradiction.
  specialize (h6 hf).
  hfirst h6.

  intro h8. 
  destruct h8 as [a' h8|a' h8]; subst. 
  rewrite  <- list_to_set_in_iff in h8. 
  pose proof (lt_increasingR_cons R l' _ _ htr h8 h5) as h9.
  pose proof (htr _ _ _ h2 h9) as h10.
  apply hir in h10; contradiction.
  destruct h8; subst.
  apply hir in h2; contradiction.
  specialize (h6 hf0).
  apply h6; auto.

  pose proof (Add_intro2 _ (Add (list_to_set l') b') a') as h7. 
  rewrite <- h3 in h7.
  pose proof (Add_intro2 _ (Add (list_to_set l) b) a) as h7'.
  rewrite h3 in h7'.
  destruct h7 as [a' h7 | a' h7],
                 h7' as [a h7' | a h7'].
  destruct h7 as [a' h7 | a' h7],
                 h7' as [a h7' | a h7'].
  rewrite <- list_to_set_in_iff in h7, h7'.
  pose proof (lt_increasingR_cons R l _ _ htr h7 h4) as h9.
  pose proof (lt_increasingR_cons R l' _ _ htr h7' h5) as h9'.
  pose proof (htr _ _ _ h1 h9) as h10.
  pose proof (htr _ _ _ h2 h9') as h11.
  pose proof (htr _ _ _ h10 h11) as h12.
  apply hir in h12; contradiction.
  destruct h7'.
  rewrite <- list_to_set_in_iff in h7.
  pose proof (lt_increasingR_cons R l _ _ htr h7 h4) as h8.
  pose proof (htr _ _ _ h8 h2) as h9.
  pose proof (htr _ _ _ h9 h1) as h10.
  apply hir in h10; contradiction.
  destruct h7.
  rewrite <- list_to_set_in_iff in h7'.
  pose proof (lt_increasingR_cons R l' _ _ htr h7' h5) as h8.
  pose proof (htr _ _ _ h2 h8) as h9.
  pose proof (htr _ _ _ h9 h1) as h10.
  apply hir in h10; contradiction.
  destruct h7, h7'.
  pose proof (htr _ _ _ h1 h2) as h7.
  apply hir in h7; contradiction. 
  destruct h7'.
  destruct h7 as [a' h7 |a' h7]; subst.
  rewrite <- list_to_set_in_iff in h7.
  pose proof (lt_increasingR_cons R l _ _ htr h7 h4) as h8.
  pose proof (htr _ _ _ h8 h1) as h9.
  apply hir in h9; contradiction.
  destruct h7; contradict h6; auto.
  destruct h7; contradict h6; auto.
  destruct h7; contradict h6; auto.
Qed.


Lemma list_to_set_injable_increasing :
  list_to_set_injable increasing.
  red; intros l l' h1 h2 h3.
  eapply list_to_set_injable_increasingR; auto.
  apply sort_dec_lt_nat.
  assumption.
  assumption.
Qed.


Lemma incrR_remove_inj :
  forall {T:Type} (pfdt:type_eq_dec T)
         (R:Relation T)
         (pfso:sort_dec R)
         (l l':list T)
         (x:T),
    In x l ->
    In x l' ->
    increasingR R l ->
    increasingR R l' ->
    remove pfdt x l =
    remove pfdt x l' ->
    l = l'.
  intros T hdt R hso l l' x hi hi' h1 h2 h3.
  eapply list_to_set_injable_increasingR.
  apply hso.
  assumption.
  assumption.
  apply Extensionality_Ensembles; red; split; red; intros c h4;
  rewrite <- list_to_set_in_iff in h4;
  rewrite <- list_to_set_in_iff;
  destruct (hdt x c) as [|h5]; subst; auto.
  pose proof (in_remove_iff hdt l c x) as h6.
  hr h6; auto.
  rewrite <- h6 in hr.
  rewrite h3 in hr.
  rewrite in_remove_iff in hr.
  destruct hr; auto.
  pose proof (in_remove_iff hdt l' c x) as h6.
  hr h6; auto.
  rewrite <- h6 in hr.
  rewrite <- h3 in hr.
  rewrite in_remove_iff in hr.
  destruct hr; auto.
Qed.



Lemma in_linds_list_to_set_injable :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    list_to_set_injable (fun xl => In xl (linds pfdt l)).
  intros T h1 l; red; intros lm ln h2 h3 h4.
  rewrite in_linds_iff in h2, h3.
  destruct h2 as [x [[h2 h5] h6]],
                 h3 as [y [[h3 h8] h7]]; subst.
  
  pose proof (increasing_linds_occ h1 l x) as h8.
  pose proof (increasing_linds_occ h1 l y) as h9.
  eapply incrR_no_dup_minimal.
  apply nat_eq_dec.
  apply h8.
  apply no_dup_linds_occ.
  assumption.
  pose proof list_to_set_injable_increasing as h10.
  red in h10.
  specialize (h10 _ _ h8 h9 h4).
  rewrite h10.
  apply refl_le_listR.
  red; intro; apply lt_irrefl.

  Grab Existential Variables.
  apply sort_dec_lt_nat.
Qed.


Lemma lec_list_to_set_injable :
  forall {T U:Type}
         (pfdt:type_eq_dec T)
         (pfdu:type_eq_dec U)
         (l :list T)
         (l':list U)
         (i j:nat)
         (pfi:i < length l)
         (pfj:j < length l'),
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l') ->
    list_to_set (lec pfdt pfi) =
    list_to_set (lec pfdu pfj) ->
    lec pfdt pfi = 
    lec pfdu pfj.
  intros T U hdt hdu l l' i j hi hj he he'.
  pose proof (in_lec hdt _ _ hi) as h1.
  pose proof (in_lec hdu _ _ hj) as h2.
  rewrite list_to_set_in_iff in h1.
  rewrite he in h1.
  rewrite <- list_to_set_in_iff in h1.
  eapply in_linds_list_to_set_injable.
  apply h1.
  assumption.
  assumption.
Qed.
  

Lemma list_to_set_linds_eq_length :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l') ->
    length l = length l'.
  intros T U h1 h2 l l' h3.
  pose proof (partition_linds h1 l) as h4.
  pose proof (partition_linds h2 l') as h5.
  eapply no_dup_partition_faml_nat_seg_list_to_set_eq.
  eapply no_dup_linds.
  eapply no_dup_linds.
  eapply no_dup_mem_linds.
  eapply no_dup_mem_linds.
  eapply in_linds_list_to_set_injable.
  eapply in_linds_list_to_set_injable.
  apply h3.
  apply partition_linds.
  apply partition_linds.
Qed.


Lemma linds_eq_length :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    linds pfdt l =
    linds pfdu l' ->
    length l = length l'.
  intros; eapply list_to_set_linds_eq_length.
  rewrite H at 1.
  reflexivity.
Qed.



Definition lastlinds {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (x:T) :=
  last (linds_occ pfdt l x) 0.
  
Definition lastlindss {T:Type} (pfdt:type_eq_dec T)
           (l:list T) : list nat :=
  map_in_dep (fun ln (pf:In ln (linds pfdt l)) =>
                last ln 0).

Lemma in_lastlindss_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat),
    In n (lastlindss pfdt l) <->
    exists ln:list nat,
      In ln (linds pfdt l) /\
      n = last ln 0.
  intros T h1 l n; split; intro h2.
  unfold lastlindss in h2.
  rewrite in_map_in_dep_iff in h2.
  destruct h2 as [ln [h2 h3]]; subst.
  exists ln; tauto.
  unfold lastlindss.
  rewrite in_map_in_dep_iff.
  destruct h2 as [ln [h2 h3]]; subst.
  exists ln, h2; auto.
Qed.


Lemma in_lastlindss_lt :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (n:nat),
    In n (lastlindss pfdt l) ->
    n < length l.
  intros T h1 l n h2.
  rewrite in_lastlindss_iff in h2.
  destruct h2 as [ln [h2 h3]]; subst.
  rewrite in_linds_iff in h2.
  destruct h2 as [x [h2 h3]].
  destruct h2 as [h2]; subst.
  eapply in_linds_occ_lt.
  apply in_last.
  apply linds_occ_in; auto.
Qed.

Lemma length_lastlindss :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T),
    length (lastlindss pfdt l) =
    length (list_singularize pfdt l).
  intros T h1 l.
  unfold lastlindss.
  rewrite length_map_in_dep.
  apply length_linds.
Qed.




Definition maxl (l:list nat) : l <> nil -> nat :=
  maxlR lt (tso_dec_lt_nat) l.

Lemma maxl_compat : 
  forall (l:list nat) (n:nat)
         (pf:In n l),
    let pfn := in_not_nil l n pf in
    n <= maxl l pfn.
  unfold maxl; intros; rewrite <- Rw_le_iff.
  apply maxlR_compat.
  red; apply lt_trans.
Qed.

Definition minl (l:list nat) : l <> nil -> nat :=
  minlR lt (tso_dec_lt_nat) l.

Lemma in_minl :
  forall (l:list nat) (pf:l <> nil),
    In (minl l pf) l.
  unfold minl; apply in_minlR.
Qed.
 
Lemma minl_compat :
  forall (l:list nat) (n:nat)
         (pf:In n l),
    let pfn := in_not_nil l n pf in
    In n l -> minl l pfn <= n.
  unfold minl; intros; rewrite <- Rw_le_iff.
  apply minlR_compat.
  red; apply lt_trans.
Qed.

Definition two_lists_to_rel {T U:Type} (l:list T) (l':list U)
           (pfe:length l = length l') : Ensemble (T * U) :=
  list_to_set (map_seg (fun m (pfm:m < length l) =>
                          (nth_lt l  m pfm,
                           nth_lt l' m (lt_eq_trans _ _ _ pfm pfe)))).

Lemma two_lists_to_rel_nil :
  forall (T U:Type),
    two_lists_to_rel (@nil T) (@nil U) (eq_refl _) = Empty_set _.
  intros; unfold two_lists_to_rel.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  rewrite <- list_to_set_in_iff in h1.
  rewrite in_map_seg_iff in h1.
  simpl in h1.
  destruct h1 as [? [? ?]]; omega.
  contradiction.
Qed.


Lemma in_two_lists_to_rel :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (x:T) (y:U),
    Inn (two_lists_to_rel l l' pfe) (x, y) ->
    exists (pfx:In x l) (pfy:In y l') (ox oy:nat)
           (pfox:ox < count_occ pfdt l x) (pfofy:oy < count_occ pfdu l' y),
      lind_occ pfdt l  ox x pfx =
      lind_occ pfdu l' oy y pfy.
  intros T U h1 h2 l l' h3 x y h4.
  unfold two_lists_to_rel in h4.
  rewrite <- list_to_set_in_iff in h4.
  rewrite in_map_seg_iff in h4.
  destruct h4 as [m [h4 h5]].
  inversion h5.
  subst.
  exists (in_nth_lt _ _ _), (in_nth_lt _ _ _).
  pose proof (ex_lind_occ_nth_lt h1 _ _ h4) as h6.
  pose proof (ex_lind_occ_nth_lt h2 _ _ (lt_eq_trans _ _ _ h4 h3)) as h7.
  destruct h6 as [ox [h6 h8]], h7 as [oy [h7 h9]].
  exists ox, oy.
  exists h6, h7.
  congruence.
Qed.

Lemma incl_graph_seg_two_lists_to_rel_in :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l')
         (x:T) (pfx:In x l)
         (f:sig_set (list_to_set l) -> U),
    Included (graph_seg f) (two_lists_to_rel l l' pfe) ->
    In (f (exist _ _ (iff1 (list_to_set_in_iff l x) pfx))) l'.
  intros T U hdt hdu l l' h1 x h2 f h3.
  red in h3.
  specialize (h3 (x, (f (exist (Inn (list_to_set l)) x (iff1 (list_to_set_in_iff l x) h2))))).
  hfirst h3.
  unfold graph_seg.
  apply Im_intro with (exist (Inn (list_to_set l)) x (iff1 (list_to_set_in_iff l x) h2)).
  constructor.
  simpl; auto.
  specialize (h3 hf).
  apply (in_two_lists_to_rel hdt hdu) in h3.
  destruct h3 as [? [? ?]]; auto.
Qed.


    
Lemma graph_seg_two_lists_to_rel_compat :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (x:T) (pfx:In x l)
         (f:sig_set (list_to_set l) -> U)
         (pfinc:Included (graph_seg f) (two_lists_to_rel l l' pfe)),
    let y := (f (exist _ _ (iff1 (list_to_set_in_iff l x) pfx))) in
    exists (ox oy:nat)
           (pfox:ox < count_occ pfdt l x) (pfofy:oy < count_occ pfdu l' y),
    let pfy := incl_graph_seg_two_lists_to_rel_in pfdt pfdu l l' pfe x pfx f pfinc in
    lind_occ pfdt l  ox x pfx =
    lind_occ pfdu l' oy y pfy.
Proof.
  intros T U hdt hdu l l' h1 x h2 f h3 y.
  pose proof (in_two_lists_to_rel hdt hdu l l' h1 x y) as h4.
  hfirst h4.
  apply h3.
  unfold graph_seg.
  apply Im_intro with (exist (Inn (list_to_set l)) x (iff1 (list_to_set_in_iff l x) h2)).
  constructor.
  simpl; auto.
  specialize (h4 hf).
  destruct h4 as [h4 [h5 [ox [oy [h6 [h7 h8]]]]]].
  exists ox, oy, h6, h7; simpl; auto.
  pose proof (proof_irrelevance _ h2 h4) as h9.
  rewrite h9 at 1.
  unfold y in h8. unfold y.
  rewrite h8.
  repeat f_equal.
  apply proof_irrelevance.
Qed.
 
Lemma incl_graph_seg_lists_to_rel_nth_lt_compat :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (i:nat)
         (pfi:i < length l)
         (f:sig_set (list_to_set l) -> U)
         (pfinc:Included (graph_seg f) (two_lists_to_rel l l' pfe)),
  exists (i':nat) (pfi':i' < length l),
    nth_lt l i pfi = nth_lt l i' pfi' /\
    let pfn := in_nth_lt l i' pfi' in
    let pfn' := iff1 (list_to_set_in_iff l _) pfn in
    f (exist _ _ pfn') =
    nth_lt l' i' (lt_eq_trans _ _ _ pfi' pfe).
  intros T U h1 h2 l l' h3 i h4 f h5.
  red in h5.
  specialize
    (h5
       ((nth_lt l i h4),
        (f (exist (Inn (list_to_set l)) (nth_lt l i h4) (iff1 (list_to_set_in_iff l _) (in_nth_lt _ _ _)))))).
  hfirst h5.
  unfold graph_seg.
  apply Im_intro with (exist (Inn (list_to_set l)) (nth_lt l i h4)
          (iff1 (list_to_set_in_iff l (nth_lt l i h4)) (in_nth_lt l i h4))).
  constructor.
  simpl; auto.
  specialize (h5 hf).
  apply (in_two_lists_to_rel h1 h2) in h5.
  destruct h5 as [h5 [h8 [ox [oy [h9 [h10 h11]]]]]].
  exists (lind_occ h1 l ox (nth_lt l i h4) h5), (lt_lind_occ _ _ _ _ _).
  split.
  rewrite lind_occ_compat; auto.
  simpl.
  pose proof (nth_lt_functional2 _ _ _ h11 (lt_eq_trans _ _ _ (lt_lind_occ _ _ _ _ _) h3)) (lt_lind_occ _ _ _ _ _) as h12.
  rewrite lind_occ_compat in h12.
  rewrite h12.
  f_equal.
  apply proj1_sig_injective; simpl.
  rewrite lind_occ_compat; auto.
Qed.



Definition two_lists_to_rel_im
           {T U:Type} (pfdt:type_eq_dec T) (l:list T) (l':list U)
           (pfe:length l = length l') (x:T) : Ensemble U :=
  list_to_set (map_in_dep (fun (n:nat)
                               (pfi:In n (linds_occ pfdt l x)) =>
                             let pflt := in_linds_occ_lt _ _ _ _ pfi in
                             nth_lt l' n (lt_eq_trans _ _ _ pflt pfe))).


Definition two_lists_to_rel_inv_im
           {T U:Type} (pfdu:type_eq_dec U) (l:list T) (l':list U)
           (pfe:length l = length l') (y:U) : Ensemble T :=
  list_to_set (map_in_dep (fun (n:nat)
                               (pfi:In n (linds_occ pfdu l' y)) =>
                let pflt := in_linds_occ_lt _ _ _ _ pfi in
                nth_lt l n (lt_eq_trans _ _ _ pflt (eq_sym pfe)))).

Lemma inh_two_lists_to_rel_im :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (x:T),
    In x l ->
    Inhabited (two_lists_to_rel_im pfdt l l' pfe x).
  intros T U hdt hdu l l' h1 x h2.
  econstructor.
  unfold two_lists_to_rel_im.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  pose proof (ex_lind_occ_nth_lt hdu _ _ (lt_eq_trans _ _ _ (lt_lind hdt _ _ h2) h1)) as h3.
  destruct h3 as [occ' [h3 h4]].
  pose proof (ex_lind_occ_nth_lt hdt _ _ (lt_lind hdt _ _ h2)) as h5.
  destruct h5 as [occ [h5 h6]].
  terml h4.
  terml h6.  
  exists x1.
  ex_goal.
  rewrite in_linds_occ_iff.
  exists occ, h2.
  rewrite lind_compat in h5.
  split; auto.
  unfold x1.
  apply lind_occ_functional; auto.
  rewrite lind_compat; auto.
  exists hex.
   unfold x1.
  rewrite <- h6 in h4.
  gen0.
  rewrite <- h4.
  intro h7. 
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat.
  reflexivity.
Qed.


Lemma inh_two_lists_to_rel_inv_im :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (y:U),
    In y l' ->
    Inhabited (two_lists_to_rel_inv_im pfdu l l' pfe y).
  intros T U hdt hdu l l' h1 x h2.
  econstructor.
  unfold two_lists_to_rel_inv_im.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_in_dep_iff.
  pose proof (ex_lind_occ_nth_lt hdt _ _ (lt_eq_trans _ _ _ (lt_lind hdu _ _ h2) (eq_sym h1))) as h3.
  destruct h3 as [occ' [h3 h4]].
  pose proof (ex_lind_occ_nth_lt hdu _ _ (lt_lind hdu _ _ h2)) as h5.
  destruct h5 as [occ [h5 h6]].
  terml h4.
  terml h6.  
  exists x1.
  ex_goal.
  rewrite in_linds_occ_iff.
  exists occ, h2.
  rewrite lind_compat in h5.
  split; auto.
  unfold x1.
  apply lind_occ_functional; auto.
  rewrite lind_compat; auto.
  exists hex.
   unfold x1.
  rewrite <- h6 in h4.
  gen0.
  rewrite <- h4.
  intro h7. 
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat.
  reflexivity.
Qed.


Lemma in_two_lists_to_rel_im :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (x:T) 
         (y:U),
    Inn (two_lists_to_rel_im pfdt l l' pfe x) y ->
    exists (pfx:In x l) (pfy:In y l') (occ:nat) (occ':nat),
      occ < count_occ pfdt l x /\
      occ' < count_occ pfdu l' y /\
      lind_occ pfdt l occ x pfx =
      lind_occ pfdu l' occ' y pfy.
  intros T U h1 h2 l l' h3 x y h5.
  unfold two_lists_to_rel_im in h5.
  rewrite <- list_to_set_in_iff in h5.
  rewrite in_map_in_dep_iff in h5.
  destruct h5 as [x' [h6 h7]].
  subst.
  pose proof h6 as h6'.
  rewrite in_linds_occ_iff in h6'.
  destruct h6' as [occ [h7 [h8 h9]]]. 
  subst.
  exists h7, (in_nth_lt _ _ _), occ.
  pose proof (ex_lind_occ_nth_lt h2 _ _ (lt_eq_trans (lind_occ h1 l occ x h7) (length l) 
             (length l') (in_linds_occ_lt h1 l x (lind_occ h1 l occ x h7) h6)
             h3)) as h9.
  destruct h9 as [occ' [h9 h10]].
  exists occ'.
  repeat split; auto.
Qed.


Lemma in_two_lists_to_rel_inv_im :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (x:T) (y:U),
    Inn (two_lists_to_rel_inv_im pfdu l l' pfe y) x ->
    exists  (pfy:In y l') (pfx:In x l) (occ:nat) (occ':nat),
      occ < count_occ pfdt l x /\
      occ' < count_occ pfdu l' y /\
      lind_occ pfdt l occ x pfx =
      lind_occ pfdu l' occ' y pfy.
  intros T U h1 h2 l l' h3 x y h5.
  unfold two_lists_to_rel_inv_im in h5.
  rewrite <- list_to_set_in_iff in h5.
  rewrite in_map_in_dep_iff in h5.
  destruct h5 as [x' [h6 h7]].
  subst.
  pose proof h6 as h6'.
  rewrite in_linds_occ_iff in h6'.
  destruct h6' as [occ [h7 [h8 h9]]]. 
  subst.
  pose proof (ex_lind_occ_nth_lt h1 _ _  (lt_eq_trans (lind_occ h2 l' occ y h7) (length l') 
             (length l)
             (in_linds_occ_lt h2 l' y (lind_occ h2 l' occ y h7) h6)
             (eq_sym h3))) as h9.
  destruct h9 as [occ' [h9 h10]]. 
  exists h7, (in_nth_lt _ _ _), occ', occ.
   repeat split; auto.
Qed.

Lemma in_two_lists_to_rel1 :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (x:T),
    In x l ->
    exists y:U,
      Inn (two_lists_to_rel_im pfdt l l' pfe x) y /\
      Inn  (two_lists_to_rel l l' pfe) (x, y).
  intros T U hdt hdu l l' h2 x h3.
  apply (inh_two_lists_to_rel_im hdt hdu l l' h2) in h3.
  destruct h3 as [y h4].
  exists y.
  split; auto. 
  apply (in_two_lists_to_rel_im hdt hdu l l' h2 x y) in h4.
  destruct h4 as [h4 [h5 [occ [occ' [h6 [h7 h8]]]]]].
  unfold two_lists_to_rel.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_seg_iff.
  exists (lind_occ hdt l occ x h4), (lt_lind_occ _ _ _ _ _).
  rewrite lind_occ_compat.
  f_equal.
  gen0.
  rewrite h8.
  intro h9.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat; auto.
Qed.


Lemma in_two_lists_to_rel2 :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (pfe:length l = length l') (y:U),
    In y l' ->
    exists x:T,
      Inn (two_lists_to_rel_inv_im pfdu l l' pfe y) x /\
      Inn (two_lists_to_rel l l' pfe) (x, y).
  intros T U hdt hdu l l' h2 y h3.
  apply (inh_two_lists_to_rel_inv_im hdt hdu l l' h2) in h3.
  destruct h3 as [x h4].
  exists x.
  split; auto. 
  apply (in_two_lists_to_rel_inv_im hdt hdu l l' h2 x y) in h4.
  destruct h4 as [h4 [h5 [occ [occ' [h6 [h7 h8]]]]]].
  unfold two_lists_to_rel.
  rewrite <- list_to_set_in_iff.
  rewrite in_map_seg_iff.
  exists (lind_occ hdu l' occ' y h4), (lt_eq_trans _ _ _ (lt_lind_occ _ _ _ _ _) (eq_sym h2)).
  f_equal.
  gen0.
  rewrite <- h8.
  intro h9.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat; auto.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat; auto.
Qed.


Definition two_lists_to_fun_in_dep
           {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
           (la:list T) (lb:list U)
           (pfl:linds pfdt la =
                linds pfdu lb) :
  fun_in_dep la U :=
  fun x (pfx:In x la) =>
    let pfl' := linds_eq_length pfdt pfdu la lb pfl in
    nth_lt lb (lind pfdt _ _ pfx) (lt_eq_trans _ _ _ (lt_lind pfdt _ _ pfx) pfl').

Lemma in_two_lists_to_fun_in_dep :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U)
         (pfl:linds pfdt la =
              linds pfdu lb)
         (x:T) (pf:In x la),
    In (two_lists_to_fun_in_dep pfdt pfdu la lb pfl x pf) lb.
  unfold two_lists_to_fun_in_dep; intros; apply in_nth_lt.
Qed.


Definition two_lists_to_sig_fun
           {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
           (la:list T) (lb:list U)
           (pfl:linds pfdt la =
                linds pfdu lb) :
  sig_set (list_to_set la) -> sig_set (list_to_set lb) :=
  fun x =>
    let pfx' := iff2 (list_to_set_in_iff la (proj1_sig x)) (proj2_sig x) in
    let pfl' := linds_eq_length pfdt pfdu la lb pfl in
    let pfi := in_two_lists_to_fun_in_dep pfdt pfdu la lb pfl in
    exist _ _ (iff1 (list_to_set_in_iff lb _) (in_nth_lt lb (lind pfdt _ _ pfx') (lt_eq_trans _ _ _ (lt_lind pfdt _ _ pfx') pfl'))).


Definition fun_linds_eq_nat 
           {U V:Type}
           (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
           (f:nat->U) (g:nat->V) : Prop :=
  forall n:nat,
    linds pfdu (map_seg (fun_to_seg f n)) =
    linds pfdv (map_seg (fun_to_seg g n)).



(*See below [linds_eq_lecs] for compatibility lemmas.*)

  
Definition list_to_segT_fun {T:Type} (l:list T) :
  length l ---> T.
  induction l as [|a l h1]; simpl. 
  intro n. destruct n as [n h1].
  apply lt_n_0 in h1. contradiction.
  intro n.
  destruct n as [n h2].
  destruct n.
  refine a.
  apply lt_S_n in h2.
  refine (h1 (exist (fun a => a < length l) _ h2)).
Defined.



Lemma list_to_segT_fun_compat :
   forall {T:Type} (l:list T),
    let f := list_to_segT_fun l in
    forall (n:(length l) @),
      In (f n) l /\
      f n =
      nth (proj1_sig n) l (hd' l (lt_length l _ (proj2_sig n))).
  intro T.
  induction l as [|a l h1]. 
  intros f n. simpl in n.  destruct n; omega.
  intros f n.
  simpl. simpl in n.
  unfold f.
  simpl.
  destruct n as [n h2].
  destruct n as [|n].
  split.
  left. reflexivity. simpl. reflexivity.
  pose proof h2 as h2'.
  apply lt_S_n in h2'.
  simpl in h1. specialize (h1 (exist (fun a => a < length l) _ h2')).
  destruct h1 as [h1 h3].
  split. right.
  pose proof (proof_irrelevance _ h2'
                                (lt_S_n _ _ h2));
  subst.
  simpl in h3. assumption. 
  simpl. simpl in h3.
  pose proof (@nth_indep _ l n (hd' l (lt_length l n h2')) (hd' (a::l) (lt_length (a::l) (S n) h2)) h2') as h4.
  simpl in h4.
  rewrite <- h4.
  rewrite <- h3.
  f_equal.
  apply proj1_sig_injective; auto.
Qed.



Lemma list_to_segT_fun_no_dup_inj :
  forall {T:Type} (l:list T),
    NoDup l ->
    Injective (list_to_segT_fun l).
  intros T l h1.
  red.
  intros m n h2.
  destruct m as [m h3], n as [n h4].
  apply proj1_sig_injective; simpl.
  pose proof (list_to_segT_fun_compat l (exist (fun a => a < length l) _  h3)) as h5.
  pose proof (list_to_segT_fun_compat l (exist (fun a => a < length l) _ h4)) as h6.
  simpl in h5, h6.
  destruct h5 as [h5 h5'].
  destruct h6 as [h6 h6'].
  rewrite h2 in h5'.
  rewrite h6' in h5'.
  apply nth_inj in h5'; auto. 
Qed.  



Definition list_to_segT_fun' {T:Type} (l:list T) :
  length l ---> {x:T | In x l}.
  intro x.
  pose proof (list_to_segT_fun_compat l x) as h1.
  destruct h1 as [h1 h2].
  refine (exist (fun a=> In a l) _ h1).
Defined.


Lemma list_to_segT_fun_prime_compat :
  forall {T:Type} (l:list T) (x:(length l) @),
    let f := list_to_segT_fun l in
    let f' := list_to_segT_fun' l in
    f x = proj1_sig (f' x).
  intros T l x f f'.
  unfold f', f, list_to_segT_fun'.
  destruct (list_to_segT_fun_compat l x) as [h1 h2].
  simpl.
  reflexivity.
Qed.


Lemma list_to_segT_fun_compat' :
   forall {T:Type} (l:list T),
    let f := list_to_segT_fun' l in
    forall (n:(length l) @),
      proj1_sig (f n) =
      nth (proj1_sig n) l (hd' l (lt_length l (proj1_sig n) (proj2_sig n))).
  intros T l f n.
  unfold f, list_to_segT_fun'.
  destruct (list_to_segT_fun_compat l n); auto.
Qed.



Lemma list_to_segT_fun_no_dup_inj' :
  forall {T:Type} (l:list T),
    NoDup l ->
    Injective (list_to_segT_fun' l).
  intros T l h1.
  pose proof (list_to_segT_fun_no_dup_inj l h1) as h2.
  red in h2. red.
  intros x y h3.
  destruct x as [x h5], y as [y h6].
  pose proof (f_equal (@proj1_sig _ _) h3) as h4.
  do 2 rewrite <- list_to_segT_fun_prime_compat in h4.
  specialize (h2 (exist (fun a => a < length l) _ h5) (exist (fun a => a < length l) _ h6) h4); auto.
Qed.

Lemma list_to_segT_fun_surj' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    surjective (list_to_segT_fun' l).
  intros T hdt l.
  red. intro y.
  destruct y as [y h1].
  pose  (exist (fun c => c < length l) (lind hdt _ _  h1) (lt_lind hdt  _ _ _)) as x.
  exists x.
  pose proof (list_to_segT_fun_compat' l x) as h2.
  apply proj1_sig_injective; simpl.
  rewrite h2.
  pose proof (nth_lt_lind_eq hdt  _ _ h1) as h3. simpl in h3.
  rewrite <- h3.
  simpl.
  rewrite nth_lt_compat.
  apply nth_indep.
  apply lt_lind.
Qed.



Definition list_to_segT_fun'' {T:Type} (l:list T) :
  length l ---> {x:T | Inn (list_to_set l) x}.
  intro x.
  pose (list_to_segT_fun' l x) as x'.
  pose (proj2_sig x') as h1. simpl in h1.
  rewrite list_to_set_in_iff in h1.
  refine (exist _ _ h1).
Defined.
 


Lemma list_to_segT_fun_compat'' :
   forall {T:Type} (l:list T),
    let f := list_to_segT_fun'' l in
    forall (n:(length l) @),
      proj1_sig (f n) =
      nth (proj1_sig n) l (hd' l (lt_length l (proj1_sig n) (proj2_sig n))).
intros T l f n.
pose proof (list_to_segT_fun_compat' l n) as h2.
simpl in h2.
unfold f, list_to_segT_fun''.
simpl.
assumption.
Qed.



Lemma list_to_segT_fun_prime_compat' :
  forall {T:Type} (l:list T) (x:(length l) @),
    let f' := list_to_segT_fun' l in
    let f'' := list_to_segT_fun'' l in
    proj1_sig (f'' x) = proj1_sig (f' x).
  intros T l x f' f''.
  unfold f', f''.
  unfold list_to_segT_fun''.
  simpl.
  reflexivity.
Qed.



Lemma list_to_segT_fun_no_dup_inj'' :
  forall {T:Type} (l:list T),
    NoDup l ->
    Injective (list_to_segT_fun'' l).
  intros T l h1.
  pose proof (list_to_segT_fun_no_dup_inj' l h1) as h2.
  red in h2. red.
  intros x y h3.
  specialize (h2 x y).
  apply h2.
  apply proj1_sig_injective.
  do 2 rewrite <- list_to_segT_fun_prime_compat'.
  f_equal; auto.
Qed.  

Lemma list_to_segT_fun_surj'' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    surjective (list_to_segT_fun'' l).
  intros T hdt l.
  red. intro y.
  pose proof (list_to_segT_fun_surj' hdt l) as h1.
  red in h1.
  destruct y as [y h2].
  pose proof h2 as h2'.
  rewrite <- list_to_set_in_iff in h2'.
  specialize (h1 (exist (fun a => In a l) _ h2')).
  destruct h1 as [x h1].
  exists x. 
  apply proj1_sig_injective. simpl.
  rewrite h1. simpl.
  reflexivity.
Qed.
  

Definition list_to_seg_one_t_fun {T:Type} (l:list T) :
  seg_one_t (length l) -> T.
  induction l as [|a l h1]; simpl.
  intro x. destruct x as [x h1]. destruct h1 as [h1 h2]. 
  assert (h3:1 <= 0). eapply le_trans. apply h1. apply h2.
  apply le_Sn_0 in h3. contradiction.
  intro x.
  destruct (seg_one_t_S_dec x) as [h2 | h2].
  refine a.
  refine (h1 (exist (fun a => 1 <= a <= length l) _ h2)).
Defined.


Lemma list_to_seg_one_t_fun_compat :
   forall {T:Type} (l:list T),
    let f := list_to_seg_one_t_fun l in
    forall (n:nat) (pf:n < length l),
      let n' := exist _ _ (iff1 (lt_sot_iff _ _) pf) in
      In (f n') l /\
      f n' =
      nth n l (hd' l (lt_length l n pf)).
  intros T l.
  induction l as [|a l h1].
  intros f n h2. unfold f, list_to_seg_one_t_fun.
  simpl in h2. pose proof (lt_n_0 _ h2). contradiction.
  intros f n h2.
  simpl. simpl in h1.
  destruct n as [|n].
  unfold f.
  simpl.
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  intuition.
  simpl in hr.
  omega.
  simpl in h2.
  pose proof (lt_S_n _ _ h2) as h3.
  specialize (h1 _ h3).
  unfold f.
  simpl. 
  lr_match_goal; fold hlr; destruct hlr as [hl | hr].
  simpl in hl.
  omega.
  simpl in hr.
  pose proof (proof_irrelevance _ hr (iff1 (lt_sot_iff _ _) h3)); subst.
  destruct h1 as [h1 h1'].   
  split.   
  right; auto.
  rewrite h1'.
  apply nth_indep; auto.
Qed.


Lemma list_to_seg_one_t_fun_no_dup_inj :
  forall {T:Type} (l:list T),
    NoDup l ->
    Injective (list_to_seg_one_t_fun l).
  intros T l h1.
  red.
  intros m n h2.
  destruct m as [m h3], n as [n h4].
  apply proj1_sig_injective; simpl.
  pose proof (list_to_seg_one_t_fun_compat l (pred m)) as h5.
  pose proof (list_to_seg_one_t_fun_compat l (pred n)) as h6.
  assert (h7:pred m < length l). omega.
  assert (h8:pred n < length l). omega.
  specialize (h5 h7). specialize (h6 h8).
  assert (h9:S (pred n) = n). omega.
  assert (h10:S (pred m) = m). omega.
  assert (h11:list_to_seg_one_t_fun l
         (exist (fun a : nat => 1 <= a <= length l) 
                (S (pred m)) (iff1 (lt_sot_iff _ _) h7)) =
              list_to_seg_one_t_fun l
                                        (exist (fun m0 : nat => 1 <= m0 <= length l) m h3)).
    f_equal. apply proj1_sig_injective; auto.
  assert (h12: list_to_seg_one_t_fun l
         (exist (fun a : nat => 1 <= a <= length l) 
                (S (pred n)) (iff1 (lt_sot_iff _ _) h8)) =
                list_to_seg_one_t_fun l
                                          (exist (fun m0 : nat => 1 <= m0 <= length l) n h4)).
    f_equal. apply proj1_sig_injective; auto.
  simpl in h5. destruct h5 as [h5 h5'].
  simpl in h6. destruct h6 as [h6 h6'].
  rewrite h12 in h6'. rewrite h11 in h5'.
  rewrite h2 in h5'.
  rewrite h6' in h5'.
  apply nth_inj in h5'; auto. omega.
Qed.  



Lemma im_list_to_segT_fun :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    Im (Full_set _) (list_to_segT_fun l) = 
    list_to_set l.
  intros T hdt l.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  clear h1.
  rewrite <- list_to_set_in_iff.
  pose proof (list_to_segT_fun_compat l x) as h1.
  destruct h1 as [h1 h2].
  rewrite h2.
  apply nth_In.
  apply proj2_sig.
  rewrite <- list_to_set_in_iff in h1.
  pose (exist (fun c => c < length l) (lind hdt _ _ h1) (lt_lind hdt _ _ h1)) as e.
  apply Im_intro with e.
  constructor.
  pose proof (list_to_segT_fun_compat l e) as h2.
  pose proof (nth_lt_lind_eq hdt _ _ h1) as h3.
  simpl in h3.
  destruct h2 as [h2a h2b].
  rewrite h2b.
  rewrite <- h3 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  apply proj2_sig.
Qed.


Definition list_to_seg_one_t_fun' {T:Type} (l:list T) :
  seg_one_t (length l) -> {x:T | In x l}.
  induction l as [|a l h1]; simpl.
  intro x. destruct x as [x h1]. destruct h1 as [h1 h2]. 
  assert (h3:1 <= 0). eapply le_trans. apply h1. apply h2.
  apply le_Sn_0 in h3. contradiction.
  intro x.
  destruct (seg_one_t_S_dec x) as [h2 | h2].
  refine (exist _ a _). left; auto.
  specialize (h1 (exist _ _ h2)).
  destruct h1 as [x' h1].
  refine (exist _ x' _). right; auto.
Defined.


Lemma list_to_seg_one_t_fun_prime_compat :
  forall {T:Type} (l:list T) (x:seg_one_t (length l)),
    let f := list_to_seg_one_t_fun l in
    let f' := list_to_seg_one_t_fun' l in
    f x = proj1_sig (f' x).
  intros T l.
  induction l as [|a l h1].
  intros x f f'.
  unfold f, f'.
  destruct x as [m h1].
  simpl in h1. omega.
  intros x f f'.
  unfold f, f'. simpl.
  destruct (seg_one_t_S_dec x) as [h2 | h2].
  simpl. reflexivity.
  rewrite h1.

  destruct ( @list_to_seg_one_t_fun' T l
            (@exist nat
               (fun m : nat => and (le (S O) m) (le m (@length T l)))
               (pred
                  (@proj1_sig nat
                     (fun m : nat =>
                      and (le (S O) m) (le m (S (@length T l)))) x)) h2)).
  simpl.
  reflexivity.
Qed.

Lemma list_to_seg_one_t_fun_compat' :
   forall {T:Type} (l:list T),
    let f := list_to_seg_one_t_fun' l in
    forall (n:nat) (pf:n < length l),
      let n' := exist _ _ (iff1 (lt_sot_iff _ _) pf) in
      proj1_sig (f n') =
      nth n l (hd' l (lt_length l n pf)).
  intros T l f n h1. simpl.
  pose proof (list_to_seg_one_t_fun_compat l n h1) as h2.
  simpl in h2.
  destruct h2 as [h2 h3].
  unfold f.
  rewrite <- list_to_seg_one_t_fun_prime_compat.
  rewrite h3.
  reflexivity.
Qed.


Lemma list_to_seg_one_t_fun_no_dup_inj' :
  forall {T:Type} (l:list T),
    NoDup l ->
    Injective (list_to_seg_one_t_fun' l).
  intros T l h1.
  pose proof (list_to_seg_one_t_fun_no_dup_inj l h1) as h2.
  red in h2. red.
  intros x y h3.
  destruct x as [x h5], y as [y h6].
  pose proof (f_equal (@proj1_sig _ _) h3) as h4.
  do 2 rewrite <- list_to_seg_one_t_fun_prime_compat in h4.
  specialize (h2 (exist _ _ h5) (exist _ _ h6) h4); auto.
Qed.


Lemma im_list_to_segT_fun' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    im_proj1_sig (Im (Full_set _) (list_to_segT_fun' l)) = 
    list_to_set l.
  intros T hdt l.
  unfold im_proj1_sig.
  rewrite im_im.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  clear h1.
  rewrite <- list_to_set_in_iff.
  pose proof (list_to_segT_fun_compat' l x) as h1.
  rewrite h1 at 1.
  apply nth_In.
  apply proj2_sig.
  rewrite <- list_to_set_in_iff in h1.
  pose (exist (fun c=> c < length l) (lind hdt _ _ h1) (lt_lind hdt _ _ h1)) as e.
  apply Im_intro with e.
  constructor.
  pose proof (list_to_segT_fun_compat' l e) as h2.
  pose proof (nth_lt_lind_eq  hdt l x h1) as h3.
  simpl in h3.
  rewrite h2 at 1.
  rewrite <- h3 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  apply proj2_sig.
Qed.


Definition list_to_seg_one_t_fun'' {T:Type} (l:list T) :
  seg_one_t (length l) -> {x:T | Inn (list_to_set l) x}.
  intro x.
  pose (list_to_seg_one_t_fun' l x) as x'.
  pose (proj2_sig x') as h1. simpl in h1.
  rewrite list_to_set_in_iff in h1.
  refine (exist _ _ h1).
Defined.

Lemma list_to_seg_one_t_fun_compat'' :
   forall {T:Type} (l:list T),
    let f := list_to_seg_one_t_fun'' l in
    forall (n:nat) (pf:n < length l),
      let n' := exist _ _ (iff1 (lt_sot_iff _ _) pf) in
      proj1_sig (f n') =
      nth n l (hd' l (lt_length l n pf)).
intros T l f n h1 n'.
pose proof (list_to_seg_one_t_fun_compat' l n) as h2.
simpl in h2.
specialize (h2 h1).
unfold f, list_to_seg_one_t_fun''.
simpl.
unfold n'.
assumption.
Qed.

Lemma list_to_seg_one_t_fun_prime_compat' :
  forall {T:Type} (l:list T) (x:seg_one_t (length l)),
    let f' := list_to_seg_one_t_fun' l in
    let f'' := list_to_seg_one_t_fun'' l in
    proj1_sig (f'' x) = proj1_sig (f' x).
  intros T l x f' f''.
  unfold f', f''.
  unfold list_to_seg_one_t_fun''.
  simpl.
  reflexivity.
Qed.


Lemma list_to_seg_one_t_fun_no_dup_inj'' :
  forall {T:Type} (l:list T),
    NoDup l ->
    Injective (list_to_seg_one_t_fun'' l).
  intros T l h1.
  pose proof (list_to_seg_one_t_fun_no_dup_inj' l h1) as h2.
  red in h2. red.
  intros x y h3.
  specialize (h2 x y).
  apply h2.
  apply proj1_sig_injective.
  do 2 rewrite <- list_to_seg_one_t_fun_prime_compat'.
  f_equal; auto.
Qed.  


Lemma im_list_to_segT_fun'' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    im_proj1_sig (Im (Full_set _) (list_to_segT_fun'' l)) = 
    list_to_set l.
  intros T hdt l.
  unfold im_proj1_sig.
  rewrite im_im.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  destruct h1 as [x h1]. subst.
  clear h1.
  rewrite <- list_to_set_in_iff.
  pose proof (list_to_segT_fun_compat'' l x) as h1.
  rewrite h1.
  apply nth_In.
  apply proj2_sig.
  rewrite <- list_to_set_in_iff in h1.
  pose (exist (fun c=> c < length l) (lind hdt _ _ h1) (lt_lind hdt _ _ h1)) as e.
  apply Im_intro with e.
  constructor.
  pose proof (list_to_segT_fun_compat'' l e) as h2.
  pose proof (nth_lt_lind_eq hdt l x h1) as h3.
  simpl in h3.
  rewrite h2.
  rewrite <- h3 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  apply proj2_sig.
Qed.

Definition list_to_segT_card_fun {T:Type} (pfdt:type_eq_dec T)
           (l:list T) : 
  card_fun1 (list_to_set l) ---> sig_set (list_to_set l) :=
  fun x => let pf0 := lt_eq_trans _ _ _ (proj2_sig x)
                                  (card_fun1_length_eq pfdt l) in
           let pf1 := in_nth_lt _ _  pf0 in
           let pf2 := iff2 (list_singularize_in_iff pfdt l _) pf1 in
           let pf3 := iff1 (list_to_set_in_iff _  (nth_lt _ _ pf0)) pf2 in
           exist (fun c => Inn (list_to_set l) c) _ pf3.

Lemma list_to_segT_card_fun_compat : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (x:(card_fun1 (list_to_set l)) @),
    let f := list_to_segT_card_fun pfdt l in
    let pf0 := lt_eq_trans _ _ _ (proj2_sig x)
                           (card_fun1_length_eq pfdt l) in
    proj1_sig (f x) = nth_lt _ _ pf0.
  intros T h1 l x f h2.
  unfold f.
  unfold list_to_segT_card_fun; simpl.
  apply nth_lt_functional3.
Qed.

Lemma list_to_segT_card_fun_inj : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    let f := list_to_segT_card_fun pfdt l in
    Injective f.
  intros T h1 l f.
  red; intros x y h2.
  unfold f, list_to_segT_card_fun in h2.
  apply exist_injective in h2.
  apply nth_lt_inj in h2.
  apply proj1_sig_injective in h2; auto.
  apply no_dup_list_singularize.
Qed.

Lemma list_to_segT_card_fun_surj : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    let f := list_to_segT_card_fun pfdt l in
    surjective f.
  intros T h1 l f.
  red; intro x.
  destruct x as [x h2].
  pose proof h2 as h2'.
  rewrite <- list_to_set_in_iff in h2'.
  rewrite  (list_singularize_in_iff h1) in h2'.
  pose proof (lt_lind h1 _ _ h2') as h3. 
  rewrite <- card_fun1_length_eq in h3.
  exists (exist (fun c => c < card_fun1 (list_to_set l))  _ h3).
  unfold f, list_to_segT_card_fun.
  simpl.
  apply proj1_sig_injective; simpl.
  gen0.
  erewrite lind_functional.
  intro h4.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  reflexivity.
  reflexivity.

  Grab Existential Variables.
  assumption.
Qed.

Lemma list_to_segT_card_fun_bij : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    let f := list_to_segT_card_fun pfdt l in
    bijective f.
intros; red; split; [apply list_to_segT_card_fun_inj; auto |
                     apply list_to_segT_card_fun_surj; auto].
Qed.
                    



Lemma im_list_to_segT_card_fun :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    im_proj1_sig (Im (Full_set _) (list_to_segT_card_fun pfdt l)) = 
    list_to_set l.
  intros T h1 l.
  unfold im_proj1_sig.
  rewrite im_im.
  unfold list_to_segT_card_fun.
  apply Extensionality_Ensembles; red; split; red; intros x h2.
  destruct h2 as [x h2]; subst.
  rewrite <- list_to_set_in_iff; simpl.
  rewrite (list_singularize_in_iff h1).
  apply in_nth_lt.
  simpl.
  rewrite <- list_to_set_in_iff in h2.
  rewrite (list_singularize_in_iff h1) in h2.
  pose proof (lt_lind h1 _ _ h2) as h3.
  rewrite <- (card_fun1_length_eq h1 l) in h3.
  apply Im_intro with (exist (fun c => c < card_fun1 (list_to_set l))  _ h3).
  constructor.
  simpl.
  symmetry.
  gen0. 
  erewrite lind_functional.
  intro.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  reflexivity.
  reflexivity.

  Grab Existential Variables.
  assumption.
Qed.


(*This defines a lexicographic ordering relation on the list of naturals,
  which is useful in the next definition.*)

(*The extra verbase [tso_dec] condition is included for compatibility 
with essential lemmas.*)
Definition le_list_nat (l m:list nat) :=
  le_listR lt (sort_dec_tso_dec _ sort_dec_lt_nat) l m.

Definition lt_list_nat (l m:list nat) : Prop :=
  lt_listR lt (sort_dec_tso_dec _ sort_dec_lt_nat)  l m.

End ListNat.


Lemma length_list_of_lists_seqs :
  forall {T:Type} (ll:faml T), 
    length (list_of_lists_seqs ll) =
    timesl (map (@length T) ll).
  intros T ll.
  induction ll as [|al ll ih]; simpl; auto.
  rewrite map_length.
  rewrite prod_length.
  f_equal; auto.
Qed.






Notation "l /= m" := (le_list_nat l m) (at level 20) : list_scope.
Notation "l /_ m" := (lt_list_nat l m) (at level 20) : list_scope.

Section ListNat'.

Lemma lt_list_nat_impl_le :
  forall (l m:list nat),
    l /_ m -> l /= m.
  unfold lt_list_nat, le_list_nat; apply lt_list_impl_leR.
Qed.

Lemma not_lt_list_nat_nil :
  forall (l:list nat),
    ~ l /_ nil.
  unfold lt_list_nat; apply not_lt_listR_nil.
Qed.
  
Lemma le_list_nat_nil :
  forall (l:list nat),
    nil /= l.
  unfold lt_list_nat, le_list_nat; apply le_listR_nil.
Qed.

Lemma le_list_nat_dec :   
  forall (l m:list nat), {l /= m} + {~l /= m}.
  unfold le_list_nat; apply le_listR_dec.
Qed.


Lemma refl_le_list_nat :
  forall (l:list nat),
    l /= l.
  unfold le_list_nat; apply refl_le_listR.
  red; apply lt_irrefl.
Qed.
 
Lemma antisym_le_list_nat :
  forall (l m:list nat),
    l /= m -> m /= l -> l = m.
  unfold le_list_nat; intros; eapply antisym_le_listR.
  apply H.
  assumption.
Qed.

Lemma antisym_not_le_list_nat :
  forall (l m:list nat),
    ~ l /= m -> ~ m /= l -> l = m.
  unfold le_list_nat; intros; eapply antisym_not_le_listR.
  apply H. apply H0.
Qed.


Lemma lt_list_nat_iff :
  forall (l m:list nat),
    l /_ m <-> ~ m /= l.
  unfold lt_list_nat, le_list_nat; intros; eapply lt_listR_iff.
Qed.

Lemma lt_list_nat_cons :
  forall (l m:list nat) (a:nat),
    (a::l) /_ (a::m) ->
    l /_ m.
  unfold lt_list_nat; intros; eapply lt_listR_cons.
  apply H.
Qed.

Lemma lt_list_nat_cons_iff :
  forall (l l':list nat) (a:nat)
         (pfn:l <> nil),
    l /_ (a::l') <->
    (hd' l pfn) < a \/
    (hd' l pfn = a /\ (tl l) /_ l').
  unfold lt_list_nat; intros; apply lt_listR_cons_iff.
Qed.

Lemma le_list_nat_dec' :
  forall (l m:list nat),
    {l /= m} + {m /_ l}.
  unfold lt_list_nat, le_list_nat; eapply le_listR_dec'.
Qed.

Lemma lt_list_nat_tso_dec :
  tso_dec lt_list_nat.
  unfold lt_list_nat; intros; apply tso_dec_lt_listR.
Qed.


Lemma trans_le_list_nat :
  Transitive _ le_list_nat.
  unfold le_list_nat; intros; apply trans_le_listR.
Qed.


Lemma trans_le_lt_list_nat :
  forall (la lb lc:list nat),
    la /= lb -> lb /_ lc -> la /_ lc.
  unfold le_list_nat, lt_list_nat; intros; eapply trans_le_lt_listR.
  apply H. assumption.
Qed.


Lemma trans_lt_le_list_nat :
  forall (la lb lc:list nat),
    la /_ lb -> lb /= lc -> la /_ lc.
  unfold le_list_nat, lt_list_nat; intros; eapply trans_lt_le_listR.
  apply H. assumption.
Qed.

Lemma trans_lt_list_nat :
  Transitive _ lt_list_nat.
  unfold lt_list_nat; intros; apply trans_lt_listR.
Qed.

Lemma irrefl_lt_list_nat :
  forall (l:list nat),
    ~ l /_ l.
  unfold lt_list_nat; intros; apply irrefl_lt_listR.
Qed.

Lemma so_lt_list_nat :
  Strict_Order lt_list_nat.
  unfold lt_list_nat; intros; apply so_lt_listR.
Qed.


Lemma le_list_nat_lt_eq_dec :
  forall (l m:list nat),
    l /= m ->
    {l /_ m} + {l = m}.
  unfold lt_list_nat, le_list_nat; intros; apply le_listR_lt_eq_dec; auto.
Qed.

Lemma lt_list_nat_sort_dec :
  sort_dec lt_list_nat.
  apply lt_listR_sort_dec.
Qed.

Lemma lt_list_nat_irrefl :
  irrefl lt_list_nat.
  red; apply irrefl_lt_listR.
Qed.

  
Lemma lt_list_nat_sortable :
  sortable lt_list_nat.
  apply sort_dec_sortable; apply lt_list_nat_sort_dec.
Qed.
               
Lemma lt_list_nat_tso_dec_eq :
  lt_list_nat_tso_dec = 
  sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec.
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality_dep.
  intro y.
  destruct (lt_list_nat_tso_dec x y) as [h1 | h1].
  destruct h1 as [h1 | h1].
  destruct ( sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec x y) as [h2 | h2].
  destruct h2 as [h2 | h2].
  repeat f_equal.
  apply proof_irrelevance.
  pose proof h1 as h1'.
  rewrite h2 in h1'.
  apply lt_list_nat_irrefl in h1'.
  contradiction.
  pose proof (trans_lt_list_nat _ _ _ h1 h2) as h3.
  apply lt_list_nat_irrefl in h3.
  contradiction.
  subst.
  destruct (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec y y) as [h1 | h1].
  destruct h1 as [h1 | h1].
  contradict (lt_list_nat_irrefl _ h1).
  repeat f_equal.
  apply proof_irrelevance.
  contradict (lt_list_nat_irrefl _ h1).
  destruct (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec x y) as [h2 | h2].
  destruct h2 as [h2 | h2].
  pose proof (trans_lt_list_nat _ _ _ h1 h2) as h3.
  apply lt_list_nat_irrefl in h3.
  contradiction.
  subst.
  contradict (lt_list_nat_irrefl _ h1).
  f_equal.
  apply proof_irrelevance.
Qed.

Lemma list_nat_eq_dec_eq' :
  list_nat_eq_dec =
  sort_dec_eq_dec _ (lt_list_nat_sort_dec).
  apply functional_extensionality_dep.
  intro x.
  apply functional_extensionality_dep.
  intro y.
  destruct (list_nat_eq_dec x y) as [h1 | h1],
                                    (sort_dec_eq_dec lt_list_nat lt_list_nat_sort_dec x y) as [h2 | h2]; subst; f_equal; try apply proof_irrelevance.
  contradict (h2 eq_refl).
  contradict (h1 eq_refl).
Qed.

Lemma Rw_lt_list_nat_iff :
  forall l l':list nat,
  Rw lt_list_nat l l'<->
  le_list_nat l l'.
  intros l l'.
  split; intro h1.
  red in h1.
  destruct h1 as [h1 |]; subst.
  apply lt_list_nat_impl_le; auto.
  apply refl_le_list_nat.
  apply le_list_nat_lt_eq_dec in h1.
  destruct h1 as [h1|]; subst.
  left; auto.
  right; auto.
Qed.


Lemma le0_nil :
  forall (l:list nat),
    l <> nil ->
    (0::nil) /= l.
  intro l.
  induction l as [|a l ih].
  intro h1; contradict h1; auto.
  intro.
  unfold le_list_nat.
  simpl.
  destruct (sort_dec_tso_dec lt sort_dec_lt_nat 0 a) as [h1 | h1].
  destruct h1; auto.
  omega.
Qed.


Lemma lt_lS :
  forall (l l':list nat),
    l' <> nil ->
    (0::l) /_ lS l'.
  intros l l'.
  destruct l' as [|a l'].
  intro h1; contradict h1; auto.
  simpl.
  split.
  simpl.
  destruct (sort_dec_tso_dec lt sort_dec_lt_nat 0 (S a)) as [h1 | h1].
  destruct h1; auto.
  discriminate.
  omega.
  intro h1.
  inversion h1.
Qed.

Lemma lt_hd0_famlS :
  forall (l:list nat) (ll:faml nat)
         (pf:famlS ll <> nil),
    ~In nil ll ->
    (0::l) /_ (hd' (famlS ll) pf).
  intros l ll h1 hnin.
  pose proof (in_hd' (famlS ll) h1) as hi.
  
  pose proof h1 as h1'. rewrite famlS_eq_nil_iff in h1'. 
  generalize h1. rewrite (hd_compat' _ h1'). 
  rewrite famlS_cons.
  intro h0.
  
  simpl.
  apply lt_lS.
  intro h2.
  rewrite <- h2 in hnin.
  contradict hnin.
  apply in_hd'.
Qed.

Lemma no_dup_famlS :
  forall (ll:faml nat),
    NoDup ll ->
    NoDup (famlS ll).
  intro ll.
  induction ll as [|al ll ih]; simpl; auto.
  intro h1.
  pose proof (no_dup_cons_nin _ _ h1) as h2.
  pose proof (no_dup_cons _ _ h1) as h3.
  apply ih in h3.
  constructor.
  contradict h2.
  rewrite in_famlS_iff in h2.
  destruct h2 as [ln' [h4 h5]].
  apply inj_lS in h5; subst; auto.
  assumption.
Qed.

Definition seg_faml (n:nat) :=
  map (fun x => (x::nil)) (seg_list n) .



Lemma sing_unq_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    sing_unq (linds pfdt l).
  intros T h1 l; red.
  intros lm ln h2 h3 h4.
  rewrite in_linds_iff in h2.
  destruct h2 as [x [[h2 h5] h8]].
  rewrite in_linds_iff in h3.
  destruct h3 as [x' [[h3 h6] h7]]; subst.
  destruct (lt_list_nat_tso_dec (linds_occ h1 l x) (linds_occ h1 l x')) as [[h9 | h9] | h9].
  symmetry.
  eapply incrR_no_dup_minimal.
  apply nat_eq_dec.
  apply increasing_linds_occ.
  apply no_dup_linds_occ.
  symmetry; auto.
  apply lt_list_nat_impl_le.
  assumption.
  assumption.
  eapply incrR_no_dup_minimal.
  apply nat_eq_dec.
  apply increasing_linds_occ.
  apply no_dup_linds_occ.
  assumption.
  apply lt_list_nat_impl_le.
  assumption.
Qed.


Lemma no_dup_in_linds_prod :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds_prod pfdt l) ->
    NoDup ln.
  intros T h1 l ln h2.
  rewrite in_linds_prod_iff in h2.
  destruct h2 as [ha hb].
  rewrite no_dup_iff. 
  rewrite dup_iff'.
  intro h2.
  destruct h2 as [m [n [h2 [h3 [h4 h5]]]]].
  pose proof h2 as h2'.
  pose proof h3 as h3'.
  rewrite ha in h2', h3'.
  pose proof (hb _ h2) as h6.
  pose proof (hb _ h3) as h7.
  destruct h6 as [h6 h8],
                 h7 as [h7 h9].
  rewrite h5 in h8. 
  pose proof (partition_linds h1 l) as h12.
  pose proof (partition_faml_dec (linds h1 l) _ _ _ h12 
                                 (in_nth_lt _ _ h6)
                                 (in_nth_lt _ _ h7)) as h13.
  destruct h13 as [h13 | h13].
  pose proof (in_linds_list_to_set_injable h1 l) as h14.
  red in h14.
  specialize (h14 (nth_lt (linds h1 l) m h6)
                  (nth_lt (linds h1 l) n h7)
                  (in_nth_lt _ _ _)
                  (in_nth_lt _ _ _)
                  h13).
  apply nth_lt_inj in h14.
  contradiction.
  apply no_dup_linds.

  do 2 red in h13.
  rewrite list_to_set_in_iff in h8, h9.
  pose proof (Intersection_intro _ _ _ _ h8 h9) as h14.
  rewrite h13 in h14.
  contradiction.

  apply nat_eq_dec.
Qed.

(*Maps corresponding parts between the partitions [linds] and 
  [linds (remove).*)
Definition linds_remove_fun
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (r:T) 
           (ln:list nat) (pfln:In ln (linds pfdt l))
: list nat :=
  if (list_nat_eq_dec ln (linds_occ pfdt l r))
  then nil
  else let v := in_linds_val _ _ _ pfln in
       linds_occ pfdt (remove pfdt r l) v.


Lemma in_linds_remove_fun :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (r:T) (pf:In r l)
         (ln:list nat) (pfln:In ln (linds pfdt l)),
    ln <> linds_occ pfdt l r ->
    In (linds_remove_fun pfdt l r ln pfln) (linds pfdt (remove pfdt r l)).
  intros T h1 l r hr ln h2 h3.
  unfold linds_remove_fun.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr'].
  contradiction.
  rewrite in_linds_iff.
  exists (in_linds_val h1 l ln h2).
  red.
  split.
  split; auto.
  rewrite in_remove_iff.
  split.
  apply in_linds_val_in. 
  intro; subst.
  rewrite linds_occ_in_linds_val in hr'.
  contradict hr'; auto.
  intros x h4.
  destruct h4 as [h4 h5].
  apply inj_linds_occ in h5; auto.
  rewrite in_remove_iff.
  split.
  apply in_linds_val_in.
  intro; subst.
  rewrite linds_occ_in_linds_val in hr'; contradict hr'; auto.
Qed.


Lemma length_linds_remove_fun' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (r:T),
    In r l ->
    forall (ln:list nat) (pfln:In ln (linds pfdt l)),
      ln <> linds_occ pfdt l r ->
      length (linds_remove_fun pfdt l r ln pfln) =
      length (linds_occ pfdt l (in_linds_val pfdt l ln pfln)).
  intros T h1 l r hir ln h0 h2.  
  unfold linds_remove_fun.   
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]. 
  contradiction.
  unfold linds_occ.
  destruct in_dec as [h3 | h3].
  destruct in_dec as [h4 | h4].
  unfold linds_occ'.
  do 2 rewrite length_map_seg.
  rewrite count_occ_remove_neq; auto.
  intro; subst.
  rewrite linds_occ_in_linds_val in h2.
  contradict h2; auto.
  pose proof h3 as h3'.
  rewrite in_remove_iff in h3'.
  destruct h3'; contradiction.
  contradict h3.
  rewrite in_remove_iff.
  split.
  apply in_linds_val_in.
  intro; subst.
  rewrite linds_occ_in_linds_val in h2.
  contradict h2; auto.
Qed.


Lemma length_linds_remove_fun :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (r:T),
    In r l ->
    forall (ln:list nat) (pfln:In ln (linds pfdt l)),
      ln <> linds_occ pfdt l r ->
      length (linds_remove_fun pfdt l r ln pfln) =
      length ln.
  intros T h1 l r h2 ln h3 h4.
  unfold linds_remove_fun.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  contradiction.
  rewrite length_linds_occ; auto.
  rewrite count_occ_remove_neq; auto.
  rewrite count_occ_in_linds_val; auto.
  intro; subst.
  rewrite linds_occ_in_linds_val in hr.
  contradict hr; auto.
Qed.


Lemma count_occ_remove_in_linds_val :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (r:T),
    In r l ->
    forall (ln:list nat) (pfln:In ln (linds pfdt l)),
      ln <> linds_occ pfdt l r ->
      count_occ pfdt (remove pfdt r l) (in_linds_val pfdt l ln pfln) =
      length ln.
  intros T h1 l r h2 ln h3 h4.
  rewrite count_occ_remove_neq; auto.
  rewrite count_occ_in_linds_val; auto.
  intro; subst.
  rewrite linds_occ_in_linds_val in h4.
  contradict h4; auto.
Qed.



(*For each [ln] in [linds_prod] and an [x] in [l], calculates the 
  element of [ln] which is the index of [x] in [l].*)
Definition in_l_domain_linds_prod_ind
           {T:Type} (pfdt:type_eq_dec T)
           (l:list T) (x:T)
           (pfx:In x l)
           (ln:list nat)
           (pfln:In ln (linds_prod pfdt l)) :
  nat :=
let pfx' := iff1 (list_singularize_in_iff pfdt l x) pfx in
let i := lind pfdt (list_singularize pfdt l) x pfx' in
let pfeq := eq_sym (in_linds_prod_length pfdt l ln pfln) in
let pflt := lt_eq_trans _ _ _ (lt_lind pfdt _ _ pfx') pfeq in
nth_lt ln i pflt.


Lemma in_l_domain_linds_prod_ind_in : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (pfx:In x l)
         (ln:list nat)
         (pfln:In ln (linds_prod pfdt l)),
    In (in_l_domain_linds_prod_ind pfdt l x pfx ln pfln) ln.
  unfold in_l_domain_linds_prod_ind; intros; apply in_nth_lt.
Qed.

Lemma lind_in_l_domain_linds_prod_ind_lt :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (pfx:In x l)
         (ln:list nat)
         (pfln:In ln (linds_prod pfdt l)),
    let pfin := in_l_domain_linds_prod_ind_in pfdt l x pfx ln pfln in
    lind nat_eq_dec _ _ pfin < length (list_singularize pfdt l).
  intros T h1 l x h2 ln h3 h4.
  eapply lt_eq_trans.
  apply lt_lind.
  erewrite in_linds_prod_length.
  reflexivity.
  assumption.
Qed.

  
Lemma in_l_domain_linds_prod_ind_lt :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (pfx:In x l)
         (ln:list nat)
         (pfln:In ln (linds_prod pfdt l)),
    in_l_domain_linds_prod_ind pfdt l x pfx ln pfln < length l.
  intros T h1 l x h2 ln h3. 
  pose proof (in_l_domain_linds_prod_ind_in h1 l x h2 ln h3) as h4. 
  eapply in_linds_prod_lt. 
  apply h3.
  assumption.
Qed.



Lemma lind_list_singularize :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (pfx:In x l)
         (ln:list nat)
         (pfln:In ln (linds_prod pfdt l)),
    let pfi := iff1 (list_singularize_in_iff pfdt _ _) pfx in
    lind pfdt _ x pfi =
    lind nat_eq_dec _ _ (in_l_domain_linds_prod_ind_in pfdt l x pfx ln pfln).
  intros T h1 l x h2 ln h3 h4.
  unfold in_l_domain_linds_prod_ind.
  symmetry.
  erewrite lind_functional.
  rewrite lind_nth_lt_no_dup.
  reflexivity.
  eapply no_dup_in_linds_prod.
  apply h3.
  apply nth_lt_functional3.

  Grab Existential Variables.
  erewrite in_linds_prod_length.
  apply lt_lind.
  assumption.
  apply in_nth_lt.
Qed.


Lemma in_l_domain_linds_prod_ind_compat :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (pfx:In x l)
         (ln:list nat)
         (pfln:In ln (linds_prod pfdt l)),
    nth_lt l  _
           (in_l_domain_linds_prod_ind_lt pfdt l x pfx ln pfln) =
    x.
  intros T h1 l x h2 ln h3. 
  rewrite (nth_lt_eq_iff h1).
  exists h2.
  pose proof (lind_list_singularize h1 l x h2 ln h3) as h5.
  simpl in h5.
  pose proof h3 as h3'. 
  rewrite in_linds_prod_iff in h3'.
  destruct h3' as [h6 h7].
  pose proof (lind_in_l_domain_linds_prod_ind_lt h1 l x h2 ln h3) as h8.
  simpl in h8.
  rewrite <- h6 in h8.
  specialize (h7 _  h8).
  destruct h7 as [h7 h9].
  unfold linds in h9.
  pose proof (nth_lt_map (list_singularize h1 l) (linds_occ h1 l)  (lind nat_eq_dec ln (in_l_domain_linds_prod_ind h1 l x h2 ln h3)
                                                                             (in_l_domain_linds_prod_ind_in h1 l x h2 ln h3))) as h10.
  hfirst h10.
  rewrite <- h6.
  apply lt_lind.
  specialize (h10 hf h7).
  rewrite h10 in h9 at 1.
  rewrite in_linds_occ_iff in h9.
  destruct h9 as [occ [h12 [h9 h11]]].
  symmetry in h11.
  erewrite nth_lt_functional3 in h11.
  erewrite lind_compat in h11.
  exists occ.
  simpl in h5.
  revert h9.
  generalize hf.
  rewrite <- h5.
  intro h13.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  intro h14.
  split; auto.
  rewrite h11.
  apply lind_occ_functional; auto.
  symmetry.
  gen0.
  rewrite <- h5.
  intro h15.
  erewrite nth_lt_functional3.
  rewrite lind_compat.
  reflexivity. 
Qed.
 

Lemma in_l_domain_linds_prod_ind_compat' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (x:T)
         (pfx:In x l)
         (ln:list nat)
         (pfln:In ln (linds_prod pfdt l)),
    let pfx' := iff1 (list_singularize_in_iff pfdt _ _) pfx in
    let pfi := lind_in_l_domain_linds_prod_ind_lt pfdt l x pfx ln pfln in
    nth_lt (list_singularize pfdt l)  _ pfi = x.
Proof.
  intros T h1 l x h2 ln h3 h4 h5.
  pose proof (lind_list_singularize h1 l x h2 ln h3) as h6.
  simpl in h6.
  gen0.
  simpl.
  rewrite <- h6.
  intro h7.
  erewrite nth_lt_functional3.
  apply lind_compat.
Qed.




Definition two_lists_linds_prod_fun 
           {T U:Type}
           {la:list T} {lb:list U}
           (pfdt:type_eq_dec T)
           (pfe:length la = length lb)
           (ln:list nat)
           (pfln:In ln (linds_prod pfdt la)) :
  fun_in_dep la U :=
  (fun x pf =>  nth_lt _ _ (lt_eq_trans _ _ _ (in_l_domain_linds_prod_ind_lt pfdt la x pf ln pfln) pfe)).

Definition two_lists_linds_sig_fun 
           {T U:Type}
           {la:list T} {lb:list U}
           (pfdt:type_eq_dec T)
           (pfe:length la = length lb)
           (ln:list nat)
           (pfln:In ln (linds_prod pfdt la)) :
  sig_set (list_to_set la) -> U :=
  (fun x =>  nth_lt _ _
                    (lt_eq_trans
                       _ _ _ (in_l_domain_linds_prod_ind_lt pfdt la (proj1_sig x) (iff2 (list_to_set_in_iff _ _) (proj2_sig x)) ln pfln) pfe)).

Lemma two_lists_linds_sig_fun_compat :
      forall {T U:Type}
             {la:list T} {lb:list U}
             (pfdt:type_eq_dec T)
             (pfe:length la = length lb)
             (ln:list nat)
             (pfln:In ln (linds_prod pfdt la))
             (x:T) (pf:In x la),
        two_lists_linds_prod_fun pfdt pfe ln pfln x pf = 
        two_lists_linds_sig_fun pfdt pfe ln pfln
                                (exist _ _ (iff1 (list_to_set_in_iff _ _) pf)).
  unfold two_lists_linds_prod_fun, two_lists_linds_sig_fun; intros; simpl.
  apply nth_lt_functional2.
  f_equal.
  apply proof_irrelevance.
Qed.



Definition sort_set_nat (E:Ensemble nat) (pf:Finite E) : list nat :=
  sort_set lt sort_dec_lt_nat E pf.

Definition is_sorted_list_nat (l:list nat) : Prop :=
  l = sort_set_nat (list_to_set l) (list_to_set_finite l).

Lemma sort_set_nat_functional :
  forall (E E':Ensemble nat) (pf:Finite E) (pf':Finite E'),
    E = E' ->
    sort_set_nat E pf = sort_set_nat E' pf'.
  unfold sort_set_nat; intros; apply sort_set_functional; auto.
Qed.

Lemma in_list_reps_sort_set_nat :
  forall (E:Ensemble nat) (pf:Finite E),
    Inn (list_reps E)  (sort_set_nat E pf).
  unfold sort_set_nat; intros; apply in_list_reps_sort_set; auto.
Qed.

Lemma in_sort_set_nat_iff :
  forall (E:Ensemble nat) (pf:Finite E) (m:nat),
    Inn E m <->
    In m (sort_set_nat E pf).
  unfold sort_set_nat; intros; apply in_sort_set_iff; auto.
Qed.

Lemma inj_sort_set_nat :
  forall (A B:Ensemble nat) (pfa:Finite A) (pfb:Finite B),
    sort_set_nat A pfa =
    sort_set_nat B pfb ->
    A = B.
  unfold sort_set_nat; intros; eapply inj_sort_set; auto.
  apply H.
Qed.

(*Search below for [inj_sort_set_nat']*)
  
Lemma no_dup_sort_set_nat :
  forall (E:Ensemble nat) (pf:Finite E),
    NoDup (sort_set_nat E pf).
  unfold sort_set_nat; intros; eapply no_dup_sort_set; auto.
Qed.
  
Lemma sort_set_nat_compat :
  forall (E:Ensemble nat) (pf:Finite E),
    list_to_set (sort_set_nat E pf) = E.
  unfold sort_set_nat; intros; eapply sort_set_compat; auto.
Qed.

Lemma sort_set_nat_empty :
  sort_set_nat (Empty_set nat) (Empty_is_finite _ ) =
  nil.
  unfold sort_set_nat; intros; eapply sort_set_empty; auto.
Qed.


Lemma sort_set_nat_nil :
  forall (A:Ensemble nat) (pf:Finite A),
    sort_set_nat A pf = nil ->
    A = Empty_set _.
  unfold sort_set_nat; intros; eapply sort_set_nil; auto.
  apply H.
Qed.

Lemma inh_sort_set_nat :
  forall (A:Ensemble nat) (pf:Finite A),
    Inhabited A ->
    sort_set_nat A pf <> nil.
  unfold sort_set_nat; intros; eapply inh_sort_set; auto.
Defined.


Lemma sort_set_nat_sing :
  forall (n:nat) (pf:Finite (Singleton n)),
    sort_set_nat (Singleton n) pf = n::nil.
  unfold sort_set_nat; intros; eapply sort_set_sing; auto.
Qed.

Lemma nin_hd_tl_sort_set_nat' :
  forall (A:Ensemble nat) (pff:Finite A)
         (pfia:Inhabited A),
    let l := sort_set_nat A pff in
    let pfic := inh_sort_set_nat A pff pfia in 
    ~In (hd' l pfic) (tl l).
  simpl; unfold sort_set_nat, inh_sort_set_nat; intros; apply nin_hd_tl_sort_set'.
Qed.


Lemma hd_sort_set_nat' :
  forall (A:Ensemble nat) (pff:Finite A)
         (pfi:Inhabited A),
    hd' (sort_set_nat A pff) (inh_sort_set_nat A pff pfi) =
    min_set_nat A pff pfi.
  unfold sort_set_nat, min_set_nat; intros; erewrite hd_functional'; rewrite  hd_sort_set' at 1.
  apply f_equal_prop_dep; auto.
Qed.
 
Lemma in_nat :
  forall (A:Ensemble nat) (pff:Finite A) (pfi:Inhabited A) (m:nat),
    Inn A m ->
    m = min_set_nat A pff pfi \/ In m (tl (sort_set_nat A pff)).
  unfold sort_set_nat, min_set_nat, lt_sortable; intros.
  pose proof (in_min lt sort_dec_lt_nat A pff pfi m H) as h1.
  simpl in h1.
  destruct h1 as [h1 | h1].
  left; rewrite h1.
  apply f_equal_prop_dep; auto.
  right; auto.
Qed.

Lemma tl_sort_set_nat :
  forall (A:Ensemble nat) (pff:Finite A) (pfi:Inhabited A),
    tl (sort_set_nat A pff) =
    remove nat_eq_dec (min_set_nat A pff pfi) (sort_set_nat A pff).
  unfold sort_set_nat, min_set_nat, lt_sortable; intros. 
  erewrite min_set_sortable_functional, sort_set_functional.
  erewrite tl_sort_set.
  f_equal. apply type_eq_dec_eq.
  reflexivity.
  reflexivity.
  Grab Existential Variables.
  apply sort_dec_lt_nat.
  assumption.
  assumption.
Qed.

Lemma in0_sort_set_nat :
  forall (A:Ensemble nat) (pf:Finite A)
         (pfi:Inn A 0),
    hd' (sort_set_nat A pf) (inh_sort_set_nat A pf (Inhabited_intro _ _ _ pfi))  = 0.
  intros A h1 h2.
  rewrite <- (hd_hd_compat' _ _ 0). 
  unfold sort_set_nat, sort_set, min_set_sortable.
  destruct constructive_definite_description as [l h3].
  simpl.
  destruct h3 as [h3 h4].
  destruct l as [|a l]; simpl; auto.
  inversion h3 as [h5].
  destruct h5 as [h5 h6]. clear h3.
  subst.
  pose proof (no_dup_cons_nin _ _ h6) as h5.
  rewrite <- list_to_set_in_iff in h2.
  destruct h2 as [|h2]; auto.
  specialize (h4 (0::a::remove nat_eq_dec 0 l)).
  hfirst h4.
  constructor.
  constructor.
  split.
  simpl.
  rewrite <- subtract_remove_compat. 
  rewrite <- add_comm.
  f_equal.
  rewrite add_sub_compat_in; auto.
  rewrite <- list_to_set_in_iff.
  assumption.
  constructor.
  intro h7.
  destruct h7 as [| h7]; subst.
  contradiction.
  apply remove_In in h7 ;auto.
  constructor.
  contradict h5.
  rewrite in_remove_iff in h5.
  destruct h5; auto.
  pose proof (no_dup_cons _ _ h6) as h7.
  apply no_dup_remove; auto.
  rewrite in_sing_iff.
  intro h7.
  inversion h7 as [h8].
  clear h7 H.
  subst.
  contradiction.
  specialize (h4 hf).
  red in h4.
  destruct h4 as [h4 h7].
  simpl in h4.
  destruct (sort_dec_tso_dec lt sort_dec_lt_nat a 0) as [h8 | h8].
  destruct h8 as [h8 | h8].
  omega.
  assumption.
  omega.
Qed.

    
Lemma incr_sort_set_nat :
  forall (C:Ensemble nat) (pf:Finite C),
    increasing (sort_set_nat C pf).
  unfold sort_set_nat; intros; apply incrR_sort_set.
Qed.


  
Lemma incr_no_dup_minimal :
  forall (l l':list nat),
    increasing l ->
    NoDup l' ->
    list_to_set l = list_to_set l' ->
    l' /= l ->
    l = l'.
  unfold le_list_nat, increasing; intros; eapply incrR_no_dup_minimal.
  apply nat_eq_dec.
  apply H.
  assumption.
  assumption.
  apply H2.
Qed.


Lemma incr_impl_sort_set_nat :
  forall (l:list nat),
    increasing l ->
    l = sort_set_nat (list_to_set l) (list_to_set_finite _).
  unfold increasing, sort_set_nat; intros; eapply incrR_impl_sort_set; auto.
Qed.


Lemma sort_set_nat_cons :
  forall (A:Ensemble nat) (pf:Finite A) (ln:list nat) (a:nat),
    sort_set_nat A pf = a::ln ->
    ln = sort_set_nat (Subtract A a) (subtract_preserves_finite _ _ pf).
  unfold sort_set_nat; intros; eapply sort_set_cons; auto.
Qed.


Lemma max_set_nat_add_lst_sort_set:
  forall (A:Ensemble nat) (m:nat)
         (pf:Finite A) (pfi:Inhabited A),
    ~Inn A m ->
    let pf' := Add_preserves_Finite _ A m pf in
    let pfi' := inh_add A m in
    let pfn := inh_sort_set_nat _ _ pfi in
    m = max_set_nat (Add A m) pf' pfi' ->
    lst (sort_set_nat A pf) pfn < m.
  simpl; unfold sort_set_nat, max_set_nat; intros; eapply max_set_sortable_add_lst_sort_set; auto.
  erewrite max_set_sortable_functional.
  apply H0.
  reflexivity.
Qed.

Lemma sort_set_nat_app_add :
  forall (A A':Ensemble nat)
         (pfa:Finite A) (pfa':Finite A') (pfi:Inhabited A') (a:nat),
    ~Inn A a ->
    A' = Add A a ->
    a = max_set_nat A' pfa' pfi ->
    sort_set_nat A' pfa' =
    sort_set_nat A pfa ++ a::nil.
  unfold sort_set_nat, max_set_nat; intros; eapply sort_set_app_add; auto.
  rewrite H1.
  apply max_set_sortable_functional; auto.
  Grab Existential Variables.
  assumption.
Qed.

Lemma sort_set_nat_seg :
  forall (n:nat),
    sort_set_nat (seg n) (finite_seg n) =
    seg_list n.
  intro n.
  erewrite sort_set_nat_functional.
  erewrite incrR_impl_sort_set.
  apply sort_set_nat_functional.
  reflexivity.
  apply incr_seg_list.
  symmetry.
  apply list_to_set_seg_list_eq.
  Grab Existential Variables.
  rewrite list_to_set_seg_list_eq.
  apply finite_seg.
Qed.

Lemma tl_sort_set_nat_add :
  forall (A:Ensemble nat) (c:nat)
         (pf:Finite (Add A c)),
    list_to_set (tl (sort_set_nat (Add A c) pf)) =
    Subtract (Add A c) (min_set_nat _ pf (inh_add _ _)).
  unfold sort_set_nat, min_set_nat; intros.
  erewrite min_set_sortable_functional; auto; eapply tl_sort_set_add.
Qed.

Lemma in_tl_sort_set_nat :
  forall (A:Ensemble nat) (pf:Finite A) (m n:nat),
    Inn A m ->
    Inn A n ->
    m < n ->
    In n (tl (sort_set_nat A pf)).
  unfold sort_set_nat; intros; eapply in_tl_sort_set.
  apply H. assumption. assumption.
Qed.


Lemma list_to_set_lpred_sort_set_nat :
  forall (A:Ensemble nat) (pf:Finite A),
    list_to_set (lpred (sort_set_nat A pf)) =
    ens_pred A.
  intros A h1.
  apply Extensionality_Ensembles; red; split; red; intros x h4.
  rewrite <- list_to_set_in_iff in h4.
  unfold lpred in h4.
  rewrite in_map_iff in h4.
  destruct h4 as [m h4]. destruct h4 as [h4 h5]; subst.
  rewrite <- in_sort_set_nat_iff in h5.
  econstructor. apply h5. reflexivity.
  destruct h4 as [x h4]; subst.
  rewrite <- list_to_set_in_iff.
  unfold lpred.
  rewrite in_map_iff.
  exists x. split; auto.
  rewrite <- in_sort_set_nat_iff; auto.
Qed.

  
Lemma list_to_set_tl_lpred_sort_set_nat :
    forall (A:Ensemble nat) (pf:Finite A),
      Inn A 0 ->
      Inn A 1 ->
      list_to_set (tl (lpred (sort_set_nat A pf))) =
      ens_pred A.
  intros A h1 h2 h3.
  rewrite <- (list_to_set_lpred_sort_set_nat _ h1).
  apply Extensionality_Ensembles; red; split; red; intros m h4.
  rewrite <- list_to_set_in_iff in h4.
  rewrite <- list_to_set_in_iff.
  apply in_tl; auto.
  rewrite <- list_to_set_in_iff in h4.
  rewrite <- list_to_set_in_iff.
  rewrite tl_lpred.
  rewrite in_lpred_iff in h4.
  rewrite in_lpred_iff.
  destruct h4 as [m' h4].
  destruct h4 as [h4 h5]; subst.
  rewrite <- in_sort_set_nat_iff in h4. 
  apply (in_nat _ h1 (Inhabited_intro _ _ _ h2)) in h4.
  rewrite min_set_nat_0 in h4; auto.
  destruct h4 as [h4 | h4]; subst.
  exists 1.
   
  split.
  rewrite (tl_sort_set_nat _ _ (Inhabited_intro _ _ _ h2)).
  rewrite in_remove_iff.
  split.
  rewrite <- in_sort_set_nat_iff; auto.
  rewrite min_set_nat_0; auto.
  reflexivity.
  exists m'.
  split; auto.
Qed.


Lemma increasing_tl_lpred_sort_set_nat :
  forall (A:Ensemble nat) (pf:Finite A),
    increasing (tl (lpred (sort_set_nat A pf))).
  intros A h1.
  pose proof (finite_set_list_no_dup _ h1) as h2.
  destruct h2 as [l [h2 h3]]; subst.
  destruct (in_dec nat_eq_dec 0 l) as [h2 | h2], (in_dec nat_eq_dec 1 l) as [h4 | h4].
  rewrite tl_lpred.
  rewrite list_to_set_in_iff in h2. pose proof h2 as h2'.
  rewrite (in_sort_set_nat_iff _ h1) in h2.
  pose proof (no_dup_sort_set_nat _ h1) as h5. 
  rewrite <- (remove_hd_no_dup' nat_eq_dec _ (in_not_nil _ _ h2) h5).
  pose proof (in_inh _ _ h4) as h6.
  erewrite hd_functional'.
  rewrite (hd_sort_set_nat' _ _ h6) at 1.
  rewrite min_set_nat_0; auto.
  apply incr_lpred_remove_in0_in1.
  assumption.
  rewrite <- in_sort_set_nat_iff.
  rewrite <- list_to_set_in_iff; auto.
  apply incr_sort_set_nat.
  apply increasing_tl.
  apply incr_lpred_in0_nin1; auto.
  rewrite <- in_sort_set_nat_iff.
  rewrite <- list_to_set_in_iff; auto.
  rewrite <- in_sort_set_nat_iff.
  rewrite <- list_to_set_in_iff; auto.
  apply incr_sort_set_nat.
  apply increasing_tl.
  apply incr_lpred_nin0.
  rewrite <- in_sort_set_nat_iff.
  rewrite <- list_to_set_in_iff; auto.
  apply incr_sort_set_nat.
  apply increasing_tl.
  apply incr_lpred_nin0.
  rewrite <- in_sort_set_nat_iff.
  rewrite <- list_to_set_in_iff; auto.
  apply incr_sort_set_nat.
Qed.

Lemma lpred_sort_set_nat_in0_in1 : 
  forall (A:Ensemble nat) (pf:Finite A),
    Inn A 0  ->
    Inn A 1 ->
    lpred (sort_set_nat A pf) =
    0::sort_set_nat (ens_pred A) (finite_ens_pred _ pf).
  intros A h1 h2 h3.
  pose proof (in0_sort_set_nat A h1 h2) as h4.
  pose proof (hd_lpred' (sort_set_nat A h1)) as h5.
  hfirst h5.
  apply neq_nil_lpred.
  intro h6.
  pose proof h2 as h2'.
  rewrite (in_sort_set_nat_iff A h1) in h2'.
  rewrite h6 in h2'.
  contradict h2'.
  specialize (h5 hf).
  rewrite (hd_compat' _ hf).
  symmetry in h5.
  erewrite hd_functional' in h5. rewrite h4 in h5 at 1.
  simpl in h5.
  rewrite <- h5.
  f_equal.
  pose proof (incr_lpred_remove_in0_in1 (sort_set_nat A h1)) as h6.
  hfirst h6.
  rewrite <- in_sort_set_nat_iff; auto.
  specialize (h6 hf0).
  hfirst h6.
  rewrite <- in_sort_set_nat_iff; auto.
  specialize (h6 hf1).
  hfirst h6.
  apply incr_sort_set_nat.
  specialize (h6 hf2).
  pose proof (increasing_tl _ hf2) as h7. 
  pose proof (list_to_set_tl_lpred_sort_set_nat A h1 h2 h3) as h8.
  
  symmetry.
  gen0.
  rewrite <- h8.
  intro h9.
  symmetry.
  erewrite sort_set_nat_functional.
  apply incr_impl_sort_set_nat.
  apply increasing_tl_lpred_sort_set_nat.
  reflexivity.
Qed.

Lemma lpred_sort_set_nat_nin0 : 
  forall (A:Ensemble nat) (pf:Finite A),
    ~Inn A 0  ->
    lpred (sort_set_nat A pf) =
    sort_set_nat (ens_pred A) (finite_ens_pred _ pf).
  intros A h1 h2.
  pose proof (incr_lpred_nin0 (sort_set_nat A h1)) as h4.
  hfirst h4.
  contradict h2.
  rewrite <- in_sort_set_nat_iff in h2; auto.
  specialize (h4 hf).
  hfirst h4.
  apply incr_sort_set_nat.
  specialize (h4 hf0).
  symmetry.
  erewrite sort_set_nat_functional.
  symmetry.
  apply incr_impl_sort_set_nat.
  assumption.
  rewrite list_to_set_lpred_sort_set_nat; auto.
Qed.


Lemma lpred_sort_set_nat_in0_nin1 :
  forall (A:Ensemble nat) (pf:Finite A),
    Inn A 0 -> ~Inn A 1  ->
    lpred (sort_set_nat A pf) =
    sort_set_nat (ens_pred A) (finite_ens_pred _ pf).
  intros A h1 h2 h3.
  pose proof (incr_lpred_in0_nin1 (sort_set_nat A h1)) as h4.
  hfirst h4.
  rewrite <- in_sort_set_nat_iff; auto.
  specialize (h4 hf).
  hfirst h4.
  contradict h3.
  rewrite <- in_sort_set_nat_iff in h3; auto.
  specialize (h4 hf0).
  hfirst h4.
  apply incr_sort_set_nat.
  specialize (h4 hf1).
  symmetry.
  erewrite sort_set_nat_functional.
  symmetry.
  apply incr_impl_sort_set_nat.
  assumption.
  rewrite list_to_set_lpred_sort_set_nat; auto.
Qed.

Lemma list_minus_sort_set_nat :
  forall (A B:Ensemble nat) (pfa:Finite A) (pfb:Finite B)
         (pfab:Finite (Setminus A B)),
    list_minus nat_eq_dec (sort_set_nat A pfa) (sort_set_nat B pfb) =
    sort_set_nat (Setminus A B) pfab.
  intros A B h1 h2 h3.
  pose proof (list_to_set_list_minus nat_eq_dec (sort_set_nat A h1) (sort_set_nat B h2)) as h4.
  do 2 rewrite sort_set_nat_compat in h4.
  symmetry.
  gen0.
  rewrite <- h4.
  intro h5.
  symmetry.
  pose proof (proof_irrelevance _ h5 (list_to_set_finite _)); subst.
  apply incr_impl_sort_set_nat.
  apply incr_list_minus.
  apply incr_sort_set_nat.
Qed.


Definition fam_nat_to_ens_sort_set (F:Family nat) (pf:finite_mem F) : Ensemble (list nat) :=
  fam_to_ens_sort_set lt sort_dec_lt_nat F pf.

Lemma fam_nat_to_ens_sort_set_empty_eq :
  forall (pf:finite_mem (Empty_set (Ensemble nat))),
    fam_nat_to_ens_sort_set _ pf = Empty_set _.
  intro; unfold fam_nat_to_ens_sort_set; apply fam_to_ens_sort_set_empty_eq.
Qed.

Lemma fam_nat_to_ens_sort_set_sing_eq :
  forall (A:Ensemble nat) (pf:Finite A) (pffm:finite_mem (Singleton A)),
    fam_nat_to_ens_sort_set (Singleton A) pffm = Singleton (sort_set_nat A pf).
  unfold fam_nat_to_ens_sort_set; apply fam_to_ens_sort_set_sing_eq.
Qed.
 
Lemma finite_fam_nat_to_ens_sort_set :
  forall (F:Family nat) (pf:finite_mem F),
    Finite F ->
    Finite (fam_nat_to_ens_sort_set F pf).
  unfold fam_nat_to_ens_sort_set; apply finite_fam_to_ens_sort_set.
Qed.

Lemma inh_fam_nat_to_ens_sort_set :
  forall (F:Family nat) (pf:finite_mem F),
    Inhabited F ->
    Inhabited (fam_nat_to_ens_sort_set F pf).
  unfold fam_nat_to_ens_sort_set; apply inh_fam_to_ens_sort_set.
Qed.

 
Lemma fam_nat_to_ens_sort_set_functional :
  forall (F F':Family nat) (pf:finite_mem F) (pf':finite_mem F'),
    F = F' ->
    fam_nat_to_ens_sort_set F  pf  =
    fam_nat_to_ens_sort_set F' pf'.
  unfold fam_nat_to_ens_sort_set; apply fam_to_ens_sort_set_functional.
Qed.


Lemma fam_nat_to_ens_sort_set_subtract :
  forall (F:Family nat) (pff:Finite F) (pffm:finite_mem F) (l:list nat),
    Inn F (list_to_set l) ->
    increasing l ->
    Subtract (fam_nat_to_ens_sort_set F pffm) l =
    fam_nat_to_ens_sort_set (Subtract F (list_to_set l))
                            (finite_mem_subtract _ _ pffm ).
  unfold fam_nat_to_ens_sort_set; apply fam_to_ens_sort_set_subtract.
Qed.

Lemma im_fam_nat_to_ens_sort_set_list_to_set :
  forall (F:Family nat) (pf:finite_mem F),
    Im (fam_nat_to_ens_sort_set F pf) list_to_set = F.
  unfold fam_nat_to_ens_sort_set; apply im_fam_to_ens_sort_set_list_to_set.
Qed.

Lemma in_fam_nat_to_ens_sort_set_impl :
  forall (F:Family nat) (pf:finite_mem F) (l:list nat),
    Inn (fam_nat_to_ens_sort_set F pf) l ->
    Inn F (list_to_set l).
  unfold fam_nat_to_ens_sort_set; apply in_fam_to_ens_sort_set_impl.
Qed.

Lemma in_fam_nat_impl :
  forall (F:Family nat) (pffm:finite_mem F) (A:Ensemble nat)
         (pfin:Inn F A),
    Inn (fam_nat_to_ens_sort_set F pffm)
        (sort_set_nat A (pffm _ pfin)).
  unfold fam_nat_to_ens_sort_set, sort_set_nat; apply in_fam_impl.
Qed.

Lemma in_fam_nat_to_ens_sort_set_incr :
  forall (F:Family nat) (pffm:finite_mem F) (l:list nat),
    Inn (fam_nat_to_ens_sort_set F pffm) l ->
    increasing l.
  unfold fam_nat_to_ens_sort_set; apply in_fam_to_ens_sort_set_incr.
Qed.

Lemma setminus_sort_set_nat_fam :
  forall (A B:Family nat)
         (pfa:finite_mem A)
         (pfb:finite_mem B),
    let f := fun S : sig_set A =>
             sort_set_nat (proj1_sig S)
                                 (pfa (proj1_sig S) (proj2_sig S)) in
    let g := fun S : sig_set B =>
             sort_set_nat (proj1_sig S)
                                 (pfb (proj1_sig S) (proj2_sig S)) in
    let f' := restriction_sig f _ (setminus_inc _ _) in
    Setminus (Im (full_sig A) f)
             (Im (full_sig B) g) =
    Im (full_sig (Setminus A B)) f'.
  simpl; unfold sort_set_nat; apply setminus_sort_set_fam.
Qed.


(*Once again, the verbose [tso_dec] is for lemma compatibility.*)
Definition min_list_nat (l m:list nat) : list nat :=
  min_listR lt (sort_dec_tso_dec _ sort_dec_lt_nat) l m.

Definition max_list_nat (l m:list nat) : list nat :=
  max_listR lt (sort_dec_tso_dec _ sort_dec_lt_nat) l m.


Lemma min_list_nat_compat1 :
  forall (l m:list nat),
    min_list_nat l m /= l.
  unfold min_list_nat, le_list_nat; eapply min_listR_compat1.
Qed.

Lemma min_list_nat_compat2 :
  forall (l m:list nat),
    min_list_nat l m /= m.
  unfold min_list_nat, le_list_nat; eapply min_listR_compat2.
Qed.

Lemma min_list_nat_eq_compat1 :
  forall (la lb lc:list nat),
    lb /= lc ->
    min_list_nat la lb /=
                 min_list_nat la lc.
  unfold min_list_nat, le_list_nat; eapply min_listR_eq_compat1.
Qed.

Lemma min_list_nat_eq_compat2 :
  forall (la lb lc:list nat),
    la /= lb ->
    min_list_nat la lc /=
                 min_list_nat lb lc.
  unfold min_list_nat, le_list_nat; eapply min_listR_eq_compat2.
Qed.


Lemma min_list_nat_nil1 :
  forall (l:list nat),
    min_list_nat nil l = nil.
  unfold min_list_nat; eapply min_listR_nil1.
Qed.

Lemma min_list_nat_nil2 :
  forall (l:list nat),
    min_list_nat l nil = nil.
  unfold min_list_nat; eapply min_listR_nil2.
Qed.


Lemma min_list_nat_eq_cons :
  forall (l m:list nat) (b:nat),
    min_list_nat (b :: l) (b :: m) = b:: min_list_nat l m.
  unfold min_list_nat; eapply min_listR_eq_cons.
Qed.


Lemma min_list_nat_comm :
  forall (l m:list nat),
    min_list_nat l m = min_list_nat m l.
  unfold min_list_nat; apply min_listR_comm.
Qed.


Lemma min_list_nat_assoc :
  forall (la lb lc:list nat),
    min_list_nat la (min_list_nat lb lc) =
    min_list_nat (min_list_nat la lb) lc.
  unfold min_list_nat; apply min_listR_assoc.
Qed.

Lemma min_list_nat_idem :
  forall (l:list nat),
    min_list_nat l l = l.
  unfold min_list_nat; apply min_listR_idem.
Qed.

Lemma min_list_nat_mono :
  forall (l la lb:list nat),
    la /= lb ->
    min_list_nat l la /= min_list_nat l lb.
  unfold min_list_nat, le_list_nat; apply min_listR_mono.
Qed.

Definition list_nat_fun : Type := list_fun nat nat. 

Definition incr_list_nat_fun (f:list_nat_fun) : Prop :=
  incrR_list_fun lt lt
                 (sort_dec_tso_dec _ sort_dec_lt_nat)
                 (sort_dec_tso_dec _ sort_dec_lt_nat)
                 f.

Definition weak_incr_list_nat_fun (f:list_nat_fun) : Prop :=
  weak_incrR_list_fun lt lt
                      (sort_dec_tso_dec _ sort_dec_lt_nat)
                      (sort_dec_tso_dec _ sort_dec_lt_nat)
                      f.

Lemma incr_list_nat_fun_impl_weak :
  forall (f:list_nat_fun),
    incr_list_nat_fun f ->
    weak_incr_list_nat_fun f. 
  intros; do 3 red; intros.
  left.
  apply H in H0; auto.
Qed.

Lemma lt_map :
  forall (f:nat->nat),
    incr_fun f ->
    forall (l l':list nat),
      l /_ l' ->
      (map f l) /_ (map f l').
  unfold lt_list_nat; intros; eapply lt_listR_map.
  red. apply lt_irrefl.
  rewrite incr_fun_iff in H.
  red in H; auto.
  apply H0.
Qed.



Lemma incr_list_nat_fun_lS :
  incr_list_nat_fun lS. 
  do 3 red; intros; unfold lS; eapply lt_listR_map. 
  red; apply lt_irrefl.
  change (incr_fun' S).
  rewrite <- incr_fun_iff.
  apply incr_fun_S; auto.
  apply H.
Qed.

  
Lemma weak_incr_list_nat_fun_lS :
  weak_incr_list_nat_fun lS.
  apply incr_list_nat_fun_impl_weak; apply incr_list_nat_fun_lS.
Qed.

Lemma incr_in0 :
  forall (l:list nat),
    increasing l ->
    In 0 l ->
    l = 0::tl l.
  intros l h1 h2.
  destruct l as [|a l]; simpl.
  contradiction.
  destruct h2 as [|h2]; subst; auto.
  pose proof (lt_increasing_cons _ _ a h2 h1).
  omega.
Qed.


  
Lemma linds_occ_nth_lt_eq_iff : 
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (n:nat)
         (pfn:n < length l)
         (ln:list nat),
    ln = linds_occ pfdt l (nth_lt _ _ pfn) <->
    (increasing ln /\
    (forall i:nat,
       (In i ln <->
        exists pfm:i < length l,
            nth_lt l i pfm =
            nth_lt l n pfn))).
  intros T h1 l n h2 ln; split; intro h3; subst.
  split.
  apply increasing_linds_occ.
  intros i.
  rewrite in_linds_occ_iff.
  split; intro h3.
  destruct h3 as [m [h5 [h3 h4]]].
  subst.
  exists (lt_lind_occ _ _ _ _ _).
  rewrite lind_occ_compat; auto.
  destruct h3 as [h3 h4].
  pose proof (ex_lind_occ_nth_lt h1 _ _ h3) as h5.
  destruct h5 as [occ [h6 h7]].
  exists occ, (in_nth_lt _ _ _).
  rewrite <- h4 at 1.
  split; auto.
  gen0. rewrite <- h4; auto.
  intro.
  rewrite <- h7.
  apply lind_occ_functional; auto.

  destruct h3 as [h3 h4].
  pose proof list_to_set_injable_increasing as h5.
  red in h5.
  apply h5; auto.
  apply increasing_linds_occ.

  apply Extensionality_Ensembles; red; split; intros i h6.
  rewrite <- list_to_set_in_iff in h6.
  rewrite h4 in h6.
  destruct h6 as [h6 h7].
  rewrite <- list_to_set_in_iff.
  rewrite in_linds_occ_iff.
  pose proof (ex_lind_occ_nth_lt h1 _ _ h6) as h8.
  destruct h8 as [occ [h8 h9]].
  exists occ, (in_nth_lt _ _ _).
  split.
  rewrite h7 in h8; auto.
  gen0.
  rewrite <- h7.
  intro h10.
  rewrite <- h9.
  apply lind_occ_functional; auto.
  rewrite <- list_to_set_in_iff.
  rewrite h4.
  rewrite <- list_to_set_in_iff in h6.
  rewrite in_linds_occ_iff in h6.
  destruct h6 as [m [h8 [h6 h7]]]; subst.
  exists (lt_lind_occ _ _ _ _ _).
  rewrite lind_occ_compat; auto.
Qed.


    
Lemma in_linds_iff_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (linds pfdt l) <->
    (increasing ln /\
     exists (n:nat) (pfn:n < length l),
       (forall i:nat,
          (In i ln <->
           exists pfm:i < length l,
             nth_lt l i pfm =
             nth_lt l n pfn))).
  intros T h1 l ln. 
  pose proof (linds_occ_nth_lt_eq_iff h1 l) as h0.
  split; intro h2. 
  rewrite in_linds_iff in h2.
  destruct h2 as [x h2].
  red in h2.
  destruct h2 as [[h2 h3] h4]; subst.
  specialize (h0 _ (lt_lind h1 _ _ h2) (linds_occ h1 l x)).
  hl h0.
  symmetry.
  f_equal.
  rewrite lind_compat; auto.
  rewrite h0  in hl.
  destruct hl as [h5 h6].
  split; auto.
  exists _, (lt_lind h1 _ _ h2); auto.
  destruct h2 as [h2 h3].
  destruct h3 as [i [h4 h5]].
  pose proof (iff2 (h0 i h4 ln) (conj h2 h5)); subst.
  apply in_linds_occ_linds.
  apply in_nth_lt.
Qed.


Lemma incr_lec :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (m:nat) (pfm:m < length l),
    increasing (lec pfdt pfm).
  unfold lec; intros; apply increasing_linds_occ.
Qed.


Lemma incr_lec_S :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T)
         (m:nat) (pfm:m < length l),
    increasing (lec_S pfdt pfm).
  unfold lec_S; intros; apply incr_lS; apply increasing_linds_occ.
Qed.

Lemma incr_lec_cons :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (m:nat) (pfm:m < length l), 
    increasing (lec_cons pfdt pfm a).
  unfold lec_cons; intros; apply incr_lec.
Qed.

  
Lemma lec_eq_iff :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T)
         (m:nat) (pfm:m < length l)
         (ln:list nat),
    lec pfdt pfm = ln <->
    (increasing ln /\
     forall (r:nat),
       In r ln <->
       exists  (pfr:r < length l),
         nth_lt l r pfr =
         nth_lt l m pfm).
  intros T h1 l m h2 ln.
  pose proof (in_linds_iff_iff h1 l ln) as h3.
  split; intro h4; subst.
  pose proof (in_lec h1 l _ h2) as h4.
  rewrite h3 in h4.
  destruct h4 as [h4 h5].
  split; auto.
  destruct h5 as [i [h5 h6]].
  intros r; split; intro h7.
  pose proof h7 as h7'.
  rewrite h6 in h7'.
  destruct h7' as [ha hb].
  exists ha; auto.
  specialize (h6 m).
  pose proof (rep_in_lec h1 l _ h2) as h10.
  rewrite h6 in h10.
  destruct h10 as [h10 h11].
  pose proof (proof_irrelevance _ h10 h2); subst; auto.
  congruence.

  destruct h7 as [h7 h8].
  rewrite in_lec_iff'.
  exists h7; auto.

  destruct h4 as [h4 h5].
  apply list_to_set_injable_increasing.
  apply incr_lec.
  assumption.
    
  apply Extensionality_Ensembles; red; split; red; intros x h6.
  rewrite <- list_to_set_in_iff in h6.
  rewrite <- list_to_set_in_iff.
  pose proof (in_lec_lt h1 l m x h2 h6) as h7.  
  rewrite (in_lec_iff _ _ _ _ _ h7) in h6.
  specialize (h5 x).
  hr h5.
  exists h7; auto.
  rewrite <- h5 in hr; auto.
  
  rewrite <- list_to_set_in_iff in h6.
  rewrite <- list_to_set_in_iff.
  rewrite in_lec_iff'.
  rewrite h5 in h6.
  destruct h6 as [h6 h7].
  exists h6; auto.
Qed.



Lemma lec_Sn_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T) (m m':nat)
         (pfm:m < length l)
         (pfsm:S m < length (a::l)),
    a <> nth_lt l m pfm ->
    lec pfdt pfsm = lS (lec pfdt pfm).
  intros T h1 l a m m' h2 h3 h4.
  rewrite lec_eq_iff.
  split.
  apply incr_lS.
  apply incr_lec.
  intros r.
  split; intro h5.
  rewrite in_lS_iff in h5.
  destruct h5 as [r' [h5 h6]]; subst.
  simpl.
  rewrite in_lec_iff' in h5.
  destruct h5 as [h5 h6].
  exists (lt_n_S _ _ h5).
  rewrite nth_lt_compat in h6.
  erewrite nth_indep in h6.
  rewrite h6 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  assumption.
  destruct h5 as [h5 h6].
  rewrite in_lS_iff.
  destruct r as [|r].
  simpl in h6.
  contradict h4.
  rewrite nth_lt_compat.
  erewrite nth_indep.
  apply h6.
  assumption.
  simpl in h6.
  exists r.
  split; auto.
  rewrite in_lec_iff'.
  simpl in h5.
  exists (lt_S_n _ _ h5).
  erewrite nth_lt_compat.
  erewrite nth_indep.
  rewrite <- h6 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  apply lt_S_n in h5; auto.
  assumption.
Qed.


Lemma nin_lec_S_cons_eq :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T) (n:nat)
         (pf:S n < length (a::l)),
    ~In a l ->
    lec_S pfdt (lt_S_n _ _ pf) =
    lec pfdt pf.
  intros T h1 l a n h3 h4.
  symmetry.
  rewrite lec_eq_iff.
  split. 
  apply incr_lec_S.  
  intro r.
  destruct r as [|r].
  split; intro h5.
  rewrite in_lec_S_iff in h5.
  destruct h5; omega.

  destruct h5 as [h5 h6].
  simpl in h6.
  contradict h4.
  rewrite h6.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  apply in_nth_lt.
  simpl in h3.
  apply lt_S_n in h3; auto.
  
  unfold lec_S.
  rewrite in_lS_iff. 
  split; intro h5.
  destruct h5 as [m h5].
  rewrite in_lec_iff' in h5.
  destruct h5 as [[h5 h6] h7].
  apply S_inj in h7; subst.
  exists (lt_n_S _ _ h5).
  simpl.
  rewrite nth_lt_compat in h6.
  erewrite nth_indep in h6.
  rewrite h6 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  let P := type of h6 in match P with nth _ _ (nth_lt _ _ ?pf) = _ => apply pf end.
  destruct h5 as [h5 h6].
  exists r.
  rewrite in_lec_iff.
  split; auto.
  simpl in h6.
  erewrite nth_lt_compat.
  erewrite nth_indep.
  rewrite <- h6 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  simpl in h5; apply lt_S_n in h5; auto.
  simpl in h3; apply lt_S_n in h3; auto.

  Grab Existential Variables.
  simpl in h5; pose proof (lt_S_n _ _ h5); auto.
  simpl in h3; apply lt_S_n in h3; auto.
Qed.

Lemma lec_cons_neq_eq :
  forall {T:Type} (pfdt:type_eq_dec T) {l:list T}
         (m:nat) (pfm:m < length l)
         (a:T),
    a <> nth_lt l m pfm ->
    lec_cons pfdt pfm a =
    lec_S pfdt pfm.
  intros T h1 l m h2 a h3.
  symmetry.
  apply list_to_set_injable_increasing.
  apply incr_lec_S.
  apply incr_lec_cons.
  apply Extensionality_Ensembles; red; split; red; intros x h4; rewrite <- list_to_set_in_iff; rewrite <- list_to_set_in_iff in h4.
  rewrite in_lec_S_iff in h4.
  destruct h4 as [h4 [h5 h6]].
  unfold lec_cons.
  rewrite in_lec_iff.
  simpl.
  destruct x as [|x].
  contradict (lt_irrefl _ h4).
  simpl in h6.
  erewrite nth_lt_compat, nth_indep in h6.
  rewrite h6 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  simpl in h5; auto. 
  assumption. 
  unfold lec_cons in h4.
  pose proof (in_lec_lt _ _ _ _ _ h4) as hlt.
  rewrite in_lec_iff in h4.
  simpl in h4.
  rewrite in_lec_S_iff.
  destruct x as [|x].
  contradict h3.
  erewrite nth_lt_compat, nth_indep.
  rewrite h4 at 1; auto.
  assumption.
  split; auto with arith.
  simpl.
  simpl in hlt.
  apply lt_S_n in hlt.
  exists hlt.
  erewrite nth_lt_compat, nth_indep.
  rewrite h4 at 1.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  assumption.

  Grab Existential Variables.
  assumption.
  simpl.
  omega.
Qed.


End ListNat'.




Lemma plusl_map_remove : 
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (f:T->nat) (x:T), 
    NoDup l -> In x l ->
    plusl (map f l) = (f x) + (plusl (map f (remove pfdt x l))).
intros T h0 l. induction l as [|a l h1]; simpl. intros; contradiction.
intros f x h2 h3.
destruct h3 as [h3l | h3r]. subst.
destruct (h0 x x) as [h3 | h4].
pose proof (no_dup_cons_nin _ _ h2) as h4.
rewrite <- remove_not_in'; auto. 
contradict h4. reflexivity.
pose proof (no_dup_cons _ _ h2) as h4.
specialize (h1 f x h4 h3r).
rewrite h1.
destruct (h0 x a) as [h5 | h6]. subst.
pose proof (no_dup_cons_nin _ _ h2). contradiction.
simpl.
rewrite plus_assoc.
rewrite (plus_comm (f a) (f x)).
rewrite plus_assoc.
reflexivity.
Qed.


Lemma plusl_nat_valued_fun_functional :
  forall {T:Type} (pfdt:type_eq_dec T) (l l':list T), 
    NoDup l -> NoDup l' ->
    list_to_set l = list_to_set l' ->
    forall f:T->nat,
      plusl (map f l) = plusl (map f l').
intros T h0 l. 
induction l as [|a l h1];simpl; auto.
intros l' h1 h2 h3. symmetry in h3.
pose proof (empty_set_nil _ h3).
subst. simpl. reflexivity.
intros l' h2 h3 h4.
pose proof (subtract_remove_compat h0 l' a) as h5.
rewrite <- h4 in h5.
pose proof (no_dup_cons_nin _ _ h2) as h6.
rewrite list_to_set_in_iff in h6.
rewrite sub_add_same_nin in h5; auto.
pose proof (no_dup_remove h0 _ a h3) as h7.
pose proof (no_dup_cons _ _ h2) as h8.
specialize (h1 _ h8 h7 h5).
intro f. rewrite h1.
rewrite (plusl_map_remove h0 _ _ a h3).
reflexivity.
pose proof (Add_intro2 _ (list_to_set l) a) as h9.
rewrite h4 in h9.
rewrite list_to_set_in_iff.
assumption.
Qed.

Lemma plusl_map_remove_no_dup :
  forall {T:Type} (pfdt:type_eq_dec T)
         (f:T->nat) (l:list T) (x:T)
         (pf:In x l),
    NoDup l ->
    plusl (map f l) =
    f x + plusl (map f (remove pfdt x l)).
  intros T h1 f l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros x h2 h3.
  destruct (h1 x a) as [|h4]; subst.
  f_equal.
  rewrite <- remove_not_in'; auto.
  apply no_dup_cons_nin in h3; auto.
  simpl.
  rewrite plus_assoc.
  rewrite (plus_comm (f x) (f a)).
  rewrite <- plus_assoc.
  f_equal.
  apply ih.
  destruct h2 as [|h2]; subst.
  contradict h4; auto.
  assumption.
  apply no_dup_cons in h3; auto.
Qed.


Lemma length_eq_plusl :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    length l =
    plusl (map (count_occ pfdt l) (list_singularize pfdt l)).
  intros T hdt l.
  pose proof (partition_linds hdt l) as h2. 
  pose proof (partition_faml_length nat_eq_dec _ _ (no_dup_seg_list (length l))
                                    (no_dup_linds hdt l)
                                    (no_dup_mem_linds _ _)
                                    (in_linds_list_to_set_injable hdt l) h2) as h3.
  rewrite length_seg_list in h3.
  rewrite <- h3.
  f_equal.
  unfold linds.
  rewrite map_map.
  apply map_ext_in.
  intros.
  rewrite length_linds_occ; auto.
Qed.


Lemma plusl_sort_list :
  forall (ln:list nat),
    plusl (sort_list _ tso_dec_lt_nat ln) =
    plusl ln.
  intro ln.
  induction ln as [|n ln ih]; simpl; auto.
  pose proof (sort_list_cons' _ sort_dec_lt_nat ln n) as h1.
  simpl in h1.
  pose proof (sort_dec_tso_dec_eq _ sort_dec_lt_nat) as h2.
  rewrite <- h2 at 1.
  rewrite h1.
  unfold weak_incrR_add.
  rewrite plusl_app; simpl.
  rewrite plus_assoc.
  rewrite (plus_comm _ n).
  rewrite <- plus_assoc.
  f_equal.
  rewrite <- ih.
  rewrite <- plusl_app.
  f_equal.
  symmetry.
  erewrite <- sort_list_decompose_ub_wlb.
  f_equal; auto.
Qed.

Lemma count_occs_eq_inv_iml :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T)
         (m:nat),
    (forall x:T, count_occ pfdt l x = count_occ pfdt l' x) ->
    list_to_set (inv_iml nat_eq_dec (count_occ pfdt l) (list_singularize pfdt l) m) =
    list_to_set (inv_iml nat_eq_dec (count_occ pfdt l') (list_singularize pfdt l') m).
  intros T h1 l l' m h2.
  pose proof (count_occs_eq_list_to_set _ _ _ h2) as h3.
  apply Extensionality_Ensembles; red; split; red; intros x h5;
  rewrite <- list_to_set_in_iff in h5;
  rewrite <- list_to_set_in_iff;
  rewrite in_inv_iml_iff in h5;
  rewrite in_inv_iml_iff;
  destruct h5 as [h5 h6]; subst.
  rewrite <- list_singularize_in_iff in h5;
    split;
    try rewrite <- list_singularize_in_iff;
    rewrite list_to_set_in_iff in h5;
    try rewrite list_to_set_in_iff.
  rewrite h3 in h5; auto.
  symmetry.
  apply h2; auto.
  rewrite <- list_singularize_in_iff in h5.
  split.
  rewrite <- list_singularize_in_iff.
  rewrite list_to_set_in_iff in h5.
  rewrite list_to_set_in_iff.
  rewrite <- h3 in h5; auto.
  apply h2; auto.
Qed.


Lemma count_occs_eq_sort_list_map :
  forall {T:Type}
         (pfdt:type_eq_dec T)
         (l:list T)
         (l':list T),
    (forall x:T, count_occ pfdt l x = count_occ pfdt l' x) ->
    sort_list lt tso_dec_lt_nat
              (map (count_occ pfdt l') (list_singularize pfdt l')) =
    sort_list lt tso_dec_lt_nat
              (map (count_occ pfdt l) (list_singularize pfdt l)).
  intros T hdt l l' h2. 
  pose proof (count_occs_eq_impl _ _ _ h2) as h3.
  pose proof (count_occs_eq_list_to_set _ _ _ h2) as h4.
  eapply weak_incrR_length_count_occ_eq.
  apply sort_dec_lt_nat.
  erewrite <- sort_dec_tso_dec_eq at 1.
  apply weak_incrR_sort_list.
  erewrite <- sort_dec_tso_dec_eq at 1.
  apply weak_incrR_sort_list.
  rewrite length_sort_list.
  rewrite length_sort_list.
  do 2 rewrite map_length; auto. 
  intro m.
  erewrite <- sort_dec_tso_dec_eq at 1.
  erewrite count_occ_sort_list.
  symmetry.
  erewrite <- sort_dec_tso_dec_eq at 1. 
  erewrite count_occ_sort_list.
  pose proof (type_eq_dec_eq nat_eq_dec) as ht.
  rewrite <- ht at 1.
  symmetry.
  rewrite <- ht at 1.
  rewrite count_occ_map; auto.
  rewrite count_occ_map; auto.
  apply no_dup_list_to_set_eq_length; auto.
  eapply no_dup_coarse_list.
  apply coarse_list_inv_iml.
  apply no_dup_list_singularize.
  eapply no_dup_coarse_list.
  apply coarse_list_inv_iml.
  apply no_dup_list_singularize.
  apply count_occs_eq_inv_iml; auto.

  Grab Existential Variables.
  apply sort_dec_lt_nat.
  apply sort_dec_lt_nat.
  apply sort_dec_lt_nat.
Qed.


Lemma count_occs_eq_plusl_map :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    (forall x:T, count_occ pfdt l x = count_occ pfdt l' x) ->
    plusl
      (map (count_occ pfdt l) (list_singularize pfdt l)) = 
    plusl
      (map (count_occ pfdt l') (list_singularize pfdt l')).
  intros T hdt l l' h1.
  rewrite <- plusl_sort_list.
  symmetry.
  rewrite <- plusl_sort_list.
  rewrite (count_occs_eq_sort_list_map hdt l l'); auto.
Qed.


Lemma count_occs_eq_length :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l l':list T),
    (forall x:T, count_occ pfdt l x = count_occ pfdt l' x) ->
    length l = length l'.
  intros T hdt l l' h2. 
  rewrite (length_eq_plusl hdt).
  rewrite (length_eq_plusl hdt).
  apply count_occs_eq_plusl_map; auto.
Qed.



Lemma sort_lists_eq_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l l':list T),
    let pfdt := sort_dec_eq_dec _ pfso in
    sort_list R (sort_dec_tso_dec _ pfso) l =
    sort_list R (sort_dec_tso_dec _ pfso) l' <->
    (forall x,
       count_occ pfdt l  x =
       count_occ pfdt l' x).
  intros T R hso l l' hdt; split; intro h1.
  intro x.
  pose proof (count_occ_sort_list R hso l x) as h2.
  pose proof (count_occ_sort_list R hso l' x) as h3.
  simpl in h2, h3.
  unfold hdt.
  rewrite <- h2, <- h3.
  rewrite h1; auto.
  pose proof h1 as hc.
  apply count_occs_eq_list_to_set in h1; auto.
  rewrite (list_to_set_sort_list R hso) in h1.
  symmetry in h1.
  rewrite (list_to_set_sort_list R hso) in h1.
 
  eapply weak_incrR_length_count_occ_eq.
  apply hso.
  apply weak_incrR_sort_list.
  apply weak_incrR_sort_list.
  do 2 rewrite length_sort_list. 
  eapply list_to_sets_eq_length in h1.
  apply (count_occs_eq_length hdt); auto.
  intro x.
  do 2 rewrite count_occ_sort_list.
  apply hc.

  Grab Existential Variables.
  assumption.
Qed.

Lemma  list_to_sets_eq_iff :
  forall {T:Type} (R:Relation T) (pfso:sort_dec R) (l l':list T),
    NoDup l ->
    NoDup l' ->
    (list_to_set l =
     list_to_set l' <->
     sort_list R (sort_dec_tso_dec _ pfso) l =
     sort_list R (sort_dec_tso_dec _ pfso) l').
  intros T R h1 l l' h2 h4; split; intro h5.
  rewrite sort_lists_eq_iff.
  intro x.
  destruct (in_dec (sort_dec_eq_dec _ h1) x l) as [h6 | h6],
                              (in_dec (sort_dec_eq_dec _ h1) x l') as [h7 | h7].
  
  rewrite count_occ_no_dup; auto.
  rewrite count_occ_no_dup; auto.
  rewrite list_to_set_in_iff in h6.
  rewrite h5 in h6.
  rewrite <- list_to_set_in_iff in h6.
  contradiction.
  rewrite list_to_set_in_iff in h7.
  rewrite <- h5 in h7.
  rewrite <- list_to_set_in_iff in h7.
  contradiction.
  rewrite (count_occ_In (sort_dec_eq_dec _ h1)) in h6.
  rewrite (count_occ_In (sort_dec_eq_dec _ h1)) in h7.
  omega.

  apply Extensionality_Ensembles; red; split; red; intros x h6; rewrite <- list_to_set_in_iff in h6; rewrite <- list_to_set_in_iff;
  rewrite <- in_sort_list_iff in h6;
  rewrite <- in_sort_list_iff.
  rewrite h5 in h6 at 1.
  apply h6.
  rewrite <- h5 in h6 at 1.
  apply h6.
Qed.
    

Section FamlNat.

  
  Lemma faml_nat_eq_dec : type_eq_dec (faml nat).
    apply list_eq_dec.
    apply list_nat_eq_dec.
  Qed.

  Definition minf (l:faml nat) (pf:l <> nil) : list nat :=
    minlR lt_list_nat lt_list_nat_tso_dec l pf.

  
Lemma in_minf :
  forall (l:faml nat)
         (pf:l <> nil),
    In (minf l pf) l.
  unfold minf; apply in_minlR.
Qed.

Lemma minf_compat :
  forall (ll:faml nat) (l:list nat)
         (pf:In l ll),
    minf ll (in_not_nil _ _ pf) /= l.
  unfold minf; intros; rewrite <- Rw_lt_list_nat_iff.
  apply minlR_compat.
  red; apply trans_lt_list_nat.
Qed.


Lemma minf_sing :
  forall (ll:faml nat) (l:list nat)
         (pfe:Singleton l = list_to_set ll),
         let pf := inh_sing l in
         let pfn := P_subst pf pfe in
         let pfn' := inh_not_nil _ pfn in
    l = minf ll pfn'.
  unfold minf; intros.
  erewrite minlR_functional'. apply minlR_sing.
  apply lt_list_nat_sort_dec.
  reflexivity.
  reflexivity.
  Grab Existential Variables.
  assumption.
  apply lt_list_nat_sort_dec.
Qed.
  

Lemma minf_in :
  forall (ll:faml nat) (l:list nat)
         (pf:In l ll),
    minf ll (in_not_nil _ _ pf) =
    minf (l::ll) (cons_neq_nil _ _). 
  unfold minf, lt_list_nat.
  intros ll l h1.
  pose proof (minlR_in lt sort_dec_lt_nat ll l h1) as h2.
  assert (h3:  (fun l0 m : list nat =>
      lt_listR lt (sort_dec_tso_dec lt sort_dec_lt_nat) l0 m) = 
               (lt_listR lt (sort_dec_tso_dec lt sort_dec_lt_nat))).
    apply functional_extensionality; intro al.
    apply functional_extensionality; intro bl.
    reflexivity.

  pose proof (minlR_functional'
                  (lt_listR lt (sort_dec_tso_dec lt sort_dec_lt_nat))
                  (fun l0 m : list nat =>
                     lt_listR lt (sort_dec_tso_dec lt sort_dec_lt_nat) l0 m) (lt_listR_sort_dec _ _)) as h4.
    hfirst h4.
    rewrite h3.
    apply lt_listR_sort_dec.
  specialize (h4 hf  (tso_dec_lt_listR _ _ ) ll ll (in_not_nil ll l h1) (in_not_nil ll l h1) h3 eq_refl).
  pose proof (sort_dec_tso_dec_eq (fun l0 m : list nat =>
                                     lt_listR lt (sort_dec_tso_dec lt sort_dec_lt_nat) l0 m) hf lt_list_nat_tso_dec) as h8.
  gen0.
  rewrite <- h8.
  intro hn.
  pose proof (proof_irrelevance _ hn
                                (in_not_nil ll l h1)); 
    subst.
  rewrite <- h4.
  rewrite h2.
  eapply minlR_functional'.  
  apply lt_listR_sort_dec.
  reflexivity.
  reflexivity.
Qed.

Definition le_faml_nat (ll lm:faml nat) : Prop :=
  le_listR lt_list_nat (sort_dec_tso_dec _ lt_list_nat_sort_dec) ll lm.

Definition lt_faml_nat (l m:faml nat) : Prop :=
  le_faml_nat l m /\ l <> m.

Notation "l //= m" := (le_faml_nat l m) (at level 20) : list_scope.
Notation "l //_ m" := (lt_faml_nat l m) (at level 20) : list_scope.

Lemma lt_faml_nat_impl_le :
  forall (l m:faml nat),
    l //_ m -> l //= m.
  unfold lt_faml_nat, le_faml_nat; apply lt_list_impl_leR.
Qed.

Lemma le_faml_nat_nil :
  forall (l:faml nat),
    nil //= l.
  intro; unfold le_list_nat; apply le_listR_nil.
Qed.


Lemma le_faml_nat_dec :
  forall (l m:faml nat), {l //= m} + {~l //= m}.
  unfold le_faml_nat; intros; apply le_listR_dec.
Qed.

Lemma refl_le_faml_nat :
  forall (l:faml nat),
    l //= l.
  unfold le_faml_nat; apply refl_le_listR.
  apply lt_list_nat_irrefl.
Qed.


Lemma antisym_le_faml_nat :
  forall (l m:faml nat),
    l //= m -> m //= l -> l = m.
  unfold le_faml_nat; apply antisym_le_listR. 
Qed.

Lemma antisym_not_le_faml_nat :
  forall (l m:faml nat),
    ~ l //= m -> ~ m //= l -> l = m.
  unfold le_faml_nat; apply antisym_not_le_listR.
Qed.

Lemma lt_faml_nat_iff :
  forall (l m:faml nat),
    l //_ m <-> ~ m //= l.
  unfold lt_faml_nat; intros; apply lt_listR_iff.
Qed.


Lemma lt_faml_nat_cons :
  forall (l m:faml nat) (al:list nat),
    (al :: l) //_ (al :: m)->
    l //_ m.
  unfold lt_faml_nat; apply lt_listR_cons.
Qed.
  
Lemma le_faml_nat_dec' :
  forall (l m:faml nat),
    {l //= m} + {m //_ l}.
  unfold le_faml_nat; apply le_listR_dec'.
Qed.

Lemma tso_lt_faml_nat :
  total_strict_order lt_faml_nat.
  unfold lt_faml_nat; apply tso_dec_total_strict_order; apply tso_dec_lt_listR.
Qed.
Lemma trans_le_faml_nat :
  Transitive _ le_faml_nat.
  unfold le_faml_nat; apply trans_le_listR.
Qed.

Lemma trans_le_lt_faml_nat :
  forall (la lb lc:faml nat),
    la //= lb -> lb //_ lc -> la //_ lc.
  unfold le_faml_nat, lt_faml_nat; eapply trans_le_lt_listR.
Qed.

Lemma trans_lt_le_faml_nat :
  forall (la lb lc:faml nat),
    la //_ lb -> lb //= lc -> la //_ lc.
  unfold lt_faml_nat, le_faml_nat; apply trans_lt_le_listR.
Qed.

Lemma trans_lt_faml_nat :
  Transitive _ lt_faml_nat.
  unfold lt_faml_nat; apply trans_lt_listR.
Qed.

Lemma irrefl_lt_faml_nat :
  forall (l:faml nat),
    ~ l //_ l.
  unfold lt_listR; apply irrefl_lt_listR.
Qed.

Lemma so_lt_faml_nat :
  Strict_Order lt_faml_nat.
  unfold lt_faml_nat; apply so_lt_listR.
Qed.

Lemma lt_faml_nat_sortable :
  sortable lt_faml_nat.
  unfold lt_faml_nat; apply lt_listR_sortable.
Qed.
 

Lemma count_occ_famlS_nin0 :
  forall (ll:faml nat) (l:list nat),
    ~In 0 l ->
    count_occ list_nat_eq_dec (famlS ll) l =
    count_occ list_nat_eq_dec ll (lpred l).
  intro ll.
  induction ll as [|al ll ih]; simpl; auto.
  intros l h0.
  destruct (list_nat_eq_dec (lS al) l) as [h1 | h1].
  subst. 
  destruct (list_nat_eq_dec al (lpred (lS al))) as [h2 | h2].
  f_equal; auto.
  
  rewrite lpred_lS in h2; contradict h2; auto.
  destruct (list_nat_eq_dec al (lpred l)) as [h2 | h2].
  rewrite h2 in h1.
  apply neq_sym in h1. rewrite <- lS_lpred_iff in h1.
  contradict h1; auto.
  auto.
Qed.

Lemma count_occ_famlS_in0 :
  forall (ll:faml nat) (l:list nat),
    In 0 l ->
    count_occ list_nat_eq_dec (famlS ll) l =
    0.
  intros ll l h1.
  assert (h2:~In l (famlS ll)).
    intro h2.
    rewrite in_famlS_iff in h2.
    destruct h2 as [ln [h2 h3]]; subst.
    rewrite in_lS_iff in h1.
    destruct h1 as [? [? ?]]; discriminate.
  rewrite (count_occ_In list_nat_eq_dec) in h2; omega.
Qed.

Definition sort_family (F:Family nat) (pff:Finite F)
           (pffm:finite_mem F) : faml nat :=
  match (finite_inh_or_empty_dec F pff) with
    | left pfi => 
      sort_set lt_list_nat lt_list_nat_sort_dec (fam_nat_to_ens_sort_set F pffm) (finite_fam_nat_to_ens_sort_set _ _ pff)
    | right _ => nil
  end.

   
Lemma sort_family_compat :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F),
    list_to_set (sort_family F pf pffm) = fam_nat_to_ens_sort_set F pffm.
  intros F h1 h2. unfold sort_family.
  apply Extensionality_Ensembles; red; split; red; intros x h3.
  rewrite <- list_to_set_in_iff in h3.
  lr_if h3; fold hlr in h3; destruct hlr as [hl | hr].
  rewrite <- in_sort_set_iff in h3; auto.
  contradict h3.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  rewrite <- list_to_set_in_iff.
  rewrite <- in_sort_set_iff; auto.
  subst.
  rewrite fam_nat_to_ens_sort_set_empty_eq in h3.
  contradict h3.
Qed.
  


Definition is_sorted_family (ll:faml nat) : Prop :=
  ll = sort_family (faml_to_fam ll) (finite_faml_to_fam _) (finite_mem_faml _).

Lemma sort_family_functional :
  forall (F F':Family nat) (pff:Finite F) (pff':Finite F')
         (pfm:finite_mem F) (pfm':finite_mem F'),
        F = F' ->
    sort_family F  pff  pfm  =
    sort_family F' pff' pfm'.
  intros F F' h1 h2 h3 h4 h5.
  subst.
  pose proof (proof_irrelevance _ h1 h2);
    pose proof (proof_irrelevance _ h3 h4);
  subst; auto.
Qed.

         
Lemma no_dup_sort_family :
  forall (F:Family nat) (pff:Finite F) (pffm:finite_mem F),
    NoDup (sort_family F pff pffm).
  intros E h1 h2.
  unfold sort_family.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  apply no_dup_sort_set.
  constructor.
Qed.
 
Lemma in_list_reps_sort_family :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F),
         Inn (list_reps (fam_nat_to_ens_sort_set F pffm)) (sort_family F pf pffm ).
  intros F h0 ha.
  unfold sort_family.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  constructor.
  split.
  rewrite sort_set_compat; auto.
  apply no_dup_sort_set.
  subst.
  rewrite fam_nat_to_ens_sort_set_empty_eq; auto.
  rewrite list_reps_empty.
  constructor.
Qed.



Lemma impl_in_sort_family :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F)
         (A:Ensemble nat) (pfa:Finite A),
    Inn F A ->
         In (sort_set_nat A pfa)
            (sort_family F pf pffm).
  intros F h1 h2 A h3 h4.
  unfold sort_family.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  rewrite <- in_sort_set_iff.
  erewrite sort_set_nat_functional.
  apply in_fam_nat_impl.
  reflexivity.
  subst.
  destruct h4.
  Grab Existential Variables.
  assumption.
Qed.

(*This is kind of weak and loses information, use the next.*)
Lemma in_sort_family_impl :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F)
         (l:list nat),
    In l (sort_family F pf pffm) ->
    Inn F (list_to_set l).
  intros F h1 h2 l h5.
  unfold sort_family in h5.
  lr_if h5; fold hlr in h5; destruct hlr as [hl | hr].
  rewrite <- in_sort_set_iff in h5.
  apply in_fam_nat_to_ens_sort_set_impl in h5; auto.
  contradict h5.
Qed.

(*This is stronger than the original.*)
Lemma in_sort_family_impl' :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F)
         (l:list nat),
    In l (sort_family F pf pffm) ->
    exists (S:Ensemble nat) (pfi:Inn F S),
      l = sort_set_nat S (pffm _ pfi).
  intros F h1 h2 l h4.
  unfold sort_family in h4. 
  lr_if h4; fold hlr in h4; destruct hlr as [hl | hr].
  rewrite <- in_sort_set_iff in h4.
  destruct h4 as [l h4]; subst.
  clear h4.
  destruct l as [l h3].
  exists _, h3.
  unfold sort_set_nat.
  apply sort_set_functional; simpl; auto.
  contradict h4.
Qed.

Lemma in_sort_family_unq :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F)
         (l l':list nat),
    list_to_set l = list_to_set l' ->
    In l  (sort_family F pf pffm) ->
    In l' (sort_family F pf pffm) ->
    l = l'.
  intros F h1 h2 l l' h4 h5 h6.
  apply in_sort_family_impl' in h5.
  apply in_sort_family_impl' in h6.
  destruct h5 as [S [h5 h5']], h6 as [S' [h6 h6']].
  subst.
  do 2 rewrite sort_set_nat_compat in h4.
  subst.
  apply sort_set_nat_functional; auto.
Qed.


Lemma sort_family_compat' :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F),
    faml_to_fam (sort_family F pf pffm) = F.
  intros F h1 h2.
  rewrite faml_to_fam_eq.
  rewrite map_im_compat.
  rewrite sort_family_compat.
  apply im_fam_nat_to_ens_sort_set_list_to_set.
Qed.


Lemma sort_family_sing :
  forall (A:Ensemble nat) (pff:Finite A) (pff':Finite (Singleton A))
         (pffm:finite_mem (Singleton A)),
         sort_family (Singleton A) pff' pffm  =
         (sort_set_nat A pff)::nil.
  intros ? h1. intros.
  unfold sort_family.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  simpl.
  gen0.
  erewrite fam_nat_to_ens_sort_set_sing_eq.
  intro.
  erewrite sort_set_sing.
  reflexivity.
  apply sing_not_empty in hr.
  contradiction.
Qed.

  
Lemma incl_sort_family_incl :
  forall (F F':Family nat)
         (pff:Finite F) (pff':Finite F')
         (pffm:finite_mem F) (pffm': finite_mem F'),
    Included F F' ->
    incl (sort_family F  pff  pffm)
          (sort_family F'  pff' pffm').
  intros F F' h1 h1' h2 h2' h4.
  rewrite incl_iff.
  do 2 rewrite sort_family_compat.
  red; intros x h5.
  destruct h5 as [x h5]. subst. clear h5.
  destruct x as [x h5]. simpl.
  pose proof (h4 _ h5) as h6.
  apply Im_intro with (exist _ _ h6).
  constructor; simpl.
  apply sort_set_functional; auto.
Qed.


Lemma in_is_sorted_family_ :
  forall (ll:faml nat) (l:list nat)
         (pf:In l ll),
    is_sorted_family ll  -> is_sorted_list_nat l.
  intros ll l h1 h2.
  red in h2. red.
  simpl in h2.
  destruct ll as [|al ll].
  red in h1. contradiction.
  pose proof h1 as h1'.
  rewrite list_to_set_in_iff in h1'.
  rewrite h2 in h1'.
  rewrite sort_family_compat in h1'.
  destruct h1' as [l h1']; subst.
  apply sort_set_nat_functional.
  rewrite sort_set_nat_compat.
  reflexivity.
Qed.


Lemma incr_impl_in_sort_family :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F)
         (l:list nat),
    increasing l ->
    Inn F (list_to_set l) ->
    In l (sort_family F pf pffm).
  intros F h1 h2 l h3 h4.
  pose proof (sort_family_compat F h1 h2) as h5.
  simpl in h5.
  destruct h5 as [h5 h6].
  pose proof (finite_inh_or_empty _ h1) as h7.
  pose proof (incrR_impl_sort_set _ sort_dec_lt_nat l h3) as h10.
  destruct h7 as [h7 | h7].
  rewrite h10.
  apply impl_in_sort_family.
  assumption.
  clear h10.
  subst; contradiction.
Qed.
 
Lemma inj_sort_family :
  forall (A B:Family nat)
         (pffa:Finite A) (pffma:finite_mem A)
         (pffb:Finite B) (pffmb:finite_mem B),
    sort_family A pffa pffma =
    sort_family B pffb pffmb ->
    A = B.
  intros A B h1 h1' h2 h2' h3.
  apply Extensionality_Ensembles; red; split; red; intros C h4.
  pose proof (impl_in_sort_family _ h1 h1' _ (h1' _ h4) h4) as hi.
  rewrite h3 in hi at 1.
  apply in_sort_family_impl in hi.
  rewrite sort_set_nat_compat in hi; auto.
  pose proof (impl_in_sort_family _ h2 h2' _ (h2' _ h4) h4) as hi.
  rewrite <- h3 in hi.
  apply in_sort_family_impl in hi.
  rewrite sort_set_nat_compat in hi; auto.
Qed.

Lemma sort_family_eq_nil  :
  forall (F:Family nat) (pf:Finite F) (pffm:finite_mem F),
    sort_family F pf pffm = nil ->
    F = Empty_set _.
  intros F h1 h2 h3.
  do 2 red in h2.
  apply eq_empty. 
  red; intros A h4.
  pose proof (impl_in_sort_family _ h1 h2 _ (h2 _ h4) h4) as h5.
  rewrite h3 in h5.
  contradiction.
Qed.


Definition increasing_faml (ll:faml nat) : Prop :=
  increasingR lt_list_nat ll.

Definition weak_increasing_faml (ll:faml nat) : Prop :=
  weak_increasingR lt_list_nat ll.

Lemma increasing_cons_faml :
  forall (ll:faml nat) (l:list nat),
    increasing_faml (l::ll) ->
    increasing_faml ll.
  unfold increasing_faml; apply increasingR_cons.
Qed.
 
Lemma weak_increasing_cons_faml :
  forall (ll:faml nat) (l:list nat),
    weak_increasing_faml (l::ll) ->
    weak_increasing_faml ll.
  unfold weak_increasing_faml; apply weak_increasingR_cons.
Qed.

Lemma increasing_impl_weak_faml :
  forall (ll:faml nat),
    increasing_faml ll ->
    weak_increasing_faml ll.
  unfold weak_increasing_faml; apply increasingR_impl_weak.
Qed.

Lemma lt_increasing_faml_cons :
  forall (ll:faml nat) (l l':list nat),
    In l ll ->
    increasing_faml (l'::ll)  ->
    l' /_ l.
  unfold increasing_faml, lt_list_nat; intros; eapply lt_increasingR_cons.
  apply sort_dec_trans. apply lt_listR_sort_dec.
  apply H.
  assumption.
Qed.

Lemma le_weak_increasing_faml_cons :
  forall (ll:faml nat) (l l':list nat),
    In l ll ->
    weak_increasing_faml (l'::ll)  ->
    l' /= l.
  unfold weak_increasing_faml, le_list_nat; intros.
  erewrite <- Rw_lt_listR_iff. eapply le_weak_increasingR_cons.
  apply trans_lt_list_nat.
  apply H.
  assumption.
Qed.


Lemma increasing_faml_cons_impl :
  forall (ll:faml nat) (l:list nat),
    increasing_faml (l::ll) ->
    increasing_faml ll.
  unfold increasing_faml. apply increasingR_cons_impl.
Qed.

Lemma weak_increasing_faml_cons_impl :
  forall (ll:faml nat) (l:list nat),
    weak_increasing_faml (l::ll) ->
    weak_increasing_faml ll.
  unfold weak_increasing_faml; apply weak_increasingR_cons_impl.
Qed.

Lemma increasing_faml_cons_impl_cons :
  forall (ll:faml nat) (l l':list nat),
    increasing_faml (l::ll) ->
    l' /_ l ->
    increasing_faml (l'::l::ll).
  unfold increasing_faml, lt_list_nat; apply increasingR_cons_impl_cons.
Qed.

Lemma waek_increasing_faml_impl_aux_cons :
  forall (ll:faml nat) (l l':list nat),
    weak_increasing_faml (l::ll) ->
    l' /= l ->
    weak_increasing_faml (l'::l::ll).
  unfold weak_increasing_faml, lt_list_nat; intros.
  rewrite <- Rw_lt_list_nat_iff in H0.
  apply weak_increasingR_cons_impl_cons; auto.
Qed.

Lemma increasing_faml_impl_cons :
  forall (ll:faml nat)
    (pf:ll <> nil) (l:list nat),
    increasing_faml ll ->
    l /_ hd' ll pf ->
    increasing_faml (l::ll).
  unfold increasing_faml, lt_list_nat; apply increasingR_impl_cons.
Qed.

Lemma weak_increasing_faml_impl_cons :
  forall (ll:faml nat)
    (pf:ll <> nil) (l:list nat),
    weak_increasing_faml ll ->
    l /= hd' ll pf ->
    weak_increasing_faml (l::ll).
  unfold weak_increasing_faml; intros.
  rewrite <- Rw_lt_list_nat_iff in H0.
  eapply weak_increasingR_impl_cons; auto.
  apply H0.
Qed.


Lemma increasing_faml_impl_cons' :
  forall (ll:faml nat) (l:list nat),
    increasing_faml ll ->
    (forall l':list nat, In l' ll -> l /_ l') ->
    increasing_faml (l::ll).
  unfold increasing; apply increasingR_impl_cons'.
Qed.

Lemma weak_increasing_faml_impl_cons' :
  forall (ll:faml nat) (l:list nat),
    weak_increasing_faml ll ->
    (forall l':list nat, In l' ll -> l /= l') ->
    weak_increasing_faml (l::ll).
  unfold weak_increasing_faml, le_list_nat; intros.
  eapply weak_increasingR_impl_cons'; auto.
  intros.
  rewrite Rw_lt_list_nat_iff.
  apply H0; auto.
Qed.

Lemma increasing_faml_tl :
  forall (ll:faml nat),
    increasing_faml ll ->
    increasing_faml (tl ll).
  unfold increasing_faml; apply increasingR_tl.
Qed.

Lemma weak_increasing_faml_tl :
  forall (ll:faml nat),
    weak_increasing_faml ll ->
    weak_increasing_faml (tl ll).
  unfold weak_increasing_faml; apply weak_increasingR_tl.
Qed.


Lemma increasing_faml_remove :
  forall (ll:faml nat) (l:list nat),
    increasing_faml ll ->
    increasing_faml (remove list_nat_eq_dec l ll).
  unfold increasing_faml; intros; eapply increasingR_remove; auto.
  apply trans_lt_list_nat.
Qed.

Lemma weak_increasing_faml_remove :
  forall (ll:faml nat) (l:list nat),
    weak_increasing_faml ll ->
    weak_increasing_faml (remove list_nat_eq_dec l ll).
  unfold weak_increasing_faml; intros; eapply weak_increasingR_remove; auto.
  apply trans_lt_list_nat.
Qed.

Lemma incr_faml_list_minus :
  forall (ll ll':faml nat),
    increasing_faml ll ->
    let lm := list_minus list_nat_eq_dec ll ll' in
    increasing_faml lm.
  unfold increasing_faml; intros; eapply incrR_list_minus; auto.
  apply trans_lt_list_nat.
Qed.

Lemma weak_incr_faml_list_minus :
  forall (ll ll':faml nat),
    weak_increasing_faml ll ->
    let lm := list_minus list_nat_eq_dec ll ll' in
    weak_increasing_faml lm.
  unfold weak_increasing_faml; intros; eapply weak_incrR_list_minus; auto.
  apply trans_lt_list_nat.
Qed.

Lemma increasing_faml_cons_lt_subst :
  forall (ll:faml nat) (l l':list nat),
    increasing_faml (l::ll) ->
    l' /_ l ->
    increasing_faml (l'::ll).
  unfold increasing_faml, lt_list_nat; intros; eapply increasingR_cons_lt_subst.
  apply trans_lt_list_nat.
  apply H.
  assumption.
Qed.

Lemma weak_increasing_faml_cons_le_subst :
  forall (ll:faml nat) (l l':list nat),
    weak_increasing_faml (l::ll) ->
    l' /= l ->
    weak_increasing_faml (l'::ll).
  unfold weak_increasing_faml, le_list_nat; intros; eapply weak_increasingR_cons_le_subst.
  apply trans_lt_list_nat.
  apply H.
  rewrite Rw_lt_list_nat_iff.
  assumption.
Qed.
  
Lemma increasing_faml_cons_no_dup :
  forall (ll:faml nat) (l:list nat),
    increasing_faml (l::ll) ->
    NoDup ll.
  unfold increasing_faml; intros; eapply increasingR_cons_no_dup.
  apply list_nat_eq_dec.
  apply so_lt_list_nat.
  apply H.
Qed.

Lemma increasing_faml_cons_nin :
  forall (ll:faml nat) (l:list nat),
    increasing_faml (l::ll) ->
    ~In l ll.
  unfold increasing_faml; intros; eapply increasingR_cons_nin.
  apply list_nat_eq_dec.
  apply so_lt_list_nat.
  apply H.
Qed.

Lemma increasing_faml_no_dup :
  forall (ll:faml nat) ,
    increasing_faml ll ->
    NoDup ll.
  unfold increasing_faml; intros; eapply increasingR_no_dup.
  apply list_nat_eq_dec.
  apply so_lt_list_nat.
  apply H.
Qed.

Lemma increasing_faml_app :
  forall (ll ll':faml nat)
         (pfl:ll <> nil)
         (pfl':ll' <> nil),
    increasing_faml ll ->
    increasing_faml ll' ->
    lst ll pfl /_ hd' ll' pfl' ->
    increasing_faml (ll++ll').
  unfold increasing_faml, lt_list_nat. intros; rewrite <- increasingR_app_iff.
  split.
  apply H. split; auto. apply H1.
  apply trans_lt_list_nat.
Qed.

Lemma weak_increasing_faml_app :
  forall (ll ll':faml nat)
         (pfl:ll <> nil)
         (pfl':ll' <> nil),
    weak_increasing_faml ll ->
    weak_increasing_faml ll' ->
    lst ll pfl /= hd' ll' pfl' ->
    weak_increasing_faml (ll++ll').
  unfold weak_increasing_faml; intros; rewrite <- weak_increasingR_app_iff; auto.
  split; auto. split; auto. 
  rewrite Rw_lt_list_nat_iff.
  apply H1.
  apply trans_lt_list_nat.
Qed.


Lemma no_dup_min_increasing_faml :
  forall  (lm:faml nat)  (C:Family nat) (pfc:Finite C) (pfm:finite_mem C),
    fam_nat_to_ens_sort_set C pfm = list_to_set lm ->
    NoDup lm ->
    (forall x : list (list nat),
       Inn (Subtract (list_reps (fam_nat_to_ens_sort_set C pfm)) lm) x ->
       lm //_ x) ->
    increasing_faml lm.
  intro lm.
  induction lm as [|al ll ih]; simpl; auto.
  intros; constructor.
  revert ih.
  destruct ll as [|bl ll].
  intros. do 2 red; auto.
  intros h1 C h0 h2 h3 h4 h5.
  pose proof (h5 (bl::al::ll)) as h6.
  hfirst h6.
  constructor.
  constructor.
  simpl.
  rewrite add_comm.
  split; auto.
  apply no_dup_comm_cons; auto.
  rewrite in_sing_iff. 
  intro h7.
  inversion h7; subst.
  pose proof (no_dup_cons_nin _ _ h4) as h8.
  contradict h8. left; auto.
  pose proof (no_dup_cons _ _ h4) as h20.
  pose proof (no_dup_cons2 _ _ _ h4) as h20'.
  pose proof (no_dup_cons_nin _ _ h4) as h21.
  pose proof (no_dup_cons2_nin _ _ _ h4) as h21'.
  split.
  specialize (h6 hf).
  red in h6.
  destruct h6 as [h6 h7].
  simpl in h6.

  do 2 red in h6.
  destruct (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec al bl) as [h9 | h9].
  destruct h9 as [h9 | h9].
  split; auto.
  pose proof (f_equal (fun D => Subtract D al) h3) as h10.
  simpl in h10.
  rewrite sub_add_same_nin in h10.
  red in h9. 
  destruct h9; auto.
  intro h11.
  destruct h11 as [al h11 | al h11].
 
  contradict h21'.
  rewrite <- list_to_set_in_iff in h11.
  assumption.
  destruct h11.
  contradict h21.
  left; auto.
  intro. subst.
  contradict h21.
  left; auto.
  subst.
  contradict h21.
  left; auto.
  contradiction.
  pose proof (f_equal (fun C => Subtract C al) h3) as h22.
  simpl in h22. 
  assert (h10:al <> bl).
    intro; subst; contradict h21; left; auto. 
  pose proof (Add_intro2 _ (Add (list_to_set ll) bl) al)  as hc.
  simpl in h3. rewrite <- h3 in hc. 
  pose proof (in_fam_nat_to_ens_sort_set_impl _ _ _ hc) as hc'.
  rewrite sub_add_same_nin in h22. 
  pose proof (Add_intro2 _ (Add (list_to_set ll) bl) al) as h11.
  rewrite <- h3 in h11.
  pose proof (in_fam_nat_to_ens_sort_set_incr _ _ _ h11) as h12.
  rewrite fam_nat_to_ens_sort_set_subtract in h22; auto.
  specialize (h1 _ (subtract_preserves_finite _  _ h0) (finite_mem_subtract _ _ h2)  h22 h20).
  hfirst h1.
  intros x h13.  
  destruct h13 as [h13 h14]. rewrite in_sing_iff in h14.
  destruct h13 as [h13]. destruct h13 as [h13 h15].
  specialize (h5 (al::x)).
  hfirst h5.
  constructor.
  constructor.
  simpl.
  rewrite <- h13.
  rewrite <- fam_nat_to_ens_sort_set_subtract.
  split.
  rewrite add_sub_eq.
  rewrite in_add_eq; auto.
  apply list_nat_eq_dec.
  constructor; auto.
  intro h16.
  rewrite list_to_set_in_iff in h16.
  rewrite <- h13 in h16.
  inversion h16 as [d h17 q h18].
  destruct d as [d hd].
  pose proof hd as hd'. rewrite h18 in hd' at 1.
  simpl in hd'.
  destruct hd' as [h19 h23].
  rewrite in_sing_iff in h23; simpl in h23.
  rewrite sort_set_nat_compat in h23.
  contradict h23; auto.
  assumption.
  assumption.
  assumption.
  rewrite in_sing_iff.
  intro h17.
  inversion h17.
  contradiction.
  eapply lt_faml_nat_cons.
  apply h5; auto.
  specialize (h1 hf0); auto.
  intro h23.
  destruct h23 as [al h23 | al h23].
  contradict h21'.
  rewrite list_to_set_in_iff; auto.
  destruct h23.
  contradict h10; auto.
Qed.

 
Lemma incr_sort_family :
  forall (C:Family nat) (pff:Finite C) (pffm:finite_mem C),
    increasing_faml (sort_family C pff pffm).
  unfold increasing_faml, sort_family; intros.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  apply incrR_sort_set.
  red; auto.
Qed.


Definition increasing_mem (ll:faml nat) : Prop :=
  p_mem increasing (list_to_set ll).

Lemma increasing_mem_sort_family :
  forall (A:Family nat) (pff:Finite A) (pffm:finite_mem A),
    increasing_mem (sort_family A pff pffm).
  intros A h1 h2.
  do 2 red.
  intros x h3.
  rewrite <- list_to_set_in_iff in h3.
  pose proof (in_sort_family_impl' _ h1 h2 _ h3) as h4.
  destruct h4 as [S [h4 h5]].
  subst.
  apply incr_sort_set_nat.
Qed.


Lemma increasing_mem_list_minus :
  forall (ll ll':faml nat),
    increasing_mem ll ->
    increasing_mem (list_minus list_nat_eq_dec ll ll').
  intros ll ll' h1.
  do 2 red in h1; do 2 red.
  intros x h2; apply h1.
  rewrite list_to_set_list_minus in h2.
  destruct h2; auto.
Qed.

Lemma increasing_mem_remove :
  forall (ll:faml nat) (l:list nat),
    increasing_mem ll ->
    increasing_mem (remove list_nat_eq_dec l ll).
  intros ll ll' h1.
  do 2 red in h1; do 2 red.
  intros x h2; apply h1.
  rewrite <- subtract_remove_compat in h2.
  destruct h2; auto.
Qed.
 


Lemma fam_nat_to_ens_list_faml_to_fam :
  forall (ll:faml nat),
    increasing_faml ll ->
    increasing_mem ll ->
    fam_nat_to_ens_sort_set (faml_to_fam ll) (finite_mem_faml ll) =
    list_to_set ll.
  intro ll. 
  induction ll as [|al ll ih]; simpl.
  rewrite fam_nat_to_ens_sort_set_empty_eq; auto.
  unfold fam_nat_to_ens_sort_set, fam_to_ens_sort_set.
  rewrite im_full_sig_add.
  simpl.
  intros h0 hp.  
  pose proof (increasing_faml_cons_impl _ _ h0) as h1.
  specialize (ih h1).
  rewrite <- ih.
  unfold fam_nat_to_ens_sort_set. 
  f_equal.
  
  apply im_ext_in.
  intros x h2.
  unfold restriction_sig.
  simpl.
  apply sort_set_nat_functional; auto.
   
  symmetry.
  pose proof (hp _ (Add_intro2 _ _ _)) as h2.
  erewrite sort_set_nat_functional.
  apply incr_impl_sort_set_nat; auto. 
  reflexivity. 
  eapply p_mem_downward_closed.
  apply incl_add.
  apply hp.
Qed.

Lemma incr_faml_no_dup_minimal :
  forall (ll ll':faml nat),
    increasing_faml ll ->
    NoDup ll' ->
    list_to_set ll = list_to_set ll' ->
    ll' //= ll ->
    ll = ll'.
  unfold increasing_faml, le_list_nat.
  apply incrR_no_dup_minimal.
  apply list_nat_eq_dec.
Qed.

Lemma incr_impl_is_sorted_family :
  forall (ll:faml nat),
    increasing_faml ll ->
    increasing_mem ll ->
    is_sorted_family ll.
  intros l h1 h0; red.
  pose proof (increasing_faml_no_dup l h1) as hn.
  unfold sort_family.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  symmetry.
  unfold sort_set, min_set_sortable.
  destruct constructive_definite_description as [l' h2]; simpl.
  destruct h2 as [h2 h3].
  destruct (tso_dec_lt_listR  _ lt_list_nat_sort_dec l l') as [h4 | h4].
  destruct h4 as [h4|]; auto. 
  specialize (h3 l). 
  hfirst h3. 
  constructor; auto.
  constructor. split; auto.
  rewrite fam_nat_to_ens_list_faml_to_fam; auto.
  intro h5; destruct h5.
  apply (irrefl_lt_listR _ _ _ h4).
  specialize (h3 hf).
  pose proof (trans_lt_listR _ _ _ _ _ h3 h4) as h5.
  pose proof (irrefl_lt_listR _ _ _  h5); contradiction.
  destruct h2 as [h2], h2 as [h2 h5].
  rewrite fam_nat_to_ens_list_faml_to_fam in h2.
  symmetry.
  eapply incr_faml_no_dup_minimal; auto.
  unfold le_faml_nat.
  apply lt_list_impl_leR; auto.
  assumption.
  assumption.
  rewrite faml_to_fam_eq in hr.
  apply empty_set_nil in hr.
  apply map_eq_nil in hr; auto.
Qed.


  
Lemma weak_incr_faml_no_dup_impl :
  forall (l:faml nat),
    weak_increasing_faml l ->
    NoDup l ->
    increasing_faml l.
  unfold weak_increasing_faml, increasing_faml; intros; apply weak_incrR_no_dup_impl; auto.
Qed.


  

Lemma weak_increasing_faml_map :
  forall (f:list_nat_fun) (ll:faml nat),
    weak_incr_list_nat_fun f ->
    weak_increasing_faml ll ->
    weak_increasing_faml (map f ll).
  unfold weak_increasing_faml; apply weak_increasingR_map.
Qed.

Lemma weak_incr_famlS :
  forall (ll:faml nat),
    weak_increasing_faml ll ->
    weak_increasing_faml (famlS ll).
  intros ll h1.
  unfold famlS.
  apply weak_increasing_faml_map; auto.
  apply weak_incr_list_nat_fun_lS; auto.
Qed.



(*Note that [linds] and [linds'] are unsorted since [list_singularize] 
filters backwards, so [list_singularize [a b c b a]] would be [c b a].
Since we wish to compare [linds]s, we need something canonical, 
i.e. sorted, to account for [list_singularize]'s permuting in 
friction against [sort]'s permuting.
So when you want to compute or sort something like [linds (a::l)], 
you will need a canonical sorted list instead.*)

Definition sorted_linds {T:Type} (pfdt:type_eq_dec T) (l:list T) :
  faml nat :=
  sort_list lt_list_nat (sort_dec_tso_dec _ lt_list_nat_sort_dec) (linds pfdt l).

Definition sorted_linds' {T:Type} (pfdt:type_eq_dec T) (l:list T) :
  faml nat :=
  sort_list lt_list_nat (sort_dec_tso_dec _ lt_list_nat_sort_dec) (linds' pfdt l).

  
Lemma sorted_linds_compat' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    sorted_linds pfdt l =
    sorted_linds' pfdt l.
  unfold sorted_linds', sorted_linds; intros; rewrite linds_compat'; auto.
Qed.


Lemma sorted_linds_nil' :
  forall {T:Type} (pfdt:type_eq_dec T),
    sorted_linds' pfdt nil = nil.
  unfold sorted_linds'; intros.
  rewrite linds_nil', sort_list_nil; auto.
Qed.

Lemma sorted_linds_nil :
  forall {T:Type} (pfdt:type_eq_dec T),
    sorted_linds pfdt nil = nil.
  intros; rewrite sorted_linds_compat'; apply sorted_linds_nil'.
Qed.

Lemma nin_nil_sorted_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    ~In nil (sorted_linds pfdt l).
  intros; intro h1.
  unfold sorted_linds in h1.
  rewrite in_sort_list_iff in h1.
  rewrite in_linds_iff in h1.
  destruct h1 as [x h2]; red in h2.
  destruct h2 as [h2 h3].
  destruct h2 as [h2 h4].
  symmetry in h4.
  unfold linds_occ in h4. 
  destruct in_dec as [h5 | h5].
  unfold linds_occ' in h4.
  apply map_seg_eq_nil in h4.
  rewrite (count_occ_In pfdt) in h2.
  omega.
  contradiction.
Qed.



Lemma nin_nil_sorted_linds' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    ~In nil (sorted_linds' pfdt l).
  intros; rewrite <- sorted_linds_compat'; apply nin_nil_sorted_linds; auto.
Qed.

Lemma sorted_linds_eq_nil_iff' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    sorted_linds' pfdt l = nil <->
    l = nil.
  intros T h1 l; split; intro h2.
  unfold sorted_linds' in h2.
  rewrite (sort_list_eq_nil_iff _ lt_list_nat_sort_dec) in h2.
  rewrite linds_eq_nil_iff' in h2; auto.
  subst.
  apply sorted_linds_nil'.
Qed.


Lemma sorted_linds_eq_nil_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    sorted_linds pfdt l = nil <->
    l = nil.
  intros; rewrite sorted_linds_compat'; apply sorted_linds_eq_nil_iff'.
Qed.

Lemma in_sorted_linds_iff :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (ln:list nat),
    In ln (sorted_linds pfdt l) <->
    In ln (linds pfdt l).
  unfold sorted_linds; intros; split; intro h1.
  rewrite in_sort_list_iff in h1; auto.
  rewrite in_sort_list_iff; auto.
Qed.


Lemma length_sorted_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    length (sorted_linds pfdt l) = length (list_singularize pfdt l).
  unfold sorted_linds; intros; rewrite length_sort_list.
  unfold linds.
  apply map_length; auto.
Qed.

Lemma incr_sorted_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    increasing_faml (sorted_linds pfdt l).
  intros T h1 l.
  unfold increasing_faml, sorted_linds.
  apply weak_incrR_no_dup_impl.
  apply weak_incrR_sort_list.
  apply no_dup_sort_list.
  apply no_dup_linds.
Qed.

  
Lemma length_sorted_linds' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    length (sorted_linds' pfdt l) = length (list_singularize pfdt l).
  intros; rewrite <- sorted_linds_compat'; apply length_sorted_linds; auto.
Qed.

Lemma sorted_linds_eq_length :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    sorted_linds pfdt l =
    sorted_linds pfdu l' ->
    length (list_singularize pfdt l) =
    length (list_singularize pfdu l').
  intros T U h1 h2 l l' h3.
  pose proof (length_sorted_linds h1 l) as h4.
  rewrite h3 in h4.
  rewrite length_sorted_linds in h4; auto.
Qed.

Lemma sorted_linds_eq_length' :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    sorted_linds' pfdt l =
    sorted_linds' pfdu l' ->
    length (list_singularize pfdt l) =
    length (list_singularize pfdu l').
  intros T U h1 h2 l l' h3.
  pose proof (length_sorted_linds' h1 l) as h4.
  rewrite h3 in h4.
  rewrite length_sorted_linds' in h4; auto.
Qed.

  
Lemma no_dup_sorted_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup (sorted_linds pfdt l).
  intros T h1 l.
  unfold sorted_linds. 
  apply no_dup_sort_list.
  apply no_dup_linds.
Qed.

  
Lemma no_dup_sorted_linds' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    NoDup (sorted_linds' pfdt l).
  intros; rewrite <- sorted_linds_compat'; apply no_dup_sorted_linds; auto.
Qed.

Lemma incl_sorted_linds' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    incl (sorted_linds' pfdt l) (linds' pfdt l).
  intros T h1 l; red; intros l' h2.
  unfold sorted_linds' in h2.
  rewrite in_sort_list_iff in h2; auto.
Qed.

Lemma incl_sorted_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    incl (sorted_linds pfdt l) (linds pfdt l).
  intros; rewrite sorted_linds_compat', linds_compat'; apply incl_sorted_linds'.
Qed.

Lemma incl_linds_sorted :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    incl (linds' pfdt l) (sorted_linds' pfdt l).
  intros T h1 l; red; intros l' h2.
  unfold sorted_linds'.
  rewrite in_sort_list_iff ; auto.
Qed.

Lemma incl_linds_sorted' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    incl  (linds pfdt l) (sorted_linds pfdt l).
  intros; rewrite sorted_linds_compat', linds_compat'; apply incl_linds_sorted.
Qed.

Lemma faml_to_fam_sorted_linds :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T),
    faml_to_fam (sorted_linds pfdt l) =
    faml_to_fam (linds pfdt l).
  intros T h1 l.
  apply Extensionality_Ensembles; red; split; red; intros ln h2; apply in_faml_to_fam_impl in h2; destruct h2 as [ln' [h2 h3]]; subst; unfold sorted_linds in h2.
  rewrite in_sort_list_iff in h2.
  apply in_faml_impl; auto.
  apply in_faml_impl.
  unfold sorted_linds.
  rewrite in_sort_list_iff; auto.
Qed.


(*We need to used [sorted_linds'] (rather than [linds']) 
to account for [list_singularize]'s permuting in 
friction against [sort]'s.*)
Lemma sorted_linds_cons_in' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T) (pf:In a l),
    let ln := linds_occ' pfdt l a pf in
    let ls := sorted_linds' pfdt l in
    sorted_linds' pfdt (a::l) =
    (0::(lS ln))::famlS (remove list_nat_eq_dec ln ls). 
  intros T h1 l a h2; simpl.  
  eapply (weak_incrR_length_count_occ_eq list_nat_eq_dec).
  apply lt_list_nat_sort_dec. 
  apply weak_incrR_sort_list.
  simpl.
  destruct (nil_dec'  (famlS
                         (remove list_nat_eq_dec (linds_occ' h1 l a h2) (sorted_linds' h1 l)))) as [h3 | h3].
  unfold famlS in h3. 
  apply map_eq_nil in h3.
  rewrite remove_eq_nil_iff in h3.
  destruct h3 as [h3 | h3].
  unfold sorted_linds' in h3.
  rewrite sort_list_eq_nil_iff in h3.
  rewrite linds_eq_nil_iff' in h3; subst.
  contradiction.
   
  apply (all_sing_remove' list_nat_eq_dec) in h3.
  rewrite h3; simpl; auto.
  rewrite (hd_compat' _ h3).
  split.
  left.
  apply lt_hd0_famlS.
  intro h4.
  rewrite in_remove_iff in h4.
  destruct h4 as [h4 h5].
  apply nin_nil_sorted_linds' in h4; auto.
  rewrite <- hd_compat'.
  apply weak_incr_famlS.
  apply weak_increasingR_remove.
  apply trans_lt_list_nat.
  unfold sorted_linds.
  apply weak_incrR_sort_list.
  unfold sorted_linds' at 1.
  rewrite length_sort_list.
  rewrite length_linds'.
  simpl.
  rewrite list_singularize_in_cons; auto.
  rewrite length_famlS.
  rewrite length_remove_no_dup.
  rewrite length_sorted_linds'.
  rewrite  list_singularize_in_iff in h2.
  rewrite <- (S_pred _ _ (not_nil_lt _ (in_not_nil _ _ h2))); auto.
  apply no_dup_sorted_linds'.
  unfold sorted_linds'.
  rewrite in_sort_list_iff.
  unfold linds'.
  rewrite in_map_in_dep_iff.
  exists a.
  ex_goal.
  rewrite <- list_singularize_in_iff; auto.
  exists hex.
  rewrite fun_in_dep_incl_compat.
  f_equal.
  apply proof_irrelevance.
  intro l'.  
  unfold sorted_linds'.
  assert (h3:list_nat_eq_dec = sort_dec_eq_dec _ lt_list_nat_sort_dec).
    apply type_eq_dec_eq.
  rewrite h3.
  rewrite count_occ_sort_list.
  unfold linds'.
  destruct (list_nat_eq_dec l' (0::lS (linds_occ' h1 l a h2))) as [he | he]; subst.
  rewrite count_occ_cons_eq.
  rewrite count_occ_map_in_dep.
  rewrite <- h3.
  rewrite count_occ_famlS_in0.
  rewrite length_inv_im_in_dep_inj; auto.
  apply no_dup_list_singularize.
  apply inj_dep_incl.
  apply inj_dep_linds_occ'.
  eapply in_not_nil.
  rewrite in_inv_im_in_dep_iff.
  pose proof (in_eq a l) as h4.
  rewrite (list_singularize_in_iff h1) in h4.
  exists h4.
  unfold fun_in_dep_incl.
  erewrite linds_occ_functional'.
  rewrite linds_occ_cons_eq' at 1.
  simpl.
  reflexivity.
  reflexivity.
  reflexivity.
  left; auto.
  assumption.
  reflexivity.

  pose proof (no_dup_linds_occ'  h1 (a :: l) _ (in_eq _ _)) as h4.
  pose proof (no_dup_map_in_dep_inj
                (fun_in_dep_incl (a :: l) (list_singularize h1 (a :: l))
                                  (linds_occ' h1 (a :: l)) (incl_list_singularize h1 (a :: l)))
                (no_dup_list_singularize _ _)) as h5.
  hfirst h5.
  apply inj_dep_incl; auto.
  apply inj_dep_linds_occ'; auto.
  specialize (h5 hf).
  let P := type of h5 in match P with NoDup ?l' => pose l' as ml end.
  destruct (in_dec list_nat_eq_dec l' ml) as [h6 | h6].
  rewrite count_occ_no_dup; auto.
  rewrite count_occ_no_dup; auto.
  constructor.
  intro h7.
  rewrite in_famlS_iff in h7.
  destruct h7 as [al [h7 h8]].
  pose proof (in_eq 0 (lS (linds_occ' h1 l a h2))) as h9.
  rewrite h8 in h9.
  rewrite in_lS_iff in h9.
  destruct h9 as [h9 [h10 h11]].
  discriminate.
  apply no_dup_map_inj.
  apply no_dup_remove.
  apply no_dup_sort_list.
  apply no_dup_map_in_dep_inj.
  apply no_dup_list_singularize.
  apply inj_dep_incl; auto.
  apply inj_dep_linds_occ'.
  apply inj_lS.
  right.
  unfold ml in h6.
  rewrite in_map_in_dep_iff in h6.
  destruct h6 as [x [h6 h7]].
  unfold fun_in_dep_incl in h7.
  pose proof h6 as h6'.
  destruct (h1 x a) as [h8 | h8]; subst.
  contradict he.
  erewrite linds_occ_functional'.
  rewrite linds_occ_cons_eq' at 1.
  reflexivity.
  reflexivity. 
  reflexivity. 
  rewrite <- list_singularize_in_iff in h6'.
  destruct h6' as [|h6']; subst.
  contradict h8; auto.
  rewrite in_famlS_iff.
  exists (linds_occ' h1 l x h6').
  split.
  rewrite in_remove_iff.
  split.
  rewrite in_sort_list_iff.
  rewrite in_map_in_dep_iff.
  exists x.
  ex_goal.
  rewrite <- list_singularize_in_iff; auto.
  exists hex.
  unfold fun_in_dep_incl.
  f_equal.
  apply proof_irrelevance.
  intro h9.
  apply inj_dep_linds_occ' in h9.
  contradiction.
  erewrite linds_occ_functional'.
  rewrite linds_occ_cons_neq'.
  reflexivity.
  apply h8.
  reflexivity.
  reflexivity.
  rewrite (count_occ_In (sort_dec_eq_dec _ lt_list_nat_sort_dec)) in h6.
  rewrite (not_lt0_iff (count_occ (sort_dec_eq_dec _ lt_list_nat_sort_dec) ml l')) in h6.
  fold ml. rewrite h6.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  rewrite hl in he.
  contradict he; auto. 
  symmetry. rewrite <- not_lt0_iff.
  rewrite <- (count_occ_In (sort_dec_eq_dec _ lt_list_nat_sort_dec)
              (famlS
        (remove (sort_dec_eq_dec lt_list_nat lt_list_nat_sort_dec)
           (linds_occ' h1 l a h2)
           (sort_list lt_list_nat
              (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec)
              (map_in_dep
                 (fun_in_dep_incl l (list_singularize h1 l)
                                   (linds_occ' h1 l) (incl_list_singularize h1 l)))))) l').
  intro h7.
  rewrite in_famlS_iff in h7.
  destruct h7 as [ln [h7 h8]].
  rewrite in_remove_iff in h7.
  destruct h7 as [h7 h9].
  rewrite in_sort_list_iff in h7.
  rewrite in_map_in_dep_iff in h7.
  destruct h7 as [x [h7 h10]].
  destruct (h1 x a) as [|h11]; subst.
  unfold fun_in_dep_incl in h9.
  contradict h9.
  f_equal; apply proof_irrelevance.
  rewrite <- (not_lt0_iff (count_occ (sort_dec_eq_dec lt_list_nat lt_list_nat_sort_dec) ml
         (lS
            (fun_in_dep_incl l (list_singularize h1 l) 
               (linds_occ' h1 l) (incl_list_singularize h1 l) x h7)))) in h6.
  rewrite <- (count_occ_In (sort_dec_eq_dec _ lt_list_nat_sort_dec)
                           ml  (lS
            (fun_in_dep_incl l (list_singularize h1 l) 
                              (linds_occ' h1 l) (incl_list_singularize h1 l) x h7))) in h6.
  contradict h6.
  unfold ml.
  rewrite in_map_in_dep_iff.
  exists x.
  ex_goal.
  rewrite <- list_singularize_in_iff.
  right; auto.
  pose proof (iff2 (list_singularize_in_iff h1 _ _) h7); auto.
  exists hex.
  unfold fun_in_dep_incl.
  erewrite linds_occ_functional'.
  rewrite linds_occ_cons_neq'.
  reflexivity.
  apply h11.
  reflexivity.
  reflexivity.
Qed.

Lemma sorted_linds_cons_in :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T) (pf:In a l),
    let ln := linds_occ pfdt l a in
    let ls := sorted_linds pfdt l in
    sorted_linds pfdt (a::l) =
    (0::(lS ln))::famlS (remove list_nat_eq_dec ln ls). 
  simpl; intros ? ? ? ? h2; do 2 rewrite sorted_linds_compat'; rewrite (linds_occ_compat' _ _ _ h2); apply sorted_linds_cons_in'.
Qed.


Lemma sorted_linds_cons_nin' :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T) (pf:~In a l),
    sorted_linds' pfdt (a::l) =
    (0::nil)::(famlS (sorted_linds' pfdt l)).
  intros T h1 l a h2.
  destruct (nil_dec' l) as [|hn]; subst.  
  unfold sorted_linds', linds', fun_in_dep_incl; simpl. 
  erewrite linds_occ_functional'.
  rewrite linds_occ_cons_eq_nin' at 1.
  rewrite sort_list_sing; auto.
  apply h2.
  reflexivity.
  reflexivity.
  eapply (weak_incrR_length_count_occ_eq list_nat_eq_dec).
  apply lt_list_nat_sort_dec.
  apply weak_incrR_sort_list.
  eapply weak_increasingR_impl_cons. 
  apply weak_incr_famlS.
  apply weak_incrR_sort_list.
  rewrite Rw_lt_list_nat_iff.
  pose proof (lt_hd0_famlS nil (sorted_linds' h1 l)) as ha.
  hfirst ha.
  contradict hn.
  rewrite famlS_eq_nil_iff in hn.
  unfold sorted_linds' in hn.
  rewrite (sort_list_eq_nil_iff _ _ (linds' h1 l))in hn at 1.
  rewrite linds_eq_nil_iff' in hn; auto.
  specialize (ha hf).
  hfirst ha.
  intro hb.
  unfold sorted_linds' in hb.
  rewrite in_sort_list_iff in hb.
  unfold linds' in hb.
  rewrite in_map_in_dep_iff in hb.
  destruct hb as [x [hc hd]].
  unfold fun_in_dep_incl in hd.
  unfold linds_occ' in hd.
  rewrite map_seg_eq_iff in hd.
  destruct hd as [hd].
  simpl in hd.
  pose proof hc as hc'.
  rewrite  <- list_singularize_in_iff in hc'.
  rewrite (count_occ_In h1) in hc'.
  omega.
  specialize (ha hf0).
  apply lt_list_nat_impl_le; auto.
  simpl.
  erewrite hd_functional'.
  apply ha.
  unfold sorted_linds'.
  unfold linds'.
  simpl.
  rewrite length_sort_list.
  rewrite length_map_in_dep.
  destruct (in_dec h1 a l) as [|h3]. 
  contradiction.
  simpl.
  f_equal.
  rewrite list_singularize_nin; auto.
  rewrite length_famlS.
  rewrite length_sort_list.
  rewrite length_map_in_dep.
  reflexivity.
  intro ln.
  unfold sorted_linds'. 
  assert (h3:list_nat_eq_dec = sort_dec_eq_dec _ lt_list_nat_sort_dec).
    apply type_eq_dec_eq.
  rewrite h3.
  rewrite count_occ_sort_list.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr].
  subst. 
  unfold linds'.
  
  rewrite count_occ_map_in_dep. 
  rewrite length_inv_im_in_dep_inj.
  f_equal.
  symmetry. rewrite  <- not_lt0_iff.
  intro h4.
  rewrite <- (count_occ_In (sort_dec_eq_dec lt_list_nat lt_list_nat_sort_dec)
               (famlS (sort_list lt_list_nat
            (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec)
            (map_in_dep
               (fun_in_dep_incl l (list_singularize h1 l)
                  (linds_occ' h1 l) (incl_list_singularize h1 l)))))
               (0 :: nil)) in h4 at 1.
  rewrite in_famlS_iff in h4.
  destruct h4 as [? [? h4]].
  pose proof (in_eq 0 nil) as h5.
  rewrite h4 in h5.
  rewrite in_lS_iff in h5.
  destruct h5 as [? [? ?]]; discriminate.
  assumption.
  simpl.
  rewrite list_singularize_nin.
  constructor.
  rewrite <- list_singularize_in_iff; auto.
  apply no_dup_list_singularize.
  assumption.
  apply inj_dep_incl.
  apply inj_dep_linds_occ'.
  eapply (in_not_nil _ a).
  rewrite in_inv_im_in_dep_iff. 
  ex_goal.
  rewrite <- list_singularize_in_iff.
  left; auto.
  exists hex.
  unfold fun_in_dep_incl.
  unfold linds_occ'.
  simpl. 
  destruct (h1 a a) as [h4 | h4].
  rewrite map_seg_eq_iff.
  simpl.
  ex_goal.
  f_equal.
  rewrite (count_occ_In h1) in h2; omega.
  exists hex0.
  intros m h5.
  destruct in_dec as [|h6].
  contradiction.
  assert (h7:m  = 0). omega.
  subst. 
  reflexivity.
  contradict h4; auto.
  assumption.
  
  destruct (in_dec nat_eq_dec 0 ln) as [h4 | h4].
  rewrite <- h3.
  rewrite count_occ_famlS_in0.
  rewrite <- not_lt0_iff.
  intro h5. 
  rewrite <- (count_occ_In list_nat_eq_dec (linds' h1 (a::l)) ln) in h5.
  rewrite <- linds_compat' in h5. 
  rewrite in_linds_iff in h5.
  destruct h5 as [x [[h5 h6] h7]]; subst.
  rewrite in_linds_occ_iff in h4.
  destruct h4 as [m [h6 [h8 h9]]].
  destruct (h1 x a) as [|h10]; subst.
  simpl in h8, h9.
  destruct (h1 a a) as [h10 | h10].
  pose proof h2 as h2'.
  rewrite (count_occ_In h1) in h2'.
  assert (m = 0). omega.
  subst.
  contradict hr.
  rewrite linds_occ_cons_eq.
  simpl; f_equal.
  symmetry.
  rewrite linds_occ_nin; auto.
  pose proof (h10 eq_refl); contradiction.
  simpl in h9.
  destruct (h1 x a) as [h11 | h11].
  contradiction.
  discriminate.
  assumption. 
  rewrite <- h3.
  rewrite count_occ_famlS_nin0; auto.
  rewrite h3.
  rewrite count_occ_sort_list.
  rewrite <- h3. apply count_occ_linds_cons_nin'; auto.
  Grab Existential Variables.
  intro h3.
  rewrite famlS_eq_nil_iff in h3.
  unfold sorted_linds' in h3.
  rewrite (sort_list_eq_nil_iff lt_list_nat lt_list_nat_sort_dec (linds' h1 l)) in h3.
  rewrite linds_eq_nil_iff' in h3.
  contradiction.
Qed.


Lemma sorted_linds_cons_nin :
  forall {T:Type} (pfdt:type_eq_dec T)
         (l:list T) (a:T) (pf:~In a l),
    sorted_linds pfdt (a::l) =
    (0::nil)::(famlS (sorted_linds pfdt l)).
  simpl; intros; do 2 rewrite sorted_linds_compat'; apply sorted_linds_cons_nin'; auto.
Qed.

Lemma sorted_linds_app_sing_in' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pf:In a l),
    let ll := sorted_linds' pfdt l in
    let ln := linds_occ' pfdt l a pf in
    sorted_linds' pfdt (l++a::nil) =
    sort_list _ lt_list_nat_tso_dec
              ((ln++length l::nil) :: (remove list_nat_eq_dec ln ll)).
  intros T h1 l a h2; simpl.
  eapply (weak_incrR_length_count_occ_eq list_nat_eq_dec).
  apply lt_list_nat_sort_dec.
  apply weak_incrR_sort_list.
  unfold sorted_linds'.
  rewrite lt_list_nat_tso_dec_eq.
  apply weak_incrR_sort_list.
  rewrite length_sorted_linds'.
  rewrite length_sort_list.
  simpl.
  rewrite list_singularize_rev_cons.
  rewrite app_length.
  simpl.
  rewrite S_compat.
  f_equal.
  rewrite length_remove_no_dup.
  rewrite length_remove_no_dup.
  rewrite length_sorted_linds'; auto.
  apply no_dup_sorted_linds'.
  unfold sorted_linds'.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds'.
  apply no_dup_list_singularize.
  rewrite <- list_singularize_in_iff.
  assumption.
  intro l'.
  
  rewrite list_nat_eq_dec_eq'.
  unfold sorted_linds'.
  rewrite lt_list_nat_tso_dec_eq.
  do 2 rewrite count_occ_sort_list.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]; subst.
  rewrite count_occ_no_dup.
  match goal with
      |- _ = S ?r' => assert (h3:r' = 0) end.
  rewrite <- not_lt0_iff.
  apply ngt_nlt.
  rewrite <- count_occ_In.
  intro h3.
  rewrite in_remove_iff in h3.
  destruct h3 as [h3 h4].
  rewrite in_sort_list_iff in h3.
  pose proof (in_app_r (in_eq (length l) nil) (linds_occ' h1 l a h2)) as h5.
  rewrite in_linds_iff' in h3.
  destruct h3 as [x [h3 h7]].
  destruct h3 as [h3 h6].
  rewrite h6 in h5.
  rewrite <- linds_occ_compat' in h5.
  apply in_linds_occ_lt in h5.
  apply lt_irrefl in h5; auto.
  rewrite h3; auto. 
  apply no_dup_linds'.
  rewrite in_linds_iff'.
  exists a.
  red.
  split.
  exists (in_app_r (in_eq a nil) _).
  symmetry.
  pose proof (proof_irrelevance _
                                (in_app_r (in_eq a nil) l)
                                (in_app_l h2 (a::nil))) as h3.
  rewrite h3. 
  rewrite (linds_occ_app_in' _ _  _ _ _ (in_eq _ _)) at 1.
  f_equal. 
  unfold linds_occ'.
  simpl.
  destruct (h1 a a); simpl.
  rewrite plus_0_r; auto.
  contradict n; auto.
  intros x h3.
  destruct h3 as [h3 h4].

  destruct (h1 a x) as [h5 | h5]; auto.
  pose proof h3 as h3'.
  apply in_app_or in h3'.
  destruct h3' as [h7 | h7].
  pose proof (proof_irrelevance _ h3 (in_app_l h7 (a::nil))).
  subst.
  rewrite (linds_occ_app_nin_r' h1  _  _ _ h7) in h4.
  pose proof (in_app_r (in_eq (length l) nil) (linds_occ' h1 l a h2)) as h8.
  rewrite h4 in h8.
  rewrite <- linds_occ_compat' in h8.
  apply in_linds_occ_lt in h8.
  apply lt_irrefl in h8.
  contradiction.
  intro h8.
  destruct h8; subst.
  contradict h5; auto.
  contradiction.
  destruct h7; auto.
  contradiction.
   
  destruct (in_dec list_nat_eq_dec  l' (linds' h1 (l++a::nil))) as [h3 | h3].
  pose proof h3 as h3'.
  rewrite in_linds_iff' in h3'.
  destruct h3' as [x h5].
  red in h5.
  destruct h5 as [h5 h7].
  destruct h5 as [h5 h8]; subst.
  pose proof h5 as h5'.
  apply in_app_or in h5'.
  destruct h5' as [h8 | h8].
  assert (h9:x <> a).
    intro; subst.
  contradict hr.
  pose proof (proof_irrelevance _ h5 (in_app_l h2 _)); subst.  
  rewrite (linds_occ_app_in' _ _ _ _ _ (in_eq _ _)).
  f_equal.
  unfold linds_occ'.
  simpl.
  destruct (h1 a a) as [h9 | h9]; simpl.
  rewrite plus_0_r; auto.
  contradict h9; auto.
  pose proof (proof_irrelevance _ h5 (in_app_l h8 _)).
  subst.
  rewrite linds_occ_app_nin_r' in hr.
  
  rewrite count_occ_no_dup; auto.
  rewrite count_occ_no_dup; auto.
  apply no_dup_remove.
  apply no_dup_sort_list.
  apply no_dup_linds'.
  rewrite in_remove_iff.
  split.
  rewrite in_sort_list_iff.
  rewrite in_linds_iff'.
  exists x.
  red.
  split.
  exists h8.
  rewrite linds_occ_app_nin_r'; auto.
  intro h10.
  inversion h10; subst.
  contradict h9; auto.
  contradiction.
  intros x' h10.
  destruct h10 as [h10 h11].
  rewrite linds_occ_app_nin_r' in h11.
  apply inj_dep_linds_occ' in h11; auto.
  intro h12.
  inversion h12; subst.
  contradict h9; auto.
  contradiction.
  rewrite linds_occ_app_nin_r'.
  intro h10.
  apply inj_dep_linds_occ' in h10.
  contradiction.
  intro h10.
  inversion h10.
  subst.
  contradict h9; auto.
  contradiction.
  apply no_dup_linds'.
  intro h10.
  inversion h10; subst.
  contradict h9; auto.
  contradiction.
  inversion h8; subst.
  pose proof (proof_irrelevance _ h5 (in_app_l h2 _)); subst.
  contradict hr.
  rewrite (linds_occ_app_in' _ _ _ _ _ (in_eq _ _)).
  f_equal.
  unfold linds_occ'.
  simpl.
  destruct (h1 x x) as [h10 | h10].
  simpl.
  rewrite plus_0_r; auto.
  contradict h10; auto.
  contradiction.

  pose proof h3 as h3'.
  rewrite (count_occ_In  (sort_dec_eq_dec lt_list_nat lt_list_nat_sort_dec)) in h3'.
  apply ngt_nlt in h3'.
  rewrite not_lt0_iff in h3'.
  rewrite h3'.
  symmetry.
  rewrite <- not_lt0_iff.
  apply ngt_nlt.
  rewrite <- (count_occ_In _).
  intro h4.
  rewrite in_remove_iff in h4.
  destruct h4 as [h4 h5].
  rewrite in_sort_list_iff in h4.
  rewrite in_linds_iff' in h4.
  destruct h4 as [x h4].
  red in h4.
  destruct h4 as [h4 h7].
  destruct h4 as [h4 h8].
  subst.
  contradict h3.
  rewrite in_linds_iff'.
  exists x.
  red.
  split.
  exists (in_app_l h4 (a::nil)).
  rewrite linds_occ_app_nin_r'; auto.
  intro h8.
  destruct h8; subst.
  contradict h5; f_equal; apply proof_irrelevance.
  contradiction.
  intros x' h8. 
  destruct h8 as [h8 h9].
  destruct (h1 x' a) as [h10 | h10]; subst.
  contradict hr; auto.
  rewrite h9.
  pose proof (proof_irrelevance _ h8 
                                (in_app_l h2 (a::nil))); subst.
  rewrite (linds_occ_app_in' _ _ _ _ _ (in_eq _ _)).
  f_equal.
  unfold linds_occ'; simpl.
  destruct (h1 a a) as [h10 | h10].
  simpl; rewrite plus_0_r; auto.
  contradict h10; auto.
  pose proof h8 as h11.
  apply in_app_or in h11.
  destruct h11 as [h11 | h12].
  pose proof (proof_irrelevance _ h8
                                (in_app_l h11 (a::nil))); subst.
  rewrite linds_occ_app_nin_r' in h9.
  apply inj_dep_linds_occ' in h9; auto.
  contradict h10.
  destruct h10; auto.
  contradiction.
  destruct h12; subst.
  contradict h10; auto.
  contradiction.
Qed.

Lemma sorted_linds_app_sing_in :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pf:In a l),
    let ll := sorted_linds pfdt l in
    let ln := linds_occ pfdt l a in
    sorted_linds pfdt (l++a::nil) =
    sort_list _ lt_list_nat_tso_dec
              ((ln++length l::nil) :: (remove list_nat_eq_dec ln ll)).
  intros T h1 l a h2; simpl.
  do 2 rewrite sorted_linds_compat'.
  rewrite (linds_occ_compat' _ _ _ h2).
  apply sorted_linds_app_sing_in'.
Qed.


Lemma sorted_linds_app_sing_nin' :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
         (pf:~In a l),
    sorted_linds' pfdt (l++a::nil) =
    sorted_linds' pfdt l ++ (length l :: nil) :: nil.
  intros T h1 l a h2.
  destruct (nil_dec' l) as [|h3]; subst; simpl. 
  unfold sorted_linds'. 
  rewrite <- linds_compat'. 
  rewrite linds_sing.
  rewrite sort_list_sing; auto.
  eapply (weak_incrR_length_count_occ_eq list_nat_eq_dec).
  apply lt_list_nat_sort_dec.
  apply weak_incrR_sort_list.
  pose proof h3 as h3'.
  rewrite <- (sorted_linds_eq_nil_iff' h1) in h3.
  pose proof (in_lst _  h3) as h4.
  assert (h5:lst (sorted_linds' h1 l) h3 <> nil).
    intro h5.
    rewrite h5 in h4.
    pose proof (nin_nil_sorted_linds' h1 l); contradiction. 
  pose proof (lt_list_nat_cons_iff  (lst (sorted_linds' h1 l) h3) nil (length l) h5) as h6. 

  pose proof (weak_increasingR_app_iff lt_list_nat trans_lt_list_nat (sorted_linds' h1 l) ((length l::nil)::nil) h3) (sing_not_nil _) as h7.
  rewrite <- h7.
  split.
  apply weak_incrR_sort_list.
  simpl.
  split; auto.
  left.
  rewrite h6.
  left.
  pose proof (in_linds_lt h1 l (lst (sorted_linds' h1 l) h3) _ (in_hd' _ h5)) as h8.
  hfirst h8.

  rewrite linds_compat'.
  apply incl_sorted_linds'.
  apply in_lst.
  eapply in_linds_lt.
  apply in_hd'.
  rewrite linds_compat'.
  eapply incl_sorted_linds'.
  apply in_lst.

  rewrite app_length.
  do 2 rewrite length_sorted_linds'.
  rewrite list_singularize_nin_app_nil.
  rewrite app_length; simpl; auto.
  assumption.

  intro l'. 
  destruct (in_dec list_nat_eq_dec l' (sorted_linds' h1 (l++a::nil))) as [h4 | h4].
  rewrite count_occ_no_dup; auto.
  pose proof h4 as h4'.
  unfold sorted_linds' in h4.
  rewrite in_sort_list_iff in h4.
  rewrite in_linds_iff' in h4.
  destruct h4 as [x [[h4 h5] h6]].
  subst.
  rewrite count_occ_app.
  simpl.
  lr_if_goal; fold hlr; destruct hlr as [hl | hr]. 
  rewrite S_compat.
  f_equal.
  symmetry.
  rewrite <- not_lt0_iff.
  apply ngt_nlt.
  rewrite <- count_occ_In.
  intro h7.
  unfold sorted_linds' in h7.
  rewrite in_sort_list_iff in h7.
  rewrite in_linds_iff' in h7.
  destruct h7 as [x' [h7 h8]].
  destruct h7 as [h7 h9].
  rewrite <- hl in h9.
  pose proof (in_linds_occ_linds' h1 l x' h7) as h10.
  rewrite <- h9 in h10.
  rewrite <- linds_compat' in h10.
  pose proof (in_linds_lt h1 l _ _ (in_eq _ _) h10) as h11.
  apply lt_irrefl in h11; contradiction.
  assert (h7:x <> a).
    intro; subst.
    pose proof (proof_irrelevance _ h4 (in_app_r (in_eq _ _) l)); subst.
    rewrite linds_occ_app_nin_l' in hr.
    unfold linds_occ' in hr.
    simpl in hr.
    destruct (h1 a a) as [h8 | h8].
    simpl in hr.
    rewrite plus_0_r in hr.
    contradict hr; auto.
    contradict (h8 eq_refl).
    assumption.
  pose proof h4 as h8.
  apply in_app_or in h8.  
  destruct h8 as [h8 | h8].
  rewrite count_occ_no_dup; auto with arith.
  apply no_dup_sorted_linds'.
  unfold sorted_linds' in h4'.
  rewrite in_sort_list_iff in h4'.
  rewrite in_linds_iff' in h4'.
  destruct h4' as [x' h10].
  red in h10.
  destruct h10 as [h10 h11].
  destruct h10 as [h12 h13].
  apply inj_dep_linds_occ' in h13.
  subst.
  unfold sorted_linds'.
  rewrite in_sort_list_iff.
  rewrite in_linds_iff'.
  exists x'.
  red.
  split.
  exists h8.
  pose proof (proof_irrelevance _ h4 (in_app_l h8 _)); subst.
  rewrite linds_occ_app_nin_r'; auto.
  contradict h7.
  destruct h7; auto.
  contradiction.
  intros x h9.
  destruct h9 as [h9 h10].
  pose proof (proof_irrelevance _ h4 (in_app_l h8 _)).
  subst.
  rewrite linds_occ_app_nin_r' in h10; auto.
  apply inj_dep_linds_occ' in h10; auto.
  contradict h7.
  destruct h7; auto. 
  contradiction.
  destruct h8; subst.
  contradict h7; auto.
  contradiction.
  apply no_dup_sorted_linds'.

  pose proof h4 as h4'.
  rewrite  (count_occ_In list_nat_eq_dec) in h4.
  apply ngt_nlt in h4.
  rewrite not_lt0_iff in h4.
  rewrite h4.
  symmetry.
  rewrite <- not_lt0_iff. 
  apply ngt_nlt.
  rewrite <- count_occ_In.
  intro h5.
  unfold sorted_linds' in h5.
  apply in_app_or in h5.
  destruct h5 as [h5 | h5]. 
  rewrite in_sort_list_iff in h5.
  contradict h4'.
  unfold sorted_linds'.
  rewrite in_sort_list_iff.
  rewrite in_linds_iff'.
  rewrite in_linds_iff' in h5.
  destruct h5 as [x h5].
  red in h5.
  destruct h5 as [h5 h6].
  destruct h5 as [h5 h7]; subst.
  exists x.
  red.
  split.
  ex_goal.
  eapply in_app_l; auto.
  exists hex.
  pose proof (proof_irrelevance _ hex
                                (in_app_l h5 (a::nil))); subst.
  rewrite linds_occ_app_nin_r'; auto.
  intro h7; destruct h7; subst.
  contradiction.
  contradiction.
  intros x' h7. 
  destruct h7 as [h7 h8].
  assert (h9:x <> a).
    intro; subst; contradiction.
  pose proof @linds_occ_app_in'.
  pose proof (linds_occ_app_nin_r' h1 l (a::nil) x h5) as h10.
  hfirst h10.
  contradict h9.
  destruct h9; auto.
  contradiction.
  specialize (h10 hf).
  simpl in h10.
  rewrite h8 in h10.
  apply inj_dep_linds_occ' in h10; auto.
  destruct h5; subst. 
  contradict h4'.
  unfold sorted_linds'.
  rewrite in_sort_list_iff.
  rewrite in_linds_iff'.
  exists a.
  red.
  split.
  exists (in_app_r (in_eq _ _) _).
  rewrite linds_occ_app_nin_l'.
  unfold linds_occ'.
  simpl.
  destruct (h1 a a) as [h5 | h5].
  simpl.
  rewrite plus_0_r; auto.
  contradict h5; auto.
  assumption.
  intros x h5. 
  destruct h5 as [h5 h6].
  rewrite h6 in h4.
  rewrite <- not_lt0_iff in h4.
  apply nlt_ngt in h4.
  rewrite <- count_occ_In in h4.
  contradict h4.
  unfold sorted_linds'.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds'.
  contradiction.
Qed.

Lemma sorted_linds_app_sing_nin :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T)
           (pf:~In a l),
    sorted_linds pfdt (l++a::nil) =
    sorted_linds pfdt l ++ (length l :: nil) :: nil.
  intros; do 2 rewrite sorted_linds_compat'; apply sorted_linds_app_sing_nin'; auto.
Qed.


Lemma list_to_set_linds_occ_eq :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (a:T) (a':U),
    list_to_set (linds pfdt (a::l)) =
    list_to_set (linds pfdu (a'::l')) ->
    linds_occ pfdt l a =
    linds_occ pfdu l' a'.
  intros T U h1 h2 l l' a a' h5.
  destruct (in_dec h1 a l) as [h3 | h3],
                              (in_dec h2 a' l') as [h4 | h4].
  pose proof (in_linds_occ_linds h1 (a::l) _ (in_eq _ _)) as h6.
  rewrite linds_occ_cons_eq in h6.
  simpl in h6.
  rewrite list_to_set_in_iff in h6.
  rewrite h5 in h6.
  rewrite <- list_to_set_in_iff in h6.
  rewrite in_linds_iff in h6.
  destruct h6 as [y h6].
  red in h6.
  destruct h6 as [h6 h7].
  destruct h6 as [h6 h8].
  destruct h6 as [|h6]; subst.
  rewrite linds_occ_cons_eq in h8.
  simpl in h8.
  inversion h8.
  apply inj_lS in H0; auto.
  destruct (h2 y a') as [|h9]; subst.
  rewrite linds_occ_cons_eq in h8.
  simpl in h8.
  inversion h8.
  apply inj_lS in H0; auto.
  rewrite linds_occ_cons_neq in h8; auto.
  pose proof (in_eq 0 (lS (linds_occ h1 l a))) as h10.
  rewrite h8 in h10.
  apply nin0_lS in h10.
  contradiction.
  pose proof (in_linds_occ_linds h1 (a::l) _ (in_eq _ _)) as h6.
  rewrite list_to_set_in_iff in h6.
  rewrite h5 in h6.
  rewrite <- list_to_set_in_iff in h6.
  rewrite linds_occ_cons_eq in h6.
  simpl in h6.
  rewrite in_linds_iff in h6.
  destruct h6 as [y h6].
  red in h6.
  destruct h6 as [h6 h7].
  destruct h6 as [h6 h8].
  destruct h6 as [|h6]; subst.
  rewrite linds_occ_cons_eq in h8.
  simpl in h8.
  inversion h8.
  apply inj_lS in H0; auto.

  assert (h9:y <> a').
    intro; subst; contradiction.
  rewrite (linds_occ_cons_neq h2 _ _ _ h9) in h8.
  pose proof (in_eq 0 (lS (linds_occ h1 l a))) as h10.
  rewrite h8 in h10.
  apply nin0_lS in h10; contradiction.

  pose proof (in_linds_occ_linds h2 (a'::l') _ (in_eq _ _)) as h6.
  rewrite list_to_set_in_iff in h6.
  rewrite <- h5 in h6.
  rewrite <- list_to_set_in_iff in h6.
  rewrite linds_occ_cons_eq in h6.
  simpl in h6.
  rewrite in_linds_iff in h6.
  destruct h6 as [x [h6 h7]].
  destruct h6 as [h6 h8].
  destruct h6 as [|h6]; subst.
  rewrite linds_occ_cons_eq in h8.
  simpl in h8.
  inversion h8.
  apply inj_lS in H0; auto.

  assert (h9:x <> a).
    intro; subst; contradiction.
  
  rewrite linds_occ_cons_neq in h8; auto.
  pose proof (in_eq 0 (lS (linds_occ h2 l' a'))) as h10.
  rewrite h8 in h10.
  apply nin0_lS in h10.
  contradiction.

  do 2 (rewrite nin_linds_occ_eq_iff; auto).
Qed.



Lemma list_to_set_linds_from_cons :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (a:T) (a':U),
    list_to_set (linds pfdt (a::l)) =
    list_to_set (linds pfdu (a'::l')) ->
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l').
  intros T U h1 h2 l l' a a' h5.
  destruct (in_dec h1 a l) as [h3 | h3],
                              (in_dec h2 a' l') as [h4 | h4].
  pose proof (list_to_set_linds_occ_eq _ _ _ _ _ _  h5) as hl.
  erewrite list_to_set_sort_list in h5.
  symmetry in h5.
  erewrite list_to_set_sort_list in h5.
  pose proof (sorted_linds_cons_in h2 _ _ h4) as h6; simpl in h6.
  pose proof (sorted_linds_cons_in h1 _ _ h3) as h7; simpl in h7.
  unfold sorted_linds in h6, h7. 
  rewrite h6, h7 in h5 at 1.
  simpl in h5.
  rewrite hl in h5.
  pose proof (f_equal (fun S => Subtract S (0::lS (linds_occ h2 l' a'))) h5) as ha.
  simpl in ha.
  rewrite sub_add_same_nin in ha.
  rewrite sub_add_same_nin in ha.
  unfold famlS in ha.
  rewrite map_im_compat in ha.
  rewrite map_im_compat in ha.
  rewrite <- subtract_remove_compat in ha.
  rewrite <- subtract_remove_compat in ha.
  rewrite <- list_to_set_sort_list in ha.
  rewrite <- list_to_set_sort_list in ha.
  rewrite im_subtract_inj in ha.
  rewrite im_subtract_inj in ha.
  apply in_sub_inj in ha.
  apply im_inj_inj in ha; auto.
  apply inj_lS.
  apply Im_intro with (linds_occ h2 l' a').
  rewrite <- list_to_set_in_iff.
  apply in_linds_occ_linds.
  assumption.
  reflexivity.
  apply Im_intro with (linds_occ h2 l' a').
  rewrite <- list_to_set_in_iff.
  rewrite <- hl.
  apply in_linds_occ_linds.
  assumption.
  reflexivity.
  apply inj_lS.
  apply inj_lS.
  intro hb.
  rewrite <- list_to_set_in_iff in hb.
  apply nin_famlS in hb; auto.
  left; auto.
  intro hb.
  rewrite <- list_to_set_in_iff in hb.
  apply nin_famlS in hb; auto.
  left; auto.

  erewrite list_to_set_sort_list in h5.
  symmetry in h5.
  erewrite list_to_set_sort_list in h5.
  pose proof (sorted_linds_cons_nin h2 _ _ h4) as h6; simpl in h6.
  pose proof (sorted_linds_cons_in h1 _ _ h3) as h7; simpl in h7.
  unfold sorted_linds in h6, h7. 
  rewrite h6, h7 in h5 at 1.
  simpl in h5.
  pose proof (Add_intro2 _ (list_to_set
            (famlS
               (sort_list lt_list_nat
                  (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec)
                  (linds h2 l')))) (0 :: nil)) as h8.
  rewrite h5 in h8.
  inversion h8; subst.
  rewrite <- list_to_set_in_iff in H.
  apply nin_famlS in H.
  contradiction.
  left; auto.
  inversion H.
  rewrite lS_eq_nil_iff in H1.
  pose proof (in_lind_linds_occ h1 l a h3) as h9.
  rewrite H1 in h9.
  contradiction.

  erewrite list_to_set_sort_list in h5.
  symmetry in h5.
  erewrite list_to_set_sort_list in h5.
  pose proof (sorted_linds_cons_in h2 _ _ h4) as h6; simpl in h6.
  pose proof (sorted_linds_cons_nin h1 _ _ h3) as h7; simpl in h7.
  unfold sorted_linds in h6, h7. 
  rewrite h6, h7 in h5 at 1.
  simpl in h5.
  pose proof (Add_intro2 _ (list_to_set
            (famlS
               (sort_list lt_list_nat
                  (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec)
                  (linds h1 l)))) (0 :: nil)) as h8.
  rewrite <- h5 in h8.
  inversion h8; subst.
  rewrite <- list_to_set_in_iff in H.
  apply nin_famlS in H.
  contradiction.
  left; auto.
  inversion H.
  rewrite lS_eq_nil_iff in H1.
  pose proof (in_lind_linds_occ h2 l' a' h4) as h9.
  rewrite H1 in h9.
  contradiction.

  
  erewrite list_to_set_sort_list in h5.
  symmetry in h5.
  erewrite list_to_set_sort_list in h5.
  pose proof (sorted_linds_cons_nin h2 _ _ h4) as h6; simpl in h6.
  pose proof (sorted_linds_cons_nin h1 _ _ h3) as h7; simpl in h7.
  unfold sorted_linds in h6, h7. 
  rewrite h6, h7 in h5 at 1.
  simpl in h5.
  apply add_nin_inj_eq in h5.
  unfold famlS in h5.
  do 2 rewrite map_im_compat in h5.
  do 2 rewrite <- list_to_set_sort_list in h5.
  apply im_inj_inj in h5; auto.
  apply inj_lS.

  intro h8.
  rewrite <- list_to_set_in_iff in h8.
  apply nin_famlS in h8; auto.
  left; auto.

  intro h8.
  rewrite <- list_to_set_in_iff in h8.
  apply nin_famlS in h8; auto.
  left; auto.
Qed.


Lemma linds_decompose_remove :
  forall {T:Type} (pfdt:type_eq_dec T) (l:list T) (a:T) (pf:In a l),
    sorted_linds pfdt l =
    sort_list _ lt_list_nat_tso_dec ((linds_occ pfdt l a) :: (linds_remove pfdt l a)).
  intros T h1 l a h2.
  unfold sorted_linds.
  eapply list_to_set_injable_increasingR.
  apply lt_list_nat_sort_dec.
  apply weak_incrR_no_dup_impl.
  apply weak_incrR_sort_list.
  apply no_dup_sort_list.
  apply no_dup_linds.
  apply weak_incrR_no_dup_impl.
  rewrite lt_list_nat_tso_dec_eq.
  apply weak_incrR_sort_list.
  rewrite lt_list_nat_tso_dec_eq.
  apply no_dup_sort_list.
  constructor.
  apply nin_linds_remove; auto.
  unfold linds_remove.
  apply no_dup_remove.
  apply no_dup_linds.
  rewrite <- list_to_set_sort_list.
  rewrite lt_list_nat_tso_dec_eq.
  rewrite <- list_to_set_sort_list.
  apply Extensionality_Ensembles; red; split; red; intros x h3; rewrite <- list_to_set_in_iff in h3; rewrite <- list_to_set_in_iff.
  rewrite in_linds_iff in h3.
  destruct h3 as [x' [h3 h4]].
  destruct h3 as [h3]; subst.
  unfold linds_remove.
  destruct (h1 x' a) as [|h5]; subst.
  left; auto.
  right.
  rewrite in_remove_iff.
  split.
  apply in_linds_occ_linds; auto.
  contradict h5.
  apply inj_linds_occ in h5; auto.

  destruct h3 as [|h3]; subst.
  apply in_linds_occ_linds; auto.
  unfold linds_remove in h3.
  rewrite in_remove_iff in h3.
  destruct h3; auto.
Qed.


Lemma partition_sorted_linds :
  forall {T:Type} (pfdt:type_eq_dec T) 
         (l:list T),
    partition_faml (sorted_linds pfdt l) (seg_list (length l)).
  intros T h1 l.
  pose proof (partition_linds h1 l) as h2.
  red in h2; red.
  unfold sorted_linds.
  pose proof (faml_to_fam_sort_list _ sort_dec_lt_nat (linds h1 l)) as h3.
  simpl in h3.
  erewrite sort_list_functional in h3.
  rewrite h3; auto.
  reflexivity.
Qed.




Lemma linds_occ_eq_l_to_inds_aux :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (a:T) (a':U)
         (pf: In a  l)
         (pf':In a' l'),
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l') ->
    linds_occ pfdt l  a =
    linds_occ pfdu l' a' ->
    l_to_inds_aux pfdt l  a  pf =
    l_to_inds_aux pfdu l' a' pf'.
  intros T U h1 h2 l.
  induction l as [|a l ih]; simpl.
  intros; contradiction.
  intros l'.
  destruct l' as [|a' l']; simpl.
  intros; contradiction.
 
  intros x x' h3 h4 hls h5.
  destruct (in_dec h1 a l) as [hi | hi]. 
  destruct (in_dec h2 a' l') as [hi' | hi'].

  
  destruct h3 as [|h3]; subst.
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.

  rewrite (l_to_inds_aux_cons_in _ _ _ hi).
  destruct h4 as [|h4]; subst.
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.
  inversion h5.
  apply inj_lS in H0.  
  rewrite (l_to_inds_aux_cons_in _ _ _ hi').
  apply ih; auto.
  eapply list_to_set_linds_from_cons.
  apply hls.
  
  destruct (h2 x' a') as [| h7]; subst. 
  rewrite linds_occ_cons_eq in h5. 
  simpl in h5.
  inversion h5.
  apply inj_lS in H0.
  rewrite (l_to_inds_aux_cons_in _ _ _ h4).
  apply ih; auto.

  eapply list_to_set_linds_from_cons.
  apply hls.
    
  rewrite linds_occ_cons_neq in h5; auto. 
  pose proof (in_eq 0%nat (lS (linds_occ h1 l x))) as h8.
  simpl in h5.
  rewrite h5 in h8.
  rewrite in_lS_iff in h8.
  destruct h8 as [? [? ?]]; discriminate.
  destruct h4 as [|h4]; subst.
  simpl in h5.
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.
  rewrite (l_to_inds_aux_cons_in _ _ _ hi').
  destruct (h1 x a) as [|he]; subst.
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.
  inversion h5.
  apply inj_lS in H0.
  rewrite (l_to_inds_aux_cons_in _ _ _ hi).
  apply ih; auto.

  eapply list_to_set_linds_from_cons.
  apply hls.
       
  rewrite linds_occ_cons_neq in h5; auto.
  pose proof (in_eq 0 (lS (linds_occ h2 l' x'))) as h6.
  rewrite <- h5 in h6.
  rewrite in_lS_iff in h6.
  destruct h6 as [? [? ?]]; discriminate.

   
  destruct (h1 x a) as [|h6]; subst.
  rewrite (l_to_inds_aux_cons_in _ _ _ h3).
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.
  destruct (h2 x' a') as [|ha]; subst.
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.
  inversion h5.
  apply inj_lS in H0.
  rewrite (l_to_inds_aux_cons_in h2 _ _ h4).
  apply ih; auto.
  
  eapply list_to_set_linds_from_cons.
  apply hls.
  
 
  rewrite linds_occ_cons_neq in h5.
  pose proof (in_eq 0 (lS (linds_occ h1 l a))) as h6.
  rewrite h5 in h6.
  rewrite in_lS_iff in h6.
  destruct h6 as [? [? ?]]; discriminate.
  assumption.
  rewrite linds_occ_cons_neq in h5; auto.
  destruct (h2 x' a') as [|he]; subst.
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.
  pose proof (in_eq 0 (lS (linds_occ h2 l' a'))) as h7.
  rewrite <- h5 in h7.
  rewrite in_lS_iff in h7.
  destruct h7 as [? [? ?]]; discriminate.
  rewrite linds_occ_cons_neq in h5; auto.
  apply inj_lS in h5.
  erewrite l_to_inds_aux_cons_in_in; auto.
  erewrite l_to_inds_aux_cons_in_in; auto.
  
  specialize (ih l' x x' h3 h4).
  hfirst ih.

    
  eapply list_to_set_linds_from_cons.
  apply hls.
  
  erewrite l_to_inds_aux_functional3.
  rewrite ih at 1.
  apply l_to_inds_aux_functional3.
  assumption.
  assumption. 
  pose proof (in_linds_occ_linds h1 (a::l) x h3) as h6.
  rewrite list_to_set_in_iff in h6.
  rewrite hls in h6.
  erewrite list_to_set_sort_list in h6.
  pose proof (sorted_linds_cons_nin h2 _ _ hi') as h7.
  unfold sorted_linds in h7.
  rewrite h7 in h6 at 1. 
  rewrite <- list_to_set_in_iff in h6. 
  destruct h3 as [|h3]; subst.
  rewrite linds_occ_cons_eq in h6.
  simpl in h6.
  destruct h6 as [h6 | h6].
  inversion h6.
  symmetry in H0.
  rewrite lS_eq_nil_iff in H0.
  pose proof (in_lind_linds_occ h1 _ _ hi) as h8.
  apply in_not_nil in h8.
  contradiction. 
  apply nin_famlS in h6.
  contradiction.
  left; auto.
  
  destruct h4 as [|h4]; subst.
  destruct (h1 x a) as [|h8]; subst.
  rewrite linds_occ_cons_eq in h6.
  simpl in h6.
  destruct h6 as [h6 | h6].
  inversion h6.
  symmetry in H0.
  rewrite lS_eq_nil_iff in H0.
  pose proof (in_lind_linds_occ h1 _ _ h3) as h8.
  apply in_not_nil in h8; contradiction.
  rewrite linds_occ_cons_eq in h5.
  rewrite linds_occ_cons_eq in h5.
  simpl in h5.
  inversion h5.
  apply inj_lS in H0.
  pose proof (in_lind_linds_occ h1 l a h3) as h8.
  rewrite H0 in h8.
  rewrite in_linds_occ_iff in h8.
  destruct h8 as [? [? ?]]; contradiction.
  rewrite linds_occ_cons_neq in h6; auto.
  destruct h6 as [h6 | h6].
  pose proof (in_eq 0 nil) as h9.
  rewrite h6 in h9.
  apply nin0_lS in h9.
  contradiction.
  rewrite in_famlS_iff in h6.
  destruct h6 as [ln' [h6 h9]].
  apply inj_lS in h9.
  subst.
  rewrite linds_occ_cons_eq in h5.
  rewrite linds_occ_cons_neq in h5; auto.
  simpl in h5.
  pose proof (in_eq 0 (lS (linds_occ h2 l' x'))) as h9.
  rewrite <- h5 in h9.
  apply nin0_lS in h9.
  contradiction.

  assert (h8:x' <> a').
    intro; subst; contradiction.
  destruct (h1 x a) as [|h9]; subst.
  rewrite linds_occ_cons_eq in h6.
  simpl in h6.
  destruct h6 as [h6 | h6].
  inversion h6.
  symmetry in H0.
  rewrite lS_eq_nil_iff in H0.
  pose proof (in_lind_linds_occ h1 _ _ h3) as h9.
  rewrite H0 in h9; contradiction.

  apply nin_famlS in h6.
  contradiction.
  left; auto.
  rewrite linds_occ_cons_neq in h6; auto.
  simpl in h6.
  destruct h6 as [h6 | h6].
  pose proof (in_eq 0 nil) as h10.
  rewrite h6 in h10.
  apply nin0_lS in h10.
  contradiction.
  rewrite in_famlS_iff in h6.
  destruct h6 as [ln [h6 h10]].
  apply inj_lS in h10; subst.
  rewrite linds_occ_cons_neq in h5; auto.
  rewrite linds_occ_cons_neq in h5; auto.
  apply inj_lS in h5.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h9); auto.
  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h8); auto.
  pose proof hls as h10.
  erewrite list_to_set_sort_list in h10.
  symmetry in h10.
  erewrite list_to_set_sort_list in h10.
  pose proof (sorted_linds_cons_in h1 l a hi) as h11.
  simpl in h11.
  unfold sorted_linds in h11.
  rewrite h11 in h10 at 1.
  pose proof (sorted_linds_cons_nin h2 l' a' hi') as h12.
  unfold sorted_linds in h12.
  rewrite h12 in h10 at 1.
  simpl in h10.
  pose proof (Add_intro2 _  (list_to_set
             (famlS
                (sort_list lt_list_nat
                   (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec)
                   (linds h2 l')))) (0 :: nil)) as h13.
  rewrite h10 in h13 at 1.
  inversion h13; subst.
  rewrite <- list_to_set_in_iff in H.
  apply nin_famlS in H.
  contradiction.
  left; auto.
  inversion H.
  rewrite lS_eq_nil_iff in H1.
  pose proof (in_lind_linds_occ h1 l a hi) as h14.
  rewrite H1 in h14.
  contradiction.

  destruct (in_dec h2 a' l') as [h6 | h6].
  pose proof hls as hls'.
  erewrite list_to_set_sort_list in hls'.
  symmetry in hls'.
  erewrite list_to_set_sort_list in hls'.
  symmetry in hls'.
  pose proof (sorted_linds_cons_nin h1 _ _ hi) as h7.
  simpl in h7.
  pose proof (sorted_linds_cons_in h2 _ _ h6) as h8.
  simpl in h8.
  unfold sorted_linds in h7, h8.
  rewrite h7, h8 in hls' at 1.
  simpl in hls'.
  pose proof (Add_intro2 _  (list_to_set
              (famlS
                 (sort_list lt_list_nat
                    (sort_dec_tso_dec lt_list_nat lt_list_nat_sort_dec)
                    (linds h1 l)))) (0 :: nil)) as h9.
  rewrite hls' in h9.
  inversion h9; subst.
  rewrite <- list_to_set_in_iff in H.
  apply nin_famlS in H.
  contradiction.
  left; auto.
  inversion H.
  rewrite lS_eq_nil_iff in H1.
  pose proof (in_lind_linds_occ h2 _ _ h6) as h10.
  rewrite H1 in h10.
  contradiction.

  pose proof hls as hls'.
  erewrite list_to_set_sort_list in hls'.
  symmetry in hls'.
  erewrite list_to_set_sort_list in hls'.
  symmetry in hls'. 
  pose proof (sorted_linds_cons_nin h1 _ _ hi) as h7.
  simpl in h7.
  pose proof (sorted_linds_cons_nin h2 _ _ h6) as h8.
  simpl in h8.
  unfold sorted_linds in h7, h8.
  rewrite h7, h8 in hls' at 1.
  simpl in hls'.

  apply add_nin_inj_eq in hls'.
  unfold famlS in hls'.
  do 2 rewrite map_im_compat in hls'.
  do 2 rewrite <- list_to_set_sort_list in hls'.
  apply im_inj_inj in hls'.
  
  destruct h3 as [|h3]; subst.
  rewrite linds_occ_cons_eq in h5.
  destruct h4 as [|h4]; subst.
  rewrite linds_occ_cons_eq in h5; auto.
  simpl in h5.
  inversion h5.
  apply inj_lS in H0.
  rewrite (l_to_inds_aux_cons_nin _ _ _ _  hi); auto.
  rewrite (l_to_inds_aux_cons_nin _ _ _ _  h6); auto.
  assert (h9:x' <> a').
    intro; subst; contradiction.

  rewrite (linds_occ_cons_neq h2 _ _ _ h9) in h5.
  simpl in h5.
  pose proof (in_eq 0 (lS (linds_occ h1 l x))) as h10.
  rewrite h5 in h10.
  rewrite in_lS_iff in h10.
  destruct h10 as [? [? ?]]; discriminate.

  destruct h4 as [|h4]; subst.
  assert (h9:x <> a).
    intro; subst; contradiction.
  rewrite linds_occ_cons_eq in h5.
  rewrite linds_occ_cons_neq in h5; auto.
  simpl in h5.
  pose proof (in_eq 0 (lS (linds_occ h2 l' x'))) as h10.
  rewrite <- h5 in h10.
  apply nin0_lS in h10; contradiction.

  assert (h9:x <> a).
    intro; subst; contradiction.

  assert (h10:x' <> a').
    intro; subst; contradiction.

  rewrite (l_to_inds_aux_cons_nin_in h1 _ _ _ h9); auto.
  rewrite (l_to_inds_aux_cons_nin_in h2 _ _ _ h10); auto.
 
  f_equal.
  apply ih; auto.
  rewrite linds_occ_cons_neq in h5; auto.
  rewrite linds_occ_cons_neq in h5; auto.
  apply inj_lS in h5; auto.
  apply inj_lS.

  intro h9.
  rewrite <- list_to_set_in_iff in h9.
  apply nin_famlS in h9; auto.
  left; auto.

  intro h9.
  rewrite <- list_to_set_in_iff in h9.
  apply nin_famlS in h9; auto.
  left; auto.

  Grab Existential Variables.
  assumption.
  assumption.
Qed.


Lemma list_to_set_linds_eq_l_to_inds : 
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T)
         (l':list U),
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l') ->
    l_to_inds pfdt l = l_to_inds pfdu l'.
  intros T U h1 h2 l.
  induction l as [|a l ih]. 
  intros l' h3.
  simpl in h3.
  symmetry in h3.
  apply empty_set_nil in h3.
  unfold l_to_inds; simpl.
  symmetry.
  rewrite linds_eq_nil_iff in h3; subst; auto.
  intros l'.
  destruct l' as [|a' l'].
  intro h3.
  simpl in h3.
    
  rewrite (list_to_set_sort_list _ lt_list_nat_sort_dec) in h3.
  destruct (in_dec h1 a l) as [h4 | h4].
  pose proof (sorted_linds_cons_in h1 l a h4) as h5.
  simpl in h5.
  unfold sorted_linds in h5.
  rewrite h5 in h3.
  simpl in h3.
  apply add_not_empty in h3; contradiction.
  pose proof (sorted_linds_cons_nin h1 l a h4) as h5.
  unfold sorted_linds in h5.
  rewrite h5 in h3.
  simpl in h3. 
  apply add_not_empty in h3; contradiction.
     
  intro h3.
  pose proof (list_to_set_linds_from_cons h1 h2 l l' a a' h3) as h4.
  pose proof (list_to_set_linds_occ_eq h1 h2 l l' a a' h3) as he.
  
  destruct (in_dec h1 a l) as [h5 | h5].
  rewrite (l_to_inds_cons_in h1 _ _ h5).
  destruct (in_dec h2 a' l') as [h6 | h6].
  rewrite (l_to_inds_cons_in h2 _ _ h6).
  pose proof (linds_occ_eq_l_to_inds_aux h1 h2 l l' a a' h5 h6 h4 he) as hl.
  rewrite hl.
  f_equal; auto.

  pose proof (in_lind_linds_occ h1 l a h5) as h7.
  rewrite he in h7.
  rewrite in_linds_occ_iff in h7.
  destruct h7 as [? [? ?]].
  contradiction.

  destruct (in_dec h2 a' l') as [h6 | h6].

  pose proof (in_lind_linds_occ h2 _ _ h6) as h7.
  rewrite <- he in h7.
  rewrite in_linds_occ_iff in h7.
  destruct h7 as [? [? ?]].
  contradiction.

  rewrite l_to_inds_cons_nin; auto.
  rewrite l_to_inds_cons_nin; auto.
  do 2 f_equal; auto.
Qed.


Lemma list_to_set_eq_linds :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l') ->
    linds pfdt l =
    linds pfdu l'.
  intros T U h1 h2  l l' h3.
  pose proof (list_to_set_linds_eq_l_to_inds h1 h2 _ _ h3) as h4.
  rewrite <- (@linds_l_to_inds T).
  rewrite <- (@linds_l_to_inds U).
  rewrite h4; auto.
Qed.

 
Lemma sorted_linds_eq_iff :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U),
    sorted_linds pfdt l =
    sorted_linds pfdu l' <->
    linds pfdt l =
    linds pfdu l'.
  intros T U h1 h2 l l'.
  unfold sorted_linds.
  pose proof (list_to_sets_eq_iff
                lt_list_nat
                lt_list_nat_sort_dec
                (linds h1 l)
                (linds h2 l')
                (no_dup_linds _ _)
                (no_dup_linds _ _)) as h3.
  simpl in h3.
  rewrite <- h3.
  split; intro h4.
  apply list_to_set_eq_linds in h4; auto.
  congruence.
Qed.


Lemma linds_eq_lecs :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U) (m:nat) (pfm:m < length l)
         (pf:linds pfdt l = linds pfdu l'),
    let pfl := linds_eq_length pfdt pfdu l l' pf in
    let pfm' := lt_eq_trans _ _ _ pfm pfl in
    lec pfdt pfm = lec pfdu pfm'.
  intros T U hdt hdu l l' m hm h1 he hm'.
  pose proof (in_lec hdt l m hm) as h2.
  rewrite h1 in h2.
  pose proof (in_lec hdu l' m hm') as h3.
  pose proof (in_linds_list_to_set_injable hdu l') as h4.
  red in h4.
  apply h4; auto.
  eapply partition_faml_eq.
  apply (partition_linds hdu).
  apply h2. apply h3.
  apply rep_in_lec.
  apply rep_in_lec.
Qed.

Lemma linds_eq_counts :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U) (i:nat)
         (pfi: i < length l)
         (pf:linds pfdt l = linds pfdu l'),
    let pfi' := lt_eq_trans _ _ _ pfi (linds_eq_length pfdt pfdu _ _ pf) in
    count_occ pfdt l  (nth_lt l i pfi) =
    count_occ pfdu l' (nth_lt l' i pfi').
  intros T U hdt hdu l l' i h1 h2 h3.
  erewrite <- length_linds_occ.
  erewrite <- length_linds_occ.
  do 2 rewrite linds_occ_nth_lt.
  pose proof (linds_eq_lecs hdt hdu _ _ i h1 h2) as h4.
  simpl in h4.
  rewrite h4.
  repeat f_equal.
Qed.



  
Lemma list_to_set_linds_eq_l_to_inds_aux :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (i:nat)
         (pflt:i < length l)
         (pflt':i < length l'),
    list_to_set (linds pfdt l) =
    list_to_set (linds pfdu l') ->
    l_to_inds_aux pfdt l  (nth_lt l  i pflt) (in_nth_lt _ _ _) =
    l_to_inds_aux pfdu l' (nth_lt l' i pflt') (in_nth_lt _ _ _).
  intros T U hdt hdu l.
  induction l as [|a l ih].
  simpl; intros; omega.
  intro l'.
  destruct l' as [|a' l'].
  intros i h1 h2 h3.
  simpl in h2.
  omega.
  intros i h1 h2 h3.
  simpl in h2. 
  pose proof h3 as h3'.
  apply (list_to_set_eq_linds hdt hdu (a::l) (a'::l')) in h3.

  rewrite <- sorted_linds_eq_iff in h3.
  simpl in h1. 
  destruct (in_dec hdt a l) as [h4 | h4].
  destruct (in_dec hdu a' l') as [h5 | h5].
  destruct (hdt a  (nth_lt (a ::l) i h1)) as [h6 | h6],
                                             (hdu a' (nth_lt (a'::l') i h2)) as [h7 | h7].
  
  gen0.
  generalize (in_nth_lt (a'::l')i h2).
  rewrite h6, h7.
  intros h8 h9.
  simpl. 
  destruct i as [|i].
  rewrite (l_to_inds_aux_cons_in _ _ _ h4); auto.
  rewrite (l_to_inds_aux_cons_in _ _ _ h5); auto.
  gen0.
  generalize h5.
  rewrite h6, h7.
  intros h10 h11. 
  simpl.
  pose proof (sorted_linds_cons_in hdt l a h4) as h12.
  simpl in h12.
  pose proof (sorted_linds_cons_in hdu l' a' h5) as h13.
  simpl in h13. 
  unfold sorted_linds in h12, h13.
  unfold sorted_linds in h3.
  rewrite h12, h13 in h3.
  inversion h3.
  apply inj_famlS in H1.
  apply inj_lS in H0.
  rewrite H0 in H1.
  eapply incrR_remove_inj in H1.
  pose proof (sorted_linds_eq_iff hdt hdu l l') as h14.
  unfold sorted_linds in h14.
  rewrite h14 in H1.
  apply linds_occ_eq_l_to_inds_aux.
  rewrite H1; auto.
  assumption.
  apply lt_list_nat_sort_dec.
  rewrite in_sort_list_iff.
  rewrite <- H0.
  apply in_linds_occ_linds.
  assumption.

  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.
  apply incr_sorted_linds.
  apply incr_sorted_linds.
  simpl in h8, h9. 
  pose proof (@nth_indep _ l i a (nth i l a)) as h10.
  pose proof h1 as h1'. pose proof h2 as h2'.
  apply lt_S_n in h1'.
  apply lt_S_n in h2'.
  specialize (h10 h1').
  gen0.
  rewrite <- h10.
  intro.
  pose proof (@nth_indep _ l' i a' (nth i l' a')) as h10'.
  specialize (h10' h2').
  symmetry.
  gen0.
  rewrite <- h10'.
  intro h11.
  rewrite (l_to_inds_aux_cons_in _ _ _ (nth_In l' a' h2')).
  rewrite (l_to_inds_aux_cons_in _ _ _ (nth_In l a h1')).
  symmetry.

  rewrite sorted_linds_cons_in in h3; auto.
  rewrite sorted_linds_cons_in in h3; auto.
  inversion h3.
  apply inj_famlS in H1.
  apply inj_lS in H0.
  rewrite H0 in H1.
  eapply incrR_remove_inj in H1.
  rewrite sorted_linds_eq_iff in H1.
  specialize (ih l' i h1' h2').
  hfirst ih.
  rewrite H1; auto.
  specialize (ih hf).
  revert ih.
  intro.
  gen0.
  erewrite nth_indep.
  intro h12.
  symmetry.
  gen0.
  erewrite nth_indep.
  intro h13.
  symmetry.
  erewrite f_equal_dep.
  rewrite ih at 1.
  apply f_equal_dep.
  rewrite nth_lt_compat.
  reflexivity.
  rewrite nth_lt_compat.
  reflexivity.
  assumption.
  assumption.
  apply lt_list_nat_sort_dec.
  rewrite <- H0.
  
  unfold sorted_linds.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.

  unfold sorted_linds.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.

  apply incr_sorted_linds.
  apply incr_sorted_linds.

  destruct i as [|i].
  simpl in h7.
  contradict h7; auto.
  simpl in h6, h7.
  simpl.
  gen0.
  generalize (in_nth_lt (a' :: l') (S i) h2).
  simpl.
  rewrite <- h6.
  intros h8 h9.
  rewrite (l_to_inds_aux_cons_in _ _ _ h4).
  apply neq_sym in h7.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h7).
  gen0.
  rewrite h6.
  intro h10.

  rewrite sorted_linds_cons_in in h3; auto.
  rewrite sorted_linds_cons_in in h3; auto.
  inversion h3.
  apply inj_famlS in H1.
  apply inj_lS in H0.
  rewrite H0 in H1.
  eapply incrR_remove_inj in H1.
  rewrite sorted_linds_eq_iff in H1.
  pose proof h1 as h1'.
  pose proof h2 as h2'.
  apply lt_S_n in h1'.
  apply lt_S_n in h2'.
  specialize (ih l' i h1' h2').
  hfirst ih.
  rewrite H1; auto.
  specialize (ih hf).
  revert ih.
  intro.
  gen0.
  erewrite nth_indep.
  intro h12.
  symmetry.
  gen0.
  erewrite nth_indep.
  intro h13.
  symmetry.
  erewrite f_equal_dep.
  rewrite ih at 1.
  apply f_equal_dep.
  rewrite nth_lt_compat.
  reflexivity.
  rewrite nth_lt_compat.
  reflexivity.
  assumption.
  assumption.
  apply lt_list_nat_sort_dec.
  rewrite <- H0.
  
  
  unfold sorted_linds.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.

  unfold sorted_linds.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.

  apply incr_sorted_linds.
  apply incr_sorted_linds.

  assumption.

  
  destruct i as [|i].
  simpl in h6.
  contradict h6; auto.
  simpl in h6, h7.
  simpl.
  symmetry.
  gen0. 
  generalize (in_nth_lt (a :: l) (S i) h1).
  simpl.
  rewrite <- h7.
  intros h8 h9.
  rewrite  (l_to_inds_aux_cons_in _ _ _ h5).
  apply neq_sym in h6.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h6).
  gen0.
  rewrite h7.
  intro h10.

  rewrite sorted_linds_cons_in in h3; auto.
  rewrite sorted_linds_cons_in in h3; auto.
  inversion h3.
  apply inj_famlS in H1.
  apply inj_lS in H0.
  rewrite H0 in H1.
  eapply incrR_remove_inj in H1.
  rewrite sorted_linds_eq_iff in H1.
  pose proof h1 as h1'.
  pose proof h2 as h2'.
  apply lt_S_n in h1'.
  apply lt_S_n in h2'.
  specialize (ih l' i h1' h2').
  hfirst ih.
  rewrite H1; auto.
  specialize (ih hf).
  revert ih.
  intro.
  gen0.
  erewrite nth_indep.
  intro h12.
  symmetry.
  gen0.
  erewrite nth_indep.
  intro h13.
  erewrite f_equal_dep.
  rewrite ih at 1.
  apply f_equal_dep.
  rewrite nth_lt_compat.
  reflexivity.
  rewrite nth_lt_compat.
  reflexivity.
  assumption.
  assumption.
  apply lt_list_nat_sort_dec.
  rewrite <- H0.
  
  
  unfold sorted_linds.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.

  unfold sorted_linds.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.

  apply incr_sorted_linds.
  apply incr_sorted_linds.

  assumption.

  destruct i as [|i].
  simpl in h6.
  contradict h6; auto.
  simpl.
  apply neq_sym in h6.
  apply neq_sym in h7.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h6); auto.
  rewrite (l_to_inds_aux_cons_in_in _ _ _ _ h7); auto.
  simpl.
  pose proof h1 as h1'.
  pose proof h2 as h2'.
  apply lt_S_n in h1'.
  apply lt_S_n in h2'.
  specialize (ih l' i h1' h2').
  rewrite sorted_linds_cons_in in h3; auto.
  rewrite sorted_linds_cons_in in h3; auto.
  inversion h3.
  apply inj_lS in H0.
  rewrite H0 in H1.
  apply inj_famlS in H1.
  eapply incrR_remove_inj in H1.
  rewrite sorted_linds_eq_iff in H1.
  hfirst ih.
  rewrite H1; auto.
  specialize (ih hf).
  revert ih.
  Gen0.  
  rewrite nth_lt_compat. 
  intro h8.
  intro h9.
  symmetry in h9.
  revert h9.
  Gen0.
  rewrite nth_lt_compat.
  intros h10 h11.
  erewrite f_equal_dep.
  rewrite <- h11 at 1.
  apply f_equal_dep.
  apply nth_indep.
  assumption.
  apply nth_indep.
  assumption.
  apply lt_list_nat_sort_dec.
  unfold sorted_linds.
  rewrite in_sort_list_iff.
  rewrite <- H0.
  apply in_linds_occ_linds.
  assumption.

  unfold sorted_linds.
  rewrite in_sort_list_iff.
  apply in_linds_occ_linds.
  assumption.
 
  apply incr_sorted_linds.
  apply incr_sorted_linds.

  rewrite sorted_linds_cons_in in h3.
  rewrite sorted_linds_cons_nin in h3.
  inversion h3.
  rewrite lS_eq_nil_iff in H0.
  pose proof (in_lind_linds_occ hdt _ _ h4) as h6.
  rewrite H0 in h6.
  contradiction.
  assumption.
  assumption.
 
  destruct (in_dec hdu a' l') as [h5 | h5]. 
  rewrite sorted_linds_cons_nin in h3; auto.
  rewrite sorted_linds_cons_in in h3; auto.
  inversion h3.
  apply inj_famlS in H1.
  symmetry in H0.
  rewrite lS_eq_nil_iff in H0.
  pose proof (in_lind_linds_occ hdu l' a' h5) as h6.
  rewrite H0 in h6.
  contradiction.

  simpl.
  destruct i as [|i].
  rewrite l_to_inds_aux_cons_nin; auto.
  rewrite l_to_inds_aux_cons_nin; auto.

  assert (h6:a <> nth i l a).
    intro h6.
    rewrite h6 in h4.
    contradict h4.
    apply nth_In; auto with arith.

  assert (h7:a' <> nth i l' a').
    intro h7.
    rewrite h7 in h5.
    contradict h5.
    apply nth_In; auto with arith.

  apply neq_sym in h6.
  apply neq_sym in h7.
    
  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h6 h4).
  rewrite (l_to_inds_aux_cons_nin_in _ _ _ _ h7 h5).
  f_equal.
  rewrite sorted_linds_cons_nin in h3; auto.
  rewrite sorted_linds_cons_nin in h3; auto.
  inversion h3.
  apply inj_famlS in H0.
  rewrite sorted_linds_eq_iff in H0.
  pose proof h1 as h1'.
  pose proof h2 as h2'.
  apply lt_S_n in h1'.
  apply lt_S_n in h2'.
    
  specialize (ih l' i h1' h2').
  hfirst ih.
  rewrite H0; auto.
  specialize (ih hf).
  revert ih.
  Gen0.
  rewrite nth_lt_compat.
  intros h8 h9.
  symmetry in h9.
  revert h9.
  Gen0.
  rewrite nth_lt_compat.
  intros h9 h10.
  erewrite f_equal_dep.
  rewrite <- h10 at 1.
  apply f_equal_dep.
  apply nth_indep.
  assumption.
  apply nth_indep.
  assumption.
Qed.

Lemma linds_eq_l_to_inds_aux :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (i:nat)
         (pflt:i < length l)
         (pflt':i < length l'),
    linds pfdt l =
    linds pfdu l' ->
    l_to_inds_aux pfdt l  (nth_lt l  i pflt) (in_nth_lt _ _ _) =
    l_to_inds_aux pfdu l' (nth_lt l' i pflt') (in_nth_lt _ _ _).
  intros; apply list_to_set_linds_eq_l_to_inds_aux.
  rewrite H; auto.
Qed.
 
   
Lemma list_to_set_linds_map_seg_eq_l_to_inds_aux :
  forall {U V:Type} (hdu:type_eq_dec U) (hdv:type_eq_dec V)
         (f:nat->U) (g:nat->V) (i n:nat)
         (pflt:i < n),
    list_to_set (linds hdu (map_seg (fun_to_seg f n))) =
    list_to_set (linds hdv (map_seg (fun_to_seg g n))) ->
      forall (pfi:In (fun_to_seg f n i pflt) (map_seg (fun_to_seg f n)))
             (pfi':In (fun_to_seg g n i pflt) (map_seg (fun_to_seg g n))),
        l_to_inds_aux hdu (map_seg (fun_to_seg f n)) (f i) pfi  =
        l_to_inds_aux hdv (map_seg (fun_to_seg g n)) (g i) pfi'.
  intros.
  pose proof (list_to_set_linds_eq_l_to_inds_aux
                hdu hdv
                (map_seg (fun_to_seg f n))
                (map_seg (fun_to_seg g n)) i) as h1.
  hfirst h1.
  rewrite length_map_seg; auto.
  specialize (h1 hf).
  hfirst h1.
  rewrite length_map_seg; auto.
  specialize (h1 hf0 H).
  erewrite f_equal_dep.
  rewrite h1 at 1.
  apply f_equal_dep.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  reflexivity.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  reflexivity.

  Grab Existential Variables.
  assumption.
  assumption.
Qed.

   
Lemma linds_map_seg_eq_l_to_inds_aux :
  forall {U V:Type} (hdu:type_eq_dec U) (hdv:type_eq_dec V)
         (f:nat->U) (g:nat->V) (i n:nat)
         (pflt:i < n),
    linds hdu (map_seg (fun_to_seg f n)) =
    linds hdv (map_seg (fun_to_seg g n)) ->
      forall (pfi:In (fun_to_seg f n i pflt) (map_seg (fun_to_seg f n)))
             (pfi':In (fun_to_seg g n i pflt) (map_seg (fun_to_seg g n))),
        l_to_inds_aux hdu (map_seg (fun_to_seg f n)) (f i) pfi  =
        l_to_inds_aux hdv (map_seg (fun_to_seg g n)) (g i) pfi'.
  intros; apply list_to_set_linds_map_seg_eq_l_to_inds_aux; auto.
  rewrite H; auto.
Qed.


Lemma linds_eq_lind_occ :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (i occ:nat)
         (pfi: i < length l)
         (pfi':i < length l'),
    linds pfdt l =
    linds pfdu l' ->
    occ < count_occ pfdt l (nth_lt _ _ pfi) ->
      lind_occ pfdt l  occ (nth_lt _ _ pfi)  (in_nth_lt _ _ _) =
      lind_occ pfdu l' occ (nth_lt _ _ pfi') (in_nth_lt _ _ _).
  intros T U h1 h2 l l' i occ h3 h4 h5 h6.
  pose proof (linds_eq_l_to_inds_aux h1 h2 _ _ i h3 h4 h5) as h11.
  pose proof (lind_occ_l_to_inds h1 l occ (nth_lt l i h3) (in_nth_lt _ _ _)
                                 h6) as h8.
  pose proof h6 as h6'.
  pose proof (linds_eq_counts h1 h2 l l' i h3 h5) as h9.
  simpl in h9.
  rewrite h9 in h6'.
  erewrite nth_lt_functional3 in h6'.
  pose proof (lind_occ_l_to_inds h2 l' occ (nth_lt l' i h4) (in_nth_lt _ _ _)
                                 h6') as h10.
  rewrite h8, h10.
  gen0.
  rewrite h11.
  intro h12.
  apply lind_occ_functional.
  apply list_to_set_linds_eq_l_to_inds.
  rewrite h5; auto.
  reflexivity.
  apply list_to_set_linds_eq_l_to_inds_aux.
  reflexivity.
Qed.


Lemma linds_map_seg_eq_lind_occ :
  forall {U V:Type} (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (f:nat->U) (g:nat->V) (i occ n:nat)
         (pflt:i < n),
    linds pfdu (map_seg (fun_to_seg f n)) =
    linds pfdv (map_seg (fun_to_seg g n)) ->
    occ < count_occ pfdu (map_seg (fun_to_seg f n)) (f i) ->
    forall (pfi:In (fun_to_seg f n i pflt) (map_seg (fun_to_seg f n)))
           (pfi':In (fun_to_seg g n i pflt) (map_seg (fun_to_seg g n))),
      lind_occ pfdu (map_seg (fun_to_seg f n)) occ (f i) pfi =
      lind_occ pfdv (map_seg (fun_to_seg g n)) occ (g i) pfi'.
  intros U V hdu hdv f g i occ n h1 h2 h3 h4 h5.
  pose proof
       (linds_eq_lind_occ
          hdu hdv
          (map_seg (fun_to_seg f n))
          (map_seg (fun_to_seg g n))
          i occ) as h6.
  hfirst h6.
  rewrite length_map_seg; auto.
  specialize (h6 hf).
  hfirst h6.
  rewrite length_map_seg; auto.
  specialize (h6 hf0 h2).
  hfirst h6.
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  rewrite fun_to_seg_compat; auto.
  specialize (h6 hf1).
  revert h6.
  Gen0. 
  generalize (in_nth_lt (map_seg (fun_to_seg g n)) i hf0). 
  erewrite nth_lt_functional3.
  rewrite nth_lt_map_seg.
  erewrite (nth_lt_functional3 (map_seg (fun_to_seg f n))).
  rewrite nth_lt_map_seg.
  do 2 rewrite fun_to_seg_compat.
  intros h6 h7 h8.
  pose proof (proof_irrelevance _ h7 h4). 
  pose proof (proof_irrelevance _ h5 h6).
  subst; auto.

  Grab Existential Variables.
  assumption.
  assumption.
  assumption.
Qed.



Lemma linds_eq_linds_occ :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U)
         (i:nat)
         (pfi: i < length l)
         (pfi':i < length l'),
    linds pfdt l =
    linds pfdu l' ->
    linds_occ pfdt l  (nth_lt _ _ pfi) =
    linds_occ pfdu l' (nth_lt _ _ pfi').
  intros T U h1 h2 l l' i h3 h4 h5.
  pose proof (@list_to_set_linds_eq_l_to_inds).
  pose proof (@linds_occ_l_to_inds).
  erewrite linds_occ_l_to_inds.
  symmetry.
  erewrite linds_occ_l_to_inds.
  erewrite list_to_set_linds_eq_l_to_inds.
  repeat f_equal.
  apply linds_eq_l_to_inds_aux; auto.
  rewrite h5; auto.
Qed.


Lemma linds_map_seg_eq_linds_occ :
  forall {U V:Type} (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (f:nat->U) (g:nat->V)
         (i n:nat)
         (pf:i < n),
    linds pfdu (map_seg (fun_to_seg f n)) =
    linds pfdv (map_seg (fun_to_seg g n)) ->
    linds_occ pfdu (map_seg (fun_to_seg f n)) (f i) =
    linds_occ pfdv (map_seg (fun_to_seg g n)) (g i).
  intros U V hdu hdv f g i n h1 h2.
  pose proof
       (linds_eq_linds_occ
          hdu hdv
          (map_seg (fun_to_seg f n))
          (map_seg (fun_to_seg g n)) i) as h3.
  hfirst h3.
  rewrite length_map_seg; auto.
  specialize (h3 hf).
  hfirst h3.
  rewrite length_map_seg; auto.
  specialize (h3 hf0 h2).
  erewrite nth_lt_functional3 in h3.
  rewrite nth_lt_map_seg in h3.
  erewrite nth_lt_functional3 in h3.
  rewrite nth_lt_map_seg in h3.
  do 2 rewrite fun_to_seg_compat in h3; auto.

  Grab Existential Variables.
  assumption.
  assumption.
Qed.
    

Lemma fun_linds_eq_nat_linds_occ :
  forall {U V:Type}
         (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (f:nat->U) (g:nat->V),
    fun_linds_eq_nat pfdu pfdv f g ->
    forall (i n:nat),
      i < n ->
      linds_occ pfdu (map_seg (fun_to_seg f n)) (f i) =
      linds_occ pfdv (map_seg (fun_to_seg g n)) (g i).
  intros U V hdu hdv f g h1 i n h2.
  red in h1.
  specialize (h1 n).
  apply linds_map_seg_eq_linds_occ; auto.
Qed.

Lemma fun_linds_eq_nat_lind_occ :
  forall {U V:Type}
         (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (f:nat->U) (g:nat->V),
    fun_linds_eq_nat pfdu pfdv f g ->
    forall (i n occ:nat) (pflt:i < n),
      occ < count_occ pfdu (map_seg (fun_to_seg f n)) (f i) ->
      forall (pfi:In (fun_to_seg f n i pflt) (map_seg (fun_to_seg f n)))
             (pfi':In (fun_to_seg g n i pflt) (map_seg (fun_to_seg g n))),
        lind_occ pfdu (map_seg (fun_to_seg f n)) occ (f i) pfi =
        lind_occ pfdv (map_seg (fun_to_seg g n)) occ (g i) pfi'.
  intros U V hdu hdv f g h1 i n occ h2 h3 h4 h5.
  red in h1.
  specialize (h1 n).
  apply linds_map_seg_eq_lind_occ; auto.
Qed.

  

Lemma linds_nth_lt_compat :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l' :list U)
         (m n:nat)
         (pf:linds pfdt l = linds pfdu l')
         (pfm:m < length l)
         (pfn:n < length l),
    let pfl := linds_eq_length pfdt pfdu l l' pf in
    let pfm' := lt_eq_trans _ _ _ pfm pfl in
    let pfn' := lt_eq_trans _ _ _ pfn pfl in
    nth_lt l m pfm  =
    nth_lt l n pfn -> 
    nth_lt l' m pfm' =
    nth_lt l' n pfn'.
  intros T U hdt hdu l l' m n h1 hm hn ha hm' hn' h2. 
  rewrite <- (lecs_eq_iff hdt) in h2.
  rewrite <- (lecs_eq_iff hdu).
  pose proof (in_lec hdt l m hm) as h3.
  pose proof (in_lec hdt l n hn) as h4. 
  pose proof (in_lec hdu l' m hm') as h3'.
  pose proof (in_lec hdu l' n hn') as h4'.  
  rewrite <- h1 in h3'.
  rewrite <- h1 in h4'.  
  rewrite in_linds_iff in h3', h4'. 
  destruct h3' as [m' h5], h4' as [n' h6].
  red in h5, h6.
  destruct h5 as [h5 h7], h6 as [h6 h8].
  destruct h5 as [h5 h9], h6 as [h6 h10]. 
  erewrite f_equal_dep; auto.
  rewrite <- (linds_eq_lecs hdt hdu l l') at 1; auto.
  symmetry.
  erewrite f_equal_dep; auto.
  rewrite <- (linds_eq_lecs hdt hdu l l') at 1.
  erewrite f_equal_dep.
  symmetry.
  erewrite f_equal_dep.
  apply h2.
  reflexivity.
  reflexivity.

  Grab Existential Variables.
  assumption.
  assumption.
  assumption.
  assumption.
Qed.


Lemma linds_nth_lt_compat_iff :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l' :list U)
         (m n:nat)
         (pf:linds pfdt l = linds pfdu l')
         (pfm:m < length l)
         (pfn:n < length l),
    let pfl := linds_eq_length pfdt pfdu l l' pf in
    let pfm' := lt_eq_trans _ _ _ pfm pfl in
    let pfn' := lt_eq_trans _ _ _ pfn pfl in
    nth_lt l m pfm  =
    nth_lt l n pfn <-> 
    nth_lt l' m pfm' =
    nth_lt l' n pfn'.
  intros T U hdt hdu l l' m n h1 hm hn hl hm' hn'.
  split.
  apply linds_nth_lt_compat; auto.
  intro h2.
  erewrite nth_lt_functional3.
  rewrite (linds_nth_lt_compat hdu hdt l' l m n (eq_sym h1) hm' hn' h2) at 1.
  apply nth_lt_functional3.
Qed.


Lemma linds_eq_iff :
  forall {T U:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l' :list U),
    linds pfdt l = linds pfdu l' <->
    l_to_inds pfdt l = l_to_inds pfdu l'.
  intros T U hdt hdu l l'; rewrite <- sorted_linds_eq_iff.
  unfold sorted_linds.
  pose proof (list_to_sets_eq_iff _ lt_list_nat_sort_dec (linds hdt l) (linds hdu l') (no_dup_linds _ _) (no_dup_linds _ _)) as h1.
  rewrite <- h1.
  split; intro h2.
  apply list_to_set_linds_eq_l_to_inds in h2; auto.
  erewrite <- linds_l_to_inds.
  rewrite h2.
  symmetry.
  erewrite <- linds_l_to_inds.
  reflexivity.
Qed.



Lemma linds_eq_iff' :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (l:list T) (l':list U) (pfe:length l = length l'),
    linds pfdt l =
    linds pfdu l' <->
    (forall (m n:nat)
            (pfm:m < length l)
            (pfn:n < length l),
       let pfm' := lt_eq_trans _ _ _ pfm pfe in
       let pfn' := lt_eq_trans _ _ _ pfn pfe in
       nth_lt l m pfm =
       nth_lt l n pfn <->
       nth_lt l' m pfm' =
       nth_lt l' n pfn').
  intros T U hdt hdu l l' he; split; intro h1.
  intros m n h2 h3 h4 h5.
  split; intro h6.
  erewrite nth_lt_functional3.
  erewrite (nth_lt_functional3 l' n).
  eapply linds_nth_lt_compat.
  apply h6.
  
  erewrite nth_lt_functional3.
  erewrite (nth_lt_functional3 l n).
  eapply linds_nth_lt_compat. 
  apply h6.

  rewrite linds_eq_iff.
  eapply nth_lts_eq_l_to_inds.
  apply h1.

  Grab Existential Variables.
  symmetry.
  apply h1.
  apply h1.
Qed.


Lemma linds_eq_firstn :
  forall {T U:Type} {pfdt:type_eq_dec T} {pfdu:type_eq_dec U}
         (l:list T) (l':list U) (m:nat),
    linds pfdt l =
    linds pfdu l' ->
    linds pfdt (firstn m l) =
    linds pfdu (firstn m l').
  intros T U hdt hdu l l' m h1.
  pose proof (linds_eq_length _ _ _ _ h1) as he.
  rewrite linds_eq_iff'.
  rewrite linds_eq_iff' in h1.
  intros r s h2 h3 h4 h5.
  do 4 rewrite nth_lt_firstn_eq.
  erewrite nth_lt_functional3.
  erewrite (nth_lt_functional3 l s).
  erewrite (nth_lt_functional3 l' r).
  erewrite (nth_lt_functional3 l' s).
  apply h1.

  Grab Existential Variables.
  eapply lt_le_trans.
  apply h3.
  apply length_firstn_le.
  eapply lt_le_trans.
  apply h2.
  apply length_firstn_le.
  apply length_eq_firstn; auto.
  assumption.
Qed.
    
  

Lemma two_lists_to_fun_in_dep_compat :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U)
         (pfl:linds pfdt la =
              linds pfdu lb)
         (i:nat) (pfi:i < length la),
    let pfe := linds_eq_length pfdt pfdu la lb pfl in
    let pfi' := lt_eq_trans _ _ _ pfi pfe in
    two_lists_to_fun_in_dep pfdt pfdu _ _ pfl _ (in_nth_lt _ _ pfi) =
    nth_lt _ _ pfi'.
  simpl; unfold two_lists_to_fun_in_dep; intros T U h1 h2 la lb h3 i h4.
  pose proof (in_lind_linds_occ h1 _ _ (in_nth_lt la i h4)) as h6.
  pose proof (in_linds_occ_linds h1 _ _ (in_nth_lt la i h4)) as h7.
  rewrite h3 in h7.
  rewrite in_linds_iff in h7. 
  destruct h7 as [y h8].
  red in h8.
  destruct h8 as [h8 h9].
  destruct h8 as [h8 h10].
  rewrite h10 in h6.
  rewrite in_linds_occ_iff in h6. 
  destruct h6 as [m [h6 h7]].
  destruct h7 as [h7 h11]. 
  gen0.
  rewrite <- h11.
  intro h12.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat.  
  apply h9.
  split.
  apply in_nth_lt.
  
  pose proof (in_linds_occ_linds h1 la (nth_lt la i h4) (in_nth_lt _ _ _)) as h13.
  rewrite h3 in h13.
  rewrite in_linds_iff in h13.
  destruct h13 as [y' [h13 h14]].
  destruct h13 as [h13 h15].
  pose proof h15 as h15'.
  rewrite h10 in h15'.
  apply inj_linds_occ in h15'; auto; subst.
  do 2 rewrite linds_occ_nth_lt.
  apply linds_eq_lecs; auto.
Qed.


Lemma two_lists_to_sig_fun_compat :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (la:list T) (lb:list U)
         (pfl:linds pfdt la =
              linds pfdu lb)
         (i:nat) (pfi:i < length la),
    let pfe := linds_eq_length pfdt pfdu la lb pfl in
    let pfi' := lt_eq_trans _ _ _ pfi pfe in
    two_lists_to_sig_fun pfdt pfdu _ _ pfl (exist _ _ (iff1 (list_to_set_in_iff _ _) (in_nth_lt _ _ pfi))) =
    exist _ _ (iff1 (list_to_set_in_iff _ _) (in_nth_lt _ _ pfi')).
  simpl; unfold two_lists_to_sig_fun; intros T U h1 h2 la lb h3 i h4.
  apply proj1_sig_injective; simpl.
  pose proof (in_lind_linds_occ h1 _ _ (in_nth_lt la i h4)) as h6.
  pose proof (in_linds_occ_linds h1 _ _ (in_nth_lt la i h4)) as h7.
  rewrite h3 in h7.
  rewrite in_linds_iff in h7. 
  destruct h7 as [y h8].
  red in h8.
  destruct h8 as [h8 h9].
  destruct h8 as [h8 h10].
  rewrite h10 in h6.
  rewrite in_linds_occ_iff in h6. 
  destruct h6 as [m [h6 h7]].
  destruct h7 as [h7 h11]. 
  gen0.
  erewrite lind_functional3 in h11.
  rewrite <- h11.
  intro h12.
  erewrite nth_lt_functional3.
  rewrite lind_occ_compat.  
  apply h9.
  split.
  apply in_nth_lt.
  
  pose proof (in_linds_occ_linds h1 la (nth_lt la i h4) (in_nth_lt _ _ _)) as h13.
  rewrite h3 in h13.
  rewrite in_linds_iff in h13.
  destruct h13 as [y' [h13 h14]].
  destruct h13 as [h13 h15].
  pose proof h15 as h15'.
  rewrite h10 in h15'.
  apply inj_linds_occ in h15'; auto; subst.
  do 2 rewrite linds_occ_nth_lt.
  apply linds_eq_lecs; auto.
Qed.
  

   
Lemma faml_to_fam_list_minus_sort_family :
  forall (A B:Family nat)
         (pffa:Finite A) (pffma:finite_mem A)
         (pffb:Finite B) (pffmb:finite_mem B),
    faml_to_fam (list_minus list_nat_eq_dec
                            (sort_family A pffa pffma) 
                            (sort_family B pffb pffmb)) =
    Setminus A B.
  intros A B h1 h2 h3 h4.
  pose proof (increasing_mem_sort_family A h1 h2) as hia.
  pose proof (increasing_mem_sort_family B h3 h4) as hib.
  apply Extensionality_Ensembles; red; split; red; intros C h5. 
  pose proof (in_faml_to_fam_impl _ _ h5) as h6.
  destruct h6 as [l h6]; subst.
  destruct h6 as [h6 h6'].
  apply (in_list_minus_impl list_nat_eq_dec) in h6.
  destruct h6 as [h6 h10]. subst.
  pose proof (in_sort_family_impl' _ _ _ _ h6) as h7.
  destruct h7 as [S [h7 h8]]; subst.
  rewrite sort_set_nat_compat.
  constructor; auto.
  contradict h10. 
  eapply impl_in_sort_family; auto.

  destruct h5 as [h5 h6].
  pose proof (h2 _ h5) as h7.
  pose proof (sort_set_nat_compat C h7) as h8.
  rewrite <- h8.
  eapply in_faml_impl.
  apply impl_in_list_minus.
  eapply impl_in_sort_family; auto.
  contradict h6.
  apply in_faml_impl in h6.
  rewrite sort_family_compat' in h6.
  rewrite sort_set_nat_compat in h6.
  assumption.
Qed.


           

Lemma list_minus_sort_family :
  forall (A B:Family nat)
         (pffa:Finite A) (pffma:finite_mem A)
         (pffb:Finite B) (pffmb:finite_mem B) 
         (pfab:Finite (Setminus A B)) (pffmab:finite_mem (Setminus A B)),
    list_minus list_nat_eq_dec
               (sort_family A pffa pffma) 
               (sort_family B pffb pffmb) =
    sort_family (Setminus A B) pfab pffmab.
  intros A B h1 h1' h2 h2' h3 h3'.
  pose proof (faml_to_fam_list_minus_sort_family A B h1 h1' h2 h2') as he.
  gterml.
  pose proof (sort_family_functional _ _ (finite_faml_to_fam x) h3 (finite_mem_faml x) h3' he) as h4.
  rewrite <- h4.
  eapply incr_impl_is_sorted_family.
  unfold x.
  apply incr_faml_list_minus.
  apply incr_sort_family.
  unfold x.
  apply increasing_mem_list_minus.
  apply increasing_mem_sort_family.
Qed.

 
Lemma faml_to_fam_remove_can_faml_nat :
  forall (A:Family nat) (B:Ensemble nat)
         (pffa:Finite A) (pffma:finite_mem A)
         (pffb:Finite B) ,
    faml_to_fam (remove list_nat_eq_dec
                        (sort_set_nat B pffb) 
                        (sort_family A pffa pffma)) = 
    Subtract A B. 
  intros A B h1 h2 h3.
  pose proof (faml_to_fam_list_minus_sort_family A (Singleton B) h1 h2 (Singleton_is_finite _ _)) as h4.
  hfirst h4.
  do 2 red; intros X h5.  destruct h5; auto.
  specialize (h4 hf).
  pose proof (sort_family_sing B h3  (Singleton_is_finite (Ensemble nat) B) hf) as h5.
  unfold Subtract. 
  rewrite <- h4.
  f_equal.
  rewrite h5.
  rewrite list_minus_sing.
  reflexivity.
Qed.

 
Lemma remove_sort_family :
  forall (A:Family nat) (B:Ensemble nat)
         (pffa:Finite A) (pffma:finite_mem A)
         (pffb:Finite B) 
         (pfab:Finite (Subtract A B)) (pffmab:finite_mem (Subtract A B)),
    remove list_nat_eq_dec
           (sort_set_nat B pffb) 
           (sort_family A pffa pffma) = 
    sort_family (Subtract A B) pfab pffmab.
  intros A B h1 h1' h2 h2' h3.
  pose proof (faml_to_fam_remove_can_faml_nat A B h1 h1' h2) as he.
  gterml.
  pose proof (sort_family_functional _ _ (finite_faml_to_fam x) h2' (finite_mem_faml x) h3 he) as h4.
  rewrite <- h4.
  eapply incr_impl_is_sorted_family.
  unfold x.
  apply increasing_faml_remove.
  apply incr_sort_family.
  unfold x.
  apply increasing_mem_remove.
  apply increasing_mem_sort_family.
Qed.


End FamlNat.




Lemma card_in_l : 
  forall {T:Type} (l:list T),
    NoDup l ->
    FiniteT_card _ (list_to_set_finitet l) = length l.
intros T l h0.
induction l as [|a l h1].
simpl. 
rewrite <- FiniteT_card_False.
pose (fun x:{_:T | False} => False_rect False (proj2_sig x)) as f.
assert (h1:invertible f). apply bijective_impl_invertible.
  red; split; red; intros; contradiction.
pose proof (FiniteT_card_bijection _ _ (list_to_set_finitet nil) f h1) as h2.
simpl in h2.
rewrite <- h2.
pose proof (proof_irrelevance
              _
              (bij_finite {_ : T | False} False f (list_to_set_finitet nil) h1)
              empty_finite) as h3.
rewrite h3. reflexivity.
pose proof (no_dup_cons_nin _ _ h0) as h0'.
simpl. 
pose proof (FiniteT_card_option _ (list_to_set_finitet l)) as h3.
pose (fun x:(option {x : T | In x l}) =>
        let f := (fun x => a = x \/ In x l) in
        match x with
          | Some e => exist f (proj1_sig e) (or_intror (proj2_sig e)) 
          | None => exist f a (or_introl (eq_refl _))
        end) as g.
assert (h4:invertible g).
  apply bijective_impl_invertible.
  red; split; red.
  intros o p h4.
  destruct o; destruct p. simpl in h4.
  apply exist_injective in h4. f_equal.  apply proj1_sig_injective. assumption.
  simpl in h4. apply exist_injective in h4.
  contradict h0'.
  destruct s. simpl in h4. subst. assumption.
  simpl in h4. apply exist_injective in h4. contradict h0'.
  destruct s. simpl in h4. subst. assumption.
  reflexivity.
  intro y.
  destruct y as [y h4].
  destruct h4 as [h4l | h4r]. exists None.
  simpl. apply proj1_sig_injective. simpl. assumption.
  exists (Some (exist _ y h4r)). simpl. reflexivity.
rewrite <- (FiniteT_card_bijection _ _ (add_finite _ (list_to_set_finitet l)) g h4) in h3.
pose proof (no_dup_cons _ _ h0) as h5.
specialize (h1 h5).
rewrite h1 in h3.
rewrite <- h3.
f_equal. apply proof_irrelevance.
Qed.

 
Section Transfer.


Lemma fun_map_sig_functional :
  forall {T U:Type} (l l':list T) (f:list {x:T |In x l}->U)
         (pf:l = l'),
    let f' := (fun c:list {x:T |In x l'} =>
                 f (transfer_dep_r (U := fun lx:list T => list {x:T |In x lx}) pf c)) in
    f (map_sig l) = f' (map_sig l').
  intros. unfold f'. clear f'.  subst. rewrite transfer_dep_r_eq_refl. 
  reflexivity.
Qed.


Lemma map_sig_transfer_dep_r_compat :
  forall {T U:Type} (l l':list T) (f:{x:T | In x l}->U)
         (pf:l = l'),
    let f' := (fun c:{x:T |In x l'} =>
                 f (transfer_dep_r (U := fun xl => {x:T |In x xl}) pf c)) in
    map f (transfer_dep_r (U := fun xl => list {x:T |In x xl}) pf (map_sig l')) =
    map f' (map_sig l').
  intros.  unfold f'.  clear f'. subst.
  rewrite transfer_dep_r_eq_refl.
  apply map_ext_in.
  intros; rewrite transfer_dep_r_eq_refl.
  reflexivity.
Qed.


Lemma map_sig_transfer_dep_compat :
  forall {T U:Type} (l l':list T) (f:{x:T | In x l'}->U)
         (pf:l = l'),
    let f' := (fun c:{x:T |In x l} =>
                 f (transfer_dep (U := fun xl => {x:T |In x xl}) pf c)) in
    map f (transfer_dep (U := fun xl => list {x:T |In x xl}) pf (map_sig l)) =
    map f' (map_sig l).
  intros.  unfold f'.  clear f'. subst.
  rewrite transfer_dep_eq_refl.
  apply map_ext_in.
  intros; rewrite transfer_dep_eq_refl.
  reflexivity.
Qed.

End Transfer.

